
//
// @(#)btyaccpar, based on byacc 1.8 (Berkeley)
// Parser skeleton modified for use in the Firebird project by Nickolay Samofatov
//
#define YYBTYACC 1

#include "firebird.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "../dsql/Nodes.h"
#include "../dsql/AggNodes.h"
#include "../dsql/DdlNodes.h"
#include "../dsql/BoolNodes.h"
#include "../dsql/ExprNodes.h"
#include "../dsql/PackageNodes.h"
#include "../dsql/StmtNodes.h"
#include "../dsql/WinNodes.h"
#include "../jrd/RecordSourceNodes.h"
#include "../common/classes/Nullable.h"
#include "gen/parse.h"
#include "../dsql/Parser.h"

/*
*	PROGRAM:	Dynamic SQL runtime support
*	MODULE:		parse.y
*	DESCRIPTION:	Dynamic SQL parser
*
* The contents of this file are subject to the Interbase Public
* License Version 1.0 (the "License"); you may not use this file
* except in compliance with the License. You may obtain a copy
* of the License at http://www.Inprise.com/IPL.html
*
* Software distributed under the License is distributed on an
* "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express
* or implied. See the License for the specific language governing
* rights and limitations under the License.
*
* The Original Code was created by Inprise Corporation
* and its predecessors. Portions created by Inprise Corporation are
* Copyright (C) Inprise Corporation.
*
* All Rights Reserved.
* Contributor(s): ______________________________________.
*
* 2002-02-24 Sean Leyne - Code Cleanup of old Win 3.1 port (WINDOWS_ONLY)
* 2001.05.20 Neil McCalden: Allow a udf to be used in a 'group by' clause.
* 2001.05.30 Claudio Valderrama: DROP TABLE and DROP VIEW lead now to two
*   different node types so DDL can tell which is which.
* 2001.06.13 Claudio Valderrama: SUBSTRING is being surfaced.
* 2001.06.30 Claudio valderrama: Feed (line,column) for each node. See node.h.
* 2001.07.10 Claudio Valderrama: Better (line,column) report and "--" for comments.
* 2001.07.28 John Bellardo: Changes to support parsing LIMIT and FIRST
* 2001.08.03 John Bellardo: Finalized syntax for LIMIT, change LIMIT to SKIP
* 2001.08.05 Claudio Valderrama: closed Bug #448062 and other spaces that appear
*   in rdb$*_source fields when altering domains plus one unexpected null pointer.
* 2001.08.12 Claudio Valderrama: adjust SUBSTRING's starting pos argument here
*   and not in gen.c; this closes Bug #450301.
* 2001.10.01 Claudio Valderrama: enable explicit GRANT...to ROLE role_name.
* 2001.10.06 Claudio Valderrama: Honor explicit USER keyword in GRANTs and REVOKEs.
* 2002.07.05 Mark O'Donohue: change keyword DEBUG to KW_DEBUG to avoid
*			clashes with normal DEBUG macro.
* 2002.07.30 Arno Brinkman:
* 2002.07.30 	Let IN predicate handle value_expressions
* 2002.07.30 	tokens CASE, NULLIF, COALESCE added
* 2002.07.30 	See block < CASE expression > what is added to value as case_expression
* 2002.07.30 	function is split up into aggregate_function, numeric_value_function, string_value_function, generate_value_function
* 2002.07.30 	new group_by_function and added to grp_column_elem
* 2002.07.30 	cast removed from function and added as cast_specification to value
* 2002.08.04 Claudio Valderrama: allow declaring and defining variables at the same time
* 2002.08.04 Dmitry Yemanov: ALTER VIEW
* 2002.08.06 Arno Brinkman: ordinal added to grp_column_elem for using positions in group by
* 2002.08.07 Dmitry Yemanov: INT64/LARGEINT are replaced with BIGINT and available in dialect 3 only
* 2002.08.31 Dmitry Yemanov: allowed user-defined index names for PK/FK/UK constraints
* 2002.09.01 Dmitry Yemanov: RECREATE VIEW
* 2002.09.28 Dmitry Yemanov: Reworked internal_info stuff, enhanced
*							exception handling in SPs/triggers,
*							implemented ROWS_AFFECTED system variable
* 2002.10.21 Nickolay Samofatov: Added support for explicit pessimistic locks
* 2002.10.29 Nickolay Samofatov: Added support for savepoints
* 2002.12.03 Dmitry Yemanov: Implemented ORDER BY clause in subqueries.
* 2002.12.18 Dmitry Yemanov: Added support for SQL-compliant labels and LEAVE statement
* 2002.12.28 Dmitry Yemanov: Added support for parametrized events.
* 2003.01.14 Dmitry Yemanov: Fixed bug with cursors in triggers.
* 2003.01.15 Dmitry Yemanov: Added support for runtime trigger action checks.
* 2003.02.10 Mike Nordell  : Undefined Microsoft introduced macros to get a clean compile.
* 2003.05.24 Nickolay Samofatov: Make SKIP and FIRST non-reserved keywords
* 2003.06.13 Nickolay Samofatov: Make INSERTING/UPDATING/DELETING non-reserved keywords
* 2003.07.01 Blas Rodriguez Somoza: Change DEBUG and IN to avoid conflicts in win32 build/bison
* 2003.08.11 Arno Brinkman: Changed GROUP BY to support all expressions and added "AS" support
*						   with table alias. Also removed group_by_function and ordinal.
* 2003.08.14 Arno Brinkman: Added support for derived tables.
* 2003.10.05 Dmitry Yemanov: Added support for explicit cursors in PSQL.
* 2004.01.16 Vlad Horsun: added support for default parameters and
*   EXECUTE BLOCK statement
* Adriano dos Santos Fernandes
*/

#include "firebird.h"
#include "dyn_consts.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

#include "gen/iberror.h"
#include "../dsql/dsql.h"
#include "ibase.h"
#include "../jrd/flags.h"
#include "../jrd/jrd.h"
#include "../jrd/DataTypeUtil.h"
#include "../dsql/errd_proto.h"
#include "../dsql/make_proto.h"
#include "../yvalve/gds_proto.h"
#include "../jrd/err_proto.h"
#include "../common/intlobj_new.h"
#include "../jrd/Attachment.h"
#include "../common/StatusArg.h"

/* This is needed here to provide backward compatibility when working with SSPI plugin*/
#include "../auth/trusted/AuthSspi.h"

/* since UNIX isn't standard, we have to define*/
/* stuff which is in <limits.h> (which isn't available on all UNIXes...*/

const long SHRT_POS_MAX			= 32767;
const long SHRT_UNSIGNED_MAX	= 65535;
const long SHRT_NEG_MAX			= 32768;
const int POSITIVE	= 0;
const int NEGATIVE	= 1;
const int UNSIGNED	= 2;

/*const int MIN_CACHE_BUFFERS	= 250;*/
/*const int DEF_CACHE_BUFFERS	= 1000;*/

#define YYSTYPE YYSTYPE
#if defined(DEBUG) || defined(DEV_BUILD)
#define YYDEBUG		1
#endif

#define YYREDUCEPOSNFUNC yyReducePosn
#define YYREDUCEPOSNFUNCARG NULL


/* ASF: Inherited attributes (aka rule parameters) are executed even when in trial mode, but action*/
/* rules ({}) are executed only when in full parse mode. NOTRIAL should be used to avoid segfaults*/
/* due to accessing invalid pointers in parameters (not yet returned from action rules).*/
#define NOTRIAL(x) (yytrial ? NULL : (x))


inline unsigned trigger_type_suffix(const unsigned slot1, const unsigned slot2, const unsigned slot3)
{
	return ((slot1 << 1) | (slot2 << 3) | (slot3 << 5));
}


#include "../dsql/chars.h"

using namespace Jrd;
using namespace Firebird;

static int yylhs[] = {                                        -1,
    0,    0,    1,    1,    1,    1,    2,    2,    2,    2,
    2,    2,    2,    2,    2,    3,    3,    3,    3,    3,
    3,    3,    3,    3,    3,    4,    4,    4,    5,    5,
    5,    5,    5,    5,    5,  648,  649,    6,  650,  652,
    7,    7,    7,    7,    7,    7,    7,  653,    7,    7,
  654,  655,    7,    8,    8,    8,    8,    8,    8,    8,
    8,    8,    8,    8,    8,    8,  651,  651,    9,    9,
  656,    9,   10,  657,   10,   11,   12,   13,   13,   13,
   13,   13,   14,   14,  658,  658,   15,   15,   16,   16,
   16,   17,   17,   18,   18,   19,   19,   19,   20,   20,
   21,   21,   22,   22,  659,  659,   23,   23,  660,   24,
  661,  662,   25,   25,   25,   25,   25,   25,   25,  663,
   25,   25,  664,   25,   25,   26,   26,   27,   27,  665,
   28,   28,  666,   28,   28,   29,   29,   29,   29,   29,
   29,   29,   30,   30,   30,   31,  667,   31,   32,   32,
   33,   33,   34,   34,   34,   35,   36,   36,  668,  669,
  670,   37,   38,   38,   38,   39,  671,   39,  672,   39,
   40,  673,   40,   41,   42,   42,   42,   42,  674,   43,
  675,   43,   44,   44,   45,   45,   45,   45,   45,   46,
   47,   47,   48,   49,  676,  677,   49,   49,   49,   49,
   49,   49,   49,   49,   49,   49,   49,   49,   49,   49,
   49,   49,   49,  678,   49,  679,   49,   50,   51,   51,
   51,   51,   51,   51,   51,   51,   51,   51,   51,   51,
   52,   53,   53,   53,   53,   53,   53,   53,   53,   53,
   53,   53,   54,   55,   56,   57,   57,   58,   58,   58,
  681,  682,   59,   60,   60,   60,   61,   61,   62,   62,
   63,  685,   63,   64,   64,  687,  688,   65,   66,   66,
   67,  689,   67,   68,   68,  686,  686,   69,   70,   70,
  690,   71,  691,  692,   72,   73,  693,   73,   74,   74,
   75,   76,  694,  694,  695,   77,   78,   78,   79,   79,
   79,  696,   80,   81,   81,   82,   82,   83,   84,   84,
   85,   86,   86,   86,   86,  697,  698,   87,   88,   88,
   88,  699,   89,   90,   91,  700,   91,   92,  701,  702,
  703,  704,   93,   94,   94,   94,   95,   95,   96,  705,
   96,   97,   97,   97,   98,   98,   99,   99,  100,  100,
  101,  101,  102,  103,  103,  103,  103,  103,  104,  104,
  104,  706,  707,  708,  105,  683,  683,  106,  107,  107,
  108,  709,  108,  109,  109,  109,  109,  109,  109,  109,
  109,  110,  110,  111,  111,  112,  112,  112,  112,  710,
  711,  113,  114,  114,  115,  712,  115,  116,  116,  713,
  713,  713,  684,  684,  684,  714,  715,  716,  117,  118,
  717,  718,  118,  119,  119,  120,  120,  121,  121,  122,
  122,  719,  720,  123,  721,  124,  722,  124,  125,  125,
  125,  125,  126,  126,  126,  127,  127,  723,  128,  128,
  724,  725,  129,  726,  129,  129,  129,  727,  728,  130,
  131,  131,  132,  729,  132,  133,  133,  134,  134,  135,
  730,  730,  731,  680,  680,  136,  136,  137,  137,  138,
  138,  138,  138,  139,  139,  139,  139,  139,  139,  140,
  732,  140,  141,  733,  141,  142,  143,  143,  143,  143,
  143,  734,  144,  145,  145,  146,  146,  146,  146,  147,
  147,  148,  148,  148,  148,  148,  149,  150,  151,  151,
  151,  151,  152,  152,  153,  154,  735,  736,  737,  155,
  156,  157,  158,  158,  158,  159,  159,  160,  160,  161,
  162,  162,  163,  164,  165,  165,  165,  166,  166,  167,
  168,  738,  169,  170,  170,  170,  171,  171,  172,  172,
  173,  174,  175,  176,  176,  177,  177,  178,  178,  179,
  180,  181,  181,  182,  182,  183,  183,  184,  184,  184,
  184,  185,  186,  186,  187,  187,  188,  188,  189,  189,
  190,  190,  191,  191,  191,  740,  741,  742,  192,  743,
  744,  193,  194,  194,  195,  739,  739,  196,  197,  197,
  197,  198,  198,  199,  200,  200,  200,  201,  201,  202,
  202,  203,  203,  203,  203,  203,  203,  203,  203,  203,
  203,  203,  203,  203,  203,  203,  203,  203,  203,  203,
  203,  203,  204,  204,  205,  205,  205,  205,  205,  206,
  207,  207,  207,  208,  745,  746,  209,  210,  210,  211,
  211,  212,  747,  748,  749,  213,  214,  215,  216,  750,
  216,  216,  217,  217,  218,  218,  219,  219,  220,  221,
  221,  751,  222,  222,  223,  223,  223,  223,  223,  223,
  223,  223,  224,  224,  225,  226,  227,  228,  229,  229,
  229,  229,  230,  231,  231,  232,  232,  233,  234,  234,
  235,  235,  236,  237,  237,  752,  753,  238,  754,  239,
  755,  239,  240,  240,  240,  240,  240,  241,  241,  241,
  242,  243,  756,  757,  244,  244,  245,  245,  245,  245,
  245,  245,  246,  246,  247,  247,  247,  248,  248,  248,
  758,  759,  760,  249,  250,  250,  251,  251,  252,  253,
  254,  255,  256,  256,  257,  257,  761,  762,  258,  763,
  259,  259,  260,  261,  261,  261,  262,  262,  262,  263,
  264,  264,  264,  264,  264,  265,  265,  266,  266,  266,
  266,  266,  266,  266,  266,  266,  266,  266,  266,  266,
  266,  266,  266,  266,  266,  266,  266,  266,  266,  266,
  266,  266,  266,  266,  266,  266,  266,  266,  266,  266,
  266,  266,  266,  266,  266,  266,  266,  266,  266,  266,
  266,  266,  267,  267,  268,  268,  268,  268,  268,  268,
  268,  268,  268,  268,  268,  268,  268,  268,  268,  269,
  269,  270,  270,  271,  764,  271,  271,  271,  271,  271,
  765,  271,  271,  271,  271,  271,  271,  271,  271,  271,
  271,  271,  271,  271,  271,  766,  767,  272,  273,  768,
  273,  274,  274,  274,  274,  274,  274,  274,  274,  274,
  275,  275,  276,  276,  276,  276,  276,  276,  276,  276,
  276,  276,  276,  276,  276,  770,  276,  276,  276,  276,
  276,  276,  276,  277,  278,  278,  278,  278,  278,  278,
  278,  278,  278,  278,  278,  278,  278,  278,  278,  278,
  278,  278,  278,  278,  278,  278,  278,  278,  278,  278,
  278,  278,  278,  278,  278,  278,  278,  278,  278,  278,
  278,  278,  278,  278,  278,  278,  278,  278,  278,  278,
  278,  278,  278,  278,  278,  278,  278,  278,  278,  278,
  278,  278,  278,  278,  278,  278,  278,  278,  278,  278,
  278,  278,  278,  278,  278,  278,  278,  278,  278,  278,
  278,  278,  278,  278,  278,  278,  278,  278,  278,  278,
  278,  278,  278,  278,  278,  278,  278,  278,  769,  769,
  279,  279,  771,  280,  280,  281,  281,  282,  282,  283,
  283,  284,  284,  285,  285,  285,  286,  286,  287,  288,
  288,  289,  289,  290,  291,  291,  292,  292,  293,  293,
  772,  294,  294,  294,  294,  294,  294,  773,  294,  294,
  294,  294,  294,  294,  294,  294,  294,  295,  295,  296,
  296,  297,  297,  298,  299,  299,  299,  300,  300,  301,
  301,  301,  301,  302,  303,  303,  303,  303,  303,  303,
  303,  303,  303,  303,  303,  303,  303,  303,  303,  303,
  303,  303,  303,  303,  303,  304,  304,  304,  305,  305,
  306,  306,  307,  307,  308,  308,  308,  309,  309,  310,
  310,  311,  311,  312,  312,  313,  313,  314,  314,  314,
  314,  314,  314,  314,  314,  314,  314,  314,  314,  314,
  314,  314,  774,  774,  775,  775,  776,  777,  315,  315,
  315,  315,  316,  316,  317,  317,  317,  318,  318,  319,
  319,  319,  320,  320,  320,  321,  321,  321,  782,  782,
  782,  781,  781,  778,  778,  778,  779,  780,  780,  322,
  323,  323,  324,  324,  324,  783,  783,  325,  325,  325,
  325,  326,  326,  327,  327,  327,  328,  329,  330,  330,
  331,  785,  785,  332,  333,  784,  784,  334,  334,  786,
  786,  787,  788,  335,  336,  337,  337,  338,  789,  790,
  339,  340,  341,  341,  342,  342,  343,  343,  343,  343,
  343,  343,  344,  344,  345,  791,  345,  346,  347,  347,
  348,  348,  348,  348,  349,  349,  349,  349,  349,  350,
  351,  351,  352,  352,  353,  792,  353,  354,  354,  354,
  354,  354,  354,  354,  354,  354,  354,  793,  794,  354,
  355,  355,  356,  356,  356,  357,  357,  357,  357,  358,
  358,  359,  359,  359,  359,  360,  360,  360,  361,  361,
  362,  362,  363,  364,  364,  365,  365,  366,  367,  367,
  367,  367,  367,  367,  795,  796,  368,  369,  369,  370,
  371,  371,  371,  371,  371,  371,  371,  371,  371,  371,
  371,  371,  371,  372,  373,  373,  373,  374,  374,  374,
  375,  376,  376,  377,  377,  378,  379,  379,  380,  380,
  381,  381,  382,  382,  383,  383,  383,  384,  384,  385,
  386,  387,  388,  388,  388,  389,  390,  391,  391,  391,
  391,  392,  392,  392,  393,  393,  393,  394,  394,  395,
  395,  396,  396,  397,  397,  398,  398,  799,  799,  399,
  400,  400,  401,  401,  402,  402,  402,  402,  403,  404,
  405,  405,  406,  406,  407,  407,  408,  408,  408,  409,
  410,  411,  411,  412,  413,  414,  414,  414,  414,  415,
  415,  416,  416,  417,  418,  418,  418,  418,  418,  800,
  800,  419,  419,  420,  420,  421,  422,  422,  423,  423,
  424,  424,  425,  425,  426,  427,  427,  428,  428,  802,
  803,  429,  801,  801,  801,  801,  801,  801,  430,  430,
  431,  431,  432,  432,  433,  804,  805,  433,  806,  433,
  434,  434,  435,  807,  435,  436,  436,  437,  438,  438,
  439,  440,  440,  440,  441,  441,  442,  442,  443,  443,
  444,  444,  808,  808,  445,  445,  809,  809,  446,  446,
  446,  810,  447,  447,  447,  448,  449,  449,  449,  450,
  450,  451,  451,  811,  812,  452,  453,  453,  813,  453,
  453,  814,  815,  816,  454,  817,  818,  455,  819,  456,
  820,  456,  456,  456,  821,  457,  822,  457,  458,  458,
  459,  460,  461,  461,  823,  462,  463,  824,  825,  826,
  464,  465,  827,  465,  466,  466,  466,  467,  468,  828,
  469,  829,  469,  470,  470,  471,  471,  472,  472,  473,
  474,  474,  475,  830,  475,  476,  477,  477,  478,  478,
  478,  479,  480,  480,  481,  482,  482,  482,  482,  482,
  482,  482,  483,  483,  483,  483,  483,  483,  483,  483,
  483,  483,  483,  483,  484,  485,  485,  485,  485,  485,
  485,  485,  485,  486,  487,  487,  487,  488,  488,  489,
  489,  490,  490,  491,  491,  491,  492,  492,  492,  492,
  492,  492,  492,  492,  493,  493,  494,  495,  496,  496,
  496,  497,  498,  498,  498,  498,  499,  499,  500,  831,
  501,  833,  502,  834,  503,  835,  504,  832,  832,  505,
  505,  506,  836,  506,  507,  507,  507,  507,  507,  507,
  507,  507,  507,  837,  507,  508,  509,  509,  510,  510,
  838,  839,  840,  511,  841,  842,  512,  843,  844,  513,
  514,  845,  846,  847,  515,  516,  517,  517,  518,  518,
  518,  519,  519,  520,  520,  520,  520,  520,  521,  521,
  522,  522,  523,  523,  524,  524,  525,  525,  526,  526,
  527,  527,  527,  528,  528,  528,  528,  528,  528,  528,
  528,  528,  528,  528,  528,  528,  528,  528,  528,  528,
  528,  528,  528,  528,  528,  528,  528,  528,  528,  528,
  528,  528,  528,  529,  529,  530,  530,  530,  530,  530,
  531,  531,  532,  532,  533,  534,  534,  535,  535,  536,
  536,  536,  536,  536,  536,  537,  537,  537,  537,  538,
  539,  539,  539,  539,  539,  540,  540,  540,  540,  540,
  541,  541,  542,  543,  543,  544,  545,  545,  545,  545,
  545,  545,  545,  546,  546,  546,  546,  546,  547,  547,
  548,  549,  549,  550,  551,  552,  553,  554,  554,  555,
  556,  556,  556,  557,  557,  557,  558,  558,  559,  559,
  559,  559,  559,  559,  559,  559,  559,  559,  559,  559,
  559,  559,  559,  559,  559,  559,  559,  559,  559,  559,
  559,  559,  559,  559,  559,  559,  559,  560,  560,  560,
  560,  560,  560,  560,  560,  560,  560,  560,  560,  560,
  560,  560,  561,  561,  561,  562,  562,  562,  563,  563,
  564,  564,  565,  566,  566,  567,  848,  849,  567,  850,
  567,  568,  568,  569,  569,  569,  570,  570,  570,  570,
  571,  571,  571,  571,  572,  572,  572,  572,  572,  573,
  573,  574,  574,  575,  576,  576,  576,  577,  578,  578,
  579,  580,  580,  581,  581,  581,  581,  581,  581,  581,
  581,  581,  581,  581,  581,  581,  581,  581,  581,  581,
  581,  581,  581,  581,  581,  581,  581,  581,  581,  581,
  581,  581,  581,  581,  581,  581,  581,  581,  581,  581,
  581,  581,  581,  581,  581,  581,  581,  581,  581,  581,
  581,  581,  581,  581,  581,  581,  581,  581,  581,  581,
  581,  581,  581,  581,  581,  581,  582,  582,  582,  582,
  582,  582,  582,  582,  582,  582,  582,  582,  582,  582,
  582,  582,  582,  583,  583,  584,  584,  585,  585,  586,
  586,  587,  587,  588,  588,  589,  589,  590,  590,  591,
  591,  592,  592,  593,  593,  594,  594,  595,  595,  595,
  596,  596,  596,  596,  597,  597,  598,  598,  599,  599,
  599,  599,  600,  600,  600,  601,  601,  601,  601,  602,
  603,  603,  604,  604,  604,  604,  604,  605,  605,  851,
  852,  853,  606,  607,  607,  608,  608,  609,  609,  610,
  610,  611,  612,  613,  614,  614,  615,  615,  616,  616,
  616,  616,  616,  616,  616,  616,  616,  616,  616,  616,
  798,  798,  797,  797,  617,  617,  618,  619,  620,  620,
  621,  621,  622,  623,  624,  625,  626,  627,  628,  629,
  630,  631,  632,  633,  634,  635,  636,  637,  638,  639,
  640,  641,  642,  643,  644,  645,  646,  646,  647,  647,
  647,  647,  647,  647,  647,  647,  647,  647,  647,  647,
  647,  647,  647,  647,  647,  647,  647,  647,  647,  647,
  647,  647,  647,  647,  647,  647,  647,  647,  647,  647,
  647,  647,  647,  647,  647,  647,  647,  647,  647,  647,
  647,  647,  647,  647,  647,  647,  647,  647,  647,  647,
  647,  647,  647,  647,  647,  647,  647,  647,  647,  647,
  647,  647,  647,  647,  647,  647,  647,  647,  647,  647,
  647,  647,  647,  647,  647,  647,  647,  647,  647,  647,
  647,  647,  647,  647,  647,  647,  647,  647,  647,  647,
  647,  647,  647,  647,  647,  647,  647,  647,  647,  647,
  647,  647,  647,  647,  647,  647,  647,  647,  647,  647,
  647,  647,  647,  647,  647,  647,  647,  647,  647,  647,
  647,  647,  647,  647,  647,  647,  647,  647,  647,  647,
  647,  647,  647,  647,  647,  647,  647,  647,  647,  647,
  647,  647,  647,  647,  647,  647,  647,  647,  647,  647,
  647,  647,  647,  647,  647,  647,  647,  647,  647,  647,
  647,  647,  647,  647,  647,  647,  647,  647,  647,  647,
  647,  647,  647,  647,  647,  647,  647,  647,  647,  647,
  647,  647,  647,  647,  647,  647,  647,  647,  647,  647,
  647,  647,  647,  647,  647,  647,  647,  647,  647,  647,
  647,  647,  647,  647,  647,  647,  647,  647,  647,  647,
  647,  647,  647,  647,  647,  647,  647,  647,  647,  647,
  647,  647,  647,  647,  647,  647,  647,  647,  647,  647,
  647,  647,  647,  647,  647,  647,  647,  647,  647,  647,
  647,  647,  647,  647,  647,  647,  647,  647,  647,  647,
  647,  647,  647,  647,  647,  647,  647,  647,  647,  647,
};
static int yylen[] = {                                         2,
    1,    2,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    0,    0,    4,    0,    0,
   10,   10,   10,   10,   10,   10,   10,    0,    8,    8,
    0,    0,    7,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    2,    1,    1,    0,    1,    1,    2,
    0,    2,    2,    0,    4,    1,    1,    1,    1,    1,
    2,    2,    2,    2,    0,    1,    2,    4,    1,    2,
    2,    2,    2,    2,    4,    1,    1,    1,    0,    3,
    0,    3,    0,    2,    2,    1,    1,    2,    0,    4,
    0,    0,   10,   10,   10,   10,   10,   10,   10,    0,
    8,    8,    0,    7,    6,    0,    3,    0,    3,    0,
    2,    2,    0,    4,    4,    2,    2,    2,    2,    2,
    2,    3,    1,    2,    2,    2,    0,    4,    1,    2,
    2,    4,    1,    2,    2,    2,    2,    3,    0,    0,
    0,   11,    1,    1,    5,    0,    0,    2,    0,    4,
    2,    0,    4,    2,    0,    2,    2,    1,    0,    2,
    0,    4,    2,    2,    0,    2,    2,    1,    3,    9,
    1,    1,    2,    2,    0,    0,    9,    2,    2,    2,
    4,    2,    2,    2,    2,    2,    2,    2,    2,    2,
    2,    2,    3,    0,    3,    0,    4,    2,    2,    2,
    2,    4,    2,    2,    2,    3,    2,    2,    2,    2,
    4,    2,    2,    2,    2,    2,    2,    2,    2,    2,
    3,    4,    2,    2,    2,    0,    1,    1,    1,    4,
    0,    0,    8,    0,    1,    1,    0,    1,    0,    4,
    0,    0,    2,    1,    2,    0,    0,    8,    0,    2,
    2,    0,    3,    1,    1,    0,    1,    2,    0,    1,
    2,    4,    0,    0,    4,    0,    0,    4,    2,    2,
    3,    3,    0,    1,    0,    4,    0,    4,    1,    2,
    2,    0,    4,    0,    4,    2,    2,    2,    0,    2,
    5,    1,    1,    2,    2,    0,    0,    4,    0,    2,
    2,    0,    6,    3,    2,    0,    4,    1,    0,    0,
    0,    0,   10,    0,    2,    5,    0,    2,    2,    0,
    3,    2,    2,    2,    2,    2,    2,    2,    2,    2,
    0,    1,    5,    3,    4,    2,    2,    2,    1,    1,
    1,    0,    0,    0,    6,    0,    1,    1,    0,    2,
    2,    0,    3,    3,    2,    2,    2,    2,    2,    3,
    4,    0,    2,    2,    3,    1,    4,    6,    3,    0,
    0,    5,    0,    2,    2,    0,    3,    3,    4,    0,
    1,    2,    0,    1,    1,    0,    0,    0,    9,    0,
    0,    0,    4,    1,    1,    3,    3,    0,    1,    2,
    2,    0,    0,    8,    0,    2,    0,    4,    0,    1,
    4,    4,    0,    3,    2,    2,    4,    0,    2,    2,
    0,    0,    7,    0,    7,    3,    2,    0,    0,    7,
    2,    1,    0,    0,    4,    3,    2,    3,    3,    4,
    1,    1,    3,    2,    1,    1,    1,    0,    2,    1,
    3,    6,    1,    1,    1,    1,    1,    1,    1,    0,
    0,    2,    2,    0,    3,    3,    1,    1,    5,    2,
    3,    0,    3,    0,    2,    3,    4,    8,    1,    0,
    4,    0,    1,    1,    2,    2,    3,    3,    1,    2,
    2,    2,    1,    1,    5,    3,    0,    0,    0,    6,
    1,    1,    0,    2,    4,    0,    5,    2,    4,    3,
    1,    4,    2,    2,    0,    2,    2,    1,    1,    5,
    3,    0,    8,    0,    2,    1,    5,    3,    0,    2,
    1,    1,    6,    1,    0,    1,    2,    3,    3,    1,
    1,    6,    5,    0,    1,    1,    2,    2,    3,    2,
    3,    2,    0,    1,    1,    2,    2,    2,    0,    1,
    1,    2,    4,    4,    4,    0,    0,    0,    8,    0,
    0,   10,    1,    1,    3,    0,    1,    7,    0,    2,
    1,    1,    1,    3,    0,    1,    2,    1,    2,    2,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    2,    1,    1,    2,    1,    1,    1,    1,    1,    5,
    2,    3,    6,    1,    0,    0,    8,    2,    2,    0,
    1,    2,    0,    0,    0,    7,    3,    5,    1,    0,
    7,    7,    2,    4,    3,    4,    2,    4,    1,    0,
    2,    0,    3,    2,    5,    3,    3,    2,    2,    3,
    3,    3,    8,    6,    3,    0,    3,    2,    1,    1,
    3,    3,    7,    0,    2,    1,    2,    2,    0,    1,
    0,    1,    3,    1,    2,    0,    0,    6,    0,    2,
    0,    4,    2,    2,    2,    2,    1,    1,    1,    1,
    2,    2,    0,    0,    7,    3,    1,    1,    1,    1,
    2,    2,    5,    7,    0,    1,    3,    0,    2,    4,
    0,    0,    0,   10,    0,    4,    2,    4,    4,    5,
    1,    1,    0,    3,    5,    3,    0,    0,    4,    0,
    4,    5,    1,    1,    1,    0,    4,    2,    2,    2,
    1,    1,    2,    2,    2,    1,    3,    2,    2,    2,
    2,    2,    2,    2,    2,    2,    2,    2,    2,    2,
    2,    2,    2,    2,    2,    2,    2,    2,    2,    2,
    2,    2,    2,    2,    2,    2,    2,    2,    2,    2,
    2,    2,    3,    2,    2,    2,    3,    3,    2,    2,
    2,    3,    1,    1,    1,    1,    1,    3,    3,    3,
    3,    3,    3,    5,    5,    5,    5,    5,    5,    0,
    2,    2,    1,    2,    0,    5,    2,    2,    2,    2,
    0,    4,    2,    2,    3,    2,    2,    2,    3,    3,
    2,    2,    3,    4,    4,    0,    0,    4,    2,    0,
    3,    2,    3,    2,    2,    2,    3,    3,    2,    2,
    2,    4,    3,    3,    3,    3,    4,    4,    5,    5,
    4,    5,    3,    4,    4,    0,    5,    4,    4,    4,
    3,    2,    2,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    2,
    1,    1,    0,    4,    2,    3,    4,    0,    2,    3,
    2,    2,    4,    0,    1,    1,    2,    2,    3,    0,
    2,    0,    2,    5,    1,    1,    1,    1,    2,    3,
    0,    3,    4,    3,    2,    2,    5,    0,    5,    1,
    4,    2,    3,    2,    2,    5,    5,    0,    2,    1,
    3,    2,    4,    1,    8,    6,    5,    2,    0,    0,
    3,    3,    3,    2,    2,    2,    2,    2,    2,    2,
    2,    2,    3,    2,    3,    2,    2,    2,    2,    5,
    2,    2,    3,    3,    4,    0,    2,    2,    1,    1,
    1,    3,    1,    1,    3,    6,    1,    1,    1,    4,
    5,    1,    3,    1,    3,    1,    2,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    2,    4,    2,
    4,    1,    1,    1,    0,    3,    0,    0,    6,    4,
    6,    5,    0,    3,    0,    2,    2,    0,    3,    4,
    1,    5,    4,    1,    4,    4,    1,    4,    1,    2,
    2,    1,    1,    1,    2,    2,    1,    1,    2,    2,
    2,    2,    0,    3,    5,    1,    1,    2,    3,    1,
    2,    0,    3,    1,    1,    1,    2,    4,    0,    1,
    5,    0,    1,    3,    3,    0,    1,    0,    2,    0,
    1,    0,    0,    5,    3,    3,    3,    4,    0,    0,
    7,    6,    1,    2,    1,    1,    1,    1,    1,    1,
    5,    5,    0,    2,    2,    0,    4,    1,    6,    5,
    0,    1,    1,    1,    0,    1,    1,    1,    1,    4,
    1,    1,    0,    2,    2,    0,    3,    2,    2,    1,
    2,    2,    3,    2,    2,    2,    3,    0,    0,    4,
    4,    1,    2,    3,    3,    1,    4,    2,    3,    1,
    1,    0,    1,    2,    2,    0,    1,    1,    1,    1,
    1,    3,    2,    0,    3,    1,    3,    4,    5,    6,
    7,    7,    6,    1,    0,    0,    9,    0,    2,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    2,    1,    1,    1,    1,    2,    2,    1,    2,    1,
    2,    1,    3,    1,    1,    3,    0,    3,    0,    2,
    0,    2,    4,    5,    0,    3,    2,    1,    3,    6,
    1,    1,    1,    4,    4,    1,   10,    0,    2,    1,
    1,    2,    4,    2,    2,    4,    2,    1,    1,    1,
    1,    1,    3,    1,    3,    1,    3,    0,    1,    2,
    1,    3,    1,    1,    1,    1,    1,    3,    6,    2,
    0,    1,    0,    3,    1,    3,    1,    1,    1,    4,
    5,    5,    5,    2,    4,    4,    2,    6,    4,    0,
    3,    1,    3,    1,    0,    1,    2,    2,    2,    0,
    1,    0,    3,    1,    3,    1,    0,    2,    0,    2,
    0,    2,    1,    3,    5,    0,    1,    0,    2,    0,
    0,    6,    0,    1,    2,    1,    1,    1,    1,    3,
    2,    1,    1,    2,    1,    0,    0,    6,    0,    5,
    1,    3,    0,    0,    5,    0,    1,    3,    1,    3,
    3,    0,    1,    1,    0,    2,    1,    1,    2,    4,
    0,    1,    1,    1,    0,    3,    1,    1,    0,    5,
    4,    0,    9,    6,    4,    3,    0,    3,    3,    1,
    3,    1,    1,    0,    0,   11,    2,    2,    0,    3,
    3,    0,    0,    0,    6,    0,    0,    6,    0,    5,
    0,    7,    2,    4,    0,    9,    0,   11,    1,    1,
    8,    5,    1,    1,    0,   10,    7,    0,    0,    0,
   16,    0,    0,    3,    0,    2,    4,    4,    3,    0,
    2,    0,    4,    3,    3,    1,    1,    0,    1,    3,
    1,    3,    0,    0,    2,    4,    1,    3,    1,    3,
    4,    1,    1,    3,    1,    1,    3,    3,    2,    3,
    3,    4,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    3,    1,    1,    1,    1,    1,
    1,    1,    1,    6,    1,    1,    1,    5,    6,    5,
    6,    3,    4,    1,    1,    2,    3,    5,    4,    6,
    4,    6,    5,    7,    3,    4,    4,    4,    1,    1,
    1,    1,    3,    3,    4,    4,    1,    3,    3,    0,
    4,    0,    5,    0,    4,    0,    5,    0,    1,    0,
    2,    2,    0,    3,    2,    2,    2,    2,    3,    3,
    1,    1,    2,    0,    5,    3,    2,    4,    3,    2,
    0,    0,    0,    5,    0,    0,    5,    0,    0,    5,
    1,    0,    0,    0,    8,    1,    2,    2,    1,    1,
    1,    1,    1,    3,    3,    3,    2,    2,    0,    2,
    3,    2,    1,    1,    1,    1,    1,    3,    1,    3,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    2,    2,    3,    3,    3,    3,
    5,    3,    3,    3,    3,    1,    1,    1,    1,    3,
    1,    1,    1,    1,    1,    1,    2,    2,    2,    2,
    0,    3,    0,    3,    4,    0,    1,    1,    3,    1,
    2,    2,    2,    2,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    2,    2,    2,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    4,    1,    1,    1,    1,    1,    1,    2,
    1,    1,    2,    1,    1,    1,    1,    1,    2,    1,
    1,    1,    1,    1,    1,    1,    1,    6,    4,    5,
    5,    5,    5,    5,    5,    5,    5,    5,    5,    6,
    6,    4,    4,    4,    4,    6,    6,    6,    6,    6,
    6,    6,    6,    6,    6,    6,    6,    3,    3,    3,
    3,    3,    4,    4,    7,    8,    6,    4,    8,    6,
    4,    4,    0,    2,    2,    1,    1,    1,    1,    1,
    3,    5,    5,    0,    3,    0,    0,    0,    4,    0,
    4,    1,    4,    2,    2,    2,    2,    2,    2,    2,
    2,    2,    2,    2,    0,    3,    3,    2,    2,    0,
    2,    1,    1,    6,    1,    1,    1,    4,    4,    4,
    4,    4,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    7,    8,    8,    8,
   12,    6,    4,    6,    6,   10,    8,    6,    4,    8,
    8,   10,    4,    1,    1,    1,    1,    0,    2,    0,
    2,    0,    2,    0,    2,    0,    2,    0,    1,    1,
    1,    0,    2,    1,    1,    1,    1,    2,    2,    2,
    1,    1,    4,    4,    7,    8,    0,    2,    7,    6,
    6,    4,    1,    1,    1,    4,    3,    6,    5,    6,
    1,    1,    6,    8,    6,    6,    8,    1,    1,    0,
    0,    0,    8,    4,    5,    0,    2,    3,    5,    4,
    5,    1,    1,    1,    3,    5,    4,    6,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    0,    1,    0,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
};
static int yydefred[] = {                                      0,
    0,    0,    0,    0,    0,    0,    0,   36,    0,    0,
  109,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    3,    4,    5,    6,   22,   24,   20,   18,   23,
   19,  843,   10,   11,   16,   21,   12, 1174, 1175, 1176,
   27,   28,   26,   33,   32,   29,   30,   35,   31,   34,
   25,   17, 1284,   13,    0,    0,    8,    0,    9,    7,
 1509, 1510,   14, 1513, 1514,   15,    0,    0,  851,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  214,    0,    0,  842, 1187,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  247,
    0,    0,    0,    0,    0,    0,  214,    0,  193,    0,
    0,    0,  156,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  214,    0, 1064,    0,  741,   37,    0,    0,   37,
    0,    0,    0, 1192,    0,    0,    0,    0,    0,    0,
    0, 2206, 2207, 2208, 2209, 2210, 2211, 2212, 2213, 2214,
 2215, 2216, 2217, 2218, 2219, 2220, 2221, 2222, 2224, 2225,
 2226, 2227, 2228, 2229, 2230, 2223, 2231, 2232, 2233, 2234,
 2235, 2236,    0, 2237, 2238, 2239, 2240, 2241, 2242, 2243,
 2244, 2245, 2246, 2247, 2248, 2249, 2250, 2251, 2252, 2253,
 2254, 2255, 2256, 2257, 2258, 2259, 2260, 2261, 2262, 2264,
 2265, 2097, 2099, 2100, 2101, 2102, 2103, 2104, 2122, 2123,
 2105, 2107, 2106, 2115, 2116, 2112, 2108, 2113, 2109, 2111,
 2114, 2110, 2118, 2119, 2117, 2120, 2121, 2125, 2124, 2126,
 2127, 2128, 2129, 2130, 2131, 2132, 2133, 2134, 2135, 2136,
 2137, 2138, 2139, 2140, 2141, 2142, 2143, 2144, 2145, 2146,
 2147, 2148, 2149, 2150, 2151, 2152, 2153, 2154, 2155, 2156,
 2157, 2158, 2159, 2160, 2161, 2162, 2163, 2164, 2165, 2166,
 2167, 2168, 2169, 2170, 2171, 2172, 2173, 2174, 2175, 2176,
 2177, 2178, 2179, 2180, 2181, 2182, 2183, 2184, 2185, 2186,
 2187, 2188, 2189, 2190, 2263, 2191, 2192, 2193, 2199, 2195,
 2194, 2196, 2197, 2198, 2200, 2201, 2202, 2203, 2204, 2205,
 2270, 2271, 2272, 2273, 2274, 2275, 2277, 2278, 2279, 2280,
 2266, 2281, 2267, 2268, 2269, 2276, 2282, 2283, 2284, 2285,
 2286, 2287, 2288, 2289, 2290, 2291, 2292, 2293, 2294, 2295,
 2296, 2297, 2298, 2299, 2302, 2304, 2301, 2305, 2306, 2307,
 2308, 2309, 2310, 2311, 2312, 2313, 2314, 2315, 2316, 2317,
 2318, 2319, 2320, 2321, 2322, 2323, 2324, 2325, 2327, 2328,
 2329, 2330, 2331, 2332, 2333, 2334, 2336, 2337, 2338, 2340,
 2341, 2342, 2343, 2344, 2345, 2346, 2347, 2348, 2349, 2350,
 2351, 2352, 2353, 2354, 2355, 2356, 2357, 2358, 2359, 2360,
 2303, 2339, 2326, 2300, 2335,    0, 1392,    0, 2089, 2098,
    0,    0, 1328,    0, 2088,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  218, 1177, 2094,    0,
    0,    2,    0,    0,    0,    0, 1333, 1336,    0,    0,
    0,    0,  363,  906,  907,  908,  909,  910,  911,  912,
  913,  914,  915,  916,  917,  918,  919,  920,  921,  922,
  923,  924,  925,  926,  934,  928,  929,  927,  961,  981,
  958,  930,  931,  932,  933,  935,  937,  940,  936,  938,
  939,  941,  942,  943,  944,  945,  946,  947,  948,  949,
  950,  951,  952,  953,  963,  975,  973,  959,  976,  954,
  960,  979,  980,  974,  962,  977,  978,  982,  983,  957,
  956,  955,  964,  965,  966,  967,  968,  969,  970,  971,
  972,  984,  985,  986,  987,  988,  989,  990,  991,  992,
  993,  994,  995,  996,  997,  998,  853,  866,  905,  844,
    0, 2078,    0,    0,  551,  538,  539,    0,  856,  542,
 2068,  861,  302, 2081,  854,    0, 2082,  521,  513,  514,
    0,  849,  517, 2086,  845, 1394,  848,    0, 2090,  858,
    0, 2091,  752,  847,    0, 1025, 1026,  857,    0, 2087,
  862,  216, 1655,  560,  850,    0, 2095, 1195,    0, 1184,
 1779,    0,  206,  362, 1781,  368,  207,    0, 2074,  194,
    0,  198,  204,  283,    0,  199, 1784,  208, 1786,    0,
  200,    0,  202,    0,  757,  211, 1620,  203,  209,  316,
  205,  210,    0, 2073,    0,  216, 1651,    0,  212, 1453,
 1454,    0,    0,  157,    0, 2079,    0, 1072, 2077, 1065,
    0, 1071, 1074, 1077, 1066, 1067,    0, 1068, 1069,    0,
 1070, 2093, 1076, 1078, 1079,  216,    0,    0, 1082,    0,
    0,    0,  742,    0, 1476,    0,    0,    0, 1185,    0,
    0, 1193, 1196,    0,    0, 1197,    0,    0,    0,    0,
    0, 1515, 1359,    0,    0,    0,    0,    0,  227,  220,
  228,  219,  221,  224,  230,  223,  229,    0,    0,  225,
    0,    0, 1290, 1291, 1296,    0, 1295,    0, 1297, 1299,
 1305,    0, 1292, 1294,    0, 1293, 1300, 1304, 1298, 1302,
 1303,    0,    0,    0,    0,    0,    0,    0, 1316,    0,
    0,    0,    0, 1341,    0,    0,    0, 1447,    0,    0,
    0, 2072,  860,    0, 2071, 1629,  859, 1624,  363,  867,
  245,  855,    0,    0,    0,    0,  419,    0,    0,  160,
  284, 1017, 1018,    0,    0,  518,  411,  764,  765,    0,
 1622,    0,    0, 1539, 1028, 1027,    0,  317, 1655,  863,
    0,    0, 1191, 1189, 1780,  363,  277,    0,  243,  284,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  214,
    0,  231,  256,  255,    0,    0,  406,    0,    0,    0,
    0,  758, 1286,    0,    0, 1651,  215,    0,  213,    0,
    0,  158,  159,    0,    0,    0,    0, 1073,    0,    0,
 1075,    0,    0, 1084, 1661, 1666, 1083,    0,    0,    0,
    0, 1724,    0, 1769,    0, 2065,    0, 1721, 1922,    0,
    0,    0,    0, 1940,    0, 1770,    0, 1764, 1752, 1753,
 1747, 1748, 1750,    0, 1751,    0,    0,    0, 1726,    0,
    0, 1754, 1755, 1765, 1766,    0, 1767, 1768,    0,    0,
    0, 1772, 1609, 1610, 1611,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1934, 1937,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1771, 1761, 1762,
 2066, 1725,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1612,    0, 1935, 1936,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1763, 1700,    0,    0, 1549, 1686, 1556, 1563, 1570, 1564,
 1565, 1566, 1567, 1568, 1571, 1572, 1573, 1574, 1569,    0,
    0,    0, 1689, 1719, 1722, 1695,    0, 1756, 1749, 1746,
 1697, 1698, 1699, 1716, 1717, 1718, 1757, 1696, 1792,    0,
    0, 1850,    0, 1793, 1794, 1882, 1883, 1885, 1886, 1887,
 1796,    0, 1893,    0,    0,    0, 1795, 2001, 2002, 1704,
 1701, 1702, 2021, 2022, 2028, 2029, 1703, 1723,    0, 1552,
    0,    0,    0,    0,    0,   38,    0,   37,    0,    0,
    0,    0,  110,  111,  123, 1183,    0,    0, 1278,    0,
 1232, 1230, 1231, 1790,    0,    0,    0, 1117, 1166, 1123,
    0, 1167,    0,    0, 1124,    0,    0,    0, 1154, 1170,
 1116, 1149,    0,    0, 1113, 1122,    0,    0, 1114, 1158,
 1205, 1098, 1106, 1099, 1108, 1109,    0, 1112, 1110, 1111,
    0, 1203, 1115,    0,    0,    0,    0,    0,    0, 1198,
 1199,    0,    0, 1530, 1393, 1329,    0, 1375, 2083,    0,
    0,  226, 1180, 1178, 1301, 1309, 1307, 1306, 1285,    0,
    0, 2085,    0,    0,    0,    0,    0, 1318, 1322,    0,
 1344, 1342,    0, 1347, 1345, 2062, 1348,    0, 1349, 1339,
    0,    0, 2064,    0,    0,    0, 1323,    0,    0, 1475,
    0,    0, 1523, 1545,    0, 1286,    0,    0,  867,    0,
    0,    0,    0,  865,    0,    0,    0,    0,    0,  541,
    0,    0,    0,  516,    0,  411,  824,  823,    0,    0,
 1286,    0, 1541,    0,    0,    0,  864, 1656, 1662,    0,
    0,    0,    0, 1089, 1090,    0,    0,    0,    0,  237,
    0,  552,  233,  238,  295,  522,  232,  763,  234,  240,
    0,  751,  236,  239,  216, 1658,  561,  235,  258,    0,
    0,  435,    0,    0,    0,  756,    0,    0,    0,  329,
  201,  422,  217, 1652,    0,    0,  160, 2070,    0,    0,
  192, 1782,  191, 2069,    0,    0, 1410,    0,    0,    0,
 1512, 1088, 1087,    0, 1085,    0,    0,    0, 1758,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1759,
 1760,    0, 1728,    0, 1730,    0,    0,    0,    0, 2030,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1727, 1729,    0,    0,    0,    0,    0,
    0,    0,    0, 1687,    0, 1331, 1332,    0,    0,    0,
    0,    0,  688,    0,    0,  733,    0,    0,    0, 1594,
 1579,    0,    0, 1580,    0, 1583,    0, 1581, 1582,    0,
    0,    0, 1576, 1577, 1578,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  169,  743,    0,   76,   77,
    0,    0,    0,    0,    0,   71,    0,    0,    0,    0,
    0, 1364, 1366, 1367, 1363, 1377, 1378, 1379, 1365,    0,
    0,    0,    0,    0,    0,   37, 1181,  313,    0,  311,
  312, 1788, 1194, 1193, 1226, 1227, 1228, 1229, 1220,    0,
 1128, 1150, 1151, 1171,    0, 1168,    0,    0, 1161, 1155,
 1156,    0,    0, 1204, 1118,    0, 1120, 1159, 1160,    0,
 1107,    0,    0,    0,    0,    0,    0,    0, 1162, 1200,
    0, 1518,    0,    0,    0, 1374,    0,  222, 1286, 1315,
 1279, 1314,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1351,    0,    0, 1352, 1354,    0,    0,    0, 1449,
    0, 1335, 1334,    0,    0,    0, 1692,    0, 1691, 1740,
 1745, 1693,    0, 1324,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1040,    0,
    0,    0,  372,  372, 1029,    0,    0, 1021,    0, 1019,
    0,    0,  357,  358,  356,  548,    0,  416,  417,    0,
    0,    0,    0,  575,    0,    0,  550,    0,    0,  303,
  284,    0,  519,    0,    0,    0,    0,    0, 1058,    0,
    0,    0,    0, 1540,    0,    0,  318,    0, 1653,    0,
    0,  364,  363,    0,  280,  266,    0,    0,  285,  284,
  244,  284, 1626, 1658,  241,    0,    0,  434,  407, 1063,
 1061, 1062,    0,    0,  759,  760, 1621, 1286,  330,    0,
 1653,    0,    0,    0, 1785, 1783,    0,    0,    0, 1424,
 1426,    0, 1427, 1419, 1420,    0, 1080,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2049, 2050, 2051, 2052,
 2053, 2054, 2059, 2060, 2055, 2058, 2056, 2057,    0,    0,
    0,    0,    0,    0, 2038,    0, 2031,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2013, 2015, 2014,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1828,    0,    0,    0,    0,    0, 1829,
 1832,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1831,    0,    0,
    0, 1847, 1846, 1848,    0, 1830, 1777, 1774, 1775, 1776,
 1778,    0,    0,    0,    0,    0, 1715,    0, 1690,  737,
    0,    0,  689,    0,  690,    0,    0,    0,    0,    0,
    0, 1605, 1617,    0, 1613,    0, 1614, 1561,    0,    0,
    0,    0,    0,    0,    0, 1596,    0, 1585, 1587, 1586,
    0,    0,    0,    0, 1710, 1709,    0,    0,    0,    0,
    0,    0,    0,    0, 1851, 2096,    0,    0,    0,    0,
 2017,    0, 1720, 1550,    0,    0,  167,    0,    0,    0,
    0,    0,    0,    0,   61,   62,   66,   57,   59,   56,
   54,   55,   63,   60,   65,   58,    0,    0,    0,    0,
    0,    0,   52,  147,    0,  146,  149,    0, 1370,    0,
    0,    0,    0, 1396,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  127,  129,    0,    0,    0,    0,
    0,    0,  315,  314, 1789,    0,    0, 1787,    0,    0,
    0,    0, 1169,    0,    0,    0,    0,    0,    0,    0,
 1208, 1209, 1207, 1202,    0,    0,    0,    0,    0,    0,
    0, 1222, 1223, 1224, 1219, 1519, 1532,    0,    0, 1531,
 1553,    0,    0, 1376,    0,    0, 1280, 1311,    0,    0,
 1313, 1283, 1343, 1346,    0,    0,    0,    0,    0,    0,
    0,    0, 1743, 1742, 1744, 1741, 2092, 1464, 1463, 1466,
 1467, 1468,    0,    0, 1478, 1479,    0, 1474,    0, 1547,
    0, 1625, 1030,    0,    0, 1035,    0, 1042, 1036,    0,
    0,    0, 1045, 1044,    0,    0,    0,    0,    0,    0,
    0,    0,  869, 1023,  354,    0,    0,    0,    0,    0,
  540,    0,  572,    0,  581,    0,    0,  576,  577,  578,
  524,  167,    0,  287,    0,  515,    0,  412,  492,    0,
    0,    0,    0,  881,    0,    0,    0,    0,  841,    0,
    0, 1623, 1542,    0,  750, 1024,    0,    0,  320,  321,
    0, 1663,    0,    0,    0,    0,  556,    0,    0,    0,
  278,  479,  474,  475,  476,  477,  478,  267,    0, 1102,
    0,    0,  287,    0,    0, 1286,  242, 1659,    0,  251,
  425,  755,    0,    0,    0,    0,    0,  407,    0,    0,
    0,    0,    0,    0,  566,  195,  169,    0,  167,    0,
 1528, 2076, 1425,    0, 1462,    0,    0,    0,    0,    0,
 1799,    0, 1607,    0, 2003,    0,    0,    0,    0,    0,
 1969, 1608,    0,    0,    0, 1732, 1734,    0,    0,    0,
    0,    0,    0, 2032,    0,    0,    0, 2047, 1888, 1889,
 1890, 2004, 1891,    0, 2012,    0,    0,    0,    0,    0,
    0,    0, 1963,    0,    0,    0, 1833,    0, 1834,    0,
 1838,    0, 1841, 1812, 1813, 1814, 1815,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1998,
 1999, 2000,    0,    0, 1842, 1773, 1973,    0,    0, 1688,
 1551,    0,    0,    0, 1735,    0,    0,    0,    0,    0,
 1615, 1616, 1562,    0,    0, 1606,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1417, 1892,    0,    0,    0,
 2016,    0,  734,    0,    0,    0,    0,   68,    0,    0,
    0,    0,    0,    0,    0,   64,   48,   48,   74,   74,
   74,    0,    0,   80,    0,   79,    0,   78,    0,   73,
   87,   94,  130,    0,  150,    0, 1368, 1401, 1399, 1397,
    0, 1484, 1398,    0,    0,    0,    0,    0,   48,    0,
    0,    0,    0,    0,   48,    0,    0,    0,    0, 1248,
 1240,    0,    0,    0, 1235,    0,    0,    0, 1130,    0,
    0, 1173,    0, 1164, 1119, 1126, 1121, 1139,    0,    0,
    0, 1140, 1143, 1145, 1146, 1148, 1201, 1200,    0,    0,
    0, 1517,    0,    0, 1330,    0,    0, 1281, 1282,    0,
    0,    0,    0, 1353, 1357, 1355, 1450,    0, 1451,    0,
    0,    0,    0, 1483,    0, 1480,    0,    0, 1546,  353,
    0,    0,    0,  264, 1034,    0, 1043,    0, 1038, 1050,
  372,    0,    0,  871,    0,    0,  874,  876,  875,    0,
    0,  872,  879,  880,  359,  360,  361,  355,  547,  590,
  586,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  632,  608,
  603,    0,    0,  602,    0,  612,  611,  635,  623,  624,
  638,    0,  621,  639,  636,  625,  618,  637,    0,  627,
  628,  626,  718,  720,  719,  619,    0,  613,  614,  617,
  615,  616,  633,  622,    0, 1537, 1536,    0,    0,  597,
    0,  582,    0,    0,    0,    0,    0,    0,    0, 1094,
 1093, 1097,    0,    0,    0,  307,  306,  520,    0,    0,
 1000,    0,    0,    0,    0,  903,  902,    0,  904,    0,
    0,    0,    0,    0,    0,    0,    0, 1056,    0,    0,
    0,    0,    0, 1657,    0,    0,    0,  553,  557,  365,
  363,    0,    0,    0,    0,    0,    0,    0,  371,    0,
    0, 1100,    0,    0,    0,    0,  289,  290,  296,    0,
    0, 1653,  367,    0,  252,    0,  438,    0,    0,    0,
    0,    0,    0, 1641,    0, 1642,    0,    0,    0,    0,
    0,    0, 1632,    0,    0,  331,  425, 1654,  568,    0,
    0,  570,    0,    0,    0,  563,    0,    0,  567,  196,
  167,  161,    0,    0,    0, 1421, 1511, 1805, 1804,    0,
    0,  470,  473, 1801, 1800,    0, 1809, 1808, 1807, 1806,
    0,    0, 1803, 1802,    0,    0,    0, 2040, 2039,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  740,  692,  691,    0,
 1619, 1618, 1588,    0,    0,    0,    0,    0,    0,    0,
 1711,    0,    0,    0, 1852,    0,    0,    0, 2019,    0,
  172,  746,    0,  747,    0,  172,    0,    0,    0,    0,
    0,    0,    0,    0,  130,  130,    0,    0,    0,   90,
   91,   82,   81,    0,    0,  148,    0,    0, 1485, 1380,
    0, 1391, 1386,    0,  130,    0,    0,    0,    0,    0,
    0,    0,  120,  120,  130, 1237, 1246,    0, 1241, 1238,
 1239, 1249,    0, 1245, 1244,    0,    0, 1242, 1252,    0,
 1132,    0, 1136, 1137,    0,    0,    0,    0,    0, 1142,
    0,    0,    0, 1533,    0, 1535,    0, 1554,    0,    0,
 1289,    0,    0,    0, 1457, 1458, 1456,    0, 1471,    0,
    0, 1548, 1033,  390,  265,    0, 1041,    0,  363,    0,
    0,    0,  873,  877,  878,  591,  587,    0,  653,    0,
    0,    0,  697,  700, 2084,  721,  722,    0,  724,  698,
    0,  634,  604,  706,  609,    0,  704,  610,    0,    0,
    0,    0,    0,  695,    0,    0,  593,  594,    0,    0,
    0,    0,  172,  525,  528,    0,    0,    0,    0,    0,
  305,    0,  308,  294,    0,  882,    0,  885,  886,  492,
    0,    0,  884, 2075,  901, 1016, 1015,  883,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  893,  461,    0,
  462,  449,    0,    0,    0,    0,    0,    0,    0,  754,
  324,    0, 1684, 1683,    0,    0,    0,    0,    0,    0,
  558,  559,    0,    0,  373,    0,    0,  379,    0,  376,
  375,  378,  377,    0,  267, 1103, 1105, 1101,  288,    0,
  299,  287,    0, 1627,    0,    0,    0,  427,  408,  436,
    0,    0,    0,    0,  771,  772,  768,  761,    0,    0,
    0,    0,    0,    0,  776,  770, 1634,    0, 1638,    0,
 1635, 1637, 1636, 1644, 1643,    0,  335,    0,    0,  569,
  571,  562,    0,    0,    0,  172,  164,    0,    0,  171,
  163,    0,    0,    0, 2020, 2048,    0, 1968, 1884,    0,
    0,    0, 2041,    0,    0, 2023, 2025,    0, 2010, 2011,
    0,    0,    0,    0,    0,    0,    0, 2026,    0, 1811,
 1810,    0,    0,    0, 1837,    0, 1840, 1816, 1817, 1818,
 1819, 1820, 1821, 1822, 1823, 1824, 1825, 1826, 1827, 1962,
 1965,    0,    0,    0, 1589,    0,    0,    0,    0, 1584,
 1798,    0,    0, 1964,    0,    0,    0,    0, 2018,    0,
    0,  534,    0,    0,   40,   40,   40,   40,   40,   40,
   40,    0,    0,    0,   75,    0,   89,    0,   88,   97,
   96,   98,   95,    0,  133,    0,    0,    0,  151,  153,
    0, 1372, 1381, 1485,    0,    0, 1382, 1383,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  130,  130,    0,
 1243,    0, 1247, 1249,    0,    0,    0, 1253, 1131,    0,
 1129, 1165,    0,    0, 1216, 1215, 1218,    0,    0, 1287,
    0,    0,    0,    0,    0, 1470, 1481, 1473,  391, 1037,
    0, 1039,    0,    0, 1047, 1046,    0,    0,    0,    0,
    0,  654,    0,    0,  685,    0,  726,  651,    0,  707,
  705,    0,    0,    0,    0,  645,    0,    0,    0,    0,
  583,    0,  601,    0,  584,  585,    0,    0,    0, 1095,
  469,    0,  546,  543, 1092,  310,  292,  495,    0,    0,
    0,  447,    0,  466,    0,  467,  899,  900,    0,  887,
    0,  888,  464,  895,  898,  894,  891,    0, 1002,    0,
    0,  449,  838,  839,  835,  834,  837,  836, 1055,  322,
 1681,    0,    0, 1677,    0, 1678, 1664,    0,    0,    0,
  384,  386,    0,  374,  380,  268,    0,    0,  291,    0,
  300,  301, 1660,  404,  405,  260,  253,    0,  438,    0,
  439,  440,  762,  774,  775,  773,    0,    0,  797,  791,
  785,  809,  782,  779,  788,  806,  794,  800,  803,  820,
  815,    0,  796,  790,  784,  808,  781,  778,  787,  805,
  793,  799,  802,  811,  819,    0,  798,  792,  786,  810,
  783,  780,  789,  807,  795,  801,  804,  812,  821,    0,
    0, 1639, 1640, 1286,    0,  332,  330,  423,  197,  249,
    0,    0,  170,  181,    0,    0,    0,    0,    0,  178,
  174,    0,    0,    0, 1432,    0, 1429,    0, 1433,    0,
  471,    0, 2005,    0,    0,    0,    0,    0,    0,    0,
 2009, 1957,    0,    0,    0,    0,    0,    0,    0, 1835,
    0,    0, 1981,    0,    0,    0,    0,    0, 1860, 1857,
    0,    0,    0, 1985,    0,  748,    0,    0,  531,    0,
  744,  130,  130,  130,  130,  130,  130,  130,    0,  133,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
  131,  132,  143,    0,    0,    0,  106,    0,   53,    0,
  154,  155, 1369,    0,    0, 1384,    0, 1388,  112,  112,
  112,  112,  112,  112,  112,    0,    0,  124,    0, 1271,
    0, 1276,    0,    0,    0, 1263, 1255, 1254,    0,    0,
 1134, 1211, 1212,    0,    0, 1516, 1646,    0, 1404,    0,
 1408,    0,    0,    0, 1049,  364, 1052, 1054,  282,    0,
  588,    0,    0,    0,    0,    0,  727,  728,  730,  729,
    0,    0,    0,  709,    0,  646,    0,    0,    0,  595,
  600,    0,  529,  530,    0,  545,    0,    0,    0,  499,
  493,    0,  441,  444,  446,  889,  890,  892,  463,    0,
  897,    0,  317, 1676, 1675, 1680, 1674,    0,  385,    0,
    0,  381,    0,    0,  271,  275,  274,  298,    0,  437,
  409,    0,  767,  813,  777,  817,  818,  822,    0,    0,
    0,    0,    0,  330,  425,    0,    0,    0,    0,    0,
  180,  173,    0,  176,  177,  190,    0, 1422, 1436, 1435,
 1439, 1431, 1434,    0, 2006,    0, 2037,    0, 2033, 2024,
 1958, 1959, 1960,    0, 2027,    0, 1967, 1844, 1845, 1836,
 1839,    0, 1971,    0, 1858, 1858,    0, 1853, 1970,    0,
  749,  172,  527,  533,    0,    0,    0,    0,    0,    0,
    0,    0,    0,   49,  137,  145,  136,  139,  144,  140,
  141,  138,    0,   50,  102,  152,  105,    0,  104,  107,
 1486,    0,    0, 1487, 1488,    0,  130,  130,  130,  130,
  130,  130,  130,  121,  122,    0,    0,    0, 1273, 1251,
 1264, 1265,    0, 1259, 1217,    0,    0,    0, 1413,    0,
 1337,  392,  391,    0,    0,    0,    0,    0,  655,    0,
    0,  640,    0,    0,    0,    0,    0,  658,    0,    0,
  536,  537,    0,    0,    0,    0,    0,    0,  452,    0,
  442,  442,  460,    0, 1003,    0,    0,    0,    0,    0,
 1671, 1670, 1665,    0, 1669,    0,  389,  273,  281,    0,
    0,  412,  414,  415, 1644, 1645,    0, 1647,    0,  336,
  333,  352,  330,  339,    0,    0,    0,    0,    0,    0,
  184,    0,  188,  183,    0, 1430,    0,    0,    0, 2034,
    0,    0,    0,    0,    0,    0,    0, 1878,    0, 1879,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
  100,  134,  135,  142,  108, 1490, 1491,    0, 1492, 1385,
    0,    0,    0,    0,    0,    0,    0, 1272, 1268, 1267,
    0, 1277, 1260, 1261, 1257,    0, 1405,    0,    0,    0,
    0, 1053,    0,  589,  643,    0,    0,    0,    0,    0,
  711,  717,    0,    0,    0,    0,  710,    0,    0,    0,
    0,    0, 1096,    0,    0,    0,  496,  451,    0,    0,
    0,    0,    0,    0,    0,    0, 1011,  326,  325,  328,
 1668, 1667, 1672, 1673,    0,  421,  420,    0,    0, 1650,
    0,  341,    0,    0,    0,    0,  342,  343,  344,  427,
    0,  430,  426,  250,  182,    0,  186,    0,  165, 1437,
 1437,  472, 2035, 1966, 1972,    0,    0,    0,    0, 1861,
 1862, 1859, 1877, 1876,    0, 1990, 1991,    0, 1989,  532,
   41,   43,   42,   44,   45,   46,   47, 1496, 1493,  113,
  115,  114,  116,  117,  118,  119, 1269, 1270, 1275, 1520,
 1414,    0,    0,    0,    0,  395,    0,    0,  656,  654,
    0,  725,  708,    0,  716,  715, 2080,  713,  714,    0,
    0,  649,  648,  702,  693,    0,    0,  497,    0,  448,
    0,  481,    0,    0, 1006, 1004,  449, 1010,    0, 1012,
    0,    0,  413, 1648, 1649,  345,  348,  347,  350,  349,
  346,    0,    0,  162,  189,    0,    0,    0,    0,    0,
    0, 1865, 1864, 1866, 1994, 1995,    0,    0, 1497, 1494,
    0,    0,  397,    0,    0,    0,  592,  660,    0,    0,
  683,  712,  647,    0,  598,    0,    0,  449,  443,    0,
    0,  445, 1007,    0,    0,  327,  388,  428,    0,    0,
    0, 1441,    0, 1440, 1868, 1867, 1870, 1869,    0, 1961,
    0,    0,    0, 1523,    0, 1415,    0,  402,  398,  654,
    0,    0,    0,    0,    0,    0,  674,  703,    0,  501,
    0,    0,  483,  484, 1013,  431,  432,    0, 1438, 1444,
    0,    0,    0, 1863,    0,    0, 1498,    0,    0, 1495,
    0, 1521,  399,    0,    0,    0,  673,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  454,  450,
  485,    0, 1442,    0, 1872, 1871, 1874, 1873,    0, 1505,
    0, 1503,    0, 1524,  655,  661,  655,  662,    0,  663,
  667,    0,    0,    0,    0,    0,    0,  680,  682,  681,
    0,    0,  498,    0,  506,    0,  505,  449,    0,    0,
    0,  488,  486,  487,    0,    0,    0,    0, 1499,    0,
    0,    0,    0,    0,    0,    0,  509,  508,  507,    0,
    0,    0,    0,  490, 1445, 1507,    0, 1504,    0, 1530,
  664,  668,    0,    0,    0,  512,  510,  511,  455,    0,
    0,    0,  457,  491,    0,    0,    0, 1501,    0,    0,
  456,    0,    0,    0,    0,    0, 1530,  458,  459,  489,
    0,    0,    0,    0,    0,    0, 1506,    0, 1508,
};
static int yydgoto[] = {                                      20,
   21,   22,   23,   24, 2389,   26, 1096, 1867, 1451, 1869,
 1452, 1453, 2250, 1454, 1870, 2251, 1455, 1871, 2252, 3291,
 2996, 3309, 3509,   27, 1103, 1104, 1105, 2982, 3301, 3302,
 1457, 1876, 2654, 2999,   28,  113,  832, 2899, 2098, 2543,
 2900, 3231, 3225, 3441, 3614,  644, 1300,   29,  109,   30,
  437,   31,  812,  610, 1260,  550,  110, 3219,  618,  815,
 1280, 2080, 3157, 2343,  607, 2844, 3147, 3415, 1635, 1636,
 3416,  613, 1639, 2073, 2497, 2446, 1264, 2499, 2852,  562,
 1610, 2024, 2447, 2783,   32, 1480,  586, 1627, 2049, 2050,
 3578, 3709,  632, 2526, 3216, 3433, 3604, 3727, 3728, 3729,
 3601,  753, 1224, 2368,  603,  604, 1632, 2059, 2489, 2480,
 2833, 3141, 2344, 3542, 3670, 3786,  621, 3421, 3592,  767,
  768, 3594, 1291, 3606, 3733,  817, 2506, 2860, 2788, 3394,
 3570, 3950, 4010, 4033, 2808, 3103, 2779, 2551, 2061, 3811,
 3870, 3913, 3993, 2789, 2790, 3391, 3697, 3946, 3947, 3948,
 4008,  568,  569,  570,  571,  572, 1267, 1609, 2227, 2435,
 2775, 3278, 3279, 2633, 3380,  555,  556,  557,  558, 3094,
  769, 1230,  559, 1263,  594, 2055, 2056, 2057,  595, 1278,
  829, 2093, 2094, 2095, 1601, 1602, 1603, 1604, 2013, 2014,
 2015, 1605, 1606, 2766, 2767, 2768, 3084, 2390, 2391, 2392,
 2393, 2394, 2395, 2396, 2397, 2398, 2399, 2400, 2401, 3688,
 3067, 3068, 2402, 2403, 2404, 3549, 3934, 3970, 3935, 3971,
 3789, 3859, 3907, 2405, 2406, 3065, 2407, 1022, 1794, 2408,
 2409, 2410, 2411, 2743, 3803, 3802, 2756, 2757, 3556, 3687,
 2412, 2413, 2414, 2415, 3373, 2416, 1023, 1406,   34, 1447,
 2224, 2634,  583, 1273,  584, 2045,  623,  624, 1655, 1269,
  780, 2510, 2511, 2867, 2874, 2875, 1239, 2876, 1621,   35,
   87,  547, 1586, 2003, 1614, 2034, 2458,  548, 3117, 3401,
 3575, 3816, 3576, 3707, 2798,  565,  762, 1221, 1590,  587,
  588,  787, 1217, 1585, 3052, 2352, 3053, 3357,  577, 1240,
  821,   36,  134,  841, 1253, 2438, 2439, 2440, 2441, 1255,
 2069, 2070, 1142, 1143, 1144, 2696, 2291, 1511, 1145, 1146,
 1147, 1148, 1149, 1499, 1150, 1496,   37,   38,   39, 1174,
   40,   41,   42,  600,   43,   44,   45,   46,   47,   48,
 1151, 1152, 1924, 2307, 2701, 3036,   49, 1935, 1489,   50,
 1112, 1483, 1906, 2285, 2688, 2689, 3028, 3665, 3337, 3661,
 3779, 3329, 3330, 3529, 3331,   51,   52,   53, 2316,  732,
  733,  734,  735,  736, 1535, 1184, 1531, 2417,  444, 1188,
  739,   55,   56,  422,  423, 1397, 1398,  446,  447,  448,
  742,  743,  744, 1198, 1543, 1544, 1545, 1546, 1956, 2320,
 1461, 1462, 1463, 1464, 3001,  698, 1167, 1465, 1466, 1467,
 1468, 3007, 3008, 1469, 1892,  416,  585, 1890, 2714, 3348,
 3349, 3045,  836, 3353, 3538, 3539, 2214, 1310, 3235, 3236,
 3237, 3238, 3452, 3881, 3884,  747,  748, 1549, 1550,  642,
 2329, 2717, 2105, 2106, 1208, 1564, 2418,   58, 1212, 2335,
 2336, 2419, 3314, 3514, 3515, 3930, 3927, 2420,   61,   62,
 2421,   64,   65, 2422, 3895, 1210,  837, 2423, 1523, 1940,
 2424,  783,  784, 1242,  750, 1214, 1979, 1024, 1025, 2425,
 1307, 1026, 1027, 1028, 1426, 1029, 1821, 1030, 1031, 1032,
 1427, 1033, 1034, 1035, 1036, 1037, 1038, 1039, 1802, 1803,
  626,  580,  757, 1270, 1657, 2085, 2523, 2711, 3429, 3598,
  827,  790, 1645,  844, 1248, 1249, 3583, 3584, 3712, 2830,
 3134, 2474, 2825, 1308, 1041, 1042, 1558, 1043, 1044, 1045,
 1333, 1335, 1046, 1692, 1047, 2063, 1048, 1049, 1050, 1051,
 1052, 1053, 1054, 1055, 1056, 1782, 1057, 1532, 1301,  619,
 1666,  620, 1910, 1481, 1482, 1058, 1059, 1060, 1061, 1062,
 3260, 1775, 1063, 1064, 2215, 2624, 3271, 3750, 3751, 3841,
 3924, 3478, 2585, 1065, 1066, 1067, 1068, 1069, 1070, 1071,
 1072, 1073, 1074, 1075, 2966, 2953, 3265, 2968, 3632, 3758,
 3759, 3847, 3848, 1076, 1770, 1077, 1078, 2911, 1079, 1731,
 1080, 1081, 1082, 1083, 1084, 1085, 2915, 3249, 1086, 1339,
 3246, 1340, 2132, 2583, 1087, 1709, 1088, 1089,  560, 1303,
  754,  633, 1090, 2793, 2101,  648,  611,  645, 3796,  614,
 3882, 1168, 2428, 1185,  573, 1877, 1091,  576,  625,  627,
 1403,  661,  438, 1092, 2216, 1093,  420,  137, 1458, 1097,
 2229, 3282, 2645, 1098, 2253, 1456, 2647, 1850, 3310,  140,
 1475, 3517, 3018, 1476, 2983, 3306, 2254, 1297, 1231, 2895,
 2544, 1847, 2970, 3226, 3437, 2540, 2893,  593,  789, 2809,
 2505, 2857, 2504, 3156, 3158,  798, 2068, 2490, 3413, 3417,
  800, 1611, 2443, 2785, 1642,  771,  788, 1246, 3403, 3821,
 1659, 3605, 2888, 3432, 3603,  796,  759, 2058, 1573, 3049,
 3354, 3783, 3856, 1283, 2081, 3160, 3422, 2449, 1660, 3435,
 2507, 3159, 2861, 3571, 3700, 3572, 3868, 3577, 3704, 2810,
 2811, 3812, 3912, 2450,  776, 1235, 2027,  770, 2431, 2737,
 3059, 3546, 2736, 3058, 3376, 3559, 3062, 3363, 3677, 3900,
 3901, 3070, 3374, 3557, 3794, 2749, 3069,  673, 1095, 1848,
  822, 1288, 2084,  777,  453,  760, 1219, 1997, 2035, 2812,
 3703, 1985, 2728, 1153, 1505, 1491, 1911, 1154, 1155, 1156,
 1157, 1258, 1159,   89, 1107,  794,  682, 2286, 1520, 1931,
 3344, 2276, 2682, 3022, 1529, 1658, 1204, 1199,  694, 2259,
 1675, 2104, 2903, 3617, 3836, 3618, 3954, 1970, 1973,  450,
 2659, 3004, 3512, 3769, 3850, 3893, 3849, 3892, 4020, 4047,
 3997, 4036, 1164, 1936, 2309, 3851, 1570, 1524, 2310,  751,
  823,  758, 1241, 1216, 2076, 2513, 3214,  828, 1661, 2051,
  791, 1629, 1646, 2502, 1630, 2475, 3408, 3476, 3625, 3475,
 1717, 2134, 2571,
};
static int yysindex[] = {                                   8192,
 5873,   -8,54885, 2326,  740,55620,  852,    0,  767,  881,
    0,   -8, 2048,135091,135589, 3615,141067, 1008, 1209,    0,
 1583,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1349, 1284,    0, 1551,    0,    0,
    0,    0,    0,    0,    0,    0, 1621, 1664,    0,127123,
141067,   -4,141067,141067,141067,141067,141067,141067,141067,141067,
141067,141067, 1536,    0,141067, 1250,    0,    0, 1778, 2308,
141067,141067,141067,141067, 1974,141067, 1803,141067,141067,    0,
141067,141067,141067,141067,141067, -156,    0,141565,    0, 1788,
 1988,141067,    0,141067,141067,141067, 2031,141067,141067,141067,
141067,141067, 1803,141067,141067,141067,141067,141067,141067,141067,
 1758,    0,142063,    0,141067,    0,    0,141067,141067,    0,
 1369,141067, 2049,    0,141067, 1655, 1881, 1992, 2085, -222,
 1763,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 2134,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 2078,    0, 2226,    0,    0,
141067, 2463,    0, 2495,    0,141067,141067,141067,141067,141067,
141067,141067,141067,141067, 1950,142561,    0,    0,    0,141067,
 6695,    0, 2133, 2153,    0,  821,    0,    0, 2182,    0,
136087, 2170,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2308,    0,141067, 1872,    0,    0,    0,  -49,    0,    0,
    0,    0,    0,    0,    0, 1286,    0,    0,    0,    0,
  -49,    0,    0,    0,    0,    0,    0, 1313,    0,    0,
 2170,    0,    0,    0, 2609,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1935,    0,    0, 2275,    0,
    0, 2255,    0,    0,    0,    0,    0, 2462,    0,    0,
 2308,    0,    0,    0, 3242,    0,    0,    0,    0,  118,
    0, 2428,    0,   37,    0,    0,    0,    0,    0,    0,
    0,    0, 2464,    0, 2363,    0,    0,146545,    0,    0,
    0, 2453,141067,    0, 2461,    0, 2359,    0,    0,    0,
141067,    0,    0,    0,    0,    0,   61,    0,    0, 2329,
    0,    0,    0,    0,    0,    0,141067,141067,    0,113544,
 2772,    0,    0,    0,    0, 2335, 1096, 2346,    0, 2408,
141067,    0,    0,  872, 2387,    0,149739,141067, 2424, 2320,
 2518,    0,    0,141067, 2463,141067,141067, 2605,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 2457,146545,    0,
 2517, 2525,    0,    0,    0, 2572,    0, 2553,    0,    0,
    0, 2562,    0,    0,141067,    0,    0,    0,    0,    0,
    0, 2610,141067,141067,141067,141067, 2580, 2458,    0,  339,
  370, 1550,    0,    0, 2690, 1770,  497,    0, 2431, 2261,
 2930,    0,    0, 2573,    0,    0,    0,    0,    0,    0,
    0,    0, 2701,  -46, 2154, 2274,    0, 2747, 2754,    0,
    0,    0,    0, 2756, 2754,    0,    0,    0,    0, 2106,
    0,141067, 2758,    0,    0,    0, 2767,    0,    0,    0,
141067, 2777,    0,    0,    0,    0,    0,149780,    0,    0,
141067,141067,141067,141067,141067,141067,141067,141067, 2443,    0,
141067,    0,    0,    0, 2776, 1504,    0, 2340, 2314, 2754,
 2823,    0,    0,136087,141067,    0,    0,141067,    0, 2834,
141067,    0,    0,123768,104353, 2725, 2431,    0, 2774, 2786,
    0, 2465,141067,    0,    0,    0,    0, 3093, 3102, 3107,
 2705,    0, 3108,    0, 3114,    0, 3115,    0,    0, 3117,
 3119,114246, 3120,    0, 3127,    0, 3129,    0,    0,    0,
    0,    0,    0,    0,    0, 3130, 2727, 2729,    0, 3132,
 3137,    0,    0,    0,    0, 3139,    0,    0,105055, 3140,
 3143,    0,    0,    0,    0, 3145, 2843, 3148, 3153, 3154,
 3156, 3157, 3159,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 3161,
 3163, 3165,    0,    0,    0,    0, 3167,    0,    0,    0,
    0,    0,    0,    0, 3169,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 3170,
 3171, 3172, 3173, 3174, 3177, 3180, 3181,    0,    0,    0,
    0,    0, 3183, 3187, 3188, 3189, 3190, 3191, 3194, 3195,
 3197, 3198, 3199, 3203, 3204, 3205, 3207, 3208,    0,    0,
    0,    0, 3209, 3211,    0,    0, 3213, 3132, 3137,    0,
    0, 3214, 3218,    0,    0, 3219,    0,    0, 3223,    0,
    0,    0,    0, 3227, 3228,    0,119160,119160,105757,141067,
    0,    0, 2880, 3200,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,78672,
 2748,    0,    0,    0,    0,    0, 3245,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2982,    0, 2676,    0,    0,    0,    0,    0,    0,    0,
    0, 3257,    0, 3258, 3261, 3262,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 3264,    0,
 3259, 3267,    0,141067, 3268,    0,  363,    0,121176, 2939,
 2940, 2945,    0,    0,    0,    0,141067,  394,    0,    0,
    0,    0,    0,    0,  -63, 3277, 2890,    0,    0,    0,
 2891,    0, 2941, 3283,    0, 2983, 3286,  873,    0,    0,
    0,    0,  705,  415,    0,    0, 2897, 3283,    0,    0,
    0,    0,    0,    0,    0,    0, 3053,    0,    0,    0,
 2912,    0,    0,  251, 3293, 3295, 3296, 3297, 3286,    0,
    0, 2387,141067,    0,    0,    0, 2269,    0,    0, 3299,
141067,    0,    0,    0,    0,    0,    0,    0,    0, 1120,
 3008,    0, 3298, 3298,    0, 3009,141067,    0,    0,114246,
    0,    0,114246,    0,    0,    0,    0,106472,    0,    0,
114246, 1284,    0, 1284,114246, 3713,    0, 2840,106472,    0,
  195, 2242,    0,    0, 3050,    0,    0, 4176,    0, 2308,
 2989, -105,   -1,    0,141067, 2308, 1467,    0, 2308,    0,
 3310,    0,    0,    0, 3310,    0,    0,    0, 1232, 2970,
    0, 2591,    0, 2916, 2905,    0,    0,    0,    0, 3086,
    0,  516, 3063,    0,    0, 3272,  289, 3297,    0,    0,
 2308,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2170,    0,    0,    0,    0,    0,    0,    0,    0, 2308,
 2308,    0, 3319, 2629, 1543,    0,    0, 1762,    0,    0,
    0,    0,    0,    0, 3095, 2998,    0,    0, 2921, 2995,
    0,    0,    0,    0, 3013,105757,    0,78672, 1358, 3003,
    0,    0,    0,141067,    0, 1853,114246, 1081,    0, 2916,
141067,114246, 1900, 1982, 3375,114246, 2916, 1984, 2338,    0,
    0, 1803,    0, 1803,    0,114246,114246,78672,  897,    0,
114246,114246,114246, 3064,114246,114246,114246,114246,114246,100102,
99387, 2338,114246,114246, 2023,114246, 3336,114246,114246,114246,
114246,114246, 3338, 3341,114246,114246,114246,114246,114246,114246,
114246,114246,114246,114246,114246,114246,114246,114246,114246,114246,
 3344, 3019, 3019,    0,    0, 2413, 3347,  775,141067,114246,
114246,105757, 2347,    0, 2347,    0,    0, 3349,    0,    0,
 2624, 3351,    0,    0,121759,    0,114246,114246,119160,    0,
    0, 3353,  789,    0,114246,    0, 1848,    0,    0,114246,
 2958, 2979,    0,    0,    0,100804,114246, -123,141067,119160,
119160,119160,119160,119160,114246, 3359,122263,114246,114246,114246,
114246,107187,133099,141067,113544,    0,    0, 3014,    0,    0,
 3033, 3036, 3038, 3311, 3125,    0,  357,136585, 3380,120672,
 2081,    0,    0,    0,    0,    0,    0,    0,    0, 3381,
 3377, 3166, 3133, 3134,  363,    0,    0,    0, 2140,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  226,
    0,    0,    0,    0, 1803,    0, 3283,  218,    0,    0,
    0, 2961, 2965,    0,    0, 2966,    0,    0,    0, 3028,
    0,126744, 3401, 1803, 1803, 1803, 1803, 1803,    0,    0,
 2103,    0,  279,141067,141067,    0, 2916,    0,    0,    0,
    0,    0, 1120,141067, 3121, 3123,141067, 1120, 3404,11125,
11844,    0, 3138, 3409,    0,    0,54274, 3411, 3415,    0,
68606,    0,    0,69325, 3024,141067,    0, -139,    0,    0,
    0,    0, 1675,    0, 3128, 3116, 3122, 3421, 2431,141067,
 2959,    0, 4176, 2973, 2975,  632, 2976,  179,    0, 3039,
 2759, 2761,    0,    0,    0,    0, 1528,    0, 2308,    0,
 3031, 3031,    0,    0,    0,    0, 2865,    0,    0,   66,
 3212, 3192,    0,    0, 3424, 3425,    0, 3444, 3096,    0,
    0, 3212,    0, 3454,  701, 3135, 3136, 3144,    0, 1803,
   37,    0,141067,    0, 3067, 2901,    0, 1193,    0, 3016,
   88,    0,    0, 5554,    0,    0,  218,  218,    0,    0,
    0,    0,    0,    0,    0,141067, 3152,    0,    0,    0,
    0,    0, 3212,141067,    0,    0,    0,    0,    0, 3465,
    0, 1357,141067, 3469,    0,    0,123768,141067,78672,    0,
    0, 3162,    0,    0,    0, 3007,    0,114246,114246,70044,
114246, 3477,114246, 3478, 3479,12563,114246,114246,114246,114246,
79391, 3480, 3245, 3481,114246,114246,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 3215, 3485,
 3487,70763, 3124,114246,    0,114246,    0, 5339, 6027, 3488,
141067,13282,14001,14720,15439,16158,    0,    0,    0, 5262,
107889,    0,    0,    0,    0,124329, 3489,  297,54263,16877,
114246,114246,71522,    0,17596,54982,18315, 5841, 7530,    0,
    0,19034,19753,20472,21191,55701,56420,57139,57858,58577,
59296,60015,60734,61453,62172,62891,63610,    0, 1271, 3221,
 3224,    0,    0,    0, 3498,    0,    0,    0,    0,    0,
    0, 3502, 3508,72201, 2900, 3513,    0,    0,    0,    0,
141067,126254,    0, 3511,    0, 3510,    0, 1099, 4251, 2563,
105757,    0,    0, 3246,    0,  962,    0,    0, 3229,119160,
 3353,114246, 3149,114246, 4715,    0,114246,    0,    0,    0,
 3523, 3229, 3229, 2815,    0,    0,  155, 1424, 1424, 2347,
 2347,78672, 3146,148537,    0,    0, 3526,72920,73639,74358,
    0, 2652,    0,    0, 3529, 2880,    0, 3176,    0,    0,
 3158,  748, 1016, 3178,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 3155, 3168, 3538, 3540,
 3541, 2773,    0,    0,141067,    0,    0, 2916,    0, 3532,
 2311, 3548, 3222,    0, 3222, 1806,114246, 3222, 3265, 3273,
114246,    0,141067, 3285,    0,    0, 3234, 3235, 3241, 3311,
 3320,  826,    0,    0,    0,    0, 1869,    0,  380, 2656,
 3201, 3577,    0, 2387, 2681, 2871, 2872, 2877,136087, 2415,
    0,    0,    0,    0, 1803, 3591, 3594, 3595, 3598, 3600,
    0,    0,    0,    0,    0,    0,    0, 2725, 2431,    0,
    0, 3581, 3599,    0, 3606,    0,    0,    0, 1120, 1120,
    0,    0,    0,    0,121176, 3217,114246,141067,123269,114246,
 3160,114246,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 5111,126254,    0,    0,108591,    0, 2703,    0,
141067,    0,    0, 3313, 3316,    0, 3322,    0,    0,  -39,
 3231,141067,    0,    0,  743,  743, 1528, 1931, 1593, 1486,
141067,149780,    0,    0,    0, 2155,141067,141067,141067,125756,
    0, 1304,    0, 3192,    0, 3387, 3396,    0,    0,    0,
    0,    0,128617,    0, -142,    0, 3176,    0,    0, 3210,
137083, 2952, 2954,    0,127123,   55, 1206, 1673,    0, 2754,
 3412,    0,    0, 3399,    0,    0, 2942, 2948,    0,    0,
 3260,    0,141067,141067, 3382,   88,    0,    0,    0, 2221,
    0,    0,    0,    0,    0,    0,    0,    0, 1849,    0,
 3620, 1860,    0,  602,    0,    0,    0,    0, 3623,    0,
    0,    0, 1313,  109,    0, 1576, 3376,    0, 3260,141067,
141067, 1357, 3386, 1419,    0,    0,    0, 3300,    0, 3388,
    0,    0,    0, 3651,    0, 2431,21910,22629,129115,23348,
    0,24067,    0,114246,    0,24786,25505,26224,26943,114948,
    0,    0,27662,28381,114246,    0,    0,114246,109293,114246,
78672, 3393, 3288,    0,114246,114246,114246,    0,    0,    0,
    0,    0,    0,114246,    0,114246,71504, 3282,114246,114246,
114246,114246,    0,64329,64329,114246,    0,114246,    0,114246,
    0,114246,    0,    0,    0,    0,    0,114246,114246,114246,
114246,114246,114246,114246,114246,114246,114246,114246,114246,    0,
    0,    0,114246,114246,    0,    0,    0,114246,114246,    0,
    0, 2763,126254,141067,    0,119160, 3666, 2764,114246, 3395,
    0,    0,    0,114246, 2674,    0, 3229,114246, 3229, 3229,
 2916,119160,114246, 3097, 3669,    0,    0,141067,114246,141067,
    0,110008,    0, 2779,141067, 3673, 3451,    0,141067,141067,
141067,141067,141067,141067,141067,    0,    0,    0,    0,    0,
    0, 3453,    0,    0, 3455,    0, 2609,    0, 2609,    0,
    0,    0,    0,136585,    0,    0,    0,    0,    0,    0,
 3389,    0,    0,121176,121176, 2790,141067, 3381,    0, 3158,
  858, 1294, 3407, 3417,    0, 1869, 3445,  127, 1051,    0,
    0, 3206, 3202, 3225,    0,   72, 3691,  380,    0,123768,
 3337,    0,  218,    0,    0,    0,    0,    0, 3301, 3305,
 3694,    0,    0,    0,    0,    0,    0,    0,    0,141067,
 3003,    0,110710,141067,    0, 3413, 3266,    0,    0, 3700,
 2311,114246, 3426,    0,    0,    0,    0, 2197,    0,78672,
 -139, 3385, 3511,    0, 2801,    0,78672,141067,    0,    0,
 2308, 2308, 3316,    0,    0, 3350,    0, 2387,    0,    0,
    0, 3442, 3340,    0, 3723, 3490,    0,    0,    0, 3430,
 3431,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 3042,141067, 1023,    0, 3740, 3232,114246, 1465,    0,
    0,141067,141067,141067,149035,141067,114246,141067,    0,    0,
    0, 3482,125258,    0, 3724,    0,    0,    0,    0,    0,
    0, 3450,    0,    0,    0,    0,    0,    0, 1083,    0,
    0,    0,    0,    0,    0,    0, 3452,    0,    0,    0,
    0,    0,    0,    0, 3728,    0,    0, 3738,    0,    0,
141067,    0,    0,    0, 2808,141067, 3432, 3524,    0,    0,
    0,    0,    0, 3363, 3515,    0,    0,    0,  701,137581,
    0, 3068,141067, 3070, 2144,    0,    0, 1591,    0, 1370,
    0, 3438, 3439, 3441, 3447, 3448, 3449,    0,    0, 3446,
 3416, 3437, 1684,    0,  171, 3766, 3770,    0,    0,    0,
    0, 2221, 3623, 3623, 2308, 3472,133597,133597,    0,    0,
  218,    0,  218, 3053,    0, 3394,    0,    0,    0, 3328,
    0,    0,    0, 2387,    0, 2829,    0, 2106,  685, 2970,
 2970, 2775, 1576,    0, 3383,    0, 2308, 3390, 2308, 2308,
 2308, 3801,    0, 3266,138079,    0,    0,    0,    0, 3785,
  149,    0, 3787,  173, 3545,    0,141067,141067,    0,    0,
    0,    0, 3805,149702, 2308,    0,    0,    0,    0, 3484,
 3811,    0,    0,    0,    0,29100,    0,    0,    0,    0,
105757,29819,    0,    0,30538,75077,75796,    0,    0,114246,
 3420,31257, 2839,65048,31976,32695,114246,114246,65767,76515,
66486,67205, 2851,114246, 3815, 3817,77234,33414, 8249, 8968,
34133,34852,35571,36290,37009,37728,38447,39166,39885,40604,
41323,42042,42761,43480,77975,77993,    0,    0,    0, 2748,
    0,    0,    0,114246, 3229,119160,114246, 3229,114246, 3818,
    0, 3819, 3592, 3003,    0, 3832, 4009, 3175,    0, 2862,
    0,    0, 3524,    0,128617,    0,    0, 3457, 3459, 3466,
 3483, 3491, 3495, 3496,    0,    0, 1866, 2254, 2301,    0,
    0,    0,    0,  232,138577,    0,147043,121176,    0,    0,
 2768,    0,    0,    0,    0,141067,141067,141067,141067,141067,
141067,141067,    0,    0,    0,    0,    0, 3391,    0,    0,
    0,    0, 1803,    0,    0, 3555,  144,    0,    0, 3500,
    0, 3867,    0,    0, 3504, 3053, 3875, 3458, 3460,    0,
 3876,141067, 2261,    0, 3007,    0,78672,    0, 1120, 3269,
    0,121176, 3650, 3602,    0,    0,    0, 3558,    0,108591,
 2431,    0,    0,    0,    0,136087,    0, 3596,    0, 3220,
 3233,114246,    0,    0,    0,    0,    0,111412,    0,114246,
 3517,78672,    0,    0,    0,    0,    0, 3601,    0,    0,
78672,    0,    0,    0,    0, 3501,    0,    0,126254,  787,
 3898,126254,114246,    0, 3524, 3883,    0,    0,  612,    0,
 3212, 3212,    0,    0,    0, 3524,139075,141067,  607, 3607,
    0,  394,    0,    0,  218,    0,141067,    0,    0,    0,
129613, 1824,    0,    0,    0,    0,    0,    0, 3580, 1803,
 3586,141067, 3672,  298, 3063,130111, 3433,    0,    0, 3913,
    0,    0, 3533, 3627, 3543, 3674, 3639, 3675, 3212,    0,
    0, 3553,    0,    0,141067, 3323, 3649,141067, 3649, 3661,
    0,    0,    0, 1197,    0, 2387, 3534,    0, 2308,    0,
    0,    0,    0, 3524,    0,    0,    0,    0,    0,  394,
    0,    0,  602,    0, 3260, 2544,    0,    0,    0,    0,
137581, 2970, 1232,  698,    0,    0,    0,    0, 3614, 5332,
 3370, 8850, 9569, 3628,    0,    0,    0, 3531,    0, 3537,
    0,    0,    0,    0,    0, 3954,    0,    0, 2878,    0,
    0,    0,122767, 2884, 3965,    0,    0, 3966, 1480,    0,
    0, 3652,134095,139573,    0,    0,    0,    0,    0,114246,
 3967,114246,    0,114246,  809,    0,    0,114246,    0,    0,
44199,44918,114246,114246,114246,114246,114246,    0,78672,    0,
    0,114246, 3693,114246,    0,114246,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,141067, 3276,114246,    0, 2748, 3229,114246, 3229,    0,
    0,114246, -160,    0,114246, 3461,141067, 3676,    0,141067,
 3950,    0,141067, 3212,    0,    0,    0,    0,    0,    0,
    0,  306,131605,  306,    0, 3453,    0, 3455,    0,    0,
    0,    0,    0, 3566,    0,  271,141067,147541,    0,    0,
 2495,    0,    0,    0,114246, 3977,    0,    0,141067, 3975,
 3706, 3712, 3715, 3719, 3721, 3725, 3729,    0,    0,  248,
    0,141067,    0,    0, 3603, 3603,  946,    0,    0, 3031,
    0,    0, 3278, 3289,    0,    0,    0, 3617, 2431,    0,
141067, 2311,114246,114246, 3304,    0,    0,    0,    0,    0,
141067,    0, 4001,141067,    0,    0, 4011, 3310, 3310, 4014,
78672,    0, 4017, 3760,    0,126254,    0,    0, 1766,    0,
    0, 3511, 3567, 3450, 3761,    0,114246, 3511,78672,  975,
    0, 3443,    0, 3776,    0,    0,141067,  975,141067,    0,
    0, 3436,    0,    0,    0,    0,    0,    0, 1430, 3672,
 3433,    0,  747,    0, 3782,    0,    0,    0, 3731,    0,
 3732,    0,    0,    0,    0,    0,    0,  643,    0, 3809,
114246,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  -17,141067,    0, 3649,    0,    0, 1197, 3803, 3739,
    0,    0, 2544,    0,    0,    0,    0, 1238,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 3316,    0,    0,
    0,    0,    0,    0,    0,    0,141067, 3680,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 3593,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 3474,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 3486,
 2291,    0,    0,    0, 2308,    0,    0,    0,    0,    0,
 3404, 4048,    0,    0, 3786,149505,149702, 1803, 1556,    0,
    0, 2308, 3162,    0,    0, 2893,    0,132103,    0,141067,
    0,78672,    0,45637,78672, 3671,114246,114246, 3788,46356,
    0,    0,47075,47794,48513,78672, 9687,10406, 2216,    0,
49232,49951,    0,114246, 4052,77975, 3229, 3245,    0,    0,
 3410,78672, 4057,    0,114246,    0, 4037, 2898,    0, 3524,
    0,    0,    0,    0,    0,    0,    0,    0, 3781,    0,
  271,141067,141067,141067,141067,141067,141067,147541,147541, 3392,
    0,    0,    0,  271, 3733,138577,    0, 3831,    0,140071,
    0,    0,    0, 2431, 3678,    0,141067,    0,    0,    0,
    0,    0,    0,    0,    0,  253,  253,    0, 4063,    0,
 3796,    0, 3734, 3682, 3456,    0,    0,    0, 3403, 3698,
    0,    0,    0,141067, 4074,    0,    0, 4075,    0,78672,
    0,141067, 2725,    0,    0,    0,    0,    0,    0, 3735,
    0,114246,115650, 3702,125756, 3511,    0,    0,    0,    0,
114246,114246, 3808,    0,125756,    0, 4083, 5554, 5554,    0,
    0, 3814,    0,    0, 4082,    0, 3793, 3795, 2609,    0,
    0,  771,    0,    0,    0,    0,    0,    0,    0,50670,
    0,    0,    0,    0,    0,    0,    0,132601,    0, 3737,
 2308,    0, 1238, 3790,    0,    0,    0,    0, 3316,    0,
    0, 1599,    0,    0,    0,    0,    0,    0, 2923,140569,
 4091, 2308,    0,    0,    0,114246,149505, 2308, 1803,   94,
    0,    0, 4094,    0,    0,    0,134095,    0,    0,    0,
    0,    0,    0, 4090,    0,114246,    0, 3722,    0,    0,
    0,    0,    0,114246,    0,114246,    0,    0,    0,    0,
    0,78672,    0, 3276,    0,    0,   -2,    0,    0, 4880,
    0,    0,    0,    0,  306,  306,  306,  306,  306,  306,
  306, 3771,131605,    0,    0,    0,    0,    0,    0,    0,
    0,    0,141067,    0,    0,    0,    0,141067,    0,    0,
    0, 3678,  687,    0,    0, 2937,    0,    0,    0,    0,
    0,    0,    0,    0,    0,141067, 1519,141067,    0,    0,
    0,    0, 1596,    0,    0,108591,114246, 4099,    0, 3881,
    0,    0,    0,141067,128617, 3176, 2944,105757,    0,78672,
125756,    0,78672,78672,141067, 1140,  908,    0, 3601, 3847,
    0,    0, 4108,141067, 2609, 2609, 3662, 3856,    0, 3899,
    0,    0,    0, 3611,    0,    0,  -18, 4125,141067,134593,
    0,    0,    0,133597,    0,136087,    0,    0,    0, 3464,
 3467,    0,    0,    0,    0,    0,141067,    0, 4109,    0,
    0,    0,    0,    0, 1211, 4129,  -70,51389, 4134, 3820,
    0,  803,    0,    0, 3053,    0, 4138,141067,141067,    0,
114246,78672,52108, 4146,101506,101506, 3535,    0, 3476,    0,
114246, 2029,141067,  271,  271,  271,  271,  271,  271,  271,
    0,    0,    0,    0,    0,    0,    0, 3629,    0,    0,
  253,  253,  253,  253,  253,  253,  253,    0,    0,    0,
 1221,    0,    0,    0,    0, 2946,    0,141067, 4151,    0,
 1033,    0, 3524,    0,    0,52827,    0, 3892, 3601,125756,
    0,    0,141067,141067,  380, 3748,    0, 3897, 3934,    0,
125756, 2916,    0, 3810, 3662, 1788,    0,    0, 3573,    0,
    0,  794,    0,  -18, 3560, 3363,    0,    0,    0,    0,
    0,    0,    0,    0, 3701,    0,    0,    0,140569,    0,
 2308,    0, 3636, 1713, 1938, 3632,    0,    0,    0,    0,
 3935,    0,    0,    0,    0, 2308,    0, 3763,    0,    0,
    0,    0,    0,    0,    0,102221, 3565, 3503, 3730,    0,
    0,    0,    0,    0,78672,    0,    0, 2313,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,148537, 1033, 3623, 3956,    0,  607, 4184,    0,    0,
125756,    0,    0,  908,    0,    0,    0,    0,    0,125756,
 3937,    0,    0,    0,    0, 4186,141067,    0, 3901,    0,
 3524,    0, 3524, 3933,    0,    0,    0,    0, 3515,    0,
141067,141067,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  -70,  523,    0,    0,141067, 3902, 3578, 3516, 3078,
 3974,    0,    0,    0,    0,    0, 4196,114246,    0,    0,
 3677, 4197,    0, 2387, 3864, 2387,    0,    0,    0, 1878,
    0,    0,    0,141067,    0, 2609,141067,    0,    0,    0,
 3958,    0,    0,    0,  218,    0,    0,    0, 3742, 3743,
 2962,    0, 4203,    0,    0,    0,    0,    0,102936,    0,
78672,  110,  135,    0, 2431,    0, 2544,    0,    0,    0,
 1878, 3822, 3939,114246, 1922,114246,    0,    0, 3882,    0,
 4204, 3958,    0,    0,    0,    0,    0,141067,    0,    0,
 3597, 3563, 3226,    0,114246, 3929,    0,114246, 1300,    0,
 2930,    0,    0, 2963, 2969,116352,    0,114246,117054,78672,
 3840, 3879, 3846,78672, 1312, 3662, 3900, 3904,    0,    0,
    0, 1445,    0,141067,    0,    0,    0,    0, 3849,    0,
 3852,    0, 3873,    0,    0,    0,    0,    0,148039,    0,
    0,78672, 3668,    0,78672, 3679,78672,    0,    0,    0,
 1406, 1406,    0, 3982,    0, 3854,    0,    0, 3945,141067,
 3662,    0,    0,    0, 2987, 3953,    0, 1320,    0,143059,
114246, 3683,114246,114246, 3835, 1663,    0,    0,    0, 4246,
  978, 3662, 2609,    0,    0,    0, 2261,    0, 3887,    0,
    0,    0,114246,78672,78672,    0,    0,    0,    0,  978,
 3857, 3515,    0,    0, 3882,    0, 3862,    0, 4260,78672,
    0,  394,  218, 3662, 2261, 4255,    0,    0,    0,    0,
 3878,108591, 4260, 4266, 3004,108591,    0, 3006,    0,
};
static int yyrindex[] = {                                   3907,
    0,  526, 1641,    0,    0,    0,    0,    0,    0,    0,
    0, 1082,    0,    0,    0,    0,    0,    0,    0,    0,
 4308,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1754,    0,    0,  296,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1333,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 3980,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1333,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,97607,    0,    0,
    0, 3910,    0, 4059,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1979,112127, 1924,    0,    0,    0,  369,
    0, 1447,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 4060,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 4060,    0,    0,    0,    0,    0,    0,  222,    0,    0,
 1447,    0,    0,    0, 4062,    0,    0,    0, 1659,    0,
    0,    0,    0,    0,    0, 4060,    0,    0, 1825,    0,
    0,    0,    0,    0,    0,    0,    0,149899,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1199,
    0, 4281,    0, 4066,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 4067,    0,    0,
    0,    0,    0,    0,    0,    0,  460,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1873,    0,    0, 1895,
    0,    0,    0,    0,    0,    0,    0, 2025,    0,  332,
    0, 6211,    0,130609,    0,    0,131107,143557,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 3936,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 4067,    0,
 2075,    0,    0,    0,    0,    0,    0,144055,    0,    0,
    0,144553,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1793,    0,    0,    0,
    0,112842,103651,    0,    0, 3940, 1061,    0, 2080, 1503,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1019,    0,    0,    0,    0,    0, 2111,    0,
    0,    0,    0,    0, 2111,    0,    0,    0,    0,  267,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 2450,    0,    0,    0,    0, 2111,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  301, 2080,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,60136,    0,    0,    0,    0,    0,
    0,    0,    0,    0,60855,    0,    0,    0,    0,    0,
    0,    0,    0, 8368,    0,    0,    0,    0,    0, 6930,
 7649,    0,    0,    0,    0,53665,    0,    0,    0,61574,
62293,    0,    0,    0,    0,63012,63731,    0,    0,    0,
    0,    0,    0, 9087, 9806,10525,11244,11963,12682,13401,
14120,14839,15558,16277,16996,17715,18434,19153,19872,54384,
55103,64450,20591,21310,22029,55822,65169,22748,23467,24186,
24905,25624,26343,27062,56541,27781,28500,29219,29938,30657,
31376,32095,32814,33533,34252,34971,35690,36409,37128,37847,
38566,39285,40004,40723,    0,    0,41442,42161,42880,65888,
66607,67326,68045,68764,69483,70202,70921,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,43599,44318,
45037,45756,71640,57260,46475,47194,57979, 6930, 7649,47913,
48632,72359,73078,49351,    0,    0,    0,    0,    0,50070,
50789,51508,52227,58698,59417,52946,    0,    0, 3907,    0,
    0,    0, 2117,76673,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  795,
    0,77392,    0,    0,    0,    0,  454,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,73797,
74516,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,75235,    0,  943,    0, 3727,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1049,
    0,    0,    0,    0, 2130, 3348, 2477,    0,    0,    0,
 3720,    0,    0, 4162,    0,    0, 4469,    0,    0,    0,
    0,    0, 3918, 4531,    0,    0, 3773, 4162,    0,    0,
    0,    0,    0,    0,    0,    0, 4106,    0,    0,    0,
    0,    0,    0, 4724, 4759,    0, 3981,  328, 4469,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 4006,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1110,    0,    0,
    0, 3907,    0,    0,    0,    0, 1796,    0,    0,    0,
 2167,    0,    0,    0,    0,    0,    0, 4071,    0,    0,
 3949,  214, 4071,    0,  184,    0,    0,    0,    0,  244,
    0,    0,    0, 3907,    0,  984,    0,    0,    0,    0,
 2006, 4531, 1020,    0,    0, 2817, 4343,    0,  220,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1447,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 4071, 1680, 1495,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 3907,    0, 3090, 4306,  231,
    0,    0,    0,    0,    0,117756,    0,117756,    0, 3907,
    0,    0,117756,117756,93000, 4307, 3907,117756,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 3915,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,117756,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 3907,79549,    0,80268,    0,    0,    0,67924,53546,
    0,    0,    0,75954,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
118458,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 4307,    0,    0,
    0,    0,    0,    0,  332,    0,    0, 6709,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 3907,
 4012,    0,    0,    0,    0,    0,    0,    0,    0,95068,
    0,    0,    0,    0, 3727,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 4162,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2193,    0,  460,    0,    0,    0, 3907,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1794,    0,
    0,    0,    0,  973,    0,    0,98267,    0, 1414,    0,
 1030,    0,    0, 1724,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 2281,    0,    0,    0, 2080,    0,
    0, 1495,    0, 4010,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1202,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 4084, 4085,    0,    0,    0,    0,145051,    0,    0,
    0,    0,    0, 2284,    0,  199,  201,  274,    0,    0,
  352, 1495,    0,    0, 2318,    0,    0,    0,    0,    0,
 4049,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1024,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 4050,    0,149530,    0,    0,    0,    0,    0,    0,
    0, 4318,    0,    0,    0,  250,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 3010,    0, 4319,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 5319, 5475, 5772,124481,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,80110,    0,78111,    0,    0,
    0,    0,    0, 2322,    0,    0, 2268,    0,93704,    0,
 3907,    0,    0,    0,    0,    0,    0,    0,90184,    0,
    0,    0,    0,    0,94408,    0,    0,    0,    0,    0,
    0,85256,85960,    0,    0,    0,78830,82425,83144,80987,
81706,  862,    0,  206,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 2117,    0, 4098, 6727, 7425,
145549,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 3995, 8131,
 4077,    0,    0,    0,    0,    0,    0, 3907,    0,    0,
 4012, 2004, 4030,    0, 4030, 4012,    0, 4030,    0,    0,
    0,96429,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 2927, 1348,    0,    0,    0,
 4917,    0,    0,    0,    0,    0,    0,    0,    0, 2355,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
98785,    0,    0,    0,    0,    0,    0,  301, 2080,    0,
    0,    0,    0,    0,    0,  991,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1604,    0,    0,    0,    0,
 1578,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1405,
    0,146047,    0, 4107,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  174,    0,    0,127621,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 2111,
 2356,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 4065,    0,  774, 2213,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2005,    0,    0,    0,  275,    0,    0,    0, 3923,    0,
    0,    0, 2455,    0, 2163, 3891,  671,    0,    0,    0,
    0, 4050,    0, 4078,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 2080,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1367,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 4344, 4344,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,90888,    0,86664,91592,
 3907,    0,    0,  227,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  960, 7480,    0, 1065,    0,  474,    0,  474,    0,
    0,    0,    0,    0,    0,95769,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,95068,    0,145549,
    0,    0,    0,    0,    0, 1348,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 3983,    0,    0,    0,    0,
 4985,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  217,    0,
  231,    0,    0,    0,    0,    0,    0,    0,    0, 1679,
  278,    0,  280,    0,    0,    0,    0,    0,    0, 1797,
    0,    0, 2357,    0,    0,    0, 3012,    0,    0,    0,
    0,    0, 1889,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 2626,    0, 2631, 2487,    0,    0,    0, 2669,
 2778, 2805,    0,    0,    0, 2841,    0,    0,    0,    0,
    0,    0, 1487,    0,    0,    0,    0,    0,    0,    0,
    0, 4329,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2999,    0,
    0,    0, 4071, 4071,    0,    0,  994,  141,  312,    0,
    0,    0,  214,  255,  407,    0,    0,    0,    0, 1620,
    0,    0,    0,  951, 2087,    0,    0,    0,    0, 3880,
 1907,  295,  330,  377,  396,  499,  503,    0, 4071,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 3923, 3923,    0,    0,    0,    0,    0, 1262,
    0,    0,    0, 4566,  220,    0,    0,    0,    0, 1261,
 1495,    0,    0,    0,    0,    0,    0,    0,    0,  -45,
 4021,    0, 3891,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 4060,    0,    0, 4060,    0,    0,    0,    0,    0,    0,
    0,    0, 3998,    0,    0,    0,    0,    0,    0, 4353,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 3907,    0,    0,    0,    0, 4354,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  162,    0,    0,    0,    0,83848,
    0,    0,    0,    0,87368,    0,    0,92296,    0,    0,
    0,    0,    0,  180,    0,    0,  198, 4061,    0,    0,
    0,    0, 4335,    0,    0,    0, 4071,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  234,    0,    0, 2769,    0,    0,    0,
 4012,    0,    0,97089,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 5049,    0,    0,    0,    0,
 2376,    0, 3971,    0,  250,    0, 1527,    0,    0,    0,
    0,    0,    0,  225,    0,    0,    0,    0,    0,    0,
 2080,    0,    0,    0,    0,    0,    0, 2293,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 4340,    0,    0,
    0, 4347,    0,    0,    0,    0,    0, 4349,    0,    0,
 4350,    0,    0,    0,    0, 4100,    0,    0,    0, 3907,
    0,    0,    0,    0,  303,    0,    0,    0, 4122,128119,
    0,    0,    0,    0,    0, 1328,    0,    0,  186,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 4373,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 2438,    0, 4101,    0, 4101,    0,
    0,    0,  988,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 2440,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1086, 1095,    0,    0,    0,
 1620,  -45,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  308,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1144,    0,    0,
    0,    0,    0,    0,149622,    0,    0,    0,  480,    0,
    0,    0, 4306,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 4111,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 4376,    0,
    0,    0, 3800,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 4378,    0,    0,84552,88072,    0,88776,    0,
    0,    0,  188,    0,    0, 4381,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  237,    0,  237,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 2442,    0, 1031,    0,    0,
 1164,    0,    0,    0,    0,    0,    0,    0,    0, 2451,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 2442,
    0,    0,    0,    0,  998,  998, 4763,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2080,    0,
    0, 1377,    0,    0, 2454,    0,    0,    0,    0,    0,
    0,    0,  874,    0,    0,    0,    0, 3949,  801,    0,
 4364,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  292,    0,    0,    0,    0,    0, 4365, 4366, 4368,
    0,    0,    0,    0,    0,    0,    0, 3044,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 4455,
 5147,    0, 3216,    0, 3023,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 2150,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 4101,    0,    0, 4101,    0,    0,    0,    0,    0,
    0,    0, 2570,    0,    0,    0, 1644,    0,    0,  275,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  223,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  812,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  958,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2469,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,119664,120168,    0,    0,    0,    0,    0,    0,
    0, 4387,    0,    0, 4013,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 3045,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  162,89480,  247,    0,    0,
 4388,  208,    0,    0,    0,    0,    0,    0,    0, 3058,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2442,    0,    0,    0,    0,    0,    0, 1031, 1187,  571,
    0,    0,    0, 2442,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  315,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 2442, 2442,    0, 4904,    0,
 5005,    0, 3983,    0,    0,    0,    0,    0,    0, 5467,
    0,    0,    0,    0,    0,    0,    0, 1257,    0,  635,
    0,    0, 3332, 2047,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  819,  989,    0,    0,    0,    0,
    0,    0,    0,    0,  819,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 3919,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2470,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 2472, 1218,    0,    0,    0,    0,    0,    0,  959,
    0,    0,    0,    0,    0,    0, 4306,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 4391,    0, 4378,    0,    0,    0,    0,    0,  264,
    0,    0,    0,    0,  237,  237,  237,  237,  237,  237,
  237,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1275,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 2640,    0,    0,
    0,    0,    0,    0,    0, 1011,    0, 3907,    0, 1161,
  819,    0, 4126, 4127,    0,    0,    0,    0, 4172,    0,
    0,    0,    0,    0,    0,    0, 2341,    0,    0,    0,
    0,    0,    0,    0,    0,  287,    0, 2475,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 2480,    0,    0,    0,    0,
    0,    0,    0,    0, 1514,    0,    0,    0,    0,    0,
    0, 3066,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  166,    0, 2442, 2442, 2442, 2442, 2442, 2442, 2442,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2442, 2442, 2442, 2442, 2442, 2442, 2442,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 2279,
    0,    0,  176,    0,    0,    0, 1391,124760, 4349,  819,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 4144,
  819, 3907,    0,    0, 2341, 3980,    0,    0,    0, 5484,
 5484,    0,  926,    0,    0,  969,    0,    0,    0,    0,
    0,    0,    0,    0,  997,    0,    0,  223,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1004,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  299,    0,    0, 4403,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  206,    0, 3923, 4002,    0, 1047,    0,    0,    0,
  819,    0,    0,    0,    0,    0,    0,    0,    0,  819,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2607,    0, 2607,    0,    0,    0,    0,    0,  407,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 3072,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
  335,    0,    0,    0, 4003,    0,    0,    0, 1618,    0,
    0,    0,    0,    0,    0, 1018,    0,    0,    0, 5729,
 1688,    0,    0,  993,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 4410,    0,    0,    0, 2080,    0, 2330,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  573,    0,
 5911, 1688,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1691,
    0,    0,    0, 1773,    0, 2341, 1698, 3270,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,124505,    0,
    0, 3079,    0, 6038, 1965,126967, 2266,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
68429,    0,    0,    0,    0,    0,  217,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1412,
    0,68429, 3905,    0,    0,    0, 3971,    0,    0,    0,
    0,    0,    0, 3080, 2378,    0,    0,    0,    0,    0,
    0,  407,    0,    0, 4505,  217,    0,    0,  307, 3081,
    0,    0,    0,68429, 3971,    0,    0,    0,    0,    0,
    0,    0,  506,    0,    0,    0,    0,    0,    0,
};
static int yycindex[] = {                                      0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 3930,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 2513,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1326,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 3714,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 3813,    0,    0,  368,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 3716,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 3976,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 4420,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 4415,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 4416,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 4170,    0,    0,
    0,    0, 4171,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 4173,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 4427, 4430,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  837,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 4179,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 4169,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1090,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  710,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 4086,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 3744,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 4175,    0,    0, 4180,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 4432,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 4220,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 4097,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1819,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 4441,    0,
    0,    0, 4194, 4195,    0,    0,    0,    0, 4128,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 4089,
 1471,    0,    0,    0,    0,    0,    0,    0, 4199,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 4189,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 4206,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 3968,    0,    0,    0,
    0,    0,    0,  942,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 4456,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 3906,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1960, 1999,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 3947,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 3911, 3912,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 4460,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 4114,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2195,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 4248,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1308,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 2248,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1224,    0,
    0,    0,    0, 4470,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,
};
static int yygindex[] = {                                      0,
    0,    0,    0,    0, 4517,    0,    0, 2618, 3046,    0,
 3048, 3049, 1880, 3054,    0, 1871, 3057,    0, 1884,-2309,
    0,  663,    0,    0,    0,    0,    0,-2393, 1027, 1041,
 3061, 2285, 1863, 1234,    0,    0,    0,-2842,    0, 2000,
 1315,    0,    0, 1106,    0,    0, 2881,    0,    0,    0,
    0,    0,    0, 4118,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1387,    0,    0,    0, 1137,-1818, 1448,
-1963,  937, 2057,    0, 1701,-1890, 3747, 1409,    0, 4474,
 2118,    0,    0,  859,    0,-2633, 4464,    0,    0,    0,
    0,  749,    0,    0,    0,    0,  965,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2089,    0,
    0, 1435,-2189,    0,    0,  791, 4148,  861,    0,-1870,
 -563,    0, 3414,    0,  750,    0, 2053, 1428, 1722,    0,
    0,    0,    0,  558,-1769,    0,-2452,    0, -313,  888,
    0,  679,    0, 1732,-3463,    0,-2356,  559,  647,  649,
  615,  101,-1853,-1820,   24,    0,    0,-1192,-1951,    0,
 1511,    0,  966,-2265, 1512,  102,-1813,-1801,  115,  815,
 -521, -627,    0,    0,  120,    0, 2943,  182,    0,    0,
 3895, 2514,    0, 2515,-1089,    0,    0, 3005,    0,    0,
 2613,-1257,-1253,    0,    0,    0,    0,-2228,-1459,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
-3189, 1062, 1864, 1868,    0,    0,    0,  629,    0,  630,
    0,    0,  729,    0,    0,    0,    0,-1076,-1658,    0,
    0,    0,    0, 2246,    0,  944,    0, 1877,    0,  842,
    0,    0,    0,    0,    0, 4637, 3230, 2793,    0,    0,
    0, 1670,  122,    0,    0,    0, -148,    0,    0,    0,
-1105,    0, 2135,    0, 1431,    0,-1822, 3405,-1872,    0,
    0,    0,    0, 2648,    0, 2198,    0, 2617,    0,    0,
    0,    0,  831,  949,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 3082,    0, 2658,    0, 1115,    0,    0,
 3041,    0,    0,    0,-1907,-2472,    0,    0, -648,    0,
 3022, 2173,-2371, -758,    0,    0,    0,-2348,    0,    0,
 -756,    0,    0, 3507,    0, -660,    0,    0,    0,    0,
    0,    0,    0, 4527,    0,    0,    0,    0,    0,    0,
    0, 3164,    0,    0,    0, 1325,    0,    0,    0,    0,
    0,    0,    0, 2398,    0, 1345,    0,    0, 1654,    0,
    0,    0, 1156,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 3518, 3948,-1363,    7,    0,    0,
    0,-1002,    0, 4264, 3990,-1609,    0,    0, 1535,    0,
    0,    0, 3944,    0, 3492,    0, 2731,    0,    0,    0,
-1309,-2000, 3244,    0,    0, 1705,    0, 3248,    0,    0,
    0,    0,    0,    0, 2444, 2259,  113, 2825,    0,    0,
 1176,    0,-1229,    0,    0, 1046,    0,-1809, 3427,    0,
 1269,    0,    0,  764,    0,-1165,    0,    0, 2762,-1519,
    0,    0, 3979, 2016,    0,    0, 4723,    0,-2560,-3296,
 2010, 4735,    0, 1227, 1229,    0,    0, 4744,    0,    0,
 4747,    0,    0, 4748,    0, -827, 3237, 2365,-1839, 2441,
    0,-2150,-2652,-1153,-2213,  823,    0,-1319, -745,-1344,
-1313,-1004,    0,    0,    0,    0,    0,    0,    0,    0,
 3339,    0,    0,    0,    0,    0,    0,    0, 2947,    0,
 4323,    0,    0,    0,-1413,    0, 2244, 2237,    0, 1043,
 3943, 3985, 3142, 3928, -650,  430,    0, 1186,    0,    0,
-1547,-1898,    0, 2521, -882, -964, 2799,    0,-1189,-1587,
 3780, 3797,    0, 3342, -950,-1019, 3398, 3239, 3243,-1155,
-1114, -721,-1576,-1539,-1508,    0,  -64,  -54,-1826, -818,
    0, -117,-1451,-1470, -667,    0,-1969,    0,    0,    0,
    0,    0,    0,    0, 1005,    0,    0, 1166,    0,    0,
    0,    0, 2642,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,-2639, 1327,    0,    0,    0,
    0,    0,    0,    0, 3419,    0,    0,    0,    0,    0,
-1917,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1557,    0,-2007,    0,    0, -312,-1501,    0, -110, 2509,
 -812, -119,   69, 2019,-2246,    0,  -67, 4691,    0,   35,
  -73,-1208,-1191, -545, -108,  -58,    6,    8,  -53,  -74,
-2001, 1513, -224,   71,-1406,  -14,    0,    0, 2990,    0,
 2542,   21,  128,    0,    0, 3360,  905,    0,    0,    0,
    0, -408, 2143,    0,-1984, 1531,    0,    0, 3527,    0,
 -744,-1393,-2467,    0,    0,    0,    0,   98, -418, 1930,
    0,    0,-2322,-2947,    0,    0,    0,-2179,    0,  875,
    0, -725,-1887,-3444,    0,    0,    0,-1217,    0,    0,
    0,-1541,    0,    0,    0,    0, -715, 1469,-1426,    0,
-2973,    0,    0,    0, 2740,    0, -531, 1237,    0,    0,
 1395, 1101,    0,    0, 1264,    0,    0,-2640,-2274,    0,
    0, 1025,  924, 2055,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,-3318,-1216,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 3633,    0,    0,    0,
    0,    0,    0,    0, 3705,    0,    0,    0,    0,    0,
    0, -513,    0, 4835,    0,    0,    0, -669,    0,-1568,
    0,    0,    0, 1827,    0, -785,    0, 1170,-1395, 1247,
    0,    0,    0,    0, 1107,    0,    0,-1688,    0,    0,
    0, 1851,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  963,    0,    0,    0,
    0, -369,    0,    0,    0,    0, 1263,    0,    0,-1485,
    0,    0,    0,    0,    0,    0,    0,    0, 1380,    0,
    0,    0,    0,
};
#define YYTABLESIZE 150645
static int yytable[] = {                                     419,
  425,  566,  439,  551,  581,  657,   54,  774,  653, 1311,
  665, 1290, 1110,  656, 1399, 1302, 1396, 1115, 1191, 1194,
  424,  418,  589, 1713,  578,  605,  670, 1915, 1628, 1720,
 1835, 1961,  792, 1539, 2357,  606, 1243, 1289, 1141, 1256,
 2426, 1257, 1613, 1218,  630, 1232, 2062,  655,  650,  775,
 1560,  660, 1394, 1394, 1400,  549,  552, 2064,  561,  564,
  567,  574,  419,  579,  582,  419,  590,  564, 1401,  663,
  597,  659, 1192, 1195, 1259, 2448,  609,  552,  561,  564,
 1251,  574, 2287,  419,  579, 1795,  582,  419,  590,  564,
  634, 1561, 2427,  597, 2065, 2703, 2652,  646, 2653,  419,
  649,  552,  820,  646,  561,  564,  567,  574,  563,  419,
  579,  582,  662,  590,  564,  634,  563, 2087,  597, 2347,
  672,  418, 2568,  419,  419, 2066,  417,  564, 2311, 1557,
  683,  658, 2067, 2192, 1393, 1395, 2746, 2747, 2748, 2005,
 2006, 2011, 3038, 1612, 1676, 2848,  418, 1234, 3096, 1254,
 1881, 1958, 2026, 2725,  654,  596, 1995, 1996, 1982,  608,
 2836, 2837, 2972,  664, 2755, 2765, 2071, 2071, 2973, 1947,
 2776, 3122, 2901, 1158, 1952, 2089,  680, 1294,  596, 1942,
 2971, 2362, 1656, 2498, 2769, 2495, 1559, 1181, 1183,  575,
 2528, 2197, 1286, 2082,  612, 3412,  616, 1653,  827,  468,
  825, 2552, 1980,  669,  637,  671, 1988, 2831, 2042, 1569,
  622,  781, 2829,  304, 2022,  711, 3149,  826,  840,  286,
 1446,  766,  410,  628, 1407, 1980,  417,  639, 1856,  667,
 1446, 2832,  526,  101,  468, 2346,   99, 2532, 1978, 3666,
 3220, 1625,  523,  840,  544, 1236, 1416,  843, 1979, 1461,
  675,  417, 2984, 1843,  309,  840, 1544,  827, 1593,  825,
  765, 2512, 1914, 2660,  840, 1407, 1059, 1854, 2655, 1909,
 2533, 3010,  304,  826,  297, 2995, 2529, 1361,  286, 1402,
  766,  410,  704, 1407, 2332, 3627, 1005, 1855, 2530, 1446,
 1514, 2995,  101, 1938,  832,   99, 3290, 3731, 1808, 1591,
 1418, 1399,  840, 1396, 1986, 3087, 1500, 3133, 1461, 1772,
 2845,  553, 3080,  309, 1525, 2333, 1944, 1684, 1361, 3628,
 1402, 1361, 1937, 3088, 1694, 1059, 3273, 1206, 1793,  833,
 1005,  735,  826,  297, 1522, 1472, 1361,  818, 1402, 1987,
 2151, 1400, 1824, 3269, 2016, 1005,  765, 3031, 2017, 3290,
  657, 1060, 1091,  832, 1222, 1091,  839,  688, 3790, 1418,
 3629,  468, 2308,  468, 1968, 1500, 3612, 1237, 2444, 3689,
 1091, 3925, 1091, 1525, 3875, 1693,  829, 1238, 1190, 1638,
 3543, 2008, 3705, 3440, 2462,  813, 1206, 1786,  833, 1396,
  735, 3146, 2733, 1522, 2678,  828, 3928, 1485, 1486, 1487,
 1874, 1021, 2133, 2053,  468, 2686,  425, 3914, 1544, 1193,
 1060,  552,  561,  564,  574,  419,  579,  582,  419,  564,
 1428,  597, 3025, 2426, 1299,  439,  424, 1400, 3227,   88,
 1572, 1429, 1021, 2826,  635,  829,  755,  526, 1479,  468,
 1484, 1243,  763,  636, 1394,  601,  602,  523, 3914,  544,
  468,  293, 2009,  736,  828, 1622, 1798, 1880,  765, 1409,
 2687, 2692,  827, 2693,  825, 1394, 1394, 1394, 1394, 1394,
  814, 3860, 1795, 1990, 2054, 2427, 2509, 1509, 1560, 2463,
  766, 3402,  765,  526, 1562,  766,  761, 1693, 1407, 3792,
  766, 1842, 3706,  523, 1521,  544, 2267,  101,  831,  827,
   99,  825,  830, 1488,  468, 1502,  596,  840,  827, 2445,
  825, 3307,  736, 1710, 1969, 1711, 3307, 1538, 1409, 1561,
  175,  689,  766,  175, 1945, 1186, 1800, 3926,  700,  702,
 1059,  766,  833, 1640, 3307, 1633, 1407,  826,  561,  468,
  838, 1361,  622, 1402,  840, 3613,  799, 1827, 1828, 1829,
 1830, 1831, 3929,  840,  706,  710, 1407,  831,  832, 2679,
 3270,  830, 2913, 1510, 1502, 3026,  827, 1059,  825, 3671,
 2356,  769,  502, 2262,  826, 1091, 1059, 1416,  523, 1825,
  827, 3936,  825,  826, 1186, 2318, 2319, 4043, 1091, 1361,
  657, 1402, 3114,  833, 3440,  832, 1407,  468, 1854, 1361,
 1361, 2620, 1402,  766,  832, 1407, 1091, 1109,  769, 1361,
 2150, 1402, 1395,  502, 2356, 1060,  502,  769, 1855, 1113,
 1566, 1091, 1448,  597, 3326, 3327, 3474,   22,  561, 2356,
  833,  502, 2482, 3404, 1406,  840,  561, 2868, 2869,  833,
  829,  826, 2718, 1543, 3142, 2321, 1592, 1407, 1059, 1361,
 1179, 1402,  846,  597, 1795,  826, 1268, 3003, 1361,  828,
 1402, 1407,  832, 1114, 1774, 3148,  567,  766, 2994, 1449,
  334, 1908, 1418, 1160, 3304, 1406,  832,  829, 1406,  425,
 1513,  425, 1169, 1446, 2532, 2863,  829,  454,  766,  769,
 2655, 2203,  840, 1406,  597, 1472,  828,  833,  766, 1165,
 1361,  424, 1402, 2541, 1615,  828, 1223, 3280,  830, 1416,
  582,  833, 1361,  835, 1361, 1793, 1402, 2533, 1182, 1182,
 1182, 1182, 1472, 2529,  284,  735, 2323, 1407, 1407,  334,
 1854, 1271, 1472, 1261, 1446, 2530, 3567, 3630,  847, 2702,
 1500, 1978, 3289, 1946,  829,  554, 1206, 1407, 1489, 2325,
 1855,  605, 1594, 1446, 1978, 4055, 2016, 1296,  829, 4058,
 2017, 1282,  831,  828, 1979,  309,  830,  609, 1543, 1843,
 2827,  468, 1461,  382,  819, 1873,  846,  828, 1941,  830,
 1361, 1361, 1402, 1402, 1114,  284,  552,  561,  564,  574,
  579,  582,  419,  564, 1738, 1543,  597,  454, 1396,  831,
 1361, 2039, 1402,  830, 1418, 1543,  468,  284,  831,  755,
  419, 1905,  830,  846, 1907, 1114,  567, 1560, 3879, 1304,
 2828, 2776, 2697, 1418, 1941,  617, 1991, 3484,  846, 1500,
  101, 1409,  382,   99, 2356, 1738, 1913, 1265, 1738, 1114,
 1409, 1538, 1265, 1988, 3308, 1394, 1988,  736, 1302, 3308,
 2198, 1506,  293, 1738, 2901, 2901, 1644, 1522, 1561,  523,
 2657, 1739,  284, 1141, 1478,  284,  831, 3308, 1446, 1874,
  830,  284, 2086, 2609,  175, 1880, 1856, 2043,  468, 1407,
  831,  596, 1477, 3104,  830, 2025,  766, 1738, 3485, 3486,
 3487, 3488, 3489, 3490, 3491, 1980, 1557, 1793, 1406, 2622,
 1446, 1643, 1739, 1262, 1266, 1739,  309, 1276,  411,  411,
  766,  526, 3694, 3695, 2074, 1854, 2075, 2060, 1186, 2099,
 1739,  523, 2803,  544, 2729, 1008,  284, 2205, 1272, 3115,
 1277, 1543, 1361, 1567, 1402, 1855,  827, 1292,  825, 1502,
 2266, 1986, 1091, 1986, 1986, 1986, 1406, 1854, 3142, 3933,
 2355, 2907, 1502, 1559, 1739, 2661, 3153, 1406, 2029,  766,
  411, 2030, 3152, 1409, 3150, 1942, 1406, 1855,  309, 1008,
 3092, 1407, 1350,   70, 3082, 3164, 1987, 2508, 1987, 1987,
 1987,  840, 1409,  319, 1008, 1445, 3116,  383, 1948, 3163,
 1470, 1951, 1009, 2722, 2355, 2078,  387, 1262, 1158,  185,
 1205, 2031, 2350,   97, 1059, 1404, 1406, 2356, 1450, 2355,
 1472,  826,  309,   24, 3633, 1406, 2855, 1538, 1020,  279,
 2071, 3102, 2847,  259, 1361, 1402, 1402,  309, 1502, 1452,
 2103, 1350,  832,  334, 2104, 3379, 1009, 2104, 1737, 1738,
  631, 1634,  319, 3568,  187,  769,  383,  652, 1233, 1091,
 3648, 1009, 2104,  114, 2104,  387, 1262, 1406, 1538,  502,
 1465, 1538, 3817, 2230,  523, 2901, 3814,  833,  363,  526,
 1452, 1406, 3673, 1452, 2103,  814, 1538, 1020,  279,  574,
 1777, 1186,  259, 1543,  672,  403, 1804, 2854, 1452, 2103,
 1302, 1778,  439, 3073,  261, 3165,  845, 3880,  138, 2040,
 3072, 1465, 2225, 3078, 2864,  544, 2298, 1233,   98, 1469,
 3247, 2312,  814,  363,  829,  605, 2608,  334,  334, 1465,
   49,  814, 1682, 3651, 3652, 3653, 3654, 3655, 3656, 3657,
  694, 1987, 1805,  828, 2231, 3390, 3552, 1406, 1406, 2275,
 1186, 1502, 1435,  337,  403, 2705, 3558, 1500,  419, 3737,
 1469, 1206, 1806,  261,  334,  605,  419, 1406, 1595, 2351,
 2317,  605, 1503, 1373,  605, 1588, 1738, 2356, 1469, 1471,
 3682, 1597,  609, 2667, 1607, 3634, 3635, 3636, 3637, 3638,
 3639, 3640,  814, 3681, 1779, 2573, 2278, 1051, 1738, 2496,
 2744, 2195,  745,  814, 2744,  523, 1641, 2807, 1714, 1715,
  334,  868,  337, 1548, 1373,  127,  745, 1373, 1396, 1501,
 1596, 3339,  139, 3683, 1548,  605,  605,  338, 1325,  659,
 3787,  816, 1373,   15, 3684, 1647, 1648, 3911, 2865, 2725,
 2278, 1394,   72, 1739, 3083, 2866,  831, 3093,  135, 3457,
  830,  334, 3248,  746, 2668, 2278, 1905, 1394,   52,   97,
 3649, 1562,  652, 2445,  694, 1739, 1403, 2104,  816, 2200,
  868,  269,  185, 1503, 1941, 1262, 3739,  816, 3009, 3378,
 2104, 2630,  845,  363,  526, 1522,  338, 2436, 2547, 1988,
 3738, 3136,  372, 1292,  317, 3166, 1350,  652, 2104, 1406,
 2501,  387, 1051, 1452, 2103,  279,  279, 1403, 1738, 1677,
 2524, 3392, 3569, 2104, 1350, 2201,  564,  187, 3679, 1826,
  544, 3085, 3086, 2610, 3097, 1403, 1193, 3685,  601,  602,
  269, 2233, 3678, 1288, 1465, 3815,  449,  372,  816, 2621,
 1716, 1262, 1188, 1845, 2234, 3395,  387,  745, 3808,  816,
 1196, 1452, 2481, 2771, 2772, 3040, 3217, 4011, 3398, 2500,
 1256, 1470, 1257, 2364,   98, 1685,  136, 2104, 3869, 3129,
 3872, 1452, 3340,  259,  701, 1739, 2232, 3280,  468,  309,
  707,  468, 1465, 1469, 1783, 1020, 1362, 1912, 1681, 2819,
 3784, 1406, 1193,  279,  317, 1538, 2032, 2033,  468, 1262,
 1797, 1188, 1465,  454, 2760, 2426, 1926, 1927, 1928, 1929,
 1930, 2355, 3042, 2355, 1262, 2426, 3138, 3366, 4048,  135,
 1796, 1193, 2468, 1448,  634, 1780, 1101, 1362, 3057, 2680,
 1362, 1469, 1836, 2148,  659,  403, 3063, 1373,  609,  561,
 1262,  969,  970, 1807,  262, 1362,  449, 1193, 3680, 1795,
 1193, 1469, 1795,  590, 3785,  672, 1628, 2427, 3620,  670,
 2278, 3793, 1503, 1504, 1448, 1193, 4037, 2427, 2963,  870,
 1254, 3854, 3805, 1530,  129, 1433, 1452, 2103,  605, 1302,
 1434, 1302, 1448,  605, 1186, 1373, 2669, 1286, 1373,  309,
  254, 1193, 1781, 1738, 4051, 1373, 1373,  372, 1262, 2681,
 1373, 3139,  659, 2047, 1630, 1373,  387, 1465, 1373,  440,
 1186, 2464,  870,  454, 1538, 1628,  330,  163, 1262, 1797,
 1169, 1844, 1373, 2355, 3281, 1738, 2739, 1452, 2761, 1182,
 1403, 1262, 1182, 1373,  605, 2235, 1534, 1616,  659, 1943,
 1471, 1373, 1452, 1452, 2004, 1373, 2342,  267,  269, 1193,
  659, 1967,  772, 3541, 1373, 1102, 1469, 2974, 3686,  814,
 1739, 3593, 1452, 1630, 1138, 1797, 1399, 1138, 1396, 1193,
 1373, 1617, 3861, 1465, 1941, 4031,  601,  602, 1403,  778,
 1534, 3863, 1193, 3723, 1111, 1943,  441, 1455, 3341, 1403,
  340, 2426, 1739, 1465, 3405, 1534, 1373, 3407, 1403, 3983,
  337,  337, 1941, 2048, 3674, 3962, 1400,  659, 1373, 2670,
 1373, 3414,  870, 1409,  969,  970, 2202, 3981,  609, 3777,
 2198,  773, 2671, 3743,  659, 4018, 2690, 3908, 1455, 2008,
  870, 1455,  468, 2278, 2104,  267,   10,  330, 1403, 4032,
 2465,  846, 1469, 2427, 4014,  166, 1455, 1403,  779,  419,
 1362,  442, 1941,  270, 1409,  254,  254, 1941,  419, 2803,
 1373, 1394, 1304, 2102,  654, 4034, 1618,  870,  316, 3778,
  443, 2083, 1409, 1266,  338,  338, 1373, 1373, 2044, 2044,
 2804, 1373, 2090,  330, 3988, 3434,  671, 1448, 1360, 1403,
 2727, 1193, 1793,  445, 1452, 1793, 1373, 4050, 1362, 1428,
 2009, 3316, 1670, 1403, 3724,  816, 3140,  503, 1362, 1362,
 1429,  340,  270, 1628, 1430, 2355,  564,  605, 1362,   61,
 2426, 1395, 1671, 1266,  330, 3909,  694,  316, 1452, 1360,
 2355, 2426,  670, 1459, 3963, 1448,   13, 2524, 2180, 2181,
 3351, 2104, 2430, 2956, 2537, 4030, 3982, 1360,  503, 3221,
 3725,  503, 3387, 2091, 4019, 1448, 1795,  340, 1362,  678,
 1452, 1286, 3229, 1317, 1193, 2138,  503, 1362,  654, 1403,
 1403,  599, 2427, 3377, 1459, 2834, 1672, 1628, 4005, 1640,
  654,  117, 1628, 2427, 2853, 2096,  609, 1797,   64, 1403,
 1634, 3726, 1459, 4017, 2268, 1998, 1138,  678,  340,  606,
 2062, 2062, 1319, 1320, 1317,  852, 1460, 1796,  694, 1362,
 2044, 2064, 2064, 2672, 1325, 2538, 4006, 2301, 3414, 1196,
   67, 1362, 1317, 1362, 3388, 1286, 2255, 1138, 1373, 1836,
 1286, 2426, 4045, 3230, 1190, 2806, 1628,  654, 1999, 3989,
 2426,  679, 2514, 1319, 1320, 3990, 2856, 1460, 2065, 2065,
  694, 1455, 1628, 1281,  654,  449, 1470, 1197,  284, 2361,
 1448, 1319, 1320, 3389,  852, 1460, 4007, 1138, 3798, 2191,
  590, 2340, 4035, 2182, 3023, 3360, 3361, 1409, 3991, 2066,
 2066,  117, 1086, 2427, 1286, 2358, 2067, 2067,  574, 1362,
 1362,  672, 2427, 1190,  605,  605, 1325, 2359, 1032, 1455,
 1286, 2799, 2491, 3048, 1081,  494, 2515, 2370, 1534, 1362,
 2371, 2516, 1477, 2491,  755,  246, 3659, 1534, 1138, 1455,
 1373, 1403,  284, 3050, 1673, 1409, 1448, 1448,  145,  272,
  270, 3660,  694, 1446, 2807, 1409, 1409,   61, 2000, 1477,
  146, 1086,  494, 1628, 3973, 1409, 1448,  246,  766, 1477,
  672, 2492, 1360, 1169,  609, 3540, 2001, 1032,  766,  671,
  601,  602, 2494, 1081,  678, 2517, 2360, 4027,  147, 1797,
 1534, 1534, 1548,  494, 1446, 1138,  634, 4002, 2466,  246,
  774, 2518, 2800,  503,  503, 1409,  598, 2349, 1321, 1796,
  503, 1286, 1446, 2002, 1409,  672,  609, 1459, 3992, 1793,
 1360, 2801, 2369,  561,  574, 2429,   64,  672, 3973,  454,
 1360, 1360, 2467, 1403,  494,  369, 4028,  272,  609, 2802,
 1360, 3268,  775, 3054, 1455, 1943,  609, 1317,  778, 1321,
 2461,  451,  678,  677, 2270, 1471, 1409, 1844,   67, 1202,
 1534, 1362, 1256, 3444, 1257, 1459,  679, 1321,  561,  574,
 1409, 3663, 1628,  494, 1628, 1628,  393,  766, 1104, 1534,
 1360,  494,  640,  372,  672, 1459, 1319, 1320,  678, 1360,
 1460,  503, 3445,  454,  369,  372, 3664, 1203, 1448, 2363,
  678,  672,  494, 1654, 1179,  561,  574, 2477,  494, 1525,
 1455, 1455,  503, 2270,  641, 1317, 1014,  779,  503,  452,
 1286, 2442, 1286, 1286,  609, 1628,  372, 1104,  372, 2455,
 1455, 1360, 1448, 2460,  679,  393, 1409, 1409, 1460, 2823,
  549,  494, 1196, 1360, 2534, 1360,  738,  148, 1810, 2639,
 1883,  503, 2640, 1362, 1319, 1320, 1409,  678, 1460, 1225,
 1014, 1410, 1884, 1179, 1448,  592, 2277, 2824, 1525,  149,
  679, 3902, 3105, 1286,  678, 1014, 1032, 1243,  150, 1001,
 1678, 1598,  679, 1971, 1885, 1470, 1470, 3118, 1032, 1196,
 1459, 2244, 1631, 3516, 1599, 2642, 1022, 2476, 3143,  549,
  599, 2519, 1811, 2520, 2521,  738, 3809, 2553, 1797,  609,
 4039, 1360, 1360, 1972,  503, 3579, 2074, 1446, 1225, 1032,
 1317, 1032, 1221, 1001,  156, 2246,  372, 1687, 1796, 1812,
  151, 1360, 1888, 2626, 2531, 2628, 2355, 4053, 1001,  679,
  609, 3346,  370, 2356,  419,  561,  574,  597,  552,  564,
  564, 1631, 1560, 1560, 2522, 1022,  679, 1651,  677, 1319,
 1320, 2278, 1455, 1460,  615, 1446, 2638, 2479,  156,  590,
 1652, 1196, 1321, 1196, 3367, 3903, 1459,  156,  617,  672,
  672, 1221,  425,  156, 3368, 1446, 2247, 3904, 1409, 1421,
 2280, 3540, 1844, 1561, 1561, 2248, 1455, 2074, 2643, 2644,
 3827,  370, 2663, 2479, 3369, 1304,  605,  605,  394, 1689,
 1526, 1695, 1196,  846, 3590, 3591, 2723, 2724,  905, 1032,
 2249,  156, 1048, 2635, 3828, 1797,  677,   94, 1455,  609,
  369, 2281, 2641,  643,  393, 2738,  156,   97, 2074, 1526,
 1321, 2074, 1525, 2088, 3905, 1943,  393,  753, 1363, 1460,
 1741,  739,  139, 1797,  676,  905, 2074,  100, 2074,  403,
 1363, 3906,  677, 1360, 1471, 1471,  766,  394, 1363, 1526,
  500,  393,  846, 1943,  677,  369,  651,  393,  142,  393,
 1409, 1048, 1363,  363, 1210, 1057, 1527,  666,  552, 2282,
 1446,  139, 2074, 1363, 1237, 2646,  142, 2745, 2102, 2102,
 2102, 2745,  647, 1797, 1238, 1214,  753,  681, 2429, 2283,
  739,  500, 2708,  363,  500,  363,  393,  142,  403, 3370,
 3371, 3372, 2284, 1943,  391, 1883, 2665,  676, 1943,  500,
 1363,  677, 2675,  684, 3915, 2887,  363, 1884,  685,  372,
  372, 3547, 2841, 1210, 1057, 1527, 2770, 2074,  677, 1633,
  605,  609,  605,  605, 2500, 1360, 1446, 1446, 3430, 1885,
 2838,  363, 2840, 2842, 1214,  609,  675, 1682, 2794,  468,
 1886,  103,  103,  106,  109,  686, 1446,  393, 1887, 1813,
  125,  687,  605, 1411,  605,  605,  605,  690,  391,  363,
  143, 2074, 2879,  691, 2881, 2882, 2883,  144,  248,  263,
 1556,  351,  582, 2843,  323, 1021, 1152, 1888,  692,  424,
  605,  372,  372, 1633,  372, 1679, 3511, 1683, 1633,  693,
 2902,  372, 1688, 1690, 1411, 3829, 1682, 1696,  468, 2635,
  103,  145, 1032, 1032, 2635,  369,  696,  370, 3107,  125,
  634, 1363, 1411,  146, 2986, 3941, 1152, 1152, 2791, 3830,
 1152, 3108,  561,  574, 1742, 3942, 3943,  248,  263,  676,
  351, 2074, 2226,  323,  697, 1152,  394, 1152,  424, 2987,
  708,  147, 1633, 1399, 2226, 1396,  393, 2226,  394, 1470,
 1048, 2870,  370, 2871, 2988, 3481, 3012,  737, 1633, 3013,
  372, 2990, 1048, 1932, 1933, 1934, 2074, 1152, 2483,  403,
  756, 1243, 4049,  394, 1032, 1032, 2872, 1032, 1446,  394,
 3000,  394, 1903, 1400, 1032, 1114, 2991,  403, 1889,  738,
  372, 2873,  372, 1048, 2796, 1048, 2797,  676, 2484,  403,
 2485, 2992, 3015, 1470, 1433, 1431,  468, 1432,  749, 1434,
 1904, 3959, 1428,  372, 3961, 2365, 2366, 2367,  394, 1941,
  609, 2486,  764, 1429,  403, 1883,  396, 1430,  403, 1941,
  403, 1624,  403,  676, 1623,  111,  112, 1884,  372, 1412,
  582,  675,  425,  672,  605,  676, 2487,  468,  782, 1633,
  468,  419,  561,  574,  597,  552,  564,  564, 3091, 1885,
  393,  393, 3002, 2690, 1790,  468,  372, 1435, 1568,  403,
 1886, 2221,  766, 3011, 2488, 2715, 2222,  403,   15,  394,
 1412,  793, 3110, 2221,  644, 2716, 2221, 3037, 2222,  630,
  396, 2222, 2221, 1048, 3468, 1435, 2289,  672, 1412, 2288,
  148,  795,  676, 2442, 3469, 3016, 3017, 1888, 3756,  675,
 3757,  755,  370,  766, 2259, 1433, 1431, 1411, 1432,  676,
 1434, 2294,  149,  766, 2293,  797, 2259,  629, 1471, 2259,
  403,  150,  393,  393, 1152,  393, 1552,  816, 1553, 3014,
 1152,  403,  393, 2339, 1797,  675, 2338, 1797, 4000, 1152,
 4001, 1152, 1152, 1152,  601,  602, 1152,  675, 1633, 1152,
 1633, 1633,  609,  634, 1796, 1411, 3076, 1796, 1371, 1225,
 1226, 1152, 2794, 3715,  605,  824,  609, 1152,  394,  825,
 1152,  831, 1471,  151, 3145, 1411, 1152,  609,  834, 1433,
 1431,  609, 1432,  835, 1434, 1697, 1698, 1699, 1700, 1701,
 1702, 1703, 1704, 2607, 2612, 1941, 2193, 1435, 1371, 1371,
 3131, 1633, 1371, 3135,  675,  842, 1106, 1094, 1889, 2632,
 1152, 1099, 2631, 2105, 2196, 1411, 1108, 1371, 1428,  403,
 2662,  675, 1114, 1435, 1411, 2105,  696, 1106, 2105, 1429,
 1152, 2721, 1161, 1430, 2720, 3090,  609, 1162, 2774, 1163,
 2110, 2773,  869,  870,  871,  872,  873, 1106,  875, 3106,
 1106, 1152, 2110,  699,  403, 2110, 1302, 1152, 1170, 2859,
 3112, 1152, 2858, 1171, 3119, 1106, 1411, 1152,  609, 2917,
 2299, 2300, 1435,  882,  883, 1173, 2271, 1176,  419,  609,
 1411, 2928,  394,  394, 2927, 1152,  257,  257, 2271,  699,
 1152, 2271, 2969, 1412, 1705, 1435, 1048, 1048, 3303,  403,
 3239, 3518, 3519, 3520, 3521, 3522, 3523,  403, 3218, 3154,
 3155, 2858, 3311, 3312, 3223, 1175, 1234, 2896, 1152, 2791,
 1706, 1177, 1152, 3448, 1941, 2616, 3447, 3263, 3483, 1428,
 1178, 3482, 1180,  403,  403, 1941, 1187,  403, 1189,  403,
 1429, 1412, 3274, 1209, 1430,  609, 1411, 1411,  609, 1424,
 1423, 1425, 1201, 3596,  394,  394, 3595,  394,  582, 1213,
  403, 1412, 3241, 1215,  394, 1211, 1411, 3650, 1048, 1048,
 1623, 1048,  582,  582, 3675, 1234, 3780, 1435, 1048, 2720,
 3845,    4,    7, 3846,  425,  403, 3283, 3284, 3285, 3286,
 3287, 3288, 3919, 3966, 1220, 3918, 3965,  419, 1227, 3968,
 1228, 1412, 3967, 1428, 3318,  403,  403, 1229,  403, 1233,
 1412, 1244, 1089,  403, 1429,  403, 3347, 4015, 1430, 3332,
 3918, 1152, 1371, 2242, 1245, 2870, 3355, 2871, 2635,  419,
 1250, 2635, 1275, 3732, 4057, 1941, 4059, 2720, 1285, 2720,
 1738, 1797, 1482, 1738, 1941, 1482, 2084, 1279, 2243, 2074,
 2872, 3358, 1412, 1089, 3561, 3562, 1089, 3932, 2244,  403,
 1616, 1796,  609, 2245,  609, 2873, 1412, 1284, 1707, 1708,
 1371, 1089, 1883, 1371,  535, 2045, 1287,  535, 2045, 1555,
 1371, 1371, 1106, 1106, 1884, 1371, 1106, 1295,  468, 1106,
 1371,  468, 2246, 1371, 1617, 1309, 2046, 1152, 1411, 2046,
 3443, 1106, 1443, 1312, 1314, 1443, 1885, 1371, 3406,  669,
  665,  666,  669,  665,  666, 1313,  674, 1886, 1371,  677,
 1555, 2260, 1316, 1555, 2263, 3005, 1371, 1424, 1423, 1425,
 1371, 1317, 1412, 1412, 2648, 2649, 1318, 1320, 1555, 1371,
  605, 1319,  419, 1321, 1322, 2635, 1323, 3385, 1324, 1326,
 3431, 1408, 1412, 2247, 1888, 1371, 1327,  605, 1328, 1329,
 1409, 1332, 2248, 1330, 3423, 1331, 1334, 3446, 1336, 1341,
 1106, 3495, 1342, 1410, 1343, 3497, 3897, 1345, 3899, 1344,
 1040, 1371, 1346, 1347, 1236, 1348, 1349, 2249, 1350, 1618,
 1351, 1106, 1352, 1371, 1353, 1371, 1355, 1106, 1356, 1357,
 1358, 1359, 1360, 1361, 1152,  279, 1362, 1411, 3496, 1363,
 1364, 3499, 1365,  419, 1412,  609, 1366, 1367, 1368, 1369,
 1370, 3000, 1413, 1371, 1372, 3510, 1373, 1374, 1375, 3501,
 1106, 3498, 1376, 1377, 1378, 3453, 1379, 1380, 1381, 1414,
 1382, 1415, 1383, 1386, 3006, 1371,  279, 1387, 1388,  279,
 1236, 1416, 1389, 1417, 1418, 1419, 1390, 1391, 3732,  504,
 1420, 1371, 1371, 1405,  279, 1889, 1371,  561,  582,  574,
  579,  582,  662,  590,  597, 1424, 1423, 1425, 1435, 1236,
 1407,  582, 1436, 1437, 1412,  582, 1438, 1439, 1089, 1089,
 1440, 1441,  609, 1442, 1443, 1089, 1472, 1446, 3454, 1473,
  504, 1421, 1444,  504, 1474, 1236, 1490, 1089, 1236, 1492,
 1493, 3611, 1495, 1497, 1494, 1498, 1508, 1510,  504, 3037,
 1512, 1418, 1515, 1236, 1516, 1517, 1518, 1836, 1527, 1408,
 1533, 1538, 1563, 1534, 1571, 1589,  605, 1127, 1409, 1608,
 2429, 1620,   15, 1555, 1626, 1631, 3587, 1634, 1649, 1236,
 2429, 1410, 1637, 1650, 1662, 1663, 1665,  605, 1667, 3502,
 1943, 1106, 1418,  605,  745, 1721, 1744, 3602, 1750, 1668,
 1943, 1751, 1325, 3610, 1768, 1769, 1089, 1776, 1127, 1787,
 1418, 1127, 1801, 3585, 1816, 1411, 1791, 1817, 1833, 1849,
 1851, 1555, 1412, 1852, 1555, 1853, 1127, 1089, 1127, 1338,
 1413, 1555, 1555, 1089, 1868, 3599, 1555, 1236, 3303, 1878,
 1891, 1555, 1893, 1371, 1555, 1894, 1916, 1414, 1919, 1415,
 1917, 1918,  419, 3645, 1424, 1423, 1425, 1236, 1555, 1416,
 1925, 1417, 1418, 1419, 1895, 1896, 1089, 1623, 1420, 1555,
 1236, 1955, 1957, 1949, 3239, 1950, 1959, 1555, 1960, 1974,
 1977, 1555, 1975,  869,  870, 1963, 1964,  873, 1976,  875,
 1555, 1981, 1984, 1986, 1989, 1992, 1908, 1993,  582, 1994,
 2007, 2010, 2019, 2020, 2021, 2012, 1555, 1408, 3644, 1421,
 2023,  279,  279,  582,  882,  883, 1409, 2028,  279, 2079,
 2046, 1422, 2052, 2044, 2088, 2036, 2037, 1555, 2097, 1410,
 1205,  419, 1555,  419, 2038, 1371, 2103, 2111, 2113, 3713,
 2121, 2122, 2114, 1555, 1555, 2126, 1555, 2127, 2125,  419,
  609, 2137, 2149, 3332, 2183, 3662, 2429, 2184, 2185, 1040,
 2102, 2130, 2186, 1411, 3741,  504,  504,  801, 2187,  609,
 1412, 3358,  504, 2190, 2193, 2194, 1943,  802, 1413, 2199,
  803, 1413, 2211, 2212, 3710, 3585, 2217, 2208, 2222, 3714,
 2226,  755, 2256, 2237, 2228, 1414, 1555, 1415, 2236,  279,
 2213, 2239, 3720, 2240, 2241, 1854, 2238, 1416, 2257, 1417,
 1418, 1419, 1555, 1555, 2258, 1418, 1420, 1555, 2269, 2264,
  279, 2270, 2271,  567,  609, 1127,  279, 2265, 2272, 2274,
 1855, 1127, 1555, 2442, 2290, 3795, 1856, 2292,  609, 2295,
 2296, 1857, 1127, 1127, 1127, 2297, 1858, 1127,  804, 1859,
 1127, 2302, 3693,  504, 2303, 2304, 2189, 1421, 2305,  279,
 2306, 2313, 1127, 1418, 2314, 1409, 2315, 2328, 1127, 2348,
 2433, 2322, 2341, 1836,  504, 2342,  605, 1127, 1410, 2434,
  504, 2345,  805, 1418, 2451, 2429, 3825,  806,  552, 3797,
 2456,  605, 2457,  807, 2470, 2469, 2429, 2493, 2473, 1422,
 2471, 3834, 3328, 2503, 2478, 1943, 2472, 3742, 2536, 2525,
 2546, 2545, 1411,  504, 2542, 2569, 1943, 1860, 3806, 1412,
 2578, 2635, 3877, 1418, 3599, 2570, 2611, 1413, 2614, 2625,
 1540, 1127, 2636, 1541, 2637, 2650, 2623, 2651, 1547, 1153,
 2673, 1551, 2677, 2658, 1414, 1554, 1415, 1861, 2684, 1547,
 2674, 2691, 1127, 2683, 2700, 2695, 1416, 2698, 1127, 1418,
 1419, 2699, 1862, 2712, 1555, 2709, 1127, 2713, 1127, 2685,
 2726,  808, 2710, 2719, 1418, 2730,  504, 1555, 2731, 1153,
 1153, 1127, 2732, 1153, 1863, 2355, 1127, 1836, 1418, 3887,
 1556, 1127, 1157, 2734, 2735, 1021, 2429,   86, 1153, 2740,
 1153, 2759, 2758, 2762, 2753, 2429, 1421, 2784, 2763, 1424,
 1423, 1425,  419, 3910, 3888, 2764, 1943,  809, 2777, 2782,
 2778, 2821, 2792, 1127, 2795, 1943, 3710,  634, 2813, 2814,
 1153, 2815, 1157, 1157, 3866, 2820, 1157, 2816, 2817, 2818,
 1864,  567, 2822, 1865, 2831, 2741, 1669, 1422, 2832, 2839,
 2850, 1157, 2878, 1157, 1418, 1418, 1555, 1680, 2851, 2880,
 2884,  810, 1686, 2890, 3953, 2891, 1691, 2892, 2896, 2102,
 2904, 2905,  567, 2914, 1418, 2930, 1712, 2931, 2960, 2961,
  811, 1718, 1719, 1157, 2962, 1722, 1723, 1724, 1725, 1726,
 1730, 1736, 2964, 1739, 1740, 2975, 1743, 2976, 1745, 1746,
 1747, 1748, 1749, 2967, 2977, 1752, 1753, 1754, 1755, 1756,
 1757, 1758, 1759, 1760, 1761, 1762, 1763, 1764, 1765, 1766,
 1767, 2978, 1127,  567, 1538, 3024, 3027, 3029, 3030, 2979,
 1784, 1785, 1669, 2980, 2981, 3032, 3021, 3957, 3041, 3035,
  426, 3974, 3043, 3033, 3044, 3034, 3046, 1040, 1799, 1866,
  427, 3051, 3066,  428, 2754, 1809, 3064, 3077, 3055,  567,
 1815, 3081, 3958, 3109, 3113, 1538, 1822, 1823, 1538, 3111,
 3095, 3056, 3121, 3494, 1967, 1832, 3124, 3123, 1040, 1838,
 1839, 1840, 1040, 1538, 3120, 1040, 3504, 3125, 3127, 3126,
 3128, 3130, 3132, 3133, 3137,  419, 1422, 1153, 1127, 3144,
 1147, 3167, 3182, 1153, 3212, 1967, 1418,   71, 3524, 3525,
 3213, 1408, 1153, 3215, 1153, 1153, 1153, 4013, 3211, 1153,
 1409,  429, 1153,  851, 3224, 3228, 3259, 3243, 3232, 3264,
 3277, 3275,   71, 1410, 1153, 3305, 3317, 2952, 3290, 3319,
 1153, 1147,   71, 1153, 1147, 3320, 3342,   71, 3321, 1153,
 1157,  430, 3322, 3336, 3323,  431, 1157, 3343, 3324, 1147,
  432, 1147, 3325, 3345, 3356, 1157,  433, 1411, 1157, 1157,
 3352, 3359, 1157, 3362, 1412, 1157,   71, 3364, 3365, 3375,
 2739, 2803, 1413, 1153, 3382, 3381, 3386, 1157, 1424, 1423,
 1425, 1147, 3399, 1157, 3396, 3397, 1157, 3410, 3411, 1414,
 3424, 1415, 1157, 1153, 3426, 1127, 3425, 3436, 3456, 3438,
 3459, 1416, 3473, 1417, 1418, 1419, 3427, 3479, 3477, 1021,
 1420, 3492, 3505, 3507, 1153, 1138, 3526, 3527, 3534, 3503,
 1153, 3513, 3531, 3536, 1153, 3533, 1157,   71, 3537, 3551,
 1153, 3555, 2687, 3560,  434, 3563,   71, 3564, 3565, 3545,
 3566, 3600, 3532, 3589, 3615, 3619, 1157, 3586, 1153, 3621,
 3641, 1421, 3668, 1153, 3669, 3691, 1138, 3692, 3696, 1138,
 3698,   71,  869,  870,  871,  872,  873, 1157,  875,  601,
  602, 1172, 3699, 1157, 1138, 3702, 1138, 1157, 3708, 3721,
  435, 1153, 3730, 1157, 3735, 1153, 3736, 3740,  877,  878,
 1538, 1538, 3716,  882,  883, 3717, 3745, 1538, 3753, 3754,
 3782, 1157, 3768, 3791, 3799, 3800, 1157, 3801, 2107, 2108,
 3807, 2110, 1172, 2112, 3810, 1172, 3826, 2116, 2117, 2118,
 2119, 3819, 3833, 3822, 3835, 2123, 2124, 3831, 3842, 3843,
 1172, 3855, 1172, 3858, 1157, 3864, 3865, 3873, 1157, 3867,
 3883, 3885, 3886,  436, 2131, 3889, 3890, 3896, 1147, 3898,
 2787, 3894, 3920, 3949, 1147, 3916, 3917, 3938, 3939, 3945,
 3955, 2147, 1172, 1147, 3956, 1147, 1147, 1147, 3960, 3978,
 1147, 2154, 2155, 1147, 3979, 3980, 3996, 3984, 1538, 3998,
 1408, 3986, 1538, 3999, 1153, 1147, 4003, 3981, 3982, 1409,
 4012, 1147, 4016, 4026, 1147, 4004, 4029, 4038, 4046, 1538,
 1147, 4023, 1410, 4042, 4052, 1538, 3761, 3762, 3763, 3764,
 3765, 3766, 3767, 1937, 4054, 4056, 1325,    1, 1452, 1327,
 1424, 1423, 1425, 3770, 3771, 3772, 3773, 3774, 3775, 3776,
  433, 1040, 1373,  418, 1147, 1538, 1411, 1157, 1538, 1060,
 2270, 1422, 2207, 1412, 2209, 1326, 1125, 2210, 1312, 2063,
  573, 1413, 1138,  523, 1147, 1423, 1395, 1736, 2043, 1031,
 1153,  555,  564,  579,  574,  969,  970, 1428, 1414, 1737,
 1415,  526,   72, 1138, 1400, 1147,   93,  554,  366, 1138,
 1416, 1147, 1417, 1418, 1419, 1147,  580, 1286, 1138, 1420,
  565, 1147, 1138, 1138, 1880, 1138, 1138,  620,  840, 1193,
  896, 1538,  168, 2104, 2007,  468, 1984, 1477,  641, 1147,
 1138, 1138,  607, 1157, 1147,  686, 1138,  650,  631, 1138,
  599, 1040,  465, 2036, 1679, 1138, 1881, 1843, 1982, 1172,
 1421, 1980,  642,  687, 1529, 1172,  535, 2008, 1875,  449,
 2042, 1983, 1147, 1574, 1172,  650, 1147, 1172, 1172,  731,
  732, 1172,  701, 1992, 1172, 1575, 3844,  400,  401, 1138,
 1993,   16,    0,   16, 2212,   28, 1172, 1153,   18,   31,
   33,   37, 1172,   40,   40, 1172,   40,   43, 1163, 1138,
   46, 1172,   55,   58,   76,   87, 1576, 1547, 1577,   82,
 1551,   79, 2330,   91,   85,   76,   33,   40,   40,  115,
  124,  112,   40,  121,  465, 2212,  133, 2337, 2212,   40,
 1138,  145,  136,  151,  502, 1172, 1138,  136,  148, 1163,
 1157,  160, 1163, 2212,  154,   33,   25, 2273, 2989, 3642,
 1897, 1409, 1898, 1899, 1138, 1172, 2985, 1163, 1900, 1163,
 1125, 1901, 2993, 3643, 1410, 1147, 1902, 3020, 2656, 3506,
 2894, 3442, 3609,  699, 3419,  502, 1172, 2100,  502, 3588,
 3393, 2849, 1172, 3151, 1274,  591, 1172, 1138, 3418, 1163,
 2781, 1138, 1172,  502, 3820, 1138,  629, 3722, 1411, 3876,
 2835, 1125, 3409, 3853, 1125, 1412, 1578,  703, 3823, 2889,
 1172, 3878, 3161, 1413, 1528, 1172, 3420, 4041, 3813, 1125,
 3951, 1125, 3162, 4044, 3987, 3985, 4009, 3383, 3760, 3384,
 1414, 3857, 1415, 1172, 2092, 2535, 1138, 2018, 2539, 1138,
 1422, 1147, 1416, 1172, 1417, 1418, 1419, 1172, 1138, 1138,
 3690, 1125, 1138, 3074, 1138, 1138, 2432, 3075, 4021, 3937,
 4022, 2750, 3071, 3804, 2556, 3862,   33, 1138, 2223, 3276,
 2562, 3428, 2862, 1619, 2354, 2565, 2786, 3874, 2566, 2567,
 2131, 2459, 3818, 2353, 1983, 2572, 1040, 2574, 3672, 2072,
 1138, 2041, 1421, 2846, 2575, 1519, 2576,  679, 3535, 2579,
 2580, 2581, 2582, 2676, 1846, 1923, 2587, 3530, 2588, 3338,
 2589, 3658, 2590, 1186,  695, 1166, 1200, 2324, 2591, 2592,
 2593, 2594, 2595, 2596, 2597, 2598, 2599, 2600, 2601, 2602,
 1565, 1536, 1879, 2603, 2604, 3313, 1138, 1882, 2605, 2606,
 2261, 2664, 3667, 3781, 2965, 3616, 1172, 3995, 1147, 2613,
 3039, 2327,   57, 1141, 2615, 1207, 1163, 1138, 2618, 3047,
 2212, 2212, 1163, 1138,   59, 1674, 1138, 2212, 3646, 2627,
 3647, 1163, 1040,   60, 1163, 1163,   63,   66, 1163, 2212,
 2704, 1163, 2752, 3964,  705, 1814, 2877, 2206, 1144, 1939,
 2885, 3824, 1256, 1163, 1141, 3711, 1138, 1141, 1293, 1163,
 1315, 2331, 1163, 1247, 1424, 1423, 1425, 1384, 1163, 1837,
  502,  502, 1141, 1773, 1141, 2077, 3852,  502, 1125, 1579,
 1580, 3752, 1172, 1965, 1125, 1385, 2586, 1966, 2694, 1144,
 3624, 1771, 1144, 1125, 3458, 3098, 1125, 1125,  652, 3500,
 1125, 2666, 1163, 1125, 1141, 1872, 3019, 1144, 2212, 1144,
 3493, 1256, 3222, 1664, 3544, 1125, 3994, 2527, 3718, 3607,
 3832, 1125, 1163, 2707, 1125, 3701, 3871, 3952, 1507, 2212,
 1125, 1138, 1138, 1138, 3099, 2212,  141, 3837, 1138, 1144,
 3333, 1587, 1422, 1163, 3315, 3626, 3931, 3719,    0, 1163,
 1138, 1581, 1582, 1163, 1583,    0,    0,    0,  502, 1163,
    0, 1584,    0,    0, 1125,    0,    0,    0, 2212,    0,
    0,    0,    0,    0,    0,    0,    0, 1163,    0,  502,
    0,    0, 1163,    0, 1125,  502,    0, 1138, 2742, 1172,
    0,    0,    0, 1250,    0,    0,    0, 2751,    0,    0,
    0,    0,    0,    0,    0, 1125, 1135,    0,    0,    0,
 1163, 1125,    0,    0, 1163, 1125,    0,    0,  502, 1138,
    0, 1125,    0,    0,    0,    0,    0,    0,    0, 1424,
 1423, 1425,    0,    0,    0,    0,    0,    0,    0, 1125,
 1138,    0,    0,    0, 1125,    0, 1138, 1135,    0,    0,
 1135,    0, 1250,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1135, 1408, 1135,    0,    0,
    0, 1141, 1125,    0, 1133, 1409, 1125, 1141,    0, 1138,
    0,  502,    0,    0,    0,    0, 1141,    0, 1410, 1141,
 1141,    0,    0, 1141, 1274,    0, 1141,    0,    0, 2212,
    0,    0,    0,    0,    0,    0, 1144,    0, 1141,    0,
    0,    0, 1144, 1163, 1141, 1133,    0, 1141, 1133,    0,
 1256, 1144, 1411, 1141, 1144, 1144,    0,    0, 1144, 1412,
    0, 1144,    0, 1133,    0, 1133,    0, 1413, 1138,    0,
    0,    0,    0, 1144,    0,    0,    0,    0,    0, 1144,
    0,    0, 1144, 1274, 1414,    0, 1415, 1141, 1144,    0,
    0,    0,    0,    0,    0,    0, 1416,    0, 1417, 1418,
 1419, 1040,    0,    0,    0, 1125,    0, 1141,    0, 1138,
 2131,    0, 1138,    0,    0,    0, 1256, 2921, 2922, 1163,
    0,    0, 1144,    0, 2929,    0,    0, 1138, 1141, 1138,
    0,    0,    0,    0, 1141,    0,    0,    0, 1141,    0,
 1138,    0, 1144,    0, 1141, 1256, 1421,    0,    0,    0,
    0,    0,    0,    0, 2955,    0,    0, 2957,    0, 2959,
    0, 1408, 1141, 1144,    0,    0, 2158, 1141,    0, 1144,
 1409, 1256,    0, 1144, 1256, 1555,    0,    0,    0, 1144,
    0, 1125,    0, 1410,    0,    0,    0,    0, 1556, 1256,
    0, 1250,    0, 1021, 1135, 1141,    0, 1144,    0, 1141,
 1135,    0, 1144,    0,    0,    0,    0, 2158,    0,    0,
 2158, 1135, 1135, 1135,    0, 1256, 1135, 1411,    0, 1135,
    0,    0,    0,    0, 1412, 2158, 1163,    0,    0,    0,
 1144, 1135, 1413,    0, 1144,    0,    0, 1135,    0,    0,
    0,    0,    0,    0,    0,    0, 1135,    0,    0, 1414,
    0, 1415,    0,    0,    0,    0,    0, 1250,    0,    0,
 2337, 1416, 1133, 1417, 1418, 1419,    0,    0, 1133,    0,
 1420,    0,    0, 1256,    0,    0,    0,    0, 3061, 1133,
 1133, 1133,    0,    0, 1133,    0, 1250, 1133, 1125,    0,
    0,    0, 1274, 1256,    0,    0,    0,    0, 1141, 1133,
 1135,    0,    0, 3079,    0, 1133, 1256,    0,    0,    0,
    0, 1421, 1250,    0, 1133, 1250,    0,    0,    0,    0,
    0, 1135, 2145,    0,    0,    0, 1138, 1135,    0,    0,
 1250,    0, 1138, 1144,    0, 1135, 1422, 1135,    0,    0,
    0, 1424, 1423, 1425, 1138, 1138,    0,    0, 1138,    0,
    0, 1138,    0,    0,    0, 1135, 1250,    0, 1274,    0,
 1135,    0,    0, 1138,    0,    0,    0,    0, 1133, 1138,
    0,    0,    0,    0, 1141,    0,    0,    0, 1138,    0,
 2122, 2122, 2059, 2122, 2122, 2122,    0, 1274,    0, 1133,
    0,    0, 1135,    0,    0, 1133,    0,    0, 2122, 2122,
 2122,    0, 2135,    0,    0, 1133,    0,    0,    0, 1144,
    0,    0,    0, 1274, 1250,    0, 1274,    0, 1424, 1423,
 1425,  851,    0, 1133,    0,    0,    0,    0, 1133, 2122,
    0, 1274, 1138,    0, 1250,    0,    0,    0,    0,    0,
    0,    0, 2158, 2158,    0,    0,    0, 1250,    0, 2158,
 3242,    0, 3244, 1138, 3245,    0,    0, 1274, 3250, 1138,
 1133, 2158,    0, 3253, 3254, 3255, 3256, 3257,    0, 1138,
    0,    0, 3258,    0, 3261,    0, 3262,    0,    0,    0,
    0, 1141,    0,    0,    0,    0, 1258, 1138,    0,    0,
    0, 1135, 1138,    0, 3266,    0,    0,    0, 3267,    0,
    0, 1422, 1040,  480,    0, 3272,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1274, 1144,    0,    0,    0,
    0,    0,    0,    0, 1138,    0,    0,    0,    0,    0,
 2158,    0,    0,    0,    0, 1274, 2123, 2123, 2060, 2123,
 2123, 2123,    0, 1408,  480, 1258,    0,  480, 1274,    0,
    0, 2158, 1409,    0, 2123, 2123, 2123, 2158,    0, 1133,
    0,    0,  480,    0,    0, 1410,    0, 1135,    0,    0,
  869,  870,  871,  872,  873,    0,  875,  601,  602,    0,
    0,    0,    0, 3350,    0, 2123,    0,    0,    0,    0,
 2158,    0,    0,    0,    0, 2144,  877,  878, 3631, 1411,
 2122,  882,  883,    0, 2122,    0, 1412,    0,    0, 2122,
    0,    0,    0,    0, 1413, 2122,    0,    0, 1555, 2122,
 1408,    0, 2122, 1138,    0,    0, 3168,    0,    0, 1409,
    0, 1414,    0, 1415, 1968, 1133,    0,    0,    0,    0,
    0,    0, 1410, 1416,    0, 1417, 1418, 1419,    0,    0,
    0, 3169, 1420,    0,    0,    0, 2122, 3170,    0,    0,
    0, 3400,    0, 2122,    0,    0,    0, 3171,    0,    0,
    0, 2122,    0,    0, 1135,    0, 1411,    0,    0,    0,
 3172,    0,    0, 1412,    0,    0,    0,    0, 2122,    0,
 2122, 1413,    0, 1421,    0,    0,    0,    0,    0, 1138,
 2122,    0, 2122, 2122, 2122,    0,    0,    0, 1414, 2122,
 1415,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1416, 2158, 1417, 1418, 1419,    0,    0,    0,    0, 1420,
    0,    0,    0,    0,    0,    0,    0,    0, 3173,    0,
    0,    0, 1133,    0,    0,    0,    0,    0,  482,    0,
 2122,    0,    0,    0, 1258,    0, 2123,    0,    0,    0,
 2123,    0,    0,    0,    0, 2123,    0,    0, 3174,    0,
 1421, 2123, 3175,  969,  970, 2123,    0, 3176, 2123,  481,
  480,    0,    0, 3177, 1969,    0,  481, 2131, 3245,  482,
    0,    0,  482,    0,    0,    0, 2122, 2122, 2122, 2122,
 2122, 2122, 2122, 2122, 3472, 3178, 1138,  482,    0,    0,
    0,    0, 2123,    0,    0, 3480,    0,    0,    0, 2123,
 1258,    0,    0,    0,    0,    0,    0, 2123,    0,    0,
    0,    0,    0, 2168, 2168, 2055, 2168, 2168, 2168,    0,
    0,    0,    0,    0, 2123,    0, 2123,    0,    0, 1258,
    0, 2168, 2168, 2168,    0,    0, 2123,    0, 2123, 2123,
 2123, 3179,    0,    0,  851, 2123,    0,  481,    0,    0,
    0,    0,    0,    0,    0, 1258,    0,    0, 1258,    0,
    0,    0, 2168, 1422,    0,    0,    0,    0,  481,    0,
  854,    0,    0, 1258,  481,    0,    0,    0,    0,    0,
    0, 2161, 1040, 3550, 2160, 2122, 2123,    0,    0,    0,
    0, 3553, 3554,    0,    0,    0,    0,  856,    0, 1258,
 1424, 1423, 1425,    0,    0,    0,    0,  481,    0,    0,
  453, 2122,    0,    0,    0,    0,    0,    0,    0,    0,
 2122,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 3180, 2123, 2123, 2123, 2123, 2123, 2123, 2123, 2123,
 1422,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 3181,  453,    0,    0,  453,    0, 3608, 1258,    0,    0,
    0,    0,    0,  866,    0,    0,    0,    0,    0,  453,
    0,    0,    0,    0,    0,    0, 2131, 1258,    0,  868,
    0,    0,    0,    0, 3622,    0, 3623,    0,    0,    0,
 1258,    0,    0,  869,  870,  871,  872,  873,    0,  875,
  601,  602,    0,    0,  484,  482,    0,    0,    0,    0,
    0,  484,    0,    0,    0,    0,    0,    0,    0,  877,
  878,  879,  880,  881,  882,  883,  884,  885,    0,    0,
    0,    0,    0, 2168,  887,  888,    0, 2168,    0,    0,
    0, 2123, 2168,  892,    0,    0,    0,    0, 2168,    0,
    0,    0, 2168,    0,    0, 2168, 2337, 3350,    0, 2122,
 2122,    0,    0,    0,    0,    0,    0, 2123, 3676,    0,
 2136,    0,    0,    0,    0,    0, 2123,    0, 2074, 2074,
 2074, 2074, 2074,    0, 2074,    0, 1424, 1423, 1425, 2168,
    0,    0,  484,    0,    0,    0, 2168, 2074, 2074, 2074,
    0,    0, 1408,    0, 2168,    0,    0,    0,    0,    0,
    0, 1409,    0,  484,    0,    0,    0,    0,    0,  484,
    0, 2168,    0, 2168, 1410,    0,    0,    0, 2074,    0,
    0,    0,    0, 2168,    0, 2168, 2168, 2168,    0,    0,
    0, 2131, 2168,    0,    0, 3749, 3749,   67,    0,    0,
    0, 3755,  484,    0,    0,    0,    0,    0, 1411,    0,
   68,    0,   69,    0,    0, 1412,    0,    0,    0,    0,
    0,    0,   70, 1413,    0,    0,    0,    0,   71,    0,
    0,    0,   72, 2168,    0,    0,  453,  453,   73,    0,
 1414,   74, 1415,  453,  968,    0,  969,  970,  971,    0,
    0,   75, 1416,    0, 1417, 1418, 1419,    0,    0,    0,
 2086, 1420,    0,    0,    0, 2123, 2123,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 2168,
 2168, 2168, 2168, 2168, 2168, 2168, 2168,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2086, 2086, 1421, 2086, 2086, 2086, 2095,  998,  999,   76,
    0,    0,    0,    0,    0,    0, 3840,    0, 2086, 2086,
    0,    0,    0, 2086,  453,    0, 1006,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1408,   77,
    0,    0,    0,   78,    0,  453,    0, 1409,   79, 2074,
    0,  453,    0, 2074,   80,    0,    0,    0, 2074,    0,
 1410,    0,    0,    0, 2074,    0,    0,    0, 2074,    0,
    0, 2074,    0,    0,    0,    0,   81,    0,    0,    0,
    0,    0,    0,    0,  453,    0,    0,    0, 2168,    0,
    0,    0,    0,    0, 1411,    0,    0,    0,    0,    0,
    0, 1412,    0,    0,    0, 2074,    0,    0,    0, 1413,
    0,    0, 2074,    0, 2168,    0,    0,    0, 3891,    0,
 2074,    0,    0, 2168,    0,    0, 1414,    0, 1415,    0,
    0,    0,   82,    0,    0,    0,    0, 2074, 1416, 2074,
 1417, 1418, 1419,    0,    0,    0,    0, 1420,    0, 2074,
    0, 2074, 2074, 2074,    0,    0,    0,    0, 2074, 3923,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 3940,    0, 3944,    0,   83,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1421,    0,
    0,    0, 1422,    0,    0,    0,    0,    0,    0, 2074,
    0,    0,    0,    0,    0,    0, 3972,    0, 3975, 3977,
    0,    0,    0,    0,    0,    0,    0, 2086,    0, 2086,
    0,    0,   84,    0, 2086, 2086,    0, 2086, 2086, 2086,
    0,    0,    0,    0, 2086,    0,    0,    0,    0, 2086,
 2086,   85, 2086,    0, 2086, 2086,    0, 2086,    0,    0,
 2086, 2086, 2086,    0,    0,    0,    0, 2086,    0, 2086,
 2086,    0, 2168, 2168, 2086,    0, 2086,    0, 2086, 2086,
    0, 3972, 2086, 4024, 4025, 2086,    0, 2086,    0, 2086,
 2086,    0, 2086, 2086, 2086,    0, 2086, 2086, 2086,    0,
    0,    0, 2086, 4040, 2086, 2086, 2086,    0,    0,    0,
 2086,    0,    0,    0, 2086,    0, 2086, 2086, 2086, 2086,
    0, 2086,    0, 2086, 2086,    0, 2086, 2086, 2086,    0,
 2086,    0, 2337,    0, 2086,    0, 2337,    0, 2086,    0,
 2086,    0, 2086,    0, 2086, 2086, 2086, 2086, 2086,    0,
 2086, 2086, 2086,    0,    0,    0, 2086,    0, 2086, 2086,
    0,    0, 2086, 2086, 2086,    0,    0, 2086,   86, 2086,
    0,    0, 2086, 2086, 2086, 2086,    0, 2086,    0, 2086,
 2086, 2086, 2086, 2086, 2086, 2086, 2086,    0, 1422,    0,
 2086,    0, 2086, 2086,    0,    0, 2086,    0,    0, 2074,
    0,    0,    0, 2086,    0, 2086, 2092, 2086, 2086, 2086,
 2086, 2086, 2086, 2086, 2086, 2086, 2086, 2086, 2086, 2086,
    0, 2086, 2086, 2086, 2086,    0, 2086, 2086,    0,    0,
    0,    0,    0,    0, 2086, 2086, 2086, 2086, 2086, 2086,
 2086, 2086, 2086, 2086, 2086, 2086, 2086,    0, 2086, 2086,
 2086, 2086, 2086,    0, 2086, 2086, 2086, 2086, 2086, 2086,
 2086, 2086,    0,    0, 2086, 2086, 2086, 2086, 2086, 2086,
 2086,    0,    0, 2086, 2086, 2086, 2086, 2086, 2086, 2086,
 2086, 2086,    0, 2086,    0, 2086, 2086, 2086,    0, 2086,
 2086,    0, 2086, 2086, 2086, 2086, 2086, 2086, 2086, 2086,
 2086, 2086, 2086, 2086, 2086, 2086, 2086, 2086, 2086, 2086,
 2086, 2086, 2086, 2086,    0, 2086, 2086, 2086, 2086, 2086,
 2086,    0, 2086, 2086, 2086, 2086,    0, 2086,    0, 2086,
 2086, 2086, 2086, 2086, 2086, 2086, 2086, 2086, 2086, 2086,
 2086, 2086, 2086, 2086, 2086, 2086, 2086,    0, 2086, 2086,
 2086, 2086,    0, 2086, 2086, 2086, 2086, 2086,    0, 2086,
 2086, 2086, 2086, 2086, 2086, 2086, 2086, 2086, 2086, 2086,
 2086, 2086,    0, 2086, 2086, 2086, 2086, 2086, 2086,    0,
 2086, 2086, 2086, 2086, 2086, 2086, 2086, 2086,    0, 2086,
 2086, 2086, 2086,    0, 2086, 2086, 2086, 2086, 2086, 2086,
    0,    0, 2086, 2086, 2086, 2086, 2086, 2086, 2086, 2086,
 2086, 2086,    0, 2086, 2086, 2086, 2086, 2086, 2086, 2086,
 2086, 2086, 2086, 2086,    0, 2086, 2086, 2086, 2086, 2086,
 2086, 2086, 2086, 2086, 2086, 2086, 2086, 2086, 2086, 2086,
 2086, 2086, 2086, 2086,    0, 2086, 2086, 2086, 2086, 2086,
 2086, 2086, 2086, 2086,    0, 2086, 2086, 2086, 2086, 2086,
 2086, 2086, 2086, 2086, 2086, 2086, 2086,    0, 2086, 2086,
 2086,    0, 2086,    0, 2086, 2086, 2086, 2086, 2086, 2086,
 2086, 2086, 2086, 2086, 2086, 2086, 2086, 2086, 2086, 1731,
 2086, 2086, 2086, 2086, 2086, 2086, 2086, 2086, 2086, 2086,
 2086, 2086, 2086, 2086, 2086, 2086, 2086, 2086, 2086, 2086,
 2086,    0,    0, 2086, 2086,    0,    0, 2086,    0, 2086,
 2086, 2086, 2086, 2086, 2086,    0,    0,    0,    0,  712,
 1731, 1731, 1731, 1731, 1731,    0, 1731,    0,    0,    0,
    0,    0,    0,   85,  713,    0,    0,    0, 1731, 1731,
 1731, 1731,    0,    0,  714,    0,    0,    0,   85,    0,
  715,   86,    0,    0,  716,  717,    0,    0,   85,    0,
  718,    0,    0,  719,   85,    0,   86,    0,    0,   85,
    0,    0, 1731,  720,   85,    0,   86,   85,    0,    0,
    0,    0,   86,    0,    0,    0,    0,   86,    0,    0,
    0,    0,   86,    0,    0,   86,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  721,    0,    0,   69,    0,    0,    0,
    0,  722,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,   70,   85,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  723,    0,   86,    0,  724,    0,    0,    0,    0,
  725,    0,    0,    0,    0,   85,  726,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
   85,    0,    0,   86,    0,    0,    0,    0,  727,  728,
    0,    0,    0,    0,    0,    0,    0,    0,   86,    0,
    0,    0,   85,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
   86,    0,    0,    0,    0,    0, 1731, 1731, 1731,    0,
    0, 1731,    0, 1731, 1731, 1731,    0, 1731, 1731,    0,
 1731,    0,    0,    0,  729, 1731, 1731,  730, 1731, 1731,
 1731, 1731, 1731, 1731,    0,    0, 1731,    0,   85, 1731,
    0,   85,    0,    0,    0,    0, 1731,    0, 1731, 1731,
 1731, 1731, 1731, 1731, 1731, 1731,   86,    0, 1731,   86,
    0, 1731,    0, 1731, 1731,    0,    0, 1731, 1731, 1731,
    0, 1731, 1731, 1731, 1731, 1731, 1731, 1731,    0,    0,
    0, 1731, 1731, 1731, 1731, 1731,    0,    0,    0, 1731,
    0,    0,    0,    0,    0,    0,    0, 1731, 1731, 1731,
 1731, 1731, 1731,    0,    0,    0, 1731, 1731,    0, 1731,
    0, 1731,    0, 1731, 1731, 1731,    0, 1731,    0, 1731,
 1731, 1731,    0, 1731, 1731, 1731, 1731, 1731,    0, 1731,
 1731,    0,    0,  731, 1731, 1731,    0, 1731, 1731,    0,
 1731, 1731, 1731, 1731,    0,    0, 1731,   85, 1731,    0,
 1731, 1731, 1731, 1731, 1731,    0, 1731,    0, 1731,    0,
 1731, 1731, 1731, 1731, 1731,   86,    0, 1731, 1731, 1731,
    0, 1731, 1731, 1731,    0,    0,    0,    0,    0,    0,
    0,    0, 1731, 1731, 1731,    0, 1731, 1731, 1731,    0,
    0,    0,    0,    0, 1731,    0,    0,    0, 1731,    0,
 1731, 1731, 1731, 1731,    0, 1731,    0, 1731, 1731, 1731,
 1731, 1731, 1731, 1731, 1731,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1731, 1731,    0, 1731, 1731, 1731,
    0,    0,    0,    0, 1731, 1731, 1731, 1731, 1731,    0,
 1731,    0,    0, 1731, 1731,    0,    0,    0, 1731, 1731,
    0,    0, 1731, 1731, 1731, 1731, 1731, 1731, 1731, 1731,
 1731,    0, 1731,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1731, 1731, 1731, 1731, 1731, 1731, 1731, 1731,
 1731, 1731, 1731, 1731, 1731, 1731, 1731, 1731, 1731, 1731,
 1731, 1731, 1731,    0, 1731, 1731, 1731, 1731, 1731, 1731,
    0, 1731, 1731, 1731, 1731,    0, 1731,    0, 1731, 1731,
 1731, 1731, 1731, 1731, 1731, 1731, 1731, 1731, 1731, 1731,
 1731, 1731, 1731, 1731, 1731, 1731,    0, 1731, 1731, 1731,
 1731,    0, 1731, 1731, 1731, 1731, 1731,    0, 1731, 1731,
 1731, 1731, 1731, 1731, 1731, 1731, 1731, 1731, 1731, 1731,
 1731, 1731, 1731, 1731, 1731, 1731, 1731, 1731,    0, 1731,
 1731, 1731, 1731, 1731, 1731, 1731, 1731,    0, 1731, 1731,
    0,    0,    0, 1731, 1731, 1731, 1731, 1731, 1731,    0,
    0, 1731, 1731, 1731, 1731, 1731, 1731, 1731, 1731, 1731,
 2163,    0,    0, 2162,    0, 1731,    0, 1731, 1731, 1731,
 1731, 1731, 1731,    0, 1731,    0,    0,    0,    0, 1424,
 1423, 1425,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1731, 1731,    0, 1731, 1731, 1731, 1731, 1731, 1731,
 1731, 1731, 1731,    0, 1731, 1731, 1731, 1731, 1731, 1731,
 1731, 1731, 1731, 1731, 1731, 1731,    0, 1731, 1731, 1731,
    0, 1731,    0,    0,    0, 1731, 1731, 1731, 1731, 1731,
 1731, 1731, 1731, 1731, 1731, 1731, 1731, 1731, 1733, 1731,
 1731,    0,    0,    0,    0,    0, 1731, 1731, 1731, 1731,
 1731, 1731, 1731, 1731, 1731, 1731, 1731, 1731, 1731, 1731,
 1731, 1731, 1731, 1731,    0,    0, 1731,    0, 1731, 1731,
 1731, 1731, 1731, 1731,    0,    0,    0,    0,    0, 1733,
 1733, 1733, 1733, 1733,    0, 1733,    0,    0,    0,   83,
    0,    0,    0,    0,    0,    0,    0, 1733, 1733, 1733,
 1733,    0,    0,    0,   92,    0,    0,    0,    0,    0,
    0,    0,    0,    0,   83,    0,    0,    0,    0,    0,
   83,    0,    0,    0,    0,   83,    0,    0,    0,    0,
   83, 1733,    0,   83,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,   89,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,   96,
    0,    0,    0,    0,    0,    0,    0,    0,    0,   89,
    0,    0,    0,    0,    0,   89,    0,    0,    0,    0,
   89, 1408,    0,    0,    0,   89,    0,    0,   89,    0,
 1409,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,   83,    0, 1410,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,   83,    0,    0,    0,    0,    0, 1411,    0,    0,
    0,    0,    0,    0, 1412,    0,   83,    0,    0,    0,
    0,    0, 1413,    0,    0,    0,   89,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,   83, 1414,
    0, 1415,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1416,    0, 1417, 1418, 1419,   89,    0,    0,    0,
 1420,    0,    0,    0,    0, 1733, 1733, 1733,    0,    0,
 1733,   89, 1733, 1733, 1733,    0, 1733, 1733,    0, 1733,
    0,    0,    0,    0, 1733, 1733,    0, 1733, 1733, 1733,
 1733, 1733, 1733,   89,   83, 1733,    0,   83, 1733,    0,
    0, 1421,    0,    0,    0, 1733,    0, 1733, 1733, 1733,
 1733, 1733, 1733, 1733, 1733,    0,    0, 1733,    0,    0,
 1733,    0, 1733, 1733,    0,    0, 1733, 1733, 1733,    0,
 1733, 1733, 1733, 1733, 1733, 1733, 1733,    0,    0,    0,
 1733, 1733, 1733, 1733, 1733,    0,    0,    0, 1733,   89,
    0,    0,   89,    0,    0,    0, 1733, 1733, 1733, 1733,
 1733, 1733,    0,    0,    0, 1733, 1733,    0, 1733,    0,
 1733,    0, 1733, 1733, 1733,    0, 1733,    0, 1733, 1733,
 1733,    0, 1733, 1733, 1733, 1733, 1733,    0, 1733, 1733,
    0,    0,    0, 1733, 1733,    0, 1733, 1733,    0, 1733,
 1733, 1733, 1733,   83,    0, 1733,    0, 1733,    0, 1733,
 1733, 1733, 1733, 1733,    0, 1733,    0, 1733,    0, 1733,
 1733, 1733, 1733, 1733,    0,    0, 1733, 1733, 1733,    0,
 1733, 1733, 1733,    0,    0,    0,    0,    0,    0,    0,
    0, 1733, 1733, 1733,    0, 1733, 1733, 1733,    0,    0,
    0,    0,    0, 1733,    0,    0,    0, 1733,   89, 1733,
 1733, 1733, 1733,    0, 1733,    0, 1733, 1733, 1733, 1733,
 1733, 1733, 1733, 1733,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1733, 1733,    0, 1733, 1733, 1733,    0,
    0, 1422,    0, 1733, 1733, 1733, 1733, 1733,    0, 1733,
    0,    0, 1733, 1733,    0,    0,    0, 1733, 1733,    0,
    0, 1733, 1733, 1733, 1733, 1733, 1733, 1733, 1733, 1733,
    0, 1733,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1733, 1733, 1733, 1733, 1733, 1733, 1733, 1733, 1733,
 1733, 1733, 1733, 1733, 1733, 1733, 1733, 1733, 1733, 1733,
 1733, 1733,    0, 1733, 1733, 1733, 1733, 1733, 1733,    0,
 1733, 1733, 1733, 1733,    0, 1733,    0, 1733, 1733, 1733,
 1733, 1733, 1733, 1733, 1733, 1733, 1733, 1733, 1733, 1733,
 1733, 1733, 1733, 1733, 1733,    0, 1733, 1733, 1733, 1733,
    0, 1733, 1733, 1733, 1733, 1733,    0, 1733, 1733, 1733,
 1733, 1733, 1733, 1733, 1733, 1733, 1733, 1733, 1733, 1733,
 1733, 1733, 1733, 1733, 1733, 1733, 1733,    0, 1733, 1733,
 1733, 1733, 1733, 1733, 1733, 1733,    0, 1733, 1733,    0,
    0,    0, 1733, 1733, 1733, 1733, 1733, 1733,    0,    0,
 1733, 1733, 1733, 1733, 1733, 1733, 1733, 1733, 1733, 2935,
    0,    0, 2934,    0, 1733,    0, 1733, 1733, 1733, 1733,
 1733, 1733,    0, 1733,    0,    0,    0,    0, 1424, 1423,
 1425,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1733, 1733,    0, 1733, 1733, 1733, 1733, 1733, 1733, 1733,
 1733, 1733,    0, 1733, 1733, 1733, 1733, 1733, 1733, 1733,
 1733, 1733, 1733, 1733, 1733,    0, 1733, 1733, 1733,    0,
 1733,    0,    0,    0, 1733, 1733, 1733, 1733, 1733, 1733,
 1733, 1733, 1733, 1733, 1733, 1733, 1733, 2097, 1733, 1733,
    0,    0,    0,    0,    0, 1733, 1733, 1733, 1733, 1733,
 1733, 1733, 1733, 1733, 1733, 1733, 1733, 1733, 1733, 1733,
 1733, 1733, 1733,    0,    0, 1733,    0, 1733, 1733, 1733,
 1733, 1733, 1733,    0,    0,   84,    0, 2067, 2097, 2097,
 2097, 2097, 2097, 2097, 2097,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 2097, 2097, 2097, 2097, 2097,
   84,    0,    0,    0,    0,    0,   84,    0,    0,    0,
    0,   84,    0,    0,    0,    0,   84,    0,    0,   84,
    0,    0,    1,    0,    0,    0,    0,    0, 2097,    0,
 2097,    0,    0,    0,    0,    0,    0,    0,    0,    2,
    0,    0,    0,    0,    0,    0,    0,    3,    0,    0,
    0,    0,    0,    0,    0,    4,    0,    5,    0,    0,
    0,    0,    6,    0,    0,    0,    0,    0,    7,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1408,    0,    8,    0,    0,    0,    0,   84,    0, 1409,
    0,    9,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1410,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,   10,   84,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,   84,    0,    0,    0, 1411,    0,    0,    0,
    0,    0,    0, 1412,    0,    0,    0,    0,    0,    0,
    0, 1413,    0,    0,   84,    0,    0,   11,    0,   12,
    0,    0,   13,    0,    0,    0,    0,    0, 1414,    0,
 1415,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1416,    0, 1417, 1418, 1419,    0,   14,    0,    0, 1420,
    0,    0,    0,    0, 2097,    0, 2097,    0,   15, 2097,
    0, 2097, 2097, 2097,    0, 2097, 2097,    0, 2097,    0,
   84,    0,    0,   84, 2097,    0, 2097, 2097, 2097, 2097,
    0, 2097,    0,    0, 2097,    0,    0, 2097,    0,    0,
 1421,    0,    0,    0, 2097,    0, 2097, 2097,   16, 2097,
 2097, 2097, 2097, 2097,    0,    0, 2097,    0,    0, 2097,
    0, 2097, 2097,   17,   18, 2097, 2097, 2097,    0, 2097,
 2097, 2097, 2097, 2097, 2097, 2097,    0,    0,    0, 2097,
 2097, 2097, 2097, 2097,    0,   19,    0, 2097,    0,    0,
    0,    0,    0,    0,    0, 2097, 2097, 2097, 2097, 2097,
 2097,    0,    0,    0, 2097, 2097,    0, 2097,    0, 2097,
    0, 2097, 2097, 2097,    0, 2097,    0, 2097, 2097, 2097,
    0, 2097, 2097, 2097, 2097, 2097,    0, 2097, 2097,   84,
    0,    0,    0, 2097,    0, 2097, 2097,    0,    0, 2097,
 2097, 2097,    0,    0, 2097,    0, 2097,    0,    0, 2097,
 2097, 2097, 2097,    0, 2097,    0, 2097,    0, 2097, 2097,
 2097, 2097, 2097,    0,    0, 2097, 2097, 2097,    0, 2097,
 2097,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2097, 2097, 2097,    0, 2097, 2097, 2097,    0,    0,    0,
    0,    0, 2097,    0,    0,    0, 2097,    0, 2097, 2097,
 2097, 2097,    0, 2097,    0, 2097, 2097, 2097, 2097, 2097,
 2097, 2097, 2097,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 2097, 2097,    0, 2097, 2097, 2097,    0,    0,
 1422,    0, 2097, 2097, 2097, 2097, 2097,    0, 2097,    0,
    0, 2097, 2097,    0,    0,    0, 2097, 2097,    0,    0,
 2097, 2097, 2097, 2097, 2097, 2097, 2097, 2097, 2097,    0,
 2097,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2097, 2097, 2097, 2097, 2097, 2097, 2097, 2097, 2097, 2097,
 2097, 2097, 2097, 2097, 2097, 2097, 2097, 2097, 2097, 2097,
 2097,    0, 2097, 2097, 2097, 2097, 2097, 2097,    0, 2097,
 2097, 2097, 2097,    0, 2097,    0, 2097, 2097, 2097, 2097,
 2097, 2097, 2097, 2097, 2097, 2097, 2097, 2097, 2097, 2097,
 2097, 2097, 2097, 2097,    0, 2097, 2097, 2097, 2097,    0,
 2097, 2097, 2097, 2097, 2097,    0, 2097, 2097, 2097, 2097,
 2097, 2097, 2097, 2097, 2097, 2097, 2097, 2097, 2097, 2097,
 2097, 2097, 2097, 2097, 2097, 2097, 2097, 2097, 2097, 2097,
 2097, 2097, 2097, 2097, 2097,    0, 2097, 2097,    0,    0,
    0, 2097, 2097, 2097, 2097, 2097, 2097,    0,    0, 2097,
 2097, 2097, 2097, 2097, 2097, 2097, 2097, 2097, 2937,    0,
    0, 2936,    0, 2097,    0, 2097, 2097, 2097, 2097, 2097,
 2097,    0, 2097,    0,    0,    0,    0, 1424, 1423, 1425,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 2097,
 2097,    0, 2097, 2097, 2097, 2097, 2097, 2097, 2097, 2097,
 2097,    0, 2097, 2097, 2097, 2097, 2097, 2097, 2097, 2097,
 2097, 2097, 2097, 2097,    0, 2097, 2097, 2097,    0, 2097,
    0,    0,    0, 2097, 2097, 2097, 2097, 2097, 2097, 2097,
 2097, 2097, 2097, 2097, 2097, 2097, 2134, 2097, 2097,    0,
    0,    0,    0,    0, 2097, 2097, 2097, 2097, 2097, 2097,
 2097, 2097, 2097, 2097, 2097, 2097, 2097, 2097, 2097, 2097,
 2097, 2097,    0,    0, 2097,    0, 2097, 2097, 2097, 2097,
 2097, 2097,    0,    0,    0,    0, 1894, 2134, 2134, 2134,
 2134, 2134, 2134, 2134,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 2134, 2134, 2134, 2134, 2134, 3183,
    0,    0,    0,    0,    0, 3184,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 3185,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 2134, 3186, 2134,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1408,
    0,    0,    0,    0,    0,    0, 3187,    0, 1409,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1410,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 3188,    0,    0,    0,
 3189,    0,    0,    0,    0, 3190,    0,    0,    0,    0,
    0, 3191,    0,    0,    0, 1411,    0,    0,    0,    0,
    0,    0, 1412,    0,    0,    0,    0,    0,    0,    0,
 1413,    0,    0, 3192,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1414,    0, 1415,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1416,
    0, 1417, 1418, 1419,    0,    0,    0,    0, 1420,    0,
    0,    0,    0, 2134,    0, 2134,    0,    0, 2134,    0,
 2134, 2134, 2134,    0, 2134, 2134,    0, 2134,    0, 3193,
    0,    0, 3194, 2134,    0, 2134, 2134, 2134, 2134,    0,
 2134,    0,    0, 2134,    0,    0, 2134,    0,    0, 1421,
    0,    0,    0, 2134,    0, 2134, 2134,    0, 2134, 2134,
 2134, 2134, 2134,    0,    0, 2134,    0,    0, 2134,    0,
 2134, 2134,    0,    0, 2134, 2134, 2134,    0, 2134, 2134,
 2134, 2134, 2134, 2134, 2134,    0,    0,    0, 2134, 2134,
 2134, 2134, 2134,    0,    0,    0, 2134,    0,    0,    0,
    0,    0,    0,    0, 2134, 2134, 2134, 2134, 2134, 2134,
    0,    0,    0, 2134, 2134,    0, 2134,    0, 2134, 3195,
 2134, 2134, 2134,    0, 2134,    0, 2134, 2134, 2134,    0,
 2134, 2134, 2134, 2134, 2134,    0, 2134, 2134, 3196,    0,
    0,    0, 2134,    0, 2134, 2134,    0,    0, 2134, 2134,
 2134,    0,    0, 2134,    0, 2134,    0,    0, 2134, 2134,
 2134, 2134,    0, 2134,    0, 2134,    0, 2134, 2134, 2134,
 2134, 2134,    0,    0, 2134, 2134, 2134,    0, 2134, 2134,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 2134,
 2134, 2134,    0, 2134, 2134, 2134,    0,    0,    0,    0,
    0, 2134,    0,    0,    0, 2134,    0, 2134, 2134, 2134,
 2134,    0, 2134,    0, 2134, 2134, 2134, 2134, 2134, 2134,
 2134, 2134,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 2134, 2134,    0, 2134, 2134, 2134,    0,    0, 1422,
    0, 2134, 2134, 2134, 2134, 2134,    0, 2134,    0,    0,
 2134, 2134,    0,    0,    0, 2134, 2134,    0,    0, 2134,
 2134, 2134, 2134, 2134, 2134, 2134, 2134, 2134,    0, 2134,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 2134,
 2134, 2134, 2134, 2134, 2134, 2134, 2134, 2134, 2134, 2134,
 2134, 2134, 2134, 2134, 2134, 2134, 2134, 2134, 2134, 2134,
    0, 2134, 2134, 2134, 2134, 2134, 2134,    0, 2134, 2134,
 2134, 2134,    0, 2134,    0, 2134, 2134, 2134, 2134, 2134,
 2134, 2134, 2134, 2134, 2134, 2134, 2134, 2134, 2134, 2134,
 2134, 2134, 2134,    0, 2134, 2134, 2134, 2134,    0, 2134,
 2134, 2134, 2134, 2134,    0, 2134, 2134, 2134, 2134, 2134,
 2134, 2134, 2134, 2134, 2134, 2134, 2134, 2134, 2134, 2134,
 2134, 2134, 2134, 2134, 2134, 2134, 2134, 2134, 2134, 2134,
 2134, 2134, 2134, 2134,    0, 2134, 2134,    0,    0,    0,
 2134, 2134, 2134, 2134, 2134, 2134,    0,    0, 2134, 2134,
 2134, 2134, 2134, 2134, 2134, 2134, 2134, 3465,    0,    0,
 3464,    0, 2134,    0, 2134, 2134, 2134, 2134, 2134, 2134,
    0, 2134,    0,    0,    0,    0, 1424, 1423, 1425,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2134, 2134,
    0, 2134, 2134, 2134, 2134, 2134, 2134, 2134, 2134, 2134,
    0, 2134, 2134, 2134, 2134, 2134, 2134, 2134, 2134, 2134,
 2134, 2134, 2134,    0, 2134, 2134, 2134,    0, 2134,    0,
    0,    0, 2134, 2134, 2134, 2134, 2134, 2134, 2134, 2134,
 2134, 2134, 2134, 2134, 2134, 2136, 2134, 2134,    0,    0,
    0,    0,    0, 2134, 2134, 2134, 2134, 2134, 2134, 2134,
 2134, 2134, 2134, 2134, 2134, 2134, 2134, 2134, 2134, 2134,
 2134,    0,    0, 2134,    0, 2134, 2134, 2134, 2134, 2134,
 2134,    0,    0,    0,    0, 1895, 2136, 2136, 2136, 2136,
 2136, 2136, 2136,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 2136, 2136, 2136, 2136, 2136, 3197,    0,
    0,    0,    0,    0, 3198,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 3199,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2136, 3200, 2136,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1408,    0,
    0,    0,    0,    0,    0, 3201,    0, 1409,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1410,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 3202,    0,    0,    0, 3203,
    0,    0,    0,    0, 3204,    0,    0,    0,    0,    0,
 3205,    0,    0,    0, 1411,    0,    0,    0,    0,    0,
    0, 1412,    0,    0,    0,    0,    0,    0,    0, 1413,
    0,    0, 3206,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1414,    0, 1415,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1416,    0,
 1417, 1418, 1419,    0,    0,    0,    0, 1420,    0,    0,
    0,    0, 2136,    0, 2136,    0,    0, 2136,    0, 2136,
 2136, 2136,    0, 2136, 2136,    0, 2136,    0, 3207,    0,
    0, 3208, 2136,    0, 2136, 2136, 2136, 2136,    0, 2136,
    0,    0, 2136,    0,    0, 2136,    0,    0, 1421,    0,
    0,    0, 2136,    0, 2136, 2136,    0, 2136, 2136, 2136,
 2136, 2136,    0,    0, 2136,    0,    0, 2136,    0, 2136,
 2136,    0,    0, 2136, 2136, 2136,    0, 2136, 2136, 2136,
 2136, 2136, 2136, 2136,    0,    0,    0, 2136, 2136, 2136,
 2136, 2136,    0,    0,    0, 2136,    0,    0,    0,    0,
    0,    0,    0, 2136, 2136, 2136, 2136, 2136, 2136,    0,
    0,    0, 2136, 2136,    0, 2136,    0, 2136, 3209, 2136,
 2136, 2136,    0, 2136,    0, 2136, 2136, 2136,    0, 2136,
 2136, 2136, 2136, 2136,    0, 2136, 2136, 3210,    0,    0,
    0, 2136,    0, 2136, 2136,    0,    0, 2136, 2136, 2136,
    0,    0, 2136,    0, 2136,    0,    0, 2136, 2136, 2136,
 2136,    0, 2136,    0, 2136,    0, 2136, 2136, 2136, 2136,
 2136,    0,    0, 2136, 2136, 2136,    0, 2136, 2136,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2136, 2136,
 2136,    0, 2136, 2136, 2136,    0,    0,    0,    0,    0,
 2136,    0,    0,    0, 2136,    0, 2136, 2136, 2136, 2136,
    0, 2136,    0, 2136, 2136, 2136, 2136, 2136, 2136, 2136,
 2136,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2136, 2136,    0, 2136, 2136, 2136,    0,    0, 1422,    0,
 2136, 2136, 2136, 2136, 2136,    0, 2136,    0,    0, 2136,
 2136,    0,    0,    0, 2136, 2136,    0,    0, 2136, 2136,
 2136, 2136, 2136, 2136, 2136, 2136, 2136,    0, 2136,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2136, 2136,
 2136, 2136, 2136, 2136, 2136, 2136, 2136, 2136, 2136, 2136,
 2136, 2136, 2136, 2136, 2136, 2136, 2136, 2136, 2136,    0,
 2136, 2136, 2136, 2136, 2136, 2136,    0, 2136, 2136, 2136,
 2136,    0, 2136,    0, 2136, 2136, 2136, 2136, 2136, 2136,
 2136, 2136, 2136, 2136, 2136, 2136, 2136, 2136, 2136, 2136,
 2136, 2136,    0, 2136, 2136, 2136, 2136,    0, 2136, 2136,
 2136, 2136, 2136,    0, 2136, 2136, 2136, 2136, 2136, 2136,
 2136, 2136, 2136, 2136, 2136, 2136, 2136, 2136, 2136, 2136,
 2136, 2136, 2136, 2136, 2136, 2136, 2136, 2136, 2136, 2136,
 2136, 2136, 2136,    0, 2136, 2136,    0,    0,    0, 2136,
 2136, 2136, 2136, 2136, 2136,    0,    0, 2136, 2136, 2136,
 2136, 2136, 2136, 2136, 2136, 2136, 3467,    0,    0,    0,
    0, 2136,    0, 2136, 2136, 2136, 2136, 2136, 2136,    0,
 2136,    0,    0,    0,    0, 1424, 1423, 1425,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 2136, 2136,    0,
 2136, 2136, 2136, 2136, 2136, 2136, 2136, 2136, 2136,    0,
 2136, 2136, 2136, 2136, 2136, 2136, 2136, 2136, 2136, 2136,
 2136, 2136,    0, 2136, 2136, 2136,    0, 2136,    0,    0,
    0, 2136, 2136, 2136, 2136, 2136, 2136, 2136, 2136, 2136,
 2136, 2136, 2136, 2136, 2138, 2136, 2136,    0,    0,    0,
    0,    0, 2136, 2136, 2136, 2136, 2136, 2136, 2136, 2136,
 2136, 2136, 2136, 2136, 2136, 2136, 2136, 2136, 2136, 2136,
    0,    0, 2136,    0, 2136, 2136, 2136, 2136, 2136, 2136,
    0,    0,    0,    0, 1897, 2138, 2138, 2138, 2138, 2138,
 2138, 2138,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 2138, 2138, 2138, 2138, 2138,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 2138,    0, 2138,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1408,    0,    0,
    0,    0,    0,    0,    0,    0, 1409,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1410,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 3466,    0,    0,
    0,    0,    0, 1411,    0,    0,    0,    0,    0,    0,
 1412,    0,    0,    0,    0,    0,    0,    0, 1413,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1414,    0, 1415,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1416,    0, 1417,
 1418, 1419,    0,    0,    0,    0, 1420,    0,    0,    0,
    0, 2138,    0, 2138,    0,    0, 2138,    0, 2138, 2138,
 2138,    0, 2138, 2138,    0, 2138,    0,    0,    0,    0,
    0, 2138,    0, 2138, 2138, 2138, 2138,    0, 2138,    0,
    0, 2138,    0,    0, 2138,    0,    0, 1421,    0,    0,
    0, 2138,    0, 2138, 2138,    0, 2138, 2138, 2138, 2138,
 2138,    0,    0, 2138,    0,    0, 2138,    0, 2138, 2138,
    0,    0, 2138, 2138, 2138,    0, 2138, 2138, 2138, 2138,
 2138, 2138, 2138,    0,    0,    0, 2138, 2138, 2138, 2138,
 2138,    0,    0,    0, 2138,    0,    0,    0,    0,    0,
    0,    0, 2138, 2138, 2138, 2138, 2138, 2138,    0,    0,
    0, 2138, 2138,    0, 2138,    0, 2138,    0, 2138, 2138,
 2138,    0, 2138,    0, 2138, 2138, 2138,    0, 2138, 2138,
 2138, 2138, 2138,    0, 2138, 2138,    0,    0,    0,    0,
 2138,    0, 2138, 2138,    0,    0, 2138, 2138, 2138,    0,
    0, 2138,    0, 2138,    0,    0, 2138, 2138, 2138, 2138,
    0, 2138,    0, 2138,    0, 2138, 2138, 2138, 2138, 2138,
    0,    0, 2138, 2138, 2138,    0, 2138, 2138,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 2138, 2138, 2138,
    0, 2138, 2138, 2138,    0,    0,    0,    0,    0, 2138,
    0,    0,    0, 2138,    0, 2138, 2138, 2138, 2138,    0,
 2138,    0, 2138, 2138, 2138, 2138, 2138, 2138, 2138, 2138,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 2138,
 2138,    0, 2138, 2138, 2138,    0,    0, 1422,    0, 2138,
 2138, 2138, 2138, 2138,    0, 2138,    0,    0, 2138, 2138,
    0,    0,    0, 2138, 2138,    0,    0, 2138, 2138, 2138,
 2138, 2138, 2138, 2138, 2138, 2138,    0, 2138,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 2138, 2138, 2138,
 2138, 2138, 2138, 2138, 2138, 2138, 2138, 2138, 2138, 2138,
 2138, 2138, 2138, 2138, 2138, 2138, 2138, 2138,    0, 2138,
 2138, 2138, 2138, 2138, 2138,    0, 2138, 2138, 2138, 2138,
    0, 2138,    0, 2138, 2138, 2138, 2138, 2138, 2138, 2138,
 2138, 2138, 2138, 2138, 2138, 2138, 2138, 2138, 2138, 2138,
 2138,    0, 2138, 2138, 2138, 2138,    0, 2138, 2138, 2138,
 2138, 2138,    0, 2138, 2138, 2138, 2138, 2138, 2138, 2138,
 2138, 2138, 2138, 2138, 2138, 2138, 2138, 2138, 2138, 2138,
 2138, 2138, 2138, 2138, 2138, 2138, 2138, 2138, 2138, 2138,
 2138, 2138,    0, 2138, 2138,    0,    0,    0, 2138, 2138,
 2138, 2138, 2138, 2138,    0,    0, 2138, 2138, 2138, 2138,
 2138, 2138, 2138, 2138, 2138, 1953,    0,    0,    0,    0,
 2138,    0, 2138, 2138, 2138, 2138, 2138, 2138,    0, 2138,
    0,    0,    0,    0, 1424, 1423, 1425,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2138, 2138,    0, 2138,
 2138, 2138, 2138, 2138, 2138, 2138, 2138, 2138,    0, 2138,
 2138, 2138, 2138, 2138, 2138, 2138, 2138, 2138, 2138, 2138,
 2138,    0, 2138, 2138, 2138,    0, 2138,    0,    0,    0,
 2138, 2138, 2138, 2138, 2138, 2138, 2138, 2138, 2138, 2138,
 2138, 2138, 2138, 2139, 2138, 2138,    0,    0,    0,    0,
    0, 2138, 2138, 2138, 2138, 2138, 2138, 2138, 2138, 2138,
 2138, 2138, 2138, 2138, 2138, 2138, 2138, 2138, 2138,    0,
    0, 2138,    0, 2138, 2138, 2138, 2138, 2138, 2138,    0,
    0,    0,    0, 1898, 2139, 2139, 2139, 2139, 2139, 2139,
 2139,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 2139, 2139, 2139, 2139, 2139,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 2139,    0, 2139,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1408,    0,    0,    0,
    0,    0,    0,    0,    0, 1409,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1410,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1411,    0,    0,    0,    0,    0,    0, 1412,
    0,    0,    0,    0,    0,    0,    0, 1413,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1414,    0, 1415,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1416,    0, 1417, 1418,
 1419,    0,    0,    0,    0, 1420,    0,    0,    0,    0,
 2139,    0, 2139,    0,    0, 2139,    0, 2139, 2139, 2139,
    0, 2139, 2139,    0, 2139,    0,    0,    0,    0,    0,
 2139,    0, 2139, 2139, 2139, 2139,    0, 2139,    0,    0,
 2139,    0,    0, 2139,    0,    0, 1421,    0,    0,    0,
 2139,    0, 2139, 2139,    0, 2139, 2139, 2139, 2139, 2139,
    0,    0, 2139,    0,    0, 2139,    0, 2139, 2139,    0,
    0, 2139, 2139, 2139,    0, 2139, 2139, 2139, 2139, 2139,
 2139, 2139,    0,    0,    0, 2139, 2139, 2139, 2139, 2139,
    0,    0,    0, 2139,    0,    0,    0,    0,    0,    0,
    0, 2139, 2139, 2139, 2139, 2139, 2139,    0,    0,    0,
 2139, 2139,    0, 2139,    0, 2139,    0, 2139, 2139, 2139,
    0, 2139,    0, 2139, 2139, 2139,    0, 2139, 2139, 2139,
 2139, 2139,    0, 2139, 2139,    0,    0,    0,    0, 2139,
    0, 2139, 2139,    0,    0, 2139, 2139, 2139,    0,    0,
 2139,    0, 2139,    0,    0, 2139, 2139, 2139, 2139,    0,
 2139,    0, 2139,    0, 2139, 2139, 2139, 2139, 2139,    0,
    0, 2139, 2139, 2139,    0, 2139, 2139,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2139, 2139, 2139,    0,
 2139, 2139, 2139,    0,    0,    0,    0,    0, 2139,    0,
    0,    0, 2139,    0, 2139, 2139, 2139, 2139,    0, 2139,
    0, 2139, 2139, 2139, 2139, 2139, 2139, 2139, 2139,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2139, 2139,
    0, 2139, 2139, 2139,    0,    0, 1422,    0, 2139, 2139,
 2139, 2139, 2139,    0, 2139,    0,    0, 2139, 2139,    0,
    0,    0, 2139, 2139,    0,    0, 2139, 2139, 2139, 2139,
 2139, 2139, 2139, 2139, 2139,    0, 2139,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2139, 2139, 2139, 2139,
 2139, 2139, 2139, 2139, 2139, 2139, 2139, 2139, 2139, 2139,
 2139, 2139, 2139, 2139, 2139, 2139, 2139,    0, 2139, 2139,
 2139, 2139, 2139, 2139,    0, 2139, 2139, 2139, 2139,    0,
 2139,    0, 2139, 2139, 2139, 2139, 2139, 2139, 2139, 2139,
 2139, 2139, 2139, 2139, 2139, 2139, 2139, 2139, 2139, 2139,
    0, 2139, 2139, 2139, 2139,    0, 2139, 2139, 2139, 2139,
 2139,    0, 2139, 2139, 2139, 2139, 2139, 2139, 2139, 2139,
 2139, 2139, 2139, 2139, 2139, 2139, 2139, 2139, 2139, 2139,
 2139, 2139, 2139, 2139, 2139, 2139, 2139, 2139, 2139, 2139,
 2139,    0, 2139, 2139,    0,    0,    0, 2139, 2139, 2139,
 2139, 2139, 2139,    0,    0, 2139, 2139, 2139, 2139, 2139,
 2139, 2139, 2139, 2139, 1954,    0,    0,    0,    0, 2139,
    0, 2139, 2139, 2139, 2139, 2139, 2139,    0, 2139,    0,
    0,    0,    0, 1424, 1423, 1425,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 2139, 2139,    0, 2139, 2139,
 2139, 2139, 2139, 2139, 2139, 2139, 2139,    0, 2139, 2139,
 2139, 2139, 2139, 2139, 2139, 2139, 2139, 2139, 2139, 2139,
    0, 2139, 2139, 2139,    0, 2139,    0,    0,    0, 2139,
 2139, 2139, 2139, 2139, 2139, 2139, 2139, 2139, 2139, 2139,
 2139, 2139, 2140, 2139, 2139,    0,    0,    0,    0,    0,
 2139, 2139, 2139, 2139, 2139, 2139, 2139, 2139, 2139, 2139,
 2139, 2139, 2139, 2139, 2139, 2139, 2139, 2139,    0,    0,
 2139,    0, 2139, 2139, 2139, 2139, 2139, 2139,    0,    0,
    0,    0, 1899, 2140, 2140, 2140, 2140, 2140, 2140, 2140,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2140, 2140, 2140, 2140, 2140,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 2140,    0, 2140,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1408,    0,    0,    0,    0,
    0,    0,    0,    0, 1409,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1410,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1411,    0,    0,    0,    0,    0,    0, 1412,    0,
    0,    0,    0,    0,    0,    0, 1413,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1414,    0, 1415,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1416,    0, 1417, 1418, 1419,
    0,    0,    0,    0, 1420,    0,    0,    0,    0, 2140,
    0, 2140,    0,    0, 2140,    0, 2140, 2140, 2140,    0,
 2140, 2140,    0, 2140,    0,    0,    0,    0,    0, 2140,
    0, 2140, 2140, 2140, 2140,    0, 2140,    0,    0, 2140,
    0,    0, 2140,    0,    0, 1421,    0,    0,    0, 2140,
    0, 2140, 2140,    0, 2140, 2140, 2140, 2140, 2140,    0,
    0, 2140,    0,    0, 2140,    0, 2140, 2140,    0,    0,
 2140, 2140, 2140,    0, 2140, 2140, 2140, 2140, 2140, 2140,
 2140,    0,    0,    0, 2140, 2140, 2140, 2140, 2140,    0,
    0,    0, 2140,    0,    0,    0,    0,    0,    0,    0,
 2140, 2140, 2140, 2140, 2140, 2140,    0,    0,    0, 2140,
 2140,    0, 2140,    0, 2140,    0, 2140, 2140, 2140,    0,
 2140,    0, 2140, 2140, 2140,    0, 2140, 2140, 2140, 2140,
 2140,    0, 2140, 2140,    0,    0,    0,    0, 2140,    0,
 2140, 2140,    0,    0, 2140, 2140, 2140,    0,    0, 2140,
    0, 2140,    0,    0, 2140, 2140, 2140, 2140,    0, 2140,
    0, 2140,    0, 2140, 2140, 2140, 2140, 2140,    0,    0,
 2140, 2140, 2140,    0, 2140, 2140,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 2140, 2140, 2140,    0, 2140,
 2140, 2140,    0,    0,    0,    0,    0, 2140,    0,    0,
    0, 2140,    0, 2140, 2140, 2140, 2140,    0, 2140,    0,
 2140, 2140, 2140, 2140, 2140, 2140, 2140, 2140,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 2140, 2140,    0,
 2140, 2140, 2140,    0,    0, 1422,    0, 2140, 2140, 2140,
 2140, 2140,    0, 2140,    0,    0, 2140, 2140,    0,    0,
    0, 2140, 2140,    0,    0, 2140, 2140, 2140, 2140, 2140,
 2140, 2140, 2140, 2140,    0, 2140,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 2140, 2140, 2140, 2140, 2140,
 2140, 2140, 2140, 2140, 2140, 2140, 2140, 2140, 2140, 2140,
 2140, 2140, 2140, 2140, 2140, 2140,    0, 2140, 2140, 2140,
 2140, 2140, 2140,    0, 2140, 2140, 2140, 2140,    0, 2140,
    0, 2140, 2140, 2140, 2140, 2140, 2140, 2140, 2140, 2140,
 2140, 2140, 2140, 2140, 2140, 2140, 2140, 2140, 2140,    0,
 2140, 2140, 2140, 2140,    0, 2140, 2140, 2140, 2140, 2140,
    0, 2140, 2140, 2140, 2140, 2140, 2140, 2140, 2140, 2140,
 2140, 2140, 2140, 2140, 2140, 2140, 2140, 2140, 2140, 2140,
 2140, 2140, 2140, 2140, 2140, 2140, 2140, 2140, 2140, 2140,
    0, 2140, 2140,    0,    0,    0, 2140, 2140, 2140, 2140,
 2140, 2140,    0,    0, 2140, 2140, 2140, 2140, 2140, 2140,
 2140, 2140, 2140, 2115,    0,    0,    0,    0, 2140,    0,
 2140, 2140, 2140, 2140, 2140, 2140,    0, 2140,    0,    0,
    0,    0, 1424, 1423, 1425,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 2140, 2140,    0, 2140, 2140, 2140,
 2140, 2140, 2140, 2140, 2140, 2140,    0, 2140, 2140, 2140,
 2140, 2140, 2140, 2140, 2140, 2140, 2140, 2140, 2140,    0,
 2140, 2140, 2140,    0, 2140,    0,    0,    0, 2140, 2140,
 2140, 2140, 2140, 2140, 2140, 2140, 2140, 2140, 2140, 2140,
 2140, 2141, 2140, 2140,    0,    0,    0,    0,    0, 2140,
 2140, 2140, 2140, 2140, 2140, 2140, 2140, 2140, 2140, 2140,
 2140, 2140, 2140, 2140, 2140, 2140, 2140,    0,    0, 2140,
    0, 2140, 2140, 2140, 2140, 2140, 2140,    0,    0,    0,
    0, 1901, 2141, 2141, 2141, 2141, 2141, 2141, 2141,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 2141,
 2141, 2141, 2141, 2141,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 2141,    0, 2141,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1408,    0,    0,    0,    0,    0,
    0,    0,    0, 1409,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1410,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1411,    0,    0,    0,    0,    0,    0, 1412,    0,    0,
    0,    0,    0,    0,    0, 1413,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1414,    0, 1415,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1416,    0, 1417, 1418, 1419,    0,
    0,    0,    0, 1420,    0,    0,    0,    0, 2141,    0,
 2141,    0,    0, 2141,    0, 2141, 2141, 2141,    0, 2141,
 2141,    0, 2141,    0,    0,    0,    0,    0, 2141,    0,
 2141, 2141, 2141, 2141,    0, 2141,    0,    0, 2141,    0,
    0, 2141,    0,    0, 1421,    0,    0,    0, 2141,    0,
 2141, 2141,    0, 2141, 2141, 2141, 2141, 2141,    0,    0,
 2141,    0,    0, 2141,    0, 2141, 2141,    0,    0, 2141,
 2141, 2141,    0, 2141, 2141, 2141, 2141, 2141, 2141, 2141,
    0,    0,    0, 2141, 2141, 2141, 2141, 2141,    0,    0,
    0, 2141,    0,    0,    0,    0,    0,    0,    0, 2141,
 2141, 2141, 2141, 2141, 2141,    0,    0,    0, 2141, 2141,
    0, 2141,    0, 2141,    0, 2141, 2141, 2141,    0, 2141,
    0, 2141, 2141, 2141,    0, 2141, 2141, 2141, 2141, 2141,
    0, 2141, 2141,    0,    0,    0,    0, 2141,    0, 2141,
 2141,    0,    0, 2141, 2141, 2141,    0,    0, 2141,    0,
 2141,    0,    0, 2141, 2141, 2141, 2141,    0, 2141,    0,
 2141,    0, 2141, 2141, 2141, 2141, 2141,    0,    0, 2141,
 2141, 2141,    0, 2141, 2141,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 2141, 2141, 2141,    0, 2141, 2141,
 2141,    0,    0,    0,    0,    0, 2141,    0,    0,    0,
 2141,    0, 2141, 2141, 2141, 2141,    0, 2141,    0, 2141,
 2141, 2141, 2141, 2141, 2141, 2141, 2141,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2141, 2141,    0, 2141,
 2141, 2141,    0,    0, 1422,    0, 2141, 2141, 2141, 2141,
 2141,    0, 2141,    0,    0, 2141, 2141,    0,    0,    0,
 2141, 2141,    0,    0, 2141, 2141, 2141, 2141, 2141, 2141,
 2141, 2141, 2141,    0, 2141,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 2141, 2141, 2141, 2141, 2141, 2141,
 2141, 2141, 2141, 2141, 2141, 2141, 2141, 2141, 2141, 2141,
 2141, 2141, 2141, 2141, 2141,    0, 2141, 2141, 2141, 2141,
 2141, 2141,    0, 2141, 2141, 2141, 2141,    0, 2141,    0,
 2141, 2141, 2141, 2141, 2141, 2141, 2141, 2141, 2141, 2141,
 2141, 2141, 2141, 2141, 2141, 2141, 2141, 2141,    0, 2141,
 2141, 2141, 2141,    0, 2141, 2141, 2141, 2141, 2141,    0,
 2141, 2141, 2141, 2141, 2141, 2141, 2141, 2141, 2141, 2141,
 2141, 2141, 2141, 2141, 2141, 2141, 2141, 2141, 2141, 2141,
 2141, 2141, 2141, 2141, 2141, 2141, 2141, 2141, 2141,    0,
 2141, 2141,    0,    0,    0, 2141, 2141, 2141, 2141, 2141,
 2141,    0,    0, 2141, 2141, 2141, 2141, 2141, 2141, 2141,
 2141, 2141, 2139,    0,    0,    0,    0, 2141,    0, 2141,
 2141, 2141, 2141, 2141, 2141,    0, 2141,    0,    0,    0,
    0, 1424, 1423, 1425,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 2141, 2141,    0, 2141, 2141, 2141, 2141,
 2141, 2141, 2141, 2141, 2141,    0, 2141, 2141, 2141, 2141,
 2141, 2141, 2141, 2141, 2141, 2141, 2141, 2141,    0, 2141,
 2141, 2141,    0, 2141,    0,    0,    0, 2141, 2141, 2141,
 2141, 2141, 2141, 2141, 2141, 2141, 2141, 2141, 2141, 2141,
 2142, 2141, 2141,    0,    0,    0,    0,    0, 2141, 2141,
 2141, 2141, 2141, 2141, 2141, 2141, 2141, 2141, 2141, 2141,
 2141, 2141, 2141, 2141, 2141, 2141,    0,    0, 2141,    0,
 2141, 2141, 2141, 2141, 2141, 2141,    0,    0,    0,    0,
 1902, 2142, 2142, 2142, 2142, 2142, 2142, 2142,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2142, 2142,
 2142, 2142, 2142,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 2142,    0, 2142,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1408,    0,    0,    0,    0,    0,    0,
    0,    0, 1409,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1410,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1411,
    0,    0,    0,    0,    0,    0, 1412,    0,    0,    0,
    0,    0,    0,    0, 1413,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1414,    0, 1415,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1416,    0, 1417, 1418, 1419,    0,    0,
    0,    0, 1420,    0,    0,    0,    0, 2142,    0, 2142,
    0,    0, 2142,    0, 2142, 2142, 2142,    0, 2142, 2142,
    0, 2142,    0,    0,    0,    0,    0, 2142,    0, 2142,
 2142, 2142, 2142,    0, 2142,    0,    0, 2142,    0,    0,
 2142,    0,    0, 1421,    0,    0,    0, 2142,    0, 2142,
 2142,    0, 2142, 2142, 2142, 2142, 2142,    0,    0, 2142,
    0,    0, 2142,    0, 2142, 2142,    0,    0, 2142, 2142,
 2142,    0, 2142, 2142, 2142, 2142, 2142, 2142, 2142,    0,
    0,    0, 2142, 2142, 2142, 2142, 2142,    0,    0,    0,
 2142,    0,    0,    0,    0,    0,    0,    0, 2142, 2142,
 2142, 2142, 2142, 2142,    0,    0,    0, 2142, 2142,    0,
 2142,    0, 2142,    0, 2142, 2142, 2142,    0, 2142,    0,
 2142, 2142, 2142,    0, 2142, 2142, 2142, 2142, 2142,    0,
 2142, 2142,    0,    0,    0,    0, 2142,    0, 2142, 2142,
    0,    0, 2142, 2142, 2142,    0,    0, 2142,    0, 2142,
    0,    0, 2142, 2142, 2142, 2142,    0, 2142,    0, 2142,
    0, 2142, 2142, 2142, 2142, 2142,    0,    0, 2142, 2142,
 2142,    0, 2142, 2142,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 2142, 2142, 2142,    0, 2142, 2142, 2142,
    0,    0,    0,    0,    0, 2142,    0,    0,    0, 2142,
    0, 2142, 2142, 2142, 2142,    0, 2142,    0, 2142, 2142,
 2142, 2142, 2142, 2142, 2142, 2142,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 2142, 2142,    0, 2142, 2142,
 2142,    0,    0, 1422,    0, 2142, 2142, 2142, 2142, 2142,
    0, 2142,    0,    0, 2142, 2142,    0,    0,    0, 2142,
 2142,    0,    0, 2142, 2142, 2142, 2142, 2142, 2142, 2142,
 2142, 2142,    0, 2142,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 2142, 2142, 2142, 2142, 2142, 2142, 2142,
 2142, 2142, 2142, 2142, 2142, 2142, 2142, 2142, 2142, 2142,
 2142, 2142, 2142, 2142,    0, 2142, 2142, 2142, 2142, 2142,
 2142,    0, 2142, 2142, 2142, 2142,    0, 2142,    0, 2142,
 2142, 2142, 2142, 2142, 2142, 2142, 2142, 2142, 2142, 2142,
 2142, 2142, 2142, 2142, 2142, 2142, 2142,    0, 2142, 2142,
 2142, 2142,    0, 2142, 2142, 2142, 2142, 2142,    0, 2142,
 2142, 2142, 2142, 2142, 2142, 2142, 2142, 2142, 2142, 2142,
 2142, 2142, 2142, 2142, 2142, 2142, 2142, 2142, 2142, 2142,
 2142, 2142, 2142, 2142, 2142, 2142, 2142, 2142,    0, 2142,
 2142,    0,    0,    0, 2142, 2142, 2142, 2142, 2142, 2142,
    0,    0, 2142, 2142, 2142, 2142, 2142, 2142, 2142, 2142,
 2142, 2140,    0,    0,    0,    0, 2142,    0, 2142, 2142,
 2142, 2142, 2142, 2142,    0, 2142,    0,    0,    0,    0,
 1424, 1423, 1425,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 2142, 2142,    0, 2142, 2142, 2142, 2142, 2142,
 2142, 2142, 2142, 2142,    0, 2142, 2142, 2142, 2142, 2142,
 2142, 2142, 2142, 2142, 2142, 2142, 2142,    0, 2142, 2142,
 2142,    0, 2142,    0,    0,    0, 2142, 2142, 2142, 2142,
 2142, 2142, 2142, 2142, 2142, 2142, 2142, 2142, 2142, 2143,
 2142, 2142,    0,    0,    0,    0,    0, 2142, 2142, 2142,
 2142, 2142, 2142, 2142, 2142, 2142, 2142, 2142, 2142, 2142,
 2142, 2142, 2142, 2142, 2142,    0,    0, 2142,    0, 2142,
 2142, 2142, 2142, 2142, 2142,    0,    0,    0,    0, 1906,
 2143, 2143, 2143, 2143, 2143, 2143, 2143,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 2143, 2143, 2143,
 2143, 2143,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2143,    0, 2143,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1408,    0,    0,    0,    0,    0,    0,    0,
    0, 1409,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1410,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1411,    0,
    0,    0,    0,    0,    0, 1412,    0,    0,    0,    0,
    0,    0,    0, 1413,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1414,    0, 1415,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1416,    0, 1417, 1418, 1419,    0,    0,    0,
    0, 1420,    0,    0,    0,    0, 2143,    0, 2143,    0,
    0, 2143,    0, 2143, 2143, 2143,    0, 2143, 2143,    0,
 2143,    0,    0,    0,    0,    0, 2143,    0, 2143, 2143,
 2143, 2143,    0, 2143,    0,    0, 2143,    0,    0, 2143,
    0,    0, 1421,    0,    0,    0, 2143,    0, 2143, 2143,
    0, 2143, 2143, 2143, 2143, 2143,    0,    0, 2143,    0,
    0, 2143,    0, 2143, 2143,    0,    0, 2143, 2143, 2143,
    0, 2143, 2143, 2143, 2143, 2143, 2143, 2143,    0,    0,
    0, 2143, 2143, 2143, 2143, 2143,    0,    0,    0, 2143,
    0,    0,    0,    0,    0,    0,    0, 2143, 2143, 2143,
 2143, 2143, 2143,    0,    0,    0, 2143, 2143,    0, 2143,
    0, 2143,    0, 2143, 2143, 2143,    0, 2143,    0, 2143,
 2143, 2143,    0, 2143, 2143, 2143, 2143, 2143,    0, 2143,
 2143,    0,    0,    0,    0, 2143,    0, 2143, 2143,    0,
    0, 2143, 2143, 2143,    0,    0, 2143,    0, 2143,    0,
    0, 2143, 2143, 2143, 2143,    0, 2143,    0, 2143,    0,
 2143, 2143, 2143, 2143, 2143,    0,    0, 2143, 2143, 2143,
    0, 2143, 2143,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 2143, 2143, 2143,    0, 2143, 2143, 2143,    0,
    0,    0,    0,    0, 2143,    0,    0,    0, 2143,    0,
 2143, 2143, 2143, 2143,    0, 2143,    0, 2143, 2143, 2143,
 2143, 2143, 2143, 2143, 2143,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 2143, 2143,    0, 2143, 2143, 2143,
    0,    0, 1422,    0, 2143, 2143, 2143, 2143, 2143,    0,
 2143,    0,    0, 2143, 2143,    0,    0,    0, 2143, 2143,
    0,    0, 2143, 2143, 2143, 2143, 2143, 2143, 2143, 2143,
 2143,    0, 2143,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 2143, 2143, 2143, 2143, 2143, 2143, 2143, 2143,
 2143, 2143, 2143, 2143, 2143, 2143, 2143, 2143, 2143, 2143,
 2143, 2143, 2143,    0, 2143, 2143, 2143, 2143, 2143, 2143,
    0, 2143, 2143, 2143, 2143,    0, 2143,    0, 2143, 2143,
 2143, 2143, 2143, 2143, 2143, 2143, 2143, 2143, 2143, 2143,
 2143, 2143, 2143, 2143, 2143, 2143,    0, 2143, 2143, 2143,
 2143,    0, 2143, 2143, 2143, 2143, 2143,    0, 2143, 2143,
 2143, 2143, 2143, 2143, 2143, 2143, 2143, 2143, 2143, 2143,
 2143, 2143, 2143, 2143, 2143, 2143, 2143, 2143, 2143, 2143,
 2143, 2143, 2143, 2143, 2143, 2143, 2143,    0, 2143, 2143,
    0,    0,    0, 2143, 2143, 2143, 2143, 2143, 2143,    0,
    0, 2143, 2143, 2143, 2143, 2143, 2143, 2143, 2143, 2143,
 2141,    0,    0,    0,    0, 2143,    0, 2143, 2143, 2143,
 2143, 2143, 2143,    0, 2143,    0,    0,    0,    0, 1424,
 1423, 1425,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 2143, 2143,    0, 2143, 2143, 2143, 2143, 2143, 2143,
 2143, 2143, 2143,    0, 2143, 2143, 2143, 2143, 2143, 2143,
 2143, 2143, 2143, 2143, 2143, 2143,    0, 2143, 2143, 2143,
    0, 2143,    0,    0,    0, 2143, 2143, 2143, 2143, 2143,
 2143, 2143, 2143, 2143, 2143, 2143, 2143, 2143, 2144, 2143,
 2143,    0,    0,    0,    0,    0, 2143, 2143, 2143, 2143,
 2143, 2143, 2143, 2143, 2143, 2143, 2143, 2143, 2143, 2143,
 2143, 2143, 2143, 2143,    0,    0, 2143,    0, 2143, 2143,
 2143, 2143, 2143, 2143,    0,    0,    0,    0, 1908, 2144,
 2144, 2144, 2144, 2144, 2144, 2144,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2144, 2144, 2144, 2144,
 2144,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 2144,
    0, 2144,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1408,    0,    0,    0,    0,    0,    0,    0,    0,
 1409,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1410,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1411,    0,    0,
    0,    0,    0,    0, 1412,    0,    0,    0,    0,    0,
    0,    0, 1413,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1414,
    0, 1415,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1416,    0, 1417, 1418, 1419,    0,    0,    0,    0,
 1420,    0,    0,    0,    0, 2144,    0, 2144,    0,    0,
 2144,    0, 2144, 2144, 2144,    0, 2144, 2144,    0, 2144,
    0,    0,    0,    0,    0, 2144,    0, 2144, 2144, 2144,
 2144,    0, 2144,    0,    0, 2144,    0,    0, 2144,    0,
    0, 1421,    0,    0,    0, 2144,    0, 2144, 2144,    0,
 2144, 2144, 2144, 2144, 2144,    0,    0, 2144,    0,    0,
 2144,    0, 2144, 2144,    0,    0, 2144, 2144, 2144,    0,
 2144, 2144, 2144, 2144, 2144, 2144, 2144,    0,    0,    0,
 2144, 2144, 2144, 2144, 2144,    0,    0,    0, 2144,    0,
    0,    0,    0,    0,    0,    0, 2144, 2144, 2144, 2144,
 2144, 2144,    0,    0,    0, 2144, 2144,    0, 2144,    0,
 2144,    0, 2144, 2144, 2144,    0, 2144,    0, 2144, 2144,
 2144,    0, 2144, 2144, 2144, 2144, 2144,    0, 2144, 2144,
    0,    0,    0,    0, 2144,    0, 2144, 2144,    0,    0,
 2144, 2144, 2144,    0,    0, 2144,    0, 2144,    0,    0,
 2144, 2144, 2144, 2144,    0, 2144,    0, 2144,    0, 2144,
 2144, 2144, 2144, 2144,    0,    0, 2144, 2144, 2144,    0,
 2144, 2144,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 2144, 2144, 2144,    0, 2144, 2144, 2144,    0,    0,
    0,    0,    0, 2144,    0,    0,    0, 2144,    0, 2144,
 2144, 2144, 2144,    0, 2144,    0, 2144, 2144, 2144, 2144,
 2144, 2144, 2144, 2144,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 2144, 2144,    0, 2144, 2144, 2144,    0,
    0, 1422,    0, 2144, 2144, 2144, 2144, 2144,    0, 2144,
    0,    0, 2144, 2144,    0,    0,    0, 2144, 2144,    0,
    0, 2144, 2144, 2144, 2144, 2144, 2144, 2144, 2144, 2144,
    0, 2144,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 2144, 2144, 2144, 2144, 2144, 2144, 2144, 2144, 2144,
 2144, 2144, 2144, 2144, 2144, 2144, 2144, 2144, 2144, 2144,
 2144, 2144,    0, 2144, 2144, 2144, 2144, 2144, 2144,    0,
 2144, 2144, 2144, 2144,    0, 2144,    0, 2144, 2144, 2144,
 2144, 2144, 2144, 2144, 2144, 2144, 2144, 2144, 2144, 2144,
 2144, 2144, 2144, 2144, 2144,    0, 2144, 2144, 2144, 2144,
    0, 2144, 2144, 2144, 2144, 2144,    0, 2144, 2144, 2144,
 2144, 2144, 2144, 2144, 2144, 2144, 2144, 2144, 2144, 2144,
 2144, 2144, 2144, 2144, 2144, 2144, 2144, 2144, 2144, 2144,
 2144, 2144, 2144, 2144, 2144, 2144,    0, 2144, 2144,    0,
    0,    0, 2144, 2144, 2144, 2144, 2144, 2144,    0,    0,
 2144, 2144, 2144, 2144, 2144, 2144, 2144, 2144, 2144, 2142,
    0,    0,    0,    0, 2144,    0, 2144, 2144, 2144, 2144,
 2144, 2144,    0, 2144,    0,    0,    0,    0, 1424, 1423,
 1425,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2144, 2144,    0, 2144, 2144, 2144, 2144, 2144, 2144, 2144,
 2144, 2144,    0, 2144, 2144, 2144, 2144, 2144, 2144, 2144,
 2144, 2144, 2144, 2144, 2144,    0, 2144, 2144, 2144,    0,
 2144,    0,    0,    0, 2144, 2144, 2144, 2144, 2144, 2144,
 2144, 2144, 2144, 2144, 2144, 2144, 2144, 2145, 2144, 2144,
    0,    0,    0,    0,    0, 2144, 2144, 2144, 2144, 2144,
 2144, 2144, 2144, 2144, 2144, 2144, 2144, 2144, 2144, 2144,
 2144, 2144, 2144,    0,    0, 2144,    0, 2144, 2144, 2144,
 2144, 2144, 2144,    0,    0,    0,    0, 1909, 2145, 2145,
 2145, 2145, 2145, 2145, 2145,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 2145, 2145, 2145, 2145, 2145,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2145,    0,
 2145,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1408,    0,    0,    0,    0,    0,    0,    0,    0, 1409,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1410,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1411,    0,    0,    0,
    0,    0,    0, 1412,    0,    0,    0,    0,    0,    0,
    0, 1413,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1414,    0,
 1415,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1416,    0, 1417, 1418, 1419,    0,    0,    0,    0, 1420,
    0,    0,    0,    0, 2145,    0, 2145,    0,    0, 2145,
    0, 2145, 2145, 2145,    0, 2145, 2145,    0, 2145,    0,
    0,    0,    0,    0, 2145,    0, 2145, 2145, 2145, 2145,
    0, 2145,    0,    0, 2145,    0,    0, 2145,    0,    0,
 1421,    0,    0,    0, 2145,    0, 2145, 2145,    0, 2145,
 2145, 2145, 2145, 2145,    0,    0, 2145,    0,    0, 2145,
    0, 2145, 2145,    0,    0, 2145, 2145, 2145,    0, 2145,
 2145, 2145, 2145, 2145, 2145, 2145,    0,    0,    0, 2145,
 2145, 2145, 2145, 2145,    0,    0,    0, 2145,    0,    0,
    0,    0,    0,    0,    0, 2145, 2145, 2145, 2145, 2145,
 2145,    0,    0,    0, 2145, 2145,    0, 2145,    0, 2145,
    0, 2145, 2145, 2145,    0, 2145,    0, 2145, 2145, 2145,
    0, 2145, 2145, 2145, 2145, 2145,    0, 2145, 2145,    0,
    0,    0,    0, 2145,    0, 2145, 2145,    0,    0, 2145,
 2145, 2145,    0,    0, 2145,    0, 2145,    0,    0, 2145,
 2145, 2145, 2145,    0, 2145,    0, 2145,    0, 2145, 2145,
 2145, 2145, 2145,    0,    0, 2145, 2145, 2145,    0, 2145,
 2145,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2145, 2145, 2145,    0, 2145, 2145, 2145,    0,    0,    0,
    0,    0, 2145,    0,    0,    0, 2145,    0, 2145, 2145,
 2145, 2145,    0, 2145,    0, 2145, 2145, 2145, 2145, 2145,
 2145, 2145, 2145,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 2145, 2145,    0, 2145, 2145, 2145,    0,    0,
 1422,    0, 2145, 2145, 2145, 2145, 2145,    0, 2145,    0,
    0, 2145, 2145,    0,    0,    0, 2145, 2145,    0,    0,
 2145, 2145, 2145, 2145, 2145, 2145, 2145, 2145, 2145,    0,
 2145,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2145, 2145, 2145, 2145, 2145, 2145, 2145, 2145, 2145, 2145,
 2145, 2145, 2145, 2145, 2145, 2145, 2145, 2145, 2145, 2145,
 2145,    0, 2145, 2145, 2145, 2145, 2145, 2145,    0, 2145,
 2145, 2145, 2145,    0, 2145,    0, 2145, 2145, 2145, 2145,
 2145, 2145, 2145, 2145, 2145, 2145, 2145, 2145, 2145, 2145,
 2145, 2145, 2145, 2145,    0, 2145, 2145, 2145, 2145,    0,
 2145, 2145, 2145, 2145, 2145,    0, 2145, 2145, 2145, 2145,
 2145, 2145, 2145, 2145, 2145, 2145, 2145, 2145, 2145, 2145,
 2145, 2145, 2145, 2145, 2145, 2145, 2145, 2145, 2145, 2145,
 2145, 2145, 2145, 2145, 2145,    0, 2145, 2145,    0,    0,
    0, 2145, 2145, 2145, 2145, 2145, 2145,    0,    0, 2145,
 2145, 2145, 2145, 2145, 2145, 2145, 2145, 2145, 2143,    0,
    0,    0,    0, 2145,    0, 2145, 2145, 2145, 2145, 2145,
 2145,    0, 2145,    0,    0,    0,    0, 1424, 1423, 1425,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 2145,
 2145,    0, 2145, 2145, 2145, 2145, 2145, 2145, 2145, 2145,
 2145,    0, 2145, 2145, 2145, 2145, 2145, 2145, 2145, 2145,
 2145, 2145, 2145, 2145,    0, 2145, 2145, 2145,    0, 2145,
    0,    0,    0, 2145, 2145, 2145, 2145, 2145, 2145, 2145,
 2145, 2145, 2145, 2145, 2145, 2145, 2146, 2145, 2145,    0,
    0,    0,    0,    0, 2145, 2145, 2145, 2145, 2145, 2145,
 2145, 2145, 2145, 2145, 2145, 2145, 2145, 2145, 2145, 2145,
 2145, 2145,    0,    0, 2145,    0, 2145, 2145, 2145, 2145,
 2145, 2145,    0,    0,    0,    0, 1910, 2146, 2146, 2146,
 2146, 2146, 2146, 2146,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 2146, 2146, 2146, 2146, 2146,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 2146,    0, 2146,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1408,
    0,    0,    0,    0,    0,    0,    0,    0, 1409,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1410,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1411,    0,    0,    0,    0,
    0,    0, 1412,    0,    0,    0,    0,    0,    0,    0,
 1413,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1414,    0, 1415,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1416,
    0, 1417, 1418, 1419,    0,    0,    0,    0, 1420,    0,
    0,    0,    0, 2146,    0, 2146,    0,    0, 2146,    0,
 2146, 2146, 2146,    0, 2146, 2146,    0, 2146,    0,    0,
    0,    0,    0, 2146,    0, 2146, 2146, 2146, 2146,    0,
 2146,    0,    0, 2146,    0,    0, 2146,    0,    0, 1421,
    0,    0,    0, 2146,    0, 2146, 2146,    0, 2146, 2146,
 2146, 2146, 2146,    0,    0, 2146,    0,    0, 2146,    0,
 2146, 2146,    0,    0, 2146, 2146, 2146,    0, 2146, 2146,
 2146, 2146, 2146, 2146, 2146,    0,    0,    0, 2146, 2146,
 2146, 2146, 2146,    0,    0,    0, 2146,    0,    0,    0,
    0,    0,    0,    0, 2146, 2146, 2146, 2146, 2146, 2146,
    0,    0,    0, 2146, 2146,    0, 2146,    0, 2146,    0,
 2146, 2146, 2146,    0, 2146,    0, 2146, 2146, 2146,    0,
 2146, 2146, 2146, 2146, 2146,    0, 2146, 2146,    0,    0,
    0,    0, 2146,    0, 2146, 2146,    0,    0, 2146, 2146,
 2146,    0,    0, 2146,    0, 2146,    0,    0, 2146, 2146,
 2146, 2146,    0, 2146,    0, 2146,    0, 2146, 2146, 2146,
 2146, 2146,    0,    0, 2146, 2146, 2146,    0, 2146, 2146,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 2146,
 2146, 2146,    0, 2146, 2146, 2146,    0,    0,    0,    0,
    0, 2146,    0,    0,    0, 2146,    0, 2146, 2146, 2146,
 2146,    0, 2146,    0, 2146, 2146, 2146, 2146, 2146, 2146,
 2146, 2146,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 2146, 2146,    0, 2146, 2146, 2146,    0,    0, 1422,
    0, 2146, 2146, 2146, 2146, 2146,    0, 2146,    0,    0,
 2146, 2146,    0,    0,    0, 2146, 2146,    0,    0, 2146,
 2146, 2146, 2146, 2146, 2146, 2146, 2146, 2146,    0, 2146,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 2146,
 2146, 2146, 2146, 2146, 2146, 2146, 2146, 2146, 2146, 2146,
 2146, 2146, 2146, 2146, 2146, 2146, 2146, 2146, 2146, 2146,
    0, 2146, 2146, 2146, 2146, 2146, 2146,    0, 2146, 2146,
 2146, 2146,    0, 2146,    0, 2146, 2146, 2146, 2146, 2146,
 2146, 2146, 2146, 2146, 2146, 2146, 2146, 2146, 2146, 2146,
 2146, 2146, 2146,    0, 2146, 2146, 2146, 2146,    0, 2146,
 2146, 2146, 2146, 2146,    0, 2146, 2146, 2146, 2146, 2146,
 2146, 2146, 2146, 2146, 2146, 2146, 2146, 2146, 2146, 2146,
 2146, 2146, 2146, 2146, 2146, 2146, 2146, 2146, 2146, 2146,
 2146, 2146, 2146, 2146,    0, 2146, 2146,    0,    0,    0,
 2146, 2146, 2146, 2146, 2146, 2146,    0,    0, 2146, 2146,
 2146, 2146, 2146, 2146, 2146, 2146, 2146, 2153,    0,    0,
    0,    0, 2146,    0, 2146, 2146, 2146, 2146, 2146, 2146,
    0, 2146,    0,    0,    0,    0, 1424, 1423, 1425,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2146, 2146,
    0, 2146, 2146, 2146, 2146, 2146, 2146, 2146, 2146, 2146,
    0, 2146, 2146, 2146, 2146, 2146, 2146, 2146, 2146, 2146,
 2146, 2146, 2146,    0, 2146, 2146, 2146,    0, 2146,    0,
    0,    0, 2146, 2146, 2146, 2146, 2146, 2146, 2146, 2146,
 2146, 2146, 2146, 2146, 2146, 2147, 2146, 2146,    0,    0,
    0,    0,    0, 2146, 2146, 2146, 2146, 2146, 2146, 2146,
 2146, 2146, 2146, 2146, 2146, 2146, 2146, 2146, 2146, 2146,
 2146,    0,    0, 2146,    0, 2146, 2146, 2146, 2146, 2146,
 2146,    0,    0,    0,    0, 1911, 2147, 2147, 2147, 2147,
 2147, 2147, 2147,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 2147, 2147, 2147, 2147, 2147,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2147,    0, 2147,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1408,    0,
    0,    0,    0,    0,    0,    0,    0, 1409,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1410,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1411,    0,    0,    0,    0,    0,
    0, 1412,    0,    0,    0,    0,    0,    0,    0, 1413,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1414,    0, 1415,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1416,    0,
 1417, 1418, 1419,    0,    0,    0,    0, 1420,    0,    0,
    0,    0, 2147,    0, 2147,    0,    0, 2147,    0, 2147,
 2147, 2147,    0, 2147, 2147,    0, 2147,    0,    0,    0,
    0,    0, 2147,    0, 2147, 2147, 2147, 2147,    0, 2147,
    0,    0, 2147,    0,    0, 2147,    0,    0, 1421,    0,
    0,    0, 2147,    0, 2147, 2147,    0, 2147, 2147, 2147,
 2147, 2147,    0,    0, 2147,    0,    0, 2147,    0, 2147,
 2147,    0,    0, 2147, 2147, 2147,    0, 2147, 2147, 2147,
 2147, 2147, 2147, 2147,    0,    0,    0, 2147, 2147, 2147,
 2147, 2147,    0,    0,    0, 2147,    0,    0,    0,    0,
    0,    0,    0, 2147, 2147, 2147, 2147, 2147, 2147,    0,
    0,    0, 2147, 2147,    0, 2147,    0, 2147,    0, 2147,
 2147, 2147,    0, 2147,    0, 2147, 2147, 2147,    0, 2147,
 2147, 2147, 2147, 2147,    0, 2147, 2147,    0,    0,    0,
    0, 2147,    0, 2147, 2147,    0,    0, 2147, 2147, 2147,
    0,    0, 2147,    0, 2147,    0,    0, 2147, 2147, 2147,
 2147,    0, 2147,    0, 2147,    0, 2147, 2147, 2147, 2147,
 2147,    0,    0, 2147, 2147, 2147,    0, 2147, 2147,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2147, 2147,
 2147,    0, 2147, 2147, 2147,    0,    0,    0,    0,    0,
 2147,    0,    0,    0, 2147,    0, 2147, 2147, 2147, 2147,
    0, 2147,    0, 2147, 2147, 2147, 2147, 2147, 2147, 2147,
 2147,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2147, 2147,    0, 2147, 2147, 2147,    0,    0, 1422,    0,
 2147, 2147, 2147, 2147, 2147,    0, 2147,    0,    0, 2147,
 2147,    0,    0,    0, 2147, 2147,    0,    0, 2147, 2147,
 2147, 2147, 2147, 2147, 2147, 2147, 2147,    0, 2147,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2147, 2147,
 2147, 2147, 2147, 2147, 2147, 2147, 2147, 2147, 2147, 2147,
 2147, 2147, 2147, 2147, 2147, 2147, 2147, 2147, 2147,    0,
 2147, 2147, 2147, 2147, 2147, 2147,    0, 2147, 2147, 2147,
 2147,    0, 2147,    0, 2147, 2147, 2147, 2147, 2147, 2147,
 2147, 2147, 2147, 2147, 2147, 2147, 2147, 2147, 2147, 2147,
 2147, 2147,    0, 2147, 2147, 2147, 2147,    0, 2147, 2147,
 2147, 2147, 2147,    0, 2147, 2147, 2147, 2147, 2147, 2147,
 2147, 2147, 2147, 2147, 2147, 2147, 2147, 2147, 2147, 2147,
 2147, 2147, 2147, 2147, 2147, 2147, 2147, 2147, 2147, 2147,
 2147, 2147, 2147,    0, 2147, 2147,    0,    0,    0, 2147,
 2147, 2147, 2147, 2147, 2147,    0,    0, 2147, 2147, 2147,
 2147, 2147, 2147, 2147, 2147, 2147, 2157,    0,    0,    0,
    0, 2147,    0, 2147, 2147, 2147, 2147, 2147, 2147,    0,
 2147,    0,    0,    0,    0, 1424, 1423, 1425,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 2147, 2147,    0,
 2147, 2147, 2147, 2147, 2147, 2147, 2147, 2147, 2147,    0,
 2147, 2147, 2147, 2147, 2147, 2147, 2147, 2147, 2147, 2147,
 2147, 2147,    0, 2147, 2147, 2147,    0, 2147,    0,    0,
    0, 2147, 2147, 2147, 2147, 2147, 2147, 2147, 2147, 2147,
 2147, 2147, 2147, 2147, 2148, 2147, 2147,    0,    0,    0,
    0,    0, 2147, 2147, 2147, 2147, 2147, 2147, 2147, 2147,
 2147, 2147, 2147, 2147, 2147, 2147, 2147, 2147, 2147, 2147,
    0,    0, 2147,    0, 2147, 2147, 2147, 2147, 2147, 2147,
    0,    0,    0,    0, 1912, 2148, 2148, 2148, 2148, 2148,
 2148, 2148,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 2148, 2148, 2148, 2148, 2148,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 2148,    0, 2148,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1408,    0,    0,
    0,    0,    0,    0,    0,    0, 1409,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1410,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1411,    0,    0,    0,    0,    0,    0,
 1412,    0,    0,    0,    0,    0,    0,    0, 1413,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1414,    0, 1415,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1416,    0, 1417,
 1418, 1419,    0,    0,    0,    0, 1420,    0,    0,    0,
    0, 2148,    0, 2148,    0,    0, 2148,    0, 2148, 2148,
 2148,    0, 2148, 2148,    0, 2148,    0,    0,    0,    0,
    0, 2148,    0, 2148, 2148, 2148, 2148,    0, 2148,    0,
    0, 2148,    0,    0, 2148,    0,    0, 1421,    0,    0,
    0, 2148,    0, 2148, 2148,    0, 2148, 2148, 2148, 2148,
 2148,    0,    0, 2148,    0,    0, 2148,    0, 2148, 2148,
    0,    0, 2148, 2148, 2148,    0, 2148, 2148, 2148, 2148,
 2148, 2148, 2148,    0,    0,    0, 2148, 2148, 2148, 2148,
 2148,    0,    0,    0, 2148,    0,    0,    0,    0,    0,
    0,    0, 2148, 2148, 2148, 2148, 2148, 2148,    0,    0,
    0, 2148, 2148,    0, 2148,    0, 2148,    0, 2148, 2148,
 2148,    0, 2148,    0, 2148, 2148, 2148,    0, 2148, 2148,
 2148, 2148, 2148,    0, 2148, 2148,    0,    0,    0,    0,
 2148,    0, 2148, 2148,    0,    0, 2148, 2148, 2148,    0,
    0, 2148,    0, 2148,    0,    0, 2148, 2148, 2148, 2148,
    0, 2148,    0, 2148,    0, 2148, 2148, 2148, 2148, 2148,
    0,    0, 2148, 2148, 2148,    0, 2148, 2148,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 2148, 2148, 2148,
    0, 2148, 2148, 2148,    0,    0,    0,    0,    0, 2148,
    0,    0,    0, 2148,    0, 2148, 2148, 2148, 2148,    0,
 2148,    0, 2148, 2148, 2148, 2148, 2148, 2148, 2148, 2148,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 2148,
 2148,    0, 2148, 2148, 2148,    0,    0, 1422,    0, 2148,
 2148, 2148, 2148, 2148,    0, 2148,    0,    0, 2148, 2148,
    0,    0,    0, 2148, 2148,    0,    0, 2148, 2148, 2148,
 2148, 2148, 2148, 2148, 2148, 2148,    0, 2148,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 2148, 2148, 2148,
 2148, 2148, 2148, 2148, 2148, 2148, 2148, 2148, 2148, 2148,
 2148, 2148, 2148, 2148, 2148, 2148, 2148, 2148,    0, 2148,
 2148, 2148, 2148, 2148, 2148,    0, 2148, 2148, 2148, 2148,
    0, 2148,    0, 2148, 2148, 2148, 2148, 2148, 2148, 2148,
 2148, 2148, 2148, 2148, 2148, 2148, 2148, 2148, 2148, 2148,
 2148,    0, 2148, 2148, 2148, 2148,    0, 2148, 2148, 2148,
 2148, 2148,    0, 2148, 2148, 2148, 2148, 2148, 2148, 2148,
 2148, 2148, 2148, 2148, 2148, 2148, 2148, 2148, 2148, 2148,
 2148, 2148, 2148, 2148, 2148, 2148, 2148, 2148, 2148, 2148,
 2148, 2148,    0, 2148, 2148,    0,    0,    0, 2148, 2148,
 2148, 2148, 2148, 2148,    0,    0, 2148, 2148, 2148, 2148,
 2148, 2148, 2148, 2148, 2148, 2159,    0,    0,    0,    0,
 2148,    0, 2148, 2148, 2148, 2148, 2148, 2148,    0, 2148,
    0,    0,    0,    0, 1424, 1423, 1425,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2148, 2148,    0, 2148,
 2148, 2148, 2148, 2148, 2148, 2148, 2148, 2148,    0, 2148,
 2148, 2148, 2148, 2148, 2148, 2148, 2148, 2148, 2148, 2148,
 2148,    0, 2148, 2148, 2148,    0, 2148,    0,    0,    0,
 2148, 2148, 2148, 2148, 2148, 2148, 2148, 2148, 2148, 2148,
 2148, 2148, 2148, 2149, 2148, 2148,    0,    0,    0,    0,
    0, 2148, 2148, 2148, 2148, 2148, 2148, 2148, 2148, 2148,
 2148, 2148, 2148, 2148, 2148, 2148, 2148, 2148, 2148,    0,
    0, 2148,    0, 2148, 2148, 2148, 2148, 2148, 2148,    0,
    0,    0,    0, 1914, 2149, 2149, 2149, 2149, 2149, 2149,
 2149,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 2149, 2149, 2149, 2149, 2149,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 2149,    0, 2149,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1408,    0,    0,    0,
    0,    0,    0,    0,    0, 1409,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1410,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1411,    0,    0,    0,    0,    0,    0, 1412,
    0,    0,    0,    0,    0,    0,    0, 1413,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1414,    0, 1415,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1416,    0, 1417, 1418,
 1419,    0,    0,    0,    0, 1420,    0,    0,    0,    0,
 2149,    0, 2149,    0,    0, 2149,    0, 2149, 2149, 2149,
    0, 2149, 2149,    0, 2149,    0,    0,    0,    0,    0,
 2149,    0, 2149, 2149, 2149, 2149,    0, 2149,    0,    0,
 2149,    0,    0, 2149,    0,    0, 1421,    0,    0,    0,
 2149,    0, 2149, 2149,    0, 2149, 2149, 2149, 2149, 2149,
    0,    0, 2149,    0,    0, 2149,    0, 2149, 2149,    0,
    0, 2149, 2149, 2149,    0, 2149, 2149, 2149, 2149, 2149,
 2149, 2149,    0,    0,    0, 2149, 2149, 2149, 2149, 2149,
    0,    0,    0, 2149,    0,    0,    0,    0,    0,    0,
    0, 2149, 2149, 2149, 2149, 2149, 2149,    0,    0,    0,
 2149, 2149,    0, 2149,    0, 2149,    0, 2149, 2149, 2149,
    0, 2149,    0, 2149, 2149, 2149,    0, 2149, 2149, 2149,
 2149, 2149,    0, 2149, 2149,    0,    0,    0,    0, 2149,
    0, 2149, 2149,    0,    0, 2149, 2149, 2149,    0,    0,
 2149,    0, 2149,    0,    0, 2149, 2149, 2149, 2149,    0,
 2149,    0, 2149,    0, 2149, 2149, 2149, 2149, 2149,    0,
    0, 2149, 2149, 2149,    0, 2149, 2149,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2149, 2149, 2149,    0,
 2149, 2149, 2149,    0,    0,    0,    0,    0, 2149,    0,
    0,    0, 2149,    0, 2149, 2149, 2149, 2149,    0, 2149,
    0, 2149, 2149, 2149, 2149, 2149, 2149, 2149, 2149,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2149, 2149,
    0, 2149, 2149, 2149,    0,    0, 1422,    0, 2149, 2149,
 2149, 2149, 2149,    0, 2149,    0,    0, 2149, 2149,    0,
    0,    0, 2149, 2149,    0,    0, 2149, 2149, 2149, 2149,
 2149, 2149, 2149, 2149, 2149,    0, 2149,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2149, 2149, 2149, 2149,
 2149, 2149, 2149, 2149, 2149, 2149, 2149, 2149, 2149, 2149,
 2149, 2149, 2149, 2149, 2149, 2149, 2149,    0, 2149, 2149,
 2149, 2149, 2149, 2149,    0, 2149, 2149, 2149, 2149,    0,
 2149,    0, 2149, 2149, 2149, 2149, 2149, 2149, 2149, 2149,
 2149, 2149, 2149, 2149, 2149, 2149, 2149, 2149, 2149, 2149,
    0, 2149, 2149, 2149, 2149,    0, 2149, 2149, 2149, 2149,
 2149,    0, 2149, 2149, 2149, 2149, 2149, 2149, 2149, 2149,
 2149, 2149, 2149, 2149, 2149, 2149, 2149, 2149, 2149, 2149,
 2149, 2149, 2149, 2149, 2149, 2149, 2149, 2149, 2149, 2149,
 2149,    0, 2149, 2149,    0,    0,    0, 2149, 2149, 2149,
 2149, 2149, 2149,    0,    0, 2149, 2149, 2149, 2149, 2149,
 2149, 2149, 2149, 2149, 2164,    0,    0,    0,    0, 2149,
    0, 2149, 2149, 2149, 2149, 2149, 2149,    0, 2149,    0,
    0,    0,    0, 1424, 1423, 1425,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 2149, 2149,    0, 2149, 2149,
 2149, 2149, 2149, 2149, 2149, 2149, 2149,    0, 2149, 2149,
 2149, 2149, 2149, 2149, 2149, 2149, 2149, 2149, 2149, 2149,
    0, 2149, 2149, 2149,    0, 2149,    0,    0,    0, 2149,
 2149, 2149, 2149, 2149, 2149, 2149, 2149, 2149, 2149, 2149,
 2149, 2149, 2150, 2149, 2149,    0,    0,    0,    0,    0,
 2149, 2149, 2149, 2149, 2149, 2149, 2149, 2149, 2149, 2149,
 2149, 2149, 2149, 2149, 2149, 2149, 2149, 2149,    0,    0,
 2149,    0, 2149, 2149, 2149, 2149, 2149, 2149,    0,    0,
    0,    0, 1915, 2150, 2150, 2150, 2150, 2150, 2150, 2150,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2150, 2150, 2150, 2150, 2150,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 2150,    0, 2150,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1408,    0,    0,    0,    0,
    0,    0,    0,    0, 1409,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1410,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1411,    0,    0,    0,    0,    0,    0, 1412,    0,
    0,    0,    0,    0,    0,    0, 1413,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1414,    0, 1415,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1416,    0, 1417, 1418, 1419,
    0,    0,    0,    0, 1420,    0,    0,    0,    0, 2150,
    0, 2150,    0,    0, 2150,    0, 2150, 2150, 2150,    0,
 2150, 2150,    0, 2150,    0,    0,    0,    0,    0, 2150,
    0, 2150, 2150, 2150, 2150,    0, 2150,    0,    0, 2150,
    0,    0, 2150,    0,    0, 1421,    0,    0,    0, 2150,
    0, 2150, 2150,    0, 2150, 2150, 2150, 2150, 2150,    0,
    0, 2150,    0,    0, 2150,    0, 2150, 2150,    0,    0,
 2150, 2150, 2150,    0, 2150, 2150, 2150, 2150, 2150, 2150,
 2150,    0,    0,    0, 2150, 2150, 2150, 2150, 2150,    0,
    0,    0, 2150,    0,    0,    0,    0,    0,    0,    0,
 2150, 2150, 2150, 2150, 2150, 2150,    0,    0,    0, 2150,
 2150,    0, 2150,    0, 2150,    0, 2150, 2150, 2150,    0,
 2150,    0, 2150, 2150, 2150,    0, 2150, 2150, 2150, 2150,
 2150,    0, 2150, 2150,    0,    0,    0,    0, 2150,    0,
 2150, 2150,    0,    0, 2150, 2150, 2150,    0,    0, 2150,
    0, 2150,    0,    0, 2150, 2150, 2150, 2150,    0, 2150,
    0, 2150,    0, 2150, 2150, 2150, 2150, 2150,    0,    0,
 2150, 2150, 2150,    0, 2150, 2150,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 2150, 2150, 2150,    0, 2150,
 2150, 2150,    0,    0,    0,    0,    0, 2150,    0,    0,
    0, 2150,    0, 2150, 2150, 2150, 2150,    0, 2150,    0,
 2150, 2150, 2150, 2150, 2150, 2150, 2150, 2150,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 2150, 2150,    0,
 2150, 2150, 2150,    0,    0, 1422,    0, 2150, 2150, 2150,
 2150, 2150,    0, 2150,    0,    0, 2150, 2150,    0,    0,
    0, 2150, 2150,    0,    0, 2150, 2150, 2150, 2150, 2150,
 2150, 2150, 2150, 2150,    0, 2150,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 2150, 2150, 2150, 2150, 2150,
 2150, 2150, 2150, 2150, 2150, 2150, 2150, 2150, 2150, 2150,
 2150, 2150, 2150, 2150, 2150, 2150,    0, 2150, 2150, 2150,
 2150, 2150, 2150,    0, 2150, 2150, 2150, 2150,    0, 2150,
    0, 2150, 2150, 2150, 2150, 2150, 2150, 2150, 2150, 2150,
 2150, 2150, 2150, 2150, 2150, 2150, 2150, 2150, 2150,    0,
 2150, 2150, 2150, 2150,    0, 2150, 2150, 2150, 2150, 2150,
    0, 2150, 2150, 2150, 2150, 2150, 2150, 2150, 2150, 2150,
 2150, 2150, 2150, 2150, 2150, 2150, 2150, 2150, 2150, 2150,
 2150, 2150, 2150, 2150, 2150, 2150, 2150, 2150, 2150, 2150,
    0, 2150, 2150,    0,    0,    0, 2150, 2150, 2150, 2150,
 2150, 2150,    0,    0, 2150, 2150, 2150, 2150, 2150, 2150,
 2150, 2150, 2150, 2165,    0,    0,    0,    0, 2150,    0,
 2150, 2150, 2150, 2150, 2150, 2150,    0, 2150,    0,    0,
    0,    0, 1424, 1423, 1425,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 2150, 2150,    0, 2150, 2150, 2150,
 2150, 2150, 2150, 2150, 2150, 2150,    0, 2150, 2150, 2150,
 2150, 2150, 2150, 2150, 2150, 2150, 2150, 2150, 2150,    0,
 2150, 2150, 2150,    0, 2150,    0,    0,    0, 2150, 2150,
 2150, 2150, 2150, 2150, 2150, 2150, 2150, 2150, 2150, 2150,
 2150, 2151, 2150, 2150,    0,    0,    0,    0,    0, 2150,
 2150, 2150, 2150, 2150, 2150, 2150, 2150, 2150, 2150, 2150,
 2150, 2150, 2150, 2150, 2150, 2150, 2150,    0,    0, 2150,
    0, 2150, 2150, 2150, 2150, 2150, 2150,    0,    0,    0,
    0, 1916, 2151, 2151, 2151, 2151, 2151, 2151, 2151,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 2151,
 2151, 2151, 2151, 2151,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 2151,    0, 2151,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1408,    0,    0,    0,    0,    0,
    0,    0,    0, 1409,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1410,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1411,    0,    0,    0,    0,    0,    0, 1412,    0,    0,
    0,    0,    0,    0,    0, 1413,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1414,    0, 1415,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1416,    0, 1417, 1418, 1419,    0,
    0,    0,    0, 1420,    0,    0,    0,    0, 2151,    0,
 2151,    0,    0, 2151,    0, 2151, 2151, 2151,    0, 2151,
 2151,    0, 2151,    0,    0,    0,    0,    0, 2151,    0,
 2151, 2151, 2151, 2151,    0, 2151,    0,    0, 2151,    0,
    0, 2151,    0,    0, 1421,    0,    0,    0, 2151,    0,
 2151, 2151,    0, 2151, 2151, 2151, 2151, 2151,    0,    0,
 2151,    0,    0, 2151,    0, 2151, 2151,    0,    0, 2151,
 2151, 2151,    0, 2151, 2151, 2151, 2151, 2151, 2151, 2151,
    0,    0,    0, 2151, 2151, 2151, 2151, 2151,    0,    0,
    0, 2151,    0,    0,    0,    0,    0,    0,    0, 2151,
 2151, 2151, 2151, 2151, 2151,    0,    0,    0, 2151, 2151,
    0, 2151,    0, 2151,    0, 2151, 2151, 2151,    0, 2151,
    0, 2151, 2151, 2151,    0, 2151, 2151, 2151, 2151, 2151,
    0, 2151, 2151,    0,    0,    0,    0, 2151,    0, 2151,
 2151,    0,    0, 2151, 2151, 2151,    0,    0, 2151,    0,
 2151,    0,    0, 2151, 2151, 2151, 2151,    0, 2151,    0,
 2151,    0, 2151, 2151, 2151, 2151, 2151,    0,    0, 2151,
 2151, 2151,    0, 2151, 2151,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 2151, 2151, 2151,    0, 2151, 2151,
 2151,    0,    0,    0,    0,    0, 2151,    0,    0,    0,
 2151,    0, 2151, 2151, 2151, 2151,    0, 2151,    0, 2151,
 2151, 2151, 2151, 2151, 2151, 2151, 2151,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2151, 2151,    0, 2151,
 2151, 2151,    0,    0, 1422,    0, 2151, 2151, 2151, 2151,
 2151,    0, 2151,    0,    0, 2151, 2151,    0,    0,    0,
 2151, 2151,    0,    0, 2151, 2151, 2151, 2151, 2151, 2151,
 2151, 2151, 2151,    0, 2151,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 2151, 2151, 2151, 2151, 2151, 2151,
 2151, 2151, 2151, 2151, 2151, 2151, 2151, 2151, 2151, 2151,
 2151, 2151, 2151, 2151, 2151,    0, 2151, 2151, 2151, 2151,
 2151, 2151,    0, 2151, 2151, 2151, 2151,    0, 2151,    0,
 2151, 2151, 2151, 2151, 2151, 2151, 2151, 2151, 2151, 2151,
 2151, 2151, 2151, 2151, 2151, 2151, 2151, 2151,    0, 2151,
 2151, 2151, 2151,    0, 2151, 2151, 2151, 2151, 2151,    0,
 2151, 2151, 2151, 2151, 2151, 2151, 2151, 2151, 2151, 2151,
 2151, 2151, 2151, 2151, 2151, 2151, 2151, 2151, 2151, 2151,
 2151, 2151, 2151, 2151, 2151, 2151, 2151, 2151, 2151,    0,
 2151, 2151,    0,    0,    0, 2151, 2151, 2151, 2151, 2151,
 2151,    0,    0, 2151, 2151, 2151, 2151, 2151, 2151, 2151,
 2151, 2151, 2166,    0,    0,    0,    0, 2151,    0, 2151,
 2151, 2151, 2151, 2151, 2151,    0, 2151,    0,    0,    0,
    0, 1424, 1423, 1425,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 2151, 2151,    0, 2151, 2151, 2151, 2151,
 2151, 2151, 2151, 2151, 2151,    0, 2151, 2151, 2151, 2151,
 2151, 2151, 2151, 2151, 2151, 2151, 2151, 2151,    0, 2151,
 2151, 2151,    0, 2151,    0,    0,    0, 2151, 2151, 2151,
 2151, 2151, 2151, 2151, 2151, 2151, 2151, 2151, 2151, 2151,
 2155, 2151, 2151,    0,    0,    0,    0,    0, 2151, 2151,
 2151, 2151, 2151, 2151, 2151, 2151, 2151, 2151, 2151, 2151,
 2151, 2151, 2151, 2151, 2151, 2151,    0,    0, 2151,    0,
 2151, 2151, 2151, 2151, 2151, 2151,    0,    0,    0,    0,
 1917, 2155, 2155, 2155, 2155, 2155, 2155, 2155,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2155, 2155,
 2155, 2155, 2155,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 2155,    0, 2155,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1408,    0,    0,    0,    0,    0,    0,
    0,    0, 1409,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1410,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1411,
    0,    0,    0,    0,    0,    0, 1412,    0,    0,    0,
    0,    0,    0,    0, 1413,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1414,    0, 1415,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1416,    0, 1417, 1418, 1419,    0,    0,
    0,    0, 1420,    0,    0,    0,    0, 2155,    0, 2155,
    0,    0, 2155,    0, 2155, 2155, 2155,    0, 2155, 2155,
    0, 2155,    0,    0,    0,    0,    0, 2155,    0, 2155,
 2155, 2155, 2155,    0, 2155,    0,    0, 2155,    0,    0,
 2155,    0,    0, 1421,    0,    0,    0, 2155,    0, 2155,
 2155,    0, 2155, 2155, 2155, 2155, 2155,    0,    0, 2155,
    0,    0, 2155,    0, 2155, 2155,    0,    0, 2155, 2155,
 2155,    0, 2155, 2155, 2155, 2155, 2155, 2155, 2155,    0,
    0,    0, 2155, 2155, 2155, 2155, 2155,    0,    0,    0,
 2155,    0,    0,    0,    0,    0,    0,    0, 2155, 2155,
 2155, 2155, 2155, 2155,    0,    0,    0, 2155, 2155,    0,
 2155,    0, 2155,    0, 2155, 2155, 2155,    0, 2155,    0,
 2155, 2155, 2155,    0, 2155, 2155, 2155, 2155, 2155,    0,
 2155, 2155,    0,    0,    0,    0, 2155,    0, 2155, 2155,
    0,    0, 2155, 2155, 2155,    0,    0, 2155,    0, 2155,
    0,    0, 2155, 2155, 2155, 2155,    0, 2155,    0, 2155,
    0, 2155, 2155, 2155, 2155, 2155,    0,    0, 2155, 2155,
 2155,    0, 2155, 2155,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 2155, 2155, 2155,    0, 2155, 2155, 2155,
    0,    0,    0,    0,    0, 2155,    0,    0,    0, 2155,
    0, 2155, 2155, 2155, 2155,    0, 2155,    0, 2155, 2155,
 2155, 2155, 2155, 2155, 2155, 2155,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 2155, 2155,    0, 2155, 2155,
 2155,    0,    0, 1422,    0, 2155, 2155, 2155, 2155, 2155,
    0, 2155,    0,    0, 2155, 2155,    0,    0,    0, 2155,
 2155,    0,    0, 2155, 2155, 2155, 2155, 2155, 2155, 2155,
 2155, 2155,    0, 2155,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 2155, 2155, 2155, 2155, 2155, 2155, 2155,
 2155, 2155, 2155, 2155, 2155, 2155, 2155, 2155, 2155, 2155,
 2155, 2155, 2155, 2155,    0, 2155, 2155, 2155, 2155, 2155,
 2155,    0, 2155, 2155, 2155, 2155,    0, 2155,    0, 2155,
 2155, 2155, 2155, 2155, 2155, 2155, 2155, 2155, 2155, 2155,
 2155, 2155, 2155, 2155, 2155, 2155, 2155,    0, 2155, 2155,
 2155, 2155,    0, 2155, 2155, 2155, 2155, 2155,    0, 2155,
 2155, 2155, 2155, 2155, 2155, 2155, 2155, 2155, 2155, 2155,
 2155, 2155, 2155, 2155, 2155, 2155, 2155, 2155, 2155, 2155,
 2155, 2155, 2155, 2155, 2155, 2155, 2155, 2155,    0, 2155,
 2155,    0,    0,    0, 2155, 2155, 2155, 2155, 2155, 2155,
    0,    0, 2155, 2155, 2155, 2155, 2155, 2155, 2155, 2155,
 2155, 2167,    0,    0,    0,    0, 2155,    0, 2155, 2155,
 2155, 2155, 2155, 2155,    0, 2155,    0,    0,    0,    0,
 1424, 1423, 1425,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 2155, 2155,    0, 2155, 2155, 2155, 2155, 2155,
 2155, 2155, 2155, 2155,    0, 2155, 2155, 2155, 2155, 2155,
 2155, 2155, 2155, 2155, 2155, 2155, 2155,    0, 2155, 2155,
 2155,    0, 2155,    0,    0,    0, 2155, 2155, 2155, 2155,
 2155, 2155, 2155, 2155, 2155, 2155, 2155, 2155, 2155, 2156,
 2155, 2155,    0,    0,    0,    0,    0, 2155, 2155, 2155,
 2155, 2155, 2155, 2155, 2155, 2155, 2155, 2155, 2155, 2155,
 2155, 2155, 2155, 2155, 2155,    0,    0, 2155,    0, 2155,
 2155, 2155, 2155, 2155, 2155,    0,    0,    0,    0, 1918,
 2156, 2156, 2156, 2156, 2156, 2156, 2156,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 2156, 2156, 2156,
 2156, 2156,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2156,    0, 2156,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1408,    0,    0,    0,    0,    0,    0,    0,
    0, 1409,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1410,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1411,    0,
    0,    0,    0,    0,    0, 1412,    0,    0,    0,    0,
    0,    0,    0, 1413,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1414,    0, 1415,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1416,    0, 1417, 1418, 1419,    0,    0,    0,
    0, 1420,    0,    0,    0,    0, 2156,    0, 2156,    0,
    0, 2156,    0, 2156, 2156, 2156,    0, 2156, 2156,    0,
 2156,    0,    0,    0,    0,    0, 2156,    0, 2156, 2156,
 2156, 2156,    0, 2156,    0,    0, 2156,    0,    0, 2156,
    0,    0, 1421,    0,    0,    0, 2156,    0, 2156, 2156,
    0, 2156, 2156, 2156, 2156, 2156,    0,    0, 2156,    0,
    0, 2156,    0, 2156, 2156,    0,    0, 2156, 2156, 2156,
    0, 2156, 2156, 2156, 2156, 2156, 2156, 2156,    0,    0,
    0, 2156, 2156, 2156, 2156, 2156,    0,    0,    0, 2156,
    0,    0,    0,    0,    0,    0,    0, 2156, 2156, 2156,
 2156, 2156, 2156,    0,    0,    0, 2156, 2156,    0, 2156,
    0, 2156,    0, 2156, 2156, 2156,    0, 2156,    0, 2156,
 2156, 2156,    0, 2156, 2156, 2156, 2156, 2156,    0, 2156,
 2156,    0,    0,    0,    0, 2156,    0, 2156, 2156,    0,
    0, 2156, 2156, 2156,    0,    0, 2156,    0, 2156,    0,
    0, 2156, 2156, 2156, 2156,    0, 2156,    0, 2156,    0,
 2156, 2156, 2156, 2156, 2156,    0,    0, 2156, 2156, 2156,
    0, 2156, 2156,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 2156, 2156, 2156,    0, 2156, 2156, 2156,    0,
    0,    0,    0,    0, 2156,    0,    0,    0, 2156,    0,
 2156, 2156, 2156, 2156,    0, 2156,    0, 2156, 2156, 2156,
 2156, 2156, 2156, 2156, 2156,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 2156, 2156,    0, 2156, 2156, 2156,
    0,    0, 1422,    0, 2156, 2156, 2156, 2156, 2156,    0,
 2156,    0,    0, 2156, 2156,    0,    0,    0, 2156, 2156,
    0,    0, 2156, 2156, 2156, 2156, 2156, 2156, 2156, 2156,
 2156,    0, 2156,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 2156, 2156, 2156, 2156, 2156, 2156, 2156, 2156,
 2156, 2156, 2156, 2156, 2156, 2156, 2156, 2156, 2156, 2156,
 2156, 2156, 2156,    0, 2156, 2156, 2156, 2156, 2156, 2156,
    0, 2156, 2156, 2156, 2156,    0, 2156,    0, 2156, 2156,
 2156, 2156, 2156, 2156, 2156, 2156, 2156, 2156, 2156, 2156,
 2156, 2156, 2156, 2156, 2156, 2156,    0, 2156, 2156, 2156,
 2156,    0, 2156, 2156, 2156, 2156, 2156,    0, 2156, 2156,
 2156, 2156, 2156, 2156, 2156, 2156, 2156, 2156, 2156, 2156,
 2156, 2156, 2156, 2156, 2156, 2156, 2156, 2156, 2156, 2156,
 2156, 2156, 2156, 2156, 2156, 2156, 2156,    0, 2156, 2156,
    0,    0,    0, 2156, 2156, 2156, 2156, 2156, 2156,    0,
    0, 2156, 2156, 2156, 2156, 2156, 2156, 2156, 2156, 2156,
 2548,    0,    0,    0,    0, 2156,    0, 2156, 2156, 2156,
 2156, 2156, 2156,    0, 2156,    0,    0,    0,    0, 1424,
 1423, 1425,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 2156, 2156,    0, 2156, 2156, 2156, 2156, 2156, 2156,
 2156, 2156, 2156,    0, 2156, 2156, 2156, 2156, 2156, 2156,
 2156, 2156, 2156, 2156, 2156, 2156,    0, 2156, 2156, 2156,
    0, 2156,    0,    0,    0, 2156, 2156, 2156, 2156, 2156,
 2156, 2156, 2156, 2156, 2156, 2156, 2156, 2156, 2157, 2156,
 2156,    0,    0,    0,    0,    0, 2156, 2156, 2156, 2156,
 2156, 2156, 2156, 2156, 2156, 2156, 2156, 2156, 2156, 2156,
 2156, 2156, 2156, 2156,    0,    0, 2156,    0, 2156, 2156,
 2156, 2156, 2156, 2156,    0,    0,    0,    0, 1919, 2157,
 2157, 2157, 2157, 2157, 2157, 2157,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2157, 2157, 2157, 2157,
 2157,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 2157,
    0, 2157,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1408,    0,    0,    0,    0,    0,    0,    0,    0,
 1409,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1410,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1411,    0,    0,
    0,    0,    0,    0, 1412,    0,    0,    0,    0,    0,
    0,    0, 1413,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1414,
    0, 1415,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1416,    0, 1417, 1418, 1419,    0,    0,    0,    0,
 1420,    0,    0,    0,    0, 2157,    0, 2157,    0,    0,
 2157,    0, 2157, 2157, 2157,    0, 2157, 2157,    0, 2157,
    0,    0,    0,    0,    0, 2157,    0, 2157, 2157, 2157,
 2157,    0, 2157,    0,    0, 2157,    0,    0, 2157,    0,
    0, 1421,    0,    0,    0, 2157,    0, 2157, 2157,    0,
 2157, 2157, 2157, 2157, 2157,    0,    0, 2157,    0,    0,
 2157,    0, 2157, 2157,    0,    0, 2157, 2157, 2157,    0,
 2157, 2157, 2157, 2157, 2157, 2157, 2157,    0,    0,    0,
 2157, 2157, 2157, 2157, 2157,    0,    0,    0, 2157,    0,
    0,    0,    0,    0,    0,    0, 2157, 2157, 2157, 2157,
 2157, 2157,    0,    0,    0, 2157, 2157,    0, 2157,    0,
 2157,    0, 2157, 2157, 2157,    0, 2157,    0, 2157, 2157,
 2157,    0, 2157, 2157, 2157, 2157, 2157,    0, 2157, 2157,
    0,    0,    0,    0, 2157,    0, 2157, 2157,    0,    0,
 2157, 2157, 2157,    0,    0, 2157,    0, 2157,    0,    0,
 2157, 2157, 2157, 2157,    0, 2157,    0, 2157,    0, 2157,
 2157, 2157, 2157, 2157,    0,    0, 2157, 2157, 2157,    0,
 2157, 2157,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 2157, 2157, 2157,    0, 2157, 2157, 2157,    0,    0,
    0,    0,    0, 2157,    0,    0,    0, 2157,    0, 2157,
 2157, 2157, 2157,    0, 2157,    0, 2157, 2157, 2157, 2157,
 2157, 2157, 2157, 2157,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 2157, 2157,    0, 2157, 2157, 2157,    0,
    0, 1422,    0, 2157, 2157, 2157, 2157, 2157,    0, 2157,
    0,    0, 2157, 2157,    0,    0,    0, 2157, 2157,    0,
    0, 2157, 2157, 2157, 2157, 2157, 2157, 2157, 2157, 2157,
    0, 2157,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 2157, 2157, 2157, 2157, 2157, 2157, 2157, 2157, 2157,
 2157, 2157, 2157, 2157, 2157, 2157, 2157, 2157, 2157, 2157,
 2157, 2157,    0, 2157, 2157, 2157, 2157, 2157, 2157,    0,
 2157, 2157, 2157, 2157,    0, 2157,    0, 2157, 2157, 2157,
 2157, 2157, 2157, 2157, 2157, 2157, 2157, 2157, 2157, 2157,
 2157, 2157, 2157, 2157, 2157,    0, 2157, 2157, 2157, 2157,
    0, 2157, 2157, 2157, 2157, 2157,    0, 2157, 2157, 2157,
 2157, 2157, 2157, 2157, 2157, 2157, 2157, 2157, 2157, 2157,
 2157, 2157, 2157, 2157, 2157, 2157, 2157, 2157, 2157, 2157,
 2157, 2157, 2157, 2157, 2157, 2157,    0, 2157, 2157,    0,
    0,    0, 2157, 2157, 2157, 2157, 2157, 2157,    0,    0,
 2157, 2157, 2157, 2157, 2157, 2157, 2157, 2157, 2157, 2549,
    0,    0,    0,    0, 2157,    0, 2157, 2157, 2157, 2157,
 2157, 2157,    0, 2157,    0,    0,    0,    0, 1424, 1423,
 1425,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2157, 2157,    0, 2157, 2157, 2157, 2157, 2157, 2157, 2157,
 2157, 2157,    0, 2157, 2157, 2157, 2157, 2157, 2157, 2157,
 2157, 2157, 2157, 2157, 2157,    0, 2157, 2157, 2157,    0,
 2157,    0,    0,    0, 2157, 2157, 2157, 2157, 2157, 2157,
 2157, 2157, 2157, 2157, 2157, 2157, 2157, 2161, 2157, 2157,
    0,    0,    0,    0,    0, 2157, 2157, 2157, 2157, 2157,
 2157, 2157, 2157, 2157, 2157, 2157, 2157, 2157, 2157, 2157,
 2157, 2157, 2157,    0,    0, 2157,    0, 2157, 2157, 2157,
 2157, 2157, 2157,    0,    0,    0,    0, 1923, 2161, 2161,
 2161, 2161, 2161, 2161, 2161,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 2161, 2161, 2161, 2161, 2161,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2161,    0,
 2161,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1408,    0,    0,    0,    0,    0,    0,    0,    0, 1409,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1410,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1411,    0,    0,    0,
    0,    0,    0, 1412,    0,    0,    0,    0,    0,    0,
    0, 1413,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1414,    0,
 1415,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1416,    0, 1417, 1418, 1419,    0,    0,    0,    0, 1420,
    0,    0,    0,    0, 2161,    0, 2161,    0,    0, 2161,
    0, 2161, 2161, 2161,    0, 2161, 2161,    0, 2161,    0,
    0,    0,    0,    0, 2161,    0, 2161, 2161, 2161, 2161,
    0, 2161,    0,    0, 2161,    0,    0, 2161,    0,    0,
 1421,    0,    0,    0, 2161,    0, 2161, 2161,    0, 2161,
 2161, 2161, 2161, 2161,    0,    0, 2161,    0,    0, 2161,
    0, 2161, 2161,    0,    0, 2161, 2161, 2161,    0, 2161,
 2161, 2161, 2161, 2161, 2161, 2161,    0,    0,    0, 2161,
 2161, 2161, 2161, 2161,    0,    0,    0, 2161,    0,    0,
    0,    0,    0,    0,    0, 2161, 2161, 2161, 2161, 2161,
 2161,    0,    0,    0, 2161, 2161,    0, 2161,    0, 2161,
    0, 2161, 2161, 2161,    0, 2161,    0, 2161, 2161, 2161,
    0, 2161, 2161, 2161, 2161, 2161,    0, 2161, 2161,    0,
    0,    0,    0, 2161,    0, 2161, 2161,    0,    0, 2161,
 2161, 2161,    0,    0, 2161,    0, 2161,    0,    0, 2161,
 2161, 2161, 2161,    0, 2161,    0, 2161,    0, 2161, 2161,
 2161, 2161, 2161,    0,    0, 2161, 2161, 2161,    0, 2161,
 2161,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2161, 2161, 2161,    0, 2161, 2161, 2161,    0,    0,    0,
    0,    0, 2161,    0,    0,    0, 2161,    0, 2161, 2161,
 2161, 2161,    0, 2161,    0, 2161, 2161, 2161, 2161, 2161,
 2161, 2161, 2161,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 2161, 2161,    0, 2161, 2161, 2161,    0,    0,
 1422,    0, 2161, 2161, 2161, 2161, 2161,    0, 2161,    0,
    0, 2161, 2161,    0,    0,    0, 2161, 2161,    0,    0,
 2161, 2161, 2161, 2161, 2161, 2161, 2161, 2161, 2161,    0,
 2161,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2161, 2161, 2161, 2161, 2161, 2161, 2161, 2161, 2161, 2161,
 2161, 2161, 2161, 2161, 2161, 2161, 2161, 2161, 2161, 2161,
 2161,    0, 2161, 2161, 2161, 2161, 2161, 2161,    0, 2161,
 2161, 2161, 2161,    0, 2161,    0, 2161, 2161, 2161, 2161,
 2161, 2161, 2161, 2161, 2161, 2161, 2161, 2161, 2161, 2161,
 2161, 2161, 2161, 2161,    0, 2161, 2161, 2161, 2161,    0,
 2161, 2161, 2161, 2161, 2161,    0, 2161, 2161, 2161, 2161,
 2161, 2161, 2161, 2161, 2161, 2161, 2161, 2161, 2161, 2161,
 2161, 2161, 2161, 2161, 2161, 2161, 2161, 2161, 2161, 2161,
 2161, 2161, 2161, 2161, 2161,    0, 2161, 2161,    0,    0,
    0, 2161, 2161, 2161, 2161, 2161, 2161,    0,    0, 2161,
 2161, 2161, 2161, 2161, 2161, 2161, 2161, 2161, 2554,    0,
    0,    0,    0, 2161,    0, 2161, 2161, 2161, 2161, 2161,
 2161,    0, 2161,    0,    0,    0,    0, 1424, 1423, 1425,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 2161,
 2161,    0, 2161, 2161, 2161, 2161, 2161, 2161, 2161, 2161,
 2161,    0, 2161, 2161, 2161, 2161, 2161, 2161, 2161, 2161,
 2161, 2161, 2161, 2161,    0, 2161, 2161, 2161,    0, 2161,
    0,    0,    0, 2161, 2161, 2161, 2161, 2161, 2161, 2161,
 2161, 2161, 2161, 2161, 2161, 2161, 2162, 2161, 2161,    0,
    0,    0,    0,    0, 2161, 2161, 2161, 2161, 2161, 2161,
 2161, 2161, 2161, 2161, 2161, 2161, 2161, 2161, 2161, 2161,
 2161, 2161,    0,    0, 2161,    0, 2161, 2161, 2161, 2161,
 2161, 2161,    0,    0,    0,    0, 1924, 2162, 2162, 2162,
 2162, 2162, 2162, 2162,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 2162, 2162, 2162, 2162, 2162,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 2162,    0, 2162,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1408,
    0,    0,    0,    0,    0,    0,    0,    0, 1409,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1410,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1411,    0,    0,    0,    0,
    0,    0, 1412,    0,    0,    0,    0,    0,    0,    0,
 1413,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1414,    0, 1415,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1416,
    0, 1417, 1418, 1419,    0,    0,    0,    0, 1420,    0,
    0,    0,    0, 2162,    0, 2162,    0,    0, 2162,    0,
 2162, 2162, 2162,    0, 2162, 2162,    0, 2162,    0,    0,
    0,    0,    0, 2162,    0, 2162, 2162, 2162, 2162,    0,
 2162,    0,    0, 2162,    0,    0, 2162,    0,    0, 1421,
    0,    0,    0, 2162,    0, 2162, 2162,    0, 2162, 2162,
 2162, 2162, 2162,    0,    0, 2162,    0,    0, 2162,    0,
 2162, 2162,    0,    0, 2162, 2162, 2162,    0, 2162, 2162,
 2162, 2162, 2162, 2162, 2162,    0,    0,    0, 2162, 2162,
 2162, 2162, 2162,    0,    0,    0, 2162,    0,    0,    0,
    0,    0,    0,    0, 2162, 2162, 2162, 2162, 2162, 2162,
    0,    0,    0, 2162, 2162,    0, 2162,    0, 2162,    0,
 2162, 2162, 2162,    0, 2162,    0, 2162, 2162, 2162,    0,
 2162, 2162, 2162, 2162, 2162,    0, 2162, 2162,    0,    0,
    0,    0, 2162,    0, 2162, 2162,    0,    0, 2162, 2162,
 2162,    0,    0, 2162,    0, 2162,    0,    0, 2162, 2162,
 2162, 2162,    0, 2162,    0, 2162,    0, 2162, 2162, 2162,
 2162, 2162,    0,    0, 2162, 2162, 2162,    0, 2162, 2162,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 2162,
 2162, 2162,    0, 2162, 2162, 2162,    0,    0,    0,    0,
    0, 2162,    0,    0,    0, 2162,    0, 2162, 2162, 2162,
 2162,    0, 2162,    0, 2162, 2162, 2162, 2162, 2162, 2162,
 2162, 2162,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 2162, 2162,    0, 2162, 2162, 2162,    0,    0, 1422,
    0, 2162, 2162, 2162, 2162, 2162,    0, 2162,    0,    0,
 2162, 2162,    0,    0,    0, 2162, 2162,    0,    0, 2162,
 2162, 2162, 2162, 2162, 2162, 2162, 2162, 2162,    0, 2162,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 2162,
 2162, 2162, 2162, 2162, 2162, 2162, 2162, 2162, 2162, 2162,
 2162, 2162, 2162, 2162, 2162, 2162, 2162, 2162, 2162, 2162,
    0, 2162, 2162, 2162, 2162, 2162, 2162,    0, 2162, 2162,
 2162, 2162,    0, 2162,    0, 2162, 2162, 2162, 2162, 2162,
 2162, 2162, 2162, 2162, 2162, 2162, 2162, 2162, 2162, 2162,
 2162, 2162, 2162,    0, 2162, 2162, 2162, 2162,    0, 2162,
 2162, 2162, 2162, 2162,    0, 2162, 2162, 2162, 2162, 2162,
 2162, 2162, 2162, 2162, 2162, 2162, 2162, 2162, 2162, 2162,
 2162, 2162, 2162, 2162, 2162, 2162, 2162, 2162, 2162, 2162,
 2162, 2162, 2162, 2162,    0, 2162, 2162,    0,    0,    0,
 2162, 2162, 2162, 2162, 2162, 2162,    0,    0, 2162, 2162,
 2162, 2162, 2162, 2162, 2162, 2162, 2162, 2555,    0,    0,
    0,    0, 2162,    0, 2162, 2162, 2162, 2162, 2162, 2162,
    0, 2162,    0,    0,    0,    0, 1424, 1423, 1425,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2162, 2162,
    0, 2162, 2162, 2162, 2162, 2162, 2162, 2162, 2162, 2162,
    0, 2162, 2162, 2162, 2162, 2162, 2162, 2162, 2162, 2162,
 2162, 2162, 2162,    0, 2162, 2162, 2162,    0, 2162,    0,
    0,    0, 2162, 2162, 2162, 2162, 2162, 2162, 2162, 2162,
 2162, 2162, 2162, 2162, 2162, 2163, 2162, 2162,    0,    0,
    0,    0,    0, 2162, 2162, 2162, 2162, 2162, 2162, 2162,
 2162, 2162, 2162, 2162, 2162, 2162, 2162, 2162, 2162, 2162,
 2162,    0,    0, 2162,    0, 2162, 2162, 2162, 2162, 2162,
 2162,    0,    0,    0,    0, 1925, 2163, 2163, 2163, 2163,
 2163, 2163, 2163,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 2163, 2163, 2163, 2163, 2163,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2163,    0, 2163,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1408,    0,
    0,    0,    0,    0,    0,    0,    0, 1409,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1410,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1411,    0,    0,    0,    0,    0,
    0, 1412,    0,    0,    0,    0,    0,    0,    0, 1413,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1414,    0, 1415,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1416,    0,
 1417, 1418, 1419,    0,    0,    0,    0, 1420,    0,    0,
    0,    0, 2163,    0, 2163,    0,    0, 2163,    0, 2163,
 2163, 2163,    0, 2163, 2163,    0, 2163,    0,    0,    0,
    0,    0, 2163,    0, 2163, 2163, 2163, 2163,    0, 2163,
    0,    0, 2163,    0,    0, 2163,    0,    0, 1421,    0,
    0,    0, 2163,    0, 2163, 2163,    0, 2163, 2163, 2163,
 2163, 2163,    0,    0, 2163,    0,    0, 2163,    0, 2163,
 2163,    0,    0, 2163, 2163, 2163,    0, 2163, 2163, 2163,
 2163, 2163, 2163, 2163,    0,    0,    0, 2163, 2163, 2163,
 2163, 2163,    0,    0,    0, 2163,    0,    0,    0,    0,
    0,    0,    0, 2163, 2163, 2163, 2163, 2163, 2163,    0,
    0,    0, 2163, 2163,    0, 2163,    0, 2163,    0, 2163,
 2163, 2163,    0, 2163,    0, 2163, 2163, 2163,    0, 2163,
 2163, 2163, 2163, 2163,    0, 2163, 2163,    0,    0,    0,
    0, 2163,    0, 2163, 2163,    0,    0, 2163, 2163, 2163,
    0,    0, 2163,    0, 2163,    0,    0, 2163, 2163, 2163,
 2163,    0, 2163,    0, 2163,    0, 2163, 2163, 2163, 2163,
 2163,    0,    0, 2163, 2163, 2163,    0, 2163, 2163,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2163, 2163,
 2163,    0, 2163, 2163, 2163,    0,    0,    0,    0,    0,
 2163,    0,    0,    0, 2163,    0, 2163, 2163, 2163, 2163,
    0, 2163,    0, 2163, 2163, 2163, 2163, 2163, 2163, 2163,
 2163,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2163, 2163,    0, 2163, 2163, 2163,    0,    0, 1422,    0,
 2163, 2163, 2163, 2163, 2163,    0, 2163,    0,    0, 2163,
 2163,    0,    0,    0, 2163, 2163,    0,    0, 2163, 2163,
 2163, 2163, 2163, 2163, 2163, 2163, 2163,    0, 2163,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2163, 2163,
 2163, 2163, 2163, 2163, 2163, 2163, 2163, 2163, 2163, 2163,
 2163, 2163, 2163, 2163, 2163, 2163, 2163, 2163, 2163,    0,
 2163, 2163, 2163, 2163, 2163, 2163,    0, 2163, 2163, 2163,
 2163,    0, 2163,    0, 2163, 2163, 2163, 2163, 2163, 2163,
 2163, 2163, 2163, 2163, 2163, 2163, 2163, 2163, 2163, 2163,
 2163, 2163,    0, 2163, 2163, 2163, 2163,    0, 2163, 2163,
 2163, 2163, 2163,    0, 2163, 2163, 2163, 2163, 2163, 2163,
 2163, 2163, 2163, 2163, 2163, 2163, 2163, 2163, 2163, 2163,
 2163, 2163, 2163, 2163, 2163, 2163, 2163, 2163, 2163, 2163,
 2163, 2163, 2163,    0, 2163, 2163,    0,    0,    0, 2163,
 2163, 2163, 2163, 2163, 2163,    0,    0, 2163, 2163, 2163,
 2163, 2163, 2163, 2163, 2163, 2163, 2557,    0,    0,    0,
    0, 2163,    0, 2163, 2163, 2163, 2163, 2163, 2163,    0,
 2163,    0,    0,    0,    0, 1424, 1423, 1425,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 2163, 2163,    0,
 2163, 2163, 2163, 2163, 2163, 2163, 2163, 2163, 2163,    0,
 2163, 2163, 2163, 2163, 2163, 2163, 2163, 2163, 2163, 2163,
 2163, 2163,    0, 2163, 2163, 2163,    0, 2163,    0,    0,
    0, 2163, 2163, 2163, 2163, 2163, 2163, 2163, 2163, 2163,
 2163, 2163, 2163, 2163, 2164, 2163, 2163,    0,    0,    0,
    0,    0, 2163, 2163, 2163, 2163, 2163, 2163, 2163, 2163,
 2163, 2163, 2163, 2163, 2163, 2163, 2163, 2163, 2163, 2163,
    0,    0, 2163,    0, 2163, 2163, 2163, 2163, 2163, 2163,
    0,    0,    0,    0, 1926, 2164, 2164, 2164, 2164, 2164,
 2164, 2164,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 2164, 2164, 2164, 2164, 2164,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 2164,    0, 2164,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1408,    0,    0,
    0,    0,    0,    0,    0,    0, 1409,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1410,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1411,    0,    0,    0,    0,    0,    0,
 1412,    0,    0,    0,    0,    0,    0,    0, 1413,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1414,    0, 1415,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1416,    0, 1417,
 1418, 1419,    0,    0,    0,    0, 1420,    0,    0,    0,
    0, 2164,    0, 2164,    0,    0, 2164,    0, 2164, 2164,
 2164,    0, 2164, 2164,    0, 2164,    0,    0,    0,    0,
    0, 2164,    0, 2164, 2164, 2164, 2164,    0, 2164,    0,
    0, 2164,    0,    0, 2164,    0,    0, 1421,    0,    0,
    0, 2164,    0, 2164, 2164,    0, 2164, 2164, 2164, 2164,
 2164,    0,    0, 2164,    0,    0, 2164,    0, 2164, 2164,
    0,    0, 2164, 2164, 2164,    0, 2164, 2164, 2164, 2164,
 2164, 2164, 2164,    0,    0,    0, 2164, 2164, 2164, 2164,
 2164,    0,    0,    0, 2164,    0,    0,    0,    0,    0,
    0,    0, 2164, 2164, 2164, 2164, 2164, 2164,    0,    0,
    0, 2164, 2164,    0, 2164,    0, 2164,    0, 2164, 2164,
 2164,    0, 2164,    0, 2164, 2164, 2164,    0, 2164, 2164,
 2164, 2164, 2164,    0, 2164, 2164,    0,    0,    0,    0,
 2164,    0, 2164, 2164,    0,    0, 2164, 2164, 2164,    0,
    0, 2164,    0, 2164,    0,    0, 2164, 2164, 2164, 2164,
    0, 2164,    0, 2164,    0, 2164, 2164, 2164, 2164, 2164,
    0,    0, 2164, 2164, 2164,    0, 2164, 2164,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 2164, 2164, 2164,
    0, 2164, 2164, 2164,    0,    0,    0,    0,    0, 2164,
    0,    0,    0, 2164,    0, 2164, 2164, 2164, 2164,    0,
 2164,    0, 2164, 2164, 2164, 2164, 2164, 2164, 2164, 2164,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 2164,
 2164,    0, 2164, 2164, 2164,    0,    0, 1422,    0, 2164,
 2164, 2164, 2164, 2164,    0, 2164,    0,    0, 2164, 2164,
    0,    0,    0, 2164, 2164,    0,    0, 2164, 2164, 2164,
 2164, 2164, 2164, 2164, 2164, 2164,    0, 2164,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 2164, 2164, 2164,
 2164, 2164, 2164, 2164, 2164, 2164, 2164, 2164, 2164, 2164,
 2164, 2164, 2164, 2164, 2164, 2164, 2164, 2164,    0, 2164,
 2164, 2164, 2164, 2164, 2164,    0, 2164, 2164, 2164, 2164,
    0, 2164,    0, 2164, 2164, 2164, 2164, 2164, 2164, 2164,
 2164, 2164, 2164, 2164, 2164, 2164, 2164, 2164, 2164, 2164,
 2164,    0, 2164, 2164, 2164, 2164,    0, 2164, 2164, 2164,
 2164, 2164,    0, 2164, 2164, 2164, 2164, 2164, 2164, 2164,
 2164, 2164, 2164, 2164, 2164, 2164, 2164, 2164, 2164, 2164,
 2164, 2164, 2164, 2164, 2164, 2164, 2164, 2164, 2164, 2164,
 2164, 2164,    0, 2164, 2164,    0,    0,    0, 2164, 2164,
 2164, 2164, 2164, 2164,    0,    0, 2164, 2164, 2164, 2164,
 2164, 2164, 2164, 2164, 2164, 2558,    0,    0,    0,    0,
 2164,    0, 2164, 2164, 2164, 2164, 2164, 2164,    0, 2164,
    0,    0,    0,    0, 1424, 1423, 1425,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2164, 2164,    0, 2164,
 2164, 2164, 2164, 2164, 2164, 2164, 2164, 2164,    0, 2164,
 2164, 2164, 2164, 2164, 2164, 2164, 2164, 2164, 2164, 2164,
 2164,    0, 2164, 2164, 2164,    0, 2164,    0,    0,    0,
 2164, 2164, 2164, 2164, 2164, 2164, 2164, 2164, 2164, 2164,
 2164, 2164, 2164, 2167, 2164, 2164,    0,    0,    0,    0,
    0, 2164, 2164, 2164, 2164, 2164, 2164, 2164, 2164, 2164,
 2164, 2164, 2164, 2164, 2164, 2164, 2164, 2164, 2164,    0,
    0, 2164,    0, 2164, 2164, 2164, 2164, 2164, 2164,    0,
    0,    0,    0, 1928, 2167, 2167, 2167, 2167, 2167, 2167,
 2167,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 2167, 2167, 2167, 2167, 2167,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 2167,    0, 2167,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1408,    0,    0,    0,
    0,    0,    0,    0,    0, 1409,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1410,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1411,    0,    0,    0,    0,    0,    0, 1412,
    0,    0,    0,    0,    0,    0,    0, 1413,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1414,    0, 1415,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1416,    0, 1417, 1418,
 1419,    0,    0,    0,    0, 1420,    0,    0,    0,    0,
 2167,    0, 2167,    0,    0, 2167,    0, 2167, 2167, 2167,
    0, 2167, 2167,    0, 2167,    0,    0,    0,    0,    0,
 2167,    0, 2167, 2167, 2167, 2167,    0, 2167,    0,    0,
 2167,    0,    0, 2167,    0,    0, 1421,    0,    0,    0,
 2167,    0, 2167, 2167,    0, 2167, 2167, 2167, 2167, 2167,
    0,    0, 2167,    0,    0, 2167,    0, 2167, 2167,    0,
    0, 2167, 2167, 2167,    0, 2167, 2167, 2167, 2167, 2167,
 2167, 2167,    0,    0,    0, 2167, 2167, 2167, 2167, 2167,
    0,    0,    0, 2167,    0,    0,    0,    0,    0,    0,
    0, 2167, 2167, 2167, 2167, 2167, 2167,    0,    0,    0,
 2167, 2167,    0, 2167,    0, 2167,    0, 2167, 2167, 2167,
    0, 2167,    0, 2167, 2167, 2167,    0, 2167, 2167, 2167,
 2167, 2167,    0, 2167, 2167,    0,    0,    0,    0, 2167,
    0, 2167, 2167,    0,    0, 2167, 2167, 2167,    0,    0,
 2167,    0, 2167,    0,    0, 2167, 2167, 2167, 2167,    0,
 2167,    0, 2167,    0, 2167, 2167, 2167, 2167, 2167,    0,
    0, 2167, 2167, 2167,    0, 2167, 2167,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2167, 2167, 2167,    0,
 2167, 2167, 2167,    0,    0,    0,    0,    0, 2167,    0,
    0,    0, 2167,    0, 2167, 2167, 2167, 2167,    0, 2167,
    0, 2167, 2167, 2167, 2167, 2167, 2167, 2167, 2167,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2167, 2167,
    0, 2167, 2167, 2167,    0,    0, 1422,    0, 2167, 2167,
 2167, 2167, 2167,    0, 2167,    0,    0, 2167, 2167,    0,
    0,    0, 2167, 2167,    0,    0, 2167, 2167, 2167, 2167,
 2167, 2167, 2167, 2167, 2167,    0, 2167,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2167, 2167, 2167, 2167,
 2167, 2167, 2167, 2167, 2167, 2167, 2167, 2167, 2167, 2167,
 2167, 2167, 2167, 2167, 2167, 2167, 2167,    0, 2167, 2167,
 2167, 2167, 2167, 2167,    0, 2167, 2167, 2167, 2167,    0,
 2167,    0, 2167, 2167, 2167, 2167, 2167, 2167, 2167, 2167,
 2167, 2167, 2167, 2167, 2167, 2167, 2167, 2167, 2167, 2167,
    0, 2167, 2167, 2167, 2167,    0, 2167, 2167, 2167, 2167,
 2167,    0, 2167, 2167, 2167, 2167, 2167, 2167, 2167, 2167,
 2167, 2167, 2167, 2167, 2167, 2167, 2167, 2167, 2167, 2167,
 2167, 2167, 2167, 2167, 2167, 2167, 2167, 2167, 2167, 2167,
 2167,    0, 2167, 2167,    0,    0,    0, 2167, 2167, 2167,
 2167, 2167, 2167,    0,    0, 2167, 2167, 2167, 2167, 2167,
 2167, 2167, 2167, 2167, 2559,    0,    0,    0,    0, 2167,
    0, 2167, 2167, 2167, 2167, 2167, 2167,    0, 2167,    0,
    0,    0,    0, 1424, 1423, 1425,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 2167, 2167,    0, 2167, 2167,
 2167, 2167, 2167, 2167, 2167, 2167, 2167,    0, 2167, 2167,
 2167, 2167, 2167, 2167, 2167, 2167, 2167, 2167, 2167, 2167,
    0, 2167, 2167, 2167,    0, 2167,    0,    0,    0, 2167,
 2167, 2167, 2167, 2167, 2167, 2167, 2167, 2167, 2167, 2167,
 2167, 2167, 2169, 2167, 2167,    0,    0,    0,    0,    0,
 2167, 2167, 2167, 2167, 2167, 2167, 2167, 2167, 2167, 2167,
 2167, 2167, 2167, 2167, 2167, 2167, 2167, 2167,    0,    0,
 2167,    0, 2167, 2167, 2167, 2167, 2167, 2167,    0,    0,
    0,    0, 1929, 2169, 2169, 2169, 2169, 2169, 2169, 2169,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2169, 2169, 2169, 2169, 2169,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 2169,    0, 2169,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1408,    0,    0,    0,    0,
    0,    0,    0,    0, 1409,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1410,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1411,    0,    0,    0,    0,    0,    0, 1412,    0,
    0,    0,    0,    0,    0,    0, 1413,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1414,    0, 1415,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1416,    0, 1417, 1418, 1419,
    0,    0,    0,    0, 1420,    0,    0,    0,    0, 2169,
    0, 2169,    0,    0, 2169,    0, 2169, 2169, 2169,    0,
 2169, 2169,    0, 2169,    0,    0,    0,    0,    0, 2169,
    0, 2169, 2169, 2169, 2169,    0, 2169,    0,    0, 2169,
    0,    0, 2169,    0,    0, 1421,    0,    0,    0, 2169,
    0, 2169, 2169,    0, 2169, 2169, 2169, 2169, 2169,    0,
    0, 2169,    0,    0, 2169,    0, 2169, 2169,    0,    0,
 2169, 2169, 2169,    0, 2169, 2169, 2169, 2169, 2169, 2169,
 2169,    0,    0,    0, 2169, 2169, 2169, 2169, 2169,    0,
    0,    0, 2169,    0,    0,    0,    0,    0,    0,    0,
 2169, 2169, 2169, 2169, 2169, 2169,    0,    0,    0, 2169,
 2169,    0, 2169,    0, 2169,    0, 2169, 2169, 2169,    0,
 2169,    0, 2169, 2169, 2169,    0, 2169, 2169, 2169, 2169,
 2169,    0, 2169, 2169,    0,    0,    0,    0, 2169,    0,
 2169, 2169,    0,    0, 2169, 2169, 2169,    0,    0, 2169,
    0, 2169,    0,    0, 2169, 2169, 2169, 2169,    0, 2169,
    0, 2169,    0, 2169, 2169, 2169, 2169, 2169,    0,    0,
 2169, 2169, 2169,    0, 2169, 2169,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 2169, 2169, 2169,    0, 2169,
 2169, 2169,    0,    0,    0,    0,    0, 2169,    0,    0,
    0, 2169,    0, 2169, 2169, 2169, 2169,    0, 2169,    0,
 2169, 2169, 2169, 2169, 2169, 2169, 2169, 2169,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 2169, 2169,    0,
 2169, 2169, 2169,    0,    0, 1422,    0, 2169, 2169, 2169,
 2169, 2169,    0, 2169,    0,    0, 2169, 2169,    0,    0,
    0, 2169, 2169,    0,    0, 2169, 2169, 2169, 2169, 2169,
 2169, 2169, 2169, 2169,    0, 2169,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 2169, 2169, 2169, 2169, 2169,
 2169, 2169, 2169, 2169, 2169, 2169, 2169, 2169, 2169, 2169,
 2169, 2169, 2169, 2169, 2169, 2169,    0, 2169, 2169, 2169,
 2169, 2169, 2169,    0, 2169, 2169, 2169, 2169,    0, 2169,
    0, 2169, 2169, 2169, 2169, 2169, 2169, 2169, 2169, 2169,
 2169, 2169, 2169, 2169, 2169, 2169, 2169, 2169, 2169,    0,
 2169, 2169, 2169, 2169,    0, 2169, 2169, 2169, 2169, 2169,
    0, 2169, 2169, 2169, 2169, 2169, 2169, 2169, 2169, 2169,
 2169, 2169, 2169, 2169, 2169, 2169, 2169, 2169, 2169, 2169,
 2169, 2169, 2169, 2169, 2169, 2169, 2169, 2169, 2169, 2169,
    0, 2169, 2169,    0,    0,    0, 2169, 2169, 2169, 2169,
 2169, 2169,    0,    0, 2169, 2169, 2169, 2169, 2169, 2169,
 2169, 2169, 2169, 2560,    0,    0,    0,    0, 2169,    0,
 2169, 2169, 2169, 2169, 2169, 2169,    0, 2169,    0,    0,
    0,    0, 1424, 1423, 1425,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 2169, 2169,    0, 2169, 2169, 2169,
 2169, 2169, 2169, 2169, 2169, 2169,    0, 2169, 2169, 2169,
 2169, 2169, 2169, 2169, 2169, 2169, 2169, 2169, 2169,    0,
 2169, 2169, 2169,    0, 2169,    0,    0,    0, 2169, 2169,
 2169, 2169, 2169, 2169, 2169, 2169, 2169, 2169, 2169, 2169,
 2169, 2170, 2169, 2169,    0,    0,    0,    0,    0, 2169,
 2169, 2169, 2169, 2169, 2169, 2169, 2169, 2169, 2169, 2169,
 2169, 2169, 2169, 2169, 2169, 2169, 2169,    0,    0, 2169,
    0, 2169, 2169, 2169, 2169, 2169, 2169,    0,    0,    0,
    0, 1930, 2170, 2170, 2170, 2170, 2170, 2170, 2170,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 2170,
 2170, 2170, 2170, 2170,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 2170,    0, 2170,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1408,    0,    0,    0,    0,    0,
    0,    0,    0, 1409,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1410,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1411,    0,    0,    0,    0,    0,    0, 1412,    0,    0,
    0,    0,    0,    0,    0, 1413,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1414,    0, 1415,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1416,    0, 1417, 1418, 1419,    0,
    0,    0,    0, 1420,    0,    0,    0,    0, 2170,    0,
 2170,    0,    0, 2170,    0, 2170, 2170, 2170,    0, 2170,
 2170,    0, 2170,    0,    0,    0,    0,    0, 2170,    0,
 2170, 2170, 2170, 2170,    0, 2170,    0,    0, 2170,    0,
    0, 2170,    0,    0, 1421,    0,    0,    0, 2170,    0,
 2170, 2170,    0, 2170, 2170, 2170, 2170, 2170,    0,    0,
 2170,    0,    0, 2170,    0, 2170, 2170,    0,    0, 2170,
 2170, 2170,    0, 2170, 2170, 2170, 2170, 2170, 2170, 2170,
    0,    0,    0, 2170, 2170, 2170, 2170, 2170,    0,    0,
    0, 2170,    0,    0,    0,    0,    0,    0,    0, 2170,
 2170, 2170, 2170, 2170, 2170,    0,    0,    0, 2170, 2170,
    0, 2170,    0, 2170,    0, 2170, 2170, 2170,    0, 2170,
    0, 2170, 2170, 2170,    0, 2170, 2170, 2170, 2170, 2170,
    0, 2170, 2170,    0,    0,    0,    0, 2170,    0, 2170,
 2170,    0,    0, 2170, 2170, 2170,    0,    0, 2170,    0,
 2170,    0,    0, 2170, 2170, 2170, 2170,    0, 2170,    0,
 2170,    0, 2170, 2170, 2170, 2170, 2170,    0,    0, 2170,
 2170, 2170,    0, 2170, 2170,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 2170, 2170, 2170,    0, 2170, 2170,
 2170,    0,    0,    0,    0,    0, 2170,    0,    0,    0,
 2170,    0, 2170, 2170, 2170, 2170,    0, 2170,    0, 2170,
 2170, 2170, 2170, 2170, 2170, 2170, 2170,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2170, 2170,    0, 2170,
 2170, 2170,    0,    0, 1422,    0, 2170, 2170, 2170, 2170,
 2170,    0, 2170,    0,    0, 2170, 2170,    0,    0,    0,
 2170, 2170,    0,    0, 2170, 2170, 2170, 2170, 2170, 2170,
 2170, 2170, 2170,    0, 2170,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 2170, 2170, 2170, 2170, 2170, 2170,
 2170, 2170, 2170, 2170, 2170, 2170, 2170, 2170, 2170, 2170,
 2170, 2170, 2170, 2170, 2170,    0, 2170, 2170, 2170, 2170,
 2170, 2170,    0, 2170, 2170, 2170, 2170,    0, 2170,    0,
 2170, 2170, 2170, 2170, 2170, 2170, 2170, 2170, 2170, 2170,
 2170, 2170, 2170, 2170, 2170, 2170, 2170, 2170,    0, 2170,
 2170, 2170, 2170,    0, 2170, 2170, 2170, 2170, 2170,    0,
 2170, 2170, 2170, 2170, 2170, 2170, 2170, 2170, 2170, 2170,
 2170, 2170, 2170, 2170, 2170, 2170, 2170, 2170, 2170, 2170,
 2170, 2170, 2170, 2170, 2170, 2170, 2170, 2170, 2170,    0,
 2170, 2170,    0,    0,    0, 2170, 2170, 2170, 2170, 2170,
 2170,    0,    0, 2170, 2170, 2170, 2170, 2170, 2170, 2170,
 2170, 2170, 2563,    0,    0,    0,    0, 2170,    0, 2170,
 2170, 2170, 2170, 2170, 2170,    0, 2170,    0,    0,    0,
    0, 1424, 1423, 1425,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 2170, 2170,    0, 2170, 2170, 2170, 2170,
 2170, 2170, 2170, 2170, 2170,    0, 2170, 2170, 2170, 2170,
 2170, 2170, 2170, 2170, 2170, 2170, 2170, 2170,    0, 2170,
 2170, 2170,    0, 2170,    0,    0,    0, 2170, 2170, 2170,
 2170, 2170, 2170, 2170, 2170, 2170, 2170, 2170, 2170, 2170,
 2173, 2170, 2170,    0,    0,    0,    0,    0, 2170, 2170,
 2170, 2170, 2170, 2170, 2170, 2170, 2170, 2170, 2170, 2170,
 2170, 2170, 2170, 2170, 2170, 2170,    0,    0, 2170,    0,
 2170, 2170, 2170, 2170, 2170, 2170,    0,    0,    0,    0,
 1931, 2173, 2173, 2173, 2173, 2173, 2173, 2173,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2173, 2173,
 2173, 2173, 2173,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 2173,    0, 2173,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1408,    0,    0,    0,    0,    0,    0,
    0,    0, 1409,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1410,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1411,
    0,    0,    0,    0,    0,    0, 1412,    0,    0,    0,
    0,    0,    0,    0, 1413,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1414,    0, 1415,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1416,    0, 1417, 1418, 1419,    0,    0,
    0,    0, 1420,    0,    0,    0,    0, 2173,    0, 2173,
    0,    0, 2173,    0, 2173, 2173, 2173,    0, 2173, 2173,
    0, 2173,    0,    0,    0,    0,    0, 2173,    0, 2173,
 2173, 2173, 2173,    0, 2173,    0,    0, 2173,    0,    0,
 2173,    0,    0, 1421,    0,    0,    0, 2173,    0, 2173,
 2173,    0, 2173, 2173, 2173, 2173, 2173,    0,    0, 2173,
    0,    0, 2173,    0, 2173, 2173,    0,    0, 2173, 2173,
 2173,    0, 2173, 2173, 2173, 2173, 2173, 2173, 2173,    0,
    0,    0, 2173, 2173, 2173, 2173, 2173,    0,    0,    0,
 2173,    0,    0,    0,    0,    0,    0,    0, 2173, 2173,
 2173, 2173, 2173, 2173,    0,    0,    0, 2173, 2173,    0,
 2173,    0, 2173,    0, 2173, 2173, 2173,    0, 2173,    0,
 2173, 2173, 2173,    0, 2173, 2173, 2173, 2173, 2173,    0,
 2173, 2173,    0,    0,    0,    0, 2173,    0, 2173, 2173,
    0,    0, 2173, 2173, 2173,    0,    0, 2173,    0, 2173,
    0,    0, 2173, 2173, 2173, 2173,    0, 2173,    0, 2173,
    0, 2173, 2173, 2173, 2173, 2173,    0,    0, 2173, 2173,
 2173,    0, 2173, 2173,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 2173, 2173, 2173,    0, 2173, 2173, 2173,
    0,    0,    0,    0,    0, 2173,    0,    0,    0, 2173,
    0, 2173, 2173, 2173, 2173,    0, 2173,    0, 2173, 2173,
 2173, 2173, 2173, 2173, 2173, 2173,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 2173, 2173,    0, 2173, 2173,
 2173,    0,    0, 1422,    0, 2173, 2173, 2173, 2173, 2173,
    0, 2173,    0,    0, 2173, 2173,    0,    0,    0, 2173,
 2173,    0,    0, 2173, 2173, 2173, 2173, 2173, 2173, 2173,
 2173, 2173,    0, 2173,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 2173, 2173, 2173, 2173, 2173, 2173, 2173,
 2173, 2173, 2173, 2173, 2173, 2173, 2173, 2173, 2173, 2173,
 2173, 2173, 2173, 2173,    0, 2173, 2173, 2173, 2173, 2173,
 2173,    0, 2173, 2173, 2173, 2173,    0, 2173,    0, 2173,
 2173, 2173, 2173, 2173, 2173, 2173, 2173, 2173, 2173, 2173,
 2173, 2173, 2173, 2173, 2173, 2173, 2173,    0, 2173, 2173,
 2173, 2173,    0, 2173, 2173, 2173, 2173, 2173,    0, 2173,
 2173, 2173, 2173, 2173, 2173, 2173, 2173, 2173, 2173, 2173,
 2173, 2173, 2173, 2173, 2173, 2173, 2173, 2173, 2173, 2173,
 2173, 2173, 2173, 2173, 2173, 2173, 2173, 2173,    0, 2173,
 2173,    0,    0,    0, 2173, 2173, 2173, 2173, 2173, 2173,
    0,    0, 2173, 2173, 2173, 2173, 2173, 2173, 2173, 2173,
 2173, 2564,    0,    0,    0,    0, 2173,    0, 2173, 2173,
 2173, 2173, 2173, 2173,    0, 2173,    0,    0,    0,    0,
 1424, 1423, 1425,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 2173, 2173,    0, 2173, 2173, 2173, 2173, 2173,
 2173, 2173, 2173, 2173,    0, 2173, 2173, 2173, 2173, 2173,
 2173, 2173, 2173, 2173, 2173, 2173, 2173,    0, 2173, 2173,
 2173,    0, 2173,    0,    0,    0, 2173, 2173, 2173, 2173,
 2173, 2173, 2173, 2173, 2173, 2173, 2173, 2173, 2173, 2175,
 2173, 2173,    0,    0,    0,    0,    0, 2173, 2173, 2173,
 2173, 2173, 2173, 2173, 2173, 2173, 2173, 2173, 2173, 2173,
 2173, 2173, 2173, 2173, 2173,    0,    0, 2173,    0, 2173,
 2173, 2173, 2173, 2173, 2173,    0,    0,    0,    0, 1932,
 2175, 2175, 2175, 2175, 2175, 2175, 2175,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 2175, 2175, 2175,
 2175, 2175,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2175,    0, 2175,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1408,    0,    0,    0,    0,    0,    0,    0,
    0, 1409,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1410,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1411,    0,
    0,    0,    0,    0,    0, 1412,    0,    0,    0,    0,
    0,    0,    0, 1413,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1414,    0, 1415,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1416,    0, 1417, 1418, 1419,    0,    0,    0,
    0, 1420,    0,    0,    0,    0, 2175,    0, 2175,    0,
    0, 2175,    0, 2175, 2175, 2175,    0, 2175, 2175,    0,
 2175,    0,    0,    0,    0,    0, 2175,    0, 2175, 2175,
 2175, 2175,    0, 2175,    0,    0, 2175,    0,    0, 2175,
    0,    0, 1421,    0,    0,    0, 2175,    0, 2175, 2175,
    0, 2175, 2175, 2175, 2175, 2175,    0,    0, 2175,    0,
    0, 2175,    0, 2175, 2175,    0,    0, 2175, 2175, 2175,
    0, 2175, 2175, 2175, 2175, 2175, 2175, 2175,    0,    0,
    0, 2175, 2175, 2175, 2175, 2175,    0,    0,    0, 2175,
    0,    0,    0,    0,    0,    0,    0, 2175, 2175, 2175,
 2175, 2175, 2175,    0,    0,    0, 2175, 2175,    0, 2175,
    0, 2175,    0, 2175, 2175, 2175,    0, 2175,    0, 2175,
 2175, 2175,    0, 2175, 2175, 2175, 2175, 2175,    0, 2175,
 2175,    0,    0,    0,    0, 2175,    0, 2175, 2175,    0,
    0, 2175, 2175, 2175,    0,    0, 2175,    0, 2175,    0,
    0, 2175, 2175, 2175, 2175,    0, 2175,    0, 2175,    0,
 2175, 2175, 2175, 2175, 2175,    0,    0, 2175, 2175, 2175,
    0, 2175, 2175,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 2175, 2175, 2175,    0, 2175, 2175, 2175,    0,
    0,    0,    0,    0, 2175,    0,    0,    0, 2175,    0,
 2175, 2175, 2175, 2175,    0, 2175,    0, 2175, 2175, 2175,
 2175, 2175, 2175, 2175, 2175,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 2175, 2175,    0, 2175, 2175, 2175,
    0,    0, 1422,    0, 2175, 2175, 2175, 2175, 2175,    0,
 2175,    0,    0, 2175, 2175,    0,    0,    0, 2175, 2175,
    0,    0, 2175, 2175, 2175, 2175, 2175, 2175, 2175, 2175,
 2175,    0, 2175,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 2175, 2175, 2175, 2175, 2175, 2175, 2175, 2175,
 2175, 2175, 2175, 2175, 2175, 2175, 2175, 2175, 2175, 2175,
 2175, 2175, 2175,    0, 2175, 2175, 2175, 2175, 2175, 2175,
    0, 2175, 2175, 2175, 2175,    0, 2175,    0, 2175, 2175,
 2175, 2175, 2175, 2175, 2175, 2175, 2175, 2175, 2175, 2175,
 2175, 2175, 2175, 2175, 2175, 2175,    0, 2175, 2175, 2175,
 2175,    0, 2175, 2175, 2175, 2175, 2175,    0, 2175, 2175,
 2175, 2175, 2175, 2175, 2175, 2175, 2175, 2175, 2175, 2175,
 2175, 2175, 2175, 2175, 2175, 2175, 2175, 2175, 2175, 2175,
 2175, 2175, 2175, 2175, 2175, 2175, 2175,    0, 2175, 2175,
    0,    0,    0, 2175, 2175, 2175, 2175, 2175, 2175,    0,
    0, 2175, 2175, 2175, 2175, 2175, 2175, 2175, 2175, 2175,
 2906,    0,    0,    0,    0, 2175,    0, 2175, 2175, 2175,
 2175, 2175, 2175,    0, 2175,    0,    0,    0,    0, 1424,
 1423, 1425,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 2175, 2175,    0, 2175, 2175, 2175, 2175, 2175, 2175,
 2175, 2175, 2175,    0, 2175, 2175, 2175, 2175, 2175, 2175,
 2175, 2175, 2175, 2175, 2175, 2175,    0, 2175, 2175, 2175,
    0, 2175,    0,    0,    0, 2175, 2175, 2175, 2175, 2175,
 2175, 2175, 2175, 2175, 2175, 2175, 2175, 2175, 2177, 2175,
 2175,    0,    0,    0,    0,    0, 2175, 2175, 2175, 2175,
 2175, 2175, 2175, 2175, 2175, 2175, 2175, 2175, 2175, 2175,
 2175, 2175, 2175, 2175,    0,    0, 2175,    0, 2175, 2175,
 2175, 2175, 2175, 2175,    0,    0,    0,    0, 1933, 2177,
 2177, 2177, 2177, 2177, 2177, 2177,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2177, 2177, 2177, 2177,
 2177,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 2177,
    0, 2177,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1408,    0,    0,    0,    0,    0,    0,    0,    0,
 1409,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1410,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1411,    0,    0,
    0,    0,    0,    0, 1412,    0,    0,    0,    0,    0,
    0,    0, 1413,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1414,
    0, 1415,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1416,    0, 1417, 1418, 1419,    0,    0,    0,    0,
 1420,    0,    0,    0,    0, 2177,    0, 2177,    0,    0,
 2177,    0, 2177, 2177, 2177,    0, 2177, 2177,    0, 2177,
    0,    0,    0,    0,    0, 2177,    0, 2177, 2177, 2177,
 2177,    0, 2177,    0,    0, 2177,    0,    0, 2177,    0,
    0, 1421,    0,    0,    0, 2177,    0, 2177, 2177,    0,
 2177, 2177, 2177, 2177, 2177,    0,    0, 2177,    0,    0,
 2177,    0, 2177, 2177,    0,    0, 2177, 2177, 2177,    0,
 2177, 2177, 2177, 2177, 2177, 2177, 2177,    0,    0,    0,
 2177, 2177, 2177, 2177, 2177,    0,    0,    0, 2177,    0,
    0,    0,    0,    0,    0,    0, 2177, 2177, 2177, 2177,
 2177, 2177,    0,    0,    0, 2177, 2177,    0, 2177,    0,
 2177,    0, 2177, 2177, 2177,    0, 2177,    0, 2177, 2177,
 2177,    0, 2177, 2177, 2177, 2177, 2177,    0, 2177, 2177,
    0,    0,    0,    0, 2177,    0, 2177, 2177,    0,    0,
 2177, 2177, 2177,    0,    0, 2177,    0, 2177,    0,    0,
 2177, 2177, 2177, 2177,    0, 2177,    0, 2177,    0, 2177,
 2177, 2177, 2177, 2177,    0,    0, 2177, 2177, 2177,    0,
 2177, 2177,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 2177, 2177, 2177,    0, 2177, 2177, 2177,    0,    0,
    0,    0,    0, 2177,    0,    0,    0, 2177,    0, 2177,
 2177, 2177, 2177,    0, 2177,    0, 2177, 2177, 2177, 2177,
 2177, 2177, 2177, 2177,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 2177, 2177,    0, 2177, 2177, 2177,    0,
    0, 1422,    0, 2177, 2177, 2177, 2177, 2177,    0, 2177,
    0,    0, 2177, 2177,    0,    0,    0, 2177, 2177,    0,
    0, 2177, 2177, 2177, 2177, 2177, 2177, 2177, 2177, 2177,
    0, 2177,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 2177, 2177, 2177, 2177, 2177, 2177, 2177, 2177, 2177,
 2177, 2177, 2177, 2177, 2177, 2177, 2177, 2177, 2177, 2177,
 2177, 2177,    0, 2177, 2177, 2177, 2177, 2177, 2177,    0,
 2177, 2177, 2177, 2177,    0, 2177,    0, 2177, 2177, 2177,
 2177, 2177, 2177, 2177, 2177, 2177, 2177, 2177, 2177, 2177,
 2177, 2177, 2177, 2177, 2177,    0, 2177, 2177, 2177, 2177,
    0, 2177, 2177, 2177, 2177, 2177,    0, 2177, 2177, 2177,
 2177, 2177, 2177, 2177, 2177, 2177, 2177, 2177, 2177, 2177,
 2177, 2177, 2177, 2177, 2177, 2177, 2177, 2177, 2177, 2177,
 2177, 2177, 2177, 2177, 2177, 2177,    0, 2177, 2177,    0,
    0,    0, 2177, 2177, 2177, 2177, 2177, 2177,    0,    0,
 2177, 2177, 2177, 2177, 2177, 2177, 2177, 2177, 2177, 2908,
    0,    0,    0,    0, 2177,    0, 2177, 2177, 2177, 2177,
 2177, 2177,    0, 2177,    0,    0,    0,    0, 1424, 1423,
 1425,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2177, 2177,    0, 2177, 2177, 2177, 2177, 2177, 2177, 2177,
 2177, 2177,    0, 2177, 2177, 2177, 2177, 2177, 2177, 2177,
 2177, 2177, 2177, 2177, 2177,    0, 2177, 2177, 2177,    0,
 2177,    0,    0,    0, 2177, 2177, 2177, 2177, 2177, 2177,
 2177, 2177, 2177, 2177, 2177, 2177, 2177, 2178, 2177, 2177,
    0,    0,    0,    0,    0, 2177, 2177, 2177, 2177, 2177,
 2177, 2177, 2177, 2177, 2177, 2177, 2177, 2177, 2177, 2177,
 2177, 2177, 2177,    0,    0, 2177,    0, 2177, 2177, 2177,
 2177, 2177, 2177,    0,    0,    0,    0, 1938, 2178, 2178,
 2178, 2178, 2178, 2178, 2178,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 2178, 2178, 2178, 2178, 2178,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2178,    0,
 2178,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1408,    0,    0,    0,    0,    0,    0,    0,    0, 1409,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1410,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1411,    0,    0,    0,
    0,    0,    0, 1412,    0,    0,    0,    0,    0,    0,
    0, 1413,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1414,    0,
 1415,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1416,    0, 1417, 1418, 1419,    0,    0,    0,    0, 1420,
    0,    0,    0,    0, 2178,    0, 2178,    0,    0, 2178,
    0, 2178, 2178, 2178,    0, 2178, 2178,    0, 2178,    0,
    0,    0,    0,    0, 2178,    0, 2178, 2178, 2178, 2178,
    0, 2178,    0,    0, 2178,    0,    0, 2178,    0,    0,
 1421,    0,    0,    0, 2178,    0, 2178, 2178,    0, 2178,
 2178, 2178, 2178, 2178,    0,    0, 2178,    0,    0, 2178,
    0, 2178, 2178,    0,    0, 2178, 2178, 2178,    0, 2178,
 2178, 2178, 2178, 2178, 2178, 2178,    0,    0,    0, 2178,
 2178, 2178, 2178, 2178,    0,    0,    0, 2178,    0,    0,
    0,    0,    0,    0,    0, 2178, 2178, 2178, 2178, 2178,
 2178,    0,    0,    0, 2178, 2178,    0, 2178,    0, 2178,
    0, 2178, 2178, 2178,    0, 2178,    0, 2178, 2178, 2178,
    0, 2178, 2178, 2178, 2178, 2178,    0, 2178, 2178,    0,
    0,    0,    0, 2178,    0, 2178, 2178,    0,    0, 2178,
 2178, 2178,    0,    0, 2178,    0, 2178,    0,    0, 2178,
 2178, 2178, 2178,    0, 2178,    0, 2178,    0, 2178, 2178,
 2178, 2178, 2178,    0,    0, 2178, 2178, 2178,    0, 2178,
 2178,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2178, 2178, 2178,    0, 2178, 2178, 2178,    0,    0,    0,
    0,    0, 2178,    0,    0,    0, 2178,    0, 2178, 2178,
 2178, 2178,    0, 2178,    0, 2178, 2178, 2178, 2178, 2178,
 2178, 2178, 2178,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 2178, 2178,    0, 2178, 2178, 2178,    0,    0,
 1422,    0, 2178, 2178, 2178, 2178, 2178,    0, 2178,    0,
    0, 2178, 2178,    0,    0,    0, 2178, 2178,    0,    0,
 2178, 2178, 2178, 2178, 2178, 2178, 2178, 2178, 2178,    0,
 2178,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2178, 2178, 2178, 2178, 2178, 2178, 2178, 2178, 2178, 2178,
 2178, 2178, 2178, 2178, 2178, 2178, 2178, 2178, 2178, 2178,
 2178,    0, 2178, 2178, 2178, 2178, 2178, 2178,    0, 2178,
 2178, 2178, 2178,    0, 2178,    0, 2178, 2178, 2178, 2178,
 2178, 2178, 2178, 2178, 2178, 2178, 2178, 2178, 2178, 2178,
 2178, 2178, 2178, 2178,    0, 2178, 2178, 2178, 2178,    0,
 2178, 2178, 2178, 2178, 2178,    0, 2178, 2178, 2178, 2178,
 2178, 2178, 2178, 2178, 2178, 2178, 2178, 2178, 2178, 2178,
 2178, 2178, 2178, 2178, 2178, 2178, 2178, 2178, 2178, 2178,
 2178, 2178, 2178, 2178, 2178,    0, 2178, 2178,    0,    0,
    0, 2178, 2178, 2178, 2178, 2178, 2178,    0,    0, 2178,
 2178, 2178, 2178, 2178, 2178, 2178, 2178, 2178, 2909,    0,
    0,    0,    0, 2178,    0, 2178, 2178, 2178, 2178, 2178,
 2178,    0, 2178,    0,    0,    0,    0, 1424, 1423, 1425,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 2178,
 2178,    0, 2178, 2178, 2178, 2178, 2178, 2178, 2178, 2178,
 2178,    0, 2178, 2178, 2178, 2178, 2178, 2178, 2178, 2178,
 2178, 2178, 2178, 2178,    0, 2178, 2178, 2178,    0, 2178,
    0,    0,    0, 2178, 2178, 2178, 2178, 2178, 2178, 2178,
 2178, 2178, 2178, 2178, 2178, 2178, 2179, 2178, 2178,    0,
    0,    0,    0,    0, 2178, 2178, 2178, 2178, 2178, 2178,
 2178, 2178, 2178, 2178, 2178, 2178, 2178, 2178, 2178, 2178,
 2178, 2178,    0,    0, 2178,    0, 2178, 2178, 2178, 2178,
 2178, 2178,    0,    0,    0,    0, 1939, 2179, 2179, 2179,
 2179, 2179, 2179, 2179,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 2179, 2179, 2179, 2179, 2179,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 2179,    0, 2179,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1408,
    0,    0,    0,    0,    0,    0,    0,    0, 1409,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1410,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1411,    0,    0,    0,    0,
    0,    0, 1412,    0,    0,    0,    0,    0,    0,    0,
 1413,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1414,    0, 1415,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1416,
    0, 1417, 1418, 1419,    0,    0,    0,    0, 1420,    0,
    0,    0,    0, 2179,    0, 2179,    0,    0, 2179,    0,
 2179, 2179, 2179,    0, 2179, 2179,    0, 2179,    0,    0,
    0,    0,    0, 2179,    0, 2179, 2179, 2179, 2179,    0,
 2179,    0,    0, 2179,    0,    0, 2179,    0,    0, 1421,
    0,    0,    0, 2179,    0, 2179, 2179,    0, 2179, 2179,
 2179, 2179, 2179,    0,    0, 2179,    0,    0, 2179,    0,
 2179, 2179,    0,    0, 2179, 2179, 2179,    0, 2179, 2179,
 2179, 2179, 2179, 2179, 2179,    0,    0,    0, 2179, 2179,
 2179, 2179, 2179,    0,    0,    0, 2179,    0,    0,    0,
    0,    0,    0,    0, 2179, 2179, 2179, 2179, 2179, 2179,
    0,    0,    0, 2179, 2179,    0, 2179,    0, 2179,    0,
 2179, 2179, 2179,    0, 2179,    0, 2179, 2179, 2179,    0,
 2179, 2179, 2179, 2179, 2179,    0, 2179, 2179,    0,    0,
    0,    0, 2179,    0, 2179, 2179,    0,    0, 2179, 2179,
 2179,    0,    0, 2179,    0, 2179,    0,    0, 2179, 2179,
 2179, 2179,    0, 2179,    0, 2179,    0, 2179, 2179, 2179,
 2179, 2179,    0,    0, 2179, 2179, 2179,    0, 2179, 2179,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 2179,
 2179, 2179,    0, 2179, 2179, 2179,    0,    0,    0,    0,
    0, 2179,    0,    0,    0, 2179,    0, 2179, 2179, 2179,
 2179,    0, 2179,    0, 2179, 2179, 2179, 2179, 2179, 2179,
 2179, 2179,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 2179, 2179,    0, 2179, 2179, 2179,    0,    0, 1422,
    0, 2179, 2179, 2179, 2179, 2179,    0, 2179,    0,    0,
 2179, 2179,    0,    0,    0, 2179, 2179,    0,    0, 2179,
 2179, 2179, 2179, 2179, 2179, 2179, 2179, 2179,    0, 2179,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 2179,
 2179, 2179, 2179, 2179, 2179, 2179, 2179, 2179, 2179, 2179,
 2179, 2179, 2179, 2179, 2179, 2179, 2179, 2179, 2179, 2179,
    0, 2179, 2179, 2179, 2179, 2179, 2179,    0, 2179, 2179,
 2179, 2179,    0, 2179,    0, 2179, 2179, 2179, 2179, 2179,
 2179, 2179, 2179, 2179, 2179, 2179, 2179, 2179, 2179, 2179,
 2179, 2179, 2179,    0, 2179, 2179, 2179, 2179,    0, 2179,
 2179, 2179, 2179, 2179,    0, 2179, 2179, 2179, 2179, 2179,
 2179, 2179, 2179, 2179, 2179, 2179, 2179, 2179, 2179, 2179,
 2179, 2179, 2179, 2179, 2179, 2179, 2179, 2179, 2179, 2179,
 2179, 2179, 2179, 2179,    0, 2179, 2179,    0,    0,    0,
 2179, 2179, 2179, 2179, 2179, 2179,    0,    0, 2179, 2179,
 2179, 2179, 2179, 2179, 2179, 2179, 2179, 2916,    0,    0,
    0,    0, 2179,    0, 2179, 2179, 2179, 2179, 2179, 2179,
    0, 2179,    0,    0,    0,    0, 1424, 1423, 1425,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2179, 2179,
    0, 2179, 2179, 2179, 2179, 2179, 2179, 2179, 2179, 2179,
    0, 2179, 2179, 2179, 2179, 2179, 2179, 2179, 2179, 2179,
 2179, 2179, 2179,    0, 2179, 2179, 2179,    0, 2179,    0,
    0,    0, 2179, 2179, 2179, 2179, 2179, 2179, 2179, 2179,
 2179, 2179, 2179, 2179, 2179, 2180, 2179, 2179,    0,    0,
    0,    0,    0, 2179, 2179, 2179, 2179, 2179, 2179, 2179,
 2179, 2179, 2179, 2179, 2179, 2179, 2179, 2179, 2179, 2179,
 2179,    0,    0, 2179,    0, 2179, 2179, 2179, 2179, 2179,
 2179,    0,    0,    0,    0, 1941, 2180, 2180, 2180, 2180,
 2180, 2180, 2180,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 2180, 2180, 2180, 2180, 2180,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2180,    0, 2180,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1408,    0,
    0,    0,    0,    0,    0,    0,    0, 1409,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1410,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1411,    0,    0,    0,    0,    0,
    0, 1412,    0,    0,    0,    0,    0,    0,    0, 1413,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1414,    0, 1415,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1416,    0,
 1417, 1418, 1419,    0,    0,    0,    0, 1420,    0,    0,
    0,    0, 2180,    0, 2180,    0,    0, 2180,    0, 2180,
 2180, 2180,    0, 2180, 2180,    0, 2180,    0,    0,    0,
    0,    0, 2180,    0, 2180, 2180, 2180, 2180,    0, 2180,
    0,    0, 2180,    0,    0, 2180,    0,    0, 1421,    0,
    0,    0, 2180,    0, 2180, 2180,    0, 2180, 2180, 2180,
 2180, 2180,    0,    0, 2180,    0,    0, 2180,    0, 2180,
 2180,    0,    0, 2180, 2180, 2180,    0, 2180, 2180, 2180,
 2180, 2180, 2180, 2180,    0,    0,    0, 2180, 2180, 2180,
 2180, 2180,    0,    0,    0, 2180,    0,    0,    0,    0,
    0,    0,    0, 2180, 2180, 2180, 2180, 2180, 2180,    0,
    0,    0, 2180, 2180,    0, 2180,    0, 2180,    0, 2180,
 2180, 2180,    0, 2180,    0, 2180, 2180, 2180,    0, 2180,
 2180, 2180, 2180, 2180,    0, 2180, 2180,    0,    0,    0,
    0, 2180,    0, 2180, 2180,    0,    0, 2180, 2180, 2180,
    0,    0, 2180,    0, 2180,    0,    0, 2180, 2180, 2180,
 2180,    0, 2180,    0, 2180,    0, 2180, 2180, 2180, 2180,
 2180,    0,    0, 2180, 2180, 2180,    0, 2180, 2180,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2180, 2180,
 2180,    0, 2180, 2180, 2180,    0,    0,    0,    0,    0,
 2180,    0,    0,    0, 2180,    0, 2180, 2180, 2180, 2180,
    0, 2180,    0, 2180, 2180, 2180, 2180, 2180, 2180, 2180,
 2180,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2180, 2180,    0, 2180, 2180, 2180,    0,    0, 1422,    0,
 2180, 2180, 2180, 2180, 2180,    0, 2180,    0,    0, 2180,
 2180,    0,    0,    0, 2180, 2180,    0,    0, 2180, 2180,
 2180, 2180, 2180, 2180, 2180, 2180, 2180,    0, 2180,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2180, 2180,
 2180, 2180, 2180, 2180, 2180, 2180, 2180, 2180, 2180, 2180,
 2180, 2180, 2180, 2180, 2180, 2180, 2180, 2180, 2180,    0,
 2180, 2180, 2180, 2180, 2180, 2180,    0, 2180, 2180, 2180,
 2180,    0, 2180,    0, 2180, 2180, 2180, 2180, 2180, 2180,
 2180, 2180, 2180, 2180, 2180, 2180, 2180, 2180, 2180, 2180,
 2180, 2180,    0, 2180, 2180, 2180, 2180,    0, 2180, 2180,
 2180, 2180, 2180,    0, 2180, 2180, 2180, 2180, 2180, 2180,
 2180, 2180, 2180, 2180, 2180, 2180, 2180, 2180, 2180, 2180,
 2180, 2180, 2180, 2180, 2180, 2180, 2180, 2180, 2180, 2180,
 2180, 2180, 2180,    0, 2180, 2180,    0,    0,    0, 2180,
 2180, 2180, 2180, 2180, 2180,    0,    0, 2180, 2180, 2180,
 2180, 2180, 2180, 2180, 2180, 2180, 2919,    0,    0,    0,
    0, 2180,    0, 2180, 2180, 2180, 2180, 2180, 2180,    0,
 2180,    0,    0,    0,    0, 1424, 1423, 1425,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 2180, 2180,    0,
 2180, 2180, 2180, 2180, 2180, 2180, 2180, 2180, 2180,    0,
 2180, 2180, 2180, 2180, 2180, 2180, 2180, 2180, 2180, 2180,
 2180, 2180,    0, 2180, 2180, 2180,    0, 2180,    0,    0,
    0, 2180, 2180, 2180, 2180, 2180, 2180, 2180, 2180, 2180,
 2180, 2180, 2180, 2180, 2181, 2180, 2180,    0,    0,    0,
    0,    0, 2180, 2180, 2180, 2180, 2180, 2180, 2180, 2180,
 2180, 2180, 2180, 2180, 2180, 2180, 2180, 2180, 2180, 2180,
    0,    0, 2180,    0, 2180, 2180, 2180, 2180, 2180, 2180,
    0,    0,    0,    0, 1942, 2181, 2181, 2181, 2181, 2181,
 2181, 2181,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 2181, 2181, 2181, 2181, 2181,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 2181,    0, 2181,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1408,    0,    0,
    0,    0,    0,    0,    0,    0, 1409,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1410,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1411,    0,    0,    0,    0,    0,    0,
 1412,    0,    0,    0,    0,    0,    0,    0, 1413,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1414,    0, 1415,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1416,    0, 1417,
 1418, 1419,    0,    0,    0,    0, 1420,    0,    0,    0,
    0, 2181,    0, 2181,    0,    0, 2181,    0, 2181, 2181,
 2181,    0, 2181, 2181,    0, 2181,    0,    0,    0,    0,
    0, 2181,    0, 2181, 2181, 2181, 2181,    0, 2181,    0,
    0, 2181,    0,    0, 2181,    0,    0, 1421,    0,    0,
    0, 2181,    0, 2181, 2181,    0, 2181, 2181, 2181, 2181,
 2181,    0,    0, 2181,    0,    0, 2181,    0, 2181, 2181,
    0,    0, 2181, 2181, 2181,    0, 2181, 2181, 2181, 2181,
 2181, 2181, 2181,    0,    0,    0, 2181, 2181, 2181, 2181,
 2181,    0,    0,    0, 2181,    0,    0,    0,    0,    0,
    0,    0, 2181, 2181, 2181, 2181, 2181, 2181,    0,    0,
    0, 2181, 2181,    0, 2181,    0, 2181,    0, 2181, 2181,
 2181,    0, 2181,    0, 2181, 2181, 2181,    0, 2181, 2181,
 2181, 2181, 2181,    0, 2181, 2181,    0,    0,    0,    0,
 2181,    0, 2181, 2181,    0,    0, 2181, 2181, 2181,    0,
    0, 2181,    0, 2181,    0,    0, 2181, 2181, 2181, 2181,
    0, 2181,    0, 2181,    0, 2181, 2181, 2181, 2181, 2181,
    0,    0, 2181, 2181, 2181,    0, 2181, 2181,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 2181, 2181, 2181,
    0, 2181, 2181, 2181,    0,    0,    0,    0,    0, 2181,
    0,    0,    0, 2181,    0, 2181, 2181, 2181, 2181,    0,
 2181,    0, 2181, 2181, 2181, 2181, 2181, 2181, 2181, 2181,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 2181,
 2181,    0, 2181, 2181, 2181,    0,    0, 1422,    0, 2181,
 2181, 2181, 2181, 2181,    0, 2181,    0,    0, 2181, 2181,
    0,    0,    0, 2181, 2181,    0,    0, 2181, 2181, 2181,
 2181, 2181, 2181, 2181, 2181, 2181,    0, 2181,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 2181, 2181, 2181,
 2181, 2181, 2181, 2181, 2181, 2181, 2181, 2181, 2181, 2181,
 2181, 2181, 2181, 2181, 2181, 2181, 2181, 2181,    0, 2181,
 2181, 2181, 2181, 2181, 2181,    0, 2181, 2181, 2181, 2181,
    0, 2181,    0, 2181, 2181, 2181, 2181, 2181, 2181, 2181,
 2181, 2181, 2181, 2181, 2181, 2181, 2181, 2181, 2181, 2181,
 2181,    0, 2181, 2181, 2181, 2181,    0, 2181, 2181, 2181,
 2181, 2181,    0, 2181, 2181, 2181, 2181, 2181, 2181, 2181,
 2181, 2181, 2181, 2181, 2181, 2181, 2181, 2181, 2181, 2181,
 2181, 2181, 2181, 2181, 2181, 2181, 2181, 2181, 2181, 2181,
 2181, 2181,    0, 2181, 2181,    0,    0,    0, 2181, 2181,
 2181, 2181, 2181, 2181,    0,    0, 2181, 2181, 2181, 2181,
 2181, 2181, 2181, 2181, 2181, 2920,    0,    0,    0,    0,
 2181,    0, 2181, 2181, 2181, 2181, 2181, 2181,    0, 2181,
    0,    0,    0,    0, 1424, 1423, 1425,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2181, 2181,    0, 2181,
 2181, 2181, 2181, 2181, 2181, 2181, 2181, 2181,    0, 2181,
 2181, 2181, 2181, 2181, 2181, 2181, 2181, 2181, 2181, 2181,
 2181,    0, 2181, 2181, 2181,    0, 2181,    0,    0,    0,
 2181, 2181, 2181, 2181, 2181, 2181, 2181, 2181, 2181, 2181,
 2181, 2181, 2181, 2182, 2181, 2181,    0,    0,    0,    0,
    0, 2181, 2181, 2181, 2181, 2181, 2181, 2181, 2181, 2181,
 2181, 2181, 2181, 2181, 2181, 2181, 2181, 2181, 2181,    0,
    0, 2181,    0, 2181, 2181, 2181, 2181, 2181, 2181,    0,
    0,    0,    0, 1945, 2182, 2182, 2182, 2182, 2182, 2182,
 2182,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 2182, 2182, 2182, 2182, 2182,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 2182,    0, 2182,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1408,    0,    0,    0,
    0,    0,    0,    0,    0, 1409,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1410,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1411,    0,    0,    0,    0,    0,    0, 1412,
    0,    0,    0,    0,    0,    0,    0, 1413,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1414,    0, 1415,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1416,    0, 1417, 1418,
 1419,    0,    0,    0,    0, 1420,    0,    0,    0,    0,
 2182,    0, 2182,    0,    0, 2182,    0, 2182, 2182, 2182,
    0, 2182, 2182,    0, 2182,    0,    0,    0,    0,    0,
 2182,    0, 2182, 2182, 2182, 2182,    0, 2182,    0,    0,
 2182,    0,    0, 2182,    0,    0, 1421,    0,    0,    0,
 2182,    0, 2182, 2182,    0, 2182, 2182, 2182, 2182, 2182,
    0,    0, 2182,    0,    0, 2182,    0, 2182, 2182,    0,
    0, 2182, 2182, 2182,    0, 2182, 2182, 2182, 2182, 2182,
 2182, 2182,    0,    0,    0, 2182, 2182, 2182, 2182, 2182,
    0,    0,    0, 2182,    0,    0,    0,    0,    0,    0,
    0, 2182, 2182, 2182, 2182, 2182, 2182,    0,    0,    0,
 2182, 2182,    0, 2182,    0, 2182,    0, 2182, 2182, 2182,
    0, 2182,    0, 2182, 2182, 2182,    0, 2182, 2182, 2182,
 2182, 2182,    0, 2182, 2182,    0,    0,    0,    0, 2182,
    0, 2182, 2182,    0,    0, 2182, 2182, 2182,    0,    0,
 2182,    0, 2182,    0,    0, 2182, 2182, 2182, 2182,    0,
 2182,    0, 2182,    0, 2182, 2182, 2182, 2182, 2182,    0,
    0, 2182, 2182, 2182,    0, 2182, 2182,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2182, 2182, 2182,    0,
 2182, 2182, 2182,    0,    0,    0,    0,    0, 2182,    0,
    0,    0, 2182,    0, 2182, 2182, 2182, 2182,    0, 2182,
    0, 2182, 2182, 2182, 2182, 2182, 2182, 2182, 2182,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2182, 2182,
    0, 2182, 2182, 2182,    0,    0, 1422,    0, 2182, 2182,
 2182, 2182, 2182,    0, 2182,    0,    0, 2182, 2182,    0,
    0,    0, 2182, 2182,    0,    0, 2182, 2182, 2182, 2182,
 2182, 2182, 2182, 2182, 2182,    0, 2182,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2182, 2182, 2182, 2182,
 2182, 2182, 2182, 2182, 2182, 2182, 2182, 2182, 2182, 2182,
 2182, 2182, 2182, 2182, 2182, 2182, 2182,    0, 2182, 2182,
 2182, 2182, 2182, 2182,    0, 2182, 2182, 2182, 2182,    0,
 2182,    0, 2182, 2182, 2182, 2182, 2182, 2182, 2182, 2182,
 2182, 2182, 2182, 2182, 2182, 2182, 2182, 2182, 2182, 2182,
    0, 2182, 2182, 2182, 2182,    0, 2182, 2182, 2182, 2182,
 2182,    0, 2182, 2182, 2182, 2182, 2182, 2182, 2182, 2182,
 2182, 2182, 2182, 2182, 2182, 2182, 2182, 2182, 2182, 2182,
 2182, 2182, 2182, 2182, 2182, 2182, 2182, 2182, 2182, 2182,
 2182,    0, 2182, 2182,    0,    0,    0, 2182, 2182, 2182,
 2182, 2182, 2182,    0,    0, 2182, 2182, 2182, 2182, 2182,
 2182, 2182, 2182, 2182, 2933,    0,    0,    0,    0, 2182,
    0, 2182, 2182, 2182, 2182, 2182, 2182,    0, 2182,    0,
    0,    0,    0, 1424, 1423, 1425,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 2182, 2182,    0, 2182, 2182,
 2182, 2182, 2182, 2182, 2182, 2182, 2182,    0, 2182, 2182,
 2182, 2182, 2182, 2182, 2182, 2182, 2182, 2182, 2182, 2182,
    0, 2182, 2182, 2182,    0, 2182,    0,    0,    0, 2182,
 2182, 2182, 2182, 2182, 2182, 2182, 2182, 2182, 2182, 2182,
 2182, 2182, 2183, 2182, 2182,    0,    0,    0,    0,    0,
 2182, 2182, 2182, 2182, 2182, 2182, 2182, 2182, 2182, 2182,
 2182, 2182, 2182, 2182, 2182, 2182, 2182, 2182,    0,    0,
 2182,    0, 2182, 2182, 2182, 2182, 2182, 2182,    0,    0,
    0,    0, 1946, 2183, 2183, 2183, 2183, 2183, 2183, 2183,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2183, 2183, 2183, 2183, 2183,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 2183,    0, 2183,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1408,    0,    0,    0,    0,
    0,    0,    0,    0, 1409,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1410,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1411,    0,    0,    0,    0,    0,    0, 1412,    0,
    0,    0,    0,    0,    0,    0, 1413,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1414,    0, 1415,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1416,    0, 1417, 1418, 1419,
    0,    0,    0,    0, 1420,    0,    0,    0,    0, 2183,
    0, 2183,    0,    0, 2183,    0, 2183, 2183, 2183,    0,
 2183, 2183,    0, 2183,    0,    0,    0,    0,    0, 2183,
    0, 2183, 2183, 2183, 2183,    0, 2183,    0,    0, 2183,
    0,    0, 2183,    0,    0, 1421,    0,    0,    0, 2183,
    0, 2183, 2183,    0, 2183, 2183, 2183, 2183, 2183,    0,
    0, 2183,    0,    0, 2183,    0, 2183, 2183,    0,    0,
 2183, 2183, 2183,    0, 2183, 2183, 2183, 2183, 2183, 2183,
 2183,    0,    0,    0, 2183, 2183, 2183, 2183, 2183,    0,
    0,    0, 2183,    0,    0,    0,    0,    0,    0,    0,
 2183, 2183, 2183, 2183, 2183, 2183,    0,    0,    0, 2183,
 2183,    0, 2183,    0, 2183,    0, 2183, 2183, 2183,    0,
 2183,    0, 2183, 2183, 2183,    0, 2183, 2183, 2183, 2183,
 2183,    0, 2183, 2183,    0,    0,    0,    0, 2183,    0,
 2183, 2183,    0,    0, 2183, 2183, 2183,    0,    0, 2183,
    0, 2183,    0,    0, 2183, 2183, 2183, 2183,    0, 2183,
    0, 2183,    0, 2183, 2183, 2183, 2183, 2183,    0,    0,
 2183, 2183, 2183,    0, 2183, 2183,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 2183, 2183, 2183,    0, 2183,
 2183, 2183,    0,    0,    0,    0,    0, 2183,    0,    0,
    0, 2183,    0, 2183, 2183, 2183, 2183,    0, 2183,    0,
 2183, 2183, 2183, 2183, 2183, 2183, 2183, 2183,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 2183, 2183,    0,
 2183, 2183, 2183,    0,    0, 1422,    0, 2183, 2183, 2183,
 2183, 2183,    0, 2183,    0,    0, 2183, 2183,    0,    0,
    0, 2183, 2183,    0,    0, 2183, 2183, 2183, 2183, 2183,
 2183, 2183, 2183, 2183,    0, 2183,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 2183, 2183, 2183, 2183, 2183,
 2183, 2183, 2183, 2183, 2183, 2183, 2183, 2183, 2183, 2183,
 2183, 2183, 2183, 2183, 2183, 2183,    0, 2183, 2183, 2183,
 2183, 2183, 2183,    0, 2183, 2183, 2183, 2183,    0, 2183,
    0, 2183, 2183, 2183, 2183, 2183, 2183, 2183, 2183, 2183,
 2183, 2183, 2183, 2183, 2183, 2183, 2183, 2183, 2183,    0,
 2183, 2183, 2183, 2183,    0, 2183, 2183, 2183, 2183, 2183,
    0, 2183, 2183, 2183, 2183, 2183, 2183, 2183, 2183, 2183,
 2183, 2183, 2183, 2183, 2183, 2183, 2183, 2183, 2183, 2183,
 2183, 2183, 2183, 2183, 2183, 2183, 2183, 2183, 2183, 2183,
    0, 2183, 2183,    0,    0,    0, 2183, 2183, 2183, 2183,
 2183, 2183,    0,    0, 2183, 2183, 2183, 2183, 2183, 2183,
 2183, 2183, 2183, 2938,    0,    0,    0,    0, 2183,    0,
 2183, 2183, 2183, 2183, 2183, 2183,    0, 2183,    0,    0,
    0,    0, 1424, 1423, 1425,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 2183, 2183,    0, 2183, 2183, 2183,
 2183, 2183, 2183, 2183, 2183, 2183,    0, 2183, 2183, 2183,
 2183, 2183, 2183, 2183, 2183, 2183, 2183, 2183, 2183,    0,
 2183, 2183, 2183,    0, 2183,    0,    0,    0, 2183, 2183,
 2183, 2183, 2183, 2183, 2183, 2183, 2183, 2183, 2183, 2183,
 2183, 2184, 2183, 2183,    0,    0,    0,    0,    0, 2183,
 2183, 2183, 2183, 2183, 2183, 2183, 2183, 2183, 2183, 2183,
 2183, 2183, 2183, 2183, 2183, 2183, 2183,    0,    0, 2183,
    0, 2183, 2183, 2183, 2183, 2183, 2183,    0,    0,    0,
    0, 1947, 2184, 2184, 2184, 2184, 2184, 2184, 2184,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 2184,
 2184, 2184, 2184, 2184,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 2184,    0, 2184,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1408,    0,    0,    0,    0,    0,
    0,    0,    0, 1409,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1410,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1411,    0,    0,    0,    0,    0,    0, 1412,    0,    0,
    0,    0,    0,    0,    0, 1413,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1414,    0, 1415,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1416,    0, 1417, 1418, 1419,    0,
    0,    0,    0, 1420,    0,    0,    0,    0, 2184,    0,
 2184,    0,    0, 2184,    0, 2184, 2184, 2184,    0, 2184,
 2184,    0, 2184,    0,    0,    0,    0,    0, 2184,    0,
 2184, 2184, 2184, 2184,    0, 2184,    0,    0, 2184,    0,
    0, 2184,    0,    0, 1421,    0,    0,    0, 2184,    0,
 2184, 2184,    0, 2184, 2184, 2184, 2184, 2184,    0,    0,
 2184,    0,    0, 2184,    0, 2184, 2184,    0,    0, 2184,
 2184, 2184,    0, 2184, 2184, 2184, 2184, 2184, 2184, 2184,
    0,    0,    0, 2184, 2184, 2184, 2184, 2184,    0,    0,
    0, 2184,    0,    0,    0,    0,    0,    0,    0, 2184,
 2184, 2184, 2184, 2184, 2184,    0,    0,    0, 2184, 2184,
    0, 2184,    0, 2184,    0, 2184, 2184, 2184,    0, 2184,
    0, 2184, 2184, 2184,    0, 2184, 2184, 2184, 2184, 2184,
    0, 2184, 2184,    0,    0,    0,    0, 2184,    0, 2184,
 2184,    0,    0, 2184, 2184, 2184,    0,    0, 2184,    0,
 2184,    0,    0, 2184, 2184, 2184, 2184,    0, 2184,    0,
 2184,    0, 2184, 2184, 2184, 2184, 2184,    0,    0, 2184,
 2184, 2184,    0, 2184, 2184,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 2184, 2184, 2184,    0, 2184, 2184,
 2184,    0,    0,    0,    0,    0, 2184,    0,    0,    0,
 2184,    0, 2184, 2184, 2184, 2184,    0, 2184,    0, 2184,
 2184, 2184, 2184, 2184, 2184, 2184, 2184,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2184, 2184,    0, 2184,
 2184, 2184,    0,    0, 1422,    0, 2184, 2184, 2184, 2184,
 2184,    0, 2184,    0,    0, 2184, 2184,    0,    0,    0,
 2184, 2184,    0,    0, 2184, 2184, 2184, 2184, 2184, 2184,
 2184, 2184, 2184,    0, 2184,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 2184, 2184, 2184, 2184, 2184, 2184,
 2184, 2184, 2184, 2184, 2184, 2184, 2184, 2184, 2184, 2184,
 2184, 2184, 2184, 2184, 2184,    0, 2184, 2184, 2184, 2184,
 2184, 2184,    0, 2184, 2184, 2184, 2184,    0, 2184,    0,
 2184, 2184, 2184, 2184, 2184, 2184, 2184, 2184, 2184, 2184,
 2184, 2184, 2184, 2184, 2184, 2184, 2184, 2184,    0, 2184,
 2184, 2184, 2184,    0, 2184, 2184, 2184, 2184, 2184,    0,
 2184, 2184, 2184, 2184, 2184, 2184, 2184, 2184, 2184, 2184,
 2184, 2184, 2184, 2184, 2184, 2184, 2184, 2184, 2184, 2184,
 2184, 2184, 2184, 2184, 2184, 2184, 2184, 2184, 2184,    0,
 2184, 2184,    0,    0,    0, 2184, 2184, 2184, 2184, 2184,
 2184,    0,    0, 2184, 2184, 2184, 2184, 2184, 2184, 2184,
 2184, 2184, 2939,    0,    0,    0,    0, 2184,    0, 2184,
 2184, 2184, 2184, 2184, 2184,    0, 2184,    0,    0,    0,
    0, 1424, 1423, 1425,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 2184, 2184,    0, 2184, 2184, 2184, 2184,
 2184, 2184, 2184, 2184, 2184,    0, 2184, 2184, 2184, 2184,
 2184, 2184, 2184, 2184, 2184, 2184, 2184, 2184,    0, 2184,
 2184, 2184,    0, 2184,    0,    0,    0, 2184, 2184, 2184,
 2184, 2184, 2184, 2184, 2184, 2184, 2184, 2184, 2184, 2184,
 2186, 2184, 2184,    0,    0,    0,    0,    0, 2184, 2184,
 2184, 2184, 2184, 2184, 2184, 2184, 2184, 2184, 2184, 2184,
 2184, 2184, 2184, 2184, 2184, 2184,    0,    0, 2184,    0,
 2184, 2184, 2184, 2184, 2184, 2184,    0,    0,    0,    0,
 1948, 2186, 2186, 2186, 2186, 2186, 2186, 2186,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2186, 2186,
 2186, 2186, 2186,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 2186,    0, 2186,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1408,    0,    0,    0,    0,    0,    0,
    0,    0, 1409,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1410,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1411,
    0,    0,    0,    0,    0,    0, 1412,    0,    0,    0,
    0,    0,    0,    0, 1413,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1414,    0, 1415,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1416,    0, 1417, 1418, 1419,    0,    0,
    0,    0, 1420,    0,    0,    0,    0, 2186,    0, 2186,
    0,    0, 2186,    0, 2186, 2186, 2186,    0, 2186, 2186,
    0, 2186,    0,    0,    0,    0,    0, 2186,    0, 2186,
 2186, 2186, 2186,    0, 2186,    0,    0, 2186,    0,    0,
 2186,    0,    0, 1421,    0,    0,    0, 2186,    0, 2186,
 2186,    0, 2186, 2186, 2186, 2186, 2186,    0,    0, 2186,
    0,    0, 2186,    0, 2186, 2186,    0,    0, 2186, 2186,
 2186,    0, 2186, 2186, 2186, 2186, 2186, 2186, 2186,    0,
    0,    0, 2186, 2186, 2186, 2186, 2186,    0,    0,    0,
 2186,    0,    0,    0,    0,    0,    0,    0, 2186, 2186,
 2186, 2186, 2186, 2186,    0,    0,    0, 2186, 2186,    0,
 2186,    0, 2186,    0, 2186, 2186, 2186,    0, 2186,    0,
 2186, 2186, 2186,    0, 2186, 2186, 2186, 2186, 2186,    0,
 2186, 2186,    0,    0,    0,    0, 2186,    0, 2186, 2186,
    0,    0, 2186, 2186, 2186,    0,    0, 2186,    0, 2186,
    0,    0, 2186, 2186, 2186, 2186,    0, 2186,    0, 2186,
    0, 2186, 2186, 2186, 2186, 2186,    0,    0, 2186, 2186,
 2186,    0, 2186, 2186,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 2186, 2186, 2186,    0, 2186, 2186, 2186,
    0,    0,    0,    0,    0, 2186,    0,    0,    0, 2186,
    0, 2186, 2186, 2186, 2186,    0, 2186,    0, 2186, 2186,
 2186, 2186, 2186, 2186, 2186, 2186,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 2186, 2186,    0, 2186, 2186,
 2186,    0,    0, 1422,    0, 2186, 2186, 2186, 2186, 2186,
    0, 2186,    0,    0, 2186, 2186,    0,    0,    0, 2186,
 2186,    0,    0, 2186, 2186, 2186, 2186, 2186, 2186, 2186,
 2186, 2186,    0, 2186,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 2186, 2186, 2186, 2186, 2186, 2186, 2186,
 2186, 2186, 2186, 2186, 2186, 2186, 2186, 2186, 2186, 2186,
 2186, 2186, 2186, 2186,    0, 2186, 2186, 2186, 2186, 2186,
 2186,    0, 2186, 2186, 2186, 2186,    0, 2186,    0, 2186,
 2186, 2186, 2186, 2186, 2186, 2186, 2186, 2186, 2186, 2186,
 2186, 2186, 2186, 2186, 2186, 2186, 2186,    0, 2186, 2186,
 2186, 2186,    0, 2186, 2186, 2186, 2186, 2186,    0, 2186,
 2186, 2186, 2186, 2186, 2186, 2186, 2186, 2186, 2186, 2186,
 2186, 2186, 2186, 2186, 2186, 2186, 2186, 2186, 2186, 2186,
 2186, 2186, 2186, 2186, 2186, 2186, 2186, 2186,    0, 2186,
 2186,    0,    0,    0, 2186, 2186, 2186, 2186, 2186, 2186,
    0,    0, 2186, 2186, 2186, 2186, 2186, 2186, 2186, 2186,
 2186, 2940,    0,    0,    0,    0, 2186,    0, 2186, 2186,
 2186, 2186, 2186, 2186,    0, 2186,    0,    0,    0,    0,
 1424, 1423, 1425,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 2186, 2186,    0, 2186, 2186, 2186, 2186, 2186,
 2186, 2186, 2186, 2186,    0, 2186, 2186, 2186, 2186, 2186,
 2186, 2186, 2186, 2186, 2186, 2186, 2186,    0, 2186, 2186,
 2186,    0, 2186,    0,    0,    0, 2186, 2186, 2186, 2186,
 2186, 2186, 2186, 2186, 2186, 2186, 2186, 2186, 2186, 2187,
 2186, 2186,    0,    0,    0,    0,    0, 2186, 2186, 2186,
 2186, 2186, 2186, 2186, 2186, 2186, 2186, 2186, 2186, 2186,
 2186, 2186, 2186, 2186, 2186,    0,    0, 2186,    0, 2186,
 2186, 2186, 2186, 2186, 2186,    0,    0,    0,    0, 1949,
 2187, 2187, 2187, 2187, 2187, 2187, 2187,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 2187, 2187, 2187,
 2187, 2187,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2187,    0, 2187,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1408,    0,    0,    0,    0,    0,    0,    0,
    0, 1409,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1410,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1411,    0,
    0,    0,    0,    0,    0, 1412,    0,    0,    0,    0,
    0,    0,    0, 1413,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1414,    0, 1415,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1416,    0, 1417, 1418, 1419,    0,    0,    0,
    0, 1420,    0,    0,    0,    0, 2187,    0, 2187,    0,
    0, 2187,    0, 2187, 2187, 2187,    0, 2187, 2187,    0,
 2187,    0,    0,    0,    0,    0, 2187,    0, 2187, 2187,
 2187, 2187,    0, 2187,    0,    0, 2187,    0,    0, 2187,
    0,    0, 1421,    0,    0,    0, 2187,    0, 2187, 2187,
    0, 2187, 2187, 2187, 2187, 2187,    0,    0, 2187,    0,
    0, 2187,    0, 2187, 2187,    0,    0, 2187, 2187, 2187,
    0, 2187, 2187, 2187, 2187, 2187, 2187, 2187,    0,    0,
    0, 2187, 2187, 2187, 2187, 2187,    0,    0,    0, 2187,
    0,    0,    0,    0,    0,    0,    0, 2187, 2187, 2187,
 2187, 2187, 2187,    0,    0,    0, 2187, 2187,    0, 2187,
    0, 2187,    0, 2187, 2187, 2187,    0, 2187,    0, 2187,
 2187, 2187,    0, 2187, 2187, 2187, 2187, 2187,    0, 2187,
 2187,    0,    0,    0,    0, 2187,    0, 2187, 2187,    0,
    0, 2187, 2187, 2187,    0,    0, 2187,    0, 2187,    0,
    0, 2187, 2187, 2187, 2187,    0, 2187,    0, 2187,    0,
 2187, 2187, 2187, 2187, 2187,    0,    0, 2187, 2187, 2187,
    0, 2187, 2187,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 2187, 2187, 2187,    0, 2187, 2187, 2187,    0,
    0,    0,    0,    0, 2187,    0,    0,    0, 2187,    0,
 2187, 2187, 2187, 2187,    0, 2187,    0, 2187, 2187, 2187,
 2187, 2187, 2187, 2187, 2187,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 2187, 2187,    0, 2187, 2187, 2187,
    0,    0, 1422,    0, 2187, 2187, 2187, 2187, 2187,    0,
 2187,    0,    0, 2187, 2187,    0,    0,    0, 2187, 2187,
    0,    0, 2187, 2187, 2187, 2187, 2187, 2187, 2187, 2187,
 2187,    0, 2187,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 2187, 2187, 2187, 2187, 2187, 2187, 2187, 2187,
 2187, 2187, 2187, 2187, 2187, 2187, 2187, 2187, 2187, 2187,
 2187, 2187, 2187,    0, 2187, 2187, 2187, 2187, 2187, 2187,
    0, 2187, 2187, 2187, 2187,    0, 2187,    0, 2187, 2187,
 2187, 2187, 2187, 2187, 2187, 2187, 2187, 2187, 2187, 2187,
 2187, 2187, 2187, 2187, 2187, 2187,    0, 2187, 2187, 2187,
 2187,    0, 2187, 2187, 2187, 2187, 2187,    0, 2187, 2187,
 2187, 2187, 2187, 2187, 2187, 2187, 2187, 2187, 2187, 2187,
 2187, 2187, 2187, 2187, 2187, 2187, 2187, 2187, 2187, 2187,
 2187, 2187, 2187, 2187, 2187, 2187, 2187,    0, 2187, 2187,
    0,    0,    0, 2187, 2187, 2187, 2187, 2187, 2187,    0,
    0, 2187, 2187, 2187, 2187, 2187, 2187, 2187, 2187, 2187,
 2941,    0,    0,    0,    0, 2187,    0, 2187, 2187, 2187,
 2187, 2187, 2187,    0, 2187,    0,    0,    0,    0, 1424,
 1423, 1425,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 2187, 2187,    0, 2187, 2187, 2187, 2187, 2187, 2187,
 2187, 2187, 2187,    0, 2187, 2187, 2187, 2187, 2187, 2187,
 2187, 2187, 2187, 2187, 2187, 2187,    0, 2187, 2187, 2187,
    0, 2187,    0,    0,    0, 2187, 2187, 2187, 2187, 2187,
 2187, 2187, 2187, 2187, 2187, 2187, 2187, 2187, 2188, 2187,
 2187,    0,    0,    0,    0,    0, 2187, 2187, 2187, 2187,
 2187, 2187, 2187, 2187, 2187, 2187, 2187, 2187, 2187, 2187,
 2187, 2187, 2187, 2187,    0,    0, 2187,    0, 2187, 2187,
 2187, 2187, 2187, 2187,    0,    0,    0,    0, 1950, 2188,
 2188, 2188, 2188, 2188, 2188, 2188,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2188, 2188, 2188, 2188,
 2188,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 2188,
    0, 2188,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1408,    0,    0,    0,    0,    0,    0,    0,    0,
 1409,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1410,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1411,    0,    0,
    0,    0,    0,    0, 1412,    0,    0,    0,    0,    0,
    0,    0, 1413,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1414,
    0, 1415,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1416,    0, 1417, 1418, 1419,    0,    0,    0,    0,
 1420,    0,    0,    0,    0, 2188,    0, 2188,    0,    0,
 2188,    0, 2188, 2188, 2188,    0, 2188, 2188,    0, 2188,
    0,    0,    0,    0,    0, 2188,    0, 2188, 2188, 2188,
 2188,    0, 2188,    0,    0, 2188,    0,    0, 2188,    0,
    0, 1421,    0,    0,    0, 2188,    0, 2188, 2188,    0,
 2188, 2188, 2188, 2188, 2188,    0,    0, 2188,    0,    0,
 2188,    0, 2188, 2188,    0,    0, 2188, 2188, 2188,    0,
 2188, 2188, 2188, 2188, 2188, 2188, 2188,    0,    0,    0,
 2188, 2188, 2188, 2188, 2188,    0,    0,    0, 2188,    0,
    0,    0,    0,    0,    0,    0, 2188, 2188, 2188, 2188,
 2188, 2188,    0,    0,    0, 2188, 2188,    0, 2188,    0,
 2188,    0, 2188, 2188, 2188,    0, 2188,    0, 2188, 2188,
 2188,    0, 2188, 2188, 2188, 2188, 2188,    0, 2188, 2188,
    0,    0,    0,    0, 2188,    0, 2188, 2188,    0,    0,
 2188, 2188, 2188,    0,    0, 2188,    0, 2188,    0,    0,
 2188, 2188, 2188, 2188,    0, 2188,    0, 2188,    0, 2188,
 2188, 2188, 2188, 2188,    0,    0, 2188, 2188, 2188,    0,
 2188, 2188,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 2188, 2188, 2188,    0, 2188, 2188, 2188,    0,    0,
    0,    0,    0, 2188,    0,    0,    0, 2188,    0, 2188,
 2188, 2188, 2188,    0, 2188,    0, 2188, 2188, 2188, 2188,
 2188, 2188, 2188, 2188,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 2188, 2188,    0, 2188, 2188, 2188,    0,
    0, 1422,    0, 2188, 2188, 2188, 2188, 2188,    0, 2188,
    0,    0, 2188, 2188,    0,    0,    0, 2188, 2188,    0,
    0, 2188, 2188, 2188, 2188, 2188, 2188, 2188, 2188, 2188,
    0, 2188,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 2188, 2188, 2188, 2188, 2188, 2188, 2188, 2188, 2188,
 2188, 2188, 2188, 2188, 2188, 2188, 2188, 2188, 2188, 2188,
 2188, 2188,    0, 2188, 2188, 2188, 2188, 2188, 2188,    0,
 2188, 2188, 2188, 2188,    0, 2188,    0, 2188, 2188, 2188,
 2188, 2188, 2188, 2188, 2188, 2188, 2188, 2188, 2188, 2188,
 2188, 2188, 2188, 2188, 2188,    0, 2188, 2188, 2188, 2188,
    0, 2188, 2188, 2188, 2188, 2188,    0, 2188, 2188, 2188,
 2188, 2188, 2188, 2188, 2188, 2188, 2188, 2188, 2188, 2188,
 2188, 2188, 2188, 2188, 2188, 2188, 2188, 2188, 2188, 2188,
 2188, 2188, 2188, 2188, 2188, 2188,    0, 2188, 2188,    0,
    0,    0, 2188, 2188, 2188, 2188, 2188, 2188,    0,    0,
 2188, 2188, 2188, 2188, 2188, 2188, 2188, 2188, 2188, 2942,
    0,    0,    0,    0, 2188,    0, 2188, 2188, 2188, 2188,
 2188, 2188,    0, 2188,    0,    0,    0,    0, 1424, 1423,
 1425,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2188, 2188,    0, 2188, 2188, 2188, 2188, 2188, 2188, 2188,
 2188, 2188,    0, 2188, 2188, 2188, 2188, 2188, 2188, 2188,
 2188, 2188, 2188, 2188, 2188,    0, 2188, 2188, 2188,    0,
 2188,    0,    0,    0, 2188, 2188, 2188, 2188, 2188, 2188,
 2188, 2188, 2188, 2188, 2188, 2188, 2188, 2190, 2188, 2188,
    0,    0,    0,    0,    0, 2188, 2188, 2188, 2188, 2188,
 2188, 2188, 2188, 2188, 2188, 2188, 2188, 2188, 2188, 2188,
 2188, 2188, 2188,    0,    0, 2188,    0, 2188, 2188, 2188,
 2188, 2188, 2188,    0,    0,    0,    0, 1951, 2190, 2190,
 2190, 2190, 2190, 2190, 2190,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 2190, 2190, 2190, 2190, 2190,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2190,    0,
 2190,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1408,    0,    0,    0,    0,    0,    0,    0,    0, 1409,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1410,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1411,    0,    0,    0,
    0,    0,    0, 1412,    0,    0,    0,    0,    0,    0,
    0, 1413,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1414,    0,
 1415,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1416,    0, 1417, 1418, 1419,    0,    0,    0,    0, 1420,
    0,    0,    0,    0, 2190,    0, 2190,    0,    0, 2190,
    0, 2190, 2190, 2190,    0, 2190, 2190,    0, 2190,    0,
    0,    0,    0,    0, 2190,    0, 2190, 2190, 2190, 2190,
    0, 2190,    0,    0, 2190,    0,    0, 2190,    0,    0,
 1421,    0,    0,    0, 2190,    0, 2190, 2190,    0, 2190,
 2190, 2190, 2190, 2190,    0,    0, 2190,    0,    0, 2190,
    0, 2190, 2190,    0,    0, 2190, 2190, 2190,    0, 2190,
 2190, 2190, 2190, 2190, 2190, 2190,    0,    0,    0, 2190,
 2190, 2190, 2190, 2190,    0,    0,    0, 2190,    0,    0,
    0,    0,    0,    0,    0, 2190, 2190, 2190, 2190, 2190,
 2190,    0,    0,    0, 2190, 2190,    0, 2190,    0, 2190,
    0, 2190, 2190, 2190,    0, 2190,    0, 2190, 2190, 2190,
    0, 2190, 2190, 2190, 2190, 2190,    0, 2190, 2190,    0,
    0,    0,    0, 2190,    0, 2190, 2190,    0,    0, 2190,
 2190, 2190,    0,    0, 2190,    0, 2190,    0,    0, 2190,
 2190, 2190, 2190,    0, 2190,    0, 2190,    0, 2190, 2190,
 2190, 2190, 2190,    0,    0, 2190, 2190, 2190,    0, 2190,
 2190,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2190, 2190, 2190,    0, 2190, 2190, 2190,    0,    0,    0,
    0,    0, 2190,    0,    0,    0, 2190,    0, 2190, 2190,
 2190, 2190,    0, 2190,    0, 2190, 2190, 2190, 2190, 2190,
 2190, 2190, 2190,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 2190, 2190,    0, 2190, 2190, 2190,    0,    0,
 1422,    0, 2190, 2190, 2190, 2190, 2190,    0, 2190,    0,
    0, 2190, 2190,    0,    0,    0, 2190, 2190,    0,    0,
 2190, 2190, 2190, 2190, 2190, 2190, 2190, 2190, 2190,    0,
 2190,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2190, 2190, 2190, 2190, 2190, 2190, 2190, 2190, 2190, 2190,
 2190, 2190, 2190, 2190, 2190, 2190, 2190, 2190, 2190, 2190,
 2190,    0, 2190, 2190, 2190, 2190, 2190, 2190,    0, 2190,
 2190, 2190, 2190,    0, 2190,    0, 2190, 2190, 2190, 2190,
 2190, 2190, 2190, 2190, 2190, 2190, 2190, 2190, 2190, 2190,
 2190, 2190, 2190, 2190,    0, 2190, 2190, 2190, 2190,    0,
 2190, 2190, 2190, 2190, 2190,    0, 2190, 2190, 2190, 2190,
 2190, 2190, 2190, 2190, 2190, 2190, 2190, 2190, 2190, 2190,
 2190, 2190, 2190, 2190, 2190, 2190, 2190, 2190, 2190, 2190,
 2190, 2190, 2190, 2190, 2190,    0, 2190, 2190,    0,    0,
    0, 2190, 2190, 2190, 2190, 2190, 2190,    0,    0, 2190,
 2190, 2190, 2190, 2190, 2190, 2190, 2190, 2190, 2943,    0,
    0,    0,    0, 2190,    0, 2190, 2190, 2190, 2190, 2190,
 2190,    0, 2190,    0,    0,    0,    0, 1424, 1423, 1425,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 2190,
 2190,    0, 2190, 2190, 2190, 2190, 2190, 2190, 2190, 2190,
 2190,    0, 2190, 2190, 2190, 2190, 2190, 2190, 2190, 2190,
 2190, 2190, 2190, 2190,    0, 2190, 2190, 2190,    0, 2190,
    0,    0,    0, 2190, 2190, 2190, 2190, 2190, 2190, 2190,
 2190, 2190, 2190, 2190, 2190, 2190, 2192, 2190, 2190,    0,
    0,    0,    0,    0, 2190, 2190, 2190, 2190, 2190, 2190,
 2190, 2190, 2190, 2190, 2190, 2190, 2190, 2190, 2190, 2190,
 2190, 2190,    0,    0, 2190,    0, 2190, 2190, 2190, 2190,
 2190, 2190,    0,    0,    0,    0, 1913, 2192, 2192, 2192,
 2192, 2192, 2192, 2192,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 2192, 2192, 2192, 2192, 2192,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 2192,    0, 2192,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1408,
    0,    0,    0,    0,    0,    0,    0,    0, 1409,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1410,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1411,    0,    0,    0,    0,
    0,    0, 1412,    0,    0,    0,    0,    0,    0,    0,
 1413,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1414,    0, 1415,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1416,
    0, 1417, 1418, 1419,    0,    0,    0,    0, 1420,    0,
    0,    0,    0, 2192,    0, 2192,    0,    0, 2192,    0,
 2192, 2192, 2192,    0, 2192, 2192,    0, 2192,    0,    0,
    0,    0,    0, 2192,    0, 2192, 2192, 2192, 2192,    0,
 2192,    0,    0, 2192,    0,    0, 2192,    0,    0, 1421,
    0,    0,    0, 2192,    0, 2192, 2192,    0, 2192, 2192,
 2192, 2192, 2192,    0,    0, 2192,    0,    0, 2192,    0,
 2192, 2192,    0,    0, 2192, 2192, 2192,    0, 2192, 2192,
 2192, 2192, 2192, 2192, 2192,    0,    0,    0, 2192, 2192,
 2192, 2192, 2192,    0,    0,    0, 2192,    0,    0,    0,
    0,    0,    0,    0, 2192, 2192, 2192, 2192, 2192, 2192,
    0,    0,    0, 2192, 2192,    0, 2192,    0, 2192,    0,
 2192, 2192, 2192,    0, 2192,    0, 2192, 2192, 2192,    0,
 2192, 2192, 2192, 2192, 2192,    0, 2192, 2192,    0,    0,
    0,    0, 2192,    0, 2192, 2192,    0,    0, 2192, 2192,
 2192,    0,    0, 2192,    0, 2192,    0,    0, 2192, 2192,
 2192, 2192,    0, 2192,    0, 2192,    0, 2192, 2192, 2192,
 2192, 2192,    0,    0, 2192, 2192, 2192,    0, 2192, 2192,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 2192,
 2192, 2192,    0, 2192, 2192, 2192,    0,    0,    0,    0,
    0, 2192,    0,    0,    0, 2192,    0, 2192, 2192, 2192,
 2192,    0, 2192,    0, 2192, 2192, 2192, 2192, 2192, 2192,
 2192, 2192,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 2192, 2192,    0, 2192, 2192, 2192,    0,    0, 1422,
    0, 2192, 2192, 2192, 2192, 2192,    0, 2192,    0,    0,
 2192, 2192,    0,    0,    0, 2192, 2192,    0,    0, 2192,
 2192, 2192, 2192, 2192, 2192, 2192, 2192, 2192,    0, 2192,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 2192,
 2192, 2192, 2192, 2192, 2192, 2192, 2192, 2192, 2192, 2192,
 2192, 2192, 2192, 2192, 2192, 2192, 2192, 2192, 2192, 2192,
    0, 2192, 2192, 2192, 2192, 2192, 2192,    0, 2192, 2192,
 2192, 2192,    0, 2192,    0, 2192, 2192, 2192, 2192, 2192,
 2192, 2192, 2192, 2192, 2192, 2192, 2192, 2192, 2192, 2192,
 2192, 2192, 2192,    0, 2192, 2192, 2192, 2192,    0, 2192,
 2192, 2192, 2192, 2192,    0, 2192, 2192, 2192, 2192, 2192,
 2192, 2192, 2192, 2192, 2192, 2192, 2192, 2192, 2192, 2192,
 2192, 2192, 2192, 2192, 2192, 2192, 2192, 2192, 2192, 2192,
 2192, 2192, 2192, 2192,    0, 2192, 2192,    0,    0,    0,
 2192, 2192, 2192, 2192, 2192, 2192,    0,    0, 2192, 2192,
 2192, 2192, 2192, 2192, 2192, 2192, 2192, 2944,    0,    0,
    0,    0, 2192,    0, 2192, 2192, 2192, 2192, 2192, 2192,
    0, 2192,    0,    0,    0,    0, 1424, 1423, 1425,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2192, 2192,
    0, 2192, 2192, 2192, 2192, 2192, 2192, 2192, 2192, 2192,
    0, 2192, 2192, 2192, 2192, 2192, 2192, 2192, 2192, 2192,
 2192, 2192, 2192,    0, 2192, 2192, 2192,    0, 2192,    0,
    0,    0, 2192, 2192, 2192, 2192, 2192, 2192, 2192, 2192,
 2192, 2192, 2192, 2192, 2192, 2198, 2192, 2192,    0,    0,
    0,    0,    0, 2192, 2192, 2192, 2192, 2192, 2192, 2192,
 2192, 2192, 2192, 2192, 2192, 2192, 2192, 2192, 2192, 2192,
 2192,    0,    0, 2192,    0, 2192, 2192, 2192, 2192, 2192,
 2192,    0,    0,    0,    0, 1952, 2198, 2198, 2198, 2198,
 2198, 2198, 2198,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 2198, 2198, 2198, 2198, 2198,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2198,    0, 2198,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1408,    0,
    0,    0,    0,    0,    0,    0,    0, 1409,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1410,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1411,    0,    0,    0,    0,    0,
    0, 1412,    0,    0,    0,    0,    0,    0,    0, 1413,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1414,    0, 1415,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1416,    0,
 1417, 1418, 1419,    0,    0,    0,    0, 1420,    0,    0,
    0,    0, 2198,    0, 2198,    0,    0, 2198,    0, 2198,
 2198, 2198,    0, 2198, 2198,    0, 2198,    0,    0,    0,
    0,    0, 2198,    0, 2198, 2198, 2198, 2198,    0, 2198,
    0,    0, 2198,    0,    0, 2198,    0,    0, 1421,    0,
    0,    0, 2198,    0, 2198, 2198,    0, 2198, 2198, 2198,
 2198, 2198,    0,    0, 2198,    0,    0, 2198,    0, 2198,
 2198,    0,    0, 2198, 2198, 2198,    0, 2198, 2198, 2198,
 2198, 2198, 2198, 2198,    0,    0,    0, 2198, 2198, 2198,
 2198, 2198,    0,    0,    0, 2198,    0,    0,    0,    0,
    0,    0,    0, 2198, 2198, 2198, 2198, 2198, 2198,    0,
    0,    0, 2198, 2198,    0, 2198,    0, 2198,    0, 2198,
 2198, 2198,    0, 2198,    0, 2198, 2198, 2198,    0, 2198,
 2198, 2198, 2198, 2198,    0, 2198, 2198,    0,    0,    0,
    0, 2198,    0, 2198, 2198,    0,    0, 2198, 2198, 2198,
    0,    0, 2198,    0, 2198,    0,    0, 2198, 2198, 2198,
 2198,    0, 2198,    0, 2198,    0, 2198, 2198, 2198, 2198,
 2198,    0,    0, 2198, 2198, 2198,    0, 2198, 2198,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2198, 2198,
 2198,    0, 2198, 2198, 2198,    0,    0,    0,    0,    0,
 2198,    0,    0,    0, 2198,    0, 2198, 2198, 2198, 2198,
    0, 2198,    0, 2198, 2198, 2198, 2198, 2198, 2198, 2198,
 2198,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2198, 2198,    0, 2198, 2198, 2198,    0,    0, 1422,    0,
 2198, 2198, 2198, 2198, 2198,    0, 2198,    0,    0, 2198,
 2198,    0,    0,    0, 2198, 2198,    0,    0, 2198, 2198,
 2198, 2198, 2198, 2198, 2198, 2198, 2198,    0, 2198,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2198, 2198,
 2198, 2198, 2198, 2198, 2198, 2198, 2198, 2198, 2198, 2198,
 2198, 2198, 2198, 2198, 2198, 2198, 2198, 2198, 2198,    0,
 2198, 2198, 2198, 2198, 2198, 2198,    0, 2198, 2198, 2198,
 2198,    0, 2198,    0, 2198, 2198, 2198, 2198, 2198, 2198,
 2198, 2198, 2198, 2198, 2198, 2198, 2198, 2198, 2198, 2198,
 2198, 2198,    0, 2198, 2198, 2198, 2198,    0, 2198, 2198,
 2198, 2198, 2198,    0, 2198, 2198, 2198, 2198, 2198, 2198,
 2198, 2198, 2198, 2198, 2198, 2198, 2198, 2198, 2198, 2198,
 2198, 2198, 2198, 2198, 2198, 2198, 2198, 2198, 2198, 2198,
 2198, 2198, 2198,    0, 2198, 2198,    0,    0,    0, 2198,
 2198, 2198, 2198, 2198, 2198,    0,    0, 2198, 2198, 2198,
 2198, 2198, 2198, 2198, 2198, 2198, 2945,    0,    0,    0,
    0, 2198,    0, 2198, 2198, 2198, 2198, 2198, 2198,    0,
 2198,    0,    0,    0,    0, 1424, 1423, 1425,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 2198, 2198,    0,
 2198, 2198, 2198, 2198, 2198, 2198, 2198, 2198, 2198,    0,
 2198, 2198, 2198, 2198, 2198, 2198, 2198, 2198, 2198, 2198,
 2198, 2198,    0, 2198, 2198, 2198,    0, 2198,    0,    0,
    0, 2198, 2198, 2198, 2198, 2198, 2198, 2198, 2198, 2198,
 2198, 2198, 2198, 2198, 2205, 2198, 2198,    0,    0,    0,
    0,    0, 2198, 2198, 2198, 2198, 2198, 2198, 2198, 2198,
 2198, 2198, 2198, 2198, 2198, 2198, 2198, 2198, 2198, 2198,
    0,    0, 2198,    0, 2198, 2198, 2198, 2198, 2198, 2198,
    0,    0,    0,    0, 1907, 2205, 2205, 2205, 2205, 2205,
 2205, 2205,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 2205, 2205, 2205, 2205, 2205,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 2205,    0, 2205,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1408,    0,    0,
    0,    0,    0,    0,    0,    0, 1409,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1410,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1411,    0,    0,    0,    0,    0,    0,
 1412,    0,    0,    0,    0,    0,    0,    0, 1413,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1414,    0, 1415,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1416,    0, 1417,
 1418, 1419,    0,    0,    0,    0, 1420,    0,    0,    0,
    0, 2205,    0, 2205,    0,    0, 2205,    0, 2205, 2205,
 2205,    0, 2205, 2205,    0, 2205,    0,    0,    0,    0,
    0, 2205,    0, 2205, 2205, 2205, 2205,    0, 2205,    0,
    0, 2205,    0,    0, 2205,    0,    0, 1421,    0,    0,
    0, 2205,    0, 2205, 2205,    0, 2205, 2205, 2205, 2205,
 2205,    0,    0, 2205,    0,    0, 2205,    0, 2205, 2205,
    0,    0, 2205, 2205, 2205,    0, 2205, 2205, 2205, 2205,
 2205, 2205, 2205,    0,    0,    0, 2205, 2205, 2205, 2205,
 2205,    0,    0,    0, 2205,    0,    0,    0,    0,    0,
    0,    0, 2205, 2205, 2205, 2205, 2205, 2205,    0,    0,
    0, 2205, 2205,    0, 2205,    0, 2205,    0, 2205, 2205,
 2205,    0, 2205,    0, 2205, 2205, 2205,    0, 2205, 2205,
 2205, 2205, 2205,    0, 2205, 2205,    0,    0,    0,    0,
 2205,    0, 2205, 2205,    0,    0, 2205, 2205, 2205,    0,
    0, 2205,    0, 2205,    0,    0, 2205, 2205, 2205, 2205,
    0, 2205,    0, 2205,    0, 2205, 2205, 2205, 2205, 2205,
    0,    0, 2205, 2205, 2205,    0, 2205, 2205,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 2205, 2205, 2205,
    0, 2205, 2205, 2205,    0,    0,    0,    0,    0, 2205,
    0,    0,    0, 2205,    0, 2205, 2205, 2205, 2205,    0,
 2205,    0, 2205, 2205, 2205, 2205, 2205, 2205, 2205, 2205,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 2205,
 2205,    0, 2205, 2205, 2205,    0,    0, 1422,    0, 2205,
 2205, 2205, 2205, 2205,    0, 2205,    0,    0, 2205, 2205,
    0,    0,    0, 2205, 2205,    0,    0, 2205, 2205, 2205,
 2205, 2205, 2205, 2205, 2205, 2205,    0, 2205,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 2205, 2205, 2205,
 2205, 2205, 2205, 2205, 2205, 2205, 2205, 2205, 2205, 2205,
 2205, 2205, 2205, 2205, 2205, 2205, 2205, 2205,    0, 2205,
 2205, 2205, 2205, 2205, 2205,    0, 2205, 2205, 2205, 2205,
    0, 2205,    0, 2205, 2205, 2205, 2205, 2205, 2205, 2205,
 2205, 2205, 2205, 2205, 2205, 2205, 2205, 2205, 2205, 2205,
 2205,    0, 2205, 2205, 2205, 2205,    0, 2205, 2205, 2205,
 2205, 2205,    0, 2205, 2205, 2205, 2205, 2205, 2205, 2205,
 2205, 2205, 2205, 2205, 2205, 2205, 2205, 2205, 2205, 2205,
 2205, 2205, 2205, 2205, 2205, 2205, 2205, 2205, 2205, 2205,
 2205, 2205,    0, 2205, 2205,    0,    0,    0, 2205, 2205,
 2205, 2205, 2205, 2205,    0,    0, 2205, 2205, 2205, 2205,
 2205, 2205, 2205, 2205, 2205, 2946,    0,    0,    0,    0,
 2205,    0, 2205, 2205, 2205, 2205, 2205, 2205,    0, 2205,
    0,    0,    0,    0, 1424, 1423, 1425,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2205, 2205,    0, 2205,
 2205, 2205, 2205, 2205, 2205, 2205, 2205, 2205,    0, 2205,
 2205, 2205, 2205, 2205, 2205, 2205, 2205, 2205, 2205, 2205,
 2205,    0, 2205, 2205, 2205,    0, 2205,    0,    0,    0,
 2205, 2205, 2205, 2205, 2205, 2205, 2205, 2205, 2205, 2205,
 2205, 2205, 2205, 2273, 2205, 2205,    0,    0,    0,    0,
    0, 2205, 2205, 2205, 2205, 2205, 2205, 2205, 2205, 2205,
 2205, 2205, 2205, 2205, 2205, 2205, 2205, 2205, 2205,    0,
    0, 2205,    0, 2205, 2205, 2205, 2205, 2205, 2205,    0,
    0,    0,    0, 1997, 2273, 2273, 2273, 2273, 2273, 2273,
 2273,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 2273, 2273, 2273, 2273, 2273,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 2273,    0, 2273,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1408,    0,    0,    0,
    0,    0,    0,    0,    0, 1409,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1410,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1411,    0,    0,    0,    0,    0,    0, 1412,
    0,    0,    0,    0,    0,    0,    0, 1413,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1414,    0, 1415,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1416,    0, 1417, 1418,
 1419,    0,    0,    0,    0, 1420,    0,    0,    0,    0,
 2273,    0, 2273,    0,    0, 2273,    0, 2273, 2273, 2273,
    0, 2273, 2273,    0, 2273,    0,    0,    0,    0,    0,
 2273,    0, 2273, 2273, 2273, 2273,    0, 2273,    0,    0,
 2273,    0,    0, 2273,    0,    0, 1421,    0,    0,    0,
 2273,    0, 2273, 2273,    0, 2273, 2273, 2273, 2273, 2273,
    0,    0, 2273,    0,    0, 2273,    0, 2273, 2273,    0,
    0, 2273, 2273, 2273,    0, 2273, 2273, 2273, 2273, 2273,
 2273, 2273,    0,    0,    0, 2273, 2273, 2273, 2273, 2273,
    0,    0,    0, 2273,    0,    0,    0,    0,    0,    0,
    0, 2273, 2273, 2273, 2273, 2273, 2273,    0,    0,    0,
 2273, 2273,    0, 2273,    0, 2273,    0, 2273, 2273, 2273,
    0, 2273,    0, 2273, 2273, 2273,    0, 2273, 2273, 2273,
 2273, 2273,    0, 2273, 2273,    0,    0,    0,    0, 2273,
    0, 2273, 2273,    0,    0, 2273, 2273, 2273,    0,    0,
 2273,    0, 2273,    0,    0, 2273, 2273, 2273, 2273,    0,
 2273,    0, 2273,    0, 2273, 2273, 2273, 2273, 2273,    0,
    0, 2273, 2273, 2273,    0, 2273, 2273,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2273, 2273, 2273,    0,
 2273, 2273, 2273,    0,    0,    0,    0,    0, 2273,    0,
    0,    0, 2273,    0, 2273, 2273, 2273, 2273,    0, 2273,
    0, 2273, 2273, 2273, 2273, 2273, 2273, 2273, 2273,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2273, 2273,
    0, 2273, 2273, 2273,    0,    0, 1422,    0, 2273, 2273,
 2273, 2273, 2273,    0, 2273,    0,    0, 2273, 2273,    0,
    0,    0, 2273, 2273,    0,    0, 2273, 2273, 2273, 2273,
 2273, 2273, 2273, 2273, 2273,    0, 2273,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2273, 2273, 2273, 2273,
 2273, 2273, 2273, 2273, 2273, 2273, 2273, 2273, 2273, 2273,
 2273, 2273, 2273, 2273, 2273, 2273, 2273,    0, 2273, 2273,
 2273, 2273, 2273, 2273,    0, 2273, 2273, 2273, 2273,    0,
 2273,    0, 2273, 2273, 2273, 2273, 2273, 2273, 2273, 2273,
 2273, 2273, 2273, 2273, 2273, 2273, 2273, 2273, 2273, 2273,
    0, 2273, 2273, 2273, 2273,    0, 2273, 2273, 2273, 2273,
 2273,    0, 2273, 2273, 2273, 2273, 2273, 2273, 2273, 2273,
 2273, 2273, 2273, 2273, 2273, 2273, 2273, 2273, 2273, 2273,
 2273, 2273, 2273, 2273, 2273, 2273, 2273, 2273, 2273, 2273,
 2273,    0, 2273, 2273,    0,    0,    0, 2273, 2273, 2273,
 2273, 2273, 2273,    0,    0, 2273, 2273, 2273, 2273, 2273,
 2273, 2273, 2273, 2273, 2947,    0,    0,    0,    0, 2273,
    0, 2273, 2273, 2273, 2273, 2273, 2273,    0, 2273,    0,
    0,    0,    0, 1424, 1423, 1425,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 2273, 2273,    0, 2273, 2273,
 2273, 2273, 2273, 2273, 2273, 2273, 2273,    0, 2273, 2273,
 2273, 2273, 2273, 2273, 2273, 2273, 2273, 2273, 2273, 2273,
    0, 2273, 2273, 2273,    0, 2273,    0,    0,    0, 2273,
 2273, 2273, 2273, 2273, 2273, 2273, 2273, 2273, 2273, 2273,
 2273, 2273, 2274, 2273, 2273,    0,    0,    0,    0,    0,
 2273, 2273, 2273, 2273, 2273, 2273, 2273, 2273, 2273, 2273,
 2273, 2273, 2273, 2273, 2273, 2273, 2273, 2273,    0,    0,
 2273,    0, 2273, 2273, 2273, 2273, 2273, 2273,    0,    0,
    0,    0, 1996, 2274, 2274, 2274, 2274, 2274, 2274, 2274,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2274, 2274, 2274, 2274, 2274,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 2274,    0, 2274,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1408,    0,    0,    0,    0,
    0,    0,    0,    0, 1409,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1410,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1411,    0,    0,    0,    0,    0,    0, 1412,    0,
    0,    0,    0,    0,    0,    0, 1413,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1414,    0, 1415,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1416,    0, 1417, 1418, 1419,
    0,    0,    0,    0, 1420,    0,    0,    0,    0, 2274,
    0, 2274,    0,    0, 2274,    0, 2274, 2274, 2274,    0,
 2274, 2274,    0, 2274,    0,    0,    0,    0,    0, 2274,
    0, 2274, 2274, 2274, 2274,    0, 2274,    0,    0, 2274,
    0,    0, 2274,    0,    0, 1421,    0,    0,    0, 2274,
    0, 2274, 2274,    0, 2274, 2274, 2274, 2274, 2274,    0,
    0, 2274,    0,    0, 2274,    0, 2274, 2274,    0,    0,
 2274, 2274, 2274,    0, 2274, 2274, 2274, 2274, 2274, 2274,
 2274,    0,    0,    0, 2274, 2274, 2274, 2274, 2274,    0,
    0,    0, 2274,    0,    0,    0,    0,    0,    0,    0,
 2274, 2274, 2274, 2274, 2274, 2274,    0,    0,    0, 2274,
 2274,    0, 2274,    0, 2274,    0, 2274, 2274, 2274,    0,
 2274,    0, 2274, 2274, 2274,    0, 2274, 2274, 2274, 2274,
 2274,    0, 2274, 2274,    0,    0,    0,    0, 2274,    0,
 2274, 2274,    0,    0, 2274, 2274, 2274,    0,    0, 2274,
    0, 2274,    0,    0, 2274, 2274, 2274, 2274,    0, 2274,
    0, 2274,    0, 2274, 2274, 2274, 2274, 2274,    0,    0,
 2274, 2274, 2274,    0, 2274, 2274,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 2274, 2274, 2274,    0, 2274,
 2274, 2274,    0,    0,    0,    0,    0, 2274,    0,    0,
    0, 2274,    0, 2274, 2274, 2274, 2274,    0, 2274,    0,
 2274, 2274, 2274, 2274, 2274, 2274, 2274, 2274,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 2274, 2274,    0,
 2274, 2274, 2274,    0,    0, 1422,    0, 2274, 2274, 2274,
 2274, 2274,    0, 2274,    0,    0, 2274, 2274,    0,    0,
    0, 2274, 2274,    0,    0, 2274, 2274, 2274, 2274, 2274,
 2274, 2274, 2274, 2274,    0, 2274,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 2274, 2274, 2274, 2274, 2274,
 2274, 2274, 2274, 2274, 2274, 2274, 2274, 2274, 2274, 2274,
 2274, 2274, 2274, 2274, 2274, 2274,    0, 2274, 2274, 2274,
 2274, 2274, 2274,    0, 2274, 2274, 2274, 2274,    0, 2274,
    0, 2274, 2274, 2274, 2274, 2274, 2274, 2274, 2274, 2274,
 2274, 2274, 2274, 2274, 2274, 2274, 2274, 2274, 2274,    0,
 2274, 2274, 2274, 2274,    0, 2274, 2274, 2274, 2274, 2274,
    0, 2274, 2274, 2274, 2274, 2274, 2274, 2274, 2274, 2274,
 2274, 2274, 2274, 2274, 2274, 2274, 2274, 2274, 2274, 2274,
 2274, 2274, 2274, 2274, 2274, 2274, 2274, 2274, 2274, 2274,
    0, 2274, 2274,    0,    0,    0, 2274, 2274, 2274, 2274,
 2274, 2274,    0,    0, 2274, 2274, 2274, 2274, 2274, 2274,
 2274, 2274, 2274, 2948,    0,    0,    0,    0, 2274,    0,
 2274, 2274, 2274, 2274, 2274, 2274,    0, 2274,    0,    0,
    0,    0, 1424, 1423, 1425,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 2274, 2274,    0, 2274, 2274, 2274,
 2274, 2274, 2274, 2274, 2274, 2274,    0, 2274, 2274, 2274,
 2274, 2274, 2274, 2274, 2274, 2274, 2274, 2274, 2274,    0,
 2274, 2274, 2274,    0, 2274,    0,    0,    0, 2274, 2274,
 2274, 2274, 2274, 2274, 2274, 2274, 2274, 2274, 2274, 2274,
 2274, 2267, 2274, 2274,    0,    0,    0,    0,    0, 2274,
 2274, 2274, 2274, 2274, 2274, 2274, 2274, 2274, 2274, 2274,
 2274, 2274, 2274, 2274, 2274, 2274, 2274,    0,    0, 2274,
    0, 2274, 2274, 2274, 2274, 2274, 2274,    0,    0,    0,
    0, 1896, 2267, 2267, 2267, 2267, 2267, 2267, 2267,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 2267,
 2267, 2267, 2267, 2267,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 2267,    0, 2267,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1408,    0,    0,    0,    0,    0,
    0,    0,    0, 1409,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1410,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1411,    0,    0,    0,    0,    0,    0, 1412,    0,    0,
    0,    0,    0,    0,    0, 1413,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1414,    0, 1415,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1416,    0, 1417, 1418, 1419,    0,
    0,    0,    0, 1420,    0,    0,    0,    0, 2267,    0,
 2267,    0,    0, 2267,    0, 2267, 2267, 2267,    0, 2267,
 2267,    0, 2267,    0,    0,    0,    0,    0, 2267,    0,
 2267, 2267, 2267, 2267,    0, 2267,    0,    0, 2267,    0,
    0, 2267,    0,    0, 1421,    0,    0,    0, 2267,    0,
 2267, 2267,    0, 2267, 2267, 2267, 2267, 2267,    0,    0,
 2267,    0,    0, 2267,    0, 2267, 2267,    0,    0, 2267,
 2267, 2267,    0, 2267, 2267, 2267, 2267, 2267, 2267, 2267,
    0,    0,    0, 2267, 2267, 2267, 2267, 2267,    0,    0,
    0, 2267,    0,    0,    0,    0,    0,    0,    0, 2267,
 2267, 2267, 2267, 2267, 2267,    0,    0,    0, 2267, 2267,
    0, 2267,    0, 2267,    0, 2267, 2267, 2267,    0, 2267,
    0, 2267, 2267, 2267,    0, 2267, 2267, 2267, 2267, 2267,
    0, 2267, 2267,    0,    0,    0,    0, 2267,    0, 2267,
 2267,    0,    0, 2267, 2267, 2267,    0,    0, 2267,    0,
 2267,    0,    0, 2267, 2267, 2267, 2267,    0, 2267,    0,
 2267,    0, 2267, 2267, 2267, 2267, 2267,    0,    0, 2267,
 2267, 2267,    0, 2267, 2267,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 2267, 2267, 2267,    0, 2267, 2267,
 2267,    0,    0,    0,    0,    0, 2267,    0,    0,    0,
 2267,    0, 2267, 2267, 2267, 2267,    0, 2267,    0, 2267,
 2267, 2267, 2267, 2267, 2267, 2267, 2267,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2267, 2267,    0, 2267,
 2267, 2267,    0,    0, 1422,    0, 2267, 2267, 2267, 2267,
 2267,    0, 2267,    0,    0, 2267, 2267,    0,    0,    0,
 2267, 2267,    0,    0, 2267, 2267, 2267, 2267, 2267, 2267,
 2267, 2267, 2267,    0, 2267,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 2267, 2267, 2267, 2267, 2267, 2267,
 2267, 2267, 2267, 2267, 2267, 2267, 2267, 2267, 2267, 2267,
 2267, 2267, 2267, 2267, 2267,    0, 2267, 2267, 2267, 2267,
 2267, 2267,    0, 2267, 2267, 2267, 2267,    0, 2267,    0,
 2267, 2267, 2267, 2267, 2267, 2267, 2267, 2267, 2267, 2267,
 2267, 2267, 2267, 2267, 2267, 2267, 2267, 2267,    0, 2267,
 2267, 2267, 2267,    0, 2267, 2267, 2267, 2267, 2267,    0,
 2267, 2267, 2267, 2267, 2267, 2267, 2267, 2267, 2267, 2267,
 2267, 2267, 2267, 2267, 2267, 2267, 2267, 2267, 2267, 2267,
 2267, 2267, 2267, 2267, 2267, 2267, 2267, 2267, 2267,    0,
 2267, 2267,    0,    0,    0, 2267, 2267, 2267, 2267, 2267,
 2267,    0,    0, 2267, 2267, 2267, 2267, 2267, 2267, 2267,
 2267, 2267, 2949,    0,    0,    0,    0, 2267,    0, 2267,
 2267, 2267, 2267, 2267, 2267,    0, 2267,    0,    0,    0,
    0, 1424, 1423, 1425,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 2267, 2267,    0, 2267, 2267, 2267, 2267,
 2267, 2267, 2267, 2267, 2267,    0, 2267, 2267, 2267, 2267,
 2267, 2267, 2267, 2267, 2267, 2267, 2267, 2267,    0, 2267,
 2267, 2267,    0, 2267,    0,    0,    0, 2267, 2267, 2267,
 2267, 2267, 2267, 2267, 2267, 2267, 2267, 2267, 2267, 2267,
 2268, 2267, 2267,    0,    0,    0,    0,    0, 2267, 2267,
 2267, 2267, 2267, 2267, 2267, 2267, 2267, 2267, 2267, 2267,
 2267, 2267, 2267, 2267, 2267, 2267,    0,    0, 2267,    0,
 2267, 2267, 2267, 2267, 2267, 2267,    0,    0,    0,    0,
 1900, 2268, 2268, 2268, 2268, 2268, 2268, 2268,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2268, 2268,
 2268, 2268, 2268,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 2268,    0, 2268,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1408,    0,    0,    0,    0,    0,    0,
    0,    0, 1409,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1410,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1411,
    0,    0,    0,    0,    0,    0, 1412,    0,    0,    0,
    0,    0,    0,    0, 1413,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1414,    0, 1415,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1416,    0, 1417, 1418, 1419,    0,    0,
    0,    0, 1420,    0,    0,    0,    0, 2268,    0, 2268,
    0,    0, 2268,    0, 2268, 2268, 2268,    0, 2268, 2268,
    0, 2268,    0,    0,    0,    0,    0, 2268,    0, 2268,
 2268, 2268, 2268,    0, 2268,    0,    0, 2268,    0,    0,
 2268,    0,    0, 1421,    0,    0,    0, 2268,    0, 2268,
 2268,    0, 2268, 2268, 2268, 2268, 2268,    0,    0, 2268,
    0,    0, 2268,    0, 2268, 2268,    0,    0, 2268, 2268,
 2268,    0, 2268, 2268, 2268, 2268, 2268, 2268, 2268,    0,
    0,    0, 2268, 2268, 2268, 2268, 2268,    0,    0,    0,
 2268,    0,    0,    0,    0,    0,    0,    0, 2268, 2268,
 2268, 2268, 2268, 2268,    0,    0,    0, 2268, 2268,    0,
 2268,    0, 2268,    0, 2268, 2268, 2268,    0, 2268,    0,
 2268, 2268, 2268,    0, 2268, 2268, 2268, 2268, 2268,    0,
 2268, 2268,    0,    0,    0,    0, 2268,    0, 2268, 2268,
    0,    0, 2268, 2268, 2268,    0,    0, 2268,    0, 2268,
    0,    0, 2268, 2268, 2268, 2268,    0, 2268,    0, 2268,
    0, 2268, 2268, 2268, 2268, 2268,    0,    0, 2268, 2268,
 2268,    0, 2268, 2268,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 2268, 2268, 2268,    0, 2268, 2268, 2268,
    0,    0,    0,    0,    0, 2268,    0,    0,    0, 2268,
    0, 2268, 2268, 2268, 2268,    0, 2268,    0, 2268, 2268,
 2268, 2268, 2268, 2268, 2268, 2268,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 2268, 2268,    0, 2268, 2268,
 2268,    0,    0, 1422,    0, 2268, 2268, 2268, 2268, 2268,
    0, 2268,    0,    0, 2268, 2268,    0,    0,    0, 2268,
 2268,    0,    0, 2268, 2268, 2268, 2268, 2268, 2268, 2268,
 2268, 2268,    0, 2268,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 2268, 2268, 2268, 2268, 2268, 2268, 2268,
 2268, 2268, 2268, 2268, 2268, 2268, 2268, 2268, 2268, 2268,
 2268, 2268, 2268, 2268,    0, 2268, 2268, 2268, 2268, 2268,
 2268,    0, 2268, 2268, 2268, 2268,    0, 2268,    0, 2268,
 2268, 2268, 2268, 2268, 2268, 2268, 2268, 2268, 2268, 2268,
 2268, 2268, 2268, 2268, 2268, 2268, 2268,    0, 2268, 2268,
 2268, 2268,    0, 2268, 2268, 2268, 2268, 2268,    0, 2268,
 2268, 2268, 2268, 2268, 2268, 2268, 2268, 2268, 2268, 2268,
 2268, 2268, 2268, 2268, 2268, 2268, 2268, 2268, 2268, 2268,
 2268, 2268, 2268, 2268, 2268, 2268, 2268, 2268,    0, 2268,
 2268,    0,    0,    0, 2268, 2268, 2268, 2268, 2268, 2268,
    0,    0, 2268, 2268, 2268, 2268, 2268, 2268, 2268, 2268,
 2268, 2950,    0,    0,    0,    0, 2268,    0, 2268, 2268,
 2268, 2268, 2268, 2268,    0, 2268,    0,    0,    0,    0,
 1424, 1423, 1425,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 2268, 2268,    0, 2268, 2268, 2268, 2268, 2268,
 2268, 2268, 2268, 2268,    0, 2268, 2268, 2268, 2268, 2268,
 2268, 2268, 2268, 2268, 2268, 2268, 2268,    0, 2268, 2268,
 2268,    0, 2268,    0,    0,    0, 2268, 2268, 2268, 2268,
 2268, 2268, 2268, 2268, 2268, 2268, 2268, 2268, 2268, 2269,
 2268, 2268,    0,    0,    0,    0,    0, 2268, 2268, 2268,
 2268, 2268, 2268, 2268, 2268, 2268, 2268, 2268, 2268, 2268,
 2268, 2268, 2268, 2268, 2268,    0,    0, 2268,    0, 2268,
 2268, 2268, 2268, 2268, 2268,    0,    0,    0,    0, 1903,
 2269, 2269, 2269, 2269, 2269, 2269, 2269,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 2269, 2269, 2269,
 2269, 2269,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2269,    0, 2269,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1408,    0,    0,    0,    0,    0,    0,    0,
    0, 1409,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1410,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1411,    0,
    0,    0,    0,    0,    0, 1412,    0,    0,    0,    0,
    0,    0,    0, 1413,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1414,    0, 1415,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1416,    0, 1417, 1418, 1419,    0,    0,    0,
    0, 1420,    0,    0,    0,    0, 2269,    0, 2269,    0,
    0, 2269,    0, 2269, 2269, 2269,    0, 2269, 2269,    0,
 2269,    0,    0,    0,    0,    0, 2269,    0, 2269, 2269,
 2269, 2269,    0, 2269,    0,    0, 2269,    0,    0, 2269,
    0,    0, 1421,    0,    0,    0, 2269,    0, 2269, 2269,
    0, 2269, 2269, 2269, 2269, 2269,    0,    0, 2269,    0,
    0, 2269,    0, 2269, 2269,    0,    0, 2269, 2269, 2269,
    0, 2269, 2269, 2269, 2269, 2269, 2269, 2269,    0,    0,
    0, 2269, 2269, 2269, 2269, 2269,    0,    0,    0, 2269,
    0,    0,    0,    0,    0,    0,    0, 2269, 2269, 2269,
 2269, 2269, 2269,    0,    0,    0, 2269, 2269,    0, 2269,
    0, 2269,    0, 2269, 2269, 2269,    0, 2269,    0, 2269,
 2269, 2269,    0, 2269, 2269, 2269, 2269, 2269,    0, 2269,
 2269,    0,    0,    0,    0, 2269,    0, 2269, 2269,    0,
    0, 2269, 2269, 2269,    0,    0, 2269,    0, 2269,    0,
    0, 2269, 2269, 2269, 2269,    0, 2269,    0, 2269,    0,
 2269, 2269, 2269, 2269, 2269,    0,    0, 2269, 2269, 2269,
    0, 2269, 2269,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 2269, 2269, 2269,    0, 2269, 2269, 2269,    0,
    0,    0,    0,    0, 2269,    0,    0,    0, 2269,    0,
 2269, 2269, 2269, 2269,    0, 2269,    0, 2269, 2269, 2269,
 2269, 2269, 2269, 2269, 2269,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 2269, 2269,    0, 2269, 2269, 2269,
    0,    0, 1422,    0, 2269, 2269, 2269, 2269, 2269,    0,
 2269,    0,    0, 2269, 2269,    0,    0,    0, 2269, 2269,
    0,    0, 2269, 2269, 2269, 2269, 2269, 2269, 2269, 2269,
 2269,    0, 2269,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 2269, 2269, 2269, 2269, 2269, 2269, 2269, 2269,
 2269, 2269, 2269, 2269, 2269, 2269, 2269, 2269, 2269, 2269,
 2269, 2269, 2269,    0, 2269, 2269, 2269, 2269, 2269, 2269,
    0, 2269, 2269, 2269, 2269,    0, 2269,    0, 2269, 2269,
 2269, 2269, 2269, 2269, 2269, 2269, 2269, 2269, 2269, 2269,
 2269, 2269, 2269, 2269, 2269, 2269,    0, 2269, 2269, 2269,
 2269,    0, 2269, 2269, 2269, 2269, 2269,    0, 2269, 2269,
 2269, 2269, 2269, 2269, 2269, 2269, 2269, 2269, 2269, 2269,
 2269, 2269, 2269, 2269, 2269, 2269, 2269, 2269, 2269, 2269,
 2269, 2269, 2269, 2269, 2269, 2269, 2269,    0, 2269, 2269,
    0,    0,    0, 2269, 2269, 2269, 2269, 2269, 2269,    0,
    0, 2269, 2269, 2269, 2269, 2269, 2269, 2269, 2269, 2269,
 2951,    0,    0,    0,    0, 2269,    0, 2269, 2269, 2269,
 2269, 2269, 2269,    0, 2269,    0,    0,    0,    0, 1424,
 1423, 1425,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 2269, 2269,    0, 2269, 2269, 2269, 2269, 2269, 2269,
 2269, 2269, 2269,    0, 2269, 2269, 2269, 2269, 2269, 2269,
 2269, 2269, 2269, 2269, 2269, 2269,    0, 2269, 2269, 2269,
    0, 2269,    0,    0,    0, 2269, 2269, 2269, 2269, 2269,
 2269, 2269, 2269, 2269, 2269, 2269, 2269, 2269, 2297, 2269,
 2269,    0,    0,    0,    0,    0, 2269, 2269, 2269, 2269,
 2269, 2269, 2269, 2269, 2269, 2269, 2269, 2269, 2269, 2269,
 2269, 2269, 2269, 2269,    0,    0, 2269,    0, 2269, 2269,
 2269, 2269, 2269, 2269,    0,    0,    0,    0, 1904, 2297,
 2297, 2297, 2297, 2297, 2297, 2297,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2297, 2297, 2297, 2297,
 2297,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 2297,
    0, 2297,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1408,    0,    0,    0,    0,    0,    0,    0,    0,
 1409,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1410,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1411,    0,    0,
    0,    0,    0,    0, 1412,    0,    0,    0,    0,    0,
    0,    0, 1413,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1414,
    0, 1415,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1416,    0, 1417, 1418, 1419,    0,    0,    0,    0,
 1420,    0,    0,    0,    0, 2297,    0, 2297,    0,    0,
 2297,    0, 2297, 2297, 2297,    0, 2297, 2297,    0, 2297,
    0,    0,    0,    0,    0, 2297,    0, 2297, 2297, 2297,
 2297,    0, 2297,    0,    0, 2297,    0,    0, 2297,    0,
    0, 1421,    0,    0,    0, 2297,    0, 2297, 2297,    0,
 2297, 2297, 2297, 2297, 2297,    0,    0, 2297,    0,    0,
 2297,    0, 2297, 2297,    0,    0, 2297, 2297, 2297,    0,
 2297, 2297, 2297, 2297, 2297, 2297, 2297,    0,    0,    0,
 2297, 2297, 2297, 2297, 2297,    0,    0,    0, 2297,    0,
    0,    0,    0,    0,    0,    0, 2297, 2297, 2297, 2297,
 2297, 2297,    0,    0,    0, 2297, 2297,    0, 2297,    0,
 2297,    0, 2297, 2297, 2297,    0, 2297,    0, 2297, 2297,
 2297,    0, 2297, 2297, 2297, 2297, 2297,    0, 2297, 2297,
    0,    0,    0,    0, 2297,    0, 2297, 2297,    0,    0,
 2297, 2297, 2297,    0,    0, 2297,    0, 2297,    0,    0,
 2297, 2297, 2297, 2297,    0, 2297,    0, 2297,    0, 2297,
 2297, 2297, 2297, 2297,    0,    0, 2297, 2297, 2297,    0,
 2297, 2297,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 2297, 2297, 2297,    0, 2297, 2297, 2297,    0,    0,
    0,    0,    0, 2297,    0,    0,    0, 2297,    0, 2297,
 2297, 2297, 2297,    0, 2297,    0, 2297, 2297, 2297, 2297,
 2297, 2297, 2297, 2297,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 2297, 2297,    0, 2297, 2297, 2297,    0,
    0, 1422,    0, 2297, 2297, 2297, 2297, 2297,    0, 2297,
    0,    0, 2297, 2297,    0,    0,    0, 2297, 2297,    0,
    0, 2297, 2297, 2297, 2297, 2297, 2297, 2297, 2297, 2297,
    0, 2297,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 2297, 2297, 2297, 2297, 2297, 2297, 2297, 2297, 2297,
 2297, 2297, 2297, 2297, 2297, 2297, 2297, 2297, 2297, 2297,
 2297, 2297,    0, 2297, 2297, 2297, 2297, 2297, 2297,    0,
 2297, 2297, 2297, 2297,    0, 2297,    0, 2297, 2297, 2297,
 2297, 2297, 2297, 2297, 2297, 2297, 2297, 2297, 2297, 2297,
 2297, 2297, 2297, 2297, 2297,    0, 2297, 2297, 2297, 2297,
    0, 2297, 2297, 2297, 2297, 2297,    0, 2297, 2297, 2297,
 2297, 2297, 2297, 2297, 2297, 2297, 2297, 2297, 2297, 2297,
 2297, 2297, 2297, 2297, 2297, 2297, 2297, 2297, 2297, 2297,
 2297, 2297, 2297, 2297, 2297, 2297,    0, 2297, 2297,    0,
    0,    0, 2297, 2297, 2297, 2297, 2297, 2297,    0,    0,
 2297, 2297, 2297, 2297, 2297, 2297, 2297, 2297, 2297, 3251,
    0,    0,    0,    0, 2297,    0, 2297, 2297, 2297, 2297,
 2297, 2297,    0, 2297,    0,    0,    0,    0, 1424, 1423,
 1425,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2297, 2297,    0, 2297, 2297, 2297, 2297, 2297, 2297, 2297,
 2297, 2297,    0, 2297, 2297, 2297, 2297, 2297, 2297, 2297,
 2297, 2297, 2297, 2297, 2297,    0, 2297, 2297, 2297,    0,
 2297,    0,    0,    0, 2297, 2297, 2297, 2297, 2297, 2297,
 2297, 2297, 2297, 2297, 2297, 2297, 2297, 2298, 2297, 2297,
    0,    0,    0,    0,    0, 2297, 2297, 2297, 2297, 2297,
 2297, 2297, 2297, 2297, 2297, 2297, 2297, 2297, 2297, 2297,
 2297, 2297, 2297,    0,    0, 2297,    0, 2297, 2297, 2297,
 2297, 2297, 2297,    0,    0,    0,    0, 1905, 2298, 2298,
 2298, 2298, 2298, 2298, 2298,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 2298, 2298, 2298, 2298, 2298,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2298,    0,
 2298,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1408,    0,    0,    0,    0,    0,    0,    0,    0, 1409,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1410,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1411,    0,    0,    0,
    0,    0,    0, 1412,    0,    0,    0,    0,    0,    0,
    0, 1413,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1414,    0,
 1415,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1416,    0, 1417, 1418, 1419,    0,    0,    0,    0, 1420,
    0,    0,    0,    0, 2298,    0, 2298,    0,    0, 2298,
    0, 2298, 2298, 2298,    0, 2298, 2298,    0, 2298,    0,
    0,    0,    0,    0, 2298,    0, 2298, 2298, 2298, 2298,
    0, 2298,    0,    0, 2298,    0,    0, 2298,    0,    0,
 1421,    0,    0,    0, 2298,    0, 2298, 2298,    0, 2298,
 2298, 2298, 2298, 2298,    0,    0, 2298,    0,    0, 2298,
    0, 2298, 2298,    0,    0, 2298, 2298, 2298,    0, 2298,
 2298, 2298, 2298, 2298, 2298, 2298,    0,    0,    0, 2298,
 2298, 2298, 2298, 2298,    0,    0,    0, 2298,    0,    0,
    0,    0,    0,    0,    0, 2298, 2298, 2298, 2298, 2298,
 2298,    0,    0,    0, 2298, 2298,    0, 2298,    0, 2298,
    0, 2298, 2298, 2298,    0, 2298,    0, 2298, 2298, 2298,
    0, 2298, 2298, 2298, 2298, 2298,    0, 2298, 2298,    0,
    0,    0,    0, 2298,    0, 2298, 2298,    0,    0, 2298,
 2298, 2298,    0,    0, 2298,    0, 2298,    0,    0, 2298,
 2298, 2298, 2298,    0, 2298,    0, 2298,    0, 2298, 2298,
 2298, 2298, 2298,    0,    0, 2298, 2298, 2298,    0, 2298,
 2298,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2298, 2298, 2298,    0, 2298, 2298, 2298,    0,    0,    0,
    0,    0, 2298,    0,    0,    0, 2298,    0, 2298, 2298,
 2298, 2298,    0, 2298,    0, 2298, 2298, 2298, 2298, 2298,
 2298, 2298, 2298,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 2298, 2298,    0, 2298, 2298, 2298,    0,    0,
 1422,    0, 2298, 2298, 2298, 2298, 2298,    0, 2298,    0,
    0, 2298, 2298,    0,    0,    0, 2298, 2298,    0,    0,
 2298, 2298, 2298, 2298, 2298, 2298, 2298, 2298, 2298,    0,
 2298,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2298, 2298, 2298, 2298, 2298, 2298, 2298, 2298, 2298, 2298,
 2298, 2298, 2298, 2298, 2298, 2298, 2298, 2298, 2298, 2298,
 2298,    0, 2298, 2298, 2298, 2298, 2298, 2298,    0, 2298,
 2298, 2298, 2298,    0, 2298,    0, 2298, 2298, 2298, 2298,
 2298, 2298, 2298, 2298, 2298, 2298, 2298, 2298, 2298, 2298,
 2298, 2298, 2298, 2298,    0, 2298, 2298, 2298, 2298,    0,
 2298, 2298, 2298, 2298, 2298,    0, 2298, 2298, 2298, 2298,
 2298, 2298, 2298, 2298, 2298, 2298, 2298, 2298, 2298, 2298,
 2298, 2298, 2298, 2298, 2298, 2298, 2298, 2298, 2298, 2298,
 2298, 2298, 2298, 2298, 2298,    0, 2298, 2298,    0,    0,
    0, 2298, 2298, 2298, 2298, 2298, 2298,    0,    0, 2298,
 2298, 2298, 2298, 2298, 2298, 2298, 2298, 2298, 3252,    0,
    0,    0,    0, 2298,    0, 2298, 2298, 2298, 2298, 2298,
 2298,    0, 2298,    0,    0,    0,    0, 1424, 1423, 1425,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 2298,
 2298,    0, 2298, 2298, 2298, 2298, 2298, 2298, 2298, 2298,
 2298,    0, 2298, 2298, 2298, 2298, 2298, 2298, 2298, 2298,
 2298, 2298, 2298, 2298,    0, 2298, 2298, 2298,    0, 2298,
    0,    0,    0, 2298, 2298, 2298, 2298, 2298, 2298, 2298,
 2298, 2298, 2298, 2298, 2298, 2298, 2302, 2298, 2298,    0,
    0,    0,    0,    0, 2298, 2298, 2298, 2298, 2298, 2298,
 2298, 2298, 2298, 2298, 2298, 2298, 2298, 2298, 2298, 2298,
 2298, 2298,    0,    0, 2298,    0, 2298, 2298, 2298, 2298,
 2298, 2298,    0,    0,    0,    0, 1956, 2302, 2302, 2302,
 2302, 2302, 2302, 2302,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 2302, 2302, 2302, 2302, 2302,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 2302,    0, 2302,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1408,
    0,    0,    0,    0,    0,    0,    0,    0, 1409,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1410,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1411,    0,    0,    0,    0,
    0,    0, 1412,    0,    0,    0,    0,    0,    0,    0,
 1413,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1414,    0, 1415,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1416,
    0, 1417, 1418, 1419,    0,    0,    0,    0, 1420,    0,
    0,    0,    0, 2302,    0, 2302,    0,    0, 2302,    0,
 2302, 2302, 2302,    0, 2302, 2302,    0, 2302,    0,    0,
    0,    0,    0, 2302,    0, 2302, 2302, 2302, 2302,    0,
 2302,    0,    0, 2302,    0,    0, 2302,    0,    0, 1421,
    0,    0,    0, 2302,    0, 2302, 2302,    0, 2302, 2302,
 2302, 2302, 2302,    0,    0, 2302,    0,    0, 2302,    0,
 2302, 2302,    0,    0, 2302, 2302, 2302,    0, 2302, 2302,
 2302, 2302, 2302, 2302, 2302,    0,    0,    0, 2302, 2302,
 2302, 2302, 2302,    0,    0,    0, 2302,    0,    0,    0,
    0,    0,    0,    0, 2302, 2302, 2302, 2302, 2302, 2302,
    0,    0,    0, 2302, 2302,    0, 2302,    0, 2302,    0,
 2302, 2302, 2302,    0, 2302,    0, 2302, 2302, 2302,    0,
 2302, 2302, 2302, 2302, 2302,    0, 2302, 2302,    0,    0,
    0,    0, 2302,    0, 2302, 2302,    0,    0, 2302, 2302,
 2302,    0,    0, 2302,    0, 2302,    0,    0, 2302, 2302,
 2302, 2302,    0, 2302,    0, 2302,    0, 2302, 2302, 2302,
 2302, 2302,    0,    0, 2302, 2302, 2302,    0, 2302, 2302,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 2302,
 2302, 2302,    0, 2302, 2302, 2302,    0,    0,    0,    0,
    0, 2302,    0,    0,    0, 2302,    0, 2302, 2302, 2302,
 2302,    0, 2302,    0, 2302, 2302, 2302, 2302, 2302, 2302,
 2302, 2302,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 2302, 2302,    0, 2302, 2302, 2302,    0,    0, 1422,
    0, 2302, 2302, 2302, 2302, 2302,    0, 2302,    0,    0,
 2302, 2302,    0,    0,    0, 2302, 2302,    0,    0, 2302,
 2302, 2302, 2302, 2302, 2302, 2302, 2302, 2302,    0, 2302,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 2302,
 2302, 2302, 2302, 2302, 2302, 2302, 2302, 2302, 2302, 2302,
 2302, 2302, 2302, 2302, 2302, 2302, 2302, 2302, 2302, 2302,
    0, 2302, 2302, 2302, 2302, 2302, 2302,    0, 2302, 2302,
 2302, 2302,    0, 2302,    0, 2302, 2302, 2302, 2302, 2302,
 2302, 2302, 2302, 2302, 2302, 2302, 2302, 2302, 2302, 2302,
 2302, 2302, 2302,    0, 2302, 2302, 2302, 2302,    0, 2302,
 2302, 2302, 2302, 2302,    0, 2302, 2302, 2302, 2302, 2302,
 2302, 2302, 2302, 2302, 2302, 2302, 2302, 2302, 2302, 2302,
 2302, 2302, 2302, 2302, 2302, 2302, 2302, 2302, 2302, 2302,
 2302, 2302, 2302, 2302,    0, 2302, 2302,    0,    0,    0,
 2302, 2302, 2302, 2302, 2302, 2302,    0,    0, 2302, 2302,
 2302, 2302, 2302, 2302, 2302, 2302, 2302, 3455,    0,    0,
    0,    0, 2302,    0, 2302, 2302, 2302, 2302, 2302, 2302,
    0, 2302,    0,    0,    0,    0, 1424, 1423, 1425,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2302, 2302,
    0, 2302, 2302, 2302, 2302, 2302, 2302, 2302, 2302, 2302,
    0, 2302, 2302, 2302, 2302, 2302, 2302, 2302, 2302, 2302,
 2302, 2302, 2302,    0, 2302, 2302, 2302,    0, 2302,    0,
    0,    0, 2302, 2302, 2302, 2302, 2302, 2302, 2302, 2302,
 2302, 2302, 2302, 2302, 2302, 2305, 2302, 2302,    0,    0,
    0,    0,    0, 2302, 2302, 2302, 2302, 2302, 2302, 2302,
 2302, 2302, 2302, 2302, 2302, 2302, 2302, 2302, 2302, 2302,
 2302,    0,    0, 2302,    0, 2302, 2302, 2302, 2302, 2302,
 2302,    0,    0,    0,    0, 1975, 2305, 2305, 2305, 2305,
 2305, 2305, 2305,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 2305, 2305, 2305, 2305, 2305,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2305,    0, 2305,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1408,    0,
    0,    0,    0,    0,    0,    0,    0, 1409,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1410,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1411,    0,    0,    0,    0,    0,
    0, 1412,    0,    0,    0,    0,    0,    0,    0, 1413,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1414,    0, 1415,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1416,    0,
 1417, 1418, 1419,    0,    0,    0,    0, 1420,    0,    0,
    0,    0, 2305,    0, 2305,    0,    0, 2305,    0, 2305,
 2305, 2305,    0, 2305, 2305,    0, 2305,    0,    0,    0,
    0,    0, 2305,    0, 2305, 2305, 2305, 2305,    0, 2305,
    0,    0, 2305,    0,    0, 2305,    0,    0, 1421,    0,
    0,    0, 2305,    0, 2305, 2305,    0, 2305, 2305, 2305,
 2305, 2305,    0,    0, 2305,    0,    0, 2305,    0, 2305,
 2305,    0,    0, 2305, 2305, 2305,    0, 2305, 2305, 2305,
 2305, 2305, 2305, 2305,    0,    0,    0, 2305, 2305, 2305,
 2305, 2305,    0,    0,    0, 2305,    0,    0,    0,    0,
    0,    0,    0, 2305, 2305, 2305, 2305, 2305, 2305,    0,
    0,    0, 2305, 2305,    0, 2305,    0, 2305,    0, 2305,
 2305, 2305,    0, 2305,    0, 2305, 2305, 2305,    0, 2305,
 2305, 2305, 2305, 2305,    0, 2305, 2305,    0,    0,    0,
    0, 2305,    0, 2305, 2305,    0,    0, 2305, 2305, 2305,
    0,    0, 2305,    0, 2305,    0,    0, 2305, 2305, 2305,
 2305,    0, 2305,    0, 2305,    0, 2305, 2305, 2305, 2305,
 2305,    0,    0, 2305, 2305, 2305,    0, 2305, 2305,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2305, 2305,
 2305,    0, 2305, 2305, 2305,    0,    0,    0,    0,    0,
 2305,    0,    0,    0, 2305,    0, 2305, 2305, 2305, 2305,
    0, 2305,    0, 2305, 2305, 2305, 2305, 2305, 2305, 2305,
 2305,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2305, 2305,    0, 2305, 2305, 2305,    0,    0, 1422,    0,
 2305, 2305, 2305, 2305, 2305,    0, 2305,    0,    0, 2305,
 2305,    0,    0,    0, 2305, 2305,    0,    0, 2305, 2305,
 2305, 2305, 2305, 2305, 2305, 2305, 2305,    0, 2305,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2305, 2305,
 2305, 2305, 2305, 2305, 2305, 2305, 2305, 2305, 2305, 2305,
 2305, 2305, 2305, 2305, 2305, 2305, 2305, 2305, 2305,    0,
 2305, 2305, 2305, 2305, 2305, 2305,    0, 2305, 2305, 2305,
 2305,    0, 2305,    0, 2305, 2305, 2305, 2305, 2305, 2305,
 2305, 2305, 2305, 2305, 2305, 2305, 2305, 2305, 2305, 2305,
 2305, 2305,    0, 2305, 2305, 2305, 2305,    0, 2305, 2305,
 2305, 2305, 2305,    0, 2305, 2305, 2305, 2305, 2305, 2305,
 2305, 2305, 2305, 2305, 2305, 2305, 2305, 2305, 2305, 2305,
 2305, 2305, 2305, 2305, 2305, 2305, 2305, 2305, 2305, 2305,
 2305, 2305, 2305,    0, 2305, 2305,    0,    0,    0, 2305,
 2305, 2305, 2305, 2305, 2305,    0,    0, 2305, 2305, 2305,
 2305, 2305, 2305, 2305, 2305, 2305, 3460,    0,    0,    0,
    0, 2305,    0, 2305, 2305, 2305, 2305, 2305, 2305,    0,
 2305,    0,    0,    0,    0, 1424, 1423, 1425,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 2305, 2305,    0,
 2305, 2305, 2305, 2305, 2305, 2305, 2305, 2305, 2305,    0,
 2305, 2305, 2305, 2305, 2305, 2305, 2305, 2305, 2305, 2305,
 2305, 2305,    0, 2305, 2305, 2305,    0, 2305,    0,    0,
    0, 2305, 2305, 2305, 2305, 2305, 2305, 2305, 2305, 2305,
 2305, 2305, 2305, 2305, 2318, 2305, 2305,    0,    0,    0,
    0,    0, 2305, 2305, 2305, 2305, 2305, 2305, 2305, 2305,
 2305, 2305, 2305, 2305, 2305, 2305, 2305, 2305, 2305, 2305,
    0,    0, 2305,    0, 2305, 2305, 2305, 2305, 2305, 2305,
    0,    0,    0,    0, 1920, 2318, 2318, 2318, 2318, 2318,
 2318, 2318,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 2318, 2318, 2318, 2318, 2318,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 2318,    0, 2318,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1408,    0,    0,
    0,    0,    0,    0,    0,    0, 1409,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1410,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1411,    0,    0,    0,    0,    0,    0,
 1412,    0,    0,    0,    0,    0,    0,    0, 1413,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1414,    0, 1415,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1416,    0, 1417,
 1418, 1419,    0,    0,    0,    0, 1420,    0,    0,    0,
    0, 2318,    0, 2318,    0,    0, 2318,    0, 2318, 2318,
 2318,    0, 2318, 2318,    0, 2318,    0,    0,    0,    0,
    0, 2318,    0, 2318, 2318, 2318, 2318,    0, 2318,    0,
    0, 2318,    0,    0, 2318,    0,    0, 1421,    0,    0,
    0, 2318,    0, 2318, 2318,    0, 2318, 2318, 2318, 2318,
 2318,    0,    0, 2318,    0,    0, 2318,    0, 2318, 2318,
    0,    0, 2318, 2318, 2318,    0, 2318, 2318, 2318, 2318,
 2318, 2318, 2318,    0,    0,    0, 2318, 2318, 2318, 2318,
 2318,    0,    0,    0, 2318,    0,    0,    0,    0,    0,
    0,    0, 2318, 2318, 2318, 2318, 2318, 2318,    0,    0,
    0, 2318, 2318,    0, 2318,    0, 2318,    0, 2318, 2318,
 2318,    0, 2318,    0, 2318, 2318, 2318,    0, 2318, 2318,
 2318, 2318, 2318,    0, 2318, 2318,    0,    0,    0,    0,
 2318,    0, 2318, 2318,    0,    0, 2318, 2318, 2318,    0,
    0, 2318,    0, 2318,    0,    0, 2318, 2318, 2318, 2318,
    0, 2318,    0, 2318,    0, 2318, 2318, 2318, 2318, 2318,
    0,    0, 2318, 2318, 2318,    0, 2318, 2318,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 2318, 2318, 2318,
    0, 2318, 2318, 2318,    0,    0,    0,    0,    0, 2318,
    0,    0,    0, 2318,    0, 2318, 2318, 2318, 2318,    0,
 2318,    0, 2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 2318,
 2318,    0, 2318, 2318, 2318,    0,    0, 1422,    0, 2318,
 2318, 2318, 2318, 2318,    0, 2318,    0,    0, 2318, 2318,
    0,    0,    0, 2318, 2318,    0,    0, 2318, 2318, 2318,
 2318, 2318, 2318, 2318, 2318, 2318,    0, 2318,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 2318, 2318, 2318,
 2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318,
 2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318,    0, 2318,
 2318, 2318, 2318, 2318, 2318,    0, 2318, 2318, 2318, 2318,
    0, 2318,    0, 2318, 2318, 2318, 2318, 2318, 2318, 2318,
 2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318,
 2318,    0, 2318, 2318, 2318, 2318,    0, 2318, 2318, 2318,
 2318, 2318,    0, 2318, 2318, 2318, 2318, 2318, 2318, 2318,
 2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318,
 2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318,
 2318, 2318,    0, 2318, 2318,    0,    0,    0, 2318, 2318,
 2318, 2318, 2318, 2318,    0,    0, 2318, 2318, 2318, 2318,
 2318, 2318, 2318, 2318, 2318, 3461,    0,    0,    0,    0,
 2318,    0, 2318, 2318, 2318, 2318, 2318, 2318,    0, 2318,
    0,    0,    0,    0, 1424, 1423, 1425,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2318, 2318,    0, 2318,
 2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318,    0, 2318,
 2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318,
 2318,    0, 2318, 2318, 2318,    0, 2318,    0,    0,    0,
 2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318,
 2318, 2318, 2318, 2319, 2318, 2318,    0,    0,    0,    0,
    0, 2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318,
 2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318,    0,
    0, 2318,    0, 2318, 2318, 2318, 2318, 2318, 2318,    0,
    0,    0,    0, 1921, 2319, 2319, 2319, 2319, 2319, 2319,
 2319,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 2319, 2319, 2319, 2319, 2319,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 2319,    0, 2319,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1408,    0,    0,    0,
    0,    0,    0,    0,    0, 1409,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1410,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1411,    0,    0,    0,    0,    0,    0, 1412,
    0,    0,    0,    0,    0,    0,    0, 1413,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1414,    0, 1415,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1416,    0, 1417, 1418,
 1419,    0,    0,    0,    0, 1420,    0,    0,    0,    0,
 2319,    0, 2319,    0,    0, 2319,    0, 2319, 2319, 2319,
    0, 2319, 2319,    0, 2319,    0,    0,    0,    0,    0,
 2319,    0, 2319, 2319, 2319, 2319,    0, 2319,    0,    0,
 2319,    0,    0, 2319,    0,    0, 1421,    0,    0,    0,
 2319,    0, 2319, 2319,    0, 2319, 2319, 2319, 2319, 2319,
    0,    0, 2319,    0,    0, 2319,    0, 2319, 2319,    0,
    0, 2319, 2319, 2319,    0, 2319, 2319, 2319, 2319, 2319,
 2319, 2319,    0,    0,    0, 2319, 2319, 2319, 2319, 2319,
    0,    0,    0, 2319,    0,    0,    0,    0,    0,    0,
    0, 2319, 2319, 2319, 2319, 2319, 2319,    0,    0,    0,
 2319, 2319,    0, 2319,    0, 2319,    0, 2319, 2319, 2319,
    0, 2319,    0, 2319, 2319, 2319,    0, 2319, 2319, 2319,
 2319, 2319,    0, 2319, 2319,    0,    0,    0,    0, 2319,
    0, 2319, 2319,    0,    0, 2319, 2319, 2319,    0,    0,
 2319,    0, 2319,    0,    0, 2319, 2319, 2319, 2319,    0,
 2319,    0, 2319,    0, 2319, 2319, 2319, 2319, 2319,    0,
    0, 2319, 2319, 2319,    0, 2319, 2319,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2319, 2319, 2319,    0,
 2319, 2319, 2319,    0,    0,    0,    0,    0, 2319,    0,
    0,    0, 2319,    0, 2319, 2319, 2319, 2319,    0, 2319,
    0, 2319, 2319, 2319, 2319, 2319, 2319, 2319, 2319,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2319, 2319,
    0, 2319, 2319, 2319,    0,    0, 1422,    0, 2319, 2319,
 2319, 2319, 2319,    0, 2319,    0,    0, 2319, 2319,    0,
    0,    0, 2319, 2319,    0,    0, 2319, 2319, 2319, 2319,
 2319, 2319, 2319, 2319, 2319,    0, 2319,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2319, 2319, 2319, 2319,
 2319, 2319, 2319, 2319, 2319, 2319, 2319, 2319, 2319, 2319,
 2319, 2319, 2319, 2319, 2319, 2319, 2319,    0, 2319, 2319,
 2319, 2319, 2319, 2319,    0, 2319, 2319, 2319, 2319,    0,
 2319,    0, 2319, 2319, 2319, 2319, 2319, 2319, 2319, 2319,
 2319, 2319, 2319, 2319, 2319, 2319, 2319, 2319, 2319, 2319,
    0, 2319, 2319, 2319, 2319,    0, 2319, 2319, 2319, 2319,
 2319,    0, 2319, 2319, 2319, 2319, 2319, 2319, 2319, 2319,
 2319, 2319, 2319, 2319, 2319, 2319, 2319, 2319, 2319, 2319,
 2319, 2319, 2319, 2319, 2319, 2319, 2319, 2319, 2319, 2319,
 2319,    0, 2319, 2319,    0,    0,    0, 2319, 2319, 2319,
 2319, 2319, 2319,    0,    0, 2319, 2319, 2319, 2319, 2319,
 2319, 2319, 2319, 2319, 3462,    0,    0,    0,    0, 2319,
    0, 2319, 2319, 2319, 2319, 2319, 2319,    0, 2319,    0,
    0,    0,    0, 1424, 1423, 1425,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 2319, 2319,    0, 2319, 2319,
 2319, 2319, 2319, 2319, 2319, 2319, 2319,    0, 2319, 2319,
 2319, 2319, 2319, 2319, 2319, 2319, 2319, 2319, 2319, 2319,
    0, 2319, 2319, 2319,    0, 2319,    0,    0,    0, 2319,
 2319, 2319, 2319, 2319, 2319, 2319, 2319, 2319, 2319, 2319,
 2319, 2319, 2328, 2319, 2319,    0,    0,    0,    0,    0,
 2319, 2319, 2319, 2319, 2319, 2319, 2319, 2319, 2319, 2319,
 2319, 2319, 2319, 2319, 2319, 2319, 2319, 2319,    0,    0,
 2319,    0, 2319, 2319, 2319, 2319, 2319, 2319,    0,    0,
    0,    0, 1927, 2328, 2328, 2328, 2328, 2328, 2328, 2328,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2328, 2328, 2328, 2328, 2328,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 2328,    0, 2328,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1408,    0,    0,    0,    0,
    0,    0,    0,    0, 1409,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1410,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1411,    0,    0,    0,    0,    0,    0, 1412,    0,
    0,    0,    0,    0,    0,    0, 1413,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1414,    0, 1415,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1416,    0, 1417, 1418, 1419,
    0,    0,    0,    0, 1420,    0,    0,    0,    0, 2328,
    0, 2328,    0,    0, 2328,    0, 2328, 2328, 2328,    0,
 2328, 2328,    0, 2328,    0,    0,    0,    0,    0, 2328,
    0, 2328, 2328, 2328, 2328,    0, 2328,    0,    0, 2328,
    0,    0, 2328,    0,    0, 1421,    0,    0,    0, 2328,
    0, 2328, 2328,    0, 2328, 2328, 2328, 2328, 2328,    0,
    0, 2328,    0,    0, 2328,    0, 2328, 2328,    0,    0,
 2328, 2328, 2328,    0, 2328, 2328, 2328, 2328, 2328, 2328,
 2328,    0,    0,    0, 2328, 2328, 2328, 2328, 2328,    0,
    0,    0, 2328,    0,    0,    0,    0,    0,    0,    0,
 2328, 2328, 2328, 2328, 2328, 2328,    0,    0,    0, 2328,
 2328,    0, 2328,    0, 2328,    0, 2328, 2328, 2328,    0,
 2328,    0, 2328, 2328, 2328,    0, 2328, 2328, 2328, 2328,
 2328,    0, 2328, 2328,    0,    0,    0,    0, 2328,    0,
 2328, 2328,    0,    0, 2328, 2328, 2328,    0,    0, 2328,
    0, 2328,    0,    0, 2328, 2328, 2328, 2328,    0, 2328,
    0, 2328,    0, 2328, 2328, 2328, 2328, 2328,    0,    0,
 2328, 2328, 2328,    0, 2328, 2328,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 2328, 2328, 2328,    0, 2328,
 2328, 2328,    0,    0,    0,    0,    0, 2328,    0,    0,
    0, 2328,    0, 2328, 2328, 2328, 2328,    0, 2328,    0,
 2328, 2328, 2328, 2328, 2328, 2328, 2328, 2328,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 2328, 2328,    0,
 2328, 2328, 2328,    0,    0, 1422,    0, 2328, 2328, 2328,
 2328, 2328,    0, 2328,    0,    0, 2328, 2328,    0,    0,
    0, 2328, 2328,    0,    0, 2328, 2328, 2328, 2328, 2328,
 2328, 2328, 2328, 2328,    0, 2328,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 2328, 2328, 2328, 2328, 2328,
 2328, 2328, 2328, 2328, 2328, 2328, 2328, 2328, 2328, 2328,
 2328, 2328, 2328, 2328, 2328, 2328,    0, 2328, 2328, 2328,
 2328, 2328, 2328,    0, 2328, 2328, 2328, 2328,    0, 2328,
    0, 2328, 2328, 2328, 2328, 2328, 2328, 2328, 2328, 2328,
 2328, 2328, 2328, 2328, 2328, 2328, 2328, 2328, 2328,    0,
 2328, 2328, 2328, 2328,    0, 2328, 2328, 2328, 2328, 2328,
    0, 2328, 2328, 2328, 2328, 2328, 2328, 2328, 2328, 2328,
 2328, 2328, 2328, 2328, 2328, 2328, 2328, 2328, 2328, 2328,
 2328, 2328, 2328, 2328, 2328, 2328, 2328, 2328, 2328, 2328,
    0, 2328, 2328,    0,    0,    0, 2328, 2328, 2328, 2328,
 2328, 2328,    0,    0, 2328, 2328, 2328, 2328, 2328, 2328,
 2328, 2328, 2328, 3463,    0,    0,    0,    0, 2328,    0,
 2328, 2328, 2328, 2328, 2328, 2328,    0, 2328,    0,    0,
    0,    0, 1424, 1423, 1425,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 2328, 2328,    0, 2328, 2328, 2328,
 2328, 2328, 2328, 2328, 2328, 2328,    0, 2328, 2328, 2328,
 2328, 2328, 2328, 2328, 2328, 2328, 2328, 2328, 2328,    0,
 2328, 2328, 2328,    0, 2328,    0,    0,    0, 2328, 2328,
 2328, 2328, 2328, 2328, 2328, 2328, 2328, 2328, 2328, 2328,
 2328, 2332, 2328, 2328,    0,    0,    0,    0,    0, 2328,
 2328, 2328, 2328, 2328, 2328, 2328, 2328, 2328, 2328, 2328,
 2328, 2328, 2328, 2328, 2328, 2328, 2328,    0,    0, 2328,
    0, 2328, 2328, 2328, 2328, 2328, 2328,    0,    0,    0,
    0, 1955, 2332, 2332, 2332, 2332, 2332, 2332, 2332,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 2332,
 2332, 2332, 2332, 2332,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 2332,    0, 2332,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1408,    0,    0,    0,    0,    0,
    0,    0,    0, 1409,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1410,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1411,    0,    0,    0,    0,    0,    0, 1412,    0,    0,
    0,    0,    0,    0,    0, 1413,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1414,    0, 1415,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1416,    0, 1417, 1418, 1419,    0,
    0,    0,    0, 1420,    0,    0,    0,    0, 2332,    0,
 2332,    0,    0, 2332,    0, 2332, 2332, 2332,    0, 2332,
 2332,    0, 2332,    0,    0,    0,    0,    0, 2332,    0,
 2332, 2332, 2332, 2332,    0, 2332,    0,    0, 2332,    0,
    0, 2332,    0,    0, 1421,    0,    0,    0, 2332,    0,
 2332, 2332,    0, 2332, 2332, 2332, 2332, 2332,    0,    0,
 2332,    0,    0, 2332,    0, 2332, 2332,    0,    0, 2332,
 2332, 2332,    0, 2332, 2332, 2332, 2332, 2332, 2332, 2332,
    0,    0,    0, 2332, 2332, 2332, 2332, 2332,    0,    0,
    0, 2332,    0,    0,    0,    0,    0,    0,    0, 2332,
 2332, 2332, 2332, 2332, 2332,    0,    0,    0, 2332, 2332,
    0, 2332,    0, 2332,    0, 2332, 2332, 2332,    0, 2332,
    0, 2332, 2332, 2332,    0, 2332, 2332, 2332, 2332, 2332,
    0, 2332, 2332,    0,    0,    0,    0, 2332,    0, 2332,
 2332,    0,    0, 2332, 2332, 2332,    0,    0, 2332,    0,
 2332,    0,    0, 2332, 2332, 2332, 2332,    0, 2332,    0,
 2332,    0, 2332, 2332, 2332, 2332, 2332,    0,    0, 2332,
 2332, 2332,    0, 2332, 2332,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 2332, 2332, 2332,    0, 2332, 2332,
 2332,    0,    0,    0,    0,    0, 2332,    0,    0,    0,
 2332,    0, 2332, 2332, 2332, 2332,    0, 2332,    0, 2332,
 2332, 2332, 2332, 2332, 2332, 2332, 2332,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2332, 2332,    0, 2332,
 2332, 2332,    0,    0, 1422,    0, 2332, 2332, 2332, 2332,
 2332,    0, 2332,    0,    0, 2332, 2332,    0,    0,    0,
 2332, 2332,    0,    0, 2332, 2332, 2332, 2332, 2332, 2332,
 2332, 2332, 2332,    0, 2332,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 2332, 2332, 2332, 2332, 2332, 2332,
 2332, 2332, 2332, 2332, 2332, 2332, 2332, 2332, 2332, 2332,
 2332, 2332, 2332, 2332, 2332,    0, 2332, 2332, 2332, 2332,
 2332, 2332,    0, 2332, 2332, 2332, 2332,    0, 2332,    0,
 2332, 2332, 2332, 2332, 2332, 2332, 2332, 2332, 2332, 2332,
 2332, 2332, 2332, 2332, 2332, 2332, 2332, 2332,    0, 2332,
 2332, 2332, 2332,    0, 2332, 2332, 2332, 2332, 2332,    0,
 2332, 2332, 2332, 2332, 2332, 2332, 2332, 2332, 2332, 2332,
 2332, 2332, 2332, 2332, 2332, 2332, 2332, 2332, 2332, 2332,
 2332, 2332, 2332, 2332, 2332, 2332, 2332, 2332, 2332,    0,
 2332, 2332,    0,    0,    0, 2332, 2332, 2332, 2332, 2332,
 2332,    0,    0, 2332, 2332, 2332, 2332, 2332, 2332, 2332,
 2332, 2332, 3470,    0,    0,    0,    0, 2332,    0, 2332,
 2332, 2332, 2332, 2332, 2332,    0, 2332,    0,    0,    0,
    0, 1424, 1423, 1425,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 2332, 2332,    0, 2332, 2332, 2332, 2332,
 2332, 2332, 2332, 2332, 2332,    0, 2332, 2332, 2332, 2332,
 2332, 2332, 2332, 2332, 2332, 2332, 2332, 2332,    0, 2332,
 2332, 2332,    0, 2332,    0,    0,    0, 2332, 2332, 2332,
 2332, 2332, 2332, 2332, 2332, 2332, 2332, 2332, 2332, 2332,
 2342, 2332, 2332,    0,    0,    0,    0,    0, 2332, 2332,
 2332, 2332, 2332, 2332, 2332, 2332, 2332, 2332, 2332, 2332,
 2332, 2332, 2332, 2332, 2332, 2332,    0,    0, 2332,    0,
 2332, 2332, 2332, 2332, 2332, 2332,    0,    0,    0,    0,
 1953, 2342, 2342, 2342, 2342, 2342, 2342, 2342,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2342, 2342,
 2342, 2342, 2342,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 2342,    0, 2342,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1408,    0,    0,    0,    0,    0,    0,
    0,    0, 1409,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1410,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1411,
    0,    0,    0,    0,    0,    0, 1412,    0,    0,    0,
    0,    0,    0,    0, 1413,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1414,    0, 1415,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1416,    0, 1417, 1418, 1419,    0,    0,
    0,    0, 1420,    0,    0,    0,    0, 2342,    0, 2342,
    0,    0, 2342,    0, 2342, 2342, 2342,    0, 2342, 2342,
    0, 2342,    0,    0,    0,    0,    0, 2342,    0, 2342,
 2342, 2342, 2342,    0, 2342,    0,    0, 2342,    0,    0,
 2342,    0,    0, 1421,    0,    0,    0, 2342,    0, 2342,
 2342,    0, 2342, 2342, 2342, 2342, 2342,    0,    0, 2342,
    0,    0, 2342,    0, 2342, 2342,    0,    0, 2342, 2342,
 2342,    0, 2342, 2342, 2342, 2342, 2342, 2342, 2342,    0,
    0,    0, 2342, 2342, 2342, 2342, 2342,    0,    0,    0,
 2342,    0,    0,    0,    0,    0,    0,    0, 2342, 2342,
 2342, 2342, 2342, 2342,    0,    0,    0, 2342, 2342,    0,
 2342,    0, 2342,    0, 2342, 2342, 2342,    0, 2342,    0,
 2342, 2342, 2342,    0, 2342, 2342, 2342, 2342, 2342,    0,
 2342, 2342,    0,    0,    0,    0, 2342,    0, 2342, 2342,
    0,    0, 2342, 2342, 2342,    0,    0, 2342,    0, 2342,
    0,    0, 2342, 2342, 2342, 2342,    0, 2342,    0, 2342,
    0, 2342, 2342, 2342, 2342, 2342,    0,    0, 2342, 2342,
 2342,    0, 2342, 2342,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 2342, 2342, 2342,    0, 2342, 2342, 2342,
    0,    0,    0,    0,    0, 2342,    0,    0,    0, 2342,
    0, 2342, 2342, 2342, 2342,    0, 2342,    0, 2342, 2342,
 2342, 2342, 2342, 2342, 2342, 2342,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 2342, 2342,    0, 2342, 2342,
 2342,    0,    0, 1422,    0, 2342, 2342, 2342, 2342, 2342,
    0, 2342,    0,    0, 2342, 2342,    0,    0,    0, 2342,
 2342,    0,    0, 2342, 2342, 2342, 2342, 2342, 2342, 2342,
 2342, 2342,    0, 2342,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 2342, 2342, 2342, 2342, 2342, 2342, 2342,
 2342, 2342, 2342, 2342, 2342, 2342, 2342, 2342, 2342, 2342,
 2342, 2342, 2342, 2342,    0, 2342, 2342, 2342, 2342, 2342,
 2342,    0, 2342, 2342, 2342, 2342,    0, 2342,    0, 2342,
 2342, 2342, 2342, 2342, 2342, 2342, 2342, 2342, 2342, 2342,
 2342, 2342, 2342, 2342, 2342, 2342, 2342,    0, 2342, 2342,
 2342, 2342,    0, 2342, 2342, 2342, 2342, 2342,    0, 2342,
 2342, 2342, 2342, 2342, 2342, 2342, 2342, 2342, 2342, 2342,
 2342, 2342, 2342, 2342, 2342, 2342, 2342, 2342, 2342, 2342,
 2342, 2342, 2342, 2342, 2342, 2342, 2342, 2342,    0, 2342,
 2342,    0,    0,    0, 2342, 2342, 2342, 2342, 2342, 2342,
    0,    0, 2342, 2342, 2342, 2342, 2342, 2342, 2342, 2342,
 2342, 3471,    0,    0,    0,    0, 2342,    0, 2342, 2342,
 2342, 2342, 2342, 2342,    0, 2342,    0,    0,    0,    0,
 1424, 1423, 1425,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 2342, 2342,    0, 2342, 2342, 2342, 2342, 2342,
 2342, 2342, 2342, 2342,    0, 2342, 2342, 2342, 2342, 2342,
 2342, 2342, 2342, 2342, 2342, 2342, 2342,    0, 2342, 2342,
 2342,    0, 2342,    0,    0,    0, 2342, 2342, 2342, 2342,
 2342, 2342, 2342, 2342, 2342, 2342, 2342, 2342, 2342, 2345,
 2342, 2342,    0,    0,    0,    0,    0, 2342, 2342, 2342,
 2342, 2342, 2342, 2342, 2342, 2342, 2342, 2342, 2342, 2342,
 2342, 2342, 2342, 2342, 2342,    0,    0, 2342,    0, 2342,
 2342, 2342, 2342, 2342, 2342,    0,    0,    0,    0, 1976,
 2345, 2345, 2345, 2345, 2345, 2345, 2345,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 2345, 2345, 2345,
 2345, 2345,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2345,    0, 2345,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1408,    0,    0,    0,    0,    0,    0,    0,
    0, 1409,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1410,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1411,    0,
    0,    0,    0,    0,    0, 1412,    0,    0,    0,    0,
    0,    0,    0, 1413,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1414,    0, 1415,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1416,    0, 1417, 1418, 1419,    0,    0,    0,
    0, 1420,    0,    0,    0,    0, 2345,    0, 2345,    0,
    0, 2345,    0, 2345, 2345, 2345,    0, 2345, 2345,    0,
 2345,    0,    0,    0,    0,    0, 2345,    0, 2345, 2345,
 2345, 2345,    0, 2345,    0,    0, 2345,    0,    0, 2345,
    0,    0, 1421,    0,    0,    0, 2345,    0, 2345, 2345,
    0, 2345, 2345, 2345, 2345, 2345,    0,    0, 2345,    0,
    0, 2345,    0, 2345, 2345,    0,    0, 2345, 2345, 2345,
    0, 2345, 2345, 2345, 2345, 2345, 2345, 2345,    0,    0,
    0, 2345, 2345, 2345, 2345, 2345,    0,    0,    0, 2345,
    0,    0,    0,    0,    0,    0,    0, 2345, 2345, 2345,
 2345, 2345, 2345,    0,    0,    0, 2345, 2345,    0, 2345,
    0, 2345,    0, 2345, 2345, 2345,    0, 2345,    0, 2345,
 2345, 2345,    0, 2345, 2345, 2345, 2345, 2345,    0, 2345,
 2345,    0,    0,    0,    0, 2345,    0, 2345, 2345,    0,
    0, 2345, 2345, 2345,    0,    0, 2345,    0, 2345,    0,
    0, 2345, 2345, 2345, 2345,    0, 2345,    0, 2345,    0,
 2345, 2345, 2345, 2345, 2345,    0,    0, 2345, 2345, 2345,
    0, 2345, 2345,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 2345, 2345, 2345,    0, 2345, 2345, 2345,    0,
    0,    0,    0,    0, 2345,    0,    0,    0, 2345,    0,
 2345, 2345, 2345, 2345,    0, 2345,    0, 2345, 2345, 2345,
 2345, 2345, 2345, 2345, 2345,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 2345, 2345,    0, 2345, 2345, 2345,
    0,    0, 1422,    0, 2345, 2345, 2345, 2345, 2345,    0,
 2345,    0,    0, 2345, 2345,    0,    0,    0, 2345, 2345,
    0,    0, 2345, 2345, 2345, 2345, 2345, 2345, 2345, 2345,
 2345,    0, 2345,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 2345, 2345, 2345, 2345, 2345, 2345, 2345, 2345,
 2345, 2345, 2345, 2345, 2345, 2345, 2345, 2345, 2345, 2345,
 2345, 2345, 2345,    0, 2345, 2345, 2345, 2345, 2345, 2345,
    0, 2345, 2345, 2345, 2345,    0, 2345,    0, 2345, 2345,
 2345, 2345, 2345, 2345, 2345, 2345, 2345, 2345, 2345, 2345,
 2345, 2345, 2345, 2345, 2345, 2345,    0, 2345, 2345, 2345,
 2345,    0, 2345, 2345, 2345, 2345, 2345,    0, 2345, 2345,
 2345, 2345, 2345, 2345, 2345, 2345, 2345, 2345, 2345, 2345,
 2345, 2345, 2345, 2345, 2345, 2345, 2345, 2345, 2345, 2345,
 2345, 2345, 2345, 2345, 2345, 2345, 2345,    0, 2345, 2345,
    0,    0,    0, 2345, 2345, 2345, 2345, 2345, 2345,    0,
    0, 2345, 2345, 2345, 2345, 2345, 2345, 2345, 2345, 2345,
 3573,    0,    0,    0,    0, 2345,    0, 2345, 2345, 2345,
 2345, 2345, 2345,    0, 2345,    0,    0,    0,    0, 1424,
 1423, 1425,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 2345, 2345,    0, 2345, 2345, 2345, 2345, 2345, 2345,
 2345, 2345, 2345,    0, 2345, 2345, 2345, 2345, 2345, 2345,
 2345, 2345, 2345, 2345, 2345, 2345,    0, 2345, 2345, 2345,
    0, 2345,    0,    0,    0, 2345, 2345, 2345, 2345, 2345,
 2345, 2345, 2345, 2345, 2345, 2345, 2345, 2345, 2346, 2345,
 2345,    0,    0,    0,    0,    0, 2345, 2345, 2345, 2345,
 2345, 2345, 2345, 2345, 2345, 2345, 2345, 2345, 2345, 2345,
 2345, 2345, 2345, 2345,    0,    0, 2345,    0, 2345, 2345,
 2345, 2345, 2345, 2345,    0,    0,    0,    0, 1977, 2346,
 2346, 2346, 2346, 2346, 2346, 2346,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2346, 2346, 2346, 2346,
 2346,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 2346,
    0, 2346,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1408,    0,    0,    0,    0,    0,    0,    0,    0,
 1409,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1410,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1411,    0,    0,
    0,    0,    0,    0, 1412,    0,    0,    0,    0,    0,
    0,    0, 1413,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1414,
    0, 1415,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1416,    0, 1417, 1418, 1419,    0,    0,    0,    0,
 1420,    0,    0,    0,    0, 2346,    0, 2346,    0,    0,
 2346,    0, 2346, 2346, 2346,    0, 2346, 2346,    0, 2346,
    0,    0,    0,    0,    0, 2346,    0, 2346, 2346, 2346,
 2346,    0, 2346,    0,    0, 2346,    0,    0, 2346,    0,
    0, 1421,    0,    0,    0, 2346,    0, 2346, 2346,    0,
 2346, 2346, 2346, 2346, 2346,    0,    0, 2346,    0,    0,
 2346,    0, 2346, 2346,    0,    0, 2346, 2346, 2346,    0,
 2346, 2346, 2346, 2346, 2346, 2346, 2346,    0,    0,    0,
 2346, 2346, 2346, 2346, 2346,    0,    0,    0, 2346,    0,
    0,    0,    0,    0,    0,    0, 2346, 2346, 2346, 2346,
 2346, 2346,    0,    0,    0, 2346, 2346,    0, 2346,    0,
 2346,    0, 2346, 2346, 2346,    0, 2346,    0, 2346, 2346,
 2346,    0, 2346, 2346, 2346, 2346, 2346,    0, 2346, 2346,
    0,    0,    0,    0, 2346,    0, 2346, 2346,    0,    0,
 2346, 2346, 2346,    0,    0, 2346,    0, 2346,    0,    0,
 2346, 2346, 2346, 2346,    0, 2346,    0, 2346,    0, 2346,
 2346, 2346, 2346, 2346,    0,    0, 2346, 2346, 2346,    0,
 2346, 2346,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 2346, 2346, 2346,    0, 2346, 2346, 2346,    0,    0,
    0,    0,    0, 2346,    0,    0,    0, 2346,    0, 2346,
 2346, 2346, 2346,    0, 2346,    0, 2346, 2346, 2346, 2346,
 2346, 2346, 2346, 2346,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 2346, 2346,    0, 2346, 2346, 2346,    0,
    0, 1422,    0, 2346, 2346, 2346, 2346, 2346,    0, 2346,
    0,    0, 2346, 2346,    0,    0,    0, 2346, 2346,    0,
    0, 2346, 2346, 2346, 2346, 2346, 2346, 2346, 2346, 2346,
    0, 2346,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 2346, 2346, 2346, 2346, 2346, 2346, 2346, 2346, 2346,
 2346, 2346, 2346, 2346, 2346, 2346, 2346, 2346, 2346, 2346,
 2346, 2346,    0, 2346, 2346, 2346, 2346, 2346, 2346,    0,
 2346, 2346, 2346, 2346,    0, 2346,    0, 2346, 2346, 2346,
 2346, 2346, 2346, 2346, 2346, 2346, 2346, 2346, 2346, 2346,
 2346, 2346, 2346, 2346, 2346,    0, 2346, 2346, 2346, 2346,
    0, 2346, 2346, 2346, 2346, 2346,    0, 2346, 2346, 2346,
 2346, 2346, 2346, 2346, 2346, 2346, 2346, 2346, 2346, 2346,
 2346, 2346, 2346, 2346, 2346, 2346, 2346, 2346, 2346, 2346,
 2346, 2346, 2346, 2346, 2346, 2346,    0, 2346, 2346,    0,
    0,    0, 2346, 2346, 2346, 2346, 2346, 2346,    0,    0,
 2346, 2346, 2346, 2346, 2346, 2346, 2346, 2346, 2346, 3734,
    0,    0,    0,    0, 2346,    0, 2346, 2346, 2346, 2346,
 2346, 2346,    0, 2346,    0,    0,    0,    0, 1424, 1423,
 1425,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2346, 2346,    0, 2346, 2346, 2346, 2346, 2346, 2346, 2346,
 2346, 2346,    0, 2346, 2346, 2346, 2346, 2346, 2346, 2346,
 2346, 2346, 2346, 2346, 2346,    0, 2346, 2346, 2346,    0,
 2346,    0,    0,    0, 2346, 2346, 2346, 2346, 2346, 2346,
 2346, 2346, 2346, 2346, 2346, 2346, 2346, 2347, 2346, 2346,
    0,    0,    0,    0,    0, 2346, 2346, 2346, 2346, 2346,
 2346, 2346, 2346, 2346, 2346, 2346, 2346, 2346, 2346, 2346,
 2346, 2346, 2346,    0,    0, 2346,    0, 2346, 2346, 2346,
 2346, 2346, 2346,    0,    0,    0,    0, 1943, 2347, 2347,
 2347, 2347, 2347, 2347, 2347,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 2347, 2347, 2347, 2347, 2347,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2347,    0,
 2347,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1408,    0,    0,    0,    0,    0,    0,    0,    0, 1409,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1410,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1411,    0,    0,    0,
    0,    0,    0, 1412,    0,    0,    0,    0,    0,    0,
    0, 1413,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1414,    0,
 1415,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1416,    0, 1417, 1418, 1419,    0,    0,    0,    0, 1420,
    0,    0,    0,    0, 2347,    0, 2347,    0,    0, 2347,
    0, 2347, 2347, 2347,    0, 2347, 2347,    0, 2347,    0,
    0,    0,    0,    0, 2347,    0, 2347, 2347, 2347, 2347,
    0, 2347,    0,    0, 2347,    0,    0, 2347,    0,    0,
 1421,    0,    0,    0, 2347,    0, 2347, 2347,    0, 2347,
 2347, 2347, 2347, 2347,    0,    0, 2347,    0,    0, 2347,
    0, 2347, 2347,    0,    0, 2347, 2347, 2347,    0, 2347,
 2347, 2347, 2347, 2347, 2347, 2347,    0,    0,    0, 2347,
 2347, 2347, 2347, 2347,    0,    0,    0, 2347,    0,    0,
    0,    0,    0,    0,    0, 2347, 2347, 2347, 2347, 2347,
 2347,    0,    0,    0, 2347, 2347,    0, 2347,    0, 2347,
    0, 2347, 2347, 2347,    0, 2347,    0, 2347, 2347, 2347,
    0, 2347, 2347, 2347, 2347, 2347,    0, 2347, 2347,    0,
    0,    0,    0, 2347,    0, 2347, 2347,    0,    0, 2347,
 2347, 2347,    0,    0, 2347,    0, 2347,    0,    0, 2347,
 2347, 2347, 2347,    0, 2347,    0, 2347,    0, 2347, 2347,
 2347, 2347, 2347,    0,    0, 2347, 2347, 2347,    0, 2347,
 2347,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2347, 2347, 2347,    0, 2347, 2347, 2347,    0,    0,    0,
    0,    0, 2347,    0,    0,    0, 2347,    0, 2347, 2347,
 2347, 2347,    0, 2347,    0, 2347, 2347, 2347, 2347, 2347,
 2347, 2347, 2347,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 2347, 2347,    0, 2347, 2347, 2347,    0,    0,
 1422,    0, 2347, 2347, 2347, 2347, 2347,    0, 2347,    0,
    0, 2347, 2347,    0,    0,    0, 2347, 2347,    0,    0,
 2347, 2347, 2347, 2347, 2347, 2347, 2347, 2347, 2347,    0,
 2347,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2347, 2347, 2347, 2347, 2347, 2347, 2347, 2347, 2347, 2347,
 2347, 2347, 2347, 2347, 2347, 2347, 2347, 2347, 2347, 2347,
 2347,    0, 2347, 2347, 2347, 2347, 2347, 2347,    0, 2347,
 2347, 2347, 2347,    0, 2347,    0, 2347, 2347, 2347, 2347,
 2347, 2347, 2347, 2347, 2347, 2347, 2347, 2347, 2347, 2347,
 2347, 2347, 2347, 2347,    0, 2347, 2347, 2347, 2347,    0,
 2347, 2347, 2347, 2347, 2347,    0, 2347, 2347, 2347, 2347,
 2347, 2347, 2347, 2347, 2347, 2347, 2347, 2347, 2347, 2347,
 2347, 2347, 2347, 2347, 2347, 2347, 2347, 2347, 2347, 2347,
 2347, 2347, 2347, 2347, 2347,    0, 2347, 2347,    0,    0,
    0, 2347, 2347, 2347, 2347, 2347, 2347,    0,    0, 2347,
 2347, 2347, 2347, 2347, 2347, 2347, 2347, 2347, 3744,    0,
    0,    0,    0, 2347,    0, 2347, 2347, 2347, 2347, 2347,
 2347,    0, 2347,    0,    0,    0,    0, 1424, 1423, 1425,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 2347,
 2347,    0, 2347, 2347, 2347, 2347, 2347, 2347, 2347, 2347,
 2347,    0, 2347, 2347, 2347, 2347, 2347, 2347, 2347, 2347,
 2347, 2347, 2347, 2347,    0, 2347, 2347, 2347,    0, 2347,
    0,    0,    0, 2347, 2347, 2347, 2347, 2347, 2347, 2347,
 2347, 2347, 2347, 2347, 2347, 2347, 2348, 2347, 2347,    0,
    0,    0,    0,    0, 2347, 2347, 2347, 2347, 2347, 2347,
 2347, 2347, 2347, 2347, 2347, 2347, 2347, 2347, 2347, 2347,
 2347, 2347,    0,    0, 2347,    0, 2347, 2347, 2347, 2347,
 2347, 2347,    0,    0,    0,    0, 1944, 2348, 2348, 2348,
 2348, 2348, 2348, 2348,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 2348, 2348, 2348, 2348, 2348,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 2348,    0, 2348,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1408,
    0,    0,    0,    0,    0,    0,    0,    0, 1409,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1410,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1411,    0,    0,    0,    0,
    0,    0, 1412,    0,    0,    0,    0,    0,    0,    0,
 1413,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1414,    0, 1415,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1416,
    0, 1417, 1418, 1419,    0,    0,    0,    0, 1420,    0,
    0,    0,    0, 2348,    0, 2348,    0,    0, 2348,    0,
 2348, 2348, 2348,    0, 2348, 2348,    0, 2348,    0,    0,
    0,    0,    0, 2348,    0, 2348, 2348, 2348, 2348,    0,
 2348,    0,    0, 2348,    0,    0, 2348,    0,    0, 1421,
    0,    0,    0, 2348,    0, 2348, 2348,    0, 2348, 2348,
 2348, 2348, 2348,    0,    0, 2348,    0,    0, 2348,    0,
 2348, 2348,    0,    0, 2348, 2348, 2348,    0, 2348, 2348,
 2348, 2348, 2348, 2348, 2348,    0,    0,    0, 2348, 2348,
 2348, 2348, 2348,    0,    0,    0, 2348,    0,    0,    0,
    0,    0,    0,    0, 2348, 2348, 2348, 2348, 2348, 2348,
    0,    0,    0, 2348, 2348,    0, 2348,    0, 2348,    0,
 2348, 2348, 2348,    0, 2348,    0, 2348, 2348, 2348,    0,
 2348, 2348, 2348, 2348, 2348,    0, 2348, 2348,    0,    0,
    0,    0, 2348,    0, 2348, 2348,    0,    0, 2348, 2348,
 2348,    0,    0, 2348,    0, 2348,    0,    0, 2348, 2348,
 2348, 2348,    0, 2348,    0, 2348,    0, 2348, 2348, 2348,
 2348, 2348,    0,    0, 2348, 2348, 2348,    0, 2348, 2348,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 2348,
 2348, 2348,    0, 2348, 2348, 2348,    0,    0,    0,    0,
    0, 2348,    0,    0,    0, 2348,    0, 2348, 2348, 2348,
 2348,    0, 2348,    0, 2348, 2348, 2348, 2348, 2348, 2348,
 2348, 2348,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 2348, 2348,    0, 2348, 2348, 2348,    0,    0, 1422,
    0, 2348, 2348, 2348, 2348, 2348,    0, 2348,    0,    0,
 2348, 2348,    0,    0,    0, 2348, 2348,    0,    0, 2348,
 2348, 2348, 2348, 2348, 2348, 2348, 2348, 2348,    0, 2348,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 2348,
 2348, 2348, 2348, 2348, 2348, 2348, 2348, 2348, 2348, 2348,
 2348, 2348, 2348, 2348, 2348, 2348, 2348, 2348, 2348, 2348,
    0, 2348, 2348, 2348, 2348, 2348, 2348,    0, 2348, 2348,
 2348, 2348,    0, 2348,    0, 2348, 2348, 2348, 2348, 2348,
 2348, 2348, 2348, 2348, 2348, 2348, 2348, 2348, 2348, 2348,
 2348, 2348, 2348,    0, 2348, 2348, 2348, 2348,    0, 2348,
 2348, 2348, 2348, 2348,    0, 2348, 2348, 2348, 2348, 2348,
 2348, 2348, 2348, 2348, 2348, 2348, 2348, 2348, 2348, 2348,
 2348, 2348, 2348, 2348, 2348, 2348, 2348, 2348, 2348, 2348,
 2348, 2348, 2348, 2348,    0, 2348, 2348,    0,    0,    0,
 2348, 2348, 2348, 2348, 2348, 2348,    0,    0, 2348, 2348,
 2348, 2348, 2348, 2348, 2348, 2348, 2348, 3788,    0,    0,
    0,    0, 2348,    0, 2348, 2348, 2348, 2348, 2348, 2348,
    0, 2348,    0,    0,    0,    0, 1424, 1423, 1425,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2348, 2348,
    0, 2348, 2348, 2348, 2348, 2348, 2348, 2348, 2348, 2348,
    0, 2348, 2348, 2348, 2348, 2348, 2348, 2348, 2348, 2348,
 2348, 2348, 2348,    0, 2348, 2348, 2348,    0, 2348,    0,
    0,    0, 2348, 2348, 2348, 2348, 2348, 2348, 2348, 2348,
 2348, 2348, 2348, 2348, 2348, 2358, 2348, 2348,    0,    0,
    0,    0,    0, 2348, 2348, 2348, 2348, 2348, 2348, 2348,
 2348, 2348, 2348, 2348, 2348, 2348, 2348, 2348, 2348, 2348,
 2348,    0,    0, 2348,    0, 2348, 2348, 2348, 2348, 2348,
 2348,    0,    0,    0,    0, 1954, 2358, 2358, 2358, 2358,
 2358, 2358, 2358,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 2358, 2358, 2358, 2358, 2358,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2358,    0, 2358,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1408,    0,
    0,    0,    0,    0,    0,    0,    0, 1409,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1410,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1411,    0,    0,    0,    0,    0,
    0, 1412,    0,    0,    0,    0,    0,    0,    0, 1413,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1414,    0, 1415,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1416,    0,
 1417, 1418, 1419,    0,    0,    0,    0, 1420,    0,    0,
    0,    0, 2358,    0, 2358,    0,    0, 2358,    0, 2358,
 2358, 2358,    0, 2358, 2358,    0, 2358,    0,    0,    0,
    0,    0, 2358,    0, 2358, 2358, 2358, 2358,    0, 2358,
    0,    0, 2358,    0,    0, 2358,    0,    0, 1421,    0,
    0,    0, 2358,    0, 2358, 2358,    0, 2358, 2358, 2358,
 2358, 2358,    0,    0, 2358,    0,    0, 2358,    0, 2358,
 2358,    0,    0, 2358, 2358, 2358,    0, 2358, 2358, 2358,
 2358, 2358, 2358, 2358,    0,    0,    0, 2358, 2358, 2358,
 2358, 2358,    0,    0,    0, 2358,    0,    0,    0,    0,
    0,    0,    0, 2358, 2358, 2358, 2358, 2358, 2358,    0,
    0,    0, 2358, 2358,    0, 2358,    0, 2358,    0, 2358,
 2358, 2358,    0, 2358,    0, 2358, 2358, 2358,    0, 2358,
 2358, 2358, 2358, 2358,    0, 2358, 2358,    0,    0,    0,
    0, 2358,    0, 2358, 2358,    0,    0, 2358, 2358, 2358,
    0,    0, 2358,    0, 2358,    0,    0, 2358, 2358, 2358,
 2358,    0, 2358,    0, 2358,    0, 2358, 2358, 2358, 2358,
 2358,    0,    0, 2358, 2358, 2358,    0, 2358, 2358,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2358, 2358,
 2358,    0, 2358, 2358, 2358,    0,    0,    0,    0,    0,
 2358,    0,    0,    0, 2358,    0, 2358, 2358, 2358, 2358,
    0, 2358,    0, 2358, 2358, 2358, 2358, 2358, 2358, 2358,
 2358,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2358, 2358,    0, 2358, 2358, 2358,    0,    0, 1422,    0,
 2358, 2358, 2358, 2358, 2358,    0, 2358,    0,    0, 2358,
 2358,    0,    0,    0, 2358, 2358,    0,    0, 2358, 2358,
 2358, 2358, 2358, 2358, 2358, 2358, 2358,    0, 2358,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2358, 2358,
 2358, 2358, 2358, 2358, 2358, 2358, 2358, 2358, 2358, 2358,
 2358, 2358, 2358, 2358, 2358, 2358, 2358, 2358, 2358,    0,
 2358, 2358, 2358, 2358, 2358, 2358,    0, 2358, 2358, 2358,
 2358,    0, 2358,    0, 2358, 2358, 2358, 2358, 2358, 2358,
 2358, 2358, 2358, 2358, 2358, 2358, 2358, 2358, 2358, 2358,
 2358, 2358,    0, 2358, 2358, 2358, 2358,    0, 2358, 2358,
 2358, 2358, 2358,    0, 2358, 2358, 2358, 2358, 2358, 2358,
 2358, 2358, 2358, 2358, 2358, 2358, 2358, 2358, 2358, 2358,
 2358, 2358, 2358, 2358, 2358, 2358, 2358, 2358, 2358, 2358,
 2358, 2358, 2358,    0, 2358, 2358,    0,    0,    0, 2358,
 2358, 2358, 2358, 2358, 2358,    0,    0, 2358, 2358, 2358,
 2358, 2358, 2358, 2358, 2358, 2358,    0, 1687, 1687, 1685,
 1687, 2358, 1687, 2358, 2358, 2358, 2358, 2358, 2358,    0,
 2358,    0,    0,    0,    0, 1685, 1685, 1685,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 2358, 2358,    0,
 2358, 2358, 2358, 2358, 2358, 2358, 2358, 2358, 2358,    0,
 2358, 2358, 2358, 2358, 2358, 2358, 2358, 2358, 2358, 2358,
 2358, 2358,    0, 2358, 2358, 2358,    0, 2358,    0,    0,
    0, 2358, 2358, 2358, 2358, 2358, 2358, 2358, 2358, 2358,
 2358, 2358, 2358, 2358, 2107, 2358, 2358,    0,    0,    0,
    0,    0, 2358, 2358, 2358, 2358, 2358, 2358, 2358, 2358,
 2358, 2358, 2358, 2358, 2358, 2358, 2358, 2358, 2358, 2358,
    0,    0, 2358,    0, 2358, 2358, 2358, 2358, 2358, 2358,
    0,    0,    0,    0,    0, 2107, 2107, 2107, 2107, 2107,
 2107, 2107,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 2107, 2107, 2107, 2107, 2107,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 2107,    0, 2107,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1685,    0,    0,
    0, 1687,    0,    0,    0,    0, 1685,    0,    0,    0,
    0,    0, 1687,    0,    0,    0, 1687,    0,    0, 1685,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1685,    0,    0,    0,    0,    0,    0,
 1685,    0,    0,    0,    0,    0,    0,    0, 1685,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1685,    0, 1685,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1685,    0, 1685,
 1685, 1685,    0,    0,    0,    0, 1685,    0,    0,    0,
    0, 2107,    0, 2107,    0,    0, 2107,    0, 2107, 2107,
 2107,    0, 2107, 2107,    0, 2107,    0,    0,    0,    0,
    0, 2107,    0, 2107, 2107, 2107, 2107,    0, 2107,    0,
    0, 2107,    0,    0, 2107,    0,    0, 1685,    0,    0,
    0, 2107,    0, 2107, 2107,    0, 2107, 2107, 2107, 2107,
 2107,    0,    0, 2107,    0,    0, 2107,    0, 2107, 2107,
    0,    0, 2107, 2107, 2107,    0, 2107, 2107, 2107, 2107,
 2107, 2107, 2107,    0,    0,    0, 2107, 2107, 2107, 2107,
 2107,    0,    0,    0, 2107,    0,    0,    0,    0,    0,
    0,    0, 2107, 2107, 2107, 2107, 2107, 2107,    0,    0,
    0, 2107, 2107,    0, 2107,    0, 2107,    0, 2107, 2107,
 2107,    0, 2107,    0, 2107, 2107, 2107,    0, 2107, 2107,
 2107, 2107, 2107,    0, 2107, 2107,    0,    0,    0,    0,
 2107,    0, 2107, 2107,    0,    0, 2107, 2107, 2107,    0,
    0, 2107,    0, 2107,    0,    0, 2107, 2107, 2107, 2107,
    0, 2107,    0, 2107,    0, 2107, 2107, 2107, 2107, 2107,
    0,    0, 2107, 2107, 2107,    0, 2107, 2107,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 2107, 2107, 2107,
    0, 2107, 2107, 2107,    0,    0,    0,    0,    0, 2107,
    0,    0,    0, 2107,    0, 2107, 2107, 2107, 2107,    0,
 2107,    0, 2107, 2107, 2107, 2107, 2107, 2107, 2107, 2107,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 2107,
 2107,    0, 2107, 2107, 2107,    0,    0, 1685,    0, 2107,
 2107, 2107, 2107, 2107,    0, 2107,    0,    0, 2107, 2107,
    0,    0,    0, 2107, 2107,    0,    0, 2107, 2107, 2107,
 2107, 2107, 2107, 2107, 2107, 2107,    0, 2107,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 2107, 2107, 2107,
 2107, 2107, 2107, 2107, 2107, 2107, 2107, 2107, 2107, 2107,
 2107, 2107, 2107, 2107, 2107, 2107, 2107, 2107,    0, 2107,
 2107, 2107, 2107, 2107, 2107,    0, 2107, 2107, 2107, 2107,
    0, 2107,    0, 2107, 2107, 2107, 2107, 2107, 2107, 2107,
 2107, 2107, 2107, 2107, 2107, 2107, 2107, 2107, 2107, 2107,
 2107,    0, 2107, 2107, 2107, 2107,    0, 2107, 2107, 2107,
 2107, 2107,    0, 2107, 2107, 2107, 2107, 2107, 2107, 2107,
 2107, 2107, 2107, 2107, 2107, 2107, 2107, 2107, 2107, 2107,
 2107, 2107, 2107, 2107, 2107, 2107, 2107, 2107, 2107, 2107,
 2107, 2107,    0, 2107, 2107,    0,    0,    0, 2107, 2107,
 2107, 2107, 2107, 2107,    0,    0, 2107, 2107, 2107, 2107,
 2107, 2107, 2107, 2107, 2107,    0, 2152,    0,    0,    0,
 2107,    0, 2107, 2107, 2107, 2107, 2107, 2107,    0, 2107,
    0,    0, 1424, 1423, 1425,    0,    0,    0,    0,    0,
    0,    0,    0, 1424, 1423, 1425, 2107, 2107,    0, 2107,
 2107, 2107, 2107, 2107, 2107, 2107, 2107, 2107,    0, 2107,
 2107, 2107, 2107, 2107, 2107, 2107, 2107, 2107, 2107, 2107,
 2107,    0, 2107, 2107, 2107,    0, 2107,    0,    0,    0,
 2107, 2107, 2107, 2107, 2107, 2107, 2107, 2107, 2107, 2107,
 2107, 2107, 2107, 2152, 2107, 2107,    0,    0,    0,    0,
    0, 2107, 2107, 2107, 2107, 2107, 2107, 2107, 2107, 2107,
 2107, 2107, 2107, 2107, 2107, 2107, 2107, 2107, 2107,    0,
    0, 2107,    0, 2107, 2107, 2107, 2107, 2107, 2107,    0,
    0,    0,    0,    0, 2152, 2152, 2152, 2152, 2152, 2152,
 2152,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 2152, 2152, 2152, 2152, 2152,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 2152,    0, 2152,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1408,    0,    0,    0,    0,    0,
    0,    0,    0, 1409,    0, 1408,    0,  693,    0,    0,
    0,    0,    0,    0, 1409,    0, 1410,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1411,    0,    0,    0,    0,    0,    0, 1412,    0,    0,
    0, 1411,    0,    0,    0, 1413,    0,    0, 1412,    0,
    0,    0,    0,    0,    0,    0, 1413,    0,    0,    0,
    0,    0, 1414,    0, 1415,    0,    0,    0,    0,    0,
    0,    0,    0, 1414, 1416, 1415, 1417, 1418, 1419,    0,
    0,    0,    0, 1420,    0, 1416,    0, 1417, 1418, 1419,
 2152,    0, 2152,    0, 1420, 2152,    0, 2152, 2152, 2152,
    0, 2152, 2152,    0, 2152,    0,    0,    0,    0,    0,
 2152,    0, 2152, 2152, 2152, 2152,    0, 2152,    0,    0,
 2152,    0,    0, 2152, 1421,    0,    0,    0,    0,    0,
 2152,    0, 2152, 2152,    0, 2152, 2152, 2152, 2152, 2152,
    0,    0, 2152,    0,    0, 2152,    0, 2152, 2152,    0,
    0, 2152, 2152, 2152,    0, 2152, 2152, 2152, 2152, 2152,
 2152, 2152,    0,    0,    0, 2152, 2152, 2152, 2152, 2152,
    0,    0,    0, 2152,    0,    0,    0,    0,    0,    0,
    0, 2152, 2152, 2152, 2152, 2152, 2152,    0,    0,    0,
 2152, 2152,    0, 2152,    0, 2152,    0, 2152, 2152, 2152,
    0, 2152,    0, 2152, 2152, 2152,    0, 2152, 2152, 2152,
 2152, 2152,    0, 2152, 2152,    0,    0,    0,    0, 2152,
    0, 2152, 2152,    0,    0, 2152, 2152, 2152,    0,    0,
 2152,    0, 2152,    0,    0, 2152, 2152, 2152, 2152,    0,
 2152,    0, 2152,    0, 2152, 2152, 2152, 2152, 2152,    0,
    0, 2152, 2152, 2152,    0, 2152, 2152,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2152, 2152, 2152,    0,
 2152, 2152, 2152,    0,    0,    0,    0,    0, 2152,    0,
    0,    0, 2152,    0, 2152, 2152, 2152, 2152,    0, 2152,
    0, 2152, 2152, 2152, 2152, 2152, 2152, 2152, 2152,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2152, 2152,
    0, 2152, 2152, 2152, 1422,    0,    0,    0, 2152, 2152,
 2152, 2152, 2152,    0, 2152, 1422,    0, 2152, 2152,    0,
    0,    0, 2152, 2152,    0,    0, 2152, 2152, 2152, 2152,
 2152, 2152, 2152, 2152, 2152,    0, 2152,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2152, 2152, 2152, 2152,
 2152, 2152, 2152, 2152, 2152, 2152, 2152, 2152, 2152, 2152,
 2152, 2152, 2152, 2152, 2152, 2152, 2152,    0, 2152, 2152,
 2152, 2152, 2152, 2152,    0, 2152, 2152, 2152, 2152,    0,
 2152,    0, 2152, 2152, 2152, 2152, 2152, 2152, 2152, 2152,
 2152, 2152, 2152, 2152, 2152, 2152, 2152, 2152, 2152, 2152,
    0, 2152, 2152, 2152, 2152,    0, 2152, 2152, 2152, 2152,
 2152,    0, 2152, 2152, 2152, 2152, 2152, 2152, 2152, 2152,
 2152, 2152, 2152, 2152, 2152, 2152, 2152, 2152, 2152, 2152,
 2152, 2152, 2152, 2152, 2152, 2152, 2152, 2152, 2152, 2152,
 2152,    0, 2152, 2152,    0,    0,    0, 2152, 2152, 2152,
 2152, 2152, 2152,    0,    0, 2152, 2152, 2152, 2152, 2152,
 2152, 2152, 2152, 2152,    0, 2158,    0,    0,    0, 2152,
    0, 2152, 2152, 2152, 2152, 2152, 2152,    0, 2152,    0,
    0, 1424, 1423, 1425,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 2152, 2152,    0, 2152, 2152,
 2152, 2152, 2152, 2152, 2152, 2152, 2152,    0, 2152, 2152,
 2152, 2152, 2152, 2152, 2152, 2152, 2152, 2152, 2152, 2152,
    0, 2152, 2152, 2152,    0, 2152,    0,    0,    0, 2152,
 2152, 2152, 2152, 2152, 2152, 2152, 2152, 2152, 2152, 2152,
 2152, 2152, 2153, 2152, 2152,    0,    0,    0,    0,    0,
 2152, 2152, 2152, 2152, 2152, 2152, 2152, 2152, 2152, 2152,
 2152, 2152, 2152, 2152, 2152, 2152, 2152, 2152,    0,    0,
 2152,    0, 2152, 2152, 2152, 2152, 2152, 2152,    0,    0,
    0,    0,    0, 2153, 2153, 2153, 2153, 2153, 2153, 2153,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2153, 2153, 2153, 2153, 2153,    0,    0,    0,    0,    0,
    0,    0,    0,    0,   90,    0,    0,    0,    0,    0,
    0,    0,    0,    0,   91,    0,    0,    0,    0,    0,
   92,    0,    0, 2153,    0, 2153,    0,    0,    0,    0,
   93,    0,    0,   94,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1408,    0,    0,    0,    0,    0,    0,
    0,    0, 1409,    0,    0,   95,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1410,    0,    0,    0,    0,
    0,   96,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,   97,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1411,
    0,   98,    0,    0,    0,   99, 1412,    0,  100,    0,
  101,    0,    0,    0, 1413,    0,  102,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1414,    0, 1415,    0,    0,    0,    0,  103,    0,
    0,    0,    0, 1416,    0, 1417, 1418, 1419,    0,    0,
    0,    0, 1420,    0,    0,    0,    0,    0,    0, 2153,
    0, 2153,    0,    0, 2153,    0, 2153, 2153, 2153,    0,
 2153, 2153,    0, 2153,    0,    0,    0,    0,    0, 2153,
    0, 2153, 2153, 2153, 2153,    0, 2153,    0,    0, 2153,
    0,    0, 2153, 1421,  104,    0,    0,  105,    0, 2153,
    0, 2153, 2153,    0, 2153, 2153, 2153, 2153, 2153,    0,
    0, 2153,    0,    0, 2153,    0, 2153, 2153,    0,    0,
 2153, 2153, 2153,    0, 2153, 2153, 2153, 2153, 2153, 2153,
 2153,    0,    0,    0, 2153, 2153, 2153, 2153, 2153,    0,
  106,    0, 2153,    0,    0,    0,    0,    0,    0,    0,
 2153, 2153, 2153, 2153, 2153, 2153,    0,    0,    0, 2153,
 2153,    0, 2153,    0, 2153,    0, 2153, 2153, 2153,    0,
 2153,    0, 2153, 2153, 2153,    0, 2153, 2153, 2153, 2153,
 2153,    0, 2153, 2153,  107,    0,    0,    0, 2153,    0,
 2153, 2153,    0,    0, 2153, 2153, 2153,    0,    0, 2153,
    0, 2153,    0,  108, 2153, 2153, 2153, 2153,    0, 2153,
    0, 2153,    0, 2153, 2153, 2153, 2153, 2153,    0,    0,
 2153, 2153, 2153,    0, 2153, 2153,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 2153, 2153, 2153,    0, 2153,
 2153, 2153,    0,    0,    0,    0,    0, 2153,    0,    0,
    0, 2153,    0, 2153, 2153, 2153, 2153,    0, 2153,    0,
 2153, 2153, 2153, 2153, 2153, 2153, 2153, 2153,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 2153, 2153,    0,
 2153, 2153, 2153, 1422,    0,    0,    0, 2153, 2153, 2153,
 2153, 2153,    0, 2153,    0,    0, 2153, 2153,    0,    0,
    0, 2153, 2153,    0,    0, 2153, 2153, 2153, 2153, 2153,
 2153, 2153, 2153, 2153,    0, 2153,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 2153, 2153, 2153, 2153, 2153,
 2153, 2153, 2153, 2153, 2153, 2153, 2153, 2153, 2153, 2153,
 2153, 2153, 2153, 2153, 2153, 2153,    0, 2153, 2153, 2153,
 2153, 2153, 2153,    0, 2153, 2153, 2153, 2153,    0, 2153,
    0, 2153, 2153, 2153, 2153, 2153, 2153, 2153, 2153, 2153,
 2153, 2153, 2153, 2153, 2153, 2153, 2153, 2153, 2153,    0,
 2153, 2153, 2153, 2153,    0, 2153, 2153, 2153, 2153, 2153,
    0, 2153, 2153, 2153, 2153, 2153, 2153, 2153, 2153, 2153,
 2153, 2153, 2153, 2153, 2153, 2153, 2153, 2153, 2153, 2153,
 2153, 2153, 2153, 2153, 2153, 2153, 2153, 2153, 2153, 2153,
    0, 2153, 2153,    0,    0,    0, 2153, 2153, 2153, 2153,
 2153, 2153,    0,    0, 2153, 2153, 2153, 2153, 2153, 2153,
 2153, 2153, 2153,    0, 2168,    0,    0,    0, 2153,    0,
 2153, 2153, 2153, 2153, 2153, 2153,    0, 2153,    0,    0,
 1424, 1423, 1425,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 2153, 2153,    0, 2153, 2153, 2153,
 2153, 2153, 2153, 2153, 2153, 2153,    0, 2153, 2153, 2153,
 2153, 2153, 2153, 2153, 2153, 2153, 2153, 2153, 2153,    0,
 2153, 2153, 2153,    0, 2153,    0,    0,    0, 2153, 2153,
 2153, 2153, 2153, 2153, 2153, 2153, 2153, 2153, 2153, 2153,
 2153, 2159, 2153, 2153,    0,    0,    0,    0,    0, 2153,
 2153, 2153, 2153, 2153, 2153, 2153, 2153, 2153, 2153, 2153,
 2153, 2153, 2153, 2153, 2153, 2153, 2153,    0,    0, 2153,
    0, 2153, 2153, 2153, 2153, 2153, 2153,    0,    0,    0,
    0,    0, 2159, 2159, 2159, 2159, 2159, 2159, 2159,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 2159,
 2159, 2159, 2159, 2159,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 2159,    0, 2159,    0,    0,    0,    0,  115,
    0,    0,    0,    0,    0,  116,    0,    0,    0,  117,
  118,    0,    0,    0,    0,  119,    0,    0,  120,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  121,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1408,    0,    0,    0,    0,    0,    0,    0,
    0, 1409,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1410,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  122,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1411,    0,
    0,  123,    0,    0,    0, 1412,    0,    0,    0,    0,
    0,    0,    0, 1413,    0,    0,  124,    0,    0,    0,
  125,    0,    0,    0,    0,  126,    0,    0,    0,    0,
 1414,  127, 1415,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1416,    0, 1417, 1418, 1419,    0,    0,    0,
    0, 1420,    0,  128,    0,    0,    0,    0, 2159,    0,
 2159,    0,    0, 2159,    0, 2159, 2159, 2159,    0, 2159,
 2159,    0, 2159,    0,    0,    0,    0,    0, 2159,    0,
 2159, 2159, 2159, 2159,    0, 2159,    0,    0, 2159,    0,
    0, 2159, 1421,    0,    0,    0,    0,    0, 2159,    0,
 2159, 2159,    0, 2159, 2159, 2159, 2159, 2159,    0,  129,
 2159,    0,  130, 2159,    0, 2159, 2159,    0,    0, 2159,
 2159, 2159,    0, 2159, 2159, 2159, 2159, 2159, 2159, 2159,
    0,    0,    0, 2159, 2159, 2159, 2159, 2159,    0,    0,
    0, 2159,    0,    0,    0,    0,    0,    0,    0, 2159,
 2159, 2159, 2159, 2159, 2159,  131,    0,    0, 2159, 2159,
    0, 2159,    0, 2159,    0, 2159, 2159, 2159,    0, 2159,
    0, 2159, 2159, 2159,    0, 2159, 2159, 2159, 2159, 2159,
    0, 2159, 2159,    0,    0,    0,    0, 2159,    0, 2159,
 2159,    0,    0, 2159, 2159, 2159,    0,    0, 2159,  132,
 2159,    0,    0, 2159, 2159, 2159, 2159,    0, 2159,    0,
 2159,    0, 2159, 2159, 2159, 2159, 2159,    0,  133, 2159,
 2159, 2159,    0, 2159, 2159,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 2159, 2159, 2159,    0, 2159, 2159,
 2159,    0,    0,    0,    0,    0, 2159,    0,    0,    0,
 2159,    0, 2159, 2159, 2159, 2159,    0, 2159,    0, 2159,
 2159, 2159, 2159, 2159, 2159, 2159, 2159,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2159, 2159,    0, 2159,
 2159, 2159, 1422,    0,    0,    0, 2159, 2159, 2159, 2159,
 2159,    0, 2159,    0,    0, 2159, 2159,    0,    0,    0,
 2159, 2159,    0,    0, 2159, 2159, 2159, 2159, 2159, 2159,
 2159, 2159, 2159,    0, 2159,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 2159, 2159, 2159, 2159, 2159, 2159,
 2159, 2159, 2159, 2159, 2159, 2159, 2159, 2159, 2159, 2159,
 2159, 2159, 2159, 2159, 2159,    0, 2159, 2159, 2159, 2159,
 2159, 2159,    0, 2159, 2159, 2159, 2159,    0, 2159,    0,
 2159, 2159, 2159, 2159, 2159, 2159, 2159, 2159, 2159, 2159,
 2159, 2159, 2159, 2159, 2159, 2159, 2159, 2159,    0, 2159,
 2159, 2159, 2159,    0, 2159, 2159, 2159, 2159, 2159,    0,
 2159, 2159, 2159, 2159, 2159, 2159, 2159, 2159, 2159, 2159,
 2159, 2159, 2159, 2159, 2159, 2159, 2159, 2159, 2159, 2159,
 2159, 2159, 2159, 2159, 2159, 2159, 2159, 2159, 2159,    0,
 2159, 2159,    0,    0,    0, 2159, 2159, 2159, 2159, 2159,
 2159,    0,    0, 2159, 2159, 2159, 2159, 2159, 2159, 2159,
 2159, 2159,    0, 2169,    0,    0,    0, 2159,    0, 2159,
 2159, 2159, 2159, 2159, 2159,    0, 2159,    0,    0, 1424,
 1423, 1425,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 2159, 2159,    0, 2159, 2159, 2159, 2159,
 2159, 2159, 2159, 2159, 2159,    0, 2159, 2159, 2159, 2159,
 2159, 2159, 2159, 2159, 2159, 2159, 2159, 2159,    0, 2159,
 2159, 2159,    0, 2159,    0,    0,    0, 2159, 2159, 2159,
 2159, 2159, 2159, 2159, 2159, 2159, 2159, 2159, 2159, 2159,
 2171, 2159, 2159,    0,    0,    0,    0,    0, 2159, 2159,
 2159, 2159, 2159, 2159, 2159, 2159, 2159, 2159, 2159, 2159,
 2159, 2159, 2159, 2159, 2159, 2159,    0,    0, 2159,    0,
 2159, 2159, 2159, 2159, 2159, 2159,    0,    0,    0,    0,
    0, 2171, 2171, 2171, 2171, 2171, 2171, 2171,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2171, 2171,
 2171, 2171, 2171,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 2171,    0, 2171,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1408,    0,    0,    0,    0,    0,    0,    0,    0,
 1409,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1410,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1411,    0,    0,
    0,    0,    0,    0, 1412,    0,    0,    0,    0,    0,
    0,    0, 1413,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1414,
    0, 1415,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1416,    0, 1417, 1418, 1419,    0,    0,    0,    0,
 1420,    0,    0,    0,    0,    0,    0, 2171,    0, 2171,
    0,    0, 2171,    0, 2171, 2171, 2171,    0, 2171, 2171,
    0, 2171,    0,    0,    0,    0,    0, 2171,    0, 2171,
 2171, 2171, 2171,    0, 2171,    0,    0, 2171,    0,    0,
 2171, 1421,    0,    0,    0,    0,    0, 2171,    0, 2171,
 2171,    0, 2171, 2171, 2171, 2171, 2171,    0,    0, 2171,
    0,    0, 2171,    0, 2171, 2171,    0,    0, 2171, 2171,
 2171,    0, 2171, 2171, 2171, 2171, 2171, 2171, 2171,    0,
    0,    0, 2171, 2171, 2171, 2171, 2171,    0,    0,    0,
 2171,    0,    0,    0,    0,    0,    0,    0, 2171, 2171,
 2171, 2171, 2171, 2171,    0,    0,    0, 2171, 2171,    0,
 2171,    0, 2171,    0, 2171, 2171, 2171,    0, 2171,    0,
 2171, 2171, 2171,    0, 2171, 2171, 2171, 2171, 2171,    0,
 2171, 2171,    0,    0,    0,    0, 2171,    0, 2171, 2171,
    0,    0, 2171, 2171, 2171,    0,    0, 2171,    0, 2171,
    0,    0, 2171, 2171, 2171, 2171,    0, 2171,    0, 2171,
    0, 2171, 2171, 2171, 2171, 2171,    0,    0, 2171, 2171,
 2171,    0, 2171, 2171,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 2171, 2171, 2171,    0, 2171, 2171, 2171,
    0,    0,    0,    0,    0, 2171,    0,    0,    0, 2171,
    0, 2171, 2171, 2171, 2171,    0, 2171,    0, 2171, 2171,
 2171, 2171, 2171, 2171, 2171, 2171,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 2171, 2171,    0, 2171, 2171,
 2171, 1422,    0,    0,    0, 2171, 2171, 2171, 2171, 2171,
    0, 2171,    0,    0, 2171, 2171,    0,    0,    0, 2171,
 2171,    0,    0, 2171, 2171, 2171, 2171, 2171, 2171, 2171,
 2171, 2171,    0, 2171,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 2171, 2171, 2171, 2171, 2171, 2171, 2171,
 2171, 2171, 2171, 2171, 2171, 2171, 2171, 2171, 2171, 2171,
 2171, 2171, 2171, 2171,    0, 2171, 2171, 2171, 2171, 2171,
 2171,    0, 2171, 2171, 2171, 2171,    0, 2171,    0, 2171,
 2171, 2171, 2171, 2171, 2171, 2171, 2171, 2171, 2171, 2171,
 2171, 2171, 2171, 2171, 2171, 2171, 2171,    0, 2171, 2171,
 2171, 2171,    0, 2171, 2171, 2171, 2171, 2171,    0, 2171,
 2171, 2171, 2171, 2171, 2171, 2171, 2171, 2171, 2171, 2171,
 2171, 2171, 2171, 2171, 2171, 2171, 2171, 2171, 2171, 2171,
 2171, 2171, 2171, 2171, 2171, 2171, 2171, 2171,    0, 2171,
 2171,    0,    0,    0, 2171, 2171, 2171, 2171, 2171, 2171,
    0,    0, 2171, 2171, 2171, 2171, 2171, 2171, 2171, 2171,
 2171,    0, 2170,    0,    0,    0, 2171,    0, 2171, 2171,
 2171, 2171, 2171, 2171,    0, 2171,    0,    0, 1424, 1423,
 1425,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 2171, 2171,    0, 2171, 2171, 2171, 2171, 2171,
 2171, 2171, 2171, 2171,    0, 2171, 2171, 2171, 2171, 2171,
 2171, 2171, 2171, 2171, 2171, 2171, 2171,    0, 2171, 2171,
 2171,    0, 2171,    0,    0,    0, 2171, 2171, 2171, 2171,
 2171, 2171, 2171, 2171, 2171, 2171, 2171, 2171, 2171, 2316,
 2171, 2171,    0,    0,    0,    0,    0, 2171, 2171, 2171,
 2171, 2171, 2171, 2171, 2171, 2171, 2171, 2171, 2171, 2171,
 2171, 2171, 2171, 2171, 2171,    0,    0, 2171,    0, 2171,
 2171, 2171, 2171, 2171, 2171,    0,    0,    0,    0,    0,
 2316, 2316, 2316, 2316, 2316, 2316, 2316,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 2316, 2316, 2316,
 2316, 2316,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2316,    0, 2316,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1408,    0,    0,    0,    0,    0,    0,    0,    0, 1409,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1410,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1411,    0,    0,    0,
    0,    0,    0, 1412,    0,    0,    0,    0,    0,    0,
    0, 1413,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1414,    0,
 1415,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1416,    0, 1417, 1418, 1419,    0,    0,    0,    0, 1420,
    0,    0,    0,    0,    0,    0, 2316,    0, 2316,    0,
    0, 2316,    0, 2316, 2316, 2316,    0, 2316, 2316,    0,
 2316,    0,    0,    0,    0,    0, 2316,    0, 2316, 2316,
 2316, 2316,    0, 2316,    0,    0, 2316,    0,    0, 2316,
 1421,    0,    0,    0,    0,    0, 2316,    0, 2316, 2316,
    0, 2316, 2316, 2316, 2316, 2316,    0,    0, 2316,    0,
    0, 2316,    0, 2316, 2316,    0,    0, 2316, 2316, 2316,
    0, 2316, 2316, 2316, 2316, 2316, 2316, 2316,    0,    0,
    0, 2316, 2316, 2316, 2316, 2316,    0,    0,    0, 2316,
    0,    0,    0,    0,    0,    0,    0, 2316, 2316, 2316,
 2316, 2316, 2316,    0,    0,    0, 2316, 2316,    0, 2316,
    0, 2316,    0, 2316, 2316, 2316,    0, 2316,    0, 2316,
 2316, 2316,    0, 2316, 2316, 2316, 2316, 2316,    0, 2316,
 2316,    0,    0,    0,    0, 2316,    0, 2316, 2316,    0,
    0, 2316, 2316, 2316,    0,    0, 2316,    0, 2316,    0,
    0, 2316, 2316, 2316, 2316,    0, 2316,    0, 2316,    0,
 2316, 2316, 2316, 2316, 2316,    0,    0, 2316, 2316, 2316,
    0, 2316, 2316,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 2316, 2316, 2316,    0, 2316, 2316, 2316,    0,
    0,    0,    0,    0, 2316,    0,    0,    0, 2316,    0,
 2316, 2316, 2316, 2316,    0, 2316,    0, 2316, 2316, 2316,
 2316, 2316, 2316, 2316, 2316,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 2316, 2316,    0, 2316, 2316, 2316,
 1422,    0,    0,    0, 2316, 2316, 2316, 2316, 2316,    0,
 2316,    0,    0, 2316, 2316,    0,    0,    0, 2316, 2316,
    0,    0, 2316, 2316, 2316, 2316, 2316, 2316, 2316, 2316,
 2316,    0, 2316,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 2316, 2316, 2316, 2316, 2316, 2316, 2316, 2316,
 2316, 2316, 2316, 2316, 2316, 2316, 2316, 2316, 2316, 2316,
 2316, 2316, 2316,    0, 2316, 2316, 2316, 2316, 2316, 2316,
    0, 2316, 2316, 2316, 2316,    0, 2316,    0, 2316, 2316,
 2316, 2316, 2316, 2316, 2316, 2316, 2316, 2316, 2316, 2316,
 2316, 2316, 2316, 2316, 2316, 2316,    0, 2316, 2316, 2316,
 2316,    0, 2316, 2316, 2316, 2316, 2316,    0, 2316, 2316,
 2316, 2316, 2316, 2316, 2316, 2316, 2316, 2316, 2316, 2316,
 2316, 2316, 2316, 2316, 2316, 2316, 2316, 2316, 2316, 2316,
 2316, 2316, 2316, 2316, 2316, 2316, 2316,    0, 2316, 2316,
    0,    0,    0, 2316, 2316, 2316, 2316, 2316, 2316,    0,
    0, 2316, 2316, 2316, 2316, 2316, 2316, 2316, 2316, 2316,
    0, 2171,    0,    0,    0, 2316,    0, 2316, 2316, 2316,
 2316, 2316, 2316,    0, 2316,    0,    0, 1424, 1423, 1425,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 2316, 2316,    0, 2316, 2316, 2316, 2316, 2316, 2316,
 2316, 2316, 2316,    0, 2316, 2316, 2316, 2316, 2316, 2316,
 2316, 2316, 2316, 2316, 2316, 2316,    0, 2316, 2316, 2316,
    0, 2316,    0,    0,    0, 2316, 2316, 2316, 2316, 2316,
 2316, 2316, 2316, 2316, 2316, 2316, 2316, 2316, 2324, 2316,
 2316,    0,    0,    0,    0,    0, 2316, 2316, 2316, 2316,
 2316, 2316, 2316, 2316, 2316, 2316, 2316, 2316, 2316, 2316,
 2316, 2316, 2316, 2316,    0,    0, 2316,    0, 2316, 2316,
 2316, 2316, 2316, 2316,    0,    0,    0,    0,    0, 2324,
 2324, 2324, 2324, 2324, 2324, 2324,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2324, 2324, 2324, 2324,
 2324,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 2324,
    0, 2324,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1408,
    0,    0,    0,    0,    0,    0,    0,    0, 1409,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1410,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1411,    0,    0,    0,    0,
    0,    0, 1412,    0,    0,    0,    0,    0,    0,    0,
 1413,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1414,    0, 1415,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1416,
    0, 1417, 1418, 1419,    0,    0,    0,    0, 1420,    0,
    0,    0,    0,    0,    0, 2324,    0, 2324,    0,    0,
 2324,    0, 2324, 2324, 2324,    0, 2324, 2324,    0, 2324,
    0,    0,    0,    0,    0, 2324,    0, 2324, 2324, 2324,
 2324,    0, 2324,    0,    0, 2324,    0,    0, 2324, 1421,
    0,    0,    0,    0,    0, 2324,    0, 2324, 2324,    0,
 2324, 2324, 2324, 2324, 2324,    0,    0, 2324,    0,    0,
 2324,    0, 2324, 2324,    0,    0, 2324, 2324, 2324,    0,
 2324, 2324, 2324, 2324, 2324, 2324, 2324,    0,    0,    0,
 2324, 2324, 2324, 2324, 2324,    0,    0,    0, 2324,    0,
    0,    0,    0,    0,    0,    0, 2324, 2324, 2324, 2324,
 2324, 2324,    0,    0,    0, 2324, 2324,    0, 2324,    0,
 2324,    0, 2324, 2324, 2324,    0, 2324,    0, 2324, 2324,
 2324,    0, 2324, 2324, 2324, 2324, 2324,    0, 2324, 2324,
    0,    0,    0,    0, 2324,    0, 2324, 2324,    0,    0,
 2324, 2324, 2324,    0,    0, 2324,    0, 2324,    0,    0,
 2324, 2324, 2324, 2324,    0, 2324,    0, 2324,    0, 2324,
 2324, 2324, 2324, 2324,    0,    0, 2324, 2324, 2324,    0,
 2324, 2324,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 2324, 2324, 2324,    0, 2324, 2324, 2324,    0,    0,
    0,    0,    0, 2324,    0,    0,    0, 2324,    0, 2324,
 2324, 2324, 2324,    0, 2324,    0, 2324, 2324, 2324, 2324,
 2324, 2324, 2324, 2324,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 2324, 2324,    0, 2324, 2324, 2324, 1422,
    0,    0,    0, 2324, 2324, 2324, 2324, 2324,    0, 2324,
    0,    0, 2324, 2324,    0,    0,    0, 2324, 2324,    0,
    0, 2324, 2324, 2324, 2324, 2324, 2324, 2324, 2324, 2324,
    0, 2324,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 2324, 2324, 2324, 2324, 2324, 2324, 2324, 2324, 2324,
 2324, 2324, 2324, 2324, 2324, 2324, 2324, 2324, 2324, 2324,
 2324, 2324,    0, 2324, 2324, 2324, 2324, 2324, 2324,    0,
 2324, 2324, 2324, 2324,    0, 2324,    0, 2324, 2324, 2324,
 2324, 2324, 2324, 2324, 2324, 2324, 2324, 2324, 2324, 2324,
 2324, 2324, 2324, 2324, 2324,    0, 2324, 2324, 2324, 2324,
    0, 2324, 2324, 2324, 2324, 2324,    0, 2324, 2324, 2324,
 2324, 2324, 2324, 2324, 2324, 2324, 2324, 2324, 2324, 2324,
 2324, 2324, 2324, 2324, 2324, 2324, 2324, 2324, 2324, 2324,
 2324, 2324, 2324, 2324, 2324, 2324,    0, 2324, 2324,    0,
    0,    0, 2324, 2324, 2324, 2324, 2324, 2324,    0,    0,
 2324, 2324, 2324, 2324, 2324, 2324, 2324, 2324, 2324,    0,
 2172,    0,    0,    0, 2324,    0, 2324, 2324, 2324, 2324,
 2324, 2324,    0, 2324,    0,    0, 1424, 1423, 1425,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2324, 2324,    0, 2324, 2324, 2324, 2324, 2324, 2324, 2324,
 2324, 2324,    0, 2324, 2324, 2324, 2324, 2324, 2324, 2324,
 2324, 2324, 2324, 2324, 2324,    0, 2324, 2324, 2324,    0,
 2324,    0,    0,    0, 2324, 2324, 2324, 2324, 2324, 2324,
 2324, 2324, 2324, 2324, 2324, 2324, 2324, 2349, 2324, 2324,
    0,    0,    0,    0,    0, 2324, 2324, 2324, 2324, 2324,
 2324, 2324, 2324, 2324, 2324, 2324, 2324, 2324, 2324, 2324,
 2324, 2324, 2324,    0,    0, 2324,    0, 2324, 2324, 2324,
 2324, 2324, 2324,    0,    0,    0,    0,    0, 2349, 2349,
 2349, 2349, 2349, 2349, 2349,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 2349, 2349, 2349, 2349, 2349,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2349,    0,
 2349,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1408,    0,
    0,    0,    0,    0,    0,    0,    0, 1409,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1410,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1411,    0,    0,    0,    0,    0,
    0, 1412,    0,    0,    0,    0,    0,    0,    0, 1413,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1414,    0, 1415,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1416,    0,
 1417, 1418, 1419,    0,    0,    0,    0, 1420,    0,    0,
    0,    0,    0,    0, 2349,    0, 2349,    0,    0, 2349,
    0, 2349, 2349, 2349,    0, 2349, 2349,    0, 2349,    0,
    0,    0,    0,    0, 2349,    0, 2349, 2349, 2349, 2349,
    0, 2349,    0,    0, 2349,    0,    0, 2349, 1421,    0,
    0,    0,    0,    0, 2349,    0, 2349, 2349,    0, 2349,
 2349, 2349, 2349, 2349,    0,    0, 2349,    0,    0, 2349,
    0, 2349, 2349,    0,    0, 2349, 2349, 2349,    0, 2349,
 2349, 2349, 2349, 2349, 2349, 2349,    0,    0,    0, 2349,
 2349, 2349, 2349, 2349,    0,    0,    0, 2349,    0,    0,
    0,    0,    0,    0,    0, 2349, 2349, 2349, 2349, 2349,
 2349,    0,    0,    0, 2349, 2349,    0, 2349,    0, 2349,
    0, 2349, 2349, 2349,    0, 2349,    0, 2349, 2349, 2349,
    0, 2349, 2349, 2349, 2349, 2349,    0, 2349, 2349,    0,
    0,    0,    0, 2349,    0, 2349, 2349,    0,    0, 2349,
 2349, 2349,    0,    0, 2349,    0, 2349,    0,    0, 2349,
 2349, 2349, 2349,    0, 2349,    0, 2349,    0, 2349, 2349,
 2349, 2349, 2349,    0,    0, 2349, 2349, 2349,    0, 2349,
 2349,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2349, 2349, 2349,    0, 2349, 2349, 2349,    0,    0,    0,
    0,    0, 2349,    0,    0,    0, 2349,    0, 2349, 2349,
 2349, 2349,    0, 2349,    0, 2349, 2349, 2349, 2349, 2349,
 2349, 2349, 2349,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 2349, 2349,    0, 2349, 2349, 2349, 1422,    0,
    0,    0, 2349, 2349, 2349, 2349, 2349,    0, 2349,    0,
    0, 2349, 2349,    0,    0,    0, 2349, 2349,    0,    0,
 2349, 2349, 2349, 2349, 2349, 2349, 2349, 2349, 2349,    0,
 2349,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2349, 2349, 2349, 2349, 2349, 2349, 2349, 2349, 2349, 2349,
 2349, 2349, 2349, 2349, 2349, 2349, 2349, 2349, 2349, 2349,
 2349,    0, 2349, 2349, 2349, 2349, 2349, 2349,    0, 2349,
 2349, 2349, 2349,    0, 2349,    0, 2349, 2349, 2349, 2349,
 2349, 2349, 2349, 2349, 2349, 2349, 2349, 2349, 2349, 2349,
 2349, 2349, 2349, 2349,    0, 2349, 2349, 2349, 2349,    0,
 2349, 2349, 2349, 2349, 2349,    0, 2349, 2349, 2349, 2349,
 2349, 2349, 2349, 2349, 2349, 2349, 2349, 2349, 2349, 2349,
 2349, 2349, 2349, 2349, 2349, 2349, 2349, 2349, 2349, 2349,
 2349, 2349, 2349, 2349, 2349,    0, 2349, 2349,    0,    0,
    0, 2349, 2349, 2349, 2349, 2349, 2349,    0,    0, 2349,
 2349, 2349, 2349, 2349, 2349, 2349, 2349, 2349,    0, 2173,
    0,    0,    0, 2349,    0, 2349, 2349, 2349, 2349, 2349,
 2349,    0, 2349,    0,    0, 1424, 1423, 1425,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 2349,
 2349,    0, 2349, 2349, 2349, 2349, 2349, 2349, 2349, 2349,
 2349,    0, 2349, 2349, 2349, 2349, 2349, 2349, 2349, 2349,
 2349, 2349, 2349, 2349,    0, 2349, 2349, 2349,    0, 2349,
    0,    0,    0, 2349, 2349, 2349, 2349, 2349, 2349, 2349,
 2349, 2349, 2349, 2349, 2349, 2349, 2350, 2349, 2349,    0,
    0,    0,    0,    0, 2349, 2349, 2349, 2349, 2349, 2349,
 2349, 2349, 2349, 2349, 2349, 2349, 2349, 2349, 2349, 2349,
 2349, 2349,    0,    0, 2349,    0, 2349, 2349, 2349, 2349,
 2349, 2349,    0,    0,    0,    0,    0, 2350, 2350, 2350,
 2350, 2350, 2350, 2350,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 2350, 2350, 2350, 2350, 2350,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 2350,    0, 2350,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1408,    0,    0,
    0,    0,    0,    0,    0,    0, 1409,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1410,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1411,    0,    0,    0,    0,    0,    0,
 1412,    0,    0,    0,    0,    0,    0,    0, 1413,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1414,    0, 1415,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1416,    0, 1417,
 1418, 1419,    0,    0,    0,    0, 1420,    0,    0,    0,
    0,    0,    0, 2350,    0, 2350,    0,    0, 2350,    0,
 2350, 2350, 2350,    0, 2350, 2350,    0, 2350,    0,    0,
    0,    0,    0, 2350,    0, 2350, 2350, 2350, 2350,    0,
 2350,    0,    0, 2350,    0,    0, 2350, 1421,    0,    0,
    0,    0,    0, 2350,    0, 2350, 2350,    0, 2350, 2350,
 2350, 2350, 2350,    0,    0, 2350,    0,    0, 2350,    0,
 2350, 2350,    0,    0, 2350, 2350, 2350,    0, 2350, 2350,
 2350, 2350, 2350, 2350, 2350,    0,    0,    0, 2350, 2350,
 2350, 2350, 2350,    0,    0,    0, 2350,    0,    0,    0,
    0,    0,    0,    0, 2350, 2350, 2350, 2350, 2350, 2350,
    0,    0,    0, 2350, 2350,    0, 2350,    0, 2350,    0,
 2350, 2350, 2350,    0, 2350,    0, 2350, 2350, 2350,    0,
 2350, 2350, 2350, 2350, 2350,    0, 2350, 2350,    0,    0,
    0,    0, 2350,    0, 2350, 2350,    0,    0, 2350, 2350,
 2350,    0,    0, 2350,    0, 2350,    0,    0, 2350, 2350,
 2350, 2350,    0, 2350,    0, 2350,    0, 2350, 2350, 2350,
 2350, 2350,    0,    0, 2350, 2350, 2350,    0, 2350, 2350,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 2350,
 2350, 2350,    0, 2350, 2350, 2350,    0,    0,    0,    0,
    0, 2350,    0,    0,    0, 2350,    0, 2350, 2350, 2350,
 2350,    0, 2350,    0, 2350, 2350, 2350, 2350, 2350, 2350,
 2350, 2350,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 2350, 2350,    0, 2350, 2350, 2350, 1422,    0,    0,
    0, 2350, 2350, 2350, 2350, 2350,    0, 2350,    0,    0,
 2350, 2350,    0,    0,    0, 2350, 2350,    0,    0, 2350,
 2350, 2350, 2350, 2350, 2350, 2350, 2350, 2350,    0, 2350,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 2350,
 2350, 2350, 2350, 2350, 2350, 2350, 2350, 2350, 2350, 2350,
 2350, 2350, 2350, 2350, 2350, 2350, 2350, 2350, 2350, 2350,
    0, 2350, 2350, 2350, 2350, 2350, 2350,    0, 2350, 2350,
 2350, 2350,    0, 2350,    0, 2350, 2350, 2350, 2350, 2350,
 2350, 2350, 2350, 2350, 2350, 2350, 2350, 2350, 2350, 2350,
 2350, 2350, 2350,    0, 2350, 2350, 2350, 2350,    0, 2350,
 2350, 2350, 2350, 2350,    0, 2350, 2350, 2350, 2350, 2350,
 2350, 2350, 2350, 2350, 2350, 2350, 2350, 2350, 2350, 2350,
 2350, 2350, 2350, 2350, 2350, 2350, 2350, 2350, 2350, 2350,
 2350, 2350, 2350, 2350,    0, 2350, 2350,    0,    0,    0,
 2350, 2350, 2350, 2350, 2350, 2350,    0,    0, 2350, 2350,
 2350, 2350, 2350, 2350, 2350, 2350, 2350,    0, 2174,    0,
    0,    0, 2350,    0, 2350, 2350, 2350, 2350, 2350, 2350,
    0, 2350,    0,    0, 1424, 1423, 1425,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2350, 2350,
    0, 2350, 2350, 2350, 2350, 2350, 2350, 2350, 2350, 2350,
    0, 2350, 2350, 2350, 2350, 2350, 2350, 2350, 2350, 2350,
 2350, 2350, 2350,    0, 2350, 2350, 2350,    0, 2350,    0,
    0,    0, 2350, 2350, 2350, 2350, 2350, 2350, 2350, 2350,
 2350, 2350, 2350, 2350, 2350, 2225, 2350, 2350,    0,    0,
    0,    0,    0, 2350, 2350, 2350, 2350, 2350, 2350, 2350,
 2350, 2350, 2350, 2350, 2350, 2350, 2350, 2350, 2350, 2350,
 2350,    0,    0, 2350,    0, 2350, 2350, 2350, 2350, 2350,
 2350,    0,    0,    0,    0,    0, 2225, 2225, 2225, 2225,
 2225, 2225, 2225,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 2225, 2225, 2225, 2225,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2225,    0, 2225,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1408,    0,    0,    0,
    0,    0,    0,    0,    0, 1409,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1410,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1411,    0,    0,    0,    0,    0,    0, 1412,
    0,    0,    0,    0,    0,    0,    0, 1413,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1414,    0, 1415,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1416,    0, 1417, 1418,
 1419,    0,    0,    0,    0, 1420,    0,    0,    0,    0,
    0,    0, 2225,    0, 2225,    0,    0, 2225,    0, 2225,
 2225, 2225,    0, 2225, 2225,    0, 2225,    0,    0,    0,
    0,    0, 2225,    0, 2225, 2225, 2225, 2225,    0, 2225,
    0,    0, 2225,    0,    0, 2225, 1421,    0,    0,    0,
    0,    0, 2225,    0, 2225, 2225,    0, 2225, 2225, 2225,
 2225, 2225,    0,    0, 2225,    0,    0, 2225,    0, 2225,
 2225,    0,    0, 2225, 2225, 2225,    0, 2225, 2225, 2225,
 2225, 2225, 2225, 2225,    0,    0,    0, 2225, 2225, 2225,
 2225, 2225,    0,    0,    0, 2225,    0,    0,    0,    0,
    0,    0,    0, 2225, 2225, 2225, 2225, 2225, 2225,    0,
    0,    0, 2225, 2225,    0, 2225,    0, 2225,    0, 2225,
 2225, 2225,    0, 2225,    0, 2225, 2225, 2225,    0, 2225,
 2225, 2225, 2225, 2225,    0, 2225, 2225,    0,    0,    0,
    0, 2225,    0, 2225, 2225,    0,    0, 2225, 2225, 2225,
    0,    0, 2225,    0, 2225,    0,    0, 2225, 2225, 2225,
 2225,    0, 2225,    0, 2225,    0, 2225, 2225, 2225, 2225,
 2225,    0,    0, 2225, 2225, 2225,    0, 2225, 2225,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2225, 2225,
 2225,    0, 2225, 2225, 2225,    0,    0,    0,    0,    0,
 2225,    0,    0,    0, 2225,    0, 2225, 2225, 2225, 2225,
    0, 2225,    0, 2225, 2225, 2225, 2225, 2225, 2225, 2225,
 2225,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2225, 2225,    0, 2225, 2225, 2225, 1422,    0,    0,    0,
 2225, 2225, 2225, 2225, 2225,    0, 2225,    0,    0, 2225,
 2225,    0,    0,    0, 2225, 2225,    0,    0, 2225, 2225,
 2225, 2225, 2225, 2225, 2225, 2225, 2225,    0, 2225,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2225, 2225,
 2225, 2225, 2225, 2225, 2225, 2225, 2225, 2225, 2225, 2225,
 2225, 2225, 2225, 2225, 2225, 2225, 2225, 2225, 2225,    0,
 2225, 2225, 2225, 2225, 2225, 2225,    0, 2225, 2225, 2225,
 2225,    0, 2225,    0, 2225, 2225, 2225, 2225, 2225, 2225,
 2225, 2225, 2225, 2225, 2225, 2225, 2225, 2225, 2225, 2225,
 2225, 2225,    0, 2225, 2225, 2225, 2225,    0, 2225, 2225,
 2225, 2225, 2225,    0, 2225, 2225, 2225, 2225, 2225, 2225,
 2225, 2225, 2225, 2225, 2225, 2225, 2225, 2225, 2225, 2225,
 2225, 2225, 2225, 2225, 2225, 2225, 2225, 2225, 2225, 2225,
 2225, 2225, 2225,    0, 2225, 2225,    0,    0,    0, 2225,
 2225, 2225, 2225, 2225, 2225,    0,    0, 2225, 2225, 2225,
 2225, 2225, 2225, 2225, 2225, 2225,    0, 2175,    0,    0,
    0, 2225,    0, 2225, 2225, 2225, 2225, 2225, 2225,    0,
 2225,    0,    0, 1424, 1423, 1425,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 2225, 2225,    0,
 2225, 2225, 2225, 2225, 2225, 2225, 2225, 2225, 2225,    0,
 2225, 2225, 2225, 2225, 2225, 2225, 2225, 2225, 2225, 2225,
 2225, 2225,    0, 2225, 2225, 2225,    0, 2225,    0,    0,
    0, 2225, 2225, 2225, 2225, 2225, 2225, 2225, 2225, 2225,
 2225, 2225, 2225, 2225, 2251, 2225, 2225,    0,    0,    0,
    0,    0, 2225, 2225, 2225, 2225, 2225, 2225, 2225, 2225,
 2225, 2225, 2225, 2225, 2225, 2225, 2225, 2225, 2225, 2225,
    0,    0, 2225,    0, 2225, 2225, 2225, 2225, 2225, 2225,
    0,    0,    0,    0,    0, 2251, 2251, 2251, 2251, 2251,
 2251, 2251,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 2251, 2251, 2251, 2251,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 2251,    0, 2251,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1408,    0,    0,    0,    0,
    0,    0,    0,    0, 1409,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1410,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1411,    0,    0,    0,    0,    0,    0, 1412,    0,
    0,    0,    0,    0,    0,    0, 1413,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1414,    0, 1415,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1416,    0, 1417, 1418, 1419,
    0,    0,    0,    0, 1420,    0,    0,    0,    0,    0,
    0, 2251,    0, 2251,    0,    0, 2251,    0, 2251, 2251,
 2251,    0, 2251, 2251,    0, 2251,    0,    0,    0,    0,
    0, 2251,    0, 2251, 2251, 2251, 2251,    0, 2251,    0,
    0, 2251,    0,    0, 2251, 1421,    0,    0,    0,    0,
    0, 2251,    0, 2251, 2251,    0, 2251, 2251, 2251, 2251,
 2251,    0,    0, 2251,    0,    0, 2251,    0, 2251, 2251,
    0,    0, 2251, 2251, 2251,    0, 2251, 2251, 2251, 2251,
 2251, 2251, 2251,    0,    0,    0, 2251, 2251, 2251, 2251,
 2251,    0,    0,    0, 2251,    0,    0,    0,    0,    0,
    0,    0, 2251, 2251, 2251, 2251, 2251, 2251,    0,    0,
    0, 2251, 2251,    0, 2251,    0, 2251,    0, 2251, 2251,
 2251,    0, 2251,    0, 2251, 2251, 2251,    0, 2251, 2251,
 2251, 2251, 2251,    0, 2251, 2251,    0,    0,    0,    0,
 2251,    0, 2251, 2251,    0,    0, 2251, 2251, 2251,    0,
    0, 2251,    0, 2251,    0,    0, 2251, 2251, 2251, 2251,
    0, 2251,    0, 2251,    0, 2251, 2251, 2251, 2251, 2251,
    0,    0, 2251, 2251, 2251,    0, 2251, 2251,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 2251, 2251, 2251,
    0, 2251, 2251, 2251,    0,    0,    0,    0,    0, 2251,
    0,    0,    0, 2251,    0, 2251, 2251, 2251, 2251,    0,
 2251,    0, 2251, 2251, 2251, 2251, 2251, 2251, 2251, 2251,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 2251,
 2251,    0, 2251, 2251, 2251, 1422,    0,    0,    0, 2251,
 2251, 2251, 2251, 2251,    0, 2251,    0,    0, 2251, 2251,
    0,    0,    0, 2251, 2251,    0,    0, 2251, 2251, 2251,
 2251, 2251, 2251, 2251, 2251, 2251,    0, 2251,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 2251, 2251, 2251,
 2251, 2251, 2251, 2251, 2251, 2251, 2251, 2251, 2251, 2251,
 2251, 2251, 2251, 2251, 2251, 2251, 2251, 2251,    0, 2251,
 2251, 2251, 2251, 2251, 2251,    0, 2251, 2251, 2251, 2251,
    0, 2251,    0, 2251, 2251, 2251, 2251, 2251, 2251, 2251,
 2251, 2251, 2251, 2251, 2251, 2251, 2251, 2251, 2251, 2251,
 2251,    0, 2251, 2251, 2251, 2251,    0, 2251, 2251, 2251,
 2251, 2251,    0, 2251, 2251, 2251, 2251, 2251, 2251, 2251,
 2251, 2251, 2251, 2251, 2251, 2251, 2251, 2251, 2251, 2251,
 2251, 2251, 2251, 2251, 2251, 2251, 2251, 2251, 2251, 2251,
 2251, 2251,    0, 2251, 2251,    0,    0,    0, 2251, 2251,
 2251, 2251, 2251, 2251,    0,    0, 2251, 2251, 2251, 2251,
 2251, 2251, 2251, 2251, 2251,    0, 2176,    0,    0,    0,
 2251,    0, 2251, 2251, 2251, 2251, 2251, 2251,    0, 2251,
    0,    0, 1424, 1423, 1425,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2251, 2251,    0, 2251,
 2251, 2251, 2251, 2251, 2251, 2251, 2251, 2251,    0, 2251,
 2251, 2251, 2251, 2251, 2251, 2251, 2251, 2251, 2251, 2251,
 2251,    0, 2251, 2251, 2251,    0, 2251,    0,    0,    0,
 2251, 2251, 2251, 2251, 2251, 2251, 2251, 2251, 2251, 2251,
 2251, 2251, 2251, 2112, 2251, 2251,    0,    0,    0,    0,
    0, 2251, 2251, 2251, 2251, 2251, 2251, 2251, 2251, 2251,
 2251, 2251, 2251, 2251, 2251, 2251, 2251, 2251, 2251,    0,
    0, 2251,    0, 2251, 2251, 2251, 2251, 2251, 2251,    0,
    0,    0,    0,    0, 2112, 2112, 2112, 2112, 2112, 2112,
 2112,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 2112, 2112, 2112, 2112,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 2112,    0, 2112,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1408,    0,    0,    0,    0,    0,
    0,    0,    0, 1409,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1410,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1411,    0,    0,    0,    0,    0,    0, 1412,    0,    0,
    0,    0,    0,    0,    0, 1413,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1414,    0, 1415,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1416,    0, 1417, 1418, 1419,    0,
    0,    0,    0, 1420,    0,    0,    0,    0,    0,    0,
 2112,    0, 2112,    0,    0, 2112,    0, 2112, 2112, 2112,
    0, 2112, 2112,    0, 2112,    0,    0,    0,    0,    0,
 2112,    0, 2112, 2112, 2112, 2112,    0, 2112,    0,    0,
 2112,    0,    0, 2112, 1421,    0,    0,    0,    0,    0,
 2112,    0, 2112, 2112,    0, 2112, 2112, 2112, 2112, 2112,
    0,    0, 2112,    0,    0, 2112,    0, 2112, 2112,    0,
    0, 2112, 2112, 2112,    0, 2112, 2112, 2112, 2112, 2112,
 2112, 2112,    0,    0,    0, 2112, 2112, 2112, 2112, 2112,
    0,    0,    0, 2112,    0,    0,    0,    0,    0,    0,
    0, 2112, 2112, 2112, 2112, 2112, 2112,    0,    0,    0,
 2112, 2112,    0, 2112,    0, 2112,    0, 2112, 2112, 2112,
    0, 2112,    0, 2112, 2112, 2112,    0, 2112, 2112, 2112,
 2112, 2112,    0, 2112, 2112,    0,    0,    0,    0, 2112,
    0, 2112, 2112,    0,    0, 2112, 2112, 2112,    0,    0,
 2112,    0, 2112,    0,    0, 2112, 2112, 2112, 2112,    0,
 2112,    0, 2112,    0, 2112, 2112, 2112, 2112, 2112,    0,
    0, 2112, 2112, 2112,    0, 2112, 2112,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2112, 2112, 2112,    0,
 2112, 2112, 2112,    0,    0,    0,    0,    0, 2112,    0,
    0,    0, 2112,    0, 2112, 2112, 2112, 2112,    0, 2112,
    0, 2112, 2112, 2112, 2112, 2112, 2112, 2112, 2112,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2112, 2112,
    0, 2112, 2112, 2112, 1422,    0,    0,    0, 2112, 2112,
 2112, 2112, 2112,    0, 2112,    0,    0, 2112, 2112,    0,
    0,    0, 2112, 2112,    0,    0, 2112, 2112, 2112, 2112,
 2112, 2112, 2112, 2112, 2112,    0, 2112,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2112, 2112, 2112, 2112,
 2112, 2112, 2112, 2112, 2112, 2112, 2112, 2112, 2112, 2112,
 2112, 2112, 2112, 2112, 2112, 2112, 2112,    0, 2112, 2112,
 2112, 2112, 2112, 2112,    0, 2112, 2112, 2112, 2112,    0,
 2112,    0, 2112, 2112, 2112, 2112, 2112, 2112, 2112, 2112,
 2112, 2112, 2112, 2112, 2112, 2112, 2112, 2112, 2112, 2112,
    0, 2112, 2112, 2112, 2112,    0, 2112, 2112, 2112, 2112,
 2112,    0, 2112, 2112, 2112, 2112, 2112, 2112, 2112, 2112,
 2112, 2112, 2112, 2112, 2112, 2112, 2112, 2112, 2112, 2112,
 2112, 2112, 2112, 2112, 2112, 2112, 2112, 2112, 2112, 2112,
 2112,    0, 2112, 2112,    0,    0,    0, 2112, 2112, 2112,
 2112, 2112, 2112,    0,    0, 2112, 2112, 2112, 2112, 2112,
 2112, 2112, 2112, 2112,    0, 2177,    0,    0,    0, 2112,
    0, 2112, 2112, 2112, 2112, 2112, 2112,    0, 2112,    0,
    0, 1424, 1423, 1425,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 2112, 2112,    0, 2112, 2112,
 2112, 2112, 2112, 2112, 2112, 2112, 2112,    0, 2112, 2112,
 2112, 2112, 2112, 2112, 2112, 2112, 2112, 2112, 2112, 2112,
    0, 2112, 2112, 2112,    0, 2112,    0,    0,    0, 2112,
 2112, 2112, 2112, 2112, 2112, 2112, 2112, 2112, 2112, 2112,
 2112, 2112, 2108, 2112, 2112,    0,    0,    0,    0,    0,
 2112, 2112, 2112, 2112, 2112, 2112, 2112, 2112, 2112, 2112,
 2112, 2112, 2112, 2112, 2112, 2112, 2112, 2112,    0,    0,
 2112,    0, 2112, 2112, 2112, 2112, 2112, 2112,    0,    0,
    0,    0,    0, 2108, 2108, 2108, 2108, 2108, 2108, 2108,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 2108, 2108, 2108, 2108,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 2108,    0, 2108,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1408,    0,    0,    0,    0,    0,    0,
    0,    0, 1409,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1410,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1411,
    0,    0,    0,    0,    0,    0, 1412,    0,    0,    0,
    0,    0,    0,    0, 1413,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1414,    0, 1415,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1416,    0, 1417, 1418, 1419,    0,    0,
    0,    0, 1420,    0,    0,    0,    0,    0,    0, 2108,
    0, 2108,    0,    0, 2108,    0, 2108, 2108, 2108,    0,
 2108, 2108,    0, 2108,    0,    0,    0,    0,    0, 2108,
    0, 2108, 2108, 2108, 2108,    0, 2108,    0,    0, 2108,
    0,    0, 2108, 1421,    0,    0,    0,    0,    0, 2108,
    0, 2108, 2108,    0, 2108, 2108, 2108, 2108, 2108,    0,
    0, 2108,    0,    0, 2108,    0, 2108, 2108,    0,    0,
 2108, 2108, 2108,    0, 2108, 2108, 2108, 2108, 2108, 2108,
 2108,    0,    0,    0, 2108, 2108, 2108, 2108, 2108,    0,
    0,    0, 2108,    0,    0,    0,    0,    0,    0,    0,
 2108, 2108, 2108, 2108, 2108, 2108,    0,    0,    0, 2108,
 2108,    0, 2108,    0, 2108,    0, 2108, 2108, 2108,    0,
 2108,    0, 2108, 2108, 2108,    0, 2108, 2108, 2108, 2108,
 2108,    0, 2108, 2108,    0,    0,    0,    0, 2108,    0,
 2108, 2108,    0,    0, 2108, 2108, 2108,    0,    0, 2108,
    0, 2108,    0,    0, 2108, 2108, 2108, 2108,    0, 2108,
    0, 2108,    0, 2108, 2108, 2108, 2108, 2108,    0,    0,
 2108, 2108, 2108,    0, 2108, 2108,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 2108, 2108, 2108,    0, 2108,
 2108, 2108,    0,    0,    0,    0,    0, 2108,    0,    0,
    0, 2108,    0, 2108, 2108, 2108, 2108,    0, 2108,    0,
 2108, 2108, 2108, 2108, 2108, 2108, 2108, 2108,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 2108, 2108,    0,
 2108, 2108, 2108, 1422,    0,    0,    0, 2108, 2108, 2108,
 2108, 2108,    0, 2108,    0,    0, 2108, 2108,    0,    0,
    0, 2108, 2108,    0,    0, 2108, 2108, 2108, 2108, 2108,
 2108, 2108, 2108, 2108,    0, 2108,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 2108, 2108, 2108, 2108, 2108,
 2108, 2108, 2108, 2108, 2108, 2108, 2108, 2108, 2108, 2108,
 2108, 2108, 2108, 2108, 2108, 2108,    0, 2108, 2108, 2108,
 2108, 2108, 2108,    0, 2108, 2108, 2108, 2108,    0, 2108,
    0, 2108, 2108, 2108, 2108, 2108, 2108, 2108, 2108, 2108,
 2108, 2108, 2108, 2108, 2108, 2108, 2108, 2108, 2108,    0,
 2108, 2108, 2108, 2108,    0, 2108, 2108, 2108, 2108, 2108,
    0, 2108, 2108, 2108, 2108, 2108, 2108, 2108, 2108, 2108,
 2108, 2108, 2108, 2108, 2108, 2108, 2108, 2108, 2108, 2108,
 2108, 2108, 2108, 2108, 2108, 2108, 2108, 2108, 2108, 2108,
    0, 2108, 2108,    0,    0,    0, 2108, 2108, 2108, 2108,
 2108, 2108,    0,    0, 2108, 2108, 2108, 2108, 2108, 2108,
 2108, 2108, 2108,    0, 2178,    0,    0,    0, 2108,    0,
 2108, 2108, 2108, 2108, 2108, 2108,    0, 2108,    0,    0,
 1424, 1423, 1425,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 2108, 2108,    0, 2108, 2108, 2108,
 2108, 2108, 2108, 2108, 2108, 2108,    0, 2108, 2108, 2108,
 2108, 2108, 2108, 2108, 2108, 2108, 2108, 2108, 2108,    0,
 2108, 2108, 2108,    0, 2108,    0,    0,    0, 2108, 2108,
 2108, 2108, 2108, 2108, 2108, 2108, 2108, 2108, 2108, 2108,
 2108, 2120, 2108, 2108,    0,    0,    0,    0,    0, 2108,
 2108, 2108, 2108, 2108, 2108, 2108, 2108, 2108, 2108, 2108,
 2108, 2108, 2108, 2108, 2108, 2108, 2108,    0,    0, 2108,
    0, 2108, 2108, 2108, 2108, 2108, 2108,    0,    0,    0,
    0,    0, 2120, 2120, 2120, 2120, 2120, 2120, 2120,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2120, 2120, 2120, 2120,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 2120,    0, 2120,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1408,    0,    0,    0,    0,    0,    0,    0,
    0, 1409,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1410,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1411,    0,
    0,    0,    0,    0,    0, 1412,    0,    0,    0,    0,
    0,    0,    0, 1413,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1414,    0, 1415,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1416,    0, 1417, 1418, 1419,    0,    0,    0,
    0, 1420,    0,    0,    0,    0,    0,    0, 2120,    0,
 2120,    0,    0, 2120,    0, 2120, 2120, 2120,    0, 2120,
 2120,    0, 2120,    0,    0,    0,    0,    0, 2120,    0,
 2120, 2120, 2120, 2120,    0, 2120,    0,    0, 2120,    0,
    0, 2120, 1421,    0,    0,    0,    0,    0, 2120,    0,
 2120, 2120,    0, 2120, 2120, 2120, 2120, 2120,    0,    0,
 2120,    0,    0, 2120,    0, 2120, 2120,    0,    0, 2120,
 2120, 2120,    0, 2120, 2120, 2120, 2120, 2120, 2120, 2120,
    0,    0,    0, 2120, 2120, 2120, 2120, 2120,    0,    0,
    0, 2120,    0,    0,    0,    0,    0,    0,    0, 2120,
 2120, 2120, 2120, 2120, 2120,    0,    0,    0, 2120, 2120,
    0, 2120,    0, 2120,    0, 2120, 2120, 2120,    0, 2120,
    0, 2120, 2120, 2120,    0, 2120, 2120, 2120, 2120, 2120,
    0, 2120, 2120,    0,    0,    0,    0, 2120,    0, 2120,
 2120,    0,    0, 2120, 2120, 2120,    0,    0, 2120,    0,
 2120,    0,    0, 2120, 2120, 2120, 2120,    0, 2120,    0,
 2120,    0, 2120, 2120, 2120, 2120, 2120,    0,    0, 2120,
 2120, 2120,    0, 2120, 2120,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 2120, 2120, 2120,    0, 2120, 2120,
 2120,    0,    0,    0,    0,    0, 2120,    0,    0,    0,
 2120,    0, 2120, 2120, 2120, 2120,    0, 2120,    0, 2120,
 2120, 2120, 2120, 2120, 2120, 2120, 2120,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2120, 2120,    0, 2120,
 2120, 2120, 1422,    0,    0,    0, 2120, 2120, 2120, 2120,
 2120,    0, 2120,    0,    0, 2120, 2120,    0,    0,    0,
 2120, 2120,    0,    0, 2120, 2120, 2120, 2120, 2120, 2120,
 2120, 2120, 2120,    0, 2120,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 2120, 2120, 2120, 2120, 2120, 2120,
 2120, 2120, 2120, 2120, 2120, 2120, 2120, 2120, 2120, 2120,
 2120, 2120, 2120, 2120, 2120,    0, 2120, 2120, 2120, 2120,
 2120, 2120,    0, 2120, 2120, 2120, 2120,    0, 2120,    0,
 2120, 2120, 2120, 2120, 2120, 2120, 2120, 2120, 2120, 2120,
 2120, 2120, 2120, 2120, 2120, 2120, 2120, 2120,    0, 2120,
 2120, 2120, 2120,    0, 2120, 2120, 2120, 2120, 2120,    0,
 2120, 2120, 2120, 2120, 2120, 2120, 2120, 2120, 2120, 2120,
 2120, 2120, 2120, 2120, 2120, 2120, 2120, 2120, 2120, 2120,
 2120, 2120, 2120, 2120, 2120, 2120, 2120, 2120, 2120,    0,
 2120, 2120,    0,    0,    0, 2120, 2120, 2120, 2120, 2120,
 2120,    0,    0, 2120, 2120, 2120, 2120, 2120, 2120, 2120,
 2120, 2120,    0, 2179,    0,    0,    0, 2120,    0, 2120,
 2120, 2120, 2120, 2120, 2120,    0, 2120,    0,    0, 1424,
 1423, 1425,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 2120, 2120,    0, 2120, 2120, 2120, 2120,
 2120, 2120, 2120, 2120, 2120,    0, 2120, 2120, 2120, 2120,
 2120, 2120, 2120, 2120, 2120, 2120, 2120, 2120,    0, 2120,
 2120, 2120,    0, 2120,    0,    0,    0, 2120, 2120, 2120,
 2120, 2120, 2120, 2120, 2120, 2120, 2120, 2120, 2120, 2120,
 2125, 2120, 2120,    0,    0,    0,    0,    0, 2120, 2120,
 2120, 2120, 2120, 2120, 2120, 2120, 2120, 2120, 2120, 2120,
 2120, 2120, 2120, 2120, 2120, 2120,    0,    0, 2120,    0,
 2120, 2120, 2120, 2120, 2120, 2120,    0,    0,    0,    0,
    0, 2125, 2125, 2125, 2125, 2125, 2125, 2125,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 2125,
 2125, 2125, 2125,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 2125,    0, 2125,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1408,    0,    0,    0,    0,    0,    0,    0,    0,
 1409,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1410,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1411,    0,    0,
    0,    0,    0,    0, 1412,    0,    0,    0,    0,    0,
    0,    0, 1413,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1414,
    0, 1415,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1416,    0, 1417, 1418, 1419,    0,    0,    0,    0,
 1420,    0,    0,    0,    0,    0,    0, 2125,    0, 2125,
    0,    0, 2125,    0, 2125, 2125, 2125,    0, 2125, 2125,
    0, 2125,    0,    0,    0,    0,    0, 2125,    0, 2125,
 2125, 2125, 2125,    0, 2125,    0,    0, 2125,    0,    0,
 2125, 1421,    0,    0,    0,    0,    0, 2125,    0, 2125,
 2125,    0, 2125, 2125, 2125, 2125, 2125,    0,    0, 2125,
    0,    0, 2125,    0, 2125, 2125,    0,    0, 2125, 2125,
 2125,    0, 2125, 2125, 2125, 2125, 2125, 2125, 2125,    0,
    0,    0, 2125, 2125, 2125, 2125, 2125,    0,    0,    0,
 2125,    0,    0,    0,    0,    0,    0,    0, 2125, 2125,
 2125, 2125, 2125, 2125,    0,    0,    0, 2125, 2125,    0,
 2125,    0, 2125,    0, 2125, 2125, 2125,    0, 2125,    0,
 2125, 2125, 2125,    0, 2125, 2125, 2125, 2125, 2125,    0,
 2125, 2125,    0,    0,    0,    0, 2125,    0, 2125, 2125,
    0,    0, 2125, 2125, 2125,    0,    0, 2125,    0, 2125,
    0,    0, 2125, 2125, 2125, 2125,    0, 2125,    0, 2125,
    0, 2125, 2125, 2125, 2125, 2125,    0,    0, 2125, 2125,
 2125,    0, 2125, 2125,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 2125, 2125, 2125,    0, 2125, 2125, 2125,
    0,    0,    0,    0,    0, 2125,    0,    0,    0, 2125,
    0, 2125, 2125, 2125, 2125,    0, 2125,    0, 2125, 2125,
 2125, 2125, 2125, 2125, 2125, 2125,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 2125, 2125,    0, 2125, 2125,
 2125, 1422,    0,    0,    0, 2125, 2125, 2125, 2125, 2125,
    0, 2125,    0,    0, 2125, 2125,    0,    0,    0, 2125,
 2125,    0,    0, 2125, 2125, 2125, 2125, 2125, 2125, 2125,
 2125, 2125,    0, 2125,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 2125, 2125, 2125, 2125, 2125, 2125, 2125,
 2125, 2125, 2125, 2125, 2125, 2125, 2125, 2125, 2125, 2125,
 2125, 2125, 2125, 2125,    0, 2125, 2125, 2125, 2125, 2125,
 2125,    0, 2125, 2125, 2125, 2125,    0, 2125,    0, 2125,
 2125, 2125, 2125, 2125, 2125, 2125, 2125, 2125, 2125, 2125,
 2125, 2125, 2125, 2125, 2125, 2125, 2125,    0, 2125, 2125,
 2125, 2125,    0, 2125, 2125, 2125, 2125, 2125,    0, 2125,
 2125, 2125, 2125, 2125, 2125, 2125, 2125, 2125, 2125, 2125,
 2125, 2125, 2125, 2125, 2125, 2125, 2125, 2125, 2125, 2125,
 2125, 2125, 2125, 2125, 2125, 2125, 2125, 2125,    0, 2125,
 2125,    0,    0,    0, 2125, 2125, 2125, 2125, 2125, 2125,
    0,    0, 2125, 2125, 2125, 2125, 2125, 2125, 2125, 2125,
 2125,    0, 2584,    0,    0,    0, 2125,    0, 2125, 2125,
 2125, 2125, 2125, 2125,    0, 2125,    0,    0, 1424, 1423,
 1425,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 2125, 2125,    0, 2125, 2125, 2125, 2125, 2125,
 2125, 2125, 2125, 2125,    0, 2125, 2125, 2125, 2125, 2125,
 2125, 2125, 2125, 2125, 2125, 2125, 2125,    0, 2125, 2125,
 2125,    0, 2125,    0,    0,    0, 2125, 2125, 2125, 2125,
 2125, 2125, 2125, 2125, 2125, 2125, 2125, 2125, 2125, 2154,
 2125, 2125,    0,    0,    0,    0,    0, 2125, 2125, 2125,
 2125, 2125, 2125, 2125, 2125, 2125, 2125, 2125, 2125, 2125,
 2125, 2125, 2125, 2125, 2125,    0,    0, 2125,    0, 2125,
 2125, 2125, 2125, 2125, 2125,    0,    0,    0,    0,    0,
 2154, 2154, 2154, 2154, 2154, 2154, 2154,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2154, 2154,
 2154, 2154,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2154,    0, 2154,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1408,    0,    0,    0,    0,    0,    0,    0,    0, 1409,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1410,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1411,    0,    0,    0,
    0,    0,    0, 1412,    0,    0,    0,    0,    0,    0,
    0, 1413,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1414,    0,
 1415,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1416,    0, 1417, 1418, 1419,    0,    0,    0,    0, 1420,
    0,    0,    0,    0,    0,    0, 2154,    0, 2154,    0,
    0, 2154,    0, 2154, 2154, 2154,    0, 2154, 2154,    0,
 2154,    0,    0,    0,    0,    0, 2154,    0, 2154, 2154,
 2154, 2154,    0, 2154,    0,    0, 2154,    0,    0, 2154,
 1421,    0,    0,    0,    0,    0, 2154,    0, 2154, 2154,
    0, 2154, 2154, 2154, 2154, 2154,    0,    0, 2154,    0,
    0, 2154,    0, 2154, 2154,    0,    0, 2154, 2154, 2154,
    0, 2154, 2154, 2154, 2154, 2154, 2154, 2154,    0,    0,
    0, 2154, 2154, 2154, 2154, 2154,    0,    0,    0, 2154,
    0,    0,    0,    0,    0,    0,    0, 2154, 2154, 2154,
 2154, 2154, 2154,    0,    0,    0, 2154, 2154,    0, 2154,
    0, 2154,    0, 2154, 2154, 2154,    0, 2154,    0, 2154,
 2154, 2154,    0, 2154, 2154, 2154, 2154, 2154,    0, 2154,
 2154,    0,    0,    0,    0, 2154,    0, 2154, 2154,    0,
    0, 2154, 2154, 2154,    0,    0, 2154,    0, 2154,    0,
    0, 2154, 2154, 2154, 2154,    0, 2154,    0, 2154,    0,
 2154, 2154, 2154, 2154, 2154,    0,    0, 2154, 2154, 2154,
    0, 2154, 2154,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 2154, 2154, 2154,    0, 2154, 2154, 2154,    0,
    0,    0,    0,    0, 2154,    0,    0,    0, 2154,    0,
 2154, 2154, 2154, 2154,    0, 2154,    0, 2154, 2154, 2154,
 2154, 2154, 2154, 2154, 2154,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 2154, 2154,    0, 2154, 2154, 2154,
 1422,    0,    0,    0, 2154, 2154, 2154, 2154, 2154,    0,
 2154,    0,    0, 2154, 2154,    0,    0,    0, 2154, 2154,
    0,    0, 2154, 2154, 2154, 2154, 2154, 2154, 2154, 2154,
 2154,    0, 2154,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 2154, 2154, 2154, 2154, 2154, 2154, 2154, 2154,
 2154, 2154, 2154, 2154, 2154, 2154, 2154, 2154, 2154, 2154,
 2154, 2154, 2154,    0, 2154, 2154, 2154, 2154, 2154, 2154,
    0, 2154, 2154, 2154, 2154,    0, 2154,    0, 2154, 2154,
 2154, 2154, 2154, 2154, 2154, 2154, 2154, 2154, 2154, 2154,
 2154, 2154, 2154, 2154, 2154, 2154,    0, 2154, 2154, 2154,
 2154,    0, 2154, 2154, 2154, 2154, 2154,    0, 2154, 2154,
 2154, 2154, 2154, 2154, 2154, 2154, 2154, 2154, 2154, 2154,
 2154, 2154, 2154, 2154, 2154, 2154, 2154, 2154, 2154, 2154,
 2154, 2154, 2154, 2154, 2154, 2154, 2154,    0, 2154, 2154,
    0,    0,    0, 2154, 2154, 2154, 2154, 2154, 2154,    0,
    0, 2154, 2154, 2154, 2154, 2154, 2154, 2154, 2154, 2154,
    0, 2918,    0,    0,    0, 2154,    0, 2154, 2154, 2154,
 2154, 2154, 2154,    0, 2154,    0,    0, 1424, 1423, 1425,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 2154, 2154,    0, 2154, 2154, 2154, 2154, 2154, 2154,
 2154, 2154, 2154,    0, 2154, 2154, 2154, 2154, 2154, 2154,
 2154, 2154, 2154, 2154, 2154, 2154,    0, 2154, 2154, 2154,
    0, 2154,    0,    0,    0, 2154, 2154, 2154, 2154, 2154,
 2154, 2154, 2154, 2154, 2154, 2154, 2154, 2154, 2160, 2154,
 2154,    0,    0,    0,    0,    0, 2154, 2154, 2154, 2154,
 2154, 2154, 2154, 2154, 2154, 2154, 2154, 2154, 2154, 2154,
 2154, 2154, 2154, 2154,    0,    0, 2154,    0, 2154, 2154,
 2154, 2154, 2154, 2154,    0,    0,    0,    0,    0, 2160,
 2160, 2160, 2160, 2160, 2160, 2160,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 2160, 2160, 2160,
 2160,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 2160,
    0, 2160,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1408,
    0,    0,    0,    0,    0,    0,    0,    0, 1409,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1410,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1411,    0,    0,    0,    0,
    0,    0, 1412,    0,    0,    0,    0,    0,    0,    0,
 1413,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1414,    0, 1415,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1416,
    0, 1417, 1418, 1419,    0,    0,    0,    0, 1420,    0,
    0,    0,    0,    0,    0, 2160,    0, 2160,    0,    0,
 2160,    0, 2160, 2160, 2160,    0, 2160, 2160,    0, 2160,
    0,    0,    0,    0,    0, 2160,    0, 2160, 2160, 2160,
 2160,    0, 2160,    0,    0, 2160,    0,    0, 2160, 1421,
    0,    0,    0,    0,    0, 2160,    0, 2160, 2160,    0,
 2160, 2160, 2160, 2160, 2160,    0,    0, 2160,    0,    0,
 2160,    0, 2160, 2160,    0,    0, 2160, 2160, 2160,    0,
 2160, 2160, 2160, 2160, 2160, 2160, 2160,    0,    0,    0,
 2160, 2160, 2160, 2160, 2160,    0,    0,    0, 2160,    0,
    0,    0,    0,    0,    0,    0, 2160, 2160, 2160, 2160,
 2160, 2160,    0,    0,    0, 2160, 2160,    0, 2160,    0,
 2160,    0, 2160, 2160, 2160,    0, 2160,    0, 2160, 2160,
 2160,    0, 2160, 2160, 2160, 2160, 2160,    0, 2160, 2160,
    0,    0,    0,    0, 2160,    0, 2160, 2160,    0,    0,
 2160, 2160, 2160,    0,    0, 2160,    0, 2160,    0,    0,
 2160, 2160, 2160, 2160,    0, 2160,    0, 2160,    0, 2160,
 2160, 2160, 2160, 2160,    0,    0, 2160, 2160, 2160,    0,
 2160, 2160,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 2160, 2160, 2160,    0, 2160, 2160, 2160,    0,    0,
    0,    0,    0, 2160,    0,    0,    0, 2160,    0, 2160,
 2160, 2160, 2160,    0, 2160,    0, 2160, 2160, 2160, 2160,
 2160, 2160, 2160, 2160,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 2160, 2160,    0, 2160, 2160, 2160, 1422,
    0,    0,    0, 2160, 2160, 2160, 2160, 2160,    0, 2160,
    0,    0, 2160, 2160,    0,    0,    0, 2160, 2160,    0,
    0, 2160, 2160, 2160, 2160, 2160, 2160, 2160, 2160, 2160,
    0, 2160,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 2160, 2160, 2160, 2160, 2160, 2160, 2160, 2160, 2160,
 2160, 2160, 2160, 2160, 2160, 2160, 2160, 2160, 2160, 2160,
 2160, 2160,    0, 2160, 2160, 2160, 2160, 2160, 2160,    0,
 2160, 2160, 2160, 2160,    0, 2160,    0, 2160, 2160, 2160,
 2160, 2160, 2160, 2160, 2160, 2160, 2160, 2160, 2160, 2160,
 2160, 2160, 2160, 2160, 2160,    0, 2160, 2160, 2160, 2160,
    0, 2160, 2160, 2160, 2160, 2160,    0, 2160, 2160, 2160,
 2160, 2160, 2160, 2160, 2160, 2160, 2160, 2160, 2160, 2160,
 2160, 2160, 2160, 2160, 2160, 2160, 2160, 2160, 2160, 2160,
 2160, 2160, 2160, 2160, 2160, 2160,    0, 2160, 2160,    0,
    0,    0, 2160, 2160, 2160, 2160, 2160, 2160,    0,    0,
 2160, 2160, 2160, 2160, 2160, 2160, 2160, 2160, 2160,    0,
 2923,    0,    0,    0, 2160,    0, 2160, 2160, 2160, 2160,
 2160, 2160,    0, 2160,    0,    0, 1424, 1423, 1425,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2160, 2160,    0, 2160, 2160, 2160, 2160, 2160, 2160, 2160,
 2160, 2160,    0, 2160, 2160, 2160, 2160, 2160, 2160, 2160,
 2160, 2160, 2160, 2160, 2160,    0, 2160, 2160, 2160,    0,
 2160,    0,    0,    0, 2160, 2160, 2160, 2160, 2160, 2160,
 2160, 2160, 2160, 2160, 2160, 2160, 2160, 2282, 2160, 2160,
    0,    0,    0,    0,    0, 2160, 2160, 2160, 2160, 2160,
 2160, 2160, 2160, 2160, 2160, 2160, 2160, 2160, 2160, 2160,
 2160, 2160, 2160,    0,    0, 2160,    0, 2160, 2160, 2160,
 2160, 2160, 2160,    0,    0,    0,    0,    0, 2282, 2282,
 2282, 2282, 2282, 2282, 2282,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2282, 2282, 2282, 2282,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2282,    0,
 2282,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1408,    0,
    0,    0,    0,    0,    0,    0,    0, 1409,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1410,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1411,    0,    0,    0,    0,    0,
    0, 1412,    0,    0,    0,    0,    0,    0,    0, 1413,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1414,    0, 1415,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1416,    0,
 1417, 1418, 1419,    0,    0,    0,    0, 1420,    0,    0,
    0,    0,    0,    0, 2282,    0, 2282,    0,    0, 2282,
    0, 2282, 2282, 2282,    0, 2282, 2282,    0, 2282,    0,
    0,    0,    0,    0, 2282,    0, 2282, 2282, 2282, 2282,
    0, 2282,    0,    0, 2282,    0,    0, 2282, 1421,    0,
    0,    0,    0,    0, 2282,    0, 2282, 2282,    0, 2282,
 2282, 2282, 2282, 2282,    0,    0, 2282,    0,    0, 2282,
    0, 2282, 2282,    0,    0, 2282, 2282, 2282,    0, 2282,
 2282, 2282, 2282, 2282, 2282, 2282,    0,    0,    0, 2282,
 2282, 2282, 2282, 2282,    0,    0,    0, 2282,    0,    0,
    0,    0,    0,    0,    0, 2282, 2282, 2282, 2282, 2282,
 2282,    0,    0,    0, 2282, 2282,    0, 2282,    0, 2282,
    0, 2282, 2282, 2282,    0, 2282,    0, 2282, 2282, 2282,
    0, 2282, 2282, 2282, 2282, 2282,    0, 2282, 2282,    0,
    0,    0,    0, 2282,    0, 2282, 2282,    0,    0, 2282,
 2282, 2282,    0,    0, 2282,    0, 2282,    0,    0, 2282,
 2282, 2282, 2282,    0, 2282,    0, 2282,    0, 2282, 2282,
 2282, 2282, 2282,    0,    0, 2282, 2282, 2282,    0, 2282,
 2282,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2282, 2282, 2282,    0, 2282, 2282, 2282,    0,    0,    0,
    0,    0, 2282,    0,    0,    0, 2282,    0, 2282, 2282,
 2282, 2282,    0, 2282,    0, 2282, 2282, 2282, 2282, 2282,
 2282, 2282, 2282,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 2282, 2282,    0, 2282, 2282, 2282, 1422,    0,
    0,    0, 2282, 2282, 2282, 2282, 2282,    0, 2282,    0,
    0, 2282, 2282,    0,    0,    0, 2282, 2282,    0,    0,
 2282, 2282, 2282, 2282, 2282, 2282, 2282, 2282, 2282,    0,
 2282,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2282, 2282, 2282, 2282, 2282, 2282, 2282, 2282, 2282, 2282,
 2282, 2282, 2282, 2282, 2282, 2282, 2282, 2282, 2282, 2282,
 2282,    0, 2282, 2282, 2282, 2282, 2282, 2282,    0, 2282,
 2282, 2282, 2282,    0, 2282,    0, 2282, 2282, 2282, 2282,
 2282, 2282, 2282, 2282, 2282, 2282, 2282, 2282, 2282, 2282,
 2282, 2282, 2282, 2282,    0, 2282, 2282, 2282, 2282,    0,
 2282, 2282, 2282, 2282, 2282,    0, 2282, 2282, 2282, 2282,
 2282, 2282, 2282, 2282, 2282, 2282, 2282, 2282, 2282, 2282,
 2282, 2282, 2282, 2282, 2282, 2282, 2282, 2282, 2282, 2282,
 2282, 2282, 2282, 2282, 2282,    0, 2282, 2282,    0,    0,
    0, 2282, 2282, 2282, 2282, 2282, 2282,    0,    0, 2282,
 2282, 2282, 2282, 2282, 2282, 2282, 2282, 2282,    0, 2925,
    0,    0,    0, 2282,    0, 2282, 2282, 2282, 2282, 2282,
 2282,    0, 2282,    0,    0, 1424, 1423, 1425,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 2282,
 2282,    0, 2282, 2282, 2282, 2282, 2282, 2282, 2282, 2282,
 2282,    0, 2282, 2282, 2282, 2282, 2282, 2282, 2282, 2282,
 2282, 2282, 2282, 2282,    0, 2282, 2282, 2282,    0, 2282,
    0,    0,    0, 2282, 2282, 2282, 2282, 2282, 2282, 2282,
 2282, 2282, 2282, 2282, 2282, 2282, 2283, 2282, 2282,    0,
    0,    0,    0,    0, 2282, 2282, 2282, 2282, 2282, 2282,
 2282, 2282, 2282, 2282, 2282, 2282, 2282, 2282, 2282, 2282,
 2282, 2282,    0,    0, 2282,    0, 2282, 2282, 2282, 2282,
 2282, 2282,    0,    0,    0,    0,    0, 2283, 2283, 2283,
 2283, 2283, 2283, 2283,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 2283, 2283, 2283, 2283,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 2283,    0, 2283,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1408,    0,    0,
    0,    0,    0,    0,    0,    0, 1409,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1410,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1411,    0,    0,    0,    0,    0,    0,
 1412,    0,    0,    0,    0,    0,    0,    0, 1413,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1414,    0, 1415,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1416,    0, 1417,
 1418, 1419,    0,    0,    0,    0, 1420,    0,    0,    0,
    0,    0,    0, 2283,    0, 2283,    0,    0, 2283,    0,
 2283, 2283, 2283,    0, 2283, 2283,    0, 2283,    0,    0,
    0,    0,    0, 2283,    0, 2283, 2283, 2283, 2283,    0,
 2283,    0,    0, 2283,    0,    0, 2283, 1421,    0,    0,
    0,    0,    0, 2283,    0, 2283, 2283,    0, 2283, 2283,
 2283, 2283, 2283,    0,    0, 2283,    0,    0, 2283,    0,
 2283, 2283,    0,    0, 2283, 2283, 2283,    0, 2283, 2283,
 2283, 2283, 2283, 2283, 2283,    0,    0,    0, 2283, 2283,
 2283, 2283, 2283,    0,    0,    0, 2283,    0,    0,    0,
    0,    0,    0,    0, 2283, 2283, 2283, 2283, 2283, 2283,
    0,    0,    0, 2283, 2283,    0, 2283,    0, 2283,    0,
 2283, 2283, 2283,    0, 2283,    0, 2283, 2283, 2283,    0,
 2283, 2283, 2283, 2283, 2283,    0, 2283, 2283,    0,    0,
    0,    0, 2283,    0, 2283, 2283,    0,    0, 2283, 2283,
 2283,    0,    0, 2283,    0, 2283,    0,    0, 2283, 2283,
 2283, 2283,    0, 2283,    0, 2283,    0, 2283, 2283, 2283,
 2283, 2283,    0,    0, 2283, 2283, 2283,    0, 2283, 2283,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 2283,
 2283, 2283,    0, 2283, 2283, 2283,    0,    0,    0,    0,
    0, 2283,    0,    0,    0, 2283,    0, 2283, 2283, 2283,
 2283,    0, 2283,    0, 2283, 2283, 2283, 2283, 2283, 2283,
 2283, 2283,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 2283, 2283,    0, 2283, 2283, 2283, 1422,    0,    0,
    0, 2283, 2283, 2283, 2283, 2283,    0, 2283,    0,    0,
 2283, 2283,    0,    0,    0, 2283, 2283,    0,    0, 2283,
 2283, 2283, 2283, 2283, 2283, 2283, 2283, 2283,    0, 2283,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 2283,
 2283, 2283, 2283, 2283, 2283, 2283, 2283, 2283, 2283, 2283,
 2283, 2283, 2283, 2283, 2283, 2283, 2283, 2283, 2283, 2283,
    0, 2283, 2283, 2283, 2283, 2283, 2283,    0, 2283, 2283,
 2283, 2283,    0, 2283,    0, 2283, 2283, 2283, 2283, 2283,
 2283, 2283, 2283, 2283, 2283, 2283, 2283, 2283, 2283, 2283,
 2283, 2283, 2283,    0, 2283, 2283, 2283, 2283,    0, 2283,
 2283, 2283, 2283, 2283,    0, 2283, 2283, 2283, 2283, 2283,
 2283, 2283, 2283, 2283, 2283, 2283, 2283, 2283, 2283, 2283,
 2283, 2283, 2283, 2283, 2283, 2283, 2283, 2283, 2283, 2283,
 2283, 2283, 2283, 2283,    0, 2283, 2283,    0,    0,    0,
 2283, 2283, 2283, 2283, 2283, 2283,    0,    0, 2283, 2283,
 2283, 2283, 2283, 2283, 2283, 2283, 2283,    0, 2926,    0,
    0,    0, 2283,    0, 2283, 2283, 2283, 2283, 2283, 2283,
    0, 2283,    0,    0, 1424, 1423, 1425,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2283, 2283,
    0, 2283, 2283, 2283, 2283, 2283, 2283, 2283, 2283, 2283,
    0, 2283, 2283, 2283, 2283, 2283, 2283, 2283, 2283, 2283,
 2283, 2283, 2283,    0, 2283, 2283, 2283,    0, 2283,    0,
    0,    0, 2283, 2283, 2283, 2283, 2283, 2283, 2283, 2283,
 2283, 2283, 2283, 2283, 2283, 2284, 2283, 2283,    0,    0,
    0,    0,    0, 2283, 2283, 2283, 2283, 2283, 2283, 2283,
 2283, 2283, 2283, 2283, 2283, 2283, 2283, 2283, 2283, 2283,
 2283,    0,    0, 2283,    0, 2283, 2283, 2283, 2283, 2283,
 2283,    0,    0,    0,    0,    0, 2284, 2284, 2284, 2284,
 2284, 2284, 2284,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 2284, 2284, 2284, 2284,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2284,    0, 2284,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1408,    0,    0,    0,
    0,    0,    0,    0,    0, 1409,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1410,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1411,    0,    0,    0,    0,    0,    0, 1412,
    0,    0,    0,    0,    0,    0,    0, 1413,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1414,    0, 1415,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1416,    0, 1417, 1418,
 1419,    0,    0,    0,    0, 1420,    0,    0,    0,    0,
    0,    0, 2284,    0, 2284,    0,    0, 2284,    0, 2284,
 2284, 2284,    0, 2284, 2284,    0, 2284,    0,    0,    0,
    0,    0, 2284,    0, 2284, 2284, 2284, 2284,    0, 2284,
    0,    0, 2284,    0,    0, 2284, 1421,    0,    0,    0,
    0,    0, 2284,    0, 2284, 2284,    0, 2284, 2284, 2284,
 2284, 2284,    0,    0, 2284,    0,    0, 2284,    0, 2284,
 2284,    0,    0, 2284, 2284, 2284,    0, 2284, 2284, 2284,
 2284, 2284, 2284, 2284,    0,    0,    0, 2284, 2284, 2284,
 2284, 2284,    0,    0,    0, 2284,    0,    0,    0,    0,
    0,    0,    0, 2284, 2284, 2284, 2284, 2284, 2284,    0,
    0,    0, 2284, 2284,    0, 2284,    0, 2284,    0, 2284,
 2284, 2284,    0, 2284,    0, 2284, 2284, 2284,    0, 2284,
 2284, 2284, 2284, 2284,    0, 2284, 2284,    0,    0,    0,
    0, 2284,    0, 2284, 2284,    0,    0, 2284, 2284, 2284,
    0,    0, 2284,    0, 2284,    0,    0, 2284, 2284, 2284,
 2284,    0, 2284,    0, 2284,    0, 2284, 2284, 2284, 2284,
 2284,    0,    0, 2284, 2284, 2284,    0, 2284, 2284,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2284, 2284,
 2284,    0, 2284, 2284, 2284,    0,    0,    0,    0,    0,
 2284,    0,    0,    0, 2284,    0, 2284, 2284, 2284, 2284,
    0, 2284,    0, 2284, 2284, 2284, 2284, 2284, 2284, 2284,
 2284,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2284, 2284,    0, 2284, 2284, 2284, 1422,    0,    0,    0,
 2284, 2284, 2284, 2284, 2284,    0, 2284,    0,    0, 2284,
 2284,    0,    0,    0, 2284, 2284,    0,    0, 2284, 2284,
 2284, 2284, 2284, 2284, 2284, 2284, 2284,    0, 2284,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2284, 2284,
 2284, 2284, 2284, 2284, 2284, 2284, 2284, 2284, 2284, 2284,
 2284, 2284, 2284, 2284, 2284, 2284, 2284, 2284, 2284,    0,
 2284, 2284, 2284, 2284, 2284, 2284,    0, 2284, 2284, 2284,
 2284,    0, 2284,    0, 2284, 2284, 2284, 2284, 2284, 2284,
 2284, 2284, 2284, 2284, 2284, 2284, 2284, 2284, 2284, 2284,
 2284, 2284,    0, 2284, 2284, 2284, 2284,    0, 2284, 2284,
 2284, 2284, 2284,    0, 2284, 2284, 2284, 2284, 2284, 2284,
 2284, 2284, 2284, 2284, 2284, 2284, 2284, 2284, 2284, 2284,
 2284, 2284, 2284, 2284, 2284, 2284, 2284, 2284, 2284, 2284,
 2284, 2284, 2284,    0, 2284, 2284,    0,    0,    0, 2284,
 2284, 2284, 2284, 2284, 2284,    0,    0, 2284, 2284, 2284,
 2284, 2284, 2284, 2284, 2284, 2284,    0, 1686,    0,    0,
    0, 2284,    0, 2284, 2284, 2284, 2284, 2284, 2284,    0,
 2284,    0,    0, 1686, 1686, 1686,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 2284, 2284,    0,
 2284, 2284, 2284, 2284, 2284, 2284, 2284, 2284, 2284,    0,
 2284, 2284, 2284, 2284, 2284, 2284, 2284, 2284, 2284, 2284,
 2284, 2284,    0, 2284, 2284, 2284,    0, 2284,    0,    0,
    0, 2284, 2284, 2284, 2284, 2284, 2284, 2284, 2284, 2284,
 2284, 2284, 2284, 2284, 2285, 2284, 2284,    0,    0,    0,
    0,    0, 2284, 2284, 2284, 2284, 2284, 2284, 2284, 2284,
 2284, 2284, 2284, 2284, 2284, 2284, 2284, 2284, 2284, 2284,
    0,    0, 2284,    0, 2284, 2284, 2284, 2284, 2284, 2284,
    0,    0,    0,    0,    0, 2285, 2285, 2285, 2285, 2285,
 2285, 2285,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 2285, 2285, 2285, 2285,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 2285,    0, 2285,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1686,    0,    0,    0,    0,
    0,    0,    0,    0, 1686,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1686,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1686,    0,    0,    0,    0,    0,    0, 1686,    0,
    0,    0,    0,    0,    0,    0, 1686,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1686,    0, 1686,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1686,    0, 1686, 1686, 1686,
    0,    0,    0,    0, 1686,    0,    0,    0,    0,    0,
    0, 2285,    0, 2285,    0,    0, 2285,    0, 2285, 2285,
 2285,    0, 2285, 2285,    0, 2285,    0,    0,    0,    0,
    0, 2285,    0, 2285, 2285, 2285, 2285,    0, 2285,    0,
    0, 2285,    0,    0, 2285, 1686,    0,    0,    0,    0,
    0, 2285,    0, 2285, 2285,    0, 2285, 2285, 2285, 2285,
 2285,    0,    0, 2285,    0,    0, 2285,    0, 2285, 2285,
    0,    0, 2285, 2285, 2285,    0, 2285, 2285, 2285, 2285,
 2285, 2285, 2285,    0,    0,    0, 2285, 2285, 2285, 2285,
 2285,    0,    0,    0, 2285,    0,    0,    0,    0,    0,
    0,    0, 2285, 2285, 2285, 2285, 2285, 2285,    0,    0,
    0, 2285, 2285,    0, 2285,    0, 2285,    0, 2285, 2285,
 2285,    0, 2285,    0, 2285, 2285, 2285,    0, 2285, 2285,
 2285, 2285, 2285,    0, 2285, 2285,    0,    0,  500,    0,
 2285,    0, 2285, 2285,    0,    0, 2285, 2285, 2285,    0,
    0, 2285,    0, 2285,    0,    0, 2285, 2285, 2285, 2285,
    0, 2285,    0, 2285,    0, 2285, 2285, 2285, 2285, 2285,
    0,    0, 2285, 2285, 2285,    0, 2285, 2285,    0,  500,
    0,    0,  500,    0,    0,    0,    0, 2285, 2285, 2285,
    0, 2285, 2285, 2285,    0,    0,    0,  500,    0, 2285,
    0,    0,    0, 2285,    0, 2285, 2285, 2285, 2285,    0,
 2285,    0, 2285, 2285, 2285, 2285, 2285, 2285, 2285, 2285,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 2285,
 2285,    0, 2285, 2285, 2285, 1686,    0,    0,    0, 2285,
 2285, 2285, 2285, 2285,    0, 2285,    0,    0, 2285, 2285,
    0,    0,    0, 2285, 2285,    0,    0, 2285, 2285, 2285,
 2285, 2285, 2285, 2285, 2285, 2285,    0, 2285,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 2285, 2285, 2285,
 2285, 2285, 2285, 2285, 2285, 2285, 2285, 2285, 2285, 2285,
 2285, 2285, 2285, 2285, 2285, 2285, 2285, 2285,    0, 2285,
 2285, 2285, 2285, 2285, 2285,    0, 2285, 2285, 2285, 2285,
    0, 2285,    0, 2285, 2285, 2285, 2285, 2285, 2285, 2285,
 2285, 2285, 2285, 2285, 2285, 2285, 2285, 2285, 2285, 2285,
 2285,    0, 2285, 2285, 2285, 2285,    0, 2285, 2285, 2285,
 2285, 2285,    0, 2285, 2285, 2285, 2285, 2285, 2285, 2285,
 2285, 2285, 2285, 2285, 2285, 2285, 2285, 2285, 2285, 2285,
 2285, 2285, 2285, 2285, 2285, 2285, 2285, 2285, 2285, 2285,
 2285, 2285,    0, 2285, 2285, 1424, 1423, 1425, 2285, 2285,
 2285, 2285, 2285, 2285,    0,    0, 2285, 2285, 2285, 2285,
 2285, 2285, 2285, 2285, 2285,    0,    0,    0,    0,    0,
 2285,    0, 2285, 2285, 2285, 2285, 2285, 2285,    0, 2285,
    0,    0,    0,    0,  500,  500,    0,    0,    0,    0,
    0,  500,    0,    0,    0,    0, 2285, 2285,    0, 2285,
 2285, 2285, 2285, 2285, 2285, 2285, 2285, 2285,    0, 2285,
 2285, 2285, 2285, 2285, 2285, 2285, 2285, 2285, 2285, 2285,
 2285,    0, 2285, 2285, 2285,    0, 2285,    0,    0,    0,
 2285, 2285, 2285, 2285, 2285, 2285, 2285, 2285, 2285, 2285,
 2285, 2285, 2285, 2286, 2285, 2285,    0,    0,    0,    0,
    0, 2285, 2285, 2285, 2285, 2285, 2285, 2285, 2285, 2285,
 2285, 2285, 2285, 2285, 2285, 2285, 2285, 2285, 2285,    0,
    0, 2285,  500, 2285, 2285, 2285, 2285, 2285, 2285,    0,
    0,    0,    0,    0, 2286, 2286, 2286, 2286, 2286, 2286,
 2286,    0,    0,  500,    0,    0,    0,    0,    0,  500,
    0,    0, 2286, 2286, 2286, 2286,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,  500,    0, 2286,    0, 2286,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1408,    0,    0,
  640,    0,    0,    0,    0,    0, 1409,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1410,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,  641,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1411,    0,    0,    0,    0,    0,    0,
 1412,    0,    0,    0,    0,    0,    0,    0, 1413,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1414,    0, 1415,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1416,    0, 1417,
 1418, 1419,    0,    0,    0,    0, 1420,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1421,    0,    0,
 2286,    0, 2286,    0,    0, 2286,    0, 2286, 2286, 2286,
    0, 2286, 2286,    0, 2286,    0,    0,    0,    0,    0,
 2286,    0, 2286, 2286, 2286, 2286,    0, 2286,    0,    0,
 2286,    0,    0, 2286,    0,    0,    0,    0,    0,    0,
 2286,    0, 2286, 2286,    0, 2286, 2286, 2286, 2286, 2286,
    0,    0, 2286,    0,    0, 2286,    0, 2286, 2286,    0,
    0, 2286, 2286, 2286,    0, 2286, 2286, 2286, 2286, 2286,
 2286, 2286,    0,    0,    0, 2286, 2286, 2286, 2286, 2286,
    0,    0,    0, 2286,    0,    0,    0,    0,    0,    0,
    0, 2286, 2286, 2286, 2286, 2286, 2286,    0,    0,    0,
 2286, 2286,    0, 2286,    0, 2286,    0, 2286, 2286, 2286,
    0, 2286,    0, 2286, 2286, 2286,    0, 2286, 2286, 2286,
 2286, 2286,    0, 2286, 2286,    0,    0,    0,    0, 2286,
    0, 2286, 2286,    0,    0, 2286, 2286, 2286,    0,    0,
 2286,    0, 2286,    0,    0, 2286, 2286, 2286, 2286,    0,
 2286,    0, 2286,    0, 2286, 2286, 2286, 2286, 2286,    0,
    0, 2286, 2286, 2286,    0, 2286, 2286,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2286, 2286, 2286,    0,
 2286, 2286, 2286,    0,    0,    0,    0, 1422, 2286,    0,
    0,    0, 2286,    0, 2286, 2286, 2286, 2286,    0, 2286,
    0, 2286, 2286, 2286, 2286, 2286, 2286, 2286, 2286,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2286, 2286,
    0, 2286, 2286, 2286,    0,    0,    0,    0, 2286, 2286,
 2286, 2286, 2286,    0, 2286,    0,    0, 2286, 2286,    0,
    0,    0, 2286, 2286,    0,    0, 2286, 2286, 2286, 2286,
 2286, 2286, 2286, 2286, 2286,    0, 2286,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2286, 2286, 2286, 2286,
 2286, 2286, 2286, 2286, 2286, 2286, 2286, 2286, 2286, 2286,
 2286, 2286, 2286, 2286, 2286, 2286, 2286,    0, 2286, 2286,
 2286, 2286, 2286, 2286,    0, 2286, 2286, 2286, 2286,    0,
 2286,    0, 2286, 2286, 2286, 2286, 2286, 2286, 2286, 2286,
 2286, 2286, 2286, 2286, 2286, 2286, 2286, 2286, 2286, 2286,
    0, 2286, 2286, 2286, 2286,    0, 2286, 2286, 2286, 2286,
 2286,    0, 2286, 2286, 2286, 2286, 2286, 2286, 2286, 2286,
 2286, 2286, 2286, 2286, 2286, 2286, 2286, 2286, 2286, 2286,
 2286, 2286, 2286, 2286, 2286, 2286, 2286, 2286, 2286, 2286,
 2286,    0, 2286, 2286, 1424, 1423, 1425, 2286, 2286, 2286,
 2286, 2286, 2286,    0,    0, 2286, 2286, 2286, 2286, 2286,
 2286, 2286, 2286, 2286,    0,    0,    0,    0,    0, 2286,
    0, 2286, 2286, 2286, 2286, 2286, 2286,    0, 2286,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 2286, 2286,    0, 2286, 2286,
 2286, 2286, 2286, 2286, 2286, 2286, 2286,    0, 2286, 2286,
 2286, 2286, 2286, 2286, 2286, 2286, 2286, 2286, 2286, 2286,
    0, 2286, 2286, 2286,    0, 2286,    0,    0,    0, 2286,
 2286, 2286, 2286, 2286, 2286, 2286, 2286, 2286, 2286, 2286,
 2286, 2286, 2287, 2286, 2286,    0,    0,    0,    0,    0,
 2286, 2286, 2286, 2286, 2286, 2286, 2286, 2286, 2286, 2286,
 2286, 2286, 2286, 2286, 2286, 2286, 2286, 2286,    0,    0,
 2286,    0, 2286, 2286, 2286, 2286, 2286, 2286,    0,    0,
    0,    0,    0, 2287, 2287, 2287, 2287, 2287, 2287, 2287,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 2287, 2287, 2287, 2287,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 2287,    0, 2287,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1408,    0,    0,    0,
    0,    0,    0,    0,    0, 1409,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1410,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1411,    0,    0,    0,    0,    0,    0, 1412,
    0,    0,    0,    0,    0,    0,    0, 1413,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1414,    0, 1415,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1416,    0, 1417, 1418,
 1419,    0,    0,    0,    0, 1420,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1421,    0,    0, 2287,
    0, 2287,    0, 1962, 2287,    0, 2287, 2287, 2287,    0,
 2287, 2287,    0, 2287,    0,    0,    0,    0,    0, 2287,
    0, 2287, 2287, 2287, 2287,    0, 2287,    0,    0, 2287,
    0,    0, 2287,    0,    0,    0,    0,    0,    0, 2287,
    0, 2287, 2287,    0, 2287, 2287, 2287, 2287, 2287,    0,
    0, 2287,    0,    0, 2287,    0, 2287, 2287,    0,    0,
 2287, 2287, 2287,    0, 2287, 2287, 2287, 2287, 2287, 2287,
 2287,    0,    0,    0, 2287, 2287, 2287, 2287, 2287,    0,
    0,    0, 2287,    0,    0,    0,    0,    0,    0,    0,
 2287, 2287, 2287, 2287, 2287, 2287,    0,    0,    0, 2287,
 2287,    0, 2287,    0, 2287,    0, 2287, 2287, 2287,    0,
 2287,    0, 2287, 2287, 2287,    0, 2287, 2287, 2287, 2287,
 2287,    0, 2287, 2287,    0,    0,    0,    0, 2287,    0,
 2287, 2287,    0,    0, 2287, 2287, 2287,    0,    0, 2287,
    0, 2287,    0,    0, 2287, 2287, 2287, 2287,    0, 2287,
    0, 2287,    0, 2287, 2287, 2287, 2287, 2287,    0,    0,
 2287, 2287, 2287,    0, 2287, 2287,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 2287, 2287, 2287,    0, 2287,
 2287, 2287,    0,    0,    0,    0, 1422, 2287,    0,    0,
    0, 2287,    0, 2287, 2287, 2287, 2287,    0, 2287,    0,
 2287, 2287, 2287, 2287, 2287, 2287, 2287, 2287,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 2287, 2287,    0,
 2287, 2287, 2287,    0,    0,    0,    0, 2287, 2287, 2287,
 2287, 2287,    0, 2287,    0,    0, 2287, 2287,    0,    0,
    0, 2287, 2287,    0,    0, 2287, 2287, 2287, 2287, 2287,
 2287, 2287, 2287, 2287,    0, 2287,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 2287, 2287, 2287, 2287, 2287,
 2287, 2287, 2287, 2287, 2287, 2287, 2287, 2287, 2287, 2287,
 2287, 2287, 2287, 2287, 2287, 2287,    0, 2287, 2287, 2287,
 2287, 2287, 2287,    0, 2287, 2287, 2287, 2287,    0, 2287,
    0, 2287, 2287, 2287, 2287, 2287, 2287, 2287, 2287, 2287,
 2287, 2287, 2287, 2287, 2287, 2287, 2287, 2287, 2287,    0,
 2287, 2287, 2287, 2287,    0, 2287, 2287, 2287, 2287, 2287,
    0, 2287, 2287, 2287, 2287, 2287, 2287, 2287, 2287, 2287,
 2287, 2287, 2287, 2287, 2287, 2287, 2287, 2287, 2287, 2287,
 2287, 2287, 2287, 2287, 2287, 2287, 2287, 2287, 2287, 2287,
    0, 2287, 2287, 1424, 1423, 1425, 2287, 2287, 2287, 2287,
 2287, 2287,    0,    0, 2287, 2287, 2287, 2287, 2287, 2287,
 2287, 2287, 2287,    0,    0,    0,    0,    0, 2287,    0,
 2287, 2287, 2287, 2287, 2287, 2287,    0, 2287,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 2287, 2287,    0, 2287, 2287, 2287,
 2287, 2287, 2287, 2287, 2287, 2287,    0, 2287, 2287, 2287,
 2287, 2287, 2287, 2287, 2287, 2287, 2287, 2287, 2287,    0,
 2287, 2287, 2287,    0, 2287,    0,    0,    0, 2287, 2287,
 2287, 2287, 2287, 2287, 2287, 2287, 2287, 2287, 2287, 2287,
 2287, 2288, 2287, 2287,    0,    0,    0,    0,    0, 2287,
 2287, 2287, 2287, 2287, 2287, 2287, 2287, 2287, 2287, 2287,
 2287, 2287, 2287, 2287, 2287, 2287, 2287,    0,    0, 2287,
    0, 2287, 2287, 2287, 2287, 2287, 2287,    0,    0,    0,
    0,    0, 2288, 2288, 2288, 2288, 2288, 2288, 2288,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2288, 2288, 2288, 2288,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 2288,    0, 2288,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1408,    0, 2109,    0,    0,
    0,    0,    0,    0, 1409,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1410,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1411,    0,    0,    0,    0,    0,    0, 1412,    0,
    0,    0,    0,    0,    0,    0, 1413,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1414,    0, 1415,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1416,    0, 1417, 1418, 1419,
    0,    0,    0,    0, 1420,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1421,    0,    0, 2288,    0,
 2288,    0,    0, 2288,    0, 2288, 2288, 2288,    0, 2288,
 2288,    0, 2288,    0,    0,    0,    0,    0, 2288,    0,
 2288, 2288, 2288, 2288,    0, 2288,    0,    0, 2288,    0,
    0, 2288,    0,    0,    0,    0,    0,    0, 2288,    0,
 2288, 2288,    0, 2288, 2288, 2288, 2288, 2288,    0,    0,
 2288,    0,    0, 2288,    0, 2288, 2288,    0,    0, 2288,
 2288, 2288,    0, 2288, 2288, 2288, 2288, 2288, 2288, 2288,
    0,    0,    0, 2288, 2288, 2288, 2288, 2288,    0,    0,
    0, 2288,    0,    0,    0,    0,    0,    0,    0, 2288,
 2288, 2288, 2288, 2288, 2288,    0,    0,    0, 2288, 2288,
    0, 2288,    0, 2288,    0, 2288, 2288, 2288,    0, 2288,
    0, 2288, 2288, 2288,    0, 2288, 2288, 2288, 2288, 2288,
    0, 2288, 2288,    0,    0,    0,    0, 2288,    0, 2288,
 2288,    0,    0, 2288, 2288, 2288,    0,    0, 2288,    0,
 2288,    0,    0, 2288, 2288, 2288, 2288,    0, 2288,    0,
 2288,    0, 2288, 2288, 2288, 2288, 2288,    0,    0, 2288,
 2288, 2288,    0, 2288, 2288,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 2288, 2288, 2288,    0, 2288, 2288,
 2288,    0,    0,    0,    0, 1422, 2288,    0,    0,    0,
 2288,    0, 2288, 2288, 2288, 2288,    0, 2288,    0, 2288,
 2288, 2288, 2288, 2288, 2288, 2288, 2288,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2288, 2288,    0, 2288,
 2288, 2288,    0,    0,    0,    0, 2288, 2288, 2288, 2288,
 2288,    0, 2288,    0,    0, 2288, 2288,    0,    0,    0,
 2288, 2288,    0,    0, 2288, 2288, 2288, 2288, 2288, 2288,
 2288, 2288, 2288,    0, 2288,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 2288, 2288, 2288, 2288, 2288, 2288,
 2288, 2288, 2288, 2288, 2288, 2288, 2288, 2288, 2288, 2288,
 2288, 2288, 2288, 2288, 2288,    0, 2288, 2288, 2288, 2288,
 2288, 2288,    0, 2288, 2288, 2288, 2288,    0, 2288,    0,
 2288, 2288, 2288, 2288, 2288, 2288, 2288, 2288, 2288, 2288,
 2288, 2288, 2288, 2288, 2288, 2288, 2288, 2288,    0, 2288,
 2288, 2288, 2288,    0, 2288, 2288, 2288, 2288, 2288,    0,
 2288, 2288, 2288, 2288, 2288, 2288, 2288, 2288, 2288, 2288,
 2288, 2288, 2288, 2288, 2288, 2288, 2288, 2288, 2288, 2288,
 2288, 2288, 2288, 2288, 2288, 2288, 2288, 2288, 2288,    0,
 2288, 2288, 1424, 1423, 1425, 2288, 2288, 2288, 2288, 2288,
 2288,    0,    0, 2288, 2288, 2288, 2288, 2288, 2288, 2288,
 2288, 2288,    0,    0,    0,    0,    0, 2288,    0, 2288,
 2288, 2288, 2288, 2288, 2288,    0, 2288,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 2288, 2288,    0, 2288, 2288, 2288, 2288,
 2288, 2288, 2288, 2288, 2288,    0, 2288, 2288, 2288, 2288,
 2288, 2288, 2288, 2288, 2288, 2288, 2288, 2288,    0, 2288,
 2288, 2288,    0, 2288,    0,    0,    0, 2288, 2288, 2288,
 2288, 2288, 2288, 2288, 2288, 2288, 2288, 2288, 2288, 2288,
 2289, 2288, 2288,    0,    0,    0,    0,    0, 2288, 2288,
 2288, 2288, 2288, 2288, 2288, 2288, 2288, 2288, 2288, 2288,
 2288, 2288, 2288, 2288, 2288, 2288,    0,    0, 2288,    0,
 2288, 2288, 2288, 2288, 2288, 2288,    0,    0,    0,    0,
    0, 2289, 2289, 2289, 2289, 2289, 2289, 2289,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 2289,
 2289, 2289, 2289,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 2289,    0, 2289,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1408,    0,    0,    0,    0,    0,
    0,    0,    0, 1409,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1410,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2128,    0,    0,    0,
 1411,    0,    0,    0,    0,    0,    0, 1412,    0,    0,
    0,    0,    0,    0,    0, 1413,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1414,    0, 1415,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1416,    0, 1417, 1418, 1419,    0,
    0,    0,    0, 1420,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1421,    0,    0, 2289,    0, 2289,
    0,    0, 2289,    0, 2289, 2289, 2289,    0, 2289, 2289,
    0, 2289,    0,    0,    0,    0,    0, 2289,    0, 2289,
 2289, 2289, 2289,    0, 2289,    0,    0, 2289,    0,    0,
 2289,    0,    0,    0,    0,    0,    0, 2289,    0, 2289,
 2289,    0, 2289, 2289, 2289, 2289, 2289,    0,    0, 2289,
    0,    0, 2289,    0, 2289, 2289,    0,    0, 2289, 2289,
 2289,    0, 2289, 2289, 2289, 2289, 2289, 2289, 2289,    0,
    0,    0, 2289, 2289, 2289, 2289, 2289,    0,    0,    0,
 2289,    0,    0,    0,    0,    0,    0,    0, 2289, 2289,
 2289, 2289, 2289, 2289,    0,    0,    0, 2289, 2289,    0,
 2289,    0, 2289,    0, 2289, 2289, 2289,    0, 2289,    0,
 2289, 2289, 2289,    0, 2289, 2289, 2289, 2289, 2289,    0,
 2289, 2289,    0,    0,    0,    0, 2289,    0, 2289, 2289,
    0,    0, 2289, 2289, 2289,    0,    0, 2289,    0, 2289,
    0,    0, 2289, 2289, 2289, 2289,    0, 2289,    0, 2289,
    0, 2289, 2289, 2289, 2289, 2289,    0,    0, 2289, 2289,
 2289,    0, 2289, 2289,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 2289, 2289, 2289,    0, 2289, 2289, 2289,
    0,    0,    0,    0, 2129, 2289,    0,    0,    0, 2289,
    0, 2289, 2289, 2289, 2289,    0, 2289,    0, 2289, 2289,
 2289, 2289, 2289, 2289, 2289, 2289,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 2289, 2289,    0, 2289, 2289,
 2289,    0,    0,    0,    0, 2289, 2289, 2289, 2289, 2289,
    0, 2289,    0,    0, 2289, 2289,    0,    0,    0, 2289,
 2289,    0,    0, 2289, 2289, 2289, 2289, 2289, 2289, 2289,
 2289, 2289,    0, 2289,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 2289, 2289, 2289, 2289, 2289, 2289, 2289,
 2289, 2289, 2289, 2289, 2289, 2289, 2289, 2289, 2289, 2289,
 2289, 2289, 2289, 2289,    0, 2289, 2289, 2289, 2289, 2289,
 2289,    0, 2289, 2289, 2289, 2289,    0, 2289,    0, 2289,
 2289, 2289, 2289, 2289, 2289, 2289, 2289, 2289, 2289, 2289,
 2289, 2289, 2289, 2289, 2289, 2289, 2289,    0, 2289, 2289,
 2289, 2289,    0, 2289, 2289, 2289, 2289, 2289,    0, 2289,
 2289, 2289, 2289, 2289, 2289, 2289, 2289, 2289, 2289, 2289,
 2289, 2289, 2289, 2289, 2289, 2289, 2289, 2289, 2289, 2289,
 2289, 2289, 2289, 2289, 2289, 2289, 2289, 2289,    0, 2289,
 2289,    0,    0,    0, 2289, 2289, 2289, 2289, 2289, 2289,
    0,    0, 2289, 2289, 2289, 2289, 2289, 2289, 2289, 2289,
 2289,    0,    0, 1424, 1423, 1425, 2289,    0, 2289, 2289,
 2289, 2289, 2289, 2289,    0, 2289,    0,    0,    0,    0,
    0, 1424, 1423, 1425,    0,    0,    0,    0,    0,    0,
    0,    0, 2289, 2289,    0, 2289, 2289, 2289, 2289, 2289,
 2289, 2289, 2289, 2289,    0, 2289, 2289, 2289, 2289, 2289,
 2289, 2289, 2289, 2289, 2289, 2289, 2289,    0, 2289, 2289,
 2289,    0, 2289,    0,    0,    0, 2289, 2289, 2289, 2289,
 2289, 2289, 2289, 2289, 2289, 2289, 2289, 2289, 2289, 2309,
 2289, 2289,    0,    0,    0,    0,    0, 2289, 2289, 2289,
 2289, 2289, 2289, 2289, 2289, 2289, 2289, 2289, 2289, 2289,
 2289, 2289, 2289, 2289, 2289,    0,    0, 2289,    0, 2289,
 2289, 2289, 2289, 2289, 2289,    0,    0,    0,    0,    0,
 2309, 2309, 2309, 2309, 2309, 2309, 2309,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2309, 2309,
 2309, 2309,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2309,    0, 2309,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1408,    0,    0,    0,    0,
    0,    0,    0,    0, 1409,    0,    0,    0,    0,    0,
    0,    0,    0, 1408,    0,    0,    0, 1410,    0,    0,
    0,    0, 1409,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1410,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 2577,    0,    0,
    0, 1411,    0,    0,    0,    0,    0,    0, 1412,    0,
    0,    0,    0,    0,    0,    0, 1413,    0,    0, 1411,
    0,    0,    0,    0,    0,    0, 1412,    0,    0,    0,
    0,    0,    0, 1414, 1413, 1415,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1416,    0, 1417, 1418, 1419,
    0, 1414,    0, 1415, 1420,    0,    0,    0,    0,    0,
    0,    0,    0, 1416,    0, 1417, 1418, 1419,    0,    0,
    0,    0, 1420,    0,    0,    0, 2309,    0, 2309,    0,
    0, 2309,    0, 2309, 2309, 2309,    0, 2309, 2309,    0,
 2309,    0,    0,    0,    0, 1421, 2309,    0, 2309, 2309,
 2309, 2309,    0, 2309,    0,    0, 2309,    0,    0, 2309,
    0,    0,    0, 1421,    0,    0, 2309,    0, 2309, 2309,
    0, 2309, 2309, 2309, 2309, 2309,    0,    0, 2309,    0,
    0, 2309,    0, 2309, 2309,    0,    0, 2309, 2309, 2309,
    0, 2309, 2309, 2309, 2309, 2309, 2309, 2309,    0,    0,
    0, 2309, 2309, 2309, 2309, 2309,    0,    0,    0, 2309,
    0,    0,    0,    0,    0,    0,    0, 2309, 2309, 2309,
 2309, 2309, 2309,    0,    0,    0, 2309, 2309,    0, 2309,
    0, 2309,    0, 2309, 2309, 2309,    0, 2309,    0, 2309,
 2309, 2309,    0, 2309, 2309, 2309, 2309, 2309,    0, 2309,
 2309,    0,    0,    0,    0, 2309,    0, 2309, 2309,    0,
    0, 2309, 2309, 2309,    0,    0, 2309,    0, 2309,    0,
    0, 2309, 2309, 2309, 2309,    0, 2309,    0, 2309,    0,
 2309, 2309, 2309, 2309, 2309,    0,    0, 2309, 2309, 2309,
    0, 2309, 2309,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 2309, 2309, 2309,    0, 2309, 2309, 2309,    0,
    0,    0,    0,    0, 2309,    0,    0,    0, 2309,    0,
 2309, 2309, 2309, 2309, 2156, 2309,    0, 2309, 2309, 2309,
 2309, 2309, 2309, 2309, 2309, 1422,    0,    0,    0,    0,
    0,    0,    0,    0, 2309, 2309,    0, 2309, 2309, 2309,
    0,    0,    0, 1422, 2309, 2309, 2309, 2309, 2309,    0,
 2309,    0,    0, 2309, 2309,    0,    0,    0, 2309, 2309,
    0,    0, 2309, 2309, 2309, 2309, 2309, 2309, 2309, 2309,
 2309,    0, 2309,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 2309, 2309, 2309, 2309, 2309, 2309, 2309, 2309,
 2309, 2309, 2309, 2309, 2309, 2309, 2309, 2309, 2309, 2309,
 2309, 2309, 2309,    0, 2309, 2309, 2309, 2309, 2309, 2309,
    0, 2309, 2309, 2309, 2309,    0, 2309,    0, 2309, 2309,
 2309, 2309, 2309, 2309, 2309, 2309, 2309, 2309, 2309, 2309,
 2309, 2309, 2309, 2309, 2309, 2309,    0, 2309, 2309, 2309,
 2309,    0, 2309, 2309, 2309, 2309, 2309,    0, 2309, 2309,
 2309, 2309, 2309, 2309, 2309, 2309, 2309, 2309, 2309, 2309,
 2309, 2309, 2309, 2309, 2309, 2309, 2309, 2309, 2309, 2309,
 2309, 2309, 2309, 2309, 2309, 2309, 2309,    0, 2309, 2309,
 1424, 1423, 1425, 2309, 2309, 2309, 2309, 2309, 2309,    0,
    0, 2309, 2309, 2309, 2309, 2309, 2309, 2309, 2309, 2309,
    0,    0,    0,    0,    0, 2309,    0, 2309, 2309, 2309,
 2309, 2309, 2309,    0, 2309,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 2309, 2309,    0, 2309, 2309, 2309, 2309, 2309, 2309,
 2309, 2309, 2309,    0, 2309, 2309, 2309, 2309, 2309, 2309,
 2309, 2309, 2309, 2309, 2309, 2309,    0, 2309, 2309, 2309,
    0, 2309,    0,    0,    0, 2309, 2309, 2309, 2309, 2309,
 2309, 2309, 2309, 2309, 2309, 2309, 2309, 2309, 2333, 2309,
 2309,    0,    0,    0,    0,    0, 2309, 2309, 2309, 2309,
 2309, 2309, 2309, 2309, 2309, 2309, 2309, 2309, 2309, 2309,
 2309, 2309, 2309, 2309,    0,    0, 2309,    0, 2309, 2309,
 2309, 2309, 2309, 2309,    0,    0,    0,    0,    0, 2333,
 2333, 2333, 2333, 2333, 2333, 2333,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 2333, 2333, 2333,
 2333,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 2333,
    0, 2333,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1408,    0,    0,    0,    0,    0,    0,    0,
    0, 1409,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1410,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1411,    0,
    0,    0,    0,    0,    0, 1412,    0,    0,    0,    0,
    0,    0,    0, 1413,    0,    0, 2188,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1414,    0, 1415,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1416,    0, 1417, 1418, 1419,    0,    0,    0,
    0, 1420,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1421,    0,    0, 2333,    0, 2333,    0,    0,
 2333,    0, 2333, 2333, 2333,    0, 2333, 2333,    0, 2333,
    0,    0,    0,    0,    0, 2333,    0, 2333, 2333, 2333,
 2333,    0, 2333,    0,    0, 2333,    0,    0, 2333,    0,
    0,    0,    0,    0,    0, 2333,    0, 2333, 2333,    0,
 2333, 2333, 2333, 2333, 2333,    0,    0, 2333,    0,    0,
 2333,    0, 2333, 2333,    0,    0, 2333, 2333, 2333,    0,
 2333, 2333, 2333, 2333, 2333, 2333, 2333,    0,    0,    0,
 2333, 2333, 2333, 2333, 2333,    0,    0,    0, 2333,    0,
    0,    0,    0,    0,    0,    0, 2333, 2333, 2333, 2333,
 2333, 2333,    0,    0,    0, 2333, 2333,    0, 2333,    0,
 2333,    0, 2333, 2333, 2333,    0, 2333,    0, 2333, 2333,
 2333,    0, 2333, 2333, 2333, 2333, 2333,    0, 2333, 2333,
    0,    0,    0,    0, 2333,    0, 2333, 2333,    0,    0,
 2333, 2333, 2333,    0,    0, 2333,    0, 2333,    0,    0,
 2333, 2333, 2333, 2333,    0, 2333,    0, 2333,    0, 2333,
 2333, 2333, 2333, 2333,    0,    0, 2333, 2333, 2333,    0,
 2333, 2333,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 2333, 2333, 2333,    0, 2333, 2333, 2333,    0,    0,
    0,    0, 1422, 2333,    0,    0,    0, 2333,    0, 2333,
 2333, 2333, 2333,    0, 2333,    0, 2333, 2333, 2333, 2333,
 2333, 2333, 2333, 2333,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 2333, 2333,    0, 2333, 2333, 2333,    0,
    0,    0,    0, 2333, 2333, 2333, 2333, 2333,    0, 2333,
    0,    0, 2333, 2333,    0,    0,    0, 2333, 2333,    0,
    0, 2333, 2333, 2333, 2333, 2333, 2333, 2333, 2333, 2333,
    0, 2333,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 2333, 2333, 2333, 2333, 2333, 2333, 2333, 2333, 2333,
 2333, 2333, 2333, 2333, 2333, 2333, 2333, 2333, 2333, 2333,
 2333, 2333,    0, 2333, 2333, 2333, 2333, 2333, 2333,    0,
 2333, 2333, 2333, 2333,    0, 2333,    0, 2333, 2333, 2333,
 2333, 2333, 2333, 2333, 2333, 2333, 2333, 2333, 2333, 2333,
 2333, 2333, 2333, 2333, 2333,    0, 2333, 2333, 2333, 2333,
    0, 2333, 2333, 2333, 2333, 2333,    0, 2333, 2333, 2333,
 2333, 2333, 2333, 2333, 2333, 2333, 2333, 2333, 2333, 2333,
 2333, 2333, 2333, 2333, 2333, 2333, 2333, 2333, 2333, 2333,
 2333, 2333, 2333, 2333, 2333, 2333,    0, 2333, 2333, 1424,
 1423, 1425, 2333, 2333, 2333, 2333, 2333, 2333,    0,    0,
 2333, 2333, 2333, 2333, 2333, 2333, 2333, 2333, 2333,    0,
    0,    0,    0,    0, 2333,    0, 2333, 2333, 2333, 2333,
 2333, 2333,    0, 2333,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2333, 2333,    0, 2333, 2333, 2333, 2333, 2333, 2333, 2333,
 2333, 2333,    0, 2333, 2333, 2333, 2333, 2333, 2333, 2333,
 2333, 2333, 2333, 2333, 2333,    0, 2333, 2333, 2333,    0,
 2333,    0,    0,    0, 2333, 2333, 2333, 2333, 2333, 2333,
 2333, 2333, 2333, 2333, 2333, 2333, 2333, 2338, 2333, 2333,
    0,    0,    0,    0,    0, 2333, 2333, 2333, 2333, 2333,
 2333, 2333, 2333, 2333, 2333, 2333, 2333, 2333, 2333, 2333,
 2333, 2333, 2333,    0,    0, 2333,    0, 2333, 2333, 2333,
 2333, 2333, 2333,    0,    0,    0,    0,    0, 2338, 2338,
 2338, 2338, 2338, 2338, 2338,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2338, 2338, 2338, 2338,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2338,    0,
 2338,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1408,    0,    0,    0,    0,    0,    0,    0,    0,
 1409,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1410,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1411,    0,    0,
    0,    0,    0,    0, 1412,    0,    0,    0,    0,    0,
    0,    0, 1413,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1414,
    0, 1415,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1416,    0, 1417, 1418, 1419,    0,    0,    0,    0,
 1420,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1421,    0,    0, 2338,    0, 2338,    0,    0, 2338,
    0, 2338, 2338, 2338,    0, 2338, 2338,    0, 2338,    0,
    0,    0,    0,    0, 2338,    0, 2338, 2338, 2338, 2338,
    0, 2338,    0,    0, 2338,    0,    0, 2338,    0,    0,
    0,    0,    0,    0, 2338,    0, 2338, 2338,    0, 2338,
 2338, 2338, 2338, 2338,    0,    0, 2338,    0,    0, 2338,
    0, 2338, 2338,    0,    0, 2338, 2338, 2338,    0, 2338,
 2338, 2338, 2338, 2338, 2338, 2338, 2218,    0,    0, 2338,
 2338, 2338, 2338, 2338,    0,    0,    0, 2338,    0,    0,
    0,    0,    0,    0,    0, 2338, 2338, 2338, 2338, 2338,
 2338,    0,    0,    0, 2338, 2338,    0, 2338,    0, 2338,
    0, 2338, 2338, 2338,    0, 2338,    0, 2338, 2338, 2338,
    0, 2338, 2338, 2338, 2338, 2338,    0, 2338, 2338,    0,
    0,    0,    0, 2338,    0, 2338, 2338,    0,    0, 2338,
 2338, 2338,    0,    0, 2338,    0, 2338,    0,    0, 2338,
 2338, 2338, 2338,    0, 2338,    0, 2338,    0, 2338, 2338,
 2338, 2338, 2338,    0,    0, 2338, 2338, 2338,    0, 2338,
 2338,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2338, 2338, 2338,    0, 2338, 2338, 2338,    0,    0,    0,
    0, 1422, 2338,    0,    0,    0, 2338,    0, 2338, 2338,
 2338, 2338,    0, 2338,    0, 2338, 2338, 2338, 2338, 2338,
 2338, 2338, 2338,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 2338, 2338,    0, 2338, 2338, 2338,    0,    0,
    0,    0, 2338, 2338, 2338, 2338, 2338,    0, 2338,    0,
    0, 2338, 2338,    0,    0,    0, 2338, 2338,    0,    0,
 2338, 2338, 2338, 2338, 2338, 2338, 2338, 2338, 2338,    0,
 2338,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2338, 2338, 2338, 2338, 2338, 2338, 2338, 2338, 2338, 2338,
 2338, 2338, 2338, 2338, 2338, 2338, 2338, 2338, 2338, 2338,
 2338,    0, 2338, 2338, 2338, 2338, 2338, 2338,    0, 2338,
 2338, 2338, 2338,    0, 2338,    0, 2338, 2338, 2338, 2338,
 2338, 2338, 2338, 2338, 2338, 2338, 2338, 2338, 2338, 2338,
 2338, 2338, 2338, 2338,    0, 2338, 2338, 2338, 2338,    0,
 2338, 2338, 2338, 2338, 2338,    0, 2338, 2338, 2338, 2338,
 2338, 2338, 2338, 2338, 2338, 2338, 2338, 2338, 2338, 2338,
 2338, 2338, 2338, 2338, 2338, 2338, 2338, 2338, 2338, 2338,
 2338, 2338, 2338, 2338, 2338,    0, 2338, 2338, 1424, 1423,
 1425, 2338, 2338, 2338, 2338, 2338, 2338,    0,    0, 2338,
 2338, 2338, 2338, 2338, 2338, 2338, 2338, 2338,    0,    0,
    0,    0,    0, 2338,    0, 2338, 2338, 2338, 2338, 2338,
 2338,    0, 2338,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 2338,
 2338,    0, 2338, 2338, 2338, 2338, 2338, 2338, 2338, 2338,
 2338,    0, 2338, 2338, 2338, 2338, 2338, 2338, 2338, 2338,
 2338, 2338, 2338, 2338,    0, 2338, 2338, 2338,    0, 2338,
    0,    0,    0, 2338, 2338, 2338, 2338, 2338, 2338, 2338,
 2338, 2338, 2338, 2338, 2338, 2338, 1791, 2338, 2338,    0,
    0,    0,    0,    0, 2338, 2338, 2338, 2338, 2338, 2338,
 2338, 2338, 2338, 2338, 2338, 2338, 2338, 2338, 2338, 2338,
 2338, 2338,    0,    0, 2338,    0, 2338, 2338, 2338, 2338,
 2338, 2338,    0,    0,    0,    0,    0, 1791, 1791, 1791,
 1791, 1791,    0, 1791,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1791, 1791, 1791, 1791,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1791,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1408,    0,    0,    0,    0,    0,    0,    0,    0, 1409,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1410,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1411,    0,    0,    0,
    0,    0,    0, 1412,    0,    0,    0,    0,    0,    0,
    0, 1413,    0,    0, 2219,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1414,    0,
 1415,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1416,    0, 1417, 1418, 1419,    0,    0,    0,    0, 1420,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1421,    0,    0, 1791,    0, 1791,    0,    0, 1791,    0,
 1791, 1791, 1791,    0, 1791, 1791,    0, 1791,    0,    0,
    0,    0,    0, 1791,    0, 1791, 1791, 1791, 1791,    0,
 1791,    0,    0, 1791,    0,    0, 1791,    0,    0,    0,
    0,    0,    0, 1791,    0, 1791, 1791,    0, 1791, 1791,
 1791, 1791, 1791,    0,    0, 1791,    0,    0, 1791,    0,
 1791, 1791,    0,    0, 1791, 1791, 1791,    0, 1791, 1791,
 1791, 1791, 1791, 1791, 1791,    0,    0,    0, 1791, 1791,
 1791, 1791, 1791,    0,    0,    0, 1791,    0,    0,    0,
    0,    0,    0,    0, 1791, 1791, 1791, 1791, 1791, 1791,
    0,    0,    0, 1791, 1791,    0, 1791,    0, 1791,    0,
 1791, 1791, 1791,    0, 1791,    0, 1791, 1791, 1791,    0,
 1791, 1791, 1791, 1791, 1791,    0, 1791, 1791,    0,    0,
    0,    0, 1791,    0, 1791, 1791,    0,    0, 1791, 1791,
 1791,    0,    0, 1791,    0, 1791,    0,    0, 1791, 1791,
 1791, 1791,    0, 1791,    0, 1791,    0, 1791, 1791, 1791,
 1791, 1791,    0,    0, 1791, 1791, 1791,    0, 1791, 1791,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1791,
 1791, 1791,    0, 1791, 1791, 1791,    0,    0,    0,    0,
 1422, 1791,    0,    0,    0, 1791,    0, 1791, 1791, 1791,
 1791,    0, 1791,    0, 1791, 1791, 1791, 1791, 1791, 1791,
 1791, 1791,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1791, 1791,    0, 1791, 1791, 1791,    0,    0,    0,
    0, 1791, 1791, 1791, 1791, 1791,    0, 1791,    0,    0,
 1791, 1791,    0,    0,    0, 1791, 1791,    0,    0, 1791,
 1791, 1791, 1791, 1791, 1791, 1791, 1791, 1791,    0, 1791,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1791,
 1791, 1791, 1791, 1791, 1791, 1791, 1791, 1791, 1791, 1791,
 1791, 1791, 1791, 1791, 1791, 1791, 1791, 1791, 1791, 1791,
    0, 1791, 1791, 1791, 1791, 1791, 1791,    0, 1791, 1791,
 1791, 1791,    0, 1791,    0, 1791, 1791, 1791, 1791, 1791,
 1791, 1791, 1791, 1791, 1791, 1791, 1791, 1791, 1791, 1791,
 1791, 1791, 1791,    0, 1791, 1791, 1791, 1791,    0, 1791,
 1791, 1791, 1791, 1791,    0, 1791, 1791, 1791, 1791, 1791,
 1791, 1791, 1791, 1791, 1791, 1791, 1791, 1791, 1791, 1791,
 1791, 1791, 1791, 1791, 1791,    0, 1791, 1791, 1791, 1791,
 1791, 1791, 1791, 1791, 1849, 1791, 1791, 1424, 1423, 1425,
 1791, 1791, 1791, 1791, 1791, 1791,    0,    0, 1791, 1791,
 1791, 1791, 1791, 1791, 1791, 1791, 1791,    0,    0,    0,
    0,    0, 1791,    0, 1791, 1791, 1791, 1791, 1791, 1791,
    0, 1791,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1791, 1791,
    0, 1791, 1791, 1791, 1791, 1791, 1791, 1791, 1791, 1791,
    0, 1791, 1791, 1791, 1791, 1791, 1791, 1791, 1791, 1791,
 1791, 1791, 1791,    0, 1791, 1791, 1791,    0, 1791,    0,
    0,    0, 1791, 1791, 1791, 1791, 1791, 1791, 1791, 1791,
 1791, 1791, 1791, 1791, 1791, 1797, 1791, 1791,    0,    0,
    0,    0,    0, 1791, 1791, 1791, 1791, 1791, 1791, 1791,
 1791, 1791, 1791, 1791, 1791, 1791, 1791, 1791, 1791, 1791,
 1791,    0,    0, 1791,    0, 1791, 1791, 1791, 1791, 1791,
 1791,    0,    0,    0,    0,    0, 1797, 1797, 1797, 1797,
 1797,    0, 1797,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1797, 1797, 1797, 1797,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1797,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1408,
    0,    0,    0,    0,    0,    0,    0,    0, 1409,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1410,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1411,    0,    0,    0,    0,
    0,    0, 1412,    0,    0,    0,    0,    0,    0,    0,
 1413,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1414,    0, 1415,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1416,
    0, 1417, 1418, 1419,    0,    0,    0,    0, 1420,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1421,
    0,    0, 1797,    0, 1797,    0,    0, 1797,    0, 1797,
 1797, 1797,    0, 1797, 1797,    0, 1797,    0,    0,    0,
    0,    0, 1797,    0, 1797, 1797, 1797, 1797,    0, 1797,
    0,    0, 1797,    0,    0, 1797,    0,    0,    0,    0,
    0,    0, 1797,    0, 1797, 1797,    0, 1797, 1797, 1797,
 1797, 1797,    0,    0, 1797,    0,    0, 1797,    0, 1797,
 1797,    0,    0, 1797, 1797, 1797,    0, 1797, 1797, 1797,
 1797, 1797, 1797, 1797, 2220,    0,    0, 1797, 1797, 1797,
 1797, 1797,    0,    0,    0, 1797,    0,    0,    0,    0,
    0,    0,    0, 1797, 1797, 1797, 1797, 1797, 1797,    0,
    0,    0, 1797, 1797,    0, 1797,    0, 1797,    0, 1797,
 1797, 1797,    0, 1797,    0, 1797, 1797, 1797,    0, 1797,
 1797, 1797, 1797, 1797,    0, 1797, 1797,    0,    0,    0,
    0, 1797,    0, 1797, 1797,    0,    0, 1797, 1797, 1797,
    0,    0, 1797,    0, 1797,    0,    0, 1797, 1797, 1797,
 1797,    0, 1797,    0, 1797,    0, 1797, 1797, 1797, 1797,
 1797,    0,    0, 1797, 1797, 1797,    0, 1797, 1797,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1797, 1797,
 1797,    0, 1797, 1797, 1797,    0,    0,    0,    0, 1422,
 1797,    0,    0,    0, 1797,    0, 1797, 1797, 1797, 1797,
    0, 1797,    0, 1797, 1797, 1797, 1797, 1797, 1797, 1797,
 1797,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1797, 1797,    0, 1797, 1797, 1797,    0,    0,    0,    0,
 1797, 1797, 1797, 1797, 1797,    0, 1797,    0,    0, 1797,
 1797,    0,    0,    0, 1797, 1797,    0,    0, 1797, 1797,
 1797, 1797, 1797, 1797, 1797, 1797, 1797,    0, 1797,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1797, 1797,
 1797, 1797, 1797, 1797, 1797, 1797, 1797, 1797, 1797, 1797,
 1797, 1797, 1797, 1797, 1797, 1797, 1797, 1797, 1797,    0,
 1797, 1797, 1797, 1797, 1797, 1797,    0, 1797, 1797, 1797,
 1797,    0, 1797,    0, 1797, 1797, 1797, 1797, 1797, 1797,
 1797, 1797, 1797, 1797, 1797, 1797, 1797, 1797, 1797, 1797,
 1797, 1797,    0, 1797, 1797, 1797, 1797,    0, 1797, 1797,
 1797, 1797, 1797,    0, 1797, 1797, 1797, 1797, 1797, 1797,
 1797, 1797, 1797, 1797, 1797, 1797, 1797, 1797, 1797, 1797,
 1797, 1797, 1797, 1797,    0, 1797, 1797, 1797, 1797, 1797,
 1797, 1797, 1797, 1797, 1797, 1797, 1424, 1423, 1425, 1797,
 1797, 1797, 1797, 1797, 1797,    0,    0, 1797, 1797, 1797,
 1797, 1797, 1797, 1797, 1797, 1797,    0,    0,    0,    0,
    0, 1797,    0, 1797, 1797, 1797, 1797, 1797, 1797,    0,
 1797,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1797, 1797,    0,
 1797, 1797, 1797, 1797, 1797, 1797, 1797, 1797, 1797,    0,
 1797, 1797, 1797, 1797, 1797, 1797, 1797, 1797, 1797, 1797,
 1797, 1797,    0, 1797, 1797, 1797,    0, 1797,    0,    0,
    0, 1797, 1797, 1797, 1797, 1797, 1797, 1797, 1797, 1797,
 1797, 1797, 1797, 1797, 2074, 1797, 1797,    0,    0,    0,
    0,    0, 1797, 1797, 1797, 1797, 1797, 1797, 1797, 1797,
 1797, 1797, 1797, 1797, 1797, 1797, 1797, 1797, 1797, 1797,
    0,    0, 1797,    0, 1797, 1797, 1797, 1797, 1797, 1797,
    0,    0,    0,    0,    0, 2074, 2074, 2074, 2074, 2074,
    0, 2074,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 2074, 2074, 2074, 2074,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 2074,    0, 2074,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1408,    0,
    0,    0,    0,    0,    0,    0,    0, 1409,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1410,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2910,    0,
    0,    0,    0,    0, 1411,    0,    0,    0,    0,    0,
    0, 1412,    0,    0,    0,    0,    0,    0,    0, 1413,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1414,    0, 1415,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1416,    0,
 1417, 1418, 1419,    0,    0,    0,    0, 1420,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1421,    0,
    0, 2074,    0, 2074,    0,    0, 2074,    0, 2074, 2074,
 2074,    0, 2074, 2074,    0, 2074,    0,    0,    0,    0,
    0, 2074,    0, 2074, 2074, 2074, 2074,    0, 2074,    0,
    0, 2074,    0,    0, 2074,    0,    0,    0,    0,    0,
    0, 2074,    0, 2074, 2074,    0, 2074, 2074, 2074, 2074,
 2074,    0,    0, 2074,    0,    0, 2074,    0, 2074, 2074,
    0,    0, 2074, 2074, 2074,    0, 2074, 2074, 2074, 2074,
 2074, 2074, 2074,    0,    0,    0, 2074, 2074, 2074, 2074,
 2074,    0,    0,    0, 2074,    0,    0,    0,    0,    0,
    0,    0, 2074, 2074, 2074, 2074, 2074, 2074,    0,    0,
    0, 2074, 2074,    0, 2074,    0, 2074,    0, 2074, 2074,
 2074,    0, 2074,    0, 2074, 2074, 2074,    0, 2074, 2074,
 2074, 2074, 2074,    0, 2074, 2074,    0,    0,    0,    0,
 2074,    0, 2074, 2074,    0,    0, 2074, 2074, 2074,    0,
    0, 2074,    0, 2074,    0,    0, 2074, 2074, 2074, 2074,
    0, 2074,    0, 2074,    0, 2074, 2074, 2074, 2074, 2074,
    0,    0, 2074, 2074, 2074,    0, 2074, 2074,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 2074, 2074, 2074,
    0, 2074, 2074, 2074,    0,    0,    0,    0, 1422, 2074,
    0,    0,    0, 2074,    0, 2074, 2074, 2074, 2074,    0,
 2074,    0, 2074, 2074, 2074, 2074, 2074, 2074, 2074, 2074,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 2074,
 2074,    0, 2074, 2074, 2074,    0,    0,    0,    0, 2074,
 2074, 2074, 2074, 2074,    0, 2074,    0,    0, 2074, 2074,
    0,    0,    0, 2074, 2074,    0,    0, 2074, 2074, 2074,
 2074, 2074, 2074, 2074, 2074, 2074,    0, 2074,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 2074, 2074, 2074,
 2074, 2074, 2074, 2074, 2074, 2074, 2074, 2074, 2074, 2074,
 2074, 2074, 2074, 2074, 2074, 2074, 2074, 2074,    0, 2074,
 2074, 2074, 2074, 2074, 2074,    0, 2074, 2074, 2074, 2074,
    0, 2074,    0, 2074, 2074, 2074, 2074, 2074, 2074, 2074,
 2074, 2074, 2074, 2074, 2074, 2074, 2074, 2074, 2074, 2074,
 2074,    0, 2074, 2074, 2074, 2074,    0, 2074, 2074, 2074,
 2074, 2074,    0, 2074, 2074, 2074, 2074, 2074, 2074, 2074,
 2074, 2074, 2074, 2074, 2074, 2074, 2074, 2074, 2074, 2074,
 2074, 2074, 2074,    0, 2074, 2074, 2074, 2074, 2074, 2074,
 2074, 2074,    0, 2074, 2074, 1424, 1423, 1425, 2074, 2074,
 2074, 2074, 2074, 2074,    0,    0, 2074, 2074, 2074, 2074,
 2074, 2074, 2074, 2074, 2074,    0,    0,    0,    0,    0,
 2074,    0, 2074, 2074, 2074, 2074, 2074, 2074,    0, 2074,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2074, 2074,    0, 2074,
 2074, 2074, 2074, 2074, 2074, 2074, 2074, 2074,    0, 2074,
 2074, 2074, 2074, 2074, 2074, 2074, 2074, 2074, 2074, 2074,
 2074,    0, 2074, 2074, 2074,    0, 2074,    0,    0,    0,
 2074, 2074, 2074, 2074, 2074, 2074, 2074, 2074, 2074, 2074,
 2074, 2074, 2074, 2092, 2074, 2074,    0,    0,    0,    0,
    0, 2074, 2074, 2074, 2074, 2074, 2074, 2074, 2074, 2074,
 2074, 2074, 2074, 2074, 2074, 2074, 2074, 2074, 2074,    0,
    0, 2074,    0, 2074, 2074, 2074, 2074, 2074, 2074,    0,
    0,    0,    0,    0, 2092, 2092, 2092, 2092, 2092, 2088,
 2092,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 2092, 2092, 2092, 2092,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2092,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1408,    0,    0,
    0,    0,    0,    0,    0,    0, 1409,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1410,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2912,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1411,    0,    0,    0,    0,    0,    0,
 1412,    0,    0,    0,    0,    0,    0,    0, 1413,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1414,    0, 1415,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1416,    0, 1417,
 1418, 1419,    0,    0,    0,    0, 1420,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1421,    0,    0,
 2092,    0, 2092,    0,    0, 2092,    0, 2092, 2092, 2092,
    0, 2092, 2092,    0, 2092,    0,    0,    0,    0,    0,
 2092,    0, 2092, 2092, 2092, 2092,    0, 2092,    0,    0,
 2092,    0,    0, 2092,    0,    0,    0,    0,    0,    0,
 2092,    0, 2092, 2092,    0, 2092, 2092, 2092, 2092, 2092,
    0,    0, 2092,    0,    0, 2092,    0, 2092, 2092,    0,
    0, 2092, 2092, 2092,    0, 2092, 2092, 2092, 2092, 2092,
 2092, 2092,    0,    0,    0, 2092, 2092, 2092, 2092, 2092,
    0,    0,    0, 2092,    0,    0,    0,    0,    0,    0,
    0, 2092, 2092, 2092, 2092, 2092, 2092,    0,    0,    0,
 2092, 2092,    0, 2092,    0, 2092,    0, 2092, 2092, 2092,
    0, 2092,    0, 2092, 2092, 2092,    0, 2092, 2092, 2092,
 2092, 2092,    0, 2092, 2092,    0,    0,    0,    0, 2092,
    0, 2092, 2092,    0,    0, 2092, 2092, 2092,    0,    0,
 2092,    0, 2092,    0,    0, 2092, 2092, 2092, 2092,    0,
 2092,    0, 2092,    0, 2092, 2092, 2092, 2092, 2092,    0,
    0, 2092, 2092, 2092,    0, 2092, 2092,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2092, 2092, 2092,    0,
 2092, 2092, 2092,    0,    0,    0,    0, 1422, 2092,    0,
    0,    0, 2092,    0, 2092, 2092, 2092, 2092,    0, 2092,
    0, 2092, 2092, 2092, 2092, 2092, 2092, 2092, 2092,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2092, 2092,
    0, 2092, 2092, 2092,    0,    0,    0,    0, 2092, 2092,
 2092, 2092, 2092,    0, 2092,    0,    0, 2092, 2092,    0,
    0,    0, 2092, 2092,    0,    0, 2092, 2092, 2092, 2092,
 2092, 2092, 2092, 2092, 2092,    0, 2092,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2092, 2092, 2092, 2092,
 2092, 2092, 2092, 2092, 2092, 2092, 2092, 2092, 2092, 2092,
 2092, 2092, 2092, 2092, 2092, 2092, 2092,    0, 2092, 2092,
 2092, 2092, 2092, 2092,    0, 2092, 2092, 2092, 2092,    0,
 2092,    0, 2092, 2092, 2092, 2092, 2092, 2092, 2092, 2092,
 2092, 2092, 2092, 2092, 2092, 2092, 2092, 2092, 2092, 2092,
    0, 2092, 2092, 2092, 2092,    0, 2092, 2092, 2092, 2092,
 2092,    0, 2092, 2092, 2092, 2092, 2092, 2092, 2092, 2092,
 2092, 2092, 2092, 2092, 2092, 2092, 2092, 2092, 2092, 2092,
 2092, 2092,    0, 2092, 2092, 2092, 2092, 2092, 2092, 2092,
 2092,    0, 2092, 2092, 1424, 1423, 1425, 2092, 2092, 2092,
 2092, 2092, 2092,    0,    0, 2092, 2092, 2092, 2092, 2092,
 2092, 2092, 2092, 2092,    0,    0,    0,    0,    0, 2092,
    0, 2092, 2092, 2092, 2092, 2092, 2092,    0, 2092,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 2092, 2092,    0, 2092, 2092,
 2092, 2092, 2092, 2092, 2092, 2092, 2092,    0, 2092, 2092,
 2092, 2092, 2092, 2092, 2092, 2092, 2092, 2092, 2092, 2092,
    0, 2092, 2092, 2092,    0, 2092,    0,    0,    0, 2092,
 2092, 2092, 2092, 2092, 2092, 2092, 2092, 2092, 2092, 2092,
 2092, 2092, 1694, 2092, 2092,    0,    0,    0,    0,    0,
 2092, 2092, 2092, 2092, 2092, 2092, 2092, 2092, 2092, 2092,
 2092, 2092, 2092, 2092, 2092, 2092, 2092, 2092,    0,    0,
 2092,    0, 2092, 2092, 2092, 2092, 2092, 2092,    0,    0,
    0,    0,    0, 1694, 1694, 1694, 1694, 1694,    0, 1694,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1694, 1694, 1694, 1694,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1694,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1408,    0,    0,    0,
    0,    0,    0,    0,    0, 1409,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1410,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1411,    0,    0,    0,    0,    0,    0, 1412,
    0,    0,    0,    0,    0,    0,    0, 1413,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1414,    0, 1415,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1416,    0, 1417, 1418,
 1419,    0,    0,    0,    0, 1420,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1421,    0,    0, 1694,
    0, 1694,    0, 2924, 1694,    0, 1694, 1694, 1694,    0,
 1694, 1694,    0, 1694,    0,    0,    0,    0,    0, 1694,
    0, 1694, 1694, 1694, 1694,    0, 1694,    0,    0, 1694,
    0,    0, 1694,    0,    0,    0,    0,    0,    0, 1694,
    0, 1694, 1694,    0, 1694, 1694, 1694, 1694, 1694,    0,
    0, 1694,    0,    0, 1694,    0, 1694, 1694,    0,    0,
 1694, 1694, 1694,    0, 1694, 1694, 1694, 1694, 1694, 1694,
 1694,    0,    0,    0, 1694, 1694, 1694, 1694, 1694,    0,
    0,    0, 1694,    0,    0,    0,    0,    0,    0,    0,
 1694, 1694, 1694, 1694, 1694, 1694,    0,    0,    0, 1694,
 1694,    0, 1694,    0, 1694,    0, 1694, 1694, 1694,    0,
 1694,    0, 1694, 1694, 1694,    0, 1694, 1694, 1694, 1694,
 1694,    0, 1694, 1694,    0,    0,    0,    0, 1694,    0,
 1694, 1694,    0,    0, 1694, 1694, 1694,    0,    0, 1694,
    0, 1694,    0,    0, 1694, 1694, 1694, 1694,    0, 1694,
    0, 1694,    0, 1694, 1694, 1694, 1694, 1694,    0,    0,
 1694, 1694, 1694,    0, 1694, 1694,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1694, 1694, 1694,    0, 1694,
 1694, 1694,    0,    0,    0,    0, 1422, 1694,    0,    0,
    0, 1694,    0, 1694, 1694, 1694, 1694,    0, 1694,    0,
 1694, 1694, 1694, 1694, 1694, 1694, 1694, 1694,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1694, 1694,    0,
 1694, 1694, 1694,    0,    0,    0,    0, 1694, 1694, 1694,
 1694, 1694,    0, 1694,    0,    0, 1694, 1694,    0,    0,
    0, 1694, 1694,    0,    0, 1694, 1694, 1694, 1694, 1694,
 1694, 1694, 1694, 1694,    0, 1694,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1694, 1694, 1694, 1694, 1694,
 1694, 1694, 1694, 1694, 1694, 1694, 1694, 1694, 1694, 1694,
 1694, 1694, 1694, 1694, 1694, 1694,    0, 1694, 1694, 1694,
 1694, 1694, 1694,    0, 1694, 1694, 1694, 1694,    0, 1694,
    0, 1694, 1694, 1694, 1694, 1694, 1694, 1694, 1694, 1694,
 1694, 1694, 1694, 1694, 1694, 1694, 1694, 1694, 1694,    0,
 1694, 1694, 1694, 1694,    0, 1694, 1694, 1694, 1694, 1694,
    0, 1694, 1694, 1694, 1694, 1694, 1694, 1694, 1694, 1694,
 1694, 1694, 1694, 1694, 1694, 1694, 1694, 1694, 1694, 1694,
 1694,    0, 1694, 1694, 1694, 1694, 1694, 1694, 1694, 1694,
    0, 1694, 1694, 1424, 1423, 1425, 1694, 1694, 1694, 1694,
 1694, 1694,    0,    0, 1694, 1694, 1694, 1694, 1694, 1694,
 1694, 1694, 1694,    0,    0,    0,    0,    0, 1694,    0,
 1694, 1694, 1694, 1694, 1694, 1694,    0, 1694,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1694, 1694,    0, 1694, 1694, 1694,
 1694, 1694, 1694, 1694, 1694, 1694,    0, 1694, 1694, 1694,
 1694, 1694, 1694, 1694, 1694, 1694, 1694, 1694, 1694,    0,
 1694, 1694, 1694,    0, 1694,    0,    0,    0, 1694, 1694,
 1694, 1694, 1694, 1694, 1694, 1694, 1694, 1694, 1694, 1694,
 1694, 1685, 1694, 1694,    0,    0,    0,    0,    0, 1694,
 1694, 1694, 1694, 1694, 1694, 1694, 1694, 1694, 1694, 1694,
 1694, 1694, 1694, 1694, 1694, 1694, 1694,    0,    0, 1694,
    0, 1694, 1694, 1694, 1694, 1694, 1694,    0,    0,    0,
    0,    0, 1685, 1687, 1687, 1685, 1687,    0, 1687,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1685, 1685, 1685, 1685,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1685,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1408,    0,    0,    0,    0,
    0,    0,    0,    0, 1409,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1410,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 2932,    0,    0,
    0, 1411,    0,    0,    0,    0,    0,    0, 1412,    0,
    0,    0,    0,    0,    0,    0, 1413,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1414,    0, 1415,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1416,    0, 1417, 1418, 1419,
    0,    0,    0,    0, 1420,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1421,    0,    0, 1685,    0,
 1685,    0,    0, 1685,    0, 1685, 1685, 1687,    0, 1685,
 1685,    0, 1685,    0,    0,    0,    0,    0, 1687,    0,
 1685, 1685, 1687, 1685,    0, 1685,    0,    0, 1685,    0,
    0, 1685,    0,    0,    0,    0,    0,    0, 1685,    0,
 1685, 1685,    0, 1685, 1685, 1685, 1685, 1685,    0,    0,
 1685,    0,    0, 1685,    0, 1685, 1685,    0,    0, 1685,
 1685, 1685,    0, 1685, 1685, 1685, 1685, 1685, 1685, 1685,
    0,    0,    0, 1685, 1685, 1685, 1685, 1685,    0,    0,
    0, 1685,    0,    0,    0,    0,    0,    0,    0, 1685,
 1685, 1685, 1685, 1685, 1685,    0,    0,    0, 1685, 1685,
    0, 1685,    0, 1685,    0, 1685, 1685, 1685,    0, 1685,
    0, 1685, 1685, 1685,    0, 1685, 1685, 1685, 1685, 1685,
    0, 1685, 1685,    0,    0,    0,    0, 1685,    0, 1685,
 1685,    0,    0, 1685, 1685, 1685,    0,    0, 1685,    0,
 1685,    0,    0, 1685, 1685, 1685, 1685,    0, 1685,    0,
 1685,    0, 1685, 1685, 1685, 1685, 1685,    0,    0, 1685,
 1685, 1685,    0, 1685, 1685,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1685, 1685, 1685,    0, 1685, 1685,
 1685,    0,    0,    0,    0, 1422, 1685,    0,    0,    0,
 1685,    0, 1685, 1685, 1685, 1685,    0, 1685,    0, 1685,
 1685, 1685, 1685, 1685, 1685, 1685, 1685,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1685, 1685,    0, 1685,
 1685, 1685,    0,    0,    0,    0, 1685, 1685, 1685, 1685,
 1685,    0, 1685,    0,    0, 1685, 1685,    0,    0,    0,
 1685, 1685,    0,    0, 1685, 1685, 1685, 1685, 1685, 1685,
 1685, 1685, 1685,    0, 1685,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1685, 1685, 1685, 1685, 1685, 1685,
 1685, 1685, 1685, 1685, 1685, 1685, 1685, 1685, 1685, 1685,
 1685, 1685, 1685, 1685, 1685,    0, 1685, 1685, 1685, 1685,
 1685, 1685,    0, 1685, 1685, 1685, 1685,    0, 1685,    0,
 1685, 1685, 1685, 1685, 1685, 1685, 1685, 1685, 1685, 1685,
 1685, 1685, 1685, 1685, 1685, 1685, 1685, 1685,    0, 1685,
 1685, 1685, 1685,    0, 1685, 1685, 1685, 1685, 1685,    0,
 1685, 1685, 1685, 1685, 1685, 1685, 1685, 1685, 1685, 1685,
 1685, 1685, 1685, 1685, 1685, 1685, 1685, 1685, 1685, 1685,
    0, 1685, 1685, 1685, 1685, 1685, 1685, 1685, 1685,    0,
 1685, 1685,    0,    0,    0, 1685, 1685, 1685, 1685, 1685,
 1685,    0,    0, 1685, 1685, 1685, 1685, 1685, 1685, 1685,
 1685, 1685,    0,    0, 1424, 1423, 1425, 1685,    0, 1685,
 1685, 1685, 1685, 1685, 1685,    0, 1685,    0,    0,    0,
    0,    0, 1424, 1423, 1425,    0,    0,    0,    0,    0,
    0,    0,    0, 1685, 1685,    0, 1685, 1685, 1685, 1685,
 1685, 1685, 1685, 1685, 1685,    0, 1685, 1685, 1685, 1685,
 1685, 1685, 1685, 1685, 1685, 1685, 1685, 1685,    0, 1685,
 1685, 1685,    0, 1685,    0,    0,    0, 1685, 1685, 1685,
 1685, 1685, 1685, 1685, 1685, 1685, 1685, 1685, 1685, 1685,
 1560, 1685, 1685,    0,    0,    0,    0,    0, 1685, 1685,
 1685, 1685, 1685, 1685, 1685, 1685, 1685, 1685, 1685, 1685,
 1685, 1685, 1685, 1685, 1685, 1685,    0,    0, 1685,    0,
 1685, 1685, 1685, 1685, 1685, 1685,    0,    0,    0,    0,
    0, 1560, 1688, 1688, 1560, 1688,    0, 1688,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1560,
 1560, 1560, 1560,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1560,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1408,    0,    0,    0,
    0,    0,    0,    0,    0, 1409,    0,    0,    0,    0,
    0,    0,    0,    0, 1408,    0,    0,    0, 1410,    0,
    0,    0,    0, 1409,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1410,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1411,    0,    0,    0,    0,    0,    0, 1412,
    0,    0,    0,    0,    0,    0,    0, 1413,    0,    0,
 1411,    0,    0,    0,    0,    0,    0, 1412,    0,    0,
    0,    0,    0,    0, 1414, 1413, 1415,    0, 2954,    0,
    0,    0,    0,    0,    0,    0, 1416,    0, 1417, 1418,
 1419,    0, 1414,    0, 1415, 1420,    0,    0,    0,    0,
    0,    0,    0,    0, 1416,    0, 1417, 1418, 1419,    0,
    0,    0,    0, 1420,    0,    0,    0, 1560,    0, 1560,
    0,    0, 1560,    0, 1560, 1560, 1688,    0, 1560, 1560,
    0, 1560,    0,    0,    0,    0, 1421, 1688,    0, 1560,
 1560, 1688, 1560,    0, 1560,    0,    0, 1560,    0,    0,
 1560,    0,    0,    0, 1421,    0,    0, 1560,    0, 1560,
 1560,    0, 1560, 1560, 1560, 1560, 1560,    0,    0, 1560,
    0,    0, 1560,    0, 1560, 1560,    0,    0, 1560, 1560,
 1560,    0, 1560, 1560, 1560, 1560, 1560, 1560, 1560,    0,
    0,    0, 1560, 1560, 1560, 1560, 1560,    0,    0,    0,
 1560,    0,    0,    0,    0,    0,    0,    0, 1560, 1560,
 1560, 1560, 1560, 1560,    0,    0,    0, 1560, 1560,    0,
 1560,    0, 1560,    0, 1560, 1560, 1560,    0, 1560,    0,
 1560, 1560, 1560,    0, 1560, 1560, 1560, 1560, 1560,    0,
 1560, 1560,    0,    0,    0,    0, 1560,    0, 1560, 1560,
    0,    0, 1560, 1560, 1560,    0,    0, 1560,    0, 1560,
    0,    0, 1560, 1560, 1560, 1560,    0, 1560,    0, 1560,
    0, 1560, 1560, 1560, 1560, 1560,    0,    0, 1560, 1560,
 1560, 2952, 1560, 1560,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1560, 1560, 1560,    0, 1560, 1560, 1560,
    0,    0,    0,    0,    0, 1560,    0,    0,    0, 1560,
    0, 1560, 1560, 1560, 1560,    0, 1560,    0, 1560, 1560,
 1560, 1560, 1560, 1560, 1560, 1560, 1422,    0,    0,    0,
    0,    0,    0,    0,    0, 1560, 1560,    0, 1560, 1560,
 1560,    0,    0,    0, 1422, 1560, 1560, 1560, 1560, 1560,
    0, 1560,    0,    0, 1560, 1560,    0,    0,    0, 1560,
 1560,    0,    0, 1560, 1560, 1560, 1560, 1560, 1560, 1560,
 1560, 1560,    0, 1560,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1560, 1560, 1560, 1560, 1560, 1560, 1560,
 1560, 1560, 1560, 1560, 1560, 1560, 1560, 1560, 1560, 1560,
 1560, 1560, 1560, 1560,    0, 1560, 1560, 1560, 1560, 1560,
 1560,    0, 1560, 1560, 1560, 1560,    0, 1560,    0, 1560,
 1560, 1560, 1560, 1560, 1560, 1560, 1560, 1560, 1560, 1560,
 1560, 1560, 1560, 1560, 1560, 1560, 1560,    0, 1560, 1560,
 1560, 1560,    0, 1560, 1560, 1560, 1560, 1560,    0, 1560,
 1560, 1560, 1560, 1560, 1560, 1560, 1560, 1560, 1560, 1560,
 1560, 1560, 1560, 1560, 1560, 1560, 1560, 1560, 1560,    0,
 1560, 1560, 1560, 1560, 1560, 1560, 1560, 1560,    0, 1560,
 1560, 1424, 1423, 1425, 1560, 1560, 1560, 1560, 1560, 1560,
    0,    0, 1560, 1560, 1560, 1560, 1560, 1560, 1560, 1560,
 1560,    0,    0,    0,    0,    0, 1560,    0, 1560, 1560,
 1560, 1560, 1560, 1560,    0, 1560,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1560, 1560,    0, 1560, 1560, 1560, 1560, 1560,
 1560, 1560, 1560, 1560,    0, 1560, 1560, 1560, 1560, 1560,
 1560, 1560, 1560, 1560, 1560, 1560, 1560,    0, 1560, 1560,
 1560,    0, 1560,    0,    0,    0, 1560, 1560, 1560, 1560,
 1560, 1560, 1560, 1560, 1560, 1560, 1560, 1560, 1560, 1708,
 1560, 1560,    0,    0,    0,    0,    0, 1560, 1560, 1560,
 1560, 1560, 1560, 1560, 1560, 1560, 1560, 1560, 1560, 1560,
 1560, 1560, 1560, 1560, 1560,    0,    0, 1560,    0, 1560,
 1560, 1560, 1560, 1560, 1560,    0,    0,    0,    0,    0,
 1708, 1708, 1708, 1708, 1708,    0, 1708,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1708, 1708,
 1708, 1708,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1708,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1408,    0,    0,    0,    0,    0,    0,
    0,    0, 1409,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1410,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1411,
    0,    0,    0,    0,    0,    0, 1412,    0,    0,    0,
    0,    0,    0,    0, 1413,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1414,    0, 1415,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1416,    0, 1417, 1418, 1419,    0,    0,
    0,    0, 1420,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1421,    0,    0, 1708,    0, 1708,    0,
    0, 1708,    0, 1708, 1708,    0,    0, 1708, 1708,    0,
 1708,    0,    0,    0,    0,    0,    0,    0, 1708, 1708,
 1708, 1708,    0, 1708,    0,    0, 1708,    0,    0, 1708,
    0,    0,    0,    0,    0,    0, 1708,    0, 1708, 1708,
    0, 1708, 1708, 1708, 1708, 1708,    0,    0, 1708,    0,
    0, 1708,    0, 1708, 1708,    0,    0, 1708, 1708, 1708,
    0, 1708, 1708, 1708, 1708, 1708, 1708, 1708,    0,    0,
    0, 1708, 1708, 1708, 1708, 1708,    0,    0,    0, 1708,
    0,    0,    0,    0,    0,    0,    0, 1708, 1708, 1708,
 1708, 1708, 1708,    0,    0,    0, 1708, 1708,    0, 1708,
    0, 1708,    0, 1708, 1708, 1708,    0, 1708,    0, 1708,
 1708, 1708,    0, 1708, 1708, 1708, 1708, 1708,    0, 1708,
 1708,    0,    0,    0,    0, 1708,    0, 1708, 1708,    0,
    0, 1708, 1708, 1708,    0,    0, 1708,    0, 1708,    0,
    0, 1708, 1708, 1708, 1708,    0, 1708,    0, 1708,    0,
 1708, 1708, 1708, 1708, 1708,    0,    0, 1708, 1708, 1708,
    0, 1708, 1708,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1708, 1708, 1708,    0, 1708, 1708, 1708,    0,
    0,    0,    0, 1422, 1708,    0,    0,    0, 1708,    0,
 1708, 1708, 1708, 1708,    0, 1708,    0, 1708, 1708, 1708,
 1708, 1708, 1708, 1708, 1708,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1708, 1708,    0, 1708, 1708, 1708,
    0,    0,    0,    0, 1708, 1708, 1708, 1708, 1708,    0,
 1708,    0,    0, 1708, 1708,    0,    0,    0, 1708, 1708,
    0,    0, 1708, 1708, 1708, 1708, 1708, 1708, 1708, 1708,
 1708,    0, 1708,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1708, 1708, 1708, 1708, 1708, 1708, 1708, 1708,
 1708, 1708, 1708, 1708, 1708, 1708, 1708, 1708, 1708, 1708,
 1708, 1708, 1708,    0, 1708, 1708, 1708, 1708, 1708, 1708,
    0, 1708, 1708, 1708, 1708,    0, 1708,    0, 1708, 1708,
 1708, 1708, 1708, 1708, 1708, 1708, 1708, 1708, 1708, 1708,
 1708, 1708, 1708, 1708, 1708, 1708,    0, 1708, 1708, 1708,
 1708,    0, 1708, 1708, 1708, 1708, 1708,    0, 1708, 1708,
 1708, 1708, 1708, 1708, 1708, 1708, 1708, 1708, 1708, 1708,
 1708, 1708, 1708, 1708, 1708, 1708, 1708, 1708,    0, 1708,
 1708, 1708, 1708, 1708, 1708, 1708, 1708,    0, 1708, 1708,
 1424, 1423, 1425, 1708, 1708, 1708, 1708, 1708, 1708,    0,
    0, 1708, 1708, 1708, 1708, 1708, 1708, 1708, 1708, 1708,
    0,    0,    0,    0,    0, 1708,    0, 1708, 1708, 1708,
 1708, 1708, 1708,    0, 1708,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1708, 1708,    0, 1708, 1708, 1708, 1708, 1708, 1708,
 1708, 1708, 1708,    0, 1708, 1708, 1708, 1708, 1708, 1708,
 1708, 1708, 1708, 1708, 1708, 1708,    0, 1708, 1708, 1708,
    0, 1708,    0,    0,    0, 1708, 1708, 1708, 1708, 1708,
 1708, 1708, 1708, 1708, 1708, 1708, 1708, 1708, 1706, 1708,
 1708,    0,    0,    0,    0,    0, 1708, 1708, 1708, 1708,
 1708, 1708, 1708, 1708, 1708, 1708, 1708, 1708, 1708, 1708,
 1708, 1708, 1708, 1708,    0,    0, 1708,    0, 1708, 1708,
 1708, 1708, 1708, 1708,    0,    0,    0,    0,    0, 1706,
 1706, 1706, 1706, 1706,    0, 1706,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1706, 1706, 1706,
 1706,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1706,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1408,    0,    0,    0,    0,    0,    0,    0,
    0, 1409,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1410,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1411,    0,
    0,    0,    0,    0,    0, 2120,    0,    0,    0,    0,
    0,    0,    0, 1413,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1414,    0, 1415,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1416,    0, 1417, 1418, 1419,    0,    0,    0,
    0, 1420,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1421,    0,    0, 1706,    0, 1706,    0,    0,
 1706,    0, 1706, 1706,    0,    0, 1706, 1706,    0, 1706,
    0,    0,    0,    0,    0,    0,    0, 1706, 1706,    0,
 1706,    0, 1706,    0,    0, 1706,    0,    0, 1706,    0,
    0,    0,    0,    0,    0, 1706,    0, 1706, 1706,    0,
 1706, 1706, 1706, 1706, 1706,    0,    0, 1706,    0,    0,
 1706,    0, 1706, 1706,    0,    0, 1706, 1706, 1706,    0,
 1706, 1706, 1706, 1706, 1706, 1706, 1706,    0,    0,    0,
 1706, 1706, 1706, 1706, 1706,    0,    0,    0, 1706,    0,
    0,    0,    0,    0,    0,    0, 1706, 1706, 1706, 1706,
 1706, 1706,    0,    0,    0, 1706, 1706,    0, 1706,    0,
 1706,    0, 1706, 1706, 1706,    0, 1706,    0, 1706, 1706,
 1706,    0, 1706, 1706, 1706, 1706, 1706,    0, 1706, 1706,
    0,    0,    0,    0, 1706,    0, 1706, 1706,    0,    0,
 1706, 1706, 1706,    0,    0, 1706,    0, 1706,    0,    0,
 1706, 1706, 1706, 1706,    0, 1706,    0, 1706,    0, 1706,
 1706, 1706, 1706, 1706,    0,    0, 1706, 1706, 1706,    0,
 1706, 1706,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1706, 1706, 1706,    0, 1706, 1706, 1706,    0,    0,
    0,    0, 1422, 1706,    0,    0,    0, 1706,    0, 1706,
 1706, 1706, 1706,    0, 1706,    0, 1706, 1706, 1706, 1706,
 1706, 1706, 1706, 1706,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1706, 1706,    0, 1706, 1706, 1706,    0,
    0,    0,    0, 1706, 1706, 1706, 1706, 1706,    0, 1706,
    0,    0, 1706, 1706,    0,    0,    0, 1706, 1706,    0,
    0, 1706, 1706, 1706, 1706, 1706, 1706, 1706, 1706, 1706,
    0, 1706,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1706, 1706, 1706, 1706, 1706, 1706, 1706, 1706, 1706,
 1706, 1706, 1706, 1706, 1706, 1706, 1706, 1706, 1706, 1706,
 1706, 1706,    0, 1706, 1706, 1706, 1706, 1706, 1706,    0,
 1706, 1706, 1706, 1706,    0, 1706,    0, 1706, 1706, 1706,
 1706, 1706, 1706, 1706, 1706, 1706, 1706, 1706, 1706, 1706,
 1706, 1706, 1706, 1706, 1706,    0, 1706, 1706, 1706, 1706,
    0, 1706, 1706, 1706, 1706, 1706,    0, 1706, 1706, 1706,
 1706, 1706, 1706, 1706, 1706, 1706, 1706, 1706, 1706, 1706,
 1706, 1706, 1706, 1706, 1706, 1706, 1706,    0, 1706, 1706,
 1706, 1706, 1706, 1706, 1706, 1706,    0, 1706, 1706, 1686,
 1686, 1686, 1706, 1706, 1706, 1706, 1706, 1706,    0,    0,
 1706, 1706, 1706, 1706, 1706, 1706, 1706, 1706, 1706,    0,
    0,    0,    0,    0, 1706,    0, 1706, 1706, 1706, 1706,
 1706, 1706,    0, 1706,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1706, 1706,    0, 1706, 1706, 1706, 1706, 1706, 1706, 1706,
 1706, 1706,    0, 1706, 1706, 1706, 1706, 1706, 1706, 1706,
 1706, 1706, 1706, 1706, 1706,    0, 1706, 1706, 1706,    0,
 1706,    0,    0,    0, 1706, 1706, 1706, 1706, 1706, 1706,
 1706, 1706, 1706, 1706, 1706, 1706, 1706, 1705, 1706, 1706,
    0,    0,    0,    0,    0, 1706, 1706, 1706, 1706, 1706,
 1706, 1706, 1706, 1706, 1706, 1706, 1706, 1706, 1706, 1706,
 1706, 1706, 1706,    0,    0, 1706,    0, 1706, 1706, 1706,
 1706, 1706, 1706,    0,    0,    0,    0,    0, 1705, 1705,
 1705, 1705, 1705,    0, 1705,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1705, 1705, 1705, 1705,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1705,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1686,    0,    0,    0,    0,    0,    0,    0,    0,
 1686,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1686,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1686,    0,    0,
    0,    0,    0,    0, 1686,    0,    0,    0,    0,    0,
    0,    0, 1686,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1686,
    0, 1686,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1686,    0, 1686, 1686, 1686,    0,    0,    0,    0,
 1686,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1686,    0,    0, 1705,    0, 1705,    0,    0, 1705,
    0, 1705, 1705,    0,    0, 1705, 1705,    0, 1705,    0,
    0,    0,    0,    0,    0,    0, 1705, 1705,    0, 1705,
    0, 1705,    0,    0, 1705,    0,    0, 1705,    0,    0,
    0,    0,    0,    0, 1705,    0, 1705, 1705,    0, 1705,
 1705, 1705, 1705, 1705,    0,    0, 1705,    0,    0, 1705,
    0, 1705, 1705,    0,    0, 1705, 1705, 1705,    0, 1705,
 1705, 1705, 1705, 1705, 1705, 1705,    0,    0,    0, 1705,
 1705, 1705, 1705, 1705,    0,    0,    0, 1705,    0,    0,
    0,    0,    0,    0,    0, 1705, 1705, 1705, 1705, 1705,
 1705,    0,    0,    0, 1705, 1705,    0, 1705,    0, 1705,
    0, 1705, 1705, 1705,    0, 1705,    0, 1705, 1705, 1705,
    0, 1705, 1705, 1705, 1705, 1705,    0, 1705, 1705,    0,
    0,    0,    0, 1705,    0, 1705, 1705,    0,    0, 1705,
 1705, 1705,    0,    0, 1705,    0, 1705,    0,    0, 1705,
 1705, 1705, 1705,    0, 1705,    0, 1705,    0, 1705, 1705,
 1705, 1705, 1705,    0,    0, 1705, 1705, 1705,    0, 1705,
 1705,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1705, 1705, 1705,    0, 1705, 1705, 1705,    0,    0,    0,
    0, 1686, 1705,    0,    0,    0, 1705,    0, 1705, 1705,
 1705, 1705,    0, 1705,    0, 1705, 1705, 1705, 1705, 1705,
 1705, 1705, 1705,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1705, 1705,    0, 1705, 1705, 1705,    0,    0,
    0,    0, 1705, 1705, 1705, 1705, 1705,    0, 1705,    0,
    0, 1705, 1705,    0,    0,    0, 1705, 1705,    0,    0,
 1705, 1705, 1705, 1705, 1705, 1705, 1705, 1705, 1705,    0,
 1705,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1705, 1705, 1705, 1705, 1705, 1705, 1705, 1705, 1705, 1705,
 1705, 1705, 1705, 1705, 1705, 1705, 1705, 1705, 1705, 1705,
 1705,    0, 1705, 1705, 1705, 1705, 1705, 1705,    0, 1705,
 1705, 1705, 1705,    0, 1705,    0, 1705, 1705, 1705, 1705,
 1705, 1705, 1705, 1705, 1705, 1705, 1705, 1705, 1705, 1705,
 1705, 1705, 1705, 1705,    0, 1705, 1705, 1705, 1705,    0,
 1705, 1705, 1705, 1705, 1705,    0, 1705, 1705, 1705, 1705,
 1705, 1705, 1705, 1705, 1705, 1705, 1705, 1705, 1705, 1705,
 1705, 1705, 1705, 1705, 1705, 1705,    0, 1705, 1705, 1705,
 1705, 1705, 1705, 1705, 1705,    0, 1705, 1705,    0,    0,
    0, 1705, 1705, 1705, 1705, 1705, 1705,    0,    0, 1705,
 1705, 1705, 1705, 1705, 1705, 1705, 1705, 1705,    0,    0,
    0,    0,    0, 1705,    0, 1705, 1705, 1705, 1705, 1705,
 1705,    0, 1705,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1705,
 1705,    0, 1705, 1705, 1705, 1705, 1705, 1705, 1705, 1705,
 1705,    0, 1705, 1705, 1705, 1705, 1705, 1705, 1705, 1705,
 1705, 1705, 1705, 1705,    0, 1705, 1705, 1705,    0, 1705,
    0,    0,    0, 1705, 1705, 1705, 1705, 1705, 1705, 1705,
 1705, 1705, 1705, 1705, 1705, 1705, 1713, 1705, 1705,    0,
    0,    0,    0,    0, 1705, 1705, 1705, 1705, 1705, 1705,
 1705, 1705, 1705, 1705, 1705, 1705, 1705, 1705, 1705, 1705,
 1705, 1705,    0,    0, 1705,    0, 1705, 1705, 1705, 1705,
 1705, 1705,    0,    0,    0,    0,    0, 1713, 1713, 1713,
 1713, 1713,    0, 1713,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1713, 1713, 1713, 1713,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1713,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1713,    0, 1713,    0,    0, 1713,    0,
 1713, 1713,    0,    0, 1713, 1713,    0, 1713,    0,    0,
    0,    0,    0,    0,    0, 1713, 1713,    0, 1713,    0,
 1713,    0,    0, 1713,    0,    0, 1713,    0,    0,    0,
    0,    0,    0, 1713,    0, 1713, 1713,    0, 1713, 1713,
 1713, 1713, 1713,    0,    0, 1713,    0,    0, 1713,    0,
 1713, 1713,    0,    0, 1713, 1713, 1713,    0, 1713, 1713,
 1713, 1713, 1713, 1713, 1713,    0,    0,    0, 1713, 1713,
 1713, 1713, 1713,    0,    0,    0, 1713,    0,    0,    0,
    0,    0,    0,    0, 1713, 1713, 1713, 1713, 1713, 1713,
    0,    0,    0, 1713, 1713,    0, 1713,    0, 1713,    0,
 1713, 1713, 1713,    0, 1713,    0, 1713, 1713, 1713,    0,
 1713, 1713, 1713, 1713, 1713,    0, 1713, 1713,    0,    0,
    0,    0, 1713,    0, 1713, 1713,    0,    0, 1713, 1713,
 1713,    0,    0, 1713,    0, 1713,    0,    0, 1713, 1713,
 1713, 1713,    0, 1713,    0, 1713,    0, 1713, 1713, 1713,
 1713, 1713,    0,    0, 1713, 1713, 1713,    0, 1713, 1713,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1713,
 1713, 1713,    0, 1713, 1713, 1713,    0,    0,    0,    0,
    0, 1713,    0,    0,    0, 1713,    0, 1713, 1713, 1713,
 1713,    0, 1713,    0, 1713, 1713, 1713, 1713, 1713, 1713,
 1713, 1713,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1713, 1713,    0, 1713, 1713, 1713,    0,    0,    0,
    0, 1713, 1713, 1713, 1713, 1713,    0, 1713,    0,    0,
 1713, 1713,    0,    0,    0, 1713, 1713,    0,    0, 1713,
 1713, 1713, 1713, 1713, 1713, 1713, 1713, 1713,    0, 1713,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1713,
 1713, 1713, 1713, 1713, 1713, 1713, 1713, 1713, 1713, 1713,
 1713, 1713, 1713, 1713, 1713, 1713, 1713, 1713, 1713, 1713,
    0, 1713, 1713, 1713, 1713, 1713, 1713,    0, 1713, 1713,
 1713, 1713,    0, 1713,    0, 1713, 1713, 1713, 1713, 1713,
 1713, 1713, 1713, 1713, 1713, 1713, 1713, 1713, 1713, 1713,
 1713, 1713, 1713,    0, 1713, 1713, 1713, 1713,    0, 1713,
 1713, 1713, 1713, 1713,    0, 1713, 1713, 1713, 1713, 1713,
 1713, 1713, 1713, 1713, 1713, 1713, 1713, 1713, 1713, 1713,
 1713, 1713, 1713, 1713, 1713,    0, 1713, 1713, 1713, 1713,
 1713, 1713, 1713, 1713,    0, 1713, 1713,    0,    0,    0,
 1713, 1713, 1713, 1713, 1713, 1713,    0,    0, 1713, 1713,
 1713, 1713, 1713, 1713, 1713, 1713, 1713,    0,    0,    0,
    0,    0, 1713,    0, 1713, 1713, 1713, 1713, 1713, 1713,
    0, 1713,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1713, 1713,
    0, 1713, 1713, 1713, 1713, 1713, 1713, 1713, 1713, 1713,
    0, 1713, 1713, 1713, 1713, 1713, 1713, 1713, 1713, 1713,
 1713, 1713, 1713,    0, 1713, 1713, 1713,    0, 1713,    0,
    0,    0, 1713, 1713, 1713, 1713, 1713, 1713, 1713, 1713,
 1713, 1713, 1713, 1713, 1713, 1714, 1713, 1713,    0,    0,
    0,    0,    0, 1713, 1713, 1713, 1713, 1713, 1713, 1713,
 1713, 1713, 1713, 1713, 1713, 1713, 1713, 1713, 1713, 1713,
 1713,    0,    0, 1713,    0, 1713, 1713, 1713, 1713, 1713,
 1713,    0,    0,    0,    0,    0, 1714, 1714, 1714, 1714,
 1714,    0, 1714,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1714, 1714, 1714, 1714,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1714,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1714,    0, 1714,    0,    0, 1714,    0, 1714,
 1714,    0,    0, 1714, 1714,    0, 1714,    0,    0,    0,
    0,    0,    0,    0, 1714, 1714,    0, 1714,    0, 1714,
    0,    0, 1714,    0,    0, 1714,    0,    0,    0,    0,
    0,    0, 1714,    0, 1714, 1714,    0, 1714, 1714, 1714,
 1714, 1714,    0,    0, 1714,    0,    0, 1714,    0, 1714,
 1714,    0,    0, 1714, 1714, 1714,    0, 1714, 1714, 1714,
 1714, 1714, 1714, 1714,    0,    0,    0, 1714, 1714, 1714,
 1714, 1714,    0,    0,    0, 1714,    0,    0,    0,    0,
    0,    0,    0, 1714, 1714, 1714, 1714, 1714, 1714,    0,
    0,    0, 1714, 1714,    0, 1714,    0, 1714,    0, 1714,
 1714, 1714,    0, 1714,    0, 1714, 1714, 1714,    0, 1714,
 1714, 1714, 1714, 1714,    0, 1714, 1714,    0,    0,    0,
    0, 1714,    0, 1714, 1714,    0,    0, 1714, 1714, 1714,
    0,    0, 1714,    0, 1714,    0,    0, 1714, 1714, 1714,
 1714,    0, 1714,    0, 1714,    0, 1714, 1714, 1714, 1714,
 1714,    0,    0, 1714, 1714, 1714,    0, 1714, 1714,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1714, 1714,
 1714,    0, 1714, 1714, 1714,    0,    0,    0,    0,    0,
 1714,    0,    0,    0, 1714,    0, 1714, 1714, 1714, 1714,
    0, 1714,    0, 1714, 1714, 1714, 1714, 1714, 1714, 1714,
 1714,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1714, 1714,    0, 1714, 1714, 1714,    0,    0,    0,    0,
 1714, 1714, 1714, 1714, 1714,    0, 1714,    0,    0, 1714,
 1714,    0,    0,    0, 1714, 1714,    0,    0, 1714, 1714,
 1714, 1714, 1714, 1714, 1714, 1714, 1714,    0, 1714,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1714, 1714,
 1714, 1714, 1714, 1714, 1714, 1714, 1714, 1714, 1714, 1714,
 1714, 1714, 1714, 1714, 1714, 1714, 1714, 1714, 1714,    0,
 1714, 1714, 1714, 1714, 1714, 1714,    0, 1714, 1714, 1714,
 1714,    0, 1714,    0, 1714, 1714, 1714, 1714, 1714, 1714,
 1714, 1714, 1714, 1714, 1714, 1714, 1714, 1714, 1714, 1714,
 1714, 1714,    0, 1714, 1714, 1714, 1714,    0, 1714, 1714,
 1714, 1714, 1714,    0, 1714, 1714, 1714, 1714, 1714, 1714,
 1714, 1714, 1714, 1714, 1714, 1714, 1714, 1714, 1714, 1714,
 1714, 1714, 1714, 1714,    0, 1714, 1714, 1714, 1714, 1714,
 1714, 1714, 1714,    0, 1714, 1714,    0,    0,    0, 1714,
 1714, 1714, 1714, 1714, 1714,    0,    0, 1714, 1714, 1714,
 1714, 1714, 1714, 1714, 1714, 1714,    0,    0,    0,    0,
    0, 1714,    0, 1714, 1714, 1714, 1714, 1714, 1714,    0,
 1714,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1714, 1714,    0,
 1714, 1714, 1714, 1714, 1714, 1714, 1714, 1714, 1714,    0,
 1714, 1714, 1714, 1714, 1714, 1714, 1714, 1714, 1714, 1714,
 1714, 1714,    0, 1714, 1714, 1714,    0, 1714,    0,    0,
    0, 1714, 1714, 1714, 1714, 1714, 1714, 1714, 1714, 1714,
 1714, 1714, 1714, 1714, 1707, 1714, 1714,    0,    0,    0,
    0,    0, 1714, 1714, 1714, 1714, 1714, 1714, 1714, 1714,
 1714, 1714, 1714, 1714, 1714, 1714, 1714, 1714, 1714, 1714,
    0,    0, 1714,    0, 1714, 1714, 1714, 1714, 1714, 1714,
    0,    0,    0,    0,    0, 1707,    0, 1707, 1707, 1707,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1707, 1707, 1707, 1707,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1707,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1707,    0, 1707,    0,    0, 1707,    0, 1707, 1707,
    0,    0, 1707, 1707,    0, 1707,    0,    0,    0,    0,
    0,    0,    0, 1707, 1707,    0, 1707,    0, 1707,    0,
    0, 1707,    0,    0, 1707,    0,    0,    0,    0,    0,
    0, 1707,    0, 1707, 1707,    0, 1707, 1707, 1707, 1707,
 1707,    0,    0, 1707,    0,    0, 1707,    0, 1707, 1707,
    0,    0, 1707, 1707, 1707,    0, 1707, 1707, 1707, 1707,
 1707, 1707, 1707,    0,    0,    0, 1707, 1707, 1707, 1707,
 1707,    0,    0,    0, 1707,    0,    0,    0,    0,    0,
    0,    0, 1707, 1707, 1707, 1707, 1707, 1707,    0,    0,
    0, 1707, 1707,    0, 1707,    0, 1707,    0, 1707, 1707,
 1707,    0, 1707,    0, 1707, 1707, 1707,    0, 1707, 1707,
 1707, 1707, 1707,    0, 1707, 1707,    0,    0,    0,    0,
 1707,    0, 1707, 1707,    0,    0, 1707, 1707, 1707,    0,
    0, 1707,    0, 1707,    0,    0, 1707, 1707, 1707, 1707,
    0, 1707,    0, 1707,    0, 1707, 1707, 1707, 1707, 1707,
    0,    0, 1707, 1707, 1707,    0, 1707, 1707,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1707, 1707, 1707,
    0, 1707, 1707, 1707,    0,    0,    0,    0,    0, 1707,
    0,    0,    0, 1707,    0, 1707, 1707, 1707, 1707,    0,
 1707,    0, 1707, 1707, 1707, 1707, 1707, 1707, 1707, 1707,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1707,
 1707,    0, 1707, 1707, 1707,    0,    0,    0,    0, 1707,
 1707, 1707, 1707, 1707,    0, 1707,    0,    0, 1707, 1707,
    0,    0,    0, 1707, 1707,    0,    0, 1707, 1707, 1707,
 1707, 1707, 1707, 1707, 1707, 1707,    0, 1707,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1707, 1707, 1707,
 1707, 1707, 1707, 1707, 1707, 1707, 1707, 1707, 1707, 1707,
 1707, 1707, 1707, 1707, 1707, 1707, 1707, 1707,    0, 1707,
 1707, 1707, 1707, 1707, 1707,    0, 1707, 1707, 1707, 1707,
    0, 1707,    0, 1707, 1707, 1707, 1707, 1707, 1707, 1707,
 1707, 1707, 1707, 1707, 1707, 1707, 1707, 1707, 1707, 1707,
 1707,    0, 1707, 1707, 1707, 1707,    0, 1707, 1707, 1707,
 1707, 1707,    0, 1707, 1707, 1707, 1707, 1707, 1707, 1707,
 1707, 1707, 1707, 1707, 1707, 1707, 1707, 1707, 1707, 1707,
 1707, 1707, 1707,    0, 1707, 1707, 1707, 1707, 1707, 1707,
 1707, 1707,    0, 1707, 1707,    0,    0,    0, 1707, 1707,
 1707, 1707, 1707, 1707,    0,    0, 1707, 1707, 1707, 1707,
 1707, 1707, 1707, 1707, 1707,    0,    0,    0,    0,    0,
 1707,    0, 1707, 1707, 1707, 1707, 1707, 1707,    0, 1707,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1707, 1707,    0, 1707,
 1707, 1707, 1707, 1707, 1707, 1707, 1707, 1707,    0, 1707,
 1707, 1707, 1707, 1707, 1707, 1707, 1707, 1707, 1707, 1707,
 1707,    0, 1707, 1707, 1707,    0, 1707,    0,    0,    0,
 1707, 1707, 1707, 1707, 1707, 1707, 1707, 1707, 1707, 1707,
 1707, 1707, 1707, 1712, 1707, 1707,    0,    0,    0,    0,
    0, 1707, 1707, 1707, 1707, 1707, 1707, 1707, 1707, 1707,
 1707, 1707, 1707, 1707, 1707, 1707, 1707, 1707, 1707,    0,
    0, 1707,    0, 1707, 1707, 1707, 1707, 1707, 1707,    0,
    0,    0,    0,    0, 1712,    0, 1712, 1712, 1712,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1712, 1712, 1712, 1712,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1712,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1712,    0, 1712,    0,    0, 1712,    0, 1712, 1712,    0,
    0, 1712, 1712,    0, 1712,    0,    0,    0,    0,    0,
    0,    0, 1712, 1712,    0, 1712,    0, 1712,    0,    0,
 1712,    0,    0, 1712,    0,    0,    0,    0,    0,    0,
 1712,    0, 1712, 1712,    0, 1712, 1712, 1712, 1712, 1712,
    0,    0, 1712,    0,    0, 1712,    0, 1712, 1712,    0,
    0, 1712, 1712, 1712,    0, 1712, 1712, 1712, 1712, 1712,
 1712, 1712,    0,    0,    0, 1712, 1712, 1712, 1712, 1712,
    0,    0,    0, 1712,    0,    0,    0,    0,    0,    0,
    0, 1712, 1712, 1712, 1712, 1712, 1712,    0,    0,    0,
 1712, 1712,    0, 1712,    0, 1712,    0, 1712, 1712, 1712,
    0, 1712,    0, 1712, 1712, 1712,    0, 1712, 1712, 1712,
 1712, 1712,    0, 1712, 1712,    0,    0,    0,    0, 1712,
    0, 1712, 1712,    0,    0, 1712, 1712, 1712,    0,    0,
 1712,    0, 1712,    0,    0, 1712, 1712, 1712, 1712,    0,
 1712,    0, 1712,    0, 1712, 1712, 1712, 1712, 1712,    0,
    0, 1712, 1712, 1712,    0, 1712, 1712,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1712, 1712, 1712,    0,
 1712, 1712, 1712,    0,    0,    0,    0,    0, 1712,    0,
    0,    0, 1712,    0, 1712, 1712, 1712, 1712,    0, 1712,
    0, 1712, 1712, 1712, 1712, 1712, 1712, 1712, 1712,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1712, 1712,
    0, 1712, 1712, 1712,    0,    0,    0,    0, 1712, 1712,
 1712, 1712, 1712,    0, 1712,    0,    0, 1712, 1712,    0,
    0,    0, 1712, 1712,    0,    0, 1712, 1712, 1712, 1712,
 1712, 1712, 1712, 1712, 1712,    0, 1712,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1712, 1712, 1712, 1712,
 1712, 1712, 1712, 1712, 1712, 1712, 1712, 1712, 1712, 1712,
 1712, 1712, 1712, 1712, 1712, 1712, 1712,    0, 1712, 1712,
 1712, 1712, 1712, 1712,    0, 1712, 1712, 1712, 1712,    0,
 1712,    0, 1712, 1712, 1712, 1712, 1712, 1712, 1712, 1712,
 1712, 1712, 1712, 1712, 1712, 1712, 1712, 1712, 1712, 1712,
    0, 1712, 1712, 1712, 1712,    0, 1712, 1712, 1712, 1712,
 1712,    0, 1712, 1712, 1712, 1712, 1712, 1712, 1712, 1712,
 1712, 1712, 1712, 1712, 1712, 1712, 1712, 1712, 1712, 1712,
 1712, 1712,    0, 1712, 1712, 1712, 1712, 1712, 1712, 1712,
 1712,    0, 1712, 1712,    0,    0,    0, 1712, 1712, 1712,
 1712, 1712, 1712,    0,    0, 1712, 1712, 1712, 1712, 1712,
 1712, 1712, 1712, 1712,    0,    0,    0,    0,    0, 1712,
    0, 1712, 1712, 1712, 1712, 1712, 1712,    0, 1712,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1712, 1712,    0, 1712, 1712,
 1712, 1712, 1712, 1712, 1712, 1712, 1712,    0, 1712, 1712,
 1712, 1712, 1712, 1712, 1712, 1712, 1712, 1712, 1712, 1712,
    0, 1712, 1712, 1712,    0, 1712,    0, 1590,    0, 1712,
 1712, 1712, 1712, 1712, 1712, 1712, 1712, 1712, 1712, 1712,
 1712, 1712,    0, 1712, 1712,    0,    0,    0,    0,    0,
 1712, 1712, 1712, 1712, 1712, 1712, 1712, 1712, 1712, 1712,
 1712, 1712, 1712, 1712, 1712, 1712, 1712, 1712, 1590,    0,
 1712, 1590, 1712, 1712, 1712, 1712, 1712, 1712,    0,    0,
    0,    0,    0,    0,    0,    0, 1590, 1590, 1590, 1590,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1590,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1590,    0, 1590,    0,    0, 1590,
    0, 1590, 1590,    0,    0, 1590, 1590,    0, 1590,    0,
    0,    0,    0,    0,    0,    0, 1590, 1590,    0, 1590,
    0, 1590,    0,    0, 1590,    0,    0, 1590,    0,    0,
    0,    0,    0,    0, 1590,    0, 1590, 1590,    0, 1590,
 1590, 1590, 1590, 1590,    0,    0, 1590,    0,    0, 1590,
    0, 1590, 1590,    0,    0, 1590, 1590, 1590,    0, 1590,
 1590, 1590, 1590, 1590, 1590, 1590,    0,    0,    0, 1590,
 1590, 1590, 1590, 1590,    0,    0,    0, 1590,    0,    0,
    0,    0,    0,    0,    0, 1590, 1590, 1590, 1590, 1590,
 1590,    0,    0,    0, 1590, 1590,    0, 1590,    0, 1590,
    0, 1590, 1590, 1590,    0, 1590,    0, 1590, 1590, 1590,
    0, 1590, 1590, 1590, 1590, 1590,    0, 1590, 1590,    0,
    0,    0,    0, 1590,    0, 1590, 1590,    0,    0, 1590,
 1590, 1590,    0,    0, 1590,    0, 1590,    0,    0, 1590,
 1590, 1590, 1590,    0, 1590,    0, 1590,    0, 1590, 1590,
 1590, 1590, 1590,    0,    0, 1590, 1590, 1590,    0, 1590,
 1590,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1590, 1590, 1590,    0, 1590, 1590, 1590,    0,    0,    0,
    0,    0, 1590,    0,    0,    0, 1590,    0, 1590, 1590,
 1590, 1590,    0, 1590,    0, 1590, 1590, 1590, 1590, 1590,
 1590, 1590, 1590,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1590, 1590,    0, 1590, 1590, 1590,    0,    0,
    0,    0, 1590, 1590, 1590, 1590, 1590,    0, 1590,    0,
    0, 1590, 1590,    0,    0,    0, 1590, 1590,    0,    0,
 1590, 1590, 1590, 1590, 1590, 1590, 1590, 1590, 1590,    0,
 1590,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1590, 1590, 1590, 1590, 1590, 1590, 1590, 1590, 1590, 1590,
 1590, 1590, 1590, 1590, 1590, 1590, 1590, 1590, 1590, 1590,
 1590,    0, 1590, 1590, 1590, 1590, 1590, 1590,    0, 1590,
 1590, 1590, 1590,    0, 1590,    0, 1590, 1590, 1590, 1590,
 1590, 1590, 1590, 1590, 1590, 1590, 1590, 1590, 1590, 1590,
 1590, 1590, 1590, 1590,    0, 1590, 1590, 1590, 1590,    0,
 1590, 1590, 1590, 1590, 1590,    0, 1590, 1590, 1590, 1590,
 1590, 1590, 1590, 1590, 1590, 1590, 1590, 1590, 1590, 1590,
 1590, 1590, 1590, 1590, 1590, 1590,    0, 1590, 1590, 1590,
 1590, 1590, 1590, 1590, 1590,    0, 1590, 1590,    0,    0,
    0, 1590, 1590, 1590, 1590, 1590, 1590,    0,    0, 1590,
 1590, 1590, 1590, 1590, 1590, 1590, 1590, 1590,    0,    0,
    0,    0,    0, 1590,    0, 1590, 1590, 1590, 1590, 1590,
 1590,    0, 1590,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1590,
 1590,    0, 1590, 1590, 1590, 1590, 1590, 1590, 1590, 1590,
 1590,    0, 1590, 1590, 1590, 1590, 1590, 1590, 1590, 1590,
 1590, 1590, 1590, 1590,    0, 1590, 1590, 1590,    0, 1590,
    0, 1591,    0, 1590, 1590, 1590, 1590, 1590, 1590, 1590,
 1590, 1590, 1590, 1590, 1590, 1590,    0, 1590, 1590,    0,
    0,    0,    0,    0, 1590, 1590, 1590, 1590, 1590, 1590,
 1590, 1590, 1590, 1590, 1590, 1590, 1590, 1590, 1590, 1590,
 1590, 1590, 1591,    0, 1590, 1591, 1590, 1590, 1590, 1590,
 1590, 1590,    0,    0,    0,    0,    0,    0,    0,    0,
 1591, 1591, 1591, 1591,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1591,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1591,    0,
 1591,    0,    0, 1591,    0, 1591, 1591,    0,    0, 1591,
 1591,    0, 1591,    0,    0,    0,    0,    0,    0,    0,
 1591, 1591,    0, 1591,    0, 1591,    0,    0, 1591,    0,
    0, 1591,    0,    0,    0,    0,    0,    0, 1591,    0,
 1591, 1591,    0, 1591, 1591, 1591, 1591, 1591,    0,    0,
 1591,    0,    0, 1591,    0, 1591, 1591,    0,    0, 1591,
 1591, 1591,    0, 1591, 1591, 1591, 1591, 1591, 1591, 1591,
    0,    0,    0, 1591, 1591, 1591, 1591, 1591,    0,    0,
    0, 1591,    0,    0,    0,    0,    0,    0,    0, 1591,
 1591, 1591, 1591, 1591, 1591,    0,    0,    0, 1591, 1591,
    0, 1591,    0, 1591,    0, 1591, 1591, 1591,    0, 1591,
    0, 1591, 1591, 1591,    0, 1591, 1591, 1591, 1591, 1591,
    0, 1591, 1591,    0,    0,    0,    0, 1591,    0, 1591,
 1591,    0,    0, 1591, 1591, 1591,    0,    0, 1591,    0,
 1591,    0,    0, 1591, 1591, 1591, 1591,    0, 1591,    0,
 1591,    0, 1591, 1591, 1591, 1591, 1591,    0,    0, 1591,
 1591, 1591,    0, 1591, 1591,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1591, 1591, 1591,    0, 1591, 1591,
 1591,    0,    0,    0,    0,    0, 1591,    0,    0,    0,
 1591,    0, 1591, 1591, 1591, 1591,    0, 1591,    0, 1591,
 1591, 1591, 1591, 1591, 1591, 1591, 1591,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1591, 1591,    0, 1591,
 1591, 1591,    0,    0,    0,    0, 1591, 1591, 1591, 1591,
 1591,    0, 1591,    0,    0, 1591, 1591,    0,    0,    0,
 1591, 1591,    0,    0, 1591, 1591, 1591, 1591, 1591, 1591,
 1591, 1591, 1591,    0, 1591,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1591, 1591, 1591, 1591, 1591, 1591,
 1591, 1591, 1591, 1591, 1591, 1591, 1591, 1591, 1591, 1591,
 1591, 1591, 1591, 1591, 1591,    0, 1591, 1591, 1591, 1591,
 1591, 1591,    0, 1591, 1591, 1591, 1591,    0, 1591,    0,
 1591, 1591, 1591, 1591, 1591, 1591, 1591, 1591, 1591, 1591,
 1591, 1591, 1591, 1591, 1591, 1591, 1591, 1591,    0, 1591,
 1591, 1591, 1591,    0, 1591, 1591, 1591, 1591, 1591,    0,
 1591, 1591, 1591, 1591, 1591, 1591, 1591, 1591, 1591, 1591,
 1591, 1591, 1591, 1591, 1591, 1591, 1591, 1591, 1591, 1591,
    0, 1591, 1591, 1591, 1591, 1591, 1591, 1591, 1591,    0,
 1591, 1591,    0,    0,    0, 1591, 1591, 1591, 1591, 1591,
 1591,    0,    0, 1591, 1591, 1591, 1591, 1591, 1591, 1591,
 1591, 1591,    0,    0,    0,    0,    0, 1591,    0, 1591,
 1591, 1591, 1591, 1591, 1591,    0, 1591,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1591, 1591,    0, 1591, 1591, 1591, 1591,
 1591, 1591, 1591, 1591, 1591,    0, 1591, 1591, 1591, 1591,
 1591, 1591, 1591, 1591, 1591, 1591, 1591, 1591,    0, 1591,
 1591, 1591,    0, 1591,    0, 1575,    0, 1591, 1591, 1591,
 1591, 1591, 1591, 1591, 1591, 1591, 1591, 1591, 1591, 1591,
    0, 1591, 1591,    0,    0,    0,    0,    0, 1591, 1591,
 1591, 1591, 1591, 1591, 1591, 1591, 1591, 1591, 1591, 1591,
 1591, 1591, 1591, 1591, 1591, 1591, 1575,    0, 1591, 1575,
 1591, 1591, 1591, 1591, 1591, 1591,    0,    0,    0,    0,
    0,    0,    0,    0, 1575, 1575, 1575, 1575,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1575,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1575,    0, 1575,    0,    0, 1575,    0, 1575,
 1575,    0,    0, 1575, 1575,    0, 1575,    0,    0,    0,
    0,    0,    0,    0, 1575, 1575,    0, 1575,    0, 1575,
    0,    0, 1575,    0,    0, 1575,    0,    0,    0,    0,
    0,    0, 1575,    0, 1575, 1575,    0, 1575, 1575, 1575,
 1575, 1575,    0,    0, 1575,    0,    0, 1575,    0, 1575,
 1575,    0,    0, 1575, 1575, 1575,    0, 1575, 1575, 1575,
 1575, 1575, 1575, 1575,    0,    0,    0, 1575,    0, 1575,
 1575, 1575,    0,    0,    0, 1575,    0,    0,    0,    0,
    0,    0,    0, 1575, 1575, 1575, 1575, 1575, 1575,    0,
    0,    0, 1575, 1575,    0, 1575,    0, 1575,    0, 1575,
 1575, 1575,    0, 1575,    0, 1575, 1575, 1575,    0, 1575,
 1575, 1575, 1575, 1575,    0, 1575, 1575,    0,    0,    0,
    0, 1575,    0, 1575, 1575,    0,    0, 1575, 1575, 1575,
    0,    0, 1575,    0, 1575,    0,    0, 1575, 1575, 1575,
 1575,    0, 1575,    0, 1575,    0, 1575, 1575, 1575, 1575,
 1575,    0,    0, 1575, 1575, 1575,    0, 1575, 1575,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1575, 1575,
 1575,    0, 1575, 1575, 1575,    0,    0,    0,    0,    0,
 1575,    0,    0,    0, 1575,    0, 1575, 1575, 1575, 1575,
    0, 1575,    0, 1575, 1575, 1575, 1575, 1575, 1575, 1575,
 1575,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1575, 1575,    0, 1575, 1575, 1575,    0,    0,    0,    0,
 1575, 1575, 1575, 1575, 1575,    0, 1575,    0,    0, 1575,
 1575,    0,    0,    0, 1575, 1575,    0,    0, 1575, 1575,
 1575, 1575, 1575, 1575, 1575, 1575, 1575,    0, 1575,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1575, 1575,
 1575, 1575, 1575, 1575, 1575, 1575, 1575, 1575, 1575, 1575,
 1575, 1575, 1575, 1575, 1575, 1575, 1575, 1575, 1575,    0,
 1575, 1575, 1575, 1575, 1575, 1575,    0, 1575, 1575, 1575,
 1575,    0, 1575,    0, 1575, 1575, 1575, 1575, 1575, 1575,
 1575, 1575, 1575, 1575, 1575, 1575, 1575, 1575, 1575, 1575,
 1575, 1575,    0, 1575, 1575, 1575, 1575,    0, 1575, 1575,
 1575, 1575, 1575,    0, 1575, 1575, 1575, 1575, 1575, 1575,
 1575, 1575, 1575, 1575, 1575, 1575, 1575, 1575, 1575, 1575,
 1575, 1575, 1575, 1575,    0, 1575, 1575, 1575, 1575, 1575,
 1575, 1575, 1575,    0, 1575, 1575,    0,    0,    0, 1575,
 1575, 1575, 1575, 1575, 1575,    0,    0, 1575, 1575, 1575,
 1575, 1575, 1575, 1575, 1575, 1575,    0,    0,    0,    0,
    0, 1575,    0, 1575, 1575, 1575, 1575, 1575, 1575,    0,
 1575,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1575, 1575,    0,
 1575, 1575, 1575, 1575, 1575, 1575, 1575, 1575, 1575,    0,
 1575, 1575, 1575, 1575, 1575, 1575, 1575, 1575, 1575, 1575,
 1575, 1575,    0, 1575, 1575, 1575,    0, 1575,    0, 1592,
    0, 1575, 1575, 1575, 1575, 1575, 1575, 1575, 1575, 1575,
 1575, 1575, 1575, 1575,    0, 1575, 1575,    0,    0,    0,
    0,    0, 1575, 1575, 1575, 1575, 1575, 1575, 1575, 1575,
 1575, 1575, 1575, 1575, 1575, 1575, 1575, 1575, 1575, 1575,
 1592,    0, 1575, 1592, 1575, 1575, 1575, 1575, 1575, 1575,
    0,    0,    0,    0,    0,    0,    0,    0, 1592, 1592,
 1592, 1592,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1592,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1592,    0, 1592,    0,
    0, 1592,    0, 1592, 1592,    0,    0, 1592, 1592,    0,
 1592,    0,    0,    0,    0,    0,    0,    0, 1592, 1592,
    0, 1592,    0, 1592,    0,    0, 1592,    0,    0, 1592,
    0,    0,    0,    0,    0,    0, 1592,    0, 1592, 1592,
    0, 1592, 1592, 1592, 1592, 1592,    0,    0, 1592,    0,
    0, 1592,    0, 1592, 1592,    0,    0, 1592, 1592, 1592,
    0, 1592, 1592, 1592, 1592, 1592, 1592, 1592,    0,    0,
    0, 1592,    0, 1592, 1592, 1592,    0,    0,    0, 1592,
    0,    0,    0,    0,    0,    0,    0, 1592, 1592, 1592,
 1592, 1592, 1592,    0,    0,    0, 1592, 1592,    0, 1592,
    0, 1592,    0, 1592, 1592, 1592,    0, 1592,    0, 1592,
 1592, 1592,    0, 1592, 1592, 1592, 1592, 1592,    0, 1592,
 1592,    0,    0,    0,    0, 1592,    0, 1592, 1592,    0,
    0, 1592, 1592, 1592,    0,    0, 1592,    0, 1592,    0,
    0, 1592, 1592, 1592, 1592,    0, 1592,    0, 1592,    0,
 1592, 1592, 1592, 1592, 1592,    0,    0, 1592, 1592, 1592,
    0, 1592, 1592,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1592, 1592, 1592,    0, 1592, 1592, 1592,    0,
    0,    0,    0,    0, 1592,    0,    0,    0, 1592,    0,
 1592, 1592, 1592, 1592,    0, 1592,    0, 1592, 1592, 1592,
 1592, 1592, 1592, 1592, 1592,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1592, 1592,    0, 1592, 1592, 1592,
    0,    0,    0,    0, 1592, 1592, 1592, 1592, 1592,    0,
 1592,    0,    0, 1592, 1592,    0,    0,    0, 1592, 1592,
    0,    0, 1592, 1592, 1592, 1592, 1592, 1592, 1592, 1592,
 1592,    0, 1592,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1592, 1592, 1592, 1592, 1592, 1592, 1592, 1592,
 1592, 1592, 1592, 1592, 1592, 1592, 1592, 1592, 1592, 1592,
 1592, 1592, 1592,    0, 1592, 1592, 1592, 1592, 1592, 1592,
    0, 1592, 1592, 1592, 1592,    0, 1592,    0, 1592, 1592,
 1592, 1592, 1592, 1592, 1592, 1592, 1592, 1592, 1592, 1592,
 1592, 1592, 1592, 1592, 1592, 1592,    0, 1592, 1592, 1592,
 1592,    0, 1592, 1592, 1592, 1592, 1592,    0, 1592, 1592,
 1592, 1592, 1592, 1592, 1592, 1592, 1592, 1592, 1592, 1592,
 1592, 1592, 1592, 1592, 1592, 1592, 1592, 1592,    0, 1592,
 1592, 1592, 1592, 1592, 1592, 1592, 1592,    0, 1592, 1592,
    0,    0,    0, 1592, 1592, 1592, 1592, 1592, 1592,    0,
    0, 1592, 1592, 1592, 1592, 1592, 1592, 1592, 1592, 1592,
    0,    0,    0,    0,    0, 1592,    0, 1592, 1592, 1592,
 1592, 1592, 1592,    0, 1592,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1592, 1592,    0, 1592, 1592, 1592, 1592, 1592, 1592,
 1592, 1592, 1592,    0, 1592, 1592, 1592, 1592, 1592, 1592,
 1592, 1592, 1592, 1592, 1592, 1592,    0, 1592, 1592, 1592,
    0, 1592,    0, 1593,    0, 1592, 1592, 1592, 1592, 1592,
 1592, 1592, 1592, 1592, 1592, 1592, 1592, 1592,    0, 1592,
 1592,    0,    0,    0,    0,    0, 1592, 1592, 1592, 1592,
 1592, 1592, 1592, 1592, 1592, 1592, 1592, 1592, 1592, 1592,
 1592, 1592, 1592, 1592, 1593,    0, 1592, 1593, 1592, 1592,
 1592, 1592, 1592, 1592,    0,    0,    0,    0,    0,    0,
    0,    0, 1593, 1593, 1593, 1593,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1593,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1593,    0, 1593,    0,    0, 1593,    0, 1593, 1593,    0,
    0, 1593, 1593,    0, 1593,    0,    0,    0,    0,    0,
    0,    0, 1593, 1593,    0, 1593,    0, 1593,    0,    0,
 1593,    0,    0, 1593,    0,    0,    0,    0,    0,    0,
 1593,    0, 1593, 1593,    0, 1593, 1593, 1593, 1593, 1593,
    0,    0, 1593,    0,    0, 1593,    0, 1593, 1593,    0,
    0, 1593, 1593, 1593,    0, 1593, 1593, 1593, 1593, 1593,
 1593, 1593,    0,    0,    0, 1593,    0, 1593, 1593, 1593,
    0,    0,    0, 1593,    0,    0,    0,    0,    0,    0,
    0, 1593, 1593, 1593, 1593, 1593, 1593,    0,    0,    0,
 1593, 1593,    0, 1593,    0, 1593,    0, 1593, 1593, 1593,
    0, 1593,    0, 1593, 1593, 1593,    0, 1593, 1593, 1593,
 1593, 1593,    0, 1593, 1593,    0,    0,    0,    0, 1593,
    0, 1593, 1593,    0,    0, 1593, 1593, 1593,    0,    0,
 1593,    0, 1593,    0,    0, 1593, 1593, 1593, 1593,    0,
 1593,    0, 1593,    0, 1593, 1593, 1593, 1593, 1593,    0,
    0, 1593, 1593, 1593,    0, 1593, 1593,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1593, 1593, 1593,    0,
 1593, 1593, 1593,    0,    0,    0,    0,    0, 1593,    0,
    0,    0, 1593,    0, 1593, 1593, 1593, 1593,    0, 1593,
    0, 1593, 1593, 1593, 1593, 1593, 1593, 1593, 1593,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1593, 1593,
    0, 1593, 1593, 1593,    0,    0,    0,    0, 1593, 1593,
 1593, 1593, 1593,    0, 1593,    0,    0, 1593, 1593,    0,
    0,    0, 1593, 1593,    0,    0, 1593, 1593, 1593, 1593,
 1593, 1593, 1593, 1593, 1593,    0, 1593,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1593, 1593, 1593, 1593,
 1593, 1593, 1593, 1593, 1593, 1593, 1593, 1593, 1593, 1593,
 1593, 1593, 1593, 1593, 1593, 1593, 1593,    0, 1593, 1593,
 1593, 1593, 1593, 1593,    0, 1593, 1593, 1593, 1593,    0,
 1593,    0, 1593, 1593, 1593, 1593, 1593, 1593, 1593, 1593,
 1593, 1593, 1593, 1593, 1593, 1593, 1593, 1593, 1593, 1593,
    0, 1593, 1593, 1593, 1593,    0, 1593, 1593, 1593, 1593,
 1593,    0, 1593, 1593, 1593, 1593, 1593, 1593, 1593, 1593,
 1593, 1593, 1593, 1593, 1593, 1593, 1593, 1593, 1593, 1593,
 1593, 1593,    0, 1593, 1593, 1593, 1593, 1593, 1593, 1593,
 1593,    0, 1593, 1593,    0,    0,    0, 1593, 1593, 1593,
 1593, 1593, 1593,    0,    0, 1593, 1593, 1593, 1593, 1593,
 1593, 1593, 1593, 1593,    0,    0,    0,    0,    0, 1593,
    0, 1593, 1593, 1593, 1593, 1593, 1593,    0, 1593,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1593, 1593,    0, 1593, 1593,
 1593, 1593, 1593, 1593, 1593, 1593, 1593,    0, 1593, 1593,
 1593, 1593, 1593, 1593, 1593, 1593, 1593, 1593, 1593, 1593,
    0, 1593, 1593, 1593,    0, 1593,    0, 1598,    0, 1593,
 1593, 1593, 1593, 1593, 1593, 1593, 1593, 1593, 1593, 1593,
 1593, 1593,    0, 1593, 1593,    0,    0,    0,    0,    0,
 1593, 1593, 1593, 1593, 1593, 1593, 1593, 1593, 1593, 1593,
 1593, 1593, 1593, 1593, 1593, 1593, 1593, 1593, 1598,    0,
 1593, 1598, 1593, 1593, 1593, 1593, 1593, 1593,    0,    0,
    0,    0,    0,    0,    0,    0, 1598, 1598, 1598, 1598,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1598,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1598,    0, 1598,    0,    0, 1598,
    0, 1598, 1598,    0,    0, 1598, 1598,    0, 1598,    0,
    0,    0,    0,    0,    0,    0, 1598, 1598,    0, 1598,
    0, 1598,    0,    0, 1598,    0,    0, 1598,    0,    0,
    0,    0,    0,    0, 1598,    0, 1598, 1598,    0, 1598,
 1598, 1598, 1598, 1598,    0,    0, 1598,    0,    0, 1598,
    0, 1598, 1598,    0,    0, 1598, 1598, 1598,    0, 1598,
 1598, 1598, 1598, 1598, 1598, 1598,    0,    0,    0, 1598,
    0, 1598, 1598, 1598,    0,    0,    0, 1598,    0,    0,
    0,    0,    0,    0,    0, 1598, 1598, 1598, 1598, 1598,
 1598,    0,    0,    0, 1598, 1598,    0, 1598,    0, 1598,
    0, 1598, 1598, 1598,    0, 1598,    0, 1598, 1598, 1598,
    0, 1598, 1598, 1598, 1598, 1598,    0, 1598, 1598,    0,
    0,    0,    0, 1598,    0, 1598, 1598,    0,    0, 1598,
 1598, 1598,    0,    0, 1598,    0, 1598,    0,    0, 1598,
 1598, 1598, 1598,    0, 1598,    0, 1598,    0, 1598, 1598,
 1598, 1598, 1598,    0,    0, 1598, 1598, 1598,    0, 1598,
 1598,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1598, 1598, 1598,    0, 1598, 1598, 1598,    0,    0,    0,
    0,    0, 1598,    0,    0,    0, 1598,    0, 1598, 1598,
 1598, 1598,    0, 1598,    0, 1598, 1598, 1598, 1598, 1598,
 1598, 1598, 1598,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1598, 1598,    0, 1598, 1598, 1598,    0,    0,
    0,    0, 1598, 1598, 1598, 1598, 1598,    0, 1598,    0,
    0, 1598, 1598,    0,    0,    0, 1598, 1598,    0,    0,
 1598, 1598, 1598, 1598, 1598, 1598, 1598, 1598, 1598,    0,
 1598,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1598, 1598, 1598, 1598, 1598, 1598, 1598, 1598, 1598, 1598,
 1598, 1598, 1598, 1598, 1598, 1598, 1598, 1598, 1598, 1598,
 1598,    0, 1598, 1598, 1598, 1598, 1598, 1598,    0, 1598,
 1598, 1598, 1598,    0, 1598,    0, 1598, 1598, 1598, 1598,
 1598, 1598, 1598, 1598, 1598, 1598, 1598, 1598, 1598, 1598,
 1598, 1598, 1598, 1598,    0, 1598, 1598, 1598, 1598,    0,
 1598, 1598, 1598, 1598, 1598,    0, 1598, 1598, 1598, 1598,
 1598, 1598, 1598, 1598, 1598, 1598, 1598, 1598, 1598, 1598,
 1598, 1598, 1598, 1598, 1598, 1598,    0, 1598, 1598, 1598,
 1598, 1598, 1598, 1598, 1598,    0, 1598, 1598,    0,    0,
    0, 1598, 1598, 1598, 1598, 1598, 1598,    0,    0, 1598,
 1598, 1598, 1598, 1598, 1598, 1598, 1598, 1598,    0,    0,
    0,    0,    0, 1598,    0, 1598, 1598, 1598, 1598, 1598,
 1598,    0, 1598,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1598,
 1598,    0, 1598, 1598, 1598, 1598, 1598, 1598, 1598, 1598,
 1598,    0, 1598, 1598, 1598, 1598, 1598, 1598, 1598, 1598,
 1598, 1598, 1598, 1598,    0, 1598, 1598, 1598,    0, 1598,
    0, 1600,    0, 1598, 1598, 1598, 1598, 1598, 1598, 1598,
 1598, 1598, 1598, 1598, 1598, 1598,    0, 1598, 1598,    0,
    0,    0,    0,    0, 1598, 1598, 1598, 1598, 1598, 1598,
 1598, 1598, 1598, 1598, 1598, 1598, 1598, 1598, 1598, 1598,
 1598, 1598, 1600,    0, 1598, 1600, 1598, 1598, 1598, 1598,
 1598, 1598,    0,    0,    0,    0,    0,    0,    0,    0,
 1600, 1600, 1600, 1600,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1600,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1600,    0,
 1600,    0,    0, 1600,    0, 1600, 1600,    0,    0, 1600,
 1600,    0, 1600,    0,    0,    0,    0,    0,    0,    0,
 1600, 1600,    0, 1600,    0, 1600,    0,    0, 1600,    0,
    0, 1600,    0,    0,    0,    0,    0,    0, 1600,    0,
 1600, 1600,    0, 1600, 1600, 1600, 1600, 1600,    0,    0,
 1600,    0,    0, 1600,    0, 1600, 1600,    0,    0, 1600,
 1600, 1600,    0, 1600, 1600, 1600, 1600, 1600, 1600, 1600,
    0,    0,    0, 1600,    0, 1600, 1600, 1600,    0,    0,
    0, 1600,    0,    0,    0,    0,    0,    0,    0, 1600,
 1600, 1600, 1600, 1600, 1600,    0,    0,    0, 1600, 1600,
    0, 1600,    0, 1600,    0, 1600, 1600, 1600,    0, 1600,
    0, 1600, 1600, 1600,    0, 1600, 1600, 1600, 1600, 1600,
    0, 1600, 1600,    0,    0,    0,    0, 1600,    0, 1600,
 1600,    0,    0, 1600, 1600, 1600,    0,    0, 1600,    0,
 1600,    0,    0, 1600, 1600, 1600, 1600,    0, 1600,    0,
 1600,    0, 1600, 1600, 1600, 1600, 1600,    0,    0, 1600,
 1600, 1600,    0, 1600, 1600,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1600, 1600, 1600,    0, 1600, 1600,
 1600,    0,    0,    0,    0,    0, 1600,    0,    0,    0,
 1600,    0, 1600, 1600, 1600, 1600,    0, 1600,    0, 1600,
 1600, 1600, 1600, 1600, 1600, 1600, 1600,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1600, 1600,    0, 1600,
 1600, 1600,    0,    0,    0,    0, 1600, 1600, 1600, 1600,
 1600,    0, 1600,    0,    0, 1600, 1600,    0,    0,    0,
 1600, 1600,    0,    0, 1600, 1600, 1600, 1600, 1600, 1600,
 1600, 1600, 1600,    0, 1600,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1600, 1600, 1600, 1600, 1600, 1600,
 1600, 1600, 1600, 1600, 1600, 1600, 1600, 1600, 1600, 1600,
 1600, 1600, 1600, 1600, 1600,    0, 1600, 1600, 1600, 1600,
 1600, 1600,    0, 1600, 1600, 1600, 1600,    0, 1600,    0,
 1600, 1600, 1600, 1600, 1600, 1600, 1600, 1600, 1600, 1600,
 1600, 1600, 1600, 1600, 1600, 1600, 1600, 1600,    0, 1600,
 1600, 1600, 1600,    0, 1600, 1600, 1600, 1600, 1600,    0,
 1600, 1600, 1600, 1600, 1600, 1600, 1600, 1600, 1600, 1600,
 1600, 1600, 1600, 1600, 1600, 1600, 1600, 1600, 1600, 1600,
    0, 1600, 1600, 1600, 1600, 1600, 1600, 1600, 1600,    0,
 1600, 1600,    0,    0,    0, 1600, 1600, 1600, 1600, 1600,
 1600,    0,    0, 1600, 1600, 1600, 1600, 1600, 1600, 1600,
 1600, 1600,    0,    0,    0,    0,    0, 1600,    0, 1600,
 1600, 1600, 1600, 1600, 1600,    0, 1600,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1600, 1600,    0, 1600, 1600, 1600, 1600,
 1600, 1600, 1600, 1600, 1600,    0, 1600, 1600, 1600, 1600,
 1600, 1600, 1600, 1600, 1600, 1600, 1600, 1600,    0, 1600,
 1600, 1600,    0, 1600,    0, 1602,    0, 1600, 1600, 1600,
 1600, 1600, 1600, 1600, 1600, 1600, 1600, 1600, 1600, 1600,
    0, 1600, 1600,    0,    0,    0,    0,    0, 1600, 1600,
 1600, 1600, 1600, 1600, 1600, 1600, 1600, 1600, 1600, 1600,
 1600, 1600, 1600, 1600, 1600, 1600, 1602,    0, 1600, 1602,
 1600, 1600, 1600, 1600, 1600, 1600,    0,    0,    0,    0,
    0,    0,    0,    0, 1602, 1602, 1602, 1602,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1602,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1602,    0, 1602,    0,    0, 1602,    0, 1602,
 1602,    0,    0, 1602, 1602,    0, 1602,    0,    0,    0,
    0,    0,    0,    0, 1602, 1602,    0, 1602,    0, 1602,
    0,    0, 1602,    0,    0, 1602,    0,    0,    0,    0,
    0,    0, 1602,    0, 1602, 1602,    0, 1602, 1602, 1602,
 1602, 1602,    0,    0, 1602,    0,    0, 1602,    0, 1602,
 1602,    0,    0, 1602, 1602, 1602,    0, 1602, 1602, 1602,
 1602, 1602, 1602, 1602,    0,    0,    0, 1602,    0, 1602,
 1602, 1602,    0,    0,    0, 1602,    0,    0,    0,    0,
    0,    0,    0, 1602, 1602, 1602, 1602, 1602, 1602,    0,
    0,    0, 1602, 1602,    0, 1602,    0, 1602,    0, 1602,
 1602, 1602,    0, 1602,    0, 1602, 1602, 1602,    0, 1602,
 1602, 1602, 1602, 1602,    0, 1602, 1602,    0,    0,    0,
    0, 1602,    0, 1602, 1602,    0,    0, 1602, 1602, 1602,
    0,    0, 1602,    0, 1602,    0,    0, 1602, 1602, 1602,
 1602,    0, 1602,    0, 1602,    0, 1602, 1602, 1602, 1602,
 1602,    0,    0, 1602, 1602, 1602,    0, 1602, 1602,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1602, 1602,
 1602,    0, 1602, 1602, 1602,    0,    0,    0,    0,    0,
 1602,    0,    0,    0, 1602,    0, 1602, 1602, 1602, 1602,
    0, 1602,    0, 1602, 1602, 1602, 1602, 1602, 1602, 1602,
 1602,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1602, 1602,    0, 1602, 1602, 1602,    0,    0,    0,    0,
 1602, 1602, 1602, 1602, 1602,    0, 1602,    0,    0, 1602,
 1602,    0,    0,    0, 1602, 1602,    0,    0, 1602, 1602,
 1602, 1602, 1602, 1602, 1602, 1602, 1602,    0, 1602,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1602, 1602,
 1602, 1602, 1602, 1602, 1602, 1602, 1602, 1602, 1602, 1602,
 1602, 1602, 1602, 1602, 1602, 1602, 1602, 1602, 1602,    0,
 1602, 1602, 1602, 1602, 1602, 1602,    0, 1602, 1602, 1602,
 1602,    0, 1602,    0, 1602, 1602, 1602, 1602, 1602, 1602,
 1602, 1602, 1602, 1602, 1602, 1602, 1602, 1602, 1602, 1602,
 1602, 1602,    0, 1602, 1602, 1602, 1602,    0, 1602, 1602,
 1602, 1602, 1602,    0, 1602, 1602, 1602, 1602, 1602, 1602,
 1602, 1602, 1602, 1602, 1602, 1602, 1602, 1602, 1602, 1602,
 1602, 1602, 1602, 1602,    0, 1602, 1602, 1602, 1602, 1602,
 1602, 1602, 1602,    0, 1602, 1602,    0,    0,    0, 1602,
 1602, 1602, 1602, 1602, 1602,    0,    0, 1602, 1602, 1602,
 1602, 1602, 1602, 1602, 1602, 1602,    0,    0,    0,    0,
    0, 1602,    0, 1602, 1602, 1602, 1602, 1602, 1602,    0,
 1602,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1602, 1602,    0,
 1602, 1602, 1602, 1602, 1602, 1602, 1602, 1602, 1602,    0,
 1602, 1602, 1602, 1602, 1602, 1602, 1602, 1602, 1602, 1602,
 1602, 1602,    0, 1602, 1602, 1602,    0, 1602,    0, 1604,
    0, 1602, 1602, 1602, 1602, 1602, 1602, 1602, 1602, 1602,
 1602, 1602, 1602, 1602,    0, 1602, 1602,    0,    0,    0,
    0,    0, 1602, 1602, 1602, 1602, 1602, 1602, 1602, 1602,
 1602, 1602, 1602, 1602, 1602, 1602, 1602, 1602, 1602, 1602,
 1604,    0, 1602, 1604, 1602, 1602, 1602, 1602, 1602, 1602,
    0,    0,    0,    0,    0,    0,    0,    0, 1604, 1604,
 1604, 1604,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1604,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1604,    0, 1604,    0,
    0, 1604,    0, 1604, 1604,    0,    0, 1604, 1604,    0,
 1604,    0,    0,    0,    0,    0,    0,    0, 1604, 1604,
    0, 1604,    0, 1604,    0,    0, 1604,    0,    0, 1604,
    0,    0,    0,    0,    0,    0, 1604,    0, 1604, 1604,
    0, 1604, 1604, 1604, 1604, 1604,    0,    0, 1604,    0,
    0, 1604,    0, 1604, 1604,    0,    0, 1604, 1604, 1604,
    0, 1604, 1604, 1604, 1604, 1604, 1604, 1604,    0,    0,
    0, 1604,    0, 1604, 1604, 1604,    0,    0,    0, 1604,
    0,    0,    0,    0,    0,    0,    0, 1604, 1604, 1604,
 1604, 1604, 1604,    0,    0,    0, 1604, 1604,    0, 1604,
    0, 1604,    0, 1604, 1604, 1604,    0, 1604,    0, 1604,
 1604, 1604,    0, 1604, 1604, 1604, 1604, 1604,    0, 1604,
 1604,    0,    0,    0,    0, 1604,    0, 1604, 1604,    0,
    0, 1604, 1604, 1604,    0,    0, 1604,    0, 1604,    0,
    0, 1604, 1604, 1604, 1604,    0, 1604,    0, 1604,    0,
 1604, 1604, 1604, 1604, 1604,    0,    0, 1604, 1604, 1604,
    0, 1604, 1604,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1604, 1604, 1604,    0, 1604, 1604, 1604,    0,
    0,    0,    0,    0, 1604,    0,    0,    0, 1604,    0,
 1604, 1604, 1604, 1604,    0, 1604,    0, 1604, 1604, 1604,
 1604, 1604, 1604, 1604, 1604,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1604, 1604,    0, 1604, 1604, 1604,
    0,    0,    0,    0, 1604, 1604, 1604, 1604, 1604,    0,
 1604,    0,    0, 1604, 1604,    0,    0,    0, 1604, 1604,
    0,    0, 1604, 1604, 1604, 1604, 1604, 1604, 1604, 1604,
 1604,    0, 1604,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1604, 1604, 1604, 1604, 1604, 1604, 1604, 1604,
 1604, 1604, 1604, 1604, 1604, 1604, 1604, 1604, 1604, 1604,
 1604, 1604, 1604,    0, 1604, 1604, 1604, 1604, 1604, 1604,
    0, 1604, 1604, 1604, 1604,    0, 1604,    0, 1604, 1604,
 1604, 1604, 1604, 1604, 1604, 1604, 1604, 1604, 1604, 1604,
 1604, 1604, 1604, 1604, 1604, 1604,    0, 1604, 1604, 1604,
 1604,    0, 1604, 1604, 1604, 1604, 1604,    0, 1604, 1604,
 1604, 1604, 1604, 1604, 1604, 1604, 1604, 1604, 1604, 1604,
 1604, 1604, 1604, 1604, 1604, 1604, 1604, 1604,    0, 1604,
 1604, 1604, 1604, 1604, 1604, 1604, 1604,    0, 1604, 1604,
    0,    0,    0, 1604, 1604, 1604, 1604, 1604, 1604,    0,
    0, 1604, 1604, 1604, 1604, 1604, 1604, 1604, 1604, 1604,
    0,    0,    0,    0,    0, 1604,    0, 1604, 1604, 1604,
 1604, 1604, 1604,    0, 1604,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1604, 1604,    0, 1604, 1604, 1604, 1604, 1604, 1604,
 1604, 1604, 1604,    0, 1604, 1604, 1604, 1604, 1604, 1604,
 1604, 1604, 1604, 1604, 1604, 1604,    0, 1604, 1604, 1604,
    0, 1604,    0, 1597,    0, 1604, 1604, 1604, 1604, 1604,
 1604, 1604, 1604, 1604, 1604, 1604, 1604, 1604,    0, 1604,
 1604,    0,    0,    0,    0,    0, 1604, 1604, 1604, 1604,
 1604, 1604, 1604, 1604, 1604, 1604, 1604, 1604, 1604, 1604,
 1604, 1604, 1604, 1604, 1597,    0, 1604, 1597, 1604, 1604,
 1604, 1604, 1604, 1604,    0,    0,    0,    0,    0,    0,
    0,    0, 1597, 1597, 1597, 1597,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1597,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1597,    0, 1597,    0,    0, 1597,    0, 1597, 1597,    0,
    0, 1597, 1597,    0, 1597,    0,    0,    0,    0,    0,
    0,    0, 1597, 1597,    0, 1597,    0, 1597,    0,    0,
 1597,    0,    0, 1597,    0,    0,    0,    0,    0,    0,
 1597,    0, 1597, 1597,    0, 1597, 1597, 1597,    0, 1597,
    0,    0, 1597,    0,    0, 1597,    0, 1597, 1597,    0,
    0, 1597, 1597, 1597,    0, 1597, 1597, 1597, 1597, 1597,
 1597, 1597,    0,    0,    0, 1597,    0, 1597, 1597, 1597,
    0,    0,    0, 1597,    0,    0,    0,    0,    0,    0,
    0, 1597, 1597, 1597, 1597, 1597, 1597,    0,    0,    0,
 1597, 1597,    0, 1597,    0, 1597,    0, 1597, 1597, 1597,
    0, 1597,    0, 1597, 1597, 1597,    0, 1597, 1597, 1597,
 1597, 1597,    0, 1597, 1597,    0,    0,    0,    0, 1597,
    0, 1597, 1597,    0,    0, 1597, 1597, 1597,    0,    0,
 1597,    0, 1597,    0,    0, 1597, 1597, 1597, 1597,    0,
 1597,    0, 1597,    0, 1597, 1597, 1597, 1597, 1597,    0,
    0, 1597, 1597, 1597,    0, 1597, 1597,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1597, 1597, 1597,    0,
 1597, 1597, 1597,    0,    0,    0,    0,    0, 1597,    0,
    0,    0, 1597,    0, 1597, 1597, 1597, 1597,    0, 1597,
    0, 1597, 1597, 1597, 1597, 1597, 1597, 1597, 1597,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1597, 1597,
    0, 1597, 1597, 1597,    0,    0,    0,    0, 1597, 1597,
 1597, 1597, 1597,    0, 1597,    0,    0, 1597, 1597,    0,
    0,    0, 1597, 1597,    0,    0, 1597, 1597, 1597, 1597,
 1597, 1597, 1597, 1597, 1597,    0, 1597,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1597, 1597, 1597, 1597,
 1597, 1597, 1597, 1597, 1597, 1597, 1597, 1597, 1597, 1597,
 1597, 1597, 1597, 1597, 1597, 1597, 1597,    0, 1597, 1597,
 1597, 1597, 1597, 1597,    0, 1597, 1597, 1597, 1597,    0,
 1597,    0, 1597, 1597, 1597, 1597, 1597, 1597, 1597, 1597,
 1597, 1597, 1597, 1597, 1597, 1597, 1597, 1597, 1597, 1597,
    0, 1597, 1597, 1597, 1597,    0, 1597, 1597, 1597, 1597,
 1597,    0, 1597, 1597, 1597, 1597, 1597, 1597, 1597, 1597,
 1597, 1597, 1597, 1597, 1597, 1597, 1597, 1597, 1597, 1597,
 1597, 1597,    0, 1597, 1597, 1597, 1597, 1597, 1597, 1597,
 1597,    0, 1597, 1597,    0,    0,    0, 1597, 1597, 1597,
 1597, 1597, 1597,    0,    0, 1597, 1597, 1597, 1597, 1597,
 1597, 1597, 1597, 1597,    0,    0,    0,    0,    0, 1597,
    0, 1597, 1597, 1597, 1597, 1597, 1597,    0, 1597,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1597, 1597,    0, 1597, 1597,
 1597, 1597, 1597, 1597, 1597, 1597, 1597,    0, 1597, 1597,
 1597, 1597, 1597, 1597, 1597, 1597, 1597, 1597, 1597, 1597,
    0, 1597, 1597, 1597,    0, 1597,    0, 1599,    0, 1597,
 1597, 1597, 1597, 1597, 1597, 1597, 1597, 1597, 1597, 1597,
 1597, 1597,    0, 1597, 1597,    0,    0,    0,    0,    0,
 1597, 1597, 1597, 1597, 1597, 1597, 1597, 1597, 1597, 1597,
 1597, 1597, 1597, 1597, 1597, 1597, 1597, 1597, 1599,    0,
 1597, 1599, 1597, 1597, 1597, 1597, 1597, 1597,    0,    0,
    0,    0,    0,    0,    0,    0, 1599, 1599, 1599, 1599,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1599,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1599,    0, 1599,    0,    0, 1599,
    0, 1599, 1599,    0,    0, 1599, 1599,    0, 1599,    0,
    0,    0,    0,    0,    0,    0, 1599, 1599,    0, 1599,
    0, 1599,    0,    0, 1599,    0,    0, 1599,    0,    0,
    0,    0,    0,    0, 1599,    0, 1599, 1599,    0, 1599,
 1599, 1599,    0, 1599,    0,    0, 1599,    0,    0, 1599,
    0, 1599, 1599,    0,    0, 1599, 1599, 1599,    0, 1599,
 1599, 1599, 1599, 1599, 1599, 1599,    0,    0,    0, 1599,
    0, 1599, 1599, 1599,    0,    0,    0, 1599,    0,    0,
    0,    0,    0,    0,    0, 1599, 1599, 1599, 1599, 1599,
 1599,    0,    0,    0, 1599, 1599,    0, 1599,    0, 1599,
    0, 1599, 1599, 1599,    0, 1599,    0, 1599, 1599, 1599,
    0, 1599, 1599, 1599, 1599, 1599,    0, 1599, 1599,    0,
    0,    0,    0, 1599,    0, 1599, 1599,    0,    0, 1599,
 1599, 1599,    0,    0, 1599,    0, 1599,    0,    0, 1599,
 1599, 1599, 1599,    0, 1599,    0, 1599,    0, 1599, 1599,
 1599, 1599, 1599,    0,    0, 1599, 1599, 1599,    0, 1599,
 1599,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1599, 1599, 1599,    0, 1599, 1599, 1599,    0,    0,    0,
    0,    0, 1599,    0,    0,    0, 1599,    0, 1599, 1599,
 1599, 1599,    0, 1599,    0, 1599, 1599, 1599, 1599, 1599,
 1599, 1599, 1599,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1599, 1599,    0, 1599, 1599, 1599,    0,    0,
    0,    0, 1599, 1599, 1599, 1599, 1599,    0, 1599,    0,
    0, 1599, 1599,    0,    0,    0, 1599, 1599,    0,    0,
 1599, 1599, 1599, 1599, 1599, 1599, 1599, 1599, 1599,    0,
 1599,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1599, 1599, 1599, 1599, 1599, 1599, 1599, 1599, 1599, 1599,
 1599, 1599, 1599, 1599, 1599, 1599, 1599, 1599, 1599, 1599,
 1599,    0, 1599, 1599, 1599, 1599, 1599, 1599,    0, 1599,
 1599, 1599, 1599,    0, 1599,    0, 1599, 1599, 1599, 1599,
 1599, 1599, 1599, 1599, 1599, 1599, 1599, 1599, 1599, 1599,
 1599, 1599, 1599, 1599,    0, 1599, 1599, 1599, 1599,    0,
 1599, 1599, 1599, 1599, 1599,    0, 1599, 1599, 1599, 1599,
 1599, 1599, 1599, 1599, 1599, 1599, 1599, 1599, 1599, 1599,
 1599, 1599, 1599, 1599, 1599, 1599,    0, 1599, 1599, 1599,
 1599, 1599, 1599, 1599, 1599,    0, 1599, 1599,    0,    0,
    0, 1599, 1599, 1599, 1599, 1599, 1599,    0,    0, 1599,
 1599, 1599, 1599, 1599, 1599, 1599, 1599, 1599,    0,    0,
    0,    0,    0, 1599,    0, 1599, 1599, 1599, 1599, 1599,
 1599,    0, 1599,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1599,
 1599,    0, 1599, 1599, 1599, 1599, 1599, 1599, 1599, 1599,
 1599,    0, 1599, 1599, 1599, 1599, 1599, 1599, 1599, 1599,
 1599, 1599, 1599, 1599,    0, 1599, 1599, 1599,    0, 1599,
    0, 1601,    0, 1599, 1599, 1599, 1599, 1599, 1599, 1599,
 1599, 1599, 1599, 1599, 1599, 1599,    0, 1599, 1599,    0,
    0,    0,    0,    0, 1599, 1599, 1599, 1599, 1599, 1599,
 1599, 1599, 1599, 1599, 1599, 1599, 1599, 1599, 1599, 1599,
 1599, 1599, 1601,    0, 1599, 1601, 1599, 1599, 1599, 1599,
 1599, 1599,    0,    0,    0,    0,    0,    0,    0,    0,
 1601, 1601, 1601, 1601,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1601,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1601,    0,
 1601,    0,    0, 1601,    0, 1601, 1601,    0,    0, 1601,
 1601,    0, 1601,    0,    0,    0,    0,    0,    0,    0,
 1601, 1601,    0, 1601,    0, 1601,    0,    0, 1601,    0,
    0, 1601,    0,    0,    0,    0,    0,    0, 1601,    0,
 1601, 1601,    0, 1601, 1601, 1601,    0, 1601,    0,    0,
 1601,    0,    0, 1601,    0, 1601, 1601,    0,    0, 1601,
 1601, 1601,    0, 1601, 1601, 1601, 1601, 1601, 1601, 1601,
    0,    0,    0, 1601,    0, 1601, 1601, 1601,    0,    0,
    0, 1601,    0,    0,    0,    0,    0,    0,    0, 1601,
 1601, 1601, 1601, 1601, 1601,    0,    0,    0, 1601, 1601,
    0, 1601,    0, 1601,    0, 1601, 1601, 1601,    0, 1601,
    0, 1601, 1601, 1601,    0, 1601, 1601, 1601, 1601, 1601,
    0, 1601, 1601,    0,    0,    0,    0, 1601,    0, 1601,
 1601,    0,    0, 1601, 1601, 1601,    0,    0, 1601,    0,
 1601,    0,    0, 1601, 1601, 1601, 1601,    0, 1601,    0,
 1601,    0, 1601, 1601, 1601, 1601, 1601,    0,    0, 1601,
 1601, 1601,    0, 1601, 1601,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1601, 1601, 1601,    0, 1601, 1601,
 1601,    0,    0,    0,    0,    0, 1601,    0,    0,    0,
 1601,    0, 1601, 1601, 1601, 1601,    0, 1601,    0, 1601,
 1601, 1601, 1601, 1601, 1601, 1601, 1601,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1601, 1601,    0, 1601,
 1601, 1601,    0,    0,    0,    0, 1601, 1601, 1601, 1601,
 1601,    0, 1601,    0,    0, 1601, 1601,    0,    0,    0,
 1601, 1601,    0,    0, 1601, 1601, 1601, 1601, 1601, 1601,
 1601, 1601, 1601,    0, 1601,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1601, 1601, 1601, 1601, 1601, 1601,
 1601, 1601, 1601, 1601, 1601, 1601, 1601, 1601, 1601, 1601,
 1601, 1601, 1601, 1601, 1601,    0, 1601, 1601, 1601, 1601,
 1601, 1601,    0, 1601, 1601, 1601, 1601,    0, 1601,    0,
 1601, 1601, 1601, 1601, 1601, 1601, 1601, 1601, 1601, 1601,
 1601, 1601, 1601, 1601, 1601, 1601, 1601, 1601,    0, 1601,
 1601, 1601, 1601,    0, 1601, 1601, 1601, 1601, 1601,    0,
 1601, 1601, 1601, 1601, 1601, 1601, 1601, 1601, 1601, 1601,
 1601, 1601, 1601, 1601, 1601, 1601, 1601, 1601, 1601, 1601,
    0, 1601, 1601, 1601, 1601, 1601, 1601, 1601, 1601,    0,
 1601, 1601,    0,    0,    0, 1601, 1601, 1601, 1601, 1601,
 1601,    0,    0, 1601, 1601, 1601, 1601, 1601, 1601, 1601,
 1601, 1601,    0,    0,    0,    0,    0, 1601,    0, 1601,
 1601, 1601, 1601, 1601, 1601,    0, 1601,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1601, 1601,    0, 1601, 1601, 1601, 1601,
 1601, 1601, 1601, 1601, 1601,    0, 1601, 1601, 1601, 1601,
 1601, 1601, 1601, 1601, 1601, 1601, 1601, 1601,    0, 1601,
 1601, 1601,    0, 1601,    0, 1603,    0, 1601, 1601, 1601,
 1601, 1601, 1601, 1601, 1601, 1601, 1601, 1601, 1601, 1601,
    0, 1601, 1601,    0,    0,    0,    0,    0, 1601, 1601,
 1601, 1601, 1601, 1601, 1601, 1601, 1601, 1601, 1601, 1601,
 1601, 1601, 1601, 1601, 1601, 1601, 1603,    0, 1601, 1603,
 1601, 1601, 1601, 1601, 1601, 1601,    0,    0,    0,    0,
    0,    0,    0,    0, 1603, 1603, 1603, 1603,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1603,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1603,    0, 1603,    0,    0, 1603,    0, 1603,
 1603,    0,    0, 1603, 1603,    0, 1603,    0,    0,    0,
    0,    0,    0,    0, 1603, 1603,    0, 1603,    0, 1603,
    0,    0, 1603,    0,    0, 1603,    0,    0,    0,    0,
    0,    0, 1603,    0, 1603, 1603,    0, 1603, 1603, 1603,
    0, 1603,    0,    0, 1603,    0,    0, 1603,    0, 1603,
 1603,    0,    0, 1603, 1603, 1603,    0, 1603, 1603, 1603,
 1603, 1603, 1603, 1603,    0,    0,    0, 1603,    0, 1603,
 1603, 1603,    0,    0,    0, 1603,    0,    0,    0,    0,
    0,    0,    0, 1603, 1603, 1603, 1603, 1603, 1603,    0,
    0,    0, 1603, 1603,    0, 1603,    0, 1603,    0, 1603,
 1603, 1603,    0, 1603,    0, 1603, 1603, 1603,    0, 1603,
 1603, 1603, 1603, 1603,    0, 1603, 1603,    0,    0,    0,
    0, 1603,    0, 1603, 1603,    0,    0, 1603, 1603, 1603,
    0,    0, 1603,    0, 1603,    0,    0, 1603, 1603, 1603,
 1603,    0, 1603,    0, 1603,    0, 1603, 1603, 1603, 1603,
 1603,    0,    0, 1603, 1603, 1603,    0, 1603, 1603,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1603, 1603,
 1603,    0, 1603, 1603, 1603,    0,    0,    0,    0,    0,
 1603,    0,    0,    0, 1603,    0, 1603, 1603, 1603, 1603,
    0, 1603,    0, 1603, 1603, 1603, 1603, 1603, 1603, 1603,
 1603,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1603, 1603,    0, 1603, 1603, 1603,    0,    0,    0,    0,
 1603, 1603, 1603, 1603, 1603,    0, 1603,    0,    0, 1603,
 1603,    0,    0,    0, 1603, 1603,    0,    0, 1603, 1603,
 1603, 1603, 1603, 1603, 1603, 1603, 1603,    0, 1603,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1603, 1603,
 1603, 1603, 1603, 1603, 1603, 1603, 1603, 1603, 1603, 1603,
 1603, 1603, 1603, 1603, 1603, 1603, 1603, 1603, 1603,    0,
 1603, 1603, 1603, 1603, 1603, 1603,    0, 1603, 1603, 1603,
 1603,    0, 1603,    0, 1603, 1603, 1603, 1603, 1603, 1603,
 1603, 1603, 1603, 1603, 1603, 1603, 1603, 1603, 1603, 1603,
 1603, 1603,    0, 1603, 1603, 1603, 1603,    0, 1603, 1603,
 1603, 1603, 1603,    0, 1603, 1603, 1603, 1603, 1603, 1603,
 1603, 1603, 1603, 1603, 1603, 1603, 1603, 1603, 1603, 1603,
 1603, 1603, 1603, 1603,    0, 1603, 1603, 1603, 1603, 1603,
 1603, 1603, 1603,    0, 1603, 1603,    0,    0,    0, 1603,
 1603, 1603, 1603, 1603, 1603,    0,    0, 1603, 1603, 1603,
 1603, 1603, 1603, 1603, 1603, 1603,    0,    0,    0,    0,
    0, 1603,    0, 1603, 1603, 1603, 1603, 1603, 1603,    0,
 1603,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1603, 1603,    0,
 1603, 1603, 1603, 1603, 1603, 1603, 1603, 1603, 1603,    0,
 1603, 1603, 1603, 1603, 1603, 1603, 1603, 1603, 1603, 1603,
 1603, 1603,    0, 1603, 1603, 1603,    0, 1603,    0, 1559,
    0, 1603, 1603, 1603, 1603, 1603, 1603, 1603, 1603, 1603,
 1603, 1603, 1603, 1603,    0, 1603, 1603,    0,    0,    0,
    0,    0, 1603, 1603, 1603, 1603, 1603, 1603, 1603, 1603,
 1603, 1603, 1603, 1603, 1603, 1603, 1603, 1603, 1603, 1603,
 1559,    0, 1603, 1559, 1603, 1603, 1603, 1603, 1603, 1603,
    0,    0,    0,    0,    0,    0,    0,    0, 1559,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1559,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1559,    0, 1559,    0,
    0, 1559,    0, 1559, 1559,    0,    0, 1559, 1559,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1559, 1559,
    0, 1559,    0,    0,    0,    0, 1559,    0,    0, 1559,
    0,    0,    0,    0,    0,    0, 1559,    0, 1559, 1559,
    0, 1559, 1559, 1559, 1559, 1559,    0,    0, 1559,    0,
    0, 1559,    0, 1559, 1559,    0,    0,    0, 1559, 1559,
    0, 1559, 1559, 1559,    0, 1559, 1559, 1559,    0,    0,
    0, 1559,    0, 1559, 1559, 1559,    0,    0,    0, 1559,
    0,    0,    0,    0,    0,    0,    0, 1559, 1559,    0,
 1559,    0, 1559,    0,    0,    0, 1559, 1559,    0, 1559,
    0,    0,    0, 1559,    0,    0,    0, 1559,    0, 1559,
 1559, 1559,    0, 1559, 1559, 1559, 1559, 1559,    0, 1559,
 1559,    0,    0,    0,    0, 1559,    0, 1559, 1559,    0,
    0, 1559, 1559, 1559,    0,    0, 1559,    0, 1559,    0,
    0, 1559, 1559, 1559, 1559,    0, 1559,    0, 1559,    0,
 1559,    0, 1559, 1559, 1559,    0,    0, 1559, 1559, 1559,
    0, 1559, 1559,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1559, 1559, 1559,    0, 1559, 1559, 1559,    0,
    0,    0,    0,    0, 1559,    0,    0,    0, 1559,    0,
 1559, 1559, 1559, 1559,    0, 1559,    0, 1559, 1559, 1559,
 1559, 1559, 1559, 1559, 1559,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1559, 1559,    0, 1559, 1559, 1559,
    0,    0,    0,    0, 1559, 1559, 1559, 1559, 1559,    0,
 1559,    0,    0, 1559, 1559,    0,    0,    0, 1559, 1559,
    0,    0, 1559, 1559, 1559, 1559, 1559, 1559, 1559, 1559,
 1559,    0, 1559,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1559, 1559, 1559, 1559, 1559, 1559, 1559, 1559,
 1559, 1559, 1559, 1559, 1559, 1559, 1559, 1559, 1559, 1559,
 1559, 1559, 1559,    0, 1559, 1559, 1559, 1559, 1559, 1559,
    0, 1559, 1559, 1559, 1559,    0, 1559,    0, 1559, 1559,
 1559, 1559, 1559, 1559, 1559, 1559, 1559, 1559, 1559, 1559,
 1559, 1559, 1559, 1559, 1559, 1559,    0, 1559, 1559, 1559,
 1559,    0, 1559, 1559, 1559, 1559, 1559,    0, 1559, 1559,
 1559, 1559, 1559, 1559, 1559, 1559, 1559, 1559, 1559, 1559,
 1559,    0, 1559, 1559, 1559, 1559, 1559, 1559,    0, 1559,
 1559, 1559, 1559, 1559, 1559, 1559, 1559,    0, 1559, 1559,
    0,    0,    0, 1559, 1559, 1559, 1559, 1559, 1559,    0,
    0, 1559, 1559, 1559, 1559, 1559, 1559, 1559, 1559, 1559,
    0,    0,    0,    0,    0, 1559,    0, 1559, 1559, 1559,
 1559, 1559, 1559,    0, 1559,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1559, 1559,    0, 1559, 1559, 1559, 1559, 1559, 1559,
 1559, 1559, 1559,    0, 1559, 1559, 1559, 1559, 1559, 1559,
 1559, 1559, 1559, 1559, 1559, 1559,    0, 1559, 1559, 1559,
    0, 1559,    0, 1558,    0, 1559, 1559, 1559, 1559, 1559,
 1559, 1559, 1559, 1559, 1559, 1559, 1559, 1559,    0, 1559,
 1559,    0,    0,    0,    0,    0, 1559, 1559, 1559, 1559,
 1559, 1559, 1559, 1559, 1559, 1559, 1559, 1559, 1559, 1559,
 1559, 1559, 1559, 1559, 1558,    0, 1559, 1558, 1559, 1559,
 1559, 1559, 1559, 1559,    0,    0,    0,    0,    0,    0,
    0,    0, 1558,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1558,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1558,    0, 1558,    0,    0, 1558,    0, 1558, 1558,    0,
    0, 1558, 1558,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1558, 1558,    0, 1558,    0,    0,    0,    0,
 1558,    0,    0, 1558,    0,    0,    0,    0,    0,    0,
 1558,    0, 1558, 1558,    0, 1558, 1558, 1558, 1558, 1558,
    0,    0, 1558,    0,    0, 1558,    0, 1558, 1558,    0,
    0,    0, 1558, 1558,    0, 1558, 1558, 1558,    0, 1558,
 1558, 1558,    0,    0,    0, 1558,    0, 1558, 1558, 1558,
    0,    0,    0, 1558,    0,    0,    0,    0,    0,    0,
    0, 1558, 1558,    0, 1558,    0, 1558,    0,    0,    0,
 1558, 1558,    0, 1558,    0,    0,    0,    0,    0,    0,
    0, 1558,    0, 1558, 1558, 1558,    0, 1558, 1558, 1558,
 1558, 1558,    0, 1558, 1558,    0,    0,    0,    0, 1558,
    0, 1558, 1558,    0,    0, 1558, 1558, 1558,    0,    0,
 1558,    0, 1558,    0,    0, 1558, 1558, 1558, 1558,    0,
 1558,    0, 1558,    0, 1558,    0, 1558, 1558, 1558,    0,
    0, 1558, 1558, 1558,    0, 1558, 1558,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1558, 1558, 1558,    0,
 1558, 1558, 1558,    0,    0,    0,    0,    0, 1558,    0,
    0,    0, 1558,    0, 1558, 1558, 1558, 1558,    0, 1558,
    0, 1558, 1558, 1558, 1558, 1558, 1558, 1558, 1558,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1558, 1558,
    0, 1558, 1558, 1558,    0,    0,    0,    0, 1558, 1558,
 1558, 1558, 1558,    0, 1558,    0,    0, 1558, 1558,    0,
    0,    0, 1558, 1558,    0,    0, 1558, 1558, 1558, 1558,
 1558, 1558, 1558, 1558, 1558,    0, 1558,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1558, 1558, 1558, 1558,
 1558, 1558, 1558, 1558, 1558, 1558, 1558, 1558, 1558, 1558,
 1558, 1558, 1558, 1558, 1558, 1558, 1558,    0, 1558, 1558,
 1558, 1558, 1558, 1558,    0, 1558, 1558, 1558, 1558,    0,
 1558,    0, 1558, 1558, 1558, 1558, 1558, 1558, 1558, 1558,
 1558, 1558, 1558, 1558, 1558, 1558, 1558, 1558, 1558, 1558,
    0, 1558, 1558, 1558, 1558,    0, 1558, 1558, 1558, 1558,
 1558,    0, 1558, 1558, 1558, 1558, 1558, 1558, 1558, 1558,
 1558, 1558, 1558, 1558, 1558,    0, 1558, 1558, 1558, 1558,
 1558, 1558,    0, 1558, 1558, 1558, 1558, 1558, 1558, 1558,
 1558,    0, 1558, 1558,    0,    0,    0, 1558, 1558, 1558,
 1558, 1558, 1558,    0,    0, 1558, 1558, 1558, 1558, 1558,
 1558, 1558, 1558, 1558,    0,    0,    0,    0,    0, 1558,
    0, 1558, 1558, 1558, 1558, 1558, 1558,    0, 1558,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1558, 1558,    0, 1558, 1558,
 1558, 1558, 1558, 1558, 1558, 1558, 1558,    0, 1558, 1558,
 1558, 1558, 1558, 1558, 1558, 1558, 1558, 1558, 1558, 1558,
    0, 1558, 1558, 1558,    0, 1558,    0, 1557,    0, 1558,
 1558, 1558, 1558, 1558, 1558, 1558, 1558, 1558, 1558, 1558,
 1558, 1558,    0, 1558, 1558,    0,    0,    0,    0,    0,
 1558, 1558, 1558, 1558, 1558, 1558, 1558, 1558, 1558, 1558,
 1558, 1558, 1558, 1558, 1558, 1558, 1558, 1558, 1557,    0,
 1558, 1557, 1558, 1558, 1558, 1558, 1558, 1558,    0,    0,
    0,    0,    0,    0,    0,    0, 1557,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1557,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1557,    0, 1557,    0,    0,    0,
    0, 1557, 1557,    0,    0, 1557, 1557,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1557, 1557,    0, 1557,
    0,    0,    0,    0, 1557,    0,    0, 1557,    0,    0,
    0,    0,    0,    0, 1557,    0, 1557, 1557,    0, 1557,
 1557, 1557, 1557, 1557,    0,    0, 1557,    0,    0, 1557,
    0, 1557, 1557,    0,    0,    0, 1557, 1557,    0, 1557,
 1557, 1557,    0, 1557, 1557, 1557,    0,    0,    0, 1557,
    0, 1557, 1557, 1557,    0,    0,    0, 1557,    0,    0,
    0,    0,    0,    0,    0, 1557, 1557,    0, 1557,    0,
 1557,    0,    0,    0, 1557, 1557,    0, 1557,    0,    0,
    0,    0,    0,    0,    0, 1557,    0, 1557, 1557, 1557,
    0, 1557, 1557, 1557, 1557, 1557,    0, 1557, 1557,    0,
    0,    0,    0, 1557,    0, 1557, 1557,    0,    0, 1557,
 1557, 1557,    0,    0, 1557,    0, 1557,    0,    0, 1557,
 1557, 1557, 1557,    0, 1557,    0, 1557,    0, 1557,    0,
 1557, 1557, 1557,    0,    0, 1557, 1557, 1557,    0, 1557,
 1557,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1557, 1557, 1557,    0, 1557, 1557, 1557,    0,    0,    0,
    0,    0, 1557,    0,    0,    0, 1557,    0, 1557, 1557,
 1557, 1557,    0, 1557,    0, 1557, 1557, 1557, 1557, 1557,
 1557, 1557, 1557,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1557, 1557,    0, 1557, 1557, 1557,    0,    0,
    0,    0, 1557, 1557, 1557, 1557, 1557,    0, 1557,    0,
    0, 1557, 1557,    0,    0,    0, 1557, 1557,    0,    0,
 1557, 1557, 1557, 1557, 1557, 1557, 1557, 1557, 1557,    0,
 1557,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1557, 1557, 1557, 1557, 1557, 1557, 1557, 1557, 1557, 1557,
 1557, 1557, 1557, 1557, 1557, 1557, 1557, 1557, 1557, 1557,
 1557,    0, 1557, 1557, 1557, 1557, 1557, 1557,    0, 1557,
 1557, 1557, 1557,    0, 1557,    0, 1557, 1557, 1557, 1557,
 1557, 1557, 1557, 1557, 1557, 1557, 1557, 1557, 1557, 1557,
 1557, 1557, 1557, 1557,    0, 1557, 1557, 1557, 1557,    0,
 1557, 1557, 1557, 1557, 1557,    0, 1557, 1557, 1557, 1557,
 1557, 1557, 1557, 1557, 1557, 1557, 1557, 1557, 1557,    0,
 1557, 1557, 1557, 1557, 1557, 1557,    0, 1557, 1557, 1557,
 1557, 1557, 1557, 1557, 1557,    0, 1557, 1557,    0,    0,
    0, 1557, 1557, 1557, 1557, 1557, 1557,    0,    0, 1557,
 1557, 1557, 1557, 1557, 1557, 1557, 1557, 1557,    0,    0,
    0,    0,    0, 1557,    0, 1557, 1557, 1557, 1557, 1557,
 1557,    0, 1557,    0,    0,    0,    0, 1390,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1557,
 1557,    0, 1557, 1557, 1557, 1557, 1557, 1557, 1557, 1557,
 1557,    0, 1557, 1557, 1557, 1557, 1557, 1557, 1557, 1557,
 1557, 1557, 1557, 1557,    0, 1557, 1557, 1557, 1390, 1557,
    0, 1390,    0, 1557, 1557, 1557, 1557, 1557, 1557, 1557,
 1557, 1557, 1557, 1557, 1557, 1557, 1390, 1557, 1557,    0,
    0,    0,    0,    0, 1557, 1557, 1557, 1557, 1557, 1557,
 1557, 1557, 1557, 1557, 1557, 1557, 1557, 1557, 1557, 1557,
 1557, 1557,    0,    0, 1557,    0, 1557, 1557, 1557, 1557,
 1557, 1557,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1390,    0, 1390,    0,    0,    0,
    0, 1390, 1390,    0,    0, 1390, 1390,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1390, 1390,    0, 1390,
    0, 1390,    0,    0, 1390,    0,    0, 1390,    0,    0,
    0,    0,    0,    0, 1390,    0, 1390, 1390,    0,    0,
    0, 1390,    0, 1390,    0,    0, 1390,    0,    0, 1390,
    0,    0, 1390,    0,    0,    0, 1390, 1390,    0, 1390,
 1390, 1390,    0, 1390, 1390, 1390,    0,    0,    0, 1390,
    0, 1390, 1390, 1390,    0,    0,    0, 1390,    0,    0,
    0,    0,    0,    0,    0, 1390, 1390,    0, 1390,    0,
 1390,    0,    0,    0, 1390, 1390,    0, 1390,    0,    0,
    0,    0,    0,    0,    0, 1390,    0, 1390,    0, 1390,
    0, 1390, 1390, 1390, 1390, 1390,    0, 1390, 1390,    0,
    0,    0,    0, 1390,    0, 1390, 1390,    0,    0, 1390,
 1390,    0,    0,    0, 1390,    0, 1390,    0,    0, 1390,
 1390, 1390, 1390,    0, 1390,    0, 1390,    0, 1390, 1390,
 1390, 1390, 1390,    0,    0,    0,    0, 1390,    0, 1390,
 1390,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1390,    0, 1390,    0, 1390, 1390, 1390,    0,    0,    0,
    0,    0, 1390,    0,    0,    0, 1390,    0, 1390, 1390,
 1390, 1390,    0, 1390,    0,    0,    0,    0,    0,    0,
    0, 1390, 1390,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1390, 1390,    0, 1390, 1390, 1390,    0,    0,
    0,    0, 1390, 1390, 1390, 1390, 1390,    0, 1390,    0,
    0, 1390, 1390,    0,    0,    0, 1390, 1390,    0,    0,
 1390, 1390, 1390, 1390, 1390, 1390, 1390, 1390, 1390,    0,
 1390,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1390, 1390, 1390, 1390, 1390, 1390, 1390, 1390, 1390, 1390,
 1390, 1390, 1390, 1390, 1390, 1390, 1390, 1390, 1390, 1390,
 1390,    0, 1390, 1390, 1390, 1390, 1390, 1390,    0, 1390,
 1390, 1390, 1390,    0, 1390,    0, 1390, 1390, 1390, 1390,
 1390, 1390, 1390, 1390, 1390, 1390, 1390, 1390, 1390, 1390,
 1390, 1390, 1390, 1390,    0, 1390, 1390, 1390, 1390,    0,
 1390, 1390, 1390, 1390, 1390,    0, 1390, 1390, 1390, 1390,
 1390, 1390, 1390, 1390, 1390, 1390, 1390, 1390, 1390,    0,
 1390, 1390, 1390, 1390, 1390, 1390,    0, 1390, 1390, 1390,
 1390, 1390, 1390, 1390, 1390,    0, 1390, 1390,    0,    0,
    0, 1390, 1390, 1390, 1390, 1390, 1390,    0,    0, 1390,
 1390, 1390, 1390, 1390, 1390, 1390, 1390, 1390,    0,    0,
    0,    0,    0, 1390,    0, 1390, 1390, 1390, 1390, 1390,
 1390,    0, 1390,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1390,
 1390,    0, 1390, 1390, 1390, 1390, 1390, 1390, 1390, 1390,
 1390,    0, 1390, 1390, 1390, 1390, 1390, 1390, 1390, 1390,
 1390, 1390, 1390, 1390,    0, 1390, 1390, 1390, 1358, 1390,
    0,    0,    0, 1390, 1390, 1390, 1390, 1390, 1390, 1390,
 1390, 1390, 1390, 1390, 1390, 1390,    0, 1390, 1390,    0,
    0,    0,    0,    0, 1390, 1390, 1390, 1390, 1390, 1390,
 1390, 1390, 1390, 1390, 1390, 1390, 1390, 1390, 1358, 1358,
 1390, 1390, 1358,    0, 1390,    0, 1390, 1390, 1390, 1390,
 1390, 1390,    0,    0,    0,    0,    0, 1358,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1358,    0, 1358,    0,    0,
    0,    0,    0, 1358,    0,    0, 1358, 1358,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1358, 1358,    0,
 1358,    0, 1358,    0,    0, 1358,    0,    0, 1358,    0,
    0,    0,    0,    0,    0, 1358,    0, 1358, 1358,    0,
    0,    0, 1358,    0, 1358,    0,    0, 1358,    0,    0,
 1358,    0,    0, 1358,    0,    0,    0, 1358, 1358,    0,
 1358, 1358, 1358,    0, 1358, 1358, 1358,    0,    0,    0,
 1358,    0, 1358, 1358, 1358,    0,    0,    0, 1358,    0,
    0,    0,    0,    0,    0,    0, 1358, 1358,    0, 1358,
    0, 1358,    0,    0,    0, 1358, 1358,    0, 1358,    0,
    0,    0,    0,    0,    0,    0, 1358,    0, 1358,    0,
 1358,    0, 1358, 1358, 1358, 1358, 1358,    0, 1358, 1358,
    0,    0,    0,    0, 1358,    0, 1358, 1358,    0,    0,
 1358, 1358,    0,    0,    0, 1358,    0, 1358,    0,    0,
 1358, 1358, 1358, 1358,    0, 1358,    0, 1358,    0, 1358,
 1358, 1358, 1358, 1358,    0,    0,    0,    0, 1358,    0,
 1358, 1358,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1358,    0, 1358,    0, 1358, 1358, 1358,    0,    0,
    0,    0,    0, 1358,    0,    0,    0, 1358,    0, 1358,
 1358, 1358, 1358,    0, 1358,    0,    0,    0,    0,    0,
    0,    0, 1358, 1358,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1358, 1358,    0, 1358, 1358, 1358,    0,
    0,    0,    0, 1358, 1358, 1358, 1358, 1358,    0, 1358,
    0,    0, 1358, 1358,    0,    0,    0, 1358, 1358,    0,
    0, 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358,
    0, 1358,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358,
 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358,
 1358, 1358,    0, 1358, 1358, 1358, 1358, 1358, 1358,    0,
 1358, 1358, 1358, 1358,    0, 1358,    0, 1358, 1358, 1358,
 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358,
 1358, 1358, 1358, 1358, 1358,    0, 1358, 1358, 1358, 1358,
    0, 1358, 1358, 1358, 1358, 1358,    0, 1358, 1358, 1358,
 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358,
    0, 1358, 1358, 1358, 1358, 1358, 1358,    0, 1358, 1358,
 1358, 1358, 1358, 1358, 1358, 1358,    0, 1358, 1358,    0,
    0,    0, 1358, 1358, 1358, 1358, 1358, 1358,    0,    0,
 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358,    0,
    0,    0,    0,    0, 1358,    0, 1358, 1358, 1358, 1358,
 1358, 1358,    0, 1358,    0,    0,    0,    0, 1387,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1358, 1358,    0, 1358, 1358, 1358, 1358, 1358, 1358, 1358,
 1358, 1358,    0, 1358, 1358, 1358, 1358, 1358, 1358, 1358,
 1358, 1358, 1358, 1358, 1358,    0, 1358, 1358, 1358, 1387,
 1358,    0, 1387,    0, 1358, 1358, 1358, 1358, 1358, 1358,
 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1387, 1358, 1358,
    0,    0,    0,    0,    0, 1358, 1358, 1358, 1358, 1358,
 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358,    0,
    0, 1358, 1358,    0,    0, 1358,    0, 1358, 1358, 1358,
 1358, 1358, 1358,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1358,    0, 1358,    0,    0,
    0,    0,    0, 1358,    0,    0, 1358, 1358,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1358, 1358,    0,
 1358,    0, 1358,    0,    0, 1358,    0,    0, 1358,    0,
    0,    0,    0,    0,    0, 1358,    0, 1358, 1358,    0,
    0,    0, 1358,    0, 1358,    0,    0, 1358,    0,    0,
 1387,    0,    0, 1387,    0,    0,    0, 1358, 1358,    0,
 1387, 1387, 1358,    0, 1358, 1387, 1358,    0,    0,    0,
 1387,    0, 1358, 1387, 1358,    0,    0,    0, 1358,    0,
    0,    0,    0,    0,    0,    0, 1358, 1387,    0, 1358,
    0, 1358,    0,    0,    0, 1358, 1358,    0, 1387,    0,
    0,    0,    0,    0,    0,    0, 1387,    0, 1358,    0,
 1387,    0, 1358, 1358, 1358, 1358, 1358,    0, 1358, 1387,
    0,    0,    0,    0, 1358,    0, 1358, 1358,    0,    0,
 1358, 1358,    0,    0,    0, 1387,    0, 1358,    0,    0,
 1358, 1358, 1358, 1358,    0, 1358,    0, 1358,    0, 1358,
 1358, 1358, 1358, 1358,    0,    0,    0,    0, 1358,    0,
 1358, 1387,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1358,    0, 1387,    0, 1387, 1358, 1358,    0,    0,
    0,    0,    0, 1358,    0,    0,    0, 1358,    0, 1358,
 1358, 1358, 1358,    0, 1358,    0,    0,    0,    0,    0,
    0,    0, 1358, 1358,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1358, 1358,    0, 1358, 1358, 1358,    0,
    0,    0,    0, 1358, 1358, 1387, 1358, 1358,    0, 1358,
    0,    0, 1358, 1358,    0,    0,    0, 1358, 1358,    0,
    0, 1387, 1387, 1358, 1358, 1358, 1387, 1358, 1358, 1358,
    0, 1358,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358,
 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358,
 1358, 1358,    0, 1358, 1358, 1358, 1358, 1358, 1358,    0,
 1358, 1358, 1358, 1358,    0, 1358,    0, 1358, 1358, 1358,
 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358,
 1358, 1358, 1358, 1358, 1358,    0, 1358, 1358, 1358, 1358,
    0, 1358, 1358, 1358, 1358, 1358,    0, 1358, 1358, 1358,
 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358,
    0, 1358, 1358, 1358, 1358, 1358, 1358,    0, 1358, 1358,
 1358, 1358, 1358, 1358, 1358, 1358,    0, 1358, 1358,    0,
    0,    0, 1358, 1358, 1358, 1358, 1358, 1358,    0,    0,
 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358,    0,
    0,    0,    0,    0, 1358,    0, 1358, 1358, 1358, 1358,
 1358, 1358,    0, 1387,    0,    0,    0,    0, 1389,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1358, 1358,    0, 1358, 1358, 1358, 1358, 1358, 1358, 1358,
 1358, 1358,    0, 1358, 1358, 1358, 1358, 1358, 1358, 1358,
 1358, 1358, 1358, 1358, 1358,    0, 1358, 1358, 1358, 1389,
 1358,    0, 1389,    0, 1358, 1358, 1358, 1358, 1358, 1358,
 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1389, 1358, 1358,
    0,    0,    0,    0,    0, 1358, 1358, 1358, 1358, 1358,
 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358,    0,
    0, 1358, 1358,    0,    0, 1387,    0, 1358, 1358, 1358,
 1358, 1358, 1358,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1358,    0, 1358,    0,    0,
    0,    0,    0, 1358,    0,    0, 1358, 1358,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1358, 1358,    0,
 1358,    0, 1358,    0,    0, 1358,    0,    0, 1358,    0,
    0,    0,    0,    0,    0, 1358,    0, 1358, 1358,    0,
    0,    0, 1358,    0, 1358,    0,    0, 1358,    0,    0,
 1389,    0,    0, 1389,    0,    0,    0, 1358, 1358,    0,
 1389, 1389, 1358,    0, 1358, 1389, 1358,    0,    0,    0,
 1389,    0, 1358, 1389, 1358,    0,    0,    0, 1358,    0,
    0,    0,    0,    0,    0,    0, 1358, 1389,    0, 1358,
    0, 1358,    0,    0,    0, 1358, 1358,    0, 1389,    0,
    0,    0,    0,    0,    0,    0, 1389,    0, 1358,    0,
 1389,    0, 1358, 1358, 1358, 1358, 1358,    0, 1358, 1389,
    0,    0,    0,    0, 1358,    0, 1358, 1358,    0,    0,
 1358, 1358,    0,    0,    0, 1389,    0, 1358,    0,    0,
 1358, 1358, 1358, 1358,    0, 1358,    0, 1358,    0, 1358,
 1358, 1358, 1358, 1358,    0,    0,    0,    0, 1358,    0,
 1358, 1389,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1358,    0, 1389,    0, 1389, 1358, 1358,    0,    0,
    0,    0,    0, 1358,    0,    0,    0, 1358,    0, 1358,
 1358, 1358, 1358,    0, 1358,    0,    0,    0,    0,    0,
    0,    0, 1358, 1358,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1358, 1358,    0, 1358, 1358, 1358,    0,
    0,    0,    0, 1358, 1358, 1389, 1358, 1358,    0, 1358,
    0,    0, 1358, 1358,    0,    0,    0, 1358, 1358,    0,
    0, 1389, 1389, 1358, 1358, 1358, 1389, 1358, 1358, 1358,
    0, 1358,    0,    0,    0,    0, 1394,    0,    0,    0,
    0,    0, 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358,
 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358,
 1358, 1358,    0, 1358, 1358, 1358, 1358, 1358, 1358,    0,
 1358, 1358, 1358, 1358,    0, 1358,    0, 1358, 1358, 1358,
 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358,
 1358, 1358, 1358, 1358, 1358, 1394, 1358, 1358, 1358, 1358,
    0, 1358, 1358, 1358, 1358, 1358,    0, 1358, 1358, 1358,
 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358,
    0, 1358, 1358, 1358, 1358, 1358, 1358,    0, 1358, 1358,
 1358, 1358, 1358, 1358, 1358, 1358,    0, 1358, 1358,    0,
    0,    0, 1358, 1358, 1358, 1358, 1358, 1358,    0,    0,
 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358,    0,
    0,    0,    0,    0, 1358,    0, 1358, 1358, 1358, 1358,
 1358, 1358,    0, 1389,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1358, 1358,    0, 1358, 1358, 1358, 1358, 1358, 1358, 1358,
 1358, 1358,    0, 1358, 1358, 1358, 1358, 1358, 1358, 1358,
 1358, 1358, 1358, 1358, 1358,    0, 1358, 1358, 1358,    0,
 1358,    0,    0,    0, 1358, 1358, 1358, 1358, 1358, 1358,
 1358, 1358, 1358, 1358, 1358, 1358, 1358,    0, 1358, 1358,
    0,    0,    0,    0,    0, 1358, 1358, 1358, 1358, 1358,
 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358,    0,
    0, 1358, 1358,    0,    0, 1389,    0, 1358, 1358, 1358,
 1358, 1358, 1358,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1358,    0, 1358,    0,    0,    0,    0,
    0, 1358,    0,    0, 1358, 1358,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1358, 1358,    0, 1358,    0,
 1358,    0,    0, 1358,    0,    0, 1358,    0,    0,    0,
    0,    0,    0, 1358,    0, 1358, 1358,    0,    0,    0,
 1358,    0, 1358,    0,    0, 1358,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1358, 1358,    0,    0,    0,
 1358,    0, 1358,    0, 1358,    0,    0,    0,    0,    0,
 1358,    0, 1358,    0,    0,    0, 1358,    0,    0,    0,
    0,    0,    0,    0, 1358,    0,    0, 1358,    0, 1358,
    0,    0,    0, 1358, 1358,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1358,    0, 1394,    0,
 1358, 1358, 1358, 1358, 1358,    0, 1358, 1394,    0,    0,
    0,    0, 1358,    0, 1358, 1358,    0,    0, 1358, 1358,
    0,    0,    0,    0,    0, 1358,    0, 1394, 1358, 1358,
 1358, 1358,    0, 1358,    0, 1358,    0, 1358, 1358, 1358,
 1358, 1358,    0,    0,    0,    0, 1358,    0, 1358,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1358,
    0, 1394,    0,    0, 1358, 1358,    0,    0,    0,    0,
    0, 1358,    0,    0,    0, 1358,    0, 1358, 1358, 1358,
 1358,    0, 1358,    0,    0,    0,    0,    0,    0,    0,
 1358, 1358,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1358, 1358,    0, 1358, 1358, 1358,    0,    0,    0,
    0, 1358, 1358, 1394, 1358, 1358,    0, 1358,    0,    0,
 1358, 1358,    0,    0,    0, 1358, 1358,    0,    0,    0,
 1394, 1358, 1358, 1358,    0, 1358, 1358, 1358,    0, 1358,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358,
 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358,
    0, 1358, 1358, 1358, 1358, 1358, 1358,    0, 1358, 1358,
 1358, 1358,    0, 1358,    0, 1358, 1358, 1358, 1358, 1358,
 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358,
 1358, 1358, 1358,    0, 1358, 1358, 1358, 1358,    0, 1358,
 1358, 1358, 1358, 1358,    0, 1358, 1358, 1358, 1358, 1358,
 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358,    0, 1358,
 1358, 1358, 1358, 1358, 1358,    0, 1358, 1358, 1358, 1358,
 1358, 1358, 1358, 1358,    0, 1358, 1358,    0,    0,    0,
 1358, 1358, 1358, 1358, 1358, 1358,    0,    0, 1358, 1358,
 1358, 1358, 1358, 1358, 1358, 1358, 1358,    0,    0,    0,
    0,    0, 1358,    0, 1358, 1358, 1358, 1358, 1358, 1358,
    0,    0,    0,    0,    0,    0, 1356,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1358, 1358,
    0, 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358,
    0, 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358,
 1358, 1358, 1358,    0, 1358, 1358, 1358,    0, 1358,    0,
 1356,    0, 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358,
 1358, 1358, 1358, 1358, 1358, 1356, 1358, 1358,    0,    0,
    0,    0,    0, 1358, 1358, 1358, 1358, 1358, 1358, 1358,
 1358, 1358, 1358, 1358, 1358, 1358, 1358,    0,    0, 1358,
 1358,    0,    0,    0,    0, 1358, 1358, 1358, 1358, 1358,
 1358,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1358,    0, 1358,    0,    0,    0,    0,
    0, 1358,    0,    0, 1358, 1358,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1358, 1358,    0, 1358,    0,
    0,    0,    0, 1358,    0,    0, 1358,    0,    0,    0,
    0,    0,    0, 1358,    0, 1358, 1358,    0,    0,    0,
 1358,    0, 1358,    0,    0, 1358,    0,    0,    0,    0,
 1356,    0,    0,    0,    0, 1358, 1358,    0,    0,    0,
 1358,    0, 1358,    0, 1358,    0,    0,    0, 1356,    0,
 1358,    0, 1358,    0,    0,    0, 1358,    0,    0,    0,
    0,    0,    0,    0, 1358,    0,    0, 1358,    0, 1358,
    0,    0,    0, 1358, 1358,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1358,    0,    0,    0,
 1358, 1358, 1358, 1358, 1358,    0, 1358,    0,    0,    0,
    0,    0, 1358,    0, 1358, 1358,    0,    0, 1358, 1358,
    0,    0,    0,    0,    0, 1358,    0,    0, 1358, 1358,
 1358, 1358,    0, 1358,    0, 1358,    0, 1358,    0, 1358,
 1358, 1358,    0,    0,    0,    0, 1358,    0, 1358,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1358,
    0,    0,    0,    0, 1358, 1358,    0,    0,    0,    0,
    0, 1358,    0,    0,    0, 1358,    0, 1358, 1358, 1358,
 1358,    0, 1358,    0,    0,    0,    0,    0,    0,    0,
 1358, 1358,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1358, 1358,    0, 1358, 1358, 1358,    0,    0,    0,
    0, 1358, 1358,    0, 1358, 1358,    0, 1358,    0,    0,
 1358, 1358,    0,    0,    0, 1358, 1358,    0,    0,    0,
    0, 1358, 1358, 1358,    0, 1358, 1358, 1358,    0, 1358,
    0,    0,    0,    0, 1213,    0,    0,    0,    0, 1358,
 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358,
 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358,
    0, 1358, 1358, 1358, 1358, 1358, 1358,    0, 1358, 1358,
 1358, 1358,    0, 1358,    0, 1358, 1358, 1358, 1358, 1358,
 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358,
 1358, 1358, 1358, 1213, 1358, 1358, 1358, 1358,    0, 1358,
 1358, 1358, 1358, 1358,    0, 1358, 1358, 1358, 1358, 1358,
 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358,    0, 1358,
 1358, 1358, 1358, 1358, 1358,    0, 1358, 1358, 1358, 1358,
 1358, 1358, 1358, 1358,    0, 1358, 1358,    0,    0,    0,
 1358, 1358, 1358, 1358, 1358, 1358,    0,    0, 1358, 1358,
 1358, 1358, 1358, 1358, 1358, 1358, 1358,    0,    0,    0,
    0,    0, 1358,    0, 1358, 1358, 1358, 1358, 1358, 1358,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1358, 1358,
    0, 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358,
    0, 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358,
 1358, 1358, 1358,    0, 1358, 1358, 1358,    0, 1358,    0,
    0,    0, 1358, 1358, 1358, 1358, 1358, 1358, 1358, 1358,
 1358, 1358, 1358, 1358, 1358,    0, 1358, 1358,    0,    0,
    0,    0,    0, 1358, 1358, 1358, 1358, 1358, 1358, 1358,
 1358, 1358, 1358, 1358, 1358, 1358, 1358,    0,    0, 1358,
 1358,    0,    0,    0,    0, 1358, 1358, 1358, 1358, 1358,
 1358,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1200,    0, 1200,    0,    0,    0,    0,    0, 1200,
    0,    0, 1200, 1200,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1200, 1200,    0, 1200,    0, 1200,    0,
    0, 1200,    0,    0, 1200,    0,    0,    0,    0,    0,
    0, 1200,    0, 1200, 1200,    0,    0,    0, 1200,    0,
 1200,    0,    0, 1200,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1200, 1200,    0,    0,    0, 1200,    0,
 1200,    0, 1200,    0,    0,    0,    0,    0, 1200,    0,
 1200,    0,    0,    0, 1200,    0,    0,    0,    0,    0,
    0,    0, 1200,    0,    0, 1200,    0, 1200,    0,    0,
    0, 1200, 1200,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1200,    0,    0,    0, 1200, 1200,
 1200, 1200, 1200,    0, 1200,    0,    0,    0,    0,    0,
 1200,    0, 1200, 1200,    0,    0, 1200, 1200,    0,    0,
    0,    0,    0, 1200,    0,    0, 1200, 1200, 1200, 1200,
    0, 1200,    0, 1200,    0, 1200, 1200, 1200, 1200, 1200,
    0,    0,    0,    0, 1200,    0, 1200,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1200,    0,    0,
    0,    0, 1200, 1200,    0,    0,    0,    0,    0, 1200,
    0,    0,    0, 1200,    0, 1200, 1200, 1200, 1200,    0,
 1200,    0,    0,    0,    0,    0,    0,    0, 1200, 1200,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1200,
 1200,    0, 1200, 1200, 1200,    0,    0,    0,    0, 1200,
 1200,    0, 1200, 1200,    0, 1200,    0,    0, 1200, 1200,
    0,    0,    0, 1200, 1200,    0,    0,    0,    0, 1200,
 1200, 1200,    0, 1200, 1200, 1200,    0, 1200,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1200, 1200, 1200,
 1200, 1200, 1200, 1200, 1200, 1200, 1200, 1200, 1200, 1200,
 1200, 1200, 1200, 1200, 1200, 1200, 1200, 1200,    0, 1200,
 1200, 1200, 1200, 1200, 1200,    0, 1200, 1200, 1200, 1200,
    0, 1200,    0, 1200, 1200, 1200, 1200, 1200, 1200, 1200,
 1200, 1200, 1200, 1200, 1200, 1200, 1200, 1200, 1200, 1200,
 1200,    0, 1200, 1200, 1200, 1200,    0, 1200, 1200, 1200,
 1200, 1200,    0, 1200, 1200, 1200, 1200, 1200, 1200, 1200,
 1200, 1200, 1200, 1200, 1200, 1200,    0, 1200, 1200, 1200,
 1200, 1200, 1200,    0, 1200, 1200, 1200, 1200, 1200, 1200,
 1200, 1200,    0, 1200, 1200,    0,    0,    0, 1200, 1200,
 1200, 1200, 1200, 1200,    0,    0, 1200, 1200, 1200, 1200,
 1200, 1200, 1200, 1200, 1200,    0, 1306,    0,    0, 1017,
 1200, 1018, 1200, 1200, 1200, 1200, 1200, 1200,    0,    0,
    0,    0,    0,    0, 1020,    0,    0,    0,    0, 1021,
    0,    0,    0,    0,    0,    0, 1200, 1200,    0, 1200,
 1200, 1200, 1200, 1200, 1200, 1200, 1200, 1200,    0, 1200,
 1200, 1200, 1200, 1200, 1200, 1200, 1200, 1200, 1200, 1200,
 1200,    0, 1200, 1200, 1200,    0, 1200,    0,    0,    0,
 1200, 1200, 1200, 1200, 1200, 1200, 1200, 1200, 1200, 1200,
 1200, 1200, 1200,    0, 1200, 1200,    0,    0,    0,    0,
    0, 1200, 1200, 1200, 1200, 1200, 1200, 1200, 1200, 1200,
 1200, 1200, 1200, 1200, 1200,    0,    0, 1200, 1200,    0,
    0,    0,    0, 1200, 1200, 1200, 1200, 1200, 1200,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  152,    0,  153,    0,    0,    0,    0,
    0,  154,    0,  848,  155,  156,    0,    0,    0,    0,
  849,    0,    0,    0,    0,  157,  158,    0,  159,    0,
  160,  850,    0,  161,    0,    0,  162,  851,  852,    0,
    0,    0,    0,  163,    0,  164,  165,    0,    0,    0,
  166,    0,  167,    0,  853,  168,    0,    0,    0,    0,
    0,    0,    0,  854,    0,  169,  855,    0,    0,    0,
  171,    0,  172,    0,  173,    0,    0,    0,    0,    0,
  174,    0,  175,    0,    0,    0,  176,    0,    0,    0,
  856,    0,  857,  858,  177,  859,    0,  178,    0,  179,
  860,    0,  861,  180,  181,    0,    0,    0,    0,    0,
  862,    0,    0,    0,    0,    0,  182,    0,    0,    0,
  184,  185,  186,  187,  188,    0,  189,    0,  863,    0,
    0,    0,  190,    0,  191,  192,    0,    0,  193,  194,
    0,    0,    0,  864,    0,  195,    0,    0,  196,  197,
  865,  199,    0,  200,    0,  201,  866,  202,  203,  204,
  205,  206,  867,    0,    0,    0,  207,    0,  208,    0,
    0,    0,  868,    0,    0,    0,    0,    0,    0,  209,
    0,    0,    0,    0,  210,  211,  869,  870,  871,  872,
  873,  874,  875,  601,  602,  213,    0,  214,  215,  216,
  217,    0,  218,  876, 1697, 1698, 1699, 1700, 1701, 1702,
 1732, 1733,  877,  878,  879,  880,  881,  882,  883,  884,
  885,  221,  886,    0,  223,  224,  225,  887,  888,    0,
  889,  890,  891,    0,  228,  229,  892,  230,    0,    0,
  231,  232,  893,  894,  895,  233,  234,    0,    0,    0,
    0,  235,  896,  237,    0,  897,  239,  240,    0,  241,
    0,  898,  899,  900,    0,  901,  902,    0,  903,  242,
  243,  244,  245,  246,  247,  904,  249,  905,  251,  906,
  907,  908,  909,  910,  911,  912,  913,  914,  915,  916,
    0,  917,  918,  919,  920,  921,  922,    0,  923,  924,
  925,  272,    0,  926,    0,  927,  928,  929,  930,  931,
  279,  280,  932, 1734,  933,  934,  935,  286,  936,  288,
  937,  290,  938,    0,  939,  940,  941,  942,    0,  943,
  944,  945,  299,  946,    0,  947,  948,  303,  949, 1735,
  306,  950,  308,  309,  310,  311,  312,  313,    0,  951,
  315,  316,  317,  318,  319,    0,  952,  321,  322,  323,
  953,  954,  326,  327,    0,  328,  329,  955,  956,    0,
  330,  331,  332,  957,  958,  959,    0,    0,  336,  960,
  961,  962,  963,  964,  965,  966,  967,  968,    0,  969,
  970,  971,  345,  972,  346,  347,  348,  349,  350,  351,
    0,    0,  973,  974,  975,  976,  977,  978,  979,  980,
  981,  982,  983,  984,  985,  986,  987,  988,  989,  990,
    0,  354,  991,  356,  357,  992,  359,  360,  361,  993,
    0,  363,  364,  365,  366,  367,  368,  994,  370,  995,
  996,  373,  374,    0,  375,  376,  997,    0,  378,    0,
  998,  999,  379, 1000,  381,  382,  383, 1001, 1002,  386,
  387,  388, 1003,  390,  391,    0, 1004,  393, 1005, 1006,
 1007, 1008, 1009,  394, 1010, 1011, 1012, 1013, 1014, 1015,
  401,  402,  403,  404,  405,  406,  407, 1707, 1708, 1016,
  409,    0,    0,    0,    0,  410,  411,  412,  413,  414,
  415, 1306,    0,    0, 1017,    0, 1018,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1020,
    0,    0,    0,    0, 1021,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  152,    0,
  153,    0,    0,    0,    0,    0,  154,    0,  848,  155,
  156,    0,    0,    0,    0,  849,    0,    0,    0,    0,
  157,  158,    0,  159,    0,  160,  850,    0,  161,    0,
    0,  162,  851,  852,    0,    0,    0,    0,  163,    0,
  164,  165,    0,    0,    0,  166,    0,  167,    0,  853,
  168,    0,    0,    0,    0,    0,    0,    0,  854,    0,
  169,  855,    0,    0,    0,  171,    0,  172,    0,  173,
    0,    0,    0,    0,    0,  174,    0,  175,    0,    0,
    0,  176,    0,    0,    0,  856,    0,  857,  858,  177,
  859,    0,  178,    0,  179,  860,    0,  861,  180,  181,
    0,    0,    0,    0,    0,  862,    0,    0,    0,    0,
    0,  182,    0,    0,    0,  184,  185,  186,  187,  188,
    0,  189,    0,  863,    0,    0,    0,  190,    0,  191,
  192,    0,    0,  193,  194,    0,    0,    0,  864,    0,
  195,    0,    0,  196,  197,  865,  199,    0,  200,    0,
  201,  866,  202,  203,  204,  205,  206,  867,    0,    0,
    0,  207,    0,  208,    0,    0,    0,  868,    0,    0,
    0,    0,    0,    0,  209,    0,    0,    0,    0,  210,
  211,  869,  870,  871,  872,  873,  874,  875,  601,  602,
  213,    0,  214,  215,  216,  217,    0,  218,  876,    0,
    0,    0,    0,    0,    0,  219,  220,  877,  878,  879,
  880,  881,  882,  883,  884,  885,  221,  886,    0,  223,
  224,  225,  887,  888,    0,  889,  890,  891,    0,  228,
  229,  892,  230,    0,    0,  231,  232,  893,  894,  895,
  233,  234,    0,    0,    0,    0,  235,  896,  237,    0,
  897,  239,  240, 1727,  241,    0,  898,  899,  900, 1728,
  901,  902, 1729,  903,  242,  243,  244,  245,  246,  247,
  904,  249,  905,  251,  906,  907,  908,  909,  910,  911,
  912,  913,  914,  915,  916,    0,  917,  918,  919,  920,
  921,  922,    0,  923,  924,  925,  272,    0,  926,    0,
  927,  928,  929,  930,  931,  279,  280,  932,  282,  933,
  934,  935,  286,  936,  288,  937,  290,  938,    0,  939,
  940,  941,  942,    0,  943,  944,  945,  299,  946,    0,
  947,  948,  303,  949,  305,  306,  950,  308,  309,  310,
  311,  312,  313,    0,  951,  315,  316,  317,  318,  319,
    0,  952,  321,  322,  323,  953,  954,  326,  327,    0,
  328,  329,  955,  956,    0,  330,  331,  332,  957,  958,
  959,    0,    0,  336,  960,  961,  962,  963,  964,  965,
  966,  967,  968,    0,  969,  970,  971,  345,  972,  346,
  347,  348,  349,  350,  351,    0,    0,  973,  974,  975,
  976,  977,  978,  979,  980,  981,  982,  983,  984,  985,
  986,  987,  988,  989,  990,    0,  354,  991,  356,  357,
  992,  359,  360,  361,  993,    0,  363,  364,  365,  366,
  367,  368,  994,  370,  995,  996,  373,  374,    0,  375,
  376,  997,    0,  378,    0,  998,  999,  379, 1000,  381,
  382,  383, 1001, 1002,  386,  387,  388, 1003,  390,  391,
    0, 1004,  393, 1005, 1006, 1007, 1008, 1009,  394, 1010,
 1011, 1012, 1013, 1014, 1015,  401,  402,  403,  404,  405,
  406,  407,    0, 1306, 1016,  409, 1017,    0, 1018,    0,
  410,  411,  412,  413,  414,  415,    0,    0,    0,    0,
    0, 1020,    0,    0,    0,    0, 1021,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
  152,    0,  153, 1818,    0,    0, 1819,    0,  154,    0,
  848,  155,  156,    0,    0,    0,    0,  849,    0,    0,
    0,    0,  157,  158,    0,  159,    0,  160,  850,    0,
  161,    0,    0,  162,  851,  852,    0,    0,    0,    0,
  163,    0,  164,  165,    0,    0,    0,  166,    0,  167,
    0,  853,  168,    0,    0,    0,    0,    0,    0,    0,
  854,    0,  169,  855,    0,    0,    0,  171,    0,  172,
    0,  173,    0,    0,    0,    0,    0,  174,    0,  175,
    0,    0,    0,  176,    0,    0,    0,  856,    0,  857,
  858,  177,  859,    0,  178,    0,  179,  860,    0,  861,
  180,  181,    0,    0,    0,    0,    0,  862,    0,    0,
    0,    0,    0,  182,    0,    0,    0,  184,  185,  186,
  187,  188,    0,  189,    0,  863,    0,    0,    0,  190,
    0,  191,  192,    0,    0,  193,  194,    0,    0,    0,
  864,    0,  195,    0,    0,  196,  197,  865,  199,    0,
  200, 1820,  201,  866,  202,  203,  204,  205,  206,  867,
    0,    0,    0,  207,    0,  208,    0,    0,    0,  868,
    0,    0,    0,    0,    0,    0,  209,    0,    0,    0,
    0,  210,  211,  869,  870,  871,  872,  873,  874,  875,
  601,  602,  213,    0,  214,  215,  216,  217,    0,  218,
  876,    0,    0,    0,    0,    0,    0,  219,  220,  877,
  878,  879,  880,  881,  882,  883,  884,  885,  221,  886,
    0,  223,  224,  225,  887,  888,    0,  889,  890,  891,
    0,  228,  229,  892,  230,    0,    0,  231,  232,  893,
  894,  895,  233,  234,    0,    0,    0,    0,  235,  896,
  237,    0,  897,  239,  240,    0,  241,    0,  898,  899,
  900,    0,  901,  902,    0,  903,  242,  243,  244,  245,
  246,  247,  904,  249,  905,  251,  906,  907,  908,  909,
  910,  911,  912,  913,  914,  915,  916,    0,  917,  918,
  919,  920,  921,  922,    0,  923,  924,  925,  272,    0,
  926,    0,  927,  928,  929,  930,  931,  279,  280,  932,
  282,  933,  934,  935,  286,  936,  288,  937,  290,  938,
    0,  939,  940,  941,  942,    0,  943,  944,  945,  299,
  946,    0,  947,  948,  303,  949,  305,  306,  950,  308,
  309,  310,  311,  312,  313,    0,  951,  315,  316,  317,
  318,  319,    0,  952,  321,  322,  323,  953,  954,  326,
  327,    0,  328,  329,  955,  956,    0,  330,  331,  332,
  957,  958,  959,    0,    0,  336,  960,  961,  962,  963,
  964,  965,  966,  967,  968,    0,  969,  970,  971,  345,
  972,  346,  347,  348,  349,  350,  351,    0,    0,  973,
  974,  975,  976,  977,  978,  979,  980,  981,  982,  983,
  984,  985,  986,  987,  988,  989,  990,    0,  354,  991,
  356,  357,  992,  359,  360,  361,  993,    0,  363,  364,
  365,  366,  367,  368,  994,  370,  995,  996,  373,  374,
    0,  375,  376,  997,    0,  378,    0,  998,  999,  379,
 1000,  381,  382,  383, 1001, 1002,  386,  387,  388, 1003,
  390,  391,    0, 1004,  393, 1005, 1006, 1007, 1008, 1009,
  394, 1010, 1011, 1012, 1013, 1014, 1015,  401,  402,  403,
  404,  405,  406,  407,    0, 1306, 1016,  409, 1017,    0,
 1018,    0,  410,  411,  412,  413,  414,  415,    0,    0,
    0,    0,    0, 1020,    0,    0,    0,    0, 1021,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,  152,    0,  153,    0,    0,    0,    0,    0,
  154,    0,  848,  155,  156,    0, 3746,    0,    0,  849,
    0,    0,    0,    0,  157,  158,    0,  159,    0,  160,
  850,    0,  161, 3747,    0,  162,  851,  852,    0,    0,
    0,    0,  163,    0,  164,  165,    0,    0,    0,  166,
    0,  167,    0,  853,  168,    0,    0,    0,    0,    0,
    0,    0,  854,    0,  169,  855,    0,    0,    0,  171,
    0,  172,    0,  173,    0,    0,    0,    0,    0,  174,
    0,  175,    0,    0,    0,  176,    0,    0,    0,  856,
    0,  857,  858,  177,  859,    0,  178,    0,  179,  860,
    0,  861,  180,  181,    0,    0,    0,    0,    0,  862,
    0,    0,    0,    0,    0,  182,    0,    0,    0,  184,
  185,  186,  187,  188,    0,  189,    0,  863,    0,    0,
    0,  190,    0,  191,  192,    0,    0,  193,  194,    0,
    0,    0,  864,    0,  195,    0,    0,  196,  197,  865,
  199,    0,  200,    0,  201,  866,  202,  203,  204,  205,
  206,  867,    0,    0,    0,  207,    0,  208,    0,    0,
    0,  868,    0,    0,    0,    0,    0,    0,  209,    0,
    0,    0,    0,  210,  211,  869,  870,  871,  872,  873,
  874,  875,  601,  602,  213,    0,  214,  215,  216,  217,
    0,  218,  876,    0,    0,    0,    0,    0,    0,  219,
  220,  877,  878,  879,  880,  881,  882,  883,  884,  885,
  221,  886,    0,  223,  224,  225,  887,  888,    0,  889,
  890,  891,    0,  228,  229,  892,  230,    0,    0,  231,
  232,  893,  894,  895,  233,  234,    0,    0,    0,    0,
  235,  896,  237,    0,  897,  239,  240,    0,  241,    0,
  898,  899,  900,    0,  901,  902,    0,  903,  242,  243,
  244,  245,  246,  247,  904,  249,  905,  251,  906,  907,
  908,  909,  910,  911,  912,  913,  914,  915,  916,    0,
  917,  918,  919,  920,  921,  922,    0,  923,  924,  925,
  272,    0,  926,    0,  927,  928,  929,  930,  931,  279,
  280,  932,  282,  933,  934,  935,  286,  936,  288,  937,
  290,  938,    0,  939,  940,  941,  942,    0,  943,  944,
  945,  299,  946,    0,  947,  948,  303,  949,  305,  306,
  950,  308,  309,  310,  311,  312,  313,    0,  951,  315,
  316,  317,  318,  319,    0,  952,  321,  322,  323,  953,
  954,  326,  327,    0,  328,  329,  955,  956,    0,  330,
  331,  332,  957,  958,  959,    0,    0,  336,  960,  961,
  962,  963,  964,  965,  966,  967,  968,    0,  969,  970,
  971,  345,  972,  346,  347,  348,  349,  350,  351,    0,
    0,  973,  974,  975,  976,  977,  978,  979,  980,  981,
  982,  983,  984,  985,  986,  987,  988,  989,  990,    0,
  354,  991,  356,  357,  992,  359,  360,  361,  993,    0,
  363,  364,  365,  366,  367,  368,  994,  370,  995,  996,
  373,  374,    0,  375,  376,  997,    0,  378,    0,  998,
  999,  379, 1000,  381,  382,  383, 1001, 1002,  386,  387,
  388, 1003,  390,  391,    0, 1004,  393, 1005, 1006, 1007,
 1008, 1009,  394, 1010, 1011, 1012, 1013, 1014, 1015,  401,
  402,  403,  404,  405,  406,  407,    0,    0, 1016,  409,
 3748,    0,    0,    0,  410,  411,  412,  413,  414,  415,
 1306,    0,    0, 1017,    0, 1018,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1020,    0,
    0,    0,    0, 1021,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  152,    0,  153,
    0,    0,    0,    0,    0,  154,    0,  848,  155,  156,
    0,    0,    0,    0,  849,    0,    0,    0,    0,  157,
  158,    0,  159,    0,  160,  850,    0,  161, 3838,    0,
  162,  851,  852,    0,    0,    0,    0,  163,    0,  164,
  165,    0,    0,    0,  166,    0,  167,    0,  853,  168,
    0,    0,    0,    0,    0,    0,    0,  854,    0,  169,
  855,    0,    0,    0,  171,    0,  172,    0,  173,    0,
    0,    0,    0,    0,  174,    0,  175,    0,    0,    0,
  176,    0,    0,    0,  856,    0,  857,  858,  177,  859,
    0,  178,    0,  179,  860,    0,  861,  180,  181,    0,
    0,    0,    0,    0,  862,    0,    0,    0,    0,    0,
  182,    0,    0,    0,  184,  185,  186,  187,  188,    0,
  189,    0,  863,    0,    0,    0,  190,    0,  191,  192,
    0,    0,  193,  194,    0,    0,    0,  864,    0,  195,
    0,    0,  196,  197,  865,  199,    0,  200,    0,  201,
  866,  202,  203,  204,  205,  206,  867,    0,    0,    0,
  207,    0,  208,    0,    0,    0,  868,    0,    0,    0,
    0,    0,    0,  209,    0,    0,    0,    0,  210,  211,
  869,  870,  871,  872,  873,  874,  875,  601,  602,  213,
    0,  214,  215,  216,  217,    0,  218,  876,    0,    0,
    0,    0,    0,    0,  219,  220,  877,  878,  879,  880,
  881,  882,  883,  884,  885,  221,  886,    0,  223,  224,
  225,  887,  888,    0,  889,  890,  891,    0,  228,  229,
  892,  230,    0,    0,  231,  232,  893,  894,  895,  233,
  234,    0,    0,    0,    0,  235,  896,  237,    0,  897,
  239,  240,    0,  241,    0,  898,  899,  900,    0,  901,
  902,    0,  903,  242,  243,  244,  245,  246,  247,  904,
  249,  905,  251,  906,  907,  908,  909,  910,  911,  912,
  913,  914,  915,  916,    0,  917,  918,  919,  920,  921,
  922,    0,  923,  924,  925,  272,    0,  926,    0,  927,
  928,  929,  930,  931,  279,  280,  932,  282,  933,  934,
  935,  286,  936,  288,  937,  290,  938,    0,  939,  940,
  941,  942,    0,  943,  944,  945,  299,  946,    0,  947,
  948,  303,  949,  305,  306,  950,  308,  309,  310,  311,
  312,  313,    0,  951,  315,  316,  317,  318,  319,    0,
  952,  321,  322,  323,  953,  954,  326,  327,    0,  328,
  329,  955,  956,    0,  330,  331,  332,  957,  958,  959,
    0,    0,  336,  960,  961,  962,  963,  964,  965,  966,
  967,  968,    0,  969,  970,  971,  345,  972,  346,  347,
  348,  349,  350,  351,    0,    0,  973,  974,  975,  976,
  977,  978,  979,  980,  981,  982,  983,  984,  985,  986,
  987,  988,  989,  990,    0,  354,  991,  356,  357,  992,
  359,  360,  361,  993,    0,  363,  364,  365,  366,  367,
  368,  994,  370,  995,  996,  373,  374,    0,  375,  376,
  997,    0,  378,    0,  998,  999,  379, 1000,  381,  382,
  383, 1001, 1002,  386,  387,  388, 1003,  390,  391,    0,
 1004,  393, 1005, 1006, 1007, 1008, 1009,  394, 1010, 1011,
 1012, 1013, 1014, 1015,  401,  402,  403,  404,  405,  406,
  407,    0,    0, 1016,  409, 3839,    0,    0,    0,  410,
  411,  412,  413,  414,  415, 1306,    0,    0, 1017,    0,
 1018,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1020,    0,    0,    0,    0, 1021,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,  152,    0,  153,    0,    0,    0,    0,    0,
  154,    0,  848,  155,  156,    0,    0,    0,    0,  849,
    0,    0,    0,    0,  157,  158,    0,  159,    0,  160,
  850,    0,  161, 3921,    0,  162,  851,  852,    0,    0,
    0,    0,  163,    0,  164,  165,    0,    0,    0,  166,
    0,  167,    0,  853,  168,    0,    0,    0,    0,    0,
    0,    0,  854,    0,  169,  855,    0,    0,    0,  171,
    0,  172,    0,  173,    0,    0,    0,    0,    0,  174,
    0,  175,    0,    0,    0,  176,    0,    0,    0,  856,
    0,  857,  858,  177,  859,    0,  178,    0,  179,  860,
    0,  861,  180,  181,    0,    0,    0,    0,    0,  862,
    0,    0,    0,    0,    0,  182,    0,    0,    0,  184,
  185,  186,  187,  188,    0,  189,    0,  863,    0,    0,
    0,  190,    0,  191,  192,    0,    0,  193,  194,    0,
    0,    0,  864,    0,  195,    0,    0,  196,  197,  865,
  199,    0,  200,    0,  201,  866,  202,  203,  204,  205,
  206,  867,    0,    0,    0,  207,    0,  208,    0,    0,
    0,  868,    0,    0,    0,    0,    0,    0,  209,    0,
    0,    0,    0,  210,  211,  869,  870,  871,  872,  873,
  874,  875,  601,  602,  213,    0,  214,  215,  216,  217,
    0,  218,  876,    0,    0,    0,    0,    0,    0,  219,
  220,  877,  878,  879,  880,  881,  882,  883,  884,  885,
  221,  886,    0,  223,  224,  225,  887,  888,    0,  889,
  890,  891,    0,  228,  229,  892,  230,    0,    0,  231,
  232,  893,  894,  895,  233,  234,    0,    0,    0,    0,
  235,  896,  237,    0,  897,  239,  240,    0,  241,    0,
  898,  899,  900,    0,  901,  902,    0,  903,  242,  243,
  244,  245,  246,  247,  904,  249,  905,  251,  906,  907,
  908,  909,  910,  911,  912,  913,  914,  915,  916,    0,
  917,  918,  919,  920,  921,  922,    0,  923,  924,  925,
  272,    0,  926,    0,  927,  928,  929,  930,  931,  279,
  280,  932,  282,  933,  934,  935,  286,  936,  288,  937,
  290,  938,    0,  939,  940,  941,  942,    0,  943,  944,
  945,  299,  946,    0,  947,  948,  303,  949,  305,  306,
  950,  308,  309,  310,  311,  312,  313,    0,  951,  315,
  316,  317,  318,  319,    0,  952,  321,  322,  323,  953,
  954,  326,  327,    0,  328,  329,  955,  956,    0,  330,
  331,  332,  957,  958,  959,    0,    0,  336,  960,  961,
  962,  963,  964,  965,  966,  967,  968,    0,  969,  970,
  971,  345,  972,  346,  347,  348,  349,  350,  351,    0,
    0,  973,  974,  975,  976,  977,  978,  979,  980,  981,
  982,  983,  984,  985,  986,  987,  988,  989,  990,    0,
  354,  991,  356,  357,  992,  359,  360,  361,  993,    0,
  363,  364,  365,  366,  367,  368,  994,  370,  995,  996,
  373,  374,    0,  375,  376,  997,    0,  378,    0,  998,
  999,  379, 1000,  381,  382,  383, 1001, 1002,  386,  387,
  388, 1003,  390,  391,    0, 1004,  393, 1005, 1006, 1007,
 1008, 1009,  394, 1010, 1011, 1012, 1013, 1014, 1015,  401,
  402,  403,  404,  405,  406,  407,    0,    0, 1016,  409,
 3922,    0,    0,    0,  410,  411,  412,  413,  414,  415,
 1340,    0, 1340, 1340,    0, 1340,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1340,    0,
    0,    0,    0, 1340,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1340,    0, 1340,
 1340,    0,    0,    0,    0, 1340,    0, 1340, 1340, 1340,
    0,    0,    0,    0, 1340,    0,    0,    0,    0, 1340,
 1340,    0, 1340,    0, 1340, 1340,    0, 1340,    0,    0,
 1340, 1340, 1340,    0,    0,    0,    0, 1340, 1340, 1340,
 1340,    0,    0,    0, 1340,    0, 1340,    0, 1340, 1340,
    0,    0,    0,    0,    0,    0,    0, 1340,    0, 1340,
 1340,    0,    0,    0, 1340,    0, 1340,    0, 1340,    0,
    0,    0,    0,    0, 1340,    0, 1340,    0,    0,    0,
 1340,    0,    0,    0, 1340,    0, 1340, 1340, 1340, 1340,
    0, 1340,    0, 1340, 1340,    0, 1340, 1340, 1340,    0,
    0,    0,    0,    0, 1340,    0,    0,    0,    0,    0,
 1340,    0,    0,    0, 1340, 1340, 1340, 1340, 1340,    0,
 1340,    0, 1340,    0,    0,    0, 1340,    0, 1340, 1340,
    0,    0, 1340, 1340,    0,    0,    0, 1340,    0, 1340,
    0,    0, 1340, 1340, 1340, 1340,    0, 1340,    0, 1340,
 1340, 1340, 1340, 1340, 1340, 1340, 1340,    0,    0,    0,
 1340,    0, 1340,    0,    0,    0, 1340,    0,    0,    0,
    0,    0,    0, 1340,    0,    0,    0,    0, 1340, 1340,
 1340, 1340, 1340, 1340, 1340, 1340, 1340, 1340, 1340, 1340,
    0, 1340, 1340, 1340, 1340,    0, 1340, 1340,    0,    0,
    0,    0,    0,    0, 1340, 1340, 1340, 1340, 1340, 1340,
 1340, 1340, 1340, 1340, 1340, 1340, 1340,    0, 1340, 1340,
    0, 1340, 1340,    0, 1340, 1340, 1340,    0, 1340, 1340,
 1340, 1340,    0,    0, 1340, 1340, 1340, 1340, 1340, 1340,
 1340,    0,    0,    0,    0, 1340, 1340, 1340,    0, 1340,
 1340, 1340,    0, 1340,    0, 1340, 1340, 1340,    0, 1340,
 1340,    0, 1340, 1340, 1340, 1340, 1340, 1340, 1340, 1340,
 1340, 1340, 1340, 1340, 1340, 1340, 1340, 1340, 1340, 1340,
 1340, 1340, 1340, 1340,    0, 1340, 1340, 1340, 1340, 1340,
 1340,    0, 1340, 1340, 1340, 1340,    0, 1340,    0, 1340,
 1340, 1340, 1340, 1340, 1340, 1340, 1340, 1340, 1340, 1340,
 1340, 1340, 1340, 1340, 1340, 1340, 1340,    0, 1340, 1340,
 1340, 1340,    0, 1340, 1340, 1340, 1340, 1340,    0, 1340,
 1340, 1340, 1340, 1340, 1340, 1340, 1340, 1340, 1340, 1340,
 1340, 1340,    0, 1340, 1340, 1340, 1340, 1340, 1340,    0,
 1340, 1340, 1340, 1340, 1340, 1340, 1340, 1340,    0, 1340,
 1340, 1340, 1340,    0, 1340, 1340, 1340, 1340, 1340, 1340,
    0,    0, 1340, 1340, 1340, 1340, 1340, 1340, 1340, 1340,
 1340, 1340,    0, 1340, 1340, 1340, 1340, 1340, 1340, 1340,
 1340, 1340, 1340, 1340,    0,    0, 1340, 1340, 1340, 1340,
 1340, 1340, 1340, 1340, 1340, 1340, 1340, 1340, 1340, 1340,
 1340, 1340, 1340, 1340,    0, 1340, 1340, 1340, 1340, 1340,
 1340, 1340, 1340, 1340,    0, 1340, 1340, 1340, 1340, 1340,
 1340, 1340, 1340, 1340, 1340, 1340, 1340,    0, 1340, 1340,
 1340,    0, 1340,    0, 1340, 1340, 1340, 1340, 1340, 1340,
 1340, 1340, 1340, 1340, 1340, 1340, 1340, 1340, 1340,    0,
 1340, 1340, 1340, 1340, 1340, 1340, 1340, 1340, 1340, 1340,
 1340, 1340, 1340, 1340, 1340, 1340, 1340, 1340, 1340, 1340,
 1340,    0, 1306, 1340, 1340, 1017,    0, 1018,    0, 1340,
 1340, 1340, 1340, 1340, 1340,    0,    0,    0,    0,    0,
 1020,    0,    0,    0,    0, 1021,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  152,
    0,  153,    0,    0,    0,    0,    0,  154,    0,  848,
  155,  156,    0,    0,    0,    0,  849,    0,    0,    0,
    0,  157,  158,    0,  159,    0,  160,  850,    0,  161,
 1305,    0,  162,  851,  852,    0,    0,    0,    0,  163,
    0,  164,  165,    0,    0,    0,  166,    0,  167,    0,
  853,  168,    0,    0,    0,    0,    0,    0,    0,  854,
    0,  169,  855,    0,    0,    0,  171,    0,  172,    0,
  173,    0,    0,    0,    0,    0,  174,    0,  175,    0,
    0,    0,  176,    0,    0,    0,  856,    0,  857,  858,
  177,  859,    0,  178,    0,  179,  860,    0,  861,  180,
  181,    0,    0,    0,    0,    0,  862,    0,    0,    0,
    0,    0,  182,    0,    0,    0,  184,  185,  186,  187,
  188,    0,  189,    0,  863,    0,    0,    0,  190,    0,
  191,  192,    0,    0,  193,  194,    0,    0,    0,  864,
    0,  195,    0,    0,  196,  197,  865,  199,    0,  200,
    0,  201,  866,  202,  203,  204,  205,  206,  867,    0,
    0,    0,  207,    0,  208,    0,    0,    0,  868,    0,
    0,    0,    0,    0,    0,  209,    0,    0,    0,    0,
  210,  211,  869,  870,  871,  872,  873,  874,  875,  601,
  602,  213,    0,  214,  215,  216,  217,    0,  218,  876,
    0,    0,    0,    0,    0,    0,  219,  220,  877,  878,
  879,  880,  881,  882,  883,  884,  885,  221,  886,    0,
  223,  224,  225,  887,  888,    0,  889,  890,  891,    0,
  228,  229,  892,  230,    0,    0,  231,  232,  893,  894,
  895,  233,  234,    0,    0,    0,    0,  235,  896,  237,
    0,  897,  239,  240,    0,  241,    0,  898,  899,  900,
    0,  901,  902,    0,  903,  242,  243,  244,  245,  246,
  247,  904,  249,  905,  251,  906,  907,  908,  909,  910,
  911,  912,  913,  914,  915,  916,    0,  917,  918,  919,
  920,  921,  922,    0,  923,  924,  925,  272,    0,  926,
    0,  927,  928,  929,  930,  931,  279,  280,  932,  282,
  933,  934,  935,  286,  936,  288,  937,  290,  938,    0,
  939,  940,  941,  942,    0,  943,  944,  945,  299,  946,
    0,  947,  948,  303,  949,  305,  306,  950,  308,  309,
  310,  311,  312,  313,    0,  951,  315,  316,  317,  318,
  319,    0,  952,  321,  322,  323,  953,  954,  326,  327,
    0,  328,  329,  955,  956,    0,  330,  331,  332,  957,
  958,  959,    0,    0,  336,  960,  961,  962,  963,  964,
  965,  966,  967,  968,    0,  969,  970,  971,  345,  972,
  346,  347,  348,  349,  350,  351,    0,    0,  973,  974,
  975,  976,  977,  978,  979,  980,  981,  982,  983,  984,
  985,  986,  987,  988,  989,  990,    0,  354,  991,  356,
  357,  992,  359,  360,  361,  993,    0,  363,  364,  365,
  366,  367,  368,  994,  370,  995,  996,  373,  374,    0,
  375,  376,  997,    0,  378,    0,  998,  999,  379, 1000,
  381,  382,  383, 1001, 1002,  386,  387,  388, 1003,  390,
  391,    0, 1004,  393, 1005, 1006, 1007, 1008, 1009,  394,
 1010, 1011, 1012, 1013, 1014, 1015,  401,  402,  403,  404,
  405,  406,  407,    0, 1306, 1016,  409, 1017,    0, 1018,
    0,  410,  411,  412,  413,  414,  415,    0,    0,    0,
    0,    0, 1020,    0,    0,    0,    0, 1021,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  152,    0,  153,    0,    0,    0,    0,    0,  154,
    0,  848,  155,  156,    0,    0,    0,    0,  849,    0,
    0,    0,    0,  157,  158,    0,  159,    0,  160,  850,
    0,  161,    0,    0,  162,  851,  852,    0,    0,    0,
    0,  163,    0,  164,  165,    0,    0,    0,  166,    0,
  167,    0,  853,  168,    0,    0,    0,    0,    0,    0,
    0,  854,    0,  169,  855,    0,    0,    0,  171,    0,
  172,    0,  173,    0,    0,    0,    0,    0,  174,    0,
  175,    0,    0,    0,  176,    0,    0,    0,  856,    0,
  857,  858,  177,  859,    0,  178,    0,  179,  860,    0,
  861,  180,  181,    0,    0,    0,    0,    0,  862,    0,
    0,    0,    0,    0,  182,    0,    0,    0,  184,  185,
  186,  187,  188,    0,  189,    0,  863,    0,    0,    0,
  190,    0,  191,  192,    0,    0,  193,  194,    0,    0,
    0,  864,    0,  195,    0,    0,  196,  197,  865,  199,
    0,  200,    0,  201,  866,  202,  203,  204,  205,  206,
  867,    0,    0,    0,  207,    0,  208,    0,    0,    0,
  868,    0,    0,    0,    0,    0,    0,  209, 1337,    0,
    0,    0,  210,  211,  869,  870,  871,  872,  873,  874,
  875,  601,  602,  213,    0,  214,  215,  216,  217,    0,
  218,  876,    0,    0,    0,    0,    0,    0,  219,  220,
  877,  878,  879,  880,  881,  882,  883,  884,  885,  221,
  886,    0,  223,  224,  225,  887,  888,    0,  889,  890,
  891,    0,  228,  229,  892,  230,    0,    0,  231,  232,
  893,  894,  895,  233,  234,    0,    0,    0,    0,  235,
  896,  237,    0,  897,  239,  240,    0,  241,    0,  898,
  899,  900,    0,  901,  902,    0,  903,  242,  243,  244,
  245,  246,  247,  904,  249,  905,  251,  906,  907,  908,
  909,  910,  911,  912,  913,  914,  915,  916,    0,  917,
  918,  919,  920,  921,  922,    0,  923,  924,  925,  272,
    0,  926,    0,  927,  928,  929,  930,  931,  279,  280,
  932,  282,  933,  934,  935,  286,  936,  288,  937,  290,
  938,    0,  939,  940,  941,  942,    0,  943,  944,  945,
  299,  946,    0,  947,  948,  303,  949,  305,  306,  950,
  308,  309,  310,  311,  312,  313,    0,  951,  315,  316,
  317,  318,  319,    0,  952,  321,  322,  323,  953,  954,
  326,  327,    0,  328,  329,  955,  956,    0,  330,  331,
  332,  957,  958,  959,    0,    0,  336,  960,  961,  962,
  963,  964,  965,  966,  967,  968,    0,  969,  970,  971,
  345,  972,  346,  347,  348,  349,  350,  351,    0,    0,
  973,  974,  975,  976,  977,  978,  979,  980,  981,  982,
  983,  984,  985,  986,  987,  988,  989,  990,    0,  354,
  991,  356,  357,  992,  359,  360,  361,  993,    0,  363,
  364,  365,  366,  367,  368,  994,  370,  995,  996,  373,
  374,    0,  375,  376,  997,    0,  378,    0,  998,  999,
  379, 1000,  381,  382,  383, 1001, 1002,  386,  387,  388,
 1003,  390,  391,    0, 1004,  393, 1005, 1006, 1007, 1008,
 1009,  394, 1010, 1011, 1012, 1013, 1014, 1015,  401,  402,
  403,  404,  405,  406,  407,    0, 1306, 1016,  409, 1017,
    0, 1018,    0,  410,  411,  412,  413,  414,  415,    0,
    0,    0,    0,    0, 1020,    0,    0,    0,    0, 1021,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  152,    0,  153,    0,    0,    0,    0,
    0,  154,    0,  848,  155,  156,    0,    0,    0,    0,
  849,    0,    0,    0,    0,  157,  158,    0,  159,    0,
  160,  850,    0,  161,    0,    0,  162,  851,  852,    0,
    0,    0,    0,  163,    0,  164,  165,    0,    0,    0,
  166,    0,  167,    0,  853,  168,    0,    0,    0,    0,
    0,    0,    0,  854,    0,  169,  855,    0,    0,    0,
  171,    0,  172,    0,  173,    0,    0,    0,    0,    0,
  174,    0,  175,    0,    0,    0,  176,    0,    0,    0,
  856,    0,  857,  858,  177,  859,    0,  178,    0,  179,
  860,    0,  861,  180,  181,    0,    0,    0,    0,    0,
  862,    0,    0,    0,    0,    0,  182,    0,    0,    0,
  184,  185,  186,  187,  188,    0,  189,    0,  863,    0,
    0,    0,  190,    0,  191,  192,    0,    0,  193,  194,
    0,    0,    0,  864,    0,  195,    0,    0,  196,  197,
  865,  199,    0,  200,    0,  201,  866,  202,  203,  204,
  205,  206,  867,    0,    0,    0,  207,    0,  208,    0,
    0,    0,  868,    0,    0,    0,    0,    0,    0,  209,
    0,    0,    0,   15,  210,  211,  869,  870,  871,  872,
  873,  874,  875,  601,  602,  213,    0,  214,  215,  216,
  217,    0,  218,  876,    0,    0,    0,    0,    0,    0,
  219,  220,  877,  878,  879,  880,  881,  882,  883,  884,
  885,  221,  886,    0,  223,  224,  225,  887,  888,    0,
  889,  890,  891,    0,  228,  229,  892,  230,    0,    0,
  231,  232,  893,  894,  895,  233,  234,    0,    0,    0,
    0,  235,  896,  237,    0,  897,  239,  240,    0,  241,
    0,  898,  899,  900,    0,  901,  902,    0,  903,  242,
  243,  244,  245,  246,  247,  904,  249,  905,  251,  906,
  907,  908,  909,  910,  911,  912,  913,  914,  915,  916,
    0,  917,  918,  919,  920,  921,  922,    0,  923,  924,
  925,  272,    0,  926,    0,  927,  928,  929,  930,  931,
  279,  280,  932,  282,  933,  934,  935,  286,  936,  288,
  937,  290,  938,    0,  939,  940,  941,  942,    0,  943,
  944,  945,  299,  946,    0,  947,  948,  303,  949,  305,
  306,  950,  308,  309,  310,  311,  312,  313,    0,  951,
  315,  316,  317,  318,  319,    0,  952,  321,  322,  323,
  953,  954,  326,  327,    0,  328,  329,  955,  956,    0,
  330,  331,  332,  957,  958,  959,    0,    0,  336,  960,
  961,  962,  963,  964,  965,  966,  967,  968,    0,  969,
  970,  971,  345,  972,  346,  347,  348,  349,  350,  351,
    0,    0,  973,  974,  975,  976,  977,  978,  979,  980,
  981,  982,  983,  984,  985,  986,  987,  988,  989,  990,
    0,  354,  991,  356,  357,  992,  359,  360,  361,  993,
    0,  363,  364,  365,  366,  367,  368,  994,  370,  995,
  996,  373,  374,    0,  375,  376,  997,    0,  378,    0,
  998,  999,  379, 1000,  381,  382,  383, 1001, 1002,  386,
  387,  388, 1003,  390,  391,    0, 1004,  393, 1005, 1006,
 1007, 1008, 1009,  394, 1010, 1011, 1012, 1013, 1014, 1015,
  401,  402,  403,  404,  405,  406,  407,    0,    0, 1016,
  409,    0,    0,    0,    0,  410,  411,  412,  413,  414,
  415, 1306,    0, 1542, 1017,    0, 1018,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1020,
    0,    0,    0,    0, 1021,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  152,    0,
  153,    0,    0,    0,    0,    0,  154,    0,  848,  155,
  156,    0,    0,    0,    0,  849,    0,    0,    0,    0,
  157,  158,    0,  159,    0,  160,  850,    0,  161,    0,
    0,  162,  851,  852,    0,    0,    0,    0,  163,    0,
  164,  165,    0,    0,    0,  166,    0,  167,    0,  853,
  168,    0,    0,    0,    0,    0,    0,    0,  854,    0,
  169,  855,    0,    0,    0,  171,    0,  172,    0,  173,
    0,    0,    0,    0,    0,  174,    0,  175,    0,    0,
    0,  176,    0,    0,    0,  856,    0,  857,  858,  177,
  859,    0,  178,    0,  179,  860,    0,  861,  180,  181,
    0,    0,    0,    0,    0,  862,    0,    0,    0,    0,
    0,  182,    0,    0,    0,  184,  185,  186,  187,  188,
    0,  189,    0,  863,    0,    0,    0,  190,    0,  191,
  192,    0,    0,  193,  194,    0,    0,    0,  864,    0,
  195,    0,    0,  196,  197,  865,  199,    0,  200,    0,
  201,  866,  202,  203,  204,  205,  206,  867,    0,    0,
    0,  207,    0,  208,    0,    0,    0,  868,    0,    0,
    0,    0,    0,    0,  209,    0,    0,    0,    0,  210,
  211,  869,  870,  871,  872,  873,  874,  875,  601,  602,
  213,    0,  214,  215,  216,  217,    0,  218,  876,    0,
    0,    0,    0,    0,    0,  219,  220,  877,  878,  879,
  880,  881,  882,  883,  884,  885,  221,  886,    0,  223,
  224,  225,  887,  888,    0,  889,  890,  891,    0,  228,
  229,  892,  230,    0,    0,  231,  232,  893,  894,  895,
  233,  234,    0,    0,    0,    0,  235,  896,  237,    0,
  897,  239,  240,    0,  241,    0,  898,  899,  900,    0,
  901,  902,    0,  903,  242,  243,  244,  245,  246,  247,
  904,  249,  905,  251,  906,  907,  908,  909,  910,  911,
  912,  913,  914,  915,  916,    0,  917,  918,  919,  920,
  921,  922,    0,  923,  924,  925,  272,    0,  926,    0,
  927,  928,  929,  930,  931,  279,  280,  932,  282,  933,
  934,  935,  286,  936,  288,  937,  290,  938,    0,  939,
  940,  941,  942,    0,  943,  944,  945,  299,  946,    0,
  947,  948,  303,  949,  305,  306,  950,  308,  309,  310,
  311,  312,  313,    0,  951,  315,  316,  317,  318,  319,
    0,  952,  321,  322,  323,  953,  954,  326,  327,    0,
  328,  329,  955,  956,    0,  330,  331,  332,  957,  958,
  959,    0,    0,  336,  960,  961,  962,  963,  964,  965,
  966,  967,  968,    0,  969,  970,  971,  345,  972,  346,
  347,  348,  349,  350,  351,    0,    0,  973,  974,  975,
  976,  977,  978,  979,  980,  981,  982,  983,  984,  985,
  986,  987,  988,  989,  990,    0,  354,  991,  356,  357,
  992,  359,  360,  361,  993,    0,  363,  364,  365,  366,
  367,  368,  994,  370,  995,  996,  373,  374,    0,  375,
  376,  997,    0,  378,    0,  998,  999,  379, 1000,  381,
  382,  383, 1001, 1002,  386,  387,  388, 1003,  390,  391,
    0, 1004,  393, 1005, 1006, 1007, 1008, 1009,  394, 1010,
 1011, 1012, 1013, 1014, 1015,  401,  402,  403,  404,  405,
  406,  407,    0,    0, 1016,  409,    0,    0,    0,    0,
  410,  411,  412,  413,  414,  415, 1306, 1841,    0, 1017,
    0, 1018,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1020,    0,    0,    0,    0, 1021,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  152,    0,  153,    0,    0,    0,    0,
    0,  154,    0,  848,  155,  156,    0,    0,    0,    0,
  849,    0,    0,    0,    0,  157,  158,    0,  159,    0,
  160,  850,    0,  161,    0,    0,  162,  851,  852,    0,
    0,    0,    0,  163,    0,  164,  165,    0,    0,    0,
  166,    0,  167,    0,  853,  168,    0,    0,    0,    0,
    0,    0,    0,  854,    0,  169,  855,    0,    0,    0,
  171,    0,  172,    0,  173,    0,    0,    0,    0,    0,
  174,    0,  175,    0,    0,    0,  176,    0,    0,    0,
  856,    0,  857,  858,  177,  859,    0,  178,    0,  179,
  860,    0,  861,  180,  181,    0,    0,    0,    0,    0,
  862,    0,    0,    0,    0,    0,  182,    0,    0,    0,
  184,  185,  186,  187,  188,    0,  189,    0,  863,    0,
    0,    0,  190,    0,  191,  192,    0,    0,  193,  194,
    0,    0,    0,  864,    0,  195,    0,    0,  196,  197,
  865,  199,    0,  200,    0,  201,  866,  202,  203,  204,
  205,  206,  867,    0,    0,    0,  207,    0,  208,    0,
    0,    0,  868,    0,    0,    0,    0,    0,    0,  209,
    0,    0,    0,    0,  210,  211,  869,  870,  871,  872,
  873,  874,  875,  601,  602,  213,    0,  214,  215,  216,
  217,    0,  218,  876,    0,    0,    0,    0,    0,    0,
  219,  220,  877,  878,  879,  880,  881,  882,  883,  884,
  885,  221,  886,    0,  223,  224,  225,  887,  888,    0,
  889,  890,  891,    0,  228,  229,  892,  230,    0,    0,
  231,  232,  893,  894,  895,  233,  234,    0,    0,    0,
    0,  235,  896,  237,    0,  897,  239,  240,    0,  241,
    0,  898,  899,  900,    0,  901,  902,    0,  903,  242,
  243,  244,  245,  246,  247,  904,  249,  905,  251,  906,
  907,  908,  909,  910,  911,  912,  913,  914,  915,  916,
    0,  917,  918,  919,  920,  921,  922,    0,  923,  924,
  925,  272,    0,  926,    0,  927,  928,  929,  930,  931,
  279,  280,  932,  282,  933,  934,  935,  286,  936,  288,
  937,  290,  938,    0,  939,  940,  941,  942,    0,  943,
  944,  945,  299,  946,    0,  947,  948,  303,  949,  305,
  306,  950,  308,  309,  310,  311,  312,  313,    0,  951,
  315,  316,  317,  318,  319,    0,  952,  321,  322,  323,
  953,  954,  326,  327,    0,  328,  329,  955,  956,    0,
  330,  331,  332,  957,  958,  959,    0,    0,  336,  960,
  961,  962,  963,  964,  965,  966,  967,  968,    0,  969,
  970,  971,  345,  972,  346,  347,  348,  349,  350,  351,
    0,    0,  973,  974,  975,  976,  977,  978,  979,  980,
  981,  982,  983,  984,  985,  986,  987,  988,  989,  990,
    0,  354,  991,  356,  357,  992,  359,  360,  361,  993,
    0,  363,  364,  365,  366,  367,  368,  994,  370,  995,
  996,  373,  374,    0,  375,  376,  997,    0,  378,    0,
  998,  999,  379, 1000,  381,  382,  383, 1001, 1002,  386,
  387,  388, 1003,  390,  391,    0, 1004,  393, 1005, 1006,
 1007, 1008, 1009,  394, 1010, 1011, 1012, 1013, 1014, 1015,
  401,  402,  403,  404,  405,  406,  407,    0, 1306, 1016,
  409, 1017,    0, 1018,    0,  410,  411,  412,  413,  414,
  415,    0,    0,    0,    0,    0, 1020,    0,    0,    0,
    0, 1021,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  152,    0,  153,    0,    0,
    0,    0,    0,  154,    0,  848,  155,  156,    0,    0,
    0,    0,  849,    0,    0,    0,    0,  157,  158,    0,
  159,    0,  160,  850,    0,  161,    0,    0,  162,  851,
  852,    0,    0,    0,    0,  163,    0,  164,  165,    0,
    0,    0,  166,    0,  167,    0,  853,  168,    0,    0,
    0,    0, 2146,    0,    0,  854,    0,  169,  855,    0,
    0,    0,  171,    0,  172,    0,  173,    0,    0,    0,
    0,    0,  174,    0,  175,    0,    0,    0,  176,    0,
    0,    0,  856,    0,  857,  858,  177,  859,    0,  178,
    0,  179,  860,    0,  861,  180,  181,    0,    0,    0,
    0,    0,  862,    0,    0,    0,    0,    0,  182,    0,
    0,    0,  184,  185,  186,  187,  188,    0,  189,    0,
  863,    0,    0,    0,  190,    0,  191,  192,    0,    0,
  193,  194,    0,    0,    0,  864,    0,  195,    0,    0,
  196,  197,  865,  199,    0,  200,    0,  201,  866,  202,
  203,  204,  205,  206,  867,    0,    0,    0,  207,    0,
  208,    0,    0,    0,  868,    0,    0,    0,    0,    0,
    0,  209,    0,    0,    0,    0,  210,  211,  869,  870,
  871,  872,  873,  874,  875,  601,  602,  213,    0,  214,
  215,  216,  217,    0,  218,  876,    0,    0,    0,    0,
    0,    0,  219,  220,  877,  878,  879,  880,  881,  882,
  883,  884,  885,  221,  886,    0,  223,  224,  225,  887,
  888,    0,  889,  890,  891,    0,  228,  229,  892,  230,
    0,    0,  231,  232,  893,  894,  895,  233,  234,    0,
    0,    0,    0,  235,  896,  237,    0,  897,  239,  240,
    0,  241,    0,  898,  899,  900,    0,  901,  902,    0,
  903,  242,  243,  244,  245,  246,  247,  904,  249,  905,
  251,  906,  907,  908,  909,  910,  911,  912,  913,  914,
  915,  916,    0,  917,  918,  919,  920,  921,  922,    0,
  923,  924,  925,  272,    0,  926,    0,  927,  928,  929,
  930,  931,  279,  280,  932,  282,  933,  934,  935,  286,
  936,  288,  937,  290,  938,    0,  939,  940,  941,  942,
    0,  943,  944,  945,  299,  946,    0,  947,  948,  303,
  949,  305,  306,  950,  308,  309,  310,  311,  312,  313,
    0,  951,  315,  316,  317,  318,  319,    0,  952,  321,
  322,  323,  953,  954,  326,  327,    0,  328,  329,  955,
  956,    0,  330,  331,  332,  957,  958,  959,    0,    0,
  336,  960,  961,  962,  963,  964,  965,  966,  967,  968,
    0,  969,  970,  971,  345,  972,  346,  347,  348,  349,
  350,  351,    0,    0,  973,  974,  975,  976,  977,  978,
  979,  980,  981,  982,  983,  984,  985,  986,  987,  988,
  989,  990,    0,  354,  991,  356,  357,  992,  359,  360,
  361,  993,    0,  363,  364,  365,  366,  367,  368,  994,
  370,  995,  996,  373,  374,    0,  375,  376,  997,    0,
  378,    0,  998,  999,  379, 1000,  381,  382,  383, 1001,
 1002,  386,  387,  388, 1003,  390,  391,    0, 1004,  393,
 1005, 1006, 1007, 1008, 1009,  394, 1010, 1011, 1012, 1013,
 1014, 1015,  401,  402,  403,  404,  405,  406,  407,    0,
 1306, 1016,  409, 1017,    0, 1018,    0,  410,  411,  412,
  413,  414,  415,    0,    0,    0,    0,    0, 1020,    0,
    0,    0,    0, 1021,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  152,    0,  153,
    0,    0,    0,    0,    0,  154,    0,  848,  155,  156,
    0,    0,    0,    0,  849,    0,    0,    0,    0,  157,
  158,    0,  159,    0,  160,  850,    0,  161,    0,    0,
  162,  851,  852,    0,    0, 2334,    0,  163,    0,  164,
  165,    0,    0,    0,  166,    0,  167,    0,  853,  168,
    0,    0,    0,    0,    0,    0,    0,  854,    0,  169,
  855,    0,    0,    0,  171,    0,  172,    0,  173,    0,
    0,    0,    0,    0,  174,    0,  175,    0,    0,    0,
  176,    0,    0,    0,  856,    0,  857,  858,  177,  859,
    0,  178,    0,  179,  860,    0,  861,  180,  181,    0,
    0,    0,    0,    0,  862,    0,    0,    0,    0,    0,
  182,    0,    0,    0,  184,  185,  186,  187,  188,    0,
  189,    0,  863,    0,    0,    0,  190,    0,  191,  192,
    0,    0,  193,  194,    0,    0,    0,  864,    0,  195,
    0,    0,  196,  197,  865,  199,    0,  200,    0,  201,
  866,  202,  203,  204,  205,  206,  867,    0,    0,    0,
  207,    0,  208,    0,    0,    0,  868,    0,    0,    0,
    0,    0,    0,  209,    0,    0,    0,    0,  210,  211,
  869,  870,  871,  872,  873,  874,  875,  601,  602,  213,
    0,  214,  215,  216,  217,    0,  218,  876,    0,    0,
    0,    0,    0,    0,  219,  220,  877,  878,  879,  880,
  881,  882,  883,  884,  885,  221,  886,    0,  223,  224,
  225,  887,  888,    0,  889,  890,  891,    0,  228,  229,
  892,  230,    0,    0,  231,  232,  893,  894,  895,  233,
  234,    0,    0,    0,    0,  235,  896,  237,    0,  897,
  239,  240,    0,  241,    0,  898,  899,  900,    0,  901,
  902,    0,  903,  242,  243,  244,  245,  246,  247,  904,
  249,  905,  251,  906,  907,  908,  909,  910,  911,  912,
  913,  914,  915,  916,    0,  917,  918,  919,  920,  921,
  922,    0,  923,  924,  925,  272,    0,  926,    0,  927,
  928,  929,  930,  931,  279,  280,  932,  282,  933,  934,
  935,  286,  936,  288,  937,  290,  938,    0,  939,  940,
  941,  942,    0,  943,  944,  945,  299,  946,    0,  947,
  948,  303,  949,  305,  306,  950,  308,  309,  310,  311,
  312,  313,    0,  951,  315,  316,  317,  318,  319,    0,
  952,  321,  322,  323,  953,  954,  326,  327,    0,  328,
  329,  955,  956,    0,  330,  331,  332,  957,  958,  959,
    0,    0,  336,  960,  961,  962,  963,  964,  965,  966,
  967,  968,    0,  969,  970,  971,  345,  972,  346,  347,
  348,  349,  350,  351,    0,    0,  973,  974,  975,  976,
  977,  978,  979,  980,  981,  982,  983,  984,  985,  986,
  987,  988,  989,  990,    0,  354,  991,  356,  357,  992,
  359,  360,  361,  993,    0,  363,  364,  365,  366,  367,
  368,  994,  370,  995,  996,  373,  374,    0,  375,  376,
  997,    0,  378,    0,  998,  999,  379, 1000,  381,  382,
  383, 1001, 1002,  386,  387,  388, 1003,  390,  391,    0,
 1004,  393, 1005, 1006, 1007, 1008, 1009,  394, 1010, 1011,
 1012, 1013, 1014, 1015,  401,  402,  403,  404,  405,  406,
  407,    0, 1306, 1016,  409, 1017,    0, 1018,    0,  410,
  411,  412,  413,  414,  415,    0,    0,    0,    0,    0,
 1020,    0,    0,    0,    0, 1021,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  152,
    0,  153,    0,    0,    0,    0,    0,  154,    0,  848,
  155,  156,    0,    0,    0,    0,  849,    0,    0,    0,
    0,  157,  158,    0,  159,    0,  160,  850,    0,  161,
    0,    0,  162,  851,  852,    0,    0,    0,    0,  163,
    0,  164,  165,    0,    0,    0,  166,    0,  167,    0,
  853,  168,    0,    0,    0,    0,    0,    0,    0,  854,
    0,  169,  855,    0,    0,    0,  171,    0,  172,    0,
  173,    0,    0,    0,    0,    0,  174,    0,  175,    0,
    0,    0,  176,    0,    0,    0,  856,    0,  857,  858,
  177,  859,    0,  178,    0,  179,  860,    0,  861,  180,
  181,    0,    0,    0,    0,    0,  862,    0,    0,    0,
    0,    0,  182,    0,    0,    0,  184,  185,  186,  187,
  188,    0,  189,    0,  863,    0,    0,    0,  190,    0,
  191,  192,    0,    0,  193,  194,    0,    0,    0,  864,
    0,  195,    0,    0,  196,  197,  865,  199,    0,  200,
    0,  201,  866,  202,  203,  204,  205,  206,  867,    0,
    0, 1817,  207,    0,  208,    0,    0,    0,  868,    0,
    0,    0,    0,    0,    0,  209,    0,    0,    0,    0,
  210,  211,  869,  870,  871,  872,  873,  874,  875,  601,
  602,  213,    0,  214,  215,  216,  217,    0,  218,  876,
    0,    0,    0,    0,    0,    0,  219,  220,  877,  878,
  879,  880,  881,  882,  883,  884,  885,  221,  886,    0,
  223,  224,  225,  887,  888,    0,  889,  890,  891,    0,
  228,  229,  892,  230,    0,    0,  231,  232,  893,  894,
  895,  233,  234,    0,    0,    0,    0,  235,  896,  237,
    0,  897,  239,  240,    0,  241,    0,  898,  899,  900,
    0,  901,  902,    0,  903,  242,  243,  244,  245,  246,
  247,  904,  249,  905,  251,  906,  907,  908,  909,  910,
  911,  912,  913,  914,  915,  916,    0,  917,  918,  919,
  920,  921,  922,    0,  923,  924,  925,  272,    0,  926,
    0,  927,  928,  929,  930,  931,  279,  280,  932,  282,
  933,  934,  935,  286,  936,  288,  937,  290,  938,    0,
  939,  940,  941,  942,    0,  943,  944,  945,  299,  946,
    0,  947,  948,  303,  949,  305,  306,  950,  308,  309,
  310,  311,  312,  313,    0,  951,  315,  316,  317,  318,
  319,    0,  952,  321,  322,  323,  953,  954,  326,  327,
    0,  328,  329,  955,  956,    0,  330,  331,  332,  957,
  958,  959,    0,    0,  336,  960,  961,  962,  963,  964,
  965,  966,  967,  968,    0,  969,  970,  971,  345,  972,
  346,  347,  348,  349,  350,  351,    0,    0,  973,  974,
  975,  976,  977,  978,  979,  980,  981,  982,  983,  984,
  985,  986,  987,  988,  989,  990,    0,  354,  991,  356,
  357,  992,  359,  360,  361,  993,    0,  363,  364,  365,
  366,  367,  368,  994,  370,  995,  996,  373,  374,    0,
  375,  376,  997,    0,  378,    0,  998,  999,  379, 1000,
  381,  382,  383, 1001, 1002,  386,  387,  388, 1003,  390,
  391,    0, 1004,  393, 1005, 1006, 1007, 1008, 1009,  394,
 1010, 1011, 1012, 1013, 1014, 1015,  401,  402,  403,  404,
  405,  406,  407,    0,    0, 1016,  409,    0,    0,    0,
    0,  410,  411,  412,  413,  414,  415, 1306, 2629,    0,
 1017,    0, 1018,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1020,    0,    0,    0,    0,
 1021,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  152,    0,  153,    0,    0,    0,
    0,    0,  154,    0,  848,  155,  156,    0,    0,    0,
    0,  849,    0,    0,    0,    0,  157,  158,    0,  159,
    0,  160,  850,    0,  161,    0,    0,  162,  851,  852,
    0,    0,    0,    0,  163,    0,  164,  165,    0,    0,
    0,  166,    0,  167,    0,  853,  168,    0,    0,    0,
    0,    0,    0,    0,  854,    0,  169,  855,    0,    0,
    0,  171,    0,  172,    0,  173,    0,    0,    0,    0,
    0,  174,    0,  175,    0,    0,    0,  176,    0,    0,
    0,  856,    0,  857,  858,  177,  859,    0,  178,    0,
  179,  860,    0,  861,  180,  181,    0,    0,    0,    0,
    0,  862,    0,    0,    0,    0,    0,  182,    0,    0,
    0,  184,  185,  186,  187,  188,    0,  189,    0,  863,
    0,    0,    0,  190,    0,  191,  192,    0,    0,  193,
  194,    0,    0,    0,  864,    0,  195,    0,    0,  196,
  197,  865,  199,    0,  200,    0,  201,  866,  202,  203,
  204,  205,  206,  867,    0,    0,    0,  207,    0,  208,
    0,    0,    0,  868,    0,    0,    0,    0,    0,    0,
  209,    0,    0,    0,    0,  210,  211,  869,  870,  871,
  872,  873,  874,  875,  601,  602,  213,    0,  214,  215,
  216,  217,    0,  218,  876,    0,    0,    0,    0,    0,
    0,  219,  220,  877,  878,  879,  880,  881,  882,  883,
  884,  885,  221,  886,    0,  223,  224,  225,  887,  888,
    0,  889,  890,  891,    0,  228,  229,  892,  230,    0,
    0,  231,  232,  893,  894,  895,  233,  234,    0,    0,
    0,    0,  235,  896,  237,    0,  897,  239,  240,    0,
  241,    0,  898,  899,  900,    0,  901,  902,    0,  903,
  242,  243,  244,  245,  246,  247,  904,  249,  905,  251,
  906,  907,  908,  909,  910,  911,  912,  913,  914,  915,
  916,    0,  917,  918,  919,  920,  921,  922,    0,  923,
  924,  925,  272,    0,  926,    0,  927,  928,  929,  930,
  931,  279,  280,  932,  282,  933,  934,  935,  286,  936,
  288,  937,  290,  938,    0,  939,  940,  941,  942,    0,
  943,  944,  945,  299,  946,    0,  947,  948,  303,  949,
  305,  306,  950,  308,  309,  310,  311,  312,  313,    0,
  951,  315,  316,  317,  318,  319,    0,  952,  321,  322,
  323,  953,  954,  326,  327,    0,  328,  329,  955,  956,
    0,  330,  331,  332,  957,  958,  959,    0,    0,  336,
  960,  961,  962,  963,  964,  965,  966,  967,  968,    0,
  969,  970,  971,  345,  972,  346,  347,  348,  349,  350,
  351,    0,    0,  973,  974,  975,  976,  977,  978,  979,
  980,  981,  982,  983,  984,  985,  986,  987,  988,  989,
  990,    0,  354,  991,  356,  357,  992,  359,  360,  361,
  993,    0,  363,  364,  365,  366,  367,  368,  994,  370,
  995,  996,  373,  374,    0,  375,  376,  997,    0,  378,
    0,  998,  999,  379, 1000,  381,  382,  383, 1001, 1002,
  386,  387,  388, 1003,  390,  391,    0, 1004,  393, 1005,
 1006, 1007, 1008, 1009,  394, 1010, 1011, 1012, 1013, 1014,
 1015,  401,  402,  403,  404,  405,  406,  407,    0, 1306,
 1016,  409, 1017,    0, 1018,    0,  410,  411,  412,  413,
  414,  415,    0,    0,    0,    0,    0, 1020,    0,    0,
    0,    0, 1021,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  152,    0,  153,    0,
    0,    0,    0,    0,  154,    0,  848,  155,  156,    0,
    0,    0,    0,  849,    0,    0,    0,    0,  157,  158,
    0,  159,    0,  160,  850,    0,  161,    0,    0,  162,
  851,  852,    0,    0, 2706,    0,  163,    0,  164,  165,
    0,    0,    0,  166,    0,  167,    0,  853,  168,    0,
    0,    0,    0,    0,    0,    0,  854,    0,  169,  855,
    0,    0,    0,  171,    0,  172,    0,  173,    0,    0,
    0,    0,    0,  174,    0,  175,    0,    0,    0,  176,
    0,    0,    0,  856,    0,  857,  858,  177,  859,    0,
  178,    0,  179,  860,    0,  861,  180,  181,    0,    0,
    0,    0,    0,  862,    0,    0,    0,    0,    0,  182,
    0,    0,    0,  184,  185,  186,  187,  188,    0,  189,
    0,  863,    0,    0,    0,  190,    0,  191,  192,    0,
    0,  193,  194,    0,    0,    0,  864,    0,  195,    0,
    0,  196,  197,  865,  199,    0,  200,    0,  201,  866,
  202,  203,  204,  205,  206,  867,    0,    0,    0,  207,
    0,  208,    0,    0,    0,  868,    0,    0,    0,    0,
    0,    0,  209,    0,    0,    0,    0,  210,  211,  869,
  870,  871,  872,  873,  874,  875,  601,  602,  213,    0,
  214,  215,  216,  217,    0,  218,  876,    0,    0,    0,
    0,    0,    0,  219,  220,  877,  878,  879,  880,  881,
  882,  883,  884,  885,  221,  886,    0,  223,  224,  225,
  887,  888,    0,  889,  890,  891,    0,  228,  229,  892,
  230,    0,    0,  231,  232,  893,  894,  895,  233,  234,
    0,    0,    0,    0,  235,  896,  237,    0,  897,  239,
  240,    0,  241,    0,  898,  899,  900,    0,  901,  902,
    0,  903,  242,  243,  244,  245,  246,  247,  904,  249,
  905,  251,  906,  907,  908,  909,  910,  911,  912,  913,
  914,  915,  916,    0,  917,  918,  919,  920,  921,  922,
    0,  923,  924,  925,  272,    0,  926,    0,  927,  928,
  929,  930,  931,  279,  280,  932,  282,  933,  934,  935,
  286,  936,  288,  937,  290,  938,    0,  939,  940,  941,
  942,    0,  943,  944,  945,  299,  946,    0,  947,  948,
  303,  949,  305,  306,  950,  308,  309,  310,  311,  312,
  313,    0,  951,  315,  316,  317,  318,  319,    0,  952,
  321,  322,  323,  953,  954,  326,  327,    0,  328,  329,
  955,  956,    0,  330,  331,  332,  957,  958,  959,    0,
    0,  336,  960,  961,  962,  963,  964,  965,  966,  967,
  968,    0,  969,  970,  971,  345,  972,  346,  347,  348,
  349,  350,  351,    0,    0,  973,  974,  975,  976,  977,
  978,  979,  980,  981,  982,  983,  984,  985,  986,  987,
  988,  989,  990,    0,  354,  991,  356,  357,  992,  359,
  360,  361,  993,    0,  363,  364,  365,  366,  367,  368,
  994,  370,  995,  996,  373,  374,    0,  375,  376,  997,
    0,  378,    0,  998,  999,  379, 1000,  381,  382,  383,
 1001, 1002,  386,  387,  388, 1003,  390,  391,    0, 1004,
  393, 1005, 1006, 1007, 1008, 1009,  394, 1010, 1011, 1012,
 1013, 1014, 1015,  401,  402,  403,  404,  405,  406,  407,
    0, 1306, 1016,  409, 1017,    0, 1018,    0,  410,  411,
  412,  413,  414,  415,    0,    0,    0,    0,    0, 1020,
    0,    0,    0,    0, 1021,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  152,    0,
  153,    0,    0,    0,    0,    0,  154,    0,  848,  155,
  156,    0,    0,    0,    0,  849,    0,    0,    0,    0,
  157,  158,    0,  159,    0,  160,  850,    0,  161,    0,
    0,  162,  851,  852,    0,    0,    0,    0,  163,    0,
  164,  165,    0,    0,    0,  166,    0,  167,    0,  853,
  168,    0,    0,    0,    0,    0,    0,    0,  854,    0,
  169,  855,    0,    0,    0,  171,    0,  172,    0,  173,
    0,    0,    0,    0,    0,  174,    0,  175,    0,    0,
    0,  176,    0,    0,    0,  856,    0,  857,  858,  177,
  859,    0,  178,    0,  179,  860,    0,  861,  180,  181,
    0,    0,    0,    0,    0,  862,    0,    0,    0,    0,
    0,  182,    0,    0,    0,  184,  185,  186,  187,  188,
    0,  189,    0,  863,    0,    0,    0,  190,    0,  191,
  192,    0,    0,  193,  194,    0,    0,    0,  864,    0,
  195,    0,    0,  196,  197,  865,  199,    0,  200,    0,
  201,  866,  202,  203,  204,  205,  206,  867,    0,    0,
    0,  207,    0,  208,    0,    0,    0,  868,    0,    0,
    0,    0,    0,    0,  209,    0,    0,    0,    0,  210,
  211,  869,  870,  871,  872,  873,  874,  875,  601,  602,
  213,    0,  214,  215,  216,  217,    0,  218,  876,    0,
    0,    0,    0,    0,    0,  219,  220,  877,  878,  879,
  880,  881,  882,  883,  884,  885,  221,  886,    0,  223,
  224,  225,  887,  888,    0,  889,  890,  891, 3060,  228,
  229,  892,  230,    0,    0,  231,  232,  893,  894,  895,
  233,  234,    0,    0,    0,    0,  235,  896,  237,    0,
  897,  239,  240,    0,  241,    0,  898,  899,  900,    0,
  901,  902,    0,  903,  242,  243,  244,  245,  246,  247,
  904,  249,  905,  251,  906,  907,  908,  909,  910,  911,
  912,  913,  914,  915,  916,    0,  917,  918,  919,  920,
  921,  922,    0,  923,  924,  925,  272,    0,  926,    0,
  927,  928,  929,  930,  931,  279,  280,  932,  282,  933,
  934,  935,  286,  936,  288,  937,  290,  938,    0,  939,
  940,  941,  942,    0,  943,  944,  945,  299,  946,    0,
  947,  948,  303,  949,  305,  306,  950,  308,  309,  310,
  311,  312,  313,    0,  951,  315,  316,  317,  318,  319,
    0,  952,  321,  322,  323,  953,  954,  326,  327,    0,
  328,  329,  955,  956,    0,  330,  331,  332,  957,  958,
  959,    0,    0,  336,  960,  961,  962,  963,  964,  965,
  966,  967,  968,    0,  969,  970,  971,  345,  972,  346,
  347,  348,  349,  350,  351,    0,    0,  973,  974,  975,
  976,  977,  978,  979,  980,  981,  982,  983,  984,  985,
  986,  987,  988,  989,  990,    0,  354,  991,  356,  357,
  992,  359,  360,  361,  993,    0,  363,  364,  365,  366,
  367,  368,  994,  370,  995,  996,  373,  374,    0,  375,
  376,  997,    0,  378,    0,  998,  999,  379, 1000,  381,
  382,  383, 1001, 1002,  386,  387,  388, 1003,  390,  391,
    0, 1004,  393, 1005, 1006, 1007, 1008, 1009,  394, 1010,
 1011, 1012, 1013, 1014, 1015,  401,  402,  403,  404,  405,
  406,  407,    0,    0, 1016,  409,    0,    0,    0,    0,
  410,  411,  412,  413,  414,  415, 1338,    0, 1338, 1338,
    0, 1338,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1338,    0,    0,    0,    0, 1338,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1338,    0, 1338, 1338,    0,    0,    0,
    0, 1338,    0, 1338, 1338, 1338,    0,    0,    0,    0,
 1338,    0,    0,    0,    0, 1338, 1338,    0, 1338,    0,
 1338, 1338,    0, 1338,    0,    0, 1338, 1338, 1338,    0,
    0,    0,    0, 1338, 1338, 1338, 1338,    0,    0,    0,
 1338,    0, 1338,    0, 1338, 1338,    0,    0,    0,    0,
    0,    0,    0, 1338,    0, 1338, 1338,    0,    0,    0,
 1338,    0, 1338,    0, 1338,    0,    0,    0,    0,    0,
 1338,    0, 1338,    0,    0,    0, 1338,    0,    0,    0,
 1338,    0, 1338, 1338, 1338, 1338,    0, 1338,    0, 1338,
 1338,    0, 1338, 1338, 1338,    0,    0,    0,    0,    0,
 1338,    0,    0,    0,    0,    0, 1338,    0,    0,    0,
 1338, 1338, 1338, 1338, 1338,    0, 1338,    0, 1338,    0,
    0,    0, 1338,    0, 1338, 1338,    0,    0, 1338, 1338,
    0,    0,    0, 1338,    0, 1338,    0,    0, 1338, 1338,
 1338, 1338,    0, 1338,    0, 1338, 1338, 1338, 1338, 1338,
 1338, 1338, 1338,    0,    0,    0, 1338,    0, 1338,    0,
    0,    0, 1338,    0,    0,    0,    0,    0,    0, 1338,
    0,    0,    0,    0, 1338, 1338, 1338, 1338, 1338, 1338,
 1338, 1338, 1338, 1338, 1338, 1338,    0, 1338, 1338, 1338,
 1338,    0, 1338, 1338,    0,    0,    0,    0,    0,    0,
 1338, 1338, 1338, 1338, 1338, 1338, 1338, 1338, 1338, 1338,
 1338, 1338, 1338,    0, 1338,    0,    0, 1338, 1338,    0,
 1338, 1338, 1338,    0, 1338, 1338, 1338, 1338,    0,    0,
 1338, 1338, 1338, 1338, 1338, 1338, 1338,    0,    0,    0,
    0, 1338, 1338, 1338,    0, 1338, 1338, 1338,    0, 1338,
    0, 1338, 1338, 1338,    0, 1338, 1338,    0, 1338, 1338,
 1338, 1338, 1338, 1338, 1338, 1338, 1338, 1338, 1338, 1338,
 1338, 1338, 1338, 1338, 1338, 1338, 1338, 1338, 1338, 1338,
    0, 1338, 1338, 1338, 1338, 1338, 1338,    0, 1338, 1338,
 1338, 1338,    0, 1338,    0, 1338, 1338, 1338, 1338, 1338,
 1338, 1338, 1338, 1338, 1338, 1338, 1338, 1338, 1338, 1338,
 1338, 1338, 1338,    0, 1338, 1338, 1338, 1338,    0, 1338,
 1338, 1338, 1338, 1338,    0, 1338, 1338, 1338, 1338, 1338,
 1338, 1338, 1338, 1338, 1338, 1338, 1338, 1338,    0, 1338,
 1338, 1338, 1338, 1338, 1338,    0, 1338, 1338, 1338, 1338,
 1338, 1338, 1338, 1338,    0, 1338, 1338, 1338, 1338,    0,
 1338, 1338, 1338, 1338, 1338, 1338,    0,    0, 1338, 1338,
 1338, 1338, 1338, 1338, 1338, 1338, 1338, 1338,    0, 1338,
 1338, 1338, 1338, 1338, 1338, 1338, 1338, 1338, 1338, 1338,
    0,    0, 1338, 1338, 1338, 1338, 1338, 1338, 1338, 1338,
 1338, 1338, 1338, 1338, 1338, 1338, 1338, 1338, 1338, 1338,
    0, 1338, 1338, 1338, 1338, 1338, 1338, 1338, 1338, 1338,
    0, 1338, 1338, 1338, 1338, 1338, 1338, 1338, 1338, 1338,
 1338, 1338, 1338,    0, 1338, 1338, 1338,    0, 1338,    0,
 1338, 1338, 1338, 1338, 1338, 1338, 1338, 1338, 1338, 1338,
 1338, 1338, 1338, 1338, 1338,    0, 1338, 1338, 1338, 1338,
 1338, 1338, 1338, 1338, 1338, 1338, 1338, 1338, 1338, 1338,
 1338, 1338, 1338, 1338, 1338, 1338, 1338,    0,    0, 1338,
 1338,    0,    0,    0,    0, 1338, 1338, 1338, 1338, 1338,
 1338, 2061,    0, 2061, 2061,    0, 2061,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 2061,
    0,    0,    0,    0, 2061,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2061,    0,
 2061,    0,    0,    0,    0,    0, 2061,    0, 2061, 2061,
 2061,    0,    0,    0,    0, 2061,    0,    0,    0,    0,
 2061, 2061,    0, 2061,    0, 2061, 2061,    0, 2061,    0,
    0, 2061, 2061, 2061,    0,    0,    0,    0, 2061,    0,
 2061, 2061,    0,    0,    0, 2061,    0, 2061,    0, 2061,
 2061,    0,    0,    0,    0,    0,    0,    0, 2061,    0,
 2061, 2061,    0,    0,    0, 2061,    0, 2061,    0, 2061,
    0,    0,    0,    0,    0, 2061,    0, 2061,    0,    0,
    0, 2061,    0,    0,    0, 2061,    0, 2061, 2061, 2061,
 2061,    0, 2061,    0, 2061, 2061,    0, 2061, 2061, 2061,
    0,    0,    0,    0,    0, 2061,    0,    0,    0,    0,
    0, 2061,    0,    0,    0, 2061, 2061, 2061, 2061, 2061,
    0, 2061,    0, 2061,    0,    0,    0, 2061,    0, 2061,
 2061,    0,    0, 2061, 2061,    0,    0,    0, 2061,    0,
 2061,    0,    0, 2061, 2061, 2061, 2061,    0, 2061,    0,
 2061, 2061, 2061, 2061, 2061, 2061, 2061, 2061,    0,    0,
    0, 2061,    0, 2061,    0,    0,    0, 2061,    0,    0,
    0,    0,    0,    0, 2061,    0,    0,    0,    0, 2061,
 2061, 2061, 2061, 2061, 2061, 2061, 2061, 2061, 2061, 2061,
 2061,    0, 2061, 2061, 2061, 2061,    0, 2061, 2061,    0,
    0,    0,    0,    0,    0, 2061, 2061, 2061, 2061, 2061,
 2061, 2061, 2061, 2061, 2061, 2061, 2061, 2061,    0, 2061,
 2061, 2061, 2061, 2061,    0, 2061, 2061, 2061,    0, 2061,
 2061, 2061, 2061,    0,    0, 2061, 2061, 2061, 2061, 2061,
 2061, 2061,    0,    0,    0,    0, 2061, 2061, 2061,    0,
 2061, 2061, 2061,    0, 2061,    0, 2061, 2061, 2061,    0,
 2061, 2061,    0, 2061, 2061, 2061, 2061, 2061, 2061, 2061,
 2061, 2061, 2061, 2061, 2061, 2061, 2061, 2061, 2061, 2061,
 2061, 2061, 2061, 2061, 2061,    0, 2061, 2061, 2061, 2061,
 2061, 2061,    0, 2061, 2061, 2061, 2061,    0, 2061,    0,
 2061, 2061, 2061, 2061, 2061, 2061, 2061, 2061, 2061, 2061,
 2061, 2061, 2061, 2061, 2061, 2061, 2061, 2061,    0, 2061,
 2061, 2061, 2061,    0, 2061, 2061, 2061, 2061, 2061,    0,
 2061, 2061, 2061, 2061, 2061, 2061, 2061, 2061, 2061, 2061,
 2061, 2061, 2061,    0, 2061, 2061, 2061, 2061, 2061, 2061,
    0, 2061, 2061, 2061, 2061, 2061, 2061, 2061, 2061,    0,
 2061, 2061, 2061, 2061,    0, 2061, 2061, 2061, 2061, 2061,
 2061,    0,    0, 2061, 2061, 2061, 2061, 2061, 2061, 2061,
 2061, 2061, 2061,    0, 2061, 2061, 2061, 2061, 2061, 2061,
 2061, 2061, 2061, 2061, 2061,    0,    0, 2061, 2061, 2061,
 2061, 2061, 2061, 2061, 2061, 2061, 2061, 2061, 2061, 2061,
 2061, 2061, 2061, 2061, 2061,    0, 2061, 2061, 2061, 2061,
 2061, 2061, 2061, 2061, 2061,    0, 2061, 2061, 2061, 2061,
 2061, 2061, 2061, 2061, 2061, 2061, 2061, 2061,    0, 2061,
 2061, 2061,    0, 2061,    0, 2061, 2061, 2061, 2061, 2061,
 2061, 2061, 2061, 2061, 2061, 2061, 2061, 2061, 2061, 2061,
    0, 2061, 2061, 2061, 2061, 2061, 2061, 2061, 2061, 2061,
 2061, 2061, 2061, 2061, 2061, 2061, 2061, 2061, 2061, 2061,
 2061, 2061,    0, 1019, 2061, 2061, 1017,    0, 1018,    0,
 2061, 2061, 2061, 2061, 2061, 2061,    0,    0,    0,    0,
    0, 1020,    0,    0,    0,    0, 1021,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
  152,    0,  153,    0,    0,    0,    0,    0,  154,    0,
  848,  155,  156,    0,    0,    0,    0,  849,    0,    0,
    0,    0,  157,  158,    0,  159,    0,  160,  850,    0,
  161,    0,    0,  162,  851,  852,    0,    0,    0,    0,
  163,    0,  164,  165,    0,    0,    0,  166,    0,  167,
    0,  853,  168,    0,    0,    0,    0,    0,    0,    0,
  854,    0,  169,  855,    0,    0,    0,  171,    0,  172,
    0,  173,    0,    0,    0,    0,    0,  174,    0,  175,
    0,    0,    0,  176,    0,    0,    0,  856,    0,  857,
  858,  177,  859,    0,  178,    0,  179,  860,    0,  861,
  180,  181,    0,    0,    0,    0,    0,  862,    0,    0,
    0,    0,    0,  182,    0,    0,    0,  184,  185,  186,
  187,  188,    0,  189,    0,  863,    0,    0,    0,  190,
    0,  191,  192,    0,    0,  193,  194,    0,    0,    0,
  864,    0,  195,    0,    0,  196,  197,  865,  199,    0,
  200,    0,  201,  866,  202,  203,  204,  205,  206,  867,
    0,    0,    0,  207,    0,  208,    0,    0,    0,  868,
    0,    0,    0,    0,    0,    0,  209,    0,    0,    0,
    0,  210,  211,  869,  870,  871,  872,  873,  874,  875,
  601,  602,  213,    0,  214,  215,  216,  217,    0,  218,
  876,    0,    0,    0,    0,    0,    0,  219,  220,  877,
  878,  879,  880,  881,  882,  883,  884,  885,  221,  886,
    0,  223,  224,  225,  887,  888,    0,  889,  890,  891,
    0,  228,  229,  892,  230,    0,    0,  231,  232,  893,
  894,  895,  233,  234,    0,    0,    0,    0,  235,  896,
  237,    0,  897,  239,  240,    0,  241,    0,  898,  899,
  900,    0,  901,  902,    0,  903,  242,  243,  244,  245,
  246,  247,  904,  249,  905,  251,  906,  907,  908,  909,
  910,  911,  912,  913,  914,  915,  916,    0,  917,  918,
  919,  920,  921,  922,    0,  923,  924,  925,  272,    0,
  926,    0,  927,  928,  929,  930,  931,  279,  280,  932,
  282,  933,  934,  935,  286,  936,  288,  937,  290,  938,
    0,  939,  940,  941,  942,    0,  943,  944,  945,  299,
  946,    0,  947,  948,  303,  949,  305,  306,  950,  308,
  309,  310,  311,  312,  313,    0,  951,  315,  316,  317,
  318,  319,    0,  952,  321,  322,  323,  953,  954,  326,
  327,    0,  328,  329,  955,  956,    0,  330,  331,  332,
  957,  958,  959,    0,    0,  336,  960,  961,  962,  963,
  964,  965,  966,  967,  968,    0,  969,  970,  971,  345,
  972,  346,  347,  348,  349,  350,  351,    0,    0,  973,
  974,  975,  976,  977,  978,  979,  980,  981,  982,  983,
  984,  985,  986,  987,  988,  989,  990,    0,  354,  991,
  356,  357,  992,  359,  360,  361,  993,    0,  363,  364,
  365,  366,  367,  368,  994,  370,  995,  996,  373,  374,
    0,  375,  376,  997,    0,  378,    0,  998,  999,  379,
 1000,  381,  382,  383, 1001, 1002,  386,  387,  388, 1003,
  390,  391,    0, 1004,  393, 1005, 1006, 1007, 1008, 1009,
  394, 1010, 1011, 1012, 1013, 1014, 1015,  401,  402,  403,
  404,  405,  406,  407,    0, 1306, 1016,  409, 1017,    0,
 1018,    0,  410,  411,  412,  413,  414,  415,    0,    0,
    0,    0,    0, 1020,    0,    0,    0,    0, 1021,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,  152,    0,  153,    0,    0,    0,    0,    0,
  154,    0,  848,  155,  156,    0,    0,    0,    0,  849,
    0,    0,    0,    0,  157,  158,    0,  159,    0,  160,
  850,    0,  161,    0,    0,  162,  851,  852,    0,    0,
    0,    0,  163,    0,  164,  165,    0,    0,    0,  166,
    0,  167,    0,  853,  168,    0,    0,    0,    0,    0,
    0,    0,  854,    0,  169,  855,    0,    0,    0,  171,
    0,  172,    0,  173,    0,    0,    0,    0,    0,  174,
    0,  175,    0,    0,    0,  176,    0,    0,    0,  856,
    0,  857,  858,  177,  859,    0,  178,    0,  179,  860,
    0,  861,  180,  181,    0,    0,    0,    0,    0,  862,
    0,    0,    0,    0,    0,  182,    0,    0,    0,  184,
  185,  186,  187,  188,    0,  189,    0,  863,    0,    0,
    0,  190,    0,  191,  192,    0,    0,  193,  194,    0,
    0,    0,  864,    0,  195,    0,    0,  196,  197,  865,
  199,    0,  200,    0,  201,  866,  202,  203,  204,  205,
  206,  867,    0,    0,    0,  207,    0,  208,    0,    0,
    0,  868,    0,    0,    0,    0,    0,    0,  209,    0,
    0,    0,    0,  210,  211,  869,  870,  871,  872,  873,
  874,  875,  601,  602,  213,    0,  214,  215,  216,  217,
    0,  218,  876,    0,    0,    0,    0,    0,    0,  219,
  220,  877,  878,  879,  880,  881,  882,  883,  884,  885,
  221,  886,    0,  223,  224,  225,  887,  888,    0,  889,
  890,  891,    0,  228,  229,  892,  230,    0,    0,  231,
  232,  893,  894,  895,  233,  234,    0,    0,    0,    0,
  235,  896,  237,    0,  897,  239,  240,    0,  241,    0,
  898,  899,  900,    0,  901,  902,    0,  903,  242,  243,
  244,  245,  246,  247,  904,  249,  905,  251,  906,  907,
  908,  909,  910,  911,  912,  913,  914,  915,  916,    0,
  917,  918,  919,  920,  921,  922,    0,  923,  924,  925,
  272,    0,  926,    0,  927,  928,  929,  930,  931,  279,
  280,  932,  282,  933,  934,  935,  286,  936,  288,  937,
  290,  938,    0,  939,  940,  941,  942,    0,  943,  944,
  945,  299,  946,    0,  947,  948,  303,  949,  305,  306,
  950,  308,  309,  310,  311,  312,  313,    0,  951,  315,
  316,  317,  318,  319,    0,  952,  321,  322,  323,  953,
  954,  326,  327,    0,  328,  329,  955,  956,    0,  330,
  331,  332,  957,  958,  959,    0,    0,  336,  960,  961,
  962,  963,  964,  965,  966,  967,  968,    0,  969,  970,
  971,  345,  972,  346,  347,  348,  349,  350,  351,    0,
    0,  973,  974,  975,  976,  977,  978,  979,  980,  981,
  982,  983,  984,  985,  986,  987,  988,  989,  990,    0,
  354,  991,  356,  357,  992,  359,  360,  361,  993,    0,
  363,  364,  365,  366,  367,  368,  994,  370,  995,  996,
  373,  374,    0,  375,  376,  997,    0,  378,    0,  998,
  999,  379, 1000,  381,  382,  383, 1001, 1002,  386,  387,
  388, 1003,  390,  391,    0, 1004,  393, 1005, 1006, 1007,
 1008, 1009,  394, 1010, 1011, 1012, 1013, 1014, 1015,  401,
  402,  403,  404,  405,  406,  407,    0, 2561, 1016,  409,
 1017,    0, 1018,    0,  410,  411,  412,  413,  414,  415,
    0,    0,    0,    0,    0, 1020,    0,    0,    0,    0,
 1021,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  152,    0,  153,    0,    0,    0,
    0,    0,  154,    0,  848,  155,  156,    0,    0,    0,
    0,  849,    0,    0,    0,    0,  157,  158,    0,  159,
    0,  160,  850,    0,  161,    0,    0,  162,  851,  852,
    0,    0,    0,    0,  163,    0,  164,  165,    0,    0,
    0,  166,    0,  167,    0,  853,  168,    0,    0,    0,
    0,    0,    0,    0,  854,    0,  169,  855,    0,    0,
    0,  171,    0,  172,    0,  173,    0,    0,    0,    0,
    0,  174,    0,  175,    0,    0,    0,  176,    0,    0,
    0,  856,    0,  857,  858,  177,  859,    0,  178,    0,
  179,  860,    0,  861,  180,  181,    0,    0,    0,    0,
    0,  862,    0,    0,    0,    0,    0,  182,    0,    0,
    0,  184,  185,  186,  187,  188,    0,  189,    0,  863,
    0,    0,    0,  190,    0,  191,  192,    0,    0,  193,
  194,    0,    0,    0,  864,    0,  195,    0,    0,  196,
  197,  865,  199,    0,  200,    0,  201,  866,  202,  203,
  204,  205,  206,  867,    0,    0,    0,  207,    0,  208,
    0,    0,    0,  868,    0,    0,    0,    0,    0,    0,
  209,    0,    0,    0,    0,  210,  211,  869,  870,  871,
  872,  873,  874,  875,  601,  602,  213,    0,  214,  215,
  216,  217,    0,  218,  876,    0,    0,    0,    0,    0,
    0,  219,  220,  877,  878,  879,  880,  881,  882,  883,
  884,  885,  221,  886,    0,  223,  224,  225,  887,  888,
    0,  889,  890,  891,    0,  228,  229,  892,  230,    0,
    0,  231,  232,  893,  894,  895,  233,  234,    0,    0,
    0,    0,  235,  896,  237,    0,  897,  239,  240,    0,
  241,    0,  898,  899,  900,    0,  901,  902,    0,  903,
  242,  243,  244,  245,  246,  247,  904,  249,  905,  251,
  906,  907,  908,  909,  910,  911,  912,  913,  914,  915,
  916,    0,  917,  918,  919,  920,  921,  922,    0,  923,
  924,  925,  272,    0,  926,    0,  927,  928,  929,  930,
  931,  279,  280,  932,  282,  933,  934,  935,  286,  936,
  288,  937,  290,  938,    0,  939,  940,  941,  942,    0,
  943,  944,  945,  299,  946,    0,  947,  948,  303,  949,
  305,  306,  950,  308,  309,  310,  311,  312,  313,    0,
  951,  315,  316,  317,  318,  319,    0,  952,  321,  322,
  323,  953,  954,  326,  327,    0,  328,  329,  955,  956,
    0,  330,  331,  332,  957,  958,  959,    0,    0,  336,
  960,  961,  962,  963,  964,  965,  966,  967,  968,    0,
  969,  970,  971,  345,  972,  346,  347,  348,  349,  350,
  351,    0,    0,  973,  974,  975,  976,  977,  978,  979,
  980,  981,  982,  983,  984,  985,  986,  987,  988,  989,
  990,    0,  354,  991,  356,  357,  992,  359,  360,  361,
  993,    0,  363,  364,  365,  366,  367,  368,  994,  370,
  995,  996,  373,  374,    0,  375,  376,  997,    0,  378,
    0,  998,  999,  379, 1000,  381,  382,  383, 1001, 1002,
  386,  387,  388, 1003,  390,  391,    0, 1004,  393, 1005,
 1006, 1007, 1008, 1009,  394, 1010, 1011, 1012, 1013, 1014,
 1015,  401,  402,  403,  404,  405,  406,  407,    0, 3548,
 1016,  409, 1017,    0, 1018,    0,  410,  411,  412,  413,
  414,  415,    0,    0,    0,    0,    0, 1020,    0,    0,
    0,    0, 1021,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  152,    0,  153,    0,
    0,    0,    0,    0,  154,    0,  848,  155,  156,    0,
    0,    0,    0,  849,    0,    0,    0,    0,  157,  158,
    0,  159,    0,  160,  850,    0,  161,    0,    0,  162,
  851,  852,    0,    0,    0,    0,  163,    0,  164,  165,
    0,    0,    0,  166,    0,  167,    0,  853,  168,    0,
    0,    0,    0,    0,    0,    0,  854,    0,  169,  855,
    0,    0,    0,  171,    0,  172,    0,  173,    0,    0,
    0,    0,    0,  174,    0,  175,    0,    0,    0,  176,
    0,    0,    0,  856,    0,  857,  858,  177,  859,    0,
  178,    0,  179,  860,    0,  861,  180,  181,    0,    0,
    0,    0,    0,  862,    0,    0,    0,    0,    0,  182,
    0,    0,    0,  184,  185,  186,  187,  188,    0,  189,
    0,  863,    0,    0,    0,  190,    0,  191,  192,    0,
    0,  193,  194,    0,    0,    0,  864,    0,  195,    0,
    0,  196,  197,  865,  199,    0,  200,    0,  201,  866,
  202,  203,  204,  205,  206,  867,    0,    0,    0,  207,
    0,  208,    0,    0,    0,  868,    0,    0,    0,    0,
    0,    0,  209,    0,    0,    0,    0,  210,  211,  869,
  870,  871,  872,  873,  874,  875,  601,  602,  213,    0,
  214,  215,  216,  217,    0,  218,  876,    0,    0,    0,
    0,    0,    0,  219,  220,  877,  878,  879,  880,  881,
  882,  883,  884,  885,  221,  886,    0,  223,  224,  225,
  887,  888,    0,  889,  890,  891,    0,  228,  229,  892,
  230,    0,    0,  231,  232,  893,  894,  895,  233,  234,
    0,    0,    0,    0,  235,  896,  237,    0,  897,  239,
  240,    0,  241,    0,  898,  899,  900,    0,  901,  902,
    0,  903,  242,  243,  244,  245,  246,  247,  904,  249,
  905,  251,  906,  907,  908,  909,  910,  911,  912,  913,
  914,  915,  916,    0,  917,  918,  919,  920,  921,  922,
    0,  923,  924,  925,  272,    0,  926,    0,  927,  928,
  929,  930,  931,  279,  280,  932,  282,  933,  934,  935,
  286,  936,  288,  937,  290,  938,    0,  939,  940,  941,
  942,    0,  943,  944,  945,  299,  946,    0,  947,  948,
  303,  949,  305,  306,  950,  308,  309,  310,  311,  312,
  313,    0,  951,  315,  316,  317,  318,  319,    0,  952,
  321,  322,  323,  953,  954,  326,  327,    0,  328,  329,
  955,  956,    0,  330,  331,  332,  957,  958,  959,    0,
    0,  336,  960,  961,  962,  963,  964,  965,  966,  967,
  968,    0,  969,  970,  971,  345,  972,  346,  347,  348,
  349,  350,  351,    0,    0,  973,  974,  975,  976,  977,
  978,  979,  980,  981,  982,  983,  984,  985,  986,  987,
  988,  989,  990,    0,  354,  991,  356,  357,  992,  359,
  360,  361,  993,    0,  363,  364,  365,  366,  367,  368,
  994,  370,  995,  996,  373,  374,    0,  375,  376,  997,
    0,  378,    0,  998,  999,  379, 1000,  381,  382,  383,
 1001, 1002,  386,  387,  388, 1003,  390,  391,    0, 1004,
  393, 1005, 1006, 1007, 1008, 1009,  394, 1010, 1011, 1012,
 1013, 1014, 1015,  401,  402,  403,  404,  405,  406,  407,
    0, 1306, 1016,  409, 1017,    0, 1018,    0,  410,  411,
  412,  413,  414,  415,    0,    0,    0,    0,    0, 1020,
    0,    0,    0,    0, 1021,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  152,    0,
  153,    0,    0,    0,    0,    0,  154,    0,  848,  155,
  156,    0,    0,    0,    0,  849,    0,    0,    0,    0,
  157,  158,    0,  159,    0,  160,  850,    0,  161,    0,
    0,  162,  851,  852,    0,    0,    0,    0,  163,    0,
  164,  165,    0,    0,    0,  166,    0,  167,    0,  853,
  168,    0,    0,    0,    0,    0,    0,    0,  854,    0,
  169,  855,    0,    0,    0,  171,    0,  172,    0,  173,
    0,    0,    0,    0,    0,  174,    0,  175,    0,    0,
    0,  176,    0,    0,    0,  856,    0,  857,  858,  177,
  859,    0,  178,    0,  179,  860,    0,  861,  180,  181,
    0,    0,    0,    0,    0,  862,    0,    0,    0,    0,
    0,  182,    0,    0,    0,  184,  185,  186,  187,  188,
    0,  189,    0,  863,    0,    0,    0,  190,    0,  191,
  192,    0,    0,  193,  194,    0,    0,    0,  864,    0,
  195,    0,    0,  196,  197,  865,  199,    0,  200,    0,
  201,  866,  202,  203,  204,  205,  206,  867,    0,    0,
    0,  207,    0,  208,    0,    0,    0,  868,    0,    0,
    0,    0,    0,    0,  209,    0,    0,    0,    0,  210,
  211,  869,  870,  871,  872,  873,  874,  875,  601,  602,
  213,    0,  214,  215,  216,  217,    0,  218,  876,    0,
    0,    0,    0,    0,    0,  219,  220,  877,  878,  879,
  880,  881,  882,  883,  884,  885,  221,  886,    0,  223,
  224,  225,  887,  888,    0,  889,  890,  891,    0,  228,
  229,  892,  230,    0,    0,  231,  232,  893,  894,  895,
  233,  234,    0,    0,    0,    0,  235,  896,  237,    0,
  897,  239,  240,    0,  241,    0,  898,  899,  900,    0,
  901,  902,    0,  903,  242,  243,  244,  245,  246,  247,
  904,  249,  905,  251,  906,  907,  908,  909,  910,  911,
  912,  913,  914,  915,  916,    0,  917,  918,  919,  920,
  921,  922,    0,  923,  924,  925,  272,    0,  926,    0,
  927,  928,  929,  930,  931,  279,  280,  932,  282,  933,
  934,  935,  286,  936,  288,  937,  290,  938,    0,  939,
  940,  941,  942,    0,  943,  944,  945,  299,  946,    0,
  947,  948,  303,  949,  305,  306,  950,  308,  309,  310,
  311,  312,  313,    0,  951,  315,  316,  317,  318,  319,
    0,  952,  321,  322,  323,  953,  954,  326,  327,    0,
  328,  329,  955,  956,    0,  330,  331,  332,  957,  958,
  959,    0,    0,  336,  960,  961,  962,  963,  964,  965,
  966,  967,  968,    0,  969,  970,  971,  345,  972,  346,
  347,  348,  349,  350,  351,    0,    0,  973,  974,  975,
  976,  977,  978,  979,  980,  981,  982,  983,  984,  985,
  986,  987,  988,  989,  990,    0,  354,  991,  356,  357,
  992,  359,  360,  361,  993,    0,  363,  364,  365, 3969,
  367,  368,  994,  370,  995,  996,  373,  374,    0,  375,
  376,  997,    0,  378,    0,  998,  999,  379, 1000,  381,
  382,  383, 1001, 1002,  386,  387,  388, 1003,  390,  391,
    0, 1004,  393, 1005, 1006, 1007, 1008, 1009,  394, 1010,
 1011, 1012, 1013, 1014, 1015,  401,  402,  403,  404,  405,
  406,  407,    0, 1306, 1016,  409, 1017,    0, 1018,    0,
  410,  411,  412,  413,  414,  415,    0,    0,    0,    0,
    0, 1020,    0,    0,    0,    0, 1021,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
  152,    0,  153,    0,    0,    0,    0,    0,  154,    0,
  848,  155,  156,    0,    0,    0,    0,  849,    0,    0,
    0,    0,  157,  158,    0,  159,    0,  160,  850,    0,
  161,    0,    0,  162,  851,  852,    0,    0,    0,    0,
  163,    0,  164,  165,    0,    0,    0,  166,    0,  167,
    0,  853,  168,    0,    0,    0,    0,    0,    0,    0,
  854,    0,  169,  855,    0,    0,    0,  171,    0,  172,
    0,  173,    0,    0,    0,    0,    0,  174,    0,  175,
    0,    0,    0,  176,    0,    0,    0,  856,    0,  857,
  858,  177,  859,    0,  178,    0,  179,  860,    0,  861,
  180,  181,    0,    0,    0,    0,    0,  862,    0,    0,
    0,    0,    0,  182,    0,    0,    0,  184,  185,  186,
  187,  188,    0,  189,    0,  863,    0,    0,    0,  190,
    0,  191,  192,    0,    0,  193,  194,    0,    0,    0,
  864,    0,  195,    0,    0,  196,  197,  865,  199,    0,
  200,    0,  201,  866,  202,  203,  204,  205,  206,  867,
    0,    0,    0,  207,    0,  208,    0,    0,    0,  868,
    0,    0,    0,    0,    0,    0,  209,    0,    0,    0,
    0,  210,  211,  869,  870,  871,  872,  873,  874,  875,
  601,  602,  213,    0,  214,  215,  216,  217,    0,  218,
  876,    0,    0,    0,    0,    0,    0,  219,  220,  877,
  878,  879,  880,  881,  882,  883,  884,  885,  221,  886,
    0,  223,  224,  225,  887,  888,    0,  889,  890,  891,
    0,  228,  229,  892,  230,    0,    0,  231,  232,  893,
  894,  895,  233,  234,    0,    0,    0,    0,  235,  896,
  237,    0,  897,  239,  240,    0,  241,    0,  898,  899,
  900,    0,  901,  902,    0,  903,  242,  243,  244,  245,
  246,  247,  904,  249,  905,  251,  906,  907,  908,  909,
  910,  911,  912,  913,  914,  915,  916,    0,  917,  918,
  919,  920,  921,  922,    0,  923,  924,  925,  272,    0,
  926,    0,  927,  928,  929,  930,  931,  279,  280,  932,
  282,  933,  934,  935,  286,  936,  288,  937,  290,  938,
    0,  939,  940,  941,  942,    0,  943,  944,  945,  299,
  946,    0,  947,  948,  303,  949,  305,  306,  950,  308,
  309,  310,  311,  312,  313,    0,  951,  315,  316, 3976,
  318,  319,    0,  952,  321,  322,  323,  953,  954,  326,
  327,    0,  328,  329,  955,  956,    0,  330,  331,  332,
  957,  958,  959,    0,    0,  336,  960,  961,  962,  963,
  964,  965,  966,  967,  968,    0,  969,  970,  971,  345,
  972,  346,  347,  348,  349,  350,  351,    0,    0,  973,
  974,  975,  976,  977,  978,  979,  980,  981,  982,  983,
  984,  985,  986,  987,  988,  989,  990,    0,  354,  991,
  356,  357,  992,  359,  360,  361,  993,    0,  363,  364,
  365,  366,  367,  368,  994,  370,  995,  996,  373,  374,
    0,  375,  376,  997,    0,  378,    0,  998,  999,  379,
 1000,  381,  382,  383, 1001, 1002,  386,  387,  388, 1003,
  390,  391,    0, 1004,  393, 1005, 1006, 1007, 1008, 1009,
  394, 1010, 1011, 1012, 1013, 1014, 1015,  401,  402,  403,
  404,  405,  406,  407,    0, 2061, 1016,  409, 2061,    0,
 2061,    0,  410,  411,  412,  413,  414,  415,    0,    0,
    0,    0,    0, 2061,    0,    0,    0,    0, 2061,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 2061,    0, 2061,    0,    0,    0,    0,    0,
 2061,    0, 2061, 2061, 2061,    0,    0,    0,    0, 2061,
    0,    0,    0,    0, 2061, 2061,    0, 2061,    0, 2061,
 2061,    0, 2061,    0,    0, 2061, 2061, 2061,    0,    0,
    0,    0, 2061,    0, 2061, 2061,    0,    0,    0, 2061,
    0, 2061,    0, 2061, 2061,    0,    0,    0,    0,    0,
    0,    0, 2061,    0, 2061, 2061,    0,    0,    0, 2061,
    0, 2061,    0, 2061,    0,    0,    0,    0,    0, 2061,
    0, 2061,    0,    0,    0, 2061,    0,    0,    0, 2061,
    0, 2061, 2061, 2061, 2061,    0, 2061,    0, 2061, 2061,
    0, 2061, 2061, 2061,    0,    0,    0,    0,    0, 2061,
    0,    0,    0,    0,    0, 2061,    0,    0,    0, 2061,
 2061, 2061, 2061, 2061,    0, 2061,    0, 2061,    0,    0,
    0, 2061,    0, 2061, 2061,    0,    0, 2061, 2061,    0,
    0,    0, 2061,    0, 2061,    0,    0, 2061, 2061, 2061,
 2061,    0, 2061,    0, 2061, 2061, 2061, 2061, 2061, 2061,
 2061, 2061,    0,    0,    0, 2061,    0, 2061,    0,    0,
    0, 2061,    0,    0,    0,    0,    0,    0, 2061,    0,
    0,    0,    0, 2061, 2061, 2061, 2061, 2061, 2061, 2061,
 2061, 2061, 2061, 2061, 2061,    0, 2061, 2061, 2061, 2061,
    0, 2061, 2061,    0,    0,    0,    0,    0,    0, 2061,
 2061, 2061, 2061, 2061, 2061, 2061, 2061, 2061, 2061, 2061,
 2061, 2061,    0, 2061, 2061, 2061, 2061, 2061,    0, 2061,
 2061, 2061,    0, 2061, 2061, 2061, 2061,    0,    0, 2061,
 2061, 2061, 2061, 2061, 2061, 2061,    0,    0,    0,    0,
 2061, 2061, 2061,    0, 2061, 2061, 2061,    0, 2061,    0,
 2061, 2061, 2061,    0, 2061, 2061,    0, 2061, 2061, 2061,
 2061, 2061, 2061, 2061, 2061, 2061, 2061, 2061, 2061, 2061,
 2061, 2061, 2061, 2061, 2061, 2061, 2061, 2061, 2061,    0,
 2061, 2061, 2061, 2061, 2061, 2061,    0, 2061, 2061, 2061,
 2061,    0, 2061,    0, 2061, 2061, 2061, 2061, 2061, 2061,
 2061, 2061, 2061, 2061, 2061, 2061, 2061, 2061, 2061, 2061,
 2061, 2061,    0, 2061, 2061, 2061, 2061,    0, 2061, 2061,
 2061, 2061, 2061,    0, 2061, 2061, 2061, 2061, 2061, 2061,
 2061, 2061, 2061, 2061, 2061, 2061, 2061,    0, 2061, 2061,
 2061, 2061, 2061, 2061,    0, 2061, 2061, 2061, 2061, 2061,
 2061, 2061, 2061,    0, 2061, 2061, 2061, 2061,    0, 2061,
 2061, 2061, 2061, 2061, 2061,    0,    0, 2061, 2061, 2061,
 2061, 2061, 2061, 2061, 2061, 2061, 2061,    0, 2061, 2061,
 2061, 2061, 2061, 2061, 2061, 2061, 2061, 2061, 2061,    0,
    0, 2061, 2061, 2061, 2061, 2061, 2061, 2061, 2061, 2061,
 2061, 2061, 2061, 2061, 2061, 2061, 2061, 2061, 2061,    0,
 2061, 2061, 2061, 2061, 2061, 2061, 2061, 2061, 2061,    0,
 2061, 2061, 2061, 2061, 2061, 2061, 2061, 2061, 2061, 2061,
 2061, 2061,    0, 2061, 2061, 2061,    0, 2061,    0, 2061,
 2061, 2061, 2061, 2061, 2061, 2061, 2061, 2061, 2061, 2061,
 2061, 2061, 2061, 2061,    0, 2061, 2061, 2061, 2061, 2061,
 2061, 2061, 2061, 2061, 2061, 2061, 2061, 2061, 2061, 2061,
 2061, 2061, 2061, 2061, 2061, 2061,    0, 1595, 2061, 2061,
 1595,    0, 1595,    0, 2061, 2061, 2061, 2061, 2061, 2061,
    0,    0,    0,    0,    0, 1595,    0,    0,    0,    0,
 1595,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1595,    0, 1595,    0,    0,    0,
    0,    0, 1595,    0, 1595, 1595, 1595,    0,    0,    0,
    0, 1595,    0,    0,    0,    0, 1595, 1595,    0, 1595,
    0, 1595, 1595,    0, 1595,    0,    0, 1595, 1595, 1595,
    0,    0,    0,    0, 1595,    0, 1595, 1595,    0,    0,
    0, 1595,    0, 1595,    0, 1595, 1595,    0,    0,    0,
    0,    0,    0,    0, 1595,    0, 1595, 1595,    0,    0,
    0, 1595,    0, 1595,    0, 1595,    0,    0,    0,    0,
    0, 1595,    0, 1595,    0,    0,    0, 1595,    0,    0,
    0, 1595,    0, 1595, 1595, 1595, 1595,    0, 1595,    0,
 1595, 1595,    0, 1595, 1595, 1595,    0,    0,    0,    0,
    0, 1595,    0,    0,    0,    0,    0, 1595,    0,    0,
    0, 1595, 1595, 1595, 1595, 1595,    0, 1595,    0, 1595,
    0,    0,    0, 1595,    0, 1595, 1595,    0,    0, 1595,
 1595,    0,    0,    0, 1595,    0, 1595,    0,    0, 1595,
 1595, 1595, 1595,    0, 1595,    0, 1595, 1595, 1595, 1595,
 1595, 1595, 1595, 1595,    0,    0,    0, 1595,    0, 1595,
    0,    0,    0, 1595,    0,    0,    0,    0,    0,    0,
 1595,    0,    0,    0,    0, 1595, 1595, 1595, 1595, 1595,
 1595, 1595, 1595, 1595, 1595, 1595, 1595,    0, 1595, 1595,
 1595, 1595,    0, 1595, 1595,    0,    0,    0,    0,    0,
    0, 1595, 1595, 1595, 1595, 1595, 1595, 1595, 1595, 1595,
 1595, 1595, 1595, 1595,    0, 1595, 1595, 1595, 1595, 1595,
    0, 1595, 1595, 1595,    0, 1595, 1595, 1595, 1595,    0,
    0, 1595, 1595, 1595, 1595, 1595, 1595, 1595,    0,    0,
    0,    0, 1595, 1595, 1595,    0, 1595, 1595, 1595,    0,
 1595,    0, 1595, 1595, 1595,    0, 1595, 1595,    0, 1595,
 1595, 1595, 1595, 1595, 1595, 1595, 1595, 1595, 1595, 1595,
 1595, 1595, 1595, 1595, 1595, 1595, 1595, 1595, 1595, 1595,
 1595,    0, 1595, 1595, 1595, 1595, 1595, 1595,    0, 1595,
 1595, 1595, 1595,    0, 1595,    0, 1595, 1595, 1595, 1595,
 1595, 1595, 1595, 1595, 1595, 1595, 1595, 1595, 1595, 1595,
 1595, 1595, 1595, 1595,    0, 1595, 1595, 1595, 1595,    0,
 1595, 1595, 1595, 1595, 1595,    0, 1595, 1595, 1595, 1595,
 1595, 1595, 1595, 1595, 1595, 1595, 1595, 1595, 1595,    0,
 1595, 1595, 1595, 1595, 1595, 1595,    0, 1595, 1595, 1595,
 1595, 1595, 1595, 1595, 1595,    0, 1595, 1595, 1595, 1595,
    0, 1595, 1595, 1595, 1595, 1595, 1595,    0,    0, 1595,
 1595, 1595, 1595, 1595, 1595, 1595, 1595, 1595, 1595,    0,
 1595, 1595, 1595, 1595, 1595, 1595, 1595, 1595, 1595, 1595,
 1595,    0,    0, 1595, 1595, 1595, 1595, 1595, 1595, 1595,
 1595, 1595, 1595, 1595, 1595, 1595, 1595, 1595, 1595, 1595,
 1595,    0, 1595, 1595, 1595, 1595, 1595, 1595, 1595, 1595,
 1595,    0, 1595, 1595, 1595, 1595, 1595, 1595, 1595, 1595,
 1595, 1595, 1595, 1595,    0, 1595, 1595, 1595,    0, 1595,
    0, 1595, 1595, 1595, 1595, 1595, 1595, 1595, 1595, 1595,
 1595, 1595, 1595, 1595, 1595, 1595,    0, 1595, 1595, 1595,
 1595, 1595, 1595, 1595, 1595, 1595, 1595, 1595, 1595, 1595,
 1595, 1595, 1595, 1595, 1595, 1595, 1595, 1595,    0, 1392,
 1595, 1595, 1017,    0, 1018,    0, 1595, 1595, 1595, 1595,
 1595, 1595,    0,    0,    0,    0,    0, 1020,    0,    0,
    0,    0, 1021,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  152,    0,  153,    0,
    0,    0,    0,    0,  154,    0,  848,  155,  156,    0,
    0,    0,    0,  849,    0,    0,    0,    0,  157,  158,
    0,  159,    0,  160,  850,    0,  161,    0,    0,  162,
  851,  852,    0,    0,    0,    0,  163,    0,  164,  165,
    0,    0,    0,  166,    0,  167,    0,    0,  168,    0,
    0,    0,    0,    0,    0,    0,  854,    0,  169,  855,
    0,    0,    0,  171,    0,  172,    0,  173,    0,    0,
    0,    0,    0,  174,    0,  175,    0,    0,    0,  176,
    0,    0,    0,  856,    0,  857,  858,  177,  859,    0,
  178,    0,  179,  860,    0,  861,  180,  181,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  182,
    0,    0,    0,  184,  185,  186,  187,  188,    0,  189,
    0,  863,    0,    0,    0,  190,    0,  191,  192,    0,
    0,  193,  194,    0,    0,    0,  864,    0,  195,    0,
    0,  196,  197,  198,  199,    0,  200,    0,  201,  866,
  202,  203,  204,  205,  206,  867,    0,    0,    0,  207,
    0,  208,    0,    0,    0,  868,    0,    0,    0,    0,
    0,    0,  209,    0,    0,    0,    0,  210,  211,  869,
  870,  871,  872,  873,  874,  875,  601,  602,  213,    0,
  214,  215,  216,  217,    0,  218,  876,    0,    0,    0,
    0,    0,    0,  219,  220,  877,  878,  879,  880,  881,
  882,  883,  884,  885,  221,  886,    0,  223,  224,  225,
  887,  888,    0,  889,  890,  891,    0,  228,  229,  892,
  230,    0,    0,  231,  232,    0,    0,    0,  233,  234,
    0,    0,    0,    0,  235,  896,  237,    0,  897,  239,
  240,    0,  241,    0,  898,  899,  900,    0,  901,  902,
    0,  903,  242,  243,  244,  245,  246,  247,  904,  249,
  905,  251,  906,  907,  908,  909,  910,  911,  912,  913,
  914,  915,  916, 1428,  917,  918,  919,  920,  921,  922,
    0,  923,  924,  925,  272,    0,  926,    0,  927,  928,
  929,  930,  931,  279,  280,  932,  282,  933,  934,  935,
  286,  936,  288,  937,  290,  938,    0,  939,  940,  941,
  942,    0,  943,  944,  945,  299,  946,    0,  947,  948,
  303,  949,  305,  306,  950,  308,  309,  310,  311,  312,
  313,    0,  951,  315,  316,  317,  318,  319,    0,  952,
  321,  322,  323,  953,  954,  326,  327,    0,  328,  329,
  955,  956,    0,  330,  331,  332,  957,  958,  959,    0,
    0,  336,  960,  961,  962,  963,  964,  965,  966,  967,
  968,    0,  969,  970,  971,  345,  972,  346,  347,  348,
  349,  350,  351,    0,    0,  973,  974,  975,  976,  977,
  978,  979,  980,  981,  982,  983,  984,  985,  986,  987,
  988,  989,  990,    0,  354,  991,  356,  357,  992,  359,
  360,  361,  993,    0,  363,  364,  365,  366,  367,  368,
  994,  370,  995,  996,  373,  374,    0,  375,  376,  997,
    0,  378,    0,  998,  999,  379, 1000,  381,  382,  383,
 1001, 1002,  386,  387,  388, 1003,  390,  391,    0, 1004,
  393,    0, 1006, 1007, 1008, 1009,  394, 1010, 1011, 1012,
 1013, 1014, 1015,  401,  402,  403,  404,  405,  406,  407,
    0,    0, 1016,  409,    0,    0,    0,    0,  410,  411,
  412,  413,  414,  415,    0,    0,    0,    0,    0,    0,
 2254,    0, 2254,    0,    0,    0,    0,    0, 2254,    0,
    0, 2254, 2254,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 2254, 2254,    0, 2254,    0, 2254,    0,    0,
 2254,    0,    0, 2254,    0,    0,    0,    0,    0,    0,
 2254,    0, 2254, 2254,    0,    0,    0, 2254,    0, 2254,
    0,    0, 2254,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 2254, 2254,    0,    0,    0, 2254,    0, 2254,
    0, 2254, 2254,    0,    0,    0,    0, 2254,    0, 2254,
    0,    0,    0, 2254,    0,    0,    0,    0,    0,    0,
    0, 2254,    0,    0, 2254,    0, 2254,    0,    0,    0,
 2254, 2254,    0, 2254,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 2254,    0, 2254,    0, 2254, 2254, 2254,
 2254, 2254,    0, 2254,    0,    0,    0,    0,    0, 2254,
    0, 2254, 2254,    0,    0, 2254, 2254,    0,    0,    0,
    0,    0, 2254,    0,    0, 2254, 2254, 2254, 2254,    0,
 2254,    0, 2254,    0, 2254, 2254, 2254, 2254, 2254,    0,
    0,    0,    0, 2254,    0, 2254,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2254,    0,    0,    0,
    0, 2254, 2254,    0,    0,    0,    0,    0, 2254,    0,
    0,    0, 2254,    0, 2254, 2254, 2254, 2254,    0, 2254,
    0,    0,    0,    0,    0,    0,    0, 2254, 2254,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2254, 2254,
    0, 2254, 2254, 2254,    0,    0,    0,    0, 2254, 2254,
    0, 2254, 2254,    0, 2254,    0,    0, 2254, 2254,    0,
    0,    0, 2254, 2254,    0,    0,    0,    0, 2254, 2254,
 2254,    0, 2254, 2254, 2254,    0, 2254,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2254, 2254, 2254, 2254,
 2254, 2254, 2254, 2254, 2254, 2254, 2254, 2254, 2254, 2254,
 2254, 2254, 2254, 2254, 2254, 2254, 2254, 1427, 2254, 2254,
 2254, 2254, 2254, 2254,    0, 2254, 2254, 2254, 2254,    0,
 2254,    0, 2254, 2254, 2254, 2254, 2254, 2254, 2254, 2254,
 2254, 2254, 2254, 2254, 2254, 2254, 2254, 2254, 2254, 2254,
    0, 2254, 2254, 2254, 2254,    0, 2254, 2254, 2254, 2254,
 2254,    0, 2254, 2254, 2254, 2254, 2254, 2254, 2254, 2254,
 2254, 2254, 2254, 2254, 2254,    0, 2254, 2254, 2254, 2254,
 2254, 2254,    0, 2254, 2254, 2254, 2254, 2254, 2254, 2254,
 2254,    0, 2254, 2254,    0,    0,    0, 2254, 2254, 2254,
 2254, 2254, 2254,    0,    0, 2254, 2254, 2254, 2254, 2254,
 2254, 2254, 2254, 2254,    0,    0,    0,    0,    0, 2254,
    0, 2254, 2254, 2254, 2254, 2254, 2254,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 2254, 2254,    0, 2254, 2254,
 2254, 2254, 2254, 2254, 2254, 2254, 2254,    0, 2254, 2254,
 2254, 2254, 2254, 2254, 2254, 2254, 2254, 2254, 2254, 2254,
    0, 2254, 2254, 2254,    0, 2254,    0,    0,    0, 2254,
 2254, 2254, 2254, 2254, 2254, 2254, 2254, 2254, 2254, 2254,
 2254, 2254,    0, 2254, 2254,    0,    0,    0,    0,    0,
 2254, 2254, 2254, 2254, 2254, 2254, 2254, 2254, 2254, 2254,
 2254, 2254, 2254, 2254,    0,    0, 2254, 2254,    0,    0,
    0,    0, 2254, 2254, 2254, 2254, 2254, 2254,    0,    0,
    0,    0,    0,    0, 2159,    0, 2159,    0,    0,    0,
    0,    0, 2159,    0,    0, 2159, 2159,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2159, 2159,    0, 2159,
    0, 2159,    0,    0, 2159,    0,    0, 2159,    0,    0,
    0,    0,    0,    0, 2159,    0, 2159, 2159,    0,    0,
    0, 2159,    0, 2159,    0,    0, 2159,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2159, 2159,    0,    0,
    0, 2159,    0, 2159,    0, 2159, 2159,    0,    0,    0,
    0, 2159,    0, 2159,    0,    0,    0, 2159,    0,    0,
    0,    0,    0,    0,    0, 2159,    0,    0, 2159,    0,
 2159,    0,    0,    0, 2159, 2159,    0, 2159,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 2159,    0, 2159,
    0, 2159, 2159, 2159, 2159, 2159,    0, 2159,    0,    0,
    0,    0,    0, 2159,    0, 2159, 2159,    0,    0, 2159,
 2159,    0,    0,    0,    0,    0, 2159,    0,    0, 2159,
 2159, 2159, 2159,    0, 2159,    0, 2159,    0, 2159, 2159,
 2159, 2159, 2159,    0,    0,    0,    0, 2159,    0, 2159,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2159,    0,    0,    0,    0, 2159, 2159,    0,    0,    0,
    0,    0, 2159,    0,    0,    0, 2159,    0, 2159, 2159,
 2159, 2159,    0, 2159,    0,    0,    0,    0,    0,    0,
    0, 2159, 2159,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 2159, 2159,    0, 2159, 2159, 2159,    0,    0,
    0,    0, 2159, 2159,    0, 2159, 2159,    0, 2159,    0,
    0, 2159, 2159,    0,    0,    0, 2159, 2159,    0,    0,
    0,    0, 2159, 2159, 2159,    0, 2159, 2159, 2159,    0,
 2159,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2159, 2159, 2159, 2159, 2159, 2159, 2159, 2159, 2159, 2159,
 2159, 2159, 2159, 2159, 2159, 2159, 2159, 2159, 2159, 2159,
 2159, 1460, 2159, 2159, 2159, 2159, 2159, 2159,    0, 2159,
 2159, 2159, 2159,    0, 2159,    0, 2159, 2159, 2159, 2159,
 2159, 2159, 2159, 2159, 2159, 2159, 2159, 2159, 2159, 2159,
 2159, 2159, 2159, 2159,    0, 2159, 2159, 2159, 2159,    0,
 2159, 2159, 2159, 2159, 2159,    0, 2159, 2159, 2159, 2159,
 2159, 2159, 2159, 2159, 2159, 2159, 2159, 2159, 2159,    0,
 2159, 2159, 2159, 2159, 2159, 2159,    0, 2159, 2159, 2159,
 2159, 2159, 2159, 2159, 2159,    0, 2159, 2159,    0,    0,
    0, 2159, 2159, 2159, 2159, 2159, 2159,    0,    0, 2159,
 2159, 2159, 2159, 2159, 2159, 2159, 2159, 2159,    0,    0,
    0,    0,    0, 2159,    0, 2159, 2159, 2159, 2159, 2159,
 2159,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 2159,
 2159,    0, 2159, 2159, 2159, 2159, 2159, 2159, 2159, 2159,
 2159,    0, 2159, 2159, 2159, 2159, 2159, 2159, 2159, 2159,
 2159, 2159, 2159, 2159,    0, 2159, 2159, 2159,    0, 2159,
    0,    0,    0, 2159, 2159, 2159, 2159, 2159, 2159, 2159,
 2159, 2159, 2159, 2159, 2159, 2159,    0, 2159, 2159,    0,
    0,    0,    0,    0, 2159, 2159, 2159, 2159, 2159, 2159,
 2159, 2159, 2159, 2159, 2159, 2159, 2159, 2159,    0,    0,
 2159, 2159,    0,    0,    0,    0, 2159, 2159, 2159, 2159,
 2159, 2159,    0,    0,    0,    0,    0,    0,  152,    0,
  153,    0,    0,    0,    0,    0,  154,    0,    0,  155,
  156,    0,    0,    0,    0,    0,    0,    0,    0,    0,
  157,  158,    0,  159,    0,  160,    0,    0,  161,    0,
    0,  162,    0,    0,    0,    0,    0,    0,  163,    0,
  164,  165,    0,    0,    0,  166,    0,  167,    0,    0,
  168,    0,    0,    0,    0,    0,    0,    0,    0,    0,
  169,  170,    0,    0,    0,  171,    0,  172,    0,  173,
    0,    0,    0,    0,    0,  174,    0,  175,    0,    0,
    0,  176,    0,    0,    0,    0,    0,    0,    0,  177,
    0,    0,  178,    0,  179,    0,    0,    0,  180,  181,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  182,    0,    0,    0,  184,  185,  186,  187,  188,
    0,  189,    0,    0,    0,    0,    0,  190,    0,  191,
  192,    0,    0,  193,  194,    0,    0,    0,    0,    0,
  195,    0,    0,  196,  197,  198,  199,    0,  200,    0,
  201,    0,  202,  203,  204,  205,  206,    0,    0,    0,
    0,  207,    0,  208,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  209,    0,    0,    0,   15,  210,
  211,    0,    0,    0,    0,    0,  212,    0,    0,    0,
  213,    0,  214,  215,  216,  217,    0,  218,    0,    0,
    0,    0,    0,    0,    0,  219,  220,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  221,  222,    0,  223,
  224,  225,    0,    0,    0,    0,  226,  227,    0,  228,
  229,    0,  230,    0,    0,  231,  232,    0,    0,    0,
  233,  234,    0,    0,    0,    0,  235,  236,  237,    0,
  238,  239,  240,    0,  241,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  242,  243,  244,  245,  246,  247,
  248,  249,  250,  251,  252,  253,  254,  255,  256,  257,
  258,  259,  260,  261,  262, 1460,  263,  264,  265,  266,
  267,  268,    0,  269,  270,  271,  272,    0,  273,    0,
  274,  275,  276,  277,  278,  279,  280,  281,  282,  283,
  284,  285,  286,  287,  288,  289,  290,  291,    0,  292,
  293,  294,  295,    0,  296,  297,  298,  299,  300,    0,
  301,  302,  303,  304,  305,  306,  307,  308,  309,  310,
  311,  312,  313,    0,  314,  315,  316,  317,  318,  319,
    0,  320,  321,  322,  323,  324,  325,  326,  327,    0,
  328,  329,    0,    0,    0,  330,  331,  332,  333,  334,
  335,    0,    0,  336,  337,  338,  339,  340,  341,  342,
  343,  344,    0,    0,    0,    0,    0,  345,    0,  346,
  347,  348,  349,  350,  351,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  352,  353,    0,  354,  355,  356,  357,
  358,  359,  360,  361,  362,    0,  363,  364,  365,  366,
  367,  368,  369,  370,  371,  372,  373,  374,    0,  375,
  376,  377, 1459,  378,    0,    0,    0,  379,  380,  381,
  382,  383,  384,  385,  386,  387,  388,  389,  390,  391,
    0,  392,  393,    0,    0,    0,    0,    0,  394,  395,
  396,  397,  398,  399,  400,  401,  402,  403,  404,  405,
  406,  407,    0,    0,  408,  409,    0,    0,    0,    0,
  410,  411,  412,  413,  414,  415,    0,    0,    0,    0,
    0,    0,  152,    0,  153,    0,    0,    0,    0,    0,
  154,    0,    0,  155,  156,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  157,  158,    0,  159,    0,  160,
    0,    0,  161,    0,    0,  162,    0,    0,    0,    0,
    0,    0,  163,    0,  164,  165,    0,    0,    0,  166,
    0,  167,    0,    0,  168,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  169,  170,    0,    0,    0,  171,
    0,  172,    0,  173,    0,    0,    0,    0,    0,  174,
    0,  175,    0,    0,    0,  176,    0,    0,    0,    0,
    0,    0,    0,  177,    0,    0,  178,    0,  179,    0,
    0,    0,  180,  181,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  182,    0,    0,    0,  184,
  185,  186,  187,  188,    0,  189,    0,    0,    0,    0,
    0,  190,    0,  191,  192,    0,    0,  193,  194,    0,
    0,    0,    0,    0,  195,    0,    0,  196,  197,  198,
  199,    0,  200,    0,  201,    0,  202,  203,  204,  205,
  206,    0,    0,    0,    0,  207,    0,  208,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  209,    0,
    0,    0,    0,  210,  211,    0,    0,    0,    0,    0,
  212,    0,    0,    0,  213,    0,  214,  215,  216,  217,
    0,  218,    0,    0,    0,    0,    0,    0,    0,  219,
  220,    0,    0,    0,    0,    0,    0,    0,    0,    0,
  221,  222,    0,  223,  224,  225,    0,    0,    0,    0,
  226,  227,    0,  228,  229,    0,  230,    0,    0,  231,
  232,    0,    0,    0,  233,  234,    0,    0,    0,    0,
  235,  236,  237,    0,  238,  239,  240,    0,  241,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  242,  243,
  244,  245,  246,  247,  248,  249,  250,  251,  252,  253,
  254,  255,  256,  257,  258,  259,  260,  261,  262,    0,
  263,  264,  265,  266,  267,  268,    0,  269,  270,  271,
  272,    0,  273,    0,  274,  275,  276,  277,  278,  279,
  280,  281,  282,  283,  284,  285,  286,  287,  288,  289,
  290,  291,    0,  292,  293,  294,  295,    0,  296,  297,
  298,  299,  300,    0,  301,  302,  303,  304,  305,  306,
  307,  308,  309,  310,  311,  312,  313,    0,  314,  315,
  316,  317,  318,  319,    0,  320,  321,  322,  323,  324,
  325,  326,  327,    0,  328,  329,    0,    0, 1792,  330,
  331,  332,  333,  334,  335,    0,    0,  336,  337,  338,
  339,  340,  341,  342,  343,  344, 1020,    0,    0,    0,
    0,  345,    0,  346,  347,  348,  349,  350,  351,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  352,  353,    0,
  354,  355,  356,  357,  358,  359,  360,  361,  362,    0,
  363,  364,  365,  366,  367,  368,  369,  370,  371,  372,
  373,  374,    0,  375,  376,  377, 1459,  378,    0,    0,
    0,  379,  380,  381,  382,  383,  384,  385,  386,  387,
  388,  389,  390,  391,    0,  392,  393,    0,    0,    0,
    0,    0,  394,  395,  396,  397,  398,  399,  400,  401,
  402,  403,  404,  405,  406,  407,    0,    0,  408,  409,
    0,    0,    0,    0,  410,  411,  412,  413,  414,  415,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  152,    0,  153,    0,    0,
    0,    0,    0,  154,    0,    0,  155,  156,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  157,  158,    0,
  159,    0,  160,    0,    0,  161,    0,    0,  162,    0,
    0,    0,    0,    0,    0,  163,    0,  164,  165,    0,
    0,    0,  166,    0,  167,    0,    0,  168,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  169,  170,    0,
    0,    0,  171,    0,  172,    0,  173,    0,    0,    0,
    0,    0,  174,    0,  175,    0,    0,    0,  176,    0,
    0,    0,    0,    0,    0,    0,  177,    0,    0,  178,
    0,  179,    0,    0,    0,  180,  181,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  182,    0,
    0,    0,  184,  185,  186,  187,  188,    0,  189,    0,
    0,    0,    0,    0,  190,    0,  191,  192,    0,    0,
  193,  194,    0,    0,    0,    0,    0,  195,    0,    0,
  196,  197,  198,  199,    0,  200,    0,  201,    0,  202,
  203,  204,  205,  206,    0,    0,    0,    0,  207,    0,
  208,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  209,    0,    0,    0,    0,  210,  211,    0,    0,
    0,    0,    0,  212,    0,    0,    0,  213,    0,  214,
  215,  216,  217,    0,  218,    0,    0,    0,    0,    0,
    0,    0,  219,  220,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  221,  222,    0,  223,  224,  225,    0,
    0,    0,    0,  226,  227,    0,  228,  229,    0,  230,
    0,    0,  231,  232,    0,    0,    0,  233,  234,    0,
    0,    0,    0,  235,  236,  237,    0,  238,  239,  240,
    0,  241,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  242,  243,  244,  245,  246,  247,  248,  249,  250,
  251,  252,  253,  254,  255,  256,  257,  258,  259,  260,
  261,  262, 1834,  263,  264,  265,  266,  267,  268,    0,
  269,  270,  271,  272,    0,  273,    0,  274,  275,  276,
  277,  278,  279,  280,  281,  282,  283,  284,  285,  286,
  287,  288,  289,  290,  291,    0,  292,  293,  294,  295,
    0,  296,  297,  298,  299,  300,    0,  301,  302,  303,
  304,  305,  306,  307,  308,  309,  310,  311,  312,  313,
    0,  314,  315,  316,  317,  318,  319,    0,  320,  321,
  322,  323,  324,  325,  326,  327,    0,  328,  329,    0,
    0,    0,  330,  331,  332,  333,  334,  335,    0,    0,
  336,  337,  338,  339,  340,  341,  342,  343,  344,    0,
    0,    0,    0,    0,  345,    0,  346,  347,  348,  349,
  350,  351,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
  352,  353,    0,  354,  355,  356,  357,  358,  359,  360,
  361,  362,    0,  363,  364,  365,  366,  367,  368,  369,
  370,  371,  372,  373,  374,    0,  375,  376,  377,    0,
  378,    0,    0,    0,  379,  380,  381,  382,  383,  384,
  385,  386,  387,  388,  389,  390,  391,    0,  392,  393,
    0,    0,    0,    0,    0,  394,  395,  396,  397,  398,
  399,  400,  401,  402,  403,  404,  405,  406,  407,    0,
    0,  408,  409,    0,    0,    0,    0,  410,  411,  412,
  413,  414,  415,    0,    0,    0,    0,    0,    0,  152,
    0,  153,    0,    0,    0,    0,    0,  154,    0,    0,
  155,  156,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  157,  158,    0,  159,    0,  160,    0,    0,  161,
    0,    0,  162,    0,    0,    0,    0,    0,    0,  163,
    0,  164,  165,    0,    0,    0,  166,    0,  167,    0,
    0,  168,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  169,  170,    0,    0,    0,  171,    0,  172,    0,
  173,    0,    0,    0,    0,    0,  174,    0,  175,    0,
    0,    0,  176,    0,    0,    0,    0,    0,    0,    0,
  177,    0,    0,  178,    0,  179,    0,    0,    0,  180,
  181,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,  182,    0,    0,    0,  184,  185,  186,  187,
  188,    0,  189,    0,    0,    0,    0,    0,  190,    0,
  191,  192,    0,    0,  193,  194,    0,    0,    0,    0,
    0,  195,    0,    0,  196,  197,  198,  199,    0,  200,
    0,  201,    0,  202,  203,  204,  205,  206,    0,    0,
    0,    0,  207,    0,  208,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  209,    0,    0,    0,    0,
  210,  211,    0,    0,    0,    0,    0,  212,    0,    0,
    0,  213,    0,  214,  215,  216,  217,    0,  218,    0,
    0,    0,    0,    0,    0,    0,  219,  220,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  221,  222,    0,
  223,  224,  225,    0,    0,    0,    0,  226,  227,    0,
  228,  229,    0,  230,    0,    0,  231,  232,    0,    0,
    0,  233,  234,    0,    0,    0,    0,  235,  236,  237,
    0,  238,  239,  240,    0,  241,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  242,  243,  244,  245,  246,
  247,  248,  249,  250,  251,  252,  253,  254,  255,  256,
  257,  258,  259,  260,  261,  262,  782,  263,  264,  265,
  266,  267,  268,    0,  269,  270,  271,  272,    0,  273,
    0,  274,  275,  276,  277,  278,  279,  280,  281,  282,
  283,  284,  285,  286,  287,  288,  289,  290,  291,    0,
  292,  293,  294,  295,    0,  296,  297,  298,  299,  300,
    0,  301,  302,  303,  304,  305,  306,  307,  308,  309,
  310,  311,  312,  313,    0,  314,  315,  316,  317,  318,
  319,    0,  320,  321,  322,  323,  324,  325,  326,  327,
    0,  328,  329,    0,    0,    0,  330,  331,  332,  333,
  334,  335,    0,    0,  336,  337,  338,  339,  340,  341,
  342,  343,  344,    0,    0,    0,    0,    0,  345,    0,
  346,  347,  348,  349,  350,  351,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  352,  353,    0,  354,  355,  356,
  357,  358,  359,  360,  361,  362,    0,  363,  364,  365,
  366,  367,  368,  369,  370,  371,  372,  373,  374,    0,
  375,  376,  377,    0,  378,    0,    0,    0,  379,  380,
  381,  382,  383,  384,  385,  386,  387,  388,  389,  390,
  391,    0,  392,  393,    0,    0,    0,    0,    0,  394,
  395,  396,  397,  398,  399,  400,  401,  402,  403,  404,
  405,  406,  407,    0,    0,  408,  409,    0,    0,    0,
    0,  410,  411,  412,  413,  414,  415,    0,    0,    0,
    0,    0,    0,  152,    0,  153,    0,    0,    0,    0,
    0,  154,    0,    0,  155,  156,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  157, 3100,    0,  159,    0,
  160,    0,    0,  161,    0,    0,  162,    0,    0,    0,
    0,    0,    0,  163,    0,  164,  165,    0,    0,    0,
  166,    0,  167,    0,    0,  168,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  169,  170,    0,    0,    0,
  171,    0,  172,    0,  173,    0,    0,    0,    0,    0,
  174,    0,  175,    0,    0,    0,  176,    0,    0,    0,
    0,    0,    0,    0,  177,    0,    0,  178,    0,  179,
    0,    0,    0,  180,  181,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  182,    0,    0,    0,
  184,  185,  186,  187,  188,    0,  189,    0,    0,    0,
    0,    0,  190,    0,  191,  192,    0,    0,  193,  194,
    0,    0,    0,    0,    0,  195,    0,    0,  196,  197,
  198,  199,    0,  200,    0,  201,    0,  202,  203,  204,
  205,  206,    0,    0,    0,    0,  207,    0,  208,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  209,
    0,    0,    0,    0,  210,  211,    0,    0,    0,    0,
    0,  212,    0,    0,    0,  213,    0,  214,  215,  216,
  217,    0,  218,    0,    0,    0,    0,    0,    0,    0,
  219,  220,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  221,  222,    0,  223,  224,  225,    0,    0,    0,
    0,  226,  227,    0,  228,  229,    0,  230,    0,    0,
  231,  232,    0,    0,    0,  233,  234,    0,    0,    0,
    0,  235,  236,  237,    0,  238,  239,  240,    0,  241,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  242,
  243,  244,  245,  246,  247,  248,  249,  250,  251,  252,
  253,  254,  255,  256,  257,  258,  259,  260,  261,  262,
 2326,  263,  264,  265,  266,  267,  268,    0,  269,  270,
  271,  272,    0,  273,    0,  274,  275,  276,  277,  278,
  279,  280,  281,  282,  283,  284,  285,  286,  287,  288,
  289,  290,  291,    0,  292,  293,  294,  295,    0,  296,
  297,  298,  299,  300,    0,  301,  302,  303,  304,  305,
  306,  307,  308,  309,  310,  311,  312,  313,    0,  314,
  315,  316,  317,  318,  319,    0,  320,  321,  322,  323,
  324,  325,  326,  327,    0,  328,  329,    0,    0,    0,
  330,  331,  332,  333,  334,  335,    0,    0,  336,  337,
  338,  339,  340,  341,  342,  343,  344,    0,    0,    0,
    0,    0,  345,    0,  346,  347,  348,  349,  350,  351,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  352,  353,
    0,  354,  355,  356,  357,  358,  359,  360,  361,  362,
    0,  363,  364,  365,  366,  367,  368,  369,  370,  371,
  372,  373,  374,    0,  375,  376,  377,    0,  378,    0,
    0,    0,  379,  380,  381,  382,  383,  384,  385,  386,
  387,  388,  389,  390,  391,    0,  392,  393,    0,    0,
    0,    0,    0,  394,  395,  396,  397,  398,  399,  400,
  401,  402,  403,  404,  405,  406,  407,    0,    0,  408,
  409,    0,    0,    0,    0,  410,  411,  412,  413,  414,
  415,    0,    0,    0,    0,  152,    0,  153,    0,    0,
    0,    0,    0,  154,    0,    0,  155,  156,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  157,  158,    0,
  159,    0,  160,    0,    0,  161,    0,    0,  162,    0,
  852,    0,    0,    0,    0,  163,    0,  164,  165,    0,
    0,    0,  166,    0,  167,    0,    0,  168,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  169,  170,    0,
    0,    0,  171,    0,  172,    0,  173,    0,    0,    0,
    0,    0,  174,    0,  175,    0,    0,    0,  176,    0,
    0,    0,    0,    0,    0,    0,  177,    0,    0,  178,
    0,  179,    0,    0,    0,  180,  181,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  182,    0,
    0,    0,  184,  185,  186,  187,  188,    0,  189,    0,
    0,    0,    0,    0,  190,    0,  191,  192,    0,    0,
  193,  194,    0,    0,    0,    0,    0,  195,    0,    0,
  196,  197,  198,  199,    0,  200,    0,  201,    0,  202,
  203,  204,  205,  206,    0,    0,    0,    0,  207,    0,
  208,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  209,    0,    0,    0,    0,  210,  211,    0,    0,
    0,    0,    0,  212,    0,    0,    0,  213,    0,  214,
  215,  216,  217,    0,  218,    0,    0,    0,    0,    0,
    0,    0,  219,  220,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  221,  222,    0,  223,  224,  225,    0,
    0,    0,    0,  226,  227,    0,  228,  229,    0,  230,
    0,    0,  231,  232,    0,    0,    0,  233,  234,    0,
    0,    0,    0,  235,  236,  237,    0,  238,  239,  240,
    0,  241,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  242,  243,  244,  245,  246,  247,  248,  249,  250,
  251,  252,  253,  254,  255,  256,  257,  258,  259,  260,
  261,  262, 1299,  263,  264,  265,  266,  267,  268,    0,
  269,  270,  271,  272,    0,  273,    0,  274,  275,  276,
  277,  278,  279,  280,  281,  282,  283,  284,  285,  286,
  287,  288,  289,  290,  291,    0,  292,  293,  294,  295,
    0,  296,  297,  298,  299,  300,    0,  301,  302,  303,
  304,  305,  306,  307,  308,  309,  310,  311,  312,  313,
    0,  314,  315,  316,  317,  318,  319,    0,  320,  321,
  322,  323,  324,  325,  326,  327,    0,  328,  329,    0,
    0,    0,  330,  331,  332,  333,  334,  335,    0,    0,
  336,  337,  338,  339,  340,  341,  342,  343,  344,    0,
    0,    0,    0,    0,  345,  972,  346,  347,  348,  349,
  350,  351,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
  352,  353,    0,  354,  355,  356,  357,  358,  359,  360,
  361,  362,    0,  363,  364,  365,  366,  367,  368,  369,
  370,  371,  372,  373,  374,    0,  375,  376,  377,    0,
  378,    0,    0,    0,  379,  380,  381,  382,  383,  384,
  385,  386,  387,  388,  389,  390,  391,    0,  392,  393,
    0,    0,    0,    0,    0,  394,  395,  396,  397,  398,
  399,  400,  401,  402,  403,  404,  405,  406,  407,    0,
    0,  408,  409,    0,    0,    0,    0,  410,  411,  412,
  413,  414,  415,    0,  152,    0,  153,    0,    0,    0,
    0,    0,  154,    0,    0,  155,  156,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  157,  158,    0,  159,
    0,  160,    0,    0,  161,    0,    0,  162,    0,    0,
    0,    0,    0,    0,  163,    0,  164,  165,    0,    0,
    0,  166,    0,  167,    0,    0,  168,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  169,  170,    0,    0,
    0,  171,    0,  172,    0,  173,    0,    0,    0,    0,
    0,  174,    0,  175,    0,    0,    0,  176,    0,    0,
    0,    0,    0,    0,    0,  177,    0,    0,  178,    0,
  179,    0,    0,    0,  180,  181,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  182,    0,    0,
    0,  184,  185,  186,  187,  188,    0,  189,    0,    0,
    0,    0,    0,  190,    0,  191,  192,    0,    0,  193,
  194,    0,    0,    0,    0,    0,  195,    0,    0,  196,
  197,  198,  199,    0,  200,    0,  201,    0,  202,  203,
  204,  205,  206,    0,    0,    0,    0,  207,    0,  208,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
  209,    0,    0,    0,    0,  210,  211,    0,    0,    0,
    0,    0,  212,  617,    0,    0,  213,    0,  214,  215,
  216,  217,    0,  218,    0,    0,    0,    0,    0,    0,
    0,  219,  220,    0,    0,    0,    0,    0,    0,    0,
    0,    0,  221,  222,    0,  223,  224,  225,    0,    0,
    0,    0,  226,  227,    0,  228,  229,    0,  230,    0,
    0,  231,  232,    0,    0,    0,  233,  234,    0,    0,
    0,    0,  235,  236,  237,    0,  238,  239,  240,    0,
  241,    0,    0,    0,    0,    0,    0,    0,    0,    0,
  242,  243,  244,  245,  246,  247,  248,  249,  250,  251,
  252,  253,  254,  255,  256,  257,  258,  259,  260,  261,
  262,    0,  263,  264,  265,  266,  267,  268,    0,  269,
  270,  271,  272,    0,  273,    0,  274,  275,  276,  277,
  278,  279,  280,  281,  282,  283,  284,  285,  286,  287,
  288,  289,  290,  291,    0,  292,  293,  294,  295,    0,
  296,  297,  298,  299,  300,    0,  301,  302,  303,  304,
  305,  306,  307,  308,  309,  310,  311,  312,  313,    0,
  314,  315,  316,  317,  318,  319,    0,  320,  321,  322,
  323,  324,  325,  326,  327,    0,  328,  329, 1424, 1423,
 1425,  330,  331,  332,  333,  334,  335,    0,    0,  336,
  337,  338,  339,  340,  341,  342,  343,  344,    0,    0,
    0,    0,    0,  345,    0,  346,  347,  348,  349,  350,
  351,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  352,
  353, 1298,  354,  355,  356,  357,  358,  359,  360,  361,
  362,    0,  363,  364,  365,  366,  367,  368,  369,  370,
  371,  372,  373,  374,    0,  375,  376,  377,    0,  378,
    0,    0,    0,  379,  380,  381,  382,  383,  384,  385,
  386,  387,  388,  389,  390,  391,    0,  392,  393,    0,
    0,    0,    0,    0,  394,  395,  396,  397,  398,  399,
  400,  401,  402,  403,  404,  405,  406,  407,    0,    0,
  408,  409,    0,    0,    0,    0,  410,  411,  412,  413,
  414,  415, 2263, 2263, 2058, 2263, 2263, 2263,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2263, 2263, 2263,    0,    0, 2313, 2313, 2313, 2313, 2313,
 2313, 2313,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 2313, 2313, 2313,    0,    0,    0,
    0, 2263,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1408,    0,    0,    0,    0, 2313,    0,    0,    0, 1409,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1410,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1411,    0,    0,    0,
    0,    0,    0, 1412,    0,    0,    0,    0,    0,    0,
    0, 1413,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1414,    0,
 1415,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1416,    0, 1417, 1418, 1419,    0,    0,    0,    0, 1420,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1421,    0, 2263,    0,    0,    0, 2263,    0,    0,    0,
    0, 2263,    0,    0,    0,    0,    0, 2263,    0,    0,
    0, 2263,    0,    0, 2263,    0, 2313,    0,    0,    0,
 2313,    0,    0,    0,    0, 2313,    0,    0,    0,    0,
    0, 2313,    0,    0,    0, 2313, 1697, 1698, 1699, 1700,
 1701, 1702, 1703, 1704,    0,    0,    0,    0, 2263,    0,
    0,    0,    0,    0,    0, 2263,    0,    0,    0,    0,
    0,    0,    0, 2263,    0,    0,    0,  684,    0,    0,
    0,    0, 2313,    0,    0,    0,    0,    0,    0, 2313,
 2263,    0, 2263,    0,    0,    0,    0, 2313,    0,    0,
    0,    0, 2263,    0, 2263, 2263, 2263,    0,    0,    0,
    0, 2263,    0,    0, 2313,    0, 2313,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2313,    0, 2313, 2313,
 2313,    0,    0,    0,    0, 2313,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 2263,    0,    0, 1705,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1706,    0,    0,    0,    0,    0,    0,    0,    0,
 1422,    0,    0,    0,    0,    0,    0,    0, 2263, 2263,
 2263, 2263, 2263, 2263, 2263, 2263,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  684,    0,  684,    0,
  684,    0,    0,    0,  684,    0,    0,  684,  684,  684,
    0,    0,    0,    0,    0,    0,    0,    0,  684,  684,
    0,  684,    0,  684,    0,    0,  684, 2263,    0,  684,
    0,    0,    0,    0,    0,  684,  684,    0,  684,  684,
    0,    0,  684,  684,    0,  684,  684,    0,  684, 1707,
 1708,  684,    0, 2263,    0,    0,    0,    0,  684,  684,
    0,    0, 2263,  684,  684,  684,    0,  684,    0,  684,
    0,    0,    0,  684,    0,  684,    0,    0,    0,  684,
    0,    0,    0,    0,    0,  684, 2313,  684,  684,    0,
  684,    0,  684, 2313,  684,    0,  684,  684,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  684,
    0,    0,    0,  684,  684,  684,  684,  684,    0,  684,
    0,  684,  684,    0,    0,  684,    0,  684,  684,    0,
    0,  684,  684,    0,    0,    0,  684,    0,  684,  684,
  684,  684,  684,  684,  684,    0,  684,    0,  684,    0,
  684,  684,  684,  684,  684,    0,    0,    0,    0,  684,
    0,  684,    0,    0,  684,    0,    0,    0,    0,    0,
    0,    0,  684,  684,    0,  684,  684,  684,  684,    0,
    0,    0,    0,    0,  684,    0,    0,    0,  684,    0,
  684,  684,  684,  684,    0,  684,  684,    0,    0,    0,
    0, 2263, 2263,  684,  684,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  684,  684,    0,  684,  684,  684,
    0,    0,    0,    0,  684,  684,    0,  684,  684,    0,
  684,    0,    0,  684,  684,    0,    0,    0,  684,  684,
  684,  684,  684,    0,  684,  684,  684,    0,  684,  684,
  684,    0,  684,    0,  684,  684,  684,    0,  684,  684,
    0,  684,  684,  684,  684,  684,  684,  684,  684,  684,
  684,  684,  684,  684,  684,  684,  684,  684,  684,  684,
  684,  684,  684,    0,  684,  684,  684,  684,  684,  684,
    0,  684,  684,  684,  684, 2388,  684,    0,  684,  684,
  684,  684,  684,  684,  684,  684,  684,  684,  684,  684,
  684,  684,  684,  684,  684,  684,    0,  684,  684,  684,
  684,    0,  684,  684,  684,  684,  684,    0,  684,  684,
  684,  684,  684,  684,  684,  684,  684,  684,  684,  684,
  684,    0,  684,  684,  684,  684,  684,  684,    0,  684,
  684,  684,  684,  684,  684,  684,  684,    0,  684,  684,
  684,  684,    0,  684,  684,  684,  684,  684,  684,  684,
    0,  684,  684,  684,  684,  684,  684,  684,  684,  684,
    0,    0,    0,    0,    0,  684,    0,  684,  684,  684,
  684,  684,  684,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  684,  684,    0,  684,  684,  684,  684,  684,  684,
  684,  684,  684,    0,  684,  684,  684,  684,  684,  684,
  684,  684,  684,  684,  684,  684,    0,  684,  684,  684,
    0,  684,    0,    0,    0,  684,  684,  684,  684,  684,
  684,  684,  684,  684,  684,  684,  684,  684,    0,  684,
  684,    0,    0,  684,  684,  684,  684,  684,  684,  684,
  684,  684,  684,  684,  684,  684,  684,  684,  684,  684,
    0,    0,  684,  684,    0,    0,    0,    0,  684,  684,
  684,  684,  684,  684,  152,    0,  153,    0, 2372,    0,
    0,    0,  154,    0,    0,  155,  156, 2010,    0,    0,
    0,    0,    0,    0,    0,    0,  157,  158,    0,  159,
    0,  160,    0,    0,  161,    0,    0,  162,    0,    0,
    0,    0,    0,    5,  163,    0,  164,  165,    0,    0,
    0,  166,    0, 2373, 2374,    0, 2375,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  169,  170,    0,    0,
    0, 2376, 2377,  172,    0,  173,    0,    9,    0,    0,
    0,  174,    0,  175,    0,    0,    0,  176,    0,    0,
    0,    0,    0,  857,    0,  177,  859,    0,  178,    0,
  179,    0,   10,    0,  180,  181,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  182,    0,    0,
    0,  184,  185,  186,  187,  188,    0,  189,    0,  863,
 2378,    0,    0,  190,    0,  191,  192,    0,    0,  193,
  194,    0,    0,    0,  864,    0,  195,    0, 2379,  196,
  197,  198,  199,    0,  200,    0,  201,    0,  202,  203,
  204,  205, 2380,    0,    0,    0,    0,  207,    0,  208,
    0,    0,   14,    0,    0,    0,    0,    0,    0,    0,
  209, 2754,    0,    0,   15,  210,  211,    0,    0,    0,
    0,    0,  874,    0,    0,    0,  213,    0,  214,  215,
  216,  217,    0,  218,  876,    0,    0,    0,    0,    0,
    0,  219,  220,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 2381,  886,    0,  223,  224,  225,    0,    0,
    0,    0,  226,  227,    0,  228,  229,    0,  230,    0,
    0,  231, 2382,    0,    0,    0,  233,  234, 2383, 2384,
 2385,    0,  235,  236,  237,    0,  238,  239,  240,    0,
  241,    0,  898,  899,  900,    0,  901,  902,    0,  903,
  242,  243,  244,  245,  246,  247,  904,  249,  905,  251,
  906,  907,  908,  909,  910,  911,  912,  913,  914,  915,
  916,    0,  917,  918,  919,  920,  921,  268,    0,  923,
  924,  925,  272, 2388,  926,    0,  274,  928,  929,  930,
  931,  279,  280,  932,  282,  933,  934,  935,  286,  936,
  288,  937,  290,  938,    0,  939,  940,  941,  942,    0,
  943,  944,  945,  299,  946,    0,  947,  948,  303,  949,
  305,  306,  950,  308,  309,  310,  311,  312,  313,    0,
  951,  315,  316,  317,  318,  319,    0,  952,  321, 2386,
  323,  953,  954,  326,  327,    0,  328,  329,  955,  956,
    0,  330,  331,  332,  957,  958,  959, 2387,    0,  336,
  337,  338,  339,  340,  341,  342,  343,  344,    0,    0,
    0,    0,    0,  345,    0,  346,  347,  348,  349,  350,
  351,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  989,
  990,    0,  354,  991,  356,  357,  992,  359,  360,  361,
  362,    0,  363,  364,  365,  366,  367,  368,  994,  370,
  995,  996,  373,  374,    0,  375,  376,  997,    0,  378,
    0,    0,    0,  379, 1000,  381,  382,  383, 1001,  385,
  386,  387,  388,  389,  390,  391,    0, 1004,  393,    0,
    0, 1007, 1008, 1009,  394, 1010, 1011, 1012, 1013, 1014,
 1015,  401,  402,  403,  404,  405,  406,  407,    0,    0,
 1016,  409,    0,    0,    0,    0,  410,  411,  412,  413,
  414,  415,  152,    0,  153,    0, 2372,    0,    0,    0,
  154,    0,    0,  155,  156, 2010,    0,    0,    0,    0,
    0,    0,    0,    0,  157,  158,    0,  159,    0,  160,
    0,    0,  161,    0,    0,  162,    0,    0,    0,    0,
    0,    5,  163,    0,  164,  165,    0,    0,    0,  166,
    0, 2373, 2374,    0, 2375,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  169,  170,    0,    0,    0, 2376,
 2377,  172,    0,  173,    0,    9,    0,    0,    0,  174,
    0,  175,    0,    0,    0,  176,    0,    0,    0,    0,
    0,  857,    0,  177,  859,    0,  178,    0,  179,    0,
   10,    0,  180,  181,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  182,    0,    0,    0,  184,
  185,  186,  187,  188,    0,  189,    0,  863, 2378,    0,
    0,  190,    0,  191,  192,    0,    0,  193,  194,    0,
    0,    0,  864,    0,  195,    0, 2379,  196,  197,  198,
  199,    0,  200,    0,  201,    0,  202,  203,  204,  205,
 2380,    0,    0,    0,    0,  207,    0,  208,    0,    0,
   14,    0,    0,    0,    0,    0,    0,    0,  209,    0,
    0,    0,   15,  210,  211,    0,    0,    0,    0,    0,
  874,    0,    0,    0,  213,    0,  214,  215,  216,  217,
    0,  218,  876,    0,    0,    0,    0,    0,    0,  219,
  220,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2381,  886,    0,  223,  224,  225,    0,    0,    0,    0,
  226,  227,    0,  228,  229,    0,  230,    0,    0,  231,
 2382,    0,    0,    0,  233,  234, 2383, 2384, 2385,    0,
  235,  236,  237,    0,  238,  239,  240,    0,  241,    0,
  898,  899,  900,    0,  901,  902,    0,  903,  242,  243,
  244,  245,  246,  247,  904,  249,  905,  251,  906,  907,
  908,  909,  910,  911,  912,  913,  914,  915,  916,    0,
  917,  918,  919,  920,  921,  268,    0,  923,  924,  925,
  272, 1020,  926,    0,  274,  928,  929,  930,  931,  279,
  280,  932,  282,  933,  934,  935,  286,  936,  288,  937,
  290,  938,    0,  939,  940,  941,  942,    0,  943,  944,
  945,  299,  946,    0,  947,  948,  303,  949,  305,  306,
  950,  308,  309,  310,  311,  312,  313,    0,  951,  315,
  316,  317,  318,  319,    0,  952,  321, 2386,  323,  953,
  954,  326,  327,    0,  328,  329,  955,  956,    0,  330,
  331,  332,  957,  958,  959, 2387,    0,  336,  337,  338,
  339,  340,  341,  342,  343,  344,    0,    0,    0,    0,
    0,  345,    0,  346,  347,  348,  349,  350,  351,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  989,  990,    0,
  354,  991,  356,  357,  992,  359,  360,  361,  362,    0,
  363,  364,  365,  366,  367,  368,  994,  370,  995,  996,
  373,  374,    0,  375,  376,  997,    0,  378,    0,    0,
    0,  379, 1000,  381,  382,  383, 1001,  385,  386,  387,
  388,  389,  390,  391,    0, 1004,  393,    0,    0, 1007,
 1008, 1009,  394, 1010, 1011, 1012, 1013, 1014, 1015,  401,
  402,  403,  404,  405,  406,  407,    0,    0, 1016,  409,
    0,    0,    0,    0,  410,  411,  412,  413,  414,  415,
  152,    0,  153,    0,    0,    0,    0,    0,  154,    0,
    0,  155,  156,    0,    0,    0,    0,    0,    0,    0,
    0,    0,  157,  158,    0,  159,    0,  160,    0,    0,
  161,    0,    0,  162,    0,    0,    0,    0,    0,    0,
  163,    0,  164,  165,    0,    0,    0,  166,    0,  167,
    0,    0,  168,    0,    0,    0,    0,    0,    0,    0,
    0,    0,  169,  170,    0,    0,    0,  171,    0,  172,
    0,  173,    0,    0,    0,    0,    0,  174,    0,  175,
    0,    0,    0,  176,    0,    0,    0,    0,    0,    0,
    0,  177,    0,    0,  178,    0,  179,    0,    0,    0,
  180,  181,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  182,    0,    0,    0,  184,  185,  186,
  187,  188,    0,  189,    0,    0,    0,    0,    0,  190,
    0,  191,  192,    0,    0,  193,  194,    0,    0,    0,
    0,    0,  195,    0,    0,  196,  197,  198,  199,    0,
  200,    0,  201,    0,  202,  203,  204,  205,  206,    0,
    0,    0,    0,  207,    0,  208,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  209,    0,    0,    0,
    0,  210,  211,    0,    0,    0,    0,    0,  212,    0,
    0,    0,  213,    0,  214,  215,  216,  217,    0,  218,
    0,    0,    0,    0,    0,    0,    0,  219,  220,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  221,  222,
    0,  223,  224,  225,    0,    0,    0,    0,  226,  227,
    0,  228,  229,    0,  230,    0,    0,  231,  232,    0,
    0,    0,  233,  234,    0,    0,    0,    0,  235,  236,
  237,    0,  238,  239,  240,    0,  241,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  242,  243,  244,  245,
  246,  247,  248,  249,  250,  251,  252,  253,  254,  255,
  256,  257,  258,  259,  260,  261,  262,    0,  263,  264,
  265,  266,  267,  268,    0,  269,  270,  271,  272,    0,
  273,    0,  274,  275,  276,  277,  278,  279,  280,  281,
  282,  283,  284,  285,  286,  287,  288,  289,  290,  291,
    0,  292,  293,  294,  295,    0,  296,  297,  298,  299,
  300,    0,  301,  302,  303,  304,  305,  306,  307,  308,
  309,  310,  311,  312,  313,    0,  314,  315,  316,  317,
  318,  319,    0,  320,  321,  322,  323,  324,  325,  326,
  327,    0,  328,  329,    0,    0,    0,  330,  331,  332,
  333,  334,  335,    0,    0,  336,  337,  338,  339,  340,
  341,  342,  343,  344,    0,    0,    0,    0,    0,  345,
    0,  346,  347,  348,  349,  350,  351,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  352,  353,    0,  354,  355,
  356,  357,  358,  359,  360,  361,  362,    0,  363,  364,
  365,  366,  367,  368,  369,  370,  371,  372,  373,  374,
    0,  375,  376,  377,    0,  378,    0,    0,    0,  379,
  380,  381,  382,  383,  384,  385,  386,  387,  388,  389,
  390,  391,    0,  392,  393,    0,    0,    0,    0,    0,
  394,  395,  396,  397,  398,  399,  400,  401,  402,  403,
  404,  405,  406,  407,    0,    0,  408,  409,    0,    0,
    0,    0,  410,  411,  412,  413,  414,  415, 2202, 2202,
    0, 2202, 2202, 2202,    0, 1116,    0,    0, 1117,    0,
    0,    0,    0,    0,    0, 2202, 2202, 2202, 2202,    0,
    0,    0,    0,    0, 1118,    0, 1119,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 2202,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1120,    0,    0,    0,    0,    0,
 1121, 1122, 1123,    0, 1124, 1125, 1126,    0, 1127,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1128,    0, 1129,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1130,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1131,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1132,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1252,
 1134,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1135,    0, 2202,    0,
 2202,    0, 2202,    0,    0,    0,    0, 2202,    0,    0,
    0,    0,    0, 2202,    0,    0,    0, 2202,    0,    0,
 2202,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 2202,    0,    0,    0,    0,    0,
    0, 2202,    0,    0,    0,    0,    0,    0, 2202, 2202,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2202,    0, 2202,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2202,    0,
 2202, 2202, 2202,    0, 2202,    0,    0, 2202,    0,    0,
    0,    0,    0,    0,    0,    0, 2202,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2202,  152,
    0,  153,    0,    0,    0, 1136,    0,  154,    0,    0,
  155,  156,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  157,  158, 2202,  159,    0,  160,    0,    0,  161,
    0,    0,  162,    0,    0,    0,    0, 1137,    0,  163,
 2202,  164,  165,    0,    0,    0,  166, 1138,  167,    0,
    0,  168,    0, 1920,    0,    0,    0,    0,    0,    0,
 1139,  169,  170,    0,    0, 1921,  171,    0,  172,    0,
  173,    0,    0, 1922,    0,    0,  174,    0,  175,    0,
    0,    0,  176,    0,    0,    0,    0,    0,    0,    0,
  177,    0,    0,  178,    0,  179,    0,    0,    0,  180,
  181,    0,    0,    0,    0,    0,    0,    0,    0, 1140,
    0,    0,  182,    0,    0,    0,  184,  185,  186,  187,
  188,    0,  189,    0,    0,    0,    0,    0,  190,    0,
  191,  192,    0,    0,  193,  194,    0,    0,    0,    0,
    0,  195,    0,    0,  196,  197,  198,  199,    0,  200,
    0,  201,    0,  202,  203,  204,  205,  206,    0,    0,
    0,    0,  207,    0,  208,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  209,    0,    0,    0,    0,
  210,  211,    0,    0,    0,    0,    0,  212, 2202,    0,
    0,  213,  454,  214,  215,  216,  217,  455,  218,  456,
  457,  458,  459,  460,  461,  462,  219,  220,  463,  464,
  465,  466,  467,    0,    0,  468,  469,  221,  222,  470,
  223,  224,  225,  471,  472,  473,  474,  226,  227,  475,
  228,  229,  476,  230,  477,  478,  231,  232,  479,  480,
  481,  233,  234,  482,  483,  484,  485,  235,  236,  237,
  486,  238,  239,  240,  487,  241,  488,  489,  490,  491,
  492,  493,  494,  495,  496,  242,  243,  244,  245,  246,
  247,  248,  249,  250,  251,  252,  253,  254,  255,  256,
  257,  258,  259,  260,  261,  262,  497,  263,  264,  265,
  266,  267,  268,  498,  269,  270,  271,  272,  499,  273,
  500,  274,  275,  276,  277,  278,  279,  280,  281,  282,
  283,  284,  285,  286,  287,  288,  289,  290,  291,  501,
  292,  293,  294,  295,  502,  296,  297,  298,  299,  300,
  503,  301,  302,  303,  304,  305,  306,  307,  308,  309,
  310,  311,  312,  313,  504,  314,  315,  316,  317,  318,
  319,    0,  320,  321,  322,  323,  324,  325,  326,  327,
  505,  328,  329,    0,    0,  506,  330,  331,  332,  333,
  334,  335,  507,  508,  336,  337,  338,  339,  340,  341,
  342,  343,  344,  509,  510,  511,  512,  513,  345,    0,
  346,  347,  348,  349,  350,  351,  514,  515,  516,  517,
  518,  519,  520,  521,  522,  523,  524,  525,  526,  527,
  528,  529,  530,  531,  352,  353,  532,  354,  355,  356,
  357,  358,  359,  360,  361,  362,  533,  363,  364,  365,
  366,  367,  368,  369,  370,  371,  372,  373,  374,  534,
  375,  376,  377,  535,  378,  536,  537,  538,  379,  380,
  381,  382,  383,  384,  385,  386,  387,  388,  389,  390,
  391,  539,  392,  393,  540,    0,    0,    0,    0,  394,
  395,  396,  397,  398,  399,  400,  401,  402,  403,  404,
  405,  406,  407,  541,  542,  408,  409,  543,  544,  545,
  546,  410,  411,  412,  413,  414,  415,  999,    0,  999,
    0,    0,    0,    0,    0,  999,    0,    0,  999,  999,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  999,
  999,    0,  999,    0,  999,    0,    0,  999,    0,    0,
  999,    0,    0,    0,    0,    0,    0,  999,    0,  999,
  999,    0,    0,    0,  999,    0,  999,    0,    0,  999,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  999,
  999,    0,    0,    0,  999,    0,  999,    0,  999,    0,
    0,    0,    0,    0,  999,    0,  999,    0,    0,    0,
  999,    0,    0,    0,    0,    0,    0,    0,  999,    0,
    0,  999,    0,  999,    0,    0,    0,  999,  999,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
  999,    0,    0,    0,  999,  999,  999,  999,  999,    0,
  999,    0,    0,    0,    0,    0,  999,    0,  999,  999,
    0,    0,  999,  999,    0,    0,    0,    0,    0,  999,
    0,    0,  999,  999,  999,  999,    0,  999,    0,  999,
    0,  999,  999,  999,  999,  999,    0,    0,    0,    0,
  999,    0,  999,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  999,    0,    0,    0,    0,  999,  999,
    0,    0,    0,    0,    0,  999,    0,    0,    0,  999,
  999,  999,  999,  999,  999,    0,  999,  999,  999,  999,
  999,  999,  999,  999,  999,  999,  999,  999,  999,  999,
  999,    0,    0,  999,  999,  999,  999,  999,  999,  999,
  999,  999,  999,  999,  999,  999,  999,  999,  999,  999,
  999,  999,  999,  999,  999,  999,  999,  999,  999,  999,
  999,  999,  999,  999,  999,  999,  999,  999,  999,  999,
  999,  999,  999,  999,  999,  999,  999,  999,  999,  999,
  999,  999,  999,  999,  999,  999,  999,  999,  999,  999,
  999,  999,  999,  999,  999,  999,  999,  999,  999,  999,
  999,  999,  999,  999,  999,  999,  999,  999,  999,  999,
  999,  999,  999,  999,  999,  999,  999,  999,  999,  999,
  999,  999,  999,  999,  999,  999,  999,  999,  999,  999,
  999,  999,  999,  999,  999,  999,  999,  999,  999,  999,
  999,  999,  999,  999,  999,  999,  999,  999,  999,  999,
  999,  999,  999,  999,  999,  999,  999,  999,  999,  999,
  999,  999,  999,  999,  999,  999,  999,  999,  999,    0,
  999,  999,  999,  999,  999,  999,  999,  999,  999,  999,
  999,    0,    0,  999,  999,  999,  999,  999,  999,  999,
  999,  999,  999,  999,  999,  999,  999,  999,  999,  999,
  999,  999,  999,  999,  999,  999,  999,    0,  999,  999,
  999,  999,  999,  999,  999,  999,  999,  999,  999,  999,
  999,  999,  999,  999,  999,  999,  999,  999,  999,  999,
  999,  999,  999,  999,  999,  999,  999,  999,  999,  999,
  999,  999,  999,  999,  999,  999,  999,  999,  999,  999,
  999,  999,  999,  999,  999,  999,  999,  999,  999,  999,
  999,  999,  999,  999,  999,  999,  999,  999,  999,  999,
  999,  999,  999,  999,  999,  999,  999,  999,  999,  999,
  999,  999,  999,    0,    0,    0,    0,  999,  999,  999,
  999,  999,  999,  999,  999,  999,  999,  999,    0,  999,
  999,  999,  999,  999,  999,  999,  999,  999,  999,  999,
  999,  999,  999,  999,  999, 2074,    0, 2074,    0,    0,
    0,    0,    0, 2074,    0,    0, 2074, 2074,    0,    0,
 2074,    0,    0, 2074,    0,    0,    0, 2074, 2074,    0,
 2074,    0, 2074,    0,    0, 2074,    0, 2076, 2074, 2074,
    0, 2074,    0,    0,    0, 2074,    0, 2074, 2074,    0,
    0,    0, 2074,    0, 2074,    0,    0, 2074,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 2074, 2074,    0,
    0,    0, 2074,    0, 2074,    0, 2074,    0,    0, 2074,
    0,    0, 2074,    0, 2074, 2074, 2074, 2074, 2074, 2074,
 2074, 2074,    0, 2074,    0,    0, 2074,    0,    0, 2074,
    0, 2074,    0,    0,    0, 2074, 2074, 2074,    0, 2074,
    0, 2076,    0,    0,    0,    0,    0,    0, 2074,    0,
    0,    0, 2074, 2074, 2074, 2074, 2074,    0, 2074,    0,
    0,    0,    0,    0, 2074,    0, 2074, 2074, 2074,    0,
 2074, 2074,    0,    0,    0,    0,    0, 2074,    0,    0,
 2074, 2074, 2074, 2074, 2074, 2074,    0, 2074,    0, 2074,
 2074, 2074, 2074, 2074,    0,    0,    0,    0, 2074,    0,
 2074,    0,    0,    0,    0,    0, 2074,    0,    0,    0,
    0, 2074,    0,    0,    0,    0, 2074, 2074,    0,    0,
    0,    0,    0, 2074,    0,    0,    0, 2074,    0, 2074,
 2074, 2074, 2074,    0, 2074,    0,    0,    0,    0,    0,
    0,    0, 2074, 2074, 2074, 2074,    0,    0,    0,    0,
    0,    0,    0, 2074, 2074,    0, 2074, 2074, 2074,    0,
    0, 2074,    0, 2074, 2074,    0, 2074, 2074,    0, 2074,
    0,    0, 2074, 2074,    0,    0,    0, 2074, 2074,    0,
    0,    0,    0, 2074, 2074, 2074,    0, 2074, 2074, 2074,
    0, 2074,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 2074, 2074, 2074, 2074, 2074, 2074, 2074, 2074, 2074,
 2074, 2074, 2074, 2074, 2074, 2074, 2074, 2074, 2074, 2074,
 2074, 2074,    0, 2074, 2074, 2074, 2074, 2074, 2074,    0,
 2074, 2074, 2074, 2074,    0, 2074,    0, 2074, 2074, 2074,
 2074, 2074, 2074, 2074, 2074, 2074, 2074, 2074, 2074, 2074,
 2074, 2074, 2074, 2074, 2074,    0, 2074, 2074, 2074, 2074,
    0, 2074, 2074, 2074, 2074, 2074,    0, 2074, 2074, 2074,
 2074, 2074, 2074, 2074, 2074, 2074, 2074, 2074, 2074, 2074,
    0, 2074, 2074, 2074, 2074, 2074, 2074,    0, 2074, 2074,
 2074, 2074, 2074, 2074, 2074, 2074,    0, 2074, 2074,    0,
    0, 2076, 2074, 2074, 2074, 2074, 2074, 2074,    0,    0,
 2074, 2074, 2074, 2074, 2074, 2074, 2074, 2074, 2074,    0,
 2074,    0,    0,    0, 2074,    0, 2074, 2074, 2074, 2074,
 2074, 2074,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 2074, 2074, 2074, 2074, 2074, 2074, 2074, 2074, 2074, 2074,
 2074, 2074, 2074, 2074, 2074, 2074, 2074, 2074, 2074, 2074,
 2074, 2074, 2074, 2074, 2074, 2074, 2074, 2074, 2074,    0,
 2074,    0,    0,    0, 2074, 2074, 2074, 2074, 2074, 2074,
 2074, 2074, 2074, 2074, 2074, 2074, 2074,    0, 2074, 2074,
    0,    0,    0,    0,    0, 2074, 2074, 2074, 2074, 2074,
 2074, 2074, 2074, 2074, 2074, 2074, 2074, 2074, 2074,    0,
    0, 2074, 2074,    0, 2074,    0,    0, 2074, 2074, 2074,
 2074, 2074, 2074,  152,    0,  153,    0,    0,    0,    0,
    0,  154,    0,    0,  155,  156,    0,    0, 1116,    0,
    0, 1117,    0,    0,    0,  157,  158,    0,  159,    0,
  160,    0,    0,  161,    0,    0,  162, 1118,    0, 1119,
    0,    0,    0,  163,    0,  164,  165,    0,    0,    0,
  166,    0,  167,    0,    0,  168,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  169,  170,    0,    0,    0,
  171,    0,  172,    0,  173,    0,    0, 1120,    0,    0,
  174,    0,  175, 1121, 1122, 1123,  176, 1124, 1125, 1126,
    0, 1127,    0,    0,  177,    0,    0,  178,    0,  179,
    0,    0,    0,  180,  181, 1128,    0, 1129,    0,    0,
    0,    0,    0,    0,    0,    0,  182,    0,    0,    0,
  184,  185,  186,  187,  188,    0,  189,    0,    0,    0,
    0,    0,  190,    0,  191,  192, 1130,    0,  193,  194,
    0,    0,    0,    0,    0,  195,    0,    0,  196,  197,
  198,  199, 1131,  200,    0,  201,    0,  202,  203,  204,
  205,  206,    0,    0,    0,    0,  207,    0,  208,    0,
    0,    0,    0,    0, 1132,    0,    0,    0,    0,  209,
    0,    0,    0,    0,  210,  211,    0,    0,    0,    0,
    0,  212,    0,    0,    0,  213,    0,  214,  215,  216,
  217,    0, 2437,    0,    0,    0,    0,    0,    0,    0,
  219,  220, 1252, 1134,    0,    0,    0,    0,    0,    0,
    0,  221,  222,    0,  223,  224,  225,    0,    0, 1135,
    0,  226,  227,    0,  228,  229,    0,  230,    0,    0,
  231,  232,    0,    0,    0,  233,  234,    0,    0,    0,
    0,  235,  236,  237,    0,  238,  239,  240,    0,  241,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  242,
  243,  244,  245,  246,  247,  248,  249,  250,  251,  252,
  253,  254,  255,  256,  257,  258,  259,  260,  261,  262,
    0,  263,  264,  265,  266,  267,  268,    0,  269,  270,
  271,  272,    0,  273,    0,  274,  275,  276,  277,  278,
  279,  280,  281,  282,  283,  284,  285,  286,  287,  288,
  289,  290,  291,    0,  292,  293,  294,  295,    0,  296,
  297,  298,  299,  300,    0,  301,  302,  303,  304,  305,
  306,  307,  308,  309,  310,  311,  312,  313,    0,  314,
  315,  316,  317,  318,  319,    0,  320,  321,  322,  323,
  324,  325,  326,  327,    0,  328,  329,    0,    0,    0,
  330,  331,  332,  333,  334,  335,    0,    0,  336,  337,
  338,  339,  340,  341,  342,  343,  344,    0, 1136,    0,
    0,    0,  345,    0,  346,  347,  348,  349,  350,  351,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  352,  353,
 1137,  354,  355,  356,  357,  358,  359,  360,  361,  362,
 1138,  363,  364,  365,  366,  367,  368,  369,  370,  371,
  372,  373,  374, 1139,  375,  376,  377,    0,  378,    0,
    0,    0,  379,  380,  381,  382,  383,  384,  385,  386,
  387,  388,  389,  390,  391,    0,  392,  393,    0,    0,
    0,    0,    0,  394,  395,  396,  397,  398,  399,  400,
  401,  402,  403,  404,  405,  406,  407,    0,    0,  408,
  409,    0, 1140,    0,    0,  410,  411,  412,  413,  414,
  415,  152,    0,  153,    0,    0,    0,    0,    0,  154,
    0,    0,  155,  156,    0,    0, 1116,    0,    0, 1117,
    0,    0,    0,  157,  158,    0,  159,    0,  160,    0,
    0,  161,    0,    0,  162, 1118,    0, 1119,    0,    0,
    0,  163,    0,  164,  165,    0,    0,    0,  166,    0,
  167,    0,    0,  168,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  169,  170,    0,    0,    0,  171,    0,
  172,    0,  173,    0,    0, 1120,    0,    0,  174,    0,
  175, 1121, 1122, 1123,  176, 1124, 1125, 1126,    0, 1127,
    0,    0,  177,    0,    0,  178,    0,  179,    0,    0,
    0,  180,  181, 1128,    0, 1129,    0,    0,    0,    0,
    0,    0,    0,    0,  182,    0,    0,    0,  184,  185,
  186,  187,  188,    0,  189,    0,    0,    0,    0,    0,
  190,    0,  191,  192, 1130,    0,  193,  194,    0,    0,
    0,    0,    0,  195,    0,    0,  196,  197,  198,  199,
 1131,  200,    0,  201,    0,  202,  203,  204,  205,  206,
    0,    0,    0,    0,  207,    0,  208,    0,    0,    0,
    0,    0, 1132,    0,    0,    0,    0,  209,    0,    0,
    0,    0,  210,  211,    0,    0,    0,    0,    0,  212,
    0,    0,    0,  213,    0,  214,  215,  216,  217,    0,
 2550,    0,    0,    0,    0,    0,    0,    0,  219,  220,
 1252, 1134,    0,    0,    0,    0,    0,    0,    0,  221,
  222,    0,  223,  224,  225,    0,    0, 1135,    0,  226,
  227,    0,  228,  229,    0,  230,    0,    0,  231,  232,
    0,    0,    0,  233,  234,    0,    0,    0,    0,  235,
  236,  237,    0,  238,  239,  240,    0,  241,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  242,  243,  244,
  245,  246,  247,  248,  249,  250,  251,  252,  253,  254,
  255,  256,  257,  258,  259,  260,  261,  262,    0,  263,
  264,  265,  266,  267,  268,    0,  269,  270,  271,  272,
    0,  273,    0,  274,  275,  276,  277,  278,  279,  280,
  281,  282,  283,  284,  285,  286,  287,  288,  289,  290,
  291,    0,  292,  293,  294,  295,    0,  296,  297,  298,
  299,  300,    0,  301,  302,  303,  304,  305,  306,  307,
  308,  309,  310,  311,  312,  313,    0,  314,  315,  316,
  317,  318,  319,    0,  320,  321,  322,  323,  324,  325,
  326,  327,    0,  328,  329,    0,    0,    0,  330,  331,
  332,  333,  334,  335,    0,    0,  336,  337,  338,  339,
  340,  341,  342,  343,  344,    0, 1136,    0,    0,    0,
  345,    0,  346,  347,  348,  349,  350,  351,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  352,  353, 1137,  354,
  355,  356,  357,  358,  359,  360,  361,  362, 1138,  363,
  364,  365,  366,  367,  368,  369,  370,  371,  372,  373,
  374, 1139,  375,  376,  377,    0,  378,    0,    0,    0,
  379,  380,  381,  382,  383,  384,  385,  386,  387,  388,
  389,  390,  391,    0,  392,  393,    0,    0,    0,    0,
    0,  394,  395,  396,  397,  398,  399,  400,  401,  402,
  403,  404,  405,  406,  407,    0,    0,  408,  409,    0,
 1140,    0,    0,  410,  411,  412,  413,  414,  415,  152,
    0,  153,    0,    0,    0,    0,    0,  154,    0,    0,
  155,  156,    0,    0, 1116,    0,    0, 1117,    0,    0,
    0,  157, 3100,    0,  159,    0,  160,    0,    0,  161,
    0,    0,  162, 1118,    0, 1119,    0,    0,    0,  163,
    0,  164,  165,    0,    0,    0,  166,    0,  167,    0,
    0,  168,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  169,  170,    0,    0,    0,  171,    0,  172,    0,
  173,    0,    0, 1120,    0,    0,  174,    0,  175, 1121,
 1122, 1123,  176, 1124, 1125, 1126,    0, 1127,    0,    0,
  177,    0,    0,  178,    0,  179,    0,    0,    0,  180,
  181, 1128,    0, 1129,    0,    0,    0,    0,    0,    0,
    0,    0,  182,    0,    0,    0,  184,  185,  186,  187,
  188,    0,  189,    0,    0,    0,    0,    0,  190,    0,
  191,  192, 1130,    0,  193,  194,    0,    0,    0,    0,
    0,  195,    0,    0,  196,  197,  198,  199, 1131,  200,
    0,  201,    0,  202,  203,  204,  205,  206,    0,    0,
    0,    0,  207,    0,  208,    0,    0,    0,    0,    0,
 1132,    0,    0,    0,    0,  209,    0,    0,    0,    0,
  210,  211,    0,    0,    0,    0,    0,  212,    0,    0,
    0,  213,    0,  214,  215,  216,  217,    0,  218,    0,
    0,    0,    0,    0,    0,    0,  219,  220, 1252, 1134,
    0,    0,    0,    0,    0,    0,    0,  221,  222,    0,
  223,  224,  225,    0,    0, 1135,    0,  226,  227,    0,
  228,  229,    0,  230,    0,    0,  231,  232,    0,    0,
    0,  233,  234,    0,    0,    0,    0,  235,  236,  237,
    0,  238,  239,  240,    0,  241,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  242,  243,  244,  245,  246,
  247,  248,  249,  250,  251,  252,  253,  254,  255,  256,
  257,  258,  259,  260,  261,  262,    0,  263,  264,  265,
  266,  267,  268,    0,  269,  270,  271, 3101,    0,  273,
    0,  274,  275,  276,  277,  278,  279,  280,  281,  282,
  283,  284,  285,  286,  287,  288,  289,  290,  291,    0,
  292,  293,  294,  295,    0,  296,  297,  298,  299,  300,
    0,  301,  302,  303,  304,  305,  306,  307,  308,  309,
  310,  311,  312,  313,    0,  314,  315,  316,  317,  318,
  319,    0,  320,  321,  322,  323,  324,  325,  326,  327,
    0,  328,  329,    0,    0,    0,  330,  331,  332,  333,
  334,  335,    0,    0,  336,  337,  338,  339,  340,  341,
  342,  343,  344,    0, 1136,    0,    0,    0,  345,    0,
  346,  347,  348,  349,  350,  351,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  352,  353, 1137,  354,  355,  356,
  357,  358,  359,  360,  361,  362, 1138,  363,  364,  365,
  366,  367,  368,  369,  370,  371,  372,  373,  374, 1139,
  375,  376,  377,    0,  378,    0,    0,    0,  379,  380,
  381,  382,  383,  384,  385,  386,  387,  388,  389,  390,
  391,    0,  392,  393,    0,    0,    0,    0,    0,  394,
  395,  396,  397,  398,  399,  400,  401,  402,  403,  404,
  405,  406,  407,    0,    0,  408,  409,    0, 1140,    0,
    0,  410,  411,  412,  413,  414,  415,  152,    0,  153,
    0,    0,    0,    0,    0,  154,    0,    0,  155,  156,
    0,    0, 1116,    0,    0, 1117,    0,    0,    0,  157,
  158,    0,  159,    0,  160,    0,    0,  161,    0,    0,
  162, 1118,    0, 1119,    0,    0,    0,  163,    0,  164,
  165,    0,    0,    0,  166,    0,  167,    0,    0,  168,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  169,
  170,    0,    0,    0,  171,    0,  172,    0,  173,    0,
    0, 1120,    0,    0,  174,    0,  175, 1121, 1122, 1123,
  176, 1124, 1125, 1126,    0, 1127,    0,    0,  177,    0,
    0,  178,    0,  179,    0,    0,    0,  180,  181, 1128,
    0, 1129,    0,    0,    0,    0,    0,    0,    0,    0,
  182,    0,    0,    0,  184,  185,  186,  187,  188,    0,
  189,    0,    0,    0,    0,    0,  190,    0,  191,  192,
 1130,    0,  193,  194,    0,    0,    0,    0,    0,  195,
    0,    0,  196,  197,  198,  199, 1131,  200,    0,  201,
    0,  202,  203,  204,  205,  206,    0,    0,    0,    0,
  207,    0,  208,    0,    0,    0,    0,    0, 1132,    0,
    0,    0,    0,  209,    0,    0,    0,    0,  210,  211,
    0,    0,    0,    0,    0,  212,    0,    0,    0,  213,
    0,  214,  215,  216,  217,    0,  218,    0,    0,    0,
    0,    0,    0,    0,  219,  220, 1252, 1134,    0,    0,
    0,    0,    0,    0,    0,  221,  222,    0,  223,  224,
  225,    0,    0, 1135,    0,  226,  227,    0,  228,  229,
    0,  230,    0,    0,  231,  232,    0,    0,    0,  233,
  234,    0,    0,    0,    0,  235,  236,  237,    0,  238,
  239,  240,    0,  241,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  242,  243,  244,  245,  246,  247,  248,
  249,  250,  251,  252,  253,  254,  255,  256,  257,  258,
  259,  260,  261,  262,    0,  263,  264,  265,  266,  267,
  268,    0,  269,  270,  271,  272,    0,  273,    0,  274,
  275,  276,  277,  278,  279,  280,  281,  282,  283,  284,
  285,  286,  287,  288,  289,  290,  291,    0,  292,  293,
  294,  295,    0,  296,  297,  298,  299,  300,    0,  301,
  302,  303,  304,  305,  306,  307,  308,  309,  310,  311,
  312,  313,    0,  314,  315,  316,  317,  318,  319,    0,
  320,  321,  322,  323,  324,  325,  326,  327,    0,  328,
  329,    0,    0,    0,  330,  331,  332,  333,  334,  335,
    0,    0,  336,  337,  338,  339,  340,  341,  342,  343,
  344,    0, 1136,    0,    0,    0,  345,    0,  346,  347,
  348,  349,  350,  351,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,  352,  353, 1137,  354,  355,  356,  357,  358,
  359,  360,  361,  362, 1138,  363,  364,  365,  366,  367,
  368,  369,  370,  371,  372,  373,  374, 1139,  375,  376,
  377,    0,  378,    0,    0,    0,  379,  380,  381,  382,
  383,  384,  385,  386,  387,  388,  389,  390,  391,    0,
  392,  393,    0,    0,    0,    0,    0,  394,  395,  396,
  397,  398,  399,  400,  401,  402,  403,  404,  405,  406,
  407,    0,    0,  408,  409,    0, 1140,    0,    0,  410,
  411,  412,  413,  414,  415,   51,    0,   51,   39,   39,
    0,    0,    0,   51,    0,    0,   51,   51,    0,    0,
    0,    0,    0,    0,    0,    0,    0,   51,   51,    0,
   51,    0,   51,    0,   39,   51,    0,    0,   51,    0,
    0,    0,    0,   51,   39,   51,    0,   51,   51,   39,
    0,    0,   51,    0,   51,   39,    0,   51,    0,    0,
    0,    0,    0,    0,    0,    0,    0,   51,   51,    0,
    0,    0,   51,    0,   51,    0,   51,    0,   39,    0,
    0,    0,   51,    0,   51,    0,    0,    0,   51,    0,
    0,    0,    0,    0,    0,    0,   51,    0,    0,   51,
    0,   51,    0,    0,    0,   51,   51,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,   51,    0,
    0,    0,   51,   51,   51,   51,   51,    0,   51,    0,
    0,    0,    0,    0,   51,    0,   51,   51,    0,   39,
   51,   51,    0,    0,    0,    0,    0,   51,   39,    0,
   51,   51,   51,   51,    0,   51,    0,   51,    0,   51,
   51,   51,   51,   51,    0,    0,    0,    0,   51,    0,
   51,    0,    0,   39,    0,    0,    0,    0,    0,    0,
    0,   51,    0,    0,    0,    0,   51,   51,    0,    0,
    0,    0,    0,   51,    0,    0,    0,   51,    0,   51,
   51,   51,   51,    0,   51,    0,    0,    0,    0,    0,
    0,    0,   51,   51,    0,    0,    0,    0,    0,    0,
    0,    0,    0,   51,   51,    0,   51,   51,   51,    0,
    0,    0,    0,   51,   51,    0,   51,   51,    0,   51,
    0,    0,   51,   51,    0,    0,    0,   51,   51,    0,
    0,    0,    0,   51,   51,   51,    0,   51,   51,   51,
    0,   51,    0,    0,    0,    0,    0,    0,    0,    0,
    0,   51,   51,   51,   51,   51,   51,   51,   51,   51,
   51,   51,   51,   51,   51,   51,   51,   51,   51,   51,
   51,   51,    0,   51,   51,   51,   51,   51,   51,    0,
   51,   51,   51,   51,    0,   51,    0,   51,   51,   51,
   51,   51,   51,   51,   51,   51,   51,   51,   51,   51,
   51,   51,   51,   51,   51,    0,   51,   51,   51,   51,
    0,   51,   51,   51,   51,   51,    0,   51,   51,   51,
   51,   51,   51,   51,   51,   51,   51,   51,   51,   51,
    0,   51,   51,   51,   51,   51,   51,    0,   51,   51,
   51,   51,   51,   51,   51,   51,    0,   51,   51,    0,
    0,    0,   51,   51,   51,   51,   51,   51,    0,    0,
   51,   51,   51,   51,   51,   51,   51,   51,   51,    0,
    0,    0,    0,    0,    0,    0,   51,   51,   51,   51,
   51,   51,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
   51,   51,    0,   51,   51,   51,   51,   51,   51,   51,
   51,   51,    0,   51,   51,   51,   51,   51,   51,   51,
   51,   51,   51,   51,   51,    0,   51,   51,   51,    0,
   51,    0,    0,    0,   51,   51,   51,   51,   51,   51,
   51,   51,   51,   51,   51,   51,   51,    0,   51,   51,
    0,    0,    0,    0,    0,   51,   51,   51,   51,   51,
   51,   51,   51,   51,   51,   51,   51,   51,   51,    0,
    0,   51,   51,    0,    0,    0,    0,   51,   51,   51,
   51,   51,   51,  128,    0,  128,    0,  126,    0,    0,
    0,  128,    0,    0,  128,  128,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  128,  128,    0,  128,    0,
  128,    0,  126,  128,    0,    0,  128,    0,    0,    0,
    0,  128,  126,  128,    0,  128,  128,  126,    0,    0,
  128,    0,  128,  126,    0,  128,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  128,  128,    0,    0,    0,
  128,    0,  128,    0,  128,    0,  126,    0,    0,    0,
  128,    0,  128,    0,    0,    0,  128,    0,    0,    0,
    0,    0,    0,    0,  128,    0,    0,  128,    0,  128,
    0,    0,    0,  128,  128,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  128,    0,    0,    0,
  128,  128,  128,  128,  128,    0,  128,    0,    0,    0,
    0,    0,  128,    0,  128,  128,    0,  126,  128,  128,
    0,    0,    0,    0,    0,  128,  126,    0,  128,  128,
  128,  128,    0,  128,    0,  128,    0,  128,  128,  128,
  128,  128,    0,    0,    0,    0,  128,    0,  128,    0,
    0,  126,    0,    0,    0,    0,    0,    0,    0,  128,
    0,    0,    0,    0,  128,  128,    0,    0,    0,    0,
    0,  128,    0,    0,    0,  128,    0,  128,  128,  128,
  128,    0,  128,    0,    0,    0,    0,    0,    0,    0,
  128,  128,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  128,  128,    0,  128,  128,  128,    0,    0,    0,
    0,  128,  128,    0,  128,  128,    0,  128,    0,    0,
  128,  128,    0,    0,    0,  128,  128,    0,    0,    0,
    0,  128,  128,  128,    0,  128,  128,  128,    0,  128,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  128,
  128,  128,  128,  128,  128,  128,  128,  128,  128,  128,
  128,  128,  128,  128,  128,  128,  128,  128,  128,  128,
    0,  128,  128,  128,  128,  128,  128,    0,  128,  128,
  128,  128,    0,  128,    0,  128,  128,  128,  128,  128,
  128,  128,  128,  128,  128,  128,  128,  128,  128,  128,
  128,  128,  128,    0,  128,  128,  128,  128,    0,  128,
  128,  128,  128,  128,    0,  128,  128,  128,  128,  128,
  128,  128,  128,  128,  128,  128,  128,  128,    0,  128,
  128,  128,  128,  128,  128,    0,  128,  128,  128,  128,
  128,  128,  128,  128,    0,  128,  128,    0,    0,    0,
  128,  128,  128,  128,  128,  128,    0,    0,  128,  128,
  128,  128,  128,  128,  128,  128,  128,    0,    0,    0,
    0,    0,    0,    0,  128,  128,  128,  128,  128,  128,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  128,  128,
    0,  128,  128,  128,  128,  128,  128,  128,  128,  128,
    0,  128,  128,  128,  128,  128,  128,  128,  128,  128,
  128,  128,  128,    0,  128,  128,  128,    0,  128,    0,
    0,    0,  128,  128,  128,  128,  128,  128,  128,  128,
  128,  128,  128,  128,  128,    0,  128,  128,    0,    0,
    0,    0,    0,  128,  128,  128,  128,  128,  128,  128,
  128,  128,  128,  128,  128,  128,  128,    0,    0,  128,
  128,    0,    0,    0,    0,  128,  128,  128,  128,  128,
  128,  152,    0,  153,    0,    0,    0,    0,    0,  154,
    0,    0,  155,  156,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  157,  158,    0,  159,    0,  160,    0,
    0,  161,    0,    0,  162,    0,    0,    0,    0,    0,
    0,  163,    0,  164,  165,    0,    0,    0,  166,    0,
  167,    0,    0,  168,    0,    0,    0,    0,    0,    0,
 3292,    0,    0,  169,  170,    0, 3293,    0,  171,    0,
  172,    0,  173,    0,    0,    0,    0,    0,  174,    0,
  175,    0,    0,    0,  176,    0,    0,    0,    0,    0,
    0,    0,  177,    0,    0,  178,    0,  179,    0,    0,
    0,  180,  181,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  182,    0,    0,    0,  184,  185,
  186,  187,  188,    0,  189,    0,    0,    0,    0,    0,
  190, 3294,  191,  192,    0,    0,  193,  194,    0,    0,
    0,    0,    0,  195,    0,    0,  196,  197,  198,  199,
    0,  200,    0,  201,    0,  202,  203,  204,  205,  206,
    0,    0,    0,    0,  207, 3295,  208,    0,    0,    0,
 3296,    0,    0,    0,    0,    0, 3297,  209,    0,    0,
    0,    0,  210,  211,    0,    0,    0,    0,    0,  212,
    0,    0,    0,  213,    0,  214,  215,  216, 3298,    0,
  218,    0,    0,    0,    0,    0,    0,    0,  219,  220,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  221,
  222,    0,  223,  224,  225,    0,    0,    0,    0,  226,
  227,    0,  228,  229,    0,  230,    0,    0,  231,  232,
    0,    0,    0,  233,  234,    0,    0,    0,    0,  235,
  236,  237,    0,  238,  239,  240,    0,  241,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  242,  243,  244,
  245,  246,  247,  248,  249,  250,  251,  252,  253,  254,
  255,  256,  257,  258,  259,  260,  261,  262,    0,  263,
  264,  265,  266,  267,  268,    0,  269,  270,  271,  272,
    0,  273,    0,  274,  275,  276,  277,  278,  279,  280,
  281,  282,  283,  284,  285,  286,  287,  288,  289,  290,
  291,    0,  292,  293,  294,  295,    0,  296,  297,  298,
  299,  300,    0,  301,  302,  303,  304,  305,  306,  307,
  308,  309,  310,  311,  312,  313,    0,  314,  315,  316,
  317,  318,  319,    0,  320,  321,  322,  323,  324,  325,
  326,  327,    0, 3299,  329,    0,    0,    0,  330,  331,
  332,  333,  334,  335,    0,    0,  336,  337,  338,  339,
  340,  341,  342,  343,  344,    0,    0,    0,    0,    0,
  345,    0,  346,  347,  348,  349,  350,  351,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  352,  353,    0,  354,
  355,  356,  357,  358,  359,  360,  361,  362,    0,  363,
  364,  365,  366,  367,  368,  369,  370,  371,  372,  373,
  374,    0,  375,  376,  377,    0,  378,    0,    0,    0,
  379,  380,  381,  382,  383,  384,  385,  386,  387,  388,
  389,  390,  391,    0,  392,  393,    0,    0,    0,    0,
    0,  394,  395,  396,  397,  398,  399,  400,  401,  402,
  403,  404,  405, 3300,  407,    0,    0,  408,  409,    0,
    0,    0,    0,  410,  411,  412,  413,  414,  415,  152,
    0,  153,    0,    0,    0,    0,    0,  154,    0,    0,
  155,  156,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  157,  158,    0,  159,    0,  160,    0,    0,  161,
    0,    0,  162,    0,    0,    0,    0,    0,    0,  163,
    0,  164,  165,    0,    0,    0,  166,    0,  167,    0,
    0,  168,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  169,  170,    0,    0,    0,  171,    0,  172,    0,
  173, 3449,    0,    0,    0,    0,  174,    0,  175,    0,
    0,    0,  176,    0,    0,    0,    0,    0,    0,    0,
  177,    0,    0,  178,    0,  179,    0,    0,    0,  180,
  181,    0, 3450,    0,    0,    0,    0,    0,    0,    0,
    0,    0,  182,    0, 3451,    0,  184,  185,  186,  187,
  188,    0,  189,    0,    0,    0,    0,    0,  190,    0,
  191,  192,    0,    0,  193,  194,    0,    0,    0,    0,
    0,  195,    0,    0,  196,  197,  198,  199,    0,  200,
    0,  201,    0,  202,  203,  204,  205,  206,    0,    0,
    0,    0,  207,    0,  208,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  209,    0,    0,    0,    0,
  210,  211,    0,    0,    0,    0,    0,  212,    0,    0,
    0,  213,    0,  214,  215,  216,  217,    0,  218,    0,
    0,    0,    0,    0,    0,    0,  219,  220,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  221,  222,    0,
  223,  224,  225,    0,    0,    0,    0,  226,  227,    0,
  228,  229,    0,  230,    0,    0,  231,  232,    0,    0,
    0,  233,  234,    0,    0,    0,    0,  235,  236,  237,
    0,  238,  239,  240,    0,  241,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  242,  243,  244,  245,  246,
  247,  248,  249,  250,  251,  252,  253,  254,  255,  256,
  257,  258,  259,  260,  261,  262,    0,  263,  264,  265,
  266,  267,  268,    0,  269,  270,  271,  272,    0,  273,
    0,  274,  275,  276,  277,  278,  279,  280,  281,  282,
  283,  284,  285,  286,  287,  288,  289,  290,  291,    0,
  292,  293,  294,  295,    0,  296,  297,  298,  299,  300,
    0,  301,  302,  303,  304,  305,  306,  307,  308,  309,
  310,  311,  312,  313,    0,  314,  315,  316,  317,  318,
  319,    0,  320,  321,  322,  323,  324,  325,  326,  327,
    0,  328,  329,    0,    0,    0,  330,  331,  332,  333,
  334,  335,    0,    0,  336,  337,  338,  339,  340,  341,
  342,  343,  344,    0,    0,    0,    0,    0,  345,    0,
  346,  347,  348,  349,  350,  351,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  352,  353,    0,  354,  355,  356,
  357,  358,  359,  360,  361,  362,    0,  363,  364,  365,
  366,  367,  368,  369,  370,  371,  372,  373,  374,    0,
  375,  376,  377,    0,  378,    0,    0,    0,  379,  380,
  381,  382,  383,  384,  385,  386,  387,  388,  389,  390,
  391,    0,  392,  393,    0,    0,    0,    0,    0,  394,
  395,  396,  397,  398,  399,  400,  401,  402,  403,  404,
  405,  406,  407,    0,    0,  408,  409,    0,    0,    0,
    0,  410,  411,  412,  413,  414,  415,  152,    0,  153,
    0,    0,    0, 3580,    0,  154,    0,    0,  155,  156,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  157,
  158,    0,  159,    0,  160,    0,    0,  161,    0,    0,
  162,    0,    0,    0,    0,    0,    0,  163,    0,  164,
  165,    0,    0,    0,  166,    0,  167,    0,    0,  168,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  169,
  170,    0, 3581,    0,  171,    0,  172,    0,  173,    0,
    0,    0,    0,    0,  174,    0,  175,    0,    0,    0,
  176,    0,    0,    0,    0,    0,    0,    0,  177,    0,
    0,  178,    0,  179,    0,    0,    0,  180,  181,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
  182,    0,    0,    0,  184,  185,  186,  187,  188,    0,
  189,    0,    0,    0,    0,    0,  190,    0,  191,  192,
    0,    0,  193,  194,    0,    0,    0,    0,    0,  195,
    0,    0,  196,  197,  198,  199,    0,  200,    0,  201,
    0,  202,  203,  204,  205,  206,    0,    0,    0,    0,
  207,    0,  208,    0,    0,    0, 3582,    0,    0,    0,
    0,    0,    0,  209,    0,    0,    0,    0,  210,  211,
    0,    0,    0,    0,    0,  212,    0,    0,    0,  213,
    0,  214,  215,  216,  217,    0,  218,    0,    0,    0,
    0,    0,    0,    0,  219,  220,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  221,  222,    0,  223,  224,
  225,    0,    0,    0,    0,  226,  227,    0,  228,  229,
    0,  230,    0,    0,  231,  232,    0,    0,    0,  233,
  234,    0,    0,    0,    0,  235,  236,  237,    0,  238,
  239,  240,    0,  241,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  242,  243,  244,  245,  246,  247,  248,
  249,  250,  251,  252,  253,  254,  255,  256,  257,  258,
  259,  260,  261,  262,    0,  263,  264,  265,  266,  267,
  268,    0,  269,  270,  271,  272,    0,  273,    0,  274,
  275,  276,  277,  278,  279,  280,  281,  282,  283,  284,
  285,  286,  287,  288,  289,  290,  291,    0,  292,  293,
  294,  295,    0,  296,  297,  298,  299,  300,    0,  301,
  302,  303,  304,  305,  306,  307,  308,  309,  310,  311,
  312,  313,    0,  314,  315,  316,  317,  318,  319,    0,
  320,  321,  322,  323,  324,  325,  326,  327,    0,  328,
  329,    0,    0,    0,  330,  331,  332,  333,  334,  335,
    0,    0,  336,  337,  338,  339,  340,  341,  342,  343,
  344,    0,    0,    0,    0,    0,  345,    0,  346,  347,
  348,  349,  350,  351,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,  352,  353,    0,  354,  355,  356,  357,  358,
  359,  360,  361,  362,    0,  363,  364,  365,  366,  367,
  368,  369,  370,  371,  372,  373,  374,    0,  375,  376,
  377,    0,  378,    0,    0,    0,  379,  380,  381,  382,
  383,  384,  385,  386,  387,  388,  389,  390,  391,    0,
  392,  393,    0,    0,    0,    0,    0,  394,  395,  396,
  397,  398,  399,  400,  401,  402,  403,  404,  405,  406,
  407,    0,    0,  408,  409,    0,    0,    0,    0,  410,
  411,  412,  413,  414,  415,  152,    0,  153,    0,    0,
    0,    0,    0,  154,    0,    0,  155,  156,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  157,  158,    0,
  159,    0,  160,    0,    0,  161,    0,    0,  162,    0,
  852,    0,    0,    0,    0,  163,    0,  164,  165,    0,
    0,    0,  166,    0,  167,    0,    0,  168,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  169,  170,    0,
    0,    0,  171,    0,  172,    0,  173,    0,    0,    0,
    0,    0,  174,    0,  175,    0,    0,    0,  176,    0,
    0,    0,    0,    0,    0,    0,  177,    0,    0,  178,
    0,  179,    0,    0,    0,  180,  181,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  182,    0,
    0,    0,  184,  185,  186,  187,  188,    0,  189,    0,
    0,    0,    0,    0,  190,    0,  191,  192,    0,    0,
  193,  194,    0,    0,    0,    0,    0,  195,    0,    0,
  196,  197,  198,  199,    0,  200,    0,  201,    0,  202,
  203,  204,  205,  206,    0,    0,    0,    0,  207,    0,
  208,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  209,    0,    0,    0,    0,  210,  211,    0,    0,
    0,    0,    0,  212,    0,    0,    0,  213,    0,  214,
  215,  216,  217,    0,  218,    0,    0,    0,    0,    0,
    0,    0,  219,  220,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  221,  222,    0,  223,  224,  225,    0,
    0,    0,    0,  226,  227,    0,  228,  229,    0,  230,
    0,    0,  231,  232,    0,    0,    0,  233,  234,    0,
    0,    0,    0,  235,  236,  237,    0,  238,  239,  240,
    0,  241,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  242,  243,  244,  245,  246,  247,  248,  249,  250,
  251,  252,  253,  254,  255,  256,  257,  258,  259,  260,
  261,  262,    0,  263,  264,  265,  266,  267,  268,    0,
  269,  270,  271,  272,    0,  273,    0,  274,  275,  276,
  277,  278,  279,  280,  281,  282,  283,  284,  285,  286,
  287,  288,  289,  290,  291,    0,  292,  293,  294,  295,
    0,  296,  297,  298,  299,  300,    0,  301,  302,  303,
  304,  305,  306,  307,  308,  309,  310,  311,  312,  313,
    0,  314,  315,  316,  317,  318,  319,    0,  320,  321,
  322,  323,  324,  325,  326,  327,    0,  328,  329,    0,
    0,    0,  330,  331,  332,  333,  334,  335,    0,    0,
  336,  337,  338,  339,  340,  341,  342,  343,  344,    0,
    0,    0,    0,    0,  345,  972,  346,  347,  348,  349,
  350,  351,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
  352,  353,    0,  354,  355,  356,  357,  358,  359,  360,
  361,  362,    0,  363,  364,  365,  366,  367,  368,  369,
  370,  371,  372,  373,  374,    0,  375,  376,  377,    0,
  378,    0,    0,    0,  379,  380,  381,  382,  383,  384,
  385,  386,  387,  388,  389,  390,  391,    0,  392,  393,
    0,    0,    0,    0,    0,  394,  395,  396,  397,  398,
  399,  400,  401,  402,  403,  404,  405,  406,  407,    0,
    0,  408,  409,    0,    0,    0,    0,  410,  411,  412,
  413,  414,  415,  152,    0,  153,    0,    0,    0,    0,
    0,  154,    0,    0,  155,  156,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  157,  158,    0,  159,    0,
  160,    0,    0,  161,    0,    0,  162,    0,    0,    0,
    0,    0,    0,  163,    0,  164,  165,    0,    0,    0,
  166,    0,  167,    0,    0,  168,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  169,  170,    0,    0,    0,
  171,    0,  172,    0,  173,    0,    0,    0,    0,    0,
  174,    0,  175,    0,    0,    0,  176,    0,    0,    0,
    0,    0,    0,    0,  177,    0,    0,  178,    0,  179,
    0,    0,    0,  180,  181,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  182,    0,    0,    0,
  184,  185,  186,  187,  188,    0,  189,    0,    0,    0,
    0,    0,  190,    0,  191,  192,    0,    0,  193,  194,
    0,    0,    0,    0,    0,  195,    0,    0,  196,  197,
  198,  199,    0,  200,    0,  201,    0,  202,  203,  204,
  205,  206,    0,    0,    0,    0,  207,    0,  208,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  209,
    0,    0,    0,    0,  210,  211,    0,    0,    0,    0,
    0,  212,    0,  601,  602,  213,    0,  214,  215,  216,
  217,    0,  218,    0,    0,    0,    0,    0,    0,    0,
  219,  220,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  221,  222,    0,  223,  224,  225,    0,    0,    0,
    0,  226,  227,    0,  228,  229,    0,  230,    0,    0,
  231,  232,    0,    0,    0,  233,  234,    0,    0,    0,
    0,  235,  236,  237,    0,  238,  239,  240,    0,  241,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  242,
  243,  244,  245,  246,  247,  248,  249,  250,  251,  252,
  253,  254,  255,  256,  257,  258,  259,  260,  261,  262,
    0,  263,  264,  265,  266,  267,  268,    0,  269,  270,
  271,  272,    0,  273,    0,  274,  275,  276,  277,  278,
  279,  280,  281,  282,  283,  284,  285,  286,  287,  288,
  289,  290,  291,    0,  292,  293,  294,  295,    0,  296,
  297,  298,  299,  300,    0,  301,  302,  303,  304,  305,
  306,  307,  308,  309,  310,  311,  312,  313,    0,  314,
  315,  316,  317,  318,  319,    0,  320,  321,  322,  323,
  324,  325,  326,  327,    0,  328,  329,    0,    0,    0,
  330,  331,  332,  333,  334,  335,    0,    0,  336,  337,
  338,  339,  340,  341,  342,  343,  344,    0,    0,    0,
    0,    0,  345,    0,  346,  347,  348,  349,  350,  351,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  352,  353,
    0,  354,  355,  356,  357,  358,  359,  360,  361,  362,
    0,  363,  364,  365,  366,  367,  368,  369,  370,  371,
  372,  373,  374,    0,  375,  376,  377,    0,  378,    0,
    0,    0,  379,  380,  381,  382,  383,  384,  385,  386,
  387,  388,  389,  390,  391,    0,  392,  393,    0,    0,
    0,    0,    0,  394,  395,  396,  397,  398,  399,  400,
  401,  402,  403,  404,  405,  406,  407,    0,    0,  408,
  409,    0,    0,    0,    0,  410,  411,  412,  413,  414,
  415,  152,    0,  153,    0,    0,    0,    0,    0,  154,
    0,    0,  155,  156,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  157,  158,    0,  159,    0,  160,    0,
    0,  161,    0,    0,  162,    0,    0,    0,    0,    0,
    0,  163,    0,  164,  165,    0,    0,    0,  166,    0,
  167,    0,    0,  168,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  169,  170,    0,    0,    0,  171,    0,
  172,    0,  173,    0,    0,    0,    0,    0,  174, 1670,
  175,    0,    0,    0,  176,    0,    0,    0,    0,    0,
    0,    0,  177,    0,    0,  178,    0,  179,    0, 1671,
    0,  180,  181,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  182,    0,    0,    0,  184,  185,
  186,  187,  188,    0,  189,    0,    0,    0,    0,    0,
  190,    0,  191,  192,    0,    0,  193,  194,    0,    0,
    0,    0,    0,  195,    0,    0,  196,  197,  198,  199,
    0,  200,    0, 3233,    0,  202,  203,  204,  205,  206,
    0,    0,    0,    0,  207,    0,  208,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  209,    0,    0,
    0,    0,  210,  211,    0,    0,    0,    0,    0,  212,
    0,    0,    0,  213,    0,  214,  215,  216,  217,    0,
  218,    0,    0,    0,    0,    0,    0,    0,  219,  220,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  221,
  222,    0,  223,  224,  225,    0,    0,    0,    0,  226,
  227,    0,  228,  229,    0,  230,    0,    0,  231,  232,
    0,    0,    0,  233,  234,    0,    0,    0,    0,  235,
  236,  237,    0,  238,  239,  240,    0,  241,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  242,  243,  244,
  245,  246,  247,  248,  249,  250,  251,  252,  253,  254,
  255,  256,  257,  258,  259,  260,  261,  262,    0,  263,
  264,  265,  266,  267,  268,    0,  269,  270,  271,  272,
    0, 3234,    0,  274,  275,  276,  277,  278,  279,  280,
  281,  282,  283,  284,  285,  286,  287,  288,  289,  290,
  291,    0,  292,  293,  294,  295,    0,  296,  297,  298,
  299,  300,    0,  301,  302,  303,  304,  305,  306,  307,
  308,  309,  310,  311,  312,  313,    0,  314,  315,  316,
  317,  318,  319,    0,  320,  321,  322,  323,  324,  325,
  326,  327,    0,  328,  329,    0,    0,    0,  330,  331,
  332,  333,  334,  335,    0,    0,  336,  337,  338,  339,
  340,  341,  342,  343,  344,    0,    0,    0,    0,    0,
  345,    0,  346,  347,  348,  349,  350,  351,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  352,  353,    0,  354,
  355,  356,  357,  358,  359,  360,  361,  362,    0,  363,
  364,  365,  366,  367,  368,  369,  370,  371,  372,  373,
  374,    0,  375,  376,  377,    0,  378,    0,    0,    0,
  379,  380,  381,  382,  383,  384,  385,  386,  387,  388,
  389,  390,  391,    0,  392,  393,    0,    0,    0,    0,
    0,  394,  395,  396,  397,  398,  399,  400,  401,  402,
  403,  404,  405,  406,  407,    0,    0,  408,  409,    0,
    0,    0,    0,  410,  411,  412,  413,  414,  415,  152,
    0,  153,    0,    0,    0,    0,    0,  154,    0,    0,
  155,  156,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  157,  158,    0,  159,    0,  160,    0,    0,  161,
    0,    0,  162,    0,    0,    0,    0,    0,    0,  163,
    0,  164,  165,    0,    0,    0,  166,    0,  167,    0,
    0,  168,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  169,  170,    0, 3581,    0,  171,    0,  172,    0,
  173,    0,    0,    0,    0,    0,  174,    0,  175,    0,
    0,    0,  176,    0,    0,    0,    0,    0,    0,    0,
  177,    0,    0,  178,    0,  179,    0,    0,    0,  180,
  181,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,  182,    0,    0,    0,  184,  185,  186,  187,
  188,    0,  189,    0,    0,    0,    0,    0,  190,    0,
  191,  192,    0,    0,  193,  194,    0,    0,    0,    0,
    0,  195,    0,    0,  196,  197,  198,  199,    0,  200,
    0,  201,    0,  202,  203,  204,  205,  206,    0,    0,
    0,    0,  207,    0,  208,    0,    0,    0, 3582,    0,
    0,    0,    0,    0,    0,  209,    0,    0,    0,    0,
  210,  211,    0,    0,    0,    0,    0,  212,    0,    0,
    0,  213,    0,  214,  215,  216,  217,    0,  218,    0,
    0,    0,    0,    0,    0,    0,  219,  220,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  221,  222,    0,
  223,  224,  225,    0,    0,    0,    0,  226,  227,    0,
  228,  229,    0,  230,    0,    0,  231,  232,    0,    0,
    0,  233,  234,    0,    0,    0,    0,  235,  236,  237,
    0,  238,  239,  240,    0,  241,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  242,  243,  244,  245,  246,
  247,  248,  249,  250,  251,  252,  253,  254,  255,  256,
  257,  258,  259,  260,  261,  262,    0,  263,  264,  265,
  266,  267,  268,    0,  269,  270,  271,  272,    0,  273,
    0,  274,  275,  276,  277,  278,  279,  280,  281,  282,
  283,  284,  285,  286,  287,  288,  289,  290,  291,    0,
  292,  293,  294,  295,    0,  296,  297,  298,  299,  300,
    0,  301,  302,  303,  304,  305,  306,  307,  308,  309,
  310,  311,  312,  313,    0,  314,  315,  316,  317,  318,
  319,    0,  320,  321,  322,  323,  324,  325,  326,  327,
    0,  328,  329,    0,    0,    0,  330,  331,  332,  333,
  334,  335,    0,    0,  336,  337,  338,  339,  340,  341,
  342,  343,  344,    0,    0,    0,    0,    0,  345,    0,
  346,  347,  348,  349,  350,  351,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  352,  353,    0,  354,  355,  356,
  357,  358,  359,  360,  361,  362,    0,  363,  364,  365,
  366,  367,  368,  369,  370,  371,  372,  373,  374,    0,
  375,  376,  377,    0,  378,    0,    0,    0,  379,  380,
  381,  382,  383,  384,  385,  386,  387,  388,  389,  390,
  391,    0,  392,  393,    0,    0,    0,    0,    0,  394,
  395,  396,  397,  398,  399,  400,  401,  402,  403,  404,
  405,  406,  407,    0,    0,  408,  409,    0,    0,    0,
    0,  410,  411,  412,  413,  414,  415,  152,    0,  153,
    0,    0,    0,    0,    0,  154,    0,    0,  155,  156,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  157,
  158,    0,  159,    0,  160,    0,    0,  161,    0,    0,
  162,    0,    0,    0,    0,    0,    0,  163,    0,  164,
  165,    0,    0,    0,  166,    0,  167,    0,    0,  168,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  169,
  170,    0,    0,    0,  171,    0,  172,    0,  173,    0,
    0,    0,    0,    0,  174,    0,  175,    0,    0,    0,
  176,    0,    0,    0,    0,    0,    0,    0,  177,    0,
    0,  178,    0,  179,    0,    0,    0,  180,  181,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
  182,  183,    0,    0,  184,  185,  186,  187,  188,    0,
  189,    0,    0,    0,    0,    0,  190,    0,  191,  192,
    0,    0,  193,  194,    0,    0,    0,    0,    0,  195,
    0,    0,  196,  197,  198,  199,    0,  200,    0,  201,
    0,  202,  203,  204,  205,  206,    0,    0,    0,    0,
  207,    0,  208,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  209,    0,    0,    0,    0,  210,  211,
    0,    0,    0,    0,    0,  212,    0,    0,    0,  213,
    0,  214,  215,  216,  217,    0,  218,    0,    0,    0,
    0,    0,    0,    0,  219,  220,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  221,  222,    0,  223,  224,
  225,    0,    0,    0,    0,  226,  227,    0,  228,  229,
    0,  230,    0,    0,  231,  232,    0,    0,    0,  233,
  234,    0,    0,    0,    0,  235,  236,  237,    0,  238,
  239,  240,    0,  241,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  242,  243,  244,  245,  246,  247,  248,
  249,  250,  251,  252,  253,  254,  255,  256,  257,  258,
  259,  260,  261,  262,    0,  263,  264,  265,  266,  267,
  268,    0,  269,  270,  271,  272,    0,  273,    0,  274,
  275,  276,  277,  278,  279,  280,  281,  282,  283,  284,
  285,  286,  287,  288,  289,  290,  291,    0,  292,  293,
  294,  295,    0,  296,  297,  298,  299,  300,    0,  301,
  302,  303,  304,  305,  306,  307,  308,  309,  310,  311,
  312,  313,    0,  314,  315,  316,  317,  318,  319,    0,
  320,  321,  322,  323,  324,  325,  326,  327,    0,  328,
  329,    0,    0,    0,  330,  331,  332,  333,  334,  335,
    0,    0,  336,  337,  338,  339,  340,  341,  342,  343,
  344,    0,    0,    0,    0,    0,  345,    0,  346,  347,
  348,  349,  350,  351,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,  352,  353,    0,  354,  355,  356,  357,  358,
  359,  360,  361,  362,    0,  363,  364,  365,  366,  367,
  368,  369,  370,  371,  372,  373,  374,    0,  375,  376,
  377,    0,  378,    0,    0,    0,  379,  380,  381,  382,
  383,  384,  385,  386,  387,  388,  389,  390,  391,    0,
  392,  393,    0,    0,    0,    0,    0,  394,  395,  396,
  397,  398,  399,  400,  401,  402,  403,  404,  405,  406,
  407,    0,    0,  408,  409,    0,    0,    0,    0,  410,
  411,  412,  413,  414,  415,  152,    0,  153,    0,    0,
    0,    0,    0,  154,    0,    0,  155,  156,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  157,  158,    0,
  159,    0,  160,    0,    0,  161,    0,    0,  162,    0,
    0,    0,    0,    0,    0,  163,    0,  164,  165,    0,
    0,    0,  166,    0,  167,    0,    0,  168,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  169,  170,    0,
    0,    0,  171,    0,  172,    0,  173,    0,    0,    0,
    0,    0,  174,    0,  175,    0,    0,    0,  176,    0,
    0,    0,    0,    0,    0,    0,  177,    0,    0,  178,
    0,  179,    0,    0,    0,  180,  181,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  182,    0,
    0,    0,  184,  185,  186,  187,  188,    0,  189,    0,
    0,    0,    0,    0,  190,    0,  191,  192,    0,    0,
  193,  194,    0,    0,    0,    0,    0,  195,    0,    0,
  196,  197,  198,  199,    0,  200,    0,  201,    0,  202,
  203,  204,  205,  206,    0,    0,    0,    0,  207,    0,
  208,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  209,    0,    0,    0,    0,  210,  211,    0,    0,
    0,    0,    0,  212,    0,    0,    0,  213,    0,  214,
  215,  216,  217,    0,  218,    0,    0,    0,    0,    0,
    0,    0,  219,  220,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  221,  222,    0,  223,  224,  225,    0,
    0,    0,    0,  226,  227,    0,  228,  229,    0,  230,
    0,    0,  231,  232,    0,    0,    0,  233,  234,    0,
    0,    0,    0,  235,  236,  237,    0,  238,  239,  240,
    0,  241,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  242,  243,  244,  245,  246,  247,  248,  249,  250,
  251,  252,  253,  254,  255,  256,  257,  258,  259,  260,
  261,  262,    0,  263,  264,  265,  266,  267,  268,    0,
  269,  270,  271,  272,    0,  273,    0,  274,  275,  276,
  277,  278,  279,  280,  281,  282,  283,  284,  285,  286,
  287,  288,  289,  290,  291,  421,  292,  293,  294,  295,
    0,  296,  297,  298,  299,  300,    0,  301,  302,  303,
  304,  305,  306,  307,  308,  309,  310,  311,  312,  313,
    0,  314,  315,  316,  317,  318,  319,    0,  320,  321,
  322,  323,  324,  325,  326,  327,    0,  328,  329,    0,
    0,    0,  330,  331,  332,  333,  334,  335,    0,    0,
  336,  337,  338,  339,  340,  341,  342,  343,  344,    0,
    0,    0,    0,    0,  345,    0,  346,  347,  348,  349,
  350,  351,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
  352,  353,    0,  354,  355,  356,  357,  358,  359,  360,
  361,  362,    0,  363,  364,  365,  366,  367,  368,  369,
  370,  371,  372,  373,  374,    0,  375,  376,  377,    0,
  378,    0,    0,    0,  379,  380,  381,  382,  383,  384,
  385,  386,  387,  388,  389,  390,  391,    0,  392,  393,
    0,    0,    0,    0,    0,  394,  395,  396,  397,  398,
  399,  400,  401,  402,  403,  404,  405,  406,  407,    0,
    0,  408,  409,    0,    0,    0,    0,  410,  411,  412,
  413,  414,  415,  152,    0,  153,    0,    0,    0,    0,
    0,  154,    0,    0,  155,  156,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  157,  158,    0,  159,    0,
  160,    0,    0,  161,    0,    0,  162,    0,    0,    0,
    0,    0,    0,  163,    0,  164,  165,    0,    0,    0,
  166,    0,  167,    0,    0,  168,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  169,  170,    0,    0,    0,
  171,    0,  172,    0,  173,    0,    0,    0,    0,    0,
  174,    0,  175,    0,    0,    0,  176,    0,    0,    0,
    0,    0,    0,    0,  177,    0,    0,  178,    0,  179,
    0,    0,    0,  180,  181,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  182,    0,    0,    0,
  184,  185,  186,  187,  188,    0,  189,    0,    0,    0,
    0,    0,  190,    0,  191,  192,    0,    0,  193,  194,
    0,    0,    0,    0,    0,  195,    0,    0,  196,  197,
  198,  199,    0,  200,    0,  201,    0,  202,  203,  204,
  205,  206,    0,    0,    0,    0,  207,    0,  208,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  209,
    0,    0,    0,    0,  210,  211,    0,    0,    0,    0,
    0,  212,    0,    0,    0,  213,    0,  214,  215,  216,
  217,    0,  218,    0,    0,    0,    0,    0,    0,    0,
  219,  220,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  221,  222,    0,  223,  224,  225,    0,    0,    0,
    0,  226,  227,    0,  228,  229,    0,  230,    0,    0,
  231,  232,    0,    0,    0,  233,  234,    0,    0,    0,
    0,  235,  236,  237,    0,  238,  239,  240,    0,  241,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  242,
  243,  244,  245,  246,  247,  248,  249,  250,  251,  252,
  253,  254,  255,  256,  257,  258,  259,  260,  261,  262,
    0,  263,  264,  265,  266,  267,  268,    0,  269,  270,
  271,  272,    0,  273,    0,  274,  275,  276,  277,  278,
  279,  280,  281,  282,  283,  284,  285,  286,  287,  288,
  289,  290,  291,    0,  292,  293,  294,  295,    0,  296,
  297,  298,  299,  300,    0,  301,  302,  303,  304,  305,
  306,  307,  308,  309,  310,  311,  312,  313,    0,  314,
  315,  316,  317,  318,  319,    0,  320,  321,  322,  323,
  324,  325,  326,  327,    0,  328,  329,    0,    0,    0,
  330,  331,  332,  333,  334,  335,    0,    0,  336,  337,
  338,  339,  340,  341,  342,  343,  344,    0,    0,    0,
    0,    0,  345,    0,  346,  347,  348,  349,  350,  351,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  352,  353,
  752,  354,  355,  356,  357,  358,  359,  360,  361,  362,
    0,  363,  364,  365,  366,  367,  368,  369,  370,  371,
  372,  373,  374,    0,  375,  376,  377,    0,  378,    0,
    0,    0,  379,  380,  381,  382,  383,  384,  385,  386,
  387,  388,  389,  390,  391,    0,  392,  393,    0,    0,
    0,    0,    0,  394,  395,  396,  397,  398,  399,  400,
  401,  402,  403,  404,  405,  406,  407,    0,    0,  408,
  409,    0,    0,    0,    0,  410,  411,  412,  413,  414,
  415,  152,    0,  153,    0,    0,    0,    0,    0,  154,
    0,    0,  155,  156,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  157,  158,    0,  159,    0,  160,    0,
    0,  161,    0,    0,  162,    0,    0,    0,    0, 1875,
    0,  163,    0,  164,  165,    0,    0,    0,  166,    0,
  167,    0,    0,  168,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  169,  170,    0,    0,    0,  171,    0,
  172,    0,  173,    0,    0,    0,    0,    0,  174,    0,
  175,    0,    0,    0,  176,    0,    0,    0,    0,    0,
    0,    0,  177,    0,    0,  178,    0,  179,    0,    0,
    0,  180,  181,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  182,    0,    0,    0,  184,  185,
  186,  187,  188,    0,  189,    0,    0,    0,    0,    0,
  190,    0,  191,  192,    0,    0,  193,  194,    0,    0,
    0,    0,    0,  195,    0,    0,  196,  197,  198,  199,
    0,  200,    0,  201,    0,  202,  203,  204,  205,  206,
    0,    0,    0,    0,  207,    0,  208,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  209,    0,    0,
    0,    0,  210,  211,    0,    0,    0,    0,    0,  212,
    0,    0,    0,  213,    0,  214,  215,  216,  217,    0,
  218,    0,    0,    0,    0,    0,    0,    0,  219,  220,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  221,
  222,    0,  223,  224,  225,    0,    0,    0,    0,  226,
  227,    0,  228,  229,    0,  230,    0,    0,  231,  232,
    0,    0,    0,  233,  234,    0,    0,    0,    0,  235,
  236,  237,    0,  238,  239,  240,    0,  241,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  242,  243,  244,
  245,  246,  247,  248,  249,  250,  251,  252,  253,  254,
  255,  256,  257,  258,  259,  260,  261,  262,    0,  263,
  264,  265,  266,  267,  268,    0,  269,  270,  271,  272,
    0,  273,    0,  274,  275,  276,  277,  278,  279,  280,
  281,  282,  283,  284,  285,  286,  287,  288,  289,  290,
  291,    0,  292,  293,  294,  295,    0,  296,  297,  298,
  299,  300,    0,  301,  302,  303,  304,  305,  306,  307,
  308,  309,  310,  311,  312,  313,    0,  314,  315,  316,
  317,  318,  319,    0,  320,  321,  322,  323,  324,  325,
  326,  327,    0,  328,  329,    0,    0,    0,  330,  331,
  332,  333,  334,  335,    0,    0,  336,  337,  338,  339,
  340,  341,  342,  343,  344,    0,    0,    0,    0,    0,
  345,    0,  346,  347,  348,  349,  350,  351,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  352,  353,    0,  354,
  355,  356,  357,  358,  359,  360,  361,  362,    0,  363,
  364,  365,  366,  367,  368,  369,  370,  371,  372,  373,
  374,    0,  375,  376,  377,    0,  378,    0,    0,    0,
  379,  380,  381,  382,  383,  384,  385,  386,  387,  388,
  389,  390,  391,    0,  392,  393,    0,    0,    0,    0,
    0,  394,  395,  396,  397,  398,  399,  400,  401,  402,
  403,  404,  405,  406,  407,    0,    0,  408,  409,    0,
    0,    0,    0,  410,  411,  412,  413,  414,  415,  152,
    0,  153,    0,    0,    0,    0,    0,  154,    0,    0,
  155,  156,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  157,  158,    0,  159, 2453,  160,    0,    0,  161,
    0,    0,  162,    0,    0,    0,    0,    0,    0,  163,
    0,  164,  165,    0,    0,    0,  166,    0,  167,    0,
    0,  168,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  169,  170,    0,    0,    0,  171,    0,  172,    0,
  173,    0,    0,    0,    0,    0,  174,    0,  175,    0,
    0,    0,  176,    0,    0,    0,    0,    0,    0,    0,
  177,    0,    0,  178,    0,  179,    0,    0,    0,  180,
  181,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,  182,    0,    0,    0,  184,  185,  186,  187,
  188,    0,  189,    0,    0,    0,    0,    0,  190,    0,
  191,  192,    0,    0,  193,  194,    0,    0,    0,    0,
    0,  195,    0,    0,  196,  197,  198,  199,    0,  200,
    0,  201,    0,  202,  203,  204,  205,  206,    0,    0,
    0,    0,  207,    0,  208,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  209,    0,    0,    0,    0,
  210,  211,    0,    0,    0,    0,    0,  212,    0,    0,
    0,  213,    0,  214,  215,  216,  217,    0,  218,    0,
    0,    0,    0,    0,    0,    0,  219,  220,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  221,  222,    0,
  223,  224,  225,    0,    0,    0,    0,  226,  227,    0,
  228,  229,    0,  230,    0,    0,  231,  232,    0,    0,
    0,  233,  234,    0,    0,    0,    0,  235,  236,  237,
    0,  238,  239,  240,    0,  241,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  242,  243,  244,  245,  246,
  247,  248,  249,  250,  251,  252,  253,  254,  255,  256,
  257,  258,  259,  260,  261,  262,    0,  263,  264,  265,
  266,  267,  268,    0,  269,  270,  271,  272,    0,  273,
    0,  274,  275,  276,  277,  278,  279,  280,  281,  282,
  283,  284,  285,  286,  287,  288,  289,  290,  291,    0,
  292,  293,  294,  295,    0,  296,  297,  298,  299,  300,
    0,  301,  302,  303,  304,  305,  306,  307,  308,  309,
  310,  311,  312,  313,    0,  314,  315,  316,  317,  318,
  319,    0,  320,  321,  322,  323,  324,  325,  326,  327,
    0,  328,  329,    0,    0,    0,  330,  331,  332,  333,
  334,  335,    0,    0,  336,  337,  338,  339,  340,  341,
  342,  343,  344,    0,    0,    0,    0,    0,  345,    0,
  346,  347,  348,  349,  350,  351,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  352,  353,    0,  354,  355,  356,
  357,  358,  359,  360,  361,  362,    0,  363,  364,  365,
  366,  367,  368,  369,  370,  371,  372,  373,  374,    0,
  375,  376,  377,    0,  378,    0,    0,    0,  379,  380,
  381,  382,  383,  384,  385,  386,  387,  388,  389,  390,
  391,    0,  392,  393,    0,    0,    0,    0,    0,  394,
  395,  396,  397,  398,  399,  400,  401,  402,  403,  404,
 2454,  406,  407,    0,    0,  408,  409,    0,    0,    0,
    0,  410,  411,  412,  413,  414,  415,  152,    0,  153,
    0,    0,    0,    0,    0,  154,    0,    0,  155,  156,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  157,
  158,    0,  159, 2787,  160,    0,    0,  161,    0,    0,
  162,    0,    0,    0,    0,    0,    0,  163,    0,  164,
  165,    0,    0,    0,  166,    0,  167,    0,    0,  168,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  169,
  170,    0,    0,    0,  171,    0,  172,    0,  173,    0,
    0,    0,    0,    0,  174,    0,  175,    0,    0,    0,
  176,    0,    0,    0,    0,    0,    0,    0,  177,    0,
    0,  178,    0,  179,    0,    0,    0,  180,  181,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
  182,    0,    0,    0,  184,  185,  186,  187,  188,    0,
  189,    0,    0,    0,    0,    0,  190,    0,  191,  192,
    0,    0,  193,  194,    0,    0,    0,    0,    0,  195,
    0,    0,  196,  197,  198,  199,    0,  200,    0,  201,
    0,  202,  203,  204,  205,  206,    0,    0,    0,    0,
  207,    0,  208,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  209,    0,    0,    0,    0,  210,  211,
    0,    0,    0,    0,    0,  212,    0,    0,    0,  213,
    0,  214,  215,  216,  217,    0,  218,    0,    0,    0,
    0,    0,    0,    0,  219,  220,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  221,  222,    0,  223,  224,
  225,    0,    0,    0,    0,  226,  227,    0,  228,  229,
    0,  230,    0,    0,  231,  232,    0,    0,    0,  233,
  234,    0,    0,    0,    0,  235,  236,  237,    0,  238,
  239,  240,    0,  241,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  242,  243,  244,  245,  246,  247,  248,
  249,  250,  251,  252,  253,  254,  255,  256,  257,  258,
  259,  260,  261,  262,    0,  263,  264,  265,  266,  267,
  268,    0,  269,  270,  271,  272,    0,  273,    0,  274,
  275,  276,  277,  278,  279,  280,  281,  282,  283,  284,
  285,  286,  287,  288,  289,  290,  291,    0,  292,  293,
  294,  295,    0,  296,  297,  298,  299,  300,    0,  301,
  302,  303,  304,  305,  306,  307,  308,  309,  310,  311,
  312,  313,    0,  314,  315,  316,  317,  318,  319,    0,
  320,  321,  322,  323,  324,  325,  326,  327,    0,  328,
  329,    0,    0,    0,  330,  331,  332,  333,  334,  335,
    0,    0,  336,  337,  338,  339,  340,  341,  342,  343,
  344,    0,    0,    0,    0,    0,  345,    0,  346,  347,
  348,  349,  350,  351,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,  352,  353,    0,  354,  355,  356,  357,  358,
  359,  360,  361,  362,    0,  363,  364,  365,  366,  367,
  368,  369,  370,  371,  372,  373,  374,    0,  375,  376,
  377,    0,  378,    0,    0,    0,  379,  380,  381,  382,
  383,  384,  385,  386,  387,  388,  389,  390,  391,    0,
  392,  393,    0,    0,    0,    0,    0,  394,  395,  396,
  397,  398,  399,  400,  401,  402,  403,  404,  405,  406,
  407,    0,    0,  408,  409,    0,    0,    0,    0,  410,
  411,  412,  413,  414,  415,  152,    0,  153,    0,    0,
    0,    0,    0,  154,    0,    0,  155,  156,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  157,  158,    0,
  159,    0,  160,    0,    0,  161,    0,    0,  162,    0,
    0,    0,    0,    0,    0,  163,    0,  164,  165,    0,
    0,    0,  166,    0,  167,    0,    0,  168, 2886,    0,
    0,    0,    0,    0,    0,    0,    0,  169,  170,    0,
    0,    0,  171,    0,  172,    0,  173,    0,    0,    0,
    0,    0,  174,    0,  175,    0,    0,    0,  176,    0,
    0,    0,    0,    0,    0,    0,  177,    0,    0,  178,
    0,  179,    0,    0,    0,  180,  181,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  182,    0,
    0,    0,  184,  185,  186,  187,  188,    0,  189,    0,
    0,    0,    0,    0,  190,    0,  191,  192,    0,    0,
  193,  194,    0,    0,    0,    0,    0,  195,    0,    0,
  196,  197,  198,  199,    0,  200,    0,  201,    0,  202,
  203,  204,  205,  206,    0,    0,    0,    0,  207,    0,
  208,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  209,    0,    0,    0,    0,  210,  211,    0,    0,
    0,    0,    0,  212,    0,    0,    0,  213,    0,  214,
  215,  216,  217,    0,  218,    0,    0,    0,    0,    0,
    0,    0,  219,  220,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  221,  222,    0,  223,  224,  225,    0,
    0,    0,    0,  226,  227,    0,  228,  229,    0,  230,
    0,    0,  231,  232,    0,    0,    0,  233,  234,    0,
    0,    0,    0,  235,  236,  237,    0,  238,  239,  240,
    0,  241,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  242,  243,  244,  245,  246,  247,  248,  249,  250,
  251,  252,  253,  254,  255,  256,  257,  258,  259,  260,
  261,  262,    0,  263,  264,  265,  266,  267,  268,    0,
  269,  270,  271,  272,    0,  273,    0,  274,  275,  276,
  277,  278,  279,  280,  281,  282,  283,  284,  285,  286,
  287,  288,  289,  290,  291,    0,  292,  293,  294,  295,
    0,  296,  297,  298,  299,  300,    0,  301,  302,  303,
  304,  305,  306,  307,  308,  309,  310,  311,  312,  313,
    0,  314,  315,  316,  317,  318,  319,    0,  320,  321,
  322,  323,  324,  325,  326,  327,    0,  328,  329,    0,
    0,    0,  330,  331,  332,  333,  334,  335,    0,    0,
  336,  337,  338,  339,  340,  341,  342,  343,  344,    0,
    0,    0,    0,    0,  345,    0,  346,  347,  348,  349,
  350,  351,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
  352,  353,    0,  354,  355,  356,  357,  358,  359,  360,
  361,  362,    0,  363,  364,  365,  366,  367,  368,  369,
  370,  371,  372,  373,  374,    0,  375,  376,  377,    0,
  378,    0,    0,    0,  379,  380,  381,  382,  383,  384,
  385,  386,  387,  388,  389,  390,  391,    0,  392,  393,
    0,    0,    0,    0,    0,  394,  395,  396,  397,  398,
  399,  400,  401,  402,  403,  404,  405,  406,  407,    0,
    0,  408,  409,    0,    0,    0,    0,  410,  411,  412,
  413,  414,  415,  152,    0,  153,    0,    0,    0,    0,
    0,  154,    0,    0,  155,  156,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  157,  158,    0,  159,    0,
  160,    0,    0,  161,    0,    0,  162,    0,    0,    0,
    0,    0,    0,  163,    0,  164,  165,    0,    0,    0,
  166,    0,  167,    0,    0,  168,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  169,  170,    0,    0,    0,
  171,    0,  172,    0,  173,    0,    0,    0,    0,    0,
  174,    0,  175,    0,    0,    0,  176,    0,    0,    0,
    0,    0,    0,    0,  177,    0,    0,  178,    0,  179,
    0,    0,    0,  180,  181,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  182,    0,    0,    0,
  184,  185,  186,  187,  188,    0,  189,    0,    0,    0,
    0,    0,  190,    0,  191,  192,    0,    0,  193,  194,
    0,    0,    0,    0,    0,  195,    0,    0,  196,  197,
  198,  199,    0,  200,    0,  201,    0,  202,  203,  204,
  205,  206,    0,    0,    0,    0,  207,    0,  208,    0,
    0,    0, 2997,    0,    0,    0,    0,    0,    0,  209,
    0,    0,    0,    0,  210,  211,    0,    0,    0,    0,
    0,  212,    0,    0,    0,  213,    0,  214,  215,  216,
 2998,    0,  218,    0,    0,    0,    0,    0,    0,    0,
  219,  220,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  221,  222,    0,  223,  224,  225,    0,    0,    0,
    0,  226,  227,    0,  228,  229,    0,  230,    0,    0,
  231,  232,    0,    0,    0,  233,  234,    0,    0,    0,
    0,  235,  236,  237,    0,  238,  239,  240,    0,  241,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  242,
  243,  244,  245,  246,  247,  248,  249,  250,  251,  252,
  253,  254,  255,  256,  257,  258,  259,  260,  261,  262,
    0,  263,  264,  265,  266,  267,  268,    0,  269,  270,
  271,  272,    0,  273,    0,  274,  275,  276,  277,  278,
  279,  280,  281,  282,  283,  284,  285,  286,  287,  288,
  289,  290,  291,    0,  292,  293,  294,  295,    0,  296,
  297,  298,  299,  300,    0,  301,  302,  303,  304,  305,
  306,  307,  308,  309,  310,  311,  312,  313,    0,  314,
  315,  316,  317,  318,  319,    0,  320,  321,  322,  323,
  324,  325,  326,  327,    0,  328,  329,    0,    0,    0,
  330,  331,  332,  333,  334,  335,    0,    0,  336,  337,
  338,  339,  340,  341,  342,  343,  344,    0,    0,    0,
    0,    0,  345,    0,  346,  347,  348,  349,  350,  351,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  352,  353,
    0,  354,  355,  356,  357,  358,  359,  360,  361,  362,
    0,  363,  364,  365,  366,  367,  368,  369,  370,  371,
  372,  373,  374,    0,  375,  376,  377,    0,  378,    0,
    0,    0,  379,  380,  381,  382,  383,  384,  385,  386,
  387,  388,  389,  390,  391,    0,  392,  393,    0,    0,
    0,    0,    0,  394,  395,  396,  397,  398,  399,  400,
  401,  402,  403,  404,  405,  406,  407,    0,    0,  408,
  409,    0,    0,    0,    0,  410,  411,  412,  413,  414,
  415,  152,    0,  153,    0,    0,    0,    0,    0,  154,
    0,    0,  155,  156,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  157,  158,    0,  159,    0,  160,    0,
    0,  161,    0,    0,  162,    0,    0,    0,    0,    0,
    0,  163,    0,  164,  165,    0,    0,    0,  166,    0,
  167,    0,    0,  168,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  169,  170,    0,    0,    0,  171,    0,
  172,    0,  173,    0,    0,    0,    0,    0,  174,    0,
  175,    0,    0,    0,  176,    0,    0,    0,    0,    0,
    0,    0,  177,    0,    0,  178,    0,  179,    0,    0,
    0,  180,  181,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  182,    0,    0,    0,  184,  185,
  186,  187,  188,    0,  189,    0,    0,    0,    0,    0,
  190,    0,  191,  192,    0,    0,  193,  194,    0,    0,
    0,    0,    0,  195,    0,    0,  196,  197,  198,  199,
    0,  200,    0,  201,    0,  202,  203,  204,  205,  206,
    0,    0,    0,    0,  207,    0,  208,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  209,    0,    0,
    0,    0,  210,  211,    0,    0,    0,    0,    0,  212,
    0,    0,    0,  213,    0,  214,  215,  216,  217, 3089,
  218,    0,    0,    0,    0,    0,    0,    0,  219,  220,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  221,
  222,    0,  223,  224,  225,    0,    0,    0,    0,  226,
  227,    0,  228,  229,    0,  230,    0,    0,  231,  232,
    0,    0,    0,  233,  234,    0,    0,    0,    0,  235,
  236,  237,    0,  238,  239,  240,    0,  241,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  242,  243,  244,
  245,  246,  247,  248,  249,  250,  251,  252,  253,  254,
  255,  256,  257,  258,  259,  260,  261,  262,    0,  263,
  264,  265,  266,  267,  268,    0,  269,  270,  271,  272,
    0,  273,    0,  274,  275,  276,  277,  278,  279,  280,
  281,  282,  283,  284,  285,  286,  287,  288,  289,  290,
  291,    0,  292,  293,  294,  295,    0,  296,  297,  298,
  299,  300,    0,  301,  302,  303,  304,  305,  306,  307,
  308,  309,  310,  311,  312,  313,    0,  314,  315,  316,
  317,  318,  319,    0,  320,  321,  322,  323,  324,  325,
  326,  327,    0,  328,  329,    0,    0,    0,  330,  331,
  332,  333,  334,  335,    0,    0,  336,  337,  338,  339,
  340,  341,  342,  343,  344,    0,    0,    0,    0,    0,
  345,    0,  346,  347,  348,  349,  350,  351,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  352,  353,    0,  354,
  355,  356,  357,  358,  359,  360,  361,  362,    0,  363,
  364,  365,  366,  367,  368,  369,  370,  371,  372,  373,
  374,    0,  375,  376,  377,    0,  378,    0,    0,    0,
  379,  380,  381,  382,  383,  384,  385,  386,  387,  388,
  389,  390,  391,    0,  392,  393,    0,    0,    0,    0,
    0,  394,  395,  396,  397,  398,  399,  400,  401,  402,
  403,  404,  405,  406,  407,    0,    0,  408,  409,    0,
    0,    0,    0,  410,  411,  412,  413,  414,  415,  152,
    0,  153,    0,    0,    0,    0,    0,  154,    0,    0,
  155,  156,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  157,  158,    0,  159,    0,  160,    0,    0,  161,
    0,    0,  162,    0,    0,    0,    0,    0,    0,  163,
    0,  164,  165,    0,    0,    0,  166,    0,  167,    0,
    0,  168,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  169,  170,    0,    0,    0,  171,    0,  172,    0,
  173,    0,    0,    0,    0,    0,  174,    0,  175,    0,
    0,    0,  176,    0,    0,    0,    0,    0,    0,    0,
  177,    0,    0,  178,    0,  179,    0,    0,    0,  180,
  181,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,  182,    0,    0,    0,  184,  185,  186,  187,
  188,    0,  189,    0,    0,    0,    0,    0,  190,    0,
  191,  192,    0,    0,  193,  194,    0,    0,    0,    0,
    0,  195,    0,    0,  196,  197,  198,  199,    0,  200,
    0,  201,    0,  202,  203,  204,  205,  206,    0,    0,
    0,    0,  207,    0,  208,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  209,    0,    0,    0,    0,
  210,  211,    0,    0,    0,    0,    0,  212,    0,    0,
    0,  213,    0,  214,  215,  216,  217, 3240,  218,    0,
    0,    0,    0,    0,    0,    0,  219,  220,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  221,  222,    0,
  223,  224,  225,    0,    0,    0,    0,  226,  227,    0,
  228,  229,    0,  230,    0,    0,  231,  232,    0,    0,
    0,  233,  234,    0,    0,    0,    0,  235,  236,  237,
    0,  238,  239,  240,    0,  241,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  242,  243,  244,  245,  246,
  247,  248,  249,  250,  251,  252,  253,  254,  255,  256,
  257,  258,  259,  260,  261,  262,    0,  263,  264,  265,
  266,  267,  268,    0,  269,  270,  271,  272,    0,  273,
    0,  274,  275,  276,  277,  278,  279,  280,  281,  282,
  283,  284,  285,  286,  287,  288,  289,  290,  291,    0,
  292,  293,  294,  295,    0,  296,  297,  298,  299,  300,
    0,  301,  302,  303,  304,  305,  306,  307,  308,  309,
  310,  311,  312,  313,    0,  314,  315,  316,  317,  318,
  319,    0,  320,  321,  322,  323,  324,  325,  326,  327,
    0,  328,  329,    0,    0,    0,  330,  331,  332,  333,
  334,  335,    0,    0,  336,  337,  338,  339,  340,  341,
  342,  343,  344,    0,    0,    0,    0,    0,  345,    0,
  346,  347,  348,  349,  350,  351,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  352,  353,    0,  354,  355,  356,
  357,  358,  359,  360,  361,  362,    0,  363,  364,  365,
  366,  367,  368,  369,  370,  371,  372,  373,  374,    0,
  375,  376,  377,    0,  378,    0,    0,    0,  379,  380,
  381,  382,  383,  384,  385,  386,  387,  388,  389,  390,
  391,    0,  392,  393,    0,    0,    0,    0,    0,  394,
  395,  396,  397,  398,  399,  400,  401,  402,  403,  404,
  405,  406,  407,    0,    0,  408,  409,    0,    0,    0,
    0,  410,  411,  412,  413,  414,  415,  152,    0,  153,
    0,    0,    0,    0,    0,  154,    0,    0,  155,  156,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  157,
  158,    0,  159,    0,  160,    0,    0,  161,    0,    0,
  162,    0,    0,    0,    0,    0,    0,  163,    0,  164,
  165,    0,    0,    0,  166,    0,  167,    0,    0,  168,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  169,
  170,    0,    0,    0,  171,    0,  172,    0,  173,    0,
    0,    0,    0,    0,  174,    0,  175,    0,    0,    0,
  176,    0,    0,    0,    0,    0,    0,    0,  177,    0,
    0,  178,    0,  179,    0,    0,    0,  180,  181,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
  182,    0,    0,    0,  184,  185,  186,  187,  188,    0,
  189,    0,    0,    0,    0,    0,  190,    0,  191,  192,
    0,    0,  193,  194,    0,    0,    0,    0,    0,  195,
    0,    0,  196,  197,  198,  199,    0,  200,    0,  201,
    0,  202,  203,  204,  205,  206,    0,    0,    0,    0,
  207,    0,  208,    0,    0,    0, 3508,    0,    0,    0,
    0,    0,    0,  209,    0,    0,    0,    0,  210,  211,
    0,    0,    0,    0,    0,  212,    0,    0,    0,  213,
    0,  214,  215,  216,  217,    0,  218,    0,    0,    0,
    0,    0,    0,    0,  219,  220,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  221,  222,    0,  223,  224,
  225,    0,    0,    0,    0,  226,  227,    0,  228,  229,
    0,  230,    0,    0,  231,  232,    0,    0,    0,  233,
  234,    0,    0,    0,    0,  235,  236,  237,    0,  238,
  239,  240,    0,  241,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  242,  243,  244,  245,  246,  247,  248,
  249,  250,  251,  252,  253,  254,  255,  256,  257,  258,
  259,  260,  261,  262,    0,  263,  264,  265,  266,  267,
  268,    0,  269,  270,  271,  272,    0,  273,    0,  274,
  275,  276,  277,  278,  279,  280,  281,  282,  283,  284,
  285,  286,  287,  288,  289,  290,  291,    0,  292,  293,
  294,  295,    0,  296,  297,  298,  299,  300,    0,  301,
  302,  303,  304,  305,  306,  307,  308,  309,  310,  311,
  312,  313,    0,  314,  315,  316,  317,  318,  319,    0,
  320,  321,  322,  323,  324,  325,  326,  327,    0,  328,
  329,    0,    0,    0,  330,  331,  332,  333,  334,  335,
    0,    0,  336,  337,  338,  339,  340,  341,  342,  343,
  344,    0,    0,    0,    0,    0,  345,    0,  346,  347,
  348,  349,  350,  351,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,  352,  353,    0,  354,  355,  356,  357,  358,
  359,  360,  361,  362,    0,  363,  364,  365,  366,  367,
  368,  369,  370,  371,  372,  373,  374,    0,  375,  376,
  377,    0,  378,    0,    0,    0,  379,  380,  381,  382,
  383,  384,  385,  386,  387,  388,  389,  390,  391,    0,
  392,  393,    0,    0,    0,    0,    0,  394,  395,  396,
  397,  398,  399,  400,  401,  402,  403,  404,  405,  406,
  407,    0,    0,  408,  409,    0,    0,    0,    0,  410,
  411,  412,  413,  414,  415,  152,    0,  153,    0,    0,
    0,    0,    0,  154,    0,    0,  155,  156,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  157,  158,    0,
  159,    0,  160,    0,    0,  161,    0,    0,  162,    0,
    0,    0,    0,    0,    0,  163,    0,  164,  165, 3597,
    0,    0,  166,    0,  167,    0,    0,  168,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  169,  170,    0,
    0,    0,  171,    0,  172,    0,  173,    0,    0,    0,
    0,    0,  174,    0,  175,    0,    0,    0,  176,    0,
    0,    0,    0,    0,    0,    0,  177,    0,    0,  178,
    0,  179,    0,    0,    0,  180,  181,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  182,    0,
    0,    0,  184,  185,  186,  187,  188,    0,  189,    0,
    0,    0,    0,    0,  190,    0,  191,  192,    0,    0,
  193,  194,    0,    0,    0,    0,    0,  195,    0,    0,
  196,  197,  198,  199,    0,  200,    0,  201,    0,  202,
  203,  204,  205,  206,    0,    0,    0,    0,  207,    0,
  208,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  209,    0,    0,    0,    0,  210,  211,    0,    0,
    0,    0,    0,  212,    0,    0,    0,  213,    0,  214,
  215,  216,  217,    0,  218,    0,    0,    0,    0,    0,
    0,    0,  219,  220,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  221,  222,    0,  223,  224,  225,    0,
    0,    0,    0,  226,  227,    0,  228,  229,    0,  230,
    0,    0,  231,  232,    0,    0,    0,  233,  234,    0,
    0,    0,    0,  235,  236,  237,    0,  238,  239,  240,
    0,  241,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  242,  243,  244,  245,  246,  247,  248,  249,  250,
  251,  252,  253,  254,  255,  256,  257,  258,  259,  260,
  261,  262,    0,  263,  264,  265,  266,  267,  268,    0,
  269,  270,  271,  272,    0,  273,    0,  274,  275,  276,
  277,  278,  279,  280,  281,  282,  283,  284,  285,  286,
  287,  288,  289,  290,  291,    0,  292,  293,  294,  295,
    0,  296,  297,  298,  299,  300,    0,  301,  302,  303,
  304,  305,  306,  307,  308,  309,  310,  311,  312,  313,
    0,  314,  315,  316,  317,  318,  319,    0,  320,  321,
  322,  323,  324,  325,  326,  327,    0,  328,  329,    0,
    0,    0,  330,  331,  332,  333,  334,  335,    0,    0,
  336,  337,  338,  339,  340,  341,  342,  343,  344,    0,
    0,    0,    0,    0,  345,    0,  346,  347,  348,  349,
  350,  351,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
  352,  353,    0,  354,  355,  356,  357,  358,  359,  360,
  361,  362,    0,  363,  364,  365,  366,  367,  368,  369,
  370,  371,  372,  373,  374,    0,  375,  376,  377,    0,
  378,    0,    0,    0,  379,  380,  381,  382,  383,  384,
  385,  386,  387,  388,  389,  390,  391,    0,  392,  393,
    0,    0,    0,    0,    0,  394,  395,  396,  397,  398,
  399,  400,  401,  402,  403,  404,  405,  406,  407,    0,
    0,  408,  409,    0,    0,    0,    0,  410,  411,  412,
  413,  414,  415,  152,    0,  153,    0,    0,    0,    0,
    0,  154,    0,    0,  155,  156,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  157,  158,    0,  159,    0,
  160,    0,    0,  161,    0,    0,  162,    0,    0,    0,
    0,    0,    0,  163,    0,  164,  165,    0,    0,    0,
  166,    0,  167,    0,    0,  168,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  169,  170,    0,    0,    0,
  171,    0,  172,    0,  173,    0,    0,    0,    0,    0,
  174,    0,  175,    0,    0,    0,  176,    0,    0,    0,
    0,    0,    0,    0,  177,    0,    0,  178,    0,  179,
    0,    0,    0,  180,  181,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  182,    0,    0,    0,
  184,  185,  186,  187,  188,    0,  189,    0,    0,    0,
    0,    0,  190,    0,  191,  192,    0,    0,  193,  194,
    0,    0,    0,    0,    0,  195,    0,    0,  196,  197,
  198,  199,    0,  200,    0,  201,    0,  202,  203,  204,
  205,  206,    0,    0,    0,    0,  207,    0,  208,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  209,
    0,    0,    0,    0,  210,  211,    0,    0,    0,    0,
    0,  212,    0,    0,    0,  213,    0,  214,  215,  216,
  217,    0,  218,    0,    0,    0,    0,    0,    0,    0,
  219,  220,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  221,  222,    0,  223,  224,  225,    0,    0,    0,
    0,  226,  227,    0,  228,  229,    0,  230,    0,    0,
  231,  232,    0,    0,    0,  233,  234,    0,    0,    0,
    0,  235,  236,  237,    0,  238,  239,  240,    0,  241,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  242,
  243,  244,  245,  246,  247,  248,  249,  250,  251,  252,
  253,  254,  255,  256,  257,  258,  259,  260,  261,  262,
    0,  263,  264,  265,  266,  267,  268,    0,  269,  270,
  271,  272,    0,  273,    0,  274,  275,  276,  277,  278,
  279,  280,  281,  282,  283,  284,  285,  286,  287,  288,
  289,  290,  291,    0,  292,  293,  294,  295,    0,  296,
  297,  298,  299,  300,    0,  301,  302,  303,  304,  305,
  306,  307,  308,  309,  310,  311,  312,  313,    0,  314,
  315,  316,  317,  318,  319,    0,  320,  321,  322,  323,
  324,  325,  326,  327,    0,  328,  329,    0,    0,    0,
  330,  331,  332,  333,  334,  335,    0,    0,  336,  337,
  338,  339,  340,  341,  342,  343,  344,    0,    0,    0,
    0,    0,  345,    0,  346,  347,  348,  349,  350,  351,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  352,  353,
    0,  354,  355,  356,  357,  358,  359,  360,  361,  362,
    0,  363,  364,  365,  366,  367,  368,  369,  370,  371,
  372,  373,  374,    0,  375,  376,  377,    0,  378,    0,
    0,    0,  379,  380,  381,  382,  383,  384,  385,  386,
  387,  388,  389,  390,  391,    0,  392,  393,    0,    0,
    0,    0,    0,  394,  395,  396,  397,  398,  399,  400,
  401,  402,  403,  404,  405,  406,  407,    0,    0,  408,
  409,    0,    0,    0,    0,  410,  411,  412,  413,  414,
  415,  152,    0,  153,    0,    0,    0,    0,    0,  154,
    0,    0,  155,  156,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  157,  158,    0,  159,    0,  160,    0,
    0,  161,    0,    0,  162,    0,    0,    0,    0,    0,
    0,  163,    0,  164,  165,    0,    0,    0,  166,    0,
  167,    0,    0,  168,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  169,  170,    0,    0,    0,  171,    0,
  172,    0,  173,    0,    0,    0,    0,    0,  174,    0,
  175,    0,    0,    0,  176,    0,    0,    0,    0,    0,
    0,    0,  177,    0,    0,  178,    0,  179,    0,    0,
    0,  180,  181,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  182,    0,    0,    0,  184,  185,
  186,  187,  188,    0,  189,    0,    0,    0,    0,    0,
  190,    0,  191,  192,    0,    0,  193,  194,    0,    0,
    0,    0,    0,  195,    0,    0,  196,  197,  198,  199,
    0,  200,    0,  201,    0,  202,  203,  204,  205,  206,
    0,    0,    0,    0,  207,    0,  208,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  209,    0,    0,
    0,    0,  210,  211,    0,    0,    0,    0,    0,  212,
    0,    0,    0,  213,    0,  214,  215,  216,  217,    0,
  218,    0,    0,    0,    0,    0,    0,    0,  219,  220,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  221,
  222,    0,  223,  224,  225,    0,    0,    0,    0,  226,
  227,    0,  228,  229,    0,  230,    0,    0,  231,  232,
    0,    0,    0,  233,  234,    0,    0,    0,    0,  235,
  236,  237,    0,  238,  239,  240,    0,  241,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  242,  243,  244,
  245,  246,  247,  248,  249,  250,  251,  252,  253,  254,
  255,  256,  257,  258,  259,  260,  261,  262,    0,  263,
  264,  265,  266,  267,  268,    0,  269,  270,  271,  272,
    0,  273,    0,  274,  275,  276,  277,  278,  279,  280,
  281,  282,  283,  284,  285,  286,  287,  288,  289,  290,
  291,    0,  292,  293,  294,  295,    0,  296,  297,  298,
  299,  300,    0,  301,  302,  303,  304,  305,  306,  307,
  308,  309,  310,  311,  312,  313,    0,  314,  315,  316,
  317,  318,  319,    0,  320,  638,  322,  323,  324,  325,
  326,  327,    0,  328,  329,    0,    0,    0,  330,  331,
  332,  333,  334,  335,    0,    0,  336,  337,  338,  339,
  340,  341,  342,  343,  344,    0,    0,    0,    0,    0,
  345,    0,  346,  347,  348,  349,  350,  351,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  352,  353,    0,  354,
  355,  356,  357,  358,  359,  360,  361,  362,    0,  363,
  364,  365,  366,  367,  368,  369,  370,  371,  372,  373,
  374,    0,  375,  376,  377,    0,  378,    0,    0,    0,
  379,  380,  381,  382,  383,  384,  385,  386,  387,  388,
  389,  390,  391,    0,  392,  393,    0,    0,    0,    0,
    0,  394,  395,  396,  397,  398,  399,  400,  401,  402,
  403,  404,  405,  406,  407,    0,    0,  408,  409,    0,
    0,    0,    0,  410,  411,  412,  413,  414,  415,  152,
    0,  153,    0,    0,    0,    0,    0,  154,    0,    0,
  155,  156,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  157,  158,    0,  159,    0,  160,    0,    0,  161,
    0,    0,  162,    0,    0,    0,    0,    0,    0,  163,
    0,  164,  165,    0,    0,    0,  166,    0,  167,    0,
    0,  168,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  169,  170,    0,    0,    0,  171,    0,  172,    0,
  173,    0,    0,    0,    0,    0,  174,    0,  175,    0,
    0,    0,  176,    0,    0,    0,    0,    0,    0,    0,
  177,    0,    0,  178,    0,  179,    0,    0,    0,  180,
  181,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,  182,    0,    0,    0,  184,  185,  186,  187,
  188,    0,  189,    0,    0,    0,    0,    0,  190,    0,
  191,  192,    0,    0,  193,  194,    0,    0,    0,    0,
    0,  195,    0,    0,  196,  197,  198,  199,    0,  200,
    0,  201,    0,  202,  203,  204,  205,  206,    0,    0,
    0,    0,  207,    0,  208,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  209,    0,    0,    0,    0,
  210,  211,    0,    0,    0,    0,    0,  212,    0,    0,
    0,  213,    0,  214,  215,  216,  217,    0,  218,    0,
    0,    0,    0,    0,    0,    0,  219,  220,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  221,  222,    0,
  223,  224,  225,    0,    0,    0,    0,  226,  227,    0,
  228,  229,    0,  230,    0,    0,  231,  232,    0,    0,
    0,  233,  234,    0,    0,    0,    0,  235,  236,  237,
    0,  238,  239,  240,    0,  241,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  242,  243,  244,  245,  246,
  247,  248,  249,  250,  251,  252,  253,  254,  255,  256,
  257,  258,  259,  260,  261,  262,    0,  263,  264,  265,
  266,  267,  268,    0,  269,  270,  271,  272,    0,  273,
    0,  274,  275,  276,  277,  278,  279,  280,  281,  282,
  283,  284,  285,  286,  287,  288,  289,  290,  291,    0,
  292,  293,  294,  295,    0,  296,  297,  298,  299,  300,
    0,  301,  302,  303,  304,  305,  306,  307,  308,  309,
  310,  311,  312,  313,    0,  314,  315,  316,  317,  318,
  319,    0,  320,  668,  322,  323,  324,  325,  326,  327,
    0,  328,  329,    0,    0,    0,  330,  331,  332,  333,
  334,  335,    0,    0,  336,  337,  338,  339,  340,  341,
  342,  343,  344,    0,    0,    0,    0,    0,  345,    0,
  346,  347,  348,  349,  350,  351,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  352,  353,    0,  354,  355,  356,
  357,  358,  359,  360,  361,  362,    0,  363,  364,  365,
  366,  367,  368,  369,  370,  371,  372,  373,  374,    0,
  375,  376,  377,    0,  378,    0,    0,    0,  379,  380,
  381,  382,  383,  384,  385,  386,  387,  388,  389,  390,
  391,    0,  392,  393,    0,    0,    0,    0,    0,  394,
  395,  396,  397,  398,  399,  400,  401,  402,  403,  404,
  405,  406,  407,    0,    0,  408,  409,    0,    0,    0,
    0,  410,  411,  412,  413,  414,  415,  152,    0,  153,
    0,    0,    0,    0,    0,  154,    0,    0,  155,  156,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  157,
  158,    0,  159,    0,  160,    0,    0,  161,    0,    0,
  162,    0,    0,    0,    0,    0,    0,  163,    0,  164,
  165,    0,    0,    0,  166,    0,  167,    0,    0,  168,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  169,
  170,    0,    0,    0,  171,    0,  172,    0,  173,    0,
    0,    0,    0,    0,  174,    0,  175,    0,    0,    0,
  176,    0,    0,    0,    0,    0,    0,    0,  177,    0,
    0,  178,    0,  179,    0,    0,    0,  180,  181,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
  182,    0,    0,    0,  184,  185,  186,  187,  188,    0,
  189,    0,    0,    0,    0,    0,  190,    0,  191,  192,
    0,    0,  193,  194,    0,    0,    0,    0,    0,  195,
    0,    0,  196,  197,  198,  199,    0,  200,    0,  201,
    0,  202,  203,  204,  205,  206,    0,    0,    0,    0,
  207,    0,  208,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  209,    0,    0,    0,    0,  210,  211,
    0,    0,    0,    0,    0,  212,    0,    0,    0,  213,
    0,  214,  215,  216,  217,    0,  218,    0,    0,    0,
    0,    0,    0,    0,  219,  220,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  221,  222,    0,  223,  224,
  225,    0,    0,    0,    0,  226,  227,    0,  228,  229,
    0,  230,    0,    0,  231,  232,    0,    0,    0,  233,
  234,    0,    0,    0,    0,  235,  236,  237,    0,  238,
  239,  240,    0,  241,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  242,  243,  244,  245,  246,  247,  248,
  249,  250,  251,  252,  253,  254,  255,  256,  257,  258,
  259,  260,  261,  262,    0,  263,  264,  265,  266,  267,
  268,    0,  269,  270,  271,  272,    0,  273,    0,  274,
  275,  276,  277,  278,  279,  280,  281,  282,  283,  284,
  285,  286,  287,  288,  289,  290,  291,    0,  292,  293,
  294,  295,    0,  296,  297,  298,  299,  300,    0,  301,
  302,  303,  304,  305,  306,  307,  308,  309,  310,  311,
  312,  313,    0,  314,  315,  316,  317,  318,  319,    0,
  320,  709,  322,  323,  324,  325,  326,  327,    0,  328,
  329,    0,    0,    0,  330,  331,  332,  333,  334,  335,
    0,    0,  336,  337,  338,  339,  340,  341,  342,  343,
  344,    0,    0,    0,    0,    0,  345,    0,  346,  347,
  348,  349,  350,  351,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,  352,  353,    0,  354,  355,  356,  357,  358,
  359,  360,  361,  362,    0,  363,  364,  365,  366,  367,
  368,  369,  370,  371,  372,  373,  374,    0,  375,  376,
  377,    0,  378,    0,    0,    0,  379,  380,  381,  382,
  383,  384,  385,  386,  387,  388,  389,  390,  391,    0,
  392,  393,    0,    0,    0,    0,    0,  394,  395,  396,
  397,  398,  399,  400,  401,  402,  403,  404,  405,  406,
  407,    0,    0,  408,  409,    0,    0,    0,    0,  410,
  411,  412,  413,  414,  415,  152,    0,  153,    0,    0,
    0,    0,    0,  154,    0,    0,  155,  156,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  157,  158,    0,
  159,    0,  160,    0,    0,  161,    0,    0,  162,    0,
    0,    0,    0,    0,    0,  163,    0,  164,  165,    0,
    0,    0,  166,    0,  167,    0,    0,  168,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  169,  170,    0,
    0,    0,  171,    0,  172,    0,  173,    0,    0,    0,
    0,    0,  174,    0,  175,    0,    0,    0,  176,    0,
    0,    0,    0,    0,    0,    0,  177,    0,    0,  178,
    0,  179,    0,    0,    0,  180,  181,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  182,    0,
    0,    0,  184,  185,  186,  187,  188,    0,  189,    0,
    0,    0,    0,    0,  190,    0,  191,  192,    0,    0,
  193,  194,    0,    0,    0,    0,    0,  195,    0,    0,
  196,  197,  198,  199,    0,  200,    0,  201,    0,  202,
  203,  204,  205,  206,    0,    0,    0,    0,  207,    0,
  208,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  209,    0,    0,    0,    0,  210,  211,    0,    0,
    0,    0,    0,  212,    0,    0,    0,  213,    0,  214,
  215,  216,  217,    0,  218,    0,    0,    0,    0,    0,
    0,    0,  219,  220,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  221,  222,    0,  223,  224,  225,    0,
    0,    0,    0,  226,  227,    0,  228,  229,    0,  230,
    0,    0,  231,  232,    0,    0,    0,  233,  234,    0,
    0,    0,    0,  235,  236,  237,    0,  238,  239,  240,
    0,  241,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  242,  243,  244,  245,  246,  247,  248,  249,  250,
  251,  252,  253,  254,  255,  256,  257,  258,  259,  260,
  261,  262,    0,  263,  264,  265,  266,  267,  268,    0,
  269,  270,  271,  272,    0,  273,    0,  274,  275,  276,
  277,  278,  279,  280,  281,  282,  283,  284,  285,  286,
  287,  288,  289,  290,  291,    0,  292,  293,  294,  295,
    0,  296,  297,  298,  299,  300,    0,  301,  302,  303,
  304,  305,  306,  307,  308,  309,  310,  311,  312,  313,
    0,  314,  315,  316,  317,  318,  319,    0,  320,  321,
  322,  323,  324,  325,  326,  327,    0,  328,  329,    0,
    0,    0,  330,  331,  332,  333,  334,  335,    0,    0,
  336,  337,  338,  339,  340,  341,  342,  343,  344,    0,
    0,    0,    0,    0,  345,    0,  346,  347,  348,  349,
  350,  351,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
  352,  353,    0,  354,  355,  356,  357,  358,  359,  360,
  361,  362,    0,  363,  364,  365, 3969,  367,  368,  369,
  370,  371,  372,  373,  374,    0,  375,  376,  377,    0,
  378,    0,    0,    0,  379,  380,  381,  382,  383,  384,
  385,  386,  387,  388,  389,  390,  391,    0,  392,  393,
    0,    0,    0,    0,    0,  394,  395,  396,  397,  398,
  399,  400,  401,  402,  403,  404,  405,  406,  407,    0,
    0,  408,  409,    0,    0,    0,    0,  410,  411,  412,
  413,  414,  415, 1182,    0, 1182,    0,    0,    0,    0,
    0, 1182,    0,    0, 1182, 1182,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1182, 1182,    0, 1182,    0,
 1182,    0,    0, 1182,    0,    0, 1182,    0,    0,    0,
    0,    0,    0, 1182,    0, 1182, 1182,    0,    0,    0,
 1182,    0, 1182,    0,    0, 1182,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1182, 1182,    0,    0,    0,
 1182,    0, 1182,    0, 1182,    0,    0,    0,    0,    0,
 1182,    0, 1182,    0,    0,    0, 1182,    0,    0,    0,
    0,    0,    0,    0, 1182,    0,    0, 1182,    0, 1182,
    0,    0,    0, 1182, 1182,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1182,    0,    0,    0,
 1182, 1182, 1182, 1182, 1182,    0, 1182,    0,    0,    0,
    0,    0, 1182,    0, 1182, 1182,    0,    0, 1182, 1182,
    0,    0,    0,    0,    0, 1182,    0,    0, 1182, 1182,
 1182, 1182,    0, 1182,    0, 1182,    0, 1182, 1182, 1182,
 1182, 1182,    0,    0,    0,    0, 1182,    0, 1182,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1182,
    0,    0,    0,    0, 1182, 1182,    0,    0,    0,    0,
    0, 1182,    0,    0,    0, 1182,    0, 1182, 1182, 1182,
 1182,    0, 1182,    0,    0,    0,    0,    0,    0,    0,
 1182, 1182,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1182, 1182,    0, 1182, 1182, 1182,    0,    0,    0,
    0, 1182, 1182,    0, 1182, 1182,    0, 1182,    0,    0,
 1182, 1182,    0,    0,    0, 1182, 1182,    0,    0,    0,
    0, 1182, 1182, 1182,    0, 1182, 1182, 1182,    0, 1182,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1182,
 1182, 1182, 1182, 1182, 1182, 1182, 1182, 1182, 1182, 1182,
 1182, 1182, 1182, 1182, 1182, 1182, 1182, 1182, 1182, 1182,
    0, 1182, 1182, 1182, 1182, 1182, 1182,    0, 1182, 1182,
 1182, 1182,    0, 1182,    0, 1182, 1182, 1182, 1182, 1182,
 1182, 1182, 1182, 1182, 1182, 1182, 1182, 1182, 1182, 1182,
 1182, 1182, 1182,    0, 1182, 1182, 1182, 1182,    0, 1182,
 1182, 1182, 1182, 1182,    0, 1182, 1182, 1182, 1182, 1182,
 1182, 1182, 1182, 1182, 1182, 1182, 1182, 1182,    0, 1182,
 1182, 1182, 1182, 1182, 1182,    0, 1182, 1182, 1182, 1182,
 1182, 1182, 1182, 1182,    0, 1182, 1182,    0,    0,    0,
 1182, 1182, 1182, 1182, 1182, 1182,    0,    0, 1182, 1182,
 1182, 1182, 1182, 1182, 1182, 1182, 1182,    0,    0,    0,
    0,    0, 1182,    0, 1182, 1182, 1182, 1182, 1182, 1182,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1182, 1182,
    0, 1182, 1182, 1182, 1182, 1182, 1182, 1182, 1182, 1182,
    0, 1182, 1182, 1182, 1182, 1182, 1182, 1182, 1182, 1182,
 1182, 1182, 1182,    0, 1182, 1182, 1182,    0, 1182,    0,
    0,    0, 1182, 1182, 1182, 1182, 1182, 1182, 1182, 1182,
 1182, 1182, 1182, 1182, 1182,    0, 1182, 1182,    0,    0,
    0,    0,    0, 1182, 1182, 1182, 1182, 1182, 1182, 1182,
 1182, 1182, 1182, 1182, 1182, 1182, 1182,    0,    0, 1182,
 1182,    0,    0,    0,    0, 1182, 1182, 1182, 1182, 1182,
 1182, 1310,    0, 1310,    0,    0,    0,    0,    0, 1310,
    0,    0, 1310, 1310,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1310, 1310,    0, 1310,    0, 1310,    0,
    0, 1310,    0,    0, 1310,    0,    0,    0,    0,    0,
    0, 1310,    0, 1310, 1310,    0,    0,    0, 1310,    0,
 1310,    0,    0, 1310,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1310, 1310,    0,    0,    0, 1310,    0,
 1310,    0, 1310,    0,    0,    0,    0,    0, 1310,    0,
 1310,    0,    0,    0, 1310,    0,    0,    0,    0,    0,
    0,    0, 1310,    0,    0, 1310,    0, 1310,    0,    0,
    0, 1310, 1310,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1310,    0,    0,    0, 1310, 1310,
 1310, 1310, 1310,    0, 1310,    0,    0,    0,    0,    0,
 1310,    0, 1310, 1310,    0,    0, 1310, 1310,    0,    0,
    0,    0,    0, 1310,    0,    0, 1310, 1310, 1310, 1310,
    0, 1310,    0, 1310,    0, 1310, 1310, 1310, 1310, 1310,
    0,    0,    0,    0, 1310,    0, 1310,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1310,    0,    0,
    0,    0, 1310, 1310,    0,    0,    0,    0,    0, 1310,
    0,    0,    0, 1310,    0, 1310, 1310, 1310, 1310,    0,
 1310,    0,    0,    0,    0,    0,    0,    0, 1310, 1310,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1310,
 1310,    0, 1310, 1310, 1310,    0,    0,    0,    0, 1310,
 1310,    0, 1310, 1310,    0, 1310,    0,    0, 1310, 1310,
    0,    0,    0, 1310, 1310,    0,    0,    0,    0, 1310,
 1310, 1310,    0, 1310, 1310, 1310,    0, 1310,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1310, 1310, 1310,
 1310, 1310, 1310, 1310, 1310, 1310, 1310, 1310, 1310, 1310,
 1310, 1310, 1310, 1310, 1310, 1310, 1310, 1310,    0, 1310,
 1310, 1310, 1310, 1310, 1310,    0, 1310, 1310, 1310, 1310,
    0, 1310,    0, 1310, 1310, 1310, 1310, 1310, 1310, 1310,
 1310, 1310, 1310, 1310, 1310, 1310, 1310, 1310, 1310, 1310,
 1310,    0, 1310, 1310, 1310, 1310,    0, 1310, 1310, 1310,
 1310, 1310,    0, 1310, 1310, 1310, 1310, 1310, 1310, 1310,
 1310, 1310, 1310, 1310, 1310, 1310,    0, 1310, 1310, 1310,
 1310, 1310, 1310,    0, 1310, 1310, 1310, 1310, 1310, 1310,
 1310, 1310,    0, 1310, 1310,    0,    0,    0, 1310, 1310,
 1310, 1310, 1310, 1310,    0,    0, 1310, 1310, 1310, 1310,
 1310, 1310, 1310, 1310, 1310,    0,    0,    0,    0,    0,
 1310,    0, 1310, 1310, 1310, 1310, 1310, 1310,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1310, 1310,    0, 1310,
 1310, 1310, 1310, 1310, 1310, 1310, 1310, 1310,    0, 1310,
 1310, 1310, 1310, 1310, 1310, 1310, 1310, 1310, 1310, 1310,
 1310,    0, 1310, 1310, 1310,    0, 1310,    0,    0,    0,
 1310, 1310, 1310, 1310, 1310, 1310, 1310, 1310, 1310, 1310,
 1310, 1310, 1310,    0, 1310, 1310,    0,    0,    0,    0,
    0, 1310, 1310, 1310, 1310, 1310, 1310, 1310, 1310, 1310,
 1310, 1310, 1310, 1310, 1310,    0,    0, 1310, 1310,    0,
    0,    0,    0, 1310, 1310, 1310, 1310, 1310, 1310, 1308,
    0, 1308,    0,    0,    0,    0,    0, 1308,    0,    0,
 1308, 1308,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1308, 1308,    0, 1308,    0, 1308,    0,    0, 1308,
    0,    0, 1308,    0,    0,    0,    0,    0,    0, 1308,
    0, 1308, 1308,    0,    0,    0, 1308,    0, 1308,    0,
    0, 1308,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1308, 1308,    0,    0,    0, 1308,    0, 1308,    0,
 1308,    0,    0,    0,    0,    0, 1308,    0, 1308,    0,
    0,    0, 1308,    0,    0,    0,    0,    0,    0,    0,
 1308,    0,    0, 1308,    0, 1308,    0,    0,    0, 1308,
 1308,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1308,    0,    0,    0, 1308, 1308, 1308, 1308,
 1308,    0, 1308,    0,    0,    0,    0,    0, 1308,    0,
 1308, 1308,    0,    0, 1308, 1308,    0,    0,    0,    0,
    0, 1308,    0,    0, 1308, 1308, 1308, 1308,    0, 1308,
    0, 1308,    0, 1308, 1308, 1308, 1308, 1308,    0,    0,
    0,    0, 1308,    0, 1308,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1308,    0,    0,    0,    0,
 1308, 1308,    0,    0,    0,    0,    0, 1308,    0,    0,
    0, 1308,    0, 1308, 1308, 1308, 1308,    0, 1308,    0,
    0,    0,    0,    0,    0,    0, 1308, 1308,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1308, 1308,    0,
 1308, 1308, 1308,    0,    0,    0,    0, 1308, 1308,    0,
 1308, 1308,    0, 1308,    0,    0, 1308, 1308,    0,    0,
    0, 1308, 1308,    0,    0,    0,    0, 1308, 1308, 1308,
    0, 1308, 1308, 1308,    0, 1308,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1308, 1308, 1308, 1308, 1308,
 1308, 1308, 1308, 1308, 1308, 1308, 1308, 1308, 1308, 1308,
 1308, 1308, 1308, 1308, 1308, 1308,    0, 1308, 1308, 1308,
 1308, 1308, 1308,    0, 1308, 1308, 1308, 1308,    0, 1308,
    0, 1308, 1308, 1308, 1308, 1308, 1308, 1308, 1308, 1308,
 1308, 1308, 1308, 1308, 1308, 1308, 1308, 1308, 1308,    0,
 1308, 1308, 1308, 1308,    0, 1308, 1308, 1308, 1308, 1308,
    0, 1308, 1308, 1308, 1308, 1308, 1308, 1308, 1308, 1308,
 1308, 1308, 1308, 1308,    0, 1308, 1308, 1308, 1308, 1308,
 1308,    0, 1308, 1308, 1308, 1308, 1308, 1308, 1308, 1308,
    0, 1308, 1308,    0,    0,    0, 1308, 1308, 1308, 1308,
 1308, 1308,    0,    0, 1308, 1308, 1308, 1308, 1308, 1308,
 1308, 1308, 1308,    0,    0,    0,    0,    0, 1308,    0,
 1308, 1308, 1308, 1308, 1308, 1308,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1308, 1308,    0, 1308, 1308, 1308,
 1308, 1308, 1308, 1308, 1308, 1308,    0, 1308, 1308, 1308,
 1308, 1308, 1308, 1308, 1308, 1308, 1308, 1308, 1308,    0,
 1308, 1308, 1308,    0, 1308,    0,    0,    0, 1308, 1308,
 1308, 1308, 1308, 1308, 1308, 1308, 1308, 1308, 1308, 1308,
 1308,    0, 1308, 1308,    0,    0,    0,    0,    0, 1308,
 1308, 1308, 1308, 1308, 1308, 1308, 1308, 1308, 1308, 1308,
 1308, 1308, 1308,    0,    0, 1308, 1308,    0,    0,    0,
    0, 1308, 1308, 1308, 1308, 1308, 1308,  169,    0,  169,
    0,    0,    0,    0,    0,  169,    0,    0,  169,  169,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  169,
  169,    0,  169,    0,  169,    0,    0,  169,    0,    0,
  169,    0,    0,    0,    0,    0,    0,  169,    0,  169,
  169,    0,    0,    0,  169,    0,  169,    0,    0,  169,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  169,
  169,    0,    0,    0,  169,    0,  169,    0,  169,    0,
    0,    0,    0,    0,  169,    0,  169,    0,    0,    0,
  169,    0,    0,    0,    0,    0,    0,    0,  169,    0,
    0,  169,    0,  169,    0,    0,    0,  169,  169,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
  169,    0,    0,    0,  169,  169,  169,  169,  169,    0,
  169,    0,    0,    0,    0,    0,  169,    0,  169,  169,
    0,    0,  169,  169,    0,    0,    0,    0,    0,  169,
    0,    0,  169,  169,  169,  169,    0,  169,    0,  169,
    0,  169,  169,  169,  169,  169,    0,    0,    0,    0,
  169,    0,  169,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  169,    0,    0,    0,    0,  169,  169,
    0,    0,    0,    0,    0,  169,    0,    0,    0,  169,
    0,  169,  169,  169,  169,    0,  169,    0,    0,    0,
    0,    0,    0,    0,  169,  169,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  169,  169,    0,  169,  169,
  169,    0,    0,    0,    0,  169,  169,    0,  169,  169,
    0,  169,    0,    0,  169,  169,    0,    0,    0,  169,
  169,    0,    0,    0,    0,  169,  169,  169,    0,  169,
  169,  169,    0,  169,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  169,  169,  169,  169,  169,  169,  169,
  169,  169,  169,  169,  169,  169,  169,  169,  169,  169,
  169,  169,  169,  169,    0,  169,  169,  169,  169,  169,
  169,    0,  169,  169,  169,  169,    0,  169,    0,  169,
  169,  169,  169,  169,  169,  169,  169,  169,  169,  169,
  169,  169,  169,  169,  169,  169,  169,    0,  169,  169,
  169,  169,    0,  169,  169,  169,  169,  169,    0,  169,
  169,  169,  169,  169,  169,  169,  169,  169,  169,  169,
  169,  169,    0,  169,  169,  169,  169,  169,  169,    0,
  169,  169,  169,  169,  169,  169,  169,  169,    0,  169,
  169,    0,    0,    0,  169,  169,  169,  169,  169,  169,
    0,    0,  169,  169,  169,  169,  169,  169,  169,  169,
  169,    0,    0,    0,    0,    0,  169,    0,  169,  169,
  169,  169,  169,  169,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,  169,  169,    0,  169,  169,  169,  169,  169,
  169,  169,  169,  169,    0,  169,  169,  169,  169,  169,
  169,  169,  169,  169,  169,  169,  169,    0,  169,  169,
  169,    0,  169,    0,    0,    0,  169,  169,  169,  169,
  169,  169,  169,  169,  169,  169,  169,  169,  169,    0,
  169,  169,    0,    0,    0,    0,    0,  169,  169,  169,
  169,  169,  169,  169,  169,  169,  169,  169,  169,  169,
  169,    0,    0,  169,  169,    0,    0,    0,    0,  169,
  169,  169,  169,  169,  169,   67,    0,   67,    0,    0,
    0,    0,    0,   67,    0,    0,   67,   67,    0,    0,
    0,    0,    0,    0,    0,    0,    0,   67,   67,    0,
   67,    0,   67,    0,    0,   67,    0,    0,   67,    0,
    0,    0,    0,    0,    0,   67,    0,   67,   67,    0,
    0,    0,   67,    0,   67,    0,    0,   67,    0,    0,
    0,    0,    0,    0,    0,    0,    0,   67,   67,    0,
    0,    0,   67,    0,   67,    0,   67,    0,    0,    0,
    0,    0,   67,    0,   67,    0,    0,    0,   67,    0,
    0,    0,    0,    0,    0,    0,   67,    0,    0,   67,
    0,   67,    0,    0,    0,   67,   67,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,   67,    0,
    0,    0,   67,   67,   67,   67,   67,    0,   67,    0,
    0,    0,    0,    0,   67,    0,   67,   67,    0,    0,
   67,   67,    0,    0,    0,    0,    0,   67,    0,    0,
   67,   67,   67,   67,    0,   67,    0,   67,    0,   67,
   67,   67,   67,   67,    0,    0,    0,    0,   67,    0,
   67,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,   67,    0,    0,    0,    0,   67,   67,    0,    0,
    0,    0,    0,   67,    0,    0,    0,   67,    0,   67,
   67,   67,   67,    0,   67,    0,    0,    0,    0,    0,
    0,    0,   67,   67,    0,    0,    0,    0,    0,    0,
    0,    0,    0,   67,   67,    0,   67,   67,   67,    0,
    0,    0,    0,   67,   67,    0,   67,   67,    0,   67,
    0,    0,   67,   67,    0,    0,    0,   67,   67,    0,
    0,    0,    0,   67,   67,   67,    0,   67,   67,   67,
    0,   67,    0,    0,    0,    0,    0,    0,    0,    0,
    0,   67,   67,   67,   67,   67,   67,   67,   67,   67,
   67,   67,   67,   67,   67,   67,   67,   67,   67,   67,
   67,   67,    0,   67,   67,   67,   67,   67,   67,    0,
   67,   67,   67,   67,    0,   67,    0,   67,   67,   67,
   67,   67,   67,   67,   67,   67,   67,   67,   67,   67,
   67,   67,   67,   67,   67,    0,   67,   67,   67,   67,
    0,   67,   67,   67,   67,   67,    0,   67,   67,   67,
   67,   67,   67,   67,   67,   67,   67,   67,   67,   67,
    0,   67,   67,   67,   67,   67,   67,    0,   67,   67,
   67,   67,   67,   67,   67,   67,    0,   67,   67,    0,
    0,    0,   67,   67,   67,   67,   67,   67,    0,    0,
   67,   67,   67,   67,   67,   67,   67,   67,   67,    0,
    0,    0,    0,    0,   67,    0,   67,   67,   67,   67,
   67,   67,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
   67,   67,    0,   67,   67,   67,   67,   67,   67,   67,
   67,   67,    0,   67,   67,   67,   67,   67,   67,   67,
   67,   67,   67,   67,   67,    0,   67,   67,   67,    0,
   67,    0,    0,    0,   67,   67,   67,   67,   67,   67,
   67,   67,   67,   67,   67,   67,   67,    0,   67,   67,
    0,    0,    0,    0,    0,   67,   67,   67,   67,   67,
   67,   67,   67,   67,   67,   67,   67,   67,   67,    0,
    0,   67,   67,    0,    0,    0,    0,   67,   67,   67,
   67,   67,   67,  596,    0,  596,    0,    0,    0,    0,
    0,  596,    0,    0,  596,  596,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  596,  596,    0,  596,    0,
  596,    0,    0,  596,    0,    0,  596,    0,    0,    0,
    0,    0,    0,  596,    0,  596,  596,    0,    0,    0,
  596,    0,  596,    0,    0,  596,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  596,  596,    0,    0,    0,
  596,    0,  596,    0,  596,    0,    0,    0,    0,    0,
  596,    0,  596,    0,    0,    0,  596,    0,    0,    0,
    0,    0,    0,    0,  596,    0,    0,  596,    0,  596,
    0,    0,    0,  596,  596,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  596,    0,    0,    0,
  596,  596,  596,  596,  596,    0,  596,    0,    0,    0,
    0,    0,  596,    0,  596,  596,    0,    0,  596,  596,
    0,    0,    0,    0,    0,  596,    0,    0,  596,  596,
  596,  596,    0,  596,    0,  596,    0,  596,  596,  596,
  596,  596,    0,    0,    0,    0,  596,    0,  596,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  596,
    0,    0,    0,    0,  596,  596,    0,    0,    0,    0,
    0,  596,    0,    0,    0,  596,    0,  596,  596,  596,
  596,    0,  596,    0,    0,    0,    0,    0,    0,    0,
  596,  596,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  596,  596,    0,  596,  596,  596,    0,    0,    0,
    0,  596,  596,    0,  596,  596,    0,  596,    0,    0,
  596,  596,    0,    0,    0,  596,  596,    0,    0,    0,
    0,  596,  596,  596,    0,  596,  596,  596,    0,  596,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  596,
  596,  596,  596,  596,  596,  596,  596,  596,  596,  596,
  596,  596,  596,  596,  596,  596,  596,  596,  596,  596,
    0,  596,  596,  596,  596,  596,  596,    0,  596,  596,
  596,  596,    0,  596,    0,  596,  596,  596,  596,  596,
  596,  596,  596,  596,  596,  596,  596,  596,  596,  596,
  596,  596,  596,    0,  596,  596,  596,  596,    0,  596,
  596,  596,  596,  596,    0,  596,  596,  596,  596,  596,
  596,  596,  596,  596,  596,  596,  596,  596,    0,  596,
  596,  596,  596,  596,  596,    0,  596,  596,  596,  596,
  596,  596,  596,  596,    0,  596,  596,    0,    0,    0,
  596,  596,  596,  596,  596,  596,    0,    0,  596,  596,
  596,  596,  596,  596,  596,  596,  596,    0,    0,    0,
    0,    0,  596,    0,  596,  596,  596,  596,  596,  596,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  596,  596,
    0,  596,  596,  596,  596,  596,  596,  596,  596,  596,
    0,  596,  596,  596,  596,  596,  596,  596,  596,  596,
  596,  596,  596,    0,  596,  596,  596,    0,  596,    0,
    0,    0,  596,  596,  596,  596,  596,  596,  596,  596,
  596,  596,  596,  596,  596,    0,  596,  596,    0,    0,
    0,    0,    0,  596,  596,  596,  596,  596,  596,  596,
  596,  596,  596,  596,  596,  596,  596,    0,    0,  596,
  596,    0,    0,    0,    0,  596,  596,  596,  596,  596,
  596,  152,    0,  153,    0,    0,    0,    0,    0,  154,
    0,    0,  155,  156,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  157,  158,    0,  159,    0,  160,    0,
    0,  161,    0,    0,  162,    0,    0,    0,    0,    0,
    0,  163,    0,  164,  165,    0,    0,    0,  166,    0,
  167,    0,    0,  168,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  169,  170,    0,    0,    0,  171,    0,
  172,    0,  173,    0,    0,    0,    0,    0,  174,    0,
  175,    0,    0,    0,  176,    0,    0,    0,    0,    0,
    0,    0,  177,    0,    0,  178,    0,  179,    0,    0,
    0,  180,  181,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  182,    0,    0,    0,  184,  185,
  186,  187,  188,    0,  189,    0,    0,    0,    0,    0,
  190,    0,  191,  192,    0,    0,  193,  194,    0,    0,
    0,    0,    0,  195,    0,    0,  196,  197,  198,  199,
    0,  200,    0,  201,    0,  202,  203,  204,  205,  206,
    0,    0,    0,    0,  207,    0,  208,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  209,    0,    0,
    0,    0,  210,  211,    0,    0,    0,    0,    0,  212,
    0,    0,    0,  213,    0,  214,  215,  216,  217,    0,
  218,    0,    0,    0,    0,    0,    0,    0,  219,  220,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  221,
  222,    0,  223,  224,  225,    0,    0,    0,    0,  226,
  227,    0,  228,  229,    0,  230,    0,    0,  231,  232,
    0,    0,    0,  233,  234,    0,    0,    0,    0,  235,
  236,  237,    0,  238,  239,  240,    0,  241,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  242,  243,  244,
  245,  246,  247,  248,  249,  250,  251,  252,  253,  254,
  255,  256,  257,  258,  259,  260,  261,  262,    0,  263,
  264,  265,  266,  267,  268,    0,  269,  270,  271,  272,
    0,  273,    0,  274,  275,  276,  277,  278,  279,  280,
  281,  282,  283,  284,  285,  286,  287,  288,  289,  290,
  291,    0,  292,  293,  294,  295,    0,  296,  297,  298,
  299,  300,    0,  301,  302,  303,  304,  305,  306,  307,
  308,  309,  310,  311,  312,  313,    0,  314,  315,  316,
  317,  318,  319,    0,  320,  321,  322,  323,  324,  325,
  326,  327,    0,  328,  329,    0,    0,    0,  330,  331,
  332,  333,  334,  335,    0,    0,  336,  337,  338,  339,
  340,  341,  342,  343,  344,    0,    0,    0,    0,    0,
  345,    0,  346,  347,  348,  349,  350,  351,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  352,  353,    0,  354,
  355,  356,  357,  358,  359,  360,  361,  362,    0,  363,
  364,  365,  366,  367,  368,  369,  370,  371,  372,  373,
  374,    0,  375,  376,  377,    0,  378,    0,    0,    0,
  379,  380,  381,  382,  383,  384,  385,  386,  387,  388,
  389,  390,  391,    0,  392,  393,    0,    0,    0,    0,
    0,  394,  395,  396,  397,  398,  399,  400,  401,  402,
  403,  404,    0,  406,  407,    0,    0,  408,  409,    0,
    0,    0,    0,  410,  411,  412,  413,  414,  415,  152,
    0,  153,    0,    0,    0,    0,    0,  154,    0,    0,
  155,  156,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  157,  158,    0,  159,    0,  160,    0,    0,  161,
    0,    0,  162,    0,    0,    0,    0,    0,    0,  163,
    0,  164,  165,    0,    0,    0,  166,    0,  167,    0,
    0,  168,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  169,  170,    0,    0,    0,  171,    0,  172,    0,
  173,    0,    0,    0,    0,    0,  174,    0,  175,    0,
    0,    0,  176,    0,    0,    0,    0,    0,    0,    0,
  177,    0,    0,  178,    0,  179,    0,    0,    0,  180,
  181,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,  182,    0,    0,    0,  184,  185,  186,  187,
  188,    0,  189,    0,    0,    0,    0,    0,  190,    0,
  191,  192,    0,    0,  193,  194,    0,    0,    0,    0,
    0,  195,    0,    0,  196,  197,  198,  199,    0,  200,
    0,  201,    0,  202,  203,  204,  205,  206,    0,    0,
    0,    0,  207,    0,  208,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  209,    0,    0,    0,    0,
  210,  211,    0,    0,    0,    0,    0,  212,    0,    0,
    0,  213,    0,  214,  215,  216,  217,    0,  218,    0,
    0,    0,    0,    0,    0,    0,  219,  220,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  221,  222,    0,
  223,  224,  225,    0,    0,    0,    0,  226,  227,    0,
  228,  229,    0,  230,    0,    0,  231,  232,    0,    0,
    0,  233,  234,    0,    0,    0,    0,  235,  236,  237,
    0,  238,  239,  240,    0,  241,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  243,  244,  245,  246,
  247,  248,  249,  250,  251,  252,  253,  254,  255,  256,
  257,  258,  259,  260,  261,  262,    0,  263,  264,  265,
  266,  267,  268,    0,  269,  270,  271,  272,    0,  273,
    0,  274,  275,  276,  277,  278,  279,  280,  281,  282,
  283,  284,  285,  286,  287,  288,  289,  290,  291,    0,
  292,  293,  294,  295,    0,  296,  297,  298,  299,  300,
    0,  301,  302,  303,  304,  305,  306,  307,  308,  309,
  310,  311,  312,  313,    0,  314,  315,  316,  317,  318,
  319,    0,  320,  321,  322,  323,  324,  325,  326,  327,
    0,  328,  329,    0,    0,    0,  330,  331,  332,  333,
  334,  335,    0,    0,  336,  337,  338,  339,  340,  341,
  342,  343,  344,    0,    0,    0,    0,    0,  345,    0,
  346,  347,  348,  349,  350,  351,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  352,  353,    0,  354,  355,  356,
  357,  358,  359,  360,  361,  362,    0,  363,  364,  365,
  366,  367,  368,  369,  370,  371,  372,  373,  374,    0,
  375,  376,  377,    0,  378,    0,    0,    0,  379,  380,
  381,  382,  383,  384,  385,  386,  387,  388,  389,  390,
  391,    0,  392,  393,    0,    0,    0,    0,    0,  394,
  395,  396,  397,  398,  399,  400,  401,  402,  403,  404,
  405,  406,  407,    0,    0,  408,  409,    0,    0,    0,
    0,  410,  411,  412,  413,  414,  415,  152,    0,  153,
    0,    0,    0,    0,    0,  154,    0,    0,  155,  156,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  157,
  158,    0,  159,    0,  160,    0,    0,  161,    0,    0,
  162,    0,    0,    0,    0,    0,    0,  163,    0,  164,
  165,    0,    0,    0,  166,    0,  167,    0,    0,  168,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  169,
  170,    0,    0,    0,  171,    0,  172,    0,  173,    0,
    0,    0,    0,    0,  174,    0,  175,    0,    0,    0,
  176,    0,    0,    0,    0,    0,    0,    0,  177,    0,
    0,  178,    0,  179,    0,    0,    0,  180,  181,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
  182,    0,    0,    0,  184,  185,  186,  187,  188,    0,
  189,    0,    0,    0,    0,    0,  190,    0,  191,  192,
    0,    0,  193,  194,    0,    0,    0,    0,    0,  195,
    0,    0,  196,  197,  198,  199,    0,  200,    0,  201,
    0,  202,  203,  204,  205,  206,    0,    0,    0,    0,
  207,    0,  208,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  209,    0,    0,    0,    0,  210,  211,
    0,    0,    0,    0,    0,  212,    0,    0,    0,  213,
    0,  214,  215,  216,  217,    0,  218,    0,    0,    0,
    0,    0,    0,    0,  219,  220,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  221,  222,    0,  223,  224,
  225,    0,    0,    0,    0,  226,  227,    0,  228,  229,
    0,  230,    0,    0,  231,  232,    0,    0,    0,  233,
  234,    0,    0,    0,    0,  235,  236,  237,    0,  238,
  239,  240,    0,  241,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  242,  243,  244,  245,  246,  247,  248,
  249,  250,  251,  252,  253,  254,  255,  256,  257,  258,
  259,  260,  261,  262,    0,  263,  264,  265,  266,  267,
  268,    0,  269,  270,  271,  272,    0,  273,    0,  274,
  275,  276,  277,  278,  279,  280,  281,  282,  283,  284,
  285,  286,  287,  288,  289,  290,  291,    0,  292,  293,
  294,  295,    0,  296,  297,  298,  299,  300,    0,  301,
  302,  303,  304,  305,  306,  307,  308,    0,  310,  311,
  312,  313,    0,  314,  315,  316,  317,  318,  319,    0,
  320,  321,  322,  323,  324,  325,  326,  327,    0,  328,
  329,    0,    0,    0,  330,  331,  332,  333,  334,  335,
    0,    0,  336,  337,  338,  339,  340,  341,  342,  343,
  344,    0,    0,    0,    0,    0,  345,    0,  346,  347,
  348,  349,  350,  351,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,  352,  353,    0,  354,  355,  356,  357,  358,
  359,  360,  361,  362,    0,  363,  364,  365,  366,  367,
  368,  369,  370,  371,  372,  373,  374,    0,  375,  376,
  377,    0,  378,    0,    0,    0,  379,  380,  381,  382,
  383,  384,  385,  386,  387,  388,  389,  390,  391,    0,
  392,  393,    0,    0,    0,    0,    0,  394,  395,  396,
  397,  398,  399,  400,  401,  402,  403,  404,  405,  406,
  407,    0,    0,  408,  409,    0,    0,    0,    0,  410,
  411,  412,  413,  414,  415,  152,    0,  153,    0,    0,
    0,    0,    0,  154,    0,    0,  155,  156,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  157,  158,    0,
  159,    0,    0,    0,    0,  161,    0,    0,  162,    0,
    0,    0,    0,    0,    0,  163,    0,  164,  165,    0,
    0,    0,  166,    0,  167,    0,    0,  168,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  169,  170,    0,
    0,    0,  171,    0,  172,    0,  173,    0,    0,    0,
    0,    0,  174,    0,  175,    0,    0,    0,  176,    0,
    0,    0,    0,    0,    0,    0,  177,    0,    0,  178,
    0,  179,    0,    0,    0,  180,  181,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  182,    0,
    0,    0,  184,  185,  186,  187,  188,    0,  189,    0,
    0,    0,    0,    0,  190,    0,  191,  192,    0,    0,
  193,  194,    0,    0,    0,    0,    0,  195,    0,    0,
  196,  197,  198,  199,    0,  200,    0,  201,    0,  202,
    0,  204,  205,  206,    0,    0,    0,    0,  207,    0,
  208,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  209,    0,    0,    0,    0,  210,  211,    0,    0,
    0,    0,    0,  212,    0,    0,    0,  213,    0,  214,
  215,  216,  217,    0,  218,    0,    0,    0,    0,    0,
    0,    0,  219,  220,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  221,  222,    0,  223,  224,  225,    0,
    0,    0,    0,  226,  227,    0,  228,  229,    0,  230,
    0,    0,  231,  232,    0,    0,    0,  233,  234,    0,
    0,    0,    0,  235,  236,  237,    0,  238,  239,  240,
    0,  241,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  242,  243,  244,  245,  246,  247,  248,  249,  250,
  251,  252,  253,  254,  255,  256,  257,  258,  259,  260,
  261,  262,    0,  263,  264,  265,  266,  267,  268,    0,
  269,  270,  271,  272,    0,  273,    0,  274,  275,  276,
  277,  278,  279,  280,  281,  282,  283,  284,  285,  286,
  287,  288,  289,  290,  291,    0,  292,  293,  294,  295,
    0,  296,  297,  298,  299,  300,    0,  301,  302,  303,
  304,  305,  306,  307,  308,  309,  310,  311,  312,  313,
    0,  314,  315,  316,  317,  318,  319,    0,  320,  321,
  322,  323,  324,  325,  326,  327,    0,  328,  329,    0,
    0,    0,  330,  331,  332,  333,  334,  335,    0,    0,
  336,  337,  338,  339,  340,  341,  342,  343,  344,    0,
    0,    0,    0,    0,  345,    0,  346,  347,  348,  349,
  350,  351,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
  352,  353,    0,  354,  355,  356,  357,  358,  359,  360,
  361,  362,    0,  363,  364,  365,  366,  367,  368,  369,
  370,  371,  372,  373,  374,    0,  375,  376,  377,    0,
  378,    0,    0,    0,  379,  380,  381,  382,  383,  384,
  385,  386,  387,  388,  389,  390,  391,    0,  392,  393,
    0,    0,    0,    0,    0,  394,  395,  396,  397,  398,
  399,  400,  401,  402,  403,  404,  405,  406,  407,    0,
    0,  408,  409,    0,    0,    0,    0,  410,  411,  412,
  413,  414,  415,  152,    0,  153,    0,    0,    0,    0,
    0,  154,    0,    0,  155,  156,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  157,  158,    0,  159,    0,
  160,    0,    0,  161,    0,    0,  162,    0,    0,    0,
    0,    0,    0,  163,    0,  164,  165,    0,    0,    0,
  166,    0,  167,    0,    0,  168,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  169,  170,    0,    0,    0,
  171,    0,  172,    0,  173,    0,    0,    0,    0,    0,
  174,    0,  175,    0,    0,    0,  176,    0,    0,    0,
    0,    0,    0,    0,  177,    0,    0,  178,    0,  179,
    0,    0,    0,  180,  181,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  182,    0,    0,    0,
  184,  185,  186,  187,  188,    0,  189,    0,    0,    0,
    0,    0,  190,    0,  191,  192,    0,    0,  193,  194,
    0,    0,    0,    0,    0,  195,    0,    0,  196,  197,
  198,  199,    0,  200,    0,  201,    0,  202,  203,  204,
  205,  206,    0,    0,    0,    0,  207,    0,  208,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  209,
    0,    0,    0,    0,  210,  211,    0,    0,    0,    0,
    0,  212,    0,    0,    0,  213,    0,  214,  215,  216,
  217,    0,  218,    0,    0,    0,    0,    0,    0,    0,
  219,  220,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  221,  222,    0,  223,  224,  225,    0,    0,    0,
    0,  226,  227,    0,  228,  229,    0,  230,    0,    0,
  231,  232,    0,    0,    0,  233,  234,    0,    0,    0,
    0,  235,  236,  237,    0,  238,  239,  240,    0,  241,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  242,
  243,  244,  245,  246,  247,  248,  249,  250,  251,  252,
  253,  254,  255,  256,  257,  258,  259,  260,  261,  262,
    0,  263,  264,  265,  266,  267,  268,    0,  269,  270,
  271,  272,    0,  273,    0,  274,  275,  276,  277,  278,
  279,  280,  281,  282,  283,  284,  285,  286,  287,  288,
  289,  290,  291,    0,  292,  293,  294,  295,    0,  296,
  297,  298,  299,  300,    0,  301,  302,  303,  304,  305,
  306,  307,  308,  309,  310,  311,  312,  313,    0,  314,
  315,  316,  317,  318,  319,    0,  320,  321,  322,  323,
  324,  325,  326,  327,    0,  328,    0,    0,    0,    0,
  330,  331,  332,  333,  334,  335,    0,    0,  336,  337,
  338,  339,  340,  341,  342,  343,  344,    0,    0,    0,
    0,    0,  345,    0,  346,  347,  348,  349,  350,  351,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  352,  353,
    0,  354,  355,  356,  357,  358,  359,  360,  361,  362,
    0,  363,  364,  365,  366,    0,  368,  369,  370,  371,
  372,  373,  374,    0,  375,  376,  377,    0,  378,    0,
    0,    0,  379,  380,  381,  382,  383,  384,  385,  386,
  387,  388,  389,  390,  391,    0,  392,    0,    0,    0,
    0,    0,    0,  394,  395,  396,  397,  398,  399,  400,
  401,  402,  403,  404,  405,  406,  407,    0,    0,  408,
  409,    0,    0,    0,    0,  410,  411,  412,  413,  414,
  415,  152,    0,  153,    0,    0,    0,    0,    0,  154,
    0,    0,  155,  156,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  157,  158,    0,  159,    0,  160,    0,
    0,  161,    0,    0,  162,    0,    0,    0,    0,    0,
    0,  163,    0,  164,  165,    0,    0,    0,  166,    0,
  167,    0,    0,  168,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  169,  170,    0,    0,    0,  171,    0,
  172,    0,  173,    0,    0,    0,    0,    0,  174,    0,
  175,    0,    0,    0,  176,    0,    0,    0,    0,    0,
    0,    0,  177,    0,    0,  178,    0,  179,    0,    0,
    0,  180,  181,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  182,    0,    0,    0,  184,  185,
  186,  187,  188,    0,  189,    0,    0,    0,    0,    0,
  190,    0,  191,  192,    0,    0,  193,  194,    0,    0,
    0,    0,    0,  195,    0,    0,  196,  197,  198,  199,
    0,  200,    0,  201,    0,  202,  203,  204,  205,  206,
    0,    0,    0,    0,  207,    0,  208,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  209,    0,    0,
    0,    0,  210,  211,    0,    0,    0,    0,    0,  212,
    0,    0,    0,  213,    0,  214,  215,  216,  217,    0,
  218,    0,    0,    0,    0,    0,    0,    0,  219,  220,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  221,
  222,    0,  223,    0,  225,    0,    0,    0,    0,  226,
  227,    0,  228,    0,    0,  230,    0,    0,  231,  232,
    0,    0,    0,  233,  234,    0,    0,    0,    0,  235,
  236,  237,    0,    0,  239,  240,    0,  241,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  242,  243,  244,
  245,  246,  247,  248,  249,  250,  251,  252,  253,  254,
  255,  256,  257,  258,  259,  260,  261,  262,    0,  263,
  264,  265,  266,  267,  268,    0,  269,  270,  271,  272,
    0,  273,    0,  274,  275,  276,  277,  278,  279,  280,
  281,  282,  283,  284,  285,  286,  287,  288,  289,  290,
  291,    0,  292,  293,  294,  295,    0,  296,  297,  298,
  299,  300,    0,  301,  302,  303,  304,  305,  306,  307,
  308,  309,  310,  311,  312,  313,    0,  314,  315,  316,
  317,  318,  319,    0,  320,  321,  322,  323,  324,  325,
  326,  327,    0,  328,  329,    0,    0,    0,    0,    0,
    0,  333,  334,  335,    0,    0,  336,  337,  338,  339,
  340,  341,  342,  343,  344,    0,    0,    0,    0,    0,
  345,    0,  346,  347,  348,  349,  350,  351,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  352,  353,    0,  354,
  355,  356,  357,  358,  359,  360,  361,  362,    0,  363,
  364,  365,  366,  367,  368,  369,  370,  371,  372,  373,
  374,    0,  375,  376,  377,    0,  378,    0,    0,    0,
  379,  380,  381,  382,  383,  384,  385,  386,  387,  388,
  389,  390,  391,    0,  392,  393,    0,    0,    0,    0,
    0,  394,  395,  396,  397,  398,  399,  400,  401,  402,
  403,  404,  405,  406,  407,    0, 2897,  408,  409, 1117,
    0,    0,    0,  410,  411,  412,  413,  414,  415,    0,
    0, 2898,    0,    0,    0, 1118,    0, 1119,    0,    0,
    0,  167,    0,    0,  167,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  167,    0,    0,    0,
  167,    0,  167,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1120,    0,    0,    0,    0,
    0, 1121, 1122, 1123,    0, 1124, 1125, 1126,    0, 1127,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
  167,    0,    0, 1128,    0, 1129,  167,  167,  167,    0,
  167,  167,  167,    0,  167,    0,    0,    0,    0,    0,
    0,    0,    0, 3439,    0,    0,    0,    0,  167,    0,
  167,    0,    0,  179, 1130,    0,  179,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  179,    0,
 1131,    0,  179,    0,  179,    0,    0,    0,    0,  167,
    0,    0,    0,    0,  166,    0,    0,    0,    0,    0,
    0,    0, 1132,    0,    0,  167,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,  179,    0,    0,    0,    0,  167,  179,  179,
  179,    0,  179,  179,  179,    0,  179,    0,    0,    0,
 1252, 1134,    0, 2897,    0,    0, 1117,    0,    0,    0,
  179,    0,  179,    0,    0,    0,    0, 1135, 2898,    0,
    0,    0, 1118,    0, 1119,  167,  167,    0,    0,    0,
  179,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1116,  179,  167, 1117,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  179,    0, 1118,
    0, 1119, 1120,    0,    0,    0,    0,    0, 1121, 1122,
 1123,    0, 1124, 1125, 1126,    0, 1127,    0,    0,  179,
    0, 1116,    0,    0, 1117,    0,    0,    0,    0,    0,
 1128,    0, 1129,    0,    0,    0,    0,    0,    0, 1120,
 1118,    0, 1119,    0,    0, 1121, 1122, 1123,    0, 1124,
 1125, 1126,    0, 1127,    0,    0,    0,  179,  179,    0,
    0, 1130,    0,    0,    0,    0,    0, 1128,    0, 1129,
    0,    0,    0,    0,  179,    0,    0, 1131,    0,    0,
 1120,    0,    0,    0,    0,    0, 1121, 1122, 1123,    0,
 1124, 1125, 1126,    0, 1127,    0,    0,    0, 1130, 1132,
    0,    0,    0,    0,    0,    0,    0,    0, 1128,    0,
 1129,    0,    0,    0, 1131,    0, 1136,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1132, 1252, 1134, 1130,
  276,  167,    0,  276,    0,    0,    0,    0, 1137,    0,
    0,    0,    0,    0, 1135, 1131,    0,    0, 1138,  276,
    0,  276,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1139,    0,  167, 1133, 1134,    0, 1132,    0,    0,
    0,    0,    0,  167,    0,    0,    0,    0,    0,    0,
    0, 1135,    0,    0,    0,    0,  167,    0,    0,  276,
    0,    0,    0,    0,    0,  276,  276,  276,    0,  276,
  276,  276,    0,  276,    0, 1252, 1134,    0,    0,    0,
 1140,    0,    0,    0,    0,    0,    0,  276,    0,  276,
    0,    0, 1135,  179,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  167,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  276,    0,
    0,    0,    0,    0,    0,  179,    0,    0,    0,    0,
    0,    0,    0,    0,  276,  179,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  179,    0,
    0,    0,    0,    0,    0,    0,  276,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1136,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  276,  276,    0,  179,    0,    0,
    0,    0,    0,    0,    0, 1137,    0,    0,    0,    0,
 1136,  276,    0,    0,    0, 1138,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1139,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1137,    0,    0,    0,    0,    0,    0,    0,
    0, 1136, 1138,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1139,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1140,    0,    0,
    0,    0,    0, 1137,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1138,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1139,    0,    0,    0,
    0,    0,    0,    0, 1140,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1140,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
  276,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,  276,    0,    0,    0,    0,    0,    0,    0,
    0,    0,  276,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  276,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  276,
};
static int yycheck[] = {                                      14,
   15,   75,   17,   71,   79,  123,    0,  571,  119,  837,
  130,  824,  682,  122, 1019,  834, 1019,  685,  740,  741,
   15,   14,   81, 1337,   78,   90,  135, 1498, 1246, 1343,
 1437, 1551,  596, 1187, 1998,   90,  782,  823,  687,  798,
 2010,  798, 1235,  759,  103,  771, 1634,  121,  116,  571,
 1206,  126, 1017, 1018, 1019,   70,   71, 1634,   73,   74,
   75,   76,   77,   78,   79,   80,   81,   82, 1019,  128,
   85,  125,  740,  741,  800, 2027,   91,   92,   93,   94,
  796,   96, 1909,   98,   99, 1405,  101,  102,  103,  104,
  105, 1206, 2010,  108, 1634, 2309, 2247,  112, 2249,  114,
  115,  116,  624,  118,  119,  120,  121,  122,   74,  124,
  125,  126,  127,  128,  129,  130,   82, 1659,  133, 1990,
  135,  114, 2130,  138,  139, 1634,   14,  142, 1938, 1206,
  145,  124, 1634, 1792, 1017, 1018, 2383, 2384, 2385, 1591,
 1592, 1601, 2703, 1233, 1310, 2494,  139,  775, 2782,  798,
 1460, 1547, 1612, 2343,  120,   85, 1583, 1584, 1572,   91,
 2483, 2484, 2635,  129, 2393, 2431, 1637, 1638, 2636, 1533,
 2436, 2812, 2544,  687, 1538, 1661,  142,  828,  108, 1524,
 2633, 2000, 1288, 2074, 2431, 2073, 1206,  733,  734,   77,
 2089, 1801,  820, 1653,   93, 3143,   96, 1287,    0,   59,
    0, 2109,   41,  133,  107,  135,   41,   59, 1622, 1212,
   98,  581,   42,    0, 1608,  440, 2850,  636,  264,    0,
   41,    0,    0,  102,    0, 1570,  114,  108,   41,  132,
    0,   59,   59,    0,   59,  275,    0, 2091,   41, 3536,
 2893, 1244,   59,    0,   59,  777,   41,  666,   41,    0,
  138,  139, 2646, 1443,    0,  301,   40,   59,  260,   59,
  310, 2084,   45, 2264,  310,   41,    0,   41, 2253,   44,
 2091, 2665,   59,    0,    0,   44, 2090,    0,   59,    0,
   59,   59,  431,   59, 1973,  288,    0,   41, 2090,   59,
   40,   44,   59, 1523,    0,   59,   44,  368, 1413,  405,
    0, 1306,   59, 1306,   41, 2773,    0,  325,   59, 1386,
 2490,  316, 2765,   59,    0, 1974, 1525, 1320,   41,  322,
   41,   44,   44, 2776, 1327,   59, 2966,    0, 1405,    0,
   44,    0,   59,   59,    0,   40,   59,  301,   59,   41,
   44, 1306,  466,  504, 1602,   59,  310, 2696, 1602,   44,
   59,    0,   41,   59,  401,   44,  296,  580, 3677,   59,
  363,   59, 1931,   61,  504,   59,  273,  259,  511, 3559,
   59,  262,   61,   59, 3819, 1326,    0,  269,   40,   91,
 3354,  316,  401, 3226,  330,  268,   59, 1392,   59, 1392,
   59, 2844, 2356,   59,  268,    0,  262,  461,  462,  463,
   44,   63, 1716,  316,  264,  334,  421, 3871,   40,   40,
   59,  426,  427,  428,  429,  430,  431,  432,  433,  434,
  266,  436,  279, 2393,   45,  440,  421, 1392, 2896,  438,
 1216,  277,   63,  263,  591,   59,  451,  264,   45,  264,
 1110, 1187,  553,  600, 1409,  447,  448,  264, 3912,  264,
  310,   45,  387,    0,   59, 1241, 1407, 1460,  310,    0,
  389, 2288,  264, 2290,  264, 1430, 1431, 1432, 1433, 1434,
  353, 3790, 1792,  295,  387, 2393,  368, 1138, 1634,  425,
  259, 3122,  310,  310, 1206,  264,  551, 1438,  264, 3679,
  269, 1442,  511,  310, 1162,  310, 1892,  264,    0,  301,
  264,  301,    0,  567,  364,    0,  436,  264,  310,  652,
  310,  264,   59, 1332,  654, 1334,  264,   44,   59, 1634,
   41,  744,  301,   44, 1527,    0, 1409,  418,  427,  429,
  264,  310,  643, 1259,  264, 1251,  312,  264,  553,  364,
  651,  264,  430,  264,  301,  452,  611, 1430, 1431, 1432,
 1433, 1434,  418,  310,  433,  436,  332,   59,  264,  433,
  721,   59, 2570,  275,   59,  422,  368,  301,  368, 3543,
    0,  264,    0, 1887,  301,  264,  310,  372,  395,  703,
  382, 3900,  382,  310,   59, 1949, 1950, 4032,  277,  312,
  299,  312,  295,  264, 3437,  301,  372,  295,  372,  322,
  323, 2211,  323,  382,  310,  381,  295,  681,  301,  332,
  314,  332,  335,   41,   44,  264,   44,  310,  372,  684,
  426,  310,  260,  638, 3018, 3019, 3266,  260,  643,   59,
  301,   59, 2059,  651,    0,  575,  651, 2510, 2511,  310,
  264,  368, 2331,  427, 2834, 1955,  752,  423,  382,  372,
  725,  372,  667,  668, 1974,  382,  805, 2658,  381,  264,
  381,  437,  368,  446, 1386, 2845,  681,  738,  437,  307,
    0,  446,  372,  688, 2984,   41,  382,  301,   44,  694,
  430,  696,  697,  504, 2538, 2508,  310,  401,  738,  382,
 2675, 1806,  738,   59,  709,  400,  301,  368,  738,  694,
  423,  696,  423, 2097, 1236,  310,  753, 2973,  638,  504,
  725,  382,  435,  435,  437, 1792,  437, 2538,  733,  734,
  735,  736,  427, 2537,  511,  394, 1956,  503,  504,   59,
  504,  806,  437,  801,  504, 2537, 3389,  740,  668, 2308,
  434, 1569,  437, 1529,  368,  750,  419,  523,  434, 1958,
  504,  816,  754,  523,  557, 4052, 2014,  831,  382, 4056,
 2014,  816,  264,  368,  557,  511,  264,  782,  400, 1959,
  600,  631,  523,    0,  738,  419,  791,  382, 1524,  709,
  503,  504,  503,  504,  446,  511,  801,  802,  803,  804,
  805,  806,  807,  808,    0,  427,  811,  511, 1801,  301,
  523, 1620,  523,  301,  504,  437,  631,  588,  310,  824,
  825, 1479,  310,  828, 1484,  446,  831, 1973,  296,  834,
  650, 3087, 2293,  523, 1570,  446,  648, 3280,  843,  523,
  597,  372,   59,  597,  264,   41, 1497,  803,   44,  446,
  381,  368,  808,  678,  597, 1810,  681,  394, 1667,  597,
 1801,  437,  446,   59, 3226, 3227, 1275,  523, 1973,   59,
 2256,    0,  588, 1512,  471,  652,  368,  597,  689,   44,
  368,  652, 1658, 2193,  395, 1878,  689, 1623,  738,  655,
  382,  811, 1107, 2791,  382, 1611,  738,   93, 3282, 3283,
 3284, 3285, 3286, 3287, 3288,  734, 1973, 1974,  264, 2213,
  721, 1271,   41,  802,  804,   44,  652,  810,  686,  687,
  738,  738, 3565, 3566, 1640,  689, 1642, 1633,  393, 1664,
   59,  738,  280,  738, 2351,    0,  652, 1810,  807,  632,
  811,  715,  655,  739,  655,  689,  738,  825,  738,  434,
 1891,  678,  631,  680,  681,  682,  312,  721, 3138, 3897,
    0, 2561,  437, 1973,   93, 2265, 2855,  323,  258,  738,
  738,  261, 2853,  504, 2852, 2310,  332,  721,    0,   44,
  364,  747,    0,  264,  363,  278,  678, 2083,  680,  681,
  682,  738,  523,    0,   59, 1094, 2805,    0, 1534, 2862,
 1099, 1537,    0, 2338,   44, 1646,    0,    0, 1512,   41,
  504,  301,  260,   44,  738, 1020,  372,  437,  646,   59,
  715,  738,   44,  646, 3482,  381, 2502,    0,    0,    0,
 2491, 2791, 2493,    0,  747, 1020,  747,   59,  523,    0,
    0,   59,  738,  363,   41,   61,   44,   44, 1351, 1352,
  104,  295,   59,  273,   41,  738,   59,   59,    0,  738,
  364,   59,   59,  314,   61,   59,   59,  423,   41,  487,
    0,   44, 3703,  316,  264, 3437,  273,  738,  295,   59,
   41,  437, 3545,   44,   44,  264,   59,   59,   59, 1094,
  306,    0,   59,  715, 1099,    0,  298, 2501,   59,   59,
 1909,  317, 1107,  307,    0,  398,  667,  575,  332, 1621,
 2759,   41, 1847, 2762,  420,   59, 1919,   59,   44,    0,
  302, 1939,  301,  340,  738, 1180, 2193,  447,  448,   59,
  284,  310,   42, 3517, 3518, 3519, 3520, 3521, 3522, 3523,
  312,  500,  344,  738,  387, 3099, 3365,  503,  504,  314,
   59,  437,   44,    0,   59, 2311, 3375,  275, 1163,  347,
   41,  655,  364,   59,  484, 1220, 1171,  523, 1223,  417,
 1946, 1226,  748,    0, 1229, 1220,  372,  597,   59, 1099,
  263, 1226, 1187,  316, 1229, 3485, 3486, 3487, 3488, 3489,
 3490, 3491,  371,   44,  410, 2136,    0,  314,  394,  588,
 2382,   93,  372,  382, 2386,  395, 1261,  555,  302,  303,
  530,    0,   59, 1198,   41,  264,  264,   44, 2211,  337,
 1225,  266,  332,  306, 1209, 1280, 1281,    0,  400,   59,
 3673,  264,   59,  437,  317, 1280, 1281, 3868,  544, 3419,
   44, 2196,  523,  372,  623,  551,  738,  631,  387, 3247,
  738,  571,  434,  423,  387,   59, 1914, 2212,  412,  290,
  564, 1973,  264,  652,  436,  394,    0,  264,  301,  298,
   59,    0,  304,  748, 2010,  268, 3615,  310, 2664,  295,
  277, 2222,  843,  500,  264, 1163,   59, 2022, 2106,  648,
  478, 2829,  295, 1171,  301,  588,  314,  299,  295,  655,
 2076,  295,  419,  264,  264,  276,  277,   41,  504, 1314,
 2086,  555,  532,  310,  332,  344, 1321,  304, 3555, 1429,
  264, 2771, 2772, 2196, 2785,   59,  268,  410,  447,  448,
   59,  306, 3551,  333,  264,  532,  401,  340,  371, 2212,
  434,  334,    0, 1444,  319, 3105,  340,  395, 3695,  382,
  260,  312, 2058, 2433, 2434, 2709, 2888, 3988, 3118, 2075,
 2109, 1460, 2109, 2002,  290, 1321,  505,  364, 3811, 2819,
 3813,  332,  417,  340,  428,  504,  619, 3633,   41,  401,
  434,   44,  312,  264, 1389,  357,    0, 1495,  298, 2469,
  348,  747,  334,  364,  401,  368,  686,  687,   61,  392,
 1405,   59,  332,  401,  312, 3365, 1514, 1515, 1516, 1517,
 1518,  451, 2712,  453,  407, 3375, 2833, 3066, 4042,  387,
 1405,  363, 2040,    0, 1429,  641,  321,   41, 2732,  369,
   44,  312, 1437, 1736,  264,  340, 2740,  264, 1443, 1444,
  433,  643,  644,  645,  340,   59,  511,  389,  299, 2759,
  392,  332, 2762, 1458,  412, 1460,    0, 3365, 3456,   59,
  264, 3680,  748,  749,   41,  407, 4017, 3375, 2624,  258,
 2109, 3784, 3691,  344,  523,   42,  437,  437, 1533, 2288,
   47, 2290,   59, 1538,  393,  312,  619,  487,  315,  511,
  282,  433,  708,  689, 4045,  322,  323,  500,  491,  439,
  327,  295,  332,  301,    0,  332,  500,  437,  335,  492,
  419,  296,  301,  511,  487,   59,  363,  284,  511, 1524,
 1525, 1443,  349,  276, 2974,  721,  494,  488,  436, 1534,
  264,  524, 1537,  360, 1589,  510,    0,  296,  368, 1524,
 1460,  368,  503,  504, 1589,  372,  340,  276,  277,  491,
  380, 1556,  257, 3353,  381,  450,  437, 2637,  641,  738,
  689, 3422,  523,   59,   41, 1570, 2561,   44, 2561,  511,
  397,  330, 3791,  503, 2310,  588,  447,  448,  312,  257,
   44, 3800,  524,  363,  703, 1570,  368,    0, 3030,  323,
  363, 3551,  721,  523, 3132,   59,  423, 3135,  332, 3946,
  447,  448, 2338,  401, 3546,  296, 2561,  437,  435,  306,
  437,  364,  401,    0,  643,  644,  645,  296, 1623,  389,
 2561,  326,  319, 3621,  454,  296, 2286, 3864,   41,  316,
  419,   44,  295,  437,  631,  364,  301,  484,  372,  652,
  425, 1646,  523, 3551, 3991,  412,   59,  381,  326, 1654,
  264,   59, 2388,    0,   41,  447,  448, 2393, 1663,  280,
  487, 2616, 1667, 1668,  264, 4012,  425,  456,    0,  439,
  312, 1654,   59,  389,  447,  448,  503,  504,  302,  303,
  301,  508,  316,  530, 3949, 3217,   59,  264,    0,  423,
 2348,  334, 2759,  400,  655, 2762,  523, 4044,  312,  266,
  387, 3005,  335,  437,  484,  738,  500,    0,  322,  323,
  277,  484,   59,  257,  281,  276, 1721,  303,  332,  620,
 3680,  335,  355,  439,  571, 3866,  312,   59,  689,   41,
  276, 3691,  332,    0,  425,  312,  401, 2513,  458,  459,
 3044,  738,  429, 2616,  316, 4010,  425,   59,   41, 2893,
  530,   44,  313,  387,  425,  332, 3066,  530,  372,   59,
  721,  257,  273,    0,  407, 1721,   59,  381,  368,  503,
  504,  393, 3680, 3077,   41, 2481,  409,  321,  363, 2495,
  380,  301,  326, 3691, 2500, 1663, 1791, 1792,  689,  523,
  295,  571,   59, 3997, 1893,  258,  273,  419,  571,  303,
 3378, 3379,    0,    0,   41,    0,    0, 1792,  312,  423,
  434, 3378, 3379,  510,  400,  387,  401, 1925,  364,  260,
  721,  435,   59,  437,  385,  321, 1875,  304,  655, 1834,
  326, 3791, 4036,  344,    0,  456,  380,  437,  301,  385,
 3800,   59,  257,   41,   41,  391, 2504,   41, 3378, 3379,
  436,  264,  396,  340,  454,  295, 1955,  298,  588,  364,
  437,   59,   59,  424,   59,   59,  451,  344, 3685, 1791,
 1875, 1981, 4013,  593, 2683, 3058, 3059,  264,  424, 3378,
 3379,  401,    0, 3791,  380,  283, 3378, 3379, 1893,  503,
  504,  264, 3800,   59, 1949, 1950,  400,  295,    0,  312,
  396,  301,   44, 2721,    0,  276,  321, 2008,  372,  523,
 2009,  326,  400,   44, 1919,  265,  388,  381,  395,  332,
  747,  655,  652, 2726,  557,  312,  503,  504,  454,  276,
  277,  403,  436,    0,  555,  322,  323,  620,  401,  427,
  466,   59,  313,  487, 3936,  332,  523,  297,  259,  437,
 1955,   93,  264, 1958, 1959, 3352,  419,   59,  269,  332,
  447,  448,   93,   59,  264,  380,  364,  295,  494, 1974,
  434,  435, 1957,  276,   41,  452, 1981, 3969,  296,  329,
 2534,  396,  382,  276,  277,  372,  727, 1992,    0, 1974,
  283,  487,   59,  456,  381,  368, 2001,  264, 3952, 3066,
  312,  401, 2007, 2008, 2009, 2010,  689,  380, 4000,  588,
  322,  323,  330,  747,  385,    0,  344,  364, 2023,  419,
  332, 2962, 2534, 2729,  437, 2010, 2031,  264,  257,   41,
 2035,  401,  332,   59,    0, 1955,  423, 1959,  721,  260,
  504,  655, 2791,  478, 2791,  312,  264,   59, 2053, 2054,
  437,  446,  596,  424,  598,  599,    0,  368,   44,  523,
  372,  364,  265,  258,  437,  332,  264,  264,  368,  381,
  264,  364,  507,  652,   59,  270,  471,  298,  655, 2001,
  380,  454,  385,  312,    0, 2090, 2091, 2054,  391,    0,
  503,  504,  385,   59,  297,  332,    0,  326,  391,  426,
  596, 2023,  598,  599, 2109,  649,  301,   93,  303, 2031,
  523,  423,  689, 2035,  332,   59,  503,  504,  312,  426,
    0,  424,  260,  435, 2091,  437,    0,  653,  271, 2230,
  315,  424, 2231,  747,  332,  332,  523,  437,  332,    0,
   44,  284,  327,   59,  721,  600,  268,  454,   59,  675,
  368,  264, 2791,  649,  454,   59,  258, 2893,  684,    0,
  298,  685,  380,  479,  349, 2264, 2265, 2806,  270,  260,
  437,  296,    0, 3317,  698, 2233,    0, 2053, 2836,   59,
  393,  596,  325,  598,  599,   59, 3696, 2109, 2193, 2194,
 4020,  503,  504,  509,  487, 3403,  280,  264,   59,  301,
  437,  303,    0,   44,    0,  330,  401,  298, 2193,  352,
  736,  523,  397, 2218, 2090, 2220,  276, 4047,   59,  437,
 2225, 3039,    0,  283, 2229, 2230, 2231, 2232, 2233, 2234,
 2235,   59, 3378, 3379,  649,   59,  454,  685,  264,  437,
  437,  363,  655,  437,  261,  312, 2229, 2056,   44, 2254,
  698,  260,  264,  260,  479,  368,  523,    0,  446, 2264,
 2265,   59, 2267,   59,  489,  332,  391,  380,  655,  412,
  392, 3668, 2194, 3378, 3379,  400,  689,    0, 2234, 2235,
  558,   59, 2267, 2092,  509, 2290, 2341, 2342,    0,  298,
    0,  298,  260,    0,  686,  687, 2341, 2342,  382,  401,
  425,   44,    0, 2225,  582, 2310,  332,  479,  721, 2314,
  295,  433, 2232,  316,  258, 2373,   59,  489,   41,   41,
  332,   44,   44,   46,  437, 2310,  270,    0,  315,  523,
  298,    0,  363, 2338,   59,  419,   59,  509,   61,    0,
  327,  454,  368,  655, 2264, 2265,  738,   59,  335,   59,
    0,  295,   59, 2338,  380,  340,  316,  301,  389,  303,
  747,   59,  349,  348,    0,    0,    0,  600, 2373,  491,
  437,  363,  456,  360,  259, 2238,  319, 2382, 2383, 2384,
 2385, 2386,  114, 2388,  269,    0,   59,  329, 2393,  511,
   59,   41, 2314,  378,   44,  380,  340,  389,   59,  624,
  625,  626,  524, 2388,  348,  315, 2269,  139, 2393,   59,
  397,  437, 2275,  749, 3875, 2525,  401,  327,  528,  614,
  615, 3362, 2487,   59,   59,   59, 2431,  511,  454,  257,
 2485, 2436, 2487, 2488, 3150,  747,  503,  504, 3214,  349,
 2485,  426, 2487, 2488,   59, 2450,   59,    0, 2453,    0,
  360,    0,  624,  625,  626,  454,  523,  401,  368,  602,
    0,  367, 2517,    0, 2519, 2520, 2521,  695,  412,  454,
  413,  555, 2517,  330, 2519, 2520, 2521,  420,    0,    0,
   58,    0, 2487, 2488,    0,   63,    0,  397,  401,    0,
 2545,  686,  687,  321,  689, 1316, 3314, 1318,  326,  264,
 2545,  696, 1323, 1324,   41,  558,   59, 1328,   59, 2431,
   59,  454,  614,  615, 2436,  500,   44,  295,  685,   59,
 2525,  508,   59,  466,  261,  594,   40,   41, 2450,  582,
   44,  698, 2537, 2538, 1355,  604,  605,   59,   59,  264,
   59,  264,   46,   59,   40,   59,  258,   61,   59,  286,
  591,  494,  380, 3548,   58, 3548,  500,   61,  270, 2658,
  258,  261,  340,  263,  301, 3277, 2667,  425,  396, 2668,
  348,  261,  270,  461,  462,  463,  299,   91,  348,    0,
  401, 3317, 4043,  295,  686,  687,  286,  689,  655,  301,
 2655,  303,  443, 3548,  696,  446,  286,  258,  508,  437,
  378,  301,  380,  301,  451,  303,  453,  332,  378,  270,
  380,  301, 2670, 2712,   42,   43,    0,   45,  427,   47,
  471, 3925,  266,  401, 3928,  461,  462,  463,  340, 3365,
 2635,  401,  751,  277,  295,  315,  348,  281,   59, 3375,
  301,   41,  303,  368,   44,  310,  311,  327,  426,    0,
 2655,  264, 2657, 2658, 2709,  380,  426,   41,   40,  487,
   44, 2666, 2667, 2668, 2669, 2670, 2671, 2672, 2778,  349,
  614,  615, 2657, 3333,   41,   59,  454,   44,  427,  340,
  360,   46,  738, 2666,  454,  479,   46,  348,  437,  401,
   41,  407, 2800,   58,   59,  489,   61, 2702,   58,   59,
  412,   61,   41,  401,  479,   44,   41, 2712,   59,   44,
  653,  447,  437, 2635,  489, 2671, 2672,  397,  680,  332,
  682, 2726,  500,  259,   46,   42,   43,  264,   45,  454,
   47,   41,  675,  269,   44,  264,   58,   59, 2658,   61,
  401,  684,  686,  687,  258,  689, 1202,  310, 1204, 2669,
  264,  412,  696,   41, 2759,  368,   44, 2762, 3965,  273,
 3967,  275,  276,  277,  447,  448,  280,  380,  596,  283,
  598,  599, 2777, 2778, 2759,  312, 2760, 2762,    0,  616,
  617,  295, 2787, 3586, 2839,  312, 2791,  301,  500,  417,
  304,  329, 2712,  736, 2839,  332,  310, 2802,  328,   42,
   43, 2806,   45,  435,   47,  458,  459,  460,  461,  462,
  463,  464,  465,   41,   41, 3551,   44,   44,   40,   41,
 2825,  649,   44, 2828,  437,  487,    0,   46,  508,   41,
  344,  487,   44,   46,  262,  372,  419,   59,  266,  500,
   41,  454,  446,   44,  381,   58,   59,  492,   61,  277,
  364,   41,  419,  281,   44, 2777, 2861,  528,   41,  332,
   46,   44,  440,  441,  442,  443,  444,   41,  446, 2791,
   44,  385,   58,   59,  295,   61, 3685,  391,  264,   41,
 2802,  395,   44,  417, 2806,   59,  423,  401, 2893,   41,
  466,  467,   44,  471,  472,  369,   46,  316, 2903, 2904,
  437,   41,  614,  615,   44,  419,  447,  448,   58,   59,
  424,   61,   41,  264,  567,   44,  614,  615, 2983,  340,
 2903, 3320, 3321, 3322, 3323, 3324, 3325,  348,   41,  376,
  377,   44, 2997, 2998,   41,  401,    0,   44,  452, 2861,
  593,  379,  456,   41, 3680,  262,   44, 2952,   41,  266,
  379,   44,  333,  614,  615, 3691,  367,  378,  491,  380,
  277,  312, 2967,  523,  281, 2970,  503,  504, 2973,   60,
   61,   62,  273,   41,  686,  687,   44,  689, 2983,   40,
  401,  332, 2904,  401,  696,  715,  523,   41,  686,  687,
   44,  689, 2997, 2998,   41,   59,   41,   44,  696,   44,
  678,  479,  480,  681, 3009,  426, 2976, 2977, 2978, 2979,
 2980, 2981,   41,   41,  304,   44,   44, 3022,  735,   41,
  264,  372,   44,  266, 3009,  686,  687,  264,  689,  264,
  381,  264,    0,  454,  277,  696, 3041,   41,  281, 3022,
   44,  555,  264,  261,  268,  261, 3051,  263, 2970, 3054,
  264, 2973,  600, 3607,   41, 3791,   41,   44,  735,   44,
   41, 3066,   41,   44, 3800,   44,   58,  282,  286,   61,
  286, 3054,  423,   41, 3378, 3379,   44, 3895,  296,  500,
  296, 3066, 3087,  301, 3089,  301,  437,  738,  741,  742,
  312,   59,  315,  315,   41,   41,  264,   44,   44,    0,
  322,  323,  276,  277,  327,  327,  280,  264,   41,  283,
  332,   44,  330,  335,  330,  381,   41,  631,  655,   44,
 3228,  295,   41,  340,  650,   44,  349,  349, 3133,   41,
   41,   41,   44,   44,   44,  340,  137,  360,  360,  140,
   41, 1885,   40,   44, 1888,  368,  368,   60,   61,   62,
  372,   40,  503,  504, 2240, 2241,   40,   40,   59,  381,
 3215,  447, 3167,   40,   40, 3087,   40, 3089,   40,   40,
 3215,  262,  523,  391,  397,  397,   40, 3232,   40,   40,
  271,   40,  400,  447, 3167,  447,   40, 3232,   40,   40,
  364, 3292,   40,  284,   40, 3294, 3854,   40, 3856,  347,
  670,  423,   40,   40,  268,   40,   40,  425,   40,  425,
   40,  385,   40,  435,   40,  437,   40,  391,   40,   40,
   40,   40,   40,   40,  738,    0,   40,  318, 3293,   40,
   40, 3296,   40, 3238,  325, 3240,   40,   40,   40,   40,
   40, 3306,  333,   40,   40, 3310,   40,   40,   40, 3298,
  424, 3295,   40,   40,   40, 3238,   40,   40,   40,  350,
   40,  352,   40,   40,  487,  487,   41,   40,   40,   44,
  334,  362,   40,  364,  365,  366,   40,   40, 3832,    0,
  371,  503,  504,  394,   59,  508,  508, 3292, 3293, 3294,
 3295, 3296, 3297, 3298, 3299,   60,   61,   62,   44,  363,
   91, 3306,  311,  618,  655, 3310,   40,   40,  276,  277,
   40,   40, 3317,   40,   46,  283,  368,   40, 3240,  370,
   41,  412,   46,   44,  370,  389,   40,  295,  392,  430,
  430, 3439,   40,  341,  384,   40,  430,  275,   59, 3344,
  419,    0,   40,  407,   40,   40,   40, 3352,   40,  262,
  333,  333,  503,   46,  295,  357, 3411,    0,  271,   40,
 3365,  382,  437,  264,  450,  270, 3411,  295,   40,  433,
 3375,  284,   91,  735,  270,  368,  446, 3432,  374, 3299,
 3365,  555,   41, 3438,  372,  312,   41, 3432,   41,  367,
 3375,   41,  862, 3438,   41,  367,  364,   41,   41,   41,
   59,   44,   40, 3408,  437,  318,   46,  419,   40,  386,
  368,  312,  325,  368,  315,  368,   59,  385,   61,  889,
  333,  322,  323,  391,  290, 3430,  327,  491, 3493,   40,
   40,  332,   46,  655,  335,  260,  466,  350,  401,  352,
  466,  466, 3447, 3508,   60,   61,   62,  511,  349,  362,
   40,  364,  365,  366,  312,  312,  424,   44,  371,  360,
  524,  314,   44,  333, 3447,  333,   46,  368,   44,  332,
   40,  372,  347,  440,  441,  442,  443,  444,  347,  446,
  381,  513,  500,  499,  499,  437,  446,  719, 3493,  719,
  616,  270,   59,   59,   41,  294,  397,  262, 3503,  412,
  395,  276,  277, 3508,  471,  472,  271,   44,  283,  348,
  600,  602,  487,  437,   40,  371,  371,  418,   40,  284,
  504, 3526,  423, 3528,  371,  747,  355,   41,   41, 3584,
   41,   41,   44,  434,  435,   41,  437,   41,  314, 3544,
 3545,   44,   44, 3526,  314, 3528, 3551,  314,   41, 1019,
 3555,  418,   41,  318, 3618,  276,  277,  306,   41, 3564,
  325, 3544,  283,   41,   44,   46, 3551,  316,  333,  314,
  319,  333,   40,  749, 3579, 3580,   41,  419,   40, 3584,
  395, 3586,   41,  419,  417,  350,  487,  352,  401,  364,
  435,   44, 3597,   44,   44,  275,  419,  362,   41,  364,
  365,  366,  503,  504,  373,  264,  371,  508,  314,  335,
  385,  368,  368, 3618, 3619,  258,  391,  335,  368,  290,
  300,  264,  523, 3545,  414, 3683,  306,   41, 3633,  749,
  749,  311,  275,  276,  277,  749,  316,  280,  387,  319,
  283,   41, 3564,  364,   41,   41,  737,  412,   41,  424,
   41,   61,  295,  312,   46,  271,   41,  488,  301,  419,
  264,  435,  340, 3668,  385,  340, 3721,  310,  284,  264,
  391,  340,  421,  332,  455, 3680, 3721,  426, 3683, 3684,
  719, 3736,  719,  432,  276,  264, 3691,   58,  419,  602,
  739, 3736, 3020,   61,  303, 3680,  739, 3619,  303,  314,
   40,  304,  318,  424,  395,  303, 3691,  387, 3692,  325,
  419, 3633, 3822,  372, 3719,  418,   41,  333,  314,   41,
 1190,  364,   40, 1193,  264,  263,  620,  263, 1198,    0,
  314, 1201,  278,  335,  350, 1205,  352,  417,  527, 1209,
  314,   41,  385,  528,   41,  399,  362,  437,  391,  365,
  366,  437,  432,   44,  655,  333,  399,  322,  401,  525,
  401,  510,  487,  369,  423,  314,  487,   45,  419,   40,
   41,  414,   40,   44,  454,  276,  419, 3782,  437,  692,
   58,  424,    0,  344,  344,   63, 3791,  736,   59,   40,
   61,  332,   59,  332,  303, 3800,  412,  273,   61,   60,
   61,   62, 3807, 3867,  717,   58, 3791,  556,  367,  437,
  277,  386,  735,  456,  735, 3800, 3821, 3822,  371,  371,
   91,  371,   40,   41, 3807,  370,   44,  371,  371,  371,
  510, 3836,  386,  513,   59,  594, 1306,  602,   59,  358,
  437,   59,  450,   61,  503,  504,  747, 1317,  511,  450,
   40,  600, 1322,   59, 3918,   59, 1326,  303,   44, 3864,
  367,   41, 3867,  434,  523,   41, 1336,   41,   41,   41,
  619, 1341, 1342,   91,  273, 1345, 1346, 1347, 1348, 1349,
 1350, 1351,   41, 1353, 1354,  419, 1356,  419, 1358, 1359,
 1360, 1361, 1362,  709,  419, 1365, 1366, 1367, 1368, 1369,
 1370, 1371, 1372, 1373, 1374, 1375, 1376, 1377, 1378, 1379,
 1380,  419,  555, 3918,    0,  351,  407,   41,  405,  419,
 1390, 1391, 1392,  419,  419,   41,  526,  692,  650,   44,
  306, 3936,  273,  466,  323,  466,  369, 1407, 1408,  619,
  316,  336,  332,  319,  434, 1415,  420,   40,  719, 3954,
 1420,   59,  717,  364,  273,   41, 1426, 1427,   44,  364,
  344,  719,   40, 3291, 3969, 1435,  330,  425, 1438, 1439,
 1440, 1441, 1442,   59,  532, 1445, 3304,  425,  330,  296,
  296,  419,  650,  325,  314, 3990,  602,  258,  631,  446,
    0,  368,  613,  264,  454, 4000,  655,  261, 3326, 3327,
  454,  262,  273,   40,  275,  276,  277, 3990,  371,  280,
  271,  387,  283,  291,   40,   40,  314,   41,  357,  734,
   61,  336,  286,  284,  295,  450,   40,  557,   44,  314,
  301,   41,  296,  304,   44,  314,  749,  301,  314,  310,
  258,  417,  314,  431,  314,  421,  264,  749,  314,   59,
  426,   61,  314,  427,   44,  273,  432,  318,  276,  277,
  747,   41,  280,   40,  325,  283,  330,   41,  299,  299,
  494,  280,  333,  344,  289,  623,  631,  295,   60,   61,
   62,   91,  264,  301,  344,  344,  304,  275,  340,  350,
  401,  352,  310,  364,  611,  738,  494,   40,  418,  304,
  303,  362,   41,  364,  365,  366,  611,   41,  689,   63,
  371,  321,  370,  273,  385,    0,   44,  312,  411,  718,
  391,  434,  431,   40,  395,  713,  344,  391,   44,  418,
  401,  314,  389,   41,  510,  312,  400,   46,  336,  395,
  336,   41,  677,  344,   41,   46,  364,  401,  419,  418,
  370,  412,   44,  424,  264,  299,   41,   40,  487,   44,
  295,  425,  440,  441,  442,  443,  444,  385,  446,  447,
  448,    0,  264,  391,   59,  555,   61,  395,   44,   61,
  556,  452,   44,  401,   41,  456,  357,   40,  466,  467,
  276,  277,  719,  471,  472,  719,   41,  283,  654,  714,
   40,  419,  564,  302,  447,  299,  424,  264, 1678, 1679,
  391, 1681,   41, 1683,  632,   44,  571, 1687, 1688, 1689,
 1690,  652,  278,  513,  452, 1695, 1696,  586,  654,  717,
   59,  266,   61,   40,  452,  289,   41,  295,  456,  329,
  329,  654,  717,  619, 1714,  262,   41,   41,  258,  376,
  283,  565,   40,   40,  264,  504,  504,  426,  310,  368,
  654, 1731,   91,  273,  692,  275,  276,  277,  330,  420,
  280, 1741, 1742,  283,  386,  420,  418,  368,  364,  418,
  262,  368,  368,  401,  555,  295,  609,  296,  425,  271,
  336,  301,  330,  449,  304,  607,   41,  401,  427,  385,
  310,  609,  284,  437,   40,  391, 3634, 3635, 3636, 3637,
 3638, 3639, 3640,   44,  427,   40,  400,    0,  329,  400,
   60,   61,   62, 3651, 3652, 3653, 3654, 3655, 3656, 3657,
   40, 1801,  264,  264,  344,  264,  318,  555,  424,  264,
  264,  602, 1812,  325, 1814,  400,  419, 1817,  333,  400,
  270,  333,    0,  395,  364,   40,  335,   41,  434,  340,
  631,  303,  303,  270,  270,  643,  644,   40,  350,   41,
  352,  264,  368,  258,  335,  385,  290,  303,  446,  264,
  362,  391,  364,  365,  366,  395,  270,  487,  273,  371,
  303,  401,  277,   41,   41,  280,   44,   59,  368,  407,
  511,  487,  395,   41,   41,   61,  336,  427,   59,  419,
  295,   59,  303,  631,  424,   59,  301,   59,   59,  304,
  289, 1891,   40,  303,  314,  310,   41,  618,   41,  258,
  412,   41,   59,   59,   59,  264,   59,   41,   41,  511,
  418,   41,  452,  258,  273,  264,  456,  276,  277,  314,
  314,  280,  299,   41,  283,  270,  717,  446,  446,  344,
   41,  738,  523,  738,    0,  480,  295,  738,  646,   40,
   46,   46,  301,  294,  294,  304,  294,   41,    0,  364,
   41,  310,  294,  305,  389,   44,  301, 1957,  303,  305,
 1960,  738, 1962,  264,  305,  389,   46,  294,  294,  401,
  523,  364,  294,  305,   40,   41,   41, 1977,   44,  294,
  395,  555,  597,   44,    0,  344,  401,  597,  597,   41,
  738,  264,   44,   59,  401,   46,    0, 1900, 2648, 3493,
 1475,  271, 1475, 1475,  419,  364, 2647,   59, 1475,   61,
    0, 1475, 2649, 3493,  284,  555, 1476, 2675, 2254, 3306,
 2541, 3227, 3437,  426, 3158,   41,  385, 1667,   44, 3413,
 3103, 2495,  391, 2853,  808,   82,  395,  452, 3150,   91,
 2443,  456,  401,   59, 3706,    0,  103, 3603,  318, 3821,
 2482,   41, 3138, 3783,   44,  325,  401,  430, 3718, 2527,
  419, 3832, 2861,  333, 1171,  424, 3159, 4030, 3701,   59,
 3912,   61, 2861, 4035, 3948, 3947, 3982, 3087, 3633, 3088,
  350, 3787,  352,  709, 1662, 2092,   41, 1603, 2094,   44,
  602,  631,  362,  452,  364,  365,  366,  456,  276,  277,
 3559,   91,  280, 2760,   59,  283, 2014, 2760, 4000, 3901,
 4001, 2386, 2756, 3690, 2114, 3794,    0,  295, 1846, 2970,
 2120, 3211, 2508, 1239, 1997, 2125, 2449, 3817, 2128, 2129,
 2130, 2035, 3704, 1996, 1573, 2135, 2136, 2137, 3544, 1638,
  555, 1621,  412, 2491, 2144, 1159, 2146,  141, 3344, 2149,
 2150, 2151, 2152, 2276, 1445, 1512, 2156, 3333, 2158, 3026,
 2160, 3526, 2162,  736,  421,  696,  743, 1957, 2168, 2169,
 2170, 2171, 2172, 2173, 2174, 2175, 2176, 2177, 2178, 2179,
 1209, 1184, 1459, 2183, 2184, 3001,  364, 1460, 2188, 2189,
 1886, 2268, 3537, 3668,  706, 3447,  555, 3954,  738, 2199,
 2705, 1960,    0,    0, 2204,  747,  258,  385, 2208, 2720,
  276,  277,  264,  391,    0, 1309,  631,  283, 3512, 2219,
 3512,  273, 2222,    0,  276,  277,    0,    0,  280,  295,
 2310,  283, 2388, 3931,  432, 1417, 2513, 1811,    0, 1523,
 2524, 3719,    0,  295,   41, 3580,  424,   44,  826,  301,
  843, 1973,  304,  789,   60,   61,   62,  998,  310, 1438,
  276,  277,   59, 1386,   61, 1644, 3782,  283,  258,  614,
  615, 3626,  631, 1555,  264,  999, 2155, 1555, 2290,   41,
 3474, 1383,   44,  273, 3248, 2787,  276,  277,  118, 3297,
  280, 2270,  344,  283,   91, 1456, 2674,   59,  364,   61,
 3290,   59, 2893, 1297, 3356,  295, 3952, 2088, 3592, 3435,
 3730,  301,  364, 2313,  304, 3572, 3812, 3914, 1134,  385,
  310,  276,  277,  738, 2790,  391,   12, 3741,  283,   91,
 3024, 1219,  602,  385, 3004, 3476, 3894, 3595,   -1,  391,
  295,  686,  687,  395,  689,   -1,   -1,   -1,  364,  401,
   -1,  696,   -1,   -1,  344,   -1,   -1,   -1,  424,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  419,   -1,  385,
   -1,   -1,  424,   -1,  364,  391,   -1,  555, 2378,  738,
   -1,   -1,   -1,    0,   -1,   -1,   -1, 2387,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  385,    0,   -1,   -1,   -1,
  452,  391,   -1,   -1,  456,  395,   -1,   -1,  424,  364,
   -1,  401,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   60,
   61,   62,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  419,
  385,   -1,   -1,   -1,  424,   -1,  391,   41,   -1,   -1,
   44,   -1,   59,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   59,  262,   61,   -1,   -1,
   -1,  258,  452,   -1,    0,  271,  456,  264,   -1,  424,
   -1,  487,   -1,   -1,   -1,   -1,  273,   -1,  284,  276,
  277,   -1,   -1,  280,    0,   -1,  283,   -1,   -1,  555,
   -1,   -1,   -1,   -1,   -1,   -1,  258,   -1,  295,   -1,
   -1,   -1,  264,  555,  301,   41,   -1,  304,   44,   -1,
  268,  273,  318,  310,  276,  277,   -1,   -1,  280,  325,
   -1,  283,   -1,   59,   -1,   61,   -1,  333,    0,   -1,
   -1,   -1,   -1,  295,   -1,   -1,   -1,   -1,   -1,  301,
   -1,   -1,  304,   59,  350,   -1,  352,  344,  310,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  362,   -1,  364,  365,
  366, 2561,   -1,   -1,   -1,  555,   -1,  364,   -1,   41,
 2570,   -1,   44,   -1,   -1,   -1,  334, 2577, 2578,  631,
   -1,   -1,  344,   -1, 2584,   -1,   -1,   59,  385,   61,
   -1,   -1,   -1,   -1,  391,   -1,   -1,   -1,  395,   -1,
  555,   -1,  364,   -1,  401,  363,  412,   -1,   -1,   -1,
   -1,   -1,   -1,   -1, 2614,   -1,   -1, 2617,   -1, 2619,
   -1,  262,  419,  385,   -1,   -1,    0,  424,   -1,  391,
  271,  389,   -1,  395,  392,   45,   -1,   -1,   -1,  401,
   -1,  631,   -1,  284,   -1,   -1,   -1,   -1,   58,  407,
   -1,  268,   -1,   63,  258,  452,   -1,  419,   -1,  456,
  264,   -1,  424,   -1,   -1,   -1,   -1,   41,   -1,   -1,
   44,  275,  276,  277,   -1,  433,  280,  318,   -1,  283,
   -1,   -1,   -1,   -1,  325,   59,  738,   -1,   -1,   -1,
  452,  295,  333,   -1,  456,   -1,   -1,  301,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  310,   -1,   -1,  350,
   -1,  352,   -1,   -1,   -1,   -1,   -1,  334,   -1,   -1,
 2720,  362,  258,  364,  365,  366,   -1,   -1,  264,   -1,
  371,   -1,   -1,  491,   -1,   -1,   -1,   -1, 2738,  275,
  276,  277,   -1,   -1,  280,   -1,  363,  283,  738,   -1,
   -1,   -1,  268,  511,   -1,   -1,   -1,   -1,  555,  295,
  364,   -1,   -1, 2763,   -1,  301,  524,   -1,   -1,   -1,
   -1,  412,  389,   -1,  310,  392,   -1,   -1,   -1,   -1,
   -1,  385,   41,   -1,   -1,   -1,  258,  391,   -1,   -1,
  407,   -1,  264,  555,   -1,  399,  602,  401,   -1,   -1,
   -1,   60,   61,   62,  276,  277,   -1,   -1,  280,   -1,
   -1,  283,   -1,   -1,   -1,  419,  433,   -1,  334,   -1,
  424,   -1,   -1,  295,   -1,   -1,   -1,   -1,  364,  301,
   -1,   -1,   -1,   -1,  631,   -1,   -1,   -1,  310,   -1,
   42,   43,   44,   45,   46,   47,   -1,  363,   -1,  385,
   -1,   -1,  456,   -1,   -1,  391,   -1,   -1,   60,   61,
   62,   -1,   44,   -1,   -1,  401,   -1,   -1,   -1,  631,
   -1,   -1,   -1,  389,  491,   -1,  392,   -1,   60,   61,
   62,  291,   -1,  419,   -1,   -1,   -1,   -1,  424,   91,
   -1,  407,  364,   -1,  511,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  276,  277,   -1,   -1,   -1,  524,   -1,  283,
 2910,   -1, 2912,  385, 2914,   -1,   -1,  433, 2918,  391,
  456,  295,   -1, 2923, 2924, 2925, 2926, 2927,   -1,  401,
   -1,   -1, 2932,   -1, 2934,   -1, 2936,   -1,   -1,   -1,
   -1,  738,   -1,   -1,   -1,   -1,    0,  419,   -1,   -1,
   -1,  555,  424,   -1, 2954,   -1,   -1,   -1, 2958,   -1,
   -1,  602, 2962,    0,   -1, 2965,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  491,  738,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  456,   -1,   -1,   -1,   -1,   -1,
  364,   -1,   -1,   -1,   -1,  511,   42,   43,   44,   45,
   46,   47,   -1,  262,   41,   59,   -1,   44,  524,   -1,
   -1,  385,  271,   -1,   60,   61,   62,  391,   -1,  555,
   -1,   -1,   59,   -1,   -1,  284,   -1,  631,   -1,   -1,
  440,  441,  442,  443,  444,   -1,  446,  447,  448,   -1,
   -1,   -1,   -1, 3043,   -1,   91,   -1,   -1,   -1,   -1,
  424,   -1,   -1,   -1,   -1,  314,  466,  467,  699,  318,
  262,  471,  472,   -1,  266,   -1,  325,   -1,   -1,  271,
   -1,   -1,   -1,   -1,  333,  277,   -1,   -1,   45,  281,
  262,   -1,  284,  555,   -1,   -1,  275,   -1,   -1,  271,
   -1,  350,   -1,  352,  504,  631,   -1,   -1,   -1,   -1,
   -1,   -1,  284,  362,   -1,  364,  365,  366,   -1,   -1,
   -1,  300,  371,   -1,   -1,   -1,  318,  306,   -1,   -1,
   -1, 3121,   -1,  325,   -1,   -1,   -1,  316,   -1,   -1,
   -1,  333,   -1,   -1,  738,   -1,  318,   -1,   -1,   -1,
  329,   -1,   -1,  325,   -1,   -1,   -1,   -1,  350,   -1,
  352,  333,   -1,  412,   -1,   -1,   -1,   -1,   -1,  631,
  362,   -1,  364,  365,  366,   -1,   -1,   -1,  350,  371,
  352,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  362,  555,  364,  365,  366,   -1,   -1,   -1,   -1,  371,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  387,   -1,
   -1,   -1,  738,   -1,   -1,   -1,   -1,   -1,    0,   -1,
  412,   -1,   -1,   -1,  268,   -1,  262,   -1,   -1,   -1,
  266,   -1,   -1,   -1,   -1,  271,   -1,   -1,  417,   -1,
  412,  277,  421,  643,  644,  281,   -1,  426,  284,  276,
  277,   -1,   -1,  432,  654,   -1,  283, 3247, 3248,   41,
   -1,   -1,   44,   -1,   -1,   -1,  458,  459,  460,  461,
  462,  463,  464,  465, 3264,  454,  738,   59,   -1,   -1,
   -1,   -1,  318,   -1,   -1, 3275,   -1,   -1,   -1,  325,
  334,   -1,   -1,   -1,   -1,   -1,   -1,  333,   -1,   -1,
   -1,   -1,   -1,   42,   43,   44,   45,   46,   47,   -1,
   -1,   -1,   -1,   -1,  350,   -1,  352,   -1,   -1,  363,
   -1,   60,   61,   62,   -1,   -1,  362,   -1,  364,  365,
  366,  510,   -1,   -1,  291,  371,   -1,  364,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  389,   -1,   -1,  392,   -1,
   -1,   -1,   91,  602,   -1,   -1,   -1,   -1,  385,   -1,
  317,   -1,   -1,  407,  391,   -1,   -1,   -1,   -1,   -1,
   -1,   41, 3362, 3363,   44,  567,  412,   -1,   -1,   -1,
   -1, 3371, 3372,   -1,   -1,   -1,   -1,  344,   -1,  433,
   60,   61,   62,   -1,   -1,   -1,   -1,  424,   -1,   -1,
    0,  593,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  602,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  600,  458,  459,  460,  461,  462,  463,  464,  465,
  602,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  619,   41,   -1,   -1,   44,   -1, 3436,  491,   -1,   -1,
   -1,   -1,   -1,  410,   -1,   -1,   -1,   -1,   -1,   59,
   -1,   -1,   -1,   -1,   -1,   -1, 3456,  511,   -1,  426,
   -1,   -1,   -1,   -1, 3464,   -1, 3466,   -1,   -1,   -1,
  524,   -1,   -1,  440,  441,  442,  443,  444,   -1,  446,
  447,  448,   -1,   -1,  276,  277,   -1,   -1,   -1,   -1,
   -1,  283,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  466,
  467,  468,  469,  470,  471,  472,  473,  474,   -1,   -1,
   -1,   -1,   -1,  262,  481,  482,   -1,  266,   -1,   -1,
   -1,  567,  271,  490,   -1,   -1,   -1,   -1,  277,   -1,
   -1,   -1,  281,   -1,   -1,  284, 3536, 3537,   -1,  741,
  742,   -1,   -1,   -1,   -1,   -1,   -1,  593, 3548,   -1,
   44,   -1,   -1,   -1,   -1,   -1,  602,   -1,   41,   42,
   43,   44,   45,   -1,   47,   -1,   60,   61,   62,  318,
   -1,   -1,  364,   -1,   -1,   -1,  325,   60,   61,   62,
   -1,   -1,  262,   -1,  333,   -1,   -1,   -1,   -1,   -1,
   -1,  271,   -1,  385,   -1,   -1,   -1,   -1,   -1,  391,
   -1,  350,   -1,  352,  284,   -1,   -1,   -1,   91,   -1,
   -1,   -1,   -1,  362,   -1,  364,  365,  366,   -1,   -1,
   -1, 3621,  371,   -1,   -1, 3625, 3626,  275,   -1,   -1,
   -1, 3631,  424,   -1,   -1,   -1,   -1,   -1,  318,   -1,
  288,   -1,  290,   -1,   -1,  325,   -1,   -1,   -1,   -1,
   -1,   -1,  300,  333,   -1,   -1,   -1,   -1,  306,   -1,
   -1,   -1,  310,  412,   -1,   -1,  276,  277,  316,   -1,
  350,  319,  352,  283,  641,   -1,  643,  644,  645,   -1,
   -1,  329,  362,   -1,  364,  365,  366,   -1,   -1,   -1,
    0,  371,   -1,   -1,   -1,  741,  742,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  458,
  459,  460,  461,  462,  463,  464,  465,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   40,   41,  412,   43,   44,   45,   46,  704,  705,  387,
   -1,   -1,   -1,   -1,   -1,   -1, 3746,   -1,   58,   59,
   -1,   -1,   -1,   63,  364,   -1,  723,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  262,  417,
   -1,   -1,   -1,  421,   -1,  385,   -1,  271,  426,  262,
   -1,  391,   -1,  266,  432,   -1,   -1,   -1,  271,   -1,
  284,   -1,   -1,   -1,  277,   -1,   -1,   -1,  281,   -1,
   -1,  284,   -1,   -1,   -1,   -1,  454,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  424,   -1,   -1,   -1,  567,   -1,
   -1,   -1,   -1,   -1,  318,   -1,   -1,   -1,   -1,   -1,
   -1,  325,   -1,   -1,   -1,  318,   -1,   -1,   -1,  333,
   -1,   -1,  325,   -1,  593,   -1,   -1,   -1, 3848,   -1,
  333,   -1,   -1,  602,   -1,   -1,  350,   -1,  352,   -1,
   -1,   -1,  510,   -1,   -1,   -1,   -1,  350,  362,  352,
  364,  365,  366,   -1,   -1,   -1,   -1,  371,   -1,  362,
   -1,  364,  365,  366,   -1,   -1,   -1,   -1,  371, 3889,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1, 3904,   -1, 3906,   -1,  556,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  412,   -1,
   -1,   -1,  602,   -1,   -1,   -1,   -1,   -1,   -1,  412,
   -1,   -1,   -1,   -1,   -1,   -1, 3936,   -1, 3938, 3939,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  257,   -1,  259,
   -1,   -1,  600,   -1,  264,  265,   -1,  267,  268,  269,
   -1,   -1,   -1,   -1,  274,   -1,   -1,   -1,   -1,  279,
  280,  619,  282,   -1,  284,  285,   -1,  287,   -1,   -1,
  290,  291,  292,   -1,   -1,   -1,   -1,  297,   -1,  299,
  300,   -1,  741,  742,  304,   -1,  306,   -1,  308,  309,
   -1, 4001,  312, 4003, 4004,  315,   -1,  317,   -1,  319,
  320,   -1,  322,  323,  324,   -1,  326,  327,  328,   -1,
   -1,   -1,  332, 4023,  334,  335,  336,   -1,   -1,   -1,
  340,   -1,   -1,   -1,  344,   -1,  346,  347,  348,  349,
   -1,  351,   -1,  353,  354,   -1,  356,  357,  358,   -1,
  360,   -1, 4052,   -1,  364,   -1, 4056,   -1,  368,   -1,
  370,   -1,  372,   -1,  374,  375,  376,  377,  378,   -1,
  380,  381,  382,   -1,   -1,   -1,  386,   -1,  388,  389,
   -1,   -1,  392,  393,  394,   -1,   -1,  397,  736,  399,
   -1,   -1,  402,  403,  404,  405,   -1,  407,   -1,  409,
  410,  411,  412,  413,  414,  415,  416,   -1,  602,   -1,
  420,   -1,  422,  423,   -1,   -1,  426,   -1,   -1,  602,
   -1,   -1,   -1,  433,   -1,  435,  609,  437,  438,  439,
  440,  441,  442,  443,  444,  445,  446,  447,  448,  449,
   -1,  451,  452,  453,  454,   -1,  456,  457,   -1,   -1,
   -1,   -1,   -1,   -1,  464,  465,  466,  467,  468,  469,
  470,  471,  472,  473,  474,  475,  476,   -1,  478,  479,
  480,  481,  482,   -1,  484,  485,  486,  487,  488,  489,
  490,  491,   -1,   -1,  494,  495,  496,  497,  498,  499,
  500,   -1,   -1,  503,  504,  505,  506,  507,  508,  509,
  510,  511,   -1,  513,   -1,  515,  516,  517,   -1,  519,
  520,   -1,  522,  523,  524,  525,  526,  527,  528,  529,
  530,  531,  532,  533,  534,  535,  536,  537,  538,  539,
  540,  541,  542,  543,   -1,  545,  546,  547,  548,  549,
  550,   -1,  552,  553,  554,  555,   -1,  557,   -1,  559,
  560,  561,  562,  563,  564,  565,  566,  567,  568,  569,
  570,  571,  572,  573,  574,  575,  576,   -1,  578,  579,
  580,  581,   -1,  583,  584,  585,  586,  587,   -1,  589,
  590,  591,  592,  593,  594,  595,  596,  597,  598,  599,
  600,  601,   -1,  603,  604,  605,  606,  607,  608,   -1,
  610,  611,  612,  613,  614,  615,  616,  617,   -1,  619,
  620,  621,  622,   -1,  624,  625,  626,  627,  628,  629,
   -1,   -1,  632,  633,  634,  635,  636,  637,  638,  639,
  640,  641,   -1,  643,  644,  645,  646,  647,  648,  649,
  650,  651,  652,  653,   -1,  655,  656,  657,  658,  659,
  660,  661,  662,  663,  664,  665,  666,  667,  668,  669,
  670,  671,  672,  673,   -1,  675,  676,  677,  678,  679,
  680,  681,  682,  683,   -1,  685,  686,  687,  688,  689,
  690,  691,  692,  693,  694,  695,  696,   -1,  698,  699,
  700,   -1,  702,   -1,  704,  705,  706,  707,  708,  709,
  710,  711,  712,  713,  714,  715,  716,  717,  718,    0,
  720,  721,  722,  723,  724,  725,  726,  727,  728,  729,
  730,  731,  732,  733,  734,  735,  736,  737,  738,  739,
  740,   -1,   -1,  743,  744,   -1,   -1,  747,   -1,  749,
  750,  751,  752,  753,  754,   -1,   -1,   -1,   -1,  275,
   41,   42,   43,   44,   45,   -1,   47,   -1,   -1,   -1,
   -1,   -1,   -1,  275,  290,   -1,   -1,   -1,   59,   60,
   61,   62,   -1,   -1,  300,   -1,   -1,   -1,  290,   -1,
  306,  275,   -1,   -1,  310,  311,   -1,   -1,  300,   -1,
  316,   -1,   -1,  319,  306,   -1,  290,   -1,   -1,  311,
   -1,   -1,   93,  329,  316,   -1,  300,  319,   -1,   -1,
   -1,   -1,  306,   -1,   -1,   -1,   -1,  311,   -1,   -1,
   -1,   -1,  316,   -1,   -1,  319,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  379,   -1,   -1,  368,   -1,   -1,   -1,
   -1,  387,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  368,  387,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  417,   -1,  387,   -1,  421,   -1,   -1,   -1,   -1,
  426,   -1,   -1,   -1,   -1,  417,  432,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  432,   -1,   -1,  417,   -1,   -1,   -1,   -1,  454,  455,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  432,   -1,
   -1,   -1,  454,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  454,   -1,   -1,   -1,   -1,   -1,  257,  258,  259,   -1,
   -1,  262,   -1,  264,  265,  266,   -1,  268,  269,   -1,
  271,   -1,   -1,   -1,  510,  276,  277,  513,  279,  280,
  281,  282,  283,  284,   -1,   -1,  287,   -1,  510,  290,
   -1,  513,   -1,   -1,   -1,   -1,  297,   -1,  299,  300,
  301,  302,  303,  304,  305,  306,  510,   -1,  309,  513,
   -1,  312,   -1,  314,  315,   -1,   -1,  318,  319,  320,
   -1,  322,  323,  324,  325,  326,  327,  328,   -1,   -1,
   -1,  332,  333,  334,  335,  336,   -1,   -1,   -1,  340,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  348,  349,  350,
  351,  352,  353,   -1,   -1,   -1,  357,  358,   -1,  360,
   -1,  362,   -1,  364,  365,  366,   -1,  368,   -1,  370,
  371,  372,   -1,  374,  375,  376,  377,  378,   -1,  380,
  381,   -1,   -1,  619,  385,  386,   -1,  388,  389,   -1,
  391,  392,  393,  394,   -1,   -1,  397,  619,  399,   -1,
  401,  402,  403,  404,  405,   -1,  407,   -1,  409,   -1,
  411,  412,  413,  414,  415,  619,   -1,  418,  419,  420,
   -1,  422,  423,  424,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  433,  434,  435,   -1,  437,  438,  439,   -1,
   -1,   -1,   -1,   -1,  445,   -1,   -1,   -1,  449,   -1,
  451,  452,  453,  454,   -1,  456,   -1,  458,  459,  460,
  461,  462,  463,  464,  465,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  475,  476,   -1,  478,  479,  480,
   -1,   -1,   -1,   -1,  485,  486,  487,  488,  489,   -1,
  491,   -1,   -1,  494,  495,   -1,   -1,   -1,  499,  500,
   -1,   -1,  503,  504,  505,  506,  507,  508,  509,  510,
  511,   -1,  513,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  523,  524,  525,  526,  527,  528,  529,  530,
  531,  532,  533,  534,  535,  536,  537,  538,  539,  540,
  541,  542,  543,   -1,  545,  546,  547,  548,  549,  550,
   -1,  552,  553,  554,  555,   -1,  557,   -1,  559,  560,
  561,  562,  563,  564,  565,  566,  567,  568,  569,  570,
  571,  572,  573,  574,  575,  576,   -1,  578,  579,  580,
  581,   -1,  583,  584,  585,  586,  587,   -1,  589,  590,
  591,  592,  593,  594,  595,  596,  597,  598,  599,  600,
  601,  602,  603,  604,  605,  606,  607,  608,   -1,  610,
  611,  612,  613,  614,  615,  616,  617,   -1,  619,  620,
   -1,   -1,   -1,  624,  625,  626,  627,  628,  629,   -1,
   -1,  632,  633,  634,  635,  636,  637,  638,  639,  640,
   41,   -1,   -1,   44,   -1,  646,   -1,  648,  649,  650,
  651,  652,  653,   -1,  655,   -1,   -1,   -1,   -1,   60,
   61,   62,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  672,  673,   -1,  675,  676,  677,  678,  679,  680,
  681,  682,  683,   -1,  685,  686,  687,  688,  689,  690,
  691,  692,  693,  694,  695,  696,   -1,  698,  699,  700,
   -1,  702,   -1,   -1,   -1,  706,  707,  708,  709,  710,
  711,  712,  713,  714,  715,  716,  717,  718,    0,  720,
  721,   -1,   -1,   -1,   -1,   -1,  727,  728,  729,  730,
  731,  732,  733,  734,  735,  736,  737,  738,  739,  740,
  741,  742,  743,  744,   -1,   -1,  747,   -1,  749,  750,
  751,  752,  753,  754,   -1,   -1,   -1,   -1,   -1,   41,
   42,   43,   44,   45,   -1,   47,   -1,   -1,   -1,  275,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   59,   60,   61,
   62,   -1,   -1,   -1,  290,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  300,   -1,   -1,   -1,   -1,   -1,
  306,   -1,   -1,   -1,   -1,  311,   -1,   -1,   -1,   -1,
  316,   93,   -1,  319,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  275,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  290,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  300,
   -1,   -1,   -1,   -1,   -1,  306,   -1,   -1,   -1,   -1,
  311,  262,   -1,   -1,   -1,  316,   -1,   -1,  319,   -1,
  271,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  387,   -1,  284,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  417,   -1,   -1,   -1,   -1,   -1,  318,   -1,   -1,
   -1,   -1,   -1,   -1,  325,   -1,  432,   -1,   -1,   -1,
   -1,   -1,  333,   -1,   -1,   -1,  387,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  454,  350,
   -1,  352,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  362,   -1,  364,  365,  366,  417,   -1,   -1,   -1,
  371,   -1,   -1,   -1,   -1,  257,  258,  259,   -1,   -1,
  262,  432,  264,  265,  266,   -1,  268,  269,   -1,  271,
   -1,   -1,   -1,   -1,  276,  277,   -1,  279,  280,  281,
  282,  283,  284,  454,  510,  287,   -1,  513,  290,   -1,
   -1,  412,   -1,   -1,   -1,  297,   -1,  299,  300,  301,
  302,  303,  304,  305,  306,   -1,   -1,  309,   -1,   -1,
  312,   -1,  314,  315,   -1,   -1,  318,  319,  320,   -1,
  322,  323,  324,  325,  326,  327,  328,   -1,   -1,   -1,
  332,  333,  334,  335,  336,   -1,   -1,   -1,  340,  510,
   -1,   -1,  513,   -1,   -1,   -1,  348,  349,  350,  351,
  352,  353,   -1,   -1,   -1,  357,  358,   -1,  360,   -1,
  362,   -1,  364,  365,  366,   -1,  368,   -1,  370,  371,
  372,   -1,  374,  375,  376,  377,  378,   -1,  380,  381,
   -1,   -1,   -1,  385,  386,   -1,  388,  389,   -1,  391,
  392,  393,  394,  619,   -1,  397,   -1,  399,   -1,  401,
  402,  403,  404,  405,   -1,  407,   -1,  409,   -1,  411,
  412,  413,  414,  415,   -1,   -1,  418,  419,  420,   -1,
  422,  423,  424,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  433,  434,  435,   -1,  437,  438,  439,   -1,   -1,
   -1,   -1,   -1,  445,   -1,   -1,   -1,  449,  619,  451,
  452,  453,  454,   -1,  456,   -1,  458,  459,  460,  461,
  462,  463,  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  475,  476,   -1,  478,  479,  480,   -1,
   -1,  602,   -1,  485,  486,  487,  488,  489,   -1,  491,
   -1,   -1,  494,  495,   -1,   -1,   -1,  499,  500,   -1,
   -1,  503,  504,  505,  506,  507,  508,  509,  510,  511,
   -1,  513,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  523,  524,  525,  526,  527,  528,  529,  530,  531,
  532,  533,  534,  535,  536,  537,  538,  539,  540,  541,
  542,  543,   -1,  545,  546,  547,  548,  549,  550,   -1,
  552,  553,  554,  555,   -1,  557,   -1,  559,  560,  561,
  562,  563,  564,  565,  566,  567,  568,  569,  570,  571,
  572,  573,  574,  575,  576,   -1,  578,  579,  580,  581,
   -1,  583,  584,  585,  586,  587,   -1,  589,  590,  591,
  592,  593,  594,  595,  596,  597,  598,  599,  600,  601,
  602,  603,  604,  605,  606,  607,  608,   -1,  610,  611,
  612,  613,  614,  615,  616,  617,   -1,  619,  620,   -1,
   -1,   -1,  624,  625,  626,  627,  628,  629,   -1,   -1,
  632,  633,  634,  635,  636,  637,  638,  639,  640,   41,
   -1,   -1,   44,   -1,  646,   -1,  648,  649,  650,  651,
  652,  653,   -1,  655,   -1,   -1,   -1,   -1,   60,   61,
   62,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  672,  673,   -1,  675,  676,  677,  678,  679,  680,  681,
  682,  683,   -1,  685,  686,  687,  688,  689,  690,  691,
  692,  693,  694,  695,  696,   -1,  698,  699,  700,   -1,
  702,   -1,   -1,   -1,  706,  707,  708,  709,  710,  711,
  712,  713,  714,  715,  716,  717,  718,    0,  720,  721,
   -1,   -1,   -1,   -1,   -1,  727,  728,  729,  730,  731,
  732,  733,  734,  735,  736,  737,  738,  739,  740,  741,
  742,  743,  744,   -1,   -1,  747,   -1,  749,  750,  751,
  752,  753,  754,   -1,   -1,  275,   -1,   40,   41,   42,
   43,   44,   45,   46,   47,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   58,   59,   60,   61,   62,
  300,   -1,   -1,   -1,   -1,   -1,  306,   -1,   -1,   -1,
   -1,  311,   -1,   -1,   -1,   -1,  316,   -1,   -1,  319,
   -1,   -1,  261,   -1,   -1,   -1,   -1,   -1,   91,   -1,
   93,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  278,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  286,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  294,   -1,  296,   -1,   -1,
   -1,   -1,  301,   -1,   -1,   -1,   -1,   -1,  307,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  262,   -1,  321,   -1,   -1,   -1,   -1,  387,   -1,  271,
   -1,  330,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  284,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  355,  417,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  432,   -1,   -1,   -1,  318,   -1,   -1,   -1,
   -1,   -1,   -1,  325,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  333,   -1,   -1,  454,   -1,   -1,  396,   -1,  398,
   -1,   -1,  401,   -1,   -1,   -1,   -1,   -1,  350,   -1,
  352,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  362,   -1,  364,  365,  366,   -1,  425,   -1,   -1,  371,
   -1,   -1,   -1,   -1,  257,   -1,  259,   -1,  437,  262,
   -1,  264,  265,  266,   -1,  268,  269,   -1,  271,   -1,
  510,   -1,   -1,  513,  277,   -1,  279,  280,  281,  282,
   -1,  284,   -1,   -1,  287,   -1,   -1,  290,   -1,   -1,
  412,   -1,   -1,   -1,  297,   -1,  299,  300,  477,  302,
  303,  304,  305,  306,   -1,   -1,  309,   -1,   -1,  312,
   -1,  314,  315,  492,  493,  318,  319,  320,   -1,  322,
  323,  324,  325,  326,  327,  328,   -1,   -1,   -1,  332,
  333,  334,  335,  336,   -1,  514,   -1,  340,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  348,  349,  350,  351,  352,
  353,   -1,   -1,   -1,  357,  358,   -1,  360,   -1,  362,
   -1,  364,  365,  366,   -1,  368,   -1,  370,  371,  372,
   -1,  374,  375,  376,  377,  378,   -1,  380,  381,  619,
   -1,   -1,   -1,  386,   -1,  388,  389,   -1,   -1,  392,
  393,  394,   -1,   -1,  397,   -1,  399,   -1,   -1,  402,
  403,  404,  405,   -1,  407,   -1,  409,   -1,  411,  412,
  413,  414,  415,   -1,   -1,  418,  419,  420,   -1,  422,
  423,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  433,  434,  435,   -1,  437,  438,  439,   -1,   -1,   -1,
   -1,   -1,  445,   -1,   -1,   -1,  449,   -1,  451,  452,
  453,  454,   -1,  456,   -1,  458,  459,  460,  461,  462,
  463,  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  475,  476,   -1,  478,  479,  480,   -1,   -1,
  602,   -1,  485,  486,  487,  488,  489,   -1,  491,   -1,
   -1,  494,  495,   -1,   -1,   -1,  499,  500,   -1,   -1,
  503,  504,  505,  506,  507,  508,  509,  510,  511,   -1,
  513,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  523,  524,  525,  526,  527,  528,  529,  530,  531,  532,
  533,  534,  535,  536,  537,  538,  539,  540,  541,  542,
  543,   -1,  545,  546,  547,  548,  549,  550,   -1,  552,
  553,  554,  555,   -1,  557,   -1,  559,  560,  561,  562,
  563,  564,  565,  566,  567,  568,  569,  570,  571,  572,
  573,  574,  575,  576,   -1,  578,  579,  580,  581,   -1,
  583,  584,  585,  586,  587,   -1,  589,  590,  591,  592,
  593,  594,  595,  596,  597,  598,  599,  600,  601,  602,
  603,  604,  605,  606,  607,  608,  609,  610,  611,  612,
  613,  614,  615,  616,  617,   -1,  619,  620,   -1,   -1,
   -1,  624,  625,  626,  627,  628,  629,   -1,   -1,  632,
  633,  634,  635,  636,  637,  638,  639,  640,   41,   -1,
   -1,   44,   -1,  646,   -1,  648,  649,  650,  651,  652,
  653,   -1,  655,   -1,   -1,   -1,   -1,   60,   61,   62,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  672,
  673,   -1,  675,  676,  677,  678,  679,  680,  681,  682,
  683,   -1,  685,  686,  687,  688,  689,  690,  691,  692,
  693,  694,  695,  696,   -1,  698,  699,  700,   -1,  702,
   -1,   -1,   -1,  706,  707,  708,  709,  710,  711,  712,
  713,  714,  715,  716,  717,  718,    0,  720,  721,   -1,
   -1,   -1,   -1,   -1,  727,  728,  729,  730,  731,  732,
  733,  734,  735,  736,  737,  738,  739,  740,  741,  742,
  743,  744,   -1,   -1,  747,   -1,  749,  750,  751,  752,
  753,  754,   -1,   -1,   -1,   -1,   40,   41,   42,   43,
   44,   45,   46,   47,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   58,   59,   60,   61,   62,  300,
   -1,   -1,   -1,   -1,   -1,  306,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  316,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   91,  329,   93,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  262,
   -1,   -1,   -1,   -1,   -1,   -1,  387,   -1,  271,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  284,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  417,   -1,   -1,   -1,
  421,   -1,   -1,   -1,   -1,  426,   -1,   -1,   -1,   -1,
   -1,  432,   -1,   -1,   -1,  318,   -1,   -1,   -1,   -1,
   -1,   -1,  325,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  333,   -1,   -1,  454,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  350,   -1,  352,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  362,
   -1,  364,  365,  366,   -1,   -1,   -1,   -1,  371,   -1,
   -1,   -1,   -1,  257,   -1,  259,   -1,   -1,  262,   -1,
  264,  265,  266,   -1,  268,  269,   -1,  271,   -1,  510,
   -1,   -1,  513,  277,   -1,  279,  280,  281,  282,   -1,
  284,   -1,   -1,  287,   -1,   -1,  290,   -1,   -1,  412,
   -1,   -1,   -1,  297,   -1,  299,  300,   -1,  302,  303,
  304,  305,  306,   -1,   -1,  309,   -1,   -1,  312,   -1,
  314,  315,   -1,   -1,  318,  319,  320,   -1,  322,  323,
  324,  325,  326,  327,  328,   -1,   -1,   -1,  332,  333,
  334,  335,  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  348,  349,  350,  351,  352,  353,
   -1,   -1,   -1,  357,  358,   -1,  360,   -1,  362,  600,
  364,  365,  366,   -1,  368,   -1,  370,  371,  372,   -1,
  374,  375,  376,  377,  378,   -1,  380,  381,  619,   -1,
   -1,   -1,  386,   -1,  388,  389,   -1,   -1,  392,  393,
  394,   -1,   -1,  397,   -1,  399,   -1,   -1,  402,  403,
  404,  405,   -1,  407,   -1,  409,   -1,  411,  412,  413,
  414,  415,   -1,   -1,  418,  419,  420,   -1,  422,  423,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  433,
  434,  435,   -1,  437,  438,  439,   -1,   -1,   -1,   -1,
   -1,  445,   -1,   -1,   -1,  449,   -1,  451,  452,  453,
  454,   -1,  456,   -1,  458,  459,  460,  461,  462,  463,
  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  475,  476,   -1,  478,  479,  480,   -1,   -1,  602,
   -1,  485,  486,  487,  488,  489,   -1,  491,   -1,   -1,
  494,  495,   -1,   -1,   -1,  499,  500,   -1,   -1,  503,
  504,  505,  506,  507,  508,  509,  510,  511,   -1,  513,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  523,
  524,  525,  526,  527,  528,  529,  530,  531,  532,  533,
  534,  535,  536,  537,  538,  539,  540,  541,  542,  543,
   -1,  545,  546,  547,  548,  549,  550,   -1,  552,  553,
  554,  555,   -1,  557,   -1,  559,  560,  561,  562,  563,
  564,  565,  566,  567,  568,  569,  570,  571,  572,  573,
  574,  575,  576,   -1,  578,  579,  580,  581,   -1,  583,
  584,  585,  586,  587,   -1,  589,  590,  591,  592,  593,
  594,  595,  596,  597,  598,  599,  600,  601,  602,  603,
  604,  605,  606,  607,  608,  609,  610,  611,  612,  613,
  614,  615,  616,  617,   -1,  619,  620,   -1,   -1,   -1,
  624,  625,  626,  627,  628,  629,   -1,   -1,  632,  633,
  634,  635,  636,  637,  638,  639,  640,   41,   -1,   -1,
   44,   -1,  646,   -1,  648,  649,  650,  651,  652,  653,
   -1,  655,   -1,   -1,   -1,   -1,   60,   61,   62,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  672,  673,
   -1,  675,  676,  677,  678,  679,  680,  681,  682,  683,
   -1,  685,  686,  687,  688,  689,  690,  691,  692,  693,
  694,  695,  696,   -1,  698,  699,  700,   -1,  702,   -1,
   -1,   -1,  706,  707,  708,  709,  710,  711,  712,  713,
  714,  715,  716,  717,  718,    0,  720,  721,   -1,   -1,
   -1,   -1,   -1,  727,  728,  729,  730,  731,  732,  733,
  734,  735,  736,  737,  738,  739,  740,  741,  742,  743,
  744,   -1,   -1,  747,   -1,  749,  750,  751,  752,  753,
  754,   -1,   -1,   -1,   -1,   40,   41,   42,   43,   44,
   45,   46,   47,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   58,   59,   60,   61,   62,  300,   -1,
   -1,   -1,   -1,   -1,  306,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  316,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   91,  329,   93,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  262,   -1,
   -1,   -1,   -1,   -1,   -1,  387,   -1,  271,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  284,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  417,   -1,   -1,   -1,  421,
   -1,   -1,   -1,   -1,  426,   -1,   -1,   -1,   -1,   -1,
  432,   -1,   -1,   -1,  318,   -1,   -1,   -1,   -1,   -1,
   -1,  325,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  333,
   -1,   -1,  454,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  350,   -1,  352,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  362,   -1,
  364,  365,  366,   -1,   -1,   -1,   -1,  371,   -1,   -1,
   -1,   -1,  257,   -1,  259,   -1,   -1,  262,   -1,  264,
  265,  266,   -1,  268,  269,   -1,  271,   -1,  510,   -1,
   -1,  513,  277,   -1,  279,  280,  281,  282,   -1,  284,
   -1,   -1,  287,   -1,   -1,  290,   -1,   -1,  412,   -1,
   -1,   -1,  297,   -1,  299,  300,   -1,  302,  303,  304,
  305,  306,   -1,   -1,  309,   -1,   -1,  312,   -1,  314,
  315,   -1,   -1,  318,  319,  320,   -1,  322,  323,  324,
  325,  326,  327,  328,   -1,   -1,   -1,  332,  333,  334,
  335,  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  348,  349,  350,  351,  352,  353,   -1,
   -1,   -1,  357,  358,   -1,  360,   -1,  362,  600,  364,
  365,  366,   -1,  368,   -1,  370,  371,  372,   -1,  374,
  375,  376,  377,  378,   -1,  380,  381,  619,   -1,   -1,
   -1,  386,   -1,  388,  389,   -1,   -1,  392,  393,  394,
   -1,   -1,  397,   -1,  399,   -1,   -1,  402,  403,  404,
  405,   -1,  407,   -1,  409,   -1,  411,  412,  413,  414,
  415,   -1,   -1,  418,  419,  420,   -1,  422,  423,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  433,  434,
  435,   -1,  437,  438,  439,   -1,   -1,   -1,   -1,   -1,
  445,   -1,   -1,   -1,  449,   -1,  451,  452,  453,  454,
   -1,  456,   -1,  458,  459,  460,  461,  462,  463,  464,
  465,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  475,  476,   -1,  478,  479,  480,   -1,   -1,  602,   -1,
  485,  486,  487,  488,  489,   -1,  491,   -1,   -1,  494,
  495,   -1,   -1,   -1,  499,  500,   -1,   -1,  503,  504,
  505,  506,  507,  508,  509,  510,  511,   -1,  513,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  523,  524,
  525,  526,  527,  528,  529,  530,  531,  532,  533,  534,
  535,  536,  537,  538,  539,  540,  541,  542,  543,   -1,
  545,  546,  547,  548,  549,  550,   -1,  552,  553,  554,
  555,   -1,  557,   -1,  559,  560,  561,  562,  563,  564,
  565,  566,  567,  568,  569,  570,  571,  572,  573,  574,
  575,  576,   -1,  578,  579,  580,  581,   -1,  583,  584,
  585,  586,  587,   -1,  589,  590,  591,  592,  593,  594,
  595,  596,  597,  598,  599,  600,  601,  602,  603,  604,
  605,  606,  607,  608,  609,  610,  611,  612,  613,  614,
  615,  616,  617,   -1,  619,  620,   -1,   -1,   -1,  624,
  625,  626,  627,  628,  629,   -1,   -1,  632,  633,  634,
  635,  636,  637,  638,  639,  640,   41,   -1,   -1,   -1,
   -1,  646,   -1,  648,  649,  650,  651,  652,  653,   -1,
  655,   -1,   -1,   -1,   -1,   60,   61,   62,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  672,  673,   -1,
  675,  676,  677,  678,  679,  680,  681,  682,  683,   -1,
  685,  686,  687,  688,  689,  690,  691,  692,  693,  694,
  695,  696,   -1,  698,  699,  700,   -1,  702,   -1,   -1,
   -1,  706,  707,  708,  709,  710,  711,  712,  713,  714,
  715,  716,  717,  718,    0,  720,  721,   -1,   -1,   -1,
   -1,   -1,  727,  728,  729,  730,  731,  732,  733,  734,
  735,  736,  737,  738,  739,  740,  741,  742,  743,  744,
   -1,   -1,  747,   -1,  749,  750,  751,  752,  753,  754,
   -1,   -1,   -1,   -1,   40,   41,   42,   43,   44,   45,
   46,   47,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   58,   59,   60,   61,   62,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   91,   -1,   93,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  262,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  271,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  284,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  312,   -1,   -1,
   -1,   -1,   -1,  318,   -1,   -1,   -1,   -1,   -1,   -1,
  325,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  333,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  350,   -1,  352,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  362,   -1,  364,
  365,  366,   -1,   -1,   -1,   -1,  371,   -1,   -1,   -1,
   -1,  257,   -1,  259,   -1,   -1,  262,   -1,  264,  265,
  266,   -1,  268,  269,   -1,  271,   -1,   -1,   -1,   -1,
   -1,  277,   -1,  279,  280,  281,  282,   -1,  284,   -1,
   -1,  287,   -1,   -1,  290,   -1,   -1,  412,   -1,   -1,
   -1,  297,   -1,  299,  300,   -1,  302,  303,  304,  305,
  306,   -1,   -1,  309,   -1,   -1,  312,   -1,  314,  315,
   -1,   -1,  318,  319,  320,   -1,  322,  323,  324,  325,
  326,  327,  328,   -1,   -1,   -1,  332,  333,  334,  335,
  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  348,  349,  350,  351,  352,  353,   -1,   -1,
   -1,  357,  358,   -1,  360,   -1,  362,   -1,  364,  365,
  366,   -1,  368,   -1,  370,  371,  372,   -1,  374,  375,
  376,  377,  378,   -1,  380,  381,   -1,   -1,   -1,   -1,
  386,   -1,  388,  389,   -1,   -1,  392,  393,  394,   -1,
   -1,  397,   -1,  399,   -1,   -1,  402,  403,  404,  405,
   -1,  407,   -1,  409,   -1,  411,  412,  413,  414,  415,
   -1,   -1,  418,  419,  420,   -1,  422,  423,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  433,  434,  435,
   -1,  437,  438,  439,   -1,   -1,   -1,   -1,   -1,  445,
   -1,   -1,   -1,  449,   -1,  451,  452,  453,  454,   -1,
  456,   -1,  458,  459,  460,  461,  462,  463,  464,  465,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  475,
  476,   -1,  478,  479,  480,   -1,   -1,  602,   -1,  485,
  486,  487,  488,  489,   -1,  491,   -1,   -1,  494,  495,
   -1,   -1,   -1,  499,  500,   -1,   -1,  503,  504,  505,
  506,  507,  508,  509,  510,  511,   -1,  513,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  523,  524,  525,
  526,  527,  528,  529,  530,  531,  532,  533,  534,  535,
  536,  537,  538,  539,  540,  541,  542,  543,   -1,  545,
  546,  547,  548,  549,  550,   -1,  552,  553,  554,  555,
   -1,  557,   -1,  559,  560,  561,  562,  563,  564,  565,
  566,  567,  568,  569,  570,  571,  572,  573,  574,  575,
  576,   -1,  578,  579,  580,  581,   -1,  583,  584,  585,
  586,  587,   -1,  589,  590,  591,  592,  593,  594,  595,
  596,  597,  598,  599,  600,  601,  602,  603,  604,  605,
  606,  607,  608,  609,  610,  611,  612,  613,  614,  615,
  616,  617,   -1,  619,  620,   -1,   -1,   -1,  624,  625,
  626,  627,  628,  629,   -1,   -1,  632,  633,  634,  635,
  636,  637,  638,  639,  640,   41,   -1,   -1,   -1,   -1,
  646,   -1,  648,  649,  650,  651,  652,  653,   -1,  655,
   -1,   -1,   -1,   -1,   60,   61,   62,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  672,  673,   -1,  675,
  676,  677,  678,  679,  680,  681,  682,  683,   -1,  685,
  686,  687,  688,  689,  690,  691,  692,  693,  694,  695,
  696,   -1,  698,  699,  700,   -1,  702,   -1,   -1,   -1,
  706,  707,  708,  709,  710,  711,  712,  713,  714,  715,
  716,  717,  718,    0,  720,  721,   -1,   -1,   -1,   -1,
   -1,  727,  728,  729,  730,  731,  732,  733,  734,  735,
  736,  737,  738,  739,  740,  741,  742,  743,  744,   -1,
   -1,  747,   -1,  749,  750,  751,  752,  753,  754,   -1,
   -1,   -1,   -1,   40,   41,   42,   43,   44,   45,   46,
   47,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   58,   59,   60,   61,   62,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   91,   -1,   93,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  262,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  271,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  284,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  318,   -1,   -1,   -1,   -1,   -1,   -1,  325,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  333,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  350,   -1,  352,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  362,   -1,  364,  365,
  366,   -1,   -1,   -1,   -1,  371,   -1,   -1,   -1,   -1,
  257,   -1,  259,   -1,   -1,  262,   -1,  264,  265,  266,
   -1,  268,  269,   -1,  271,   -1,   -1,   -1,   -1,   -1,
  277,   -1,  279,  280,  281,  282,   -1,  284,   -1,   -1,
  287,   -1,   -1,  290,   -1,   -1,  412,   -1,   -1,   -1,
  297,   -1,  299,  300,   -1,  302,  303,  304,  305,  306,
   -1,   -1,  309,   -1,   -1,  312,   -1,  314,  315,   -1,
   -1,  318,  319,  320,   -1,  322,  323,  324,  325,  326,
  327,  328,   -1,   -1,   -1,  332,  333,  334,  335,  336,
   -1,   -1,   -1,  340,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  348,  349,  350,  351,  352,  353,   -1,   -1,   -1,
  357,  358,   -1,  360,   -1,  362,   -1,  364,  365,  366,
   -1,  368,   -1,  370,  371,  372,   -1,  374,  375,  376,
  377,  378,   -1,  380,  381,   -1,   -1,   -1,   -1,  386,
   -1,  388,  389,   -1,   -1,  392,  393,  394,   -1,   -1,
  397,   -1,  399,   -1,   -1,  402,  403,  404,  405,   -1,
  407,   -1,  409,   -1,  411,  412,  413,  414,  415,   -1,
   -1,  418,  419,  420,   -1,  422,  423,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  433,  434,  435,   -1,
  437,  438,  439,   -1,   -1,   -1,   -1,   -1,  445,   -1,
   -1,   -1,  449,   -1,  451,  452,  453,  454,   -1,  456,
   -1,  458,  459,  460,  461,  462,  463,  464,  465,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  475,  476,
   -1,  478,  479,  480,   -1,   -1,  602,   -1,  485,  486,
  487,  488,  489,   -1,  491,   -1,   -1,  494,  495,   -1,
   -1,   -1,  499,  500,   -1,   -1,  503,  504,  505,  506,
  507,  508,  509,  510,  511,   -1,  513,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  523,  524,  525,  526,
  527,  528,  529,  530,  531,  532,  533,  534,  535,  536,
  537,  538,  539,  540,  541,  542,  543,   -1,  545,  546,
  547,  548,  549,  550,   -1,  552,  553,  554,  555,   -1,
  557,   -1,  559,  560,  561,  562,  563,  564,  565,  566,
  567,  568,  569,  570,  571,  572,  573,  574,  575,  576,
   -1,  578,  579,  580,  581,   -1,  583,  584,  585,  586,
  587,   -1,  589,  590,  591,  592,  593,  594,  595,  596,
  597,  598,  599,  600,  601,  602,  603,  604,  605,  606,
  607,  608,  609,  610,  611,  612,  613,  614,  615,  616,
  617,   -1,  619,  620,   -1,   -1,   -1,  624,  625,  626,
  627,  628,  629,   -1,   -1,  632,  633,  634,  635,  636,
  637,  638,  639,  640,   41,   -1,   -1,   -1,   -1,  646,
   -1,  648,  649,  650,  651,  652,  653,   -1,  655,   -1,
   -1,   -1,   -1,   60,   61,   62,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  672,  673,   -1,  675,  676,
  677,  678,  679,  680,  681,  682,  683,   -1,  685,  686,
  687,  688,  689,  690,  691,  692,  693,  694,  695,  696,
   -1,  698,  699,  700,   -1,  702,   -1,   -1,   -1,  706,
  707,  708,  709,  710,  711,  712,  713,  714,  715,  716,
  717,  718,    0,  720,  721,   -1,   -1,   -1,   -1,   -1,
  727,  728,  729,  730,  731,  732,  733,  734,  735,  736,
  737,  738,  739,  740,  741,  742,  743,  744,   -1,   -1,
  747,   -1,  749,  750,  751,  752,  753,  754,   -1,   -1,
   -1,   -1,   40,   41,   42,   43,   44,   45,   46,   47,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   58,   59,   60,   61,   62,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   91,   -1,   93,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  262,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  271,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  284,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  318,   -1,   -1,   -1,   -1,   -1,   -1,  325,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  333,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  350,   -1,  352,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  362,   -1,  364,  365,  366,
   -1,   -1,   -1,   -1,  371,   -1,   -1,   -1,   -1,  257,
   -1,  259,   -1,   -1,  262,   -1,  264,  265,  266,   -1,
  268,  269,   -1,  271,   -1,   -1,   -1,   -1,   -1,  277,
   -1,  279,  280,  281,  282,   -1,  284,   -1,   -1,  287,
   -1,   -1,  290,   -1,   -1,  412,   -1,   -1,   -1,  297,
   -1,  299,  300,   -1,  302,  303,  304,  305,  306,   -1,
   -1,  309,   -1,   -1,  312,   -1,  314,  315,   -1,   -1,
  318,  319,  320,   -1,  322,  323,  324,  325,  326,  327,
  328,   -1,   -1,   -1,  332,  333,  334,  335,  336,   -1,
   -1,   -1,  340,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  348,  349,  350,  351,  352,  353,   -1,   -1,   -1,  357,
  358,   -1,  360,   -1,  362,   -1,  364,  365,  366,   -1,
  368,   -1,  370,  371,  372,   -1,  374,  375,  376,  377,
  378,   -1,  380,  381,   -1,   -1,   -1,   -1,  386,   -1,
  388,  389,   -1,   -1,  392,  393,  394,   -1,   -1,  397,
   -1,  399,   -1,   -1,  402,  403,  404,  405,   -1,  407,
   -1,  409,   -1,  411,  412,  413,  414,  415,   -1,   -1,
  418,  419,  420,   -1,  422,  423,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  433,  434,  435,   -1,  437,
  438,  439,   -1,   -1,   -1,   -1,   -1,  445,   -1,   -1,
   -1,  449,   -1,  451,  452,  453,  454,   -1,  456,   -1,
  458,  459,  460,  461,  462,  463,  464,  465,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  475,  476,   -1,
  478,  479,  480,   -1,   -1,  602,   -1,  485,  486,  487,
  488,  489,   -1,  491,   -1,   -1,  494,  495,   -1,   -1,
   -1,  499,  500,   -1,   -1,  503,  504,  505,  506,  507,
  508,  509,  510,  511,   -1,  513,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  523,  524,  525,  526,  527,
  528,  529,  530,  531,  532,  533,  534,  535,  536,  537,
  538,  539,  540,  541,  542,  543,   -1,  545,  546,  547,
  548,  549,  550,   -1,  552,  553,  554,  555,   -1,  557,
   -1,  559,  560,  561,  562,  563,  564,  565,  566,  567,
  568,  569,  570,  571,  572,  573,  574,  575,  576,   -1,
  578,  579,  580,  581,   -1,  583,  584,  585,  586,  587,
   -1,  589,  590,  591,  592,  593,  594,  595,  596,  597,
  598,  599,  600,  601,  602,  603,  604,  605,  606,  607,
  608,  609,  610,  611,  612,  613,  614,  615,  616,  617,
   -1,  619,  620,   -1,   -1,   -1,  624,  625,  626,  627,
  628,  629,   -1,   -1,  632,  633,  634,  635,  636,  637,
  638,  639,  640,   41,   -1,   -1,   -1,   -1,  646,   -1,
  648,  649,  650,  651,  652,  653,   -1,  655,   -1,   -1,
   -1,   -1,   60,   61,   62,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  672,  673,   -1,  675,  676,  677,
  678,  679,  680,  681,  682,  683,   -1,  685,  686,  687,
  688,  689,  690,  691,  692,  693,  694,  695,  696,   -1,
  698,  699,  700,   -1,  702,   -1,   -1,   -1,  706,  707,
  708,  709,  710,  711,  712,  713,  714,  715,  716,  717,
  718,    0,  720,  721,   -1,   -1,   -1,   -1,   -1,  727,
  728,  729,  730,  731,  732,  733,  734,  735,  736,  737,
  738,  739,  740,  741,  742,  743,  744,   -1,   -1,  747,
   -1,  749,  750,  751,  752,  753,  754,   -1,   -1,   -1,
   -1,   40,   41,   42,   43,   44,   45,   46,   47,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   58,
   59,   60,   61,   62,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   91,   -1,   93,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  262,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  271,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  284,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  318,   -1,   -1,   -1,   -1,   -1,   -1,  325,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  333,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  350,   -1,  352,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  362,   -1,  364,  365,  366,   -1,
   -1,   -1,   -1,  371,   -1,   -1,   -1,   -1,  257,   -1,
  259,   -1,   -1,  262,   -1,  264,  265,  266,   -1,  268,
  269,   -1,  271,   -1,   -1,   -1,   -1,   -1,  277,   -1,
  279,  280,  281,  282,   -1,  284,   -1,   -1,  287,   -1,
   -1,  290,   -1,   -1,  412,   -1,   -1,   -1,  297,   -1,
  299,  300,   -1,  302,  303,  304,  305,  306,   -1,   -1,
  309,   -1,   -1,  312,   -1,  314,  315,   -1,   -1,  318,
  319,  320,   -1,  322,  323,  324,  325,  326,  327,  328,
   -1,   -1,   -1,  332,  333,  334,  335,  336,   -1,   -1,
   -1,  340,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  348,
  349,  350,  351,  352,  353,   -1,   -1,   -1,  357,  358,
   -1,  360,   -1,  362,   -1,  364,  365,  366,   -1,  368,
   -1,  370,  371,  372,   -1,  374,  375,  376,  377,  378,
   -1,  380,  381,   -1,   -1,   -1,   -1,  386,   -1,  388,
  389,   -1,   -1,  392,  393,  394,   -1,   -1,  397,   -1,
  399,   -1,   -1,  402,  403,  404,  405,   -1,  407,   -1,
  409,   -1,  411,  412,  413,  414,  415,   -1,   -1,  418,
  419,  420,   -1,  422,  423,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  433,  434,  435,   -1,  437,  438,
  439,   -1,   -1,   -1,   -1,   -1,  445,   -1,   -1,   -1,
  449,   -1,  451,  452,  453,  454,   -1,  456,   -1,  458,
  459,  460,  461,  462,  463,  464,  465,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  475,  476,   -1,  478,
  479,  480,   -1,   -1,  602,   -1,  485,  486,  487,  488,
  489,   -1,  491,   -1,   -1,  494,  495,   -1,   -1,   -1,
  499,  500,   -1,   -1,  503,  504,  505,  506,  507,  508,
  509,  510,  511,   -1,  513,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  523,  524,  525,  526,  527,  528,
  529,  530,  531,  532,  533,  534,  535,  536,  537,  538,
  539,  540,  541,  542,  543,   -1,  545,  546,  547,  548,
  549,  550,   -1,  552,  553,  554,  555,   -1,  557,   -1,
  559,  560,  561,  562,  563,  564,  565,  566,  567,  568,
  569,  570,  571,  572,  573,  574,  575,  576,   -1,  578,
  579,  580,  581,   -1,  583,  584,  585,  586,  587,   -1,
  589,  590,  591,  592,  593,  594,  595,  596,  597,  598,
  599,  600,  601,  602,  603,  604,  605,  606,  607,  608,
  609,  610,  611,  612,  613,  614,  615,  616,  617,   -1,
  619,  620,   -1,   -1,   -1,  624,  625,  626,  627,  628,
  629,   -1,   -1,  632,  633,  634,  635,  636,  637,  638,
  639,  640,   41,   -1,   -1,   -1,   -1,  646,   -1,  648,
  649,  650,  651,  652,  653,   -1,  655,   -1,   -1,   -1,
   -1,   60,   61,   62,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  672,  673,   -1,  675,  676,  677,  678,
  679,  680,  681,  682,  683,   -1,  685,  686,  687,  688,
  689,  690,  691,  692,  693,  694,  695,  696,   -1,  698,
  699,  700,   -1,  702,   -1,   -1,   -1,  706,  707,  708,
  709,  710,  711,  712,  713,  714,  715,  716,  717,  718,
    0,  720,  721,   -1,   -1,   -1,   -1,   -1,  727,  728,
  729,  730,  731,  732,  733,  734,  735,  736,  737,  738,
  739,  740,  741,  742,  743,  744,   -1,   -1,  747,   -1,
  749,  750,  751,  752,  753,  754,   -1,   -1,   -1,   -1,
   40,   41,   42,   43,   44,   45,   46,   47,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   58,   59,
   60,   61,   62,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   91,   -1,   93,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  262,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  271,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  284,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  318,
   -1,   -1,   -1,   -1,   -1,   -1,  325,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  333,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  350,   -1,  352,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  362,   -1,  364,  365,  366,   -1,   -1,
   -1,   -1,  371,   -1,   -1,   -1,   -1,  257,   -1,  259,
   -1,   -1,  262,   -1,  264,  265,  266,   -1,  268,  269,
   -1,  271,   -1,   -1,   -1,   -1,   -1,  277,   -1,  279,
  280,  281,  282,   -1,  284,   -1,   -1,  287,   -1,   -1,
  290,   -1,   -1,  412,   -1,   -1,   -1,  297,   -1,  299,
  300,   -1,  302,  303,  304,  305,  306,   -1,   -1,  309,
   -1,   -1,  312,   -1,  314,  315,   -1,   -1,  318,  319,
  320,   -1,  322,  323,  324,  325,  326,  327,  328,   -1,
   -1,   -1,  332,  333,  334,  335,  336,   -1,   -1,   -1,
  340,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  348,  349,
  350,  351,  352,  353,   -1,   -1,   -1,  357,  358,   -1,
  360,   -1,  362,   -1,  364,  365,  366,   -1,  368,   -1,
  370,  371,  372,   -1,  374,  375,  376,  377,  378,   -1,
  380,  381,   -1,   -1,   -1,   -1,  386,   -1,  388,  389,
   -1,   -1,  392,  393,  394,   -1,   -1,  397,   -1,  399,
   -1,   -1,  402,  403,  404,  405,   -1,  407,   -1,  409,
   -1,  411,  412,  413,  414,  415,   -1,   -1,  418,  419,
  420,   -1,  422,  423,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  433,  434,  435,   -1,  437,  438,  439,
   -1,   -1,   -1,   -1,   -1,  445,   -1,   -1,   -1,  449,
   -1,  451,  452,  453,  454,   -1,  456,   -1,  458,  459,
  460,  461,  462,  463,  464,  465,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  475,  476,   -1,  478,  479,
  480,   -1,   -1,  602,   -1,  485,  486,  487,  488,  489,
   -1,  491,   -1,   -1,  494,  495,   -1,   -1,   -1,  499,
  500,   -1,   -1,  503,  504,  505,  506,  507,  508,  509,
  510,  511,   -1,  513,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  523,  524,  525,  526,  527,  528,  529,
  530,  531,  532,  533,  534,  535,  536,  537,  538,  539,
  540,  541,  542,  543,   -1,  545,  546,  547,  548,  549,
  550,   -1,  552,  553,  554,  555,   -1,  557,   -1,  559,
  560,  561,  562,  563,  564,  565,  566,  567,  568,  569,
  570,  571,  572,  573,  574,  575,  576,   -1,  578,  579,
  580,  581,   -1,  583,  584,  585,  586,  587,   -1,  589,
  590,  591,  592,  593,  594,  595,  596,  597,  598,  599,
  600,  601,  602,  603,  604,  605,  606,  607,  608,  609,
  610,  611,  612,  613,  614,  615,  616,  617,   -1,  619,
  620,   -1,   -1,   -1,  624,  625,  626,  627,  628,  629,
   -1,   -1,  632,  633,  634,  635,  636,  637,  638,  639,
  640,   41,   -1,   -1,   -1,   -1,  646,   -1,  648,  649,
  650,  651,  652,  653,   -1,  655,   -1,   -1,   -1,   -1,
   60,   61,   62,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  672,  673,   -1,  675,  676,  677,  678,  679,
  680,  681,  682,  683,   -1,  685,  686,  687,  688,  689,
  690,  691,  692,  693,  694,  695,  696,   -1,  698,  699,
  700,   -1,  702,   -1,   -1,   -1,  706,  707,  708,  709,
  710,  711,  712,  713,  714,  715,  716,  717,  718,    0,
  720,  721,   -1,   -1,   -1,   -1,   -1,  727,  728,  729,
  730,  731,  732,  733,  734,  735,  736,  737,  738,  739,
  740,  741,  742,  743,  744,   -1,   -1,  747,   -1,  749,
  750,  751,  752,  753,  754,   -1,   -1,   -1,   -1,   40,
   41,   42,   43,   44,   45,   46,   47,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   58,   59,   60,
   61,   62,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   91,   -1,   93,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  262,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  271,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  284,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  318,   -1,
   -1,   -1,   -1,   -1,   -1,  325,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  333,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  350,   -1,  352,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  362,   -1,  364,  365,  366,   -1,   -1,   -1,
   -1,  371,   -1,   -1,   -1,   -1,  257,   -1,  259,   -1,
   -1,  262,   -1,  264,  265,  266,   -1,  268,  269,   -1,
  271,   -1,   -1,   -1,   -1,   -1,  277,   -1,  279,  280,
  281,  282,   -1,  284,   -1,   -1,  287,   -1,   -1,  290,
   -1,   -1,  412,   -1,   -1,   -1,  297,   -1,  299,  300,
   -1,  302,  303,  304,  305,  306,   -1,   -1,  309,   -1,
   -1,  312,   -1,  314,  315,   -1,   -1,  318,  319,  320,
   -1,  322,  323,  324,  325,  326,  327,  328,   -1,   -1,
   -1,  332,  333,  334,  335,  336,   -1,   -1,   -1,  340,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  348,  349,  350,
  351,  352,  353,   -1,   -1,   -1,  357,  358,   -1,  360,
   -1,  362,   -1,  364,  365,  366,   -1,  368,   -1,  370,
  371,  372,   -1,  374,  375,  376,  377,  378,   -1,  380,
  381,   -1,   -1,   -1,   -1,  386,   -1,  388,  389,   -1,
   -1,  392,  393,  394,   -1,   -1,  397,   -1,  399,   -1,
   -1,  402,  403,  404,  405,   -1,  407,   -1,  409,   -1,
  411,  412,  413,  414,  415,   -1,   -1,  418,  419,  420,
   -1,  422,  423,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  433,  434,  435,   -1,  437,  438,  439,   -1,
   -1,   -1,   -1,   -1,  445,   -1,   -1,   -1,  449,   -1,
  451,  452,  453,  454,   -1,  456,   -1,  458,  459,  460,
  461,  462,  463,  464,  465,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  475,  476,   -1,  478,  479,  480,
   -1,   -1,  602,   -1,  485,  486,  487,  488,  489,   -1,
  491,   -1,   -1,  494,  495,   -1,   -1,   -1,  499,  500,
   -1,   -1,  503,  504,  505,  506,  507,  508,  509,  510,
  511,   -1,  513,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  523,  524,  525,  526,  527,  528,  529,  530,
  531,  532,  533,  534,  535,  536,  537,  538,  539,  540,
  541,  542,  543,   -1,  545,  546,  547,  548,  549,  550,
   -1,  552,  553,  554,  555,   -1,  557,   -1,  559,  560,
  561,  562,  563,  564,  565,  566,  567,  568,  569,  570,
  571,  572,  573,  574,  575,  576,   -1,  578,  579,  580,
  581,   -1,  583,  584,  585,  586,  587,   -1,  589,  590,
  591,  592,  593,  594,  595,  596,  597,  598,  599,  600,
  601,  602,  603,  604,  605,  606,  607,  608,  609,  610,
  611,  612,  613,  614,  615,  616,  617,   -1,  619,  620,
   -1,   -1,   -1,  624,  625,  626,  627,  628,  629,   -1,
   -1,  632,  633,  634,  635,  636,  637,  638,  639,  640,
   41,   -1,   -1,   -1,   -1,  646,   -1,  648,  649,  650,
  651,  652,  653,   -1,  655,   -1,   -1,   -1,   -1,   60,
   61,   62,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  672,  673,   -1,  675,  676,  677,  678,  679,  680,
  681,  682,  683,   -1,  685,  686,  687,  688,  689,  690,
  691,  692,  693,  694,  695,  696,   -1,  698,  699,  700,
   -1,  702,   -1,   -1,   -1,  706,  707,  708,  709,  710,
  711,  712,  713,  714,  715,  716,  717,  718,    0,  720,
  721,   -1,   -1,   -1,   -1,   -1,  727,  728,  729,  730,
  731,  732,  733,  734,  735,  736,  737,  738,  739,  740,
  741,  742,  743,  744,   -1,   -1,  747,   -1,  749,  750,
  751,  752,  753,  754,   -1,   -1,   -1,   -1,   40,   41,
   42,   43,   44,   45,   46,   47,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   58,   59,   60,   61,
   62,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   91,
   -1,   93,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  262,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  271,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  284,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  318,   -1,   -1,
   -1,   -1,   -1,   -1,  325,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  333,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  350,
   -1,  352,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  362,   -1,  364,  365,  366,   -1,   -1,   -1,   -1,
  371,   -1,   -1,   -1,   -1,  257,   -1,  259,   -1,   -1,
  262,   -1,  264,  265,  266,   -1,  268,  269,   -1,  271,
   -1,   -1,   -1,   -1,   -1,  277,   -1,  279,  280,  281,
  282,   -1,  284,   -1,   -1,  287,   -1,   -1,  290,   -1,
   -1,  412,   -1,   -1,   -1,  297,   -1,  299,  300,   -1,
  302,  303,  304,  305,  306,   -1,   -1,  309,   -1,   -1,
  312,   -1,  314,  315,   -1,   -1,  318,  319,  320,   -1,
  322,  323,  324,  325,  326,  327,  328,   -1,   -1,   -1,
  332,  333,  334,  335,  336,   -1,   -1,   -1,  340,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  348,  349,  350,  351,
  352,  353,   -1,   -1,   -1,  357,  358,   -1,  360,   -1,
  362,   -1,  364,  365,  366,   -1,  368,   -1,  370,  371,
  372,   -1,  374,  375,  376,  377,  378,   -1,  380,  381,
   -1,   -1,   -1,   -1,  386,   -1,  388,  389,   -1,   -1,
  392,  393,  394,   -1,   -1,  397,   -1,  399,   -1,   -1,
  402,  403,  404,  405,   -1,  407,   -1,  409,   -1,  411,
  412,  413,  414,  415,   -1,   -1,  418,  419,  420,   -1,
  422,  423,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  433,  434,  435,   -1,  437,  438,  439,   -1,   -1,
   -1,   -1,   -1,  445,   -1,   -1,   -1,  449,   -1,  451,
  452,  453,  454,   -1,  456,   -1,  458,  459,  460,  461,
  462,  463,  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  475,  476,   -1,  478,  479,  480,   -1,
   -1,  602,   -1,  485,  486,  487,  488,  489,   -1,  491,
   -1,   -1,  494,  495,   -1,   -1,   -1,  499,  500,   -1,
   -1,  503,  504,  505,  506,  507,  508,  509,  510,  511,
   -1,  513,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  523,  524,  525,  526,  527,  528,  529,  530,  531,
  532,  533,  534,  535,  536,  537,  538,  539,  540,  541,
  542,  543,   -1,  545,  546,  547,  548,  549,  550,   -1,
  552,  553,  554,  555,   -1,  557,   -1,  559,  560,  561,
  562,  563,  564,  565,  566,  567,  568,  569,  570,  571,
  572,  573,  574,  575,  576,   -1,  578,  579,  580,  581,
   -1,  583,  584,  585,  586,  587,   -1,  589,  590,  591,
  592,  593,  594,  595,  596,  597,  598,  599,  600,  601,
  602,  603,  604,  605,  606,  607,  608,  609,  610,  611,
  612,  613,  614,  615,  616,  617,   -1,  619,  620,   -1,
   -1,   -1,  624,  625,  626,  627,  628,  629,   -1,   -1,
  632,  633,  634,  635,  636,  637,  638,  639,  640,   41,
   -1,   -1,   -1,   -1,  646,   -1,  648,  649,  650,  651,
  652,  653,   -1,  655,   -1,   -1,   -1,   -1,   60,   61,
   62,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  672,  673,   -1,  675,  676,  677,  678,  679,  680,  681,
  682,  683,   -1,  685,  686,  687,  688,  689,  690,  691,
  692,  693,  694,  695,  696,   -1,  698,  699,  700,   -1,
  702,   -1,   -1,   -1,  706,  707,  708,  709,  710,  711,
  712,  713,  714,  715,  716,  717,  718,    0,  720,  721,
   -1,   -1,   -1,   -1,   -1,  727,  728,  729,  730,  731,
  732,  733,  734,  735,  736,  737,  738,  739,  740,  741,
  742,  743,  744,   -1,   -1,  747,   -1,  749,  750,  751,
  752,  753,  754,   -1,   -1,   -1,   -1,   40,   41,   42,
   43,   44,   45,   46,   47,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   58,   59,   60,   61,   62,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   91,   -1,
   93,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  262,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  271,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  284,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  318,   -1,   -1,   -1,
   -1,   -1,   -1,  325,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  333,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  350,   -1,
  352,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  362,   -1,  364,  365,  366,   -1,   -1,   -1,   -1,  371,
   -1,   -1,   -1,   -1,  257,   -1,  259,   -1,   -1,  262,
   -1,  264,  265,  266,   -1,  268,  269,   -1,  271,   -1,
   -1,   -1,   -1,   -1,  277,   -1,  279,  280,  281,  282,
   -1,  284,   -1,   -1,  287,   -1,   -1,  290,   -1,   -1,
  412,   -1,   -1,   -1,  297,   -1,  299,  300,   -1,  302,
  303,  304,  305,  306,   -1,   -1,  309,   -1,   -1,  312,
   -1,  314,  315,   -1,   -1,  318,  319,  320,   -1,  322,
  323,  324,  325,  326,  327,  328,   -1,   -1,   -1,  332,
  333,  334,  335,  336,   -1,   -1,   -1,  340,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  348,  349,  350,  351,  352,
  353,   -1,   -1,   -1,  357,  358,   -1,  360,   -1,  362,
   -1,  364,  365,  366,   -1,  368,   -1,  370,  371,  372,
   -1,  374,  375,  376,  377,  378,   -1,  380,  381,   -1,
   -1,   -1,   -1,  386,   -1,  388,  389,   -1,   -1,  392,
  393,  394,   -1,   -1,  397,   -1,  399,   -1,   -1,  402,
  403,  404,  405,   -1,  407,   -1,  409,   -1,  411,  412,
  413,  414,  415,   -1,   -1,  418,  419,  420,   -1,  422,
  423,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  433,  434,  435,   -1,  437,  438,  439,   -1,   -1,   -1,
   -1,   -1,  445,   -1,   -1,   -1,  449,   -1,  451,  452,
  453,  454,   -1,  456,   -1,  458,  459,  460,  461,  462,
  463,  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  475,  476,   -1,  478,  479,  480,   -1,   -1,
  602,   -1,  485,  486,  487,  488,  489,   -1,  491,   -1,
   -1,  494,  495,   -1,   -1,   -1,  499,  500,   -1,   -1,
  503,  504,  505,  506,  507,  508,  509,  510,  511,   -1,
  513,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  523,  524,  525,  526,  527,  528,  529,  530,  531,  532,
  533,  534,  535,  536,  537,  538,  539,  540,  541,  542,
  543,   -1,  545,  546,  547,  548,  549,  550,   -1,  552,
  553,  554,  555,   -1,  557,   -1,  559,  560,  561,  562,
  563,  564,  565,  566,  567,  568,  569,  570,  571,  572,
  573,  574,  575,  576,   -1,  578,  579,  580,  581,   -1,
  583,  584,  585,  586,  587,   -1,  589,  590,  591,  592,
  593,  594,  595,  596,  597,  598,  599,  600,  601,  602,
  603,  604,  605,  606,  607,  608,  609,  610,  611,  612,
  613,  614,  615,  616,  617,   -1,  619,  620,   -1,   -1,
   -1,  624,  625,  626,  627,  628,  629,   -1,   -1,  632,
  633,  634,  635,  636,  637,  638,  639,  640,   41,   -1,
   -1,   -1,   -1,  646,   -1,  648,  649,  650,  651,  652,
  653,   -1,  655,   -1,   -1,   -1,   -1,   60,   61,   62,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  672,
  673,   -1,  675,  676,  677,  678,  679,  680,  681,  682,
  683,   -1,  685,  686,  687,  688,  689,  690,  691,  692,
  693,  694,  695,  696,   -1,  698,  699,  700,   -1,  702,
   -1,   -1,   -1,  706,  707,  708,  709,  710,  711,  712,
  713,  714,  715,  716,  717,  718,    0,  720,  721,   -1,
   -1,   -1,   -1,   -1,  727,  728,  729,  730,  731,  732,
  733,  734,  735,  736,  737,  738,  739,  740,  741,  742,
  743,  744,   -1,   -1,  747,   -1,  749,  750,  751,  752,
  753,  754,   -1,   -1,   -1,   -1,   40,   41,   42,   43,
   44,   45,   46,   47,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   58,   59,   60,   61,   62,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   91,   -1,   93,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  262,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  271,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  284,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  318,   -1,   -1,   -1,   -1,
   -1,   -1,  325,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  333,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  350,   -1,  352,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  362,
   -1,  364,  365,  366,   -1,   -1,   -1,   -1,  371,   -1,
   -1,   -1,   -1,  257,   -1,  259,   -1,   -1,  262,   -1,
  264,  265,  266,   -1,  268,  269,   -1,  271,   -1,   -1,
   -1,   -1,   -1,  277,   -1,  279,  280,  281,  282,   -1,
  284,   -1,   -1,  287,   -1,   -1,  290,   -1,   -1,  412,
   -1,   -1,   -1,  297,   -1,  299,  300,   -1,  302,  303,
  304,  305,  306,   -1,   -1,  309,   -1,   -1,  312,   -1,
  314,  315,   -1,   -1,  318,  319,  320,   -1,  322,  323,
  324,  325,  326,  327,  328,   -1,   -1,   -1,  332,  333,
  334,  335,  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  348,  349,  350,  351,  352,  353,
   -1,   -1,   -1,  357,  358,   -1,  360,   -1,  362,   -1,
  364,  365,  366,   -1,  368,   -1,  370,  371,  372,   -1,
  374,  375,  376,  377,  378,   -1,  380,  381,   -1,   -1,
   -1,   -1,  386,   -1,  388,  389,   -1,   -1,  392,  393,
  394,   -1,   -1,  397,   -1,  399,   -1,   -1,  402,  403,
  404,  405,   -1,  407,   -1,  409,   -1,  411,  412,  413,
  414,  415,   -1,   -1,  418,  419,  420,   -1,  422,  423,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  433,
  434,  435,   -1,  437,  438,  439,   -1,   -1,   -1,   -1,
   -1,  445,   -1,   -1,   -1,  449,   -1,  451,  452,  453,
  454,   -1,  456,   -1,  458,  459,  460,  461,  462,  463,
  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  475,  476,   -1,  478,  479,  480,   -1,   -1,  602,
   -1,  485,  486,  487,  488,  489,   -1,  491,   -1,   -1,
  494,  495,   -1,   -1,   -1,  499,  500,   -1,   -1,  503,
  504,  505,  506,  507,  508,  509,  510,  511,   -1,  513,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  523,
  524,  525,  526,  527,  528,  529,  530,  531,  532,  533,
  534,  535,  536,  537,  538,  539,  540,  541,  542,  543,
   -1,  545,  546,  547,  548,  549,  550,   -1,  552,  553,
  554,  555,   -1,  557,   -1,  559,  560,  561,  562,  563,
  564,  565,  566,  567,  568,  569,  570,  571,  572,  573,
  574,  575,  576,   -1,  578,  579,  580,  581,   -1,  583,
  584,  585,  586,  587,   -1,  589,  590,  591,  592,  593,
  594,  595,  596,  597,  598,  599,  600,  601,  602,  603,
  604,  605,  606,  607,  608,  609,  610,  611,  612,  613,
  614,  615,  616,  617,   -1,  619,  620,   -1,   -1,   -1,
  624,  625,  626,  627,  628,  629,   -1,   -1,  632,  633,
  634,  635,  636,  637,  638,  639,  640,   41,   -1,   -1,
   -1,   -1,  646,   -1,  648,  649,  650,  651,  652,  653,
   -1,  655,   -1,   -1,   -1,   -1,   60,   61,   62,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  672,  673,
   -1,  675,  676,  677,  678,  679,  680,  681,  682,  683,
   -1,  685,  686,  687,  688,  689,  690,  691,  692,  693,
  694,  695,  696,   -1,  698,  699,  700,   -1,  702,   -1,
   -1,   -1,  706,  707,  708,  709,  710,  711,  712,  713,
  714,  715,  716,  717,  718,    0,  720,  721,   -1,   -1,
   -1,   -1,   -1,  727,  728,  729,  730,  731,  732,  733,
  734,  735,  736,  737,  738,  739,  740,  741,  742,  743,
  744,   -1,   -1,  747,   -1,  749,  750,  751,  752,  753,
  754,   -1,   -1,   -1,   -1,   40,   41,   42,   43,   44,
   45,   46,   47,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   58,   59,   60,   61,   62,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   91,   -1,   93,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  262,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  271,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  284,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  318,   -1,   -1,   -1,   -1,   -1,
   -1,  325,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  333,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  350,   -1,  352,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  362,   -1,
  364,  365,  366,   -1,   -1,   -1,   -1,  371,   -1,   -1,
   -1,   -1,  257,   -1,  259,   -1,   -1,  262,   -1,  264,
  265,  266,   -1,  268,  269,   -1,  271,   -1,   -1,   -1,
   -1,   -1,  277,   -1,  279,  280,  281,  282,   -1,  284,
   -1,   -1,  287,   -1,   -1,  290,   -1,   -1,  412,   -1,
   -1,   -1,  297,   -1,  299,  300,   -1,  302,  303,  304,
  305,  306,   -1,   -1,  309,   -1,   -1,  312,   -1,  314,
  315,   -1,   -1,  318,  319,  320,   -1,  322,  323,  324,
  325,  326,  327,  328,   -1,   -1,   -1,  332,  333,  334,
  335,  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  348,  349,  350,  351,  352,  353,   -1,
   -1,   -1,  357,  358,   -1,  360,   -1,  362,   -1,  364,
  365,  366,   -1,  368,   -1,  370,  371,  372,   -1,  374,
  375,  376,  377,  378,   -1,  380,  381,   -1,   -1,   -1,
   -1,  386,   -1,  388,  389,   -1,   -1,  392,  393,  394,
   -1,   -1,  397,   -1,  399,   -1,   -1,  402,  403,  404,
  405,   -1,  407,   -1,  409,   -1,  411,  412,  413,  414,
  415,   -1,   -1,  418,  419,  420,   -1,  422,  423,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  433,  434,
  435,   -1,  437,  438,  439,   -1,   -1,   -1,   -1,   -1,
  445,   -1,   -1,   -1,  449,   -1,  451,  452,  453,  454,
   -1,  456,   -1,  458,  459,  460,  461,  462,  463,  464,
  465,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  475,  476,   -1,  478,  479,  480,   -1,   -1,  602,   -1,
  485,  486,  487,  488,  489,   -1,  491,   -1,   -1,  494,
  495,   -1,   -1,   -1,  499,  500,   -1,   -1,  503,  504,
  505,  506,  507,  508,  509,  510,  511,   -1,  513,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  523,  524,
  525,  526,  527,  528,  529,  530,  531,  532,  533,  534,
  535,  536,  537,  538,  539,  540,  541,  542,  543,   -1,
  545,  546,  547,  548,  549,  550,   -1,  552,  553,  554,
  555,   -1,  557,   -1,  559,  560,  561,  562,  563,  564,
  565,  566,  567,  568,  569,  570,  571,  572,  573,  574,
  575,  576,   -1,  578,  579,  580,  581,   -1,  583,  584,
  585,  586,  587,   -1,  589,  590,  591,  592,  593,  594,
  595,  596,  597,  598,  599,  600,  601,  602,  603,  604,
  605,  606,  607,  608,  609,  610,  611,  612,  613,  614,
  615,  616,  617,   -1,  619,  620,   -1,   -1,   -1,  624,
  625,  626,  627,  628,  629,   -1,   -1,  632,  633,  634,
  635,  636,  637,  638,  639,  640,   41,   -1,   -1,   -1,
   -1,  646,   -1,  648,  649,  650,  651,  652,  653,   -1,
  655,   -1,   -1,   -1,   -1,   60,   61,   62,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  672,  673,   -1,
  675,  676,  677,  678,  679,  680,  681,  682,  683,   -1,
  685,  686,  687,  688,  689,  690,  691,  692,  693,  694,
  695,  696,   -1,  698,  699,  700,   -1,  702,   -1,   -1,
   -1,  706,  707,  708,  709,  710,  711,  712,  713,  714,
  715,  716,  717,  718,    0,  720,  721,   -1,   -1,   -1,
   -1,   -1,  727,  728,  729,  730,  731,  732,  733,  734,
  735,  736,  737,  738,  739,  740,  741,  742,  743,  744,
   -1,   -1,  747,   -1,  749,  750,  751,  752,  753,  754,
   -1,   -1,   -1,   -1,   40,   41,   42,   43,   44,   45,
   46,   47,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   58,   59,   60,   61,   62,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   91,   -1,   93,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  262,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  271,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  284,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  318,   -1,   -1,   -1,   -1,   -1,   -1,
  325,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  333,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  350,   -1,  352,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  362,   -1,  364,
  365,  366,   -1,   -1,   -1,   -1,  371,   -1,   -1,   -1,
   -1,  257,   -1,  259,   -1,   -1,  262,   -1,  264,  265,
  266,   -1,  268,  269,   -1,  271,   -1,   -1,   -1,   -1,
   -1,  277,   -1,  279,  280,  281,  282,   -1,  284,   -1,
   -1,  287,   -1,   -1,  290,   -1,   -1,  412,   -1,   -1,
   -1,  297,   -1,  299,  300,   -1,  302,  303,  304,  305,
  306,   -1,   -1,  309,   -1,   -1,  312,   -1,  314,  315,
   -1,   -1,  318,  319,  320,   -1,  322,  323,  324,  325,
  326,  327,  328,   -1,   -1,   -1,  332,  333,  334,  335,
  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  348,  349,  350,  351,  352,  353,   -1,   -1,
   -1,  357,  358,   -1,  360,   -1,  362,   -1,  364,  365,
  366,   -1,  368,   -1,  370,  371,  372,   -1,  374,  375,
  376,  377,  378,   -1,  380,  381,   -1,   -1,   -1,   -1,
  386,   -1,  388,  389,   -1,   -1,  392,  393,  394,   -1,
   -1,  397,   -1,  399,   -1,   -1,  402,  403,  404,  405,
   -1,  407,   -1,  409,   -1,  411,  412,  413,  414,  415,
   -1,   -1,  418,  419,  420,   -1,  422,  423,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  433,  434,  435,
   -1,  437,  438,  439,   -1,   -1,   -1,   -1,   -1,  445,
   -1,   -1,   -1,  449,   -1,  451,  452,  453,  454,   -1,
  456,   -1,  458,  459,  460,  461,  462,  463,  464,  465,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  475,
  476,   -1,  478,  479,  480,   -1,   -1,  602,   -1,  485,
  486,  487,  488,  489,   -1,  491,   -1,   -1,  494,  495,
   -1,   -1,   -1,  499,  500,   -1,   -1,  503,  504,  505,
  506,  507,  508,  509,  510,  511,   -1,  513,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  523,  524,  525,
  526,  527,  528,  529,  530,  531,  532,  533,  534,  535,
  536,  537,  538,  539,  540,  541,  542,  543,   -1,  545,
  546,  547,  548,  549,  550,   -1,  552,  553,  554,  555,
   -1,  557,   -1,  559,  560,  561,  562,  563,  564,  565,
  566,  567,  568,  569,  570,  571,  572,  573,  574,  575,
  576,   -1,  578,  579,  580,  581,   -1,  583,  584,  585,
  586,  587,   -1,  589,  590,  591,  592,  593,  594,  595,
  596,  597,  598,  599,  600,  601,  602,  603,  604,  605,
  606,  607,  608,  609,  610,  611,  612,  613,  614,  615,
  616,  617,   -1,  619,  620,   -1,   -1,   -1,  624,  625,
  626,  627,  628,  629,   -1,   -1,  632,  633,  634,  635,
  636,  637,  638,  639,  640,   41,   -1,   -1,   -1,   -1,
  646,   -1,  648,  649,  650,  651,  652,  653,   -1,  655,
   -1,   -1,   -1,   -1,   60,   61,   62,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  672,  673,   -1,  675,
  676,  677,  678,  679,  680,  681,  682,  683,   -1,  685,
  686,  687,  688,  689,  690,  691,  692,  693,  694,  695,
  696,   -1,  698,  699,  700,   -1,  702,   -1,   -1,   -1,
  706,  707,  708,  709,  710,  711,  712,  713,  714,  715,
  716,  717,  718,    0,  720,  721,   -1,   -1,   -1,   -1,
   -1,  727,  728,  729,  730,  731,  732,  733,  734,  735,
  736,  737,  738,  739,  740,  741,  742,  743,  744,   -1,
   -1,  747,   -1,  749,  750,  751,  752,  753,  754,   -1,
   -1,   -1,   -1,   40,   41,   42,   43,   44,   45,   46,
   47,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   58,   59,   60,   61,   62,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   91,   -1,   93,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  262,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  271,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  284,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  318,   -1,   -1,   -1,   -1,   -1,   -1,  325,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  333,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  350,   -1,  352,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  362,   -1,  364,  365,
  366,   -1,   -1,   -1,   -1,  371,   -1,   -1,   -1,   -1,
  257,   -1,  259,   -1,   -1,  262,   -1,  264,  265,  266,
   -1,  268,  269,   -1,  271,   -1,   -1,   -1,   -1,   -1,
  277,   -1,  279,  280,  281,  282,   -1,  284,   -1,   -1,
  287,   -1,   -1,  290,   -1,   -1,  412,   -1,   -1,   -1,
  297,   -1,  299,  300,   -1,  302,  303,  304,  305,  306,
   -1,   -1,  309,   -1,   -1,  312,   -1,  314,  315,   -1,
   -1,  318,  319,  320,   -1,  322,  323,  324,  325,  326,
  327,  328,   -1,   -1,   -1,  332,  333,  334,  335,  336,
   -1,   -1,   -1,  340,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  348,  349,  350,  351,  352,  353,   -1,   -1,   -1,
  357,  358,   -1,  360,   -1,  362,   -1,  364,  365,  366,
   -1,  368,   -1,  370,  371,  372,   -1,  374,  375,  376,
  377,  378,   -1,  380,  381,   -1,   -1,   -1,   -1,  386,
   -1,  388,  389,   -1,   -1,  392,  393,  394,   -1,   -1,
  397,   -1,  399,   -1,   -1,  402,  403,  404,  405,   -1,
  407,   -1,  409,   -1,  411,  412,  413,  414,  415,   -1,
   -1,  418,  419,  420,   -1,  422,  423,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  433,  434,  435,   -1,
  437,  438,  439,   -1,   -1,   -1,   -1,   -1,  445,   -1,
   -1,   -1,  449,   -1,  451,  452,  453,  454,   -1,  456,
   -1,  458,  459,  460,  461,  462,  463,  464,  465,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  475,  476,
   -1,  478,  479,  480,   -1,   -1,  602,   -1,  485,  486,
  487,  488,  489,   -1,  491,   -1,   -1,  494,  495,   -1,
   -1,   -1,  499,  500,   -1,   -1,  503,  504,  505,  506,
  507,  508,  509,  510,  511,   -1,  513,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  523,  524,  525,  526,
  527,  528,  529,  530,  531,  532,  533,  534,  535,  536,
  537,  538,  539,  540,  541,  542,  543,   -1,  545,  546,
  547,  548,  549,  550,   -1,  552,  553,  554,  555,   -1,
  557,   -1,  559,  560,  561,  562,  563,  564,  565,  566,
  567,  568,  569,  570,  571,  572,  573,  574,  575,  576,
   -1,  578,  579,  580,  581,   -1,  583,  584,  585,  586,
  587,   -1,  589,  590,  591,  592,  593,  594,  595,  596,
  597,  598,  599,  600,  601,  602,  603,  604,  605,  606,
  607,  608,  609,  610,  611,  612,  613,  614,  615,  616,
  617,   -1,  619,  620,   -1,   -1,   -1,  624,  625,  626,
  627,  628,  629,   -1,   -1,  632,  633,  634,  635,  636,
  637,  638,  639,  640,   41,   -1,   -1,   -1,   -1,  646,
   -1,  648,  649,  650,  651,  652,  653,   -1,  655,   -1,
   -1,   -1,   -1,   60,   61,   62,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  672,  673,   -1,  675,  676,
  677,  678,  679,  680,  681,  682,  683,   -1,  685,  686,
  687,  688,  689,  690,  691,  692,  693,  694,  695,  696,
   -1,  698,  699,  700,   -1,  702,   -1,   -1,   -1,  706,
  707,  708,  709,  710,  711,  712,  713,  714,  715,  716,
  717,  718,    0,  720,  721,   -1,   -1,   -1,   -1,   -1,
  727,  728,  729,  730,  731,  732,  733,  734,  735,  736,
  737,  738,  739,  740,  741,  742,  743,  744,   -1,   -1,
  747,   -1,  749,  750,  751,  752,  753,  754,   -1,   -1,
   -1,   -1,   40,   41,   42,   43,   44,   45,   46,   47,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   58,   59,   60,   61,   62,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   91,   -1,   93,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  262,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  271,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  284,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  318,   -1,   -1,   -1,   -1,   -1,   -1,  325,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  333,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  350,   -1,  352,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  362,   -1,  364,  365,  366,
   -1,   -1,   -1,   -1,  371,   -1,   -1,   -1,   -1,  257,
   -1,  259,   -1,   -1,  262,   -1,  264,  265,  266,   -1,
  268,  269,   -1,  271,   -1,   -1,   -1,   -1,   -1,  277,
   -1,  279,  280,  281,  282,   -1,  284,   -1,   -1,  287,
   -1,   -1,  290,   -1,   -1,  412,   -1,   -1,   -1,  297,
   -1,  299,  300,   -1,  302,  303,  304,  305,  306,   -1,
   -1,  309,   -1,   -1,  312,   -1,  314,  315,   -1,   -1,
  318,  319,  320,   -1,  322,  323,  324,  325,  326,  327,
  328,   -1,   -1,   -1,  332,  333,  334,  335,  336,   -1,
   -1,   -1,  340,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  348,  349,  350,  351,  352,  353,   -1,   -1,   -1,  357,
  358,   -1,  360,   -1,  362,   -1,  364,  365,  366,   -1,
  368,   -1,  370,  371,  372,   -1,  374,  375,  376,  377,
  378,   -1,  380,  381,   -1,   -1,   -1,   -1,  386,   -1,
  388,  389,   -1,   -1,  392,  393,  394,   -1,   -1,  397,
   -1,  399,   -1,   -1,  402,  403,  404,  405,   -1,  407,
   -1,  409,   -1,  411,  412,  413,  414,  415,   -1,   -1,
  418,  419,  420,   -1,  422,  423,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  433,  434,  435,   -1,  437,
  438,  439,   -1,   -1,   -1,   -1,   -1,  445,   -1,   -1,
   -1,  449,   -1,  451,  452,  453,  454,   -1,  456,   -1,
  458,  459,  460,  461,  462,  463,  464,  465,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  475,  476,   -1,
  478,  479,  480,   -1,   -1,  602,   -1,  485,  486,  487,
  488,  489,   -1,  491,   -1,   -1,  494,  495,   -1,   -1,
   -1,  499,  500,   -1,   -1,  503,  504,  505,  506,  507,
  508,  509,  510,  511,   -1,  513,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  523,  524,  525,  526,  527,
  528,  529,  530,  531,  532,  533,  534,  535,  536,  537,
  538,  539,  540,  541,  542,  543,   -1,  545,  546,  547,
  548,  549,  550,   -1,  552,  553,  554,  555,   -1,  557,
   -1,  559,  560,  561,  562,  563,  564,  565,  566,  567,
  568,  569,  570,  571,  572,  573,  574,  575,  576,   -1,
  578,  579,  580,  581,   -1,  583,  584,  585,  586,  587,
   -1,  589,  590,  591,  592,  593,  594,  595,  596,  597,
  598,  599,  600,  601,  602,  603,  604,  605,  606,  607,
  608,  609,  610,  611,  612,  613,  614,  615,  616,  617,
   -1,  619,  620,   -1,   -1,   -1,  624,  625,  626,  627,
  628,  629,   -1,   -1,  632,  633,  634,  635,  636,  637,
  638,  639,  640,   41,   -1,   -1,   -1,   -1,  646,   -1,
  648,  649,  650,  651,  652,  653,   -1,  655,   -1,   -1,
   -1,   -1,   60,   61,   62,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  672,  673,   -1,  675,  676,  677,
  678,  679,  680,  681,  682,  683,   -1,  685,  686,  687,
  688,  689,  690,  691,  692,  693,  694,  695,  696,   -1,
  698,  699,  700,   -1,  702,   -1,   -1,   -1,  706,  707,
  708,  709,  710,  711,  712,  713,  714,  715,  716,  717,
  718,    0,  720,  721,   -1,   -1,   -1,   -1,   -1,  727,
  728,  729,  730,  731,  732,  733,  734,  735,  736,  737,
  738,  739,  740,  741,  742,  743,  744,   -1,   -1,  747,
   -1,  749,  750,  751,  752,  753,  754,   -1,   -1,   -1,
   -1,   40,   41,   42,   43,   44,   45,   46,   47,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   58,
   59,   60,   61,   62,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   91,   -1,   93,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  262,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  271,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  284,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  318,   -1,   -1,   -1,   -1,   -1,   -1,  325,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  333,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  350,   -1,  352,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  362,   -1,  364,  365,  366,   -1,
   -1,   -1,   -1,  371,   -1,   -1,   -1,   -1,  257,   -1,
  259,   -1,   -1,  262,   -1,  264,  265,  266,   -1,  268,
  269,   -1,  271,   -1,   -1,   -1,   -1,   -1,  277,   -1,
  279,  280,  281,  282,   -1,  284,   -1,   -1,  287,   -1,
   -1,  290,   -1,   -1,  412,   -1,   -1,   -1,  297,   -1,
  299,  300,   -1,  302,  303,  304,  305,  306,   -1,   -1,
  309,   -1,   -1,  312,   -1,  314,  315,   -1,   -1,  318,
  319,  320,   -1,  322,  323,  324,  325,  326,  327,  328,
   -1,   -1,   -1,  332,  333,  334,  335,  336,   -1,   -1,
   -1,  340,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  348,
  349,  350,  351,  352,  353,   -1,   -1,   -1,  357,  358,
   -1,  360,   -1,  362,   -1,  364,  365,  366,   -1,  368,
   -1,  370,  371,  372,   -1,  374,  375,  376,  377,  378,
   -1,  380,  381,   -1,   -1,   -1,   -1,  386,   -1,  388,
  389,   -1,   -1,  392,  393,  394,   -1,   -1,  397,   -1,
  399,   -1,   -1,  402,  403,  404,  405,   -1,  407,   -1,
  409,   -1,  411,  412,  413,  414,  415,   -1,   -1,  418,
  419,  420,   -1,  422,  423,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  433,  434,  435,   -1,  437,  438,
  439,   -1,   -1,   -1,   -1,   -1,  445,   -1,   -1,   -1,
  449,   -1,  451,  452,  453,  454,   -1,  456,   -1,  458,
  459,  460,  461,  462,  463,  464,  465,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  475,  476,   -1,  478,
  479,  480,   -1,   -1,  602,   -1,  485,  486,  487,  488,
  489,   -1,  491,   -1,   -1,  494,  495,   -1,   -1,   -1,
  499,  500,   -1,   -1,  503,  504,  505,  506,  507,  508,
  509,  510,  511,   -1,  513,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  523,  524,  525,  526,  527,  528,
  529,  530,  531,  532,  533,  534,  535,  536,  537,  538,
  539,  540,  541,  542,  543,   -1,  545,  546,  547,  548,
  549,  550,   -1,  552,  553,  554,  555,   -1,  557,   -1,
  559,  560,  561,  562,  563,  564,  565,  566,  567,  568,
  569,  570,  571,  572,  573,  574,  575,  576,   -1,  578,
  579,  580,  581,   -1,  583,  584,  585,  586,  587,   -1,
  589,  590,  591,  592,  593,  594,  595,  596,  597,  598,
  599,  600,  601,  602,  603,  604,  605,  606,  607,  608,
  609,  610,  611,  612,  613,  614,  615,  616,  617,   -1,
  619,  620,   -1,   -1,   -1,  624,  625,  626,  627,  628,
  629,   -1,   -1,  632,  633,  634,  635,  636,  637,  638,
  639,  640,   41,   -1,   -1,   -1,   -1,  646,   -1,  648,
  649,  650,  651,  652,  653,   -1,  655,   -1,   -1,   -1,
   -1,   60,   61,   62,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  672,  673,   -1,  675,  676,  677,  678,
  679,  680,  681,  682,  683,   -1,  685,  686,  687,  688,
  689,  690,  691,  692,  693,  694,  695,  696,   -1,  698,
  699,  700,   -1,  702,   -1,   -1,   -1,  706,  707,  708,
  709,  710,  711,  712,  713,  714,  715,  716,  717,  718,
    0,  720,  721,   -1,   -1,   -1,   -1,   -1,  727,  728,
  729,  730,  731,  732,  733,  734,  735,  736,  737,  738,
  739,  740,  741,  742,  743,  744,   -1,   -1,  747,   -1,
  749,  750,  751,  752,  753,  754,   -1,   -1,   -1,   -1,
   40,   41,   42,   43,   44,   45,   46,   47,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   58,   59,
   60,   61,   62,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   91,   -1,   93,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  262,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  271,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  284,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  318,
   -1,   -1,   -1,   -1,   -1,   -1,  325,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  333,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  350,   -1,  352,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  362,   -1,  364,  365,  366,   -1,   -1,
   -1,   -1,  371,   -1,   -1,   -1,   -1,  257,   -1,  259,
   -1,   -1,  262,   -1,  264,  265,  266,   -1,  268,  269,
   -1,  271,   -1,   -1,   -1,   -1,   -1,  277,   -1,  279,
  280,  281,  282,   -1,  284,   -1,   -1,  287,   -1,   -1,
  290,   -1,   -1,  412,   -1,   -1,   -1,  297,   -1,  299,
  300,   -1,  302,  303,  304,  305,  306,   -1,   -1,  309,
   -1,   -1,  312,   -1,  314,  315,   -1,   -1,  318,  319,
  320,   -1,  322,  323,  324,  325,  326,  327,  328,   -1,
   -1,   -1,  332,  333,  334,  335,  336,   -1,   -1,   -1,
  340,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  348,  349,
  350,  351,  352,  353,   -1,   -1,   -1,  357,  358,   -1,
  360,   -1,  362,   -1,  364,  365,  366,   -1,  368,   -1,
  370,  371,  372,   -1,  374,  375,  376,  377,  378,   -1,
  380,  381,   -1,   -1,   -1,   -1,  386,   -1,  388,  389,
   -1,   -1,  392,  393,  394,   -1,   -1,  397,   -1,  399,
   -1,   -1,  402,  403,  404,  405,   -1,  407,   -1,  409,
   -1,  411,  412,  413,  414,  415,   -1,   -1,  418,  419,
  420,   -1,  422,  423,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  433,  434,  435,   -1,  437,  438,  439,
   -1,   -1,   -1,   -1,   -1,  445,   -1,   -1,   -1,  449,
   -1,  451,  452,  453,  454,   -1,  456,   -1,  458,  459,
  460,  461,  462,  463,  464,  465,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  475,  476,   -1,  478,  479,
  480,   -1,   -1,  602,   -1,  485,  486,  487,  488,  489,
   -1,  491,   -1,   -1,  494,  495,   -1,   -1,   -1,  499,
  500,   -1,   -1,  503,  504,  505,  506,  507,  508,  509,
  510,  511,   -1,  513,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  523,  524,  525,  526,  527,  528,  529,
  530,  531,  532,  533,  534,  535,  536,  537,  538,  539,
  540,  541,  542,  543,   -1,  545,  546,  547,  548,  549,
  550,   -1,  552,  553,  554,  555,   -1,  557,   -1,  559,
  560,  561,  562,  563,  564,  565,  566,  567,  568,  569,
  570,  571,  572,  573,  574,  575,  576,   -1,  578,  579,
  580,  581,   -1,  583,  584,  585,  586,  587,   -1,  589,
  590,  591,  592,  593,  594,  595,  596,  597,  598,  599,
  600,  601,  602,  603,  604,  605,  606,  607,  608,  609,
  610,  611,  612,  613,  614,  615,  616,  617,   -1,  619,
  620,   -1,   -1,   -1,  624,  625,  626,  627,  628,  629,
   -1,   -1,  632,  633,  634,  635,  636,  637,  638,  639,
  640,   41,   -1,   -1,   -1,   -1,  646,   -1,  648,  649,
  650,  651,  652,  653,   -1,  655,   -1,   -1,   -1,   -1,
   60,   61,   62,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  672,  673,   -1,  675,  676,  677,  678,  679,
  680,  681,  682,  683,   -1,  685,  686,  687,  688,  689,
  690,  691,  692,  693,  694,  695,  696,   -1,  698,  699,
  700,   -1,  702,   -1,   -1,   -1,  706,  707,  708,  709,
  710,  711,  712,  713,  714,  715,  716,  717,  718,    0,
  720,  721,   -1,   -1,   -1,   -1,   -1,  727,  728,  729,
  730,  731,  732,  733,  734,  735,  736,  737,  738,  739,
  740,  741,  742,  743,  744,   -1,   -1,  747,   -1,  749,
  750,  751,  752,  753,  754,   -1,   -1,   -1,   -1,   40,
   41,   42,   43,   44,   45,   46,   47,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   58,   59,   60,
   61,   62,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   91,   -1,   93,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  262,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  271,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  284,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  318,   -1,
   -1,   -1,   -1,   -1,   -1,  325,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  333,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  350,   -1,  352,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  362,   -1,  364,  365,  366,   -1,   -1,   -1,
   -1,  371,   -1,   -1,   -1,   -1,  257,   -1,  259,   -1,
   -1,  262,   -1,  264,  265,  266,   -1,  268,  269,   -1,
  271,   -1,   -1,   -1,   -1,   -1,  277,   -1,  279,  280,
  281,  282,   -1,  284,   -1,   -1,  287,   -1,   -1,  290,
   -1,   -1,  412,   -1,   -1,   -1,  297,   -1,  299,  300,
   -1,  302,  303,  304,  305,  306,   -1,   -1,  309,   -1,
   -1,  312,   -1,  314,  315,   -1,   -1,  318,  319,  320,
   -1,  322,  323,  324,  325,  326,  327,  328,   -1,   -1,
   -1,  332,  333,  334,  335,  336,   -1,   -1,   -1,  340,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  348,  349,  350,
  351,  352,  353,   -1,   -1,   -1,  357,  358,   -1,  360,
   -1,  362,   -1,  364,  365,  366,   -1,  368,   -1,  370,
  371,  372,   -1,  374,  375,  376,  377,  378,   -1,  380,
  381,   -1,   -1,   -1,   -1,  386,   -1,  388,  389,   -1,
   -1,  392,  393,  394,   -1,   -1,  397,   -1,  399,   -1,
   -1,  402,  403,  404,  405,   -1,  407,   -1,  409,   -1,
  411,  412,  413,  414,  415,   -1,   -1,  418,  419,  420,
   -1,  422,  423,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  433,  434,  435,   -1,  437,  438,  439,   -1,
   -1,   -1,   -1,   -1,  445,   -1,   -1,   -1,  449,   -1,
  451,  452,  453,  454,   -1,  456,   -1,  458,  459,  460,
  461,  462,  463,  464,  465,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  475,  476,   -1,  478,  479,  480,
   -1,   -1,  602,   -1,  485,  486,  487,  488,  489,   -1,
  491,   -1,   -1,  494,  495,   -1,   -1,   -1,  499,  500,
   -1,   -1,  503,  504,  505,  506,  507,  508,  509,  510,
  511,   -1,  513,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  523,  524,  525,  526,  527,  528,  529,  530,
  531,  532,  533,  534,  535,  536,  537,  538,  539,  540,
  541,  542,  543,   -1,  545,  546,  547,  548,  549,  550,
   -1,  552,  553,  554,  555,   -1,  557,   -1,  559,  560,
  561,  562,  563,  564,  565,  566,  567,  568,  569,  570,
  571,  572,  573,  574,  575,  576,   -1,  578,  579,  580,
  581,   -1,  583,  584,  585,  586,  587,   -1,  589,  590,
  591,  592,  593,  594,  595,  596,  597,  598,  599,  600,
  601,  602,  603,  604,  605,  606,  607,  608,  609,  610,
  611,  612,  613,  614,  615,  616,  617,   -1,  619,  620,
   -1,   -1,   -1,  624,  625,  626,  627,  628,  629,   -1,
   -1,  632,  633,  634,  635,  636,  637,  638,  639,  640,
   41,   -1,   -1,   -1,   -1,  646,   -1,  648,  649,  650,
  651,  652,  653,   -1,  655,   -1,   -1,   -1,   -1,   60,
   61,   62,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  672,  673,   -1,  675,  676,  677,  678,  679,  680,
  681,  682,  683,   -1,  685,  686,  687,  688,  689,  690,
  691,  692,  693,  694,  695,  696,   -1,  698,  699,  700,
   -1,  702,   -1,   -1,   -1,  706,  707,  708,  709,  710,
  711,  712,  713,  714,  715,  716,  717,  718,    0,  720,
  721,   -1,   -1,   -1,   -1,   -1,  727,  728,  729,  730,
  731,  732,  733,  734,  735,  736,  737,  738,  739,  740,
  741,  742,  743,  744,   -1,   -1,  747,   -1,  749,  750,
  751,  752,  753,  754,   -1,   -1,   -1,   -1,   40,   41,
   42,   43,   44,   45,   46,   47,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   58,   59,   60,   61,
   62,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   91,
   -1,   93,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  262,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  271,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  284,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  318,   -1,   -1,
   -1,   -1,   -1,   -1,  325,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  333,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  350,
   -1,  352,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  362,   -1,  364,  365,  366,   -1,   -1,   -1,   -1,
  371,   -1,   -1,   -1,   -1,  257,   -1,  259,   -1,   -1,
  262,   -1,  264,  265,  266,   -1,  268,  269,   -1,  271,
   -1,   -1,   -1,   -1,   -1,  277,   -1,  279,  280,  281,
  282,   -1,  284,   -1,   -1,  287,   -1,   -1,  290,   -1,
   -1,  412,   -1,   -1,   -1,  297,   -1,  299,  300,   -1,
  302,  303,  304,  305,  306,   -1,   -1,  309,   -1,   -1,
  312,   -1,  314,  315,   -1,   -1,  318,  319,  320,   -1,
  322,  323,  324,  325,  326,  327,  328,   -1,   -1,   -1,
  332,  333,  334,  335,  336,   -1,   -1,   -1,  340,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  348,  349,  350,  351,
  352,  353,   -1,   -1,   -1,  357,  358,   -1,  360,   -1,
  362,   -1,  364,  365,  366,   -1,  368,   -1,  370,  371,
  372,   -1,  374,  375,  376,  377,  378,   -1,  380,  381,
   -1,   -1,   -1,   -1,  386,   -1,  388,  389,   -1,   -1,
  392,  393,  394,   -1,   -1,  397,   -1,  399,   -1,   -1,
  402,  403,  404,  405,   -1,  407,   -1,  409,   -1,  411,
  412,  413,  414,  415,   -1,   -1,  418,  419,  420,   -1,
  422,  423,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  433,  434,  435,   -1,  437,  438,  439,   -1,   -1,
   -1,   -1,   -1,  445,   -1,   -1,   -1,  449,   -1,  451,
  452,  453,  454,   -1,  456,   -1,  458,  459,  460,  461,
  462,  463,  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  475,  476,   -1,  478,  479,  480,   -1,
   -1,  602,   -1,  485,  486,  487,  488,  489,   -1,  491,
   -1,   -1,  494,  495,   -1,   -1,   -1,  499,  500,   -1,
   -1,  503,  504,  505,  506,  507,  508,  509,  510,  511,
   -1,  513,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  523,  524,  525,  526,  527,  528,  529,  530,  531,
  532,  533,  534,  535,  536,  537,  538,  539,  540,  541,
  542,  543,   -1,  545,  546,  547,  548,  549,  550,   -1,
  552,  553,  554,  555,   -1,  557,   -1,  559,  560,  561,
  562,  563,  564,  565,  566,  567,  568,  569,  570,  571,
  572,  573,  574,  575,  576,   -1,  578,  579,  580,  581,
   -1,  583,  584,  585,  586,  587,   -1,  589,  590,  591,
  592,  593,  594,  595,  596,  597,  598,  599,  600,  601,
  602,  603,  604,  605,  606,  607,  608,  609,  610,  611,
  612,  613,  614,  615,  616,  617,   -1,  619,  620,   -1,
   -1,   -1,  624,  625,  626,  627,  628,  629,   -1,   -1,
  632,  633,  634,  635,  636,  637,  638,  639,  640,   41,
   -1,   -1,   -1,   -1,  646,   -1,  648,  649,  650,  651,
  652,  653,   -1,  655,   -1,   -1,   -1,   -1,   60,   61,
   62,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  672,  673,   -1,  675,  676,  677,  678,  679,  680,  681,
  682,  683,   -1,  685,  686,  687,  688,  689,  690,  691,
  692,  693,  694,  695,  696,   -1,  698,  699,  700,   -1,
  702,   -1,   -1,   -1,  706,  707,  708,  709,  710,  711,
  712,  713,  714,  715,  716,  717,  718,    0,  720,  721,
   -1,   -1,   -1,   -1,   -1,  727,  728,  729,  730,  731,
  732,  733,  734,  735,  736,  737,  738,  739,  740,  741,
  742,  743,  744,   -1,   -1,  747,   -1,  749,  750,  751,
  752,  753,  754,   -1,   -1,   -1,   -1,   40,   41,   42,
   43,   44,   45,   46,   47,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   58,   59,   60,   61,   62,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   91,   -1,
   93,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  262,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  271,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  284,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  318,   -1,   -1,   -1,
   -1,   -1,   -1,  325,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  333,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  350,   -1,
  352,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  362,   -1,  364,  365,  366,   -1,   -1,   -1,   -1,  371,
   -1,   -1,   -1,   -1,  257,   -1,  259,   -1,   -1,  262,
   -1,  264,  265,  266,   -1,  268,  269,   -1,  271,   -1,
   -1,   -1,   -1,   -1,  277,   -1,  279,  280,  281,  282,
   -1,  284,   -1,   -1,  287,   -1,   -1,  290,   -1,   -1,
  412,   -1,   -1,   -1,  297,   -1,  299,  300,   -1,  302,
  303,  304,  305,  306,   -1,   -1,  309,   -1,   -1,  312,
   -1,  314,  315,   -1,   -1,  318,  319,  320,   -1,  322,
  323,  324,  325,  326,  327,  328,   -1,   -1,   -1,  332,
  333,  334,  335,  336,   -1,   -1,   -1,  340,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  348,  349,  350,  351,  352,
  353,   -1,   -1,   -1,  357,  358,   -1,  360,   -1,  362,
   -1,  364,  365,  366,   -1,  368,   -1,  370,  371,  372,
   -1,  374,  375,  376,  377,  378,   -1,  380,  381,   -1,
   -1,   -1,   -1,  386,   -1,  388,  389,   -1,   -1,  392,
  393,  394,   -1,   -1,  397,   -1,  399,   -1,   -1,  402,
  403,  404,  405,   -1,  407,   -1,  409,   -1,  411,  412,
  413,  414,  415,   -1,   -1,  418,  419,  420,   -1,  422,
  423,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  433,  434,  435,   -1,  437,  438,  439,   -1,   -1,   -1,
   -1,   -1,  445,   -1,   -1,   -1,  449,   -1,  451,  452,
  453,  454,   -1,  456,   -1,  458,  459,  460,  461,  462,
  463,  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  475,  476,   -1,  478,  479,  480,   -1,   -1,
  602,   -1,  485,  486,  487,  488,  489,   -1,  491,   -1,
   -1,  494,  495,   -1,   -1,   -1,  499,  500,   -1,   -1,
  503,  504,  505,  506,  507,  508,  509,  510,  511,   -1,
  513,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  523,  524,  525,  526,  527,  528,  529,  530,  531,  532,
  533,  534,  535,  536,  537,  538,  539,  540,  541,  542,
  543,   -1,  545,  546,  547,  548,  549,  550,   -1,  552,
  553,  554,  555,   -1,  557,   -1,  559,  560,  561,  562,
  563,  564,  565,  566,  567,  568,  569,  570,  571,  572,
  573,  574,  575,  576,   -1,  578,  579,  580,  581,   -1,
  583,  584,  585,  586,  587,   -1,  589,  590,  591,  592,
  593,  594,  595,  596,  597,  598,  599,  600,  601,  602,
  603,  604,  605,  606,  607,  608,  609,  610,  611,  612,
  613,  614,  615,  616,  617,   -1,  619,  620,   -1,   -1,
   -1,  624,  625,  626,  627,  628,  629,   -1,   -1,  632,
  633,  634,  635,  636,  637,  638,  639,  640,   41,   -1,
   -1,   -1,   -1,  646,   -1,  648,  649,  650,  651,  652,
  653,   -1,  655,   -1,   -1,   -1,   -1,   60,   61,   62,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  672,
  673,   -1,  675,  676,  677,  678,  679,  680,  681,  682,
  683,   -1,  685,  686,  687,  688,  689,  690,  691,  692,
  693,  694,  695,  696,   -1,  698,  699,  700,   -1,  702,
   -1,   -1,   -1,  706,  707,  708,  709,  710,  711,  712,
  713,  714,  715,  716,  717,  718,    0,  720,  721,   -1,
   -1,   -1,   -1,   -1,  727,  728,  729,  730,  731,  732,
  733,  734,  735,  736,  737,  738,  739,  740,  741,  742,
  743,  744,   -1,   -1,  747,   -1,  749,  750,  751,  752,
  753,  754,   -1,   -1,   -1,   -1,   40,   41,   42,   43,
   44,   45,   46,   47,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   58,   59,   60,   61,   62,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   91,   -1,   93,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  262,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  271,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  284,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  318,   -1,   -1,   -1,   -1,
   -1,   -1,  325,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  333,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  350,   -1,  352,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  362,
   -1,  364,  365,  366,   -1,   -1,   -1,   -1,  371,   -1,
   -1,   -1,   -1,  257,   -1,  259,   -1,   -1,  262,   -1,
  264,  265,  266,   -1,  268,  269,   -1,  271,   -1,   -1,
   -1,   -1,   -1,  277,   -1,  279,  280,  281,  282,   -1,
  284,   -1,   -1,  287,   -1,   -1,  290,   -1,   -1,  412,
   -1,   -1,   -1,  297,   -1,  299,  300,   -1,  302,  303,
  304,  305,  306,   -1,   -1,  309,   -1,   -1,  312,   -1,
  314,  315,   -1,   -1,  318,  319,  320,   -1,  322,  323,
  324,  325,  326,  327,  328,   -1,   -1,   -1,  332,  333,
  334,  335,  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  348,  349,  350,  351,  352,  353,
   -1,   -1,   -1,  357,  358,   -1,  360,   -1,  362,   -1,
  364,  365,  366,   -1,  368,   -1,  370,  371,  372,   -1,
  374,  375,  376,  377,  378,   -1,  380,  381,   -1,   -1,
   -1,   -1,  386,   -1,  388,  389,   -1,   -1,  392,  393,
  394,   -1,   -1,  397,   -1,  399,   -1,   -1,  402,  403,
  404,  405,   -1,  407,   -1,  409,   -1,  411,  412,  413,
  414,  415,   -1,   -1,  418,  419,  420,   -1,  422,  423,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  433,
  434,  435,   -1,  437,  438,  439,   -1,   -1,   -1,   -1,
   -1,  445,   -1,   -1,   -1,  449,   -1,  451,  452,  453,
  454,   -1,  456,   -1,  458,  459,  460,  461,  462,  463,
  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  475,  476,   -1,  478,  479,  480,   -1,   -1,  602,
   -1,  485,  486,  487,  488,  489,   -1,  491,   -1,   -1,
  494,  495,   -1,   -1,   -1,  499,  500,   -1,   -1,  503,
  504,  505,  506,  507,  508,  509,  510,  511,   -1,  513,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  523,
  524,  525,  526,  527,  528,  529,  530,  531,  532,  533,
  534,  535,  536,  537,  538,  539,  540,  541,  542,  543,
   -1,  545,  546,  547,  548,  549,  550,   -1,  552,  553,
  554,  555,   -1,  557,   -1,  559,  560,  561,  562,  563,
  564,  565,  566,  567,  568,  569,  570,  571,  572,  573,
  574,  575,  576,   -1,  578,  579,  580,  581,   -1,  583,
  584,  585,  586,  587,   -1,  589,  590,  591,  592,  593,
  594,  595,  596,  597,  598,  599,  600,  601,  602,  603,
  604,  605,  606,  607,  608,  609,  610,  611,  612,  613,
  614,  615,  616,  617,   -1,  619,  620,   -1,   -1,   -1,
  624,  625,  626,  627,  628,  629,   -1,   -1,  632,  633,
  634,  635,  636,  637,  638,  639,  640,   41,   -1,   -1,
   -1,   -1,  646,   -1,  648,  649,  650,  651,  652,  653,
   -1,  655,   -1,   -1,   -1,   -1,   60,   61,   62,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  672,  673,
   -1,  675,  676,  677,  678,  679,  680,  681,  682,  683,
   -1,  685,  686,  687,  688,  689,  690,  691,  692,  693,
  694,  695,  696,   -1,  698,  699,  700,   -1,  702,   -1,
   -1,   -1,  706,  707,  708,  709,  710,  711,  712,  713,
  714,  715,  716,  717,  718,    0,  720,  721,   -1,   -1,
   -1,   -1,   -1,  727,  728,  729,  730,  731,  732,  733,
  734,  735,  736,  737,  738,  739,  740,  741,  742,  743,
  744,   -1,   -1,  747,   -1,  749,  750,  751,  752,  753,
  754,   -1,   -1,   -1,   -1,   40,   41,   42,   43,   44,
   45,   46,   47,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   58,   59,   60,   61,   62,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   91,   -1,   93,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  262,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  271,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  284,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  318,   -1,   -1,   -1,   -1,   -1,
   -1,  325,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  333,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  350,   -1,  352,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  362,   -1,
  364,  365,  366,   -1,   -1,   -1,   -1,  371,   -1,   -1,
   -1,   -1,  257,   -1,  259,   -1,   -1,  262,   -1,  264,
  265,  266,   -1,  268,  269,   -1,  271,   -1,   -1,   -1,
   -1,   -1,  277,   -1,  279,  280,  281,  282,   -1,  284,
   -1,   -1,  287,   -1,   -1,  290,   -1,   -1,  412,   -1,
   -1,   -1,  297,   -1,  299,  300,   -1,  302,  303,  304,
  305,  306,   -1,   -1,  309,   -1,   -1,  312,   -1,  314,
  315,   -1,   -1,  318,  319,  320,   -1,  322,  323,  324,
  325,  326,  327,  328,   -1,   -1,   -1,  332,  333,  334,
  335,  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  348,  349,  350,  351,  352,  353,   -1,
   -1,   -1,  357,  358,   -1,  360,   -1,  362,   -1,  364,
  365,  366,   -1,  368,   -1,  370,  371,  372,   -1,  374,
  375,  376,  377,  378,   -1,  380,  381,   -1,   -1,   -1,
   -1,  386,   -1,  388,  389,   -1,   -1,  392,  393,  394,
   -1,   -1,  397,   -1,  399,   -1,   -1,  402,  403,  404,
  405,   -1,  407,   -1,  409,   -1,  411,  412,  413,  414,
  415,   -1,   -1,  418,  419,  420,   -1,  422,  423,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  433,  434,
  435,   -1,  437,  438,  439,   -1,   -1,   -1,   -1,   -1,
  445,   -1,   -1,   -1,  449,   -1,  451,  452,  453,  454,
   -1,  456,   -1,  458,  459,  460,  461,  462,  463,  464,
  465,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  475,  476,   -1,  478,  479,  480,   -1,   -1,  602,   -1,
  485,  486,  487,  488,  489,   -1,  491,   -1,   -1,  494,
  495,   -1,   -1,   -1,  499,  500,   -1,   -1,  503,  504,
  505,  506,  507,  508,  509,  510,  511,   -1,  513,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  523,  524,
  525,  526,  527,  528,  529,  530,  531,  532,  533,  534,
  535,  536,  537,  538,  539,  540,  541,  542,  543,   -1,
  545,  546,  547,  548,  549,  550,   -1,  552,  553,  554,
  555,   -1,  557,   -1,  559,  560,  561,  562,  563,  564,
  565,  566,  567,  568,  569,  570,  571,  572,  573,  574,
  575,  576,   -1,  578,  579,  580,  581,   -1,  583,  584,
  585,  586,  587,   -1,  589,  590,  591,  592,  593,  594,
  595,  596,  597,  598,  599,  600,  601,  602,  603,  604,
  605,  606,  607,  608,  609,  610,  611,  612,  613,  614,
  615,  616,  617,   -1,  619,  620,   -1,   -1,   -1,  624,
  625,  626,  627,  628,  629,   -1,   -1,  632,  633,  634,
  635,  636,  637,  638,  639,  640,   41,   -1,   -1,   -1,
   -1,  646,   -1,  648,  649,  650,  651,  652,  653,   -1,
  655,   -1,   -1,   -1,   -1,   60,   61,   62,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  672,  673,   -1,
  675,  676,  677,  678,  679,  680,  681,  682,  683,   -1,
  685,  686,  687,  688,  689,  690,  691,  692,  693,  694,
  695,  696,   -1,  698,  699,  700,   -1,  702,   -1,   -1,
   -1,  706,  707,  708,  709,  710,  711,  712,  713,  714,
  715,  716,  717,  718,    0,  720,  721,   -1,   -1,   -1,
   -1,   -1,  727,  728,  729,  730,  731,  732,  733,  734,
  735,  736,  737,  738,  739,  740,  741,  742,  743,  744,
   -1,   -1,  747,   -1,  749,  750,  751,  752,  753,  754,
   -1,   -1,   -1,   -1,   40,   41,   42,   43,   44,   45,
   46,   47,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   58,   59,   60,   61,   62,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   91,   -1,   93,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  262,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  271,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  284,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  318,   -1,   -1,   -1,   -1,   -1,   -1,
  325,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  333,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  350,   -1,  352,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  362,   -1,  364,
  365,  366,   -1,   -1,   -1,   -1,  371,   -1,   -1,   -1,
   -1,  257,   -1,  259,   -1,   -1,  262,   -1,  264,  265,
  266,   -1,  268,  269,   -1,  271,   -1,   -1,   -1,   -1,
   -1,  277,   -1,  279,  280,  281,  282,   -1,  284,   -1,
   -1,  287,   -1,   -1,  290,   -1,   -1,  412,   -1,   -1,
   -1,  297,   -1,  299,  300,   -1,  302,  303,  304,  305,
  306,   -1,   -1,  309,   -1,   -1,  312,   -1,  314,  315,
   -1,   -1,  318,  319,  320,   -1,  322,  323,  324,  325,
  326,  327,  328,   -1,   -1,   -1,  332,  333,  334,  335,
  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  348,  349,  350,  351,  352,  353,   -1,   -1,
   -1,  357,  358,   -1,  360,   -1,  362,   -1,  364,  365,
  366,   -1,  368,   -1,  370,  371,  372,   -1,  374,  375,
  376,  377,  378,   -1,  380,  381,   -1,   -1,   -1,   -1,
  386,   -1,  388,  389,   -1,   -1,  392,  393,  394,   -1,
   -1,  397,   -1,  399,   -1,   -1,  402,  403,  404,  405,
   -1,  407,   -1,  409,   -1,  411,  412,  413,  414,  415,
   -1,   -1,  418,  419,  420,   -1,  422,  423,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  433,  434,  435,
   -1,  437,  438,  439,   -1,   -1,   -1,   -1,   -1,  445,
   -1,   -1,   -1,  449,   -1,  451,  452,  453,  454,   -1,
  456,   -1,  458,  459,  460,  461,  462,  463,  464,  465,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  475,
  476,   -1,  478,  479,  480,   -1,   -1,  602,   -1,  485,
  486,  487,  488,  489,   -1,  491,   -1,   -1,  494,  495,
   -1,   -1,   -1,  499,  500,   -1,   -1,  503,  504,  505,
  506,  507,  508,  509,  510,  511,   -1,  513,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  523,  524,  525,
  526,  527,  528,  529,  530,  531,  532,  533,  534,  535,
  536,  537,  538,  539,  540,  541,  542,  543,   -1,  545,
  546,  547,  548,  549,  550,   -1,  552,  553,  554,  555,
   -1,  557,   -1,  559,  560,  561,  562,  563,  564,  565,
  566,  567,  568,  569,  570,  571,  572,  573,  574,  575,
  576,   -1,  578,  579,  580,  581,   -1,  583,  584,  585,
  586,  587,   -1,  589,  590,  591,  592,  593,  594,  595,
  596,  597,  598,  599,  600,  601,  602,  603,  604,  605,
  606,  607,  608,  609,  610,  611,  612,  613,  614,  615,
  616,  617,   -1,  619,  620,   -1,   -1,   -1,  624,  625,
  626,  627,  628,  629,   -1,   -1,  632,  633,  634,  635,
  636,  637,  638,  639,  640,   41,   -1,   -1,   -1,   -1,
  646,   -1,  648,  649,  650,  651,  652,  653,   -1,  655,
   -1,   -1,   -1,   -1,   60,   61,   62,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  672,  673,   -1,  675,
  676,  677,  678,  679,  680,  681,  682,  683,   -1,  685,
  686,  687,  688,  689,  690,  691,  692,  693,  694,  695,
  696,   -1,  698,  699,  700,   -1,  702,   -1,   -1,   -1,
  706,  707,  708,  709,  710,  711,  712,  713,  714,  715,
  716,  717,  718,    0,  720,  721,   -1,   -1,   -1,   -1,
   -1,  727,  728,  729,  730,  731,  732,  733,  734,  735,
  736,  737,  738,  739,  740,  741,  742,  743,  744,   -1,
   -1,  747,   -1,  749,  750,  751,  752,  753,  754,   -1,
   -1,   -1,   -1,   40,   41,   42,   43,   44,   45,   46,
   47,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   58,   59,   60,   61,   62,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   91,   -1,   93,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  262,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  271,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  284,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  318,   -1,   -1,   -1,   -1,   -1,   -1,  325,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  333,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  350,   -1,  352,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  362,   -1,  364,  365,
  366,   -1,   -1,   -1,   -1,  371,   -1,   -1,   -1,   -1,
  257,   -1,  259,   -1,   -1,  262,   -1,  264,  265,  266,
   -1,  268,  269,   -1,  271,   -1,   -1,   -1,   -1,   -1,
  277,   -1,  279,  280,  281,  282,   -1,  284,   -1,   -1,
  287,   -1,   -1,  290,   -1,   -1,  412,   -1,   -1,   -1,
  297,   -1,  299,  300,   -1,  302,  303,  304,  305,  306,
   -1,   -1,  309,   -1,   -1,  312,   -1,  314,  315,   -1,
   -1,  318,  319,  320,   -1,  322,  323,  324,  325,  326,
  327,  328,   -1,   -1,   -1,  332,  333,  334,  335,  336,
   -1,   -1,   -1,  340,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  348,  349,  350,  351,  352,  353,   -1,   -1,   -1,
  357,  358,   -1,  360,   -1,  362,   -1,  364,  365,  366,
   -1,  368,   -1,  370,  371,  372,   -1,  374,  375,  376,
  377,  378,   -1,  380,  381,   -1,   -1,   -1,   -1,  386,
   -1,  388,  389,   -1,   -1,  392,  393,  394,   -1,   -1,
  397,   -1,  399,   -1,   -1,  402,  403,  404,  405,   -1,
  407,   -1,  409,   -1,  411,  412,  413,  414,  415,   -1,
   -1,  418,  419,  420,   -1,  422,  423,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  433,  434,  435,   -1,
  437,  438,  439,   -1,   -1,   -1,   -1,   -1,  445,   -1,
   -1,   -1,  449,   -1,  451,  452,  453,  454,   -1,  456,
   -1,  458,  459,  460,  461,  462,  463,  464,  465,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  475,  476,
   -1,  478,  479,  480,   -1,   -1,  602,   -1,  485,  486,
  487,  488,  489,   -1,  491,   -1,   -1,  494,  495,   -1,
   -1,   -1,  499,  500,   -1,   -1,  503,  504,  505,  506,
  507,  508,  509,  510,  511,   -1,  513,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  523,  524,  525,  526,
  527,  528,  529,  530,  531,  532,  533,  534,  535,  536,
  537,  538,  539,  540,  541,  542,  543,   -1,  545,  546,
  547,  548,  549,  550,   -1,  552,  553,  554,  555,   -1,
  557,   -1,  559,  560,  561,  562,  563,  564,  565,  566,
  567,  568,  569,  570,  571,  572,  573,  574,  575,  576,
   -1,  578,  579,  580,  581,   -1,  583,  584,  585,  586,
  587,   -1,  589,  590,  591,  592,  593,  594,  595,  596,
  597,  598,  599,  600,  601,  602,  603,  604,  605,  606,
  607,  608,  609,  610,  611,  612,  613,  614,  615,  616,
  617,   -1,  619,  620,   -1,   -1,   -1,  624,  625,  626,
  627,  628,  629,   -1,   -1,  632,  633,  634,  635,  636,
  637,  638,  639,  640,   41,   -1,   -1,   -1,   -1,  646,
   -1,  648,  649,  650,  651,  652,  653,   -1,  655,   -1,
   -1,   -1,   -1,   60,   61,   62,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  672,  673,   -1,  675,  676,
  677,  678,  679,  680,  681,  682,  683,   -1,  685,  686,
  687,  688,  689,  690,  691,  692,  693,  694,  695,  696,
   -1,  698,  699,  700,   -1,  702,   -1,   -1,   -1,  706,
  707,  708,  709,  710,  711,  712,  713,  714,  715,  716,
  717,  718,    0,  720,  721,   -1,   -1,   -1,   -1,   -1,
  727,  728,  729,  730,  731,  732,  733,  734,  735,  736,
  737,  738,  739,  740,  741,  742,  743,  744,   -1,   -1,
  747,   -1,  749,  750,  751,  752,  753,  754,   -1,   -1,
   -1,   -1,   40,   41,   42,   43,   44,   45,   46,   47,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   58,   59,   60,   61,   62,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   91,   -1,   93,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  262,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  271,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  284,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  318,   -1,   -1,   -1,   -1,   -1,   -1,  325,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  333,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  350,   -1,  352,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  362,   -1,  364,  365,  366,
   -1,   -1,   -1,   -1,  371,   -1,   -1,   -1,   -1,  257,
   -1,  259,   -1,   -1,  262,   -1,  264,  265,  266,   -1,
  268,  269,   -1,  271,   -1,   -1,   -1,   -1,   -1,  277,
   -1,  279,  280,  281,  282,   -1,  284,   -1,   -1,  287,
   -1,   -1,  290,   -1,   -1,  412,   -1,   -1,   -1,  297,
   -1,  299,  300,   -1,  302,  303,  304,  305,  306,   -1,
   -1,  309,   -1,   -1,  312,   -1,  314,  315,   -1,   -1,
  318,  319,  320,   -1,  322,  323,  324,  325,  326,  327,
  328,   -1,   -1,   -1,  332,  333,  334,  335,  336,   -1,
   -1,   -1,  340,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  348,  349,  350,  351,  352,  353,   -1,   -1,   -1,  357,
  358,   -1,  360,   -1,  362,   -1,  364,  365,  366,   -1,
  368,   -1,  370,  371,  372,   -1,  374,  375,  376,  377,
  378,   -1,  380,  381,   -1,   -1,   -1,   -1,  386,   -1,
  388,  389,   -1,   -1,  392,  393,  394,   -1,   -1,  397,
   -1,  399,   -1,   -1,  402,  403,  404,  405,   -1,  407,
   -1,  409,   -1,  411,  412,  413,  414,  415,   -1,   -1,
  418,  419,  420,   -1,  422,  423,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  433,  434,  435,   -1,  437,
  438,  439,   -1,   -1,   -1,   -1,   -1,  445,   -1,   -1,
   -1,  449,   -1,  451,  452,  453,  454,   -1,  456,   -1,
  458,  459,  460,  461,  462,  463,  464,  465,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  475,  476,   -1,
  478,  479,  480,   -1,   -1,  602,   -1,  485,  486,  487,
  488,  489,   -1,  491,   -1,   -1,  494,  495,   -1,   -1,
   -1,  499,  500,   -1,   -1,  503,  504,  505,  506,  507,
  508,  509,  510,  511,   -1,  513,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  523,  524,  525,  526,  527,
  528,  529,  530,  531,  532,  533,  534,  535,  536,  537,
  538,  539,  540,  541,  542,  543,   -1,  545,  546,  547,
  548,  549,  550,   -1,  552,  553,  554,  555,   -1,  557,
   -1,  559,  560,  561,  562,  563,  564,  565,  566,  567,
  568,  569,  570,  571,  572,  573,  574,  575,  576,   -1,
  578,  579,  580,  581,   -1,  583,  584,  585,  586,  587,
   -1,  589,  590,  591,  592,  593,  594,  595,  596,  597,
  598,  599,  600,  601,  602,  603,  604,  605,  606,  607,
  608,  609,  610,  611,  612,  613,  614,  615,  616,  617,
   -1,  619,  620,   -1,   -1,   -1,  624,  625,  626,  627,
  628,  629,   -1,   -1,  632,  633,  634,  635,  636,  637,
  638,  639,  640,   41,   -1,   -1,   -1,   -1,  646,   -1,
  648,  649,  650,  651,  652,  653,   -1,  655,   -1,   -1,
   -1,   -1,   60,   61,   62,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  672,  673,   -1,  675,  676,  677,
  678,  679,  680,  681,  682,  683,   -1,  685,  686,  687,
  688,  689,  690,  691,  692,  693,  694,  695,  696,   -1,
  698,  699,  700,   -1,  702,   -1,   -1,   -1,  706,  707,
  708,  709,  710,  711,  712,  713,  714,  715,  716,  717,
  718,    0,  720,  721,   -1,   -1,   -1,   -1,   -1,  727,
  728,  729,  730,  731,  732,  733,  734,  735,  736,  737,
  738,  739,  740,  741,  742,  743,  744,   -1,   -1,  747,
   -1,  749,  750,  751,  752,  753,  754,   -1,   -1,   -1,
   -1,   40,   41,   42,   43,   44,   45,   46,   47,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   58,
   59,   60,   61,   62,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   91,   -1,   93,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  262,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  271,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  284,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  318,   -1,   -1,   -1,   -1,   -1,   -1,  325,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  333,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  350,   -1,  352,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  362,   -1,  364,  365,  366,   -1,
   -1,   -1,   -1,  371,   -1,   -1,   -1,   -1,  257,   -1,
  259,   -1,   -1,  262,   -1,  264,  265,  266,   -1,  268,
  269,   -1,  271,   -1,   -1,   -1,   -1,   -1,  277,   -1,
  279,  280,  281,  282,   -1,  284,   -1,   -1,  287,   -1,
   -1,  290,   -1,   -1,  412,   -1,   -1,   -1,  297,   -1,
  299,  300,   -1,  302,  303,  304,  305,  306,   -1,   -1,
  309,   -1,   -1,  312,   -1,  314,  315,   -1,   -1,  318,
  319,  320,   -1,  322,  323,  324,  325,  326,  327,  328,
   -1,   -1,   -1,  332,  333,  334,  335,  336,   -1,   -1,
   -1,  340,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  348,
  349,  350,  351,  352,  353,   -1,   -1,   -1,  357,  358,
   -1,  360,   -1,  362,   -1,  364,  365,  366,   -1,  368,
   -1,  370,  371,  372,   -1,  374,  375,  376,  377,  378,
   -1,  380,  381,   -1,   -1,   -1,   -1,  386,   -1,  388,
  389,   -1,   -1,  392,  393,  394,   -1,   -1,  397,   -1,
  399,   -1,   -1,  402,  403,  404,  405,   -1,  407,   -1,
  409,   -1,  411,  412,  413,  414,  415,   -1,   -1,  418,
  419,  420,   -1,  422,  423,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  433,  434,  435,   -1,  437,  438,
  439,   -1,   -1,   -1,   -1,   -1,  445,   -1,   -1,   -1,
  449,   -1,  451,  452,  453,  454,   -1,  456,   -1,  458,
  459,  460,  461,  462,  463,  464,  465,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  475,  476,   -1,  478,
  479,  480,   -1,   -1,  602,   -1,  485,  486,  487,  488,
  489,   -1,  491,   -1,   -1,  494,  495,   -1,   -1,   -1,
  499,  500,   -1,   -1,  503,  504,  505,  506,  507,  508,
  509,  510,  511,   -1,  513,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  523,  524,  525,  526,  527,  528,
  529,  530,  531,  532,  533,  534,  535,  536,  537,  538,
  539,  540,  541,  542,  543,   -1,  545,  546,  547,  548,
  549,  550,   -1,  552,  553,  554,  555,   -1,  557,   -1,
  559,  560,  561,  562,  563,  564,  565,  566,  567,  568,
  569,  570,  571,  572,  573,  574,  575,  576,   -1,  578,
  579,  580,  581,   -1,  583,  584,  585,  586,  587,   -1,
  589,  590,  591,  592,  593,  594,  595,  596,  597,  598,
  599,  600,  601,  602,  603,  604,  605,  606,  607,  608,
  609,  610,  611,  612,  613,  614,  615,  616,  617,   -1,
  619,  620,   -1,   -1,   -1,  624,  625,  626,  627,  628,
  629,   -1,   -1,  632,  633,  634,  635,  636,  637,  638,
  639,  640,   41,   -1,   -1,   -1,   -1,  646,   -1,  648,
  649,  650,  651,  652,  653,   -1,  655,   -1,   -1,   -1,
   -1,   60,   61,   62,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  672,  673,   -1,  675,  676,  677,  678,
  679,  680,  681,  682,  683,   -1,  685,  686,  687,  688,
  689,  690,  691,  692,  693,  694,  695,  696,   -1,  698,
  699,  700,   -1,  702,   -1,   -1,   -1,  706,  707,  708,
  709,  710,  711,  712,  713,  714,  715,  716,  717,  718,
    0,  720,  721,   -1,   -1,   -1,   -1,   -1,  727,  728,
  729,  730,  731,  732,  733,  734,  735,  736,  737,  738,
  739,  740,  741,  742,  743,  744,   -1,   -1,  747,   -1,
  749,  750,  751,  752,  753,  754,   -1,   -1,   -1,   -1,
   40,   41,   42,   43,   44,   45,   46,   47,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   58,   59,
   60,   61,   62,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   91,   -1,   93,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  262,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  271,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  284,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  318,
   -1,   -1,   -1,   -1,   -1,   -1,  325,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  333,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  350,   -1,  352,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  362,   -1,  364,  365,  366,   -1,   -1,
   -1,   -1,  371,   -1,   -1,   -1,   -1,  257,   -1,  259,
   -1,   -1,  262,   -1,  264,  265,  266,   -1,  268,  269,
   -1,  271,   -1,   -1,   -1,   -1,   -1,  277,   -1,  279,
  280,  281,  282,   -1,  284,   -1,   -1,  287,   -1,   -1,
  290,   -1,   -1,  412,   -1,   -1,   -1,  297,   -1,  299,
  300,   -1,  302,  303,  304,  305,  306,   -1,   -1,  309,
   -1,   -1,  312,   -1,  314,  315,   -1,   -1,  318,  319,
  320,   -1,  322,  323,  324,  325,  326,  327,  328,   -1,
   -1,   -1,  332,  333,  334,  335,  336,   -1,   -1,   -1,
  340,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  348,  349,
  350,  351,  352,  353,   -1,   -1,   -1,  357,  358,   -1,
  360,   -1,  362,   -1,  364,  365,  366,   -1,  368,   -1,
  370,  371,  372,   -1,  374,  375,  376,  377,  378,   -1,
  380,  381,   -1,   -1,   -1,   -1,  386,   -1,  388,  389,
   -1,   -1,  392,  393,  394,   -1,   -1,  397,   -1,  399,
   -1,   -1,  402,  403,  404,  405,   -1,  407,   -1,  409,
   -1,  411,  412,  413,  414,  415,   -1,   -1,  418,  419,
  420,   -1,  422,  423,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  433,  434,  435,   -1,  437,  438,  439,
   -1,   -1,   -1,   -1,   -1,  445,   -1,   -1,   -1,  449,
   -1,  451,  452,  453,  454,   -1,  456,   -1,  458,  459,
  460,  461,  462,  463,  464,  465,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  475,  476,   -1,  478,  479,
  480,   -1,   -1,  602,   -1,  485,  486,  487,  488,  489,
   -1,  491,   -1,   -1,  494,  495,   -1,   -1,   -1,  499,
  500,   -1,   -1,  503,  504,  505,  506,  507,  508,  509,
  510,  511,   -1,  513,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  523,  524,  525,  526,  527,  528,  529,
  530,  531,  532,  533,  534,  535,  536,  537,  538,  539,
  540,  541,  542,  543,   -1,  545,  546,  547,  548,  549,
  550,   -1,  552,  553,  554,  555,   -1,  557,   -1,  559,
  560,  561,  562,  563,  564,  565,  566,  567,  568,  569,
  570,  571,  572,  573,  574,  575,  576,   -1,  578,  579,
  580,  581,   -1,  583,  584,  585,  586,  587,   -1,  589,
  590,  591,  592,  593,  594,  595,  596,  597,  598,  599,
  600,  601,  602,  603,  604,  605,  606,  607,  608,  609,
  610,  611,  612,  613,  614,  615,  616,  617,   -1,  619,
  620,   -1,   -1,   -1,  624,  625,  626,  627,  628,  629,
   -1,   -1,  632,  633,  634,  635,  636,  637,  638,  639,
  640,   41,   -1,   -1,   -1,   -1,  646,   -1,  648,  649,
  650,  651,  652,  653,   -1,  655,   -1,   -1,   -1,   -1,
   60,   61,   62,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  672,  673,   -1,  675,  676,  677,  678,  679,
  680,  681,  682,  683,   -1,  685,  686,  687,  688,  689,
  690,  691,  692,  693,  694,  695,  696,   -1,  698,  699,
  700,   -1,  702,   -1,   -1,   -1,  706,  707,  708,  709,
  710,  711,  712,  713,  714,  715,  716,  717,  718,    0,
  720,  721,   -1,   -1,   -1,   -1,   -1,  727,  728,  729,
  730,  731,  732,  733,  734,  735,  736,  737,  738,  739,
  740,  741,  742,  743,  744,   -1,   -1,  747,   -1,  749,
  750,  751,  752,  753,  754,   -1,   -1,   -1,   -1,   40,
   41,   42,   43,   44,   45,   46,   47,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   58,   59,   60,
   61,   62,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   91,   -1,   93,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  262,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  271,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  284,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  318,   -1,
   -1,   -1,   -1,   -1,   -1,  325,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  333,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  350,   -1,  352,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  362,   -1,  364,  365,  366,   -1,   -1,   -1,
   -1,  371,   -1,   -1,   -1,   -1,  257,   -1,  259,   -1,
   -1,  262,   -1,  264,  265,  266,   -1,  268,  269,   -1,
  271,   -1,   -1,   -1,   -1,   -1,  277,   -1,  279,  280,
  281,  282,   -1,  284,   -1,   -1,  287,   -1,   -1,  290,
   -1,   -1,  412,   -1,   -1,   -1,  297,   -1,  299,  300,
   -1,  302,  303,  304,  305,  306,   -1,   -1,  309,   -1,
   -1,  312,   -1,  314,  315,   -1,   -1,  318,  319,  320,
   -1,  322,  323,  324,  325,  326,  327,  328,   -1,   -1,
   -1,  332,  333,  334,  335,  336,   -1,   -1,   -1,  340,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  348,  349,  350,
  351,  352,  353,   -1,   -1,   -1,  357,  358,   -1,  360,
   -1,  362,   -1,  364,  365,  366,   -1,  368,   -1,  370,
  371,  372,   -1,  374,  375,  376,  377,  378,   -1,  380,
  381,   -1,   -1,   -1,   -1,  386,   -1,  388,  389,   -1,
   -1,  392,  393,  394,   -1,   -1,  397,   -1,  399,   -1,
   -1,  402,  403,  404,  405,   -1,  407,   -1,  409,   -1,
  411,  412,  413,  414,  415,   -1,   -1,  418,  419,  420,
   -1,  422,  423,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  433,  434,  435,   -1,  437,  438,  439,   -1,
   -1,   -1,   -1,   -1,  445,   -1,   -1,   -1,  449,   -1,
  451,  452,  453,  454,   -1,  456,   -1,  458,  459,  460,
  461,  462,  463,  464,  465,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  475,  476,   -1,  478,  479,  480,
   -1,   -1,  602,   -1,  485,  486,  487,  488,  489,   -1,
  491,   -1,   -1,  494,  495,   -1,   -1,   -1,  499,  500,
   -1,   -1,  503,  504,  505,  506,  507,  508,  509,  510,
  511,   -1,  513,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  523,  524,  525,  526,  527,  528,  529,  530,
  531,  532,  533,  534,  535,  536,  537,  538,  539,  540,
  541,  542,  543,   -1,  545,  546,  547,  548,  549,  550,
   -1,  552,  553,  554,  555,   -1,  557,   -1,  559,  560,
  561,  562,  563,  564,  565,  566,  567,  568,  569,  570,
  571,  572,  573,  574,  575,  576,   -1,  578,  579,  580,
  581,   -1,  583,  584,  585,  586,  587,   -1,  589,  590,
  591,  592,  593,  594,  595,  596,  597,  598,  599,  600,
  601,  602,  603,  604,  605,  606,  607,  608,  609,  610,
  611,  612,  613,  614,  615,  616,  617,   -1,  619,  620,
   -1,   -1,   -1,  624,  625,  626,  627,  628,  629,   -1,
   -1,  632,  633,  634,  635,  636,  637,  638,  639,  640,
   41,   -1,   -1,   -1,   -1,  646,   -1,  648,  649,  650,
  651,  652,  653,   -1,  655,   -1,   -1,   -1,   -1,   60,
   61,   62,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  672,  673,   -1,  675,  676,  677,  678,  679,  680,
  681,  682,  683,   -1,  685,  686,  687,  688,  689,  690,
  691,  692,  693,  694,  695,  696,   -1,  698,  699,  700,
   -1,  702,   -1,   -1,   -1,  706,  707,  708,  709,  710,
  711,  712,  713,  714,  715,  716,  717,  718,    0,  720,
  721,   -1,   -1,   -1,   -1,   -1,  727,  728,  729,  730,
  731,  732,  733,  734,  735,  736,  737,  738,  739,  740,
  741,  742,  743,  744,   -1,   -1,  747,   -1,  749,  750,
  751,  752,  753,  754,   -1,   -1,   -1,   -1,   40,   41,
   42,   43,   44,   45,   46,   47,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   58,   59,   60,   61,
   62,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   91,
   -1,   93,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  262,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  271,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  284,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  318,   -1,   -1,
   -1,   -1,   -1,   -1,  325,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  333,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  350,
   -1,  352,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  362,   -1,  364,  365,  366,   -1,   -1,   -1,   -1,
  371,   -1,   -1,   -1,   -1,  257,   -1,  259,   -1,   -1,
  262,   -1,  264,  265,  266,   -1,  268,  269,   -1,  271,
   -1,   -1,   -1,   -1,   -1,  277,   -1,  279,  280,  281,
  282,   -1,  284,   -1,   -1,  287,   -1,   -1,  290,   -1,
   -1,  412,   -1,   -1,   -1,  297,   -1,  299,  300,   -1,
  302,  303,  304,  305,  306,   -1,   -1,  309,   -1,   -1,
  312,   -1,  314,  315,   -1,   -1,  318,  319,  320,   -1,
  322,  323,  324,  325,  326,  327,  328,   -1,   -1,   -1,
  332,  333,  334,  335,  336,   -1,   -1,   -1,  340,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  348,  349,  350,  351,
  352,  353,   -1,   -1,   -1,  357,  358,   -1,  360,   -1,
  362,   -1,  364,  365,  366,   -1,  368,   -1,  370,  371,
  372,   -1,  374,  375,  376,  377,  378,   -1,  380,  381,
   -1,   -1,   -1,   -1,  386,   -1,  388,  389,   -1,   -1,
  392,  393,  394,   -1,   -1,  397,   -1,  399,   -1,   -1,
  402,  403,  404,  405,   -1,  407,   -1,  409,   -1,  411,
  412,  413,  414,  415,   -1,   -1,  418,  419,  420,   -1,
  422,  423,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  433,  434,  435,   -1,  437,  438,  439,   -1,   -1,
   -1,   -1,   -1,  445,   -1,   -1,   -1,  449,   -1,  451,
  452,  453,  454,   -1,  456,   -1,  458,  459,  460,  461,
  462,  463,  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  475,  476,   -1,  478,  479,  480,   -1,
   -1,  602,   -1,  485,  486,  487,  488,  489,   -1,  491,
   -1,   -1,  494,  495,   -1,   -1,   -1,  499,  500,   -1,
   -1,  503,  504,  505,  506,  507,  508,  509,  510,  511,
   -1,  513,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  523,  524,  525,  526,  527,  528,  529,  530,  531,
  532,  533,  534,  535,  536,  537,  538,  539,  540,  541,
  542,  543,   -1,  545,  546,  547,  548,  549,  550,   -1,
  552,  553,  554,  555,   -1,  557,   -1,  559,  560,  561,
  562,  563,  564,  565,  566,  567,  568,  569,  570,  571,
  572,  573,  574,  575,  576,   -1,  578,  579,  580,  581,
   -1,  583,  584,  585,  586,  587,   -1,  589,  590,  591,
  592,  593,  594,  595,  596,  597,  598,  599,  600,  601,
  602,  603,  604,  605,  606,  607,  608,  609,  610,  611,
  612,  613,  614,  615,  616,  617,   -1,  619,  620,   -1,
   -1,   -1,  624,  625,  626,  627,  628,  629,   -1,   -1,
  632,  633,  634,  635,  636,  637,  638,  639,  640,   41,
   -1,   -1,   -1,   -1,  646,   -1,  648,  649,  650,  651,
  652,  653,   -1,  655,   -1,   -1,   -1,   -1,   60,   61,
   62,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  672,  673,   -1,  675,  676,  677,  678,  679,  680,  681,
  682,  683,   -1,  685,  686,  687,  688,  689,  690,  691,
  692,  693,  694,  695,  696,   -1,  698,  699,  700,   -1,
  702,   -1,   -1,   -1,  706,  707,  708,  709,  710,  711,
  712,  713,  714,  715,  716,  717,  718,    0,  720,  721,
   -1,   -1,   -1,   -1,   -1,  727,  728,  729,  730,  731,
  732,  733,  734,  735,  736,  737,  738,  739,  740,  741,
  742,  743,  744,   -1,   -1,  747,   -1,  749,  750,  751,
  752,  753,  754,   -1,   -1,   -1,   -1,   40,   41,   42,
   43,   44,   45,   46,   47,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   58,   59,   60,   61,   62,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   91,   -1,
   93,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  262,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  271,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  284,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  318,   -1,   -1,   -1,
   -1,   -1,   -1,  325,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  333,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  350,   -1,
  352,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  362,   -1,  364,  365,  366,   -1,   -1,   -1,   -1,  371,
   -1,   -1,   -1,   -1,  257,   -1,  259,   -1,   -1,  262,
   -1,  264,  265,  266,   -1,  268,  269,   -1,  271,   -1,
   -1,   -1,   -1,   -1,  277,   -1,  279,  280,  281,  282,
   -1,  284,   -1,   -1,  287,   -1,   -1,  290,   -1,   -1,
  412,   -1,   -1,   -1,  297,   -1,  299,  300,   -1,  302,
  303,  304,  305,  306,   -1,   -1,  309,   -1,   -1,  312,
   -1,  314,  315,   -1,   -1,  318,  319,  320,   -1,  322,
  323,  324,  325,  326,  327,  328,   -1,   -1,   -1,  332,
  333,  334,  335,  336,   -1,   -1,   -1,  340,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  348,  349,  350,  351,  352,
  353,   -1,   -1,   -1,  357,  358,   -1,  360,   -1,  362,
   -1,  364,  365,  366,   -1,  368,   -1,  370,  371,  372,
   -1,  374,  375,  376,  377,  378,   -1,  380,  381,   -1,
   -1,   -1,   -1,  386,   -1,  388,  389,   -1,   -1,  392,
  393,  394,   -1,   -1,  397,   -1,  399,   -1,   -1,  402,
  403,  404,  405,   -1,  407,   -1,  409,   -1,  411,  412,
  413,  414,  415,   -1,   -1,  418,  419,  420,   -1,  422,
  423,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  433,  434,  435,   -1,  437,  438,  439,   -1,   -1,   -1,
   -1,   -1,  445,   -1,   -1,   -1,  449,   -1,  451,  452,
  453,  454,   -1,  456,   -1,  458,  459,  460,  461,  462,
  463,  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  475,  476,   -1,  478,  479,  480,   -1,   -1,
  602,   -1,  485,  486,  487,  488,  489,   -1,  491,   -1,
   -1,  494,  495,   -1,   -1,   -1,  499,  500,   -1,   -1,
  503,  504,  505,  506,  507,  508,  509,  510,  511,   -1,
  513,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  523,  524,  525,  526,  527,  528,  529,  530,  531,  532,
  533,  534,  535,  536,  537,  538,  539,  540,  541,  542,
  543,   -1,  545,  546,  547,  548,  549,  550,   -1,  552,
  553,  554,  555,   -1,  557,   -1,  559,  560,  561,  562,
  563,  564,  565,  566,  567,  568,  569,  570,  571,  572,
  573,  574,  575,  576,   -1,  578,  579,  580,  581,   -1,
  583,  584,  585,  586,  587,   -1,  589,  590,  591,  592,
  593,  594,  595,  596,  597,  598,  599,  600,  601,  602,
  603,  604,  605,  606,  607,  608,  609,  610,  611,  612,
  613,  614,  615,  616,  617,   -1,  619,  620,   -1,   -1,
   -1,  624,  625,  626,  627,  628,  629,   -1,   -1,  632,
  633,  634,  635,  636,  637,  638,  639,  640,   41,   -1,
   -1,   -1,   -1,  646,   -1,  648,  649,  650,  651,  652,
  653,   -1,  655,   -1,   -1,   -1,   -1,   60,   61,   62,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  672,
  673,   -1,  675,  676,  677,  678,  679,  680,  681,  682,
  683,   -1,  685,  686,  687,  688,  689,  690,  691,  692,
  693,  694,  695,  696,   -1,  698,  699,  700,   -1,  702,
   -1,   -1,   -1,  706,  707,  708,  709,  710,  711,  712,
  713,  714,  715,  716,  717,  718,    0,  720,  721,   -1,
   -1,   -1,   -1,   -1,  727,  728,  729,  730,  731,  732,
  733,  734,  735,  736,  737,  738,  739,  740,  741,  742,
  743,  744,   -1,   -1,  747,   -1,  749,  750,  751,  752,
  753,  754,   -1,   -1,   -1,   -1,   40,   41,   42,   43,
   44,   45,   46,   47,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   58,   59,   60,   61,   62,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   91,   -1,   93,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  262,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  271,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  284,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  318,   -1,   -1,   -1,   -1,
   -1,   -1,  325,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  333,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  350,   -1,  352,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  362,
   -1,  364,  365,  366,   -1,   -1,   -1,   -1,  371,   -1,
   -1,   -1,   -1,  257,   -1,  259,   -1,   -1,  262,   -1,
  264,  265,  266,   -1,  268,  269,   -1,  271,   -1,   -1,
   -1,   -1,   -1,  277,   -1,  279,  280,  281,  282,   -1,
  284,   -1,   -1,  287,   -1,   -1,  290,   -1,   -1,  412,
   -1,   -1,   -1,  297,   -1,  299,  300,   -1,  302,  303,
  304,  305,  306,   -1,   -1,  309,   -1,   -1,  312,   -1,
  314,  315,   -1,   -1,  318,  319,  320,   -1,  322,  323,
  324,  325,  326,  327,  328,   -1,   -1,   -1,  332,  333,
  334,  335,  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  348,  349,  350,  351,  352,  353,
   -1,   -1,   -1,  357,  358,   -1,  360,   -1,  362,   -1,
  364,  365,  366,   -1,  368,   -1,  370,  371,  372,   -1,
  374,  375,  376,  377,  378,   -1,  380,  381,   -1,   -1,
   -1,   -1,  386,   -1,  388,  389,   -1,   -1,  392,  393,
  394,   -1,   -1,  397,   -1,  399,   -1,   -1,  402,  403,
  404,  405,   -1,  407,   -1,  409,   -1,  411,  412,  413,
  414,  415,   -1,   -1,  418,  419,  420,   -1,  422,  423,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  433,
  434,  435,   -1,  437,  438,  439,   -1,   -1,   -1,   -1,
   -1,  445,   -1,   -1,   -1,  449,   -1,  451,  452,  453,
  454,   -1,  456,   -1,  458,  459,  460,  461,  462,  463,
  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  475,  476,   -1,  478,  479,  480,   -1,   -1,  602,
   -1,  485,  486,  487,  488,  489,   -1,  491,   -1,   -1,
  494,  495,   -1,   -1,   -1,  499,  500,   -1,   -1,  503,
  504,  505,  506,  507,  508,  509,  510,  511,   -1,  513,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  523,
  524,  525,  526,  527,  528,  529,  530,  531,  532,  533,
  534,  535,  536,  537,  538,  539,  540,  541,  542,  543,
   -1,  545,  546,  547,  548,  549,  550,   -1,  552,  553,
  554,  555,   -1,  557,   -1,  559,  560,  561,  562,  563,
  564,  565,  566,  567,  568,  569,  570,  571,  572,  573,
  574,  575,  576,   -1,  578,  579,  580,  581,   -1,  583,
  584,  585,  586,  587,   -1,  589,  590,  591,  592,  593,
  594,  595,  596,  597,  598,  599,  600,  601,  602,  603,
  604,  605,  606,  607,  608,  609,  610,  611,  612,  613,
  614,  615,  616,  617,   -1,  619,  620,   -1,   -1,   -1,
  624,  625,  626,  627,  628,  629,   -1,   -1,  632,  633,
  634,  635,  636,  637,  638,  639,  640,   41,   -1,   -1,
   -1,   -1,  646,   -1,  648,  649,  650,  651,  652,  653,
   -1,  655,   -1,   -1,   -1,   -1,   60,   61,   62,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  672,  673,
   -1,  675,  676,  677,  678,  679,  680,  681,  682,  683,
   -1,  685,  686,  687,  688,  689,  690,  691,  692,  693,
  694,  695,  696,   -1,  698,  699,  700,   -1,  702,   -1,
   -1,   -1,  706,  707,  708,  709,  710,  711,  712,  713,
  714,  715,  716,  717,  718,    0,  720,  721,   -1,   -1,
   -1,   -1,   -1,  727,  728,  729,  730,  731,  732,  733,
  734,  735,  736,  737,  738,  739,  740,  741,  742,  743,
  744,   -1,   -1,  747,   -1,  749,  750,  751,  752,  753,
  754,   -1,   -1,   -1,   -1,   40,   41,   42,   43,   44,
   45,   46,   47,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   58,   59,   60,   61,   62,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   91,   -1,   93,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  262,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  271,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  284,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  318,   -1,   -1,   -1,   -1,   -1,
   -1,  325,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  333,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  350,   -1,  352,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  362,   -1,
  364,  365,  366,   -1,   -1,   -1,   -1,  371,   -1,   -1,
   -1,   -1,  257,   -1,  259,   -1,   -1,  262,   -1,  264,
  265,  266,   -1,  268,  269,   -1,  271,   -1,   -1,   -1,
   -1,   -1,  277,   -1,  279,  280,  281,  282,   -1,  284,
   -1,   -1,  287,   -1,   -1,  290,   -1,   -1,  412,   -1,
   -1,   -1,  297,   -1,  299,  300,   -1,  302,  303,  304,
  305,  306,   -1,   -1,  309,   -1,   -1,  312,   -1,  314,
  315,   -1,   -1,  318,  319,  320,   -1,  322,  323,  324,
  325,  326,  327,  328,   -1,   -1,   -1,  332,  333,  334,
  335,  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  348,  349,  350,  351,  352,  353,   -1,
   -1,   -1,  357,  358,   -1,  360,   -1,  362,   -1,  364,
  365,  366,   -1,  368,   -1,  370,  371,  372,   -1,  374,
  375,  376,  377,  378,   -1,  380,  381,   -1,   -1,   -1,
   -1,  386,   -1,  388,  389,   -1,   -1,  392,  393,  394,
   -1,   -1,  397,   -1,  399,   -1,   -1,  402,  403,  404,
  405,   -1,  407,   -1,  409,   -1,  411,  412,  413,  414,
  415,   -1,   -1,  418,  419,  420,   -1,  422,  423,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  433,  434,
  435,   -1,  437,  438,  439,   -1,   -1,   -1,   -1,   -1,
  445,   -1,   -1,   -1,  449,   -1,  451,  452,  453,  454,
   -1,  456,   -1,  458,  459,  460,  461,  462,  463,  464,
  465,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  475,  476,   -1,  478,  479,  480,   -1,   -1,  602,   -1,
  485,  486,  487,  488,  489,   -1,  491,   -1,   -1,  494,
  495,   -1,   -1,   -1,  499,  500,   -1,   -1,  503,  504,
  505,  506,  507,  508,  509,  510,  511,   -1,  513,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  523,  524,
  525,  526,  527,  528,  529,  530,  531,  532,  533,  534,
  535,  536,  537,  538,  539,  540,  541,  542,  543,   -1,
  545,  546,  547,  548,  549,  550,   -1,  552,  553,  554,
  555,   -1,  557,   -1,  559,  560,  561,  562,  563,  564,
  565,  566,  567,  568,  569,  570,  571,  572,  573,  574,
  575,  576,   -1,  578,  579,  580,  581,   -1,  583,  584,
  585,  586,  587,   -1,  589,  590,  591,  592,  593,  594,
  595,  596,  597,  598,  599,  600,  601,  602,  603,  604,
  605,  606,  607,  608,  609,  610,  611,  612,  613,  614,
  615,  616,  617,   -1,  619,  620,   -1,   -1,   -1,  624,
  625,  626,  627,  628,  629,   -1,   -1,  632,  633,  634,
  635,  636,  637,  638,  639,  640,   41,   -1,   -1,   -1,
   -1,  646,   -1,  648,  649,  650,  651,  652,  653,   -1,
  655,   -1,   -1,   -1,   -1,   60,   61,   62,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  672,  673,   -1,
  675,  676,  677,  678,  679,  680,  681,  682,  683,   -1,
  685,  686,  687,  688,  689,  690,  691,  692,  693,  694,
  695,  696,   -1,  698,  699,  700,   -1,  702,   -1,   -1,
   -1,  706,  707,  708,  709,  710,  711,  712,  713,  714,
  715,  716,  717,  718,    0,  720,  721,   -1,   -1,   -1,
   -1,   -1,  727,  728,  729,  730,  731,  732,  733,  734,
  735,  736,  737,  738,  739,  740,  741,  742,  743,  744,
   -1,   -1,  747,   -1,  749,  750,  751,  752,  753,  754,
   -1,   -1,   -1,   -1,   40,   41,   42,   43,   44,   45,
   46,   47,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   58,   59,   60,   61,   62,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   91,   -1,   93,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  262,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  271,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  284,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  318,   -1,   -1,   -1,   -1,   -1,   -1,
  325,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  333,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  350,   -1,  352,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  362,   -1,  364,
  365,  366,   -1,   -1,   -1,   -1,  371,   -1,   -1,   -1,
   -1,  257,   -1,  259,   -1,   -1,  262,   -1,  264,  265,
  266,   -1,  268,  269,   -1,  271,   -1,   -1,   -1,   -1,
   -1,  277,   -1,  279,  280,  281,  282,   -1,  284,   -1,
   -1,  287,   -1,   -1,  290,   -1,   -1,  412,   -1,   -1,
   -1,  297,   -1,  299,  300,   -1,  302,  303,  304,  305,
  306,   -1,   -1,  309,   -1,   -1,  312,   -1,  314,  315,
   -1,   -1,  318,  319,  320,   -1,  322,  323,  324,  325,
  326,  327,  328,   -1,   -1,   -1,  332,  333,  334,  335,
  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  348,  349,  350,  351,  352,  353,   -1,   -1,
   -1,  357,  358,   -1,  360,   -1,  362,   -1,  364,  365,
  366,   -1,  368,   -1,  370,  371,  372,   -1,  374,  375,
  376,  377,  378,   -1,  380,  381,   -1,   -1,   -1,   -1,
  386,   -1,  388,  389,   -1,   -1,  392,  393,  394,   -1,
   -1,  397,   -1,  399,   -1,   -1,  402,  403,  404,  405,
   -1,  407,   -1,  409,   -1,  411,  412,  413,  414,  415,
   -1,   -1,  418,  419,  420,   -1,  422,  423,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  433,  434,  435,
   -1,  437,  438,  439,   -1,   -1,   -1,   -1,   -1,  445,
   -1,   -1,   -1,  449,   -1,  451,  452,  453,  454,   -1,
  456,   -1,  458,  459,  460,  461,  462,  463,  464,  465,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  475,
  476,   -1,  478,  479,  480,   -1,   -1,  602,   -1,  485,
  486,  487,  488,  489,   -1,  491,   -1,   -1,  494,  495,
   -1,   -1,   -1,  499,  500,   -1,   -1,  503,  504,  505,
  506,  507,  508,  509,  510,  511,   -1,  513,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  523,  524,  525,
  526,  527,  528,  529,  530,  531,  532,  533,  534,  535,
  536,  537,  538,  539,  540,  541,  542,  543,   -1,  545,
  546,  547,  548,  549,  550,   -1,  552,  553,  554,  555,
   -1,  557,   -1,  559,  560,  561,  562,  563,  564,  565,
  566,  567,  568,  569,  570,  571,  572,  573,  574,  575,
  576,   -1,  578,  579,  580,  581,   -1,  583,  584,  585,
  586,  587,   -1,  589,  590,  591,  592,  593,  594,  595,
  596,  597,  598,  599,  600,  601,  602,  603,  604,  605,
  606,  607,  608,  609,  610,  611,  612,  613,  614,  615,
  616,  617,   -1,  619,  620,   -1,   -1,   -1,  624,  625,
  626,  627,  628,  629,   -1,   -1,  632,  633,  634,  635,
  636,  637,  638,  639,  640,   41,   -1,   -1,   -1,   -1,
  646,   -1,  648,  649,  650,  651,  652,  653,   -1,  655,
   -1,   -1,   -1,   -1,   60,   61,   62,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  672,  673,   -1,  675,
  676,  677,  678,  679,  680,  681,  682,  683,   -1,  685,
  686,  687,  688,  689,  690,  691,  692,  693,  694,  695,
  696,   -1,  698,  699,  700,   -1,  702,   -1,   -1,   -1,
  706,  707,  708,  709,  710,  711,  712,  713,  714,  715,
  716,  717,  718,    0,  720,  721,   -1,   -1,   -1,   -1,
   -1,  727,  728,  729,  730,  731,  732,  733,  734,  735,
  736,  737,  738,  739,  740,  741,  742,  743,  744,   -1,
   -1,  747,   -1,  749,  750,  751,  752,  753,  754,   -1,
   -1,   -1,   -1,   40,   41,   42,   43,   44,   45,   46,
   47,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   58,   59,   60,   61,   62,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   91,   -1,   93,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  262,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  271,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  284,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  318,   -1,   -1,   -1,   -1,   -1,   -1,  325,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  333,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  350,   -1,  352,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  362,   -1,  364,  365,
  366,   -1,   -1,   -1,   -1,  371,   -1,   -1,   -1,   -1,
  257,   -1,  259,   -1,   -1,  262,   -1,  264,  265,  266,
   -1,  268,  269,   -1,  271,   -1,   -1,   -1,   -1,   -1,
  277,   -1,  279,  280,  281,  282,   -1,  284,   -1,   -1,
  287,   -1,   -1,  290,   -1,   -1,  412,   -1,   -1,   -1,
  297,   -1,  299,  300,   -1,  302,  303,  304,  305,  306,
   -1,   -1,  309,   -1,   -1,  312,   -1,  314,  315,   -1,
   -1,  318,  319,  320,   -1,  322,  323,  324,  325,  326,
  327,  328,   -1,   -1,   -1,  332,  333,  334,  335,  336,
   -1,   -1,   -1,  340,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  348,  349,  350,  351,  352,  353,   -1,   -1,   -1,
  357,  358,   -1,  360,   -1,  362,   -1,  364,  365,  366,
   -1,  368,   -1,  370,  371,  372,   -1,  374,  375,  376,
  377,  378,   -1,  380,  381,   -1,   -1,   -1,   -1,  386,
   -1,  388,  389,   -1,   -1,  392,  393,  394,   -1,   -1,
  397,   -1,  399,   -1,   -1,  402,  403,  404,  405,   -1,
  407,   -1,  409,   -1,  411,  412,  413,  414,  415,   -1,
   -1,  418,  419,  420,   -1,  422,  423,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  433,  434,  435,   -1,
  437,  438,  439,   -1,   -1,   -1,   -1,   -1,  445,   -1,
   -1,   -1,  449,   -1,  451,  452,  453,  454,   -1,  456,
   -1,  458,  459,  460,  461,  462,  463,  464,  465,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  475,  476,
   -1,  478,  479,  480,   -1,   -1,  602,   -1,  485,  486,
  487,  488,  489,   -1,  491,   -1,   -1,  494,  495,   -1,
   -1,   -1,  499,  500,   -1,   -1,  503,  504,  505,  506,
  507,  508,  509,  510,  511,   -1,  513,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  523,  524,  525,  526,
  527,  528,  529,  530,  531,  532,  533,  534,  535,  536,
  537,  538,  539,  540,  541,  542,  543,   -1,  545,  546,
  547,  548,  549,  550,   -1,  552,  553,  554,  555,   -1,
  557,   -1,  559,  560,  561,  562,  563,  564,  565,  566,
  567,  568,  569,  570,  571,  572,  573,  574,  575,  576,
   -1,  578,  579,  580,  581,   -1,  583,  584,  585,  586,
  587,   -1,  589,  590,  591,  592,  593,  594,  595,  596,
  597,  598,  599,  600,  601,  602,  603,  604,  605,  606,
  607,  608,  609,  610,  611,  612,  613,  614,  615,  616,
  617,   -1,  619,  620,   -1,   -1,   -1,  624,  625,  626,
  627,  628,  629,   -1,   -1,  632,  633,  634,  635,  636,
  637,  638,  639,  640,   41,   -1,   -1,   -1,   -1,  646,
   -1,  648,  649,  650,  651,  652,  653,   -1,  655,   -1,
   -1,   -1,   -1,   60,   61,   62,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  672,  673,   -1,  675,  676,
  677,  678,  679,  680,  681,  682,  683,   -1,  685,  686,
  687,  688,  689,  690,  691,  692,  693,  694,  695,  696,
   -1,  698,  699,  700,   -1,  702,   -1,   -1,   -1,  706,
  707,  708,  709,  710,  711,  712,  713,  714,  715,  716,
  717,  718,    0,  720,  721,   -1,   -1,   -1,   -1,   -1,
  727,  728,  729,  730,  731,  732,  733,  734,  735,  736,
  737,  738,  739,  740,  741,  742,  743,  744,   -1,   -1,
  747,   -1,  749,  750,  751,  752,  753,  754,   -1,   -1,
   -1,   -1,   40,   41,   42,   43,   44,   45,   46,   47,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   58,   59,   60,   61,   62,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   91,   -1,   93,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  262,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  271,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  284,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  318,   -1,   -1,   -1,   -1,   -1,   -1,  325,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  333,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  350,   -1,  352,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  362,   -1,  364,  365,  366,
   -1,   -1,   -1,   -1,  371,   -1,   -1,   -1,   -1,  257,
   -1,  259,   -1,   -1,  262,   -1,  264,  265,  266,   -1,
  268,  269,   -1,  271,   -1,   -1,   -1,   -1,   -1,  277,
   -1,  279,  280,  281,  282,   -1,  284,   -1,   -1,  287,
   -1,   -1,  290,   -1,   -1,  412,   -1,   -1,   -1,  297,
   -1,  299,  300,   -1,  302,  303,  304,  305,  306,   -1,
   -1,  309,   -1,   -1,  312,   -1,  314,  315,   -1,   -1,
  318,  319,  320,   -1,  322,  323,  324,  325,  326,  327,
  328,   -1,   -1,   -1,  332,  333,  334,  335,  336,   -1,
   -1,   -1,  340,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  348,  349,  350,  351,  352,  353,   -1,   -1,   -1,  357,
  358,   -1,  360,   -1,  362,   -1,  364,  365,  366,   -1,
  368,   -1,  370,  371,  372,   -1,  374,  375,  376,  377,
  378,   -1,  380,  381,   -1,   -1,   -1,   -1,  386,   -1,
  388,  389,   -1,   -1,  392,  393,  394,   -1,   -1,  397,
   -1,  399,   -1,   -1,  402,  403,  404,  405,   -1,  407,
   -1,  409,   -1,  411,  412,  413,  414,  415,   -1,   -1,
  418,  419,  420,   -1,  422,  423,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  433,  434,  435,   -1,  437,
  438,  439,   -1,   -1,   -1,   -1,   -1,  445,   -1,   -1,
   -1,  449,   -1,  451,  452,  453,  454,   -1,  456,   -1,
  458,  459,  460,  461,  462,  463,  464,  465,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  475,  476,   -1,
  478,  479,  480,   -1,   -1,  602,   -1,  485,  486,  487,
  488,  489,   -1,  491,   -1,   -1,  494,  495,   -1,   -1,
   -1,  499,  500,   -1,   -1,  503,  504,  505,  506,  507,
  508,  509,  510,  511,   -1,  513,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  523,  524,  525,  526,  527,
  528,  529,  530,  531,  532,  533,  534,  535,  536,  537,
  538,  539,  540,  541,  542,  543,   -1,  545,  546,  547,
  548,  549,  550,   -1,  552,  553,  554,  555,   -1,  557,
   -1,  559,  560,  561,  562,  563,  564,  565,  566,  567,
  568,  569,  570,  571,  572,  573,  574,  575,  576,   -1,
  578,  579,  580,  581,   -1,  583,  584,  585,  586,  587,
   -1,  589,  590,  591,  592,  593,  594,  595,  596,  597,
  598,  599,  600,  601,  602,  603,  604,  605,  606,  607,
  608,  609,  610,  611,  612,  613,  614,  615,  616,  617,
   -1,  619,  620,   -1,   -1,   -1,  624,  625,  626,  627,
  628,  629,   -1,   -1,  632,  633,  634,  635,  636,  637,
  638,  639,  640,   41,   -1,   -1,   -1,   -1,  646,   -1,
  648,  649,  650,  651,  652,  653,   -1,  655,   -1,   -1,
   -1,   -1,   60,   61,   62,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  672,  673,   -1,  675,  676,  677,
  678,  679,  680,  681,  682,  683,   -1,  685,  686,  687,
  688,  689,  690,  691,  692,  693,  694,  695,  696,   -1,
  698,  699,  700,   -1,  702,   -1,   -1,   -1,  706,  707,
  708,  709,  710,  711,  712,  713,  714,  715,  716,  717,
  718,    0,  720,  721,   -1,   -1,   -1,   -1,   -1,  727,
  728,  729,  730,  731,  732,  733,  734,  735,  736,  737,
  738,  739,  740,  741,  742,  743,  744,   -1,   -1,  747,
   -1,  749,  750,  751,  752,  753,  754,   -1,   -1,   -1,
   -1,   40,   41,   42,   43,   44,   45,   46,   47,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   58,
   59,   60,   61,   62,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   91,   -1,   93,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  262,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  271,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  284,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  318,   -1,   -1,   -1,   -1,   -1,   -1,  325,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  333,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  350,   -1,  352,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  362,   -1,  364,  365,  366,   -1,
   -1,   -1,   -1,  371,   -1,   -1,   -1,   -1,  257,   -1,
  259,   -1,   -1,  262,   -1,  264,  265,  266,   -1,  268,
  269,   -1,  271,   -1,   -1,   -1,   -1,   -1,  277,   -1,
  279,  280,  281,  282,   -1,  284,   -1,   -1,  287,   -1,
   -1,  290,   -1,   -1,  412,   -1,   -1,   -1,  297,   -1,
  299,  300,   -1,  302,  303,  304,  305,  306,   -1,   -1,
  309,   -1,   -1,  312,   -1,  314,  315,   -1,   -1,  318,
  319,  320,   -1,  322,  323,  324,  325,  326,  327,  328,
   -1,   -1,   -1,  332,  333,  334,  335,  336,   -1,   -1,
   -1,  340,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  348,
  349,  350,  351,  352,  353,   -1,   -1,   -1,  357,  358,
   -1,  360,   -1,  362,   -1,  364,  365,  366,   -1,  368,
   -1,  370,  371,  372,   -1,  374,  375,  376,  377,  378,
   -1,  380,  381,   -1,   -1,   -1,   -1,  386,   -1,  388,
  389,   -1,   -1,  392,  393,  394,   -1,   -1,  397,   -1,
  399,   -1,   -1,  402,  403,  404,  405,   -1,  407,   -1,
  409,   -1,  411,  412,  413,  414,  415,   -1,   -1,  418,
  419,  420,   -1,  422,  423,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  433,  434,  435,   -1,  437,  438,
  439,   -1,   -1,   -1,   -1,   -1,  445,   -1,   -1,   -1,
  449,   -1,  451,  452,  453,  454,   -1,  456,   -1,  458,
  459,  460,  461,  462,  463,  464,  465,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  475,  476,   -1,  478,
  479,  480,   -1,   -1,  602,   -1,  485,  486,  487,  488,
  489,   -1,  491,   -1,   -1,  494,  495,   -1,   -1,   -1,
  499,  500,   -1,   -1,  503,  504,  505,  506,  507,  508,
  509,  510,  511,   -1,  513,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  523,  524,  525,  526,  527,  528,
  529,  530,  531,  532,  533,  534,  535,  536,  537,  538,
  539,  540,  541,  542,  543,   -1,  545,  546,  547,  548,
  549,  550,   -1,  552,  553,  554,  555,   -1,  557,   -1,
  559,  560,  561,  562,  563,  564,  565,  566,  567,  568,
  569,  570,  571,  572,  573,  574,  575,  576,   -1,  578,
  579,  580,  581,   -1,  583,  584,  585,  586,  587,   -1,
  589,  590,  591,  592,  593,  594,  595,  596,  597,  598,
  599,  600,  601,  602,  603,  604,  605,  606,  607,  608,
  609,  610,  611,  612,  613,  614,  615,  616,  617,   -1,
  619,  620,   -1,   -1,   -1,  624,  625,  626,  627,  628,
  629,   -1,   -1,  632,  633,  634,  635,  636,  637,  638,
  639,  640,   41,   -1,   -1,   -1,   -1,  646,   -1,  648,
  649,  650,  651,  652,  653,   -1,  655,   -1,   -1,   -1,
   -1,   60,   61,   62,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  672,  673,   -1,  675,  676,  677,  678,
  679,  680,  681,  682,  683,   -1,  685,  686,  687,  688,
  689,  690,  691,  692,  693,  694,  695,  696,   -1,  698,
  699,  700,   -1,  702,   -1,   -1,   -1,  706,  707,  708,
  709,  710,  711,  712,  713,  714,  715,  716,  717,  718,
    0,  720,  721,   -1,   -1,   -1,   -1,   -1,  727,  728,
  729,  730,  731,  732,  733,  734,  735,  736,  737,  738,
  739,  740,  741,  742,  743,  744,   -1,   -1,  747,   -1,
  749,  750,  751,  752,  753,  754,   -1,   -1,   -1,   -1,
   40,   41,   42,   43,   44,   45,   46,   47,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   58,   59,
   60,   61,   62,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   91,   -1,   93,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  262,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  271,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  284,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  318,
   -1,   -1,   -1,   -1,   -1,   -1,  325,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  333,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  350,   -1,  352,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  362,   -1,  364,  365,  366,   -1,   -1,
   -1,   -1,  371,   -1,   -1,   -1,   -1,  257,   -1,  259,
   -1,   -1,  262,   -1,  264,  265,  266,   -1,  268,  269,
   -1,  271,   -1,   -1,   -1,   -1,   -1,  277,   -1,  279,
  280,  281,  282,   -1,  284,   -1,   -1,  287,   -1,   -1,
  290,   -1,   -1,  412,   -1,   -1,   -1,  297,   -1,  299,
  300,   -1,  302,  303,  304,  305,  306,   -1,   -1,  309,
   -1,   -1,  312,   -1,  314,  315,   -1,   -1,  318,  319,
  320,   -1,  322,  323,  324,  325,  326,  327,  328,   -1,
   -1,   -1,  332,  333,  334,  335,  336,   -1,   -1,   -1,
  340,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  348,  349,
  350,  351,  352,  353,   -1,   -1,   -1,  357,  358,   -1,
  360,   -1,  362,   -1,  364,  365,  366,   -1,  368,   -1,
  370,  371,  372,   -1,  374,  375,  376,  377,  378,   -1,
  380,  381,   -1,   -1,   -1,   -1,  386,   -1,  388,  389,
   -1,   -1,  392,  393,  394,   -1,   -1,  397,   -1,  399,
   -1,   -1,  402,  403,  404,  405,   -1,  407,   -1,  409,
   -1,  411,  412,  413,  414,  415,   -1,   -1,  418,  419,
  420,   -1,  422,  423,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  433,  434,  435,   -1,  437,  438,  439,
   -1,   -1,   -1,   -1,   -1,  445,   -1,   -1,   -1,  449,
   -1,  451,  452,  453,  454,   -1,  456,   -1,  458,  459,
  460,  461,  462,  463,  464,  465,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  475,  476,   -1,  478,  479,
  480,   -1,   -1,  602,   -1,  485,  486,  487,  488,  489,
   -1,  491,   -1,   -1,  494,  495,   -1,   -1,   -1,  499,
  500,   -1,   -1,  503,  504,  505,  506,  507,  508,  509,
  510,  511,   -1,  513,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  523,  524,  525,  526,  527,  528,  529,
  530,  531,  532,  533,  534,  535,  536,  537,  538,  539,
  540,  541,  542,  543,   -1,  545,  546,  547,  548,  549,
  550,   -1,  552,  553,  554,  555,   -1,  557,   -1,  559,
  560,  561,  562,  563,  564,  565,  566,  567,  568,  569,
  570,  571,  572,  573,  574,  575,  576,   -1,  578,  579,
  580,  581,   -1,  583,  584,  585,  586,  587,   -1,  589,
  590,  591,  592,  593,  594,  595,  596,  597,  598,  599,
  600,  601,  602,  603,  604,  605,  606,  607,  608,  609,
  610,  611,  612,  613,  614,  615,  616,  617,   -1,  619,
  620,   -1,   -1,   -1,  624,  625,  626,  627,  628,  629,
   -1,   -1,  632,  633,  634,  635,  636,  637,  638,  639,
  640,   41,   -1,   -1,   -1,   -1,  646,   -1,  648,  649,
  650,  651,  652,  653,   -1,  655,   -1,   -1,   -1,   -1,
   60,   61,   62,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  672,  673,   -1,  675,  676,  677,  678,  679,
  680,  681,  682,  683,   -1,  685,  686,  687,  688,  689,
  690,  691,  692,  693,  694,  695,  696,   -1,  698,  699,
  700,   -1,  702,   -1,   -1,   -1,  706,  707,  708,  709,
  710,  711,  712,  713,  714,  715,  716,  717,  718,    0,
  720,  721,   -1,   -1,   -1,   -1,   -1,  727,  728,  729,
  730,  731,  732,  733,  734,  735,  736,  737,  738,  739,
  740,  741,  742,  743,  744,   -1,   -1,  747,   -1,  749,
  750,  751,  752,  753,  754,   -1,   -1,   -1,   -1,   40,
   41,   42,   43,   44,   45,   46,   47,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   58,   59,   60,
   61,   62,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   91,   -1,   93,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  262,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  271,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  284,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  318,   -1,
   -1,   -1,   -1,   -1,   -1,  325,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  333,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  350,   -1,  352,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  362,   -1,  364,  365,  366,   -1,   -1,   -1,
   -1,  371,   -1,   -1,   -1,   -1,  257,   -1,  259,   -1,
   -1,  262,   -1,  264,  265,  266,   -1,  268,  269,   -1,
  271,   -1,   -1,   -1,   -1,   -1,  277,   -1,  279,  280,
  281,  282,   -1,  284,   -1,   -1,  287,   -1,   -1,  290,
   -1,   -1,  412,   -1,   -1,   -1,  297,   -1,  299,  300,
   -1,  302,  303,  304,  305,  306,   -1,   -1,  309,   -1,
   -1,  312,   -1,  314,  315,   -1,   -1,  318,  319,  320,
   -1,  322,  323,  324,  325,  326,  327,  328,   -1,   -1,
   -1,  332,  333,  334,  335,  336,   -1,   -1,   -1,  340,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  348,  349,  350,
  351,  352,  353,   -1,   -1,   -1,  357,  358,   -1,  360,
   -1,  362,   -1,  364,  365,  366,   -1,  368,   -1,  370,
  371,  372,   -1,  374,  375,  376,  377,  378,   -1,  380,
  381,   -1,   -1,   -1,   -1,  386,   -1,  388,  389,   -1,
   -1,  392,  393,  394,   -1,   -1,  397,   -1,  399,   -1,
   -1,  402,  403,  404,  405,   -1,  407,   -1,  409,   -1,
  411,  412,  413,  414,  415,   -1,   -1,  418,  419,  420,
   -1,  422,  423,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  433,  434,  435,   -1,  437,  438,  439,   -1,
   -1,   -1,   -1,   -1,  445,   -1,   -1,   -1,  449,   -1,
  451,  452,  453,  454,   -1,  456,   -1,  458,  459,  460,
  461,  462,  463,  464,  465,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  475,  476,   -1,  478,  479,  480,
   -1,   -1,  602,   -1,  485,  486,  487,  488,  489,   -1,
  491,   -1,   -1,  494,  495,   -1,   -1,   -1,  499,  500,
   -1,   -1,  503,  504,  505,  506,  507,  508,  509,  510,
  511,   -1,  513,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  523,  524,  525,  526,  527,  528,  529,  530,
  531,  532,  533,  534,  535,  536,  537,  538,  539,  540,
  541,  542,  543,   -1,  545,  546,  547,  548,  549,  550,
   -1,  552,  553,  554,  555,   -1,  557,   -1,  559,  560,
  561,  562,  563,  564,  565,  566,  567,  568,  569,  570,
  571,  572,  573,  574,  575,  576,   -1,  578,  579,  580,
  581,   -1,  583,  584,  585,  586,  587,   -1,  589,  590,
  591,  592,  593,  594,  595,  596,  597,  598,  599,  600,
  601,  602,  603,  604,  605,  606,  607,  608,  609,  610,
  611,  612,  613,  614,  615,  616,  617,   -1,  619,  620,
   -1,   -1,   -1,  624,  625,  626,  627,  628,  629,   -1,
   -1,  632,  633,  634,  635,  636,  637,  638,  639,  640,
   41,   -1,   -1,   -1,   -1,  646,   -1,  648,  649,  650,
  651,  652,  653,   -1,  655,   -1,   -1,   -1,   -1,   60,
   61,   62,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  672,  673,   -1,  675,  676,  677,  678,  679,  680,
  681,  682,  683,   -1,  685,  686,  687,  688,  689,  690,
  691,  692,  693,  694,  695,  696,   -1,  698,  699,  700,
   -1,  702,   -1,   -1,   -1,  706,  707,  708,  709,  710,
  711,  712,  713,  714,  715,  716,  717,  718,    0,  720,
  721,   -1,   -1,   -1,   -1,   -1,  727,  728,  729,  730,
  731,  732,  733,  734,  735,  736,  737,  738,  739,  740,
  741,  742,  743,  744,   -1,   -1,  747,   -1,  749,  750,
  751,  752,  753,  754,   -1,   -1,   -1,   -1,   40,   41,
   42,   43,   44,   45,   46,   47,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   58,   59,   60,   61,
   62,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   91,
   -1,   93,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  262,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  271,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  284,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  318,   -1,   -1,
   -1,   -1,   -1,   -1,  325,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  333,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  350,
   -1,  352,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  362,   -1,  364,  365,  366,   -1,   -1,   -1,   -1,
  371,   -1,   -1,   -1,   -1,  257,   -1,  259,   -1,   -1,
  262,   -1,  264,  265,  266,   -1,  268,  269,   -1,  271,
   -1,   -1,   -1,   -1,   -1,  277,   -1,  279,  280,  281,
  282,   -1,  284,   -1,   -1,  287,   -1,   -1,  290,   -1,
   -1,  412,   -1,   -1,   -1,  297,   -1,  299,  300,   -1,
  302,  303,  304,  305,  306,   -1,   -1,  309,   -1,   -1,
  312,   -1,  314,  315,   -1,   -1,  318,  319,  320,   -1,
  322,  323,  324,  325,  326,  327,  328,   -1,   -1,   -1,
  332,  333,  334,  335,  336,   -1,   -1,   -1,  340,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  348,  349,  350,  351,
  352,  353,   -1,   -1,   -1,  357,  358,   -1,  360,   -1,
  362,   -1,  364,  365,  366,   -1,  368,   -1,  370,  371,
  372,   -1,  374,  375,  376,  377,  378,   -1,  380,  381,
   -1,   -1,   -1,   -1,  386,   -1,  388,  389,   -1,   -1,
  392,  393,  394,   -1,   -1,  397,   -1,  399,   -1,   -1,
  402,  403,  404,  405,   -1,  407,   -1,  409,   -1,  411,
  412,  413,  414,  415,   -1,   -1,  418,  419,  420,   -1,
  422,  423,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  433,  434,  435,   -1,  437,  438,  439,   -1,   -1,
   -1,   -1,   -1,  445,   -1,   -1,   -1,  449,   -1,  451,
  452,  453,  454,   -1,  456,   -1,  458,  459,  460,  461,
  462,  463,  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  475,  476,   -1,  478,  479,  480,   -1,
   -1,  602,   -1,  485,  486,  487,  488,  489,   -1,  491,
   -1,   -1,  494,  495,   -1,   -1,   -1,  499,  500,   -1,
   -1,  503,  504,  505,  506,  507,  508,  509,  510,  511,
   -1,  513,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  523,  524,  525,  526,  527,  528,  529,  530,  531,
  532,  533,  534,  535,  536,  537,  538,  539,  540,  541,
  542,  543,   -1,  545,  546,  547,  548,  549,  550,   -1,
  552,  553,  554,  555,   -1,  557,   -1,  559,  560,  561,
  562,  563,  564,  565,  566,  567,  568,  569,  570,  571,
  572,  573,  574,  575,  576,   -1,  578,  579,  580,  581,
   -1,  583,  584,  585,  586,  587,   -1,  589,  590,  591,
  592,  593,  594,  595,  596,  597,  598,  599,  600,  601,
  602,  603,  604,  605,  606,  607,  608,  609,  610,  611,
  612,  613,  614,  615,  616,  617,   -1,  619,  620,   -1,
   -1,   -1,  624,  625,  626,  627,  628,  629,   -1,   -1,
  632,  633,  634,  635,  636,  637,  638,  639,  640,   41,
   -1,   -1,   -1,   -1,  646,   -1,  648,  649,  650,  651,
  652,  653,   -1,  655,   -1,   -1,   -1,   -1,   60,   61,
   62,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  672,  673,   -1,  675,  676,  677,  678,  679,  680,  681,
  682,  683,   -1,  685,  686,  687,  688,  689,  690,  691,
  692,  693,  694,  695,  696,   -1,  698,  699,  700,   -1,
  702,   -1,   -1,   -1,  706,  707,  708,  709,  710,  711,
  712,  713,  714,  715,  716,  717,  718,    0,  720,  721,
   -1,   -1,   -1,   -1,   -1,  727,  728,  729,  730,  731,
  732,  733,  734,  735,  736,  737,  738,  739,  740,  741,
  742,  743,  744,   -1,   -1,  747,   -1,  749,  750,  751,
  752,  753,  754,   -1,   -1,   -1,   -1,   40,   41,   42,
   43,   44,   45,   46,   47,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   58,   59,   60,   61,   62,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   91,   -1,
   93,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  262,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  271,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  284,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  318,   -1,   -1,   -1,
   -1,   -1,   -1,  325,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  333,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  350,   -1,
  352,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  362,   -1,  364,  365,  366,   -1,   -1,   -1,   -1,  371,
   -1,   -1,   -1,   -1,  257,   -1,  259,   -1,   -1,  262,
   -1,  264,  265,  266,   -1,  268,  269,   -1,  271,   -1,
   -1,   -1,   -1,   -1,  277,   -1,  279,  280,  281,  282,
   -1,  284,   -1,   -1,  287,   -1,   -1,  290,   -1,   -1,
  412,   -1,   -1,   -1,  297,   -1,  299,  300,   -1,  302,
  303,  304,  305,  306,   -1,   -1,  309,   -1,   -1,  312,
   -1,  314,  315,   -1,   -1,  318,  319,  320,   -1,  322,
  323,  324,  325,  326,  327,  328,   -1,   -1,   -1,  332,
  333,  334,  335,  336,   -1,   -1,   -1,  340,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  348,  349,  350,  351,  352,
  353,   -1,   -1,   -1,  357,  358,   -1,  360,   -1,  362,
   -1,  364,  365,  366,   -1,  368,   -1,  370,  371,  372,
   -1,  374,  375,  376,  377,  378,   -1,  380,  381,   -1,
   -1,   -1,   -1,  386,   -1,  388,  389,   -1,   -1,  392,
  393,  394,   -1,   -1,  397,   -1,  399,   -1,   -1,  402,
  403,  404,  405,   -1,  407,   -1,  409,   -1,  411,  412,
  413,  414,  415,   -1,   -1,  418,  419,  420,   -1,  422,
  423,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  433,  434,  435,   -1,  437,  438,  439,   -1,   -1,   -1,
   -1,   -1,  445,   -1,   -1,   -1,  449,   -1,  451,  452,
  453,  454,   -1,  456,   -1,  458,  459,  460,  461,  462,
  463,  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  475,  476,   -1,  478,  479,  480,   -1,   -1,
  602,   -1,  485,  486,  487,  488,  489,   -1,  491,   -1,
   -1,  494,  495,   -1,   -1,   -1,  499,  500,   -1,   -1,
  503,  504,  505,  506,  507,  508,  509,  510,  511,   -1,
  513,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  523,  524,  525,  526,  527,  528,  529,  530,  531,  532,
  533,  534,  535,  536,  537,  538,  539,  540,  541,  542,
  543,   -1,  545,  546,  547,  548,  549,  550,   -1,  552,
  553,  554,  555,   -1,  557,   -1,  559,  560,  561,  562,
  563,  564,  565,  566,  567,  568,  569,  570,  571,  572,
  573,  574,  575,  576,   -1,  578,  579,  580,  581,   -1,
  583,  584,  585,  586,  587,   -1,  589,  590,  591,  592,
  593,  594,  595,  596,  597,  598,  599,  600,  601,  602,
  603,  604,  605,  606,  607,  608,  609,  610,  611,  612,
  613,  614,  615,  616,  617,   -1,  619,  620,   -1,   -1,
   -1,  624,  625,  626,  627,  628,  629,   -1,   -1,  632,
  633,  634,  635,  636,  637,  638,  639,  640,   41,   -1,
   -1,   -1,   -1,  646,   -1,  648,  649,  650,  651,  652,
  653,   -1,  655,   -1,   -1,   -1,   -1,   60,   61,   62,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  672,
  673,   -1,  675,  676,  677,  678,  679,  680,  681,  682,
  683,   -1,  685,  686,  687,  688,  689,  690,  691,  692,
  693,  694,  695,  696,   -1,  698,  699,  700,   -1,  702,
   -1,   -1,   -1,  706,  707,  708,  709,  710,  711,  712,
  713,  714,  715,  716,  717,  718,    0,  720,  721,   -1,
   -1,   -1,   -1,   -1,  727,  728,  729,  730,  731,  732,
  733,  734,  735,  736,  737,  738,  739,  740,  741,  742,
  743,  744,   -1,   -1,  747,   -1,  749,  750,  751,  752,
  753,  754,   -1,   -1,   -1,   -1,   40,   41,   42,   43,
   44,   45,   46,   47,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   58,   59,   60,   61,   62,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   91,   -1,   93,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  262,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  271,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  284,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  318,   -1,   -1,   -1,   -1,
   -1,   -1,  325,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  333,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  350,   -1,  352,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  362,
   -1,  364,  365,  366,   -1,   -1,   -1,   -1,  371,   -1,
   -1,   -1,   -1,  257,   -1,  259,   -1,   -1,  262,   -1,
  264,  265,  266,   -1,  268,  269,   -1,  271,   -1,   -1,
   -1,   -1,   -1,  277,   -1,  279,  280,  281,  282,   -1,
  284,   -1,   -1,  287,   -1,   -1,  290,   -1,   -1,  412,
   -1,   -1,   -1,  297,   -1,  299,  300,   -1,  302,  303,
  304,  305,  306,   -1,   -1,  309,   -1,   -1,  312,   -1,
  314,  315,   -1,   -1,  318,  319,  320,   -1,  322,  323,
  324,  325,  326,  327,  328,   -1,   -1,   -1,  332,  333,
  334,  335,  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  348,  349,  350,  351,  352,  353,
   -1,   -1,   -1,  357,  358,   -1,  360,   -1,  362,   -1,
  364,  365,  366,   -1,  368,   -1,  370,  371,  372,   -1,
  374,  375,  376,  377,  378,   -1,  380,  381,   -1,   -1,
   -1,   -1,  386,   -1,  388,  389,   -1,   -1,  392,  393,
  394,   -1,   -1,  397,   -1,  399,   -1,   -1,  402,  403,
  404,  405,   -1,  407,   -1,  409,   -1,  411,  412,  413,
  414,  415,   -1,   -1,  418,  419,  420,   -1,  422,  423,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  433,
  434,  435,   -1,  437,  438,  439,   -1,   -1,   -1,   -1,
   -1,  445,   -1,   -1,   -1,  449,   -1,  451,  452,  453,
  454,   -1,  456,   -1,  458,  459,  460,  461,  462,  463,
  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  475,  476,   -1,  478,  479,  480,   -1,   -1,  602,
   -1,  485,  486,  487,  488,  489,   -1,  491,   -1,   -1,
  494,  495,   -1,   -1,   -1,  499,  500,   -1,   -1,  503,
  504,  505,  506,  507,  508,  509,  510,  511,   -1,  513,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  523,
  524,  525,  526,  527,  528,  529,  530,  531,  532,  533,
  534,  535,  536,  537,  538,  539,  540,  541,  542,  543,
   -1,  545,  546,  547,  548,  549,  550,   -1,  552,  553,
  554,  555,   -1,  557,   -1,  559,  560,  561,  562,  563,
  564,  565,  566,  567,  568,  569,  570,  571,  572,  573,
  574,  575,  576,   -1,  578,  579,  580,  581,   -1,  583,
  584,  585,  586,  587,   -1,  589,  590,  591,  592,  593,
  594,  595,  596,  597,  598,  599,  600,  601,  602,  603,
  604,  605,  606,  607,  608,  609,  610,  611,  612,  613,
  614,  615,  616,  617,   -1,  619,  620,   -1,   -1,   -1,
  624,  625,  626,  627,  628,  629,   -1,   -1,  632,  633,
  634,  635,  636,  637,  638,  639,  640,   41,   -1,   -1,
   -1,   -1,  646,   -1,  648,  649,  650,  651,  652,  653,
   -1,  655,   -1,   -1,   -1,   -1,   60,   61,   62,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  672,  673,
   -1,  675,  676,  677,  678,  679,  680,  681,  682,  683,
   -1,  685,  686,  687,  688,  689,  690,  691,  692,  693,
  694,  695,  696,   -1,  698,  699,  700,   -1,  702,   -1,
   -1,   -1,  706,  707,  708,  709,  710,  711,  712,  713,
  714,  715,  716,  717,  718,    0,  720,  721,   -1,   -1,
   -1,   -1,   -1,  727,  728,  729,  730,  731,  732,  733,
  734,  735,  736,  737,  738,  739,  740,  741,  742,  743,
  744,   -1,   -1,  747,   -1,  749,  750,  751,  752,  753,
  754,   -1,   -1,   -1,   -1,   40,   41,   42,   43,   44,
   45,   46,   47,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   58,   59,   60,   61,   62,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   91,   -1,   93,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  262,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  271,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  284,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  318,   -1,   -1,   -1,   -1,   -1,
   -1,  325,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  333,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  350,   -1,  352,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  362,   -1,
  364,  365,  366,   -1,   -1,   -1,   -1,  371,   -1,   -1,
   -1,   -1,  257,   -1,  259,   -1,   -1,  262,   -1,  264,
  265,  266,   -1,  268,  269,   -1,  271,   -1,   -1,   -1,
   -1,   -1,  277,   -1,  279,  280,  281,  282,   -1,  284,
   -1,   -1,  287,   -1,   -1,  290,   -1,   -1,  412,   -1,
   -1,   -1,  297,   -1,  299,  300,   -1,  302,  303,  304,
  305,  306,   -1,   -1,  309,   -1,   -1,  312,   -1,  314,
  315,   -1,   -1,  318,  319,  320,   -1,  322,  323,  324,
  325,  326,  327,  328,   -1,   -1,   -1,  332,  333,  334,
  335,  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  348,  349,  350,  351,  352,  353,   -1,
   -1,   -1,  357,  358,   -1,  360,   -1,  362,   -1,  364,
  365,  366,   -1,  368,   -1,  370,  371,  372,   -1,  374,
  375,  376,  377,  378,   -1,  380,  381,   -1,   -1,   -1,
   -1,  386,   -1,  388,  389,   -1,   -1,  392,  393,  394,
   -1,   -1,  397,   -1,  399,   -1,   -1,  402,  403,  404,
  405,   -1,  407,   -1,  409,   -1,  411,  412,  413,  414,
  415,   -1,   -1,  418,  419,  420,   -1,  422,  423,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  433,  434,
  435,   -1,  437,  438,  439,   -1,   -1,   -1,   -1,   -1,
  445,   -1,   -1,   -1,  449,   -1,  451,  452,  453,  454,
   -1,  456,   -1,  458,  459,  460,  461,  462,  463,  464,
  465,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  475,  476,   -1,  478,  479,  480,   -1,   -1,  602,   -1,
  485,  486,  487,  488,  489,   -1,  491,   -1,   -1,  494,
  495,   -1,   -1,   -1,  499,  500,   -1,   -1,  503,  504,
  505,  506,  507,  508,  509,  510,  511,   -1,  513,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  523,  524,
  525,  526,  527,  528,  529,  530,  531,  532,  533,  534,
  535,  536,  537,  538,  539,  540,  541,  542,  543,   -1,
  545,  546,  547,  548,  549,  550,   -1,  552,  553,  554,
  555,   -1,  557,   -1,  559,  560,  561,  562,  563,  564,
  565,  566,  567,  568,  569,  570,  571,  572,  573,  574,
  575,  576,   -1,  578,  579,  580,  581,   -1,  583,  584,
  585,  586,  587,   -1,  589,  590,  591,  592,  593,  594,
  595,  596,  597,  598,  599,  600,  601,  602,  603,  604,
  605,  606,  607,  608,  609,  610,  611,  612,  613,  614,
  615,  616,  617,   -1,  619,  620,   -1,   -1,   -1,  624,
  625,  626,  627,  628,  629,   -1,   -1,  632,  633,  634,
  635,  636,  637,  638,  639,  640,   41,   -1,   -1,   -1,
   -1,  646,   -1,  648,  649,  650,  651,  652,  653,   -1,
  655,   -1,   -1,   -1,   -1,   60,   61,   62,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  672,  673,   -1,
  675,  676,  677,  678,  679,  680,  681,  682,  683,   -1,
  685,  686,  687,  688,  689,  690,  691,  692,  693,  694,
  695,  696,   -1,  698,  699,  700,   -1,  702,   -1,   -1,
   -1,  706,  707,  708,  709,  710,  711,  712,  713,  714,
  715,  716,  717,  718,    0,  720,  721,   -1,   -1,   -1,
   -1,   -1,  727,  728,  729,  730,  731,  732,  733,  734,
  735,  736,  737,  738,  739,  740,  741,  742,  743,  744,
   -1,   -1,  747,   -1,  749,  750,  751,  752,  753,  754,
   -1,   -1,   -1,   -1,   40,   41,   42,   43,   44,   45,
   46,   47,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   58,   59,   60,   61,   62,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   91,   -1,   93,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  262,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  271,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  284,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  318,   -1,   -1,   -1,   -1,   -1,   -1,
  325,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  333,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  350,   -1,  352,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  362,   -1,  364,
  365,  366,   -1,   -1,   -1,   -1,  371,   -1,   -1,   -1,
   -1,  257,   -1,  259,   -1,   -1,  262,   -1,  264,  265,
  266,   -1,  268,  269,   -1,  271,   -1,   -1,   -1,   -1,
   -1,  277,   -1,  279,  280,  281,  282,   -1,  284,   -1,
   -1,  287,   -1,   -1,  290,   -1,   -1,  412,   -1,   -1,
   -1,  297,   -1,  299,  300,   -1,  302,  303,  304,  305,
  306,   -1,   -1,  309,   -1,   -1,  312,   -1,  314,  315,
   -1,   -1,  318,  319,  320,   -1,  322,  323,  324,  325,
  326,  327,  328,   -1,   -1,   -1,  332,  333,  334,  335,
  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  348,  349,  350,  351,  352,  353,   -1,   -1,
   -1,  357,  358,   -1,  360,   -1,  362,   -1,  364,  365,
  366,   -1,  368,   -1,  370,  371,  372,   -1,  374,  375,
  376,  377,  378,   -1,  380,  381,   -1,   -1,   -1,   -1,
  386,   -1,  388,  389,   -1,   -1,  392,  393,  394,   -1,
   -1,  397,   -1,  399,   -1,   -1,  402,  403,  404,  405,
   -1,  407,   -1,  409,   -1,  411,  412,  413,  414,  415,
   -1,   -1,  418,  419,  420,   -1,  422,  423,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  433,  434,  435,
   -1,  437,  438,  439,   -1,   -1,   -1,   -1,   -1,  445,
   -1,   -1,   -1,  449,   -1,  451,  452,  453,  454,   -1,
  456,   -1,  458,  459,  460,  461,  462,  463,  464,  465,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  475,
  476,   -1,  478,  479,  480,   -1,   -1,  602,   -1,  485,
  486,  487,  488,  489,   -1,  491,   -1,   -1,  494,  495,
   -1,   -1,   -1,  499,  500,   -1,   -1,  503,  504,  505,
  506,  507,  508,  509,  510,  511,   -1,  513,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  523,  524,  525,
  526,  527,  528,  529,  530,  531,  532,  533,  534,  535,
  536,  537,  538,  539,  540,  541,  542,  543,   -1,  545,
  546,  547,  548,  549,  550,   -1,  552,  553,  554,  555,
   -1,  557,   -1,  559,  560,  561,  562,  563,  564,  565,
  566,  567,  568,  569,  570,  571,  572,  573,  574,  575,
  576,   -1,  578,  579,  580,  581,   -1,  583,  584,  585,
  586,  587,   -1,  589,  590,  591,  592,  593,  594,  595,
  596,  597,  598,  599,  600,  601,  602,  603,  604,  605,
  606,  607,  608,  609,  610,  611,  612,  613,  614,  615,
  616,  617,   -1,  619,  620,   -1,   -1,   -1,  624,  625,
  626,  627,  628,  629,   -1,   -1,  632,  633,  634,  635,
  636,  637,  638,  639,  640,   41,   -1,   -1,   -1,   -1,
  646,   -1,  648,  649,  650,  651,  652,  653,   -1,  655,
   -1,   -1,   -1,   -1,   60,   61,   62,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  672,  673,   -1,  675,
  676,  677,  678,  679,  680,  681,  682,  683,   -1,  685,
  686,  687,  688,  689,  690,  691,  692,  693,  694,  695,
  696,   -1,  698,  699,  700,   -1,  702,   -1,   -1,   -1,
  706,  707,  708,  709,  710,  711,  712,  713,  714,  715,
  716,  717,  718,    0,  720,  721,   -1,   -1,   -1,   -1,
   -1,  727,  728,  729,  730,  731,  732,  733,  734,  735,
  736,  737,  738,  739,  740,  741,  742,  743,  744,   -1,
   -1,  747,   -1,  749,  750,  751,  752,  753,  754,   -1,
   -1,   -1,   -1,   40,   41,   42,   43,   44,   45,   46,
   47,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   58,   59,   60,   61,   62,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   91,   -1,   93,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  262,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  271,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  284,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  318,   -1,   -1,   -1,   -1,   -1,   -1,  325,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  333,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  350,   -1,  352,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  362,   -1,  364,  365,
  366,   -1,   -1,   -1,   -1,  371,   -1,   -1,   -1,   -1,
  257,   -1,  259,   -1,   -1,  262,   -1,  264,  265,  266,
   -1,  268,  269,   -1,  271,   -1,   -1,   -1,   -1,   -1,
  277,   -1,  279,  280,  281,  282,   -1,  284,   -1,   -1,
  287,   -1,   -1,  290,   -1,   -1,  412,   -1,   -1,   -1,
  297,   -1,  299,  300,   -1,  302,  303,  304,  305,  306,
   -1,   -1,  309,   -1,   -1,  312,   -1,  314,  315,   -1,
   -1,  318,  319,  320,   -1,  322,  323,  324,  325,  326,
  327,  328,   -1,   -1,   -1,  332,  333,  334,  335,  336,
   -1,   -1,   -1,  340,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  348,  349,  350,  351,  352,  353,   -1,   -1,   -1,
  357,  358,   -1,  360,   -1,  362,   -1,  364,  365,  366,
   -1,  368,   -1,  370,  371,  372,   -1,  374,  375,  376,
  377,  378,   -1,  380,  381,   -1,   -1,   -1,   -1,  386,
   -1,  388,  389,   -1,   -1,  392,  393,  394,   -1,   -1,
  397,   -1,  399,   -1,   -1,  402,  403,  404,  405,   -1,
  407,   -1,  409,   -1,  411,  412,  413,  414,  415,   -1,
   -1,  418,  419,  420,   -1,  422,  423,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  433,  434,  435,   -1,
  437,  438,  439,   -1,   -1,   -1,   -1,   -1,  445,   -1,
   -1,   -1,  449,   -1,  451,  452,  453,  454,   -1,  456,
   -1,  458,  459,  460,  461,  462,  463,  464,  465,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  475,  476,
   -1,  478,  479,  480,   -1,   -1,  602,   -1,  485,  486,
  487,  488,  489,   -1,  491,   -1,   -1,  494,  495,   -1,
   -1,   -1,  499,  500,   -1,   -1,  503,  504,  505,  506,
  507,  508,  509,  510,  511,   -1,  513,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  523,  524,  525,  526,
  527,  528,  529,  530,  531,  532,  533,  534,  535,  536,
  537,  538,  539,  540,  541,  542,  543,   -1,  545,  546,
  547,  548,  549,  550,   -1,  552,  553,  554,  555,   -1,
  557,   -1,  559,  560,  561,  562,  563,  564,  565,  566,
  567,  568,  569,  570,  571,  572,  573,  574,  575,  576,
   -1,  578,  579,  580,  581,   -1,  583,  584,  585,  586,
  587,   -1,  589,  590,  591,  592,  593,  594,  595,  596,
  597,  598,  599,  600,  601,  602,  603,  604,  605,  606,
  607,  608,  609,  610,  611,  612,  613,  614,  615,  616,
  617,   -1,  619,  620,   -1,   -1,   -1,  624,  625,  626,
  627,  628,  629,   -1,   -1,  632,  633,  634,  635,  636,
  637,  638,  639,  640,   41,   -1,   -1,   -1,   -1,  646,
   -1,  648,  649,  650,  651,  652,  653,   -1,  655,   -1,
   -1,   -1,   -1,   60,   61,   62,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  672,  673,   -1,  675,  676,
  677,  678,  679,  680,  681,  682,  683,   -1,  685,  686,
  687,  688,  689,  690,  691,  692,  693,  694,  695,  696,
   -1,  698,  699,  700,   -1,  702,   -1,   -1,   -1,  706,
  707,  708,  709,  710,  711,  712,  713,  714,  715,  716,
  717,  718,    0,  720,  721,   -1,   -1,   -1,   -1,   -1,
  727,  728,  729,  730,  731,  732,  733,  734,  735,  736,
  737,  738,  739,  740,  741,  742,  743,  744,   -1,   -1,
  747,   -1,  749,  750,  751,  752,  753,  754,   -1,   -1,
   -1,   -1,   40,   41,   42,   43,   44,   45,   46,   47,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   58,   59,   60,   61,   62,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   91,   -1,   93,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  262,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  271,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  284,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  318,   -1,   -1,   -1,   -1,   -1,   -1,  325,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  333,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  350,   -1,  352,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  362,   -1,  364,  365,  366,
   -1,   -1,   -1,   -1,  371,   -1,   -1,   -1,   -1,  257,
   -1,  259,   -1,   -1,  262,   -1,  264,  265,  266,   -1,
  268,  269,   -1,  271,   -1,   -1,   -1,   -1,   -1,  277,
   -1,  279,  280,  281,  282,   -1,  284,   -1,   -1,  287,
   -1,   -1,  290,   -1,   -1,  412,   -1,   -1,   -1,  297,
   -1,  299,  300,   -1,  302,  303,  304,  305,  306,   -1,
   -1,  309,   -1,   -1,  312,   -1,  314,  315,   -1,   -1,
  318,  319,  320,   -1,  322,  323,  324,  325,  326,  327,
  328,   -1,   -1,   -1,  332,  333,  334,  335,  336,   -1,
   -1,   -1,  340,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  348,  349,  350,  351,  352,  353,   -1,   -1,   -1,  357,
  358,   -1,  360,   -1,  362,   -1,  364,  365,  366,   -1,
  368,   -1,  370,  371,  372,   -1,  374,  375,  376,  377,
  378,   -1,  380,  381,   -1,   -1,   -1,   -1,  386,   -1,
  388,  389,   -1,   -1,  392,  393,  394,   -1,   -1,  397,
   -1,  399,   -1,   -1,  402,  403,  404,  405,   -1,  407,
   -1,  409,   -1,  411,  412,  413,  414,  415,   -1,   -1,
  418,  419,  420,   -1,  422,  423,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  433,  434,  435,   -1,  437,
  438,  439,   -1,   -1,   -1,   -1,   -1,  445,   -1,   -1,
   -1,  449,   -1,  451,  452,  453,  454,   -1,  456,   -1,
  458,  459,  460,  461,  462,  463,  464,  465,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  475,  476,   -1,
  478,  479,  480,   -1,   -1,  602,   -1,  485,  486,  487,
  488,  489,   -1,  491,   -1,   -1,  494,  495,   -1,   -1,
   -1,  499,  500,   -1,   -1,  503,  504,  505,  506,  507,
  508,  509,  510,  511,   -1,  513,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  523,  524,  525,  526,  527,
  528,  529,  530,  531,  532,  533,  534,  535,  536,  537,
  538,  539,  540,  541,  542,  543,   -1,  545,  546,  547,
  548,  549,  550,   -1,  552,  553,  554,  555,   -1,  557,
   -1,  559,  560,  561,  562,  563,  564,  565,  566,  567,
  568,  569,  570,  571,  572,  573,  574,  575,  576,   -1,
  578,  579,  580,  581,   -1,  583,  584,  585,  586,  587,
   -1,  589,  590,  591,  592,  593,  594,  595,  596,  597,
  598,  599,  600,  601,  602,  603,  604,  605,  606,  607,
  608,  609,  610,  611,  612,  613,  614,  615,  616,  617,
   -1,  619,  620,   -1,   -1,   -1,  624,  625,  626,  627,
  628,  629,   -1,   -1,  632,  633,  634,  635,  636,  637,
  638,  639,  640,   41,   -1,   -1,   -1,   -1,  646,   -1,
  648,  649,  650,  651,  652,  653,   -1,  655,   -1,   -1,
   -1,   -1,   60,   61,   62,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  672,  673,   -1,  675,  676,  677,
  678,  679,  680,  681,  682,  683,   -1,  685,  686,  687,
  688,  689,  690,  691,  692,  693,  694,  695,  696,   -1,
  698,  699,  700,   -1,  702,   -1,   -1,   -1,  706,  707,
  708,  709,  710,  711,  712,  713,  714,  715,  716,  717,
  718,    0,  720,  721,   -1,   -1,   -1,   -1,   -1,  727,
  728,  729,  730,  731,  732,  733,  734,  735,  736,  737,
  738,  739,  740,  741,  742,  743,  744,   -1,   -1,  747,
   -1,  749,  750,  751,  752,  753,  754,   -1,   -1,   -1,
   -1,   40,   41,   42,   43,   44,   45,   46,   47,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   58,
   59,   60,   61,   62,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   91,   -1,   93,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  262,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  271,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  284,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  318,   -1,   -1,   -1,   -1,   -1,   -1,  325,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  333,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  350,   -1,  352,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  362,   -1,  364,  365,  366,   -1,
   -1,   -1,   -1,  371,   -1,   -1,   -1,   -1,  257,   -1,
  259,   -1,   -1,  262,   -1,  264,  265,  266,   -1,  268,
  269,   -1,  271,   -1,   -1,   -1,   -1,   -1,  277,   -1,
  279,  280,  281,  282,   -1,  284,   -1,   -1,  287,   -1,
   -1,  290,   -1,   -1,  412,   -1,   -1,   -1,  297,   -1,
  299,  300,   -1,  302,  303,  304,  305,  306,   -1,   -1,
  309,   -1,   -1,  312,   -1,  314,  315,   -1,   -1,  318,
  319,  320,   -1,  322,  323,  324,  325,  326,  327,  328,
   -1,   -1,   -1,  332,  333,  334,  335,  336,   -1,   -1,
   -1,  340,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  348,
  349,  350,  351,  352,  353,   -1,   -1,   -1,  357,  358,
   -1,  360,   -1,  362,   -1,  364,  365,  366,   -1,  368,
   -1,  370,  371,  372,   -1,  374,  375,  376,  377,  378,
   -1,  380,  381,   -1,   -1,   -1,   -1,  386,   -1,  388,
  389,   -1,   -1,  392,  393,  394,   -1,   -1,  397,   -1,
  399,   -1,   -1,  402,  403,  404,  405,   -1,  407,   -1,
  409,   -1,  411,  412,  413,  414,  415,   -1,   -1,  418,
  419,  420,   -1,  422,  423,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  433,  434,  435,   -1,  437,  438,
  439,   -1,   -1,   -1,   -1,   -1,  445,   -1,   -1,   -1,
  449,   -1,  451,  452,  453,  454,   -1,  456,   -1,  458,
  459,  460,  461,  462,  463,  464,  465,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  475,  476,   -1,  478,
  479,  480,   -1,   -1,  602,   -1,  485,  486,  487,  488,
  489,   -1,  491,   -1,   -1,  494,  495,   -1,   -1,   -1,
  499,  500,   -1,   -1,  503,  504,  505,  506,  507,  508,
  509,  510,  511,   -1,  513,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  523,  524,  525,  526,  527,  528,
  529,  530,  531,  532,  533,  534,  535,  536,  537,  538,
  539,  540,  541,  542,  543,   -1,  545,  546,  547,  548,
  549,  550,   -1,  552,  553,  554,  555,   -1,  557,   -1,
  559,  560,  561,  562,  563,  564,  565,  566,  567,  568,
  569,  570,  571,  572,  573,  574,  575,  576,   -1,  578,
  579,  580,  581,   -1,  583,  584,  585,  586,  587,   -1,
  589,  590,  591,  592,  593,  594,  595,  596,  597,  598,
  599,  600,  601,  602,  603,  604,  605,  606,  607,  608,
  609,  610,  611,  612,  613,  614,  615,  616,  617,   -1,
  619,  620,   -1,   -1,   -1,  624,  625,  626,  627,  628,
  629,   -1,   -1,  632,  633,  634,  635,  636,  637,  638,
  639,  640,   41,   -1,   -1,   -1,   -1,  646,   -1,  648,
  649,  650,  651,  652,  653,   -1,  655,   -1,   -1,   -1,
   -1,   60,   61,   62,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  672,  673,   -1,  675,  676,  677,  678,
  679,  680,  681,  682,  683,   -1,  685,  686,  687,  688,
  689,  690,  691,  692,  693,  694,  695,  696,   -1,  698,
  699,  700,   -1,  702,   -1,   -1,   -1,  706,  707,  708,
  709,  710,  711,  712,  713,  714,  715,  716,  717,  718,
    0,  720,  721,   -1,   -1,   -1,   -1,   -1,  727,  728,
  729,  730,  731,  732,  733,  734,  735,  736,  737,  738,
  739,  740,  741,  742,  743,  744,   -1,   -1,  747,   -1,
  749,  750,  751,  752,  753,  754,   -1,   -1,   -1,   -1,
   40,   41,   42,   43,   44,   45,   46,   47,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   58,   59,
   60,   61,   62,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   91,   -1,   93,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  262,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  271,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  284,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  318,
   -1,   -1,   -1,   -1,   -1,   -1,  325,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  333,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  350,   -1,  352,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  362,   -1,  364,  365,  366,   -1,   -1,
   -1,   -1,  371,   -1,   -1,   -1,   -1,  257,   -1,  259,
   -1,   -1,  262,   -1,  264,  265,  266,   -1,  268,  269,
   -1,  271,   -1,   -1,   -1,   -1,   -1,  277,   -1,  279,
  280,  281,  282,   -1,  284,   -1,   -1,  287,   -1,   -1,
  290,   -1,   -1,  412,   -1,   -1,   -1,  297,   -1,  299,
  300,   -1,  302,  303,  304,  305,  306,   -1,   -1,  309,
   -1,   -1,  312,   -1,  314,  315,   -1,   -1,  318,  319,
  320,   -1,  322,  323,  324,  325,  326,  327,  328,   -1,
   -1,   -1,  332,  333,  334,  335,  336,   -1,   -1,   -1,
  340,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  348,  349,
  350,  351,  352,  353,   -1,   -1,   -1,  357,  358,   -1,
  360,   -1,  362,   -1,  364,  365,  366,   -1,  368,   -1,
  370,  371,  372,   -1,  374,  375,  376,  377,  378,   -1,
  380,  381,   -1,   -1,   -1,   -1,  386,   -1,  388,  389,
   -1,   -1,  392,  393,  394,   -1,   -1,  397,   -1,  399,
   -1,   -1,  402,  403,  404,  405,   -1,  407,   -1,  409,
   -1,  411,  412,  413,  414,  415,   -1,   -1,  418,  419,
  420,   -1,  422,  423,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  433,  434,  435,   -1,  437,  438,  439,
   -1,   -1,   -1,   -1,   -1,  445,   -1,   -1,   -1,  449,
   -1,  451,  452,  453,  454,   -1,  456,   -1,  458,  459,
  460,  461,  462,  463,  464,  465,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  475,  476,   -1,  478,  479,
  480,   -1,   -1,  602,   -1,  485,  486,  487,  488,  489,
   -1,  491,   -1,   -1,  494,  495,   -1,   -1,   -1,  499,
  500,   -1,   -1,  503,  504,  505,  506,  507,  508,  509,
  510,  511,   -1,  513,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  523,  524,  525,  526,  527,  528,  529,
  530,  531,  532,  533,  534,  535,  536,  537,  538,  539,
  540,  541,  542,  543,   -1,  545,  546,  547,  548,  549,
  550,   -1,  552,  553,  554,  555,   -1,  557,   -1,  559,
  560,  561,  562,  563,  564,  565,  566,  567,  568,  569,
  570,  571,  572,  573,  574,  575,  576,   -1,  578,  579,
  580,  581,   -1,  583,  584,  585,  586,  587,   -1,  589,
  590,  591,  592,  593,  594,  595,  596,  597,  598,  599,
  600,  601,  602,  603,  604,  605,  606,  607,  608,  609,
  610,  611,  612,  613,  614,  615,  616,  617,   -1,  619,
  620,   -1,   -1,   -1,  624,  625,  626,  627,  628,  629,
   -1,   -1,  632,  633,  634,  635,  636,  637,  638,  639,
  640,   41,   -1,   -1,   -1,   -1,  646,   -1,  648,  649,
  650,  651,  652,  653,   -1,  655,   -1,   -1,   -1,   -1,
   60,   61,   62,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  672,  673,   -1,  675,  676,  677,  678,  679,
  680,  681,  682,  683,   -1,  685,  686,  687,  688,  689,
  690,  691,  692,  693,  694,  695,  696,   -1,  698,  699,
  700,   -1,  702,   -1,   -1,   -1,  706,  707,  708,  709,
  710,  711,  712,  713,  714,  715,  716,  717,  718,    0,
  720,  721,   -1,   -1,   -1,   -1,   -1,  727,  728,  729,
  730,  731,  732,  733,  734,  735,  736,  737,  738,  739,
  740,  741,  742,  743,  744,   -1,   -1,  747,   -1,  749,
  750,  751,  752,  753,  754,   -1,   -1,   -1,   -1,   40,
   41,   42,   43,   44,   45,   46,   47,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   58,   59,   60,
   61,   62,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   91,   -1,   93,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  262,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  271,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  284,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  318,   -1,
   -1,   -1,   -1,   -1,   -1,  325,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  333,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  350,   -1,  352,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  362,   -1,  364,  365,  366,   -1,   -1,   -1,
   -1,  371,   -1,   -1,   -1,   -1,  257,   -1,  259,   -1,
   -1,  262,   -1,  264,  265,  266,   -1,  268,  269,   -1,
  271,   -1,   -1,   -1,   -1,   -1,  277,   -1,  279,  280,
  281,  282,   -1,  284,   -1,   -1,  287,   -1,   -1,  290,
   -1,   -1,  412,   -1,   -1,   -1,  297,   -1,  299,  300,
   -1,  302,  303,  304,  305,  306,   -1,   -1,  309,   -1,
   -1,  312,   -1,  314,  315,   -1,   -1,  318,  319,  320,
   -1,  322,  323,  324,  325,  326,  327,  328,   -1,   -1,
   -1,  332,  333,  334,  335,  336,   -1,   -1,   -1,  340,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  348,  349,  350,
  351,  352,  353,   -1,   -1,   -1,  357,  358,   -1,  360,
   -1,  362,   -1,  364,  365,  366,   -1,  368,   -1,  370,
  371,  372,   -1,  374,  375,  376,  377,  378,   -1,  380,
  381,   -1,   -1,   -1,   -1,  386,   -1,  388,  389,   -1,
   -1,  392,  393,  394,   -1,   -1,  397,   -1,  399,   -1,
   -1,  402,  403,  404,  405,   -1,  407,   -1,  409,   -1,
  411,  412,  413,  414,  415,   -1,   -1,  418,  419,  420,
   -1,  422,  423,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  433,  434,  435,   -1,  437,  438,  439,   -1,
   -1,   -1,   -1,   -1,  445,   -1,   -1,   -1,  449,   -1,
  451,  452,  453,  454,   -1,  456,   -1,  458,  459,  460,
  461,  462,  463,  464,  465,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  475,  476,   -1,  478,  479,  480,
   -1,   -1,  602,   -1,  485,  486,  487,  488,  489,   -1,
  491,   -1,   -1,  494,  495,   -1,   -1,   -1,  499,  500,
   -1,   -1,  503,  504,  505,  506,  507,  508,  509,  510,
  511,   -1,  513,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  523,  524,  525,  526,  527,  528,  529,  530,
  531,  532,  533,  534,  535,  536,  537,  538,  539,  540,
  541,  542,  543,   -1,  545,  546,  547,  548,  549,  550,
   -1,  552,  553,  554,  555,   -1,  557,   -1,  559,  560,
  561,  562,  563,  564,  565,  566,  567,  568,  569,  570,
  571,  572,  573,  574,  575,  576,   -1,  578,  579,  580,
  581,   -1,  583,  584,  585,  586,  587,   -1,  589,  590,
  591,  592,  593,  594,  595,  596,  597,  598,  599,  600,
  601,  602,  603,  604,  605,  606,  607,  608,  609,  610,
  611,  612,  613,  614,  615,  616,  617,   -1,  619,  620,
   -1,   -1,   -1,  624,  625,  626,  627,  628,  629,   -1,
   -1,  632,  633,  634,  635,  636,  637,  638,  639,  640,
   41,   -1,   -1,   -1,   -1,  646,   -1,  648,  649,  650,
  651,  652,  653,   -1,  655,   -1,   -1,   -1,   -1,   60,
   61,   62,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  672,  673,   -1,  675,  676,  677,  678,  679,  680,
  681,  682,  683,   -1,  685,  686,  687,  688,  689,  690,
  691,  692,  693,  694,  695,  696,   -1,  698,  699,  700,
   -1,  702,   -1,   -1,   -1,  706,  707,  708,  709,  710,
  711,  712,  713,  714,  715,  716,  717,  718,    0,  720,
  721,   -1,   -1,   -1,   -1,   -1,  727,  728,  729,  730,
  731,  732,  733,  734,  735,  736,  737,  738,  739,  740,
  741,  742,  743,  744,   -1,   -1,  747,   -1,  749,  750,
  751,  752,  753,  754,   -1,   -1,   -1,   -1,   40,   41,
   42,   43,   44,   45,   46,   47,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   58,   59,   60,   61,
   62,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   91,
   -1,   93,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  262,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  271,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  284,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  318,   -1,   -1,
   -1,   -1,   -1,   -1,  325,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  333,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  350,
   -1,  352,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  362,   -1,  364,  365,  366,   -1,   -1,   -1,   -1,
  371,   -1,   -1,   -1,   -1,  257,   -1,  259,   -1,   -1,
  262,   -1,  264,  265,  266,   -1,  268,  269,   -1,  271,
   -1,   -1,   -1,   -1,   -1,  277,   -1,  279,  280,  281,
  282,   -1,  284,   -1,   -1,  287,   -1,   -1,  290,   -1,
   -1,  412,   -1,   -1,   -1,  297,   -1,  299,  300,   -1,
  302,  303,  304,  305,  306,   -1,   -1,  309,   -1,   -1,
  312,   -1,  314,  315,   -1,   -1,  318,  319,  320,   -1,
  322,  323,  324,  325,  326,  327,  328,   -1,   -1,   -1,
  332,  333,  334,  335,  336,   -1,   -1,   -1,  340,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  348,  349,  350,  351,
  352,  353,   -1,   -1,   -1,  357,  358,   -1,  360,   -1,
  362,   -1,  364,  365,  366,   -1,  368,   -1,  370,  371,
  372,   -1,  374,  375,  376,  377,  378,   -1,  380,  381,
   -1,   -1,   -1,   -1,  386,   -1,  388,  389,   -1,   -1,
  392,  393,  394,   -1,   -1,  397,   -1,  399,   -1,   -1,
  402,  403,  404,  405,   -1,  407,   -1,  409,   -1,  411,
  412,  413,  414,  415,   -1,   -1,  418,  419,  420,   -1,
  422,  423,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  433,  434,  435,   -1,  437,  438,  439,   -1,   -1,
   -1,   -1,   -1,  445,   -1,   -1,   -1,  449,   -1,  451,
  452,  453,  454,   -1,  456,   -1,  458,  459,  460,  461,
  462,  463,  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  475,  476,   -1,  478,  479,  480,   -1,
   -1,  602,   -1,  485,  486,  487,  488,  489,   -1,  491,
   -1,   -1,  494,  495,   -1,   -1,   -1,  499,  500,   -1,
   -1,  503,  504,  505,  506,  507,  508,  509,  510,  511,
   -1,  513,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  523,  524,  525,  526,  527,  528,  529,  530,  531,
  532,  533,  534,  535,  536,  537,  538,  539,  540,  541,
  542,  543,   -1,  545,  546,  547,  548,  549,  550,   -1,
  552,  553,  554,  555,   -1,  557,   -1,  559,  560,  561,
  562,  563,  564,  565,  566,  567,  568,  569,  570,  571,
  572,  573,  574,  575,  576,   -1,  578,  579,  580,  581,
   -1,  583,  584,  585,  586,  587,   -1,  589,  590,  591,
  592,  593,  594,  595,  596,  597,  598,  599,  600,  601,
  602,  603,  604,  605,  606,  607,  608,  609,  610,  611,
  612,  613,  614,  615,  616,  617,   -1,  619,  620,   -1,
   -1,   -1,  624,  625,  626,  627,  628,  629,   -1,   -1,
  632,  633,  634,  635,  636,  637,  638,  639,  640,   41,
   -1,   -1,   -1,   -1,  646,   -1,  648,  649,  650,  651,
  652,  653,   -1,  655,   -1,   -1,   -1,   -1,   60,   61,
   62,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  672,  673,   -1,  675,  676,  677,  678,  679,  680,  681,
  682,  683,   -1,  685,  686,  687,  688,  689,  690,  691,
  692,  693,  694,  695,  696,   -1,  698,  699,  700,   -1,
  702,   -1,   -1,   -1,  706,  707,  708,  709,  710,  711,
  712,  713,  714,  715,  716,  717,  718,    0,  720,  721,
   -1,   -1,   -1,   -1,   -1,  727,  728,  729,  730,  731,
  732,  733,  734,  735,  736,  737,  738,  739,  740,  741,
  742,  743,  744,   -1,   -1,  747,   -1,  749,  750,  751,
  752,  753,  754,   -1,   -1,   -1,   -1,   40,   41,   42,
   43,   44,   45,   46,   47,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   58,   59,   60,   61,   62,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   91,   -1,
   93,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  262,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  271,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  284,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  318,   -1,   -1,   -1,
   -1,   -1,   -1,  325,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  333,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  350,   -1,
  352,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  362,   -1,  364,  365,  366,   -1,   -1,   -1,   -1,  371,
   -1,   -1,   -1,   -1,  257,   -1,  259,   -1,   -1,  262,
   -1,  264,  265,  266,   -1,  268,  269,   -1,  271,   -1,
   -1,   -1,   -1,   -1,  277,   -1,  279,  280,  281,  282,
   -1,  284,   -1,   -1,  287,   -1,   -1,  290,   -1,   -1,
  412,   -1,   -1,   -1,  297,   -1,  299,  300,   -1,  302,
  303,  304,  305,  306,   -1,   -1,  309,   -1,   -1,  312,
   -1,  314,  315,   -1,   -1,  318,  319,  320,   -1,  322,
  323,  324,  325,  326,  327,  328,   -1,   -1,   -1,  332,
  333,  334,  335,  336,   -1,   -1,   -1,  340,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  348,  349,  350,  351,  352,
  353,   -1,   -1,   -1,  357,  358,   -1,  360,   -1,  362,
   -1,  364,  365,  366,   -1,  368,   -1,  370,  371,  372,
   -1,  374,  375,  376,  377,  378,   -1,  380,  381,   -1,
   -1,   -1,   -1,  386,   -1,  388,  389,   -1,   -1,  392,
  393,  394,   -1,   -1,  397,   -1,  399,   -1,   -1,  402,
  403,  404,  405,   -1,  407,   -1,  409,   -1,  411,  412,
  413,  414,  415,   -1,   -1,  418,  419,  420,   -1,  422,
  423,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  433,  434,  435,   -1,  437,  438,  439,   -1,   -1,   -1,
   -1,   -1,  445,   -1,   -1,   -1,  449,   -1,  451,  452,
  453,  454,   -1,  456,   -1,  458,  459,  460,  461,  462,
  463,  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  475,  476,   -1,  478,  479,  480,   -1,   -1,
  602,   -1,  485,  486,  487,  488,  489,   -1,  491,   -1,
   -1,  494,  495,   -1,   -1,   -1,  499,  500,   -1,   -1,
  503,  504,  505,  506,  507,  508,  509,  510,  511,   -1,
  513,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  523,  524,  525,  526,  527,  528,  529,  530,  531,  532,
  533,  534,  535,  536,  537,  538,  539,  540,  541,  542,
  543,   -1,  545,  546,  547,  548,  549,  550,   -1,  552,
  553,  554,  555,   -1,  557,   -1,  559,  560,  561,  562,
  563,  564,  565,  566,  567,  568,  569,  570,  571,  572,
  573,  574,  575,  576,   -1,  578,  579,  580,  581,   -1,
  583,  584,  585,  586,  587,   -1,  589,  590,  591,  592,
  593,  594,  595,  596,  597,  598,  599,  600,  601,  602,
  603,  604,  605,  606,  607,  608,  609,  610,  611,  612,
  613,  614,  615,  616,  617,   -1,  619,  620,   -1,   -1,
   -1,  624,  625,  626,  627,  628,  629,   -1,   -1,  632,
  633,  634,  635,  636,  637,  638,  639,  640,   41,   -1,
   -1,   -1,   -1,  646,   -1,  648,  649,  650,  651,  652,
  653,   -1,  655,   -1,   -1,   -1,   -1,   60,   61,   62,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  672,
  673,   -1,  675,  676,  677,  678,  679,  680,  681,  682,
  683,   -1,  685,  686,  687,  688,  689,  690,  691,  692,
  693,  694,  695,  696,   -1,  698,  699,  700,   -1,  702,
   -1,   -1,   -1,  706,  707,  708,  709,  710,  711,  712,
  713,  714,  715,  716,  717,  718,    0,  720,  721,   -1,
   -1,   -1,   -1,   -1,  727,  728,  729,  730,  731,  732,
  733,  734,  735,  736,  737,  738,  739,  740,  741,  742,
  743,  744,   -1,   -1,  747,   -1,  749,  750,  751,  752,
  753,  754,   -1,   -1,   -1,   -1,   40,   41,   42,   43,
   44,   45,   46,   47,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   58,   59,   60,   61,   62,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   91,   -1,   93,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  262,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  271,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  284,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  318,   -1,   -1,   -1,   -1,
   -1,   -1,  325,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  333,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  350,   -1,  352,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  362,
   -1,  364,  365,  366,   -1,   -1,   -1,   -1,  371,   -1,
   -1,   -1,   -1,  257,   -1,  259,   -1,   -1,  262,   -1,
  264,  265,  266,   -1,  268,  269,   -1,  271,   -1,   -1,
   -1,   -1,   -1,  277,   -1,  279,  280,  281,  282,   -1,
  284,   -1,   -1,  287,   -1,   -1,  290,   -1,   -1,  412,
   -1,   -1,   -1,  297,   -1,  299,  300,   -1,  302,  303,
  304,  305,  306,   -1,   -1,  309,   -1,   -1,  312,   -1,
  314,  315,   -1,   -1,  318,  319,  320,   -1,  322,  323,
  324,  325,  326,  327,  328,   -1,   -1,   -1,  332,  333,
  334,  335,  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  348,  349,  350,  351,  352,  353,
   -1,   -1,   -1,  357,  358,   -1,  360,   -1,  362,   -1,
  364,  365,  366,   -1,  368,   -1,  370,  371,  372,   -1,
  374,  375,  376,  377,  378,   -1,  380,  381,   -1,   -1,
   -1,   -1,  386,   -1,  388,  389,   -1,   -1,  392,  393,
  394,   -1,   -1,  397,   -1,  399,   -1,   -1,  402,  403,
  404,  405,   -1,  407,   -1,  409,   -1,  411,  412,  413,
  414,  415,   -1,   -1,  418,  419,  420,   -1,  422,  423,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  433,
  434,  435,   -1,  437,  438,  439,   -1,   -1,   -1,   -1,
   -1,  445,   -1,   -1,   -1,  449,   -1,  451,  452,  453,
  454,   -1,  456,   -1,  458,  459,  460,  461,  462,  463,
  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  475,  476,   -1,  478,  479,  480,   -1,   -1,  602,
   -1,  485,  486,  487,  488,  489,   -1,  491,   -1,   -1,
  494,  495,   -1,   -1,   -1,  499,  500,   -1,   -1,  503,
  504,  505,  506,  507,  508,  509,  510,  511,   -1,  513,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  523,
  524,  525,  526,  527,  528,  529,  530,  531,  532,  533,
  534,  535,  536,  537,  538,  539,  540,  541,  542,  543,
   -1,  545,  546,  547,  548,  549,  550,   -1,  552,  553,
  554,  555,   -1,  557,   -1,  559,  560,  561,  562,  563,
  564,  565,  566,  567,  568,  569,  570,  571,  572,  573,
  574,  575,  576,   -1,  578,  579,  580,  581,   -1,  583,
  584,  585,  586,  587,   -1,  589,  590,  591,  592,  593,
  594,  595,  596,  597,  598,  599,  600,  601,  602,  603,
  604,  605,  606,  607,  608,  609,  610,  611,  612,  613,
  614,  615,  616,  617,   -1,  619,  620,   -1,   -1,   -1,
  624,  625,  626,  627,  628,  629,   -1,   -1,  632,  633,
  634,  635,  636,  637,  638,  639,  640,   41,   -1,   -1,
   -1,   -1,  646,   -1,  648,  649,  650,  651,  652,  653,
   -1,  655,   -1,   -1,   -1,   -1,   60,   61,   62,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  672,  673,
   -1,  675,  676,  677,  678,  679,  680,  681,  682,  683,
   -1,  685,  686,  687,  688,  689,  690,  691,  692,  693,
  694,  695,  696,   -1,  698,  699,  700,   -1,  702,   -1,
   -1,   -1,  706,  707,  708,  709,  710,  711,  712,  713,
  714,  715,  716,  717,  718,    0,  720,  721,   -1,   -1,
   -1,   -1,   -1,  727,  728,  729,  730,  731,  732,  733,
  734,  735,  736,  737,  738,  739,  740,  741,  742,  743,
  744,   -1,   -1,  747,   -1,  749,  750,  751,  752,  753,
  754,   -1,   -1,   -1,   -1,   40,   41,   42,   43,   44,
   45,   46,   47,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   58,   59,   60,   61,   62,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   91,   -1,   93,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  262,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  271,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  284,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  318,   -1,   -1,   -1,   -1,   -1,
   -1,  325,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  333,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  350,   -1,  352,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  362,   -1,
  364,  365,  366,   -1,   -1,   -1,   -1,  371,   -1,   -1,
   -1,   -1,  257,   -1,  259,   -1,   -1,  262,   -1,  264,
  265,  266,   -1,  268,  269,   -1,  271,   -1,   -1,   -1,
   -1,   -1,  277,   -1,  279,  280,  281,  282,   -1,  284,
   -1,   -1,  287,   -1,   -1,  290,   -1,   -1,  412,   -1,
   -1,   -1,  297,   -1,  299,  300,   -1,  302,  303,  304,
  305,  306,   -1,   -1,  309,   -1,   -1,  312,   -1,  314,
  315,   -1,   -1,  318,  319,  320,   -1,  322,  323,  324,
  325,  326,  327,  328,   -1,   -1,   -1,  332,  333,  334,
  335,  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  348,  349,  350,  351,  352,  353,   -1,
   -1,   -1,  357,  358,   -1,  360,   -1,  362,   -1,  364,
  365,  366,   -1,  368,   -1,  370,  371,  372,   -1,  374,
  375,  376,  377,  378,   -1,  380,  381,   -1,   -1,   -1,
   -1,  386,   -1,  388,  389,   -1,   -1,  392,  393,  394,
   -1,   -1,  397,   -1,  399,   -1,   -1,  402,  403,  404,
  405,   -1,  407,   -1,  409,   -1,  411,  412,  413,  414,
  415,   -1,   -1,  418,  419,  420,   -1,  422,  423,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  433,  434,
  435,   -1,  437,  438,  439,   -1,   -1,   -1,   -1,   -1,
  445,   -1,   -1,   -1,  449,   -1,  451,  452,  453,  454,
   -1,  456,   -1,  458,  459,  460,  461,  462,  463,  464,
  465,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  475,  476,   -1,  478,  479,  480,   -1,   -1,  602,   -1,
  485,  486,  487,  488,  489,   -1,  491,   -1,   -1,  494,
  495,   -1,   -1,   -1,  499,  500,   -1,   -1,  503,  504,
  505,  506,  507,  508,  509,  510,  511,   -1,  513,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  523,  524,
  525,  526,  527,  528,  529,  530,  531,  532,  533,  534,
  535,  536,  537,  538,  539,  540,  541,  542,  543,   -1,
  545,  546,  547,  548,  549,  550,   -1,  552,  553,  554,
  555,   -1,  557,   -1,  559,  560,  561,  562,  563,  564,
  565,  566,  567,  568,  569,  570,  571,  572,  573,  574,
  575,  576,   -1,  578,  579,  580,  581,   -1,  583,  584,
  585,  586,  587,   -1,  589,  590,  591,  592,  593,  594,
  595,  596,  597,  598,  599,  600,  601,  602,  603,  604,
  605,  606,  607,  608,  609,  610,  611,  612,  613,  614,
  615,  616,  617,   -1,  619,  620,   -1,   -1,   -1,  624,
  625,  626,  627,  628,  629,   -1,   -1,  632,  633,  634,
  635,  636,  637,  638,  639,  640,   41,   -1,   -1,   -1,
   -1,  646,   -1,  648,  649,  650,  651,  652,  653,   -1,
  655,   -1,   -1,   -1,   -1,   60,   61,   62,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  672,  673,   -1,
  675,  676,  677,  678,  679,  680,  681,  682,  683,   -1,
  685,  686,  687,  688,  689,  690,  691,  692,  693,  694,
  695,  696,   -1,  698,  699,  700,   -1,  702,   -1,   -1,
   -1,  706,  707,  708,  709,  710,  711,  712,  713,  714,
  715,  716,  717,  718,    0,  720,  721,   -1,   -1,   -1,
   -1,   -1,  727,  728,  729,  730,  731,  732,  733,  734,
  735,  736,  737,  738,  739,  740,  741,  742,  743,  744,
   -1,   -1,  747,   -1,  749,  750,  751,  752,  753,  754,
   -1,   -1,   -1,   -1,   40,   41,   42,   43,   44,   45,
   46,   47,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   58,   59,   60,   61,   62,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   91,   -1,   93,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  262,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  271,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  284,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  318,   -1,   -1,   -1,   -1,   -1,   -1,
  325,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  333,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  350,   -1,  352,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  362,   -1,  364,
  365,  366,   -1,   -1,   -1,   -1,  371,   -1,   -1,   -1,
   -1,  257,   -1,  259,   -1,   -1,  262,   -1,  264,  265,
  266,   -1,  268,  269,   -1,  271,   -1,   -1,   -1,   -1,
   -1,  277,   -1,  279,  280,  281,  282,   -1,  284,   -1,
   -1,  287,   -1,   -1,  290,   -1,   -1,  412,   -1,   -1,
   -1,  297,   -1,  299,  300,   -1,  302,  303,  304,  305,
  306,   -1,   -1,  309,   -1,   -1,  312,   -1,  314,  315,
   -1,   -1,  318,  319,  320,   -1,  322,  323,  324,  325,
  326,  327,  328,   -1,   -1,   -1,  332,  333,  334,  335,
  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  348,  349,  350,  351,  352,  353,   -1,   -1,
   -1,  357,  358,   -1,  360,   -1,  362,   -1,  364,  365,
  366,   -1,  368,   -1,  370,  371,  372,   -1,  374,  375,
  376,  377,  378,   -1,  380,  381,   -1,   -1,   -1,   -1,
  386,   -1,  388,  389,   -1,   -1,  392,  393,  394,   -1,
   -1,  397,   -1,  399,   -1,   -1,  402,  403,  404,  405,
   -1,  407,   -1,  409,   -1,  411,  412,  413,  414,  415,
   -1,   -1,  418,  419,  420,   -1,  422,  423,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  433,  434,  435,
   -1,  437,  438,  439,   -1,   -1,   -1,   -1,   -1,  445,
   -1,   -1,   -1,  449,   -1,  451,  452,  453,  454,   -1,
  456,   -1,  458,  459,  460,  461,  462,  463,  464,  465,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  475,
  476,   -1,  478,  479,  480,   -1,   -1,  602,   -1,  485,
  486,  487,  488,  489,   -1,  491,   -1,   -1,  494,  495,
   -1,   -1,   -1,  499,  500,   -1,   -1,  503,  504,  505,
  506,  507,  508,  509,  510,  511,   -1,  513,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  523,  524,  525,
  526,  527,  528,  529,  530,  531,  532,  533,  534,  535,
  536,  537,  538,  539,  540,  541,  542,  543,   -1,  545,
  546,  547,  548,  549,  550,   -1,  552,  553,  554,  555,
   -1,  557,   -1,  559,  560,  561,  562,  563,  564,  565,
  566,  567,  568,  569,  570,  571,  572,  573,  574,  575,
  576,   -1,  578,  579,  580,  581,   -1,  583,  584,  585,
  586,  587,   -1,  589,  590,  591,  592,  593,  594,  595,
  596,  597,  598,  599,  600,  601,  602,  603,  604,  605,
  606,  607,  608,  609,  610,  611,  612,  613,  614,  615,
  616,  617,   -1,  619,  620,   -1,   -1,   -1,  624,  625,
  626,  627,  628,  629,   -1,   -1,  632,  633,  634,  635,
  636,  637,  638,  639,  640,   41,   -1,   -1,   -1,   -1,
  646,   -1,  648,  649,  650,  651,  652,  653,   -1,  655,
   -1,   -1,   -1,   -1,   60,   61,   62,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  672,  673,   -1,  675,
  676,  677,  678,  679,  680,  681,  682,  683,   -1,  685,
  686,  687,  688,  689,  690,  691,  692,  693,  694,  695,
  696,   -1,  698,  699,  700,   -1,  702,   -1,   -1,   -1,
  706,  707,  708,  709,  710,  711,  712,  713,  714,  715,
  716,  717,  718,    0,  720,  721,   -1,   -1,   -1,   -1,
   -1,  727,  728,  729,  730,  731,  732,  733,  734,  735,
  736,  737,  738,  739,  740,  741,  742,  743,  744,   -1,
   -1,  747,   -1,  749,  750,  751,  752,  753,  754,   -1,
   -1,   -1,   -1,   40,   41,   42,   43,   44,   45,   46,
   47,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   58,   59,   60,   61,   62,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   91,   -1,   93,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  262,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  271,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  284,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  318,   -1,   -1,   -1,   -1,   -1,   -1,  325,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  333,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  350,   -1,  352,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  362,   -1,  364,  365,
  366,   -1,   -1,   -1,   -1,  371,   -1,   -1,   -1,   -1,
  257,   -1,  259,   -1,   -1,  262,   -1,  264,  265,  266,
   -1,  268,  269,   -1,  271,   -1,   -1,   -1,   -1,   -1,
  277,   -1,  279,  280,  281,  282,   -1,  284,   -1,   -1,
  287,   -1,   -1,  290,   -1,   -1,  412,   -1,   -1,   -1,
  297,   -1,  299,  300,   -1,  302,  303,  304,  305,  306,
   -1,   -1,  309,   -1,   -1,  312,   -1,  314,  315,   -1,
   -1,  318,  319,  320,   -1,  322,  323,  324,  325,  326,
  327,  328,   -1,   -1,   -1,  332,  333,  334,  335,  336,
   -1,   -1,   -1,  340,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  348,  349,  350,  351,  352,  353,   -1,   -1,   -1,
  357,  358,   -1,  360,   -1,  362,   -1,  364,  365,  366,
   -1,  368,   -1,  370,  371,  372,   -1,  374,  375,  376,
  377,  378,   -1,  380,  381,   -1,   -1,   -1,   -1,  386,
   -1,  388,  389,   -1,   -1,  392,  393,  394,   -1,   -1,
  397,   -1,  399,   -1,   -1,  402,  403,  404,  405,   -1,
  407,   -1,  409,   -1,  411,  412,  413,  414,  415,   -1,
   -1,  418,  419,  420,   -1,  422,  423,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  433,  434,  435,   -1,
  437,  438,  439,   -1,   -1,   -1,   -1,   -1,  445,   -1,
   -1,   -1,  449,   -1,  451,  452,  453,  454,   -1,  456,
   -1,  458,  459,  460,  461,  462,  463,  464,  465,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  475,  476,
   -1,  478,  479,  480,   -1,   -1,  602,   -1,  485,  486,
  487,  488,  489,   -1,  491,   -1,   -1,  494,  495,   -1,
   -1,   -1,  499,  500,   -1,   -1,  503,  504,  505,  506,
  507,  508,  509,  510,  511,   -1,  513,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  523,  524,  525,  526,
  527,  528,  529,  530,  531,  532,  533,  534,  535,  536,
  537,  538,  539,  540,  541,  542,  543,   -1,  545,  546,
  547,  548,  549,  550,   -1,  552,  553,  554,  555,   -1,
  557,   -1,  559,  560,  561,  562,  563,  564,  565,  566,
  567,  568,  569,  570,  571,  572,  573,  574,  575,  576,
   -1,  578,  579,  580,  581,   -1,  583,  584,  585,  586,
  587,   -1,  589,  590,  591,  592,  593,  594,  595,  596,
  597,  598,  599,  600,  601,  602,  603,  604,  605,  606,
  607,  608,  609,  610,  611,  612,  613,  614,  615,  616,
  617,   -1,  619,  620,   -1,   -1,   -1,  624,  625,  626,
  627,  628,  629,   -1,   -1,  632,  633,  634,  635,  636,
  637,  638,  639,  640,   41,   -1,   -1,   -1,   -1,  646,
   -1,  648,  649,  650,  651,  652,  653,   -1,  655,   -1,
   -1,   -1,   -1,   60,   61,   62,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  672,  673,   -1,  675,  676,
  677,  678,  679,  680,  681,  682,  683,   -1,  685,  686,
  687,  688,  689,  690,  691,  692,  693,  694,  695,  696,
   -1,  698,  699,  700,   -1,  702,   -1,   -1,   -1,  706,
  707,  708,  709,  710,  711,  712,  713,  714,  715,  716,
  717,  718,    0,  720,  721,   -1,   -1,   -1,   -1,   -1,
  727,  728,  729,  730,  731,  732,  733,  734,  735,  736,
  737,  738,  739,  740,  741,  742,  743,  744,   -1,   -1,
  747,   -1,  749,  750,  751,  752,  753,  754,   -1,   -1,
   -1,   -1,   40,   41,   42,   43,   44,   45,   46,   47,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   58,   59,   60,   61,   62,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   91,   -1,   93,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  262,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  271,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  284,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  318,   -1,   -1,   -1,   -1,   -1,   -1,  325,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  333,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  350,   -1,  352,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  362,   -1,  364,  365,  366,
   -1,   -1,   -1,   -1,  371,   -1,   -1,   -1,   -1,  257,
   -1,  259,   -1,   -1,  262,   -1,  264,  265,  266,   -1,
  268,  269,   -1,  271,   -1,   -1,   -1,   -1,   -1,  277,
   -1,  279,  280,  281,  282,   -1,  284,   -1,   -1,  287,
   -1,   -1,  290,   -1,   -1,  412,   -1,   -1,   -1,  297,
   -1,  299,  300,   -1,  302,  303,  304,  305,  306,   -1,
   -1,  309,   -1,   -1,  312,   -1,  314,  315,   -1,   -1,
  318,  319,  320,   -1,  322,  323,  324,  325,  326,  327,
  328,   -1,   -1,   -1,  332,  333,  334,  335,  336,   -1,
   -1,   -1,  340,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  348,  349,  350,  351,  352,  353,   -1,   -1,   -1,  357,
  358,   -1,  360,   -1,  362,   -1,  364,  365,  366,   -1,
  368,   -1,  370,  371,  372,   -1,  374,  375,  376,  377,
  378,   -1,  380,  381,   -1,   -1,   -1,   -1,  386,   -1,
  388,  389,   -1,   -1,  392,  393,  394,   -1,   -1,  397,
   -1,  399,   -1,   -1,  402,  403,  404,  405,   -1,  407,
   -1,  409,   -1,  411,  412,  413,  414,  415,   -1,   -1,
  418,  419,  420,   -1,  422,  423,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  433,  434,  435,   -1,  437,
  438,  439,   -1,   -1,   -1,   -1,   -1,  445,   -1,   -1,
   -1,  449,   -1,  451,  452,  453,  454,   -1,  456,   -1,
  458,  459,  460,  461,  462,  463,  464,  465,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  475,  476,   -1,
  478,  479,  480,   -1,   -1,  602,   -1,  485,  486,  487,
  488,  489,   -1,  491,   -1,   -1,  494,  495,   -1,   -1,
   -1,  499,  500,   -1,   -1,  503,  504,  505,  506,  507,
  508,  509,  510,  511,   -1,  513,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  523,  524,  525,  526,  527,
  528,  529,  530,  531,  532,  533,  534,  535,  536,  537,
  538,  539,  540,  541,  542,  543,   -1,  545,  546,  547,
  548,  549,  550,   -1,  552,  553,  554,  555,   -1,  557,
   -1,  559,  560,  561,  562,  563,  564,  565,  566,  567,
  568,  569,  570,  571,  572,  573,  574,  575,  576,   -1,
  578,  579,  580,  581,   -1,  583,  584,  585,  586,  587,
   -1,  589,  590,  591,  592,  593,  594,  595,  596,  597,
  598,  599,  600,  601,  602,  603,  604,  605,  606,  607,
  608,  609,  610,  611,  612,  613,  614,  615,  616,  617,
   -1,  619,  620,   -1,   -1,   -1,  624,  625,  626,  627,
  628,  629,   -1,   -1,  632,  633,  634,  635,  636,  637,
  638,  639,  640,   41,   -1,   -1,   -1,   -1,  646,   -1,
  648,  649,  650,  651,  652,  653,   -1,  655,   -1,   -1,
   -1,   -1,   60,   61,   62,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  672,  673,   -1,  675,  676,  677,
  678,  679,  680,  681,  682,  683,   -1,  685,  686,  687,
  688,  689,  690,  691,  692,  693,  694,  695,  696,   -1,
  698,  699,  700,   -1,  702,   -1,   -1,   -1,  706,  707,
  708,  709,  710,  711,  712,  713,  714,  715,  716,  717,
  718,    0,  720,  721,   -1,   -1,   -1,   -1,   -1,  727,
  728,  729,  730,  731,  732,  733,  734,  735,  736,  737,
  738,  739,  740,  741,  742,  743,  744,   -1,   -1,  747,
   -1,  749,  750,  751,  752,  753,  754,   -1,   -1,   -1,
   -1,   40,   41,   42,   43,   44,   45,   46,   47,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   58,
   59,   60,   61,   62,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   91,   -1,   93,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  262,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  271,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  284,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  318,   -1,   -1,   -1,   -1,   -1,   -1,  325,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  333,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  350,   -1,  352,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  362,   -1,  364,  365,  366,   -1,
   -1,   -1,   -1,  371,   -1,   -1,   -1,   -1,  257,   -1,
  259,   -1,   -1,  262,   -1,  264,  265,  266,   -1,  268,
  269,   -1,  271,   -1,   -1,   -1,   -1,   -1,  277,   -1,
  279,  280,  281,  282,   -1,  284,   -1,   -1,  287,   -1,
   -1,  290,   -1,   -1,  412,   -1,   -1,   -1,  297,   -1,
  299,  300,   -1,  302,  303,  304,  305,  306,   -1,   -1,
  309,   -1,   -1,  312,   -1,  314,  315,   -1,   -1,  318,
  319,  320,   -1,  322,  323,  324,  325,  326,  327,  328,
   -1,   -1,   -1,  332,  333,  334,  335,  336,   -1,   -1,
   -1,  340,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  348,
  349,  350,  351,  352,  353,   -1,   -1,   -1,  357,  358,
   -1,  360,   -1,  362,   -1,  364,  365,  366,   -1,  368,
   -1,  370,  371,  372,   -1,  374,  375,  376,  377,  378,
   -1,  380,  381,   -1,   -1,   -1,   -1,  386,   -1,  388,
  389,   -1,   -1,  392,  393,  394,   -1,   -1,  397,   -1,
  399,   -1,   -1,  402,  403,  404,  405,   -1,  407,   -1,
  409,   -1,  411,  412,  413,  414,  415,   -1,   -1,  418,
  419,  420,   -1,  422,  423,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  433,  434,  435,   -1,  437,  438,
  439,   -1,   -1,   -1,   -1,   -1,  445,   -1,   -1,   -1,
  449,   -1,  451,  452,  453,  454,   -1,  456,   -1,  458,
  459,  460,  461,  462,  463,  464,  465,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  475,  476,   -1,  478,
  479,  480,   -1,   -1,  602,   -1,  485,  486,  487,  488,
  489,   -1,  491,   -1,   -1,  494,  495,   -1,   -1,   -1,
  499,  500,   -1,   -1,  503,  504,  505,  506,  507,  508,
  509,  510,  511,   -1,  513,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  523,  524,  525,  526,  527,  528,
  529,  530,  531,  532,  533,  534,  535,  536,  537,  538,
  539,  540,  541,  542,  543,   -1,  545,  546,  547,  548,
  549,  550,   -1,  552,  553,  554,  555,   -1,  557,   -1,
  559,  560,  561,  562,  563,  564,  565,  566,  567,  568,
  569,  570,  571,  572,  573,  574,  575,  576,   -1,  578,
  579,  580,  581,   -1,  583,  584,  585,  586,  587,   -1,
  589,  590,  591,  592,  593,  594,  595,  596,  597,  598,
  599,  600,  601,  602,  603,  604,  605,  606,  607,  608,
  609,  610,  611,  612,  613,  614,  615,  616,  617,   -1,
  619,  620,   -1,   -1,   -1,  624,  625,  626,  627,  628,
  629,   -1,   -1,  632,  633,  634,  635,  636,  637,  638,
  639,  640,   41,   -1,   -1,   -1,   -1,  646,   -1,  648,
  649,  650,  651,  652,  653,   -1,  655,   -1,   -1,   -1,
   -1,   60,   61,   62,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  672,  673,   -1,  675,  676,  677,  678,
  679,  680,  681,  682,  683,   -1,  685,  686,  687,  688,
  689,  690,  691,  692,  693,  694,  695,  696,   -1,  698,
  699,  700,   -1,  702,   -1,   -1,   -1,  706,  707,  708,
  709,  710,  711,  712,  713,  714,  715,  716,  717,  718,
    0,  720,  721,   -1,   -1,   -1,   -1,   -1,  727,  728,
  729,  730,  731,  732,  733,  734,  735,  736,  737,  738,
  739,  740,  741,  742,  743,  744,   -1,   -1,  747,   -1,
  749,  750,  751,  752,  753,  754,   -1,   -1,   -1,   -1,
   40,   41,   42,   43,   44,   45,   46,   47,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   58,   59,
   60,   61,   62,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   91,   -1,   93,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  262,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  271,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  284,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  318,
   -1,   -1,   -1,   -1,   -1,   -1,  325,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  333,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  350,   -1,  352,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  362,   -1,  364,  365,  366,   -1,   -1,
   -1,   -1,  371,   -1,   -1,   -1,   -1,  257,   -1,  259,
   -1,   -1,  262,   -1,  264,  265,  266,   -1,  268,  269,
   -1,  271,   -1,   -1,   -1,   -1,   -1,  277,   -1,  279,
  280,  281,  282,   -1,  284,   -1,   -1,  287,   -1,   -1,
  290,   -1,   -1,  412,   -1,   -1,   -1,  297,   -1,  299,
  300,   -1,  302,  303,  304,  305,  306,   -1,   -1,  309,
   -1,   -1,  312,   -1,  314,  315,   -1,   -1,  318,  319,
  320,   -1,  322,  323,  324,  325,  326,  327,  328,   -1,
   -1,   -1,  332,  333,  334,  335,  336,   -1,   -1,   -1,
  340,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  348,  349,
  350,  351,  352,  353,   -1,   -1,   -1,  357,  358,   -1,
  360,   -1,  362,   -1,  364,  365,  366,   -1,  368,   -1,
  370,  371,  372,   -1,  374,  375,  376,  377,  378,   -1,
  380,  381,   -1,   -1,   -1,   -1,  386,   -1,  388,  389,
   -1,   -1,  392,  393,  394,   -1,   -1,  397,   -1,  399,
   -1,   -1,  402,  403,  404,  405,   -1,  407,   -1,  409,
   -1,  411,  412,  413,  414,  415,   -1,   -1,  418,  419,
  420,   -1,  422,  423,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  433,  434,  435,   -1,  437,  438,  439,
   -1,   -1,   -1,   -1,   -1,  445,   -1,   -1,   -1,  449,
   -1,  451,  452,  453,  454,   -1,  456,   -1,  458,  459,
  460,  461,  462,  463,  464,  465,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  475,  476,   -1,  478,  479,
  480,   -1,   -1,  602,   -1,  485,  486,  487,  488,  489,
   -1,  491,   -1,   -1,  494,  495,   -1,   -1,   -1,  499,
  500,   -1,   -1,  503,  504,  505,  506,  507,  508,  509,
  510,  511,   -1,  513,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  523,  524,  525,  526,  527,  528,  529,
  530,  531,  532,  533,  534,  535,  536,  537,  538,  539,
  540,  541,  542,  543,   -1,  545,  546,  547,  548,  549,
  550,   -1,  552,  553,  554,  555,   -1,  557,   -1,  559,
  560,  561,  562,  563,  564,  565,  566,  567,  568,  569,
  570,  571,  572,  573,  574,  575,  576,   -1,  578,  579,
  580,  581,   -1,  583,  584,  585,  586,  587,   -1,  589,
  590,  591,  592,  593,  594,  595,  596,  597,  598,  599,
  600,  601,  602,  603,  604,  605,  606,  607,  608,  609,
  610,  611,  612,  613,  614,  615,  616,  617,   -1,  619,
  620,   -1,   -1,   -1,  624,  625,  626,  627,  628,  629,
   -1,   -1,  632,  633,  634,  635,  636,  637,  638,  639,
  640,   41,   -1,   -1,   -1,   -1,  646,   -1,  648,  649,
  650,  651,  652,  653,   -1,  655,   -1,   -1,   -1,   -1,
   60,   61,   62,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  672,  673,   -1,  675,  676,  677,  678,  679,
  680,  681,  682,  683,   -1,  685,  686,  687,  688,  689,
  690,  691,  692,  693,  694,  695,  696,   -1,  698,  699,
  700,   -1,  702,   -1,   -1,   -1,  706,  707,  708,  709,
  710,  711,  712,  713,  714,  715,  716,  717,  718,    0,
  720,  721,   -1,   -1,   -1,   -1,   -1,  727,  728,  729,
  730,  731,  732,  733,  734,  735,  736,  737,  738,  739,
  740,  741,  742,  743,  744,   -1,   -1,  747,   -1,  749,
  750,  751,  752,  753,  754,   -1,   -1,   -1,   -1,   40,
   41,   42,   43,   44,   45,   46,   47,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   58,   59,   60,
   61,   62,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   91,   -1,   93,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  262,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  271,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  284,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  318,   -1,
   -1,   -1,   -1,   -1,   -1,  325,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  333,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  350,   -1,  352,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  362,   -1,  364,  365,  366,   -1,   -1,   -1,
   -1,  371,   -1,   -1,   -1,   -1,  257,   -1,  259,   -1,
   -1,  262,   -1,  264,  265,  266,   -1,  268,  269,   -1,
  271,   -1,   -1,   -1,   -1,   -1,  277,   -1,  279,  280,
  281,  282,   -1,  284,   -1,   -1,  287,   -1,   -1,  290,
   -1,   -1,  412,   -1,   -1,   -1,  297,   -1,  299,  300,
   -1,  302,  303,  304,  305,  306,   -1,   -1,  309,   -1,
   -1,  312,   -1,  314,  315,   -1,   -1,  318,  319,  320,
   -1,  322,  323,  324,  325,  326,  327,  328,   -1,   -1,
   -1,  332,  333,  334,  335,  336,   -1,   -1,   -1,  340,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  348,  349,  350,
  351,  352,  353,   -1,   -1,   -1,  357,  358,   -1,  360,
   -1,  362,   -1,  364,  365,  366,   -1,  368,   -1,  370,
  371,  372,   -1,  374,  375,  376,  377,  378,   -1,  380,
  381,   -1,   -1,   -1,   -1,  386,   -1,  388,  389,   -1,
   -1,  392,  393,  394,   -1,   -1,  397,   -1,  399,   -1,
   -1,  402,  403,  404,  405,   -1,  407,   -1,  409,   -1,
  411,  412,  413,  414,  415,   -1,   -1,  418,  419,  420,
   -1,  422,  423,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  433,  434,  435,   -1,  437,  438,  439,   -1,
   -1,   -1,   -1,   -1,  445,   -1,   -1,   -1,  449,   -1,
  451,  452,  453,  454,   -1,  456,   -1,  458,  459,  460,
  461,  462,  463,  464,  465,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  475,  476,   -1,  478,  479,  480,
   -1,   -1,  602,   -1,  485,  486,  487,  488,  489,   -1,
  491,   -1,   -1,  494,  495,   -1,   -1,   -1,  499,  500,
   -1,   -1,  503,  504,  505,  506,  507,  508,  509,  510,
  511,   -1,  513,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  523,  524,  525,  526,  527,  528,  529,  530,
  531,  532,  533,  534,  535,  536,  537,  538,  539,  540,
  541,  542,  543,   -1,  545,  546,  547,  548,  549,  550,
   -1,  552,  553,  554,  555,   -1,  557,   -1,  559,  560,
  561,  562,  563,  564,  565,  566,  567,  568,  569,  570,
  571,  572,  573,  574,  575,  576,   -1,  578,  579,  580,
  581,   -1,  583,  584,  585,  586,  587,   -1,  589,  590,
  591,  592,  593,  594,  595,  596,  597,  598,  599,  600,
  601,  602,  603,  604,  605,  606,  607,  608,  609,  610,
  611,  612,  613,  614,  615,  616,  617,   -1,  619,  620,
   -1,   -1,   -1,  624,  625,  626,  627,  628,  629,   -1,
   -1,  632,  633,  634,  635,  636,  637,  638,  639,  640,
   41,   -1,   -1,   -1,   -1,  646,   -1,  648,  649,  650,
  651,  652,  653,   -1,  655,   -1,   -1,   -1,   -1,   60,
   61,   62,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  672,  673,   -1,  675,  676,  677,  678,  679,  680,
  681,  682,  683,   -1,  685,  686,  687,  688,  689,  690,
  691,  692,  693,  694,  695,  696,   -1,  698,  699,  700,
   -1,  702,   -1,   -1,   -1,  706,  707,  708,  709,  710,
  711,  712,  713,  714,  715,  716,  717,  718,    0,  720,
  721,   -1,   -1,   -1,   -1,   -1,  727,  728,  729,  730,
  731,  732,  733,  734,  735,  736,  737,  738,  739,  740,
  741,  742,  743,  744,   -1,   -1,  747,   -1,  749,  750,
  751,  752,  753,  754,   -1,   -1,   -1,   -1,   40,   41,
   42,   43,   44,   45,   46,   47,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   58,   59,   60,   61,
   62,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   91,
   -1,   93,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  262,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  271,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  284,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  318,   -1,   -1,
   -1,   -1,   -1,   -1,  325,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  333,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  350,
   -1,  352,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  362,   -1,  364,  365,  366,   -1,   -1,   -1,   -1,
  371,   -1,   -1,   -1,   -1,  257,   -1,  259,   -1,   -1,
  262,   -1,  264,  265,  266,   -1,  268,  269,   -1,  271,
   -1,   -1,   -1,   -1,   -1,  277,   -1,  279,  280,  281,
  282,   -1,  284,   -1,   -1,  287,   -1,   -1,  290,   -1,
   -1,  412,   -1,   -1,   -1,  297,   -1,  299,  300,   -1,
  302,  303,  304,  305,  306,   -1,   -1,  309,   -1,   -1,
  312,   -1,  314,  315,   -1,   -1,  318,  319,  320,   -1,
  322,  323,  324,  325,  326,  327,  328,   -1,   -1,   -1,
  332,  333,  334,  335,  336,   -1,   -1,   -1,  340,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  348,  349,  350,  351,
  352,  353,   -1,   -1,   -1,  357,  358,   -1,  360,   -1,
  362,   -1,  364,  365,  366,   -1,  368,   -1,  370,  371,
  372,   -1,  374,  375,  376,  377,  378,   -1,  380,  381,
   -1,   -1,   -1,   -1,  386,   -1,  388,  389,   -1,   -1,
  392,  393,  394,   -1,   -1,  397,   -1,  399,   -1,   -1,
  402,  403,  404,  405,   -1,  407,   -1,  409,   -1,  411,
  412,  413,  414,  415,   -1,   -1,  418,  419,  420,   -1,
  422,  423,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  433,  434,  435,   -1,  437,  438,  439,   -1,   -1,
   -1,   -1,   -1,  445,   -1,   -1,   -1,  449,   -1,  451,
  452,  453,  454,   -1,  456,   -1,  458,  459,  460,  461,
  462,  463,  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  475,  476,   -1,  478,  479,  480,   -1,
   -1,  602,   -1,  485,  486,  487,  488,  489,   -1,  491,
   -1,   -1,  494,  495,   -1,   -1,   -1,  499,  500,   -1,
   -1,  503,  504,  505,  506,  507,  508,  509,  510,  511,
   -1,  513,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  523,  524,  525,  526,  527,  528,  529,  530,  531,
  532,  533,  534,  535,  536,  537,  538,  539,  540,  541,
  542,  543,   -1,  545,  546,  547,  548,  549,  550,   -1,
  552,  553,  554,  555,   -1,  557,   -1,  559,  560,  561,
  562,  563,  564,  565,  566,  567,  568,  569,  570,  571,
  572,  573,  574,  575,  576,   -1,  578,  579,  580,  581,
   -1,  583,  584,  585,  586,  587,   -1,  589,  590,  591,
  592,  593,  594,  595,  596,  597,  598,  599,  600,  601,
  602,  603,  604,  605,  606,  607,  608,  609,  610,  611,
  612,  613,  614,  615,  616,  617,   -1,  619,  620,   -1,
   -1,   -1,  624,  625,  626,  627,  628,  629,   -1,   -1,
  632,  633,  634,  635,  636,  637,  638,  639,  640,   41,
   -1,   -1,   -1,   -1,  646,   -1,  648,  649,  650,  651,
  652,  653,   -1,  655,   -1,   -1,   -1,   -1,   60,   61,
   62,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  672,  673,   -1,  675,  676,  677,  678,  679,  680,  681,
  682,  683,   -1,  685,  686,  687,  688,  689,  690,  691,
  692,  693,  694,  695,  696,   -1,  698,  699,  700,   -1,
  702,   -1,   -1,   -1,  706,  707,  708,  709,  710,  711,
  712,  713,  714,  715,  716,  717,  718,    0,  720,  721,
   -1,   -1,   -1,   -1,   -1,  727,  728,  729,  730,  731,
  732,  733,  734,  735,  736,  737,  738,  739,  740,  741,
  742,  743,  744,   -1,   -1,  747,   -1,  749,  750,  751,
  752,  753,  754,   -1,   -1,   -1,   -1,   40,   41,   42,
   43,   44,   45,   46,   47,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   58,   59,   60,   61,   62,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   91,   -1,
   93,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  262,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  271,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  284,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  318,   -1,   -1,   -1,
   -1,   -1,   -1,  325,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  333,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  350,   -1,
  352,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  362,   -1,  364,  365,  366,   -1,   -1,   -1,   -1,  371,
   -1,   -1,   -1,   -1,  257,   -1,  259,   -1,   -1,  262,
   -1,  264,  265,  266,   -1,  268,  269,   -1,  271,   -1,
   -1,   -1,   -1,   -1,  277,   -1,  279,  280,  281,  282,
   -1,  284,   -1,   -1,  287,   -1,   -1,  290,   -1,   -1,
  412,   -1,   -1,   -1,  297,   -1,  299,  300,   -1,  302,
  303,  304,  305,  306,   -1,   -1,  309,   -1,   -1,  312,
   -1,  314,  315,   -1,   -1,  318,  319,  320,   -1,  322,
  323,  324,  325,  326,  327,  328,   -1,   -1,   -1,  332,
  333,  334,  335,  336,   -1,   -1,   -1,  340,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  348,  349,  350,  351,  352,
  353,   -1,   -1,   -1,  357,  358,   -1,  360,   -1,  362,
   -1,  364,  365,  366,   -1,  368,   -1,  370,  371,  372,
   -1,  374,  375,  376,  377,  378,   -1,  380,  381,   -1,
   -1,   -1,   -1,  386,   -1,  388,  389,   -1,   -1,  392,
  393,  394,   -1,   -1,  397,   -1,  399,   -1,   -1,  402,
  403,  404,  405,   -1,  407,   -1,  409,   -1,  411,  412,
  413,  414,  415,   -1,   -1,  418,  419,  420,   -1,  422,
  423,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  433,  434,  435,   -1,  437,  438,  439,   -1,   -1,   -1,
   -1,   -1,  445,   -1,   -1,   -1,  449,   -1,  451,  452,
  453,  454,   -1,  456,   -1,  458,  459,  460,  461,  462,
  463,  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  475,  476,   -1,  478,  479,  480,   -1,   -1,
  602,   -1,  485,  486,  487,  488,  489,   -1,  491,   -1,
   -1,  494,  495,   -1,   -1,   -1,  499,  500,   -1,   -1,
  503,  504,  505,  506,  507,  508,  509,  510,  511,   -1,
  513,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  523,  524,  525,  526,  527,  528,  529,  530,  531,  532,
  533,  534,  535,  536,  537,  538,  539,  540,  541,  542,
  543,   -1,  545,  546,  547,  548,  549,  550,   -1,  552,
  553,  554,  555,   -1,  557,   -1,  559,  560,  561,  562,
  563,  564,  565,  566,  567,  568,  569,  570,  571,  572,
  573,  574,  575,  576,   -1,  578,  579,  580,  581,   -1,
  583,  584,  585,  586,  587,   -1,  589,  590,  591,  592,
  593,  594,  595,  596,  597,  598,  599,  600,  601,  602,
  603,  604,  605,  606,  607,  608,  609,  610,  611,  612,
  613,  614,  615,  616,  617,   -1,  619,  620,   -1,   -1,
   -1,  624,  625,  626,  627,  628,  629,   -1,   -1,  632,
  633,  634,  635,  636,  637,  638,  639,  640,   41,   -1,
   -1,   -1,   -1,  646,   -1,  648,  649,  650,  651,  652,
  653,   -1,  655,   -1,   -1,   -1,   -1,   60,   61,   62,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  672,
  673,   -1,  675,  676,  677,  678,  679,  680,  681,  682,
  683,   -1,  685,  686,  687,  688,  689,  690,  691,  692,
  693,  694,  695,  696,   -1,  698,  699,  700,   -1,  702,
   -1,   -1,   -1,  706,  707,  708,  709,  710,  711,  712,
  713,  714,  715,  716,  717,  718,    0,  720,  721,   -1,
   -1,   -1,   -1,   -1,  727,  728,  729,  730,  731,  732,
  733,  734,  735,  736,  737,  738,  739,  740,  741,  742,
  743,  744,   -1,   -1,  747,   -1,  749,  750,  751,  752,
  753,  754,   -1,   -1,   -1,   -1,   40,   41,   42,   43,
   44,   45,   46,   47,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   58,   59,   60,   61,   62,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   91,   -1,   93,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  262,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  271,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  284,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  318,   -1,   -1,   -1,   -1,
   -1,   -1,  325,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  333,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  350,   -1,  352,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  362,
   -1,  364,  365,  366,   -1,   -1,   -1,   -1,  371,   -1,
   -1,   -1,   -1,  257,   -1,  259,   -1,   -1,  262,   -1,
  264,  265,  266,   -1,  268,  269,   -1,  271,   -1,   -1,
   -1,   -1,   -1,  277,   -1,  279,  280,  281,  282,   -1,
  284,   -1,   -1,  287,   -1,   -1,  290,   -1,   -1,  412,
   -1,   -1,   -1,  297,   -1,  299,  300,   -1,  302,  303,
  304,  305,  306,   -1,   -1,  309,   -1,   -1,  312,   -1,
  314,  315,   -1,   -1,  318,  319,  320,   -1,  322,  323,
  324,  325,  326,  327,  328,   -1,   -1,   -1,  332,  333,
  334,  335,  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  348,  349,  350,  351,  352,  353,
   -1,   -1,   -1,  357,  358,   -1,  360,   -1,  362,   -1,
  364,  365,  366,   -1,  368,   -1,  370,  371,  372,   -1,
  374,  375,  376,  377,  378,   -1,  380,  381,   -1,   -1,
   -1,   -1,  386,   -1,  388,  389,   -1,   -1,  392,  393,
  394,   -1,   -1,  397,   -1,  399,   -1,   -1,  402,  403,
  404,  405,   -1,  407,   -1,  409,   -1,  411,  412,  413,
  414,  415,   -1,   -1,  418,  419,  420,   -1,  422,  423,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  433,
  434,  435,   -1,  437,  438,  439,   -1,   -1,   -1,   -1,
   -1,  445,   -1,   -1,   -1,  449,   -1,  451,  452,  453,
  454,   -1,  456,   -1,  458,  459,  460,  461,  462,  463,
  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  475,  476,   -1,  478,  479,  480,   -1,   -1,  602,
   -1,  485,  486,  487,  488,  489,   -1,  491,   -1,   -1,
  494,  495,   -1,   -1,   -1,  499,  500,   -1,   -1,  503,
  504,  505,  506,  507,  508,  509,  510,  511,   -1,  513,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  523,
  524,  525,  526,  527,  528,  529,  530,  531,  532,  533,
  534,  535,  536,  537,  538,  539,  540,  541,  542,  543,
   -1,  545,  546,  547,  548,  549,  550,   -1,  552,  553,
  554,  555,   -1,  557,   -1,  559,  560,  561,  562,  563,
  564,  565,  566,  567,  568,  569,  570,  571,  572,  573,
  574,  575,  576,   -1,  578,  579,  580,  581,   -1,  583,
  584,  585,  586,  587,   -1,  589,  590,  591,  592,  593,
  594,  595,  596,  597,  598,  599,  600,  601,  602,  603,
  604,  605,  606,  607,  608,  609,  610,  611,  612,  613,
  614,  615,  616,  617,   -1,  619,  620,   -1,   -1,   -1,
  624,  625,  626,  627,  628,  629,   -1,   -1,  632,  633,
  634,  635,  636,  637,  638,  639,  640,   41,   -1,   -1,
   -1,   -1,  646,   -1,  648,  649,  650,  651,  652,  653,
   -1,  655,   -1,   -1,   -1,   -1,   60,   61,   62,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  672,  673,
   -1,  675,  676,  677,  678,  679,  680,  681,  682,  683,
   -1,  685,  686,  687,  688,  689,  690,  691,  692,  693,
  694,  695,  696,   -1,  698,  699,  700,   -1,  702,   -1,
   -1,   -1,  706,  707,  708,  709,  710,  711,  712,  713,
  714,  715,  716,  717,  718,    0,  720,  721,   -1,   -1,
   -1,   -1,   -1,  727,  728,  729,  730,  731,  732,  733,
  734,  735,  736,  737,  738,  739,  740,  741,  742,  743,
  744,   -1,   -1,  747,   -1,  749,  750,  751,  752,  753,
  754,   -1,   -1,   -1,   -1,   40,   41,   42,   43,   44,
   45,   46,   47,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   58,   59,   60,   61,   62,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   91,   -1,   93,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  262,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  271,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  284,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  318,   -1,   -1,   -1,   -1,   -1,
   -1,  325,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  333,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  350,   -1,  352,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  362,   -1,
  364,  365,  366,   -1,   -1,   -1,   -1,  371,   -1,   -1,
   -1,   -1,  257,   -1,  259,   -1,   -1,  262,   -1,  264,
  265,  266,   -1,  268,  269,   -1,  271,   -1,   -1,   -1,
   -1,   -1,  277,   -1,  279,  280,  281,  282,   -1,  284,
   -1,   -1,  287,   -1,   -1,  290,   -1,   -1,  412,   -1,
   -1,   -1,  297,   -1,  299,  300,   -1,  302,  303,  304,
  305,  306,   -1,   -1,  309,   -1,   -1,  312,   -1,  314,
  315,   -1,   -1,  318,  319,  320,   -1,  322,  323,  324,
  325,  326,  327,  328,   -1,   -1,   -1,  332,  333,  334,
  335,  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  348,  349,  350,  351,  352,  353,   -1,
   -1,   -1,  357,  358,   -1,  360,   -1,  362,   -1,  364,
  365,  366,   -1,  368,   -1,  370,  371,  372,   -1,  374,
  375,  376,  377,  378,   -1,  380,  381,   -1,   -1,   -1,
   -1,  386,   -1,  388,  389,   -1,   -1,  392,  393,  394,
   -1,   -1,  397,   -1,  399,   -1,   -1,  402,  403,  404,
  405,   -1,  407,   -1,  409,   -1,  411,  412,  413,  414,
  415,   -1,   -1,  418,  419,  420,   -1,  422,  423,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  433,  434,
  435,   -1,  437,  438,  439,   -1,   -1,   -1,   -1,   -1,
  445,   -1,   -1,   -1,  449,   -1,  451,  452,  453,  454,
   -1,  456,   -1,  458,  459,  460,  461,  462,  463,  464,
  465,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  475,  476,   -1,  478,  479,  480,   -1,   -1,  602,   -1,
  485,  486,  487,  488,  489,   -1,  491,   -1,   -1,  494,
  495,   -1,   -1,   -1,  499,  500,   -1,   -1,  503,  504,
  505,  506,  507,  508,  509,  510,  511,   -1,  513,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  523,  524,
  525,  526,  527,  528,  529,  530,  531,  532,  533,  534,
  535,  536,  537,  538,  539,  540,  541,  542,  543,   -1,
  545,  546,  547,  548,  549,  550,   -1,  552,  553,  554,
  555,   -1,  557,   -1,  559,  560,  561,  562,  563,  564,
  565,  566,  567,  568,  569,  570,  571,  572,  573,  574,
  575,  576,   -1,  578,  579,  580,  581,   -1,  583,  584,
  585,  586,  587,   -1,  589,  590,  591,  592,  593,  594,
  595,  596,  597,  598,  599,  600,  601,  602,  603,  604,
  605,  606,  607,  608,  609,  610,  611,  612,  613,  614,
  615,  616,  617,   -1,  619,  620,   -1,   -1,   -1,  624,
  625,  626,  627,  628,  629,   -1,   -1,  632,  633,  634,
  635,  636,  637,  638,  639,  640,   -1,   42,   43,   44,
   45,  646,   47,  648,  649,  650,  651,  652,  653,   -1,
  655,   -1,   -1,   -1,   -1,   60,   61,   62,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  672,  673,   -1,
  675,  676,  677,  678,  679,  680,  681,  682,  683,   -1,
  685,  686,  687,  688,  689,  690,  691,  692,  693,  694,
  695,  696,   -1,  698,  699,  700,   -1,  702,   -1,   -1,
   -1,  706,  707,  708,  709,  710,  711,  712,  713,  714,
  715,  716,  717,  718,    0,  720,  721,   -1,   -1,   -1,
   -1,   -1,  727,  728,  729,  730,  731,  732,  733,  734,
  735,  736,  737,  738,  739,  740,  741,  742,  743,  744,
   -1,   -1,  747,   -1,  749,  750,  751,  752,  753,  754,
   -1,   -1,   -1,   -1,   -1,   41,   42,   43,   44,   45,
   46,   47,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   58,   59,   60,   61,   62,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   91,   -1,   93,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  262,   -1,   -1,
   -1,  266,   -1,   -1,   -1,   -1,  271,   -1,   -1,   -1,
   -1,   -1,  277,   -1,   -1,   -1,  281,   -1,   -1,  284,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  318,   -1,   -1,   -1,   -1,   -1,   -1,
  325,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  333,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  350,   -1,  352,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  362,   -1,  364,
  365,  366,   -1,   -1,   -1,   -1,  371,   -1,   -1,   -1,
   -1,  257,   -1,  259,   -1,   -1,  262,   -1,  264,  265,
  266,   -1,  268,  269,   -1,  271,   -1,   -1,   -1,   -1,
   -1,  277,   -1,  279,  280,  281,  282,   -1,  284,   -1,
   -1,  287,   -1,   -1,  290,   -1,   -1,  412,   -1,   -1,
   -1,  297,   -1,  299,  300,   -1,  302,  303,  304,  305,
  306,   -1,   -1,  309,   -1,   -1,  312,   -1,  314,  315,
   -1,   -1,  318,  319,  320,   -1,  322,  323,  324,  325,
  326,  327,  328,   -1,   -1,   -1,  332,  333,  334,  335,
  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  348,  349,  350,  351,  352,  353,   -1,   -1,
   -1,  357,  358,   -1,  360,   -1,  362,   -1,  364,  365,
  366,   -1,  368,   -1,  370,  371,  372,   -1,  374,  375,
  376,  377,  378,   -1,  380,  381,   -1,   -1,   -1,   -1,
  386,   -1,  388,  389,   -1,   -1,  392,  393,  394,   -1,
   -1,  397,   -1,  399,   -1,   -1,  402,  403,  404,  405,
   -1,  407,   -1,  409,   -1,  411,  412,  413,  414,  415,
   -1,   -1,  418,  419,  420,   -1,  422,  423,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  433,  434,  435,
   -1,  437,  438,  439,   -1,   -1,   -1,   -1,   -1,  445,
   -1,   -1,   -1,  449,   -1,  451,  452,  453,  454,   -1,
  456,   -1,  458,  459,  460,  461,  462,  463,  464,  465,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  475,
  476,   -1,  478,  479,  480,   -1,   -1,  602,   -1,  485,
  486,  487,  488,  489,   -1,  491,   -1,   -1,  494,  495,
   -1,   -1,   -1,  499,  500,   -1,   -1,  503,  504,  505,
  506,  507,  508,  509,  510,  511,   -1,  513,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  523,  524,  525,
  526,  527,  528,  529,  530,  531,  532,  533,  534,  535,
  536,  537,  538,  539,  540,  541,  542,  543,   -1,  545,
  546,  547,  548,  549,  550,   -1,  552,  553,  554,  555,
   -1,  557,   -1,  559,  560,  561,  562,  563,  564,  565,
  566,  567,  568,  569,  570,  571,  572,  573,  574,  575,
  576,   -1,  578,  579,  580,  581,   -1,  583,  584,  585,
  586,  587,   -1,  589,  590,  591,  592,  593,  594,  595,
  596,  597,  598,  599,  600,  601,  602,  603,  604,  605,
  606,  607,  608,  609,  610,  611,  612,  613,  614,  615,
  616,  617,   -1,  619,  620,   -1,   -1,   -1,  624,  625,
  626,  627,  628,  629,   -1,   -1,  632,  633,  634,  635,
  636,  637,  638,  639,  640,   -1,   44,   -1,   -1,   -1,
  646,   -1,  648,  649,  650,  651,  652,  653,   -1,  655,
   -1,   -1,   60,   61,   62,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   60,   61,   62,  672,  673,   -1,  675,
  676,  677,  678,  679,  680,  681,  682,  683,   -1,  685,
  686,  687,  688,  689,  690,  691,  692,  693,  694,  695,
  696,   -1,  698,  699,  700,   -1,  702,   -1,   -1,   -1,
  706,  707,  708,  709,  710,  711,  712,  713,  714,  715,
  716,  717,  718,    0,  720,  721,   -1,   -1,   -1,   -1,
   -1,  727,  728,  729,  730,  731,  732,  733,  734,  735,
  736,  737,  738,  739,  740,  741,  742,  743,  744,   -1,
   -1,  747,   -1,  749,  750,  751,  752,  753,  754,   -1,
   -1,   -1,   -1,   -1,   41,   42,   43,   44,   45,   46,
   47,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   58,   59,   60,   61,   62,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   91,   -1,   93,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  262,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  271,   -1,  262,   -1,  264,   -1,   -1,
   -1,   -1,   -1,   -1,  271,   -1,  284,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  318,   -1,   -1,   -1,   -1,   -1,   -1,  325,   -1,   -1,
   -1,  318,   -1,   -1,   -1,  333,   -1,   -1,  325,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  333,   -1,   -1,   -1,
   -1,   -1,  350,   -1,  352,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  350,  362,  352,  364,  365,  366,   -1,
   -1,   -1,   -1,  371,   -1,  362,   -1,  364,  365,  366,
  257,   -1,  259,   -1,  371,  262,   -1,  264,  265,  266,
   -1,  268,  269,   -1,  271,   -1,   -1,   -1,   -1,   -1,
  277,   -1,  279,  280,  281,  282,   -1,  284,   -1,   -1,
  287,   -1,   -1,  290,  412,   -1,   -1,   -1,   -1,   -1,
  297,   -1,  299,  300,   -1,  302,  303,  304,  305,  306,
   -1,   -1,  309,   -1,   -1,  312,   -1,  314,  315,   -1,
   -1,  318,  319,  320,   -1,  322,  323,  324,  325,  326,
  327,  328,   -1,   -1,   -1,  332,  333,  334,  335,  336,
   -1,   -1,   -1,  340,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  348,  349,  350,  351,  352,  353,   -1,   -1,   -1,
  357,  358,   -1,  360,   -1,  362,   -1,  364,  365,  366,
   -1,  368,   -1,  370,  371,  372,   -1,  374,  375,  376,
  377,  378,   -1,  380,  381,   -1,   -1,   -1,   -1,  386,
   -1,  388,  389,   -1,   -1,  392,  393,  394,   -1,   -1,
  397,   -1,  399,   -1,   -1,  402,  403,  404,  405,   -1,
  407,   -1,  409,   -1,  411,  412,  413,  414,  415,   -1,
   -1,  418,  419,  420,   -1,  422,  423,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  433,  434,  435,   -1,
  437,  438,  439,   -1,   -1,   -1,   -1,   -1,  445,   -1,
   -1,   -1,  449,   -1,  451,  452,  453,  454,   -1,  456,
   -1,  458,  459,  460,  461,  462,  463,  464,  465,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  475,  476,
   -1,  478,  479,  480,  602,   -1,   -1,   -1,  485,  486,
  487,  488,  489,   -1,  491,  602,   -1,  494,  495,   -1,
   -1,   -1,  499,  500,   -1,   -1,  503,  504,  505,  506,
  507,  508,  509,  510,  511,   -1,  513,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  523,  524,  525,  526,
  527,  528,  529,  530,  531,  532,  533,  534,  535,  536,
  537,  538,  539,  540,  541,  542,  543,   -1,  545,  546,
  547,  548,  549,  550,   -1,  552,  553,  554,  555,   -1,
  557,   -1,  559,  560,  561,  562,  563,  564,  565,  566,
  567,  568,  569,  570,  571,  572,  573,  574,  575,  576,
   -1,  578,  579,  580,  581,   -1,  583,  584,  585,  586,
  587,   -1,  589,  590,  591,  592,  593,  594,  595,  596,
  597,  598,  599,  600,  601,  602,  603,  604,  605,  606,
  607,  608,  609,  610,  611,  612,  613,  614,  615,  616,
  617,   -1,  619,  620,   -1,   -1,   -1,  624,  625,  626,
  627,  628,  629,   -1,   -1,  632,  633,  634,  635,  636,
  637,  638,  639,  640,   -1,   44,   -1,   -1,   -1,  646,
   -1,  648,  649,  650,  651,  652,  653,   -1,  655,   -1,
   -1,   60,   61,   62,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  672,  673,   -1,  675,  676,
  677,  678,  679,  680,  681,  682,  683,   -1,  685,  686,
  687,  688,  689,  690,  691,  692,  693,  694,  695,  696,
   -1,  698,  699,  700,   -1,  702,   -1,   -1,   -1,  706,
  707,  708,  709,  710,  711,  712,  713,  714,  715,  716,
  717,  718,    0,  720,  721,   -1,   -1,   -1,   -1,   -1,
  727,  728,  729,  730,  731,  732,  733,  734,  735,  736,
  737,  738,  739,  740,  741,  742,  743,  744,   -1,   -1,
  747,   -1,  749,  750,  751,  752,  753,  754,   -1,   -1,
   -1,   -1,   -1,   41,   42,   43,   44,   45,   46,   47,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   58,   59,   60,   61,   62,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  290,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  300,   -1,   -1,   -1,   -1,   -1,
  306,   -1,   -1,   91,   -1,   93,   -1,   -1,   -1,   -1,
  316,   -1,   -1,  319,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  262,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  271,   -1,   -1,  371,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  284,   -1,   -1,   -1,   -1,
   -1,  387,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  402,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  318,
   -1,  417,   -1,   -1,   -1,  421,  325,   -1,  424,   -1,
  426,   -1,   -1,   -1,  333,   -1,  432,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  350,   -1,  352,   -1,   -1,   -1,   -1,  454,   -1,
   -1,   -1,   -1,  362,   -1,  364,  365,  366,   -1,   -1,
   -1,   -1,  371,   -1,   -1,   -1,   -1,   -1,   -1,  257,
   -1,  259,   -1,   -1,  262,   -1,  264,  265,  266,   -1,
  268,  269,   -1,  271,   -1,   -1,   -1,   -1,   -1,  277,
   -1,  279,  280,  281,  282,   -1,  284,   -1,   -1,  287,
   -1,   -1,  290,  412,  510,   -1,   -1,  513,   -1,  297,
   -1,  299,  300,   -1,  302,  303,  304,  305,  306,   -1,
   -1,  309,   -1,   -1,  312,   -1,  314,  315,   -1,   -1,
  318,  319,  320,   -1,  322,  323,  324,  325,  326,  327,
  328,   -1,   -1,   -1,  332,  333,  334,  335,  336,   -1,
  556,   -1,  340,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  348,  349,  350,  351,  352,  353,   -1,   -1,   -1,  357,
  358,   -1,  360,   -1,  362,   -1,  364,  365,  366,   -1,
  368,   -1,  370,  371,  372,   -1,  374,  375,  376,  377,
  378,   -1,  380,  381,  600,   -1,   -1,   -1,  386,   -1,
  388,  389,   -1,   -1,  392,  393,  394,   -1,   -1,  397,
   -1,  399,   -1,  619,  402,  403,  404,  405,   -1,  407,
   -1,  409,   -1,  411,  412,  413,  414,  415,   -1,   -1,
  418,  419,  420,   -1,  422,  423,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  433,  434,  435,   -1,  437,
  438,  439,   -1,   -1,   -1,   -1,   -1,  445,   -1,   -1,
   -1,  449,   -1,  451,  452,  453,  454,   -1,  456,   -1,
  458,  459,  460,  461,  462,  463,  464,  465,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  475,  476,   -1,
  478,  479,  480,  602,   -1,   -1,   -1,  485,  486,  487,
  488,  489,   -1,  491,   -1,   -1,  494,  495,   -1,   -1,
   -1,  499,  500,   -1,   -1,  503,  504,  505,  506,  507,
  508,  509,  510,  511,   -1,  513,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  523,  524,  525,  526,  527,
  528,  529,  530,  531,  532,  533,  534,  535,  536,  537,
  538,  539,  540,  541,  542,  543,   -1,  545,  546,  547,
  548,  549,  550,   -1,  552,  553,  554,  555,   -1,  557,
   -1,  559,  560,  561,  562,  563,  564,  565,  566,  567,
  568,  569,  570,  571,  572,  573,  574,  575,  576,   -1,
  578,  579,  580,  581,   -1,  583,  584,  585,  586,  587,
   -1,  589,  590,  591,  592,  593,  594,  595,  596,  597,
  598,  599,  600,  601,  602,  603,  604,  605,  606,  607,
  608,  609,  610,  611,  612,  613,  614,  615,  616,  617,
   -1,  619,  620,   -1,   -1,   -1,  624,  625,  626,  627,
  628,  629,   -1,   -1,  632,  633,  634,  635,  636,  637,
  638,  639,  640,   -1,   44,   -1,   -1,   -1,  646,   -1,
  648,  649,  650,  651,  652,  653,   -1,  655,   -1,   -1,
   60,   61,   62,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  672,  673,   -1,  675,  676,  677,
  678,  679,  680,  681,  682,  683,   -1,  685,  686,  687,
  688,  689,  690,  691,  692,  693,  694,  695,  696,   -1,
  698,  699,  700,   -1,  702,   -1,   -1,   -1,  706,  707,
  708,  709,  710,  711,  712,  713,  714,  715,  716,  717,
  718,    0,  720,  721,   -1,   -1,   -1,   -1,   -1,  727,
  728,  729,  730,  731,  732,  733,  734,  735,  736,  737,
  738,  739,  740,  741,  742,  743,  744,   -1,   -1,  747,
   -1,  749,  750,  751,  752,  753,  754,   -1,   -1,   -1,
   -1,   -1,   41,   42,   43,   44,   45,   46,   47,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   58,
   59,   60,   61,   62,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   91,   -1,   93,   -1,   -1,   -1,   -1,  300,
   -1,   -1,   -1,   -1,   -1,  306,   -1,   -1,   -1,  310,
  311,   -1,   -1,   -1,   -1,  316,   -1,   -1,  319,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  329,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  262,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  271,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  284,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  387,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  318,   -1,
   -1,  402,   -1,   -1,   -1,  325,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  333,   -1,   -1,  417,   -1,   -1,   -1,
  421,   -1,   -1,   -1,   -1,  426,   -1,   -1,   -1,   -1,
  350,  432,  352,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  362,   -1,  364,  365,  366,   -1,   -1,   -1,
   -1,  371,   -1,  454,   -1,   -1,   -1,   -1,  257,   -1,
  259,   -1,   -1,  262,   -1,  264,  265,  266,   -1,  268,
  269,   -1,  271,   -1,   -1,   -1,   -1,   -1,  277,   -1,
  279,  280,  281,  282,   -1,  284,   -1,   -1,  287,   -1,
   -1,  290,  412,   -1,   -1,   -1,   -1,   -1,  297,   -1,
  299,  300,   -1,  302,  303,  304,  305,  306,   -1,  510,
  309,   -1,  513,  312,   -1,  314,  315,   -1,   -1,  318,
  319,  320,   -1,  322,  323,  324,  325,  326,  327,  328,
   -1,   -1,   -1,  332,  333,  334,  335,  336,   -1,   -1,
   -1,  340,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  348,
  349,  350,  351,  352,  353,  556,   -1,   -1,  357,  358,
   -1,  360,   -1,  362,   -1,  364,  365,  366,   -1,  368,
   -1,  370,  371,  372,   -1,  374,  375,  376,  377,  378,
   -1,  380,  381,   -1,   -1,   -1,   -1,  386,   -1,  388,
  389,   -1,   -1,  392,  393,  394,   -1,   -1,  397,  600,
  399,   -1,   -1,  402,  403,  404,  405,   -1,  407,   -1,
  409,   -1,  411,  412,  413,  414,  415,   -1,  619,  418,
  419,  420,   -1,  422,  423,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  433,  434,  435,   -1,  437,  438,
  439,   -1,   -1,   -1,   -1,   -1,  445,   -1,   -1,   -1,
  449,   -1,  451,  452,  453,  454,   -1,  456,   -1,  458,
  459,  460,  461,  462,  463,  464,  465,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  475,  476,   -1,  478,
  479,  480,  602,   -1,   -1,   -1,  485,  486,  487,  488,
  489,   -1,  491,   -1,   -1,  494,  495,   -1,   -1,   -1,
  499,  500,   -1,   -1,  503,  504,  505,  506,  507,  508,
  509,  510,  511,   -1,  513,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  523,  524,  525,  526,  527,  528,
  529,  530,  531,  532,  533,  534,  535,  536,  537,  538,
  539,  540,  541,  542,  543,   -1,  545,  546,  547,  548,
  549,  550,   -1,  552,  553,  554,  555,   -1,  557,   -1,
  559,  560,  561,  562,  563,  564,  565,  566,  567,  568,
  569,  570,  571,  572,  573,  574,  575,  576,   -1,  578,
  579,  580,  581,   -1,  583,  584,  585,  586,  587,   -1,
  589,  590,  591,  592,  593,  594,  595,  596,  597,  598,
  599,  600,  601,  602,  603,  604,  605,  606,  607,  608,
  609,  610,  611,  612,  613,  614,  615,  616,  617,   -1,
  619,  620,   -1,   -1,   -1,  624,  625,  626,  627,  628,
  629,   -1,   -1,  632,  633,  634,  635,  636,  637,  638,
  639,  640,   -1,   44,   -1,   -1,   -1,  646,   -1,  648,
  649,  650,  651,  652,  653,   -1,  655,   -1,   -1,   60,
   61,   62,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  672,  673,   -1,  675,  676,  677,  678,
  679,  680,  681,  682,  683,   -1,  685,  686,  687,  688,
  689,  690,  691,  692,  693,  694,  695,  696,   -1,  698,
  699,  700,   -1,  702,   -1,   -1,   -1,  706,  707,  708,
  709,  710,  711,  712,  713,  714,  715,  716,  717,  718,
    0,  720,  721,   -1,   -1,   -1,   -1,   -1,  727,  728,
  729,  730,  731,  732,  733,  734,  735,  736,  737,  738,
  739,  740,  741,  742,  743,  744,   -1,   -1,  747,   -1,
  749,  750,  751,  752,  753,  754,   -1,   -1,   -1,   -1,
   -1,   41,   42,   43,   44,   45,   46,   47,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   58,   59,
   60,   61,   62,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   91,   -1,   93,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  262,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  271,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  284,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  318,   -1,   -1,
   -1,   -1,   -1,   -1,  325,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  333,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  350,
   -1,  352,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  362,   -1,  364,  365,  366,   -1,   -1,   -1,   -1,
  371,   -1,   -1,   -1,   -1,   -1,   -1,  257,   -1,  259,
   -1,   -1,  262,   -1,  264,  265,  266,   -1,  268,  269,
   -1,  271,   -1,   -1,   -1,   -1,   -1,  277,   -1,  279,
  280,  281,  282,   -1,  284,   -1,   -1,  287,   -1,   -1,
  290,  412,   -1,   -1,   -1,   -1,   -1,  297,   -1,  299,
  300,   -1,  302,  303,  304,  305,  306,   -1,   -1,  309,
   -1,   -1,  312,   -1,  314,  315,   -1,   -1,  318,  319,
  320,   -1,  322,  323,  324,  325,  326,  327,  328,   -1,
   -1,   -1,  332,  333,  334,  335,  336,   -1,   -1,   -1,
  340,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  348,  349,
  350,  351,  352,  353,   -1,   -1,   -1,  357,  358,   -1,
  360,   -1,  362,   -1,  364,  365,  366,   -1,  368,   -1,
  370,  371,  372,   -1,  374,  375,  376,  377,  378,   -1,
  380,  381,   -1,   -1,   -1,   -1,  386,   -1,  388,  389,
   -1,   -1,  392,  393,  394,   -1,   -1,  397,   -1,  399,
   -1,   -1,  402,  403,  404,  405,   -1,  407,   -1,  409,
   -1,  411,  412,  413,  414,  415,   -1,   -1,  418,  419,
  420,   -1,  422,  423,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  433,  434,  435,   -1,  437,  438,  439,
   -1,   -1,   -1,   -1,   -1,  445,   -1,   -1,   -1,  449,
   -1,  451,  452,  453,  454,   -1,  456,   -1,  458,  459,
  460,  461,  462,  463,  464,  465,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  475,  476,   -1,  478,  479,
  480,  602,   -1,   -1,   -1,  485,  486,  487,  488,  489,
   -1,  491,   -1,   -1,  494,  495,   -1,   -1,   -1,  499,
  500,   -1,   -1,  503,  504,  505,  506,  507,  508,  509,
  510,  511,   -1,  513,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  523,  524,  525,  526,  527,  528,  529,
  530,  531,  532,  533,  534,  535,  536,  537,  538,  539,
  540,  541,  542,  543,   -1,  545,  546,  547,  548,  549,
  550,   -1,  552,  553,  554,  555,   -1,  557,   -1,  559,
  560,  561,  562,  563,  564,  565,  566,  567,  568,  569,
  570,  571,  572,  573,  574,  575,  576,   -1,  578,  579,
  580,  581,   -1,  583,  584,  585,  586,  587,   -1,  589,
  590,  591,  592,  593,  594,  595,  596,  597,  598,  599,
  600,  601,  602,  603,  604,  605,  606,  607,  608,  609,
  610,  611,  612,  613,  614,  615,  616,  617,   -1,  619,
  620,   -1,   -1,   -1,  624,  625,  626,  627,  628,  629,
   -1,   -1,  632,  633,  634,  635,  636,  637,  638,  639,
  640,   -1,   44,   -1,   -1,   -1,  646,   -1,  648,  649,
  650,  651,  652,  653,   -1,  655,   -1,   -1,   60,   61,
   62,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  672,  673,   -1,  675,  676,  677,  678,  679,
  680,  681,  682,  683,   -1,  685,  686,  687,  688,  689,
  690,  691,  692,  693,  694,  695,  696,   -1,  698,  699,
  700,   -1,  702,   -1,   -1,   -1,  706,  707,  708,  709,
  710,  711,  712,  713,  714,  715,  716,  717,  718,    0,
  720,  721,   -1,   -1,   -1,   -1,   -1,  727,  728,  729,
  730,  731,  732,  733,  734,  735,  736,  737,  738,  739,
  740,  741,  742,  743,  744,   -1,   -1,  747,   -1,  749,
  750,  751,  752,  753,  754,   -1,   -1,   -1,   -1,   -1,
   41,   42,   43,   44,   45,   46,   47,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   58,   59,   60,
   61,   62,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   91,   -1,   93,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  262,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  271,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  284,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  318,   -1,   -1,   -1,
   -1,   -1,   -1,  325,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  333,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  350,   -1,
  352,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  362,   -1,  364,  365,  366,   -1,   -1,   -1,   -1,  371,
   -1,   -1,   -1,   -1,   -1,   -1,  257,   -1,  259,   -1,
   -1,  262,   -1,  264,  265,  266,   -1,  268,  269,   -1,
  271,   -1,   -1,   -1,   -1,   -1,  277,   -1,  279,  280,
  281,  282,   -1,  284,   -1,   -1,  287,   -1,   -1,  290,
  412,   -1,   -1,   -1,   -1,   -1,  297,   -1,  299,  300,
   -1,  302,  303,  304,  305,  306,   -1,   -1,  309,   -1,
   -1,  312,   -1,  314,  315,   -1,   -1,  318,  319,  320,
   -1,  322,  323,  324,  325,  326,  327,  328,   -1,   -1,
   -1,  332,  333,  334,  335,  336,   -1,   -1,   -1,  340,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  348,  349,  350,
  351,  352,  353,   -1,   -1,   -1,  357,  358,   -1,  360,
   -1,  362,   -1,  364,  365,  366,   -1,  368,   -1,  370,
  371,  372,   -1,  374,  375,  376,  377,  378,   -1,  380,
  381,   -1,   -1,   -1,   -1,  386,   -1,  388,  389,   -1,
   -1,  392,  393,  394,   -1,   -1,  397,   -1,  399,   -1,
   -1,  402,  403,  404,  405,   -1,  407,   -1,  409,   -1,
  411,  412,  413,  414,  415,   -1,   -1,  418,  419,  420,
   -1,  422,  423,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  433,  434,  435,   -1,  437,  438,  439,   -1,
   -1,   -1,   -1,   -1,  445,   -1,   -1,   -1,  449,   -1,
  451,  452,  453,  454,   -1,  456,   -1,  458,  459,  460,
  461,  462,  463,  464,  465,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  475,  476,   -1,  478,  479,  480,
  602,   -1,   -1,   -1,  485,  486,  487,  488,  489,   -1,
  491,   -1,   -1,  494,  495,   -1,   -1,   -1,  499,  500,
   -1,   -1,  503,  504,  505,  506,  507,  508,  509,  510,
  511,   -1,  513,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  523,  524,  525,  526,  527,  528,  529,  530,
  531,  532,  533,  534,  535,  536,  537,  538,  539,  540,
  541,  542,  543,   -1,  545,  546,  547,  548,  549,  550,
   -1,  552,  553,  554,  555,   -1,  557,   -1,  559,  560,
  561,  562,  563,  564,  565,  566,  567,  568,  569,  570,
  571,  572,  573,  574,  575,  576,   -1,  578,  579,  580,
  581,   -1,  583,  584,  585,  586,  587,   -1,  589,  590,
  591,  592,  593,  594,  595,  596,  597,  598,  599,  600,
  601,  602,  603,  604,  605,  606,  607,  608,  609,  610,
  611,  612,  613,  614,  615,  616,  617,   -1,  619,  620,
   -1,   -1,   -1,  624,  625,  626,  627,  628,  629,   -1,
   -1,  632,  633,  634,  635,  636,  637,  638,  639,  640,
   -1,   44,   -1,   -1,   -1,  646,   -1,  648,  649,  650,
  651,  652,  653,   -1,  655,   -1,   -1,   60,   61,   62,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  672,  673,   -1,  675,  676,  677,  678,  679,  680,
  681,  682,  683,   -1,  685,  686,  687,  688,  689,  690,
  691,  692,  693,  694,  695,  696,   -1,  698,  699,  700,
   -1,  702,   -1,   -1,   -1,  706,  707,  708,  709,  710,
  711,  712,  713,  714,  715,  716,  717,  718,    0,  720,
  721,   -1,   -1,   -1,   -1,   -1,  727,  728,  729,  730,
  731,  732,  733,  734,  735,  736,  737,  738,  739,  740,
  741,  742,  743,  744,   -1,   -1,  747,   -1,  749,  750,
  751,  752,  753,  754,   -1,   -1,   -1,   -1,   -1,   41,
   42,   43,   44,   45,   46,   47,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   58,   59,   60,   61,
   62,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   91,
   -1,   93,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  262,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  271,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  284,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  318,   -1,   -1,   -1,   -1,
   -1,   -1,  325,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  333,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  350,   -1,  352,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  362,
   -1,  364,  365,  366,   -1,   -1,   -1,   -1,  371,   -1,
   -1,   -1,   -1,   -1,   -1,  257,   -1,  259,   -1,   -1,
  262,   -1,  264,  265,  266,   -1,  268,  269,   -1,  271,
   -1,   -1,   -1,   -1,   -1,  277,   -1,  279,  280,  281,
  282,   -1,  284,   -1,   -1,  287,   -1,   -1,  290,  412,
   -1,   -1,   -1,   -1,   -1,  297,   -1,  299,  300,   -1,
  302,  303,  304,  305,  306,   -1,   -1,  309,   -1,   -1,
  312,   -1,  314,  315,   -1,   -1,  318,  319,  320,   -1,
  322,  323,  324,  325,  326,  327,  328,   -1,   -1,   -1,
  332,  333,  334,  335,  336,   -1,   -1,   -1,  340,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  348,  349,  350,  351,
  352,  353,   -1,   -1,   -1,  357,  358,   -1,  360,   -1,
  362,   -1,  364,  365,  366,   -1,  368,   -1,  370,  371,
  372,   -1,  374,  375,  376,  377,  378,   -1,  380,  381,
   -1,   -1,   -1,   -1,  386,   -1,  388,  389,   -1,   -1,
  392,  393,  394,   -1,   -1,  397,   -1,  399,   -1,   -1,
  402,  403,  404,  405,   -1,  407,   -1,  409,   -1,  411,
  412,  413,  414,  415,   -1,   -1,  418,  419,  420,   -1,
  422,  423,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  433,  434,  435,   -1,  437,  438,  439,   -1,   -1,
   -1,   -1,   -1,  445,   -1,   -1,   -1,  449,   -1,  451,
  452,  453,  454,   -1,  456,   -1,  458,  459,  460,  461,
  462,  463,  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  475,  476,   -1,  478,  479,  480,  602,
   -1,   -1,   -1,  485,  486,  487,  488,  489,   -1,  491,
   -1,   -1,  494,  495,   -1,   -1,   -1,  499,  500,   -1,
   -1,  503,  504,  505,  506,  507,  508,  509,  510,  511,
   -1,  513,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  523,  524,  525,  526,  527,  528,  529,  530,  531,
  532,  533,  534,  535,  536,  537,  538,  539,  540,  541,
  542,  543,   -1,  545,  546,  547,  548,  549,  550,   -1,
  552,  553,  554,  555,   -1,  557,   -1,  559,  560,  561,
  562,  563,  564,  565,  566,  567,  568,  569,  570,  571,
  572,  573,  574,  575,  576,   -1,  578,  579,  580,  581,
   -1,  583,  584,  585,  586,  587,   -1,  589,  590,  591,
  592,  593,  594,  595,  596,  597,  598,  599,  600,  601,
  602,  603,  604,  605,  606,  607,  608,  609,  610,  611,
  612,  613,  614,  615,  616,  617,   -1,  619,  620,   -1,
   -1,   -1,  624,  625,  626,  627,  628,  629,   -1,   -1,
  632,  633,  634,  635,  636,  637,  638,  639,  640,   -1,
   44,   -1,   -1,   -1,  646,   -1,  648,  649,  650,  651,
  652,  653,   -1,  655,   -1,   -1,   60,   61,   62,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  672,  673,   -1,  675,  676,  677,  678,  679,  680,  681,
  682,  683,   -1,  685,  686,  687,  688,  689,  690,  691,
  692,  693,  694,  695,  696,   -1,  698,  699,  700,   -1,
  702,   -1,   -1,   -1,  706,  707,  708,  709,  710,  711,
  712,  713,  714,  715,  716,  717,  718,    0,  720,  721,
   -1,   -1,   -1,   -1,   -1,  727,  728,  729,  730,  731,
  732,  733,  734,  735,  736,  737,  738,  739,  740,  741,
  742,  743,  744,   -1,   -1,  747,   -1,  749,  750,  751,
  752,  753,  754,   -1,   -1,   -1,   -1,   -1,   41,   42,
   43,   44,   45,   46,   47,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   58,   59,   60,   61,   62,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   91,   -1,
   93,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  262,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  271,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  284,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  318,   -1,   -1,   -1,   -1,   -1,
   -1,  325,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  333,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  350,   -1,  352,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  362,   -1,
  364,  365,  366,   -1,   -1,   -1,   -1,  371,   -1,   -1,
   -1,   -1,   -1,   -1,  257,   -1,  259,   -1,   -1,  262,
   -1,  264,  265,  266,   -1,  268,  269,   -1,  271,   -1,
   -1,   -1,   -1,   -1,  277,   -1,  279,  280,  281,  282,
   -1,  284,   -1,   -1,  287,   -1,   -1,  290,  412,   -1,
   -1,   -1,   -1,   -1,  297,   -1,  299,  300,   -1,  302,
  303,  304,  305,  306,   -1,   -1,  309,   -1,   -1,  312,
   -1,  314,  315,   -1,   -1,  318,  319,  320,   -1,  322,
  323,  324,  325,  326,  327,  328,   -1,   -1,   -1,  332,
  333,  334,  335,  336,   -1,   -1,   -1,  340,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  348,  349,  350,  351,  352,
  353,   -1,   -1,   -1,  357,  358,   -1,  360,   -1,  362,
   -1,  364,  365,  366,   -1,  368,   -1,  370,  371,  372,
   -1,  374,  375,  376,  377,  378,   -1,  380,  381,   -1,
   -1,   -1,   -1,  386,   -1,  388,  389,   -1,   -1,  392,
  393,  394,   -1,   -1,  397,   -1,  399,   -1,   -1,  402,
  403,  404,  405,   -1,  407,   -1,  409,   -1,  411,  412,
  413,  414,  415,   -1,   -1,  418,  419,  420,   -1,  422,
  423,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  433,  434,  435,   -1,  437,  438,  439,   -1,   -1,   -1,
   -1,   -1,  445,   -1,   -1,   -1,  449,   -1,  451,  452,
  453,  454,   -1,  456,   -1,  458,  459,  460,  461,  462,
  463,  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  475,  476,   -1,  478,  479,  480,  602,   -1,
   -1,   -1,  485,  486,  487,  488,  489,   -1,  491,   -1,
   -1,  494,  495,   -1,   -1,   -1,  499,  500,   -1,   -1,
  503,  504,  505,  506,  507,  508,  509,  510,  511,   -1,
  513,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  523,  524,  525,  526,  527,  528,  529,  530,  531,  532,
  533,  534,  535,  536,  537,  538,  539,  540,  541,  542,
  543,   -1,  545,  546,  547,  548,  549,  550,   -1,  552,
  553,  554,  555,   -1,  557,   -1,  559,  560,  561,  562,
  563,  564,  565,  566,  567,  568,  569,  570,  571,  572,
  573,  574,  575,  576,   -1,  578,  579,  580,  581,   -1,
  583,  584,  585,  586,  587,   -1,  589,  590,  591,  592,
  593,  594,  595,  596,  597,  598,  599,  600,  601,  602,
  603,  604,  605,  606,  607,  608,  609,  610,  611,  612,
  613,  614,  615,  616,  617,   -1,  619,  620,   -1,   -1,
   -1,  624,  625,  626,  627,  628,  629,   -1,   -1,  632,
  633,  634,  635,  636,  637,  638,  639,  640,   -1,   44,
   -1,   -1,   -1,  646,   -1,  648,  649,  650,  651,  652,
  653,   -1,  655,   -1,   -1,   60,   61,   62,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  672,
  673,   -1,  675,  676,  677,  678,  679,  680,  681,  682,
  683,   -1,  685,  686,  687,  688,  689,  690,  691,  692,
  693,  694,  695,  696,   -1,  698,  699,  700,   -1,  702,
   -1,   -1,   -1,  706,  707,  708,  709,  710,  711,  712,
  713,  714,  715,  716,  717,  718,    0,  720,  721,   -1,
   -1,   -1,   -1,   -1,  727,  728,  729,  730,  731,  732,
  733,  734,  735,  736,  737,  738,  739,  740,  741,  742,
  743,  744,   -1,   -1,  747,   -1,  749,  750,  751,  752,
  753,  754,   -1,   -1,   -1,   -1,   -1,   41,   42,   43,
   44,   45,   46,   47,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   58,   59,   60,   61,   62,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   91,   -1,   93,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  262,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  271,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  284,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  318,   -1,   -1,   -1,   -1,   -1,   -1,
  325,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  333,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  350,   -1,  352,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  362,   -1,  364,
  365,  366,   -1,   -1,   -1,   -1,  371,   -1,   -1,   -1,
   -1,   -1,   -1,  257,   -1,  259,   -1,   -1,  262,   -1,
  264,  265,  266,   -1,  268,  269,   -1,  271,   -1,   -1,
   -1,   -1,   -1,  277,   -1,  279,  280,  281,  282,   -1,
  284,   -1,   -1,  287,   -1,   -1,  290,  412,   -1,   -1,
   -1,   -1,   -1,  297,   -1,  299,  300,   -1,  302,  303,
  304,  305,  306,   -1,   -1,  309,   -1,   -1,  312,   -1,
  314,  315,   -1,   -1,  318,  319,  320,   -1,  322,  323,
  324,  325,  326,  327,  328,   -1,   -1,   -1,  332,  333,
  334,  335,  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  348,  349,  350,  351,  352,  353,
   -1,   -1,   -1,  357,  358,   -1,  360,   -1,  362,   -1,
  364,  365,  366,   -1,  368,   -1,  370,  371,  372,   -1,
  374,  375,  376,  377,  378,   -1,  380,  381,   -1,   -1,
   -1,   -1,  386,   -1,  388,  389,   -1,   -1,  392,  393,
  394,   -1,   -1,  397,   -1,  399,   -1,   -1,  402,  403,
  404,  405,   -1,  407,   -1,  409,   -1,  411,  412,  413,
  414,  415,   -1,   -1,  418,  419,  420,   -1,  422,  423,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  433,
  434,  435,   -1,  437,  438,  439,   -1,   -1,   -1,   -1,
   -1,  445,   -1,   -1,   -1,  449,   -1,  451,  452,  453,
  454,   -1,  456,   -1,  458,  459,  460,  461,  462,  463,
  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  475,  476,   -1,  478,  479,  480,  602,   -1,   -1,
   -1,  485,  486,  487,  488,  489,   -1,  491,   -1,   -1,
  494,  495,   -1,   -1,   -1,  499,  500,   -1,   -1,  503,
  504,  505,  506,  507,  508,  509,  510,  511,   -1,  513,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  523,
  524,  525,  526,  527,  528,  529,  530,  531,  532,  533,
  534,  535,  536,  537,  538,  539,  540,  541,  542,  543,
   -1,  545,  546,  547,  548,  549,  550,   -1,  552,  553,
  554,  555,   -1,  557,   -1,  559,  560,  561,  562,  563,
  564,  565,  566,  567,  568,  569,  570,  571,  572,  573,
  574,  575,  576,   -1,  578,  579,  580,  581,   -1,  583,
  584,  585,  586,  587,   -1,  589,  590,  591,  592,  593,
  594,  595,  596,  597,  598,  599,  600,  601,  602,  603,
  604,  605,  606,  607,  608,  609,  610,  611,  612,  613,
  614,  615,  616,  617,   -1,  619,  620,   -1,   -1,   -1,
  624,  625,  626,  627,  628,  629,   -1,   -1,  632,  633,
  634,  635,  636,  637,  638,  639,  640,   -1,   44,   -1,
   -1,   -1,  646,   -1,  648,  649,  650,  651,  652,  653,
   -1,  655,   -1,   -1,   60,   61,   62,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  672,  673,
   -1,  675,  676,  677,  678,  679,  680,  681,  682,  683,
   -1,  685,  686,  687,  688,  689,  690,  691,  692,  693,
  694,  695,  696,   -1,  698,  699,  700,   -1,  702,   -1,
   -1,   -1,  706,  707,  708,  709,  710,  711,  712,  713,
  714,  715,  716,  717,  718,    0,  720,  721,   -1,   -1,
   -1,   -1,   -1,  727,  728,  729,  730,  731,  732,  733,
  734,  735,  736,  737,  738,  739,  740,  741,  742,  743,
  744,   -1,   -1,  747,   -1,  749,  750,  751,  752,  753,
  754,   -1,   -1,   -1,   -1,   -1,   41,   42,   43,   44,
   45,   46,   47,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   59,   60,   61,   62,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   91,   -1,   93,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  262,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  271,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  284,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  318,   -1,   -1,   -1,   -1,   -1,   -1,  325,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  333,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  350,   -1,  352,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  362,   -1,  364,  365,
  366,   -1,   -1,   -1,   -1,  371,   -1,   -1,   -1,   -1,
   -1,   -1,  257,   -1,  259,   -1,   -1,  262,   -1,  264,
  265,  266,   -1,  268,  269,   -1,  271,   -1,   -1,   -1,
   -1,   -1,  277,   -1,  279,  280,  281,  282,   -1,  284,
   -1,   -1,  287,   -1,   -1,  290,  412,   -1,   -1,   -1,
   -1,   -1,  297,   -1,  299,  300,   -1,  302,  303,  304,
  305,  306,   -1,   -1,  309,   -1,   -1,  312,   -1,  314,
  315,   -1,   -1,  318,  319,  320,   -1,  322,  323,  324,
  325,  326,  327,  328,   -1,   -1,   -1,  332,  333,  334,
  335,  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  348,  349,  350,  351,  352,  353,   -1,
   -1,   -1,  357,  358,   -1,  360,   -1,  362,   -1,  364,
  365,  366,   -1,  368,   -1,  370,  371,  372,   -1,  374,
  375,  376,  377,  378,   -1,  380,  381,   -1,   -1,   -1,
   -1,  386,   -1,  388,  389,   -1,   -1,  392,  393,  394,
   -1,   -1,  397,   -1,  399,   -1,   -1,  402,  403,  404,
  405,   -1,  407,   -1,  409,   -1,  411,  412,  413,  414,
  415,   -1,   -1,  418,  419,  420,   -1,  422,  423,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  433,  434,
  435,   -1,  437,  438,  439,   -1,   -1,   -1,   -1,   -1,
  445,   -1,   -1,   -1,  449,   -1,  451,  452,  453,  454,
   -1,  456,   -1,  458,  459,  460,  461,  462,  463,  464,
  465,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  475,  476,   -1,  478,  479,  480,  602,   -1,   -1,   -1,
  485,  486,  487,  488,  489,   -1,  491,   -1,   -1,  494,
  495,   -1,   -1,   -1,  499,  500,   -1,   -1,  503,  504,
  505,  506,  507,  508,  509,  510,  511,   -1,  513,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  523,  524,
  525,  526,  527,  528,  529,  530,  531,  532,  533,  534,
  535,  536,  537,  538,  539,  540,  541,  542,  543,   -1,
  545,  546,  547,  548,  549,  550,   -1,  552,  553,  554,
  555,   -1,  557,   -1,  559,  560,  561,  562,  563,  564,
  565,  566,  567,  568,  569,  570,  571,  572,  573,  574,
  575,  576,   -1,  578,  579,  580,  581,   -1,  583,  584,
  585,  586,  587,   -1,  589,  590,  591,  592,  593,  594,
  595,  596,  597,  598,  599,  600,  601,  602,  603,  604,
  605,  606,  607,  608,  609,  610,  611,  612,  613,  614,
  615,  616,  617,   -1,  619,  620,   -1,   -1,   -1,  624,
  625,  626,  627,  628,  629,   -1,   -1,  632,  633,  634,
  635,  636,  637,  638,  639,  640,   -1,   44,   -1,   -1,
   -1,  646,   -1,  648,  649,  650,  651,  652,  653,   -1,
  655,   -1,   -1,   60,   61,   62,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  672,  673,   -1,
  675,  676,  677,  678,  679,  680,  681,  682,  683,   -1,
  685,  686,  687,  688,  689,  690,  691,  692,  693,  694,
  695,  696,   -1,  698,  699,  700,   -1,  702,   -1,   -1,
   -1,  706,  707,  708,  709,  710,  711,  712,  713,  714,
  715,  716,  717,  718,    0,  720,  721,   -1,   -1,   -1,
   -1,   -1,  727,  728,  729,  730,  731,  732,  733,  734,
  735,  736,  737,  738,  739,  740,  741,  742,  743,  744,
   -1,   -1,  747,   -1,  749,  750,  751,  752,  753,  754,
   -1,   -1,   -1,   -1,   -1,   41,   42,   43,   44,   45,
   46,   47,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   59,   60,   61,   62,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   91,   -1,   93,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  262,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  271,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  284,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  318,   -1,   -1,   -1,   -1,   -1,   -1,  325,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  333,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  350,   -1,  352,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  362,   -1,  364,  365,  366,
   -1,   -1,   -1,   -1,  371,   -1,   -1,   -1,   -1,   -1,
   -1,  257,   -1,  259,   -1,   -1,  262,   -1,  264,  265,
  266,   -1,  268,  269,   -1,  271,   -1,   -1,   -1,   -1,
   -1,  277,   -1,  279,  280,  281,  282,   -1,  284,   -1,
   -1,  287,   -1,   -1,  290,  412,   -1,   -1,   -1,   -1,
   -1,  297,   -1,  299,  300,   -1,  302,  303,  304,  305,
  306,   -1,   -1,  309,   -1,   -1,  312,   -1,  314,  315,
   -1,   -1,  318,  319,  320,   -1,  322,  323,  324,  325,
  326,  327,  328,   -1,   -1,   -1,  332,  333,  334,  335,
  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  348,  349,  350,  351,  352,  353,   -1,   -1,
   -1,  357,  358,   -1,  360,   -1,  362,   -1,  364,  365,
  366,   -1,  368,   -1,  370,  371,  372,   -1,  374,  375,
  376,  377,  378,   -1,  380,  381,   -1,   -1,   -1,   -1,
  386,   -1,  388,  389,   -1,   -1,  392,  393,  394,   -1,
   -1,  397,   -1,  399,   -1,   -1,  402,  403,  404,  405,
   -1,  407,   -1,  409,   -1,  411,  412,  413,  414,  415,
   -1,   -1,  418,  419,  420,   -1,  422,  423,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  433,  434,  435,
   -1,  437,  438,  439,   -1,   -1,   -1,   -1,   -1,  445,
   -1,   -1,   -1,  449,   -1,  451,  452,  453,  454,   -1,
  456,   -1,  458,  459,  460,  461,  462,  463,  464,  465,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  475,
  476,   -1,  478,  479,  480,  602,   -1,   -1,   -1,  485,
  486,  487,  488,  489,   -1,  491,   -1,   -1,  494,  495,
   -1,   -1,   -1,  499,  500,   -1,   -1,  503,  504,  505,
  506,  507,  508,  509,  510,  511,   -1,  513,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  523,  524,  525,
  526,  527,  528,  529,  530,  531,  532,  533,  534,  535,
  536,  537,  538,  539,  540,  541,  542,  543,   -1,  545,
  546,  547,  548,  549,  550,   -1,  552,  553,  554,  555,
   -1,  557,   -1,  559,  560,  561,  562,  563,  564,  565,
  566,  567,  568,  569,  570,  571,  572,  573,  574,  575,
  576,   -1,  578,  579,  580,  581,   -1,  583,  584,  585,
  586,  587,   -1,  589,  590,  591,  592,  593,  594,  595,
  596,  597,  598,  599,  600,  601,  602,  603,  604,  605,
  606,  607,  608,  609,  610,  611,  612,  613,  614,  615,
  616,  617,   -1,  619,  620,   -1,   -1,   -1,  624,  625,
  626,  627,  628,  629,   -1,   -1,  632,  633,  634,  635,
  636,  637,  638,  639,  640,   -1,   44,   -1,   -1,   -1,
  646,   -1,  648,  649,  650,  651,  652,  653,   -1,  655,
   -1,   -1,   60,   61,   62,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  672,  673,   -1,  675,
  676,  677,  678,  679,  680,  681,  682,  683,   -1,  685,
  686,  687,  688,  689,  690,  691,  692,  693,  694,  695,
  696,   -1,  698,  699,  700,   -1,  702,   -1,   -1,   -1,
  706,  707,  708,  709,  710,  711,  712,  713,  714,  715,
  716,  717,  718,    0,  720,  721,   -1,   -1,   -1,   -1,
   -1,  727,  728,  729,  730,  731,  732,  733,  734,  735,
  736,  737,  738,  739,  740,  741,  742,  743,  744,   -1,
   -1,  747,   -1,  749,  750,  751,  752,  753,  754,   -1,
   -1,   -1,   -1,   -1,   41,   42,   43,   44,   45,   46,
   47,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   59,   60,   61,   62,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   91,   -1,   93,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  262,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  271,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  284,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  318,   -1,   -1,   -1,   -1,   -1,   -1,  325,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  333,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  350,   -1,  352,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  362,   -1,  364,  365,  366,   -1,
   -1,   -1,   -1,  371,   -1,   -1,   -1,   -1,   -1,   -1,
  257,   -1,  259,   -1,   -1,  262,   -1,  264,  265,  266,
   -1,  268,  269,   -1,  271,   -1,   -1,   -1,   -1,   -1,
  277,   -1,  279,  280,  281,  282,   -1,  284,   -1,   -1,
  287,   -1,   -1,  290,  412,   -1,   -1,   -1,   -1,   -1,
  297,   -1,  299,  300,   -1,  302,  303,  304,  305,  306,
   -1,   -1,  309,   -1,   -1,  312,   -1,  314,  315,   -1,
   -1,  318,  319,  320,   -1,  322,  323,  324,  325,  326,
  327,  328,   -1,   -1,   -1,  332,  333,  334,  335,  336,
   -1,   -1,   -1,  340,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  348,  349,  350,  351,  352,  353,   -1,   -1,   -1,
  357,  358,   -1,  360,   -1,  362,   -1,  364,  365,  366,
   -1,  368,   -1,  370,  371,  372,   -1,  374,  375,  376,
  377,  378,   -1,  380,  381,   -1,   -1,   -1,   -1,  386,
   -1,  388,  389,   -1,   -1,  392,  393,  394,   -1,   -1,
  397,   -1,  399,   -1,   -1,  402,  403,  404,  405,   -1,
  407,   -1,  409,   -1,  411,  412,  413,  414,  415,   -1,
   -1,  418,  419,  420,   -1,  422,  423,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  433,  434,  435,   -1,
  437,  438,  439,   -1,   -1,   -1,   -1,   -1,  445,   -1,
   -1,   -1,  449,   -1,  451,  452,  453,  454,   -1,  456,
   -1,  458,  459,  460,  461,  462,  463,  464,  465,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  475,  476,
   -1,  478,  479,  480,  602,   -1,   -1,   -1,  485,  486,
  487,  488,  489,   -1,  491,   -1,   -1,  494,  495,   -1,
   -1,   -1,  499,  500,   -1,   -1,  503,  504,  505,  506,
  507,  508,  509,  510,  511,   -1,  513,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  523,  524,  525,  526,
  527,  528,  529,  530,  531,  532,  533,  534,  535,  536,
  537,  538,  539,  540,  541,  542,  543,   -1,  545,  546,
  547,  548,  549,  550,   -1,  552,  553,  554,  555,   -1,
  557,   -1,  559,  560,  561,  562,  563,  564,  565,  566,
  567,  568,  569,  570,  571,  572,  573,  574,  575,  576,
   -1,  578,  579,  580,  581,   -1,  583,  584,  585,  586,
  587,   -1,  589,  590,  591,  592,  593,  594,  595,  596,
  597,  598,  599,  600,  601,  602,  603,  604,  605,  606,
  607,  608,  609,  610,  611,  612,  613,  614,  615,  616,
  617,   -1,  619,  620,   -1,   -1,   -1,  624,  625,  626,
  627,  628,  629,   -1,   -1,  632,  633,  634,  635,  636,
  637,  638,  639,  640,   -1,   44,   -1,   -1,   -1,  646,
   -1,  648,  649,  650,  651,  652,  653,   -1,  655,   -1,
   -1,   60,   61,   62,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  672,  673,   -1,  675,  676,
  677,  678,  679,  680,  681,  682,  683,   -1,  685,  686,
  687,  688,  689,  690,  691,  692,  693,  694,  695,  696,
   -1,  698,  699,  700,   -1,  702,   -1,   -1,   -1,  706,
  707,  708,  709,  710,  711,  712,  713,  714,  715,  716,
  717,  718,    0,  720,  721,   -1,   -1,   -1,   -1,   -1,
  727,  728,  729,  730,  731,  732,  733,  734,  735,  736,
  737,  738,  739,  740,  741,  742,  743,  744,   -1,   -1,
  747,   -1,  749,  750,  751,  752,  753,  754,   -1,   -1,
   -1,   -1,   -1,   41,   42,   43,   44,   45,   46,   47,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   59,   60,   61,   62,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   91,   -1,   93,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  262,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  271,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  284,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  318,
   -1,   -1,   -1,   -1,   -1,   -1,  325,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  333,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  350,   -1,  352,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  362,   -1,  364,  365,  366,   -1,   -1,
   -1,   -1,  371,   -1,   -1,   -1,   -1,   -1,   -1,  257,
   -1,  259,   -1,   -1,  262,   -1,  264,  265,  266,   -1,
  268,  269,   -1,  271,   -1,   -1,   -1,   -1,   -1,  277,
   -1,  279,  280,  281,  282,   -1,  284,   -1,   -1,  287,
   -1,   -1,  290,  412,   -1,   -1,   -1,   -1,   -1,  297,
   -1,  299,  300,   -1,  302,  303,  304,  305,  306,   -1,
   -1,  309,   -1,   -1,  312,   -1,  314,  315,   -1,   -1,
  318,  319,  320,   -1,  322,  323,  324,  325,  326,  327,
  328,   -1,   -1,   -1,  332,  333,  334,  335,  336,   -1,
   -1,   -1,  340,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  348,  349,  350,  351,  352,  353,   -1,   -1,   -1,  357,
  358,   -1,  360,   -1,  362,   -1,  364,  365,  366,   -1,
  368,   -1,  370,  371,  372,   -1,  374,  375,  376,  377,
  378,   -1,  380,  381,   -1,   -1,   -1,   -1,  386,   -1,
  388,  389,   -1,   -1,  392,  393,  394,   -1,   -1,  397,
   -1,  399,   -1,   -1,  402,  403,  404,  405,   -1,  407,
   -1,  409,   -1,  411,  412,  413,  414,  415,   -1,   -1,
  418,  419,  420,   -1,  422,  423,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  433,  434,  435,   -1,  437,
  438,  439,   -1,   -1,   -1,   -1,   -1,  445,   -1,   -1,
   -1,  449,   -1,  451,  452,  453,  454,   -1,  456,   -1,
  458,  459,  460,  461,  462,  463,  464,  465,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  475,  476,   -1,
  478,  479,  480,  602,   -1,   -1,   -1,  485,  486,  487,
  488,  489,   -1,  491,   -1,   -1,  494,  495,   -1,   -1,
   -1,  499,  500,   -1,   -1,  503,  504,  505,  506,  507,
  508,  509,  510,  511,   -1,  513,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  523,  524,  525,  526,  527,
  528,  529,  530,  531,  532,  533,  534,  535,  536,  537,
  538,  539,  540,  541,  542,  543,   -1,  545,  546,  547,
  548,  549,  550,   -1,  552,  553,  554,  555,   -1,  557,
   -1,  559,  560,  561,  562,  563,  564,  565,  566,  567,
  568,  569,  570,  571,  572,  573,  574,  575,  576,   -1,
  578,  579,  580,  581,   -1,  583,  584,  585,  586,  587,
   -1,  589,  590,  591,  592,  593,  594,  595,  596,  597,
  598,  599,  600,  601,  602,  603,  604,  605,  606,  607,
  608,  609,  610,  611,  612,  613,  614,  615,  616,  617,
   -1,  619,  620,   -1,   -1,   -1,  624,  625,  626,  627,
  628,  629,   -1,   -1,  632,  633,  634,  635,  636,  637,
  638,  639,  640,   -1,   44,   -1,   -1,   -1,  646,   -1,
  648,  649,  650,  651,  652,  653,   -1,  655,   -1,   -1,
   60,   61,   62,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  672,  673,   -1,  675,  676,  677,
  678,  679,  680,  681,  682,  683,   -1,  685,  686,  687,
  688,  689,  690,  691,  692,  693,  694,  695,  696,   -1,
  698,  699,  700,   -1,  702,   -1,   -1,   -1,  706,  707,
  708,  709,  710,  711,  712,  713,  714,  715,  716,  717,
  718,    0,  720,  721,   -1,   -1,   -1,   -1,   -1,  727,
  728,  729,  730,  731,  732,  733,  734,  735,  736,  737,
  738,  739,  740,  741,  742,  743,  744,   -1,   -1,  747,
   -1,  749,  750,  751,  752,  753,  754,   -1,   -1,   -1,
   -1,   -1,   41,   42,   43,   44,   45,   46,   47,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   59,   60,   61,   62,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   91,   -1,   93,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  262,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  271,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  284,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  318,   -1,
   -1,   -1,   -1,   -1,   -1,  325,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  333,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  350,   -1,  352,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  362,   -1,  364,  365,  366,   -1,   -1,   -1,
   -1,  371,   -1,   -1,   -1,   -1,   -1,   -1,  257,   -1,
  259,   -1,   -1,  262,   -1,  264,  265,  266,   -1,  268,
  269,   -1,  271,   -1,   -1,   -1,   -1,   -1,  277,   -1,
  279,  280,  281,  282,   -1,  284,   -1,   -1,  287,   -1,
   -1,  290,  412,   -1,   -1,   -1,   -1,   -1,  297,   -1,
  299,  300,   -1,  302,  303,  304,  305,  306,   -1,   -1,
  309,   -1,   -1,  312,   -1,  314,  315,   -1,   -1,  318,
  319,  320,   -1,  322,  323,  324,  325,  326,  327,  328,
   -1,   -1,   -1,  332,  333,  334,  335,  336,   -1,   -1,
   -1,  340,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  348,
  349,  350,  351,  352,  353,   -1,   -1,   -1,  357,  358,
   -1,  360,   -1,  362,   -1,  364,  365,  366,   -1,  368,
   -1,  370,  371,  372,   -1,  374,  375,  376,  377,  378,
   -1,  380,  381,   -1,   -1,   -1,   -1,  386,   -1,  388,
  389,   -1,   -1,  392,  393,  394,   -1,   -1,  397,   -1,
  399,   -1,   -1,  402,  403,  404,  405,   -1,  407,   -1,
  409,   -1,  411,  412,  413,  414,  415,   -1,   -1,  418,
  419,  420,   -1,  422,  423,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  433,  434,  435,   -1,  437,  438,
  439,   -1,   -1,   -1,   -1,   -1,  445,   -1,   -1,   -1,
  449,   -1,  451,  452,  453,  454,   -1,  456,   -1,  458,
  459,  460,  461,  462,  463,  464,  465,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  475,  476,   -1,  478,
  479,  480,  602,   -1,   -1,   -1,  485,  486,  487,  488,
  489,   -1,  491,   -1,   -1,  494,  495,   -1,   -1,   -1,
  499,  500,   -1,   -1,  503,  504,  505,  506,  507,  508,
  509,  510,  511,   -1,  513,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  523,  524,  525,  526,  527,  528,
  529,  530,  531,  532,  533,  534,  535,  536,  537,  538,
  539,  540,  541,  542,  543,   -1,  545,  546,  547,  548,
  549,  550,   -1,  552,  553,  554,  555,   -1,  557,   -1,
  559,  560,  561,  562,  563,  564,  565,  566,  567,  568,
  569,  570,  571,  572,  573,  574,  575,  576,   -1,  578,
  579,  580,  581,   -1,  583,  584,  585,  586,  587,   -1,
  589,  590,  591,  592,  593,  594,  595,  596,  597,  598,
  599,  600,  601,  602,  603,  604,  605,  606,  607,  608,
  609,  610,  611,  612,  613,  614,  615,  616,  617,   -1,
  619,  620,   -1,   -1,   -1,  624,  625,  626,  627,  628,
  629,   -1,   -1,  632,  633,  634,  635,  636,  637,  638,
  639,  640,   -1,   44,   -1,   -1,   -1,  646,   -1,  648,
  649,  650,  651,  652,  653,   -1,  655,   -1,   -1,   60,
   61,   62,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  672,  673,   -1,  675,  676,  677,  678,
  679,  680,  681,  682,  683,   -1,  685,  686,  687,  688,
  689,  690,  691,  692,  693,  694,  695,  696,   -1,  698,
  699,  700,   -1,  702,   -1,   -1,   -1,  706,  707,  708,
  709,  710,  711,  712,  713,  714,  715,  716,  717,  718,
    0,  720,  721,   -1,   -1,   -1,   -1,   -1,  727,  728,
  729,  730,  731,  732,  733,  734,  735,  736,  737,  738,
  739,  740,  741,  742,  743,  744,   -1,   -1,  747,   -1,
  749,  750,  751,  752,  753,  754,   -1,   -1,   -1,   -1,
   -1,   41,   42,   43,   44,   45,   46,   47,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   59,
   60,   61,   62,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   91,   -1,   93,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  262,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  271,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  284,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  318,   -1,   -1,
   -1,   -1,   -1,   -1,  325,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  333,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  350,
   -1,  352,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  362,   -1,  364,  365,  366,   -1,   -1,   -1,   -1,
  371,   -1,   -1,   -1,   -1,   -1,   -1,  257,   -1,  259,
   -1,   -1,  262,   -1,  264,  265,  266,   -1,  268,  269,
   -1,  271,   -1,   -1,   -1,   -1,   -1,  277,   -1,  279,
  280,  281,  282,   -1,  284,   -1,   -1,  287,   -1,   -1,
  290,  412,   -1,   -1,   -1,   -1,   -1,  297,   -1,  299,
  300,   -1,  302,  303,  304,  305,  306,   -1,   -1,  309,
   -1,   -1,  312,   -1,  314,  315,   -1,   -1,  318,  319,
  320,   -1,  322,  323,  324,  325,  326,  327,  328,   -1,
   -1,   -1,  332,  333,  334,  335,  336,   -1,   -1,   -1,
  340,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  348,  349,
  350,  351,  352,  353,   -1,   -1,   -1,  357,  358,   -1,
  360,   -1,  362,   -1,  364,  365,  366,   -1,  368,   -1,
  370,  371,  372,   -1,  374,  375,  376,  377,  378,   -1,
  380,  381,   -1,   -1,   -1,   -1,  386,   -1,  388,  389,
   -1,   -1,  392,  393,  394,   -1,   -1,  397,   -1,  399,
   -1,   -1,  402,  403,  404,  405,   -1,  407,   -1,  409,
   -1,  411,  412,  413,  414,  415,   -1,   -1,  418,  419,
  420,   -1,  422,  423,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  433,  434,  435,   -1,  437,  438,  439,
   -1,   -1,   -1,   -1,   -1,  445,   -1,   -1,   -1,  449,
   -1,  451,  452,  453,  454,   -1,  456,   -1,  458,  459,
  460,  461,  462,  463,  464,  465,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  475,  476,   -1,  478,  479,
  480,  602,   -1,   -1,   -1,  485,  486,  487,  488,  489,
   -1,  491,   -1,   -1,  494,  495,   -1,   -1,   -1,  499,
  500,   -1,   -1,  503,  504,  505,  506,  507,  508,  509,
  510,  511,   -1,  513,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  523,  524,  525,  526,  527,  528,  529,
  530,  531,  532,  533,  534,  535,  536,  537,  538,  539,
  540,  541,  542,  543,   -1,  545,  546,  547,  548,  549,
  550,   -1,  552,  553,  554,  555,   -1,  557,   -1,  559,
  560,  561,  562,  563,  564,  565,  566,  567,  568,  569,
  570,  571,  572,  573,  574,  575,  576,   -1,  578,  579,
  580,  581,   -1,  583,  584,  585,  586,  587,   -1,  589,
  590,  591,  592,  593,  594,  595,  596,  597,  598,  599,
  600,  601,  602,  603,  604,  605,  606,  607,  608,  609,
  610,  611,  612,  613,  614,  615,  616,  617,   -1,  619,
  620,   -1,   -1,   -1,  624,  625,  626,  627,  628,  629,
   -1,   -1,  632,  633,  634,  635,  636,  637,  638,  639,
  640,   -1,   44,   -1,   -1,   -1,  646,   -1,  648,  649,
  650,  651,  652,  653,   -1,  655,   -1,   -1,   60,   61,
   62,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  672,  673,   -1,  675,  676,  677,  678,  679,
  680,  681,  682,  683,   -1,  685,  686,  687,  688,  689,
  690,  691,  692,  693,  694,  695,  696,   -1,  698,  699,
  700,   -1,  702,   -1,   -1,   -1,  706,  707,  708,  709,
  710,  711,  712,  713,  714,  715,  716,  717,  718,    0,
  720,  721,   -1,   -1,   -1,   -1,   -1,  727,  728,  729,
  730,  731,  732,  733,  734,  735,  736,  737,  738,  739,
  740,  741,  742,  743,  744,   -1,   -1,  747,   -1,  749,
  750,  751,  752,  753,  754,   -1,   -1,   -1,   -1,   -1,
   41,   42,   43,   44,   45,   46,   47,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   59,   60,
   61,   62,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   91,   -1,   93,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  262,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  271,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  284,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  318,   -1,   -1,   -1,
   -1,   -1,   -1,  325,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  333,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  350,   -1,
  352,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  362,   -1,  364,  365,  366,   -1,   -1,   -1,   -1,  371,
   -1,   -1,   -1,   -1,   -1,   -1,  257,   -1,  259,   -1,
   -1,  262,   -1,  264,  265,  266,   -1,  268,  269,   -1,
  271,   -1,   -1,   -1,   -1,   -1,  277,   -1,  279,  280,
  281,  282,   -1,  284,   -1,   -1,  287,   -1,   -1,  290,
  412,   -1,   -1,   -1,   -1,   -1,  297,   -1,  299,  300,
   -1,  302,  303,  304,  305,  306,   -1,   -1,  309,   -1,
   -1,  312,   -1,  314,  315,   -1,   -1,  318,  319,  320,
   -1,  322,  323,  324,  325,  326,  327,  328,   -1,   -1,
   -1,  332,  333,  334,  335,  336,   -1,   -1,   -1,  340,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  348,  349,  350,
  351,  352,  353,   -1,   -1,   -1,  357,  358,   -1,  360,
   -1,  362,   -1,  364,  365,  366,   -1,  368,   -1,  370,
  371,  372,   -1,  374,  375,  376,  377,  378,   -1,  380,
  381,   -1,   -1,   -1,   -1,  386,   -1,  388,  389,   -1,
   -1,  392,  393,  394,   -1,   -1,  397,   -1,  399,   -1,
   -1,  402,  403,  404,  405,   -1,  407,   -1,  409,   -1,
  411,  412,  413,  414,  415,   -1,   -1,  418,  419,  420,
   -1,  422,  423,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  433,  434,  435,   -1,  437,  438,  439,   -1,
   -1,   -1,   -1,   -1,  445,   -1,   -1,   -1,  449,   -1,
  451,  452,  453,  454,   -1,  456,   -1,  458,  459,  460,
  461,  462,  463,  464,  465,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  475,  476,   -1,  478,  479,  480,
  602,   -1,   -1,   -1,  485,  486,  487,  488,  489,   -1,
  491,   -1,   -1,  494,  495,   -1,   -1,   -1,  499,  500,
   -1,   -1,  503,  504,  505,  506,  507,  508,  509,  510,
  511,   -1,  513,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  523,  524,  525,  526,  527,  528,  529,  530,
  531,  532,  533,  534,  535,  536,  537,  538,  539,  540,
  541,  542,  543,   -1,  545,  546,  547,  548,  549,  550,
   -1,  552,  553,  554,  555,   -1,  557,   -1,  559,  560,
  561,  562,  563,  564,  565,  566,  567,  568,  569,  570,
  571,  572,  573,  574,  575,  576,   -1,  578,  579,  580,
  581,   -1,  583,  584,  585,  586,  587,   -1,  589,  590,
  591,  592,  593,  594,  595,  596,  597,  598,  599,  600,
  601,  602,  603,  604,  605,  606,  607,  608,  609,  610,
  611,  612,  613,  614,  615,  616,  617,   -1,  619,  620,
   -1,   -1,   -1,  624,  625,  626,  627,  628,  629,   -1,
   -1,  632,  633,  634,  635,  636,  637,  638,  639,  640,
   -1,   44,   -1,   -1,   -1,  646,   -1,  648,  649,  650,
  651,  652,  653,   -1,  655,   -1,   -1,   60,   61,   62,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  672,  673,   -1,  675,  676,  677,  678,  679,  680,
  681,  682,  683,   -1,  685,  686,  687,  688,  689,  690,
  691,  692,  693,  694,  695,  696,   -1,  698,  699,  700,
   -1,  702,   -1,   -1,   -1,  706,  707,  708,  709,  710,
  711,  712,  713,  714,  715,  716,  717,  718,    0,  720,
  721,   -1,   -1,   -1,   -1,   -1,  727,  728,  729,  730,
  731,  732,  733,  734,  735,  736,  737,  738,  739,  740,
  741,  742,  743,  744,   -1,   -1,  747,   -1,  749,  750,
  751,  752,  753,  754,   -1,   -1,   -1,   -1,   -1,   41,
   42,   43,   44,   45,   46,   47,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   59,   60,   61,
   62,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   91,
   -1,   93,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  262,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  271,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  284,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  318,   -1,   -1,   -1,   -1,
   -1,   -1,  325,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  333,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  350,   -1,  352,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  362,
   -1,  364,  365,  366,   -1,   -1,   -1,   -1,  371,   -1,
   -1,   -1,   -1,   -1,   -1,  257,   -1,  259,   -1,   -1,
  262,   -1,  264,  265,  266,   -1,  268,  269,   -1,  271,
   -1,   -1,   -1,   -1,   -1,  277,   -1,  279,  280,  281,
  282,   -1,  284,   -1,   -1,  287,   -1,   -1,  290,  412,
   -1,   -1,   -1,   -1,   -1,  297,   -1,  299,  300,   -1,
  302,  303,  304,  305,  306,   -1,   -1,  309,   -1,   -1,
  312,   -1,  314,  315,   -1,   -1,  318,  319,  320,   -1,
  322,  323,  324,  325,  326,  327,  328,   -1,   -1,   -1,
  332,  333,  334,  335,  336,   -1,   -1,   -1,  340,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  348,  349,  350,  351,
  352,  353,   -1,   -1,   -1,  357,  358,   -1,  360,   -1,
  362,   -1,  364,  365,  366,   -1,  368,   -1,  370,  371,
  372,   -1,  374,  375,  376,  377,  378,   -1,  380,  381,
   -1,   -1,   -1,   -1,  386,   -1,  388,  389,   -1,   -1,
  392,  393,  394,   -1,   -1,  397,   -1,  399,   -1,   -1,
  402,  403,  404,  405,   -1,  407,   -1,  409,   -1,  411,
  412,  413,  414,  415,   -1,   -1,  418,  419,  420,   -1,
  422,  423,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  433,  434,  435,   -1,  437,  438,  439,   -1,   -1,
   -1,   -1,   -1,  445,   -1,   -1,   -1,  449,   -1,  451,
  452,  453,  454,   -1,  456,   -1,  458,  459,  460,  461,
  462,  463,  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  475,  476,   -1,  478,  479,  480,  602,
   -1,   -1,   -1,  485,  486,  487,  488,  489,   -1,  491,
   -1,   -1,  494,  495,   -1,   -1,   -1,  499,  500,   -1,
   -1,  503,  504,  505,  506,  507,  508,  509,  510,  511,
   -1,  513,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  523,  524,  525,  526,  527,  528,  529,  530,  531,
  532,  533,  534,  535,  536,  537,  538,  539,  540,  541,
  542,  543,   -1,  545,  546,  547,  548,  549,  550,   -1,
  552,  553,  554,  555,   -1,  557,   -1,  559,  560,  561,
  562,  563,  564,  565,  566,  567,  568,  569,  570,  571,
  572,  573,  574,  575,  576,   -1,  578,  579,  580,  581,
   -1,  583,  584,  585,  586,  587,   -1,  589,  590,  591,
  592,  593,  594,  595,  596,  597,  598,  599,  600,  601,
  602,  603,  604,  605,  606,  607,  608,  609,  610,  611,
  612,  613,  614,  615,  616,  617,   -1,  619,  620,   -1,
   -1,   -1,  624,  625,  626,  627,  628,  629,   -1,   -1,
  632,  633,  634,  635,  636,  637,  638,  639,  640,   -1,
   44,   -1,   -1,   -1,  646,   -1,  648,  649,  650,  651,
  652,  653,   -1,  655,   -1,   -1,   60,   61,   62,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  672,  673,   -1,  675,  676,  677,  678,  679,  680,  681,
  682,  683,   -1,  685,  686,  687,  688,  689,  690,  691,
  692,  693,  694,  695,  696,   -1,  698,  699,  700,   -1,
  702,   -1,   -1,   -1,  706,  707,  708,  709,  710,  711,
  712,  713,  714,  715,  716,  717,  718,    0,  720,  721,
   -1,   -1,   -1,   -1,   -1,  727,  728,  729,  730,  731,
  732,  733,  734,  735,  736,  737,  738,  739,  740,  741,
  742,  743,  744,   -1,   -1,  747,   -1,  749,  750,  751,
  752,  753,  754,   -1,   -1,   -1,   -1,   -1,   41,   42,
   43,   44,   45,   46,   47,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   59,   60,   61,   62,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   91,   -1,
   93,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  262,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  271,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  284,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  318,   -1,   -1,   -1,   -1,   -1,
   -1,  325,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  333,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  350,   -1,  352,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  362,   -1,
  364,  365,  366,   -1,   -1,   -1,   -1,  371,   -1,   -1,
   -1,   -1,   -1,   -1,  257,   -1,  259,   -1,   -1,  262,
   -1,  264,  265,  266,   -1,  268,  269,   -1,  271,   -1,
   -1,   -1,   -1,   -1,  277,   -1,  279,  280,  281,  282,
   -1,  284,   -1,   -1,  287,   -1,   -1,  290,  412,   -1,
   -1,   -1,   -1,   -1,  297,   -1,  299,  300,   -1,  302,
  303,  304,  305,  306,   -1,   -1,  309,   -1,   -1,  312,
   -1,  314,  315,   -1,   -1,  318,  319,  320,   -1,  322,
  323,  324,  325,  326,  327,  328,   -1,   -1,   -1,  332,
  333,  334,  335,  336,   -1,   -1,   -1,  340,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  348,  349,  350,  351,  352,
  353,   -1,   -1,   -1,  357,  358,   -1,  360,   -1,  362,
   -1,  364,  365,  366,   -1,  368,   -1,  370,  371,  372,
   -1,  374,  375,  376,  377,  378,   -1,  380,  381,   -1,
   -1,   -1,   -1,  386,   -1,  388,  389,   -1,   -1,  392,
  393,  394,   -1,   -1,  397,   -1,  399,   -1,   -1,  402,
  403,  404,  405,   -1,  407,   -1,  409,   -1,  411,  412,
  413,  414,  415,   -1,   -1,  418,  419,  420,   -1,  422,
  423,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  433,  434,  435,   -1,  437,  438,  439,   -1,   -1,   -1,
   -1,   -1,  445,   -1,   -1,   -1,  449,   -1,  451,  452,
  453,  454,   -1,  456,   -1,  458,  459,  460,  461,  462,
  463,  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  475,  476,   -1,  478,  479,  480,  602,   -1,
   -1,   -1,  485,  486,  487,  488,  489,   -1,  491,   -1,
   -1,  494,  495,   -1,   -1,   -1,  499,  500,   -1,   -1,
  503,  504,  505,  506,  507,  508,  509,  510,  511,   -1,
  513,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  523,  524,  525,  526,  527,  528,  529,  530,  531,  532,
  533,  534,  535,  536,  537,  538,  539,  540,  541,  542,
  543,   -1,  545,  546,  547,  548,  549,  550,   -1,  552,
  553,  554,  555,   -1,  557,   -1,  559,  560,  561,  562,
  563,  564,  565,  566,  567,  568,  569,  570,  571,  572,
  573,  574,  575,  576,   -1,  578,  579,  580,  581,   -1,
  583,  584,  585,  586,  587,   -1,  589,  590,  591,  592,
  593,  594,  595,  596,  597,  598,  599,  600,  601,  602,
  603,  604,  605,  606,  607,  608,  609,  610,  611,  612,
  613,  614,  615,  616,  617,   -1,  619,  620,   -1,   -1,
   -1,  624,  625,  626,  627,  628,  629,   -1,   -1,  632,
  633,  634,  635,  636,  637,  638,  639,  640,   -1,   44,
   -1,   -1,   -1,  646,   -1,  648,  649,  650,  651,  652,
  653,   -1,  655,   -1,   -1,   60,   61,   62,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  672,
  673,   -1,  675,  676,  677,  678,  679,  680,  681,  682,
  683,   -1,  685,  686,  687,  688,  689,  690,  691,  692,
  693,  694,  695,  696,   -1,  698,  699,  700,   -1,  702,
   -1,   -1,   -1,  706,  707,  708,  709,  710,  711,  712,
  713,  714,  715,  716,  717,  718,    0,  720,  721,   -1,
   -1,   -1,   -1,   -1,  727,  728,  729,  730,  731,  732,
  733,  734,  735,  736,  737,  738,  739,  740,  741,  742,
  743,  744,   -1,   -1,  747,   -1,  749,  750,  751,  752,
  753,  754,   -1,   -1,   -1,   -1,   -1,   41,   42,   43,
   44,   45,   46,   47,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   59,   60,   61,   62,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   91,   -1,   93,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  262,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  271,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  284,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  318,   -1,   -1,   -1,   -1,   -1,   -1,
  325,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  333,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  350,   -1,  352,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  362,   -1,  364,
  365,  366,   -1,   -1,   -1,   -1,  371,   -1,   -1,   -1,
   -1,   -1,   -1,  257,   -1,  259,   -1,   -1,  262,   -1,
  264,  265,  266,   -1,  268,  269,   -1,  271,   -1,   -1,
   -1,   -1,   -1,  277,   -1,  279,  280,  281,  282,   -1,
  284,   -1,   -1,  287,   -1,   -1,  290,  412,   -1,   -1,
   -1,   -1,   -1,  297,   -1,  299,  300,   -1,  302,  303,
  304,  305,  306,   -1,   -1,  309,   -1,   -1,  312,   -1,
  314,  315,   -1,   -1,  318,  319,  320,   -1,  322,  323,
  324,  325,  326,  327,  328,   -1,   -1,   -1,  332,  333,
  334,  335,  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  348,  349,  350,  351,  352,  353,
   -1,   -1,   -1,  357,  358,   -1,  360,   -1,  362,   -1,
  364,  365,  366,   -1,  368,   -1,  370,  371,  372,   -1,
  374,  375,  376,  377,  378,   -1,  380,  381,   -1,   -1,
   -1,   -1,  386,   -1,  388,  389,   -1,   -1,  392,  393,
  394,   -1,   -1,  397,   -1,  399,   -1,   -1,  402,  403,
  404,  405,   -1,  407,   -1,  409,   -1,  411,  412,  413,
  414,  415,   -1,   -1,  418,  419,  420,   -1,  422,  423,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  433,
  434,  435,   -1,  437,  438,  439,   -1,   -1,   -1,   -1,
   -1,  445,   -1,   -1,   -1,  449,   -1,  451,  452,  453,
  454,   -1,  456,   -1,  458,  459,  460,  461,  462,  463,
  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  475,  476,   -1,  478,  479,  480,  602,   -1,   -1,
   -1,  485,  486,  487,  488,  489,   -1,  491,   -1,   -1,
  494,  495,   -1,   -1,   -1,  499,  500,   -1,   -1,  503,
  504,  505,  506,  507,  508,  509,  510,  511,   -1,  513,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  523,
  524,  525,  526,  527,  528,  529,  530,  531,  532,  533,
  534,  535,  536,  537,  538,  539,  540,  541,  542,  543,
   -1,  545,  546,  547,  548,  549,  550,   -1,  552,  553,
  554,  555,   -1,  557,   -1,  559,  560,  561,  562,  563,
  564,  565,  566,  567,  568,  569,  570,  571,  572,  573,
  574,  575,  576,   -1,  578,  579,  580,  581,   -1,  583,
  584,  585,  586,  587,   -1,  589,  590,  591,  592,  593,
  594,  595,  596,  597,  598,  599,  600,  601,  602,  603,
  604,  605,  606,  607,  608,  609,  610,  611,  612,  613,
  614,  615,  616,  617,   -1,  619,  620,   -1,   -1,   -1,
  624,  625,  626,  627,  628,  629,   -1,   -1,  632,  633,
  634,  635,  636,  637,  638,  639,  640,   -1,   44,   -1,
   -1,   -1,  646,   -1,  648,  649,  650,  651,  652,  653,
   -1,  655,   -1,   -1,   60,   61,   62,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  672,  673,
   -1,  675,  676,  677,  678,  679,  680,  681,  682,  683,
   -1,  685,  686,  687,  688,  689,  690,  691,  692,  693,
  694,  695,  696,   -1,  698,  699,  700,   -1,  702,   -1,
   -1,   -1,  706,  707,  708,  709,  710,  711,  712,  713,
  714,  715,  716,  717,  718,    0,  720,  721,   -1,   -1,
   -1,   -1,   -1,  727,  728,  729,  730,  731,  732,  733,
  734,  735,  736,  737,  738,  739,  740,  741,  742,  743,
  744,   -1,   -1,  747,   -1,  749,  750,  751,  752,  753,
  754,   -1,   -1,   -1,   -1,   -1,   41,   42,   43,   44,
   45,   46,   47,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   59,   60,   61,   62,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   91,   -1,   93,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  262,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  271,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  284,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  318,   -1,   -1,   -1,   -1,   -1,   -1,  325,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  333,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  350,   -1,  352,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  362,   -1,  364,  365,
  366,   -1,   -1,   -1,   -1,  371,   -1,   -1,   -1,   -1,
   -1,   -1,  257,   -1,  259,   -1,   -1,  262,   -1,  264,
  265,  266,   -1,  268,  269,   -1,  271,   -1,   -1,   -1,
   -1,   -1,  277,   -1,  279,  280,  281,  282,   -1,  284,
   -1,   -1,  287,   -1,   -1,  290,  412,   -1,   -1,   -1,
   -1,   -1,  297,   -1,  299,  300,   -1,  302,  303,  304,
  305,  306,   -1,   -1,  309,   -1,   -1,  312,   -1,  314,
  315,   -1,   -1,  318,  319,  320,   -1,  322,  323,  324,
  325,  326,  327,  328,   -1,   -1,   -1,  332,  333,  334,
  335,  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  348,  349,  350,  351,  352,  353,   -1,
   -1,   -1,  357,  358,   -1,  360,   -1,  362,   -1,  364,
  365,  366,   -1,  368,   -1,  370,  371,  372,   -1,  374,
  375,  376,  377,  378,   -1,  380,  381,   -1,   -1,   -1,
   -1,  386,   -1,  388,  389,   -1,   -1,  392,  393,  394,
   -1,   -1,  397,   -1,  399,   -1,   -1,  402,  403,  404,
  405,   -1,  407,   -1,  409,   -1,  411,  412,  413,  414,
  415,   -1,   -1,  418,  419,  420,   -1,  422,  423,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  433,  434,
  435,   -1,  437,  438,  439,   -1,   -1,   -1,   -1,   -1,
  445,   -1,   -1,   -1,  449,   -1,  451,  452,  453,  454,
   -1,  456,   -1,  458,  459,  460,  461,  462,  463,  464,
  465,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  475,  476,   -1,  478,  479,  480,  602,   -1,   -1,   -1,
  485,  486,  487,  488,  489,   -1,  491,   -1,   -1,  494,
  495,   -1,   -1,   -1,  499,  500,   -1,   -1,  503,  504,
  505,  506,  507,  508,  509,  510,  511,   -1,  513,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  523,  524,
  525,  526,  527,  528,  529,  530,  531,  532,  533,  534,
  535,  536,  537,  538,  539,  540,  541,  542,  543,   -1,
  545,  546,  547,  548,  549,  550,   -1,  552,  553,  554,
  555,   -1,  557,   -1,  559,  560,  561,  562,  563,  564,
  565,  566,  567,  568,  569,  570,  571,  572,  573,  574,
  575,  576,   -1,  578,  579,  580,  581,   -1,  583,  584,
  585,  586,  587,   -1,  589,  590,  591,  592,  593,  594,
  595,  596,  597,  598,  599,  600,  601,  602,  603,  604,
  605,  606,  607,  608,  609,  610,  611,  612,  613,  614,
  615,  616,  617,   -1,  619,  620,   -1,   -1,   -1,  624,
  625,  626,  627,  628,  629,   -1,   -1,  632,  633,  634,
  635,  636,  637,  638,  639,  640,   -1,   44,   -1,   -1,
   -1,  646,   -1,  648,  649,  650,  651,  652,  653,   -1,
  655,   -1,   -1,   60,   61,   62,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  672,  673,   -1,
  675,  676,  677,  678,  679,  680,  681,  682,  683,   -1,
  685,  686,  687,  688,  689,  690,  691,  692,  693,  694,
  695,  696,   -1,  698,  699,  700,   -1,  702,   -1,   -1,
   -1,  706,  707,  708,  709,  710,  711,  712,  713,  714,
  715,  716,  717,  718,    0,  720,  721,   -1,   -1,   -1,
   -1,   -1,  727,  728,  729,  730,  731,  732,  733,  734,
  735,  736,  737,  738,  739,  740,  741,  742,  743,  744,
   -1,   -1,  747,   -1,  749,  750,  751,  752,  753,  754,
   -1,   -1,   -1,   -1,   -1,   41,   42,   43,   44,   45,
   46,   47,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   59,   60,   61,   62,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   91,   -1,   93,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  262,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  271,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  284,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  318,   -1,   -1,   -1,   -1,   -1,   -1,  325,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  333,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  350,   -1,  352,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  362,   -1,  364,  365,  366,
   -1,   -1,   -1,   -1,  371,   -1,   -1,   -1,   -1,   -1,
   -1,  257,   -1,  259,   -1,   -1,  262,   -1,  264,  265,
  266,   -1,  268,  269,   -1,  271,   -1,   -1,   -1,   -1,
   -1,  277,   -1,  279,  280,  281,  282,   -1,  284,   -1,
   -1,  287,   -1,   -1,  290,  412,   -1,   -1,   -1,   -1,
   -1,  297,   -1,  299,  300,   -1,  302,  303,  304,  305,
  306,   -1,   -1,  309,   -1,   -1,  312,   -1,  314,  315,
   -1,   -1,  318,  319,  320,   -1,  322,  323,  324,  325,
  326,  327,  328,   -1,   -1,   -1,  332,  333,  334,  335,
  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  348,  349,  350,  351,  352,  353,   -1,   -1,
   -1,  357,  358,   -1,  360,   -1,  362,   -1,  364,  365,
  366,   -1,  368,   -1,  370,  371,  372,   -1,  374,  375,
  376,  377,  378,   -1,  380,  381,   -1,   -1,    0,   -1,
  386,   -1,  388,  389,   -1,   -1,  392,  393,  394,   -1,
   -1,  397,   -1,  399,   -1,   -1,  402,  403,  404,  405,
   -1,  407,   -1,  409,   -1,  411,  412,  413,  414,  415,
   -1,   -1,  418,  419,  420,   -1,  422,  423,   -1,   41,
   -1,   -1,   44,   -1,   -1,   -1,   -1,  433,  434,  435,
   -1,  437,  438,  439,   -1,   -1,   -1,   59,   -1,  445,
   -1,   -1,   -1,  449,   -1,  451,  452,  453,  454,   -1,
  456,   -1,  458,  459,  460,  461,  462,  463,  464,  465,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  475,
  476,   -1,  478,  479,  480,  602,   -1,   -1,   -1,  485,
  486,  487,  488,  489,   -1,  491,   -1,   -1,  494,  495,
   -1,   -1,   -1,  499,  500,   -1,   -1,  503,  504,  505,
  506,  507,  508,  509,  510,  511,   -1,  513,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  523,  524,  525,
  526,  527,  528,  529,  530,  531,  532,  533,  534,  535,
  536,  537,  538,  539,  540,  541,  542,  543,   -1,  545,
  546,  547,  548,  549,  550,   -1,  552,  553,  554,  555,
   -1,  557,   -1,  559,  560,  561,  562,  563,  564,  565,
  566,  567,  568,  569,  570,  571,  572,  573,  574,  575,
  576,   -1,  578,  579,  580,  581,   -1,  583,  584,  585,
  586,  587,   -1,  589,  590,  591,  592,  593,  594,  595,
  596,  597,  598,  599,  600,  601,  602,  603,  604,  605,
  606,  607,  608,  609,  610,  611,  612,  613,  614,  615,
  616,  617,   -1,  619,  620,   60,   61,   62,  624,  625,
  626,  627,  628,  629,   -1,   -1,  632,  633,  634,  635,
  636,  637,  638,  639,  640,   -1,   -1,   -1,   -1,   -1,
  646,   -1,  648,  649,  650,  651,  652,  653,   -1,  655,
   -1,   -1,   -1,   -1,  276,  277,   -1,   -1,   -1,   -1,
   -1,  283,   -1,   -1,   -1,   -1,  672,  673,   -1,  675,
  676,  677,  678,  679,  680,  681,  682,  683,   -1,  685,
  686,  687,  688,  689,  690,  691,  692,  693,  694,  695,
  696,   -1,  698,  699,  700,   -1,  702,   -1,   -1,   -1,
  706,  707,  708,  709,  710,  711,  712,  713,  714,  715,
  716,  717,  718,    0,  720,  721,   -1,   -1,   -1,   -1,
   -1,  727,  728,  729,  730,  731,  732,  733,  734,  735,
  736,  737,  738,  739,  740,  741,  742,  743,  744,   -1,
   -1,  747,  364,  749,  750,  751,  752,  753,  754,   -1,
   -1,   -1,   -1,   -1,   41,   42,   43,   44,   45,   46,
   47,   -1,   -1,  385,   -1,   -1,   -1,   -1,   -1,  391,
   -1,   -1,   59,   60,   61,   62,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  424,   -1,   91,   -1,   93,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  262,   -1,   -1,
  265,   -1,   -1,   -1,   -1,   -1,  271,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  284,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  297,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  318,   -1,   -1,   -1,   -1,   -1,   -1,
  325,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  333,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  350,   -1,  352,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  362,   -1,  364,
  365,  366,   -1,   -1,   -1,   -1,  371,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  412,   -1,   -1,
  257,   -1,  259,   -1,   -1,  262,   -1,  264,  265,  266,
   -1,  268,  269,   -1,  271,   -1,   -1,   -1,   -1,   -1,
  277,   -1,  279,  280,  281,  282,   -1,  284,   -1,   -1,
  287,   -1,   -1,  290,   -1,   -1,   -1,   -1,   -1,   -1,
  297,   -1,  299,  300,   -1,  302,  303,  304,  305,  306,
   -1,   -1,  309,   -1,   -1,  312,   -1,  314,  315,   -1,
   -1,  318,  319,  320,   -1,  322,  323,  324,  325,  326,
  327,  328,   -1,   -1,   -1,  332,  333,  334,  335,  336,
   -1,   -1,   -1,  340,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  348,  349,  350,  351,  352,  353,   -1,   -1,   -1,
  357,  358,   -1,  360,   -1,  362,   -1,  364,  365,  366,
   -1,  368,   -1,  370,  371,  372,   -1,  374,  375,  376,
  377,  378,   -1,  380,  381,   -1,   -1,   -1,   -1,  386,
   -1,  388,  389,   -1,   -1,  392,  393,  394,   -1,   -1,
  397,   -1,  399,   -1,   -1,  402,  403,  404,  405,   -1,
  407,   -1,  409,   -1,  411,  412,  413,  414,  415,   -1,
   -1,  418,  419,  420,   -1,  422,  423,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  433,  434,  435,   -1,
  437,  438,  439,   -1,   -1,   -1,   -1,  602,  445,   -1,
   -1,   -1,  449,   -1,  451,  452,  453,  454,   -1,  456,
   -1,  458,  459,  460,  461,  462,  463,  464,  465,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  475,  476,
   -1,  478,  479,  480,   -1,   -1,   -1,   -1,  485,  486,
  487,  488,  489,   -1,  491,   -1,   -1,  494,  495,   -1,
   -1,   -1,  499,  500,   -1,   -1,  503,  504,  505,  506,
  507,  508,  509,  510,  511,   -1,  513,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  523,  524,  525,  526,
  527,  528,  529,  530,  531,  532,  533,  534,  535,  536,
  537,  538,  539,  540,  541,  542,  543,   -1,  545,  546,
  547,  548,  549,  550,   -1,  552,  553,  554,  555,   -1,
  557,   -1,  559,  560,  561,  562,  563,  564,  565,  566,
  567,  568,  569,  570,  571,  572,  573,  574,  575,  576,
   -1,  578,  579,  580,  581,   -1,  583,  584,  585,  586,
  587,   -1,  589,  590,  591,  592,  593,  594,  595,  596,
  597,  598,  599,  600,  601,  602,  603,  604,  605,  606,
  607,  608,  609,  610,  611,  612,  613,  614,  615,  616,
  617,   -1,  619,  620,   60,   61,   62,  624,  625,  626,
  627,  628,  629,   -1,   -1,  632,  633,  634,  635,  636,
  637,  638,  639,  640,   -1,   -1,   -1,   -1,   -1,  646,
   -1,  648,  649,  650,  651,  652,  653,   -1,  655,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  672,  673,   -1,  675,  676,
  677,  678,  679,  680,  681,  682,  683,   -1,  685,  686,
  687,  688,  689,  690,  691,  692,  693,  694,  695,  696,
   -1,  698,  699,  700,   -1,  702,   -1,   -1,   -1,  706,
  707,  708,  709,  710,  711,  712,  713,  714,  715,  716,
  717,  718,    0,  720,  721,   -1,   -1,   -1,   -1,   -1,
  727,  728,  729,  730,  731,  732,  733,  734,  735,  736,
  737,  738,  739,  740,  741,  742,  743,  744,   -1,   -1,
  747,   -1,  749,  750,  751,  752,  753,  754,   -1,   -1,
   -1,   -1,   -1,   41,   42,   43,   44,   45,   46,   47,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   59,   60,   61,   62,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   91,   -1,   93,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  262,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  271,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  284,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  318,   -1,   -1,   -1,   -1,   -1,   -1,  325,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  333,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  350,   -1,  352,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  362,   -1,  364,  365,
  366,   -1,   -1,   -1,   -1,  371,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  412,   -1,   -1,  257,
   -1,  259,   -1,  419,  262,   -1,  264,  265,  266,   -1,
  268,  269,   -1,  271,   -1,   -1,   -1,   -1,   -1,  277,
   -1,  279,  280,  281,  282,   -1,  284,   -1,   -1,  287,
   -1,   -1,  290,   -1,   -1,   -1,   -1,   -1,   -1,  297,
   -1,  299,  300,   -1,  302,  303,  304,  305,  306,   -1,
   -1,  309,   -1,   -1,  312,   -1,  314,  315,   -1,   -1,
  318,  319,  320,   -1,  322,  323,  324,  325,  326,  327,
  328,   -1,   -1,   -1,  332,  333,  334,  335,  336,   -1,
   -1,   -1,  340,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  348,  349,  350,  351,  352,  353,   -1,   -1,   -1,  357,
  358,   -1,  360,   -1,  362,   -1,  364,  365,  366,   -1,
  368,   -1,  370,  371,  372,   -1,  374,  375,  376,  377,
  378,   -1,  380,  381,   -1,   -1,   -1,   -1,  386,   -1,
  388,  389,   -1,   -1,  392,  393,  394,   -1,   -1,  397,
   -1,  399,   -1,   -1,  402,  403,  404,  405,   -1,  407,
   -1,  409,   -1,  411,  412,  413,  414,  415,   -1,   -1,
  418,  419,  420,   -1,  422,  423,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  433,  434,  435,   -1,  437,
  438,  439,   -1,   -1,   -1,   -1,  602,  445,   -1,   -1,
   -1,  449,   -1,  451,  452,  453,  454,   -1,  456,   -1,
  458,  459,  460,  461,  462,  463,  464,  465,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  475,  476,   -1,
  478,  479,  480,   -1,   -1,   -1,   -1,  485,  486,  487,
  488,  489,   -1,  491,   -1,   -1,  494,  495,   -1,   -1,
   -1,  499,  500,   -1,   -1,  503,  504,  505,  506,  507,
  508,  509,  510,  511,   -1,  513,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  523,  524,  525,  526,  527,
  528,  529,  530,  531,  532,  533,  534,  535,  536,  537,
  538,  539,  540,  541,  542,  543,   -1,  545,  546,  547,
  548,  549,  550,   -1,  552,  553,  554,  555,   -1,  557,
   -1,  559,  560,  561,  562,  563,  564,  565,  566,  567,
  568,  569,  570,  571,  572,  573,  574,  575,  576,   -1,
  578,  579,  580,  581,   -1,  583,  584,  585,  586,  587,
   -1,  589,  590,  591,  592,  593,  594,  595,  596,  597,
  598,  599,  600,  601,  602,  603,  604,  605,  606,  607,
  608,  609,  610,  611,  612,  613,  614,  615,  616,  617,
   -1,  619,  620,   60,   61,   62,  624,  625,  626,  627,
  628,  629,   -1,   -1,  632,  633,  634,  635,  636,  637,
  638,  639,  640,   -1,   -1,   -1,   -1,   -1,  646,   -1,
  648,  649,  650,  651,  652,  653,   -1,  655,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  672,  673,   -1,  675,  676,  677,
  678,  679,  680,  681,  682,  683,   -1,  685,  686,  687,
  688,  689,  690,  691,  692,  693,  694,  695,  696,   -1,
  698,  699,  700,   -1,  702,   -1,   -1,   -1,  706,  707,
  708,  709,  710,  711,  712,  713,  714,  715,  716,  717,
  718,    0,  720,  721,   -1,   -1,   -1,   -1,   -1,  727,
  728,  729,  730,  731,  732,  733,  734,  735,  736,  737,
  738,  739,  740,  741,  742,  743,  744,   -1,   -1,  747,
   -1,  749,  750,  751,  752,  753,  754,   -1,   -1,   -1,
   -1,   -1,   41,   42,   43,   44,   45,   46,   47,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   59,   60,   61,   62,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   91,   -1,   93,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  262,   -1,  264,   -1,   -1,
   -1,   -1,   -1,   -1,  271,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  284,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  318,   -1,   -1,   -1,   -1,   -1,   -1,  325,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  333,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  350,   -1,  352,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  362,   -1,  364,  365,  366,
   -1,   -1,   -1,   -1,  371,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  412,   -1,   -1,  257,   -1,
  259,   -1,   -1,  262,   -1,  264,  265,  266,   -1,  268,
  269,   -1,  271,   -1,   -1,   -1,   -1,   -1,  277,   -1,
  279,  280,  281,  282,   -1,  284,   -1,   -1,  287,   -1,
   -1,  290,   -1,   -1,   -1,   -1,   -1,   -1,  297,   -1,
  299,  300,   -1,  302,  303,  304,  305,  306,   -1,   -1,
  309,   -1,   -1,  312,   -1,  314,  315,   -1,   -1,  318,
  319,  320,   -1,  322,  323,  324,  325,  326,  327,  328,
   -1,   -1,   -1,  332,  333,  334,  335,  336,   -1,   -1,
   -1,  340,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  348,
  349,  350,  351,  352,  353,   -1,   -1,   -1,  357,  358,
   -1,  360,   -1,  362,   -1,  364,  365,  366,   -1,  368,
   -1,  370,  371,  372,   -1,  374,  375,  376,  377,  378,
   -1,  380,  381,   -1,   -1,   -1,   -1,  386,   -1,  388,
  389,   -1,   -1,  392,  393,  394,   -1,   -1,  397,   -1,
  399,   -1,   -1,  402,  403,  404,  405,   -1,  407,   -1,
  409,   -1,  411,  412,  413,  414,  415,   -1,   -1,  418,
  419,  420,   -1,  422,  423,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  433,  434,  435,   -1,  437,  438,
  439,   -1,   -1,   -1,   -1,  602,  445,   -1,   -1,   -1,
  449,   -1,  451,  452,  453,  454,   -1,  456,   -1,  458,
  459,  460,  461,  462,  463,  464,  465,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  475,  476,   -1,  478,
  479,  480,   -1,   -1,   -1,   -1,  485,  486,  487,  488,
  489,   -1,  491,   -1,   -1,  494,  495,   -1,   -1,   -1,
  499,  500,   -1,   -1,  503,  504,  505,  506,  507,  508,
  509,  510,  511,   -1,  513,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  523,  524,  525,  526,  527,  528,
  529,  530,  531,  532,  533,  534,  535,  536,  537,  538,
  539,  540,  541,  542,  543,   -1,  545,  546,  547,  548,
  549,  550,   -1,  552,  553,  554,  555,   -1,  557,   -1,
  559,  560,  561,  562,  563,  564,  565,  566,  567,  568,
  569,  570,  571,  572,  573,  574,  575,  576,   -1,  578,
  579,  580,  581,   -1,  583,  584,  585,  586,  587,   -1,
  589,  590,  591,  592,  593,  594,  595,  596,  597,  598,
  599,  600,  601,  602,  603,  604,  605,  606,  607,  608,
  609,  610,  611,  612,  613,  614,  615,  616,  617,   -1,
  619,  620,   60,   61,   62,  624,  625,  626,  627,  628,
  629,   -1,   -1,  632,  633,  634,  635,  636,  637,  638,
  639,  640,   -1,   -1,   -1,   -1,   -1,  646,   -1,  648,
  649,  650,  651,  652,  653,   -1,  655,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  672,  673,   -1,  675,  676,  677,  678,
  679,  680,  681,  682,  683,   -1,  685,  686,  687,  688,
  689,  690,  691,  692,  693,  694,  695,  696,   -1,  698,
  699,  700,   -1,  702,   -1,   -1,   -1,  706,  707,  708,
  709,  710,  711,  712,  713,  714,  715,  716,  717,  718,
    0,  720,  721,   -1,   -1,   -1,   -1,   -1,  727,  728,
  729,  730,  731,  732,  733,  734,  735,  736,  737,  738,
  739,  740,  741,  742,  743,  744,   -1,   -1,  747,   -1,
  749,  750,  751,  752,  753,  754,   -1,   -1,   -1,   -1,
   -1,   41,   42,   43,   44,   45,   46,   47,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   59,
   60,   61,   62,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   91,   -1,   93,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  262,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  271,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  284,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  314,   -1,   -1,   -1,
  318,   -1,   -1,   -1,   -1,   -1,   -1,  325,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  333,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  350,   -1,  352,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  362,   -1,  364,  365,  366,   -1,
   -1,   -1,   -1,  371,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  412,   -1,   -1,  257,   -1,  259,
   -1,   -1,  262,   -1,  264,  265,  266,   -1,  268,  269,
   -1,  271,   -1,   -1,   -1,   -1,   -1,  277,   -1,  279,
  280,  281,  282,   -1,  284,   -1,   -1,  287,   -1,   -1,
  290,   -1,   -1,   -1,   -1,   -1,   -1,  297,   -1,  299,
  300,   -1,  302,  303,  304,  305,  306,   -1,   -1,  309,
   -1,   -1,  312,   -1,  314,  315,   -1,   -1,  318,  319,
  320,   -1,  322,  323,  324,  325,  326,  327,  328,   -1,
   -1,   -1,  332,  333,  334,  335,  336,   -1,   -1,   -1,
  340,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  348,  349,
  350,  351,  352,  353,   -1,   -1,   -1,  357,  358,   -1,
  360,   -1,  362,   -1,  364,  365,  366,   -1,  368,   -1,
  370,  371,  372,   -1,  374,  375,  376,  377,  378,   -1,
  380,  381,   -1,   -1,   -1,   -1,  386,   -1,  388,  389,
   -1,   -1,  392,  393,  394,   -1,   -1,  397,   -1,  399,
   -1,   -1,  402,  403,  404,  405,   -1,  407,   -1,  409,
   -1,  411,  412,  413,  414,  415,   -1,   -1,  418,  419,
  420,   -1,  422,  423,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  433,  434,  435,   -1,  437,  438,  439,
   -1,   -1,   -1,   -1,  602,  445,   -1,   -1,   -1,  449,
   -1,  451,  452,  453,  454,   -1,  456,   -1,  458,  459,
  460,  461,  462,  463,  464,  465,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  475,  476,   -1,  478,  479,
  480,   -1,   -1,   -1,   -1,  485,  486,  487,  488,  489,
   -1,  491,   -1,   -1,  494,  495,   -1,   -1,   -1,  499,
  500,   -1,   -1,  503,  504,  505,  506,  507,  508,  509,
  510,  511,   -1,  513,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  523,  524,  525,  526,  527,  528,  529,
  530,  531,  532,  533,  534,  535,  536,  537,  538,  539,
  540,  541,  542,  543,   -1,  545,  546,  547,  548,  549,
  550,   -1,  552,  553,  554,  555,   -1,  557,   -1,  559,
  560,  561,  562,  563,  564,  565,  566,  567,  568,  569,
  570,  571,  572,  573,  574,  575,  576,   -1,  578,  579,
  580,  581,   -1,  583,  584,  585,  586,  587,   -1,  589,
  590,  591,  592,  593,  594,  595,  596,  597,  598,  599,
  600,  601,  602,  603,  604,  605,  606,  607,  608,  609,
  610,  611,  612,  613,  614,  615,  616,  617,   -1,  619,
  620,   -1,   -1,   -1,  624,  625,  626,  627,  628,  629,
   -1,   -1,  632,  633,  634,  635,  636,  637,  638,  639,
  640,   -1,   -1,   60,   61,   62,  646,   -1,  648,  649,
  650,  651,  652,  653,   -1,  655,   -1,   -1,   -1,   -1,
   -1,   60,   61,   62,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  672,  673,   -1,  675,  676,  677,  678,  679,
  680,  681,  682,  683,   -1,  685,  686,  687,  688,  689,
  690,  691,  692,  693,  694,  695,  696,   -1,  698,  699,
  700,   -1,  702,   -1,   -1,   -1,  706,  707,  708,  709,
  710,  711,  712,  713,  714,  715,  716,  717,  718,    0,
  720,  721,   -1,   -1,   -1,   -1,   -1,  727,  728,  729,
  730,  731,  732,  733,  734,  735,  736,  737,  738,  739,
  740,  741,  742,  743,  744,   -1,   -1,  747,   -1,  749,
  750,  751,  752,  753,  754,   -1,   -1,   -1,   -1,   -1,
   41,   42,   43,   44,   45,   46,   47,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   59,   60,
   61,   62,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   91,   -1,   93,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  262,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  271,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  262,   -1,   -1,   -1,  284,   -1,   -1,
   -1,   -1,  271,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  284,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  314,   -1,   -1,
   -1,  318,   -1,   -1,   -1,   -1,   -1,   -1,  325,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  333,   -1,   -1,  318,
   -1,   -1,   -1,   -1,   -1,   -1,  325,   -1,   -1,   -1,
   -1,   -1,   -1,  350,  333,  352,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  362,   -1,  364,  365,  366,
   -1,  350,   -1,  352,  371,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  362,   -1,  364,  365,  366,   -1,   -1,
   -1,   -1,  371,   -1,   -1,   -1,  257,   -1,  259,   -1,
   -1,  262,   -1,  264,  265,  266,   -1,  268,  269,   -1,
  271,   -1,   -1,   -1,   -1,  412,  277,   -1,  279,  280,
  281,  282,   -1,  284,   -1,   -1,  287,   -1,   -1,  290,
   -1,   -1,   -1,  412,   -1,   -1,  297,   -1,  299,  300,
   -1,  302,  303,  304,  305,  306,   -1,   -1,  309,   -1,
   -1,  312,   -1,  314,  315,   -1,   -1,  318,  319,  320,
   -1,  322,  323,  324,  325,  326,  327,  328,   -1,   -1,
   -1,  332,  333,  334,  335,  336,   -1,   -1,   -1,  340,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  348,  349,  350,
  351,  352,  353,   -1,   -1,   -1,  357,  358,   -1,  360,
   -1,  362,   -1,  364,  365,  366,   -1,  368,   -1,  370,
  371,  372,   -1,  374,  375,  376,  377,  378,   -1,  380,
  381,   -1,   -1,   -1,   -1,  386,   -1,  388,  389,   -1,
   -1,  392,  393,  394,   -1,   -1,  397,   -1,  399,   -1,
   -1,  402,  403,  404,  405,   -1,  407,   -1,  409,   -1,
  411,  412,  413,  414,  415,   -1,   -1,  418,  419,  420,
   -1,  422,  423,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  433,  434,  435,   -1,  437,  438,  439,   -1,
   -1,   -1,   -1,   -1,  445,   -1,   -1,   -1,  449,   -1,
  451,  452,  453,  454,  573,  456,   -1,  458,  459,  460,
  461,  462,  463,  464,  465,  602,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  475,  476,   -1,  478,  479,  480,
   -1,   -1,   -1,  602,  485,  486,  487,  488,  489,   -1,
  491,   -1,   -1,  494,  495,   -1,   -1,   -1,  499,  500,
   -1,   -1,  503,  504,  505,  506,  507,  508,  509,  510,
  511,   -1,  513,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  523,  524,  525,  526,  527,  528,  529,  530,
  531,  532,  533,  534,  535,  536,  537,  538,  539,  540,
  541,  542,  543,   -1,  545,  546,  547,  548,  549,  550,
   -1,  552,  553,  554,  555,   -1,  557,   -1,  559,  560,
  561,  562,  563,  564,  565,  566,  567,  568,  569,  570,
  571,  572,  573,  574,  575,  576,   -1,  578,  579,  580,
  581,   -1,  583,  584,  585,  586,  587,   -1,  589,  590,
  591,  592,  593,  594,  595,  596,  597,  598,  599,  600,
  601,  602,  603,  604,  605,  606,  607,  608,  609,  610,
  611,  612,  613,  614,  615,  616,  617,   -1,  619,  620,
   60,   61,   62,  624,  625,  626,  627,  628,  629,   -1,
   -1,  632,  633,  634,  635,  636,  637,  638,  639,  640,
   -1,   -1,   -1,   -1,   -1,  646,   -1,  648,  649,  650,
  651,  652,  653,   -1,  655,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  672,  673,   -1,  675,  676,  677,  678,  679,  680,
  681,  682,  683,   -1,  685,  686,  687,  688,  689,  690,
  691,  692,  693,  694,  695,  696,   -1,  698,  699,  700,
   -1,  702,   -1,   -1,   -1,  706,  707,  708,  709,  710,
  711,  712,  713,  714,  715,  716,  717,  718,    0,  720,
  721,   -1,   -1,   -1,   -1,   -1,  727,  728,  729,  730,
  731,  732,  733,  734,  735,  736,  737,  738,  739,  740,
  741,  742,  743,  744,   -1,   -1,  747,   -1,  749,  750,
  751,  752,  753,  754,   -1,   -1,   -1,   -1,   -1,   41,
   42,   43,   44,   45,   46,   47,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   59,   60,   61,
   62,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   91,
   -1,   93,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  262,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  271,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  284,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  318,   -1,
   -1,   -1,   -1,   -1,   -1,  325,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  333,   -1,   -1,  336,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  350,   -1,  352,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  362,   -1,  364,  365,  366,   -1,   -1,   -1,
   -1,  371,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  412,   -1,   -1,  257,   -1,  259,   -1,   -1,
  262,   -1,  264,  265,  266,   -1,  268,  269,   -1,  271,
   -1,   -1,   -1,   -1,   -1,  277,   -1,  279,  280,  281,
  282,   -1,  284,   -1,   -1,  287,   -1,   -1,  290,   -1,
   -1,   -1,   -1,   -1,   -1,  297,   -1,  299,  300,   -1,
  302,  303,  304,  305,  306,   -1,   -1,  309,   -1,   -1,
  312,   -1,  314,  315,   -1,   -1,  318,  319,  320,   -1,
  322,  323,  324,  325,  326,  327,  328,   -1,   -1,   -1,
  332,  333,  334,  335,  336,   -1,   -1,   -1,  340,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  348,  349,  350,  351,
  352,  353,   -1,   -1,   -1,  357,  358,   -1,  360,   -1,
  362,   -1,  364,  365,  366,   -1,  368,   -1,  370,  371,
  372,   -1,  374,  375,  376,  377,  378,   -1,  380,  381,
   -1,   -1,   -1,   -1,  386,   -1,  388,  389,   -1,   -1,
  392,  393,  394,   -1,   -1,  397,   -1,  399,   -1,   -1,
  402,  403,  404,  405,   -1,  407,   -1,  409,   -1,  411,
  412,  413,  414,  415,   -1,   -1,  418,  419,  420,   -1,
  422,  423,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  433,  434,  435,   -1,  437,  438,  439,   -1,   -1,
   -1,   -1,  602,  445,   -1,   -1,   -1,  449,   -1,  451,
  452,  453,  454,   -1,  456,   -1,  458,  459,  460,  461,
  462,  463,  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  475,  476,   -1,  478,  479,  480,   -1,
   -1,   -1,   -1,  485,  486,  487,  488,  489,   -1,  491,
   -1,   -1,  494,  495,   -1,   -1,   -1,  499,  500,   -1,
   -1,  503,  504,  505,  506,  507,  508,  509,  510,  511,
   -1,  513,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  523,  524,  525,  526,  527,  528,  529,  530,  531,
  532,  533,  534,  535,  536,  537,  538,  539,  540,  541,
  542,  543,   -1,  545,  546,  547,  548,  549,  550,   -1,
  552,  553,  554,  555,   -1,  557,   -1,  559,  560,  561,
  562,  563,  564,  565,  566,  567,  568,  569,  570,  571,
  572,  573,  574,  575,  576,   -1,  578,  579,  580,  581,
   -1,  583,  584,  585,  586,  587,   -1,  589,  590,  591,
  592,  593,  594,  595,  596,  597,  598,  599,  600,  601,
  602,  603,  604,  605,  606,  607,  608,  609,  610,  611,
  612,  613,  614,  615,  616,  617,   -1,  619,  620,   60,
   61,   62,  624,  625,  626,  627,  628,  629,   -1,   -1,
  632,  633,  634,  635,  636,  637,  638,  639,  640,   -1,
   -1,   -1,   -1,   -1,  646,   -1,  648,  649,  650,  651,
  652,  653,   -1,  655,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  672,  673,   -1,  675,  676,  677,  678,  679,  680,  681,
  682,  683,   -1,  685,  686,  687,  688,  689,  690,  691,
  692,  693,  694,  695,  696,   -1,  698,  699,  700,   -1,
  702,   -1,   -1,   -1,  706,  707,  708,  709,  710,  711,
  712,  713,  714,  715,  716,  717,  718,    0,  720,  721,
   -1,   -1,   -1,   -1,   -1,  727,  728,  729,  730,  731,
  732,  733,  734,  735,  736,  737,  738,  739,  740,  741,
  742,  743,  744,   -1,   -1,  747,   -1,  749,  750,  751,
  752,  753,  754,   -1,   -1,   -1,   -1,   -1,   41,   42,
   43,   44,   45,   46,   47,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   59,   60,   61,   62,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   91,   -1,
   93,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  262,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  271,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  284,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  318,   -1,   -1,
   -1,   -1,   -1,   -1,  325,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  333,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  350,
   -1,  352,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  362,   -1,  364,  365,  366,   -1,   -1,   -1,   -1,
  371,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  412,   -1,   -1,  257,   -1,  259,   -1,   -1,  262,
   -1,  264,  265,  266,   -1,  268,  269,   -1,  271,   -1,
   -1,   -1,   -1,   -1,  277,   -1,  279,  280,  281,  282,
   -1,  284,   -1,   -1,  287,   -1,   -1,  290,   -1,   -1,
   -1,   -1,   -1,   -1,  297,   -1,  299,  300,   -1,  302,
  303,  304,  305,  306,   -1,   -1,  309,   -1,   -1,  312,
   -1,  314,  315,   -1,   -1,  318,  319,  320,   -1,  322,
  323,  324,  325,  326,  327,  328,  487,   -1,   -1,  332,
  333,  334,  335,  336,   -1,   -1,   -1,  340,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  348,  349,  350,  351,  352,
  353,   -1,   -1,   -1,  357,  358,   -1,  360,   -1,  362,
   -1,  364,  365,  366,   -1,  368,   -1,  370,  371,  372,
   -1,  374,  375,  376,  377,  378,   -1,  380,  381,   -1,
   -1,   -1,   -1,  386,   -1,  388,  389,   -1,   -1,  392,
  393,  394,   -1,   -1,  397,   -1,  399,   -1,   -1,  402,
  403,  404,  405,   -1,  407,   -1,  409,   -1,  411,  412,
  413,  414,  415,   -1,   -1,  418,  419,  420,   -1,  422,
  423,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  433,  434,  435,   -1,  437,  438,  439,   -1,   -1,   -1,
   -1,  602,  445,   -1,   -1,   -1,  449,   -1,  451,  452,
  453,  454,   -1,  456,   -1,  458,  459,  460,  461,  462,
  463,  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  475,  476,   -1,  478,  479,  480,   -1,   -1,
   -1,   -1,  485,  486,  487,  488,  489,   -1,  491,   -1,
   -1,  494,  495,   -1,   -1,   -1,  499,  500,   -1,   -1,
  503,  504,  505,  506,  507,  508,  509,  510,  511,   -1,
  513,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  523,  524,  525,  526,  527,  528,  529,  530,  531,  532,
  533,  534,  535,  536,  537,  538,  539,  540,  541,  542,
  543,   -1,  545,  546,  547,  548,  549,  550,   -1,  552,
  553,  554,  555,   -1,  557,   -1,  559,  560,  561,  562,
  563,  564,  565,  566,  567,  568,  569,  570,  571,  572,
  573,  574,  575,  576,   -1,  578,  579,  580,  581,   -1,
  583,  584,  585,  586,  587,   -1,  589,  590,  591,  592,
  593,  594,  595,  596,  597,  598,  599,  600,  601,  602,
  603,  604,  605,  606,  607,  608,  609,  610,  611,  612,
  613,  614,  615,  616,  617,   -1,  619,  620,   60,   61,
   62,  624,  625,  626,  627,  628,  629,   -1,   -1,  632,
  633,  634,  635,  636,  637,  638,  639,  640,   -1,   -1,
   -1,   -1,   -1,  646,   -1,  648,  649,  650,  651,  652,
  653,   -1,  655,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  672,
  673,   -1,  675,  676,  677,  678,  679,  680,  681,  682,
  683,   -1,  685,  686,  687,  688,  689,  690,  691,  692,
  693,  694,  695,  696,   -1,  698,  699,  700,   -1,  702,
   -1,   -1,   -1,  706,  707,  708,  709,  710,  711,  712,
  713,  714,  715,  716,  717,  718,    0,  720,  721,   -1,
   -1,   -1,   -1,   -1,  727,  728,  729,  730,  731,  732,
  733,  734,  735,  736,  737,  738,  739,  740,  741,  742,
  743,  744,   -1,   -1,  747,   -1,  749,  750,  751,  752,
  753,  754,   -1,   -1,   -1,   -1,   -1,   41,   42,   43,
   44,   45,   -1,   47,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   59,   60,   61,   62,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   93,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  262,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  271,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  284,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  318,   -1,   -1,   -1,
   -1,   -1,   -1,  325,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  333,   -1,   -1,  336,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  350,   -1,
  352,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  362,   -1,  364,  365,  366,   -1,   -1,   -1,   -1,  371,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  412,   -1,   -1,  257,   -1,  259,   -1,   -1,  262,   -1,
  264,  265,  266,   -1,  268,  269,   -1,  271,   -1,   -1,
   -1,   -1,   -1,  277,   -1,  279,  280,  281,  282,   -1,
  284,   -1,   -1,  287,   -1,   -1,  290,   -1,   -1,   -1,
   -1,   -1,   -1,  297,   -1,  299,  300,   -1,  302,  303,
  304,  305,  306,   -1,   -1,  309,   -1,   -1,  312,   -1,
  314,  315,   -1,   -1,  318,  319,  320,   -1,  322,  323,
  324,  325,  326,  327,  328,   -1,   -1,   -1,  332,  333,
  334,  335,  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  348,  349,  350,  351,  352,  353,
   -1,   -1,   -1,  357,  358,   -1,  360,   -1,  362,   -1,
  364,  365,  366,   -1,  368,   -1,  370,  371,  372,   -1,
  374,  375,  376,  377,  378,   -1,  380,  381,   -1,   -1,
   -1,   -1,  386,   -1,  388,  389,   -1,   -1,  392,  393,
  394,   -1,   -1,  397,   -1,  399,   -1,   -1,  402,  403,
  404,  405,   -1,  407,   -1,  409,   -1,  411,  412,  413,
  414,  415,   -1,   -1,  418,  419,  420,   -1,  422,  423,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  433,
  434,  435,   -1,  437,  438,  439,   -1,   -1,   -1,   -1,
  602,  445,   -1,   -1,   -1,  449,   -1,  451,  452,  453,
  454,   -1,  456,   -1,  458,  459,  460,  461,  462,  463,
  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  475,  476,   -1,  478,  479,  480,   -1,   -1,   -1,
   -1,  485,  486,  487,  488,  489,   -1,  491,   -1,   -1,
  494,  495,   -1,   -1,   -1,  499,  500,   -1,   -1,  503,
  504,  505,  506,  507,  508,  509,  510,  511,   -1,  513,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  523,
  524,  525,  526,  527,  528,  529,  530,  531,  532,  533,
  534,  535,  536,  537,  538,  539,  540,  541,  542,  543,
   -1,  545,  546,  547,  548,  549,  550,   -1,  552,  553,
  554,  555,   -1,  557,   -1,  559,  560,  561,  562,  563,
  564,  565,  566,  567,  568,  569,  570,  571,  572,  573,
  574,  575,  576,   -1,  578,  579,  580,  581,   -1,  583,
  584,  585,  586,  587,   -1,  589,  590,  591,  592,  593,
  594,  595,  596,  597,  598,  599,  600,  601,  602,  603,
  604,  605,  606,  607,  608,   -1,  610,  611,  612,  613,
  614,  615,  616,  617,  618,  619,  620,   60,   61,   62,
  624,  625,  626,  627,  628,  629,   -1,   -1,  632,  633,
  634,  635,  636,  637,  638,  639,  640,   -1,   -1,   -1,
   -1,   -1,  646,   -1,  648,  649,  650,  651,  652,  653,
   -1,  655,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  672,  673,
   -1,  675,  676,  677,  678,  679,  680,  681,  682,  683,
   -1,  685,  686,  687,  688,  689,  690,  691,  692,  693,
  694,  695,  696,   -1,  698,  699,  700,   -1,  702,   -1,
   -1,   -1,  706,  707,  708,  709,  710,  711,  712,  713,
  714,  715,  716,  717,  718,    0,  720,  721,   -1,   -1,
   -1,   -1,   -1,  727,  728,  729,  730,  731,  732,  733,
  734,  735,  736,  737,  738,  739,  740,  741,  742,  743,
  744,   -1,   -1,  747,   -1,  749,  750,  751,  752,  753,
  754,   -1,   -1,   -1,   -1,   -1,   41,   42,   43,   44,
   45,   -1,   47,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   59,   60,   61,   62,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   93,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  262,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  271,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  284,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  318,   -1,   -1,   -1,   -1,
   -1,   -1,  325,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  333,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  350,   -1,  352,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  362,
   -1,  364,  365,  366,   -1,   -1,   -1,   -1,  371,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  412,
   -1,   -1,  257,   -1,  259,   -1,   -1,  262,   -1,  264,
  265,  266,   -1,  268,  269,   -1,  271,   -1,   -1,   -1,
   -1,   -1,  277,   -1,  279,  280,  281,  282,   -1,  284,
   -1,   -1,  287,   -1,   -1,  290,   -1,   -1,   -1,   -1,
   -1,   -1,  297,   -1,  299,  300,   -1,  302,  303,  304,
  305,  306,   -1,   -1,  309,   -1,   -1,  312,   -1,  314,
  315,   -1,   -1,  318,  319,  320,   -1,  322,  323,  324,
  325,  326,  327,  328,  487,   -1,   -1,  332,  333,  334,
  335,  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  348,  349,  350,  351,  352,  353,   -1,
   -1,   -1,  357,  358,   -1,  360,   -1,  362,   -1,  364,
  365,  366,   -1,  368,   -1,  370,  371,  372,   -1,  374,
  375,  376,  377,  378,   -1,  380,  381,   -1,   -1,   -1,
   -1,  386,   -1,  388,  389,   -1,   -1,  392,  393,  394,
   -1,   -1,  397,   -1,  399,   -1,   -1,  402,  403,  404,
  405,   -1,  407,   -1,  409,   -1,  411,  412,  413,  414,
  415,   -1,   -1,  418,  419,  420,   -1,  422,  423,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  433,  434,
  435,   -1,  437,  438,  439,   -1,   -1,   -1,   -1,  602,
  445,   -1,   -1,   -1,  449,   -1,  451,  452,  453,  454,
   -1,  456,   -1,  458,  459,  460,  461,  462,  463,  464,
  465,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  475,  476,   -1,  478,  479,  480,   -1,   -1,   -1,   -1,
  485,  486,  487,  488,  489,   -1,  491,   -1,   -1,  494,
  495,   -1,   -1,   -1,  499,  500,   -1,   -1,  503,  504,
  505,  506,  507,  508,  509,  510,  511,   -1,  513,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  523,  524,
  525,  526,  527,  528,  529,  530,  531,  532,  533,  534,
  535,  536,  537,  538,  539,  540,  541,  542,  543,   -1,
  545,  546,  547,  548,  549,  550,   -1,  552,  553,  554,
  555,   -1,  557,   -1,  559,  560,  561,  562,  563,  564,
  565,  566,  567,  568,  569,  570,  571,  572,  573,  574,
  575,  576,   -1,  578,  579,  580,  581,   -1,  583,  584,
  585,  586,  587,   -1,  589,  590,  591,  592,  593,  594,
  595,  596,  597,  598,  599,  600,  601,  602,  603,  604,
  605,  606,  607,  608,   -1,  610,  611,  612,  613,  614,
  615,  616,  617,  618,  619,  620,   60,   61,   62,  624,
  625,  626,  627,  628,  629,   -1,   -1,  632,  633,  634,
  635,  636,  637,  638,  639,  640,   -1,   -1,   -1,   -1,
   -1,  646,   -1,  648,  649,  650,  651,  652,  653,   -1,
  655,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  672,  673,   -1,
  675,  676,  677,  678,  679,  680,  681,  682,  683,   -1,
  685,  686,  687,  688,  689,  690,  691,  692,  693,  694,
  695,  696,   -1,  698,  699,  700,   -1,  702,   -1,   -1,
   -1,  706,  707,  708,  709,  710,  711,  712,  713,  714,
  715,  716,  717,  718,    0,  720,  721,   -1,   -1,   -1,
   -1,   -1,  727,  728,  729,  730,  731,  732,  733,  734,
  735,  736,  737,  738,  739,  740,  741,  742,  743,  744,
   -1,   -1,  747,   -1,  749,  750,  751,  752,  753,  754,
   -1,   -1,   -1,   -1,   -1,   41,   42,   43,   44,   45,
   -1,   47,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   59,   60,   61,   62,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   91,   -1,   93,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  262,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  271,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  284,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  312,   -1,
   -1,   -1,   -1,   -1,  318,   -1,   -1,   -1,   -1,   -1,
   -1,  325,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  333,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  350,   -1,  352,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  362,   -1,
  364,  365,  366,   -1,   -1,   -1,   -1,  371,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  412,   -1,
   -1,  257,   -1,  259,   -1,   -1,  262,   -1,  264,  265,
  266,   -1,  268,  269,   -1,  271,   -1,   -1,   -1,   -1,
   -1,  277,   -1,  279,  280,  281,  282,   -1,  284,   -1,
   -1,  287,   -1,   -1,  290,   -1,   -1,   -1,   -1,   -1,
   -1,  297,   -1,  299,  300,   -1,  302,  303,  304,  305,
  306,   -1,   -1,  309,   -1,   -1,  312,   -1,  314,  315,
   -1,   -1,  318,  319,  320,   -1,  322,  323,  324,  325,
  326,  327,  328,   -1,   -1,   -1,  332,  333,  334,  335,
  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  348,  349,  350,  351,  352,  353,   -1,   -1,
   -1,  357,  358,   -1,  360,   -1,  362,   -1,  364,  365,
  366,   -1,  368,   -1,  370,  371,  372,   -1,  374,  375,
  376,  377,  378,   -1,  380,  381,   -1,   -1,   -1,   -1,
  386,   -1,  388,  389,   -1,   -1,  392,  393,  394,   -1,
   -1,  397,   -1,  399,   -1,   -1,  402,  403,  404,  405,
   -1,  407,   -1,  409,   -1,  411,  412,  413,  414,  415,
   -1,   -1,  418,  419,  420,   -1,  422,  423,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  433,  434,  435,
   -1,  437,  438,  439,   -1,   -1,   -1,   -1,  602,  445,
   -1,   -1,   -1,  449,   -1,  451,  452,  453,  454,   -1,
  456,   -1,  458,  459,  460,  461,  462,  463,  464,  465,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  475,
  476,   -1,  478,  479,  480,   -1,   -1,   -1,   -1,  485,
  486,  487,  488,  489,   -1,  491,   -1,   -1,  494,  495,
   -1,   -1,   -1,  499,  500,   -1,   -1,  503,  504,  505,
  506,  507,  508,  509,  510,  511,   -1,  513,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  523,  524,  525,
  526,  527,  528,  529,  530,  531,  532,  533,  534,  535,
  536,  537,  538,  539,  540,  541,  542,  543,   -1,  545,
  546,  547,  548,  549,  550,   -1,  552,  553,  554,  555,
   -1,  557,   -1,  559,  560,  561,  562,  563,  564,  565,
  566,  567,  568,  569,  570,  571,  572,  573,  574,  575,
  576,   -1,  578,  579,  580,  581,   -1,  583,  584,  585,
  586,  587,   -1,  589,  590,  591,  592,  593,  594,  595,
  596,  597,  598,  599,  600,  601,  602,  603,  604,  605,
  606,  607,  608,   -1,  610,  611,  612,  613,  614,  615,
  616,  617,   -1,  619,  620,   60,   61,   62,  624,  625,
  626,  627,  628,  629,   -1,   -1,  632,  633,  634,  635,
  636,  637,  638,  639,  640,   -1,   -1,   -1,   -1,   -1,
  646,   -1,  648,  649,  650,  651,  652,  653,   -1,  655,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  672,  673,   -1,  675,
  676,  677,  678,  679,  680,  681,  682,  683,   -1,  685,
  686,  687,  688,  689,  690,  691,  692,  693,  694,  695,
  696,   -1,  698,  699,  700,   -1,  702,   -1,   -1,   -1,
  706,  707,  708,  709,  710,  711,  712,  713,  714,  715,
  716,  717,  718,    0,  720,  721,   -1,   -1,   -1,   -1,
   -1,  727,  728,  729,  730,  731,  732,  733,  734,  735,
  736,  737,  738,  739,  740,  741,  742,  743,  744,   -1,
   -1,  747,   -1,  749,  750,  751,  752,  753,  754,   -1,
   -1,   -1,   -1,   -1,   41,   42,   43,   44,   45,   46,
   47,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   59,   60,   61,   62,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   93,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  262,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  271,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  284,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  305,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  318,   -1,   -1,   -1,   -1,   -1,   -1,
  325,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  333,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  350,   -1,  352,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  362,   -1,  364,
  365,  366,   -1,   -1,   -1,   -1,  371,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  412,   -1,   -1,
  257,   -1,  259,   -1,   -1,  262,   -1,  264,  265,  266,
   -1,  268,  269,   -1,  271,   -1,   -1,   -1,   -1,   -1,
  277,   -1,  279,  280,  281,  282,   -1,  284,   -1,   -1,
  287,   -1,   -1,  290,   -1,   -1,   -1,   -1,   -1,   -1,
  297,   -1,  299,  300,   -1,  302,  303,  304,  305,  306,
   -1,   -1,  309,   -1,   -1,  312,   -1,  314,  315,   -1,
   -1,  318,  319,  320,   -1,  322,  323,  324,  325,  326,
  327,  328,   -1,   -1,   -1,  332,  333,  334,  335,  336,
   -1,   -1,   -1,  340,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  348,  349,  350,  351,  352,  353,   -1,   -1,   -1,
  357,  358,   -1,  360,   -1,  362,   -1,  364,  365,  366,
   -1,  368,   -1,  370,  371,  372,   -1,  374,  375,  376,
  377,  378,   -1,  380,  381,   -1,   -1,   -1,   -1,  386,
   -1,  388,  389,   -1,   -1,  392,  393,  394,   -1,   -1,
  397,   -1,  399,   -1,   -1,  402,  403,  404,  405,   -1,
  407,   -1,  409,   -1,  411,  412,  413,  414,  415,   -1,
   -1,  418,  419,  420,   -1,  422,  423,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  433,  434,  435,   -1,
  437,  438,  439,   -1,   -1,   -1,   -1,  602,  445,   -1,
   -1,   -1,  449,   -1,  451,  452,  453,  454,   -1,  456,
   -1,  458,  459,  460,  461,  462,  463,  464,  465,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  475,  476,
   -1,  478,  479,  480,   -1,   -1,   -1,   -1,  485,  486,
  487,  488,  489,   -1,  491,   -1,   -1,  494,  495,   -1,
   -1,   -1,  499,  500,   -1,   -1,  503,  504,  505,  506,
  507,  508,  509,  510,  511,   -1,  513,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  523,  524,  525,  526,
  527,  528,  529,  530,  531,  532,  533,  534,  535,  536,
  537,  538,  539,  540,  541,  542,  543,   -1,  545,  546,
  547,  548,  549,  550,   -1,  552,  553,  554,  555,   -1,
  557,   -1,  559,  560,  561,  562,  563,  564,  565,  566,
  567,  568,  569,  570,  571,  572,  573,  574,  575,  576,
   -1,  578,  579,  580,  581,   -1,  583,  584,  585,  586,
  587,   -1,  589,  590,  591,  592,  593,  594,  595,  596,
  597,  598,  599,  600,  601,  602,  603,  604,  605,  606,
  607,  608,   -1,  610,  611,  612,  613,  614,  615,  616,
  617,   -1,  619,  620,   60,   61,   62,  624,  625,  626,
  627,  628,  629,   -1,   -1,  632,  633,  634,  635,  636,
  637,  638,  639,  640,   -1,   -1,   -1,   -1,   -1,  646,
   -1,  648,  649,  650,  651,  652,  653,   -1,  655,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  672,  673,   -1,  675,  676,
  677,  678,  679,  680,  681,  682,  683,   -1,  685,  686,
  687,  688,  689,  690,  691,  692,  693,  694,  695,  696,
   -1,  698,  699,  700,   -1,  702,   -1,   -1,   -1,  706,
  707,  708,  709,  710,  711,  712,  713,  714,  715,  716,
  717,  718,    0,  720,  721,   -1,   -1,   -1,   -1,   -1,
  727,  728,  729,  730,  731,  732,  733,  734,  735,  736,
  737,  738,  739,  740,  741,  742,  743,  744,   -1,   -1,
  747,   -1,  749,  750,  751,  752,  753,  754,   -1,   -1,
   -1,   -1,   -1,   41,   42,   43,   44,   45,   -1,   47,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   59,   60,   61,   62,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   93,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  262,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  271,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  284,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  318,   -1,   -1,   -1,   -1,   -1,   -1,  325,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  333,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  350,   -1,  352,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  362,   -1,  364,  365,
  366,   -1,   -1,   -1,   -1,  371,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  412,   -1,   -1,  257,
   -1,  259,   -1,  419,  262,   -1,  264,  265,  266,   -1,
  268,  269,   -1,  271,   -1,   -1,   -1,   -1,   -1,  277,
   -1,  279,  280,  281,  282,   -1,  284,   -1,   -1,  287,
   -1,   -1,  290,   -1,   -1,   -1,   -1,   -1,   -1,  297,
   -1,  299,  300,   -1,  302,  303,  304,  305,  306,   -1,
   -1,  309,   -1,   -1,  312,   -1,  314,  315,   -1,   -1,
  318,  319,  320,   -1,  322,  323,  324,  325,  326,  327,
  328,   -1,   -1,   -1,  332,  333,  334,  335,  336,   -1,
   -1,   -1,  340,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  348,  349,  350,  351,  352,  353,   -1,   -1,   -1,  357,
  358,   -1,  360,   -1,  362,   -1,  364,  365,  366,   -1,
  368,   -1,  370,  371,  372,   -1,  374,  375,  376,  377,
  378,   -1,  380,  381,   -1,   -1,   -1,   -1,  386,   -1,
  388,  389,   -1,   -1,  392,  393,  394,   -1,   -1,  397,
   -1,  399,   -1,   -1,  402,  403,  404,  405,   -1,  407,
   -1,  409,   -1,  411,  412,  413,  414,  415,   -1,   -1,
  418,  419,  420,   -1,  422,  423,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  433,  434,  435,   -1,  437,
  438,  439,   -1,   -1,   -1,   -1,  602,  445,   -1,   -1,
   -1,  449,   -1,  451,  452,  453,  454,   -1,  456,   -1,
  458,  459,  460,  461,  462,  463,  464,  465,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  475,  476,   -1,
  478,  479,  480,   -1,   -1,   -1,   -1,  485,  486,  487,
  488,  489,   -1,  491,   -1,   -1,  494,  495,   -1,   -1,
   -1,  499,  500,   -1,   -1,  503,  504,  505,  506,  507,
  508,  509,  510,  511,   -1,  513,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  523,  524,  525,  526,  527,
  528,  529,  530,  531,  532,  533,  534,  535,  536,  537,
  538,  539,  540,  541,  542,  543,   -1,  545,  546,  547,
  548,  549,  550,   -1,  552,  553,  554,  555,   -1,  557,
   -1,  559,  560,  561,  562,  563,  564,  565,  566,  567,
  568,  569,  570,  571,  572,  573,  574,  575,  576,   -1,
  578,  579,  580,  581,   -1,  583,  584,  585,  586,  587,
   -1,  589,  590,  591,  592,  593,  594,  595,  596,  597,
  598,  599,  600,  601,  602,  603,  604,  605,  606,  607,
  608,   -1,  610,  611,  612,  613,  614,  615,  616,  617,
   -1,  619,  620,   60,   61,   62,  624,  625,  626,  627,
  628,  629,   -1,   -1,  632,  633,  634,  635,  636,  637,
  638,  639,  640,   -1,   -1,   -1,   -1,   -1,  646,   -1,
  648,  649,  650,  651,  652,  653,   -1,  655,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  672,  673,   -1,  675,  676,  677,
  678,  679,  680,  681,  682,  683,   -1,  685,  686,  687,
  688,  689,  690,  691,  692,  693,  694,  695,  696,   -1,
  698,  699,  700,   -1,  702,   -1,   -1,   -1,  706,  707,
  708,  709,  710,  711,  712,  713,  714,  715,  716,  717,
  718,    0,  720,  721,   -1,   -1,   -1,   -1,   -1,  727,
  728,  729,  730,  731,  732,  733,  734,  735,  736,  737,
  738,  739,  740,  741,  742,  743,  744,   -1,   -1,  747,
   -1,  749,  750,  751,  752,  753,  754,   -1,   -1,   -1,
   -1,   -1,   41,   42,   43,   44,   45,   -1,   47,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   59,   60,   61,   62,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   93,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  262,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  271,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  284,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  314,   -1,   -1,
   -1,  318,   -1,   -1,   -1,   -1,   -1,   -1,  325,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  333,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  350,   -1,  352,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  362,   -1,  364,  365,  366,
   -1,   -1,   -1,   -1,  371,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  412,   -1,   -1,  257,   -1,
  259,   -1,   -1,  262,   -1,  264,  265,  266,   -1,  268,
  269,   -1,  271,   -1,   -1,   -1,   -1,   -1,  277,   -1,
  279,  280,  281,  282,   -1,  284,   -1,   -1,  287,   -1,
   -1,  290,   -1,   -1,   -1,   -1,   -1,   -1,  297,   -1,
  299,  300,   -1,  302,  303,  304,  305,  306,   -1,   -1,
  309,   -1,   -1,  312,   -1,  314,  315,   -1,   -1,  318,
  319,  320,   -1,  322,  323,  324,  325,  326,  327,  328,
   -1,   -1,   -1,  332,  333,  334,  335,  336,   -1,   -1,
   -1,  340,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  348,
  349,  350,  351,  352,  353,   -1,   -1,   -1,  357,  358,
   -1,  360,   -1,  362,   -1,  364,  365,  366,   -1,  368,
   -1,  370,  371,  372,   -1,  374,  375,  376,  377,  378,
   -1,  380,  381,   -1,   -1,   -1,   -1,  386,   -1,  388,
  389,   -1,   -1,  392,  393,  394,   -1,   -1,  397,   -1,
  399,   -1,   -1,  402,  403,  404,  405,   -1,  407,   -1,
  409,   -1,  411,  412,  413,  414,  415,   -1,   -1,  418,
  419,  420,   -1,  422,  423,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  433,  434,  435,   -1,  437,  438,
  439,   -1,   -1,   -1,   -1,  602,  445,   -1,   -1,   -1,
  449,   -1,  451,  452,  453,  454,   -1,  456,   -1,  458,
  459,  460,  461,  462,  463,  464,  465,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  475,  476,   -1,  478,
  479,  480,   -1,   -1,   -1,   -1,  485,  486,  487,  488,
  489,   -1,  491,   -1,   -1,  494,  495,   -1,   -1,   -1,
  499,  500,   -1,   -1,  503,  504,  505,  506,  507,  508,
  509,  510,  511,   -1,  513,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  523,  524,  525,  526,  527,  528,
  529,  530,  531,  532,  533,  534,  535,  536,  537,  538,
  539,  540,  541,  542,  543,   -1,  545,  546,  547,  548,
  549,  550,   -1,  552,  553,  554,  555,   -1,  557,   -1,
  559,  560,  561,  562,  563,  564,  565,  566,  567,  568,
  569,  570,  571,  572,  573,  574,  575,  576,   -1,  578,
  579,  580,  581,   -1,  583,  584,  585,  586,  587,   -1,
  589,  590,  591,  592,  593,  594,  595,  596,  597,  598,
  599,  600,  601,  602,  603,  604,  605,  606,  607,  608,
   -1,  610,  611,  612,  613,  614,  615,  616,  617,   -1,
  619,  620,   -1,   -1,   -1,  624,  625,  626,  627,  628,
  629,   -1,   -1,  632,  633,  634,  635,  636,  637,  638,
  639,  640,   -1,   -1,   60,   61,   62,  646,   -1,  648,
  649,  650,  651,  652,  653,   -1,  655,   -1,   -1,   -1,
   -1,   -1,   60,   61,   62,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  672,  673,   -1,  675,  676,  677,  678,
  679,  680,  681,  682,  683,   -1,  685,  686,  687,  688,
  689,  690,  691,  692,  693,  694,  695,  696,   -1,  698,
  699,  700,   -1,  702,   -1,   -1,   -1,  706,  707,  708,
  709,  710,  711,  712,  713,  714,  715,  716,  717,  718,
    0,  720,  721,   -1,   -1,   -1,   -1,   -1,  727,  728,
  729,  730,  731,  732,  733,  734,  735,  736,  737,  738,
  739,  740,  741,  742,  743,  744,   -1,   -1,  747,   -1,
  749,  750,  751,  752,  753,  754,   -1,   -1,   -1,   -1,
   -1,   41,   42,   43,   44,   45,   -1,   47,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   59,
   60,   61,   62,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   93,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  262,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  271,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  262,   -1,   -1,   -1,  284,   -1,
   -1,   -1,   -1,  271,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  284,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  318,   -1,   -1,   -1,   -1,   -1,   -1,  325,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  333,   -1,   -1,
  318,   -1,   -1,   -1,   -1,   -1,   -1,  325,   -1,   -1,
   -1,   -1,   -1,   -1,  350,  333,  352,   -1,  336,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  362,   -1,  364,  365,
  366,   -1,  350,   -1,  352,  371,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  362,   -1,  364,  365,  366,   -1,
   -1,   -1,   -1,  371,   -1,   -1,   -1,  257,   -1,  259,
   -1,   -1,  262,   -1,  264,  265,  266,   -1,  268,  269,
   -1,  271,   -1,   -1,   -1,   -1,  412,  277,   -1,  279,
  280,  281,  282,   -1,  284,   -1,   -1,  287,   -1,   -1,
  290,   -1,   -1,   -1,  412,   -1,   -1,  297,   -1,  299,
  300,   -1,  302,  303,  304,  305,  306,   -1,   -1,  309,
   -1,   -1,  312,   -1,  314,  315,   -1,   -1,  318,  319,
  320,   -1,  322,  323,  324,  325,  326,  327,  328,   -1,
   -1,   -1,  332,  333,  334,  335,  336,   -1,   -1,   -1,
  340,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  348,  349,
  350,  351,  352,  353,   -1,   -1,   -1,  357,  358,   -1,
  360,   -1,  362,   -1,  364,  365,  366,   -1,  368,   -1,
  370,  371,  372,   -1,  374,  375,  376,  377,  378,   -1,
  380,  381,   -1,   -1,   -1,   -1,  386,   -1,  388,  389,
   -1,   -1,  392,  393,  394,   -1,   -1,  397,   -1,  399,
   -1,   -1,  402,  403,  404,  405,   -1,  407,   -1,  409,
   -1,  411,  412,  413,  414,  415,   -1,   -1,  418,  419,
  420,  557,  422,  423,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  433,  434,  435,   -1,  437,  438,  439,
   -1,   -1,   -1,   -1,   -1,  445,   -1,   -1,   -1,  449,
   -1,  451,  452,  453,  454,   -1,  456,   -1,  458,  459,
  460,  461,  462,  463,  464,  465,  602,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  475,  476,   -1,  478,  479,
  480,   -1,   -1,   -1,  602,  485,  486,  487,  488,  489,
   -1,  491,   -1,   -1,  494,  495,   -1,   -1,   -1,  499,
  500,   -1,   -1,  503,  504,  505,  506,  507,  508,  509,
  510,  511,   -1,  513,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  523,  524,  525,  526,  527,  528,  529,
  530,  531,  532,  533,  534,  535,  536,  537,  538,  539,
  540,  541,  542,  543,   -1,  545,  546,  547,  548,  549,
  550,   -1,  552,  553,  554,  555,   -1,  557,   -1,  559,
  560,  561,  562,  563,  564,  565,  566,  567,  568,  569,
  570,  571,  572,  573,  574,  575,  576,   -1,  578,  579,
  580,  581,   -1,  583,  584,  585,  586,  587,   -1,  589,
  590,  591,  592,  593,  594,  595,  596,  597,  598,  599,
  600,  601,  602,  603,  604,  605,  606,  607,  608,   -1,
  610,  611,  612,  613,  614,  615,  616,  617,   -1,  619,
  620,   60,   61,   62,  624,  625,  626,  627,  628,  629,
   -1,   -1,  632,  633,  634,  635,  636,  637,  638,  639,
  640,   -1,   -1,   -1,   -1,   -1,  646,   -1,  648,  649,
  650,  651,  652,  653,   -1,  655,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  672,  673,   -1,  675,  676,  677,  678,  679,
  680,  681,  682,  683,   -1,  685,  686,  687,  688,  689,
  690,  691,  692,  693,  694,  695,  696,   -1,  698,  699,
  700,   -1,  702,   -1,   -1,   -1,  706,  707,  708,  709,
  710,  711,  712,  713,  714,  715,  716,  717,  718,    0,
  720,  721,   -1,   -1,   -1,   -1,   -1,  727,  728,  729,
  730,  731,  732,  733,  734,  735,  736,  737,  738,  739,
  740,  741,  742,  743,  744,   -1,   -1,  747,   -1,  749,
  750,  751,  752,  753,  754,   -1,   -1,   -1,   -1,   -1,
   41,   42,   43,   44,   45,   -1,   47,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   59,   60,
   61,   62,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   93,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  262,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  271,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  284,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  318,
   -1,   -1,   -1,   -1,   -1,   -1,  325,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  333,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  350,   -1,  352,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  362,   -1,  364,  365,  366,   -1,   -1,
   -1,   -1,  371,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  412,   -1,   -1,  257,   -1,  259,   -1,
   -1,  262,   -1,  264,  265,   -1,   -1,  268,  269,   -1,
  271,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  279,  280,
  281,  282,   -1,  284,   -1,   -1,  287,   -1,   -1,  290,
   -1,   -1,   -1,   -1,   -1,   -1,  297,   -1,  299,  300,
   -1,  302,  303,  304,  305,  306,   -1,   -1,  309,   -1,
   -1,  312,   -1,  314,  315,   -1,   -1,  318,  319,  320,
   -1,  322,  323,  324,  325,  326,  327,  328,   -1,   -1,
   -1,  332,  333,  334,  335,  336,   -1,   -1,   -1,  340,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  348,  349,  350,
  351,  352,  353,   -1,   -1,   -1,  357,  358,   -1,  360,
   -1,  362,   -1,  364,  365,  366,   -1,  368,   -1,  370,
  371,  372,   -1,  374,  375,  376,  377,  378,   -1,  380,
  381,   -1,   -1,   -1,   -1,  386,   -1,  388,  389,   -1,
   -1,  392,  393,  394,   -1,   -1,  397,   -1,  399,   -1,
   -1,  402,  403,  404,  405,   -1,  407,   -1,  409,   -1,
  411,  412,  413,  414,  415,   -1,   -1,  418,  419,  420,
   -1,  422,  423,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  433,  434,  435,   -1,  437,  438,  439,   -1,
   -1,   -1,   -1,  602,  445,   -1,   -1,   -1,  449,   -1,
  451,  452,  453,  454,   -1,  456,   -1,  458,  459,  460,
  461,  462,  463,  464,  465,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  475,  476,   -1,  478,  479,  480,
   -1,   -1,   -1,   -1,  485,  486,  487,  488,  489,   -1,
  491,   -1,   -1,  494,  495,   -1,   -1,   -1,  499,  500,
   -1,   -1,  503,  504,  505,  506,  507,  508,  509,  510,
  511,   -1,  513,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  523,  524,  525,  526,  527,  528,  529,  530,
  531,  532,  533,  534,  535,  536,  537,  538,  539,  540,
  541,  542,  543,   -1,  545,  546,  547,  548,  549,  550,
   -1,  552,  553,  554,  555,   -1,  557,   -1,  559,  560,
  561,  562,  563,  564,  565,  566,  567,  568,  569,  570,
  571,  572,  573,  574,  575,  576,   -1,  578,  579,  580,
  581,   -1,  583,  584,  585,  586,  587,   -1,  589,  590,
  591,  592,  593,  594,  595,  596,  597,  598,  599,  600,
  601,  602,  603,  604,  605,  606,  607,  608,   -1,  610,
  611,  612,  613,  614,  615,  616,  617,   -1,  619,  620,
   60,   61,   62,  624,  625,  626,  627,  628,  629,   -1,
   -1,  632,  633,  634,  635,  636,  637,  638,  639,  640,
   -1,   -1,   -1,   -1,   -1,  646,   -1,  648,  649,  650,
  651,  652,  653,   -1,  655,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  672,  673,   -1,  675,  676,  677,  678,  679,  680,
  681,  682,  683,   -1,  685,  686,  687,  688,  689,  690,
  691,  692,  693,  694,  695,  696,   -1,  698,  699,  700,
   -1,  702,   -1,   -1,   -1,  706,  707,  708,  709,  710,
  711,  712,  713,  714,  715,  716,  717,  718,    0,  720,
  721,   -1,   -1,   -1,   -1,   -1,  727,  728,  729,  730,
  731,  732,  733,  734,  735,  736,  737,  738,  739,  740,
  741,  742,  743,  744,   -1,   -1,  747,   -1,  749,  750,
  751,  752,  753,  754,   -1,   -1,   -1,   -1,   -1,   41,
   42,   43,   44,   45,   -1,   47,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   59,   60,   61,
   62,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   93,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  262,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  271,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  284,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  318,   -1,
   -1,   -1,   -1,   -1,   -1,  325,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  333,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  350,   -1,  352,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  362,   -1,  364,  365,  366,   -1,   -1,   -1,
   -1,  371,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  412,   -1,   -1,  257,   -1,  259,   -1,   -1,
  262,   -1,  264,  265,   -1,   -1,  268,  269,   -1,  271,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  279,  280,   -1,
  282,   -1,  284,   -1,   -1,  287,   -1,   -1,  290,   -1,
   -1,   -1,   -1,   -1,   -1,  297,   -1,  299,  300,   -1,
  302,  303,  304,  305,  306,   -1,   -1,  309,   -1,   -1,
  312,   -1,  314,  315,   -1,   -1,  318,  319,  320,   -1,
  322,  323,  324,  325,  326,  327,  328,   -1,   -1,   -1,
  332,  333,  334,  335,  336,   -1,   -1,   -1,  340,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  348,  349,  350,  351,
  352,  353,   -1,   -1,   -1,  357,  358,   -1,  360,   -1,
  362,   -1,  364,  365,  366,   -1,  368,   -1,  370,  371,
  372,   -1,  374,  375,  376,  377,  378,   -1,  380,  381,
   -1,   -1,   -1,   -1,  386,   -1,  388,  389,   -1,   -1,
  392,  393,  394,   -1,   -1,  397,   -1,  399,   -1,   -1,
  402,  403,  404,  405,   -1,  407,   -1,  409,   -1,  411,
  412,  413,  414,  415,   -1,   -1,  418,  419,  420,   -1,
  422,  423,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  433,  434,  435,   -1,  437,  438,  439,   -1,   -1,
   -1,   -1,  602,  445,   -1,   -1,   -1,  449,   -1,  451,
  452,  453,  454,   -1,  456,   -1,  458,  459,  460,  461,
  462,  463,  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  475,  476,   -1,  478,  479,  480,   -1,
   -1,   -1,   -1,  485,  486,  487,  488,  489,   -1,  491,
   -1,   -1,  494,  495,   -1,   -1,   -1,  499,  500,   -1,
   -1,  503,  504,  505,  506,  507,  508,  509,  510,  511,
   -1,  513,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  523,  524,  525,  526,  527,  528,  529,  530,  531,
  532,  533,  534,  535,  536,  537,  538,  539,  540,  541,
  542,  543,   -1,  545,  546,  547,  548,  549,  550,   -1,
  552,  553,  554,  555,   -1,  557,   -1,  559,  560,  561,
  562,  563,  564,  565,  566,  567,  568,  569,  570,  571,
  572,  573,  574,  575,  576,   -1,  578,  579,  580,  581,
   -1,  583,  584,  585,  586,  587,   -1,  589,  590,  591,
  592,  593,  594,  595,  596,  597,  598,  599,  600,  601,
  602,  603,  604,  605,  606,  607,  608,   -1,  610,  611,
  612,  613,  614,  615,  616,  617,   -1,  619,  620,   60,
   61,   62,  624,  625,  626,  627,  628,  629,   -1,   -1,
  632,  633,  634,  635,  636,  637,  638,  639,  640,   -1,
   -1,   -1,   -1,   -1,  646,   -1,  648,  649,  650,  651,
  652,  653,   -1,  655,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  672,  673,   -1,  675,  676,  677,  678,  679,  680,  681,
  682,  683,   -1,  685,  686,  687,  688,  689,  690,  691,
  692,  693,  694,  695,  696,   -1,  698,  699,  700,   -1,
  702,   -1,   -1,   -1,  706,  707,  708,  709,  710,  711,
  712,  713,  714,  715,  716,  717,  718,    0,  720,  721,
   -1,   -1,   -1,   -1,   -1,  727,  728,  729,  730,  731,
  732,  733,  734,  735,  736,  737,  738,  739,  740,  741,
  742,  743,  744,   -1,   -1,  747,   -1,  749,  750,  751,
  752,  753,  754,   -1,   -1,   -1,   -1,   -1,   41,   42,
   43,   44,   45,   -1,   47,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   59,   60,   61,   62,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   93,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  262,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  271,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  284,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  318,   -1,   -1,
   -1,   -1,   -1,   -1,  325,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  333,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  350,
   -1,  352,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  362,   -1,  364,  365,  366,   -1,   -1,   -1,   -1,
  371,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  412,   -1,   -1,  257,   -1,  259,   -1,   -1,  262,
   -1,  264,  265,   -1,   -1,  268,  269,   -1,  271,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  279,  280,   -1,  282,
   -1,  284,   -1,   -1,  287,   -1,   -1,  290,   -1,   -1,
   -1,   -1,   -1,   -1,  297,   -1,  299,  300,   -1,  302,
  303,  304,  305,  306,   -1,   -1,  309,   -1,   -1,  312,
   -1,  314,  315,   -1,   -1,  318,  319,  320,   -1,  322,
  323,  324,  325,  326,  327,  328,   -1,   -1,   -1,  332,
  333,  334,  335,  336,   -1,   -1,   -1,  340,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  348,  349,  350,  351,  352,
  353,   -1,   -1,   -1,  357,  358,   -1,  360,   -1,  362,
   -1,  364,  365,  366,   -1,  368,   -1,  370,  371,  372,
   -1,  374,  375,  376,  377,  378,   -1,  380,  381,   -1,
   -1,   -1,   -1,  386,   -1,  388,  389,   -1,   -1,  392,
  393,  394,   -1,   -1,  397,   -1,  399,   -1,   -1,  402,
  403,  404,  405,   -1,  407,   -1,  409,   -1,  411,  412,
  413,  414,  415,   -1,   -1,  418,  419,  420,   -1,  422,
  423,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  433,  434,  435,   -1,  437,  438,  439,   -1,   -1,   -1,
   -1,  602,  445,   -1,   -1,   -1,  449,   -1,  451,  452,
  453,  454,   -1,  456,   -1,  458,  459,  460,  461,  462,
  463,  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  475,  476,   -1,  478,  479,  480,   -1,   -1,
   -1,   -1,  485,  486,  487,  488,  489,   -1,  491,   -1,
   -1,  494,  495,   -1,   -1,   -1,  499,  500,   -1,   -1,
  503,  504,  505,  506,  507,  508,  509,  510,  511,   -1,
  513,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  523,  524,  525,  526,  527,  528,  529,  530,  531,  532,
  533,  534,  535,  536,  537,  538,  539,  540,  541,  542,
  543,   -1,  545,  546,  547,  548,  549,  550,   -1,  552,
  553,  554,  555,   -1,  557,   -1,  559,  560,  561,  562,
  563,  564,  565,  566,  567,  568,  569,  570,  571,  572,
  573,  574,  575,  576,   -1,  578,  579,  580,  581,   -1,
  583,  584,  585,  586,  587,   -1,  589,  590,  591,  592,
  593,  594,  595,  596,  597,  598,  599,  600,  601,  602,
  603,  604,  605,  606,  607,  608,   -1,  610,  611,  612,
  613,  614,  615,  616,  617,   -1,  619,  620,   -1,   -1,
   -1,  624,  625,  626,  627,  628,  629,   -1,   -1,  632,
  633,  634,  635,  636,  637,  638,  639,  640,   -1,   -1,
   -1,   -1,   -1,  646,   -1,  648,  649,  650,  651,  652,
  653,   -1,  655,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  672,
  673,   -1,  675,  676,  677,  678,  679,  680,  681,  682,
  683,   -1,  685,  686,  687,  688,  689,  690,  691,  692,
  693,  694,  695,  696,   -1,  698,  699,  700,   -1,  702,
   -1,   -1,   -1,  706,  707,  708,  709,  710,  711,  712,
  713,  714,  715,  716,  717,  718,    0,  720,  721,   -1,
   -1,   -1,   -1,   -1,  727,  728,  729,  730,  731,  732,
  733,  734,  735,  736,  737,  738,  739,  740,  741,  742,
  743,  744,   -1,   -1,  747,   -1,  749,  750,  751,  752,
  753,  754,   -1,   -1,   -1,   -1,   -1,   41,   42,   43,
   44,   45,   -1,   47,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   59,   60,   61,   62,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   93,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  257,   -1,  259,   -1,   -1,  262,   -1,
  264,  265,   -1,   -1,  268,  269,   -1,  271,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  279,  280,   -1,  282,   -1,
  284,   -1,   -1,  287,   -1,   -1,  290,   -1,   -1,   -1,
   -1,   -1,   -1,  297,   -1,  299,  300,   -1,  302,  303,
  304,  305,  306,   -1,   -1,  309,   -1,   -1,  312,   -1,
  314,  315,   -1,   -1,  318,  319,  320,   -1,  322,  323,
  324,  325,  326,  327,  328,   -1,   -1,   -1,  332,  333,
  334,  335,  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  348,  349,  350,  351,  352,  353,
   -1,   -1,   -1,  357,  358,   -1,  360,   -1,  362,   -1,
  364,  365,  366,   -1,  368,   -1,  370,  371,  372,   -1,
  374,  375,  376,  377,  378,   -1,  380,  381,   -1,   -1,
   -1,   -1,  386,   -1,  388,  389,   -1,   -1,  392,  393,
  394,   -1,   -1,  397,   -1,  399,   -1,   -1,  402,  403,
  404,  405,   -1,  407,   -1,  409,   -1,  411,  412,  413,
  414,  415,   -1,   -1,  418,  419,  420,   -1,  422,  423,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  433,
  434,  435,   -1,  437,  438,  439,   -1,   -1,   -1,   -1,
   -1,  445,   -1,   -1,   -1,  449,   -1,  451,  452,  453,
  454,   -1,  456,   -1,  458,  459,  460,  461,  462,  463,
  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  475,  476,   -1,  478,  479,  480,   -1,   -1,   -1,
   -1,  485,  486,  487,  488,  489,   -1,  491,   -1,   -1,
  494,  495,   -1,   -1,   -1,  499,  500,   -1,   -1,  503,
  504,  505,  506,  507,  508,  509,  510,  511,   -1,  513,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  523,
  524,  525,  526,  527,  528,  529,  530,  531,  532,  533,
  534,  535,  536,  537,  538,  539,  540,  541,  542,  543,
   -1,  545,  546,  547,  548,  549,  550,   -1,  552,  553,
  554,  555,   -1,  557,   -1,  559,  560,  561,  562,  563,
  564,  565,  566,  567,  568,  569,  570,  571,  572,  573,
  574,  575,  576,   -1,  578,  579,  580,  581,   -1,  583,
  584,  585,  586,  587,   -1,  589,  590,  591,  592,  593,
  594,  595,  596,  597,  598,  599,  600,  601,  602,  603,
  604,  605,  606,  607,  608,   -1,  610,  611,  612,  613,
  614,  615,  616,  617,   -1,  619,  620,   -1,   -1,   -1,
  624,  625,  626,  627,  628,  629,   -1,   -1,  632,  633,
  634,  635,  636,  637,  638,  639,  640,   -1,   -1,   -1,
   -1,   -1,  646,   -1,  648,  649,  650,  651,  652,  653,
   -1,  655,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  672,  673,
   -1,  675,  676,  677,  678,  679,  680,  681,  682,  683,
   -1,  685,  686,  687,  688,  689,  690,  691,  692,  693,
  694,  695,  696,   -1,  698,  699,  700,   -1,  702,   -1,
   -1,   -1,  706,  707,  708,  709,  710,  711,  712,  713,
  714,  715,  716,  717,  718,    0,  720,  721,   -1,   -1,
   -1,   -1,   -1,  727,  728,  729,  730,  731,  732,  733,
  734,  735,  736,  737,  738,  739,  740,  741,  742,  743,
  744,   -1,   -1,  747,   -1,  749,  750,  751,  752,  753,
  754,   -1,   -1,   -1,   -1,   -1,   41,   42,   43,   44,
   45,   -1,   47,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   59,   60,   61,   62,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   93,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  257,   -1,  259,   -1,   -1,  262,   -1,  264,
  265,   -1,   -1,  268,  269,   -1,  271,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  279,  280,   -1,  282,   -1,  284,
   -1,   -1,  287,   -1,   -1,  290,   -1,   -1,   -1,   -1,
   -1,   -1,  297,   -1,  299,  300,   -1,  302,  303,  304,
  305,  306,   -1,   -1,  309,   -1,   -1,  312,   -1,  314,
  315,   -1,   -1,  318,  319,  320,   -1,  322,  323,  324,
  325,  326,  327,  328,   -1,   -1,   -1,  332,  333,  334,
  335,  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  348,  349,  350,  351,  352,  353,   -1,
   -1,   -1,  357,  358,   -1,  360,   -1,  362,   -1,  364,
  365,  366,   -1,  368,   -1,  370,  371,  372,   -1,  374,
  375,  376,  377,  378,   -1,  380,  381,   -1,   -1,   -1,
   -1,  386,   -1,  388,  389,   -1,   -1,  392,  393,  394,
   -1,   -1,  397,   -1,  399,   -1,   -1,  402,  403,  404,
  405,   -1,  407,   -1,  409,   -1,  411,  412,  413,  414,
  415,   -1,   -1,  418,  419,  420,   -1,  422,  423,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  433,  434,
  435,   -1,  437,  438,  439,   -1,   -1,   -1,   -1,   -1,
  445,   -1,   -1,   -1,  449,   -1,  451,  452,  453,  454,
   -1,  456,   -1,  458,  459,  460,  461,  462,  463,  464,
  465,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  475,  476,   -1,  478,  479,  480,   -1,   -1,   -1,   -1,
  485,  486,  487,  488,  489,   -1,  491,   -1,   -1,  494,
  495,   -1,   -1,   -1,  499,  500,   -1,   -1,  503,  504,
  505,  506,  507,  508,  509,  510,  511,   -1,  513,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  523,  524,
  525,  526,  527,  528,  529,  530,  531,  532,  533,  534,
  535,  536,  537,  538,  539,  540,  541,  542,  543,   -1,
  545,  546,  547,  548,  549,  550,   -1,  552,  553,  554,
  555,   -1,  557,   -1,  559,  560,  561,  562,  563,  564,
  565,  566,  567,  568,  569,  570,  571,  572,  573,  574,
  575,  576,   -1,  578,  579,  580,  581,   -1,  583,  584,
  585,  586,  587,   -1,  589,  590,  591,  592,  593,  594,
  595,  596,  597,  598,  599,  600,  601,  602,  603,  604,
  605,  606,  607,  608,   -1,  610,  611,  612,  613,  614,
  615,  616,  617,   -1,  619,  620,   -1,   -1,   -1,  624,
  625,  626,  627,  628,  629,   -1,   -1,  632,  633,  634,
  635,  636,  637,  638,  639,  640,   -1,   -1,   -1,   -1,
   -1,  646,   -1,  648,  649,  650,  651,  652,  653,   -1,
  655,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  672,  673,   -1,
  675,  676,  677,  678,  679,  680,  681,  682,  683,   -1,
  685,  686,  687,  688,  689,  690,  691,  692,  693,  694,
  695,  696,   -1,  698,  699,  700,   -1,  702,   -1,   -1,
   -1,  706,  707,  708,  709,  710,  711,  712,  713,  714,
  715,  716,  717,  718,    0,  720,  721,   -1,   -1,   -1,
   -1,   -1,  727,  728,  729,  730,  731,  732,  733,  734,
  735,  736,  737,  738,  739,  740,  741,  742,  743,  744,
   -1,   -1,  747,   -1,  749,  750,  751,  752,  753,  754,
   -1,   -1,   -1,   -1,   -1,   41,   -1,   43,   44,   45,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   59,   60,   61,   62,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   93,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  257,   -1,  259,   -1,   -1,  262,   -1,  264,  265,
   -1,   -1,  268,  269,   -1,  271,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  279,  280,   -1,  282,   -1,  284,   -1,
   -1,  287,   -1,   -1,  290,   -1,   -1,   -1,   -1,   -1,
   -1,  297,   -1,  299,  300,   -1,  302,  303,  304,  305,
  306,   -1,   -1,  309,   -1,   -1,  312,   -1,  314,  315,
   -1,   -1,  318,  319,  320,   -1,  322,  323,  324,  325,
  326,  327,  328,   -1,   -1,   -1,  332,  333,  334,  335,
  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  348,  349,  350,  351,  352,  353,   -1,   -1,
   -1,  357,  358,   -1,  360,   -1,  362,   -1,  364,  365,
  366,   -1,  368,   -1,  370,  371,  372,   -1,  374,  375,
  376,  377,  378,   -1,  380,  381,   -1,   -1,   -1,   -1,
  386,   -1,  388,  389,   -1,   -1,  392,  393,  394,   -1,
   -1,  397,   -1,  399,   -1,   -1,  402,  403,  404,  405,
   -1,  407,   -1,  409,   -1,  411,  412,  413,  414,  415,
   -1,   -1,  418,  419,  420,   -1,  422,  423,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  433,  434,  435,
   -1,  437,  438,  439,   -1,   -1,   -1,   -1,   -1,  445,
   -1,   -1,   -1,  449,   -1,  451,  452,  453,  454,   -1,
  456,   -1,  458,  459,  460,  461,  462,  463,  464,  465,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  475,
  476,   -1,  478,  479,  480,   -1,   -1,   -1,   -1,  485,
  486,  487,  488,  489,   -1,  491,   -1,   -1,  494,  495,
   -1,   -1,   -1,  499,  500,   -1,   -1,  503,  504,  505,
  506,  507,  508,  509,  510,  511,   -1,  513,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  523,  524,  525,
  526,  527,  528,  529,  530,  531,  532,  533,  534,  535,
  536,  537,  538,  539,  540,  541,  542,  543,   -1,  545,
  546,  547,  548,  549,  550,   -1,  552,  553,  554,  555,
   -1,  557,   -1,  559,  560,  561,  562,  563,  564,  565,
  566,  567,  568,  569,  570,  571,  572,  573,  574,  575,
  576,   -1,  578,  579,  580,  581,   -1,  583,  584,  585,
  586,  587,   -1,  589,  590,  591,  592,  593,  594,  595,
  596,  597,  598,  599,  600,  601,  602,  603,  604,  605,
  606,  607,  608,   -1,  610,  611,  612,  613,  614,  615,
  616,  617,   -1,  619,  620,   -1,   -1,   -1,  624,  625,
  626,  627,  628,  629,   -1,   -1,  632,  633,  634,  635,
  636,  637,  638,  639,  640,   -1,   -1,   -1,   -1,   -1,
  646,   -1,  648,  649,  650,  651,  652,  653,   -1,  655,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  672,  673,   -1,  675,
  676,  677,  678,  679,  680,  681,  682,  683,   -1,  685,
  686,  687,  688,  689,  690,  691,  692,  693,  694,  695,
  696,   -1,  698,  699,  700,   -1,  702,   -1,   -1,   -1,
  706,  707,  708,  709,  710,  711,  712,  713,  714,  715,
  716,  717,  718,    0,  720,  721,   -1,   -1,   -1,   -1,
   -1,  727,  728,  729,  730,  731,  732,  733,  734,  735,
  736,  737,  738,  739,  740,  741,  742,  743,  744,   -1,
   -1,  747,   -1,  749,  750,  751,  752,  753,  754,   -1,
   -1,   -1,   -1,   -1,   41,   -1,   43,   44,   45,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   59,   60,   61,   62,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   93,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  257,   -1,  259,   -1,   -1,  262,   -1,  264,  265,   -1,
   -1,  268,  269,   -1,  271,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  279,  280,   -1,  282,   -1,  284,   -1,   -1,
  287,   -1,   -1,  290,   -1,   -1,   -1,   -1,   -1,   -1,
  297,   -1,  299,  300,   -1,  302,  303,  304,  305,  306,
   -1,   -1,  309,   -1,   -1,  312,   -1,  314,  315,   -1,
   -1,  318,  319,  320,   -1,  322,  323,  324,  325,  326,
  327,  328,   -1,   -1,   -1,  332,  333,  334,  335,  336,
   -1,   -1,   -1,  340,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  348,  349,  350,  351,  352,  353,   -1,   -1,   -1,
  357,  358,   -1,  360,   -1,  362,   -1,  364,  365,  366,
   -1,  368,   -1,  370,  371,  372,   -1,  374,  375,  376,
  377,  378,   -1,  380,  381,   -1,   -1,   -1,   -1,  386,
   -1,  388,  389,   -1,   -1,  392,  393,  394,   -1,   -1,
  397,   -1,  399,   -1,   -1,  402,  403,  404,  405,   -1,
  407,   -1,  409,   -1,  411,  412,  413,  414,  415,   -1,
   -1,  418,  419,  420,   -1,  422,  423,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  433,  434,  435,   -1,
  437,  438,  439,   -1,   -1,   -1,   -1,   -1,  445,   -1,
   -1,   -1,  449,   -1,  451,  452,  453,  454,   -1,  456,
   -1,  458,  459,  460,  461,  462,  463,  464,  465,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  475,  476,
   -1,  478,  479,  480,   -1,   -1,   -1,   -1,  485,  486,
  487,  488,  489,   -1,  491,   -1,   -1,  494,  495,   -1,
   -1,   -1,  499,  500,   -1,   -1,  503,  504,  505,  506,
  507,  508,  509,  510,  511,   -1,  513,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  523,  524,  525,  526,
  527,  528,  529,  530,  531,  532,  533,  534,  535,  536,
  537,  538,  539,  540,  541,  542,  543,   -1,  545,  546,
  547,  548,  549,  550,   -1,  552,  553,  554,  555,   -1,
  557,   -1,  559,  560,  561,  562,  563,  564,  565,  566,
  567,  568,  569,  570,  571,  572,  573,  574,  575,  576,
   -1,  578,  579,  580,  581,   -1,  583,  584,  585,  586,
  587,   -1,  589,  590,  591,  592,  593,  594,  595,  596,
  597,  598,  599,  600,  601,  602,  603,  604,  605,  606,
  607,  608,   -1,  610,  611,  612,  613,  614,  615,  616,
  617,   -1,  619,  620,   -1,   -1,   -1,  624,  625,  626,
  627,  628,  629,   -1,   -1,  632,  633,  634,  635,  636,
  637,  638,  639,  640,   -1,   -1,   -1,   -1,   -1,  646,
   -1,  648,  649,  650,  651,  652,  653,   -1,  655,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  672,  673,   -1,  675,  676,
  677,  678,  679,  680,  681,  682,  683,   -1,  685,  686,
  687,  688,  689,  690,  691,  692,  693,  694,  695,  696,
   -1,  698,  699,  700,   -1,  702,   -1,    0,   -1,  706,
  707,  708,  709,  710,  711,  712,  713,  714,  715,  716,
  717,  718,   -1,  720,  721,   -1,   -1,   -1,   -1,   -1,
  727,  728,  729,  730,  731,  732,  733,  734,  735,  736,
  737,  738,  739,  740,  741,  742,  743,  744,   41,   -1,
  747,   44,  749,  750,  751,  752,  753,  754,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   59,   60,   61,   62,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   93,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  257,   -1,  259,   -1,   -1,  262,
   -1,  264,  265,   -1,   -1,  268,  269,   -1,  271,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  279,  280,   -1,  282,
   -1,  284,   -1,   -1,  287,   -1,   -1,  290,   -1,   -1,
   -1,   -1,   -1,   -1,  297,   -1,  299,  300,   -1,  302,
  303,  304,  305,  306,   -1,   -1,  309,   -1,   -1,  312,
   -1,  314,  315,   -1,   -1,  318,  319,  320,   -1,  322,
  323,  324,  325,  326,  327,  328,   -1,   -1,   -1,  332,
  333,  334,  335,  336,   -1,   -1,   -1,  340,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  348,  349,  350,  351,  352,
  353,   -1,   -1,   -1,  357,  358,   -1,  360,   -1,  362,
   -1,  364,  365,  366,   -1,  368,   -1,  370,  371,  372,
   -1,  374,  375,  376,  377,  378,   -1,  380,  381,   -1,
   -1,   -1,   -1,  386,   -1,  388,  389,   -1,   -1,  392,
  393,  394,   -1,   -1,  397,   -1,  399,   -1,   -1,  402,
  403,  404,  405,   -1,  407,   -1,  409,   -1,  411,  412,
  413,  414,  415,   -1,   -1,  418,  419,  420,   -1,  422,
  423,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  433,  434,  435,   -1,  437,  438,  439,   -1,   -1,   -1,
   -1,   -1,  445,   -1,   -1,   -1,  449,   -1,  451,  452,
  453,  454,   -1,  456,   -1,  458,  459,  460,  461,  462,
  463,  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  475,  476,   -1,  478,  479,  480,   -1,   -1,
   -1,   -1,  485,  486,  487,  488,  489,   -1,  491,   -1,
   -1,  494,  495,   -1,   -1,   -1,  499,  500,   -1,   -1,
  503,  504,  505,  506,  507,  508,  509,  510,  511,   -1,
  513,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  523,  524,  525,  526,  527,  528,  529,  530,  531,  532,
  533,  534,  535,  536,  537,  538,  539,  540,  541,  542,
  543,   -1,  545,  546,  547,  548,  549,  550,   -1,  552,
  553,  554,  555,   -1,  557,   -1,  559,  560,  561,  562,
  563,  564,  565,  566,  567,  568,  569,  570,  571,  572,
  573,  574,  575,  576,   -1,  578,  579,  580,  581,   -1,
  583,  584,  585,  586,  587,   -1,  589,  590,  591,  592,
  593,  594,  595,  596,  597,  598,  599,  600,  601,  602,
  603,  604,  605,  606,  607,  608,   -1,  610,  611,  612,
  613,  614,  615,  616,  617,   -1,  619,  620,   -1,   -1,
   -1,  624,  625,  626,  627,  628,  629,   -1,   -1,  632,
  633,  634,  635,  636,  637,  638,  639,  640,   -1,   -1,
   -1,   -1,   -1,  646,   -1,  648,  649,  650,  651,  652,
  653,   -1,  655,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  672,
  673,   -1,  675,  676,  677,  678,  679,  680,  681,  682,
  683,   -1,  685,  686,  687,  688,  689,  690,  691,  692,
  693,  694,  695,  696,   -1,  698,  699,  700,   -1,  702,
   -1,    0,   -1,  706,  707,  708,  709,  710,  711,  712,
  713,  714,  715,  716,  717,  718,   -1,  720,  721,   -1,
   -1,   -1,   -1,   -1,  727,  728,  729,  730,  731,  732,
  733,  734,  735,  736,  737,  738,  739,  740,  741,  742,
  743,  744,   41,   -1,  747,   44,  749,  750,  751,  752,
  753,  754,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   59,   60,   61,   62,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   93,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  257,   -1,
  259,   -1,   -1,  262,   -1,  264,  265,   -1,   -1,  268,
  269,   -1,  271,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  279,  280,   -1,  282,   -1,  284,   -1,   -1,  287,   -1,
   -1,  290,   -1,   -1,   -1,   -1,   -1,   -1,  297,   -1,
  299,  300,   -1,  302,  303,  304,  305,  306,   -1,   -1,
  309,   -1,   -1,  312,   -1,  314,  315,   -1,   -1,  318,
  319,  320,   -1,  322,  323,  324,  325,  326,  327,  328,
   -1,   -1,   -1,  332,  333,  334,  335,  336,   -1,   -1,
   -1,  340,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  348,
  349,  350,  351,  352,  353,   -1,   -1,   -1,  357,  358,
   -1,  360,   -1,  362,   -1,  364,  365,  366,   -1,  368,
   -1,  370,  371,  372,   -1,  374,  375,  376,  377,  378,
   -1,  380,  381,   -1,   -1,   -1,   -1,  386,   -1,  388,
  389,   -1,   -1,  392,  393,  394,   -1,   -1,  397,   -1,
  399,   -1,   -1,  402,  403,  404,  405,   -1,  407,   -1,
  409,   -1,  411,  412,  413,  414,  415,   -1,   -1,  418,
  419,  420,   -1,  422,  423,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  433,  434,  435,   -1,  437,  438,
  439,   -1,   -1,   -1,   -1,   -1,  445,   -1,   -1,   -1,
  449,   -1,  451,  452,  453,  454,   -1,  456,   -1,  458,
  459,  460,  461,  462,  463,  464,  465,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  475,  476,   -1,  478,
  479,  480,   -1,   -1,   -1,   -1,  485,  486,  487,  488,
  489,   -1,  491,   -1,   -1,  494,  495,   -1,   -1,   -1,
  499,  500,   -1,   -1,  503,  504,  505,  506,  507,  508,
  509,  510,  511,   -1,  513,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  523,  524,  525,  526,  527,  528,
  529,  530,  531,  532,  533,  534,  535,  536,  537,  538,
  539,  540,  541,  542,  543,   -1,  545,  546,  547,  548,
  549,  550,   -1,  552,  553,  554,  555,   -1,  557,   -1,
  559,  560,  561,  562,  563,  564,  565,  566,  567,  568,
  569,  570,  571,  572,  573,  574,  575,  576,   -1,  578,
  579,  580,  581,   -1,  583,  584,  585,  586,  587,   -1,
  589,  590,  591,  592,  593,  594,  595,  596,  597,  598,
  599,  600,  601,  602,  603,  604,  605,  606,  607,  608,
   -1,  610,  611,  612,  613,  614,  615,  616,  617,   -1,
  619,  620,   -1,   -1,   -1,  624,  625,  626,  627,  628,
  629,   -1,   -1,  632,  633,  634,  635,  636,  637,  638,
  639,  640,   -1,   -1,   -1,   -1,   -1,  646,   -1,  648,
  649,  650,  651,  652,  653,   -1,  655,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  672,  673,   -1,  675,  676,  677,  678,
  679,  680,  681,  682,  683,   -1,  685,  686,  687,  688,
  689,  690,  691,  692,  693,  694,  695,  696,   -1,  698,
  699,  700,   -1,  702,   -1,    0,   -1,  706,  707,  708,
  709,  710,  711,  712,  713,  714,  715,  716,  717,  718,
   -1,  720,  721,   -1,   -1,   -1,   -1,   -1,  727,  728,
  729,  730,  731,  732,  733,  734,  735,  736,  737,  738,
  739,  740,  741,  742,  743,  744,   41,   -1,  747,   44,
  749,  750,  751,  752,  753,  754,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   59,   60,   61,   62,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   93,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  257,   -1,  259,   -1,   -1,  262,   -1,  264,
  265,   -1,   -1,  268,  269,   -1,  271,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  279,  280,   -1,  282,   -1,  284,
   -1,   -1,  287,   -1,   -1,  290,   -1,   -1,   -1,   -1,
   -1,   -1,  297,   -1,  299,  300,   -1,  302,  303,  304,
  305,  306,   -1,   -1,  309,   -1,   -1,  312,   -1,  314,
  315,   -1,   -1,  318,  319,  320,   -1,  322,  323,  324,
  325,  326,  327,  328,   -1,   -1,   -1,  332,   -1,  334,
  335,  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  348,  349,  350,  351,  352,  353,   -1,
   -1,   -1,  357,  358,   -1,  360,   -1,  362,   -1,  364,
  365,  366,   -1,  368,   -1,  370,  371,  372,   -1,  374,
  375,  376,  377,  378,   -1,  380,  381,   -1,   -1,   -1,
   -1,  386,   -1,  388,  389,   -1,   -1,  392,  393,  394,
   -1,   -1,  397,   -1,  399,   -1,   -1,  402,  403,  404,
  405,   -1,  407,   -1,  409,   -1,  411,  412,  413,  414,
  415,   -1,   -1,  418,  419,  420,   -1,  422,  423,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  433,  434,
  435,   -1,  437,  438,  439,   -1,   -1,   -1,   -1,   -1,
  445,   -1,   -1,   -1,  449,   -1,  451,  452,  453,  454,
   -1,  456,   -1,  458,  459,  460,  461,  462,  463,  464,
  465,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  475,  476,   -1,  478,  479,  480,   -1,   -1,   -1,   -1,
  485,  486,  487,  488,  489,   -1,  491,   -1,   -1,  494,
  495,   -1,   -1,   -1,  499,  500,   -1,   -1,  503,  504,
  505,  506,  507,  508,  509,  510,  511,   -1,  513,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  523,  524,
  525,  526,  527,  528,  529,  530,  531,  532,  533,  534,
  535,  536,  537,  538,  539,  540,  541,  542,  543,   -1,
  545,  546,  547,  548,  549,  550,   -1,  552,  553,  554,
  555,   -1,  557,   -1,  559,  560,  561,  562,  563,  564,
  565,  566,  567,  568,  569,  570,  571,  572,  573,  574,
  575,  576,   -1,  578,  579,  580,  581,   -1,  583,  584,
  585,  586,  587,   -1,  589,  590,  591,  592,  593,  594,
  595,  596,  597,  598,  599,  600,  601,  602,  603,  604,
  605,  606,  607,  608,   -1,  610,  611,  612,  613,  614,
  615,  616,  617,   -1,  619,  620,   -1,   -1,   -1,  624,
  625,  626,  627,  628,  629,   -1,   -1,  632,  633,  634,
  635,  636,  637,  638,  639,  640,   -1,   -1,   -1,   -1,
   -1,  646,   -1,  648,  649,  650,  651,  652,  653,   -1,
  655,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  672,  673,   -1,
  675,  676,  677,  678,  679,  680,  681,  682,  683,   -1,
  685,  686,  687,  688,  689,  690,  691,  692,  693,  694,
  695,  696,   -1,  698,  699,  700,   -1,  702,   -1,    0,
   -1,  706,  707,  708,  709,  710,  711,  712,  713,  714,
  715,  716,  717,  718,   -1,  720,  721,   -1,   -1,   -1,
   -1,   -1,  727,  728,  729,  730,  731,  732,  733,  734,
  735,  736,  737,  738,  739,  740,  741,  742,  743,  744,
   41,   -1,  747,   44,  749,  750,  751,  752,  753,  754,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   59,   60,
   61,   62,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   93,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  257,   -1,  259,   -1,
   -1,  262,   -1,  264,  265,   -1,   -1,  268,  269,   -1,
  271,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  279,  280,
   -1,  282,   -1,  284,   -1,   -1,  287,   -1,   -1,  290,
   -1,   -1,   -1,   -1,   -1,   -1,  297,   -1,  299,  300,
   -1,  302,  303,  304,  305,  306,   -1,   -1,  309,   -1,
   -1,  312,   -1,  314,  315,   -1,   -1,  318,  319,  320,
   -1,  322,  323,  324,  325,  326,  327,  328,   -1,   -1,
   -1,  332,   -1,  334,  335,  336,   -1,   -1,   -1,  340,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  348,  349,  350,
  351,  352,  353,   -1,   -1,   -1,  357,  358,   -1,  360,
   -1,  362,   -1,  364,  365,  366,   -1,  368,   -1,  370,
  371,  372,   -1,  374,  375,  376,  377,  378,   -1,  380,
  381,   -1,   -1,   -1,   -1,  386,   -1,  388,  389,   -1,
   -1,  392,  393,  394,   -1,   -1,  397,   -1,  399,   -1,
   -1,  402,  403,  404,  405,   -1,  407,   -1,  409,   -1,
  411,  412,  413,  414,  415,   -1,   -1,  418,  419,  420,
   -1,  422,  423,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  433,  434,  435,   -1,  437,  438,  439,   -1,
   -1,   -1,   -1,   -1,  445,   -1,   -1,   -1,  449,   -1,
  451,  452,  453,  454,   -1,  456,   -1,  458,  459,  460,
  461,  462,  463,  464,  465,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  475,  476,   -1,  478,  479,  480,
   -1,   -1,   -1,   -1,  485,  486,  487,  488,  489,   -1,
  491,   -1,   -1,  494,  495,   -1,   -1,   -1,  499,  500,
   -1,   -1,  503,  504,  505,  506,  507,  508,  509,  510,
  511,   -1,  513,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  523,  524,  525,  526,  527,  528,  529,  530,
  531,  532,  533,  534,  535,  536,  537,  538,  539,  540,
  541,  542,  543,   -1,  545,  546,  547,  548,  549,  550,
   -1,  552,  553,  554,  555,   -1,  557,   -1,  559,  560,
  561,  562,  563,  564,  565,  566,  567,  568,  569,  570,
  571,  572,  573,  574,  575,  576,   -1,  578,  579,  580,
  581,   -1,  583,  584,  585,  586,  587,   -1,  589,  590,
  591,  592,  593,  594,  595,  596,  597,  598,  599,  600,
  601,  602,  603,  604,  605,  606,  607,  608,   -1,  610,
  611,  612,  613,  614,  615,  616,  617,   -1,  619,  620,
   -1,   -1,   -1,  624,  625,  626,  627,  628,  629,   -1,
   -1,  632,  633,  634,  635,  636,  637,  638,  639,  640,
   -1,   -1,   -1,   -1,   -1,  646,   -1,  648,  649,  650,
  651,  652,  653,   -1,  655,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  672,  673,   -1,  675,  676,  677,  678,  679,  680,
  681,  682,  683,   -1,  685,  686,  687,  688,  689,  690,
  691,  692,  693,  694,  695,  696,   -1,  698,  699,  700,
   -1,  702,   -1,    0,   -1,  706,  707,  708,  709,  710,
  711,  712,  713,  714,  715,  716,  717,  718,   -1,  720,
  721,   -1,   -1,   -1,   -1,   -1,  727,  728,  729,  730,
  731,  732,  733,  734,  735,  736,  737,  738,  739,  740,
  741,  742,  743,  744,   41,   -1,  747,   44,  749,  750,
  751,  752,  753,  754,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   59,   60,   61,   62,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   93,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  257,   -1,  259,   -1,   -1,  262,   -1,  264,  265,   -1,
   -1,  268,  269,   -1,  271,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  279,  280,   -1,  282,   -1,  284,   -1,   -1,
  287,   -1,   -1,  290,   -1,   -1,   -1,   -1,   -1,   -1,
  297,   -1,  299,  300,   -1,  302,  303,  304,  305,  306,
   -1,   -1,  309,   -1,   -1,  312,   -1,  314,  315,   -1,
   -1,  318,  319,  320,   -1,  322,  323,  324,  325,  326,
  327,  328,   -1,   -1,   -1,  332,   -1,  334,  335,  336,
   -1,   -1,   -1,  340,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  348,  349,  350,  351,  352,  353,   -1,   -1,   -1,
  357,  358,   -1,  360,   -1,  362,   -1,  364,  365,  366,
   -1,  368,   -1,  370,  371,  372,   -1,  374,  375,  376,
  377,  378,   -1,  380,  381,   -1,   -1,   -1,   -1,  386,
   -1,  388,  389,   -1,   -1,  392,  393,  394,   -1,   -1,
  397,   -1,  399,   -1,   -1,  402,  403,  404,  405,   -1,
  407,   -1,  409,   -1,  411,  412,  413,  414,  415,   -1,
   -1,  418,  419,  420,   -1,  422,  423,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  433,  434,  435,   -1,
  437,  438,  439,   -1,   -1,   -1,   -1,   -1,  445,   -1,
   -1,   -1,  449,   -1,  451,  452,  453,  454,   -1,  456,
   -1,  458,  459,  460,  461,  462,  463,  464,  465,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  475,  476,
   -1,  478,  479,  480,   -1,   -1,   -1,   -1,  485,  486,
  487,  488,  489,   -1,  491,   -1,   -1,  494,  495,   -1,
   -1,   -1,  499,  500,   -1,   -1,  503,  504,  505,  506,
  507,  508,  509,  510,  511,   -1,  513,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  523,  524,  525,  526,
  527,  528,  529,  530,  531,  532,  533,  534,  535,  536,
  537,  538,  539,  540,  541,  542,  543,   -1,  545,  546,
  547,  548,  549,  550,   -1,  552,  553,  554,  555,   -1,
  557,   -1,  559,  560,  561,  562,  563,  564,  565,  566,
  567,  568,  569,  570,  571,  572,  573,  574,  575,  576,
   -1,  578,  579,  580,  581,   -1,  583,  584,  585,  586,
  587,   -1,  589,  590,  591,  592,  593,  594,  595,  596,
  597,  598,  599,  600,  601,  602,  603,  604,  605,  606,
  607,  608,   -1,  610,  611,  612,  613,  614,  615,  616,
  617,   -1,  619,  620,   -1,   -1,   -1,  624,  625,  626,
  627,  628,  629,   -1,   -1,  632,  633,  634,  635,  636,
  637,  638,  639,  640,   -1,   -1,   -1,   -1,   -1,  646,
   -1,  648,  649,  650,  651,  652,  653,   -1,  655,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  672,  673,   -1,  675,  676,
  677,  678,  679,  680,  681,  682,  683,   -1,  685,  686,
  687,  688,  689,  690,  691,  692,  693,  694,  695,  696,
   -1,  698,  699,  700,   -1,  702,   -1,    0,   -1,  706,
  707,  708,  709,  710,  711,  712,  713,  714,  715,  716,
  717,  718,   -1,  720,  721,   -1,   -1,   -1,   -1,   -1,
  727,  728,  729,  730,  731,  732,  733,  734,  735,  736,
  737,  738,  739,  740,  741,  742,  743,  744,   41,   -1,
  747,   44,  749,  750,  751,  752,  753,  754,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   59,   60,   61,   62,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   93,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  257,   -1,  259,   -1,   -1,  262,
   -1,  264,  265,   -1,   -1,  268,  269,   -1,  271,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  279,  280,   -1,  282,
   -1,  284,   -1,   -1,  287,   -1,   -1,  290,   -1,   -1,
   -1,   -1,   -1,   -1,  297,   -1,  299,  300,   -1,  302,
  303,  304,  305,  306,   -1,   -1,  309,   -1,   -1,  312,
   -1,  314,  315,   -1,   -1,  318,  319,  320,   -1,  322,
  323,  324,  325,  326,  327,  328,   -1,   -1,   -1,  332,
   -1,  334,  335,  336,   -1,   -1,   -1,  340,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  348,  349,  350,  351,  352,
  353,   -1,   -1,   -1,  357,  358,   -1,  360,   -1,  362,
   -1,  364,  365,  366,   -1,  368,   -1,  370,  371,  372,
   -1,  374,  375,  376,  377,  378,   -1,  380,  381,   -1,
   -1,   -1,   -1,  386,   -1,  388,  389,   -1,   -1,  392,
  393,  394,   -1,   -1,  397,   -1,  399,   -1,   -1,  402,
  403,  404,  405,   -1,  407,   -1,  409,   -1,  411,  412,
  413,  414,  415,   -1,   -1,  418,  419,  420,   -1,  422,
  423,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  433,  434,  435,   -1,  437,  438,  439,   -1,   -1,   -1,
   -1,   -1,  445,   -1,   -1,   -1,  449,   -1,  451,  452,
  453,  454,   -1,  456,   -1,  458,  459,  460,  461,  462,
  463,  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  475,  476,   -1,  478,  479,  480,   -1,   -1,
   -1,   -1,  485,  486,  487,  488,  489,   -1,  491,   -1,
   -1,  494,  495,   -1,   -1,   -1,  499,  500,   -1,   -1,
  503,  504,  505,  506,  507,  508,  509,  510,  511,   -1,
  513,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  523,  524,  525,  526,  527,  528,  529,  530,  531,  532,
  533,  534,  535,  536,  537,  538,  539,  540,  541,  542,
  543,   -1,  545,  546,  547,  548,  549,  550,   -1,  552,
  553,  554,  555,   -1,  557,   -1,  559,  560,  561,  562,
  563,  564,  565,  566,  567,  568,  569,  570,  571,  572,
  573,  574,  575,  576,   -1,  578,  579,  580,  581,   -1,
  583,  584,  585,  586,  587,   -1,  589,  590,  591,  592,
  593,  594,  595,  596,  597,  598,  599,  600,  601,  602,
  603,  604,  605,  606,  607,  608,   -1,  610,  611,  612,
  613,  614,  615,  616,  617,   -1,  619,  620,   -1,   -1,
   -1,  624,  625,  626,  627,  628,  629,   -1,   -1,  632,
  633,  634,  635,  636,  637,  638,  639,  640,   -1,   -1,
   -1,   -1,   -1,  646,   -1,  648,  649,  650,  651,  652,
  653,   -1,  655,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  672,
  673,   -1,  675,  676,  677,  678,  679,  680,  681,  682,
  683,   -1,  685,  686,  687,  688,  689,  690,  691,  692,
  693,  694,  695,  696,   -1,  698,  699,  700,   -1,  702,
   -1,    0,   -1,  706,  707,  708,  709,  710,  711,  712,
  713,  714,  715,  716,  717,  718,   -1,  720,  721,   -1,
   -1,   -1,   -1,   -1,  727,  728,  729,  730,  731,  732,
  733,  734,  735,  736,  737,  738,  739,  740,  741,  742,
  743,  744,   41,   -1,  747,   44,  749,  750,  751,  752,
  753,  754,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   59,   60,   61,   62,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   93,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  257,   -1,
  259,   -1,   -1,  262,   -1,  264,  265,   -1,   -1,  268,
  269,   -1,  271,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  279,  280,   -1,  282,   -1,  284,   -1,   -1,  287,   -1,
   -1,  290,   -1,   -1,   -1,   -1,   -1,   -1,  297,   -1,
  299,  300,   -1,  302,  303,  304,  305,  306,   -1,   -1,
  309,   -1,   -1,  312,   -1,  314,  315,   -1,   -1,  318,
  319,  320,   -1,  322,  323,  324,  325,  326,  327,  328,
   -1,   -1,   -1,  332,   -1,  334,  335,  336,   -1,   -1,
   -1,  340,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  348,
  349,  350,  351,  352,  353,   -1,   -1,   -1,  357,  358,
   -1,  360,   -1,  362,   -1,  364,  365,  366,   -1,  368,
   -1,  370,  371,  372,   -1,  374,  375,  376,  377,  378,
   -1,  380,  381,   -1,   -1,   -1,   -1,  386,   -1,  388,
  389,   -1,   -1,  392,  393,  394,   -1,   -1,  397,   -1,
  399,   -1,   -1,  402,  403,  404,  405,   -1,  407,   -1,
  409,   -1,  411,  412,  413,  414,  415,   -1,   -1,  418,
  419,  420,   -1,  422,  423,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  433,  434,  435,   -1,  437,  438,
  439,   -1,   -1,   -1,   -1,   -1,  445,   -1,   -1,   -1,
  449,   -1,  451,  452,  453,  454,   -1,  456,   -1,  458,
  459,  460,  461,  462,  463,  464,  465,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  475,  476,   -1,  478,
  479,  480,   -1,   -1,   -1,   -1,  485,  486,  487,  488,
  489,   -1,  491,   -1,   -1,  494,  495,   -1,   -1,   -1,
  499,  500,   -1,   -1,  503,  504,  505,  506,  507,  508,
  509,  510,  511,   -1,  513,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  523,  524,  525,  526,  527,  528,
  529,  530,  531,  532,  533,  534,  535,  536,  537,  538,
  539,  540,  541,  542,  543,   -1,  545,  546,  547,  548,
  549,  550,   -1,  552,  553,  554,  555,   -1,  557,   -1,
  559,  560,  561,  562,  563,  564,  565,  566,  567,  568,
  569,  570,  571,  572,  573,  574,  575,  576,   -1,  578,
  579,  580,  581,   -1,  583,  584,  585,  586,  587,   -1,
  589,  590,  591,  592,  593,  594,  595,  596,  597,  598,
  599,  600,  601,  602,  603,  604,  605,  606,  607,  608,
   -1,  610,  611,  612,  613,  614,  615,  616,  617,   -1,
  619,  620,   -1,   -1,   -1,  624,  625,  626,  627,  628,
  629,   -1,   -1,  632,  633,  634,  635,  636,  637,  638,
  639,  640,   -1,   -1,   -1,   -1,   -1,  646,   -1,  648,
  649,  650,  651,  652,  653,   -1,  655,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  672,  673,   -1,  675,  676,  677,  678,
  679,  680,  681,  682,  683,   -1,  685,  686,  687,  688,
  689,  690,  691,  692,  693,  694,  695,  696,   -1,  698,
  699,  700,   -1,  702,   -1,    0,   -1,  706,  707,  708,
  709,  710,  711,  712,  713,  714,  715,  716,  717,  718,
   -1,  720,  721,   -1,   -1,   -1,   -1,   -1,  727,  728,
  729,  730,  731,  732,  733,  734,  735,  736,  737,  738,
  739,  740,  741,  742,  743,  744,   41,   -1,  747,   44,
  749,  750,  751,  752,  753,  754,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   59,   60,   61,   62,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   93,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  257,   -1,  259,   -1,   -1,  262,   -1,  264,
  265,   -1,   -1,  268,  269,   -1,  271,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  279,  280,   -1,  282,   -1,  284,
   -1,   -1,  287,   -1,   -1,  290,   -1,   -1,   -1,   -1,
   -1,   -1,  297,   -1,  299,  300,   -1,  302,  303,  304,
  305,  306,   -1,   -1,  309,   -1,   -1,  312,   -1,  314,
  315,   -1,   -1,  318,  319,  320,   -1,  322,  323,  324,
  325,  326,  327,  328,   -1,   -1,   -1,  332,   -1,  334,
  335,  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  348,  349,  350,  351,  352,  353,   -1,
   -1,   -1,  357,  358,   -1,  360,   -1,  362,   -1,  364,
  365,  366,   -1,  368,   -1,  370,  371,  372,   -1,  374,
  375,  376,  377,  378,   -1,  380,  381,   -1,   -1,   -1,
   -1,  386,   -1,  388,  389,   -1,   -1,  392,  393,  394,
   -1,   -1,  397,   -1,  399,   -1,   -1,  402,  403,  404,
  405,   -1,  407,   -1,  409,   -1,  411,  412,  413,  414,
  415,   -1,   -1,  418,  419,  420,   -1,  422,  423,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  433,  434,
  435,   -1,  437,  438,  439,   -1,   -1,   -1,   -1,   -1,
  445,   -1,   -1,   -1,  449,   -1,  451,  452,  453,  454,
   -1,  456,   -1,  458,  459,  460,  461,  462,  463,  464,
  465,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  475,  476,   -1,  478,  479,  480,   -1,   -1,   -1,   -1,
  485,  486,  487,  488,  489,   -1,  491,   -1,   -1,  494,
  495,   -1,   -1,   -1,  499,  500,   -1,   -1,  503,  504,
  505,  506,  507,  508,  509,  510,  511,   -1,  513,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  523,  524,
  525,  526,  527,  528,  529,  530,  531,  532,  533,  534,
  535,  536,  537,  538,  539,  540,  541,  542,  543,   -1,
  545,  546,  547,  548,  549,  550,   -1,  552,  553,  554,
  555,   -1,  557,   -1,  559,  560,  561,  562,  563,  564,
  565,  566,  567,  568,  569,  570,  571,  572,  573,  574,
  575,  576,   -1,  578,  579,  580,  581,   -1,  583,  584,
  585,  586,  587,   -1,  589,  590,  591,  592,  593,  594,
  595,  596,  597,  598,  599,  600,  601,  602,  603,  604,
  605,  606,  607,  608,   -1,  610,  611,  612,  613,  614,
  615,  616,  617,   -1,  619,  620,   -1,   -1,   -1,  624,
  625,  626,  627,  628,  629,   -1,   -1,  632,  633,  634,
  635,  636,  637,  638,  639,  640,   -1,   -1,   -1,   -1,
   -1,  646,   -1,  648,  649,  650,  651,  652,  653,   -1,
  655,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  672,  673,   -1,
  675,  676,  677,  678,  679,  680,  681,  682,  683,   -1,
  685,  686,  687,  688,  689,  690,  691,  692,  693,  694,
  695,  696,   -1,  698,  699,  700,   -1,  702,   -1,    0,
   -1,  706,  707,  708,  709,  710,  711,  712,  713,  714,
  715,  716,  717,  718,   -1,  720,  721,   -1,   -1,   -1,
   -1,   -1,  727,  728,  729,  730,  731,  732,  733,  734,
  735,  736,  737,  738,  739,  740,  741,  742,  743,  744,
   41,   -1,  747,   44,  749,  750,  751,  752,  753,  754,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   59,   60,
   61,   62,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   93,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  257,   -1,  259,   -1,
   -1,  262,   -1,  264,  265,   -1,   -1,  268,  269,   -1,
  271,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  279,  280,
   -1,  282,   -1,  284,   -1,   -1,  287,   -1,   -1,  290,
   -1,   -1,   -1,   -1,   -1,   -1,  297,   -1,  299,  300,
   -1,  302,  303,  304,  305,  306,   -1,   -1,  309,   -1,
   -1,  312,   -1,  314,  315,   -1,   -1,  318,  319,  320,
   -1,  322,  323,  324,  325,  326,  327,  328,   -1,   -1,
   -1,  332,   -1,  334,  335,  336,   -1,   -1,   -1,  340,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  348,  349,  350,
  351,  352,  353,   -1,   -1,   -1,  357,  358,   -1,  360,
   -1,  362,   -1,  364,  365,  366,   -1,  368,   -1,  370,
  371,  372,   -1,  374,  375,  376,  377,  378,   -1,  380,
  381,   -1,   -1,   -1,   -1,  386,   -1,  388,  389,   -1,
   -1,  392,  393,  394,   -1,   -1,  397,   -1,  399,   -1,
   -1,  402,  403,  404,  405,   -1,  407,   -1,  409,   -1,
  411,  412,  413,  414,  415,   -1,   -1,  418,  419,  420,
   -1,  422,  423,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  433,  434,  435,   -1,  437,  438,  439,   -1,
   -1,   -1,   -1,   -1,  445,   -1,   -1,   -1,  449,   -1,
  451,  452,  453,  454,   -1,  456,   -1,  458,  459,  460,
  461,  462,  463,  464,  465,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  475,  476,   -1,  478,  479,  480,
   -1,   -1,   -1,   -1,  485,  486,  487,  488,  489,   -1,
  491,   -1,   -1,  494,  495,   -1,   -1,   -1,  499,  500,
   -1,   -1,  503,  504,  505,  506,  507,  508,  509,  510,
  511,   -1,  513,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  523,  524,  525,  526,  527,  528,  529,  530,
  531,  532,  533,  534,  535,  536,  537,  538,  539,  540,
  541,  542,  543,   -1,  545,  546,  547,  548,  549,  550,
   -1,  552,  553,  554,  555,   -1,  557,   -1,  559,  560,
  561,  562,  563,  564,  565,  566,  567,  568,  569,  570,
  571,  572,  573,  574,  575,  576,   -1,  578,  579,  580,
  581,   -1,  583,  584,  585,  586,  587,   -1,  589,  590,
  591,  592,  593,  594,  595,  596,  597,  598,  599,  600,
  601,  602,  603,  604,  605,  606,  607,  608,   -1,  610,
  611,  612,  613,  614,  615,  616,  617,   -1,  619,  620,
   -1,   -1,   -1,  624,  625,  626,  627,  628,  629,   -1,
   -1,  632,  633,  634,  635,  636,  637,  638,  639,  640,
   -1,   -1,   -1,   -1,   -1,  646,   -1,  648,  649,  650,
  651,  652,  653,   -1,  655,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  672,  673,   -1,  675,  676,  677,  678,  679,  680,
  681,  682,  683,   -1,  685,  686,  687,  688,  689,  690,
  691,  692,  693,  694,  695,  696,   -1,  698,  699,  700,
   -1,  702,   -1,    0,   -1,  706,  707,  708,  709,  710,
  711,  712,  713,  714,  715,  716,  717,  718,   -1,  720,
  721,   -1,   -1,   -1,   -1,   -1,  727,  728,  729,  730,
  731,  732,  733,  734,  735,  736,  737,  738,  739,  740,
  741,  742,  743,  744,   41,   -1,  747,   44,  749,  750,
  751,  752,  753,  754,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   59,   60,   61,   62,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   93,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  257,   -1,  259,   -1,   -1,  262,   -1,  264,  265,   -1,
   -1,  268,  269,   -1,  271,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  279,  280,   -1,  282,   -1,  284,   -1,   -1,
  287,   -1,   -1,  290,   -1,   -1,   -1,   -1,   -1,   -1,
  297,   -1,  299,  300,   -1,  302,  303,  304,   -1,  306,
   -1,   -1,  309,   -1,   -1,  312,   -1,  314,  315,   -1,
   -1,  318,  319,  320,   -1,  322,  323,  324,  325,  326,
  327,  328,   -1,   -1,   -1,  332,   -1,  334,  335,  336,
   -1,   -1,   -1,  340,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  348,  349,  350,  351,  352,  353,   -1,   -1,   -1,
  357,  358,   -1,  360,   -1,  362,   -1,  364,  365,  366,
   -1,  368,   -1,  370,  371,  372,   -1,  374,  375,  376,
  377,  378,   -1,  380,  381,   -1,   -1,   -1,   -1,  386,
   -1,  388,  389,   -1,   -1,  392,  393,  394,   -1,   -1,
  397,   -1,  399,   -1,   -1,  402,  403,  404,  405,   -1,
  407,   -1,  409,   -1,  411,  412,  413,  414,  415,   -1,
   -1,  418,  419,  420,   -1,  422,  423,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  433,  434,  435,   -1,
  437,  438,  439,   -1,   -1,   -1,   -1,   -1,  445,   -1,
   -1,   -1,  449,   -1,  451,  452,  453,  454,   -1,  456,
   -1,  458,  459,  460,  461,  462,  463,  464,  465,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  475,  476,
   -1,  478,  479,  480,   -1,   -1,   -1,   -1,  485,  486,
  487,  488,  489,   -1,  491,   -1,   -1,  494,  495,   -1,
   -1,   -1,  499,  500,   -1,   -1,  503,  504,  505,  506,
  507,  508,  509,  510,  511,   -1,  513,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  523,  524,  525,  526,
  527,  528,  529,  530,  531,  532,  533,  534,  535,  536,
  537,  538,  539,  540,  541,  542,  543,   -1,  545,  546,
  547,  548,  549,  550,   -1,  552,  553,  554,  555,   -1,
  557,   -1,  559,  560,  561,  562,  563,  564,  565,  566,
  567,  568,  569,  570,  571,  572,  573,  574,  575,  576,
   -1,  578,  579,  580,  581,   -1,  583,  584,  585,  586,
  587,   -1,  589,  590,  591,  592,  593,  594,  595,  596,
  597,  598,  599,  600,  601,  602,  603,  604,  605,  606,
  607,  608,   -1,  610,  611,  612,  613,  614,  615,  616,
  617,   -1,  619,  620,   -1,   -1,   -1,  624,  625,  626,
  627,  628,  629,   -1,   -1,  632,  633,  634,  635,  636,
  637,  638,  639,  640,   -1,   -1,   -1,   -1,   -1,  646,
   -1,  648,  649,  650,  651,  652,  653,   -1,  655,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  672,  673,   -1,  675,  676,
  677,  678,  679,  680,  681,  682,  683,   -1,  685,  686,
  687,  688,  689,  690,  691,  692,  693,  694,  695,  696,
   -1,  698,  699,  700,   -1,  702,   -1,    0,   -1,  706,
  707,  708,  709,  710,  711,  712,  713,  714,  715,  716,
  717,  718,   -1,  720,  721,   -1,   -1,   -1,   -1,   -1,
  727,  728,  729,  730,  731,  732,  733,  734,  735,  736,
  737,  738,  739,  740,  741,  742,  743,  744,   41,   -1,
  747,   44,  749,  750,  751,  752,  753,  754,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   59,   60,   61,   62,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   93,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  257,   -1,  259,   -1,   -1,  262,
   -1,  264,  265,   -1,   -1,  268,  269,   -1,  271,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  279,  280,   -1,  282,
   -1,  284,   -1,   -1,  287,   -1,   -1,  290,   -1,   -1,
   -1,   -1,   -1,   -1,  297,   -1,  299,  300,   -1,  302,
  303,  304,   -1,  306,   -1,   -1,  309,   -1,   -1,  312,
   -1,  314,  315,   -1,   -1,  318,  319,  320,   -1,  322,
  323,  324,  325,  326,  327,  328,   -1,   -1,   -1,  332,
   -1,  334,  335,  336,   -1,   -1,   -1,  340,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  348,  349,  350,  351,  352,
  353,   -1,   -1,   -1,  357,  358,   -1,  360,   -1,  362,
   -1,  364,  365,  366,   -1,  368,   -1,  370,  371,  372,
   -1,  374,  375,  376,  377,  378,   -1,  380,  381,   -1,
   -1,   -1,   -1,  386,   -1,  388,  389,   -1,   -1,  392,
  393,  394,   -1,   -1,  397,   -1,  399,   -1,   -1,  402,
  403,  404,  405,   -1,  407,   -1,  409,   -1,  411,  412,
  413,  414,  415,   -1,   -1,  418,  419,  420,   -1,  422,
  423,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  433,  434,  435,   -1,  437,  438,  439,   -1,   -1,   -1,
   -1,   -1,  445,   -1,   -1,   -1,  449,   -1,  451,  452,
  453,  454,   -1,  456,   -1,  458,  459,  460,  461,  462,
  463,  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  475,  476,   -1,  478,  479,  480,   -1,   -1,
   -1,   -1,  485,  486,  487,  488,  489,   -1,  491,   -1,
   -1,  494,  495,   -1,   -1,   -1,  499,  500,   -1,   -1,
  503,  504,  505,  506,  507,  508,  509,  510,  511,   -1,
  513,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  523,  524,  525,  526,  527,  528,  529,  530,  531,  532,
  533,  534,  535,  536,  537,  538,  539,  540,  541,  542,
  543,   -1,  545,  546,  547,  548,  549,  550,   -1,  552,
  553,  554,  555,   -1,  557,   -1,  559,  560,  561,  562,
  563,  564,  565,  566,  567,  568,  569,  570,  571,  572,
  573,  574,  575,  576,   -1,  578,  579,  580,  581,   -1,
  583,  584,  585,  586,  587,   -1,  589,  590,  591,  592,
  593,  594,  595,  596,  597,  598,  599,  600,  601,  602,
  603,  604,  605,  606,  607,  608,   -1,  610,  611,  612,
  613,  614,  615,  616,  617,   -1,  619,  620,   -1,   -1,
   -1,  624,  625,  626,  627,  628,  629,   -1,   -1,  632,
  633,  634,  635,  636,  637,  638,  639,  640,   -1,   -1,
   -1,   -1,   -1,  646,   -1,  648,  649,  650,  651,  652,
  653,   -1,  655,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  672,
  673,   -1,  675,  676,  677,  678,  679,  680,  681,  682,
  683,   -1,  685,  686,  687,  688,  689,  690,  691,  692,
  693,  694,  695,  696,   -1,  698,  699,  700,   -1,  702,
   -1,    0,   -1,  706,  707,  708,  709,  710,  711,  712,
  713,  714,  715,  716,  717,  718,   -1,  720,  721,   -1,
   -1,   -1,   -1,   -1,  727,  728,  729,  730,  731,  732,
  733,  734,  735,  736,  737,  738,  739,  740,  741,  742,
  743,  744,   41,   -1,  747,   44,  749,  750,  751,  752,
  753,  754,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   59,   60,   61,   62,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   93,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  257,   -1,
  259,   -1,   -1,  262,   -1,  264,  265,   -1,   -1,  268,
  269,   -1,  271,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  279,  280,   -1,  282,   -1,  284,   -1,   -1,  287,   -1,
   -1,  290,   -1,   -1,   -1,   -1,   -1,   -1,  297,   -1,
  299,  300,   -1,  302,  303,  304,   -1,  306,   -1,   -1,
  309,   -1,   -1,  312,   -1,  314,  315,   -1,   -1,  318,
  319,  320,   -1,  322,  323,  324,  325,  326,  327,  328,
   -1,   -1,   -1,  332,   -1,  334,  335,  336,   -1,   -1,
   -1,  340,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  348,
  349,  350,  351,  352,  353,   -1,   -1,   -1,  357,  358,
   -1,  360,   -1,  362,   -1,  364,  365,  366,   -1,  368,
   -1,  370,  371,  372,   -1,  374,  375,  376,  377,  378,
   -1,  380,  381,   -1,   -1,   -1,   -1,  386,   -1,  388,
  389,   -1,   -1,  392,  393,  394,   -1,   -1,  397,   -1,
  399,   -1,   -1,  402,  403,  404,  405,   -1,  407,   -1,
  409,   -1,  411,  412,  413,  414,  415,   -1,   -1,  418,
  419,  420,   -1,  422,  423,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  433,  434,  435,   -1,  437,  438,
  439,   -1,   -1,   -1,   -1,   -1,  445,   -1,   -1,   -1,
  449,   -1,  451,  452,  453,  454,   -1,  456,   -1,  458,
  459,  460,  461,  462,  463,  464,  465,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  475,  476,   -1,  478,
  479,  480,   -1,   -1,   -1,   -1,  485,  486,  487,  488,
  489,   -1,  491,   -1,   -1,  494,  495,   -1,   -1,   -1,
  499,  500,   -1,   -1,  503,  504,  505,  506,  507,  508,
  509,  510,  511,   -1,  513,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  523,  524,  525,  526,  527,  528,
  529,  530,  531,  532,  533,  534,  535,  536,  537,  538,
  539,  540,  541,  542,  543,   -1,  545,  546,  547,  548,
  549,  550,   -1,  552,  553,  554,  555,   -1,  557,   -1,
  559,  560,  561,  562,  563,  564,  565,  566,  567,  568,
  569,  570,  571,  572,  573,  574,  575,  576,   -1,  578,
  579,  580,  581,   -1,  583,  584,  585,  586,  587,   -1,
  589,  590,  591,  592,  593,  594,  595,  596,  597,  598,
  599,  600,  601,  602,  603,  604,  605,  606,  607,  608,
   -1,  610,  611,  612,  613,  614,  615,  616,  617,   -1,
  619,  620,   -1,   -1,   -1,  624,  625,  626,  627,  628,
  629,   -1,   -1,  632,  633,  634,  635,  636,  637,  638,
  639,  640,   -1,   -1,   -1,   -1,   -1,  646,   -1,  648,
  649,  650,  651,  652,  653,   -1,  655,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  672,  673,   -1,  675,  676,  677,  678,
  679,  680,  681,  682,  683,   -1,  685,  686,  687,  688,
  689,  690,  691,  692,  693,  694,  695,  696,   -1,  698,
  699,  700,   -1,  702,   -1,    0,   -1,  706,  707,  708,
  709,  710,  711,  712,  713,  714,  715,  716,  717,  718,
   -1,  720,  721,   -1,   -1,   -1,   -1,   -1,  727,  728,
  729,  730,  731,  732,  733,  734,  735,  736,  737,  738,
  739,  740,  741,  742,  743,  744,   41,   -1,  747,   44,
  749,  750,  751,  752,  753,  754,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   59,   60,   61,   62,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   93,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  257,   -1,  259,   -1,   -1,  262,   -1,  264,
  265,   -1,   -1,  268,  269,   -1,  271,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  279,  280,   -1,  282,   -1,  284,
   -1,   -1,  287,   -1,   -1,  290,   -1,   -1,   -1,   -1,
   -1,   -1,  297,   -1,  299,  300,   -1,  302,  303,  304,
   -1,  306,   -1,   -1,  309,   -1,   -1,  312,   -1,  314,
  315,   -1,   -1,  318,  319,  320,   -1,  322,  323,  324,
  325,  326,  327,  328,   -1,   -1,   -1,  332,   -1,  334,
  335,  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  348,  349,  350,  351,  352,  353,   -1,
   -1,   -1,  357,  358,   -1,  360,   -1,  362,   -1,  364,
  365,  366,   -1,  368,   -1,  370,  371,  372,   -1,  374,
  375,  376,  377,  378,   -1,  380,  381,   -1,   -1,   -1,
   -1,  386,   -1,  388,  389,   -1,   -1,  392,  393,  394,
   -1,   -1,  397,   -1,  399,   -1,   -1,  402,  403,  404,
  405,   -1,  407,   -1,  409,   -1,  411,  412,  413,  414,
  415,   -1,   -1,  418,  419,  420,   -1,  422,  423,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  433,  434,
  435,   -1,  437,  438,  439,   -1,   -1,   -1,   -1,   -1,
  445,   -1,   -1,   -1,  449,   -1,  451,  452,  453,  454,
   -1,  456,   -1,  458,  459,  460,  461,  462,  463,  464,
  465,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  475,  476,   -1,  478,  479,  480,   -1,   -1,   -1,   -1,
  485,  486,  487,  488,  489,   -1,  491,   -1,   -1,  494,
  495,   -1,   -1,   -1,  499,  500,   -1,   -1,  503,  504,
  505,  506,  507,  508,  509,  510,  511,   -1,  513,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  523,  524,
  525,  526,  527,  528,  529,  530,  531,  532,  533,  534,
  535,  536,  537,  538,  539,  540,  541,  542,  543,   -1,
  545,  546,  547,  548,  549,  550,   -1,  552,  553,  554,
  555,   -1,  557,   -1,  559,  560,  561,  562,  563,  564,
  565,  566,  567,  568,  569,  570,  571,  572,  573,  574,
  575,  576,   -1,  578,  579,  580,  581,   -1,  583,  584,
  585,  586,  587,   -1,  589,  590,  591,  592,  593,  594,
  595,  596,  597,  598,  599,  600,  601,  602,  603,  604,
  605,  606,  607,  608,   -1,  610,  611,  612,  613,  614,
  615,  616,  617,   -1,  619,  620,   -1,   -1,   -1,  624,
  625,  626,  627,  628,  629,   -1,   -1,  632,  633,  634,
  635,  636,  637,  638,  639,  640,   -1,   -1,   -1,   -1,
   -1,  646,   -1,  648,  649,  650,  651,  652,  653,   -1,
  655,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  672,  673,   -1,
  675,  676,  677,  678,  679,  680,  681,  682,  683,   -1,
  685,  686,  687,  688,  689,  690,  691,  692,  693,  694,
  695,  696,   -1,  698,  699,  700,   -1,  702,   -1,    0,
   -1,  706,  707,  708,  709,  710,  711,  712,  713,  714,
  715,  716,  717,  718,   -1,  720,  721,   -1,   -1,   -1,
   -1,   -1,  727,  728,  729,  730,  731,  732,  733,  734,
  735,  736,  737,  738,  739,  740,  741,  742,  743,  744,
   41,   -1,  747,   44,  749,  750,  751,  752,  753,  754,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   59,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   93,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  257,   -1,  259,   -1,
   -1,  262,   -1,  264,  265,   -1,   -1,  268,  269,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  279,  280,
   -1,  282,   -1,   -1,   -1,   -1,  287,   -1,   -1,  290,
   -1,   -1,   -1,   -1,   -1,   -1,  297,   -1,  299,  300,
   -1,  302,  303,  304,  305,  306,   -1,   -1,  309,   -1,
   -1,  312,   -1,  314,  315,   -1,   -1,   -1,  319,  320,
   -1,  322,  323,  324,   -1,  326,  327,  328,   -1,   -1,
   -1,  332,   -1,  334,  335,  336,   -1,   -1,   -1,  340,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  348,  349,   -1,
  351,   -1,  353,   -1,   -1,   -1,  357,  358,   -1,  360,
   -1,   -1,   -1,  364,   -1,   -1,   -1,  368,   -1,  370,
  371,  372,   -1,  374,  375,  376,  377,  378,   -1,  380,
  381,   -1,   -1,   -1,   -1,  386,   -1,  388,  389,   -1,
   -1,  392,  393,  394,   -1,   -1,  397,   -1,  399,   -1,
   -1,  402,  403,  404,  405,   -1,  407,   -1,  409,   -1,
  411,   -1,  413,  414,  415,   -1,   -1,  418,  419,  420,
   -1,  422,  423,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  433,  434,  435,   -1,  437,  438,  439,   -1,
   -1,   -1,   -1,   -1,  445,   -1,   -1,   -1,  449,   -1,
  451,  452,  453,  454,   -1,  456,   -1,  458,  459,  460,
  461,  462,  463,  464,  465,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  475,  476,   -1,  478,  479,  480,
   -1,   -1,   -1,   -1,  485,  486,  487,  488,  489,   -1,
  491,   -1,   -1,  494,  495,   -1,   -1,   -1,  499,  500,
   -1,   -1,  503,  504,  505,  506,  507,  508,  509,  510,
  511,   -1,  513,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  523,  524,  525,  526,  527,  528,  529,  530,
  531,  532,  533,  534,  535,  536,  537,  538,  539,  540,
  541,  542,  543,   -1,  545,  546,  547,  548,  549,  550,
   -1,  552,  553,  554,  555,   -1,  557,   -1,  559,  560,
  561,  562,  563,  564,  565,  566,  567,  568,  569,  570,
  571,  572,  573,  574,  575,  576,   -1,  578,  579,  580,
  581,   -1,  583,  584,  585,  586,  587,   -1,  589,  590,
  591,  592,  593,  594,  595,  596,  597,  598,  599,  600,
  601,   -1,  603,  604,  605,  606,  607,  608,   -1,  610,
  611,  612,  613,  614,  615,  616,  617,   -1,  619,  620,
   -1,   -1,   -1,  624,  625,  626,  627,  628,  629,   -1,
   -1,  632,  633,  634,  635,  636,  637,  638,  639,  640,
   -1,   -1,   -1,   -1,   -1,  646,   -1,  648,  649,  650,
  651,  652,  653,   -1,  655,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  672,  673,   -1,  675,  676,  677,  678,  679,  680,
  681,  682,  683,   -1,  685,  686,  687,  688,  689,  690,
  691,  692,  693,  694,  695,  696,   -1,  698,  699,  700,
   -1,  702,   -1,    0,   -1,  706,  707,  708,  709,  710,
  711,  712,  713,  714,  715,  716,  717,  718,   -1,  720,
  721,   -1,   -1,   -1,   -1,   -1,  727,  728,  729,  730,
  731,  732,  733,  734,  735,  736,  737,  738,  739,  740,
  741,  742,  743,  744,   41,   -1,  747,   44,  749,  750,
  751,  752,  753,  754,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   59,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   93,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  257,   -1,  259,   -1,   -1,  262,   -1,  264,  265,   -1,
   -1,  268,  269,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  279,  280,   -1,  282,   -1,   -1,   -1,   -1,
  287,   -1,   -1,  290,   -1,   -1,   -1,   -1,   -1,   -1,
  297,   -1,  299,  300,   -1,  302,  303,  304,  305,  306,
   -1,   -1,  309,   -1,   -1,  312,   -1,  314,  315,   -1,
   -1,   -1,  319,  320,   -1,  322,  323,  324,   -1,  326,
  327,  328,   -1,   -1,   -1,  332,   -1,  334,  335,  336,
   -1,   -1,   -1,  340,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  348,  349,   -1,  351,   -1,  353,   -1,   -1,   -1,
  357,  358,   -1,  360,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  368,   -1,  370,  371,  372,   -1,  374,  375,  376,
  377,  378,   -1,  380,  381,   -1,   -1,   -1,   -1,  386,
   -1,  388,  389,   -1,   -1,  392,  393,  394,   -1,   -1,
  397,   -1,  399,   -1,   -1,  402,  403,  404,  405,   -1,
  407,   -1,  409,   -1,  411,   -1,  413,  414,  415,   -1,
   -1,  418,  419,  420,   -1,  422,  423,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  433,  434,  435,   -1,
  437,  438,  439,   -1,   -1,   -1,   -1,   -1,  445,   -1,
   -1,   -1,  449,   -1,  451,  452,  453,  454,   -1,  456,
   -1,  458,  459,  460,  461,  462,  463,  464,  465,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  475,  476,
   -1,  478,  479,  480,   -1,   -1,   -1,   -1,  485,  486,
  487,  488,  489,   -1,  491,   -1,   -1,  494,  495,   -1,
   -1,   -1,  499,  500,   -1,   -1,  503,  504,  505,  506,
  507,  508,  509,  510,  511,   -1,  513,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  523,  524,  525,  526,
  527,  528,  529,  530,  531,  532,  533,  534,  535,  536,
  537,  538,  539,  540,  541,  542,  543,   -1,  545,  546,
  547,  548,  549,  550,   -1,  552,  553,  554,  555,   -1,
  557,   -1,  559,  560,  561,  562,  563,  564,  565,  566,
  567,  568,  569,  570,  571,  572,  573,  574,  575,  576,
   -1,  578,  579,  580,  581,   -1,  583,  584,  585,  586,
  587,   -1,  589,  590,  591,  592,  593,  594,  595,  596,
  597,  598,  599,  600,  601,   -1,  603,  604,  605,  606,
  607,  608,   -1,  610,  611,  612,  613,  614,  615,  616,
  617,   -1,  619,  620,   -1,   -1,   -1,  624,  625,  626,
  627,  628,  629,   -1,   -1,  632,  633,  634,  635,  636,
  637,  638,  639,  640,   -1,   -1,   -1,   -1,   -1,  646,
   -1,  648,  649,  650,  651,  652,  653,   -1,  655,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  672,  673,   -1,  675,  676,
  677,  678,  679,  680,  681,  682,  683,   -1,  685,  686,
  687,  688,  689,  690,  691,  692,  693,  694,  695,  696,
   -1,  698,  699,  700,   -1,  702,   -1,    0,   -1,  706,
  707,  708,  709,  710,  711,  712,  713,  714,  715,  716,
  717,  718,   -1,  720,  721,   -1,   -1,   -1,   -1,   -1,
  727,  728,  729,  730,  731,  732,  733,  734,  735,  736,
  737,  738,  739,  740,  741,  742,  743,  744,   41,   -1,
  747,   44,  749,  750,  751,  752,  753,  754,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   59,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   93,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  257,   -1,  259,   -1,   -1,   -1,
   -1,  264,  265,   -1,   -1,  268,  269,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  279,  280,   -1,  282,
   -1,   -1,   -1,   -1,  287,   -1,   -1,  290,   -1,   -1,
   -1,   -1,   -1,   -1,  297,   -1,  299,  300,   -1,  302,
  303,  304,  305,  306,   -1,   -1,  309,   -1,   -1,  312,
   -1,  314,  315,   -1,   -1,   -1,  319,  320,   -1,  322,
  323,  324,   -1,  326,  327,  328,   -1,   -1,   -1,  332,
   -1,  334,  335,  336,   -1,   -1,   -1,  340,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  348,  349,   -1,  351,   -1,
  353,   -1,   -1,   -1,  357,  358,   -1,  360,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  368,   -1,  370,  371,  372,
   -1,  374,  375,  376,  377,  378,   -1,  380,  381,   -1,
   -1,   -1,   -1,  386,   -1,  388,  389,   -1,   -1,  392,
  393,  394,   -1,   -1,  397,   -1,  399,   -1,   -1,  402,
  403,  404,  405,   -1,  407,   -1,  409,   -1,  411,   -1,
  413,  414,  415,   -1,   -1,  418,  419,  420,   -1,  422,
  423,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  433,  434,  435,   -1,  437,  438,  439,   -1,   -1,   -1,
   -1,   -1,  445,   -1,   -1,   -1,  449,   -1,  451,  452,
  453,  454,   -1,  456,   -1,  458,  459,  460,  461,  462,
  463,  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  475,  476,   -1,  478,  479,  480,   -1,   -1,
   -1,   -1,  485,  486,  487,  488,  489,   -1,  491,   -1,
   -1,  494,  495,   -1,   -1,   -1,  499,  500,   -1,   -1,
  503,  504,  505,  506,  507,  508,  509,  510,  511,   -1,
  513,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  523,  524,  525,  526,  527,  528,  529,  530,  531,  532,
  533,  534,  535,  536,  537,  538,  539,  540,  541,  542,
  543,   -1,  545,  546,  547,  548,  549,  550,   -1,  552,
  553,  554,  555,   -1,  557,   -1,  559,  560,  561,  562,
  563,  564,  565,  566,  567,  568,  569,  570,  571,  572,
  573,  574,  575,  576,   -1,  578,  579,  580,  581,   -1,
  583,  584,  585,  586,  587,   -1,  589,  590,  591,  592,
  593,  594,  595,  596,  597,  598,  599,  600,  601,   -1,
  603,  604,  605,  606,  607,  608,   -1,  610,  611,  612,
  613,  614,  615,  616,  617,   -1,  619,  620,   -1,   -1,
   -1,  624,  625,  626,  627,  628,  629,   -1,   -1,  632,
  633,  634,  635,  636,  637,  638,  639,  640,   -1,   -1,
   -1,   -1,   -1,  646,   -1,  648,  649,  650,  651,  652,
  653,   -1,  655,   -1,   -1,   -1,   -1,    0,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  672,
  673,   -1,  675,  676,  677,  678,  679,  680,  681,  682,
  683,   -1,  685,  686,  687,  688,  689,  690,  691,  692,
  693,  694,  695,  696,   -1,  698,  699,  700,   41,  702,
   -1,   44,   -1,  706,  707,  708,  709,  710,  711,  712,
  713,  714,  715,  716,  717,  718,   59,  720,  721,   -1,
   -1,   -1,   -1,   -1,  727,  728,  729,  730,  731,  732,
  733,  734,  735,  736,  737,  738,  739,  740,  741,  742,
  743,  744,   -1,   -1,  747,   -1,  749,  750,  751,  752,
  753,  754,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  257,   -1,  259,   -1,   -1,   -1,
   -1,  264,  265,   -1,   -1,  268,  269,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  279,  280,   -1,  282,
   -1,  284,   -1,   -1,  287,   -1,   -1,  290,   -1,   -1,
   -1,   -1,   -1,   -1,  297,   -1,  299,  300,   -1,   -1,
   -1,  304,   -1,  306,   -1,   -1,  309,   -1,   -1,  312,
   -1,   -1,  315,   -1,   -1,   -1,  319,  320,   -1,  322,
  323,  324,   -1,  326,  327,  328,   -1,   -1,   -1,  332,
   -1,  334,  335,  336,   -1,   -1,   -1,  340,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  348,  349,   -1,  351,   -1,
  353,   -1,   -1,   -1,  357,  358,   -1,  360,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  368,   -1,  370,   -1,  372,
   -1,  374,  375,  376,  377,  378,   -1,  380,  381,   -1,
   -1,   -1,   -1,  386,   -1,  388,  389,   -1,   -1,  392,
  393,   -1,   -1,   -1,  397,   -1,  399,   -1,   -1,  402,
  403,  404,  405,   -1,  407,   -1,  409,   -1,  411,  412,
  413,  414,  415,   -1,   -1,   -1,   -1,  420,   -1,  422,
  423,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  433,   -1,  435,   -1,  437,  438,  439,   -1,   -1,   -1,
   -1,   -1,  445,   -1,   -1,   -1,  449,   -1,  451,  452,
  453,  454,   -1,  456,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  475,  476,   -1,  478,  479,  480,   -1,   -1,
   -1,   -1,  485,  486,  487,  488,  489,   -1,  491,   -1,
   -1,  494,  495,   -1,   -1,   -1,  499,  500,   -1,   -1,
  503,  504,  505,  506,  507,  508,  509,  510,  511,   -1,
  513,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  523,  524,  525,  526,  527,  528,  529,  530,  531,  532,
  533,  534,  535,  536,  537,  538,  539,  540,  541,  542,
  543,   -1,  545,  546,  547,  548,  549,  550,   -1,  552,
  553,  554,  555,   -1,  557,   -1,  559,  560,  561,  562,
  563,  564,  565,  566,  567,  568,  569,  570,  571,  572,
  573,  574,  575,  576,   -1,  578,  579,  580,  581,   -1,
  583,  584,  585,  586,  587,   -1,  589,  590,  591,  592,
  593,  594,  595,  596,  597,  598,  599,  600,  601,   -1,
  603,  604,  605,  606,  607,  608,   -1,  610,  611,  612,
  613,  614,  615,  616,  617,   -1,  619,  620,   -1,   -1,
   -1,  624,  625,  626,  627,  628,  629,   -1,   -1,  632,
  633,  634,  635,  636,  637,  638,  639,  640,   -1,   -1,
   -1,   -1,   -1,  646,   -1,  648,  649,  650,  651,  652,
  653,   -1,  655,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  672,
  673,   -1,  675,  676,  677,  678,  679,  680,  681,  682,
  683,   -1,  685,  686,  687,  688,  689,  690,  691,  692,
  693,  694,  695,  696,   -1,  698,  699,  700,    0,  702,
   -1,   -1,   -1,  706,  707,  708,  709,  710,  711,  712,
  713,  714,  715,  716,  717,  718,   -1,  720,  721,   -1,
   -1,   -1,   -1,   -1,  727,  728,  729,  730,  731,  732,
  733,  734,  735,  736,  737,  738,  739,  740,   40,   41,
  743,  744,   44,   -1,  747,   -1,  749,  750,  751,  752,
  753,  754,   -1,   -1,   -1,   -1,   -1,   59,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  257,   -1,  259,   -1,   -1,
   -1,   -1,   -1,  265,   -1,   -1,  268,  269,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  279,  280,   -1,
  282,   -1,  284,   -1,   -1,  287,   -1,   -1,  290,   -1,
   -1,   -1,   -1,   -1,   -1,  297,   -1,  299,  300,   -1,
   -1,   -1,  304,   -1,  306,   -1,   -1,  309,   -1,   -1,
  312,   -1,   -1,  315,   -1,   -1,   -1,  319,  320,   -1,
  322,  323,  324,   -1,  326,  327,  328,   -1,   -1,   -1,
  332,   -1,  334,  335,  336,   -1,   -1,   -1,  340,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  348,  349,   -1,  351,
   -1,  353,   -1,   -1,   -1,  357,  358,   -1,  360,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  368,   -1,  370,   -1,
  372,   -1,  374,  375,  376,  377,  378,   -1,  380,  381,
   -1,   -1,   -1,   -1,  386,   -1,  388,  389,   -1,   -1,
  392,  393,   -1,   -1,   -1,  397,   -1,  399,   -1,   -1,
  402,  403,  404,  405,   -1,  407,   -1,  409,   -1,  411,
  412,  413,  414,  415,   -1,   -1,   -1,   -1,  420,   -1,
  422,  423,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  433,   -1,  435,   -1,  437,  438,  439,   -1,   -1,
   -1,   -1,   -1,  445,   -1,   -1,   -1,  449,   -1,  451,
  452,  453,  454,   -1,  456,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  475,  476,   -1,  478,  479,  480,   -1,
   -1,   -1,   -1,  485,  486,  487,  488,  489,   -1,  491,
   -1,   -1,  494,  495,   -1,   -1,   -1,  499,  500,   -1,
   -1,  503,  504,  505,  506,  507,  508,  509,  510,  511,
   -1,  513,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  523,  524,  525,  526,  527,  528,  529,  530,  531,
  532,  533,  534,  535,  536,  537,  538,  539,  540,  541,
  542,  543,   -1,  545,  546,  547,  548,  549,  550,   -1,
  552,  553,  554,  555,   -1,  557,   -1,  559,  560,  561,
  562,  563,  564,  565,  566,  567,  568,  569,  570,  571,
  572,  573,  574,  575,  576,   -1,  578,  579,  580,  581,
   -1,  583,  584,  585,  586,  587,   -1,  589,  590,  591,
  592,  593,  594,  595,  596,  597,  598,  599,  600,  601,
   -1,  603,  604,  605,  606,  607,  608,   -1,  610,  611,
  612,  613,  614,  615,  616,  617,   -1,  619,  620,   -1,
   -1,   -1,  624,  625,  626,  627,  628,  629,   -1,   -1,
  632,  633,  634,  635,  636,  637,  638,  639,  640,   -1,
   -1,   -1,   -1,   -1,  646,   -1,  648,  649,  650,  651,
  652,  653,   -1,  655,   -1,   -1,   -1,   -1,    0,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  672,  673,   -1,  675,  676,  677,  678,  679,  680,  681,
  682,  683,   -1,  685,  686,  687,  688,  689,  690,  691,
  692,  693,  694,  695,  696,   -1,  698,  699,  700,   41,
  702,   -1,   44,   -1,  706,  707,  708,  709,  710,  711,
  712,  713,  714,  715,  716,  717,  718,   59,  720,  721,
   -1,   -1,   -1,   -1,   -1,  727,  728,  729,  730,  731,
  732,  733,  734,  735,  736,  737,  738,  739,  740,   -1,
   -1,  743,  744,   -1,   -1,  747,   -1,  749,  750,  751,
  752,  753,  754,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  257,   -1,  259,   -1,   -1,
   -1,   -1,   -1,  265,   -1,   -1,  268,  269,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  279,  280,   -1,
  282,   -1,  284,   -1,   -1,  287,   -1,   -1,  290,   -1,
   -1,   -1,   -1,   -1,   -1,  297,   -1,  299,  300,   -1,
   -1,   -1,  304,   -1,  306,   -1,   -1,  309,   -1,   -1,
  312,   -1,   -1,  315,   -1,   -1,   -1,  319,  320,   -1,
  322,  323,  324,   -1,  326,  327,  328,   -1,   -1,   -1,
  332,   -1,  334,  335,  336,   -1,   -1,   -1,  340,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  348,  349,   -1,  351,
   -1,  353,   -1,   -1,   -1,  357,  358,   -1,  360,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  368,   -1,  370,   -1,
  372,   -1,  374,  375,  376,  377,  378,   -1,  380,  381,
   -1,   -1,   -1,   -1,  386,   -1,  388,  389,   -1,   -1,
  392,  393,   -1,   -1,   -1,  397,   -1,  399,   -1,   -1,
  402,  403,  404,  405,   -1,  407,   -1,  409,   -1,  411,
  412,  413,  414,  415,   -1,   -1,   -1,   -1,  420,   -1,
  422,  423,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  433,   -1,  435,   -1,  437,  438,  439,   -1,   -1,
   -1,   -1,   -1,  445,   -1,   -1,   -1,  449,   -1,  451,
  452,  453,  454,   -1,  456,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  475,  476,   -1,  478,  479,  480,   -1,
   -1,   -1,   -1,  485,  486,  487,  488,  489,   -1,  491,
   -1,   -1,  494,  495,   -1,   -1,   -1,  499,  500,   -1,
   -1,  503,  504,  505,  506,  507,  508,  509,  510,  511,
   -1,  513,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  524,  525,  526,  527,  528,  529,  530,  531,
  532,  533,  534,  535,  536,  537,  538,  539,  540,  541,
  542,  543,   -1,  545,  546,  547,  548,  549,  550,   -1,
  552,  553,  554,  555,   -1,  557,   -1,  559,  560,  561,
  562,  563,  564,  565,  566,  567,  568,  569,  570,  571,
  572,  573,  574,  575,  576,   -1,  578,  579,  580,  581,
   -1,  583,  584,  585,  586,  587,   -1,  589,  590,  591,
  592,  593,  594,  595,  596,  597,  598,  599,  600,  601,
   -1,  603,  604,  605,  606,  607,  608,   -1,  610,  611,
  612,  613,  614,  615,  616,  617,   -1,  619,  620,   -1,
   -1,   -1,  624,  625,  626,  627,  628,  629,   -1,   -1,
  632,  633,  634,  635,  636,  637,  638,  639,  640,   -1,
   -1,   -1,   -1,   -1,  646,   -1,  648,  649,  650,  651,
  652,  653,   -1,  655,   -1,   -1,   -1,   -1,    0,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  672,  673,   -1,  675,  676,  677,  678,  679,  680,  681,
  682,  683,   -1,  685,  686,  687,  688,  689,  690,  691,
  692,  693,  694,  695,  696,   -1,  698,  699,  700,   41,
  702,   -1,   44,   -1,  706,  707,  708,  709,  710,  711,
  712,  713,  714,  715,  716,  717,  718,   59,  720,  721,
   -1,   -1,   -1,   -1,   -1,  727,  728,  729,  730,  731,
  732,  733,  734,  735,  736,  737,  738,  739,  740,   -1,
   -1,  743,  744,   -1,   -1,  747,   -1,  749,  750,  751,
  752,  753,  754,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  257,   -1,  259,   -1,   -1,
   -1,   -1,   -1,  265,   -1,   -1,  268,  269,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  279,  280,   -1,
  282,   -1,  284,   -1,   -1,  287,   -1,   -1,  290,   -1,
   -1,   -1,   -1,   -1,   -1,  297,   -1,  299,  300,   -1,
   -1,   -1,  304,   -1,  306,   -1,   -1,  309,   -1,   -1,
  312,   -1,   -1,  315,   -1,   -1,   -1,  319,  320,   -1,
  322,  323,  324,   -1,  326,  327,  328,   -1,   -1,   -1,
  332,   -1,  334,  335,  336,   -1,   -1,   -1,  340,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  348,  349,   -1,  351,
   -1,  353,   -1,   -1,   -1,  357,  358,   -1,  360,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  368,   -1,  370,   -1,
  372,   -1,  374,  375,  376,  377,  378,   -1,  380,  381,
   -1,   -1,   -1,   -1,  386,   -1,  388,  389,   -1,   -1,
  392,  393,   -1,   -1,   -1,  397,   -1,  399,   -1,   -1,
  402,  403,  404,  405,   -1,  407,   -1,  409,   -1,  411,
  412,  413,  414,  415,   -1,   -1,   -1,   -1,  420,   -1,
  422,  423,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  433,   -1,  435,   -1,  437,  438,  439,   -1,   -1,
   -1,   -1,   -1,  445,   -1,   -1,   -1,  449,   -1,  451,
  452,  453,  454,   -1,  456,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  475,  476,   -1,  478,  479,  480,   -1,
   -1,   -1,   -1,  485,  486,  487,  488,  489,   -1,  491,
   -1,   -1,  494,  495,   -1,   -1,   -1,  499,  500,   -1,
   -1,  503,  504,  505,  506,  507,  508,  509,  510,  511,
   -1,  513,   -1,   -1,   -1,   -1,    0,   -1,   -1,   -1,
   -1,   -1,  524,  525,  526,  527,  528,  529,  530,  531,
  532,  533,  534,  535,  536,  537,  538,  539,  540,  541,
  542,  543,   -1,  545,  546,  547,  548,  549,  550,   -1,
  552,  553,  554,  555,   -1,  557,   -1,  559,  560,  561,
  562,  563,  564,  565,  566,  567,  568,  569,  570,  571,
  572,  573,  574,  575,  576,   59,  578,  579,  580,  581,
   -1,  583,  584,  585,  586,  587,   -1,  589,  590,  591,
  592,  593,  594,  595,  596,  597,  598,  599,  600,  601,
   -1,  603,  604,  605,  606,  607,  608,   -1,  610,  611,
  612,  613,  614,  615,  616,  617,   -1,  619,  620,   -1,
   -1,   -1,  624,  625,  626,  627,  628,  629,   -1,   -1,
  632,  633,  634,  635,  636,  637,  638,  639,  640,   -1,
   -1,   -1,   -1,   -1,  646,   -1,  648,  649,  650,  651,
  652,  653,   -1,  655,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  672,  673,   -1,  675,  676,  677,  678,  679,  680,  681,
  682,  683,   -1,  685,  686,  687,  688,  689,  690,  691,
  692,  693,  694,  695,  696,   -1,  698,  699,  700,   -1,
  702,   -1,   -1,   -1,  706,  707,  708,  709,  710,  711,
  712,  713,  714,  715,  716,  717,  718,   -1,  720,  721,
   -1,   -1,   -1,   -1,   -1,  727,  728,  729,  730,  731,
  732,  733,  734,  735,  736,  737,  738,  739,  740,   -1,
   -1,  743,  744,   -1,   -1,  747,   -1,  749,  750,  751,
  752,  753,  754,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  257,   -1,  259,   -1,   -1,   -1,   -1,
   -1,  265,   -1,   -1,  268,  269,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  279,  280,   -1,  282,   -1,
  284,   -1,   -1,  287,   -1,   -1,  290,   -1,   -1,   -1,
   -1,   -1,   -1,  297,   -1,  299,  300,   -1,   -1,   -1,
  304,   -1,  306,   -1,   -1,  309,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  319,  320,   -1,   -1,   -1,
  324,   -1,  326,   -1,  328,   -1,   -1,   -1,   -1,   -1,
  334,   -1,  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  348,   -1,   -1,  351,   -1,  353,
   -1,   -1,   -1,  357,  358,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  370,   -1,  372,   -1,
  374,  375,  376,  377,  378,   -1,  380,  381,   -1,   -1,
   -1,   -1,  386,   -1,  388,  389,   -1,   -1,  392,  393,
   -1,   -1,   -1,   -1,   -1,  399,   -1,  401,  402,  403,
  404,  405,   -1,  407,   -1,  409,   -1,  411,  412,  413,
  414,  415,   -1,   -1,   -1,   -1,  420,   -1,  422,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  433,
   -1,  435,   -1,   -1,  438,  439,   -1,   -1,   -1,   -1,
   -1,  445,   -1,   -1,   -1,  449,   -1,  451,  452,  453,
  454,   -1,  456,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  475,  476,   -1,  478,  479,  480,   -1,   -1,   -1,
   -1,  485,  486,  487,  488,  489,   -1,  491,   -1,   -1,
  494,  495,   -1,   -1,   -1,  499,  500,   -1,   -1,   -1,
  504,  505,  506,  507,   -1,  509,  510,  511,   -1,  513,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  524,  525,  526,  527,  528,  529,  530,  531,  532,  533,
  534,  535,  536,  537,  538,  539,  540,  541,  542,  543,
   -1,  545,  546,  547,  548,  549,  550,   -1,  552,  553,
  554,  555,   -1,  557,   -1,  559,  560,  561,  562,  563,
  564,  565,  566,  567,  568,  569,  570,  571,  572,  573,
  574,  575,  576,   -1,  578,  579,  580,  581,   -1,  583,
  584,  585,  586,  587,   -1,  589,  590,  591,  592,  593,
  594,  595,  596,  597,  598,  599,  600,  601,   -1,  603,
  604,  605,  606,  607,  608,   -1,  610,  611,  612,  613,
  614,  615,  616,  617,   -1,  619,  620,   -1,   -1,   -1,
  624,  625,  626,  627,  628,  629,   -1,   -1,  632,  633,
  634,  635,  636,  637,  638,  639,  640,   -1,   -1,   -1,
   -1,   -1,  646,   -1,  648,  649,  650,  651,  652,  653,
   -1,   -1,   -1,   -1,   -1,   -1,    0,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  672,  673,
   -1,  675,  676,  677,  678,  679,  680,  681,  682,  683,
   -1,  685,  686,  687,  688,  689,  690,  691,  692,  693,
  694,  695,  696,   -1,  698,  699,  700,   -1,  702,   -1,
   44,   -1,  706,  707,  708,  709,  710,  711,  712,  713,
  714,  715,  716,  717,  718,   59,  720,  721,   -1,   -1,
   -1,   -1,   -1,  727,  728,  729,  730,  731,  732,  733,
  734,  735,  736,  737,  738,  739,  740,   -1,   -1,  743,
  744,   -1,   -1,   -1,   -1,  749,  750,  751,  752,  753,
  754,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  257,   -1,  259,   -1,   -1,   -1,   -1,
   -1,  265,   -1,   -1,  268,  269,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  279,  280,   -1,  282,   -1,
   -1,   -1,   -1,  287,   -1,   -1,  290,   -1,   -1,   -1,
   -1,   -1,   -1,  297,   -1,  299,  300,   -1,   -1,   -1,
  304,   -1,  306,   -1,   -1,  309,   -1,   -1,   -1,   -1,
  314,   -1,   -1,   -1,   -1,  319,  320,   -1,   -1,   -1,
  324,   -1,  326,   -1,  328,   -1,   -1,   -1,  332,   -1,
  334,   -1,  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  348,   -1,   -1,  351,   -1,  353,
   -1,   -1,   -1,  357,  358,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  370,   -1,   -1,   -1,
  374,  375,  376,  377,  378,   -1,  380,   -1,   -1,   -1,
   -1,   -1,  386,   -1,  388,  389,   -1,   -1,  392,  393,
   -1,   -1,   -1,   -1,   -1,  399,   -1,   -1,  402,  403,
  404,  405,   -1,  407,   -1,  409,   -1,  411,   -1,  413,
  414,  415,   -1,   -1,   -1,   -1,  420,   -1,  422,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  433,
   -1,   -1,   -1,   -1,  438,  439,   -1,   -1,   -1,   -1,
   -1,  445,   -1,   -1,   -1,  449,   -1,  451,  452,  453,
  454,   -1,  456,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  475,  476,   -1,  478,  479,  480,   -1,   -1,   -1,
   -1,  485,  486,   -1,  488,  489,   -1,  491,   -1,   -1,
  494,  495,   -1,   -1,   -1,  499,  500,   -1,   -1,   -1,
   -1,  505,  506,  507,   -1,  509,  510,  511,   -1,  513,
   -1,   -1,   -1,   -1,    0,   -1,   -1,   -1,   -1,  523,
  524,  525,  526,  527,  528,  529,  530,  531,  532,  533,
  534,  535,  536,  537,  538,  539,  540,  541,  542,  543,
   -1,  545,  546,  547,  548,  549,  550,   -1,  552,  553,
  554,  555,   -1,  557,   -1,  559,  560,  561,  562,  563,
  564,  565,  566,  567,  568,  569,  570,  571,  572,  573,
  574,  575,  576,   59,  578,  579,  580,  581,   -1,  583,
  584,  585,  586,  587,   -1,  589,  590,  591,  592,  593,
  594,  595,  596,  597,  598,  599,  600,  601,   -1,  603,
  604,  605,  606,  607,  608,   -1,  610,  611,  612,  613,
  614,  615,  616,  617,   -1,  619,  620,   -1,   -1,   -1,
  624,  625,  626,  627,  628,  629,   -1,   -1,  632,  633,
  634,  635,  636,  637,  638,  639,  640,   -1,   -1,   -1,
   -1,   -1,  646,   -1,  648,  649,  650,  651,  652,  653,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  672,  673,
   -1,  675,  676,  677,  678,  679,  680,  681,  682,  683,
   -1,  685,  686,  687,  688,  689,  690,  691,  692,  693,
  694,  695,  696,   -1,  698,  699,  700,   -1,  702,   -1,
   -1,   -1,  706,  707,  708,  709,  710,  711,  712,  713,
  714,  715,  716,  717,  718,   -1,  720,  721,   -1,   -1,
   -1,   -1,   -1,  727,  728,  729,  730,  731,  732,  733,
  734,  735,  736,  737,  738,  739,  740,   -1,   -1,  743,
  744,   -1,   -1,   -1,   -1,  749,  750,  751,  752,  753,
  754,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  257,   -1,  259,   -1,   -1,   -1,   -1,   -1,  265,
   -1,   -1,  268,  269,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  279,  280,   -1,  282,   -1,  284,   -1,
   -1,  287,   -1,   -1,  290,   -1,   -1,   -1,   -1,   -1,
   -1,  297,   -1,  299,  300,   -1,   -1,   -1,  304,   -1,
  306,   -1,   -1,  309,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  319,  320,   -1,   -1,   -1,  324,   -1,
  326,   -1,  328,   -1,   -1,   -1,   -1,   -1,  334,   -1,
  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  348,   -1,   -1,  351,   -1,  353,   -1,   -1,
   -1,  357,  358,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  370,   -1,   -1,   -1,  374,  375,
  376,  377,  378,   -1,  380,   -1,   -1,   -1,   -1,   -1,
  386,   -1,  388,  389,   -1,   -1,  392,  393,   -1,   -1,
   -1,   -1,   -1,  399,   -1,   -1,  402,  403,  404,  405,
   -1,  407,   -1,  409,   -1,  411,  412,  413,  414,  415,
   -1,   -1,   -1,   -1,  420,   -1,  422,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  433,   -1,   -1,
   -1,   -1,  438,  439,   -1,   -1,   -1,   -1,   -1,  445,
   -1,   -1,   -1,  449,   -1,  451,  452,  453,  454,   -1,
  456,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  464,  465,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  475,
  476,   -1,  478,  479,  480,   -1,   -1,   -1,   -1,  485,
  486,   -1,  488,  489,   -1,  491,   -1,   -1,  494,  495,
   -1,   -1,   -1,  499,  500,   -1,   -1,   -1,   -1,  505,
  506,  507,   -1,  509,  510,  511,   -1,  513,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  523,  524,  525,
  526,  527,  528,  529,  530,  531,  532,  533,  534,  535,
  536,  537,  538,  539,  540,  541,  542,  543,   -1,  545,
  546,  547,  548,  549,  550,   -1,  552,  553,  554,  555,
   -1,  557,   -1,  559,  560,  561,  562,  563,  564,  565,
  566,  567,  568,  569,  570,  571,  572,  573,  574,  575,
  576,   -1,  578,  579,  580,  581,   -1,  583,  584,  585,
  586,  587,   -1,  589,  590,  591,  592,  593,  594,  595,
  596,  597,  598,  599,  600,  601,   -1,  603,  604,  605,
  606,  607,  608,   -1,  610,  611,  612,  613,  614,  615,
  616,  617,   -1,  619,  620,   -1,   -1,   -1,  624,  625,
  626,  627,  628,  629,   -1,   -1,  632,  633,  634,  635,
  636,  637,  638,  639,  640,   -1,   40,   -1,   -1,   43,
  646,   45,  648,  649,  650,  651,  652,  653,   -1,   -1,
   -1,   -1,   -1,   -1,   58,   -1,   -1,   -1,   -1,   63,
   -1,   -1,   -1,   -1,   -1,   -1,  672,  673,   -1,  675,
  676,  677,  678,  679,  680,  681,  682,  683,   -1,  685,
  686,  687,  688,  689,  690,  691,  692,  693,  694,  695,
  696,   -1,  698,  699,  700,   -1,  702,   -1,   -1,   -1,
  706,  707,  708,  709,  710,  711,  712,  713,  714,  715,
  716,  717,  718,   -1,  720,  721,   -1,   -1,   -1,   -1,
   -1,  727,  728,  729,  730,  731,  732,  733,  734,  735,
  736,  737,  738,  739,  740,   -1,   -1,  743,  744,   -1,
   -1,   -1,   -1,  749,  750,  751,  752,  753,  754,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  257,   -1,  259,   -1,   -1,   -1,   -1,
   -1,  265,   -1,  267,  268,  269,   -1,   -1,   -1,   -1,
  274,   -1,   -1,   -1,   -1,  279,  280,   -1,  282,   -1,
  284,  285,   -1,  287,   -1,   -1,  290,  291,  292,   -1,
   -1,   -1,   -1,  297,   -1,  299,  300,   -1,   -1,   -1,
  304,   -1,  306,   -1,  308,  309,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  317,   -1,  319,  320,   -1,   -1,   -1,
  324,   -1,  326,   -1,  328,   -1,   -1,   -1,   -1,   -1,
  334,   -1,  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,
  344,   -1,  346,  347,  348,  349,   -1,  351,   -1,  353,
  354,   -1,  356,  357,  358,   -1,   -1,   -1,   -1,   -1,
  364,   -1,   -1,   -1,   -1,   -1,  370,   -1,   -1,   -1,
  374,  375,  376,  377,  378,   -1,  380,   -1,  382,   -1,
   -1,   -1,  386,   -1,  388,  389,   -1,   -1,  392,  393,
   -1,   -1,   -1,  397,   -1,  399,   -1,   -1,  402,  403,
  404,  405,   -1,  407,   -1,  409,  410,  411,  412,  413,
  414,  415,  416,   -1,   -1,   -1,  420,   -1,  422,   -1,
   -1,   -1,  426,   -1,   -1,   -1,   -1,   -1,   -1,  433,
   -1,   -1,   -1,   -1,  438,  439,  440,  441,  442,  443,
  444,  445,  446,  447,  448,  449,   -1,  451,  452,  453,
  454,   -1,  456,  457,  458,  459,  460,  461,  462,  463,
  464,  465,  466,  467,  468,  469,  470,  471,  472,  473,
  474,  475,  476,   -1,  478,  479,  480,  481,  482,   -1,
  484,  485,  486,   -1,  488,  489,  490,  491,   -1,   -1,
  494,  495,  496,  497,  498,  499,  500,   -1,   -1,   -1,
   -1,  505,  506,  507,   -1,  509,  510,  511,   -1,  513,
   -1,  515,  516,  517,   -1,  519,  520,   -1,  522,  523,
  524,  525,  526,  527,  528,  529,  530,  531,  532,  533,
  534,  535,  536,  537,  538,  539,  540,  541,  542,  543,
   -1,  545,  546,  547,  548,  549,  550,   -1,  552,  553,
  554,  555,   -1,  557,   -1,  559,  560,  561,  562,  563,
  564,  565,  566,  567,  568,  569,  570,  571,  572,  573,
  574,  575,  576,   -1,  578,  579,  580,  581,   -1,  583,
  584,  585,  586,  587,   -1,  589,  590,  591,  592,  593,
  594,  595,  596,  597,  598,  599,  600,  601,   -1,  603,
  604,  605,  606,  607,  608,   -1,  610,  611,  612,  613,
  614,  615,  616,  617,   -1,  619,  620,  621,  622,   -1,
  624,  625,  626,  627,  628,  629,   -1,   -1,  632,  633,
  634,  635,  636,  637,  638,  639,  640,  641,   -1,  643,
  644,  645,  646,  647,  648,  649,  650,  651,  652,  653,
   -1,   -1,  656,  657,  658,  659,  660,  661,  662,  663,
  664,  665,  666,  667,  668,  669,  670,  671,  672,  673,
   -1,  675,  676,  677,  678,  679,  680,  681,  682,  683,
   -1,  685,  686,  687,  688,  689,  690,  691,  692,  693,
  694,  695,  696,   -1,  698,  699,  700,   -1,  702,   -1,
  704,  705,  706,  707,  708,  709,  710,  711,  712,  713,
  714,  715,  716,  717,  718,   -1,  720,  721,  722,  723,
  724,  725,  726,  727,  728,  729,  730,  731,  732,  733,
  734,  735,  736,  737,  738,  739,  740,  741,  742,  743,
  744,   -1,   -1,   -1,   -1,  749,  750,  751,  752,  753,
  754,   40,   -1,   -1,   43,   -1,   45,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   58,
   -1,   -1,   -1,   -1,   63,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  257,   -1,
  259,   -1,   -1,   -1,   -1,   -1,  265,   -1,  267,  268,
  269,   -1,   -1,   -1,   -1,  274,   -1,   -1,   -1,   -1,
  279,  280,   -1,  282,   -1,  284,  285,   -1,  287,   -1,
   -1,  290,  291,  292,   -1,   -1,   -1,   -1,  297,   -1,
  299,  300,   -1,   -1,   -1,  304,   -1,  306,   -1,  308,
  309,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  317,   -1,
  319,  320,   -1,   -1,   -1,  324,   -1,  326,   -1,  328,
   -1,   -1,   -1,   -1,   -1,  334,   -1,  336,   -1,   -1,
   -1,  340,   -1,   -1,   -1,  344,   -1,  346,  347,  348,
  349,   -1,  351,   -1,  353,  354,   -1,  356,  357,  358,
   -1,   -1,   -1,   -1,   -1,  364,   -1,   -1,   -1,   -1,
   -1,  370,   -1,   -1,   -1,  374,  375,  376,  377,  378,
   -1,  380,   -1,  382,   -1,   -1,   -1,  386,   -1,  388,
  389,   -1,   -1,  392,  393,   -1,   -1,   -1,  397,   -1,
  399,   -1,   -1,  402,  403,  404,  405,   -1,  407,   -1,
  409,  410,  411,  412,  413,  414,  415,  416,   -1,   -1,
   -1,  420,   -1,  422,   -1,   -1,   -1,  426,   -1,   -1,
   -1,   -1,   -1,   -1,  433,   -1,   -1,   -1,   -1,  438,
  439,  440,  441,  442,  443,  444,  445,  446,  447,  448,
  449,   -1,  451,  452,  453,  454,   -1,  456,  457,   -1,
   -1,   -1,   -1,   -1,   -1,  464,  465,  466,  467,  468,
  469,  470,  471,  472,  473,  474,  475,  476,   -1,  478,
  479,  480,  481,  482,   -1,  484,  485,  486,   -1,  488,
  489,  490,  491,   -1,   -1,  494,  495,  496,  497,  498,
  499,  500,   -1,   -1,   -1,   -1,  505,  506,  507,   -1,
  509,  510,  511,  512,  513,   -1,  515,  516,  517,  518,
  519,  520,  521,  522,  523,  524,  525,  526,  527,  528,
  529,  530,  531,  532,  533,  534,  535,  536,  537,  538,
  539,  540,  541,  542,  543,   -1,  545,  546,  547,  548,
  549,  550,   -1,  552,  553,  554,  555,   -1,  557,   -1,
  559,  560,  561,  562,  563,  564,  565,  566,  567,  568,
  569,  570,  571,  572,  573,  574,  575,  576,   -1,  578,
  579,  580,  581,   -1,  583,  584,  585,  586,  587,   -1,
  589,  590,  591,  592,  593,  594,  595,  596,  597,  598,
  599,  600,  601,   -1,  603,  604,  605,  606,  607,  608,
   -1,  610,  611,  612,  613,  614,  615,  616,  617,   -1,
  619,  620,  621,  622,   -1,  624,  625,  626,  627,  628,
  629,   -1,   -1,  632,  633,  634,  635,  636,  637,  638,
  639,  640,  641,   -1,  643,  644,  645,  646,  647,  648,
  649,  650,  651,  652,  653,   -1,   -1,  656,  657,  658,
  659,  660,  661,  662,  663,  664,  665,  666,  667,  668,
  669,  670,  671,  672,  673,   -1,  675,  676,  677,  678,
  679,  680,  681,  682,  683,   -1,  685,  686,  687,  688,
  689,  690,  691,  692,  693,  694,  695,  696,   -1,  698,
  699,  700,   -1,  702,   -1,  704,  705,  706,  707,  708,
  709,  710,  711,  712,  713,  714,  715,  716,  717,  718,
   -1,  720,  721,  722,  723,  724,  725,  726,  727,  728,
  729,  730,  731,  732,  733,  734,  735,  736,  737,  738,
  739,  740,   -1,   40,  743,  744,   43,   -1,   45,   -1,
  749,  750,  751,  752,  753,  754,   -1,   -1,   -1,   -1,
   -1,   58,   -1,   -1,   -1,   -1,   63,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  257,   -1,  259,  260,   -1,   -1,  263,   -1,  265,   -1,
  267,  268,  269,   -1,   -1,   -1,   -1,  274,   -1,   -1,
   -1,   -1,  279,  280,   -1,  282,   -1,  284,  285,   -1,
  287,   -1,   -1,  290,  291,  292,   -1,   -1,   -1,   -1,
  297,   -1,  299,  300,   -1,   -1,   -1,  304,   -1,  306,
   -1,  308,  309,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  317,   -1,  319,  320,   -1,   -1,   -1,  324,   -1,  326,
   -1,  328,   -1,   -1,   -1,   -1,   -1,  334,   -1,  336,
   -1,   -1,   -1,  340,   -1,   -1,   -1,  344,   -1,  346,
  347,  348,  349,   -1,  351,   -1,  353,  354,   -1,  356,
  357,  358,   -1,   -1,   -1,   -1,   -1,  364,   -1,   -1,
   -1,   -1,   -1,  370,   -1,   -1,   -1,  374,  375,  376,
  377,  378,   -1,  380,   -1,  382,   -1,   -1,   -1,  386,
   -1,  388,  389,   -1,   -1,  392,  393,   -1,   -1,   -1,
  397,   -1,  399,   -1,   -1,  402,  403,  404,  405,   -1,
  407,  408,  409,  410,  411,  412,  413,  414,  415,  416,
   -1,   -1,   -1,  420,   -1,  422,   -1,   -1,   -1,  426,
   -1,   -1,   -1,   -1,   -1,   -1,  433,   -1,   -1,   -1,
   -1,  438,  439,  440,  441,  442,  443,  444,  445,  446,
  447,  448,  449,   -1,  451,  452,  453,  454,   -1,  456,
  457,   -1,   -1,   -1,   -1,   -1,   -1,  464,  465,  466,
  467,  468,  469,  470,  471,  472,  473,  474,  475,  476,
   -1,  478,  479,  480,  481,  482,   -1,  484,  485,  486,
   -1,  488,  489,  490,  491,   -1,   -1,  494,  495,  496,
  497,  498,  499,  500,   -1,   -1,   -1,   -1,  505,  506,
  507,   -1,  509,  510,  511,   -1,  513,   -1,  515,  516,
  517,   -1,  519,  520,   -1,  522,  523,  524,  525,  526,
  527,  528,  529,  530,  531,  532,  533,  534,  535,  536,
  537,  538,  539,  540,  541,  542,  543,   -1,  545,  546,
  547,  548,  549,  550,   -1,  552,  553,  554,  555,   -1,
  557,   -1,  559,  560,  561,  562,  563,  564,  565,  566,
  567,  568,  569,  570,  571,  572,  573,  574,  575,  576,
   -1,  578,  579,  580,  581,   -1,  583,  584,  585,  586,
  587,   -1,  589,  590,  591,  592,  593,  594,  595,  596,
  597,  598,  599,  600,  601,   -1,  603,  604,  605,  606,
  607,  608,   -1,  610,  611,  612,  613,  614,  615,  616,
  617,   -1,  619,  620,  621,  622,   -1,  624,  625,  626,
  627,  628,  629,   -1,   -1,  632,  633,  634,  635,  636,
  637,  638,  639,  640,  641,   -1,  643,  644,  645,  646,
  647,  648,  649,  650,  651,  652,  653,   -1,   -1,  656,
  657,  658,  659,  660,  661,  662,  663,  664,  665,  666,
  667,  668,  669,  670,  671,  672,  673,   -1,  675,  676,
  677,  678,  679,  680,  681,  682,  683,   -1,  685,  686,
  687,  688,  689,  690,  691,  692,  693,  694,  695,  696,
   -1,  698,  699,  700,   -1,  702,   -1,  704,  705,  706,
  707,  708,  709,  710,  711,  712,  713,  714,  715,  716,
  717,  718,   -1,  720,  721,  722,  723,  724,  725,  726,
  727,  728,  729,  730,  731,  732,  733,  734,  735,  736,
  737,  738,  739,  740,   -1,   40,  743,  744,   43,   -1,
   45,   -1,  749,  750,  751,  752,  753,  754,   -1,   -1,
   -1,   -1,   -1,   58,   -1,   -1,   -1,   -1,   63,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  257,   -1,  259,   -1,   -1,   -1,   -1,   -1,
  265,   -1,  267,  268,  269,   -1,  271,   -1,   -1,  274,
   -1,   -1,   -1,   -1,  279,  280,   -1,  282,   -1,  284,
  285,   -1,  287,  288,   -1,  290,  291,  292,   -1,   -1,
   -1,   -1,  297,   -1,  299,  300,   -1,   -1,   -1,  304,
   -1,  306,   -1,  308,  309,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  317,   -1,  319,  320,   -1,   -1,   -1,  324,
   -1,  326,   -1,  328,   -1,   -1,   -1,   -1,   -1,  334,
   -1,  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,  344,
   -1,  346,  347,  348,  349,   -1,  351,   -1,  353,  354,
   -1,  356,  357,  358,   -1,   -1,   -1,   -1,   -1,  364,
   -1,   -1,   -1,   -1,   -1,  370,   -1,   -1,   -1,  374,
  375,  376,  377,  378,   -1,  380,   -1,  382,   -1,   -1,
   -1,  386,   -1,  388,  389,   -1,   -1,  392,  393,   -1,
   -1,   -1,  397,   -1,  399,   -1,   -1,  402,  403,  404,
  405,   -1,  407,   -1,  409,  410,  411,  412,  413,  414,
  415,  416,   -1,   -1,   -1,  420,   -1,  422,   -1,   -1,
   -1,  426,   -1,   -1,   -1,   -1,   -1,   -1,  433,   -1,
   -1,   -1,   -1,  438,  439,  440,  441,  442,  443,  444,
  445,  446,  447,  448,  449,   -1,  451,  452,  453,  454,
   -1,  456,  457,   -1,   -1,   -1,   -1,   -1,   -1,  464,
  465,  466,  467,  468,  469,  470,  471,  472,  473,  474,
  475,  476,   -1,  478,  479,  480,  481,  482,   -1,  484,
  485,  486,   -1,  488,  489,  490,  491,   -1,   -1,  494,
  495,  496,  497,  498,  499,  500,   -1,   -1,   -1,   -1,
  505,  506,  507,   -1,  509,  510,  511,   -1,  513,   -1,
  515,  516,  517,   -1,  519,  520,   -1,  522,  523,  524,
  525,  526,  527,  528,  529,  530,  531,  532,  533,  534,
  535,  536,  537,  538,  539,  540,  541,  542,  543,   -1,
  545,  546,  547,  548,  549,  550,   -1,  552,  553,  554,
  555,   -1,  557,   -1,  559,  560,  561,  562,  563,  564,
  565,  566,  567,  568,  569,  570,  571,  572,  573,  574,
  575,  576,   -1,  578,  579,  580,  581,   -1,  583,  584,
  585,  586,  587,   -1,  589,  590,  591,  592,  593,  594,
  595,  596,  597,  598,  599,  600,  601,   -1,  603,  604,
  605,  606,  607,  608,   -1,  610,  611,  612,  613,  614,
  615,  616,  617,   -1,  619,  620,  621,  622,   -1,  624,
  625,  626,  627,  628,  629,   -1,   -1,  632,  633,  634,
  635,  636,  637,  638,  639,  640,  641,   -1,  643,  644,
  645,  646,  647,  648,  649,  650,  651,  652,  653,   -1,
   -1,  656,  657,  658,  659,  660,  661,  662,  663,  664,
  665,  666,  667,  668,  669,  670,  671,  672,  673,   -1,
  675,  676,  677,  678,  679,  680,  681,  682,  683,   -1,
  685,  686,  687,  688,  689,  690,  691,  692,  693,  694,
  695,  696,   -1,  698,  699,  700,   -1,  702,   -1,  704,
  705,  706,  707,  708,  709,  710,  711,  712,  713,  714,
  715,  716,  717,  718,   -1,  720,  721,  722,  723,  724,
  725,  726,  727,  728,  729,  730,  731,  732,  733,  734,
  735,  736,  737,  738,  739,  740,   -1,   -1,  743,  744,
  745,   -1,   -1,   -1,  749,  750,  751,  752,  753,  754,
   40,   -1,   -1,   43,   -1,   45,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   58,   -1,
   -1,   -1,   -1,   63,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  257,   -1,  259,
   -1,   -1,   -1,   -1,   -1,  265,   -1,  267,  268,  269,
   -1,   -1,   -1,   -1,  274,   -1,   -1,   -1,   -1,  279,
  280,   -1,  282,   -1,  284,  285,   -1,  287,  288,   -1,
  290,  291,  292,   -1,   -1,   -1,   -1,  297,   -1,  299,
  300,   -1,   -1,   -1,  304,   -1,  306,   -1,  308,  309,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  317,   -1,  319,
  320,   -1,   -1,   -1,  324,   -1,  326,   -1,  328,   -1,
   -1,   -1,   -1,   -1,  334,   -1,  336,   -1,   -1,   -1,
  340,   -1,   -1,   -1,  344,   -1,  346,  347,  348,  349,
   -1,  351,   -1,  353,  354,   -1,  356,  357,  358,   -1,
   -1,   -1,   -1,   -1,  364,   -1,   -1,   -1,   -1,   -1,
  370,   -1,   -1,   -1,  374,  375,  376,  377,  378,   -1,
  380,   -1,  382,   -1,   -1,   -1,  386,   -1,  388,  389,
   -1,   -1,  392,  393,   -1,   -1,   -1,  397,   -1,  399,
   -1,   -1,  402,  403,  404,  405,   -1,  407,   -1,  409,
  410,  411,  412,  413,  414,  415,  416,   -1,   -1,   -1,
  420,   -1,  422,   -1,   -1,   -1,  426,   -1,   -1,   -1,
   -1,   -1,   -1,  433,   -1,   -1,   -1,   -1,  438,  439,
  440,  441,  442,  443,  444,  445,  446,  447,  448,  449,
   -1,  451,  452,  453,  454,   -1,  456,  457,   -1,   -1,
   -1,   -1,   -1,   -1,  464,  465,  466,  467,  468,  469,
  470,  471,  472,  473,  474,  475,  476,   -1,  478,  479,
  480,  481,  482,   -1,  484,  485,  486,   -1,  488,  489,
  490,  491,   -1,   -1,  494,  495,  496,  497,  498,  499,
  500,   -1,   -1,   -1,   -1,  505,  506,  507,   -1,  509,
  510,  511,   -1,  513,   -1,  515,  516,  517,   -1,  519,
  520,   -1,  522,  523,  524,  525,  526,  527,  528,  529,
  530,  531,  532,  533,  534,  535,  536,  537,  538,  539,
  540,  541,  542,  543,   -1,  545,  546,  547,  548,  549,
  550,   -1,  552,  553,  554,  555,   -1,  557,   -1,  559,
  560,  561,  562,  563,  564,  565,  566,  567,  568,  569,
  570,  571,  572,  573,  574,  575,  576,   -1,  578,  579,
  580,  581,   -1,  583,  584,  585,  586,  587,   -1,  589,
  590,  591,  592,  593,  594,  595,  596,  597,  598,  599,
  600,  601,   -1,  603,  604,  605,  606,  607,  608,   -1,
  610,  611,  612,  613,  614,  615,  616,  617,   -1,  619,
  620,  621,  622,   -1,  624,  625,  626,  627,  628,  629,
   -1,   -1,  632,  633,  634,  635,  636,  637,  638,  639,
  640,  641,   -1,  643,  644,  645,  646,  647,  648,  649,
  650,  651,  652,  653,   -1,   -1,  656,  657,  658,  659,
  660,  661,  662,  663,  664,  665,  666,  667,  668,  669,
  670,  671,  672,  673,   -1,  675,  676,  677,  678,  679,
  680,  681,  682,  683,   -1,  685,  686,  687,  688,  689,
  690,  691,  692,  693,  694,  695,  696,   -1,  698,  699,
  700,   -1,  702,   -1,  704,  705,  706,  707,  708,  709,
  710,  711,  712,  713,  714,  715,  716,  717,  718,   -1,
  720,  721,  722,  723,  724,  725,  726,  727,  728,  729,
  730,  731,  732,  733,  734,  735,  736,  737,  738,  739,
  740,   -1,   -1,  743,  744,  745,   -1,   -1,   -1,  749,
  750,  751,  752,  753,  754,   40,   -1,   -1,   43,   -1,
   45,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   58,   -1,   -1,   -1,   -1,   63,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  257,   -1,  259,   -1,   -1,   -1,   -1,   -1,
  265,   -1,  267,  268,  269,   -1,   -1,   -1,   -1,  274,
   -1,   -1,   -1,   -1,  279,  280,   -1,  282,   -1,  284,
  285,   -1,  287,  288,   -1,  290,  291,  292,   -1,   -1,
   -1,   -1,  297,   -1,  299,  300,   -1,   -1,   -1,  304,
   -1,  306,   -1,  308,  309,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  317,   -1,  319,  320,   -1,   -1,   -1,  324,
   -1,  326,   -1,  328,   -1,   -1,   -1,   -1,   -1,  334,
   -1,  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,  344,
   -1,  346,  347,  348,  349,   -1,  351,   -1,  353,  354,
   -1,  356,  357,  358,   -1,   -1,   -1,   -1,   -1,  364,
   -1,   -1,   -1,   -1,   -1,  370,   -1,   -1,   -1,  374,
  375,  376,  377,  378,   -1,  380,   -1,  382,   -1,   -1,
   -1,  386,   -1,  388,  389,   -1,   -1,  392,  393,   -1,
   -1,   -1,  397,   -1,  399,   -1,   -1,  402,  403,  404,
  405,   -1,  407,   -1,  409,  410,  411,  412,  413,  414,
  415,  416,   -1,   -1,   -1,  420,   -1,  422,   -1,   -1,
   -1,  426,   -1,   -1,   -1,   -1,   -1,   -1,  433,   -1,
   -1,   -1,   -1,  438,  439,  440,  441,  442,  443,  444,
  445,  446,  447,  448,  449,   -1,  451,  452,  453,  454,
   -1,  456,  457,   -1,   -1,   -1,   -1,   -1,   -1,  464,
  465,  466,  467,  468,  469,  470,  471,  472,  473,  474,
  475,  476,   -1,  478,  479,  480,  481,  482,   -1,  484,
  485,  486,   -1,  488,  489,  490,  491,   -1,   -1,  494,
  495,  496,  497,  498,  499,  500,   -1,   -1,   -1,   -1,
  505,  506,  507,   -1,  509,  510,  511,   -1,  513,   -1,
  515,  516,  517,   -1,  519,  520,   -1,  522,  523,  524,
  525,  526,  527,  528,  529,  530,  531,  532,  533,  534,
  535,  536,  537,  538,  539,  540,  541,  542,  543,   -1,
  545,  546,  547,  548,  549,  550,   -1,  552,  553,  554,
  555,   -1,  557,   -1,  559,  560,  561,  562,  563,  564,
  565,  566,  567,  568,  569,  570,  571,  572,  573,  574,
  575,  576,   -1,  578,  579,  580,  581,   -1,  583,  584,
  585,  586,  587,   -1,  589,  590,  591,  592,  593,  594,
  595,  596,  597,  598,  599,  600,  601,   -1,  603,  604,
  605,  606,  607,  608,   -1,  610,  611,  612,  613,  614,
  615,  616,  617,   -1,  619,  620,  621,  622,   -1,  624,
  625,  626,  627,  628,  629,   -1,   -1,  632,  633,  634,
  635,  636,  637,  638,  639,  640,  641,   -1,  643,  644,
  645,  646,  647,  648,  649,  650,  651,  652,  653,   -1,
   -1,  656,  657,  658,  659,  660,  661,  662,  663,  664,
  665,  666,  667,  668,  669,  670,  671,  672,  673,   -1,
  675,  676,  677,  678,  679,  680,  681,  682,  683,   -1,
  685,  686,  687,  688,  689,  690,  691,  692,  693,  694,
  695,  696,   -1,  698,  699,  700,   -1,  702,   -1,  704,
  705,  706,  707,  708,  709,  710,  711,  712,  713,  714,
  715,  716,  717,  718,   -1,  720,  721,  722,  723,  724,
  725,  726,  727,  728,  729,  730,  731,  732,  733,  734,
  735,  736,  737,  738,  739,  740,   -1,   -1,  743,  744,
  745,   -1,   -1,   -1,  749,  750,  751,  752,  753,  754,
   40,   -1,   42,   43,   -1,   45,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   58,   -1,
   -1,   -1,   -1,   63,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  257,   -1,  259,
  260,   -1,   -1,   -1,   -1,  265,   -1,  267,  268,  269,
   -1,   -1,   -1,   -1,  274,   -1,   -1,   -1,   -1,  279,
  280,   -1,  282,   -1,  284,  285,   -1,  287,   -1,   -1,
  290,  291,  292,   -1,   -1,   -1,   -1,  297,  298,  299,
  300,   -1,   -1,   -1,  304,   -1,  306,   -1,  308,  309,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  317,   -1,  319,
  320,   -1,   -1,   -1,  324,   -1,  326,   -1,  328,   -1,
   -1,   -1,   -1,   -1,  334,   -1,  336,   -1,   -1,   -1,
  340,   -1,   -1,   -1,  344,   -1,  346,  347,  348,  349,
   -1,  351,   -1,  353,  354,   -1,  356,  357,  358,   -1,
   -1,   -1,   -1,   -1,  364,   -1,   -1,   -1,   -1,   -1,
  370,   -1,   -1,   -1,  374,  375,  376,  377,  378,   -1,
  380,   -1,  382,   -1,   -1,   -1,  386,   -1,  388,  389,
   -1,   -1,  392,  393,   -1,   -1,   -1,  397,   -1,  399,
   -1,   -1,  402,  403,  404,  405,   -1,  407,   -1,  409,
  410,  411,  412,  413,  414,  415,  416,   -1,   -1,   -1,
  420,   -1,  422,   -1,   -1,   -1,  426,   -1,   -1,   -1,
   -1,   -1,   -1,  433,   -1,   -1,   -1,   -1,  438,  439,
  440,  441,  442,  443,  444,  445,  446,  447,  448,  449,
   -1,  451,  452,  453,  454,   -1,  456,  457,   -1,   -1,
   -1,   -1,   -1,   -1,  464,  465,  466,  467,  468,  469,
  470,  471,  472,  473,  474,  475,  476,   -1,  478,  479,
   -1,  481,  482,   -1,  484,  485,  486,   -1,  488,  489,
  490,  491,   -1,   -1,  494,  495,  496,  497,  498,  499,
  500,   -1,   -1,   -1,   -1,  505,  506,  507,   -1,  509,
  510,  511,   -1,  513,   -1,  515,  516,  517,   -1,  519,
  520,   -1,  522,  523,  524,  525,  526,  527,  528,  529,
  530,  531,  532,  533,  534,  535,  536,  537,  538,  539,
  540,  541,  542,  543,   -1,  545,  546,  547,  548,  549,
  550,   -1,  552,  553,  554,  555,   -1,  557,   -1,  559,
  560,  561,  562,  563,  564,  565,  566,  567,  568,  569,
  570,  571,  572,  573,  574,  575,  576,   -1,  578,  579,
  580,  581,   -1,  583,  584,  585,  586,  587,   -1,  589,
  590,  591,  592,  593,  594,  595,  596,  597,  598,  599,
  600,  601,   -1,  603,  604,  605,  606,  607,  608,   -1,
  610,  611,  612,  613,  614,  615,  616,  617,   -1,  619,
  620,  621,  622,   -1,  624,  625,  626,  627,  628,  629,
   -1,   -1,  632,  633,  634,  635,  636,  637,  638,  639,
  640,  641,   -1,  643,  644,  645,  646,  647,  648,  649,
  650,  651,  652,  653,   -1,   -1,  656,  657,  658,  659,
  660,  661,  662,  663,  664,  665,  666,  667,  668,  669,
  670,  671,  672,  673,   -1,  675,  676,  677,  678,  679,
  680,  681,  682,  683,   -1,  685,  686,  687,  688,  689,
  690,  691,  692,  693,  694,  695,  696,   -1,  698,  699,
  700,   -1,  702,   -1,  704,  705,  706,  707,  708,  709,
  710,  711,  712,  713,  714,  715,  716,  717,  718,   -1,
  720,  721,  722,  723,  724,  725,  726,  727,  728,  729,
  730,  731,  732,  733,  734,  735,  736,  737,  738,  739,
  740,   -1,   40,  743,  744,   43,   -1,   45,   -1,  749,
  750,  751,  752,  753,  754,   -1,   -1,   -1,   -1,   -1,
   58,   -1,   -1,   -1,   -1,   63,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  257,
   -1,  259,   -1,   -1,   -1,   -1,   -1,  265,   -1,  267,
  268,  269,   -1,   -1,   -1,   -1,  274,   -1,   -1,   -1,
   -1,  279,  280,   -1,  282,   -1,  284,  285,   -1,  287,
  288,   -1,  290,  291,  292,   -1,   -1,   -1,   -1,  297,
   -1,  299,  300,   -1,   -1,   -1,  304,   -1,  306,   -1,
  308,  309,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  317,
   -1,  319,  320,   -1,   -1,   -1,  324,   -1,  326,   -1,
  328,   -1,   -1,   -1,   -1,   -1,  334,   -1,  336,   -1,
   -1,   -1,  340,   -1,   -1,   -1,  344,   -1,  346,  347,
  348,  349,   -1,  351,   -1,  353,  354,   -1,  356,  357,
  358,   -1,   -1,   -1,   -1,   -1,  364,   -1,   -1,   -1,
   -1,   -1,  370,   -1,   -1,   -1,  374,  375,  376,  377,
  378,   -1,  380,   -1,  382,   -1,   -1,   -1,  386,   -1,
  388,  389,   -1,   -1,  392,  393,   -1,   -1,   -1,  397,
   -1,  399,   -1,   -1,  402,  403,  404,  405,   -1,  407,
   -1,  409,  410,  411,  412,  413,  414,  415,  416,   -1,
   -1,   -1,  420,   -1,  422,   -1,   -1,   -1,  426,   -1,
   -1,   -1,   -1,   -1,   -1,  433,   -1,   -1,   -1,   -1,
  438,  439,  440,  441,  442,  443,  444,  445,  446,  447,
  448,  449,   -1,  451,  452,  453,  454,   -1,  456,  457,
   -1,   -1,   -1,   -1,   -1,   -1,  464,  465,  466,  467,
  468,  469,  470,  471,  472,  473,  474,  475,  476,   -1,
  478,  479,  480,  481,  482,   -1,  484,  485,  486,   -1,
  488,  489,  490,  491,   -1,   -1,  494,  495,  496,  497,
  498,  499,  500,   -1,   -1,   -1,   -1,  505,  506,  507,
   -1,  509,  510,  511,   -1,  513,   -1,  515,  516,  517,
   -1,  519,  520,   -1,  522,  523,  524,  525,  526,  527,
  528,  529,  530,  531,  532,  533,  534,  535,  536,  537,
  538,  539,  540,  541,  542,  543,   -1,  545,  546,  547,
  548,  549,  550,   -1,  552,  553,  554,  555,   -1,  557,
   -1,  559,  560,  561,  562,  563,  564,  565,  566,  567,
  568,  569,  570,  571,  572,  573,  574,  575,  576,   -1,
  578,  579,  580,  581,   -1,  583,  584,  585,  586,  587,
   -1,  589,  590,  591,  592,  593,  594,  595,  596,  597,
  598,  599,  600,  601,   -1,  603,  604,  605,  606,  607,
  608,   -1,  610,  611,  612,  613,  614,  615,  616,  617,
   -1,  619,  620,  621,  622,   -1,  624,  625,  626,  627,
  628,  629,   -1,   -1,  632,  633,  634,  635,  636,  637,
  638,  639,  640,  641,   -1,  643,  644,  645,  646,  647,
  648,  649,  650,  651,  652,  653,   -1,   -1,  656,  657,
  658,  659,  660,  661,  662,  663,  664,  665,  666,  667,
  668,  669,  670,  671,  672,  673,   -1,  675,  676,  677,
  678,  679,  680,  681,  682,  683,   -1,  685,  686,  687,
  688,  689,  690,  691,  692,  693,  694,  695,  696,   -1,
  698,  699,  700,   -1,  702,   -1,  704,  705,  706,  707,
  708,  709,  710,  711,  712,  713,  714,  715,  716,  717,
  718,   -1,  720,  721,  722,  723,  724,  725,  726,  727,
  728,  729,  730,  731,  732,  733,  734,  735,  736,  737,
  738,  739,  740,   -1,   40,  743,  744,   43,   -1,   45,
   -1,  749,  750,  751,  752,  753,  754,   -1,   -1,   -1,
   -1,   -1,   58,   -1,   -1,   -1,   -1,   63,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  257,   -1,  259,   -1,   -1,   -1,   -1,   -1,  265,
   -1,  267,  268,  269,   -1,   -1,   -1,   -1,  274,   -1,
   -1,   -1,   -1,  279,  280,   -1,  282,   -1,  284,  285,
   -1,  287,   -1,   -1,  290,  291,  292,   -1,   -1,   -1,
   -1,  297,   -1,  299,  300,   -1,   -1,   -1,  304,   -1,
  306,   -1,  308,  309,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  317,   -1,  319,  320,   -1,   -1,   -1,  324,   -1,
  326,   -1,  328,   -1,   -1,   -1,   -1,   -1,  334,   -1,
  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,  344,   -1,
  346,  347,  348,  349,   -1,  351,   -1,  353,  354,   -1,
  356,  357,  358,   -1,   -1,   -1,   -1,   -1,  364,   -1,
   -1,   -1,   -1,   -1,  370,   -1,   -1,   -1,  374,  375,
  376,  377,  378,   -1,  380,   -1,  382,   -1,   -1,   -1,
  386,   -1,  388,  389,   -1,   -1,  392,  393,   -1,   -1,
   -1,  397,   -1,  399,   -1,   -1,  402,  403,  404,  405,
   -1,  407,   -1,  409,  410,  411,  412,  413,  414,  415,
  416,   -1,   -1,   -1,  420,   -1,  422,   -1,   -1,   -1,
  426,   -1,   -1,   -1,   -1,   -1,   -1,  433,  434,   -1,
   -1,   -1,  438,  439,  440,  441,  442,  443,  444,  445,
  446,  447,  448,  449,   -1,  451,  452,  453,  454,   -1,
  456,  457,   -1,   -1,   -1,   -1,   -1,   -1,  464,  465,
  466,  467,  468,  469,  470,  471,  472,  473,  474,  475,
  476,   -1,  478,  479,  480,  481,  482,   -1,  484,  485,
  486,   -1,  488,  489,  490,  491,   -1,   -1,  494,  495,
  496,  497,  498,  499,  500,   -1,   -1,   -1,   -1,  505,
  506,  507,   -1,  509,  510,  511,   -1,  513,   -1,  515,
  516,  517,   -1,  519,  520,   -1,  522,  523,  524,  525,
  526,  527,  528,  529,  530,  531,  532,  533,  534,  535,
  536,  537,  538,  539,  540,  541,  542,  543,   -1,  545,
  546,  547,  548,  549,  550,   -1,  552,  553,  554,  555,
   -1,  557,   -1,  559,  560,  561,  562,  563,  564,  565,
  566,  567,  568,  569,  570,  571,  572,  573,  574,  575,
  576,   -1,  578,  579,  580,  581,   -1,  583,  584,  585,
  586,  587,   -1,  589,  590,  591,  592,  593,  594,  595,
  596,  597,  598,  599,  600,  601,   -1,  603,  604,  605,
  606,  607,  608,   -1,  610,  611,  612,  613,  614,  615,
  616,  617,   -1,  619,  620,  621,  622,   -1,  624,  625,
  626,  627,  628,  629,   -1,   -1,  632,  633,  634,  635,
  636,  637,  638,  639,  640,  641,   -1,  643,  644,  645,
  646,  647,  648,  649,  650,  651,  652,  653,   -1,   -1,
  656,  657,  658,  659,  660,  661,  662,  663,  664,  665,
  666,  667,  668,  669,  670,  671,  672,  673,   -1,  675,
  676,  677,  678,  679,  680,  681,  682,  683,   -1,  685,
  686,  687,  688,  689,  690,  691,  692,  693,  694,  695,
  696,   -1,  698,  699,  700,   -1,  702,   -1,  704,  705,
  706,  707,  708,  709,  710,  711,  712,  713,  714,  715,
  716,  717,  718,   -1,  720,  721,  722,  723,  724,  725,
  726,  727,  728,  729,  730,  731,  732,  733,  734,  735,
  736,  737,  738,  739,  740,   -1,   40,  743,  744,   43,
   -1,   45,   -1,  749,  750,  751,  752,  753,  754,   -1,
   -1,   -1,   -1,   -1,   58,   -1,   -1,   -1,   -1,   63,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  257,   -1,  259,   -1,   -1,   -1,   -1,
   -1,  265,   -1,  267,  268,  269,   -1,   -1,   -1,   -1,
  274,   -1,   -1,   -1,   -1,  279,  280,   -1,  282,   -1,
  284,  285,   -1,  287,   -1,   -1,  290,  291,  292,   -1,
   -1,   -1,   -1,  297,   -1,  299,  300,   -1,   -1,   -1,
  304,   -1,  306,   -1,  308,  309,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  317,   -1,  319,  320,   -1,   -1,   -1,
  324,   -1,  326,   -1,  328,   -1,   -1,   -1,   -1,   -1,
  334,   -1,  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,
  344,   -1,  346,  347,  348,  349,   -1,  351,   -1,  353,
  354,   -1,  356,  357,  358,   -1,   -1,   -1,   -1,   -1,
  364,   -1,   -1,   -1,   -1,   -1,  370,   -1,   -1,   -1,
  374,  375,  376,  377,  378,   -1,  380,   -1,  382,   -1,
   -1,   -1,  386,   -1,  388,  389,   -1,   -1,  392,  393,
   -1,   -1,   -1,  397,   -1,  399,   -1,   -1,  402,  403,
  404,  405,   -1,  407,   -1,  409,  410,  411,  412,  413,
  414,  415,  416,   -1,   -1,   -1,  420,   -1,  422,   -1,
   -1,   -1,  426,   -1,   -1,   -1,   -1,   -1,   -1,  433,
   -1,   -1,   -1,  437,  438,  439,  440,  441,  442,  443,
  444,  445,  446,  447,  448,  449,   -1,  451,  452,  453,
  454,   -1,  456,  457,   -1,   -1,   -1,   -1,   -1,   -1,
  464,  465,  466,  467,  468,  469,  470,  471,  472,  473,
  474,  475,  476,   -1,  478,  479,  480,  481,  482,   -1,
  484,  485,  486,   -1,  488,  489,  490,  491,   -1,   -1,
  494,  495,  496,  497,  498,  499,  500,   -1,   -1,   -1,
   -1,  505,  506,  507,   -1,  509,  510,  511,   -1,  513,
   -1,  515,  516,  517,   -1,  519,  520,   -1,  522,  523,
  524,  525,  526,  527,  528,  529,  530,  531,  532,  533,
  534,  535,  536,  537,  538,  539,  540,  541,  542,  543,
   -1,  545,  546,  547,  548,  549,  550,   -1,  552,  553,
  554,  555,   -1,  557,   -1,  559,  560,  561,  562,  563,
  564,  565,  566,  567,  568,  569,  570,  571,  572,  573,
  574,  575,  576,   -1,  578,  579,  580,  581,   -1,  583,
  584,  585,  586,  587,   -1,  589,  590,  591,  592,  593,
  594,  595,  596,  597,  598,  599,  600,  601,   -1,  603,
  604,  605,  606,  607,  608,   -1,  610,  611,  612,  613,
  614,  615,  616,  617,   -1,  619,  620,  621,  622,   -1,
  624,  625,  626,  627,  628,  629,   -1,   -1,  632,  633,
  634,  635,  636,  637,  638,  639,  640,  641,   -1,  643,
  644,  645,  646,  647,  648,  649,  650,  651,  652,  653,
   -1,   -1,  656,  657,  658,  659,  660,  661,  662,  663,
  664,  665,  666,  667,  668,  669,  670,  671,  672,  673,
   -1,  675,  676,  677,  678,  679,  680,  681,  682,  683,
   -1,  685,  686,  687,  688,  689,  690,  691,  692,  693,
  694,  695,  696,   -1,  698,  699,  700,   -1,  702,   -1,
  704,  705,  706,  707,  708,  709,  710,  711,  712,  713,
  714,  715,  716,  717,  718,   -1,  720,  721,  722,  723,
  724,  725,  726,  727,  728,  729,  730,  731,  732,  733,
  734,  735,  736,  737,  738,  739,  740,   -1,   -1,  743,
  744,   -1,   -1,   -1,   -1,  749,  750,  751,  752,  753,
  754,   40,   -1,   42,   43,   -1,   45,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   58,
   -1,   -1,   -1,   -1,   63,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  257,   -1,
  259,   -1,   -1,   -1,   -1,   -1,  265,   -1,  267,  268,
  269,   -1,   -1,   -1,   -1,  274,   -1,   -1,   -1,   -1,
  279,  280,   -1,  282,   -1,  284,  285,   -1,  287,   -1,
   -1,  290,  291,  292,   -1,   -1,   -1,   -1,  297,   -1,
  299,  300,   -1,   -1,   -1,  304,   -1,  306,   -1,  308,
  309,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  317,   -1,
  319,  320,   -1,   -1,   -1,  324,   -1,  326,   -1,  328,
   -1,   -1,   -1,   -1,   -1,  334,   -1,  336,   -1,   -1,
   -1,  340,   -1,   -1,   -1,  344,   -1,  346,  347,  348,
  349,   -1,  351,   -1,  353,  354,   -1,  356,  357,  358,
   -1,   -1,   -1,   -1,   -1,  364,   -1,   -1,   -1,   -1,
   -1,  370,   -1,   -1,   -1,  374,  375,  376,  377,  378,
   -1,  380,   -1,  382,   -1,   -1,   -1,  386,   -1,  388,
  389,   -1,   -1,  392,  393,   -1,   -1,   -1,  397,   -1,
  399,   -1,   -1,  402,  403,  404,  405,   -1,  407,   -1,
  409,  410,  411,  412,  413,  414,  415,  416,   -1,   -1,
   -1,  420,   -1,  422,   -1,   -1,   -1,  426,   -1,   -1,
   -1,   -1,   -1,   -1,  433,   -1,   -1,   -1,   -1,  438,
  439,  440,  441,  442,  443,  444,  445,  446,  447,  448,
  449,   -1,  451,  452,  453,  454,   -1,  456,  457,   -1,
   -1,   -1,   -1,   -1,   -1,  464,  465,  466,  467,  468,
  469,  470,  471,  472,  473,  474,  475,  476,   -1,  478,
  479,  480,  481,  482,   -1,  484,  485,  486,   -1,  488,
  489,  490,  491,   -1,   -1,  494,  495,  496,  497,  498,
  499,  500,   -1,   -1,   -1,   -1,  505,  506,  507,   -1,
  509,  510,  511,   -1,  513,   -1,  515,  516,  517,   -1,
  519,  520,   -1,  522,  523,  524,  525,  526,  527,  528,
  529,  530,  531,  532,  533,  534,  535,  536,  537,  538,
  539,  540,  541,  542,  543,   -1,  545,  546,  547,  548,
  549,  550,   -1,  552,  553,  554,  555,   -1,  557,   -1,
  559,  560,  561,  562,  563,  564,  565,  566,  567,  568,
  569,  570,  571,  572,  573,  574,  575,  576,   -1,  578,
  579,  580,  581,   -1,  583,  584,  585,  586,  587,   -1,
  589,  590,  591,  592,  593,  594,  595,  596,  597,  598,
  599,  600,  601,   -1,  603,  604,  605,  606,  607,  608,
   -1,  610,  611,  612,  613,  614,  615,  616,  617,   -1,
  619,  620,  621,  622,   -1,  624,  625,  626,  627,  628,
  629,   -1,   -1,  632,  633,  634,  635,  636,  637,  638,
  639,  640,  641,   -1,  643,  644,  645,  646,  647,  648,
  649,  650,  651,  652,  653,   -1,   -1,  656,  657,  658,
  659,  660,  661,  662,  663,  664,  665,  666,  667,  668,
  669,  670,  671,  672,  673,   -1,  675,  676,  677,  678,
  679,  680,  681,  682,  683,   -1,  685,  686,  687,  688,
  689,  690,  691,  692,  693,  694,  695,  696,   -1,  698,
  699,  700,   -1,  702,   -1,  704,  705,  706,  707,  708,
  709,  710,  711,  712,  713,  714,  715,  716,  717,  718,
   -1,  720,  721,  722,  723,  724,  725,  726,  727,  728,
  729,  730,  731,  732,  733,  734,  735,  736,  737,  738,
  739,  740,   -1,   -1,  743,  744,   -1,   -1,   -1,   -1,
  749,  750,  751,  752,  753,  754,   40,   41,   -1,   43,
   -1,   45,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   58,   -1,   -1,   -1,   -1,   63,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  257,   -1,  259,   -1,   -1,   -1,   -1,
   -1,  265,   -1,  267,  268,  269,   -1,   -1,   -1,   -1,
  274,   -1,   -1,   -1,   -1,  279,  280,   -1,  282,   -1,
  284,  285,   -1,  287,   -1,   -1,  290,  291,  292,   -1,
   -1,   -1,   -1,  297,   -1,  299,  300,   -1,   -1,   -1,
  304,   -1,  306,   -1,  308,  309,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  317,   -1,  319,  320,   -1,   -1,   -1,
  324,   -1,  326,   -1,  328,   -1,   -1,   -1,   -1,   -1,
  334,   -1,  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,
  344,   -1,  346,  347,  348,  349,   -1,  351,   -1,  353,
  354,   -1,  356,  357,  358,   -1,   -1,   -1,   -1,   -1,
  364,   -1,   -1,   -1,   -1,   -1,  370,   -1,   -1,   -1,
  374,  375,  376,  377,  378,   -1,  380,   -1,  382,   -1,
   -1,   -1,  386,   -1,  388,  389,   -1,   -1,  392,  393,
   -1,   -1,   -1,  397,   -1,  399,   -1,   -1,  402,  403,
  404,  405,   -1,  407,   -1,  409,  410,  411,  412,  413,
  414,  415,  416,   -1,   -1,   -1,  420,   -1,  422,   -1,
   -1,   -1,  426,   -1,   -1,   -1,   -1,   -1,   -1,  433,
   -1,   -1,   -1,   -1,  438,  439,  440,  441,  442,  443,
  444,  445,  446,  447,  448,  449,   -1,  451,  452,  453,
  454,   -1,  456,  457,   -1,   -1,   -1,   -1,   -1,   -1,
  464,  465,  466,  467,  468,  469,  470,  471,  472,  473,
  474,  475,  476,   -1,  478,  479,  480,  481,  482,   -1,
  484,  485,  486,   -1,  488,  489,  490,  491,   -1,   -1,
  494,  495,  496,  497,  498,  499,  500,   -1,   -1,   -1,
   -1,  505,  506,  507,   -1,  509,  510,  511,   -1,  513,
   -1,  515,  516,  517,   -1,  519,  520,   -1,  522,  523,
  524,  525,  526,  527,  528,  529,  530,  531,  532,  533,
  534,  535,  536,  537,  538,  539,  540,  541,  542,  543,
   -1,  545,  546,  547,  548,  549,  550,   -1,  552,  553,
  554,  555,   -1,  557,   -1,  559,  560,  561,  562,  563,
  564,  565,  566,  567,  568,  569,  570,  571,  572,  573,
  574,  575,  576,   -1,  578,  579,  580,  581,   -1,  583,
  584,  585,  586,  587,   -1,  589,  590,  591,  592,  593,
  594,  595,  596,  597,  598,  599,  600,  601,   -1,  603,
  604,  605,  606,  607,  608,   -1,  610,  611,  612,  613,
  614,  615,  616,  617,   -1,  619,  620,  621,  622,   -1,
  624,  625,  626,  627,  628,  629,   -1,   -1,  632,  633,
  634,  635,  636,  637,  638,  639,  640,  641,   -1,  643,
  644,  645,  646,  647,  648,  649,  650,  651,  652,  653,
   -1,   -1,  656,  657,  658,  659,  660,  661,  662,  663,
  664,  665,  666,  667,  668,  669,  670,  671,  672,  673,
   -1,  675,  676,  677,  678,  679,  680,  681,  682,  683,
   -1,  685,  686,  687,  688,  689,  690,  691,  692,  693,
  694,  695,  696,   -1,  698,  699,  700,   -1,  702,   -1,
  704,  705,  706,  707,  708,  709,  710,  711,  712,  713,
  714,  715,  716,  717,  718,   -1,  720,  721,  722,  723,
  724,  725,  726,  727,  728,  729,  730,  731,  732,  733,
  734,  735,  736,  737,  738,  739,  740,   -1,   40,  743,
  744,   43,   -1,   45,   -1,  749,  750,  751,  752,  753,
  754,   -1,   -1,   -1,   -1,   -1,   58,   -1,   -1,   -1,
   -1,   63,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  257,   -1,  259,   -1,   -1,
   -1,   -1,   -1,  265,   -1,  267,  268,  269,   -1,   -1,
   -1,   -1,  274,   -1,   -1,   -1,   -1,  279,  280,   -1,
  282,   -1,  284,  285,   -1,  287,   -1,   -1,  290,  291,
  292,   -1,   -1,   -1,   -1,  297,   -1,  299,  300,   -1,
   -1,   -1,  304,   -1,  306,   -1,  308,  309,   -1,   -1,
   -1,   -1,  314,   -1,   -1,  317,   -1,  319,  320,   -1,
   -1,   -1,  324,   -1,  326,   -1,  328,   -1,   -1,   -1,
   -1,   -1,  334,   -1,  336,   -1,   -1,   -1,  340,   -1,
   -1,   -1,  344,   -1,  346,  347,  348,  349,   -1,  351,
   -1,  353,  354,   -1,  356,  357,  358,   -1,   -1,   -1,
   -1,   -1,  364,   -1,   -1,   -1,   -1,   -1,  370,   -1,
   -1,   -1,  374,  375,  376,  377,  378,   -1,  380,   -1,
  382,   -1,   -1,   -1,  386,   -1,  388,  389,   -1,   -1,
  392,  393,   -1,   -1,   -1,  397,   -1,  399,   -1,   -1,
  402,  403,  404,  405,   -1,  407,   -1,  409,  410,  411,
  412,  413,  414,  415,  416,   -1,   -1,   -1,  420,   -1,
  422,   -1,   -1,   -1,  426,   -1,   -1,   -1,   -1,   -1,
   -1,  433,   -1,   -1,   -1,   -1,  438,  439,  440,  441,
  442,  443,  444,  445,  446,  447,  448,  449,   -1,  451,
  452,  453,  454,   -1,  456,  457,   -1,   -1,   -1,   -1,
   -1,   -1,  464,  465,  466,  467,  468,  469,  470,  471,
  472,  473,  474,  475,  476,   -1,  478,  479,  480,  481,
  482,   -1,  484,  485,  486,   -1,  488,  489,  490,  491,
   -1,   -1,  494,  495,  496,  497,  498,  499,  500,   -1,
   -1,   -1,   -1,  505,  506,  507,   -1,  509,  510,  511,
   -1,  513,   -1,  515,  516,  517,   -1,  519,  520,   -1,
  522,  523,  524,  525,  526,  527,  528,  529,  530,  531,
  532,  533,  534,  535,  536,  537,  538,  539,  540,  541,
  542,  543,   -1,  545,  546,  547,  548,  549,  550,   -1,
  552,  553,  554,  555,   -1,  557,   -1,  559,  560,  561,
  562,  563,  564,  565,  566,  567,  568,  569,  570,  571,
  572,  573,  574,  575,  576,   -1,  578,  579,  580,  581,
   -1,  583,  584,  585,  586,  587,   -1,  589,  590,  591,
  592,  593,  594,  595,  596,  597,  598,  599,  600,  601,
   -1,  603,  604,  605,  606,  607,  608,   -1,  610,  611,
  612,  613,  614,  615,  616,  617,   -1,  619,  620,  621,
  622,   -1,  624,  625,  626,  627,  628,  629,   -1,   -1,
  632,  633,  634,  635,  636,  637,  638,  639,  640,  641,
   -1,  643,  644,  645,  646,  647,  648,  649,  650,  651,
  652,  653,   -1,   -1,  656,  657,  658,  659,  660,  661,
  662,  663,  664,  665,  666,  667,  668,  669,  670,  671,
  672,  673,   -1,  675,  676,  677,  678,  679,  680,  681,
  682,  683,   -1,  685,  686,  687,  688,  689,  690,  691,
  692,  693,  694,  695,  696,   -1,  698,  699,  700,   -1,
  702,   -1,  704,  705,  706,  707,  708,  709,  710,  711,
  712,  713,  714,  715,  716,  717,  718,   -1,  720,  721,
  722,  723,  724,  725,  726,  727,  728,  729,  730,  731,
  732,  733,  734,  735,  736,  737,  738,  739,  740,   -1,
   40,  743,  744,   43,   -1,   45,   -1,  749,  750,  751,
  752,  753,  754,   -1,   -1,   -1,   -1,   -1,   58,   -1,
   -1,   -1,   -1,   63,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  257,   -1,  259,
   -1,   -1,   -1,   -1,   -1,  265,   -1,  267,  268,  269,
   -1,   -1,   -1,   -1,  274,   -1,   -1,   -1,   -1,  279,
  280,   -1,  282,   -1,  284,  285,   -1,  287,   -1,   -1,
  290,  291,  292,   -1,   -1,  295,   -1,  297,   -1,  299,
  300,   -1,   -1,   -1,  304,   -1,  306,   -1,  308,  309,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  317,   -1,  319,
  320,   -1,   -1,   -1,  324,   -1,  326,   -1,  328,   -1,
   -1,   -1,   -1,   -1,  334,   -1,  336,   -1,   -1,   -1,
  340,   -1,   -1,   -1,  344,   -1,  346,  347,  348,  349,
   -1,  351,   -1,  353,  354,   -1,  356,  357,  358,   -1,
   -1,   -1,   -1,   -1,  364,   -1,   -1,   -1,   -1,   -1,
  370,   -1,   -1,   -1,  374,  375,  376,  377,  378,   -1,
  380,   -1,  382,   -1,   -1,   -1,  386,   -1,  388,  389,
   -1,   -1,  392,  393,   -1,   -1,   -1,  397,   -1,  399,
   -1,   -1,  402,  403,  404,  405,   -1,  407,   -1,  409,
  410,  411,  412,  413,  414,  415,  416,   -1,   -1,   -1,
  420,   -1,  422,   -1,   -1,   -1,  426,   -1,   -1,   -1,
   -1,   -1,   -1,  433,   -1,   -1,   -1,   -1,  438,  439,
  440,  441,  442,  443,  444,  445,  446,  447,  448,  449,
   -1,  451,  452,  453,  454,   -1,  456,  457,   -1,   -1,
   -1,   -1,   -1,   -1,  464,  465,  466,  467,  468,  469,
  470,  471,  472,  473,  474,  475,  476,   -1,  478,  479,
  480,  481,  482,   -1,  484,  485,  486,   -1,  488,  489,
  490,  491,   -1,   -1,  494,  495,  496,  497,  498,  499,
  500,   -1,   -1,   -1,   -1,  505,  506,  507,   -1,  509,
  510,  511,   -1,  513,   -1,  515,  516,  517,   -1,  519,
  520,   -1,  522,  523,  524,  525,  526,  527,  528,  529,
  530,  531,  532,  533,  534,  535,  536,  537,  538,  539,
  540,  541,  542,  543,   -1,  545,  546,  547,  548,  549,
  550,   -1,  552,  553,  554,  555,   -1,  557,   -1,  559,
  560,  561,  562,  563,  564,  565,  566,  567,  568,  569,
  570,  571,  572,  573,  574,  575,  576,   -1,  578,  579,
  580,  581,   -1,  583,  584,  585,  586,  587,   -1,  589,
  590,  591,  592,  593,  594,  595,  596,  597,  598,  599,
  600,  601,   -1,  603,  604,  605,  606,  607,  608,   -1,
  610,  611,  612,  613,  614,  615,  616,  617,   -1,  619,
  620,  621,  622,   -1,  624,  625,  626,  627,  628,  629,
   -1,   -1,  632,  633,  634,  635,  636,  637,  638,  639,
  640,  641,   -1,  643,  644,  645,  646,  647,  648,  649,
  650,  651,  652,  653,   -1,   -1,  656,  657,  658,  659,
  660,  661,  662,  663,  664,  665,  666,  667,  668,  669,
  670,  671,  672,  673,   -1,  675,  676,  677,  678,  679,
  680,  681,  682,  683,   -1,  685,  686,  687,  688,  689,
  690,  691,  692,  693,  694,  695,  696,   -1,  698,  699,
  700,   -1,  702,   -1,  704,  705,  706,  707,  708,  709,
  710,  711,  712,  713,  714,  715,  716,  717,  718,   -1,
  720,  721,  722,  723,  724,  725,  726,  727,  728,  729,
  730,  731,  732,  733,  734,  735,  736,  737,  738,  739,
  740,   -1,   40,  743,  744,   43,   -1,   45,   -1,  749,
  750,  751,  752,  753,  754,   -1,   -1,   -1,   -1,   -1,
   58,   -1,   -1,   -1,   -1,   63,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  257,
   -1,  259,   -1,   -1,   -1,   -1,   -1,  265,   -1,  267,
  268,  269,   -1,   -1,   -1,   -1,  274,   -1,   -1,   -1,
   -1,  279,  280,   -1,  282,   -1,  284,  285,   -1,  287,
   -1,   -1,  290,  291,  292,   -1,   -1,   -1,   -1,  297,
   -1,  299,  300,   -1,   -1,   -1,  304,   -1,  306,   -1,
  308,  309,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  317,
   -1,  319,  320,   -1,   -1,   -1,  324,   -1,  326,   -1,
  328,   -1,   -1,   -1,   -1,   -1,  334,   -1,  336,   -1,
   -1,   -1,  340,   -1,   -1,   -1,  344,   -1,  346,  347,
  348,  349,   -1,  351,   -1,  353,  354,   -1,  356,  357,
  358,   -1,   -1,   -1,   -1,   -1,  364,   -1,   -1,   -1,
   -1,   -1,  370,   -1,   -1,   -1,  374,  375,  376,  377,
  378,   -1,  380,   -1,  382,   -1,   -1,   -1,  386,   -1,
  388,  389,   -1,   -1,  392,  393,   -1,   -1,   -1,  397,
   -1,  399,   -1,   -1,  402,  403,  404,  405,   -1,  407,
   -1,  409,  410,  411,  412,  413,  414,  415,  416,   -1,
   -1,  419,  420,   -1,  422,   -1,   -1,   -1,  426,   -1,
   -1,   -1,   -1,   -1,   -1,  433,   -1,   -1,   -1,   -1,
  438,  439,  440,  441,  442,  443,  444,  445,  446,  447,
  448,  449,   -1,  451,  452,  453,  454,   -1,  456,  457,
   -1,   -1,   -1,   -1,   -1,   -1,  464,  465,  466,  467,
  468,  469,  470,  471,  472,  473,  474,  475,  476,   -1,
  478,  479,  480,  481,  482,   -1,  484,  485,  486,   -1,
  488,  489,  490,  491,   -1,   -1,  494,  495,  496,  497,
  498,  499,  500,   -1,   -1,   -1,   -1,  505,  506,  507,
   -1,  509,  510,  511,   -1,  513,   -1,  515,  516,  517,
   -1,  519,  520,   -1,  522,  523,  524,  525,  526,  527,
  528,  529,  530,  531,  532,  533,  534,  535,  536,  537,
  538,  539,  540,  541,  542,  543,   -1,  545,  546,  547,
  548,  549,  550,   -1,  552,  553,  554,  555,   -1,  557,
   -1,  559,  560,  561,  562,  563,  564,  565,  566,  567,
  568,  569,  570,  571,  572,  573,  574,  575,  576,   -1,
  578,  579,  580,  581,   -1,  583,  584,  585,  586,  587,
   -1,  589,  590,  591,  592,  593,  594,  595,  596,  597,
  598,  599,  600,  601,   -1,  603,  604,  605,  606,  607,
  608,   -1,  610,  611,  612,  613,  614,  615,  616,  617,
   -1,  619,  620,  621,  622,   -1,  624,  625,  626,  627,
  628,  629,   -1,   -1,  632,  633,  634,  635,  636,  637,
  638,  639,  640,  641,   -1,  643,  644,  645,  646,  647,
  648,  649,  650,  651,  652,  653,   -1,   -1,  656,  657,
  658,  659,  660,  661,  662,  663,  664,  665,  666,  667,
  668,  669,  670,  671,  672,  673,   -1,  675,  676,  677,
  678,  679,  680,  681,  682,  683,   -1,  685,  686,  687,
  688,  689,  690,  691,  692,  693,  694,  695,  696,   -1,
  698,  699,  700,   -1,  702,   -1,  704,  705,  706,  707,
  708,  709,  710,  711,  712,  713,  714,  715,  716,  717,
  718,   -1,  720,  721,  722,  723,  724,  725,  726,  727,
  728,  729,  730,  731,  732,  733,  734,  735,  736,  737,
  738,  739,  740,   -1,   -1,  743,  744,   -1,   -1,   -1,
   -1,  749,  750,  751,  752,  753,  754,   40,   41,   -1,
   43,   -1,   45,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   58,   -1,   -1,   -1,   -1,
   63,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  257,   -1,  259,   -1,   -1,   -1,
   -1,   -1,  265,   -1,  267,  268,  269,   -1,   -1,   -1,
   -1,  274,   -1,   -1,   -1,   -1,  279,  280,   -1,  282,
   -1,  284,  285,   -1,  287,   -1,   -1,  290,  291,  292,
   -1,   -1,   -1,   -1,  297,   -1,  299,  300,   -1,   -1,
   -1,  304,   -1,  306,   -1,  308,  309,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  317,   -1,  319,  320,   -1,   -1,
   -1,  324,   -1,  326,   -1,  328,   -1,   -1,   -1,   -1,
   -1,  334,   -1,  336,   -1,   -1,   -1,  340,   -1,   -1,
   -1,  344,   -1,  346,  347,  348,  349,   -1,  351,   -1,
  353,  354,   -1,  356,  357,  358,   -1,   -1,   -1,   -1,
   -1,  364,   -1,   -1,   -1,   -1,   -1,  370,   -1,   -1,
   -1,  374,  375,  376,  377,  378,   -1,  380,   -1,  382,
   -1,   -1,   -1,  386,   -1,  388,  389,   -1,   -1,  392,
  393,   -1,   -1,   -1,  397,   -1,  399,   -1,   -1,  402,
  403,  404,  405,   -1,  407,   -1,  409,  410,  411,  412,
  413,  414,  415,  416,   -1,   -1,   -1,  420,   -1,  422,
   -1,   -1,   -1,  426,   -1,   -1,   -1,   -1,   -1,   -1,
  433,   -1,   -1,   -1,   -1,  438,  439,  440,  441,  442,
  443,  444,  445,  446,  447,  448,  449,   -1,  451,  452,
  453,  454,   -1,  456,  457,   -1,   -1,   -1,   -1,   -1,
   -1,  464,  465,  466,  467,  468,  469,  470,  471,  472,
  473,  474,  475,  476,   -1,  478,  479,  480,  481,  482,
   -1,  484,  485,  486,   -1,  488,  489,  490,  491,   -1,
   -1,  494,  495,  496,  497,  498,  499,  500,   -1,   -1,
   -1,   -1,  505,  506,  507,   -1,  509,  510,  511,   -1,
  513,   -1,  515,  516,  517,   -1,  519,  520,   -1,  522,
  523,  524,  525,  526,  527,  528,  529,  530,  531,  532,
  533,  534,  535,  536,  537,  538,  539,  540,  541,  542,
  543,   -1,  545,  546,  547,  548,  549,  550,   -1,  552,
  553,  554,  555,   -1,  557,   -1,  559,  560,  561,  562,
  563,  564,  565,  566,  567,  568,  569,  570,  571,  572,
  573,  574,  575,  576,   -1,  578,  579,  580,  581,   -1,
  583,  584,  585,  586,  587,   -1,  589,  590,  591,  592,
  593,  594,  595,  596,  597,  598,  599,  600,  601,   -1,
  603,  604,  605,  606,  607,  608,   -1,  610,  611,  612,
  613,  614,  615,  616,  617,   -1,  619,  620,  621,  622,
   -1,  624,  625,  626,  627,  628,  629,   -1,   -1,  632,
  633,  634,  635,  636,  637,  638,  639,  640,  641,   -1,
  643,  644,  645,  646,  647,  648,  649,  650,  651,  652,
  653,   -1,   -1,  656,  657,  658,  659,  660,  661,  662,
  663,  664,  665,  666,  667,  668,  669,  670,  671,  672,
  673,   -1,  675,  676,  677,  678,  679,  680,  681,  682,
  683,   -1,  685,  686,  687,  688,  689,  690,  691,  692,
  693,  694,  695,  696,   -1,  698,  699,  700,   -1,  702,
   -1,  704,  705,  706,  707,  708,  709,  710,  711,  712,
  713,  714,  715,  716,  717,  718,   -1,  720,  721,  722,
  723,  724,  725,  726,  727,  728,  729,  730,  731,  732,
  733,  734,  735,  736,  737,  738,  739,  740,   -1,   40,
  743,  744,   43,   -1,   45,   -1,  749,  750,  751,  752,
  753,  754,   -1,   -1,   -1,   -1,   -1,   58,   -1,   -1,
   -1,   -1,   63,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  257,   -1,  259,   -1,
   -1,   -1,   -1,   -1,  265,   -1,  267,  268,  269,   -1,
   -1,   -1,   -1,  274,   -1,   -1,   -1,   -1,  279,  280,
   -1,  282,   -1,  284,  285,   -1,  287,   -1,   -1,  290,
  291,  292,   -1,   -1,  295,   -1,  297,   -1,  299,  300,
   -1,   -1,   -1,  304,   -1,  306,   -1,  308,  309,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  317,   -1,  319,  320,
   -1,   -1,   -1,  324,   -1,  326,   -1,  328,   -1,   -1,
   -1,   -1,   -1,  334,   -1,  336,   -1,   -1,   -1,  340,
   -1,   -1,   -1,  344,   -1,  346,  347,  348,  349,   -1,
  351,   -1,  353,  354,   -1,  356,  357,  358,   -1,   -1,
   -1,   -1,   -1,  364,   -1,   -1,   -1,   -1,   -1,  370,
   -1,   -1,   -1,  374,  375,  376,  377,  378,   -1,  380,
   -1,  382,   -1,   -1,   -1,  386,   -1,  388,  389,   -1,
   -1,  392,  393,   -1,   -1,   -1,  397,   -1,  399,   -1,
   -1,  402,  403,  404,  405,   -1,  407,   -1,  409,  410,
  411,  412,  413,  414,  415,  416,   -1,   -1,   -1,  420,
   -1,  422,   -1,   -1,   -1,  426,   -1,   -1,   -1,   -1,
   -1,   -1,  433,   -1,   -1,   -1,   -1,  438,  439,  440,
  441,  442,  443,  444,  445,  446,  447,  448,  449,   -1,
  451,  452,  453,  454,   -1,  456,  457,   -1,   -1,   -1,
   -1,   -1,   -1,  464,  465,  466,  467,  468,  469,  470,
  471,  472,  473,  474,  475,  476,   -1,  478,  479,  480,
  481,  482,   -1,  484,  485,  486,   -1,  488,  489,  490,
  491,   -1,   -1,  494,  495,  496,  497,  498,  499,  500,
   -1,   -1,   -1,   -1,  505,  506,  507,   -1,  509,  510,
  511,   -1,  513,   -1,  515,  516,  517,   -1,  519,  520,
   -1,  522,  523,  524,  525,  526,  527,  528,  529,  530,
  531,  532,  533,  534,  535,  536,  537,  538,  539,  540,
  541,  542,  543,   -1,  545,  546,  547,  548,  549,  550,
   -1,  552,  553,  554,  555,   -1,  557,   -1,  559,  560,
  561,  562,  563,  564,  565,  566,  567,  568,  569,  570,
  571,  572,  573,  574,  575,  576,   -1,  578,  579,  580,
  581,   -1,  583,  584,  585,  586,  587,   -1,  589,  590,
  591,  592,  593,  594,  595,  596,  597,  598,  599,  600,
  601,   -1,  603,  604,  605,  606,  607,  608,   -1,  610,
  611,  612,  613,  614,  615,  616,  617,   -1,  619,  620,
  621,  622,   -1,  624,  625,  626,  627,  628,  629,   -1,
   -1,  632,  633,  634,  635,  636,  637,  638,  639,  640,
  641,   -1,  643,  644,  645,  646,  647,  648,  649,  650,
  651,  652,  653,   -1,   -1,  656,  657,  658,  659,  660,
  661,  662,  663,  664,  665,  666,  667,  668,  669,  670,
  671,  672,  673,   -1,  675,  676,  677,  678,  679,  680,
  681,  682,  683,   -1,  685,  686,  687,  688,  689,  690,
  691,  692,  693,  694,  695,  696,   -1,  698,  699,  700,
   -1,  702,   -1,  704,  705,  706,  707,  708,  709,  710,
  711,  712,  713,  714,  715,  716,  717,  718,   -1,  720,
  721,  722,  723,  724,  725,  726,  727,  728,  729,  730,
  731,  732,  733,  734,  735,  736,  737,  738,  739,  740,
   -1,   40,  743,  744,   43,   -1,   45,   -1,  749,  750,
  751,  752,  753,  754,   -1,   -1,   -1,   -1,   -1,   58,
   -1,   -1,   -1,   -1,   63,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  257,   -1,
  259,   -1,   -1,   -1,   -1,   -1,  265,   -1,  267,  268,
  269,   -1,   -1,   -1,   -1,  274,   -1,   -1,   -1,   -1,
  279,  280,   -1,  282,   -1,  284,  285,   -1,  287,   -1,
   -1,  290,  291,  292,   -1,   -1,   -1,   -1,  297,   -1,
  299,  300,   -1,   -1,   -1,  304,   -1,  306,   -1,  308,
  309,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  317,   -1,
  319,  320,   -1,   -1,   -1,  324,   -1,  326,   -1,  328,
   -1,   -1,   -1,   -1,   -1,  334,   -1,  336,   -1,   -1,
   -1,  340,   -1,   -1,   -1,  344,   -1,  346,  347,  348,
  349,   -1,  351,   -1,  353,  354,   -1,  356,  357,  358,
   -1,   -1,   -1,   -1,   -1,  364,   -1,   -1,   -1,   -1,
   -1,  370,   -1,   -1,   -1,  374,  375,  376,  377,  378,
   -1,  380,   -1,  382,   -1,   -1,   -1,  386,   -1,  388,
  389,   -1,   -1,  392,  393,   -1,   -1,   -1,  397,   -1,
  399,   -1,   -1,  402,  403,  404,  405,   -1,  407,   -1,
  409,  410,  411,  412,  413,  414,  415,  416,   -1,   -1,
   -1,  420,   -1,  422,   -1,   -1,   -1,  426,   -1,   -1,
   -1,   -1,   -1,   -1,  433,   -1,   -1,   -1,   -1,  438,
  439,  440,  441,  442,  443,  444,  445,  446,  447,  448,
  449,   -1,  451,  452,  453,  454,   -1,  456,  457,   -1,
   -1,   -1,   -1,   -1,   -1,  464,  465,  466,  467,  468,
  469,  470,  471,  472,  473,  474,  475,  476,   -1,  478,
  479,  480,  481,  482,   -1,  484,  485,  486,  487,  488,
  489,  490,  491,   -1,   -1,  494,  495,  496,  497,  498,
  499,  500,   -1,   -1,   -1,   -1,  505,  506,  507,   -1,
  509,  510,  511,   -1,  513,   -1,  515,  516,  517,   -1,
  519,  520,   -1,  522,  523,  524,  525,  526,  527,  528,
  529,  530,  531,  532,  533,  534,  535,  536,  537,  538,
  539,  540,  541,  542,  543,   -1,  545,  546,  547,  548,
  549,  550,   -1,  552,  553,  554,  555,   -1,  557,   -1,
  559,  560,  561,  562,  563,  564,  565,  566,  567,  568,
  569,  570,  571,  572,  573,  574,  575,  576,   -1,  578,
  579,  580,  581,   -1,  583,  584,  585,  586,  587,   -1,
  589,  590,  591,  592,  593,  594,  595,  596,  597,  598,
  599,  600,  601,   -1,  603,  604,  605,  606,  607,  608,
   -1,  610,  611,  612,  613,  614,  615,  616,  617,   -1,
  619,  620,  621,  622,   -1,  624,  625,  626,  627,  628,
  629,   -1,   -1,  632,  633,  634,  635,  636,  637,  638,
  639,  640,  641,   -1,  643,  644,  645,  646,  647,  648,
  649,  650,  651,  652,  653,   -1,   -1,  656,  657,  658,
  659,  660,  661,  662,  663,  664,  665,  666,  667,  668,
  669,  670,  671,  672,  673,   -1,  675,  676,  677,  678,
  679,  680,  681,  682,  683,   -1,  685,  686,  687,  688,
  689,  690,  691,  692,  693,  694,  695,  696,   -1,  698,
  699,  700,   -1,  702,   -1,  704,  705,  706,  707,  708,
  709,  710,  711,  712,  713,  714,  715,  716,  717,  718,
   -1,  720,  721,  722,  723,  724,  725,  726,  727,  728,
  729,  730,  731,  732,  733,  734,  735,  736,  737,  738,
  739,  740,   -1,   -1,  743,  744,   -1,   -1,   -1,   -1,
  749,  750,  751,  752,  753,  754,   40,   -1,   42,   43,
   -1,   45,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   58,   -1,   -1,   -1,   -1,   63,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  257,   -1,  259,  260,   -1,   -1,   -1,
   -1,  265,   -1,  267,  268,  269,   -1,   -1,   -1,   -1,
  274,   -1,   -1,   -1,   -1,  279,  280,   -1,  282,   -1,
  284,  285,   -1,  287,   -1,   -1,  290,  291,  292,   -1,
   -1,   -1,   -1,  297,  298,  299,  300,   -1,   -1,   -1,
  304,   -1,  306,   -1,  308,  309,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  317,   -1,  319,  320,   -1,   -1,   -1,
  324,   -1,  326,   -1,  328,   -1,   -1,   -1,   -1,   -1,
  334,   -1,  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,
  344,   -1,  346,  347,  348,  349,   -1,  351,   -1,  353,
  354,   -1,  356,  357,  358,   -1,   -1,   -1,   -1,   -1,
  364,   -1,   -1,   -1,   -1,   -1,  370,   -1,   -1,   -1,
  374,  375,  376,  377,  378,   -1,  380,   -1,  382,   -1,
   -1,   -1,  386,   -1,  388,  389,   -1,   -1,  392,  393,
   -1,   -1,   -1,  397,   -1,  399,   -1,   -1,  402,  403,
  404,  405,   -1,  407,   -1,  409,  410,  411,  412,  413,
  414,  415,  416,   -1,   -1,   -1,  420,   -1,  422,   -1,
   -1,   -1,  426,   -1,   -1,   -1,   -1,   -1,   -1,  433,
   -1,   -1,   -1,   -1,  438,  439,  440,  441,  442,  443,
  444,  445,  446,  447,  448,  449,   -1,  451,  452,  453,
  454,   -1,  456,  457,   -1,   -1,   -1,   -1,   -1,   -1,
  464,  465,  466,  467,  468,  469,  470,  471,  472,  473,
  474,  475,  476,   -1,  478,   -1,   -1,  481,  482,   -1,
  484,  485,  486,   -1,  488,  489,  490,  491,   -1,   -1,
  494,  495,  496,  497,  498,  499,  500,   -1,   -1,   -1,
   -1,  505,  506,  507,   -1,  509,  510,  511,   -1,  513,
   -1,  515,  516,  517,   -1,  519,  520,   -1,  522,  523,
  524,  525,  526,  527,  528,  529,  530,  531,  532,  533,
  534,  535,  536,  537,  538,  539,  540,  541,  542,  543,
   -1,  545,  546,  547,  548,  549,  550,   -1,  552,  553,
  554,  555,   -1,  557,   -1,  559,  560,  561,  562,  563,
  564,  565,  566,  567,  568,  569,  570,  571,  572,  573,
  574,  575,  576,   -1,  578,  579,  580,  581,   -1,  583,
  584,  585,  586,  587,   -1,  589,  590,  591,  592,  593,
  594,  595,  596,  597,  598,  599,  600,  601,   -1,  603,
  604,  605,  606,  607,  608,   -1,  610,  611,  612,  613,
  614,  615,  616,  617,   -1,  619,  620,  621,  622,   -1,
  624,  625,  626,  627,  628,  629,   -1,   -1,  632,  633,
  634,  635,  636,  637,  638,  639,  640,  641,   -1,  643,
  644,  645,  646,  647,  648,  649,  650,  651,  652,  653,
   -1,   -1,  656,  657,  658,  659,  660,  661,  662,  663,
  664,  665,  666,  667,  668,  669,  670,  671,  672,  673,
   -1,  675,  676,  677,  678,  679,  680,  681,  682,  683,
   -1,  685,  686,  687,  688,  689,  690,  691,  692,  693,
  694,  695,  696,   -1,  698,  699,  700,   -1,  702,   -1,
  704,  705,  706,  707,  708,  709,  710,  711,  712,  713,
  714,  715,  716,  717,  718,   -1,  720,  721,  722,  723,
  724,  725,  726,  727,  728,  729,  730,  731,  732,  733,
  734,  735,  736,  737,  738,  739,  740,   -1,   -1,  743,
  744,   -1,   -1,   -1,   -1,  749,  750,  751,  752,  753,
  754,   40,   -1,   42,   43,   -1,   45,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   58,
   -1,   -1,   -1,   -1,   63,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  257,   -1,
  259,   -1,   -1,   -1,   -1,   -1,  265,   -1,  267,  268,
  269,   -1,   -1,   -1,   -1,  274,   -1,   -1,   -1,   -1,
  279,  280,   -1,  282,   -1,  284,  285,   -1,  287,   -1,
   -1,  290,  291,  292,   -1,   -1,   -1,   -1,  297,   -1,
  299,  300,   -1,   -1,   -1,  304,   -1,  306,   -1,  308,
  309,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  317,   -1,
  319,  320,   -1,   -1,   -1,  324,   -1,  326,   -1,  328,
   -1,   -1,   -1,   -1,   -1,  334,   -1,  336,   -1,   -1,
   -1,  340,   -1,   -1,   -1,  344,   -1,  346,  347,  348,
  349,   -1,  351,   -1,  353,  354,   -1,  356,  357,  358,
   -1,   -1,   -1,   -1,   -1,  364,   -1,   -1,   -1,   -1,
   -1,  370,   -1,   -1,   -1,  374,  375,  376,  377,  378,
   -1,  380,   -1,  382,   -1,   -1,   -1,  386,   -1,  388,
  389,   -1,   -1,  392,  393,   -1,   -1,   -1,  397,   -1,
  399,   -1,   -1,  402,  403,  404,  405,   -1,  407,   -1,
  409,  410,  411,  412,  413,  414,  415,  416,   -1,   -1,
   -1,  420,   -1,  422,   -1,   -1,   -1,  426,   -1,   -1,
   -1,   -1,   -1,   -1,  433,   -1,   -1,   -1,   -1,  438,
  439,  440,  441,  442,  443,  444,  445,  446,  447,  448,
  449,   -1,  451,  452,  453,  454,   -1,  456,  457,   -1,
   -1,   -1,   -1,   -1,   -1,  464,  465,  466,  467,  468,
  469,  470,  471,  472,  473,  474,  475,  476,   -1,  478,
  479,  480,  481,  482,   -1,  484,  485,  486,   -1,  488,
  489,  490,  491,   -1,   -1,  494,  495,  496,  497,  498,
  499,  500,   -1,   -1,   -1,   -1,  505,  506,  507,   -1,
  509,  510,  511,   -1,  513,   -1,  515,  516,  517,   -1,
  519,  520,   -1,  522,  523,  524,  525,  526,  527,  528,
  529,  530,  531,  532,  533,  534,  535,  536,  537,  538,
  539,  540,  541,  542,  543,   -1,  545,  546,  547,  548,
  549,  550,   -1,  552,  553,  554,  555,   -1,  557,   -1,
  559,  560,  561,  562,  563,  564,  565,  566,  567,  568,
  569,  570,  571,  572,  573,  574,  575,  576,   -1,  578,
  579,  580,  581,   -1,  583,  584,  585,  586,  587,   -1,
  589,  590,  591,  592,  593,  594,  595,  596,  597,  598,
  599,  600,  601,   -1,  603,  604,  605,  606,  607,  608,
   -1,  610,  611,  612,  613,  614,  615,  616,  617,   -1,
  619,  620,  621,  622,   -1,  624,  625,  626,  627,  628,
  629,   -1,   -1,  632,  633,  634,  635,  636,  637,  638,
  639,  640,  641,   -1,  643,  644,  645,  646,  647,  648,
  649,  650,  651,  652,  653,   -1,   -1,  656,  657,  658,
  659,  660,  661,  662,  663,  664,  665,  666,  667,  668,
  669,  670,  671,  672,  673,   -1,  675,  676,  677,  678,
  679,  680,  681,  682,  683,   -1,  685,  686,  687,  688,
  689,  690,  691,  692,  693,  694,  695,  696,   -1,  698,
  699,  700,   -1,  702,   -1,  704,  705,  706,  707,  708,
  709,  710,  711,  712,  713,  714,  715,  716,  717,  718,
   -1,  720,  721,  722,  723,  724,  725,  726,  727,  728,
  729,  730,  731,  732,  733,  734,  735,  736,  737,  738,
  739,  740,   -1,   40,  743,  744,   43,   -1,   45,   -1,
  749,  750,  751,  752,  753,  754,   -1,   -1,   -1,   -1,
   -1,   58,   -1,   -1,   -1,   -1,   63,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  257,   -1,  259,   -1,   -1,   -1,   -1,   -1,  265,   -1,
  267,  268,  269,   -1,   -1,   -1,   -1,  274,   -1,   -1,
   -1,   -1,  279,  280,   -1,  282,   -1,  284,  285,   -1,
  287,   -1,   -1,  290,  291,  292,   -1,   -1,   -1,   -1,
  297,   -1,  299,  300,   -1,   -1,   -1,  304,   -1,  306,
   -1,  308,  309,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  317,   -1,  319,  320,   -1,   -1,   -1,  324,   -1,  326,
   -1,  328,   -1,   -1,   -1,   -1,   -1,  334,   -1,  336,
   -1,   -1,   -1,  340,   -1,   -1,   -1,  344,   -1,  346,
  347,  348,  349,   -1,  351,   -1,  353,  354,   -1,  356,
  357,  358,   -1,   -1,   -1,   -1,   -1,  364,   -1,   -1,
   -1,   -1,   -1,  370,   -1,   -1,   -1,  374,  375,  376,
  377,  378,   -1,  380,   -1,  382,   -1,   -1,   -1,  386,
   -1,  388,  389,   -1,   -1,  392,  393,   -1,   -1,   -1,
  397,   -1,  399,   -1,   -1,  402,  403,  404,  405,   -1,
  407,   -1,  409,  410,  411,  412,  413,  414,  415,  416,
   -1,   -1,   -1,  420,   -1,  422,   -1,   -1,   -1,  426,
   -1,   -1,   -1,   -1,   -1,   -1,  433,   -1,   -1,   -1,
   -1,  438,  439,  440,  441,  442,  443,  444,  445,  446,
  447,  448,  449,   -1,  451,  452,  453,  454,   -1,  456,
  457,   -1,   -1,   -1,   -1,   -1,   -1,  464,  465,  466,
  467,  468,  469,  470,  471,  472,  473,  474,  475,  476,
   -1,  478,  479,  480,  481,  482,   -1,  484,  485,  486,
   -1,  488,  489,  490,  491,   -1,   -1,  494,  495,  496,
  497,  498,  499,  500,   -1,   -1,   -1,   -1,  505,  506,
  507,   -1,  509,  510,  511,   -1,  513,   -1,  515,  516,
  517,   -1,  519,  520,   -1,  522,  523,  524,  525,  526,
  527,  528,  529,  530,  531,  532,  533,  534,  535,  536,
  537,  538,  539,  540,  541,  542,  543,   -1,  545,  546,
  547,  548,  549,  550,   -1,  552,  553,  554,  555,   -1,
  557,   -1,  559,  560,  561,  562,  563,  564,  565,  566,
  567,  568,  569,  570,  571,  572,  573,  574,  575,  576,
   -1,  578,  579,  580,  581,   -1,  583,  584,  585,  586,
  587,   -1,  589,  590,  591,  592,  593,  594,  595,  596,
  597,  598,  599,  600,  601,   -1,  603,  604,  605,  606,
  607,  608,   -1,  610,  611,  612,  613,  614,  615,  616,
  617,   -1,  619,  620,  621,  622,   -1,  624,  625,  626,
  627,  628,  629,   -1,   -1,  632,  633,  634,  635,  636,
  637,  638,  639,  640,  641,   -1,  643,  644,  645,  646,
  647,  648,  649,  650,  651,  652,  653,   -1,   -1,  656,
  657,  658,  659,  660,  661,  662,  663,  664,  665,  666,
  667,  668,  669,  670,  671,  672,  673,   -1,  675,  676,
  677,  678,  679,  680,  681,  682,  683,   -1,  685,  686,
  687,  688,  689,  690,  691,  692,  693,  694,  695,  696,
   -1,  698,  699,  700,   -1,  702,   -1,  704,  705,  706,
  707,  708,  709,  710,  711,  712,  713,  714,  715,  716,
  717,  718,   -1,  720,  721,  722,  723,  724,  725,  726,
  727,  728,  729,  730,  731,  732,  733,  734,  735,  736,
  737,  738,  739,  740,   -1,   40,  743,  744,   43,   -1,
   45,   -1,  749,  750,  751,  752,  753,  754,   -1,   -1,
   -1,   -1,   -1,   58,   -1,   -1,   -1,   -1,   63,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  257,   -1,  259,   -1,   -1,   -1,   -1,   -1,
  265,   -1,  267,  268,  269,   -1,   -1,   -1,   -1,  274,
   -1,   -1,   -1,   -1,  279,  280,   -1,  282,   -1,  284,
  285,   -1,  287,   -1,   -1,  290,  291,  292,   -1,   -1,
   -1,   -1,  297,   -1,  299,  300,   -1,   -1,   -1,  304,
   -1,  306,   -1,  308,  309,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  317,   -1,  319,  320,   -1,   -1,   -1,  324,
   -1,  326,   -1,  328,   -1,   -1,   -1,   -1,   -1,  334,
   -1,  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,  344,
   -1,  346,  347,  348,  349,   -1,  351,   -1,  353,  354,
   -1,  356,  357,  358,   -1,   -1,   -1,   -1,   -1,  364,
   -1,   -1,   -1,   -1,   -1,  370,   -1,   -1,   -1,  374,
  375,  376,  377,  378,   -1,  380,   -1,  382,   -1,   -1,
   -1,  386,   -1,  388,  389,   -1,   -1,  392,  393,   -1,
   -1,   -1,  397,   -1,  399,   -1,   -1,  402,  403,  404,
  405,   -1,  407,   -1,  409,  410,  411,  412,  413,  414,
  415,  416,   -1,   -1,   -1,  420,   -1,  422,   -1,   -1,
   -1,  426,   -1,   -1,   -1,   -1,   -1,   -1,  433,   -1,
   -1,   -1,   -1,  438,  439,  440,  441,  442,  443,  444,
  445,  446,  447,  448,  449,   -1,  451,  452,  453,  454,
   -1,  456,  457,   -1,   -1,   -1,   -1,   -1,   -1,  464,
  465,  466,  467,  468,  469,  470,  471,  472,  473,  474,
  475,  476,   -1,  478,  479,  480,  481,  482,   -1,  484,
  485,  486,   -1,  488,  489,  490,  491,   -1,   -1,  494,
  495,  496,  497,  498,  499,  500,   -1,   -1,   -1,   -1,
  505,  506,  507,   -1,  509,  510,  511,   -1,  513,   -1,
  515,  516,  517,   -1,  519,  520,   -1,  522,  523,  524,
  525,  526,  527,  528,  529,  530,  531,  532,  533,  534,
  535,  536,  537,  538,  539,  540,  541,  542,  543,   -1,
  545,  546,  547,  548,  549,  550,   -1,  552,  553,  554,
  555,   -1,  557,   -1,  559,  560,  561,  562,  563,  564,
  565,  566,  567,  568,  569,  570,  571,  572,  573,  574,
  575,  576,   -1,  578,  579,  580,  581,   -1,  583,  584,
  585,  586,  587,   -1,  589,  590,  591,  592,  593,  594,
  595,  596,  597,  598,  599,  600,  601,   -1,  603,  604,
  605,  606,  607,  608,   -1,  610,  611,  612,  613,  614,
  615,  616,  617,   -1,  619,  620,  621,  622,   -1,  624,
  625,  626,  627,  628,  629,   -1,   -1,  632,  633,  634,
  635,  636,  637,  638,  639,  640,  641,   -1,  643,  644,
  645,  646,  647,  648,  649,  650,  651,  652,  653,   -1,
   -1,  656,  657,  658,  659,  660,  661,  662,  663,  664,
  665,  666,  667,  668,  669,  670,  671,  672,  673,   -1,
  675,  676,  677,  678,  679,  680,  681,  682,  683,   -1,
  685,  686,  687,  688,  689,  690,  691,  692,  693,  694,
  695,  696,   -1,  698,  699,  700,   -1,  702,   -1,  704,
  705,  706,  707,  708,  709,  710,  711,  712,  713,  714,
  715,  716,  717,  718,   -1,  720,  721,  722,  723,  724,
  725,  726,  727,  728,  729,  730,  731,  732,  733,  734,
  735,  736,  737,  738,  739,  740,   -1,   40,  743,  744,
   43,   -1,   45,   -1,  749,  750,  751,  752,  753,  754,
   -1,   -1,   -1,   -1,   -1,   58,   -1,   -1,   -1,   -1,
   63,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  257,   -1,  259,   -1,   -1,   -1,
   -1,   -1,  265,   -1,  267,  268,  269,   -1,   -1,   -1,
   -1,  274,   -1,   -1,   -1,   -1,  279,  280,   -1,  282,
   -1,  284,  285,   -1,  287,   -1,   -1,  290,  291,  292,
   -1,   -1,   -1,   -1,  297,   -1,  299,  300,   -1,   -1,
   -1,  304,   -1,  306,   -1,  308,  309,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  317,   -1,  319,  320,   -1,   -1,
   -1,  324,   -1,  326,   -1,  328,   -1,   -1,   -1,   -1,
   -1,  334,   -1,  336,   -1,   -1,   -1,  340,   -1,   -1,
   -1,  344,   -1,  346,  347,  348,  349,   -1,  351,   -1,
  353,  354,   -1,  356,  357,  358,   -1,   -1,   -1,   -1,
   -1,  364,   -1,   -1,   -1,   -1,   -1,  370,   -1,   -1,
   -1,  374,  375,  376,  377,  378,   -1,  380,   -1,  382,
   -1,   -1,   -1,  386,   -1,  388,  389,   -1,   -1,  392,
  393,   -1,   -1,   -1,  397,   -1,  399,   -1,   -1,  402,
  403,  404,  405,   -1,  407,   -1,  409,  410,  411,  412,
  413,  414,  415,  416,   -1,   -1,   -1,  420,   -1,  422,
   -1,   -1,   -1,  426,   -1,   -1,   -1,   -1,   -1,   -1,
  433,   -1,   -1,   -1,   -1,  438,  439,  440,  441,  442,
  443,  444,  445,  446,  447,  448,  449,   -1,  451,  452,
  453,  454,   -1,  456,  457,   -1,   -1,   -1,   -1,   -1,
   -1,  464,  465,  466,  467,  468,  469,  470,  471,  472,
  473,  474,  475,  476,   -1,  478,  479,  480,  481,  482,
   -1,  484,  485,  486,   -1,  488,  489,  490,  491,   -1,
   -1,  494,  495,  496,  497,  498,  499,  500,   -1,   -1,
   -1,   -1,  505,  506,  507,   -1,  509,  510,  511,   -1,
  513,   -1,  515,  516,  517,   -1,  519,  520,   -1,  522,
  523,  524,  525,  526,  527,  528,  529,  530,  531,  532,
  533,  534,  535,  536,  537,  538,  539,  540,  541,  542,
  543,   -1,  545,  546,  547,  548,  549,  550,   -1,  552,
  553,  554,  555,   -1,  557,   -1,  559,  560,  561,  562,
  563,  564,  565,  566,  567,  568,  569,  570,  571,  572,
  573,  574,  575,  576,   -1,  578,  579,  580,  581,   -1,
  583,  584,  585,  586,  587,   -1,  589,  590,  591,  592,
  593,  594,  595,  596,  597,  598,  599,  600,  601,   -1,
  603,  604,  605,  606,  607,  608,   -1,  610,  611,  612,
  613,  614,  615,  616,  617,   -1,  619,  620,  621,  622,
   -1,  624,  625,  626,  627,  628,  629,   -1,   -1,  632,
  633,  634,  635,  636,  637,  638,  639,  640,  641,   -1,
  643,  644,  645,  646,  647,  648,  649,  650,  651,  652,
  653,   -1,   -1,  656,  657,  658,  659,  660,  661,  662,
  663,  664,  665,  666,  667,  668,  669,  670,  671,  672,
  673,   -1,  675,  676,  677,  678,  679,  680,  681,  682,
  683,   -1,  685,  686,  687,  688,  689,  690,  691,  692,
  693,  694,  695,  696,   -1,  698,  699,  700,   -1,  702,
   -1,  704,  705,  706,  707,  708,  709,  710,  711,  712,
  713,  714,  715,  716,  717,  718,   -1,  720,  721,  722,
  723,  724,  725,  726,  727,  728,  729,  730,  731,  732,
  733,  734,  735,  736,  737,  738,  739,  740,   -1,   40,
  743,  744,   43,   -1,   45,   -1,  749,  750,  751,  752,
  753,  754,   -1,   -1,   -1,   -1,   -1,   58,   -1,   -1,
   -1,   -1,   63,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  257,   -1,  259,   -1,
   -1,   -1,   -1,   -1,  265,   -1,  267,  268,  269,   -1,
   -1,   -1,   -1,  274,   -1,   -1,   -1,   -1,  279,  280,
   -1,  282,   -1,  284,  285,   -1,  287,   -1,   -1,  290,
  291,  292,   -1,   -1,   -1,   -1,  297,   -1,  299,  300,
   -1,   -1,   -1,  304,   -1,  306,   -1,  308,  309,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  317,   -1,  319,  320,
   -1,   -1,   -1,  324,   -1,  326,   -1,  328,   -1,   -1,
   -1,   -1,   -1,  334,   -1,  336,   -1,   -1,   -1,  340,
   -1,   -1,   -1,  344,   -1,  346,  347,  348,  349,   -1,
  351,   -1,  353,  354,   -1,  356,  357,  358,   -1,   -1,
   -1,   -1,   -1,  364,   -1,   -1,   -1,   -1,   -1,  370,
   -1,   -1,   -1,  374,  375,  376,  377,  378,   -1,  380,
   -1,  382,   -1,   -1,   -1,  386,   -1,  388,  389,   -1,
   -1,  392,  393,   -1,   -1,   -1,  397,   -1,  399,   -1,
   -1,  402,  403,  404,  405,   -1,  407,   -1,  409,  410,
  411,  412,  413,  414,  415,  416,   -1,   -1,   -1,  420,
   -1,  422,   -1,   -1,   -1,  426,   -1,   -1,   -1,   -1,
   -1,   -1,  433,   -1,   -1,   -1,   -1,  438,  439,  440,
  441,  442,  443,  444,  445,  446,  447,  448,  449,   -1,
  451,  452,  453,  454,   -1,  456,  457,   -1,   -1,   -1,
   -1,   -1,   -1,  464,  465,  466,  467,  468,  469,  470,
  471,  472,  473,  474,  475,  476,   -1,  478,  479,  480,
  481,  482,   -1,  484,  485,  486,   -1,  488,  489,  490,
  491,   -1,   -1,  494,  495,  496,  497,  498,  499,  500,
   -1,   -1,   -1,   -1,  505,  506,  507,   -1,  509,  510,
  511,   -1,  513,   -1,  515,  516,  517,   -1,  519,  520,
   -1,  522,  523,  524,  525,  526,  527,  528,  529,  530,
  531,  532,  533,  534,  535,  536,  537,  538,  539,  540,
  541,  542,  543,   -1,  545,  546,  547,  548,  549,  550,
   -1,  552,  553,  554,  555,   -1,  557,   -1,  559,  560,
  561,  562,  563,  564,  565,  566,  567,  568,  569,  570,
  571,  572,  573,  574,  575,  576,   -1,  578,  579,  580,
  581,   -1,  583,  584,  585,  586,  587,   -1,  589,  590,
  591,  592,  593,  594,  595,  596,  597,  598,  599,  600,
  601,   -1,  603,  604,  605,  606,  607,  608,   -1,  610,
  611,  612,  613,  614,  615,  616,  617,   -1,  619,  620,
  621,  622,   -1,  624,  625,  626,  627,  628,  629,   -1,
   -1,  632,  633,  634,  635,  636,  637,  638,  639,  640,
  641,   -1,  643,  644,  645,  646,  647,  648,  649,  650,
  651,  652,  653,   -1,   -1,  656,  657,  658,  659,  660,
  661,  662,  663,  664,  665,  666,  667,  668,  669,  670,
  671,  672,  673,   -1,  675,  676,  677,  678,  679,  680,
  681,  682,  683,   -1,  685,  686,  687,  688,  689,  690,
  691,  692,  693,  694,  695,  696,   -1,  698,  699,  700,
   -1,  702,   -1,  704,  705,  706,  707,  708,  709,  710,
  711,  712,  713,  714,  715,  716,  717,  718,   -1,  720,
  721,  722,  723,  724,  725,  726,  727,  728,  729,  730,
  731,  732,  733,  734,  735,  736,  737,  738,  739,  740,
   -1,   40,  743,  744,   43,   -1,   45,   -1,  749,  750,
  751,  752,  753,  754,   -1,   -1,   -1,   -1,   -1,   58,
   -1,   -1,   -1,   -1,   63,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  257,   -1,
  259,   -1,   -1,   -1,   -1,   -1,  265,   -1,  267,  268,
  269,   -1,   -1,   -1,   -1,  274,   -1,   -1,   -1,   -1,
  279,  280,   -1,  282,   -1,  284,  285,   -1,  287,   -1,
   -1,  290,  291,  292,   -1,   -1,   -1,   -1,  297,   -1,
  299,  300,   -1,   -1,   -1,  304,   -1,  306,   -1,  308,
  309,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  317,   -1,
  319,  320,   -1,   -1,   -1,  324,   -1,  326,   -1,  328,
   -1,   -1,   -1,   -1,   -1,  334,   -1,  336,   -1,   -1,
   -1,  340,   -1,   -1,   -1,  344,   -1,  346,  347,  348,
  349,   -1,  351,   -1,  353,  354,   -1,  356,  357,  358,
   -1,   -1,   -1,   -1,   -1,  364,   -1,   -1,   -1,   -1,
   -1,  370,   -1,   -1,   -1,  374,  375,  376,  377,  378,
   -1,  380,   -1,  382,   -1,   -1,   -1,  386,   -1,  388,
  389,   -1,   -1,  392,  393,   -1,   -1,   -1,  397,   -1,
  399,   -1,   -1,  402,  403,  404,  405,   -1,  407,   -1,
  409,  410,  411,  412,  413,  414,  415,  416,   -1,   -1,
   -1,  420,   -1,  422,   -1,   -1,   -1,  426,   -1,   -1,
   -1,   -1,   -1,   -1,  433,   -1,   -1,   -1,   -1,  438,
  439,  440,  441,  442,  443,  444,  445,  446,  447,  448,
  449,   -1,  451,  452,  453,  454,   -1,  456,  457,   -1,
   -1,   -1,   -1,   -1,   -1,  464,  465,  466,  467,  468,
  469,  470,  471,  472,  473,  474,  475,  476,   -1,  478,
  479,  480,  481,  482,   -1,  484,  485,  486,   -1,  488,
  489,  490,  491,   -1,   -1,  494,  495,  496,  497,  498,
  499,  500,   -1,   -1,   -1,   -1,  505,  506,  507,   -1,
  509,  510,  511,   -1,  513,   -1,  515,  516,  517,   -1,
  519,  520,   -1,  522,  523,  524,  525,  526,  527,  528,
  529,  530,  531,  532,  533,  534,  535,  536,  537,  538,
  539,  540,  541,  542,  543,   -1,  545,  546,  547,  548,
  549,  550,   -1,  552,  553,  554,  555,   -1,  557,   -1,
  559,  560,  561,  562,  563,  564,  565,  566,  567,  568,
  569,  570,  571,  572,  573,  574,  575,  576,   -1,  578,
  579,  580,  581,   -1,  583,  584,  585,  586,  587,   -1,
  589,  590,  591,  592,  593,  594,  595,  596,  597,  598,
  599,  600,  601,   -1,  603,  604,  605,  606,  607,  608,
   -1,  610,  611,  612,  613,  614,  615,  616,  617,   -1,
  619,  620,  621,  622,   -1,  624,  625,  626,  627,  628,
  629,   -1,   -1,  632,  633,  634,  635,  636,  637,  638,
  639,  640,  641,   -1,  643,  644,  645,  646,  647,  648,
  649,  650,  651,  652,  653,   -1,   -1,  656,  657,  658,
  659,  660,  661,  662,  663,  664,  665,  666,  667,  668,
  669,  670,  671,  672,  673,   -1,  675,  676,  677,  678,
  679,  680,  681,  682,  683,   -1,  685,  686,  687,  688,
  689,  690,  691,  692,  693,  694,  695,  696,   -1,  698,
  699,  700,   -1,  702,   -1,  704,  705,  706,  707,  708,
  709,  710,  711,  712,  713,  714,  715,  716,  717,  718,
   -1,  720,  721,  722,  723,  724,  725,  726,  727,  728,
  729,  730,  731,  732,  733,  734,  735,  736,  737,  738,
  739,  740,   -1,   40,  743,  744,   43,   -1,   45,   -1,
  749,  750,  751,  752,  753,  754,   -1,   -1,   -1,   -1,
   -1,   58,   -1,   -1,   -1,   -1,   63,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  257,   -1,  259,   -1,   -1,   -1,   -1,   -1,  265,   -1,
  267,  268,  269,   -1,   -1,   -1,   -1,  274,   -1,   -1,
   -1,   -1,  279,  280,   -1,  282,   -1,  284,  285,   -1,
  287,   -1,   -1,  290,  291,  292,   -1,   -1,   -1,   -1,
  297,   -1,  299,  300,   -1,   -1,   -1,  304,   -1,  306,
   -1,  308,  309,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  317,   -1,  319,  320,   -1,   -1,   -1,  324,   -1,  326,
   -1,  328,   -1,   -1,   -1,   -1,   -1,  334,   -1,  336,
   -1,   -1,   -1,  340,   -1,   -1,   -1,  344,   -1,  346,
  347,  348,  349,   -1,  351,   -1,  353,  354,   -1,  356,
  357,  358,   -1,   -1,   -1,   -1,   -1,  364,   -1,   -1,
   -1,   -1,   -1,  370,   -1,   -1,   -1,  374,  375,  376,
  377,  378,   -1,  380,   -1,  382,   -1,   -1,   -1,  386,
   -1,  388,  389,   -1,   -1,  392,  393,   -1,   -1,   -1,
  397,   -1,  399,   -1,   -1,  402,  403,  404,  405,   -1,
  407,   -1,  409,  410,  411,  412,  413,  414,  415,  416,
   -1,   -1,   -1,  420,   -1,  422,   -1,   -1,   -1,  426,
   -1,   -1,   -1,   -1,   -1,   -1,  433,   -1,   -1,   -1,
   -1,  438,  439,  440,  441,  442,  443,  444,  445,  446,
  447,  448,  449,   -1,  451,  452,  453,  454,   -1,  456,
  457,   -1,   -1,   -1,   -1,   -1,   -1,  464,  465,  466,
  467,  468,  469,  470,  471,  472,  473,  474,  475,  476,
   -1,  478,  479,  480,  481,  482,   -1,  484,  485,  486,
   -1,  488,  489,  490,  491,   -1,   -1,  494,  495,  496,
  497,  498,  499,  500,   -1,   -1,   -1,   -1,  505,  506,
  507,   -1,  509,  510,  511,   -1,  513,   -1,  515,  516,
  517,   -1,  519,  520,   -1,  522,  523,  524,  525,  526,
  527,  528,  529,  530,  531,  532,  533,  534,  535,  536,
  537,  538,  539,  540,  541,  542,  543,   -1,  545,  546,
  547,  548,  549,  550,   -1,  552,  553,  554,  555,   -1,
  557,   -1,  559,  560,  561,  562,  563,  564,  565,  566,
  567,  568,  569,  570,  571,  572,  573,  574,  575,  576,
   -1,  578,  579,  580,  581,   -1,  583,  584,  585,  586,
  587,   -1,  589,  590,  591,  592,  593,  594,  595,  596,
  597,  598,  599,  600,  601,   -1,  603,  604,  605,  606,
  607,  608,   -1,  610,  611,  612,  613,  614,  615,  616,
  617,   -1,  619,  620,  621,  622,   -1,  624,  625,  626,
  627,  628,  629,   -1,   -1,  632,  633,  634,  635,  636,
  637,  638,  639,  640,  641,   -1,  643,  644,  645,  646,
  647,  648,  649,  650,  651,  652,  653,   -1,   -1,  656,
  657,  658,  659,  660,  661,  662,  663,  664,  665,  666,
  667,  668,  669,  670,  671,  672,  673,   -1,  675,  676,
  677,  678,  679,  680,  681,  682,  683,   -1,  685,  686,
  687,  688,  689,  690,  691,  692,  693,  694,  695,  696,
   -1,  698,  699,  700,   -1,  702,   -1,  704,  705,  706,
  707,  708,  709,  710,  711,  712,  713,  714,  715,  716,
  717,  718,   -1,  720,  721,  722,  723,  724,  725,  726,
  727,  728,  729,  730,  731,  732,  733,  734,  735,  736,
  737,  738,  739,  740,   -1,   40,  743,  744,   43,   -1,
   45,   -1,  749,  750,  751,  752,  753,  754,   -1,   -1,
   -1,   -1,   -1,   58,   -1,   -1,   -1,   -1,   63,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  257,   -1,  259,   -1,   -1,   -1,   -1,   -1,
  265,   -1,  267,  268,  269,   -1,   -1,   -1,   -1,  274,
   -1,   -1,   -1,   -1,  279,  280,   -1,  282,   -1,  284,
  285,   -1,  287,   -1,   -1,  290,  291,  292,   -1,   -1,
   -1,   -1,  297,   -1,  299,  300,   -1,   -1,   -1,  304,
   -1,  306,   -1,  308,  309,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  317,   -1,  319,  320,   -1,   -1,   -1,  324,
   -1,  326,   -1,  328,   -1,   -1,   -1,   -1,   -1,  334,
   -1,  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,  344,
   -1,  346,  347,  348,  349,   -1,  351,   -1,  353,  354,
   -1,  356,  357,  358,   -1,   -1,   -1,   -1,   -1,  364,
   -1,   -1,   -1,   -1,   -1,  370,   -1,   -1,   -1,  374,
  375,  376,  377,  378,   -1,  380,   -1,  382,   -1,   -1,
   -1,  386,   -1,  388,  389,   -1,   -1,  392,  393,   -1,
   -1,   -1,  397,   -1,  399,   -1,   -1,  402,  403,  404,
  405,   -1,  407,   -1,  409,  410,  411,  412,  413,  414,
  415,  416,   -1,   -1,   -1,  420,   -1,  422,   -1,   -1,
   -1,  426,   -1,   -1,   -1,   -1,   -1,   -1,  433,   -1,
   -1,   -1,   -1,  438,  439,  440,  441,  442,  443,  444,
  445,  446,  447,  448,  449,   -1,  451,  452,  453,  454,
   -1,  456,  457,   -1,   -1,   -1,   -1,   -1,   -1,  464,
  465,  466,  467,  468,  469,  470,  471,  472,  473,  474,
  475,  476,   -1,  478,  479,  480,  481,  482,   -1,  484,
  485,  486,   -1,  488,  489,  490,  491,   -1,   -1,  494,
  495,  496,  497,  498,  499,  500,   -1,   -1,   -1,   -1,
  505,  506,  507,   -1,  509,  510,  511,   -1,  513,   -1,
  515,  516,  517,   -1,  519,  520,   -1,  522,  523,  524,
  525,  526,  527,  528,  529,  530,  531,  532,  533,  534,
  535,  536,  537,  538,  539,  540,  541,  542,  543,   -1,
  545,  546,  547,  548,  549,  550,   -1,  552,  553,  554,
  555,   -1,  557,   -1,  559,  560,  561,  562,  563,  564,
  565,  566,  567,  568,  569,  570,  571,  572,  573,  574,
  575,  576,   -1,  578,  579,  580,  581,   -1,  583,  584,
  585,  586,  587,   -1,  589,  590,  591,  592,  593,  594,
  595,  596,  597,  598,  599,  600,  601,   -1,  603,  604,
  605,  606,  607,  608,   -1,  610,  611,  612,  613,  614,
  615,  616,  617,   -1,  619,  620,  621,  622,   -1,  624,
  625,  626,  627,  628,  629,   -1,   -1,  632,  633,  634,
  635,  636,  637,  638,  639,  640,  641,   -1,  643,  644,
  645,  646,  647,  648,  649,  650,  651,  652,  653,   -1,
   -1,  656,  657,  658,  659,  660,  661,  662,  663,  664,
  665,  666,  667,  668,  669,  670,  671,  672,  673,   -1,
  675,  676,  677,  678,  679,  680,  681,  682,  683,   -1,
  685,  686,  687,  688,  689,  690,  691,  692,  693,  694,
  695,  696,   -1,  698,  699,  700,   -1,  702,   -1,  704,
  705,  706,  707,  708,  709,  710,  711,  712,  713,  714,
  715,  716,  717,  718,   -1,  720,  721,  722,  723,  724,
  725,  726,  727,  728,  729,  730,  731,  732,  733,  734,
  735,  736,  737,  738,  739,  740,   -1,   40,  743,  744,
   43,   -1,   45,   -1,  749,  750,  751,  752,  753,  754,
   -1,   -1,   -1,   -1,   -1,   58,   -1,   -1,   -1,   -1,
   63,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  257,   -1,  259,   -1,   -1,   -1,
   -1,   -1,  265,   -1,  267,  268,  269,   -1,   -1,   -1,
   -1,  274,   -1,   -1,   -1,   -1,  279,  280,   -1,  282,
   -1,  284,  285,   -1,  287,   -1,   -1,  290,  291,  292,
   -1,   -1,   -1,   -1,  297,   -1,  299,  300,   -1,   -1,
   -1,  304,   -1,  306,   -1,  308,  309,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  317,   -1,  319,  320,   -1,   -1,
   -1,  324,   -1,  326,   -1,  328,   -1,   -1,   -1,   -1,
   -1,  334,   -1,  336,   -1,   -1,   -1,  340,   -1,   -1,
   -1,  344,   -1,  346,  347,  348,  349,   -1,  351,   -1,
  353,  354,   -1,  356,  357,  358,   -1,   -1,   -1,   -1,
   -1,  364,   -1,   -1,   -1,   -1,   -1,  370,   -1,   -1,
   -1,  374,  375,  376,  377,  378,   -1,  380,   -1,  382,
   -1,   -1,   -1,  386,   -1,  388,  389,   -1,   -1,  392,
  393,   -1,   -1,   -1,  397,   -1,  399,   -1,   -1,  402,
  403,  404,  405,   -1,  407,   -1,  409,  410,  411,  412,
  413,  414,  415,  416,   -1,   -1,   -1,  420,   -1,  422,
   -1,   -1,   -1,  426,   -1,   -1,   -1,   -1,   -1,   -1,
  433,   -1,   -1,   -1,   -1,  438,  439,  440,  441,  442,
  443,  444,  445,  446,  447,  448,  449,   -1,  451,  452,
  453,  454,   -1,  456,  457,   -1,   -1,   -1,   -1,   -1,
   -1,  464,  465,  466,  467,  468,  469,  470,  471,  472,
  473,  474,  475,  476,   -1,  478,  479,  480,  481,  482,
   -1,  484,  485,  486,   -1,  488,  489,  490,  491,   -1,
   -1,  494,  495,  496,  497,  498,  499,  500,   -1,   -1,
   -1,   -1,  505,  506,  507,   -1,  509,  510,  511,   -1,
  513,   -1,  515,  516,  517,   -1,  519,  520,   -1,  522,
  523,  524,  525,  526,  527,  528,  529,  530,  531,  532,
  533,  534,  535,  536,  537,  538,  539,  540,  541,  542,
  543,   -1,  545,  546,  547,  548,  549,  550,   -1,  552,
  553,  554,  555,   -1,  557,   -1,  559,  560,  561,  562,
  563,  564,  565,  566,  567,  568,  569,  570,  571,  572,
  573,  574,  575,  576,   -1,  578,  579,  580,  581,   -1,
  583,  584,  585,  586,  587,   -1,  589,  590,  591,  592,
  593,  594,  595,  596,  597,  598,  599,  600,  601,   -1,
  603,  604,  605,  606,  607,  608,   -1,  610,  611,  612,
  613,  614,  615,  616,  617,   -1,  619,  620,  621,  622,
   -1,  624,  625,  626,  627,  628,  629,   -1,   -1,  632,
  633,  634,  635,  636,  637,  638,  639,  640,  641,   -1,
  643,  644,  645,  646,  647,  648,  649,  650,  651,  652,
  653,   -1,   -1,  656,  657,  658,  659,  660,  661,  662,
  663,  664,  665,  666,  667,  668,  669,  670,  671,  672,
  673,   -1,  675,  676,  677,  678,  679,  680,  681,  682,
  683,   -1,  685,  686,  687,  688,  689,  690,  691,  692,
  693,  694,  695,  696,   -1,  698,  699,  700,   -1,  702,
   -1,  704,  705,  706,  707,  708,  709,  710,  711,  712,
  713,  714,  715,  716,  717,  718,   -1,  720,  721,  722,
  723,  724,  725,  726,  727,  728,  729,  730,  731,  732,
  733,  734,  735,  736,  737,  738,  739,  740,   -1,   40,
  743,  744,   43,   -1,   45,   -1,  749,  750,  751,  752,
  753,  754,   -1,   -1,   -1,   -1,   -1,   58,   -1,   -1,
   -1,   -1,   63,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  257,   -1,  259,   -1,
   -1,   -1,   -1,   -1,  265,   -1,  267,  268,  269,   -1,
   -1,   -1,   -1,  274,   -1,   -1,   -1,   -1,  279,  280,
   -1,  282,   -1,  284,  285,   -1,  287,   -1,   -1,  290,
  291,  292,   -1,   -1,   -1,   -1,  297,   -1,  299,  300,
   -1,   -1,   -1,  304,   -1,  306,   -1,   -1,  309,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  317,   -1,  319,  320,
   -1,   -1,   -1,  324,   -1,  326,   -1,  328,   -1,   -1,
   -1,   -1,   -1,  334,   -1,  336,   -1,   -1,   -1,  340,
   -1,   -1,   -1,  344,   -1,  346,  347,  348,  349,   -1,
  351,   -1,  353,  354,   -1,  356,  357,  358,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  370,
   -1,   -1,   -1,  374,  375,  376,  377,  378,   -1,  380,
   -1,  382,   -1,   -1,   -1,  386,   -1,  388,  389,   -1,
   -1,  392,  393,   -1,   -1,   -1,  397,   -1,  399,   -1,
   -1,  402,  403,  404,  405,   -1,  407,   -1,  409,  410,
  411,  412,  413,  414,  415,  416,   -1,   -1,   -1,  420,
   -1,  422,   -1,   -1,   -1,  426,   -1,   -1,   -1,   -1,
   -1,   -1,  433,   -1,   -1,   -1,   -1,  438,  439,  440,
  441,  442,  443,  444,  445,  446,  447,  448,  449,   -1,
  451,  452,  453,  454,   -1,  456,  457,   -1,   -1,   -1,
   -1,   -1,   -1,  464,  465,  466,  467,  468,  469,  470,
  471,  472,  473,  474,  475,  476,   -1,  478,  479,  480,
  481,  482,   -1,  484,  485,  486,   -1,  488,  489,  490,
  491,   -1,   -1,  494,  495,   -1,   -1,   -1,  499,  500,
   -1,   -1,   -1,   -1,  505,  506,  507,   -1,  509,  510,
  511,   -1,  513,   -1,  515,  516,  517,   -1,  519,  520,
   -1,  522,  523,  524,  525,  526,  527,  528,  529,  530,
  531,  532,  533,  534,  535,  536,  537,  538,  539,  540,
  541,  542,  543,   40,  545,  546,  547,  548,  549,  550,
   -1,  552,  553,  554,  555,   -1,  557,   -1,  559,  560,
  561,  562,  563,  564,  565,  566,  567,  568,  569,  570,
  571,  572,  573,  574,  575,  576,   -1,  578,  579,  580,
  581,   -1,  583,  584,  585,  586,  587,   -1,  589,  590,
  591,  592,  593,  594,  595,  596,  597,  598,  599,  600,
  601,   -1,  603,  604,  605,  606,  607,  608,   -1,  610,
  611,  612,  613,  614,  615,  616,  617,   -1,  619,  620,
  621,  622,   -1,  624,  625,  626,  627,  628,  629,   -1,
   -1,  632,  633,  634,  635,  636,  637,  638,  639,  640,
  641,   -1,  643,  644,  645,  646,  647,  648,  649,  650,
  651,  652,  653,   -1,   -1,  656,  657,  658,  659,  660,
  661,  662,  663,  664,  665,  666,  667,  668,  669,  670,
  671,  672,  673,   -1,  675,  676,  677,  678,  679,  680,
  681,  682,  683,   -1,  685,  686,  687,  688,  689,  690,
  691,  692,  693,  694,  695,  696,   -1,  698,  699,  700,
   -1,  702,   -1,  704,  705,  706,  707,  708,  709,  710,
  711,  712,  713,  714,  715,  716,  717,  718,   -1,  720,
  721,   -1,  723,  724,  725,  726,  727,  728,  729,  730,
  731,  732,  733,  734,  735,  736,  737,  738,  739,  740,
   -1,   -1,  743,  744,   -1,   -1,   -1,   -1,  749,  750,
  751,  752,  753,  754,   -1,   -1,   -1,   -1,   -1,   -1,
  257,   -1,  259,   -1,   -1,   -1,   -1,   -1,  265,   -1,
   -1,  268,  269,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  279,  280,   -1,  282,   -1,  284,   -1,   -1,
  287,   -1,   -1,  290,   -1,   -1,   -1,   -1,   -1,   -1,
  297,   -1,  299,  300,   -1,   -1,   -1,  304,   -1,  306,
   -1,   -1,  309,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  319,  320,   -1,   -1,   -1,  324,   -1,  326,
   -1,  328,  329,   -1,   -1,   -1,   -1,  334,   -1,  336,
   -1,   -1,   -1,  340,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  348,   -1,   -1,  351,   -1,  353,   -1,   -1,   -1,
  357,  358,   -1,  360,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  370,   -1,  372,   -1,  374,  375,  376,
  377,  378,   -1,  380,   -1,   -1,   -1,   -1,   -1,  386,
   -1,  388,  389,   -1,   -1,  392,  393,   -1,   -1,   -1,
   -1,   -1,  399,   -1,   -1,  402,  403,  404,  405,   -1,
  407,   -1,  409,   -1,  411,  412,  413,  414,  415,   -1,
   -1,   -1,   -1,  420,   -1,  422,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  433,   -1,   -1,   -1,
   -1,  438,  439,   -1,   -1,   -1,   -1,   -1,  445,   -1,
   -1,   -1,  449,   -1,  451,  452,  453,  454,   -1,  456,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  464,  465,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  475,  476,
   -1,  478,  479,  480,   -1,   -1,   -1,   -1,  485,  486,
   -1,  488,  489,   -1,  491,   -1,   -1,  494,  495,   -1,
   -1,   -1,  499,  500,   -1,   -1,   -1,   -1,  505,  506,
  507,   -1,  509,  510,  511,   -1,  513,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  523,  524,  525,  526,
  527,  528,  529,  530,  531,  532,  533,  534,  535,  536,
  537,  538,  539,  540,  541,  542,  543,   40,  545,  546,
  547,  548,  549,  550,   -1,  552,  553,  554,  555,   -1,
  557,   -1,  559,  560,  561,  562,  563,  564,  565,  566,
  567,  568,  569,  570,  571,  572,  573,  574,  575,  576,
   -1,  578,  579,  580,  581,   -1,  583,  584,  585,  586,
  587,   -1,  589,  590,  591,  592,  593,  594,  595,  596,
  597,  598,  599,  600,  601,   -1,  603,  604,  605,  606,
  607,  608,   -1,  610,  611,  612,  613,  614,  615,  616,
  617,   -1,  619,  620,   -1,   -1,   -1,  624,  625,  626,
  627,  628,  629,   -1,   -1,  632,  633,  634,  635,  636,
  637,  638,  639,  640,   -1,   -1,   -1,   -1,   -1,  646,
   -1,  648,  649,  650,  651,  652,  653,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  672,  673,   -1,  675,  676,
  677,  678,  679,  680,  681,  682,  683,   -1,  685,  686,
  687,  688,  689,  690,  691,  692,  693,  694,  695,  696,
   -1,  698,  699,  700,   -1,  702,   -1,   -1,   -1,  706,
  707,  708,  709,  710,  711,  712,  713,  714,  715,  716,
  717,  718,   -1,  720,  721,   -1,   -1,   -1,   -1,   -1,
  727,  728,  729,  730,  731,  732,  733,  734,  735,  736,
  737,  738,  739,  740,   -1,   -1,  743,  744,   -1,   -1,
   -1,   -1,  749,  750,  751,  752,  753,  754,   -1,   -1,
   -1,   -1,   -1,   -1,  257,   -1,  259,   -1,   -1,   -1,
   -1,   -1,  265,   -1,   -1,  268,  269,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  279,  280,   -1,  282,
   -1,  284,   -1,   -1,  287,   -1,   -1,  290,   -1,   -1,
   -1,   -1,   -1,   -1,  297,   -1,  299,  300,   -1,   -1,
   -1,  304,   -1,  306,   -1,   -1,  309,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  319,  320,   -1,   -1,
   -1,  324,   -1,  326,   -1,  328,  329,   -1,   -1,   -1,
   -1,  334,   -1,  336,   -1,   -1,   -1,  340,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  348,   -1,   -1,  351,   -1,
  353,   -1,   -1,   -1,  357,  358,   -1,  360,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  370,   -1,  372,
   -1,  374,  375,  376,  377,  378,   -1,  380,   -1,   -1,
   -1,   -1,   -1,  386,   -1,  388,  389,   -1,   -1,  392,
  393,   -1,   -1,   -1,   -1,   -1,  399,   -1,   -1,  402,
  403,  404,  405,   -1,  407,   -1,  409,   -1,  411,  412,
  413,  414,  415,   -1,   -1,   -1,   -1,  420,   -1,  422,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  433,   -1,   -1,   -1,   -1,  438,  439,   -1,   -1,   -1,
   -1,   -1,  445,   -1,   -1,   -1,  449,   -1,  451,  452,
  453,  454,   -1,  456,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  475,  476,   -1,  478,  479,  480,   -1,   -1,
   -1,   -1,  485,  486,   -1,  488,  489,   -1,  491,   -1,
   -1,  494,  495,   -1,   -1,   -1,  499,  500,   -1,   -1,
   -1,   -1,  505,  506,  507,   -1,  509,  510,  511,   -1,
  513,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  523,  524,  525,  526,  527,  528,  529,  530,  531,  532,
  533,  534,  535,  536,  537,  538,  539,  540,  541,  542,
  543,   40,  545,  546,  547,  548,  549,  550,   -1,  552,
  553,  554,  555,   -1,  557,   -1,  559,  560,  561,  562,
  563,  564,  565,  566,  567,  568,  569,  570,  571,  572,
  573,  574,  575,  576,   -1,  578,  579,  580,  581,   -1,
  583,  584,  585,  586,  587,   -1,  589,  590,  591,  592,
  593,  594,  595,  596,  597,  598,  599,  600,  601,   -1,
  603,  604,  605,  606,  607,  608,   -1,  610,  611,  612,
  613,  614,  615,  616,  617,   -1,  619,  620,   -1,   -1,
   -1,  624,  625,  626,  627,  628,  629,   -1,   -1,  632,
  633,  634,  635,  636,  637,  638,  639,  640,   -1,   -1,
   -1,   -1,   -1,  646,   -1,  648,  649,  650,  651,  652,
  653,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  672,
  673,   -1,  675,  676,  677,  678,  679,  680,  681,  682,
  683,   -1,  685,  686,  687,  688,  689,  690,  691,  692,
  693,  694,  695,  696,   -1,  698,  699,  700,   -1,  702,
   -1,   -1,   -1,  706,  707,  708,  709,  710,  711,  712,
  713,  714,  715,  716,  717,  718,   -1,  720,  721,   -1,
   -1,   -1,   -1,   -1,  727,  728,  729,  730,  731,  732,
  733,  734,  735,  736,  737,  738,  739,  740,   -1,   -1,
  743,  744,   -1,   -1,   -1,   -1,  749,  750,  751,  752,
  753,  754,   -1,   -1,   -1,   -1,   -1,   -1,  257,   -1,
  259,   -1,   -1,   -1,   -1,   -1,  265,   -1,   -1,  268,
  269,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  279,  280,   -1,  282,   -1,  284,   -1,   -1,  287,   -1,
   -1,  290,   -1,   -1,   -1,   -1,   -1,   -1,  297,   -1,
  299,  300,   -1,   -1,   -1,  304,   -1,  306,   -1,   -1,
  309,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  319,  320,   -1,   -1,   -1,  324,   -1,  326,   -1,  328,
   -1,   -1,   -1,   -1,   -1,  334,   -1,  336,   -1,   -1,
   -1,  340,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  348,
   -1,   -1,  351,   -1,  353,   -1,   -1,   -1,  357,  358,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  370,   -1,   -1,   -1,  374,  375,  376,  377,  378,
   -1,  380,   -1,   -1,   -1,   -1,   -1,  386,   -1,  388,
  389,   -1,   -1,  392,  393,   -1,   -1,   -1,   -1,   -1,
  399,   -1,   -1,  402,  403,  404,  405,   -1,  407,   -1,
  409,   -1,  411,  412,  413,  414,  415,   -1,   -1,   -1,
   -1,  420,   -1,  422,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  433,   -1,   -1,   -1,  437,  438,
  439,   -1,   -1,   -1,   -1,   -1,  445,   -1,   -1,   -1,
  449,   -1,  451,  452,  453,  454,   -1,  456,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  464,  465,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  475,  476,   -1,  478,
  479,  480,   -1,   -1,   -1,   -1,  485,  486,   -1,  488,
  489,   -1,  491,   -1,   -1,  494,  495,   -1,   -1,   -1,
  499,  500,   -1,   -1,   -1,   -1,  505,  506,  507,   -1,
  509,  510,  511,   -1,  513,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  523,  524,  525,  526,  527,  528,
  529,  530,  531,  532,  533,  534,  535,  536,  537,  538,
  539,  540,  541,  542,  543,   40,  545,  546,  547,  548,
  549,  550,   -1,  552,  553,  554,  555,   -1,  557,   -1,
  559,  560,  561,  562,  563,  564,  565,  566,  567,  568,
  569,  570,  571,  572,  573,  574,  575,  576,   -1,  578,
  579,  580,  581,   -1,  583,  584,  585,  586,  587,   -1,
  589,  590,  591,  592,  593,  594,  595,  596,  597,  598,
  599,  600,  601,   -1,  603,  604,  605,  606,  607,  608,
   -1,  610,  611,  612,  613,  614,  615,  616,  617,   -1,
  619,  620,   -1,   -1,   -1,  624,  625,  626,  627,  628,
  629,   -1,   -1,  632,  633,  634,  635,  636,  637,  638,
  639,  640,   -1,   -1,   -1,   -1,   -1,  646,   -1,  648,
  649,  650,  651,  652,  653,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  672,  673,   -1,  675,  676,  677,  678,
  679,  680,  681,  682,  683,   -1,  685,  686,  687,  688,
  689,  690,  691,  692,  693,  694,  695,  696,   -1,  698,
  699,  700,  701,  702,   -1,   -1,   -1,  706,  707,  708,
  709,  710,  711,  712,  713,  714,  715,  716,  717,  718,
   -1,  720,  721,   -1,   -1,   -1,   -1,   -1,  727,  728,
  729,  730,  731,  732,  733,  734,  735,  736,  737,  738,
  739,  740,   -1,   -1,  743,  744,   -1,   -1,   -1,   -1,
  749,  750,  751,  752,  753,  754,   -1,   -1,   -1,   -1,
   -1,   -1,  257,   -1,  259,   -1,   -1,   -1,   -1,   -1,
  265,   -1,   -1,  268,  269,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  279,  280,   -1,  282,   -1,  284,
   -1,   -1,  287,   -1,   -1,  290,   -1,   -1,   -1,   -1,
   -1,   -1,  297,   -1,  299,  300,   -1,   -1,   -1,  304,
   -1,  306,   -1,   -1,  309,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  319,  320,   -1,   -1,   -1,  324,
   -1,  326,   -1,  328,   -1,   -1,   -1,   -1,   -1,  334,
   -1,  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  348,   -1,   -1,  351,   -1,  353,   -1,
   -1,   -1,  357,  358,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  370,   -1,   -1,   -1,  374,
  375,  376,  377,  378,   -1,  380,   -1,   -1,   -1,   -1,
   -1,  386,   -1,  388,  389,   -1,   -1,  392,  393,   -1,
   -1,   -1,   -1,   -1,  399,   -1,   -1,  402,  403,  404,
  405,   -1,  407,   -1,  409,   -1,  411,  412,  413,  414,
  415,   -1,   -1,   -1,   -1,  420,   -1,  422,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  433,   -1,
   -1,   -1,   -1,  438,  439,   -1,   -1,   -1,   -1,   -1,
  445,   -1,   -1,   -1,  449,   -1,  451,  452,  453,  454,
   -1,  456,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  464,
  465,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  475,  476,   -1,  478,  479,  480,   -1,   -1,   -1,   -1,
  485,  486,   -1,  488,  489,   -1,  491,   -1,   -1,  494,
  495,   -1,   -1,   -1,  499,  500,   -1,   -1,   -1,   -1,
  505,  506,  507,   -1,  509,  510,  511,   -1,  513,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  523,  524,
  525,  526,  527,  528,  529,  530,  531,  532,  533,  534,
  535,  536,  537,  538,  539,  540,  541,  542,  543,   -1,
  545,  546,  547,  548,  549,  550,   -1,  552,  553,  554,
  555,   -1,  557,   -1,  559,  560,  561,  562,  563,  564,
  565,  566,  567,  568,  569,  570,  571,  572,  573,  574,
  575,  576,   -1,  578,  579,  580,  581,   -1,  583,  584,
  585,  586,  587,   -1,  589,  590,  591,  592,  593,  594,
  595,  596,  597,  598,  599,  600,  601,   -1,  603,  604,
  605,  606,  607,  608,   -1,  610,  611,  612,  613,  614,
  615,  616,  617,   -1,  619,  620,   -1,   -1,   40,  624,
  625,  626,  627,  628,  629,   -1,   -1,  632,  633,  634,
  635,  636,  637,  638,  639,  640,   58,   -1,   -1,   -1,
   -1,  646,   -1,  648,  649,  650,  651,  652,  653,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  672,  673,   -1,
  675,  676,  677,  678,  679,  680,  681,  682,  683,   -1,
  685,  686,  687,  688,  689,  690,  691,  692,  693,  694,
  695,  696,   -1,  698,  699,  700,  701,  702,   -1,   -1,
   -1,  706,  707,  708,  709,  710,  711,  712,  713,  714,
  715,  716,  717,  718,   -1,  720,  721,   -1,   -1,   -1,
   -1,   -1,  727,  728,  729,  730,  731,  732,  733,  734,
  735,  736,  737,  738,  739,  740,   -1,   -1,  743,  744,
   -1,   -1,   -1,   -1,  749,  750,  751,  752,  753,  754,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  257,   -1,  259,   -1,   -1,
   -1,   -1,   -1,  265,   -1,   -1,  268,  269,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  279,  280,   -1,
  282,   -1,  284,   -1,   -1,  287,   -1,   -1,  290,   -1,
   -1,   -1,   -1,   -1,   -1,  297,   -1,  299,  300,   -1,
   -1,   -1,  304,   -1,  306,   -1,   -1,  309,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  319,  320,   -1,
   -1,   -1,  324,   -1,  326,   -1,  328,   -1,   -1,   -1,
   -1,   -1,  334,   -1,  336,   -1,   -1,   -1,  340,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  348,   -1,   -1,  351,
   -1,  353,   -1,   -1,   -1,  357,  358,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  370,   -1,
   -1,   -1,  374,  375,  376,  377,  378,   -1,  380,   -1,
   -1,   -1,   -1,   -1,  386,   -1,  388,  389,   -1,   -1,
  392,  393,   -1,   -1,   -1,   -1,   -1,  399,   -1,   -1,
  402,  403,  404,  405,   -1,  407,   -1,  409,   -1,  411,
  412,  413,  414,  415,   -1,   -1,   -1,   -1,  420,   -1,
  422,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  433,   -1,   -1,   -1,   -1,  438,  439,   -1,   -1,
   -1,   -1,   -1,  445,   -1,   -1,   -1,  449,   -1,  451,
  452,  453,  454,   -1,  456,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  475,  476,   -1,  478,  479,  480,   -1,
   -1,   -1,   -1,  485,  486,   -1,  488,  489,   -1,  491,
   -1,   -1,  494,  495,   -1,   -1,   -1,  499,  500,   -1,
   -1,   -1,   -1,  505,  506,  507,   -1,  509,  510,  511,
   -1,  513,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  523,  524,  525,  526,  527,  528,  529,  530,  531,
  532,  533,  534,  535,  536,  537,  538,  539,  540,  541,
  542,  543,   40,  545,  546,  547,  548,  549,  550,   -1,
  552,  553,  554,  555,   -1,  557,   -1,  559,  560,  561,
  562,  563,  564,  565,  566,  567,  568,  569,  570,  571,
  572,  573,  574,  575,  576,   -1,  578,  579,  580,  581,
   -1,  583,  584,  585,  586,  587,   -1,  589,  590,  591,
  592,  593,  594,  595,  596,  597,  598,  599,  600,  601,
   -1,  603,  604,  605,  606,  607,  608,   -1,  610,  611,
  612,  613,  614,  615,  616,  617,   -1,  619,  620,   -1,
   -1,   -1,  624,  625,  626,  627,  628,  629,   -1,   -1,
  632,  633,  634,  635,  636,  637,  638,  639,  640,   -1,
   -1,   -1,   -1,   -1,  646,   -1,  648,  649,  650,  651,
  652,  653,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  672,  673,   -1,  675,  676,  677,  678,  679,  680,  681,
  682,  683,   -1,  685,  686,  687,  688,  689,  690,  691,
  692,  693,  694,  695,  696,   -1,  698,  699,  700,   -1,
  702,   -1,   -1,   -1,  706,  707,  708,  709,  710,  711,
  712,  713,  714,  715,  716,  717,  718,   -1,  720,  721,
   -1,   -1,   -1,   -1,   -1,  727,  728,  729,  730,  731,
  732,  733,  734,  735,  736,  737,  738,  739,  740,   -1,
   -1,  743,  744,   -1,   -1,   -1,   -1,  749,  750,  751,
  752,  753,  754,   -1,   -1,   -1,   -1,   -1,   -1,  257,
   -1,  259,   -1,   -1,   -1,   -1,   -1,  265,   -1,   -1,
  268,  269,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  279,  280,   -1,  282,   -1,  284,   -1,   -1,  287,
   -1,   -1,  290,   -1,   -1,   -1,   -1,   -1,   -1,  297,
   -1,  299,  300,   -1,   -1,   -1,  304,   -1,  306,   -1,
   -1,  309,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  319,  320,   -1,   -1,   -1,  324,   -1,  326,   -1,
  328,   -1,   -1,   -1,   -1,   -1,  334,   -1,  336,   -1,
   -1,   -1,  340,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  348,   -1,   -1,  351,   -1,  353,   -1,   -1,   -1,  357,
  358,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  370,   -1,   -1,   -1,  374,  375,  376,  377,
  378,   -1,  380,   -1,   -1,   -1,   -1,   -1,  386,   -1,
  388,  389,   -1,   -1,  392,  393,   -1,   -1,   -1,   -1,
   -1,  399,   -1,   -1,  402,  403,  404,  405,   -1,  407,
   -1,  409,   -1,  411,  412,  413,  414,  415,   -1,   -1,
   -1,   -1,  420,   -1,  422,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  433,   -1,   -1,   -1,   -1,
  438,  439,   -1,   -1,   -1,   -1,   -1,  445,   -1,   -1,
   -1,  449,   -1,  451,  452,  453,  454,   -1,  456,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  464,  465,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  475,  476,   -1,
  478,  479,  480,   -1,   -1,   -1,   -1,  485,  486,   -1,
  488,  489,   -1,  491,   -1,   -1,  494,  495,   -1,   -1,
   -1,  499,  500,   -1,   -1,   -1,   -1,  505,  506,  507,
   -1,  509,  510,  511,   -1,  513,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  523,  524,  525,  526,  527,
  528,  529,  530,  531,  532,  533,  534,  535,  536,  537,
  538,  539,  540,  541,  542,  543,   40,  545,  546,  547,
  548,  549,  550,   -1,  552,  553,  554,  555,   -1,  557,
   -1,  559,  560,  561,  562,  563,  564,  565,  566,  567,
  568,  569,  570,  571,  572,  573,  574,  575,  576,   -1,
  578,  579,  580,  581,   -1,  583,  584,  585,  586,  587,
   -1,  589,  590,  591,  592,  593,  594,  595,  596,  597,
  598,  599,  600,  601,   -1,  603,  604,  605,  606,  607,
  608,   -1,  610,  611,  612,  613,  614,  615,  616,  617,
   -1,  619,  620,   -1,   -1,   -1,  624,  625,  626,  627,
  628,  629,   -1,   -1,  632,  633,  634,  635,  636,  637,
  638,  639,  640,   -1,   -1,   -1,   -1,   -1,  646,   -1,
  648,  649,  650,  651,  652,  653,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  672,  673,   -1,  675,  676,  677,
  678,  679,  680,  681,  682,  683,   -1,  685,  686,  687,
  688,  689,  690,  691,  692,  693,  694,  695,  696,   -1,
  698,  699,  700,   -1,  702,   -1,   -1,   -1,  706,  707,
  708,  709,  710,  711,  712,  713,  714,  715,  716,  717,
  718,   -1,  720,  721,   -1,   -1,   -1,   -1,   -1,  727,
  728,  729,  730,  731,  732,  733,  734,  735,  736,  737,
  738,  739,  740,   -1,   -1,  743,  744,   -1,   -1,   -1,
   -1,  749,  750,  751,  752,  753,  754,   -1,   -1,   -1,
   -1,   -1,   -1,  257,   -1,  259,   -1,   -1,   -1,   -1,
   -1,  265,   -1,   -1,  268,  269,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  279,  280,   -1,  282,   -1,
  284,   -1,   -1,  287,   -1,   -1,  290,   -1,   -1,   -1,
   -1,   -1,   -1,  297,   -1,  299,  300,   -1,   -1,   -1,
  304,   -1,  306,   -1,   -1,  309,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  319,  320,   -1,   -1,   -1,
  324,   -1,  326,   -1,  328,   -1,   -1,   -1,   -1,   -1,
  334,   -1,  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  348,   -1,   -1,  351,   -1,  353,
   -1,   -1,   -1,  357,  358,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  370,   -1,   -1,   -1,
  374,  375,  376,  377,  378,   -1,  380,   -1,   -1,   -1,
   -1,   -1,  386,   -1,  388,  389,   -1,   -1,  392,  393,
   -1,   -1,   -1,   -1,   -1,  399,   -1,   -1,  402,  403,
  404,  405,   -1,  407,   -1,  409,   -1,  411,  412,  413,
  414,  415,   -1,   -1,   -1,   -1,  420,   -1,  422,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  433,
   -1,   -1,   -1,   -1,  438,  439,   -1,   -1,   -1,   -1,
   -1,  445,   -1,   -1,   -1,  449,   -1,  451,  452,  453,
  454,   -1,  456,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  475,  476,   -1,  478,  479,  480,   -1,   -1,   -1,
   -1,  485,  486,   -1,  488,  489,   -1,  491,   -1,   -1,
  494,  495,   -1,   -1,   -1,  499,  500,   -1,   -1,   -1,
   -1,  505,  506,  507,   -1,  509,  510,  511,   -1,  513,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  523,
  524,  525,  526,  527,  528,  529,  530,  531,  532,  533,
  534,  535,  536,  537,  538,  539,  540,  541,  542,  543,
   42,  545,  546,  547,  548,  549,  550,   -1,  552,  553,
  554,  555,   -1,  557,   -1,  559,  560,  561,  562,  563,
  564,  565,  566,  567,  568,  569,  570,  571,  572,  573,
  574,  575,  576,   -1,  578,  579,  580,  581,   -1,  583,
  584,  585,  586,  587,   -1,  589,  590,  591,  592,  593,
  594,  595,  596,  597,  598,  599,  600,  601,   -1,  603,
  604,  605,  606,  607,  608,   -1,  610,  611,  612,  613,
  614,  615,  616,  617,   -1,  619,  620,   -1,   -1,   -1,
  624,  625,  626,  627,  628,  629,   -1,   -1,  632,  633,
  634,  635,  636,  637,  638,  639,  640,   -1,   -1,   -1,
   -1,   -1,  646,   -1,  648,  649,  650,  651,  652,  653,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  672,  673,
   -1,  675,  676,  677,  678,  679,  680,  681,  682,  683,
   -1,  685,  686,  687,  688,  689,  690,  691,  692,  693,
  694,  695,  696,   -1,  698,  699,  700,   -1,  702,   -1,
   -1,   -1,  706,  707,  708,  709,  710,  711,  712,  713,
  714,  715,  716,  717,  718,   -1,  720,  721,   -1,   -1,
   -1,   -1,   -1,  727,  728,  729,  730,  731,  732,  733,
  734,  735,  736,  737,  738,  739,  740,   -1,   -1,  743,
  744,   -1,   -1,   -1,   -1,  749,  750,  751,  752,  753,
  754,   -1,   -1,   -1,   -1,  257,   -1,  259,   -1,   -1,
   -1,   -1,   -1,  265,   -1,   -1,  268,  269,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  279,  280,   -1,
  282,   -1,  284,   -1,   -1,  287,   -1,   -1,  290,   -1,
  292,   -1,   -1,   -1,   -1,  297,   -1,  299,  300,   -1,
   -1,   -1,  304,   -1,  306,   -1,   -1,  309,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  319,  320,   -1,
   -1,   -1,  324,   -1,  326,   -1,  328,   -1,   -1,   -1,
   -1,   -1,  334,   -1,  336,   -1,   -1,   -1,  340,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  348,   -1,   -1,  351,
   -1,  353,   -1,   -1,   -1,  357,  358,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  370,   -1,
   -1,   -1,  374,  375,  376,  377,  378,   -1,  380,   -1,
   -1,   -1,   -1,   -1,  386,   -1,  388,  389,   -1,   -1,
  392,  393,   -1,   -1,   -1,   -1,   -1,  399,   -1,   -1,
  402,  403,  404,  405,   -1,  407,   -1,  409,   -1,  411,
  412,  413,  414,  415,   -1,   -1,   -1,   -1,  420,   -1,
  422,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  433,   -1,   -1,   -1,   -1,  438,  439,   -1,   -1,
   -1,   -1,   -1,  445,   -1,   -1,   -1,  449,   -1,  451,
  452,  453,  454,   -1,  456,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  475,  476,   -1,  478,  479,  480,   -1,
   -1,   -1,   -1,  485,  486,   -1,  488,  489,   -1,  491,
   -1,   -1,  494,  495,   -1,   -1,   -1,  499,  500,   -1,
   -1,   -1,   -1,  505,  506,  507,   -1,  509,  510,  511,
   -1,  513,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  523,  524,  525,  526,  527,  528,  529,  530,  531,
  532,  533,  534,  535,  536,  537,  538,  539,  540,  541,
  542,  543,   45,  545,  546,  547,  548,  549,  550,   -1,
  552,  553,  554,  555,   -1,  557,   -1,  559,  560,  561,
  562,  563,  564,  565,  566,  567,  568,  569,  570,  571,
  572,  573,  574,  575,  576,   -1,  578,  579,  580,  581,
   -1,  583,  584,  585,  586,  587,   -1,  589,  590,  591,
  592,  593,  594,  595,  596,  597,  598,  599,  600,  601,
   -1,  603,  604,  605,  606,  607,  608,   -1,  610,  611,
  612,  613,  614,  615,  616,  617,   -1,  619,  620,   -1,
   -1,   -1,  624,  625,  626,  627,  628,  629,   -1,   -1,
  632,  633,  634,  635,  636,  637,  638,  639,  640,   -1,
   -1,   -1,   -1,   -1,  646,  647,  648,  649,  650,  651,
  652,  653,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  672,  673,   -1,  675,  676,  677,  678,  679,  680,  681,
  682,  683,   -1,  685,  686,  687,  688,  689,  690,  691,
  692,  693,  694,  695,  696,   -1,  698,  699,  700,   -1,
  702,   -1,   -1,   -1,  706,  707,  708,  709,  710,  711,
  712,  713,  714,  715,  716,  717,  718,   -1,  720,  721,
   -1,   -1,   -1,   -1,   -1,  727,  728,  729,  730,  731,
  732,  733,  734,  735,  736,  737,  738,  739,  740,   -1,
   -1,  743,  744,   -1,   -1,   -1,   -1,  749,  750,  751,
  752,  753,  754,   -1,  257,   -1,  259,   -1,   -1,   -1,
   -1,   -1,  265,   -1,   -1,  268,  269,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  279,  280,   -1,  282,
   -1,  284,   -1,   -1,  287,   -1,   -1,  290,   -1,   -1,
   -1,   -1,   -1,   -1,  297,   -1,  299,  300,   -1,   -1,
   -1,  304,   -1,  306,   -1,   -1,  309,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  319,  320,   -1,   -1,
   -1,  324,   -1,  326,   -1,  328,   -1,   -1,   -1,   -1,
   -1,  334,   -1,  336,   -1,   -1,   -1,  340,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  348,   -1,   -1,  351,   -1,
  353,   -1,   -1,   -1,  357,  358,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  370,   -1,   -1,
   -1,  374,  375,  376,  377,  378,   -1,  380,   -1,   -1,
   -1,   -1,   -1,  386,   -1,  388,  389,   -1,   -1,  392,
  393,   -1,   -1,   -1,   -1,   -1,  399,   -1,   -1,  402,
  403,  404,  405,   -1,  407,   -1,  409,   -1,  411,  412,
  413,  414,  415,   -1,   -1,   -1,   -1,  420,   -1,  422,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  433,   -1,   -1,   -1,   -1,  438,  439,   -1,   -1,   -1,
   -1,   -1,  445,  446,   -1,   -1,  449,   -1,  451,  452,
  453,  454,   -1,  456,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  475,  476,   -1,  478,  479,  480,   -1,   -1,
   -1,   -1,  485,  486,   -1,  488,  489,   -1,  491,   -1,
   -1,  494,  495,   -1,   -1,   -1,  499,  500,   -1,   -1,
   -1,   -1,  505,  506,  507,   -1,  509,  510,  511,   -1,
  513,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  523,  524,  525,  526,  527,  528,  529,  530,  531,  532,
  533,  534,  535,  536,  537,  538,  539,  540,  541,  542,
  543,   -1,  545,  546,  547,  548,  549,  550,   -1,  552,
  553,  554,  555,   -1,  557,   -1,  559,  560,  561,  562,
  563,  564,  565,  566,  567,  568,  569,  570,  571,  572,
  573,  574,  575,  576,   -1,  578,  579,  580,  581,   -1,
  583,  584,  585,  586,  587,   -1,  589,  590,  591,  592,
  593,  594,  595,  596,  597,  598,  599,  600,  601,   -1,
  603,  604,  605,  606,  607,  608,   -1,  610,  611,  612,
  613,  614,  615,  616,  617,   -1,  619,  620,   60,   61,
   62,  624,  625,  626,  627,  628,  629,   -1,   -1,  632,
  633,  634,  635,  636,  637,  638,  639,  640,   -1,   -1,
   -1,   -1,   -1,  646,   -1,  648,  649,  650,  651,  652,
  653,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  672,
  673,  674,  675,  676,  677,  678,  679,  680,  681,  682,
  683,   -1,  685,  686,  687,  688,  689,  690,  691,  692,
  693,  694,  695,  696,   -1,  698,  699,  700,   -1,  702,
   -1,   -1,   -1,  706,  707,  708,  709,  710,  711,  712,
  713,  714,  715,  716,  717,  718,   -1,  720,  721,   -1,
   -1,   -1,   -1,   -1,  727,  728,  729,  730,  731,  732,
  733,  734,  735,  736,  737,  738,  739,  740,   -1,   -1,
  743,  744,   -1,   -1,   -1,   -1,  749,  750,  751,  752,
  753,  754,   42,   43,   44,   45,   46,   47,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   60,   61,   62,   -1,   -1,   41,   42,   43,   44,   45,
   46,   47,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   60,   61,   62,   -1,   -1,   -1,
   -1,   91,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  262,   -1,   -1,   -1,   -1,   91,   -1,   -1,   -1,  271,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  284,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  318,   -1,   -1,   -1,
   -1,   -1,   -1,  325,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  333,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  350,   -1,
  352,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  362,   -1,  364,  365,  366,   -1,   -1,   -1,   -1,  371,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  412,   -1,  262,   -1,   -1,   -1,  266,   -1,   -1,   -1,
   -1,  271,   -1,   -1,   -1,   -1,   -1,  277,   -1,   -1,
   -1,  281,   -1,   -1,  284,   -1,  262,   -1,   -1,   -1,
  266,   -1,   -1,   -1,   -1,  271,   -1,   -1,   -1,   -1,
   -1,  277,   -1,   -1,   -1,  281,  458,  459,  460,  461,
  462,  463,  464,  465,   -1,   -1,   -1,   -1,  318,   -1,
   -1,   -1,   -1,   -1,   -1,  325,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  333,   -1,   -1,   -1,   58,   -1,   -1,
   -1,   -1,  318,   -1,   -1,   -1,   -1,   -1,   -1,  325,
  350,   -1,  352,   -1,   -1,   -1,   -1,  333,   -1,   -1,
   -1,   -1,  362,   -1,  364,  365,  366,   -1,   -1,   -1,
   -1,  371,   -1,   -1,  350,   -1,  352,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  362,   -1,  364,  365,
  366,   -1,   -1,   -1,   -1,  371,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  412,   -1,   -1,  567,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  593,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  602,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  458,  459,
  460,  461,  462,  463,  464,  465,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  257,   -1,  259,   -1,
  261,   -1,   -1,   -1,  265,   -1,   -1,  268,  269,  270,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  279,  280,
   -1,  282,   -1,  284,   -1,   -1,  287,  567,   -1,  290,
   -1,   -1,   -1,   -1,   -1,  296,  297,   -1,  299,  300,
   -1,   -1,  303,  304,   -1,  306,  307,   -1,  309,  741,
  742,  312,   -1,  593,   -1,   -1,   -1,   -1,  319,  320,
   -1,   -1,  602,  324,  325,  326,   -1,  328,   -1,  330,
   -1,   -1,   -1,  334,   -1,  336,   -1,   -1,   -1,  340,
   -1,   -1,   -1,   -1,   -1,  346,  602,  348,  349,   -1,
  351,   -1,  353,  609,  355,   -1,  357,  358,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  370,
   -1,   -1,   -1,  374,  375,  376,  377,  378,   -1,  380,
   -1,  382,  383,   -1,   -1,  386,   -1,  388,  389,   -1,
   -1,  392,  393,   -1,   -1,   -1,  397,   -1,  399,  400,
  401,  402,  403,  404,  405,   -1,  407,   -1,  409,   -1,
  411,  412,  413,  414,  415,   -1,   -1,   -1,   -1,  420,
   -1,  422,   -1,   -1,  425,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  433,  434,   -1,  436,  437,  438,  439,   -1,
   -1,   -1,   -1,   -1,  445,   -1,   -1,   -1,  449,   -1,
  451,  452,  453,  454,   -1,  456,  457,   -1,   -1,   -1,
   -1,  741,  742,  464,  465,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  475,  476,   -1,  478,  479,  480,
   -1,   -1,   -1,   -1,  485,  486,   -1,  488,  489,   -1,
  491,   -1,   -1,  494,  495,   -1,   -1,   -1,  499,  500,
  501,  502,  503,   -1,  505,  506,  507,   -1,  509,  510,
  511,   -1,  513,   -1,  515,  516,  517,   -1,  519,  520,
   -1,  522,  523,  524,  525,  526,  527,  528,  529,  530,
  531,  532,  533,  534,  535,  536,  537,  538,  539,  540,
  541,  542,  543,   -1,  545,  546,  547,  548,  549,  550,
   -1,  552,  553,  554,  555,   58,  557,   -1,  559,  560,
  561,  562,  563,  564,  565,  566,  567,  568,  569,  570,
  571,  572,  573,  574,  575,  576,   -1,  578,  579,  580,
  581,   -1,  583,  584,  585,  586,  587,   -1,  589,  590,
  591,  592,  593,  594,  595,  596,  597,  598,  599,  600,
  601,   -1,  603,  604,  605,  606,  607,  608,   -1,  610,
  611,  612,  613,  614,  615,  616,  617,   -1,  619,  620,
  621,  622,   -1,  624,  625,  626,  627,  628,  629,  630,
   -1,  632,  633,  634,  635,  636,  637,  638,  639,  640,
   -1,   -1,   -1,   -1,   -1,  646,   -1,  648,  649,  650,
  651,  652,  653,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  672,  673,   -1,  675,  676,  677,  678,  679,  680,
  681,  682,  683,   -1,  685,  686,  687,  688,  689,  690,
  691,  692,  693,  694,  695,  696,   -1,  698,  699,  700,
   -1,  702,   -1,   -1,   -1,  706,  707,  708,  709,  710,
  711,  712,  713,  714,  715,  716,  717,  718,   -1,  720,
  721,   -1,   -1,  724,  725,  726,  727,  728,  729,  730,
  731,  732,  733,  734,  735,  736,  737,  738,  739,  740,
   -1,   -1,  743,  744,   -1,   -1,   -1,   -1,  749,  750,
  751,  752,  753,  754,  257,   -1,  259,   -1,  261,   -1,
   -1,   -1,  265,   -1,   -1,  268,  269,  270,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  279,  280,   -1,  282,
   -1,  284,   -1,   -1,  287,   -1,   -1,  290,   -1,   -1,
   -1,   -1,   -1,  296,  297,   -1,  299,  300,   -1,   -1,
   -1,  304,   -1,  306,  307,   -1,  309,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  319,  320,   -1,   -1,
   -1,  324,  325,  326,   -1,  328,   -1,  330,   -1,   -1,
   -1,  334,   -1,  336,   -1,   -1,   -1,  340,   -1,   -1,
   -1,   -1,   -1,  346,   -1,  348,  349,   -1,  351,   -1,
  353,   -1,  355,   -1,  357,  358,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  370,   -1,   -1,
   -1,  374,  375,  376,  377,  378,   -1,  380,   -1,  382,
  383,   -1,   -1,  386,   -1,  388,  389,   -1,   -1,  392,
  393,   -1,   -1,   -1,  397,   -1,  399,   -1,  401,  402,
  403,  404,  405,   -1,  407,   -1,  409,   -1,  411,  412,
  413,  414,  415,   -1,   -1,   -1,   -1,  420,   -1,  422,
   -1,   -1,  425,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  433,  434,   -1,   -1,  437,  438,  439,   -1,   -1,   -1,
   -1,   -1,  445,   -1,   -1,   -1,  449,   -1,  451,  452,
  453,  454,   -1,  456,  457,   -1,   -1,   -1,   -1,   -1,
   -1,  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  475,  476,   -1,  478,  479,  480,   -1,   -1,
   -1,   -1,  485,  486,   -1,  488,  489,   -1,  491,   -1,
   -1,  494,  495,   -1,   -1,   -1,  499,  500,  501,  502,
  503,   -1,  505,  506,  507,   -1,  509,  510,  511,   -1,
  513,   -1,  515,  516,  517,   -1,  519,  520,   -1,  522,
  523,  524,  525,  526,  527,  528,  529,  530,  531,  532,
  533,  534,  535,  536,  537,  538,  539,  540,  541,  542,
  543,   -1,  545,  546,  547,  548,  549,  550,   -1,  552,
  553,  554,  555,   58,  557,   -1,  559,  560,  561,  562,
  563,  564,  565,  566,  567,  568,  569,  570,  571,  572,
  573,  574,  575,  576,   -1,  578,  579,  580,  581,   -1,
  583,  584,  585,  586,  587,   -1,  589,  590,  591,  592,
  593,  594,  595,  596,  597,  598,  599,  600,  601,   -1,
  603,  604,  605,  606,  607,  608,   -1,  610,  611,  612,
  613,  614,  615,  616,  617,   -1,  619,  620,  621,  622,
   -1,  624,  625,  626,  627,  628,  629,  630,   -1,  632,
  633,  634,  635,  636,  637,  638,  639,  640,   -1,   -1,
   -1,   -1,   -1,  646,   -1,  648,  649,  650,  651,  652,
  653,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  672,
  673,   -1,  675,  676,  677,  678,  679,  680,  681,  682,
  683,   -1,  685,  686,  687,  688,  689,  690,  691,  692,
  693,  694,  695,  696,   -1,  698,  699,  700,   -1,  702,
   -1,   -1,   -1,  706,  707,  708,  709,  710,  711,  712,
  713,  714,  715,  716,  717,  718,   -1,  720,  721,   -1,
   -1,  724,  725,  726,  727,  728,  729,  730,  731,  732,
  733,  734,  735,  736,  737,  738,  739,  740,   -1,   -1,
  743,  744,   -1,   -1,   -1,   -1,  749,  750,  751,  752,
  753,  754,  257,   -1,  259,   -1,  261,   -1,   -1,   -1,
  265,   -1,   -1,  268,  269,  270,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  279,  280,   -1,  282,   -1,  284,
   -1,   -1,  287,   -1,   -1,  290,   -1,   -1,   -1,   -1,
   -1,  296,  297,   -1,  299,  300,   -1,   -1,   -1,  304,
   -1,  306,  307,   -1,  309,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  319,  320,   -1,   -1,   -1,  324,
  325,  326,   -1,  328,   -1,  330,   -1,   -1,   -1,  334,
   -1,  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,   -1,
   -1,  346,   -1,  348,  349,   -1,  351,   -1,  353,   -1,
  355,   -1,  357,  358,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  370,   -1,   -1,   -1,  374,
  375,  376,  377,  378,   -1,  380,   -1,  382,  383,   -1,
   -1,  386,   -1,  388,  389,   -1,   -1,  392,  393,   -1,
   -1,   -1,  397,   -1,  399,   -1,  401,  402,  403,  404,
  405,   -1,  407,   -1,  409,   -1,  411,  412,  413,  414,
  415,   -1,   -1,   -1,   -1,  420,   -1,  422,   -1,   -1,
  425,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  433,   -1,
   -1,   -1,  437,  438,  439,   -1,   -1,   -1,   -1,   -1,
  445,   -1,   -1,   -1,  449,   -1,  451,  452,  453,  454,
   -1,  456,  457,   -1,   -1,   -1,   -1,   -1,   -1,  464,
  465,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  475,  476,   -1,  478,  479,  480,   -1,   -1,   -1,   -1,
  485,  486,   -1,  488,  489,   -1,  491,   -1,   -1,  494,
  495,   -1,   -1,   -1,  499,  500,  501,  502,  503,   -1,
  505,  506,  507,   -1,  509,  510,  511,   -1,  513,   -1,
  515,  516,  517,   -1,  519,  520,   -1,  522,  523,  524,
  525,  526,  527,  528,  529,  530,  531,  532,  533,  534,
  535,  536,  537,  538,  539,  540,  541,  542,  543,   -1,
  545,  546,  547,  548,  549,  550,   -1,  552,  553,  554,
  555,   58,  557,   -1,  559,  560,  561,  562,  563,  564,
  565,  566,  567,  568,  569,  570,  571,  572,  573,  574,
  575,  576,   -1,  578,  579,  580,  581,   -1,  583,  584,
  585,  586,  587,   -1,  589,  590,  591,  592,  593,  594,
  595,  596,  597,  598,  599,  600,  601,   -1,  603,  604,
  605,  606,  607,  608,   -1,  610,  611,  612,  613,  614,
  615,  616,  617,   -1,  619,  620,  621,  622,   -1,  624,
  625,  626,  627,  628,  629,  630,   -1,  632,  633,  634,
  635,  636,  637,  638,  639,  640,   -1,   -1,   -1,   -1,
   -1,  646,   -1,  648,  649,  650,  651,  652,  653,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  672,  673,   -1,
  675,  676,  677,  678,  679,  680,  681,  682,  683,   -1,
  685,  686,  687,  688,  689,  690,  691,  692,  693,  694,
  695,  696,   -1,  698,  699,  700,   -1,  702,   -1,   -1,
   -1,  706,  707,  708,  709,  710,  711,  712,  713,  714,
  715,  716,  717,  718,   -1,  720,  721,   -1,   -1,  724,
  725,  726,  727,  728,  729,  730,  731,  732,  733,  734,
  735,  736,  737,  738,  739,  740,   -1,   -1,  743,  744,
   -1,   -1,   -1,   -1,  749,  750,  751,  752,  753,  754,
  257,   -1,  259,   -1,   -1,   -1,   -1,   -1,  265,   -1,
   -1,  268,  269,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  279,  280,   -1,  282,   -1,  284,   -1,   -1,
  287,   -1,   -1,  290,   -1,   -1,   -1,   -1,   -1,   -1,
  297,   -1,  299,  300,   -1,   -1,   -1,  304,   -1,  306,
   -1,   -1,  309,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  319,  320,   -1,   -1,   -1,  324,   -1,  326,
   -1,  328,   -1,   -1,   -1,   -1,   -1,  334,   -1,  336,
   -1,   -1,   -1,  340,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  348,   -1,   -1,  351,   -1,  353,   -1,   -1,   -1,
  357,  358,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  370,   -1,   -1,   -1,  374,  375,  376,
  377,  378,   -1,  380,   -1,   -1,   -1,   -1,   -1,  386,
   -1,  388,  389,   -1,   -1,  392,  393,   -1,   -1,   -1,
   -1,   -1,  399,   -1,   -1,  402,  403,  404,  405,   -1,
  407,   -1,  409,   -1,  411,  412,  413,  414,  415,   -1,
   -1,   -1,   -1,  420,   -1,  422,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  433,   -1,   -1,   -1,
   -1,  438,  439,   -1,   -1,   -1,   -1,   -1,  445,   -1,
   -1,   -1,  449,   -1,  451,  452,  453,  454,   -1,  456,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  464,  465,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  475,  476,
   -1,  478,  479,  480,   -1,   -1,   -1,   -1,  485,  486,
   -1,  488,  489,   -1,  491,   -1,   -1,  494,  495,   -1,
   -1,   -1,  499,  500,   -1,   -1,   -1,   -1,  505,  506,
  507,   -1,  509,  510,  511,   -1,  513,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  523,  524,  525,  526,
  527,  528,  529,  530,  531,  532,  533,  534,  535,  536,
  537,  538,  539,  540,  541,  542,  543,   -1,  545,  546,
  547,  548,  549,  550,   -1,  552,  553,  554,  555,   -1,
  557,   -1,  559,  560,  561,  562,  563,  564,  565,  566,
  567,  568,  569,  570,  571,  572,  573,  574,  575,  576,
   -1,  578,  579,  580,  581,   -1,  583,  584,  585,  586,
  587,   -1,  589,  590,  591,  592,  593,  594,  595,  596,
  597,  598,  599,  600,  601,   -1,  603,  604,  605,  606,
  607,  608,   -1,  610,  611,  612,  613,  614,  615,  616,
  617,   -1,  619,  620,   -1,   -1,   -1,  624,  625,  626,
  627,  628,  629,   -1,   -1,  632,  633,  634,  635,  636,
  637,  638,  639,  640,   -1,   -1,   -1,   -1,   -1,  646,
   -1,  648,  649,  650,  651,  652,  653,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  672,  673,   -1,  675,  676,
  677,  678,  679,  680,  681,  682,  683,   -1,  685,  686,
  687,  688,  689,  690,  691,  692,  693,  694,  695,  696,
   -1,  698,  699,  700,   -1,  702,   -1,   -1,   -1,  706,
  707,  708,  709,  710,  711,  712,  713,  714,  715,  716,
  717,  718,   -1,  720,  721,   -1,   -1,   -1,   -1,   -1,
  727,  728,  729,  730,  731,  732,  733,  734,  735,  736,
  737,  738,  739,  740,   -1,   -1,  743,  744,   -1,   -1,
   -1,   -1,  749,  750,  751,  752,  753,  754,   42,   43,
   -1,   45,   46,   47,   -1,  272,   -1,   -1,  275,   -1,
   -1,   -1,   -1,   -1,   -1,   59,   60,   61,   62,   -1,
   -1,   -1,   -1,   -1,  291,   -1,  293,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   91,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  331,   -1,   -1,   -1,   -1,   -1,
  337,  338,  339,   -1,  341,  342,  343,   -1,  345,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  359,   -1,  361,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  390,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  406,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  428,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  466,
  467,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  483,   -1,  262,   -1,
  264,   -1,  266,   -1,   -1,   -1,   -1,  271,   -1,   -1,
   -1,   -1,   -1,  277,   -1,   -1,   -1,  281,   -1,   -1,
  284,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  318,   -1,   -1,   -1,   -1,   -1,
   -1,  325,   -1,   -1,   -1,   -1,   -1,   -1,  332,  333,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  350,   -1,  352,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  362,   -1,
  364,  365,  366,   -1,  368,   -1,   -1,  371,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  380,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  412,  257,
   -1,  259,   -1,   -1,   -1,  642,   -1,  265,   -1,   -1,
  268,  269,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  279,  280,  437,  282,   -1,  284,   -1,   -1,  287,
   -1,   -1,  290,   -1,   -1,   -1,   -1,  674,   -1,  297,
  454,  299,  300,   -1,   -1,   -1,  304,  684,  306,   -1,
   -1,  309,   -1,  690,   -1,   -1,   -1,   -1,   -1,   -1,
  697,  319,  320,   -1,   -1,  702,  324,   -1,  326,   -1,
  328,   -1,   -1,  710,   -1,   -1,  334,   -1,  336,   -1,
   -1,   -1,  340,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  348,   -1,   -1,  351,   -1,  353,   -1,   -1,   -1,  357,
  358,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  746,
   -1,   -1,  370,   -1,   -1,   -1,  374,  375,  376,  377,
  378,   -1,  380,   -1,   -1,   -1,   -1,   -1,  386,   -1,
  388,  389,   -1,   -1,  392,  393,   -1,   -1,   -1,   -1,
   -1,  399,   -1,   -1,  402,  403,  404,  405,   -1,  407,
   -1,  409,   -1,  411,  412,  413,  414,  415,   -1,   -1,
   -1,   -1,  420,   -1,  422,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  433,   -1,   -1,   -1,   -1,
  438,  439,   -1,   -1,   -1,   -1,   -1,  445,  602,   -1,
   -1,  449,  450,  451,  452,  453,  454,  455,  456,  457,
  458,  459,  460,  461,  462,  463,  464,  465,  466,  467,
  468,  469,  470,   -1,   -1,  473,  474,  475,  476,  477,
  478,  479,  480,  481,  482,  483,  484,  485,  486,  487,
  488,  489,  490,  491,  492,  493,  494,  495,  496,  497,
  498,  499,  500,  501,  502,  503,  504,  505,  506,  507,
  508,  509,  510,  511,  512,  513,  514,  515,  516,  517,
  518,  519,  520,  521,  522,  523,  524,  525,  526,  527,
  528,  529,  530,  531,  532,  533,  534,  535,  536,  537,
  538,  539,  540,  541,  542,  543,  544,  545,  546,  547,
  548,  549,  550,  551,  552,  553,  554,  555,  556,  557,
  558,  559,  560,  561,  562,  563,  564,  565,  566,  567,
  568,  569,  570,  571,  572,  573,  574,  575,  576,  577,
  578,  579,  580,  581,  582,  583,  584,  585,  586,  587,
  588,  589,  590,  591,  592,  593,  594,  595,  596,  597,
  598,  599,  600,  601,  602,  603,  604,  605,  606,  607,
  608,   -1,  610,  611,  612,  613,  614,  615,  616,  617,
  618,  619,  620,   -1,   -1,  623,  624,  625,  626,  627,
  628,  629,  630,  631,  632,  633,  634,  635,  636,  637,
  638,  639,  640,  641,  642,  643,  644,  645,  646,   -1,
  648,  649,  650,  651,  652,  653,  654,  655,  656,  657,
  658,  659,  660,  661,  662,  663,  664,  665,  666,  667,
  668,  669,  670,  671,  672,  673,  674,  675,  676,  677,
  678,  679,  680,  681,  682,  683,  684,  685,  686,  687,
  688,  689,  690,  691,  692,  693,  694,  695,  696,  697,
  698,  699,  700,  701,  702,  703,  704,  705,  706,  707,
  708,  709,  710,  711,  712,  713,  714,  715,  716,  717,
  718,  719,  720,  721,  722,   -1,   -1,   -1,   -1,  727,
  728,  729,  730,  731,  732,  733,  734,  735,  736,  737,
  738,  739,  740,  741,  742,  743,  744,  745,  746,  747,
  748,  749,  750,  751,  752,  753,  754,  257,   -1,  259,
   -1,   -1,   -1,   -1,   -1,  265,   -1,   -1,  268,  269,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  279,
  280,   -1,  282,   -1,  284,   -1,   -1,  287,   -1,   -1,
  290,   -1,   -1,   -1,   -1,   -1,   -1,  297,   -1,  299,
  300,   -1,   -1,   -1,  304,   -1,  306,   -1,   -1,  309,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  319,
  320,   -1,   -1,   -1,  324,   -1,  326,   -1,  328,   -1,
   -1,   -1,   -1,   -1,  334,   -1,  336,   -1,   -1,   -1,
  340,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  348,   -1,
   -1,  351,   -1,  353,   -1,   -1,   -1,  357,  358,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  370,   -1,   -1,   -1,  374,  375,  376,  377,  378,   -1,
  380,   -1,   -1,   -1,   -1,   -1,  386,   -1,  388,  389,
   -1,   -1,  392,  393,   -1,   -1,   -1,   -1,   -1,  399,
   -1,   -1,  402,  403,  404,  405,   -1,  407,   -1,  409,
   -1,  411,  412,  413,  414,  415,   -1,   -1,   -1,   -1,
  420,   -1,  422,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  433,   -1,   -1,   -1,   -1,  438,  439,
   -1,   -1,   -1,   -1,   -1,  445,   -1,   -1,   -1,  449,
  450,  451,  452,  453,  454,   -1,  456,  457,  458,  459,
  460,  461,  462,  463,  464,  465,  466,  467,  468,  469,
  470,   -1,   -1,  473,  474,  475,  476,  477,  478,  479,
  480,  481,  482,  483,  484,  485,  486,  487,  488,  489,
  490,  491,  492,  493,  494,  495,  496,  497,  498,  499,
  500,  501,  502,  503,  504,  505,  506,  507,  508,  509,
  510,  511,  512,  513,  514,  515,  516,  517,  518,  519,
  520,  521,  522,  523,  524,  525,  526,  527,  528,  529,
  530,  531,  532,  533,  534,  535,  536,  537,  538,  539,
  540,  541,  542,  543,  544,  545,  546,  547,  548,  549,
  550,  551,  552,  553,  554,  555,  556,  557,  558,  559,
  560,  561,  562,  563,  564,  565,  566,  567,  568,  569,
  570,  571,  572,  573,  574,  575,  576,  577,  578,  579,
  580,  581,  582,  583,  584,  585,  586,  587,  588,  589,
  590,  591,  592,  593,  594,  595,  596,  597,  598,  599,
  600,  601,  602,  603,  604,  605,  606,  607,  608,   -1,
  610,  611,  612,  613,  614,  615,  616,  617,  618,  619,
  620,   -1,   -1,  623,  624,  625,  626,  627,  628,  629,
  630,  631,  632,  633,  634,  635,  636,  637,  638,  639,
  640,  641,  642,  643,  644,  645,  646,   -1,  648,  649,
  650,  651,  652,  653,  654,  655,  656,  657,  658,  659,
  660,  661,  662,  663,  664,  665,  666,  667,  668,  669,
  670,  671,  672,  673,  674,  675,  676,  677,  678,  679,
  680,  681,  682,  683,  684,  685,  686,  687,  688,  689,
  690,  691,  692,  693,  694,  695,  696,  697,  698,  699,
  700,  701,  702,  703,  704,  705,  706,  707,  708,  709,
  710,  711,  712,  713,  714,  715,  716,  717,  718,  719,
  720,  721,  722,   -1,   -1,   -1,   -1,  727,  728,  729,
  730,  731,  732,  733,  734,  735,  736,  737,   -1,  739,
  740,  741,  742,  743,  744,  745,  746,  747,  748,  749,
  750,  751,  752,  753,  754,  257,   -1,  259,   -1,   -1,
   -1,   -1,   -1,  265,   -1,   -1,  268,  269,   -1,   -1,
  272,   -1,   -1,  275,   -1,   -1,   -1,  279,  280,   -1,
  282,   -1,  284,   -1,   -1,  287,   -1,  289,  290,  291,
   -1,  293,   -1,   -1,   -1,  297,   -1,  299,  300,   -1,
   -1,   -1,  304,   -1,  306,   -1,   -1,  309,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  319,  320,   -1,
   -1,   -1,  324,   -1,  326,   -1,  328,   -1,   -1,  331,
   -1,   -1,  334,   -1,  336,  337,  338,  339,  340,  341,
  342,  343,   -1,  345,   -1,   -1,  348,   -1,   -1,  351,
   -1,  353,   -1,   -1,   -1,  357,  358,  359,   -1,  361,
   -1,  363,   -1,   -1,   -1,   -1,   -1,   -1,  370,   -1,
   -1,   -1,  374,  375,  376,  377,  378,   -1,  380,   -1,
   -1,   -1,   -1,   -1,  386,   -1,  388,  389,  390,   -1,
  392,  393,   -1,   -1,   -1,   -1,   -1,  399,   -1,   -1,
  402,  403,  404,  405,  406,  407,   -1,  409,   -1,  411,
  412,  413,  414,  415,   -1,   -1,   -1,   -1,  420,   -1,
  422,   -1,   -1,   -1,   -1,   -1,  428,   -1,   -1,   -1,
   -1,  433,   -1,   -1,   -1,   -1,  438,  439,   -1,   -1,
   -1,   -1,   -1,  445,   -1,   -1,   -1,  449,   -1,  451,
  452,  453,  454,   -1,  456,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  464,  465,  466,  467,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  475,  476,   -1,  478,  479,  480,   -1,
   -1,  483,   -1,  485,  486,   -1,  488,  489,   -1,  491,
   -1,   -1,  494,  495,   -1,   -1,   -1,  499,  500,   -1,
   -1,   -1,   -1,  505,  506,  507,   -1,  509,  510,  511,
   -1,  513,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  523,  524,  525,  526,  527,  528,  529,  530,  531,
  532,  533,  534,  535,  536,  537,  538,  539,  540,  541,
  542,  543,   -1,  545,  546,  547,  548,  549,  550,   -1,
  552,  553,  554,  555,   -1,  557,   -1,  559,  560,  561,
  562,  563,  564,  565,  566,  567,  568,  569,  570,  571,
  572,  573,  574,  575,  576,   -1,  578,  579,  580,  581,
   -1,  583,  584,  585,  586,  587,   -1,  589,  590,  591,
  592,  593,  594,  595,  596,  597,  598,  599,  600,  601,
   -1,  603,  604,  605,  606,  607,  608,   -1,  610,  611,
  612,  613,  614,  615,  616,  617,   -1,  619,  620,   -1,
   -1,  623,  624,  625,  626,  627,  628,  629,   -1,   -1,
  632,  633,  634,  635,  636,  637,  638,  639,  640,   -1,
  642,   -1,   -1,   -1,  646,   -1,  648,  649,  650,  651,
  652,  653,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  672,  673,  674,  675,  676,  677,  678,  679,  680,  681,
  682,  683,  684,  685,  686,  687,  688,  689,  690,  691,
  692,  693,  694,  695,  696,  697,  698,  699,  700,   -1,
  702,   -1,   -1,   -1,  706,  707,  708,  709,  710,  711,
  712,  713,  714,  715,  716,  717,  718,   -1,  720,  721,
   -1,   -1,   -1,   -1,   -1,  727,  728,  729,  730,  731,
  732,  733,  734,  735,  736,  737,  738,  739,  740,   -1,
   -1,  743,  744,   -1,  746,   -1,   -1,  749,  750,  751,
  752,  753,  754,  257,   -1,  259,   -1,   -1,   -1,   -1,
   -1,  265,   -1,   -1,  268,  269,   -1,   -1,  272,   -1,
   -1,  275,   -1,   -1,   -1,  279,  280,   -1,  282,   -1,
  284,   -1,   -1,  287,   -1,   -1,  290,  291,   -1,  293,
   -1,   -1,   -1,  297,   -1,  299,  300,   -1,   -1,   -1,
  304,   -1,  306,   -1,   -1,  309,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  319,  320,   -1,   -1,   -1,
  324,   -1,  326,   -1,  328,   -1,   -1,  331,   -1,   -1,
  334,   -1,  336,  337,  338,  339,  340,  341,  342,  343,
   -1,  345,   -1,   -1,  348,   -1,   -1,  351,   -1,  353,
   -1,   -1,   -1,  357,  358,  359,   -1,  361,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  370,   -1,   -1,   -1,
  374,  375,  376,  377,  378,   -1,  380,   -1,   -1,   -1,
   -1,   -1,  386,   -1,  388,  389,  390,   -1,  392,  393,
   -1,   -1,   -1,   -1,   -1,  399,   -1,   -1,  402,  403,
  404,  405,  406,  407,   -1,  409,   -1,  411,  412,  413,
  414,  415,   -1,   -1,   -1,   -1,  420,   -1,  422,   -1,
   -1,   -1,   -1,   -1,  428,   -1,   -1,   -1,   -1,  433,
   -1,   -1,   -1,   -1,  438,  439,   -1,   -1,   -1,   -1,
   -1,  445,   -1,   -1,   -1,  449,   -1,  451,  452,  453,
  454,   -1,  456,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  464,  465,  466,  467,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  475,  476,   -1,  478,  479,  480,   -1,   -1,  483,
   -1,  485,  486,   -1,  488,  489,   -1,  491,   -1,   -1,
  494,  495,   -1,   -1,   -1,  499,  500,   -1,   -1,   -1,
   -1,  505,  506,  507,   -1,  509,  510,  511,   -1,  513,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  523,
  524,  525,  526,  527,  528,  529,  530,  531,  532,  533,
  534,  535,  536,  537,  538,  539,  540,  541,  542,  543,
   -1,  545,  546,  547,  548,  549,  550,   -1,  552,  553,
  554,  555,   -1,  557,   -1,  559,  560,  561,  562,  563,
  564,  565,  566,  567,  568,  569,  570,  571,  572,  573,
  574,  575,  576,   -1,  578,  579,  580,  581,   -1,  583,
  584,  585,  586,  587,   -1,  589,  590,  591,  592,  593,
  594,  595,  596,  597,  598,  599,  600,  601,   -1,  603,
  604,  605,  606,  607,  608,   -1,  610,  611,  612,  613,
  614,  615,  616,  617,   -1,  619,  620,   -1,   -1,   -1,
  624,  625,  626,  627,  628,  629,   -1,   -1,  632,  633,
  634,  635,  636,  637,  638,  639,  640,   -1,  642,   -1,
   -1,   -1,  646,   -1,  648,  649,  650,  651,  652,  653,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  672,  673,
  674,  675,  676,  677,  678,  679,  680,  681,  682,  683,
  684,  685,  686,  687,  688,  689,  690,  691,  692,  693,
  694,  695,  696,  697,  698,  699,  700,   -1,  702,   -1,
   -1,   -1,  706,  707,  708,  709,  710,  711,  712,  713,
  714,  715,  716,  717,  718,   -1,  720,  721,   -1,   -1,
   -1,   -1,   -1,  727,  728,  729,  730,  731,  732,  733,
  734,  735,  736,  737,  738,  739,  740,   -1,   -1,  743,
  744,   -1,  746,   -1,   -1,  749,  750,  751,  752,  753,
  754,  257,   -1,  259,   -1,   -1,   -1,   -1,   -1,  265,
   -1,   -1,  268,  269,   -1,   -1,  272,   -1,   -1,  275,
   -1,   -1,   -1,  279,  280,   -1,  282,   -1,  284,   -1,
   -1,  287,   -1,   -1,  290,  291,   -1,  293,   -1,   -1,
   -1,  297,   -1,  299,  300,   -1,   -1,   -1,  304,   -1,
  306,   -1,   -1,  309,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  319,  320,   -1,   -1,   -1,  324,   -1,
  326,   -1,  328,   -1,   -1,  331,   -1,   -1,  334,   -1,
  336,  337,  338,  339,  340,  341,  342,  343,   -1,  345,
   -1,   -1,  348,   -1,   -1,  351,   -1,  353,   -1,   -1,
   -1,  357,  358,  359,   -1,  361,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  370,   -1,   -1,   -1,  374,  375,
  376,  377,  378,   -1,  380,   -1,   -1,   -1,   -1,   -1,
  386,   -1,  388,  389,  390,   -1,  392,  393,   -1,   -1,
   -1,   -1,   -1,  399,   -1,   -1,  402,  403,  404,  405,
  406,  407,   -1,  409,   -1,  411,  412,  413,  414,  415,
   -1,   -1,   -1,   -1,  420,   -1,  422,   -1,   -1,   -1,
   -1,   -1,  428,   -1,   -1,   -1,   -1,  433,   -1,   -1,
   -1,   -1,  438,  439,   -1,   -1,   -1,   -1,   -1,  445,
   -1,   -1,   -1,  449,   -1,  451,  452,  453,  454,   -1,
  456,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  464,  465,
  466,  467,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  475,
  476,   -1,  478,  479,  480,   -1,   -1,  483,   -1,  485,
  486,   -1,  488,  489,   -1,  491,   -1,   -1,  494,  495,
   -1,   -1,   -1,  499,  500,   -1,   -1,   -1,   -1,  505,
  506,  507,   -1,  509,  510,  511,   -1,  513,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  523,  524,  525,
  526,  527,  528,  529,  530,  531,  532,  533,  534,  535,
  536,  537,  538,  539,  540,  541,  542,  543,   -1,  545,
  546,  547,  548,  549,  550,   -1,  552,  553,  554,  555,
   -1,  557,   -1,  559,  560,  561,  562,  563,  564,  565,
  566,  567,  568,  569,  570,  571,  572,  573,  574,  575,
  576,   -1,  578,  579,  580,  581,   -1,  583,  584,  585,
  586,  587,   -1,  589,  590,  591,  592,  593,  594,  595,
  596,  597,  598,  599,  600,  601,   -1,  603,  604,  605,
  606,  607,  608,   -1,  610,  611,  612,  613,  614,  615,
  616,  617,   -1,  619,  620,   -1,   -1,   -1,  624,  625,
  626,  627,  628,  629,   -1,   -1,  632,  633,  634,  635,
  636,  637,  638,  639,  640,   -1,  642,   -1,   -1,   -1,
  646,   -1,  648,  649,  650,  651,  652,  653,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  672,  673,  674,  675,
  676,  677,  678,  679,  680,  681,  682,  683,  684,  685,
  686,  687,  688,  689,  690,  691,  692,  693,  694,  695,
  696,  697,  698,  699,  700,   -1,  702,   -1,   -1,   -1,
  706,  707,  708,  709,  710,  711,  712,  713,  714,  715,
  716,  717,  718,   -1,  720,  721,   -1,   -1,   -1,   -1,
   -1,  727,  728,  729,  730,  731,  732,  733,  734,  735,
  736,  737,  738,  739,  740,   -1,   -1,  743,  744,   -1,
  746,   -1,   -1,  749,  750,  751,  752,  753,  754,  257,
   -1,  259,   -1,   -1,   -1,   -1,   -1,  265,   -1,   -1,
  268,  269,   -1,   -1,  272,   -1,   -1,  275,   -1,   -1,
   -1,  279,  280,   -1,  282,   -1,  284,   -1,   -1,  287,
   -1,   -1,  290,  291,   -1,  293,   -1,   -1,   -1,  297,
   -1,  299,  300,   -1,   -1,   -1,  304,   -1,  306,   -1,
   -1,  309,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  319,  320,   -1,   -1,   -1,  324,   -1,  326,   -1,
  328,   -1,   -1,  331,   -1,   -1,  334,   -1,  336,  337,
  338,  339,  340,  341,  342,  343,   -1,  345,   -1,   -1,
  348,   -1,   -1,  351,   -1,  353,   -1,   -1,   -1,  357,
  358,  359,   -1,  361,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  370,   -1,   -1,   -1,  374,  375,  376,  377,
  378,   -1,  380,   -1,   -1,   -1,   -1,   -1,  386,   -1,
  388,  389,  390,   -1,  392,  393,   -1,   -1,   -1,   -1,
   -1,  399,   -1,   -1,  402,  403,  404,  405,  406,  407,
   -1,  409,   -1,  411,  412,  413,  414,  415,   -1,   -1,
   -1,   -1,  420,   -1,  422,   -1,   -1,   -1,   -1,   -1,
  428,   -1,   -1,   -1,   -1,  433,   -1,   -1,   -1,   -1,
  438,  439,   -1,   -1,   -1,   -1,   -1,  445,   -1,   -1,
   -1,  449,   -1,  451,  452,  453,  454,   -1,  456,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  464,  465,  466,  467,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  475,  476,   -1,
  478,  479,  480,   -1,   -1,  483,   -1,  485,  486,   -1,
  488,  489,   -1,  491,   -1,   -1,  494,  495,   -1,   -1,
   -1,  499,  500,   -1,   -1,   -1,   -1,  505,  506,  507,
   -1,  509,  510,  511,   -1,  513,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  523,  524,  525,  526,  527,
  528,  529,  530,  531,  532,  533,  534,  535,  536,  537,
  538,  539,  540,  541,  542,  543,   -1,  545,  546,  547,
  548,  549,  550,   -1,  552,  553,  554,  555,   -1,  557,
   -1,  559,  560,  561,  562,  563,  564,  565,  566,  567,
  568,  569,  570,  571,  572,  573,  574,  575,  576,   -1,
  578,  579,  580,  581,   -1,  583,  584,  585,  586,  587,
   -1,  589,  590,  591,  592,  593,  594,  595,  596,  597,
  598,  599,  600,  601,   -1,  603,  604,  605,  606,  607,
  608,   -1,  610,  611,  612,  613,  614,  615,  616,  617,
   -1,  619,  620,   -1,   -1,   -1,  624,  625,  626,  627,
  628,  629,   -1,   -1,  632,  633,  634,  635,  636,  637,
  638,  639,  640,   -1,  642,   -1,   -1,   -1,  646,   -1,
  648,  649,  650,  651,  652,  653,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  672,  673,  674,  675,  676,  677,
  678,  679,  680,  681,  682,  683,  684,  685,  686,  687,
  688,  689,  690,  691,  692,  693,  694,  695,  696,  697,
  698,  699,  700,   -1,  702,   -1,   -1,   -1,  706,  707,
  708,  709,  710,  711,  712,  713,  714,  715,  716,  717,
  718,   -1,  720,  721,   -1,   -1,   -1,   -1,   -1,  727,
  728,  729,  730,  731,  732,  733,  734,  735,  736,  737,
  738,  739,  740,   -1,   -1,  743,  744,   -1,  746,   -1,
   -1,  749,  750,  751,  752,  753,  754,  257,   -1,  259,
   -1,   -1,   -1,   -1,   -1,  265,   -1,   -1,  268,  269,
   -1,   -1,  272,   -1,   -1,  275,   -1,   -1,   -1,  279,
  280,   -1,  282,   -1,  284,   -1,   -1,  287,   -1,   -1,
  290,  291,   -1,  293,   -1,   -1,   -1,  297,   -1,  299,
  300,   -1,   -1,   -1,  304,   -1,  306,   -1,   -1,  309,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  319,
  320,   -1,   -1,   -1,  324,   -1,  326,   -1,  328,   -1,
   -1,  331,   -1,   -1,  334,   -1,  336,  337,  338,  339,
  340,  341,  342,  343,   -1,  345,   -1,   -1,  348,   -1,
   -1,  351,   -1,  353,   -1,   -1,   -1,  357,  358,  359,
   -1,  361,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  370,   -1,   -1,   -1,  374,  375,  376,  377,  378,   -1,
  380,   -1,   -1,   -1,   -1,   -1,  386,   -1,  388,  389,
  390,   -1,  392,  393,   -1,   -1,   -1,   -1,   -1,  399,
   -1,   -1,  402,  403,  404,  405,  406,  407,   -1,  409,
   -1,  411,  412,  413,  414,  415,   -1,   -1,   -1,   -1,
  420,   -1,  422,   -1,   -1,   -1,   -1,   -1,  428,   -1,
   -1,   -1,   -1,  433,   -1,   -1,   -1,   -1,  438,  439,
   -1,   -1,   -1,   -1,   -1,  445,   -1,   -1,   -1,  449,
   -1,  451,  452,  453,  454,   -1,  456,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  464,  465,  466,  467,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  475,  476,   -1,  478,  479,
  480,   -1,   -1,  483,   -1,  485,  486,   -1,  488,  489,
   -1,  491,   -1,   -1,  494,  495,   -1,   -1,   -1,  499,
  500,   -1,   -1,   -1,   -1,  505,  506,  507,   -1,  509,
  510,  511,   -1,  513,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  523,  524,  525,  526,  527,  528,  529,
  530,  531,  532,  533,  534,  535,  536,  537,  538,  539,
  540,  541,  542,  543,   -1,  545,  546,  547,  548,  549,
  550,   -1,  552,  553,  554,  555,   -1,  557,   -1,  559,
  560,  561,  562,  563,  564,  565,  566,  567,  568,  569,
  570,  571,  572,  573,  574,  575,  576,   -1,  578,  579,
  580,  581,   -1,  583,  584,  585,  586,  587,   -1,  589,
  590,  591,  592,  593,  594,  595,  596,  597,  598,  599,
  600,  601,   -1,  603,  604,  605,  606,  607,  608,   -1,
  610,  611,  612,  613,  614,  615,  616,  617,   -1,  619,
  620,   -1,   -1,   -1,  624,  625,  626,  627,  628,  629,
   -1,   -1,  632,  633,  634,  635,  636,  637,  638,  639,
  640,   -1,  642,   -1,   -1,   -1,  646,   -1,  648,  649,
  650,  651,  652,  653,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  672,  673,  674,  675,  676,  677,  678,  679,
  680,  681,  682,  683,  684,  685,  686,  687,  688,  689,
  690,  691,  692,  693,  694,  695,  696,  697,  698,  699,
  700,   -1,  702,   -1,   -1,   -1,  706,  707,  708,  709,
  710,  711,  712,  713,  714,  715,  716,  717,  718,   -1,
  720,  721,   -1,   -1,   -1,   -1,   -1,  727,  728,  729,
  730,  731,  732,  733,  734,  735,  736,  737,  738,  739,
  740,   -1,   -1,  743,  744,   -1,  746,   -1,   -1,  749,
  750,  751,  752,  753,  754,  257,   -1,  259,  260,  261,
   -1,   -1,   -1,  265,   -1,   -1,  268,  269,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  279,  280,   -1,
  282,   -1,  284,   -1,  286,  287,   -1,   -1,  290,   -1,
   -1,   -1,   -1,  295,  296,  297,   -1,  299,  300,  301,
   -1,   -1,  304,   -1,  306,  307,   -1,  309,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  319,  320,   -1,
   -1,   -1,  324,   -1,  326,   -1,  328,   -1,  330,   -1,
   -1,   -1,  334,   -1,  336,   -1,   -1,   -1,  340,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  348,   -1,   -1,  351,
   -1,  353,   -1,   -1,   -1,  357,  358,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  370,   -1,
   -1,   -1,  374,  375,  376,  377,  378,   -1,  380,   -1,
   -1,   -1,   -1,   -1,  386,   -1,  388,  389,   -1,  391,
  392,  393,   -1,   -1,   -1,   -1,   -1,  399,  400,   -1,
  402,  403,  404,  405,   -1,  407,   -1,  409,   -1,  411,
  412,  413,  414,  415,   -1,   -1,   -1,   -1,  420,   -1,
  422,   -1,   -1,  425,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  433,   -1,   -1,   -1,   -1,  438,  439,   -1,   -1,
   -1,   -1,   -1,  445,   -1,   -1,   -1,  449,   -1,  451,
  452,  453,  454,   -1,  456,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  475,  476,   -1,  478,  479,  480,   -1,
   -1,   -1,   -1,  485,  486,   -1,  488,  489,   -1,  491,
   -1,   -1,  494,  495,   -1,   -1,   -1,  499,  500,   -1,
   -1,   -1,   -1,  505,  506,  507,   -1,  509,  510,  511,
   -1,  513,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  523,  524,  525,  526,  527,  528,  529,  530,  531,
  532,  533,  534,  535,  536,  537,  538,  539,  540,  541,
  542,  543,   -1,  545,  546,  547,  548,  549,  550,   -1,
  552,  553,  554,  555,   -1,  557,   -1,  559,  560,  561,
  562,  563,  564,  565,  566,  567,  568,  569,  570,  571,
  572,  573,  574,  575,  576,   -1,  578,  579,  580,  581,
   -1,  583,  584,  585,  586,  587,   -1,  589,  590,  591,
  592,  593,  594,  595,  596,  597,  598,  599,  600,  601,
   -1,  603,  604,  605,  606,  607,  608,   -1,  610,  611,
  612,  613,  614,  615,  616,  617,   -1,  619,  620,   -1,
   -1,   -1,  624,  625,  626,  627,  628,  629,   -1,   -1,
  632,  633,  634,  635,  636,  637,  638,  639,  640,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  648,  649,  650,  651,
  652,  653,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  672,  673,   -1,  675,  676,  677,  678,  679,  680,  681,
  682,  683,   -1,  685,  686,  687,  688,  689,  690,  691,
  692,  693,  694,  695,  696,   -1,  698,  699,  700,   -1,
  702,   -1,   -1,   -1,  706,  707,  708,  709,  710,  711,
  712,  713,  714,  715,  716,  717,  718,   -1,  720,  721,
   -1,   -1,   -1,   -1,   -1,  727,  728,  729,  730,  731,
  732,  733,  734,  735,  736,  737,  738,  739,  740,   -1,
   -1,  743,  744,   -1,   -1,   -1,   -1,  749,  750,  751,
  752,  753,  754,  257,   -1,  259,   -1,  261,   -1,   -1,
   -1,  265,   -1,   -1,  268,  269,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  279,  280,   -1,  282,   -1,
  284,   -1,  286,  287,   -1,   -1,  290,   -1,   -1,   -1,
   -1,  295,  296,  297,   -1,  299,  300,  301,   -1,   -1,
  304,   -1,  306,  307,   -1,  309,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  319,  320,   -1,   -1,   -1,
  324,   -1,  326,   -1,  328,   -1,  330,   -1,   -1,   -1,
  334,   -1,  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  348,   -1,   -1,  351,   -1,  353,
   -1,   -1,   -1,  357,  358,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  370,   -1,   -1,   -1,
  374,  375,  376,  377,  378,   -1,  380,   -1,   -1,   -1,
   -1,   -1,  386,   -1,  388,  389,   -1,  391,  392,  393,
   -1,   -1,   -1,   -1,   -1,  399,  400,   -1,  402,  403,
  404,  405,   -1,  407,   -1,  409,   -1,  411,  412,  413,
  414,  415,   -1,   -1,   -1,   -1,  420,   -1,  422,   -1,
   -1,  425,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  433,
   -1,   -1,   -1,   -1,  438,  439,   -1,   -1,   -1,   -1,
   -1,  445,   -1,   -1,   -1,  449,   -1,  451,  452,  453,
  454,   -1,  456,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  475,  476,   -1,  478,  479,  480,   -1,   -1,   -1,
   -1,  485,  486,   -1,  488,  489,   -1,  491,   -1,   -1,
  494,  495,   -1,   -1,   -1,  499,  500,   -1,   -1,   -1,
   -1,  505,  506,  507,   -1,  509,  510,  511,   -1,  513,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  523,
  524,  525,  526,  527,  528,  529,  530,  531,  532,  533,
  534,  535,  536,  537,  538,  539,  540,  541,  542,  543,
   -1,  545,  546,  547,  548,  549,  550,   -1,  552,  553,
  554,  555,   -1,  557,   -1,  559,  560,  561,  562,  563,
  564,  565,  566,  567,  568,  569,  570,  571,  572,  573,
  574,  575,  576,   -1,  578,  579,  580,  581,   -1,  583,
  584,  585,  586,  587,   -1,  589,  590,  591,  592,  593,
  594,  595,  596,  597,  598,  599,  600,  601,   -1,  603,
  604,  605,  606,  607,  608,   -1,  610,  611,  612,  613,
  614,  615,  616,  617,   -1,  619,  620,   -1,   -1,   -1,
  624,  625,  626,  627,  628,  629,   -1,   -1,  632,  633,
  634,  635,  636,  637,  638,  639,  640,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  648,  649,  650,  651,  652,  653,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  672,  673,
   -1,  675,  676,  677,  678,  679,  680,  681,  682,  683,
   -1,  685,  686,  687,  688,  689,  690,  691,  692,  693,
  694,  695,  696,   -1,  698,  699,  700,   -1,  702,   -1,
   -1,   -1,  706,  707,  708,  709,  710,  711,  712,  713,
  714,  715,  716,  717,  718,   -1,  720,  721,   -1,   -1,
   -1,   -1,   -1,  727,  728,  729,  730,  731,  732,  733,
  734,  735,  736,  737,  738,  739,  740,   -1,   -1,  743,
  744,   -1,   -1,   -1,   -1,  749,  750,  751,  752,  753,
  754,  257,   -1,  259,   -1,   -1,   -1,   -1,   -1,  265,
   -1,   -1,  268,  269,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  279,  280,   -1,  282,   -1,  284,   -1,
   -1,  287,   -1,   -1,  290,   -1,   -1,   -1,   -1,   -1,
   -1,  297,   -1,  299,  300,   -1,   -1,   -1,  304,   -1,
  306,   -1,   -1,  309,   -1,   -1,   -1,   -1,   -1,   -1,
  316,   -1,   -1,  319,  320,   -1,  322,   -1,  324,   -1,
  326,   -1,  328,   -1,   -1,   -1,   -1,   -1,  334,   -1,
  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  348,   -1,   -1,  351,   -1,  353,   -1,   -1,
   -1,  357,  358,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  370,   -1,   -1,   -1,  374,  375,
  376,  377,  378,   -1,  380,   -1,   -1,   -1,   -1,   -1,
  386,  387,  388,  389,   -1,   -1,  392,  393,   -1,   -1,
   -1,   -1,   -1,  399,   -1,   -1,  402,  403,  404,  405,
   -1,  407,   -1,  409,   -1,  411,  412,  413,  414,  415,
   -1,   -1,   -1,   -1,  420,  421,  422,   -1,   -1,   -1,
  426,   -1,   -1,   -1,   -1,   -1,  432,  433,   -1,   -1,
   -1,   -1,  438,  439,   -1,   -1,   -1,   -1,   -1,  445,
   -1,   -1,   -1,  449,   -1,  451,  452,  453,  454,   -1,
  456,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  464,  465,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  475,
  476,   -1,  478,  479,  480,   -1,   -1,   -1,   -1,  485,
  486,   -1,  488,  489,   -1,  491,   -1,   -1,  494,  495,
   -1,   -1,   -1,  499,  500,   -1,   -1,   -1,   -1,  505,
  506,  507,   -1,  509,  510,  511,   -1,  513,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  523,  524,  525,
  526,  527,  528,  529,  530,  531,  532,  533,  534,  535,
  536,  537,  538,  539,  540,  541,  542,  543,   -1,  545,
  546,  547,  548,  549,  550,   -1,  552,  553,  554,  555,
   -1,  557,   -1,  559,  560,  561,  562,  563,  564,  565,
  566,  567,  568,  569,  570,  571,  572,  573,  574,  575,
  576,   -1,  578,  579,  580,  581,   -1,  583,  584,  585,
  586,  587,   -1,  589,  590,  591,  592,  593,  594,  595,
  596,  597,  598,  599,  600,  601,   -1,  603,  604,  605,
  606,  607,  608,   -1,  610,  611,  612,  613,  614,  615,
  616,  617,   -1,  619,  620,   -1,   -1,   -1,  624,  625,
  626,  627,  628,  629,   -1,   -1,  632,  633,  634,  635,
  636,  637,  638,  639,  640,   -1,   -1,   -1,   -1,   -1,
  646,   -1,  648,  649,  650,  651,  652,  653,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  672,  673,   -1,  675,
  676,  677,  678,  679,  680,  681,  682,  683,   -1,  685,
  686,  687,  688,  689,  690,  691,  692,  693,  694,  695,
  696,   -1,  698,  699,  700,   -1,  702,   -1,   -1,   -1,
  706,  707,  708,  709,  710,  711,  712,  713,  714,  715,
  716,  717,  718,   -1,  720,  721,   -1,   -1,   -1,   -1,
   -1,  727,  728,  729,  730,  731,  732,  733,  734,  735,
  736,  737,  738,  739,  740,   -1,   -1,  743,  744,   -1,
   -1,   -1,   -1,  749,  750,  751,  752,  753,  754,  257,
   -1,  259,   -1,   -1,   -1,   -1,   -1,  265,   -1,   -1,
  268,  269,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  279,  280,   -1,  282,   -1,  284,   -1,   -1,  287,
   -1,   -1,  290,   -1,   -1,   -1,   -1,   -1,   -1,  297,
   -1,  299,  300,   -1,   -1,   -1,  304,   -1,  306,   -1,
   -1,  309,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  319,  320,   -1,   -1,   -1,  324,   -1,  326,   -1,
  328,  329,   -1,   -1,   -1,   -1,  334,   -1,  336,   -1,
   -1,   -1,  340,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  348,   -1,   -1,  351,   -1,  353,   -1,   -1,   -1,  357,
  358,   -1,  360,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  370,   -1,  372,   -1,  374,  375,  376,  377,
  378,   -1,  380,   -1,   -1,   -1,   -1,   -1,  386,   -1,
  388,  389,   -1,   -1,  392,  393,   -1,   -1,   -1,   -1,
   -1,  399,   -1,   -1,  402,  403,  404,  405,   -1,  407,
   -1,  409,   -1,  411,  412,  413,  414,  415,   -1,   -1,
   -1,   -1,  420,   -1,  422,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  433,   -1,   -1,   -1,   -1,
  438,  439,   -1,   -1,   -1,   -1,   -1,  445,   -1,   -1,
   -1,  449,   -1,  451,  452,  453,  454,   -1,  456,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  464,  465,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  475,  476,   -1,
  478,  479,  480,   -1,   -1,   -1,   -1,  485,  486,   -1,
  488,  489,   -1,  491,   -1,   -1,  494,  495,   -1,   -1,
   -1,  499,  500,   -1,   -1,   -1,   -1,  505,  506,  507,
   -1,  509,  510,  511,   -1,  513,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  523,  524,  525,  526,  527,
  528,  529,  530,  531,  532,  533,  534,  535,  536,  537,
  538,  539,  540,  541,  542,  543,   -1,  545,  546,  547,
  548,  549,  550,   -1,  552,  553,  554,  555,   -1,  557,
   -1,  559,  560,  561,  562,  563,  564,  565,  566,  567,
  568,  569,  570,  571,  572,  573,  574,  575,  576,   -1,
  578,  579,  580,  581,   -1,  583,  584,  585,  586,  587,
   -1,  589,  590,  591,  592,  593,  594,  595,  596,  597,
  598,  599,  600,  601,   -1,  603,  604,  605,  606,  607,
  608,   -1,  610,  611,  612,  613,  614,  615,  616,  617,
   -1,  619,  620,   -1,   -1,   -1,  624,  625,  626,  627,
  628,  629,   -1,   -1,  632,  633,  634,  635,  636,  637,
  638,  639,  640,   -1,   -1,   -1,   -1,   -1,  646,   -1,
  648,  649,  650,  651,  652,  653,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  672,  673,   -1,  675,  676,  677,
  678,  679,  680,  681,  682,  683,   -1,  685,  686,  687,
  688,  689,  690,  691,  692,  693,  694,  695,  696,   -1,
  698,  699,  700,   -1,  702,   -1,   -1,   -1,  706,  707,
  708,  709,  710,  711,  712,  713,  714,  715,  716,  717,
  718,   -1,  720,  721,   -1,   -1,   -1,   -1,   -1,  727,
  728,  729,  730,  731,  732,  733,  734,  735,  736,  737,
  738,  739,  740,   -1,   -1,  743,  744,   -1,   -1,   -1,
   -1,  749,  750,  751,  752,  753,  754,  257,   -1,  259,
   -1,   -1,   -1,  263,   -1,  265,   -1,   -1,  268,  269,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  279,
  280,   -1,  282,   -1,  284,   -1,   -1,  287,   -1,   -1,
  290,   -1,   -1,   -1,   -1,   -1,   -1,  297,   -1,  299,
  300,   -1,   -1,   -1,  304,   -1,  306,   -1,   -1,  309,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  319,
  320,   -1,  322,   -1,  324,   -1,  326,   -1,  328,   -1,
   -1,   -1,   -1,   -1,  334,   -1,  336,   -1,   -1,   -1,
  340,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  348,   -1,
   -1,  351,   -1,  353,   -1,   -1,   -1,  357,  358,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  370,   -1,   -1,   -1,  374,  375,  376,  377,  378,   -1,
  380,   -1,   -1,   -1,   -1,   -1,  386,   -1,  388,  389,
   -1,   -1,  392,  393,   -1,   -1,   -1,   -1,   -1,  399,
   -1,   -1,  402,  403,  404,  405,   -1,  407,   -1,  409,
   -1,  411,  412,  413,  414,  415,   -1,   -1,   -1,   -1,
  420,   -1,  422,   -1,   -1,   -1,  426,   -1,   -1,   -1,
   -1,   -1,   -1,  433,   -1,   -1,   -1,   -1,  438,  439,
   -1,   -1,   -1,   -1,   -1,  445,   -1,   -1,   -1,  449,
   -1,  451,  452,  453,  454,   -1,  456,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  464,  465,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  475,  476,   -1,  478,  479,
  480,   -1,   -1,   -1,   -1,  485,  486,   -1,  488,  489,
   -1,  491,   -1,   -1,  494,  495,   -1,   -1,   -1,  499,
  500,   -1,   -1,   -1,   -1,  505,  506,  507,   -1,  509,
  510,  511,   -1,  513,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  523,  524,  525,  526,  527,  528,  529,
  530,  531,  532,  533,  534,  535,  536,  537,  538,  539,
  540,  541,  542,  543,   -1,  545,  546,  547,  548,  549,
  550,   -1,  552,  553,  554,  555,   -1,  557,   -1,  559,
  560,  561,  562,  563,  564,  565,  566,  567,  568,  569,
  570,  571,  572,  573,  574,  575,  576,   -1,  578,  579,
  580,  581,   -1,  583,  584,  585,  586,  587,   -1,  589,
  590,  591,  592,  593,  594,  595,  596,  597,  598,  599,
  600,  601,   -1,  603,  604,  605,  606,  607,  608,   -1,
  610,  611,  612,  613,  614,  615,  616,  617,   -1,  619,
  620,   -1,   -1,   -1,  624,  625,  626,  627,  628,  629,
   -1,   -1,  632,  633,  634,  635,  636,  637,  638,  639,
  640,   -1,   -1,   -1,   -1,   -1,  646,   -1,  648,  649,
  650,  651,  652,  653,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  672,  673,   -1,  675,  676,  677,  678,  679,
  680,  681,  682,  683,   -1,  685,  686,  687,  688,  689,
  690,  691,  692,  693,  694,  695,  696,   -1,  698,  699,
  700,   -1,  702,   -1,   -1,   -1,  706,  707,  708,  709,
  710,  711,  712,  713,  714,  715,  716,  717,  718,   -1,
  720,  721,   -1,   -1,   -1,   -1,   -1,  727,  728,  729,
  730,  731,  732,  733,  734,  735,  736,  737,  738,  739,
  740,   -1,   -1,  743,  744,   -1,   -1,   -1,   -1,  749,
  750,  751,  752,  753,  754,  257,   -1,  259,   -1,   -1,
   -1,   -1,   -1,  265,   -1,   -1,  268,  269,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  279,  280,   -1,
  282,   -1,  284,   -1,   -1,  287,   -1,   -1,  290,   -1,
  292,   -1,   -1,   -1,   -1,  297,   -1,  299,  300,   -1,
   -1,   -1,  304,   -1,  306,   -1,   -1,  309,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  319,  320,   -1,
   -1,   -1,  324,   -1,  326,   -1,  328,   -1,   -1,   -1,
   -1,   -1,  334,   -1,  336,   -1,   -1,   -1,  340,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  348,   -1,   -1,  351,
   -1,  353,   -1,   -1,   -1,  357,  358,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  370,   -1,
   -1,   -1,  374,  375,  376,  377,  378,   -1,  380,   -1,
   -1,   -1,   -1,   -1,  386,   -1,  388,  389,   -1,   -1,
  392,  393,   -1,   -1,   -1,   -1,   -1,  399,   -1,   -1,
  402,  403,  404,  405,   -1,  407,   -1,  409,   -1,  411,
  412,  413,  414,  415,   -1,   -1,   -1,   -1,  420,   -1,
  422,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  433,   -1,   -1,   -1,   -1,  438,  439,   -1,   -1,
   -1,   -1,   -1,  445,   -1,   -1,   -1,  449,   -1,  451,
  452,  453,  454,   -1,  456,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  475,  476,   -1,  478,  479,  480,   -1,
   -1,   -1,   -1,  485,  486,   -1,  488,  489,   -1,  491,
   -1,   -1,  494,  495,   -1,   -1,   -1,  499,  500,   -1,
   -1,   -1,   -1,  505,  506,  507,   -1,  509,  510,  511,
   -1,  513,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  523,  524,  525,  526,  527,  528,  529,  530,  531,
  532,  533,  534,  535,  536,  537,  538,  539,  540,  541,
  542,  543,   -1,  545,  546,  547,  548,  549,  550,   -1,
  552,  553,  554,  555,   -1,  557,   -1,  559,  560,  561,
  562,  563,  564,  565,  566,  567,  568,  569,  570,  571,
  572,  573,  574,  575,  576,   -1,  578,  579,  580,  581,
   -1,  583,  584,  585,  586,  587,   -1,  589,  590,  591,
  592,  593,  594,  595,  596,  597,  598,  599,  600,  601,
   -1,  603,  604,  605,  606,  607,  608,   -1,  610,  611,
  612,  613,  614,  615,  616,  617,   -1,  619,  620,   -1,
   -1,   -1,  624,  625,  626,  627,  628,  629,   -1,   -1,
  632,  633,  634,  635,  636,  637,  638,  639,  640,   -1,
   -1,   -1,   -1,   -1,  646,  647,  648,  649,  650,  651,
  652,  653,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  672,  673,   -1,  675,  676,  677,  678,  679,  680,  681,
  682,  683,   -1,  685,  686,  687,  688,  689,  690,  691,
  692,  693,  694,  695,  696,   -1,  698,  699,  700,   -1,
  702,   -1,   -1,   -1,  706,  707,  708,  709,  710,  711,
  712,  713,  714,  715,  716,  717,  718,   -1,  720,  721,
   -1,   -1,   -1,   -1,   -1,  727,  728,  729,  730,  731,
  732,  733,  734,  735,  736,  737,  738,  739,  740,   -1,
   -1,  743,  744,   -1,   -1,   -1,   -1,  749,  750,  751,
  752,  753,  754,  257,   -1,  259,   -1,   -1,   -1,   -1,
   -1,  265,   -1,   -1,  268,  269,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  279,  280,   -1,  282,   -1,
  284,   -1,   -1,  287,   -1,   -1,  290,   -1,   -1,   -1,
   -1,   -1,   -1,  297,   -1,  299,  300,   -1,   -1,   -1,
  304,   -1,  306,   -1,   -1,  309,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  319,  320,   -1,   -1,   -1,
  324,   -1,  326,   -1,  328,   -1,   -1,   -1,   -1,   -1,
  334,   -1,  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  348,   -1,   -1,  351,   -1,  353,
   -1,   -1,   -1,  357,  358,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  370,   -1,   -1,   -1,
  374,  375,  376,  377,  378,   -1,  380,   -1,   -1,   -1,
   -1,   -1,  386,   -1,  388,  389,   -1,   -1,  392,  393,
   -1,   -1,   -1,   -1,   -1,  399,   -1,   -1,  402,  403,
  404,  405,   -1,  407,   -1,  409,   -1,  411,  412,  413,
  414,  415,   -1,   -1,   -1,   -1,  420,   -1,  422,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  433,
   -1,   -1,   -1,   -1,  438,  439,   -1,   -1,   -1,   -1,
   -1,  445,   -1,  447,  448,  449,   -1,  451,  452,  453,
  454,   -1,  456,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  475,  476,   -1,  478,  479,  480,   -1,   -1,   -1,
   -1,  485,  486,   -1,  488,  489,   -1,  491,   -1,   -1,
  494,  495,   -1,   -1,   -1,  499,  500,   -1,   -1,   -1,
   -1,  505,  506,  507,   -1,  509,  510,  511,   -1,  513,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  523,
  524,  525,  526,  527,  528,  529,  530,  531,  532,  533,
  534,  535,  536,  537,  538,  539,  540,  541,  542,  543,
   -1,  545,  546,  547,  548,  549,  550,   -1,  552,  553,
  554,  555,   -1,  557,   -1,  559,  560,  561,  562,  563,
  564,  565,  566,  567,  568,  569,  570,  571,  572,  573,
  574,  575,  576,   -1,  578,  579,  580,  581,   -1,  583,
  584,  585,  586,  587,   -1,  589,  590,  591,  592,  593,
  594,  595,  596,  597,  598,  599,  600,  601,   -1,  603,
  604,  605,  606,  607,  608,   -1,  610,  611,  612,  613,
  614,  615,  616,  617,   -1,  619,  620,   -1,   -1,   -1,
  624,  625,  626,  627,  628,  629,   -1,   -1,  632,  633,
  634,  635,  636,  637,  638,  639,  640,   -1,   -1,   -1,
   -1,   -1,  646,   -1,  648,  649,  650,  651,  652,  653,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  672,  673,
   -1,  675,  676,  677,  678,  679,  680,  681,  682,  683,
   -1,  685,  686,  687,  688,  689,  690,  691,  692,  693,
  694,  695,  696,   -1,  698,  699,  700,   -1,  702,   -1,
   -1,   -1,  706,  707,  708,  709,  710,  711,  712,  713,
  714,  715,  716,  717,  718,   -1,  720,  721,   -1,   -1,
   -1,   -1,   -1,  727,  728,  729,  730,  731,  732,  733,
  734,  735,  736,  737,  738,  739,  740,   -1,   -1,  743,
  744,   -1,   -1,   -1,   -1,  749,  750,  751,  752,  753,
  754,  257,   -1,  259,   -1,   -1,   -1,   -1,   -1,  265,
   -1,   -1,  268,  269,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  279,  280,   -1,  282,   -1,  284,   -1,
   -1,  287,   -1,   -1,  290,   -1,   -1,   -1,   -1,   -1,
   -1,  297,   -1,  299,  300,   -1,   -1,   -1,  304,   -1,
  306,   -1,   -1,  309,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  319,  320,   -1,   -1,   -1,  324,   -1,
  326,   -1,  328,   -1,   -1,   -1,   -1,   -1,  334,  335,
  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  348,   -1,   -1,  351,   -1,  353,   -1,  355,
   -1,  357,  358,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  370,   -1,   -1,   -1,  374,  375,
  376,  377,  378,   -1,  380,   -1,   -1,   -1,   -1,   -1,
  386,   -1,  388,  389,   -1,   -1,  392,  393,   -1,   -1,
   -1,   -1,   -1,  399,   -1,   -1,  402,  403,  404,  405,
   -1,  407,   -1,  409,   -1,  411,  412,  413,  414,  415,
   -1,   -1,   -1,   -1,  420,   -1,  422,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  433,   -1,   -1,
   -1,   -1,  438,  439,   -1,   -1,   -1,   -1,   -1,  445,
   -1,   -1,   -1,  449,   -1,  451,  452,  453,  454,   -1,
  456,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  464,  465,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  475,
  476,   -1,  478,  479,  480,   -1,   -1,   -1,   -1,  485,
  486,   -1,  488,  489,   -1,  491,   -1,   -1,  494,  495,
   -1,   -1,   -1,  499,  500,   -1,   -1,   -1,   -1,  505,
  506,  507,   -1,  509,  510,  511,   -1,  513,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  523,  524,  525,
  526,  527,  528,  529,  530,  531,  532,  533,  534,  535,
  536,  537,  538,  539,  540,  541,  542,  543,   -1,  545,
  546,  547,  548,  549,  550,   -1,  552,  553,  554,  555,
   -1,  557,   -1,  559,  560,  561,  562,  563,  564,  565,
  566,  567,  568,  569,  570,  571,  572,  573,  574,  575,
  576,   -1,  578,  579,  580,  581,   -1,  583,  584,  585,
  586,  587,   -1,  589,  590,  591,  592,  593,  594,  595,
  596,  597,  598,  599,  600,  601,   -1,  603,  604,  605,
  606,  607,  608,   -1,  610,  611,  612,  613,  614,  615,
  616,  617,   -1,  619,  620,   -1,   -1,   -1,  624,  625,
  626,  627,  628,  629,   -1,   -1,  632,  633,  634,  635,
  636,  637,  638,  639,  640,   -1,   -1,   -1,   -1,   -1,
  646,   -1,  648,  649,  650,  651,  652,  653,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  672,  673,   -1,  675,
  676,  677,  678,  679,  680,  681,  682,  683,   -1,  685,
  686,  687,  688,  689,  690,  691,  692,  693,  694,  695,
  696,   -1,  698,  699,  700,   -1,  702,   -1,   -1,   -1,
  706,  707,  708,  709,  710,  711,  712,  713,  714,  715,
  716,  717,  718,   -1,  720,  721,   -1,   -1,   -1,   -1,
   -1,  727,  728,  729,  730,  731,  732,  733,  734,  735,
  736,  737,  738,  739,  740,   -1,   -1,  743,  744,   -1,
   -1,   -1,   -1,  749,  750,  751,  752,  753,  754,  257,
   -1,  259,   -1,   -1,   -1,   -1,   -1,  265,   -1,   -1,
  268,  269,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  279,  280,   -1,  282,   -1,  284,   -1,   -1,  287,
   -1,   -1,  290,   -1,   -1,   -1,   -1,   -1,   -1,  297,
   -1,  299,  300,   -1,   -1,   -1,  304,   -1,  306,   -1,
   -1,  309,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  319,  320,   -1,  322,   -1,  324,   -1,  326,   -1,
  328,   -1,   -1,   -1,   -1,   -1,  334,   -1,  336,   -1,
   -1,   -1,  340,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  348,   -1,   -1,  351,   -1,  353,   -1,   -1,   -1,  357,
  358,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  370,   -1,   -1,   -1,  374,  375,  376,  377,
  378,   -1,  380,   -1,   -1,   -1,   -1,   -1,  386,   -1,
  388,  389,   -1,   -1,  392,  393,   -1,   -1,   -1,   -1,
   -1,  399,   -1,   -1,  402,  403,  404,  405,   -1,  407,
   -1,  409,   -1,  411,  412,  413,  414,  415,   -1,   -1,
   -1,   -1,  420,   -1,  422,   -1,   -1,   -1,  426,   -1,
   -1,   -1,   -1,   -1,   -1,  433,   -1,   -1,   -1,   -1,
  438,  439,   -1,   -1,   -1,   -1,   -1,  445,   -1,   -1,
   -1,  449,   -1,  451,  452,  453,  454,   -1,  456,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  464,  465,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  475,  476,   -1,
  478,  479,  480,   -1,   -1,   -1,   -1,  485,  486,   -1,
  488,  489,   -1,  491,   -1,   -1,  494,  495,   -1,   -1,
   -1,  499,  500,   -1,   -1,   -1,   -1,  505,  506,  507,
   -1,  509,  510,  511,   -1,  513,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  523,  524,  525,  526,  527,
  528,  529,  530,  531,  532,  533,  534,  535,  536,  537,
  538,  539,  540,  541,  542,  543,   -1,  545,  546,  547,
  548,  549,  550,   -1,  552,  553,  554,  555,   -1,  557,
   -1,  559,  560,  561,  562,  563,  564,  565,  566,  567,
  568,  569,  570,  571,  572,  573,  574,  575,  576,   -1,
  578,  579,  580,  581,   -1,  583,  584,  585,  586,  587,
   -1,  589,  590,  591,  592,  593,  594,  595,  596,  597,
  598,  599,  600,  601,   -1,  603,  604,  605,  606,  607,
  608,   -1,  610,  611,  612,  613,  614,  615,  616,  617,
   -1,  619,  620,   -1,   -1,   -1,  624,  625,  626,  627,
  628,  629,   -1,   -1,  632,  633,  634,  635,  636,  637,
  638,  639,  640,   -1,   -1,   -1,   -1,   -1,  646,   -1,
  648,  649,  650,  651,  652,  653,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  672,  673,   -1,  675,  676,  677,
  678,  679,  680,  681,  682,  683,   -1,  685,  686,  687,
  688,  689,  690,  691,  692,  693,  694,  695,  696,   -1,
  698,  699,  700,   -1,  702,   -1,   -1,   -1,  706,  707,
  708,  709,  710,  711,  712,  713,  714,  715,  716,  717,
  718,   -1,  720,  721,   -1,   -1,   -1,   -1,   -1,  727,
  728,  729,  730,  731,  732,  733,  734,  735,  736,  737,
  738,  739,  740,   -1,   -1,  743,  744,   -1,   -1,   -1,
   -1,  749,  750,  751,  752,  753,  754,  257,   -1,  259,
   -1,   -1,   -1,   -1,   -1,  265,   -1,   -1,  268,  269,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  279,
  280,   -1,  282,   -1,  284,   -1,   -1,  287,   -1,   -1,
  290,   -1,   -1,   -1,   -1,   -1,   -1,  297,   -1,  299,
  300,   -1,   -1,   -1,  304,   -1,  306,   -1,   -1,  309,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  319,
  320,   -1,   -1,   -1,  324,   -1,  326,   -1,  328,   -1,
   -1,   -1,   -1,   -1,  334,   -1,  336,   -1,   -1,   -1,
  340,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  348,   -1,
   -1,  351,   -1,  353,   -1,   -1,   -1,  357,  358,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  370,  371,   -1,   -1,  374,  375,  376,  377,  378,   -1,
  380,   -1,   -1,   -1,   -1,   -1,  386,   -1,  388,  389,
   -1,   -1,  392,  393,   -1,   -1,   -1,   -1,   -1,  399,
   -1,   -1,  402,  403,  404,  405,   -1,  407,   -1,  409,
   -1,  411,  412,  413,  414,  415,   -1,   -1,   -1,   -1,
  420,   -1,  422,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  433,   -1,   -1,   -1,   -1,  438,  439,
   -1,   -1,   -1,   -1,   -1,  445,   -1,   -1,   -1,  449,
   -1,  451,  452,  453,  454,   -1,  456,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  464,  465,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  475,  476,   -1,  478,  479,
  480,   -1,   -1,   -1,   -1,  485,  486,   -1,  488,  489,
   -1,  491,   -1,   -1,  494,  495,   -1,   -1,   -1,  499,
  500,   -1,   -1,   -1,   -1,  505,  506,  507,   -1,  509,
  510,  511,   -1,  513,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  523,  524,  525,  526,  527,  528,  529,
  530,  531,  532,  533,  534,  535,  536,  537,  538,  539,
  540,  541,  542,  543,   -1,  545,  546,  547,  548,  549,
  550,   -1,  552,  553,  554,  555,   -1,  557,   -1,  559,
  560,  561,  562,  563,  564,  565,  566,  567,  568,  569,
  570,  571,  572,  573,  574,  575,  576,   -1,  578,  579,
  580,  581,   -1,  583,  584,  585,  586,  587,   -1,  589,
  590,  591,  592,  593,  594,  595,  596,  597,  598,  599,
  600,  601,   -1,  603,  604,  605,  606,  607,  608,   -1,
  610,  611,  612,  613,  614,  615,  616,  617,   -1,  619,
  620,   -1,   -1,   -1,  624,  625,  626,  627,  628,  629,
   -1,   -1,  632,  633,  634,  635,  636,  637,  638,  639,
  640,   -1,   -1,   -1,   -1,   -1,  646,   -1,  648,  649,
  650,  651,  652,  653,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  672,  673,   -1,  675,  676,  677,  678,  679,
  680,  681,  682,  683,   -1,  685,  686,  687,  688,  689,
  690,  691,  692,  693,  694,  695,  696,   -1,  698,  699,
  700,   -1,  702,   -1,   -1,   -1,  706,  707,  708,  709,
  710,  711,  712,  713,  714,  715,  716,  717,  718,   -1,
  720,  721,   -1,   -1,   -1,   -1,   -1,  727,  728,  729,
  730,  731,  732,  733,  734,  735,  736,  737,  738,  739,
  740,   -1,   -1,  743,  744,   -1,   -1,   -1,   -1,  749,
  750,  751,  752,  753,  754,  257,   -1,  259,   -1,   -1,
   -1,   -1,   -1,  265,   -1,   -1,  268,  269,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  279,  280,   -1,
  282,   -1,  284,   -1,   -1,  287,   -1,   -1,  290,   -1,
   -1,   -1,   -1,   -1,   -1,  297,   -1,  299,  300,   -1,
   -1,   -1,  304,   -1,  306,   -1,   -1,  309,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  319,  320,   -1,
   -1,   -1,  324,   -1,  326,   -1,  328,   -1,   -1,   -1,
   -1,   -1,  334,   -1,  336,   -1,   -1,   -1,  340,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  348,   -1,   -1,  351,
   -1,  353,   -1,   -1,   -1,  357,  358,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  370,   -1,
   -1,   -1,  374,  375,  376,  377,  378,   -1,  380,   -1,
   -1,   -1,   -1,   -1,  386,   -1,  388,  389,   -1,   -1,
  392,  393,   -1,   -1,   -1,   -1,   -1,  399,   -1,   -1,
  402,  403,  404,  405,   -1,  407,   -1,  409,   -1,  411,
  412,  413,  414,  415,   -1,   -1,   -1,   -1,  420,   -1,
  422,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  433,   -1,   -1,   -1,   -1,  438,  439,   -1,   -1,
   -1,   -1,   -1,  445,   -1,   -1,   -1,  449,   -1,  451,
  452,  453,  454,   -1,  456,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  475,  476,   -1,  478,  479,  480,   -1,
   -1,   -1,   -1,  485,  486,   -1,  488,  489,   -1,  491,
   -1,   -1,  494,  495,   -1,   -1,   -1,  499,  500,   -1,
   -1,   -1,   -1,  505,  506,  507,   -1,  509,  510,  511,
   -1,  513,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  523,  524,  525,  526,  527,  528,  529,  530,  531,
  532,  533,  534,  535,  536,  537,  538,  539,  540,  541,
  542,  543,   -1,  545,  546,  547,  548,  549,  550,   -1,
  552,  553,  554,  555,   -1,  557,   -1,  559,  560,  561,
  562,  563,  564,  565,  566,  567,  568,  569,  570,  571,
  572,  573,  574,  575,  576,  577,  578,  579,  580,  581,
   -1,  583,  584,  585,  586,  587,   -1,  589,  590,  591,
  592,  593,  594,  595,  596,  597,  598,  599,  600,  601,
   -1,  603,  604,  605,  606,  607,  608,   -1,  610,  611,
  612,  613,  614,  615,  616,  617,   -1,  619,  620,   -1,
   -1,   -1,  624,  625,  626,  627,  628,  629,   -1,   -1,
  632,  633,  634,  635,  636,  637,  638,  639,  640,   -1,
   -1,   -1,   -1,   -1,  646,   -1,  648,  649,  650,  651,
  652,  653,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  672,  673,   -1,  675,  676,  677,  678,  679,  680,  681,
  682,  683,   -1,  685,  686,  687,  688,  689,  690,  691,
  692,  693,  694,  695,  696,   -1,  698,  699,  700,   -1,
  702,   -1,   -1,   -1,  706,  707,  708,  709,  710,  711,
  712,  713,  714,  715,  716,  717,  718,   -1,  720,  721,
   -1,   -1,   -1,   -1,   -1,  727,  728,  729,  730,  731,
  732,  733,  734,  735,  736,  737,  738,  739,  740,   -1,
   -1,  743,  744,   -1,   -1,   -1,   -1,  749,  750,  751,
  752,  753,  754,  257,   -1,  259,   -1,   -1,   -1,   -1,
   -1,  265,   -1,   -1,  268,  269,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  279,  280,   -1,  282,   -1,
  284,   -1,   -1,  287,   -1,   -1,  290,   -1,   -1,   -1,
   -1,   -1,   -1,  297,   -1,  299,  300,   -1,   -1,   -1,
  304,   -1,  306,   -1,   -1,  309,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  319,  320,   -1,   -1,   -1,
  324,   -1,  326,   -1,  328,   -1,   -1,   -1,   -1,   -1,
  334,   -1,  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  348,   -1,   -1,  351,   -1,  353,
   -1,   -1,   -1,  357,  358,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  370,   -1,   -1,   -1,
  374,  375,  376,  377,  378,   -1,  380,   -1,   -1,   -1,
   -1,   -1,  386,   -1,  388,  389,   -1,   -1,  392,  393,
   -1,   -1,   -1,   -1,   -1,  399,   -1,   -1,  402,  403,
  404,  405,   -1,  407,   -1,  409,   -1,  411,  412,  413,
  414,  415,   -1,   -1,   -1,   -1,  420,   -1,  422,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  433,
   -1,   -1,   -1,   -1,  438,  439,   -1,   -1,   -1,   -1,
   -1,  445,   -1,   -1,   -1,  449,   -1,  451,  452,  453,
  454,   -1,  456,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  475,  476,   -1,  478,  479,  480,   -1,   -1,   -1,
   -1,  485,  486,   -1,  488,  489,   -1,  491,   -1,   -1,
  494,  495,   -1,   -1,   -1,  499,  500,   -1,   -1,   -1,
   -1,  505,  506,  507,   -1,  509,  510,  511,   -1,  513,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  523,
  524,  525,  526,  527,  528,  529,  530,  531,  532,  533,
  534,  535,  536,  537,  538,  539,  540,  541,  542,  543,
   -1,  545,  546,  547,  548,  549,  550,   -1,  552,  553,
  554,  555,   -1,  557,   -1,  559,  560,  561,  562,  563,
  564,  565,  566,  567,  568,  569,  570,  571,  572,  573,
  574,  575,  576,   -1,  578,  579,  580,  581,   -1,  583,
  584,  585,  586,  587,   -1,  589,  590,  591,  592,  593,
  594,  595,  596,  597,  598,  599,  600,  601,   -1,  603,
  604,  605,  606,  607,  608,   -1,  610,  611,  612,  613,
  614,  615,  616,  617,   -1,  619,  620,   -1,   -1,   -1,
  624,  625,  626,  627,  628,  629,   -1,   -1,  632,  633,
  634,  635,  636,  637,  638,  639,  640,   -1,   -1,   -1,
   -1,   -1,  646,   -1,  648,  649,  650,  651,  652,  653,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  672,  673,
  674,  675,  676,  677,  678,  679,  680,  681,  682,  683,
   -1,  685,  686,  687,  688,  689,  690,  691,  692,  693,
  694,  695,  696,   -1,  698,  699,  700,   -1,  702,   -1,
   -1,   -1,  706,  707,  708,  709,  710,  711,  712,  713,
  714,  715,  716,  717,  718,   -1,  720,  721,   -1,   -1,
   -1,   -1,   -1,  727,  728,  729,  730,  731,  732,  733,
  734,  735,  736,  737,  738,  739,  740,   -1,   -1,  743,
  744,   -1,   -1,   -1,   -1,  749,  750,  751,  752,  753,
  754,  257,   -1,  259,   -1,   -1,   -1,   -1,   -1,  265,
   -1,   -1,  268,  269,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  279,  280,   -1,  282,   -1,  284,   -1,
   -1,  287,   -1,   -1,  290,   -1,   -1,   -1,   -1,  295,
   -1,  297,   -1,  299,  300,   -1,   -1,   -1,  304,   -1,
  306,   -1,   -1,  309,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  319,  320,   -1,   -1,   -1,  324,   -1,
  326,   -1,  328,   -1,   -1,   -1,   -1,   -1,  334,   -1,
  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  348,   -1,   -1,  351,   -1,  353,   -1,   -1,
   -1,  357,  358,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  370,   -1,   -1,   -1,  374,  375,
  376,  377,  378,   -1,  380,   -1,   -1,   -1,   -1,   -1,
  386,   -1,  388,  389,   -1,   -1,  392,  393,   -1,   -1,
   -1,   -1,   -1,  399,   -1,   -1,  402,  403,  404,  405,
   -1,  407,   -1,  409,   -1,  411,  412,  413,  414,  415,
   -1,   -1,   -1,   -1,  420,   -1,  422,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  433,   -1,   -1,
   -1,   -1,  438,  439,   -1,   -1,   -1,   -1,   -1,  445,
   -1,   -1,   -1,  449,   -1,  451,  452,  453,  454,   -1,
  456,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  464,  465,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  475,
  476,   -1,  478,  479,  480,   -1,   -1,   -1,   -1,  485,
  486,   -1,  488,  489,   -1,  491,   -1,   -1,  494,  495,
   -1,   -1,   -1,  499,  500,   -1,   -1,   -1,   -1,  505,
  506,  507,   -1,  509,  510,  511,   -1,  513,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  523,  524,  525,
  526,  527,  528,  529,  530,  531,  532,  533,  534,  535,
  536,  537,  538,  539,  540,  541,  542,  543,   -1,  545,
  546,  547,  548,  549,  550,   -1,  552,  553,  554,  555,
   -1,  557,   -1,  559,  560,  561,  562,  563,  564,  565,
  566,  567,  568,  569,  570,  571,  572,  573,  574,  575,
  576,   -1,  578,  579,  580,  581,   -1,  583,  584,  585,
  586,  587,   -1,  589,  590,  591,  592,  593,  594,  595,
  596,  597,  598,  599,  600,  601,   -1,  603,  604,  605,
  606,  607,  608,   -1,  610,  611,  612,  613,  614,  615,
  616,  617,   -1,  619,  620,   -1,   -1,   -1,  624,  625,
  626,  627,  628,  629,   -1,   -1,  632,  633,  634,  635,
  636,  637,  638,  639,  640,   -1,   -1,   -1,   -1,   -1,
  646,   -1,  648,  649,  650,  651,  652,  653,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  672,  673,   -1,  675,
  676,  677,  678,  679,  680,  681,  682,  683,   -1,  685,
  686,  687,  688,  689,  690,  691,  692,  693,  694,  695,
  696,   -1,  698,  699,  700,   -1,  702,   -1,   -1,   -1,
  706,  707,  708,  709,  710,  711,  712,  713,  714,  715,
  716,  717,  718,   -1,  720,  721,   -1,   -1,   -1,   -1,
   -1,  727,  728,  729,  730,  731,  732,  733,  734,  735,
  736,  737,  738,  739,  740,   -1,   -1,  743,  744,   -1,
   -1,   -1,   -1,  749,  750,  751,  752,  753,  754,  257,
   -1,  259,   -1,   -1,   -1,   -1,   -1,  265,   -1,   -1,
  268,  269,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  279,  280,   -1,  282,  283,  284,   -1,   -1,  287,
   -1,   -1,  290,   -1,   -1,   -1,   -1,   -1,   -1,  297,
   -1,  299,  300,   -1,   -1,   -1,  304,   -1,  306,   -1,
   -1,  309,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  319,  320,   -1,   -1,   -1,  324,   -1,  326,   -1,
  328,   -1,   -1,   -1,   -1,   -1,  334,   -1,  336,   -1,
   -1,   -1,  340,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  348,   -1,   -1,  351,   -1,  353,   -1,   -1,   -1,  357,
  358,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  370,   -1,   -1,   -1,  374,  375,  376,  377,
  378,   -1,  380,   -1,   -1,   -1,   -1,   -1,  386,   -1,
  388,  389,   -1,   -1,  392,  393,   -1,   -1,   -1,   -1,
   -1,  399,   -1,   -1,  402,  403,  404,  405,   -1,  407,
   -1,  409,   -1,  411,  412,  413,  414,  415,   -1,   -1,
   -1,   -1,  420,   -1,  422,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  433,   -1,   -1,   -1,   -1,
  438,  439,   -1,   -1,   -1,   -1,   -1,  445,   -1,   -1,
   -1,  449,   -1,  451,  452,  453,  454,   -1,  456,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  464,  465,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  475,  476,   -1,
  478,  479,  480,   -1,   -1,   -1,   -1,  485,  486,   -1,
  488,  489,   -1,  491,   -1,   -1,  494,  495,   -1,   -1,
   -1,  499,  500,   -1,   -1,   -1,   -1,  505,  506,  507,
   -1,  509,  510,  511,   -1,  513,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  523,  524,  525,  526,  527,
  528,  529,  530,  531,  532,  533,  534,  535,  536,  537,
  538,  539,  540,  541,  542,  543,   -1,  545,  546,  547,
  548,  549,  550,   -1,  552,  553,  554,  555,   -1,  557,
   -1,  559,  560,  561,  562,  563,  564,  565,  566,  567,
  568,  569,  570,  571,  572,  573,  574,  575,  576,   -1,
  578,  579,  580,  581,   -1,  583,  584,  585,  586,  587,
   -1,  589,  590,  591,  592,  593,  594,  595,  596,  597,
  598,  599,  600,  601,   -1,  603,  604,  605,  606,  607,
  608,   -1,  610,  611,  612,  613,  614,  615,  616,  617,
   -1,  619,  620,   -1,   -1,   -1,  624,  625,  626,  627,
  628,  629,   -1,   -1,  632,  633,  634,  635,  636,  637,
  638,  639,  640,   -1,   -1,   -1,   -1,   -1,  646,   -1,
  648,  649,  650,  651,  652,  653,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  672,  673,   -1,  675,  676,  677,
  678,  679,  680,  681,  682,  683,   -1,  685,  686,  687,
  688,  689,  690,  691,  692,  693,  694,  695,  696,   -1,
  698,  699,  700,   -1,  702,   -1,   -1,   -1,  706,  707,
  708,  709,  710,  711,  712,  713,  714,  715,  716,  717,
  718,   -1,  720,  721,   -1,   -1,   -1,   -1,   -1,  727,
  728,  729,  730,  731,  732,  733,  734,  735,  736,  737,
  738,  739,  740,   -1,   -1,  743,  744,   -1,   -1,   -1,
   -1,  749,  750,  751,  752,  753,  754,  257,   -1,  259,
   -1,   -1,   -1,   -1,   -1,  265,   -1,   -1,  268,  269,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  279,
  280,   -1,  282,  283,  284,   -1,   -1,  287,   -1,   -1,
  290,   -1,   -1,   -1,   -1,   -1,   -1,  297,   -1,  299,
  300,   -1,   -1,   -1,  304,   -1,  306,   -1,   -1,  309,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  319,
  320,   -1,   -1,   -1,  324,   -1,  326,   -1,  328,   -1,
   -1,   -1,   -1,   -1,  334,   -1,  336,   -1,   -1,   -1,
  340,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  348,   -1,
   -1,  351,   -1,  353,   -1,   -1,   -1,  357,  358,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  370,   -1,   -1,   -1,  374,  375,  376,  377,  378,   -1,
  380,   -1,   -1,   -1,   -1,   -1,  386,   -1,  388,  389,
   -1,   -1,  392,  393,   -1,   -1,   -1,   -1,   -1,  399,
   -1,   -1,  402,  403,  404,  405,   -1,  407,   -1,  409,
   -1,  411,  412,  413,  414,  415,   -1,   -1,   -1,   -1,
  420,   -1,  422,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  433,   -1,   -1,   -1,   -1,  438,  439,
   -1,   -1,   -1,   -1,   -1,  445,   -1,   -1,   -1,  449,
   -1,  451,  452,  453,  454,   -1,  456,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  464,  465,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  475,  476,   -1,  478,  479,
  480,   -1,   -1,   -1,   -1,  485,  486,   -1,  488,  489,
   -1,  491,   -1,   -1,  494,  495,   -1,   -1,   -1,  499,
  500,   -1,   -1,   -1,   -1,  505,  506,  507,   -1,  509,
  510,  511,   -1,  513,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  523,  524,  525,  526,  527,  528,  529,
  530,  531,  532,  533,  534,  535,  536,  537,  538,  539,
  540,  541,  542,  543,   -1,  545,  546,  547,  548,  549,
  550,   -1,  552,  553,  554,  555,   -1,  557,   -1,  559,
  560,  561,  562,  563,  564,  565,  566,  567,  568,  569,
  570,  571,  572,  573,  574,  575,  576,   -1,  578,  579,
  580,  581,   -1,  583,  584,  585,  586,  587,   -1,  589,
  590,  591,  592,  593,  594,  595,  596,  597,  598,  599,
  600,  601,   -1,  603,  604,  605,  606,  607,  608,   -1,
  610,  611,  612,  613,  614,  615,  616,  617,   -1,  619,
  620,   -1,   -1,   -1,  624,  625,  626,  627,  628,  629,
   -1,   -1,  632,  633,  634,  635,  636,  637,  638,  639,
  640,   -1,   -1,   -1,   -1,   -1,  646,   -1,  648,  649,
  650,  651,  652,  653,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  672,  673,   -1,  675,  676,  677,  678,  679,
  680,  681,  682,  683,   -1,  685,  686,  687,  688,  689,
  690,  691,  692,  693,  694,  695,  696,   -1,  698,  699,
  700,   -1,  702,   -1,   -1,   -1,  706,  707,  708,  709,
  710,  711,  712,  713,  714,  715,  716,  717,  718,   -1,
  720,  721,   -1,   -1,   -1,   -1,   -1,  727,  728,  729,
  730,  731,  732,  733,  734,  735,  736,  737,  738,  739,
  740,   -1,   -1,  743,  744,   -1,   -1,   -1,   -1,  749,
  750,  751,  752,  753,  754,  257,   -1,  259,   -1,   -1,
   -1,   -1,   -1,  265,   -1,   -1,  268,  269,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  279,  280,   -1,
  282,   -1,  284,   -1,   -1,  287,   -1,   -1,  290,   -1,
   -1,   -1,   -1,   -1,   -1,  297,   -1,  299,  300,   -1,
   -1,   -1,  304,   -1,  306,   -1,   -1,  309,  310,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  319,  320,   -1,
   -1,   -1,  324,   -1,  326,   -1,  328,   -1,   -1,   -1,
   -1,   -1,  334,   -1,  336,   -1,   -1,   -1,  340,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  348,   -1,   -1,  351,
   -1,  353,   -1,   -1,   -1,  357,  358,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  370,   -1,
   -1,   -1,  374,  375,  376,  377,  378,   -1,  380,   -1,
   -1,   -1,   -1,   -1,  386,   -1,  388,  389,   -1,   -1,
  392,  393,   -1,   -1,   -1,   -1,   -1,  399,   -1,   -1,
  402,  403,  404,  405,   -1,  407,   -1,  409,   -1,  411,
  412,  413,  414,  415,   -1,   -1,   -1,   -1,  420,   -1,
  422,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  433,   -1,   -1,   -1,   -1,  438,  439,   -1,   -1,
   -1,   -1,   -1,  445,   -1,   -1,   -1,  449,   -1,  451,
  452,  453,  454,   -1,  456,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  475,  476,   -1,  478,  479,  480,   -1,
   -1,   -1,   -1,  485,  486,   -1,  488,  489,   -1,  491,
   -1,   -1,  494,  495,   -1,   -1,   -1,  499,  500,   -1,
   -1,   -1,   -1,  505,  506,  507,   -1,  509,  510,  511,
   -1,  513,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  523,  524,  525,  526,  527,  528,  529,  530,  531,
  532,  533,  534,  535,  536,  537,  538,  539,  540,  541,
  542,  543,   -1,  545,  546,  547,  548,  549,  550,   -1,
  552,  553,  554,  555,   -1,  557,   -1,  559,  560,  561,
  562,  563,  564,  565,  566,  567,  568,  569,  570,  571,
  572,  573,  574,  575,  576,   -1,  578,  579,  580,  581,
   -1,  583,  584,  585,  586,  587,   -1,  589,  590,  591,
  592,  593,  594,  595,  596,  597,  598,  599,  600,  601,
   -1,  603,  604,  605,  606,  607,  608,   -1,  610,  611,
  612,  613,  614,  615,  616,  617,   -1,  619,  620,   -1,
   -1,   -1,  624,  625,  626,  627,  628,  629,   -1,   -1,
  632,  633,  634,  635,  636,  637,  638,  639,  640,   -1,
   -1,   -1,   -1,   -1,  646,   -1,  648,  649,  650,  651,
  652,  653,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  672,  673,   -1,  675,  676,  677,  678,  679,  680,  681,
  682,  683,   -1,  685,  686,  687,  688,  689,  690,  691,
  692,  693,  694,  695,  696,   -1,  698,  699,  700,   -1,
  702,   -1,   -1,   -1,  706,  707,  708,  709,  710,  711,
  712,  713,  714,  715,  716,  717,  718,   -1,  720,  721,
   -1,   -1,   -1,   -1,   -1,  727,  728,  729,  730,  731,
  732,  733,  734,  735,  736,  737,  738,  739,  740,   -1,
   -1,  743,  744,   -1,   -1,   -1,   -1,  749,  750,  751,
  752,  753,  754,  257,   -1,  259,   -1,   -1,   -1,   -1,
   -1,  265,   -1,   -1,  268,  269,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  279,  280,   -1,  282,   -1,
  284,   -1,   -1,  287,   -1,   -1,  290,   -1,   -1,   -1,
   -1,   -1,   -1,  297,   -1,  299,  300,   -1,   -1,   -1,
  304,   -1,  306,   -1,   -1,  309,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  319,  320,   -1,   -1,   -1,
  324,   -1,  326,   -1,  328,   -1,   -1,   -1,   -1,   -1,
  334,   -1,  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  348,   -1,   -1,  351,   -1,  353,
   -1,   -1,   -1,  357,  358,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  370,   -1,   -1,   -1,
  374,  375,  376,  377,  378,   -1,  380,   -1,   -1,   -1,
   -1,   -1,  386,   -1,  388,  389,   -1,   -1,  392,  393,
   -1,   -1,   -1,   -1,   -1,  399,   -1,   -1,  402,  403,
  404,  405,   -1,  407,   -1,  409,   -1,  411,  412,  413,
  414,  415,   -1,   -1,   -1,   -1,  420,   -1,  422,   -1,
   -1,   -1,  426,   -1,   -1,   -1,   -1,   -1,   -1,  433,
   -1,   -1,   -1,   -1,  438,  439,   -1,   -1,   -1,   -1,
   -1,  445,   -1,   -1,   -1,  449,   -1,  451,  452,  453,
  454,   -1,  456,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  475,  476,   -1,  478,  479,  480,   -1,   -1,   -1,
   -1,  485,  486,   -1,  488,  489,   -1,  491,   -1,   -1,
  494,  495,   -1,   -1,   -1,  499,  500,   -1,   -1,   -1,
   -1,  505,  506,  507,   -1,  509,  510,  511,   -1,  513,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  523,
  524,  525,  526,  527,  528,  529,  530,  531,  532,  533,
  534,  535,  536,  537,  538,  539,  540,  541,  542,  543,
   -1,  545,  546,  547,  548,  549,  550,   -1,  552,  553,
  554,  555,   -1,  557,   -1,  559,  560,  561,  562,  563,
  564,  565,  566,  567,  568,  569,  570,  571,  572,  573,
  574,  575,  576,   -1,  578,  579,  580,  581,   -1,  583,
  584,  585,  586,  587,   -1,  589,  590,  591,  592,  593,
  594,  595,  596,  597,  598,  599,  600,  601,   -1,  603,
  604,  605,  606,  607,  608,   -1,  610,  611,  612,  613,
  614,  615,  616,  617,   -1,  619,  620,   -1,   -1,   -1,
  624,  625,  626,  627,  628,  629,   -1,   -1,  632,  633,
  634,  635,  636,  637,  638,  639,  640,   -1,   -1,   -1,
   -1,   -1,  646,   -1,  648,  649,  650,  651,  652,  653,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  672,  673,
   -1,  675,  676,  677,  678,  679,  680,  681,  682,  683,
   -1,  685,  686,  687,  688,  689,  690,  691,  692,  693,
  694,  695,  696,   -1,  698,  699,  700,   -1,  702,   -1,
   -1,   -1,  706,  707,  708,  709,  710,  711,  712,  713,
  714,  715,  716,  717,  718,   -1,  720,  721,   -1,   -1,
   -1,   -1,   -1,  727,  728,  729,  730,  731,  732,  733,
  734,  735,  736,  737,  738,  739,  740,   -1,   -1,  743,
  744,   -1,   -1,   -1,   -1,  749,  750,  751,  752,  753,
  754,  257,   -1,  259,   -1,   -1,   -1,   -1,   -1,  265,
   -1,   -1,  268,  269,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  279,  280,   -1,  282,   -1,  284,   -1,
   -1,  287,   -1,   -1,  290,   -1,   -1,   -1,   -1,   -1,
   -1,  297,   -1,  299,  300,   -1,   -1,   -1,  304,   -1,
  306,   -1,   -1,  309,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  319,  320,   -1,   -1,   -1,  324,   -1,
  326,   -1,  328,   -1,   -1,   -1,   -1,   -1,  334,   -1,
  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  348,   -1,   -1,  351,   -1,  353,   -1,   -1,
   -1,  357,  358,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  370,   -1,   -1,   -1,  374,  375,
  376,  377,  378,   -1,  380,   -1,   -1,   -1,   -1,   -1,
  386,   -1,  388,  389,   -1,   -1,  392,  393,   -1,   -1,
   -1,   -1,   -1,  399,   -1,   -1,  402,  403,  404,  405,
   -1,  407,   -1,  409,   -1,  411,  412,  413,  414,  415,
   -1,   -1,   -1,   -1,  420,   -1,  422,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  433,   -1,   -1,
   -1,   -1,  438,  439,   -1,   -1,   -1,   -1,   -1,  445,
   -1,   -1,   -1,  449,   -1,  451,  452,  453,  454,  455,
  456,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  464,  465,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  475,
  476,   -1,  478,  479,  480,   -1,   -1,   -1,   -1,  485,
  486,   -1,  488,  489,   -1,  491,   -1,   -1,  494,  495,
   -1,   -1,   -1,  499,  500,   -1,   -1,   -1,   -1,  505,
  506,  507,   -1,  509,  510,  511,   -1,  513,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  523,  524,  525,
  526,  527,  528,  529,  530,  531,  532,  533,  534,  535,
  536,  537,  538,  539,  540,  541,  542,  543,   -1,  545,
  546,  547,  548,  549,  550,   -1,  552,  553,  554,  555,
   -1,  557,   -1,  559,  560,  561,  562,  563,  564,  565,
  566,  567,  568,  569,  570,  571,  572,  573,  574,  575,
  576,   -1,  578,  579,  580,  581,   -1,  583,  584,  585,
  586,  587,   -1,  589,  590,  591,  592,  593,  594,  595,
  596,  597,  598,  599,  600,  601,   -1,  603,  604,  605,
  606,  607,  608,   -1,  610,  611,  612,  613,  614,  615,
  616,  617,   -1,  619,  620,   -1,   -1,   -1,  624,  625,
  626,  627,  628,  629,   -1,   -1,  632,  633,  634,  635,
  636,  637,  638,  639,  640,   -1,   -1,   -1,   -1,   -1,
  646,   -1,  648,  649,  650,  651,  652,  653,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  672,  673,   -1,  675,
  676,  677,  678,  679,  680,  681,  682,  683,   -1,  685,
  686,  687,  688,  689,  690,  691,  692,  693,  694,  695,
  696,   -1,  698,  699,  700,   -1,  702,   -1,   -1,   -1,
  706,  707,  708,  709,  710,  711,  712,  713,  714,  715,
  716,  717,  718,   -1,  720,  721,   -1,   -1,   -1,   -1,
   -1,  727,  728,  729,  730,  731,  732,  733,  734,  735,
  736,  737,  738,  739,  740,   -1,   -1,  743,  744,   -1,
   -1,   -1,   -1,  749,  750,  751,  752,  753,  754,  257,
   -1,  259,   -1,   -1,   -1,   -1,   -1,  265,   -1,   -1,
  268,  269,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  279,  280,   -1,  282,   -1,  284,   -1,   -1,  287,
   -1,   -1,  290,   -1,   -1,   -1,   -1,   -1,   -1,  297,
   -1,  299,  300,   -1,   -1,   -1,  304,   -1,  306,   -1,
   -1,  309,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  319,  320,   -1,   -1,   -1,  324,   -1,  326,   -1,
  328,   -1,   -1,   -1,   -1,   -1,  334,   -1,  336,   -1,
   -1,   -1,  340,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  348,   -1,   -1,  351,   -1,  353,   -1,   -1,   -1,  357,
  358,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  370,   -1,   -1,   -1,  374,  375,  376,  377,
  378,   -1,  380,   -1,   -1,   -1,   -1,   -1,  386,   -1,
  388,  389,   -1,   -1,  392,  393,   -1,   -1,   -1,   -1,
   -1,  399,   -1,   -1,  402,  403,  404,  405,   -1,  407,
   -1,  409,   -1,  411,  412,  413,  414,  415,   -1,   -1,
   -1,   -1,  420,   -1,  422,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  433,   -1,   -1,   -1,   -1,
  438,  439,   -1,   -1,   -1,   -1,   -1,  445,   -1,   -1,
   -1,  449,   -1,  451,  452,  453,  454,  455,  456,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  464,  465,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  475,  476,   -1,
  478,  479,  480,   -1,   -1,   -1,   -1,  485,  486,   -1,
  488,  489,   -1,  491,   -1,   -1,  494,  495,   -1,   -1,
   -1,  499,  500,   -1,   -1,   -1,   -1,  505,  506,  507,
   -1,  509,  510,  511,   -1,  513,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  523,  524,  525,  526,  527,
  528,  529,  530,  531,  532,  533,  534,  535,  536,  537,
  538,  539,  540,  541,  542,  543,   -1,  545,  546,  547,
  548,  549,  550,   -1,  552,  553,  554,  555,   -1,  557,
   -1,  559,  560,  561,  562,  563,  564,  565,  566,  567,
  568,  569,  570,  571,  572,  573,  574,  575,  576,   -1,
  578,  579,  580,  581,   -1,  583,  584,  585,  586,  587,
   -1,  589,  590,  591,  592,  593,  594,  595,  596,  597,
  598,  599,  600,  601,   -1,  603,  604,  605,  606,  607,
  608,   -1,  610,  611,  612,  613,  614,  615,  616,  617,
   -1,  619,  620,   -1,   -1,   -1,  624,  625,  626,  627,
  628,  629,   -1,   -1,  632,  633,  634,  635,  636,  637,
  638,  639,  640,   -1,   -1,   -1,   -1,   -1,  646,   -1,
  648,  649,  650,  651,  652,  653,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  672,  673,   -1,  675,  676,  677,
  678,  679,  680,  681,  682,  683,   -1,  685,  686,  687,
  688,  689,  690,  691,  692,  693,  694,  695,  696,   -1,
  698,  699,  700,   -1,  702,   -1,   -1,   -1,  706,  707,
  708,  709,  710,  711,  712,  713,  714,  715,  716,  717,
  718,   -1,  720,  721,   -1,   -1,   -1,   -1,   -1,  727,
  728,  729,  730,  731,  732,  733,  734,  735,  736,  737,
  738,  739,  740,   -1,   -1,  743,  744,   -1,   -1,   -1,
   -1,  749,  750,  751,  752,  753,  754,  257,   -1,  259,
   -1,   -1,   -1,   -1,   -1,  265,   -1,   -1,  268,  269,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  279,
  280,   -1,  282,   -1,  284,   -1,   -1,  287,   -1,   -1,
  290,   -1,   -1,   -1,   -1,   -1,   -1,  297,   -1,  299,
  300,   -1,   -1,   -1,  304,   -1,  306,   -1,   -1,  309,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  319,
  320,   -1,   -1,   -1,  324,   -1,  326,   -1,  328,   -1,
   -1,   -1,   -1,   -1,  334,   -1,  336,   -1,   -1,   -1,
  340,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  348,   -1,
   -1,  351,   -1,  353,   -1,   -1,   -1,  357,  358,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  370,   -1,   -1,   -1,  374,  375,  376,  377,  378,   -1,
  380,   -1,   -1,   -1,   -1,   -1,  386,   -1,  388,  389,
   -1,   -1,  392,  393,   -1,   -1,   -1,   -1,   -1,  399,
   -1,   -1,  402,  403,  404,  405,   -1,  407,   -1,  409,
   -1,  411,  412,  413,  414,  415,   -1,   -1,   -1,   -1,
  420,   -1,  422,   -1,   -1,   -1,  426,   -1,   -1,   -1,
   -1,   -1,   -1,  433,   -1,   -1,   -1,   -1,  438,  439,
   -1,   -1,   -1,   -1,   -1,  445,   -1,   -1,   -1,  449,
   -1,  451,  452,  453,  454,   -1,  456,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  464,  465,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  475,  476,   -1,  478,  479,
  480,   -1,   -1,   -1,   -1,  485,  486,   -1,  488,  489,
   -1,  491,   -1,   -1,  494,  495,   -1,   -1,   -1,  499,
  500,   -1,   -1,   -1,   -1,  505,  506,  507,   -1,  509,
  510,  511,   -1,  513,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  523,  524,  525,  526,  527,  528,  529,
  530,  531,  532,  533,  534,  535,  536,  537,  538,  539,
  540,  541,  542,  543,   -1,  545,  546,  547,  548,  549,
  550,   -1,  552,  553,  554,  555,   -1,  557,   -1,  559,
  560,  561,  562,  563,  564,  565,  566,  567,  568,  569,
  570,  571,  572,  573,  574,  575,  576,   -1,  578,  579,
  580,  581,   -1,  583,  584,  585,  586,  587,   -1,  589,
  590,  591,  592,  593,  594,  595,  596,  597,  598,  599,
  600,  601,   -1,  603,  604,  605,  606,  607,  608,   -1,
  610,  611,  612,  613,  614,  615,  616,  617,   -1,  619,
  620,   -1,   -1,   -1,  624,  625,  626,  627,  628,  629,
   -1,   -1,  632,  633,  634,  635,  636,  637,  638,  639,
  640,   -1,   -1,   -1,   -1,   -1,  646,   -1,  648,  649,
  650,  651,  652,  653,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  672,  673,   -1,  675,  676,  677,  678,  679,
  680,  681,  682,  683,   -1,  685,  686,  687,  688,  689,
  690,  691,  692,  693,  694,  695,  696,   -1,  698,  699,
  700,   -1,  702,   -1,   -1,   -1,  706,  707,  708,  709,
  710,  711,  712,  713,  714,  715,  716,  717,  718,   -1,
  720,  721,   -1,   -1,   -1,   -1,   -1,  727,  728,  729,
  730,  731,  732,  733,  734,  735,  736,  737,  738,  739,
  740,   -1,   -1,  743,  744,   -1,   -1,   -1,   -1,  749,
  750,  751,  752,  753,  754,  257,   -1,  259,   -1,   -1,
   -1,   -1,   -1,  265,   -1,   -1,  268,  269,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  279,  280,   -1,
  282,   -1,  284,   -1,   -1,  287,   -1,   -1,  290,   -1,
   -1,   -1,   -1,   -1,   -1,  297,   -1,  299,  300,  301,
   -1,   -1,  304,   -1,  306,   -1,   -1,  309,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  319,  320,   -1,
   -1,   -1,  324,   -1,  326,   -1,  328,   -1,   -1,   -1,
   -1,   -1,  334,   -1,  336,   -1,   -1,   -1,  340,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  348,   -1,   -1,  351,
   -1,  353,   -1,   -1,   -1,  357,  358,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  370,   -1,
   -1,   -1,  374,  375,  376,  377,  378,   -1,  380,   -1,
   -1,   -1,   -1,   -1,  386,   -1,  388,  389,   -1,   -1,
  392,  393,   -1,   -1,   -1,   -1,   -1,  399,   -1,   -1,
  402,  403,  404,  405,   -1,  407,   -1,  409,   -1,  411,
  412,  413,  414,  415,   -1,   -1,   -1,   -1,  420,   -1,
  422,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  433,   -1,   -1,   -1,   -1,  438,  439,   -1,   -1,
   -1,   -1,   -1,  445,   -1,   -1,   -1,  449,   -1,  451,
  452,  453,  454,   -1,  456,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  475,  476,   -1,  478,  479,  480,   -1,
   -1,   -1,   -1,  485,  486,   -1,  488,  489,   -1,  491,
   -1,   -1,  494,  495,   -1,   -1,   -1,  499,  500,   -1,
   -1,   -1,   -1,  505,  506,  507,   -1,  509,  510,  511,
   -1,  513,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  523,  524,  525,  526,  527,  528,  529,  530,  531,
  532,  533,  534,  535,  536,  537,  538,  539,  540,  541,
  542,  543,   -1,  545,  546,  547,  548,  549,  550,   -1,
  552,  553,  554,  555,   -1,  557,   -1,  559,  560,  561,
  562,  563,  564,  565,  566,  567,  568,  569,  570,  571,
  572,  573,  574,  575,  576,   -1,  578,  579,  580,  581,
   -1,  583,  584,  585,  586,  587,   -1,  589,  590,  591,
  592,  593,  594,  595,  596,  597,  598,  599,  600,  601,
   -1,  603,  604,  605,  606,  607,  608,   -1,  610,  611,
  612,  613,  614,  615,  616,  617,   -1,  619,  620,   -1,
   -1,   -1,  624,  625,  626,  627,  628,  629,   -1,   -1,
  632,  633,  634,  635,  636,  637,  638,  639,  640,   -1,
   -1,   -1,   -1,   -1,  646,   -1,  648,  649,  650,  651,
  652,  653,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  672,  673,   -1,  675,  676,  677,  678,  679,  680,  681,
  682,  683,   -1,  685,  686,  687,  688,  689,  690,  691,
  692,  693,  694,  695,  696,   -1,  698,  699,  700,   -1,
  702,   -1,   -1,   -1,  706,  707,  708,  709,  710,  711,
  712,  713,  714,  715,  716,  717,  718,   -1,  720,  721,
   -1,   -1,   -1,   -1,   -1,  727,  728,  729,  730,  731,
  732,  733,  734,  735,  736,  737,  738,  739,  740,   -1,
   -1,  743,  744,   -1,   -1,   -1,   -1,  749,  750,  751,
  752,  753,  754,  257,   -1,  259,   -1,   -1,   -1,   -1,
   -1,  265,   -1,   -1,  268,  269,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  279,  280,   -1,  282,   -1,
  284,   -1,   -1,  287,   -1,   -1,  290,   -1,   -1,   -1,
   -1,   -1,   -1,  297,   -1,  299,  300,   -1,   -1,   -1,
  304,   -1,  306,   -1,   -1,  309,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  319,  320,   -1,   -1,   -1,
  324,   -1,  326,   -1,  328,   -1,   -1,   -1,   -1,   -1,
  334,   -1,  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  348,   -1,   -1,  351,   -1,  353,
   -1,   -1,   -1,  357,  358,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  370,   -1,   -1,   -1,
  374,  375,  376,  377,  378,   -1,  380,   -1,   -1,   -1,
   -1,   -1,  386,   -1,  388,  389,   -1,   -1,  392,  393,
   -1,   -1,   -1,   -1,   -1,  399,   -1,   -1,  402,  403,
  404,  405,   -1,  407,   -1,  409,   -1,  411,  412,  413,
  414,  415,   -1,   -1,   -1,   -1,  420,   -1,  422,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  433,
   -1,   -1,   -1,   -1,  438,  439,   -1,   -1,   -1,   -1,
   -1,  445,   -1,   -1,   -1,  449,   -1,  451,  452,  453,
  454,   -1,  456,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  475,  476,   -1,  478,  479,  480,   -1,   -1,   -1,
   -1,  485,  486,   -1,  488,  489,   -1,  491,   -1,   -1,
  494,  495,   -1,   -1,   -1,  499,  500,   -1,   -1,   -1,
   -1,  505,  506,  507,   -1,  509,  510,  511,   -1,  513,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  523,
  524,  525,  526,  527,  528,  529,  530,  531,  532,  533,
  534,  535,  536,  537,  538,  539,  540,  541,  542,  543,
   -1,  545,  546,  547,  548,  549,  550,   -1,  552,  553,
  554,  555,   -1,  557,   -1,  559,  560,  561,  562,  563,
  564,  565,  566,  567,  568,  569,  570,  571,  572,  573,
  574,  575,  576,   -1,  578,  579,  580,  581,   -1,  583,
  584,  585,  586,  587,   -1,  589,  590,  591,  592,  593,
  594,  595,  596,  597,  598,  599,  600,  601,   -1,  603,
  604,  605,  606,  607,  608,   -1,  610,  611,  612,  613,
  614,  615,  616,  617,   -1,  619,  620,   -1,   -1,   -1,
  624,  625,  626,  627,  628,  629,   -1,   -1,  632,  633,
  634,  635,  636,  637,  638,  639,  640,   -1,   -1,   -1,
   -1,   -1,  646,   -1,  648,  649,  650,  651,  652,  653,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  672,  673,
   -1,  675,  676,  677,  678,  679,  680,  681,  682,  683,
   -1,  685,  686,  687,  688,  689,  690,  691,  692,  693,
  694,  695,  696,   -1,  698,  699,  700,   -1,  702,   -1,
   -1,   -1,  706,  707,  708,  709,  710,  711,  712,  713,
  714,  715,  716,  717,  718,   -1,  720,  721,   -1,   -1,
   -1,   -1,   -1,  727,  728,  729,  730,  731,  732,  733,
  734,  735,  736,  737,  738,  739,  740,   -1,   -1,  743,
  744,   -1,   -1,   -1,   -1,  749,  750,  751,  752,  753,
  754,  257,   -1,  259,   -1,   -1,   -1,   -1,   -1,  265,
   -1,   -1,  268,  269,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  279,  280,   -1,  282,   -1,  284,   -1,
   -1,  287,   -1,   -1,  290,   -1,   -1,   -1,   -1,   -1,
   -1,  297,   -1,  299,  300,   -1,   -1,   -1,  304,   -1,
  306,   -1,   -1,  309,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  319,  320,   -1,   -1,   -1,  324,   -1,
  326,   -1,  328,   -1,   -1,   -1,   -1,   -1,  334,   -1,
  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  348,   -1,   -1,  351,   -1,  353,   -1,   -1,
   -1,  357,  358,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  370,   -1,   -1,   -1,  374,  375,
  376,  377,  378,   -1,  380,   -1,   -1,   -1,   -1,   -1,
  386,   -1,  388,  389,   -1,   -1,  392,  393,   -1,   -1,
   -1,   -1,   -1,  399,   -1,   -1,  402,  403,  404,  405,
   -1,  407,   -1,  409,   -1,  411,  412,  413,  414,  415,
   -1,   -1,   -1,   -1,  420,   -1,  422,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  433,   -1,   -1,
   -1,   -1,  438,  439,   -1,   -1,   -1,   -1,   -1,  445,
   -1,   -1,   -1,  449,   -1,  451,  452,  453,  454,   -1,
  456,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  464,  465,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  475,
  476,   -1,  478,  479,  480,   -1,   -1,   -1,   -1,  485,
  486,   -1,  488,  489,   -1,  491,   -1,   -1,  494,  495,
   -1,   -1,   -1,  499,  500,   -1,   -1,   -1,   -1,  505,
  506,  507,   -1,  509,  510,  511,   -1,  513,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  523,  524,  525,
  526,  527,  528,  529,  530,  531,  532,  533,  534,  535,
  536,  537,  538,  539,  540,  541,  542,  543,   -1,  545,
  546,  547,  548,  549,  550,   -1,  552,  553,  554,  555,
   -1,  557,   -1,  559,  560,  561,  562,  563,  564,  565,
  566,  567,  568,  569,  570,  571,  572,  573,  574,  575,
  576,   -1,  578,  579,  580,  581,   -1,  583,  584,  585,
  586,  587,   -1,  589,  590,  591,  592,  593,  594,  595,
  596,  597,  598,  599,  600,  601,   -1,  603,  604,  605,
  606,  607,  608,   -1,  610,  611,  612,  613,  614,  615,
  616,  617,   -1,  619,  620,   -1,   -1,   -1,  624,  625,
  626,  627,  628,  629,   -1,   -1,  632,  633,  634,  635,
  636,  637,  638,  639,  640,   -1,   -1,   -1,   -1,   -1,
  646,   -1,  648,  649,  650,  651,  652,  653,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  672,  673,   -1,  675,
  676,  677,  678,  679,  680,  681,  682,  683,   -1,  685,
  686,  687,  688,  689,  690,  691,  692,  693,  694,  695,
  696,   -1,  698,  699,  700,   -1,  702,   -1,   -1,   -1,
  706,  707,  708,  709,  710,  711,  712,  713,  714,  715,
  716,  717,  718,   -1,  720,  721,   -1,   -1,   -1,   -1,
   -1,  727,  728,  729,  730,  731,  732,  733,  734,  735,
  736,  737,  738,  739,  740,   -1,   -1,  743,  744,   -1,
   -1,   -1,   -1,  749,  750,  751,  752,  753,  754,  257,
   -1,  259,   -1,   -1,   -1,   -1,   -1,  265,   -1,   -1,
  268,  269,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  279,  280,   -1,  282,   -1,  284,   -1,   -1,  287,
   -1,   -1,  290,   -1,   -1,   -1,   -1,   -1,   -1,  297,
   -1,  299,  300,   -1,   -1,   -1,  304,   -1,  306,   -1,
   -1,  309,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  319,  320,   -1,   -1,   -1,  324,   -1,  326,   -1,
  328,   -1,   -1,   -1,   -1,   -1,  334,   -1,  336,   -1,
   -1,   -1,  340,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  348,   -1,   -1,  351,   -1,  353,   -1,   -1,   -1,  357,
  358,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  370,   -1,   -1,   -1,  374,  375,  376,  377,
  378,   -1,  380,   -1,   -1,   -1,   -1,   -1,  386,   -1,
  388,  389,   -1,   -1,  392,  393,   -1,   -1,   -1,   -1,
   -1,  399,   -1,   -1,  402,  403,  404,  405,   -1,  407,
   -1,  409,   -1,  411,  412,  413,  414,  415,   -1,   -1,
   -1,   -1,  420,   -1,  422,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  433,   -1,   -1,   -1,   -1,
  438,  439,   -1,   -1,   -1,   -1,   -1,  445,   -1,   -1,
   -1,  449,   -1,  451,  452,  453,  454,   -1,  456,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  464,  465,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  475,  476,   -1,
  478,  479,  480,   -1,   -1,   -1,   -1,  485,  486,   -1,
  488,  489,   -1,  491,   -1,   -1,  494,  495,   -1,   -1,
   -1,  499,  500,   -1,   -1,   -1,   -1,  505,  506,  507,
   -1,  509,  510,  511,   -1,  513,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  523,  524,  525,  526,  527,
  528,  529,  530,  531,  532,  533,  534,  535,  536,  537,
  538,  539,  540,  541,  542,  543,   -1,  545,  546,  547,
  548,  549,  550,   -1,  552,  553,  554,  555,   -1,  557,
   -1,  559,  560,  561,  562,  563,  564,  565,  566,  567,
  568,  569,  570,  571,  572,  573,  574,  575,  576,   -1,
  578,  579,  580,  581,   -1,  583,  584,  585,  586,  587,
   -1,  589,  590,  591,  592,  593,  594,  595,  596,  597,
  598,  599,  600,  601,   -1,  603,  604,  605,  606,  607,
  608,   -1,  610,  611,  612,  613,  614,  615,  616,  617,
   -1,  619,  620,   -1,   -1,   -1,  624,  625,  626,  627,
  628,  629,   -1,   -1,  632,  633,  634,  635,  636,  637,
  638,  639,  640,   -1,   -1,   -1,   -1,   -1,  646,   -1,
  648,  649,  650,  651,  652,  653,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  672,  673,   -1,  675,  676,  677,
  678,  679,  680,  681,  682,  683,   -1,  685,  686,  687,
  688,  689,  690,  691,  692,  693,  694,  695,  696,   -1,
  698,  699,  700,   -1,  702,   -1,   -1,   -1,  706,  707,
  708,  709,  710,  711,  712,  713,  714,  715,  716,  717,
  718,   -1,  720,  721,   -1,   -1,   -1,   -1,   -1,  727,
  728,  729,  730,  731,  732,  733,  734,  735,  736,  737,
  738,  739,  740,   -1,   -1,  743,  744,   -1,   -1,   -1,
   -1,  749,  750,  751,  752,  753,  754,  257,   -1,  259,
   -1,   -1,   -1,   -1,   -1,  265,   -1,   -1,  268,  269,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  279,
  280,   -1,  282,   -1,  284,   -1,   -1,  287,   -1,   -1,
  290,   -1,   -1,   -1,   -1,   -1,   -1,  297,   -1,  299,
  300,   -1,   -1,   -1,  304,   -1,  306,   -1,   -1,  309,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  319,
  320,   -1,   -1,   -1,  324,   -1,  326,   -1,  328,   -1,
   -1,   -1,   -1,   -1,  334,   -1,  336,   -1,   -1,   -1,
  340,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  348,   -1,
   -1,  351,   -1,  353,   -1,   -1,   -1,  357,  358,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  370,   -1,   -1,   -1,  374,  375,  376,  377,  378,   -1,
  380,   -1,   -1,   -1,   -1,   -1,  386,   -1,  388,  389,
   -1,   -1,  392,  393,   -1,   -1,   -1,   -1,   -1,  399,
   -1,   -1,  402,  403,  404,  405,   -1,  407,   -1,  409,
   -1,  411,  412,  413,  414,  415,   -1,   -1,   -1,   -1,
  420,   -1,  422,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  433,   -1,   -1,   -1,   -1,  438,  439,
   -1,   -1,   -1,   -1,   -1,  445,   -1,   -1,   -1,  449,
   -1,  451,  452,  453,  454,   -1,  456,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  464,  465,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  475,  476,   -1,  478,  479,
  480,   -1,   -1,   -1,   -1,  485,  486,   -1,  488,  489,
   -1,  491,   -1,   -1,  494,  495,   -1,   -1,   -1,  499,
  500,   -1,   -1,   -1,   -1,  505,  506,  507,   -1,  509,
  510,  511,   -1,  513,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  523,  524,  525,  526,  527,  528,  529,
  530,  531,  532,  533,  534,  535,  536,  537,  538,  539,
  540,  541,  542,  543,   -1,  545,  546,  547,  548,  549,
  550,   -1,  552,  553,  554,  555,   -1,  557,   -1,  559,
  560,  561,  562,  563,  564,  565,  566,  567,  568,  569,
  570,  571,  572,  573,  574,  575,  576,   -1,  578,  579,
  580,  581,   -1,  583,  584,  585,  586,  587,   -1,  589,
  590,  591,  592,  593,  594,  595,  596,  597,  598,  599,
  600,  601,   -1,  603,  604,  605,  606,  607,  608,   -1,
  610,  611,  612,  613,  614,  615,  616,  617,   -1,  619,
  620,   -1,   -1,   -1,  624,  625,  626,  627,  628,  629,
   -1,   -1,  632,  633,  634,  635,  636,  637,  638,  639,
  640,   -1,   -1,   -1,   -1,   -1,  646,   -1,  648,  649,
  650,  651,  652,  653,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  672,  673,   -1,  675,  676,  677,  678,  679,
  680,  681,  682,  683,   -1,  685,  686,  687,  688,  689,
  690,  691,  692,  693,  694,  695,  696,   -1,  698,  699,
  700,   -1,  702,   -1,   -1,   -1,  706,  707,  708,  709,
  710,  711,  712,  713,  714,  715,  716,  717,  718,   -1,
  720,  721,   -1,   -1,   -1,   -1,   -1,  727,  728,  729,
  730,  731,  732,  733,  734,  735,  736,  737,  738,  739,
  740,   -1,   -1,  743,  744,   -1,   -1,   -1,   -1,  749,
  750,  751,  752,  753,  754,  257,   -1,  259,   -1,   -1,
   -1,   -1,   -1,  265,   -1,   -1,  268,  269,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  279,  280,   -1,
  282,   -1,  284,   -1,   -1,  287,   -1,   -1,  290,   -1,
   -1,   -1,   -1,   -1,   -1,  297,   -1,  299,  300,   -1,
   -1,   -1,  304,   -1,  306,   -1,   -1,  309,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  319,  320,   -1,
   -1,   -1,  324,   -1,  326,   -1,  328,   -1,   -1,   -1,
   -1,   -1,  334,   -1,  336,   -1,   -1,   -1,  340,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  348,   -1,   -1,  351,
   -1,  353,   -1,   -1,   -1,  357,  358,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  370,   -1,
   -1,   -1,  374,  375,  376,  377,  378,   -1,  380,   -1,
   -1,   -1,   -1,   -1,  386,   -1,  388,  389,   -1,   -1,
  392,  393,   -1,   -1,   -1,   -1,   -1,  399,   -1,   -1,
  402,  403,  404,  405,   -1,  407,   -1,  409,   -1,  411,
  412,  413,  414,  415,   -1,   -1,   -1,   -1,  420,   -1,
  422,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  433,   -1,   -1,   -1,   -1,  438,  439,   -1,   -1,
   -1,   -1,   -1,  445,   -1,   -1,   -1,  449,   -1,  451,
  452,  453,  454,   -1,  456,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  475,  476,   -1,  478,  479,  480,   -1,
   -1,   -1,   -1,  485,  486,   -1,  488,  489,   -1,  491,
   -1,   -1,  494,  495,   -1,   -1,   -1,  499,  500,   -1,
   -1,   -1,   -1,  505,  506,  507,   -1,  509,  510,  511,
   -1,  513,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  523,  524,  525,  526,  527,  528,  529,  530,  531,
  532,  533,  534,  535,  536,  537,  538,  539,  540,  541,
  542,  543,   -1,  545,  546,  547,  548,  549,  550,   -1,
  552,  553,  554,  555,   -1,  557,   -1,  559,  560,  561,
  562,  563,  564,  565,  566,  567,  568,  569,  570,  571,
  572,  573,  574,  575,  576,   -1,  578,  579,  580,  581,
   -1,  583,  584,  585,  586,  587,   -1,  589,  590,  591,
  592,  593,  594,  595,  596,  597,  598,  599,  600,  601,
   -1,  603,  604,  605,  606,  607,  608,   -1,  610,  611,
  612,  613,  614,  615,  616,  617,   -1,  619,  620,   -1,
   -1,   -1,  624,  625,  626,  627,  628,  629,   -1,   -1,
  632,  633,  634,  635,  636,  637,  638,  639,  640,   -1,
   -1,   -1,   -1,   -1,  646,   -1,  648,  649,  650,  651,
  652,  653,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  672,  673,   -1,  675,  676,  677,  678,  679,  680,  681,
  682,  683,   -1,  685,  686,  687,  688,  689,  690,  691,
  692,  693,  694,  695,  696,   -1,  698,  699,  700,   -1,
  702,   -1,   -1,   -1,  706,  707,  708,  709,  710,  711,
  712,  713,  714,  715,  716,  717,  718,   -1,  720,  721,
   -1,   -1,   -1,   -1,   -1,  727,  728,  729,  730,  731,
  732,  733,  734,  735,  736,  737,  738,  739,  740,   -1,
   -1,  743,  744,   -1,   -1,   -1,   -1,  749,  750,  751,
  752,  753,  754,  257,   -1,  259,   -1,   -1,   -1,   -1,
   -1,  265,   -1,   -1,  268,  269,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  279,  280,   -1,  282,   -1,
  284,   -1,   -1,  287,   -1,   -1,  290,   -1,   -1,   -1,
   -1,   -1,   -1,  297,   -1,  299,  300,   -1,   -1,   -1,
  304,   -1,  306,   -1,   -1,  309,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  319,  320,   -1,   -1,   -1,
  324,   -1,  326,   -1,  328,   -1,   -1,   -1,   -1,   -1,
  334,   -1,  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  348,   -1,   -1,  351,   -1,  353,
   -1,   -1,   -1,  357,  358,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  370,   -1,   -1,   -1,
  374,  375,  376,  377,  378,   -1,  380,   -1,   -1,   -1,
   -1,   -1,  386,   -1,  388,  389,   -1,   -1,  392,  393,
   -1,   -1,   -1,   -1,   -1,  399,   -1,   -1,  402,  403,
  404,  405,   -1,  407,   -1,  409,   -1,  411,  412,  413,
  414,  415,   -1,   -1,   -1,   -1,  420,   -1,  422,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  433,
   -1,   -1,   -1,   -1,  438,  439,   -1,   -1,   -1,   -1,
   -1,  445,   -1,   -1,   -1,  449,   -1,  451,  452,  453,
  454,   -1,  456,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  475,  476,   -1,  478,  479,  480,   -1,   -1,   -1,
   -1,  485,  486,   -1,  488,  489,   -1,  491,   -1,   -1,
  494,  495,   -1,   -1,   -1,  499,  500,   -1,   -1,   -1,
   -1,  505,  506,  507,   -1,  509,  510,  511,   -1,  513,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  523,
  524,  525,  526,  527,  528,  529,  530,  531,  532,  533,
  534,  535,  536,  537,  538,  539,  540,  541,  542,  543,
   -1,  545,  546,  547,  548,  549,  550,   -1,  552,  553,
  554,  555,   -1,  557,   -1,  559,  560,  561,  562,  563,
  564,  565,  566,  567,  568,  569,  570,  571,  572,  573,
  574,  575,  576,   -1,  578,  579,  580,  581,   -1,  583,
  584,  585,  586,  587,   -1,  589,  590,  591,  592,  593,
  594,  595,  596,  597,  598,  599,  600,  601,   -1,  603,
  604,  605,  606,  607,  608,   -1,  610,  611,  612,  613,
  614,  615,  616,  617,   -1,  619,  620,   -1,   -1,   -1,
  624,  625,  626,  627,  628,  629,   -1,   -1,  632,  633,
  634,  635,  636,  637,  638,  639,  640,   -1,   -1,   -1,
   -1,   -1,  646,   -1,  648,  649,  650,  651,  652,  653,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  672,  673,
   -1,  675,  676,  677,  678,  679,  680,  681,  682,  683,
   -1,  685,  686,  687,  688,  689,  690,  691,  692,  693,
  694,  695,  696,   -1,  698,  699,  700,   -1,  702,   -1,
   -1,   -1,  706,  707,  708,  709,  710,  711,  712,  713,
  714,  715,  716,  717,  718,   -1,  720,  721,   -1,   -1,
   -1,   -1,   -1,  727,  728,  729,  730,  731,  732,  733,
  734,  735,  736,  737,  738,  739,  740,   -1,   -1,  743,
  744,   -1,   -1,   -1,   -1,  749,  750,  751,  752,  753,
  754,  257,   -1,  259,   -1,   -1,   -1,   -1,   -1,  265,
   -1,   -1,  268,  269,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  279,  280,   -1,  282,   -1,  284,   -1,
   -1,  287,   -1,   -1,  290,   -1,   -1,   -1,   -1,   -1,
   -1,  297,   -1,  299,  300,   -1,   -1,   -1,  304,   -1,
  306,   -1,   -1,  309,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  319,  320,   -1,   -1,   -1,  324,   -1,
  326,   -1,  328,   -1,   -1,   -1,   -1,   -1,  334,   -1,
  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  348,   -1,   -1,  351,   -1,  353,   -1,   -1,
   -1,  357,  358,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  370,   -1,   -1,   -1,  374,  375,
  376,  377,  378,   -1,  380,   -1,   -1,   -1,   -1,   -1,
  386,   -1,  388,  389,   -1,   -1,  392,  393,   -1,   -1,
   -1,   -1,   -1,  399,   -1,   -1,  402,  403,  404,  405,
   -1,  407,   -1,  409,   -1,  411,  412,  413,  414,  415,
   -1,   -1,   -1,   -1,  420,   -1,  422,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  433,   -1,   -1,
   -1,   -1,  438,  439,   -1,   -1,   -1,   -1,   -1,  445,
   -1,   -1,   -1,  449,   -1,  451,  452,  453,  454,   -1,
  456,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  464,  465,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  475,
  476,   -1,  478,  479,  480,   -1,   -1,   -1,   -1,  485,
  486,   -1,  488,  489,   -1,  491,   -1,   -1,  494,  495,
   -1,   -1,   -1,  499,  500,   -1,   -1,   -1,   -1,  505,
  506,  507,   -1,  509,  510,  511,   -1,  513,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  523,  524,  525,
  526,  527,  528,  529,  530,  531,  532,  533,  534,  535,
  536,  537,  538,  539,  540,  541,  542,  543,   -1,  545,
  546,  547,  548,  549,  550,   -1,  552,  553,  554,  555,
   -1,  557,   -1,  559,  560,  561,  562,  563,  564,  565,
  566,  567,  568,  569,  570,  571,  572,  573,  574,  575,
  576,   -1,  578,  579,  580,  581,   -1,  583,  584,  585,
  586,  587,   -1,  589,  590,  591,  592,  593,  594,  595,
  596,  597,  598,  599,  600,  601,   -1,  603,  604,  605,
  606,  607,  608,   -1,  610,  611,  612,  613,  614,  615,
  616,  617,   -1,  619,  620,   -1,   -1,   -1,  624,  625,
  626,  627,  628,  629,   -1,   -1,  632,  633,  634,  635,
  636,  637,  638,  639,  640,   -1,   -1,   -1,   -1,   -1,
  646,   -1,  648,  649,  650,  651,  652,  653,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  672,  673,   -1,  675,
  676,  677,  678,  679,  680,  681,  682,  683,   -1,  685,
  686,  687,  688,  689,  690,  691,  692,  693,  694,  695,
  696,   -1,  698,  699,  700,   -1,  702,   -1,   -1,   -1,
  706,  707,  708,  709,  710,  711,  712,  713,  714,  715,
  716,  717,  718,   -1,  720,  721,   -1,   -1,   -1,   -1,
   -1,  727,  728,  729,  730,  731,  732,  733,  734,  735,
  736,  737,  738,  739,  740,   -1,   -1,  743,  744,   -1,
   -1,   -1,   -1,  749,  750,  751,  752,  753,  754,  257,
   -1,  259,   -1,   -1,   -1,   -1,   -1,  265,   -1,   -1,
  268,  269,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  279,  280,   -1,  282,   -1,  284,   -1,   -1,  287,
   -1,   -1,  290,   -1,   -1,   -1,   -1,   -1,   -1,  297,
   -1,  299,  300,   -1,   -1,   -1,  304,   -1,  306,   -1,
   -1,  309,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  319,  320,   -1,   -1,   -1,  324,   -1,  326,   -1,
  328,   -1,   -1,   -1,   -1,   -1,  334,   -1,  336,   -1,
   -1,   -1,  340,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  348,   -1,   -1,  351,   -1,  353,   -1,   -1,   -1,  357,
  358,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  370,   -1,   -1,   -1,  374,  375,  376,  377,
  378,   -1,  380,   -1,   -1,   -1,   -1,   -1,  386,   -1,
  388,  389,   -1,   -1,  392,  393,   -1,   -1,   -1,   -1,
   -1,  399,   -1,   -1,  402,  403,  404,  405,   -1,  407,
   -1,  409,   -1,  411,  412,  413,  414,  415,   -1,   -1,
   -1,   -1,  420,   -1,  422,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  433,   -1,   -1,   -1,   -1,
  438,  439,   -1,   -1,   -1,   -1,   -1,  445,   -1,   -1,
   -1,  449,   -1,  451,  452,  453,  454,   -1,  456,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  464,  465,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  475,  476,   -1,
  478,  479,  480,   -1,   -1,   -1,   -1,  485,  486,   -1,
  488,  489,   -1,  491,   -1,   -1,  494,  495,   -1,   -1,
   -1,  499,  500,   -1,   -1,   -1,   -1,  505,  506,  507,
   -1,  509,  510,  511,   -1,  513,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  523,  524,  525,  526,  527,
  528,  529,  530,  531,  532,  533,  534,  535,  536,  537,
  538,  539,  540,  541,  542,  543,   -1,  545,  546,  547,
  548,  549,  550,   -1,  552,  553,  554,  555,   -1,  557,
   -1,  559,  560,  561,  562,  563,  564,  565,  566,  567,
  568,  569,  570,  571,  572,  573,  574,  575,  576,   -1,
  578,  579,  580,  581,   -1,  583,  584,  585,  586,  587,
   -1,  589,  590,  591,  592,  593,  594,  595,  596,  597,
  598,  599,  600,  601,   -1,  603,  604,  605,  606,  607,
  608,   -1,  610,  611,  612,  613,  614,  615,  616,  617,
   -1,  619,  620,   -1,   -1,   -1,  624,  625,  626,  627,
  628,  629,   -1,   -1,  632,  633,  634,  635,  636,  637,
  638,  639,  640,   -1,   -1,   -1,   -1,   -1,  646,   -1,
  648,  649,  650,  651,  652,  653,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  672,  673,   -1,  675,  676,  677,
  678,  679,  680,  681,  682,  683,   -1,  685,  686,  687,
  688,  689,  690,  691,  692,  693,  694,  695,  696,   -1,
  698,  699,  700,   -1,  702,   -1,   -1,   -1,  706,  707,
  708,  709,  710,  711,  712,  713,  714,  715,  716,  717,
  718,   -1,  720,  721,   -1,   -1,   -1,   -1,   -1,  727,
  728,  729,  730,  731,  732,  733,  734,  735,  736,  737,
  738,  739,  740,   -1,   -1,  743,  744,   -1,   -1,   -1,
   -1,  749,  750,  751,  752,  753,  754,  257,   -1,  259,
   -1,   -1,   -1,   -1,   -1,  265,   -1,   -1,  268,  269,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  279,
  280,   -1,  282,   -1,  284,   -1,   -1,  287,   -1,   -1,
  290,   -1,   -1,   -1,   -1,   -1,   -1,  297,   -1,  299,
  300,   -1,   -1,   -1,  304,   -1,  306,   -1,   -1,  309,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  319,
  320,   -1,   -1,   -1,  324,   -1,  326,   -1,  328,   -1,
   -1,   -1,   -1,   -1,  334,   -1,  336,   -1,   -1,   -1,
  340,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  348,   -1,
   -1,  351,   -1,  353,   -1,   -1,   -1,  357,  358,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  370,   -1,   -1,   -1,  374,  375,  376,  377,  378,   -1,
  380,   -1,   -1,   -1,   -1,   -1,  386,   -1,  388,  389,
   -1,   -1,  392,  393,   -1,   -1,   -1,   -1,   -1,  399,
   -1,   -1,  402,  403,  404,  405,   -1,  407,   -1,  409,
   -1,  411,  412,  413,  414,  415,   -1,   -1,   -1,   -1,
  420,   -1,  422,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  433,   -1,   -1,   -1,   -1,  438,  439,
   -1,   -1,   -1,   -1,   -1,  445,   -1,   -1,   -1,  449,
   -1,  451,  452,  453,  454,   -1,  456,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  464,  465,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  475,  476,   -1,  478,  479,
  480,   -1,   -1,   -1,   -1,  485,  486,   -1,  488,  489,
   -1,  491,   -1,   -1,  494,  495,   -1,   -1,   -1,  499,
  500,   -1,   -1,   -1,   -1,  505,  506,  507,   -1,  509,
  510,  511,   -1,  513,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  523,  524,  525,  526,  527,  528,  529,
  530,  531,  532,  533,  534,  535,  536,  537,  538,  539,
  540,  541,  542,  543,   -1,  545,  546,  547,  548,  549,
  550,   -1,  552,  553,  554,  555,   -1,  557,   -1,  559,
  560,  561,  562,  563,  564,  565,  566,  567,  568,  569,
  570,  571,  572,  573,  574,  575,  576,   -1,  578,  579,
  580,  581,   -1,  583,  584,  585,  586,  587,   -1,  589,
  590,  591,  592,  593,  594,  595,  596,  597,  598,  599,
  600,  601,   -1,  603,  604,  605,  606,  607,  608,   -1,
  610,  611,  612,  613,  614,  615,  616,  617,   -1,  619,
  620,   -1,   -1,   -1,  624,  625,  626,  627,  628,  629,
   -1,   -1,  632,  633,  634,  635,  636,  637,  638,  639,
  640,   -1,   -1,   -1,   -1,   -1,  646,   -1,  648,  649,
  650,  651,  652,  653,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  672,  673,   -1,  675,  676,  677,  678,  679,
  680,  681,  682,  683,   -1,  685,  686,  687,  688,  689,
  690,  691,  692,  693,  694,  695,  696,   -1,  698,  699,
  700,   -1,  702,   -1,   -1,   -1,  706,  707,  708,  709,
  710,  711,  712,  713,  714,  715,  716,  717,  718,   -1,
  720,  721,   -1,   -1,   -1,   -1,   -1,  727,  728,  729,
  730,  731,  732,  733,  734,  735,  736,  737,  738,  739,
  740,   -1,   -1,  743,  744,   -1,   -1,   -1,   -1,  749,
  750,  751,  752,  753,  754,  257,   -1,  259,   -1,   -1,
   -1,   -1,   -1,  265,   -1,   -1,  268,  269,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  279,  280,   -1,
  282,   -1,  284,   -1,   -1,  287,   -1,   -1,  290,   -1,
   -1,   -1,   -1,   -1,   -1,  297,   -1,  299,  300,   -1,
   -1,   -1,  304,   -1,  306,   -1,   -1,  309,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  319,  320,   -1,
   -1,   -1,  324,   -1,  326,   -1,  328,   -1,   -1,   -1,
   -1,   -1,  334,   -1,  336,   -1,   -1,   -1,  340,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  348,   -1,   -1,  351,
   -1,  353,   -1,   -1,   -1,  357,  358,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  370,   -1,
   -1,   -1,  374,  375,  376,  377,  378,   -1,  380,   -1,
   -1,   -1,   -1,   -1,  386,   -1,  388,  389,   -1,   -1,
  392,  393,   -1,   -1,   -1,   -1,   -1,  399,   -1,   -1,
  402,  403,  404,  405,   -1,  407,   -1,  409,   -1,  411,
  412,  413,  414,  415,   -1,   -1,   -1,   -1,  420,   -1,
  422,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  433,   -1,   -1,   -1,   -1,  438,  439,   -1,   -1,
   -1,   -1,   -1,  445,   -1,   -1,   -1,  449,   -1,  451,
  452,  453,  454,   -1,  456,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  475,  476,   -1,  478,  479,  480,   -1,
   -1,   -1,   -1,  485,  486,   -1,  488,  489,   -1,  491,
   -1,   -1,  494,  495,   -1,   -1,   -1,  499,  500,   -1,
   -1,   -1,   -1,  505,  506,  507,   -1,  509,  510,  511,
   -1,  513,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  523,  524,  525,  526,  527,  528,  529,  530,  531,
  532,  533,  534,  535,  536,  537,  538,  539,  540,  541,
  542,  543,   -1,  545,  546,  547,  548,  549,  550,   -1,
  552,  553,  554,  555,   -1,  557,   -1,  559,  560,  561,
  562,  563,  564,  565,  566,  567,  568,  569,  570,  571,
  572,  573,  574,  575,  576,   -1,  578,  579,  580,  581,
   -1,  583,  584,  585,  586,  587,   -1,  589,  590,  591,
  592,  593,  594,  595,  596,  597,  598,  599,  600,  601,
   -1,  603,  604,  605,  606,  607,  608,   -1,  610,  611,
  612,  613,  614,  615,  616,  617,   -1,  619,  620,   -1,
   -1,   -1,  624,  625,  626,  627,  628,  629,   -1,   -1,
  632,  633,  634,  635,  636,  637,  638,  639,  640,   -1,
   -1,   -1,   -1,   -1,  646,   -1,  648,  649,  650,  651,
  652,  653,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  672,  673,   -1,  675,  676,  677,  678,  679,  680,  681,
  682,  683,   -1,  685,  686,  687,  688,  689,  690,  691,
  692,  693,  694,  695,  696,   -1,  698,  699,  700,   -1,
  702,   -1,   -1,   -1,  706,  707,  708,  709,  710,  711,
  712,  713,  714,  715,  716,  717,  718,   -1,  720,  721,
   -1,   -1,   -1,   -1,   -1,  727,  728,  729,  730,  731,
  732,  733,  734,  735,  736,  737,  738,  739,  740,   -1,
   -1,  743,  744,   -1,   -1,   -1,   -1,  749,  750,  751,
  752,  753,  754,  257,   -1,  259,   -1,   -1,   -1,   -1,
   -1,  265,   -1,   -1,  268,  269,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  279,  280,   -1,  282,   -1,
  284,   -1,   -1,  287,   -1,   -1,  290,   -1,   -1,   -1,
   -1,   -1,   -1,  297,   -1,  299,  300,   -1,   -1,   -1,
  304,   -1,  306,   -1,   -1,  309,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  319,  320,   -1,   -1,   -1,
  324,   -1,  326,   -1,  328,   -1,   -1,   -1,   -1,   -1,
  334,   -1,  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  348,   -1,   -1,  351,   -1,  353,
   -1,   -1,   -1,  357,  358,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  370,   -1,   -1,   -1,
  374,  375,  376,  377,  378,   -1,  380,   -1,   -1,   -1,
   -1,   -1,  386,   -1,  388,  389,   -1,   -1,  392,  393,
   -1,   -1,   -1,   -1,   -1,  399,   -1,   -1,  402,  403,
  404,  405,   -1,  407,   -1,  409,   -1,  411,  412,  413,
  414,  415,   -1,   -1,   -1,   -1,  420,   -1,  422,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  433,
   -1,   -1,   -1,   -1,  438,  439,   -1,   -1,   -1,   -1,
   -1,  445,   -1,   -1,   -1,  449,   -1,  451,  452,  453,
  454,   -1,  456,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  475,  476,   -1,  478,  479,  480,   -1,   -1,   -1,
   -1,  485,  486,   -1,  488,  489,   -1,  491,   -1,   -1,
  494,  495,   -1,   -1,   -1,  499,  500,   -1,   -1,   -1,
   -1,  505,  506,  507,   -1,  509,  510,  511,   -1,  513,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  523,
  524,  525,  526,  527,  528,  529,  530,  531,  532,  533,
  534,  535,  536,  537,  538,  539,  540,  541,  542,  543,
   -1,  545,  546,  547,  548,  549,  550,   -1,  552,  553,
  554,  555,   -1,  557,   -1,  559,  560,  561,  562,  563,
  564,  565,  566,  567,  568,  569,  570,  571,  572,  573,
  574,  575,  576,   -1,  578,  579,  580,  581,   -1,  583,
  584,  585,  586,  587,   -1,  589,  590,  591,  592,  593,
  594,  595,  596,  597,  598,  599,  600,  601,   -1,  603,
  604,  605,  606,  607,  608,   -1,  610,  611,  612,  613,
  614,  615,  616,  617,   -1,  619,  620,   -1,   -1,   -1,
  624,  625,  626,  627,  628,  629,   -1,   -1,  632,  633,
  634,  635,  636,  637,  638,  639,  640,   -1,   -1,   -1,
   -1,   -1,  646,   -1,  648,  649,  650,  651,  652,  653,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  672,  673,
   -1,  675,  676,  677,  678,  679,  680,  681,  682,  683,
   -1,  685,  686,  687,  688,  689,  690,  691,  692,  693,
  694,  695,  696,   -1,  698,  699,  700,   -1,  702,   -1,
   -1,   -1,  706,  707,  708,  709,  710,  711,  712,  713,
  714,  715,  716,  717,  718,   -1,  720,  721,   -1,   -1,
   -1,   -1,   -1,  727,  728,  729,  730,  731,  732,  733,
  734,  735,  736,  737,  738,  739,  740,   -1,   -1,  743,
  744,   -1,   -1,   -1,   -1,  749,  750,  751,  752,  753,
  754,  257,   -1,  259,   -1,   -1,   -1,   -1,   -1,  265,
   -1,   -1,  268,  269,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  279,  280,   -1,  282,   -1,  284,   -1,
   -1,  287,   -1,   -1,  290,   -1,   -1,   -1,   -1,   -1,
   -1,  297,   -1,  299,  300,   -1,   -1,   -1,  304,   -1,
  306,   -1,   -1,  309,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  319,  320,   -1,   -1,   -1,  324,   -1,
  326,   -1,  328,   -1,   -1,   -1,   -1,   -1,  334,   -1,
  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  348,   -1,   -1,  351,   -1,  353,   -1,   -1,
   -1,  357,  358,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  370,   -1,   -1,   -1,  374,  375,
  376,  377,  378,   -1,  380,   -1,   -1,   -1,   -1,   -1,
  386,   -1,  388,  389,   -1,   -1,  392,  393,   -1,   -1,
   -1,   -1,   -1,  399,   -1,   -1,  402,  403,  404,  405,
   -1,  407,   -1,  409,   -1,  411,  412,  413,  414,  415,
   -1,   -1,   -1,   -1,  420,   -1,  422,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  433,   -1,   -1,
   -1,   -1,  438,  439,   -1,   -1,   -1,   -1,   -1,  445,
   -1,   -1,   -1,  449,   -1,  451,  452,  453,  454,   -1,
  456,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  464,  465,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  475,
  476,   -1,  478,  479,  480,   -1,   -1,   -1,   -1,  485,
  486,   -1,  488,  489,   -1,  491,   -1,   -1,  494,  495,
   -1,   -1,   -1,  499,  500,   -1,   -1,   -1,   -1,  505,
  506,  507,   -1,  509,  510,  511,   -1,  513,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  523,  524,  525,
  526,  527,  528,  529,  530,  531,  532,  533,  534,  535,
  536,  537,  538,  539,  540,  541,  542,  543,   -1,  545,
  546,  547,  548,  549,  550,   -1,  552,  553,  554,  555,
   -1,  557,   -1,  559,  560,  561,  562,  563,  564,  565,
  566,  567,  568,  569,  570,  571,  572,  573,  574,  575,
  576,   -1,  578,  579,  580,  581,   -1,  583,  584,  585,
  586,  587,   -1,  589,  590,  591,  592,  593,  594,  595,
  596,  597,  598,  599,  600,  601,   -1,  603,  604,  605,
  606,  607,  608,   -1,  610,  611,  612,  613,  614,  615,
  616,  617,   -1,  619,  620,   -1,   -1,   -1,  624,  625,
  626,  627,  628,  629,   -1,   -1,  632,  633,  634,  635,
  636,  637,  638,  639,  640,   -1,   -1,   -1,   -1,   -1,
  646,   -1,  648,  649,  650,  651,  652,  653,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  672,  673,   -1,  675,
  676,  677,  678,  679,  680,  681,  682,  683,   -1,  685,
  686,  687,  688,  689,  690,  691,  692,  693,  694,  695,
  696,   -1,  698,  699,  700,   -1,  702,   -1,   -1,   -1,
  706,  707,  708,  709,  710,  711,  712,  713,  714,  715,
  716,  717,  718,   -1,  720,  721,   -1,   -1,   -1,   -1,
   -1,  727,  728,  729,  730,  731,  732,  733,  734,  735,
  736,  737,   -1,  739,  740,   -1,   -1,  743,  744,   -1,
   -1,   -1,   -1,  749,  750,  751,  752,  753,  754,  257,
   -1,  259,   -1,   -1,   -1,   -1,   -1,  265,   -1,   -1,
  268,  269,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  279,  280,   -1,  282,   -1,  284,   -1,   -1,  287,
   -1,   -1,  290,   -1,   -1,   -1,   -1,   -1,   -1,  297,
   -1,  299,  300,   -1,   -1,   -1,  304,   -1,  306,   -1,
   -1,  309,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  319,  320,   -1,   -1,   -1,  324,   -1,  326,   -1,
  328,   -1,   -1,   -1,   -1,   -1,  334,   -1,  336,   -1,
   -1,   -1,  340,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  348,   -1,   -1,  351,   -1,  353,   -1,   -1,   -1,  357,
  358,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  370,   -1,   -1,   -1,  374,  375,  376,  377,
  378,   -1,  380,   -1,   -1,   -1,   -1,   -1,  386,   -1,
  388,  389,   -1,   -1,  392,  393,   -1,   -1,   -1,   -1,
   -1,  399,   -1,   -1,  402,  403,  404,  405,   -1,  407,
   -1,  409,   -1,  411,  412,  413,  414,  415,   -1,   -1,
   -1,   -1,  420,   -1,  422,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  433,   -1,   -1,   -1,   -1,
  438,  439,   -1,   -1,   -1,   -1,   -1,  445,   -1,   -1,
   -1,  449,   -1,  451,  452,  453,  454,   -1,  456,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  464,  465,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  475,  476,   -1,
  478,  479,  480,   -1,   -1,   -1,   -1,  485,  486,   -1,
  488,  489,   -1,  491,   -1,   -1,  494,  495,   -1,   -1,
   -1,  499,  500,   -1,   -1,   -1,   -1,  505,  506,  507,
   -1,  509,  510,  511,   -1,  513,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  524,  525,  526,  527,
  528,  529,  530,  531,  532,  533,  534,  535,  536,  537,
  538,  539,  540,  541,  542,  543,   -1,  545,  546,  547,
  548,  549,  550,   -1,  552,  553,  554,  555,   -1,  557,
   -1,  559,  560,  561,  562,  563,  564,  565,  566,  567,
  568,  569,  570,  571,  572,  573,  574,  575,  576,   -1,
  578,  579,  580,  581,   -1,  583,  584,  585,  586,  587,
   -1,  589,  590,  591,  592,  593,  594,  595,  596,  597,
  598,  599,  600,  601,   -1,  603,  604,  605,  606,  607,
  608,   -1,  610,  611,  612,  613,  614,  615,  616,  617,
   -1,  619,  620,   -1,   -1,   -1,  624,  625,  626,  627,
  628,  629,   -1,   -1,  632,  633,  634,  635,  636,  637,
  638,  639,  640,   -1,   -1,   -1,   -1,   -1,  646,   -1,
  648,  649,  650,  651,  652,  653,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  672,  673,   -1,  675,  676,  677,
  678,  679,  680,  681,  682,  683,   -1,  685,  686,  687,
  688,  689,  690,  691,  692,  693,  694,  695,  696,   -1,
  698,  699,  700,   -1,  702,   -1,   -1,   -1,  706,  707,
  708,  709,  710,  711,  712,  713,  714,  715,  716,  717,
  718,   -1,  720,  721,   -1,   -1,   -1,   -1,   -1,  727,
  728,  729,  730,  731,  732,  733,  734,  735,  736,  737,
  738,  739,  740,   -1,   -1,  743,  744,   -1,   -1,   -1,
   -1,  749,  750,  751,  752,  753,  754,  257,   -1,  259,
   -1,   -1,   -1,   -1,   -1,  265,   -1,   -1,  268,  269,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  279,
  280,   -1,  282,   -1,  284,   -1,   -1,  287,   -1,   -1,
  290,   -1,   -1,   -1,   -1,   -1,   -1,  297,   -1,  299,
  300,   -1,   -1,   -1,  304,   -1,  306,   -1,   -1,  309,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  319,
  320,   -1,   -1,   -1,  324,   -1,  326,   -1,  328,   -1,
   -1,   -1,   -1,   -1,  334,   -1,  336,   -1,   -1,   -1,
  340,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  348,   -1,
   -1,  351,   -1,  353,   -1,   -1,   -1,  357,  358,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  370,   -1,   -1,   -1,  374,  375,  376,  377,  378,   -1,
  380,   -1,   -1,   -1,   -1,   -1,  386,   -1,  388,  389,
   -1,   -1,  392,  393,   -1,   -1,   -1,   -1,   -1,  399,
   -1,   -1,  402,  403,  404,  405,   -1,  407,   -1,  409,
   -1,  411,  412,  413,  414,  415,   -1,   -1,   -1,   -1,
  420,   -1,  422,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  433,   -1,   -1,   -1,   -1,  438,  439,
   -1,   -1,   -1,   -1,   -1,  445,   -1,   -1,   -1,  449,
   -1,  451,  452,  453,  454,   -1,  456,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  464,  465,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  475,  476,   -1,  478,  479,
  480,   -1,   -1,   -1,   -1,  485,  486,   -1,  488,  489,
   -1,  491,   -1,   -1,  494,  495,   -1,   -1,   -1,  499,
  500,   -1,   -1,   -1,   -1,  505,  506,  507,   -1,  509,
  510,  511,   -1,  513,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  523,  524,  525,  526,  527,  528,  529,
  530,  531,  532,  533,  534,  535,  536,  537,  538,  539,
  540,  541,  542,  543,   -1,  545,  546,  547,  548,  549,
  550,   -1,  552,  553,  554,  555,   -1,  557,   -1,  559,
  560,  561,  562,  563,  564,  565,  566,  567,  568,  569,
  570,  571,  572,  573,  574,  575,  576,   -1,  578,  579,
  580,  581,   -1,  583,  584,  585,  586,  587,   -1,  589,
  590,  591,  592,  593,  594,  595,  596,   -1,  598,  599,
  600,  601,   -1,  603,  604,  605,  606,  607,  608,   -1,
  610,  611,  612,  613,  614,  615,  616,  617,   -1,  619,
  620,   -1,   -1,   -1,  624,  625,  626,  627,  628,  629,
   -1,   -1,  632,  633,  634,  635,  636,  637,  638,  639,
  640,   -1,   -1,   -1,   -1,   -1,  646,   -1,  648,  649,
  650,  651,  652,  653,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  672,  673,   -1,  675,  676,  677,  678,  679,
  680,  681,  682,  683,   -1,  685,  686,  687,  688,  689,
  690,  691,  692,  693,  694,  695,  696,   -1,  698,  699,
  700,   -1,  702,   -1,   -1,   -1,  706,  707,  708,  709,
  710,  711,  712,  713,  714,  715,  716,  717,  718,   -1,
  720,  721,   -1,   -1,   -1,   -1,   -1,  727,  728,  729,
  730,  731,  732,  733,  734,  735,  736,  737,  738,  739,
  740,   -1,   -1,  743,  744,   -1,   -1,   -1,   -1,  749,
  750,  751,  752,  753,  754,  257,   -1,  259,   -1,   -1,
   -1,   -1,   -1,  265,   -1,   -1,  268,  269,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  279,  280,   -1,
  282,   -1,   -1,   -1,   -1,  287,   -1,   -1,  290,   -1,
   -1,   -1,   -1,   -1,   -1,  297,   -1,  299,  300,   -1,
   -1,   -1,  304,   -1,  306,   -1,   -1,  309,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  319,  320,   -1,
   -1,   -1,  324,   -1,  326,   -1,  328,   -1,   -1,   -1,
   -1,   -1,  334,   -1,  336,   -1,   -1,   -1,  340,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  348,   -1,   -1,  351,
   -1,  353,   -1,   -1,   -1,  357,  358,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  370,   -1,
   -1,   -1,  374,  375,  376,  377,  378,   -1,  380,   -1,
   -1,   -1,   -1,   -1,  386,   -1,  388,  389,   -1,   -1,
  392,  393,   -1,   -1,   -1,   -1,   -1,  399,   -1,   -1,
  402,  403,  404,  405,   -1,  407,   -1,  409,   -1,  411,
   -1,  413,  414,  415,   -1,   -1,   -1,   -1,  420,   -1,
  422,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  433,   -1,   -1,   -1,   -1,  438,  439,   -1,   -1,
   -1,   -1,   -1,  445,   -1,   -1,   -1,  449,   -1,  451,
  452,  453,  454,   -1,  456,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  475,  476,   -1,  478,  479,  480,   -1,
   -1,   -1,   -1,  485,  486,   -1,  488,  489,   -1,  491,
   -1,   -1,  494,  495,   -1,   -1,   -1,  499,  500,   -1,
   -1,   -1,   -1,  505,  506,  507,   -1,  509,  510,  511,
   -1,  513,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  523,  524,  525,  526,  527,  528,  529,  530,  531,
  532,  533,  534,  535,  536,  537,  538,  539,  540,  541,
  542,  543,   -1,  545,  546,  547,  548,  549,  550,   -1,
  552,  553,  554,  555,   -1,  557,   -1,  559,  560,  561,
  562,  563,  564,  565,  566,  567,  568,  569,  570,  571,
  572,  573,  574,  575,  576,   -1,  578,  579,  580,  581,
   -1,  583,  584,  585,  586,  587,   -1,  589,  590,  591,
  592,  593,  594,  595,  596,  597,  598,  599,  600,  601,
   -1,  603,  604,  605,  606,  607,  608,   -1,  610,  611,
  612,  613,  614,  615,  616,  617,   -1,  619,  620,   -1,
   -1,   -1,  624,  625,  626,  627,  628,  629,   -1,   -1,
  632,  633,  634,  635,  636,  637,  638,  639,  640,   -1,
   -1,   -1,   -1,   -1,  646,   -1,  648,  649,  650,  651,
  652,  653,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  672,  673,   -1,  675,  676,  677,  678,  679,  680,  681,
  682,  683,   -1,  685,  686,  687,  688,  689,  690,  691,
  692,  693,  694,  695,  696,   -1,  698,  699,  700,   -1,
  702,   -1,   -1,   -1,  706,  707,  708,  709,  710,  711,
  712,  713,  714,  715,  716,  717,  718,   -1,  720,  721,
   -1,   -1,   -1,   -1,   -1,  727,  728,  729,  730,  731,
  732,  733,  734,  735,  736,  737,  738,  739,  740,   -1,
   -1,  743,  744,   -1,   -1,   -1,   -1,  749,  750,  751,
  752,  753,  754,  257,   -1,  259,   -1,   -1,   -1,   -1,
   -1,  265,   -1,   -1,  268,  269,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  279,  280,   -1,  282,   -1,
  284,   -1,   -1,  287,   -1,   -1,  290,   -1,   -1,   -1,
   -1,   -1,   -1,  297,   -1,  299,  300,   -1,   -1,   -1,
  304,   -1,  306,   -1,   -1,  309,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  319,  320,   -1,   -1,   -1,
  324,   -1,  326,   -1,  328,   -1,   -1,   -1,   -1,   -1,
  334,   -1,  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  348,   -1,   -1,  351,   -1,  353,
   -1,   -1,   -1,  357,  358,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  370,   -1,   -1,   -1,
  374,  375,  376,  377,  378,   -1,  380,   -1,   -1,   -1,
   -1,   -1,  386,   -1,  388,  389,   -1,   -1,  392,  393,
   -1,   -1,   -1,   -1,   -1,  399,   -1,   -1,  402,  403,
  404,  405,   -1,  407,   -1,  409,   -1,  411,  412,  413,
  414,  415,   -1,   -1,   -1,   -1,  420,   -1,  422,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  433,
   -1,   -1,   -1,   -1,  438,  439,   -1,   -1,   -1,   -1,
   -1,  445,   -1,   -1,   -1,  449,   -1,  451,  452,  453,
  454,   -1,  456,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  464,  465,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  475,  476,   -1,  478,  479,  480,   -1,   -1,   -1,
   -1,  485,  486,   -1,  488,  489,   -1,  491,   -1,   -1,
  494,  495,   -1,   -1,   -1,  499,  500,   -1,   -1,   -1,
   -1,  505,  506,  507,   -1,  509,  510,  511,   -1,  513,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  523,
  524,  525,  526,  527,  528,  529,  530,  531,  532,  533,
  534,  535,  536,  537,  538,  539,  540,  541,  542,  543,
   -1,  545,  546,  547,  548,  549,  550,   -1,  552,  553,
  554,  555,   -1,  557,   -1,  559,  560,  561,  562,  563,
  564,  565,  566,  567,  568,  569,  570,  571,  572,  573,
  574,  575,  576,   -1,  578,  579,  580,  581,   -1,  583,
  584,  585,  586,  587,   -1,  589,  590,  591,  592,  593,
  594,  595,  596,  597,  598,  599,  600,  601,   -1,  603,
  604,  605,  606,  607,  608,   -1,  610,  611,  612,  613,
  614,  615,  616,  617,   -1,  619,   -1,   -1,   -1,   -1,
  624,  625,  626,  627,  628,  629,   -1,   -1,  632,  633,
  634,  635,  636,  637,  638,  639,  640,   -1,   -1,   -1,
   -1,   -1,  646,   -1,  648,  649,  650,  651,  652,  653,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  672,  673,
   -1,  675,  676,  677,  678,  679,  680,  681,  682,  683,
   -1,  685,  686,  687,  688,   -1,  690,  691,  692,  693,
  694,  695,  696,   -1,  698,  699,  700,   -1,  702,   -1,
   -1,   -1,  706,  707,  708,  709,  710,  711,  712,  713,
  714,  715,  716,  717,  718,   -1,  720,   -1,   -1,   -1,
   -1,   -1,   -1,  727,  728,  729,  730,  731,  732,  733,
  734,  735,  736,  737,  738,  739,  740,   -1,   -1,  743,
  744,   -1,   -1,   -1,   -1,  749,  750,  751,  752,  753,
  754,  257,   -1,  259,   -1,   -1,   -1,   -1,   -1,  265,
   -1,   -1,  268,  269,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  279,  280,   -1,  282,   -1,  284,   -1,
   -1,  287,   -1,   -1,  290,   -1,   -1,   -1,   -1,   -1,
   -1,  297,   -1,  299,  300,   -1,   -1,   -1,  304,   -1,
  306,   -1,   -1,  309,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  319,  320,   -1,   -1,   -1,  324,   -1,
  326,   -1,  328,   -1,   -1,   -1,   -1,   -1,  334,   -1,
  336,   -1,   -1,   -1,  340,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  348,   -1,   -1,  351,   -1,  353,   -1,   -1,
   -1,  357,  358,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  370,   -1,   -1,   -1,  374,  375,
  376,  377,  378,   -1,  380,   -1,   -1,   -1,   -1,   -1,
  386,   -1,  388,  389,   -1,   -1,  392,  393,   -1,   -1,
   -1,   -1,   -1,  399,   -1,   -1,  402,  403,  404,  405,
   -1,  407,   -1,  409,   -1,  411,  412,  413,  414,  415,
   -1,   -1,   -1,   -1,  420,   -1,  422,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  433,   -1,   -1,
   -1,   -1,  438,  439,   -1,   -1,   -1,   -1,   -1,  445,
   -1,   -1,   -1,  449,   -1,  451,  452,  453,  454,   -1,
  456,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  464,  465,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  475,
  476,   -1,  478,   -1,  480,   -1,   -1,   -1,   -1,  485,
  486,   -1,  488,   -1,   -1,  491,   -1,   -1,  494,  495,
   -1,   -1,   -1,  499,  500,   -1,   -1,   -1,   -1,  505,
  506,  507,   -1,   -1,  510,  511,   -1,  513,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  523,  524,  525,
  526,  527,  528,  529,  530,  531,  532,  533,  534,  535,
  536,  537,  538,  539,  540,  541,  542,  543,   -1,  545,
  546,  547,  548,  549,  550,   -1,  552,  553,  554,  555,
   -1,  557,   -1,  559,  560,  561,  562,  563,  564,  565,
  566,  567,  568,  569,  570,  571,  572,  573,  574,  575,
  576,   -1,  578,  579,  580,  581,   -1,  583,  584,  585,
  586,  587,   -1,  589,  590,  591,  592,  593,  594,  595,
  596,  597,  598,  599,  600,  601,   -1,  603,  604,  605,
  606,  607,  608,   -1,  610,  611,  612,  613,  614,  615,
  616,  617,   -1,  619,  620,   -1,   -1,   -1,   -1,   -1,
   -1,  627,  628,  629,   -1,   -1,  632,  633,  634,  635,
  636,  637,  638,  639,  640,   -1,   -1,   -1,   -1,   -1,
  646,   -1,  648,  649,  650,  651,  652,  653,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  672,  673,   -1,  675,
  676,  677,  678,  679,  680,  681,  682,  683,   -1,  685,
  686,  687,  688,  689,  690,  691,  692,  693,  694,  695,
  696,   -1,  698,  699,  700,   -1,  702,   -1,   -1,   -1,
  706,  707,  708,  709,  710,  711,  712,  713,  714,  715,
  716,  717,  718,   -1,  720,  721,   -1,   -1,   -1,   -1,
   -1,  727,  728,  729,  730,  731,  732,  733,  734,  735,
  736,  737,  738,  739,  740,   -1,  272,  743,  744,  275,
   -1,   -1,   -1,  749,  750,  751,  752,  753,  754,   -1,
   -1,  287,   -1,   -1,   -1,  291,   -1,  293,   -1,   -1,
   -1,  272,   -1,   -1,  275,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  287,   -1,   -1,   -1,
  291,   -1,  293,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  331,   -1,   -1,   -1,   -1,
   -1,  337,  338,  339,   -1,  341,  342,  343,   -1,  345,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  331,   -1,   -1,  359,   -1,  361,  337,  338,  339,   -1,
  341,  342,  343,   -1,  345,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  379,   -1,   -1,   -1,   -1,  359,   -1,
  361,   -1,   -1,  272,  390,   -1,  275,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  287,   -1,
  406,   -1,  291,   -1,  293,   -1,   -1,   -1,   -1,  390,
   -1,   -1,   -1,   -1,  395,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  428,   -1,   -1,  406,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  331,   -1,   -1,   -1,   -1,  428,  337,  338,
  339,   -1,  341,  342,  343,   -1,  345,   -1,   -1,   -1,
  466,  467,   -1,  272,   -1,   -1,  275,   -1,   -1,   -1,
  359,   -1,  361,   -1,   -1,   -1,   -1,  483,  287,   -1,
   -1,   -1,  291,   -1,  293,  466,  467,   -1,   -1,   -1,
  379,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  272,  390,  483,  275,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  406,   -1,  291,
   -1,  293,  331,   -1,   -1,   -1,   -1,   -1,  337,  338,
  339,   -1,  341,  342,  343,   -1,  345,   -1,   -1,  428,
   -1,  272,   -1,   -1,  275,   -1,   -1,   -1,   -1,   -1,
  359,   -1,  361,   -1,   -1,   -1,   -1,   -1,   -1,  331,
  291,   -1,  293,   -1,   -1,  337,  338,  339,   -1,  341,
  342,  343,   -1,  345,   -1,   -1,   -1,  466,  467,   -1,
   -1,  390,   -1,   -1,   -1,   -1,   -1,  359,   -1,  361,
   -1,   -1,   -1,   -1,  483,   -1,   -1,  406,   -1,   -1,
  331,   -1,   -1,   -1,   -1,   -1,  337,  338,  339,   -1,
  341,  342,  343,   -1,  345,   -1,   -1,   -1,  390,  428,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  359,   -1,
  361,   -1,   -1,   -1,  406,   -1,  642,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  428,  466,  467,  390,
  272,  642,   -1,  275,   -1,   -1,   -1,   -1,  674,   -1,
   -1,   -1,   -1,   -1,  483,  406,   -1,   -1,  684,  291,
   -1,  293,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  697,   -1,  674,  466,  467,   -1,  428,   -1,   -1,
   -1,   -1,   -1,  684,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  483,   -1,   -1,   -1,   -1,  697,   -1,   -1,  331,
   -1,   -1,   -1,   -1,   -1,  337,  338,  339,   -1,  341,
  342,  343,   -1,  345,   -1,  466,  467,   -1,   -1,   -1,
  746,   -1,   -1,   -1,   -1,   -1,   -1,  359,   -1,  361,
   -1,   -1,  483,  642,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  746,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  390,   -1,
   -1,   -1,   -1,   -1,   -1,  674,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  406,  684,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  697,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  428,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  642,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  466,  467,   -1,  746,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  674,   -1,   -1,   -1,   -1,
  642,  483,   -1,   -1,   -1,  684,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  697,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  674,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  642,  684,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  697,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  746,   -1,   -1,
   -1,   -1,   -1,  674,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  684,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  697,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  746,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  746,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  642,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  674,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  684,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  697,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  746,
};
static int yyctable[] = {                                     -1,
 1358, 1394,   -1,  740, 1338,   -1,  741, 1338,   -1,  785,
  316,   -1,  786,  316,   -1,  405, 2270,   -1,   39,   51,
   -1, 1100,  126,   -1,  126,  128,   -1,  741, 1340,   -1,
 1354, 1974,   -1, 2088, 2095,   -1, 1537, 1312,   -1, 1600,
  573,   -1, 1788, 1686,   -1, 1789, 1685,   -1, 1410, 1358,
   -1, 1421, 1358,   -1, 1600,  574,   -1, 2204, 1597,   -1,
  329, 1416,   -1,  367, 1416,   -1,  393, 1416,   -1,  693,
 1387,   -1, 1358, 1387,   -1, 2279, 1193,   -1, 2452,  999,
   -1, 2617, 1599,   -1, 2619, 1601,   -1,   89,   96,   -1,
  693, 1358,   -1,  224,  723,   -1,  229,  723,   -1,  238,
  723,   -1,  330,  723,   -1,  331,  723,   -1,  332,  723,
   -1, 2780, 1091,   -1, 2805,  896,   -1,  905, 2074,   -1,
 2958, 1603,   -1,  242, 1371,   -1,  693, 1389,   -1, 1358,
 1389,   -1, 2611, 1332,   -1,  309, 2103,   -1, 3334, 1262,
   -1, 3335, 1262,   -1, 2807, 1089,   -1,  309, 2278,   -1,
 3528, 1274,   -1, 3574,  449,   -1,  418,  429,   -1, 3801,
  651,   -1,  160, 2313,   -1,  203, 2313,   -1,
};
#define YYFINAL 20
#ifndef YYDEBUG
#define YYDEBUG 0
#endif
#define YYMAXTOKEN 756
#if YYDEBUG
static const char *yyname[] = {
"end-of-file",0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,"'('","')'","'*'","'+'","','","'-'","'.'","'/'",0,0,0,0,0,0,0,0,0,0,
"':'","';'","'<'","'='","'>'","'?'",0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,"'['",0,"']'",0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,"ACTIVE","ADD","AFTER","ALL","ALTER","AND","ANY",
"AS","ASC","AT","AVG","AUTO","BEFORE","BEGIN","BETWEEN","BLOB","BY","CAST",
"CHARACTER","CHECK","COLLATE","COMMIT","COMMITTED","COMPUTED","CONCATENATE",
"CONDITIONAL","CONSTRAINT","CONTAINING","COUNT","CREATE","CSTRING","CURRENT",
"CURSOR","DATABASE","DATE","DB_KEY","DECIMAL","DECLARE","DEFAULT","DELETE",
"DESC","DISTINCT","DO","DOMAIN","DROP","ELSE","END","ENTRY_POINT","ESCAPE",
"EXCEPTION","EXECUTE","EXISTS","EXIT","EXTERNAL","FILTER","FOR","FOREIGN",
"FROM","FULL","FUNCTION","GDSCODE","GEQ","GENERATOR","GEN_ID","GRANT","GROUP",
"HAVING","IF","IN","INACTIVE","INNER","INPUT_TYPE","INDEX","INSERT","INTEGER",
"INTO","IS","ISOLATION","JOIN","KEY","CHAR","DEC","DOUBLE","FILE","FLOAT","INT",
"LONG","NULL","NUMERIC","UPPER","VALUE","LENGTH","LEFT","LEQ","LEVEL","LIKE",
"MANUAL","MAXIMUM","MERGE","MINIMUM","MODULE_NAME","NAMES","NATIONAL","NATURAL",
"NCHAR","NEQ","NO","NOT","NOT_GTR","NOT_LSS","OF","ON","ONLY","OPTION","OR",
"ORDER","OUTER","OUTPUT_TYPE","OVERFLOW","PAGE","PAGES","PAGE_SIZE","PARAMETER",
"PASSWORD","PLAN","POSITION","POST_EVENT","PRECISION","PRIMARY","PRIVILEGES",
"PROCEDURE","PROTECTED","READ","REAL","REFERENCES","RESERVING","RETAIN",
"RETURNING_VALUES","RETURNS","REVOKE","RIGHT","ROLLBACK","SEGMENT","SELECT",
"SET","SHADOW","SHARED","SINGULAR","SIZE","SMALLINT","SNAPSHOT","SOME","SORT",
"SQLCODE","STABILITY","STARTING","STATISTICS","SUB_TYPE","SUSPEND","SUM",
"TABLE","THEN","TO","TRANSACTION","TRIGGER","UNCOMMITTED","UNION","UNIQUE",
"UPDATE","USER","VALUES","VARCHAR","VARIABLE","VARYING","VERSION","VIEW","WAIT",
"WHEN","WHERE","WHILE","WITH","WORK","WRITE","FLOAT_NUMBER","DECIMAL_NUMBER",
"LIMIT64_NUMBER","LIMIT64_INT","NUM128","SYMBOL","NUMBER32BIT","STRING",
"INTRODUCER","ACTION","ADMIN","CASCADE","FREE_IT","RESTRICT","ROLE","COLUMN",
"TYPE","EXTRACT","YEAR","MONTH","DAY","HOUR","MINUTE","SECOND","WEEKDAY",
"YEARDAY","TIME","TIMESTAMP","CURRENT_DATE","CURRENT_TIME","CURRENT_TIMESTAMP",
"NUMBER64BIT","SCALEDINT","CURRENT_USER","CURRENT_ROLE","BREAK","SUBSTRING",
"RECREATE","DESCRIPTOR","FIRST","SKIP","CURRENT_CONNECTION",
"CURRENT_TRANSACTION","BIGINT","CASE","NULLIF","COALESCE","USING","NULLS",
"LAST","ROW_COUNT","LOCK","SAVEPOINT","RELEASE","STATEMENT","LEAVE","INSERTING",
"UPDATING","DELETING","BACKUP","DIFFERENCE","OPEN","CLOSE","FETCH","ROWS",
"BLOCK","IIF","SCALAR_ARRAY","CROSS","NEXT","SEQUENCE","RESTART","BOTH",
"COLLATION","COMMENT","BIT_LENGTH","CHAR_LENGTH","CHARACTER_LENGTH","LEADING",
"LOWER","OCTET_LENGTH","TRAILING","TRIM","RETURNING","IGNORE","LIMBO","UNDO",
"REQUESTS","TIMEOUT","ABS","ACCENT","ACOS","ALWAYS","ASCII_CHAR","ASCII_VAL",
"ASIN","ATAN","ATAN2","BIN_AND","BIN_OR","BIN_SHL","BIN_SHR","BIN_XOR","CEIL",
"CONNECT","COS","COSH","COT","DATEADD","DATEDIFF","DECODE","DISCONNECT","EXP",
"FLOOR","GEN_UUID","GENERATED","GLOBAL","HASH","INSENSITIVE","LIST","LN","LOG",
"LOG10","LPAD","MATCHED","MATCHING","MAXVALUE","MILLISECOND","MINVALUE","MOD",
"OVERLAY","PAD","PI","PLACING","POWER","PRESERVE","RAND","RECURSIVE","REPLACE",
"REVERSE","ROUND","RPAD","SENSITIVE","SIGN","SIN","SINH","SPACE","SQRT","START",
"TAN","TANH","TEMPORARY","TRUNC","WEEK","AUTONOMOUS","CHAR_TO_UUID","FIRSTNAME",
"GRANTED","LASTNAME","MIDDLENAME","MAPPING","OS_NAME","SIMILAR","UUID_TO_CHAR",
"CALLER","COMMON","DATA","SOURCE","TWO_PHASE","BIND_PARAM","BIN_NOT","BODY",
"CONTINUE","DDL","DECRYPT","ENCRYPT","ENGINE","NAME","OVER","PACKAGE",
"PARTITION","RDB_GET_CONTEXT","RDB_SET_CONTEXT","SCROLL","PRIOR","ABSOLUTE",
"RELATIVE","ACOSH","ASINH","ATANH","RETURN","DETERMINISTIC","IDENTITY",
"DENSE_RANK","FIRST_VALUE","NTH_VALUE","LAST_VALUE","LAG","LEAD","RANK",
"ROW_NUMBER","SQLSTATE","BOOLEAN","FALSE","TRUE","UNKNOWN","USAGE",
"RDB_RECORD_VERSION","LINGER","TAGS","PLUGIN","SERVERWIDE","INCREMENT",
"TRUSTED","ROW","OFFSET","STDDEV_SAMP","STDDEV_POP","VAR_SAMP","VAR_POP",
"COVAR_SAMP","COVAR_POP","CORR","REGR_AVGX","REGR_AVGY","REGR_COUNT",
"REGR_INTERCEPT","REGR_R2","REGR_SLOPE","REGR_SXX","REGR_SXY","REGR_SYY",
"BASE64_DECODE","BASE64_ENCODE","BINARY","BIND","COMPARE_DECFLOAT",
"CONSISTENCY","COUNTER","CRYPT_HASH","CTR_BIG_ENDIAN","CTR_LENGTH",
"CTR_LITTLE_ENDIAN","CUME_DIST","DECFLOAT","DEFINER","DISABLE","ENABLE",
"EXCESS","EXCLUDE","EXTENDED","FIRST_DAY","FOLLOWING","HEX_DECODE","HEX_ENCODE",
"IDLE","INCLUDE","INT128","INVOKER","IV","LAST_DAY","LATERAL","LEGACY","LOCAL",
"LOCALTIME","LOCALTIMESTAMP","LPARAM","MAKE_DBKEY","MESSAGE","MODE","NATIVE",
"NORMALIZE_DECFLOAT","NTILE","NUMBER","OTHERS","OVERRIDING","PERCENT_RANK",
"PRECEDING","PRIVILEGE","PUBLICATION","QUANTIZE","RANGE","RESETTING",
"RDB_ERROR","RDB_GET_TRANSACTION_CN","RDB_ROLE_IN_USE","RDB_SYSTEM_PRIVILEGE",
"RESET","RSA_DECRYPT","RSA_ENCRYPT","RSA_PRIVATE","RSA_PUBLIC","RSA_SIGN_HASH",
"RSA_VERIFY_HASH","SALT_LENGTH","SECURITY","SESSION","SIGNATURE","SQL","SYSTEM",
"TIES","TIMEZONE_HOUR","TIMEZONE_MINUTE","TOTALORDER","TRAPS","UNBOUNDED",
"VARBINARY","WINDOW","WITHOUT","ZONE","CONNECTIONS","POOL","LIFETIME","CLEAR",
"OLDEST","UMINUS","UPLUS",
};
static const char *yyrule[] = {
"$accept : top",
"top : statement",
"top : statement ';'",
"statement : dml_statement",
"statement : ddl_statement",
"statement : tra_statement",
"statement : mng_statement",
"dml_statement : delete",
"dml_statement : insert",
"dml_statement : merge",
"dml_statement : exec_procedure",
"dml_statement : exec_block",
"dml_statement : savepoint",
"dml_statement : select",
"dml_statement : update",
"dml_statement : update_or_insert",
"ddl_statement : alter",
"ddl_statement : comment",
"ddl_statement : create",
"ddl_statement : create_or_alter",
"ddl_statement : declare",
"ddl_statement : drop",
"ddl_statement : grant",
"ddl_statement : recreate",
"ddl_statement : revoke",
"ddl_statement : set_statistics",
"tra_statement : set_transaction",
"tra_statement : commit",
"tra_statement : rollback",
"mng_statement : set_decfloat_round",
"mng_statement : set_decfloat_traps",
"mng_statement : session_statement",
"mng_statement : set_role",
"mng_statement : session_reset",
"mng_statement : set_time_zone",
"mng_statement : set_bind",
"$$1 :",
"$$2 :",
"grant : GRANT $$1 $$2 grant0",
"$$3 :",
"$$4 :",
"grant0 : $$3 privileges ON table_noise symbol_table_name TO $$4 non_role_grantee_list grant_option granted_by",
"grant0 : $$3 execute_privilege ON PROCEDURE symbol_procedure_name TO $$4 non_role_grantee_list grant_option granted_by",
"grant0 : $$3 execute_privilege ON FUNCTION symbol_UDF_name TO $$4 non_role_grantee_list grant_option granted_by",
"grant0 : $$3 execute_privilege ON PACKAGE symbol_package_name TO $$4 non_role_grantee_list grant_option granted_by",
"grant0 : $$3 usage_privilege ON EXCEPTION symbol_exception_name TO $$4 non_role_grantee_list grant_option granted_by",
"grant0 : $$3 usage_privilege ON GENERATOR symbol_generator_name TO $$4 non_role_grantee_list grant_option granted_by",
"grant0 : $$3 usage_privilege ON SEQUENCE symbol_generator_name TO $$4 non_role_grantee_list grant_option granted_by",
"$$5 :",
"grant0 : $$3 ddl_privileges object TO $$5 non_role_grantee_list grant_option granted_by",
"grant0 : $$3 db_ddl_privileges DATABASE TO $$5 non_role_grantee_list grant_option granted_by",
"$$6 :",
"$$7 :",
"grant0 : $$6 role_name_list TO $$7 role_grantee_list role_admin_option granted_by",
"object : TABLE",
"object : VIEW",
"object : PROCEDURE",
"object : FUNCTION",
"object : PACKAGE",
"object : GENERATOR",
"object : SEQUENCE",
"object : DOMAIN",
"object : EXCEPTION",
"object : ROLE",
"object : CHARACTER SET",
"object : COLLATION",
"object : FILTER",
"table_noise :",
"table_noise : TABLE",
"privileges : ALL",
"privileges : ALL PRIVILEGES",
"$$8 :",
"privileges : $$8 privilege_list",
"privilege_list : $$8 privilege",
"$$9 :",
"privilege_list : privilege_list ',' $$9 privilege",
"execute_privilege : EXECUTE",
"usage_privilege : USAGE",
"privilege : SELECT",
"privilege : INSERT",
"privilege : DELETE",
"privilege : UPDATE column_parens_opt",
"privilege : REFERENCES column_parens_opt",
"ddl_privileges : ALL privileges_opt",
"ddl_privileges : $$8 ddl_privilege_list",
"privileges_opt :",
"privileges_opt : PRIVILEGES",
"ddl_privilege_list : $$8 ddl_privilege",
"ddl_privilege_list : ddl_privilege_list ',' $$9 ddl_privilege",
"ddl_privilege : CREATE",
"ddl_privilege : ALTER ANY",
"ddl_privilege : DROP ANY",
"db_ddl_privileges : ALL privileges_opt",
"db_ddl_privileges : $$8 db_ddl_privilege_list",
"db_ddl_privilege_list : $$8 db_ddl_privilege",
"db_ddl_privilege_list : db_ddl_privilege_list ',' $$9 db_ddl_privilege",
"db_ddl_privilege : CREATE",
"db_ddl_privilege : ALTER",
"db_ddl_privilege : DROP",
"grant_option :",
"grant_option : WITH GRANT OPTION",
"role_admin_option :",
"role_admin_option : WITH ADMIN OPTION",
"granted_by :",
"granted_by : granted_by_text grantor",
"granted_by_text : GRANTED BY",
"granted_by_text : AS",
"grantor : symbol_user_name",
"grantor : USER symbol_user_name",
"$$10 :",
"revoke : REVOKE $$10 $$2 revoke0",
"$$11 :",
"$$12 :",
"revoke0 : rev_grant_option $$11 privileges ON table_noise symbol_table_name FROM $$12 non_role_grantee_list granted_by",
"revoke0 : rev_grant_option $$11 execute_privilege ON PROCEDURE symbol_procedure_name FROM $$12 non_role_grantee_list granted_by",
"revoke0 : rev_grant_option $$11 execute_privilege ON FUNCTION symbol_UDF_name FROM $$12 non_role_grantee_list granted_by",
"revoke0 : rev_grant_option $$11 execute_privilege ON PACKAGE symbol_package_name FROM $$12 non_role_grantee_list granted_by",
"revoke0 : rev_grant_option $$11 usage_privilege ON EXCEPTION symbol_exception_name FROM $$12 non_role_grantee_list granted_by",
"revoke0 : rev_grant_option $$11 usage_privilege ON GENERATOR symbol_generator_name FROM $$12 non_role_grantee_list granted_by",
"revoke0 : rev_grant_option $$11 usage_privilege ON SEQUENCE symbol_generator_name FROM $$12 non_role_grantee_list granted_by",
"$$13 :",
"revoke0 : rev_grant_option $$11 ddl_privileges object FROM $$13 non_role_grantee_list granted_by",
"revoke0 : rev_grant_option $$11 db_ddl_privileges DATABASE FROM $$13 non_role_grantee_list granted_by",
"$$14 :",
"revoke0 : rev_admin_option $$14 role_name_list FROM $$5 role_grantee_list granted_by",
"revoke0 : ALL ON ALL FROM $$5 non_role_grantee_list",
"rev_grant_option :",
"rev_grant_option : GRANT OPTION FOR",
"rev_admin_option :",
"rev_admin_option : ADMIN OPTION FOR",
"$$15 :",
"non_role_grantee_list : $$15 grantee",
"non_role_grantee_list : $$15 user_grantee",
"$$16 :",
"non_role_grantee_list : non_role_grantee_list ',' $$16 grantee",
"non_role_grantee_list : non_role_grantee_list ',' $$16 user_grantee",
"grantee : PROCEDURE symbol_procedure_name",
"grantee : FUNCTION symbol_UDF_name",
"grantee : PACKAGE symbol_package_name",
"grantee : TRIGGER symbol_trigger_name",
"grantee : VIEW symbol_view_name",
"grantee : ROLE symbol_role_name",
"grantee : SYSTEM PRIVILEGE valid_symbol_name",
"user_grantee : symbol_user_name",
"user_grantee : USER symbol_user_name",
"user_grantee : GROUP symbol_user_name",
"role_name_list : $$2 role_name",
"$$17 :",
"role_name_list : role_name_list ',' $$17 role_name",
"role_name : symbol_role_name",
"role_name : DEFAULT symbol_role_name",
"role_grantee_list : $$15 role_grantee",
"role_grantee_list : role_grantee_list ',' $$16 role_grantee",
"role_grantee : symbol_user_name",
"role_grantee : USER symbol_user_name",
"role_grantee : ROLE symbol_user_name",
"declare : DECLARE declare_clause",
"declare_clause : FILTER filter_decl_clause",
"declare_clause : EXTERNAL FUNCTION udf_decl_clause",
"$$18 :",
"$$19 :",
"$$20 :",
"udf_decl_clause : symbol_UDF_name $$18 $$19 arg_desc_list1 RETURNS $$20 return_value1 ENTRY_POINT utf_string MODULE_NAME utf_string",
"udf_data_type : simple_type",
"udf_data_type : BLOB",
"udf_data_type : CSTRING '(' pos_short_integer ')' charset_clause",
"arg_desc_list1 :",
"$$21 :",
"arg_desc_list1 : $$21 arg_desc_list",
"$$22 :",
"arg_desc_list1 : '(' $$22 arg_desc_list ')'",
"arg_desc_list : $$21 arg_desc",
"$$23 :",
"arg_desc_list : arg_desc_list ',' $$23 arg_desc",
"arg_desc : udf_data_type param_mechanism",
"param_mechanism :",
"param_mechanism : BY DESCRIPTOR",
"param_mechanism : BY SCALAR_ARRAY",
"param_mechanism : NULL",
"$$24 :",
"return_value1 : $$24 return_value",
"$$25 :",
"return_value1 : '(' $$25 return_value ')'",
"return_value : udf_data_type return_mechanism",
"return_value : PARAMETER pos_short_integer",
"return_mechanism :",
"return_mechanism : BY VALUE",
"return_mechanism : BY DESCRIPTOR",
"return_mechanism : FREE_IT",
"return_mechanism : BY DESCRIPTOR FREE_IT",
"filter_decl_clause : symbol_filter_name INPUT_TYPE blob_filter_subtype OUTPUT_TYPE blob_filter_subtype ENTRY_POINT utf_string MODULE_NAME utf_string",
"blob_filter_subtype : symbol_blob_subtype_name",
"blob_filter_subtype : signed_short_integer",
"create : CREATE create_clause",
"create_clause : EXCEPTION exception_clause",
"$$26 :",
"$$27 :",
"create_clause : unique_opt order_direction INDEX symbol_index_name ON simple_table_name $$26 $$27 index_definition",
"create_clause : FUNCTION function_clause",
"create_clause : PROCEDURE procedure_clause",
"create_clause : TABLE table_clause",
"create_clause : GLOBAL TEMPORARY TABLE gtt_table_clause",
"create_clause : TRIGGER trigger_clause",
"create_clause : VIEW view_clause",
"create_clause : GENERATOR generator_clause",
"create_clause : SEQUENCE generator_clause",
"create_clause : DATABASE db_clause",
"create_clause : DOMAIN domain_clause",
"create_clause : SHADOW shadow_clause",
"create_clause : ROLE role_clause",
"create_clause : COLLATION collation_clause",
"create_clause : USER create_user_clause",
"create_clause : PACKAGE package_clause",
"create_clause : PACKAGE BODY package_body_clause",
"$$28 :",
"create_clause : MAPPING $$28 create_map_clause",
"$$29 :",
"create_clause : GLOBAL MAPPING $$29 create_map_clause",
"recreate : RECREATE recreate_clause",
"recreate_clause : PROCEDURE procedure_clause",
"recreate_clause : FUNCTION function_clause",
"recreate_clause : TABLE table_clause",
"recreate_clause : GLOBAL TEMPORARY TABLE gtt_table_clause",
"recreate_clause : VIEW view_clause",
"recreate_clause : TRIGGER trigger_clause",
"recreate_clause : PACKAGE package_clause",
"recreate_clause : PACKAGE BODY package_body_clause",
"recreate_clause : EXCEPTION exception_clause",
"recreate_clause : GENERATOR generator_clause",
"recreate_clause : SEQUENCE generator_clause",
"recreate_clause : USER create_user_clause",
"create_or_alter : CREATE OR ALTER replace_clause",
"replace_clause : PROCEDURE replace_procedure_clause",
"replace_clause : FUNCTION replace_function_clause",
"replace_clause : TRIGGER replace_trigger_clause",
"replace_clause : PACKAGE replace_package_clause",
"replace_clause : VIEW replace_view_clause",
"replace_clause : EXCEPTION replace_exception_clause",
"replace_clause : GENERATOR replace_sequence_clause",
"replace_clause : SEQUENCE replace_sequence_clause",
"replace_clause : USER replace_user_clause",
"replace_clause : MAPPING $$28 replace_map_clause",
"replace_clause : GLOBAL MAPPING $$29 replace_map_clause",
"exception_clause : symbol_exception_name sql_string",
"replace_exception_clause : symbol_exception_name sql_string",
"alter_exception_clause : symbol_exception_name sql_string",
"unique_opt :",
"unique_opt : UNIQUE",
"index_definition : column_list",
"index_definition : column_parens",
"index_definition : computed_by '(' value ')'",
"$$30 :",
"$$31 :",
"shadow_clause : pos_short_integer manual_auto conditional utf_string first_file_length $$30 $$31 sec_shadow_files",
"manual_auto :",
"manual_auto : MANUAL",
"manual_auto : AUTO",
"conditional :",
"conditional : CONDITIONAL",
"first_file_length :",
"first_file_length : LENGTH equals long_integer page_noise",
"sec_shadow_files :",
"$$32 :",
"sec_shadow_files : $$32 db_file_list",
"db_file_list : db_file",
"db_file_list : db_file_list db_file",
"$$33 :",
"$$34 :",
"domain_clause : symbol_column_name as_opt data_type domain_default_opt $$33 $$34 domain_constraints_opt collate_clause",
"domain_constraints_opt :",
"domain_constraints_opt : $$34 domain_constraints",
"domain_constraints : $$34 domain_constraint",
"$$35 :",
"domain_constraints : domain_constraints $$35 domain_constraint",
"domain_constraint : null_constraint",
"domain_constraint : check_constraint",
"as_opt :",
"as_opt : AS",
"domain_default : DEFAULT default_value",
"domain_default_opt :",
"domain_default_opt : domain_default",
"null_constraint : NOT NULL",
"check_constraint : CHECK '(' search_condition ')'",
"$$36 :",
"$$37 :",
"generator_clause : symbol_generator_name $$36 $$37 create_sequence_options",
"create_sequence_options :",
"$$38 :",
"create_sequence_options : $$37 create_seq_option $$38 create_sequence_options",
"create_seq_option : $$37 start_with_opt",
"create_seq_option : $$37 step_option",
"start_with_opt : START WITH sequence_value",
"step_option : INCREMENT by_noise signed_long_integer",
"by_noise :",
"by_noise : BY",
"$$39 :",
"replace_sequence_clause : symbol_generator_name $$39 $$37 replace_sequence_options",
"replace_sequence_options :",
"replace_sequence_options : $$37 replace_seq_option $$38 replace_sequence_options",
"replace_seq_option : RESTART",
"replace_seq_option : $$37 start_with_opt",
"replace_seq_option : $$37 step_option",
"$$40 :",
"alter_sequence_clause : symbol_generator_name $$40 $$37 alter_sequence_options",
"alter_sequence_options :",
"alter_sequence_options : $$37 alter_seq_option $$38 alter_sequence_options",
"alter_seq_option : $$37 restart_option",
"alter_seq_option : $$37 step_option",
"restart_option : RESTART with_opt",
"with_opt :",
"with_opt : WITH sequence_value",
"set_generator_clause : SET GENERATOR symbol_generator_name TO sequence_value",
"sequence_value : signed_long_integer",
"sequence_value : NUMBER64BIT",
"sequence_value : '-' NUMBER64BIT",
"sequence_value : '-' LIMIT64_INT",
"$$41 :",
"$$42 :",
"role_clause : symbol_role_name $$41 $$42 opt_system_privileges",
"opt_system_privileges :",
"opt_system_privileges : $$42 set_system_privileges",
"opt_system_privileges : $$42 drop_system_privileges",
"$$43 :",
"set_system_privileges : SET SYSTEM PRIVILEGES TO $$43 system_privileges_list",
"drop_system_privileges : DROP SYSTEM PRIVILEGES",
"system_privileges_list : $$42 system_privilege",
"$$44 :",
"system_privileges_list : system_privileges_list ',' $$44 system_privilege",
"system_privilege : valid_symbol_name",
"$$45 :",
"$$46 :",
"$$47 :",
"$$48 :",
"collation_clause : symbol_collation_name FOR symbol_character_set_name $$45 $$46 collation_sequence_definition $$47 collation_attribute_list_opt $$48 collation_specific_attribute_opt",
"collation_sequence_definition :",
"collation_sequence_definition : FROM symbol_collation_name",
"collation_sequence_definition : FROM EXTERNAL '(' utf_string ')'",
"collation_attribute_list_opt :",
"collation_attribute_list_opt : $$46 collation_attribute_list",
"collation_attribute_list : $$46 collation_attribute",
"$$49 :",
"collation_attribute_list : collation_attribute_list $$49 collation_attribute",
"collation_attribute : $$46 collation_pad_attribute",
"collation_attribute : $$46 collation_case_attribute",
"collation_attribute : $$46 collation_accent_attribute",
"collation_pad_attribute : NO PAD",
"collation_pad_attribute : PAD SPACE",
"collation_case_attribute : CASE SENSITIVE",
"collation_case_attribute : CASE INSENSITIVE",
"collation_accent_attribute : ACCENT SENSITIVE",
"collation_accent_attribute : ACCENT INSENSITIVE",
"collation_specific_attribute_opt :",
"collation_specific_attribute_opt : utf_string",
"alter_charset_clause : symbol_character_set_name SET DEFAULT COLLATION symbol_collation_name",
"alter_eds_conn_pool_clause : SET SIZE unsigned_short_integer",
"alter_eds_conn_pool_clause : SET LIFETIME unsigned_short_integer eds_pool_lifetime_mult",
"alter_eds_conn_pool_clause : CLEAR sql_string",
"alter_eds_conn_pool_clause : CLEAR ALL",
"alter_eds_conn_pool_clause : CLEAR OLDEST",
"eds_pool_lifetime_mult : HOUR",
"eds_pool_lifetime_mult : MINUTE",
"eds_pool_lifetime_mult : SECOND",
"$$50 :",
"$$51 :",
"$$52 :",
"db_clause : db_name $$50 $$51 db_initial_desc1 $$52 db_rem_desc1",
"equals :",
"equals : '='",
"db_name : utf_string",
"db_initial_desc1 :",
"db_initial_desc1 : $$51 db_initial_desc",
"db_initial_desc : $$51 db_initial_option",
"$$53 :",
"db_initial_desc : db_initial_desc $$53 db_initial_option",
"db_initial_option : PAGE_SIZE equals NUMBER32BIT",
"db_initial_option : USER symbol_user_name",
"db_initial_option : USER utf_string",
"db_initial_option : ROLE valid_symbol_name",
"db_initial_option : ROLE utf_string",
"db_initial_option : PASSWORD utf_string",
"db_initial_option : SET NAMES utf_string",
"db_initial_option : LENGTH equals long_integer page_noise",
"db_rem_desc1 :",
"db_rem_desc1 : $$51 db_rem_desc",
"db_rem_desc : $$51 db_rem_option",
"db_rem_desc : db_rem_desc $$53 db_rem_option",
"db_rem_option : db_file",
"db_rem_option : DEFAULT CHARACTER SET symbol_character_set_name",
"db_rem_option : DEFAULT CHARACTER SET symbol_character_set_name COLLATION symbol_collation_name",
"db_rem_option : DIFFERENCE FILE utf_string",
"$$54 :",
"$$55 :",
"db_file : FILE utf_string $$54 $$55 file_desc1",
"file_desc1 :",
"file_desc1 : $$55 file_desc",
"file_desc : $$55 file_clause",
"$$56 :",
"file_desc : file_desc $$56 file_clause",
"file_clause : STARTING file_clause_noise long_integer",
"file_clause : LENGTH equals long_integer page_noise",
"file_clause_noise :",
"file_clause_noise : AT",
"file_clause_noise : AT PAGE",
"page_noise :",
"page_noise : PAGE",
"page_noise : PAGES",
"$$57 :",
"$$58 :",
"$$59 :",
"table_clause : simple_table_name external_file $$57 '(' $$58 table_elements ')' $$59 table_attributes",
"table_attributes :",
"$$60 :",
"$$61 :",
"table_attributes : $$60 table_attribute $$61 table_attributes",
"table_attribute : sql_security_clause",
"table_attribute : publication_state",
"sql_security_clause : SQL SECURITY DEFINER",
"sql_security_clause : SQL SECURITY INVOKER",
"sql_security_clause_opt :",
"sql_security_clause_opt : sql_security_clause",
"publication_state : ENABLE PUBLICATION",
"publication_state : DISABLE PUBLICATION",
"$$62 :",
"$$63 :",
"gtt_table_clause : simple_table_name $$62 '(' $$58 table_elements ')' $$63 gtt_ops",
"$$64 :",
"gtt_ops : $$64 gtt_op",
"$$65 :",
"gtt_ops : gtt_ops ',' $$65 gtt_op",
"gtt_op :",
"gtt_op : sql_security_clause_opt",
"gtt_op : ON COMMIT DELETE ROWS",
"gtt_op : ON COMMIT PRESERVE ROWS",
"external_file :",
"external_file : EXTERNAL FILE utf_string",
"external_file : EXTERNAL utf_string",
"table_elements : $$64 table_element",
"table_elements : table_elements ',' $$65 table_element",
"$$66 :",
"table_element : $$66 column_def",
"table_element : $$66 table_constraint_definition",
"$$67 :",
"$$68 :",
"column_def : symbol_column_name data_type_or_domain domain_default_opt $$67 $$68 column_constraint_clause collate_clause",
"$$69 :",
"column_def : symbol_column_name data_type_or_domain identity_clause $$69 $$68 column_constraint_clause collate_clause",
"column_def : symbol_column_name non_array_type def_computed",
"column_def : symbol_column_name def_computed",
"$$70 :",
"$$71 :",
"identity_clause : GENERATED identity_clause_type AS IDENTITY $$70 $$71 identity_clause_options_opt",
"identity_clause_type : BY DEFAULT",
"identity_clause_type : ALWAYS",
"identity_clause_options_opt :",
"$$72 :",
"identity_clause_options_opt : '(' $$72 identity_clause_options ')'",
"identity_clause_options : identity_clause_options $$72 identity_clause_option",
"identity_clause_options : $$71 identity_clause_option",
"identity_clause_option : START WITH sequence_value",
"identity_clause_option : INCREMENT by_noise signed_long_integer",
"def_computed : computed_clause '(' value ')'",
"computed_clause : computed_by",
"computed_clause : generated_always_clause",
"generated_always_clause : GENERATED ALWAYS AS",
"computed_by : COMPUTED BY",
"computed_by : COMPUTED",
"data_type_or_domain : data_type",
"data_type_or_domain : symbol_column_name",
"collate_clause :",
"collate_clause : COLLATE symbol_collation_name",
"data_type_descriptor : data_type",
"data_type_descriptor : TYPE OF symbol_column_name",
"data_type_descriptor : TYPE OF COLUMN symbol_column_name '.' symbol_column_name",
"data_type_descriptor : symbol_column_name",
"default_value : constant",
"default_value : current_user",
"default_value : current_role",
"default_value : internal_info",
"default_value : null_value",
"default_value : datetime_value_expression",
"column_constraint_clause :",
"$$73 :",
"column_constraint_clause : $$73 column_constraint_list",
"column_constraint_list : $$73 column_constraint_def",
"$$74 :",
"column_constraint_list : column_constraint_list $$74 column_constraint_def",
"column_constraint_def : constraint_name_opt $$74 column_constraint",
"column_constraint : null_constraint",
"column_constraint : check_constraint",
"column_constraint : REFERENCES symbol_table_name column_parens_opt referential_trigger_action constraint_index_opt",
"column_constraint : UNIQUE constraint_index_opt",
"column_constraint : PRIMARY KEY constraint_index_opt",
"$$75 :",
"table_constraint_definition : constraint_name_opt $$75 table_constraint",
"constraint_name_opt :",
"constraint_name_opt : CONSTRAINT symbol_constraint_name",
"table_constraint : UNIQUE column_parens constraint_index_opt",
"table_constraint : PRIMARY KEY column_parens constraint_index_opt",
"table_constraint : FOREIGN KEY column_parens REFERENCES symbol_table_name column_parens_opt referential_trigger_action constraint_index_opt",
"table_constraint : check_constraint",
"constraint_index_opt :",
"constraint_index_opt : USING order_direction INDEX symbol_index_name",
"referential_trigger_action :",
"referential_trigger_action : update_rule",
"referential_trigger_action : delete_rule",
"referential_trigger_action : delete_rule update_rule",
"referential_trigger_action : update_rule delete_rule",
"update_rule : ON UPDATE referential_action",
"delete_rule : ON DELETE referential_action",
"referential_action : CASCADE",
"referential_action : SET DEFAULT",
"referential_action : SET NULL",
"referential_action : NO ACTION",
"procedure_clause : psql_procedure_clause",
"procedure_clause : external_procedure_clause",
"psql_procedure_clause : procedure_clause_start sql_security_clause_opt AS local_declarations_opt full_proc_block",
"external_procedure_clause : procedure_clause_start external_clause external_body_clause_opt",
"$$76 :",
"$$77 :",
"$$78 :",
"procedure_clause_start : symbol_procedure_name $$76 $$77 input_parameters $$78 output_parameters",
"alter_procedure_clause : procedure_clause",
"replace_procedure_clause : procedure_clause",
"input_parameters :",
"input_parameters : '(' ')'",
"input_parameters : '(' $$22 input_proc_parameters ')'",
"output_parameters :",
"output_parameters : RETURNS '(' $$23 output_proc_parameters ')'",
"input_proc_parameters : $$21 input_proc_parameter",
"input_proc_parameters : input_proc_parameters ',' $$23 input_proc_parameter",
"input_proc_parameter : column_domain_or_non_array_type collate_clause default_par_opt",
"output_proc_parameters : output_proc_parameter",
"output_proc_parameters : output_proc_parameters ',' $$23 output_proc_parameter",
"output_proc_parameter : column_domain_or_non_array_type collate_clause",
"column_domain_or_non_array_type : symbol_column_name domain_or_non_array_type",
"default_par_opt :",
"default_par_opt : DEFAULT default_value",
"default_par_opt : '=' default_value",
"function_clause : psql_function_clause",
"function_clause : external_function_clause",
"psql_function_clause : function_clause_start sql_security_clause_opt AS local_declarations_opt full_proc_block",
"external_function_clause : function_clause_start external_clause external_body_clause_opt",
"$$79 :",
"function_clause_start : symbol_UDF_name $$79 $$19 input_parameters RETURNS domain_or_non_array_type collate_clause deterministic_opt",
"deterministic_opt :",
"deterministic_opt : NOT DETERMINISTIC",
"deterministic_opt : DETERMINISTIC",
"external_clause : EXTERNAL NAME utf_string ENGINE valid_symbol_name",
"external_clause : EXTERNAL ENGINE valid_symbol_name",
"external_body_clause_opt :",
"external_body_clause_opt : AS utf_string",
"alter_function_clause : function_clause",
"replace_function_clause : function_clause",
"package_clause : symbol_package_name sql_security_clause_opt AS BEGIN package_items_opt END",
"package_items_opt : package_items",
"package_items_opt :",
"package_items : package_item",
"package_items : package_items package_item",
"package_item : FUNCTION function_clause_start ';'",
"package_item : PROCEDURE procedure_clause_start ';'",
"alter_package_clause : package_clause",
"replace_package_clause : package_clause",
"package_body_clause : symbol_package_name AS BEGIN package_items package_body_items_opt END",
"package_body_clause : symbol_package_name AS BEGIN package_body_items_opt END",
"package_body_items_opt :",
"package_body_items_opt : package_body_items",
"package_body_items : package_body_item",
"package_body_items : package_body_items package_body_item",
"package_body_item : FUNCTION psql_function_clause",
"package_body_item : FUNCTION external_function_clause ';'",
"package_body_item : PROCEDURE psql_procedure_clause",
"package_body_item : PROCEDURE external_procedure_clause ';'",
"local_declarations_opt : local_forward_declarations_opt local_nonforward_declarations_opt",
"local_forward_declarations_opt :",
"local_forward_declarations_opt : local_forward_declarations",
"local_forward_declarations : local_forward_declaration",
"local_forward_declarations : local_forward_declarations local_forward_declaration",
"local_forward_declaration : local_declaration_subproc_start ';'",
"local_forward_declaration : local_declaration_subfunc_start ';'",
"local_nonforward_declarations_opt :",
"local_nonforward_declarations_opt : local_nonforward_declarations",
"local_nonforward_declarations : local_nonforward_declaration",
"local_nonforward_declarations : local_nonforward_declarations local_nonforward_declaration",
"local_nonforward_declaration : DECLARE var_decl_opt local_declaration_item ';'",
"local_nonforward_declaration : local_declaration_subproc_start AS local_declarations_opt full_proc_block",
"local_nonforward_declaration : local_declaration_subfunc_start AS local_declarations_opt full_proc_block",
"$$80 :",
"$$81 :",
"$$82 :",
"local_declaration_subproc_start : DECLARE PROCEDURE symbol_procedure_name $$80 $$81 input_parameters $$82 output_parameters",
"$$83 :",
"$$84 :",
"local_declaration_subfunc_start : DECLARE FUNCTION symbol_UDF_name $$83 $$84 input_parameters RETURNS domain_or_non_array_type collate_clause deterministic_opt",
"local_declaration_item : var_declaration_item",
"local_declaration_item : cursor_declaration_item",
"var_declaration_item : column_domain_or_non_array_type collate_clause default_par_opt",
"var_decl_opt :",
"var_decl_opt : VARIABLE",
"cursor_declaration_item : symbol_cursor_name scroll_opt CURSOR FOR '(' select ')'",
"scroll_opt :",
"scroll_opt : NO SCROLL",
"scroll_opt : SCROLL",
"proc_block : proc_statement",
"proc_block : full_proc_block",
"full_proc_block : BEGIN full_proc_block_body END",
"full_proc_block_body :",
"full_proc_block_body : proc_statements",
"full_proc_block_body : proc_statements excp_hndl_statements",
"proc_statements : proc_block",
"proc_statements : proc_statements proc_block",
"proc_statement : simple_proc_statement ';'",
"proc_statement : complex_proc_statement",
"simple_proc_statement : assignment_statement",
"simple_proc_statement : insert",
"simple_proc_statement : merge",
"simple_proc_statement : update",
"simple_proc_statement : update_or_insert",
"simple_proc_statement : delete",
"simple_proc_statement : singleton_select",
"simple_proc_statement : exec_procedure",
"simple_proc_statement : exec_sql",
"simple_proc_statement : exec_into",
"simple_proc_statement : exec_function",
"simple_proc_statement : excp_statement",
"simple_proc_statement : raise_statement",
"simple_proc_statement : post_event",
"simple_proc_statement : cursor_statement",
"simple_proc_statement : breakleave",
"simple_proc_statement : continue",
"simple_proc_statement : SUSPEND",
"simple_proc_statement : EXIT",
"simple_proc_statement : RETURN value",
"simple_proc_statement : mng_statement",
"assignment_statement : assignment",
"assignment_statement : ':' assignment",
"complex_proc_statement : in_autonomous_transaction",
"complex_proc_statement : if_then_else",
"complex_proc_statement : while",
"complex_proc_statement : for_select",
"complex_proc_statement : for_exec_into",
"in_autonomous_transaction : IN AUTONOMOUS TRANSACTION DO proc_block",
"excp_statement : EXCEPTION symbol_exception_name",
"excp_statement : EXCEPTION symbol_exception_name value",
"excp_statement : EXCEPTION symbol_exception_name USING '(' value_list ')'",
"raise_statement : EXCEPTION",
"$$85 :",
"$$86 :",
"for_select : label_def_opt FOR select $$85 $$86 for_select_into_cursor DO proc_block",
"for_select_into_cursor : into_variable_list cursor_def_opt",
"for_select_into_cursor : into_variable_list_opt cursor_def",
"into_variable_list_opt :",
"into_variable_list_opt : into_variable_list",
"into_variable_list : INTO variable_list",
"$$87 :",
"$$88 :",
"$$89 :",
"exec_sql : EXECUTE STATEMENT $$87 $$88 exec_stmt_inputs $$89 exec_stmt_options",
"exec_into : exec_sql INTO variable_list",
"for_exec_into : label_def_opt FOR exec_into DO proc_block",
"exec_stmt_inputs : value",
"$$90 :",
"exec_stmt_inputs : '(' value ')' '(' $$90 named_params_list ')'",
"exec_stmt_inputs : '(' value ')' '(' $$90 not_named_params_list ')'",
"named_params_list : $$88 named_param",
"named_params_list : named_params_list ',' $$89 named_param",
"named_param : symbol_variable_name BIND_PARAM value",
"named_param : EXCESS symbol_variable_name BIND_PARAM value",
"not_named_params_list : $$88 not_named_param",
"not_named_params_list : not_named_params_list ',' $$89 not_named_param",
"not_named_param : value",
"exec_stmt_options :",
"exec_stmt_options : $$88 exec_stmt_options_list",
"$$91 :",
"exec_stmt_options_list : exec_stmt_options_list $$91 exec_stmt_option",
"exec_stmt_options_list : $$88 exec_stmt_option",
"exec_stmt_option : ON EXTERNAL DATA SOURCE value",
"exec_stmt_option : ON EXTERNAL value",
"exec_stmt_option : AS USER value",
"exec_stmt_option : PASSWORD value",
"exec_stmt_option : ROLE value",
"exec_stmt_option : WITH AUTONOMOUS TRANSACTION",
"exec_stmt_option : WITH COMMON TRANSACTION",
"exec_stmt_option : WITH CALLER PRIVILEGES",
"if_then_else : IF '(' search_condition ')' THEN proc_block ELSE proc_block",
"if_then_else : IF '(' search_condition ')' THEN proc_block",
"post_event : POST_EVENT value event_argument_opt",
"event_argument_opt :",
"singleton_select : select INTO variable_list",
"variable : ':' symbol_variable_name",
"variable_list : variable",
"variable_list : column_name",
"variable_list : variable_list ',' column_name",
"variable_list : variable_list ',' variable",
"while : label_def_opt WHILE '(' search_condition ')' DO proc_block",
"label_def_opt :",
"label_def_opt : symbol_label_name ':'",
"breakleave : BREAK",
"breakleave : LEAVE label_use_opt",
"continue : CONTINUE label_use_opt",
"label_use_opt :",
"label_use_opt : symbol_label_name",
"cursor_def_opt :",
"cursor_def_opt : cursor_def",
"cursor_def : AS CURSOR symbol_cursor_name",
"excp_hndl_statements : excp_hndl_statement",
"excp_hndl_statements : excp_hndl_statements excp_hndl_statement",
"$$92 :",
"$$93 :",
"excp_hndl_statement : WHEN $$92 $$93 errors DO proc_block",
"$$94 :",
"errors : $$94 err",
"$$95 :",
"errors : errors ',' $$95 err",
"err : SQLCODE signed_short_integer",
"err : SQLSTATE STRING",
"err : GDSCODE symbol_gdscode_name",
"err : EXCEPTION symbol_exception_name",
"err : ANY",
"cursor_statement : open_cursor",
"cursor_statement : fetch_cursor",
"cursor_statement : close_cursor",
"open_cursor : OPEN symbol_cursor_name",
"close_cursor : CLOSE symbol_cursor_name",
"$$96 :",
"$$97 :",
"fetch_cursor : FETCH $$96 $$97 fetch_scroll FROM symbol_cursor_name into_variable_list_opt",
"fetch_cursor : FETCH symbol_cursor_name into_variable_list_opt",
"fetch_scroll : FIRST",
"fetch_scroll : LAST",
"fetch_scroll : PRIOR",
"fetch_scroll : NEXT",
"fetch_scroll : ABSOLUTE value",
"fetch_scroll : RELATIVE value",
"exec_procedure : EXECUTE PROCEDURE symbol_procedure_name proc_inputs proc_outputs_opt",
"exec_procedure : EXECUTE PROCEDURE symbol_package_name '.' symbol_procedure_name proc_inputs proc_outputs_opt",
"proc_inputs :",
"proc_inputs : value_list",
"proc_inputs : '(' value_list ')'",
"proc_outputs_opt :",
"proc_outputs_opt : RETURNING_VALUES variable_list",
"proc_outputs_opt : RETURNING_VALUES '(' variable_list ')'",
"$$98 :",
"$$99 :",
"$$100 :",
"exec_block : EXECUTE BLOCK $$98 $$99 block_input_params $$100 output_parameters AS local_declarations_opt full_proc_block",
"block_input_params :",
"block_input_params : '(' $$22 block_parameters ')'",
"block_parameters : $$21 block_parameter",
"block_parameters : block_parameters ',' $$23 block_parameter",
"block_parameter : column_domain_or_non_array_type collate_clause '=' parameter",
"view_clause : simple_table_name column_parens_opt AS select_expr check_opt",
"replace_view_clause : view_clause",
"alter_view_clause : view_clause",
"check_opt :",
"check_opt : WITH CHECK OPTION",
"trigger_clause : create_trigger_start trg_sql_security_clause AS local_declarations_opt full_proc_block",
"trigger_clause : create_trigger_start external_clause external_body_clause_opt",
"$$101 :",
"$$102 :",
"create_trigger_start : symbol_trigger_name $$101 $$102 create_trigger_common",
"$$103 :",
"create_trigger_common : trigger_active $$103 trigger_type trigger_position",
"create_trigger_common : FOR symbol_table_name trigger_active table_trigger_type trigger_position",
"replace_trigger_clause : trigger_clause",
"trigger_active : ACTIVE",
"trigger_active : INACTIVE",
"trigger_active :",
"trigger_type : table_trigger_type trigger_position ON symbol_table_name",
"trigger_type : ON trigger_db_type",
"trigger_type : trigger_type_prefix trigger_ddl_type",
"table_trigger_type : trigger_type_prefix trigger_type_suffix",
"trigger_db_type : CONNECT",
"trigger_db_type : DISCONNECT",
"trigger_db_type : TRANSACTION START",
"trigger_db_type : TRANSACTION COMMIT",
"trigger_db_type : TRANSACTION ROLLBACK",
"trigger_ddl_type : trigger_ddl_type_items",
"trigger_ddl_type : ANY DDL STATEMENT",
"trigger_ddl_type_items : CREATE TABLE",
"trigger_ddl_type_items : ALTER TABLE",
"trigger_ddl_type_items : DROP TABLE",
"trigger_ddl_type_items : CREATE PROCEDURE",
"trigger_ddl_type_items : ALTER PROCEDURE",
"trigger_ddl_type_items : DROP PROCEDURE",
"trigger_ddl_type_items : CREATE FUNCTION",
"trigger_ddl_type_items : ALTER FUNCTION",
"trigger_ddl_type_items : DROP FUNCTION",
"trigger_ddl_type_items : CREATE TRIGGER",
"trigger_ddl_type_items : ALTER TRIGGER",
"trigger_ddl_type_items : DROP TRIGGER",
"trigger_ddl_type_items : CREATE EXCEPTION",
"trigger_ddl_type_items : ALTER EXCEPTION",
"trigger_ddl_type_items : DROP EXCEPTION",
"trigger_ddl_type_items : CREATE VIEW",
"trigger_ddl_type_items : ALTER VIEW",
"trigger_ddl_type_items : DROP VIEW",
"trigger_ddl_type_items : CREATE DOMAIN",
"trigger_ddl_type_items : ALTER DOMAIN",
"trigger_ddl_type_items : DROP DOMAIN",
"trigger_ddl_type_items : CREATE ROLE",
"trigger_ddl_type_items : ALTER ROLE",
"trigger_ddl_type_items : DROP ROLE",
"trigger_ddl_type_items : CREATE SEQUENCE",
"trigger_ddl_type_items : ALTER SEQUENCE",
"trigger_ddl_type_items : DROP SEQUENCE",
"trigger_ddl_type_items : CREATE USER",
"trigger_ddl_type_items : ALTER USER",
"trigger_ddl_type_items : DROP USER",
"trigger_ddl_type_items : CREATE INDEX",
"trigger_ddl_type_items : ALTER INDEX",
"trigger_ddl_type_items : DROP INDEX",
"trigger_ddl_type_items : CREATE COLLATION",
"trigger_ddl_type_items : DROP COLLATION",
"trigger_ddl_type_items : ALTER CHARACTER SET",
"trigger_ddl_type_items : CREATE PACKAGE",
"trigger_ddl_type_items : ALTER PACKAGE",
"trigger_ddl_type_items : DROP PACKAGE",
"trigger_ddl_type_items : CREATE PACKAGE BODY",
"trigger_ddl_type_items : DROP PACKAGE BODY",
"trigger_ddl_type_items : CREATE MAPPING",
"trigger_ddl_type_items : ALTER MAPPING",
"trigger_ddl_type_items : DROP MAPPING",
"trigger_ddl_type_items : trigger_ddl_type OR trigger_ddl_type",
"trigger_type_prefix : BEFORE",
"trigger_type_prefix : AFTER",
"trigger_type_suffix : INSERT",
"trigger_type_suffix : UPDATE",
"trigger_type_suffix : DELETE",
"trigger_type_suffix : INSERT OR UPDATE",
"trigger_type_suffix : INSERT OR DELETE",
"trigger_type_suffix : UPDATE OR INSERT",
"trigger_type_suffix : UPDATE OR DELETE",
"trigger_type_suffix : DELETE OR INSERT",
"trigger_type_suffix : DELETE OR UPDATE",
"trigger_type_suffix : INSERT OR UPDATE OR DELETE",
"trigger_type_suffix : INSERT OR DELETE OR UPDATE",
"trigger_type_suffix : UPDATE OR INSERT OR DELETE",
"trigger_type_suffix : UPDATE OR DELETE OR INSERT",
"trigger_type_suffix : DELETE OR INSERT OR UPDATE",
"trigger_type_suffix : DELETE OR UPDATE OR INSERT",
"trigger_position :",
"trigger_position : POSITION nonneg_short_integer",
"alter : ALTER alter_clause",
"alter : set_generator_clause",
"alter_clause : EXCEPTION alter_exception_clause",
"$$104 :",
"alter_clause : TABLE simple_table_name $$104 $$60 alter_ops",
"alter_clause : VIEW alter_view_clause",
"alter_clause : TRIGGER alter_trigger_clause",
"alter_clause : PROCEDURE alter_procedure_clause",
"alter_clause : PACKAGE alter_package_clause",
"$$105 :",
"alter_clause : DATABASE $$105 $$51 alter_db",
"alter_clause : DOMAIN alter_domain",
"alter_clause : INDEX alter_index_clause",
"alter_clause : EXTERNAL FUNCTION alter_udf_clause",
"alter_clause : FUNCTION alter_function_clause",
"alter_clause : ROLE alter_role_clause",
"alter_clause : USER alter_user_clause",
"alter_clause : CURRENT USER alter_cur_user_clause",
"alter_clause : CHARACTER SET alter_charset_clause",
"alter_clause : GENERATOR alter_sequence_clause",
"alter_clause : SEQUENCE alter_sequence_clause",
"alter_clause : MAPPING $$28 alter_map_clause",
"alter_clause : GLOBAL MAPPING $$29 alter_map_clause",
"alter_clause : EXTERNAL CONNECTIONS POOL alter_eds_conn_pool_clause",
"$$106 :",
"$$107 :",
"alter_domain : keyword_or_column $$106 $$107 alter_domain_ops",
"alter_domain_ops : $$107 alter_domain_op",
"$$108 :",
"alter_domain_ops : alter_domain_ops $$108 alter_domain_op",
"alter_domain_op : SET domain_default",
"alter_domain_op : ADD CONSTRAINT check_constraint",
"alter_domain_op : ADD check_constraint",
"alter_domain_op : DROP DEFAULT",
"alter_domain_op : DROP CONSTRAINT",
"alter_domain_op : DROP NOT NULL",
"alter_domain_op : SET NOT NULL",
"alter_domain_op : TO symbol_column_name",
"alter_domain_op : TYPE non_array_type",
"alter_ops : $$60 alter_op",
"alter_ops : alter_ops ',' $$61 alter_op",
"alter_op : DROP symbol_column_name drop_behaviour",
"alter_op : DROP CONSTRAINT symbol_constraint_name",
"alter_op : ADD $$75 column_def",
"alter_op : ADD $$75 table_constraint_definition",
"alter_op : col_opt alter_column_name POSITION pos_short_integer",
"alter_op : col_opt alter_column_name TO symbol_column_name",
"alter_op : col_opt alter_column_name DROP NOT NULL",
"alter_op : col_opt alter_column_name SET NOT NULL",
"alter_op : col_opt symbol_column_name TYPE alter_data_type_or_domain",
"alter_op : col_opt symbol_column_name TYPE non_array_type def_computed",
"alter_op : col_opt symbol_column_name def_computed",
"alter_op : col_opt symbol_column_name SET domain_default",
"alter_op : col_opt symbol_column_name DROP DEFAULT",
"$$109 :",
"alter_op : col_opt symbol_column_name $$109 $$71 alter_identity_clause_spec",
"alter_op : col_opt symbol_column_name DROP IDENTITY",
"alter_op : ALTER SQL SECURITY DEFINER",
"alter_op : ALTER SQL SECURITY INVOKER",
"alter_op : DROP SQL SECURITY",
"alter_op : ENABLE PUBLICATION",
"alter_op : DISABLE PUBLICATION",
"alter_column_name : keyword_or_column",
"keyword_or_column : valid_symbol_name",
"keyword_or_column : ADMIN",
"keyword_or_column : COLUMN",
"keyword_or_column : EXTRACT",
"keyword_or_column : YEAR",
"keyword_or_column : MONTH",
"keyword_or_column : DAY",
"keyword_or_column : HOUR",
"keyword_or_column : MINUTE",
"keyword_or_column : SECOND",
"keyword_or_column : TIME",
"keyword_or_column : TIMESTAMP",
"keyword_or_column : CURRENT_DATE",
"keyword_or_column : CURRENT_TIME",
"keyword_or_column : CURRENT_TIMESTAMP",
"keyword_or_column : CURRENT_USER",
"keyword_or_column : CURRENT_ROLE",
"keyword_or_column : RECREATE",
"keyword_or_column : CURRENT_CONNECTION",
"keyword_or_column : CURRENT_TRANSACTION",
"keyword_or_column : BIGINT",
"keyword_or_column : CASE",
"keyword_or_column : RELEASE",
"keyword_or_column : ROW_COUNT",
"keyword_or_column : SAVEPOINT",
"keyword_or_column : OPEN",
"keyword_or_column : CLOSE",
"keyword_or_column : FETCH",
"keyword_or_column : ROWS",
"keyword_or_column : USING",
"keyword_or_column : CROSS",
"keyword_or_column : BIT_LENGTH",
"keyword_or_column : BOTH",
"keyword_or_column : CHAR_LENGTH",
"keyword_or_column : CHARACTER_LENGTH",
"keyword_or_column : COMMENT",
"keyword_or_column : LEADING",
"keyword_or_column : LOWER",
"keyword_or_column : OCTET_LENGTH",
"keyword_or_column : TRAILING",
"keyword_or_column : TRIM",
"keyword_or_column : CONNECT",
"keyword_or_column : DISCONNECT",
"keyword_or_column : GLOBAL",
"keyword_or_column : INSENSITIVE",
"keyword_or_column : RECURSIVE",
"keyword_or_column : SENSITIVE",
"keyword_or_column : START",
"keyword_or_column : SIMILAR",
"keyword_or_column : BOOLEAN",
"keyword_or_column : CORR",
"keyword_or_column : COVAR_POP",
"keyword_or_column : COVAR_SAMP",
"keyword_or_column : DELETING",
"keyword_or_column : DETERMINISTIC",
"keyword_or_column : FALSE",
"keyword_or_column : INSERTING",
"keyword_or_column : OFFSET",
"keyword_or_column : OVER",
"keyword_or_column : REGR_AVGX",
"keyword_or_column : REGR_AVGY",
"keyword_or_column : REGR_COUNT",
"keyword_or_column : REGR_INTERCEPT",
"keyword_or_column : REGR_R2",
"keyword_or_column : REGR_SLOPE",
"keyword_or_column : REGR_SXX",
"keyword_or_column : REGR_SXY",
"keyword_or_column : REGR_SYY",
"keyword_or_column : RETURN",
"keyword_or_column : ROW",
"keyword_or_column : SCROLL",
"keyword_or_column : SQLSTATE",
"keyword_or_column : STDDEV_SAMP",
"keyword_or_column : STDDEV_POP",
"keyword_or_column : TRUE",
"keyword_or_column : UNKNOWN",
"keyword_or_column : UPDATING",
"keyword_or_column : VAR_SAMP",
"keyword_or_column : VAR_POP",
"keyword_or_column : BINARY",
"keyword_or_column : DECFLOAT",
"keyword_or_column : INT128",
"keyword_or_column : LATERAL",
"keyword_or_column : LOCAL",
"keyword_or_column : LOCALTIME",
"keyword_or_column : LOCALTIMESTAMP",
"keyword_or_column : PUBLICATION",
"keyword_or_column : RESETTING",
"keyword_or_column : TIMEZONE_HOUR",
"keyword_or_column : TIMEZONE_MINUTE",
"keyword_or_column : UNBOUNDED",
"keyword_or_column : VARBINARY",
"keyword_or_column : WINDOW",
"keyword_or_column : WITHOUT",
"col_opt : ALTER",
"col_opt : ALTER COLUMN",
"alter_data_type_or_domain : non_array_type",
"alter_data_type_or_domain : symbol_column_name",
"$$110 :",
"alter_identity_clause_spec : $$71 alter_identity_clause_generation $$110 alter_identity_clause_options_opt",
"alter_identity_clause_spec : $$71 alter_identity_clause_options",
"alter_identity_clause_generation : SET GENERATED ALWAYS",
"alter_identity_clause_generation : SET GENERATED BY DEFAULT",
"alter_identity_clause_options_opt :",
"alter_identity_clause_options_opt : $$71 alter_identity_clause_options",
"alter_identity_clause_options : alter_identity_clause_options $$72 alter_identity_clause_option",
"alter_identity_clause_options : $$71 alter_identity_clause_option",
"alter_identity_clause_option : RESTART with_opt",
"alter_identity_clause_option : SET INCREMENT by_noise signed_long_integer",
"drop_behaviour :",
"drop_behaviour : RESTRICT",
"drop_behaviour : CASCADE",
"alter_index_clause : symbol_index_name ACTIVE",
"alter_index_clause : symbol_index_name INACTIVE",
"alter_udf_clause : symbol_UDF_name entry_op module_op",
"entry_op :",
"entry_op : ENTRY_POINT utf_string",
"module_op :",
"module_op : MODULE_NAME utf_string",
"alter_role_2X_compatibility : symbol_role_name alter_role_enable AUTO ADMIN MAPPING",
"alter_role_clause : role_clause",
"alter_role_clause : alter_role_2X_compatibility",
"alter_role_enable : SET",
"alter_role_enable : DROP",
"alter_db : $$51 db_alter_clause",
"alter_db : alter_db $$53 db_alter_clause",
"$$111 :",
"db_alter_clause : ADD $$111 db_file_list",
"db_alter_clause : ADD DIFFERENCE FILE utf_string",
"db_alter_clause : DROP DIFFERENCE FILE",
"db_alter_clause : BEGIN BACKUP",
"db_alter_clause : END BACKUP",
"db_alter_clause : SET DEFAULT CHARACTER SET symbol_character_set_name",
"$$112 :",
"db_alter_clause : ENCRYPT WITH valid_symbol_name $$112 crypt_key_clause",
"db_alter_clause : DECRYPT",
"db_alter_clause : SET LINGER TO long_integer",
"db_alter_clause : DROP LINGER",
"db_alter_clause : SET DEFAULT sql_security_clause",
"db_alter_clause : ENABLE PUBLICATION",
"db_alter_clause : DISABLE PUBLICATION",
"db_alter_clause : INCLUDE $$53 pub_table_filter TO PUBLICATION",
"db_alter_clause : EXCLUDE $$53 pub_table_filter FROM PUBLICATION",
"crypt_key_clause :",
"crypt_key_clause : KEY valid_symbol_name",
"pub_table_filter : ALL",
"pub_table_filter : TABLE $$53 pub_table_list",
"pub_table_list : $$51 pub_table_clause",
"pub_table_list : pub_table_list ',' $$52 pub_table_clause",
"pub_table_clause : symbol_table_name",
"alter_trigger_clause : symbol_trigger_name trigger_active trigger_type_opt trigger_position trg_sql_security_clause AS local_declarations_opt full_proc_block",
"alter_trigger_clause : symbol_trigger_name trigger_active trigger_type_opt trigger_position external_clause external_body_clause_opt",
"alter_trigger_clause : symbol_trigger_name trigger_active trigger_type_opt trigger_position trg_sql_security_clause",
"trigger_type_opt : trigger_type_prefix trigger_type_suffix",
"trigger_type_opt :",
"trg_sql_security_clause :",
"trg_sql_security_clause : SQL SECURITY DEFINER",
"trg_sql_security_clause : SQL SECURITY INVOKER",
"trg_sql_security_clause : DROP SQL SECURITY",
"drop : DROP drop_clause",
"drop_clause : EXCEPTION symbol_exception_name",
"drop_clause : INDEX symbol_index_name",
"drop_clause : PROCEDURE symbol_procedure_name",
"drop_clause : TABLE symbol_table_name",
"drop_clause : TRIGGER symbol_trigger_name",
"drop_clause : VIEW symbol_view_name",
"drop_clause : FILTER symbol_filter_name",
"drop_clause : DOMAIN symbol_domain_name",
"drop_clause : EXTERNAL FUNCTION symbol_UDF_name",
"drop_clause : FUNCTION symbol_UDF_name",
"drop_clause : SHADOW pos_short_integer opt_no_file_delete",
"drop_clause : ROLE symbol_role_name",
"drop_clause : GENERATOR symbol_generator_name",
"drop_clause : SEQUENCE symbol_generator_name",
"drop_clause : COLLATION symbol_collation_name",
"drop_clause : USER symbol_user_name USING PLUGIN valid_symbol_name",
"drop_clause : USER symbol_user_name",
"drop_clause : PACKAGE symbol_package_name",
"drop_clause : PACKAGE BODY symbol_package_name",
"drop_clause : MAPPING $$28 drop_map_clause",
"drop_clause : GLOBAL MAPPING $$29 drop_map_clause",
"opt_no_file_delete :",
"opt_no_file_delete : PRESERVE FILE",
"opt_no_file_delete : DELETE FILE",
"data_type : non_array_type",
"data_type : array_type",
"domain_or_non_array_type : domain_or_non_array_type_name",
"domain_or_non_array_type : domain_or_non_array_type_name NOT NULL",
"domain_or_non_array_type_name : non_array_type",
"domain_or_non_array_type_name : domain_type",
"domain_type : TYPE OF symbol_column_name",
"domain_type : TYPE OF COLUMN symbol_column_name '.' symbol_column_name",
"domain_type : symbol_column_name",
"non_array_type : simple_type",
"non_array_type : blob_type",
"array_type : non_charset_simple_type '[' array_spec ']'",
"array_type : character_type '[' array_spec ']' charset_clause",
"array_spec : array_range",
"array_spec : array_spec ',' array_range",
"array_range : signed_long_integer",
"array_range : signed_long_integer ':' signed_long_integer",
"simple_type : non_charset_simple_type",
"simple_type : character_type charset_clause",
"non_charset_simple_type : national_character_type",
"non_charset_simple_type : binary_character_type",
"non_charset_simple_type : numeric_type",
"non_charset_simple_type : float_type",
"non_charset_simple_type : decfloat_type",
"non_charset_simple_type : BIGINT",
"non_charset_simple_type : INT128",
"non_charset_simple_type : integer_keyword",
"non_charset_simple_type : SMALLINT",
"non_charset_simple_type : DATE",
"non_charset_simple_type : TIME without_time_zone_opt",
"non_charset_simple_type : TIME WITH TIME ZONE",
"non_charset_simple_type : TIMESTAMP without_time_zone_opt",
"non_charset_simple_type : TIMESTAMP WITH TIME ZONE",
"non_charset_simple_type : BOOLEAN",
"integer_keyword : INTEGER",
"integer_keyword : INT",
"without_time_zone_opt :",
"without_time_zone_opt : WITHOUT TIME ZONE",
"$$113 :",
"$$114 :",
"blob_type : BLOB $$113 $$114 blob_subtype blob_segsize charset_clause",
"blob_type : BLOB '(' unsigned_short_integer ')'",
"blob_type : BLOB '(' unsigned_short_integer ',' signed_short_integer ')'",
"blob_type : BLOB '(' ',' signed_short_integer ')'",
"blob_segsize :",
"blob_segsize : SEGMENT SIZE unsigned_short_integer",
"blob_subtype :",
"blob_subtype : SUB_TYPE signed_short_integer",
"blob_subtype : SUB_TYPE symbol_blob_subtype_name",
"charset_clause :",
"charset_clause : CHARACTER SET symbol_character_set_name",
"national_character_type : national_character_keyword '(' pos_short_integer ')'",
"national_character_type : national_character_keyword",
"national_character_type : national_character_keyword VARYING '(' pos_short_integer ')'",
"binary_character_type : binary_character_keyword '(' pos_short_integer ')'",
"binary_character_type : binary_character_keyword",
"binary_character_type : varbinary_character_keyword '(' pos_short_integer ')'",
"character_type : character_keyword '(' pos_short_integer ')'",
"character_type : character_keyword",
"character_type : varying_keyword '(' pos_short_integer ')'",
"varying_keyword : VARCHAR",
"varying_keyword : CHARACTER VARYING",
"varying_keyword : CHAR VARYING",
"character_keyword : CHARACTER",
"character_keyword : CHAR",
"national_character_keyword : NCHAR",
"national_character_keyword : NATIONAL CHARACTER",
"national_character_keyword : NATIONAL CHAR",
"binary_character_keyword : BINARY",
"varbinary_character_keyword : VARBINARY",
"varbinary_character_keyword : BINARY VARYING",
"decfloat_type : DECFLOAT precision_opt_nz",
"numeric_type : NUMERIC prec_scale",
"numeric_type : decimal_keyword prec_scale",
"prec_scale :",
"prec_scale : '(' signed_long_integer ')'",
"prec_scale : '(' signed_long_integer ',' signed_long_integer ')'",
"decimal_keyword : DECIMAL",
"decimal_keyword : DEC",
"float_type : FLOAT precision_opt_nz",
"float_type : LONG FLOAT precision_opt_nz",
"float_type : REAL",
"float_type : DOUBLE PRECISION",
"precision_opt_nz :",
"precision_opt_nz : '(' pos_short_integer ')'",
"savepoint : set_savepoint",
"savepoint : release_savepoint",
"savepoint : undo_savepoint",
"set_savepoint : SAVEPOINT symbol_savepoint_name",
"release_savepoint : RELEASE SAVEPOINT symbol_savepoint_name release_only_opt",
"release_only_opt :",
"release_only_opt : ONLY",
"undo_savepoint : ROLLBACK optional_work TO optional_savepoint symbol_savepoint_name",
"optional_savepoint :",
"optional_savepoint : SAVEPOINT",
"commit : COMMIT optional_work optional_retain",
"rollback : ROLLBACK optional_work optional_retain",
"optional_work :",
"optional_work : WORK",
"optional_retain :",
"optional_retain : RETAIN opt_snapshot",
"opt_snapshot :",
"opt_snapshot : SNAPSHOT",
"$$115 :",
"$$116 :",
"set_transaction : SET TRANSACTION $$115 $$116 tran_option_list_opt",
"session_reset : ALTER SESSION RESET",
"set_role : SET ROLE valid_symbol_name",
"set_role : SET TRUSTED ROLE",
"set_decfloat_round : SET DECFLOAT ROUND valid_symbol_name",
"$$117 :",
"$$118 :",
"set_decfloat_traps : SET DECFLOAT TRAPS TO $$117 $$118 decfloat_traps_list_opt",
"set_bind : SET BIND OF set_bind_from TO set_bind_to",
"set_bind_from : bind_type",
"set_bind_from : TIME ZONE",
"bind_type : non_array_type",
"bind_type : varying_keyword",
"set_bind_to : bind_type",
"set_bind_to : LEGACY",
"set_bind_to : NATIVE",
"set_bind_to : EXTENDED",
"set_bind_to : EXTENDED TIME WITH TIME ZONE",
"set_bind_to : EXTENDED TIMESTAMP WITH TIME ZONE",
"decfloat_traps_list_opt :",
"decfloat_traps_list_opt : $$118 decfloat_traps_list",
"decfloat_traps_list : $$118 decfloat_trap",
"$$119 :",
"decfloat_traps_list : decfloat_traps_list ',' $$119 decfloat_trap",
"decfloat_trap : valid_symbol_name",
"session_statement : SET SESSION IDLE TIMEOUT long_integer timepart_sesion_idle_tout",
"session_statement : SET STATEMENT TIMEOUT long_integer timepart_ses_stmt_tout",
"timepart_sesion_idle_tout :",
"timepart_sesion_idle_tout : HOUR",
"timepart_sesion_idle_tout : MINUTE",
"timepart_sesion_idle_tout : SECOND",
"timepart_ses_stmt_tout :",
"timepart_ses_stmt_tout : HOUR",
"timepart_ses_stmt_tout : MINUTE",
"timepart_ses_stmt_tout : SECOND",
"timepart_ses_stmt_tout : MILLISECOND",
"set_time_zone : SET TIME ZONE set_time_zone_option",
"set_time_zone_option : sql_string",
"set_time_zone_option : LOCAL",
"tran_option_list_opt :",
"tran_option_list_opt : $$116 tran_option_list",
"tran_option_list : $$116 tran_option",
"$$120 :",
"tran_option_list : tran_option_list $$120 tran_option",
"tran_option : READ ONLY",
"tran_option : READ WRITE",
"tran_option : WAIT",
"tran_option : NO WAIT",
"tran_option : $$116 isolation_mode",
"tran_option : NO AUTO UNDO",
"tran_option : IGNORE LIMBO",
"tran_option : RESTART REQUESTS",
"tran_option : AUTO COMMIT",
"tran_option : LOCK TIMEOUT nonneg_short_integer",
"$$121 :",
"$$122 :",
"tran_option : RESERVING $$121 $$122 restr_list",
"isolation_mode : ISOLATION LEVEL $$122 iso_mode",
"isolation_mode : iso_mode",
"iso_mode : $$116 snap_shot",
"iso_mode : READ UNCOMMITTED version_mode",
"iso_mode : READ COMMITTED version_mode",
"snap_shot : SNAPSHOT",
"snap_shot : SNAPSHOT AT NUMBER snapshot_number",
"snap_shot : SNAPSHOT TABLE",
"snap_shot : SNAPSHOT TABLE STABILITY",
"snapshot_number : NUMBER32BIT",
"snapshot_number : NUMBER64BIT",
"version_mode :",
"version_mode : VERSION",
"version_mode : NO VERSION",
"version_mode : READ CONSISTENCY",
"lock_type :",
"lock_type : SHARED",
"lock_type : PROTECTED",
"lock_mode : READ",
"lock_mode : WRITE",
"restr_list : restr_option",
"restr_list : restr_list ',' restr_option",
"restr_option : table_list table_lock",
"table_lock :",
"table_lock : FOR lock_type lock_mode",
"table_list : symbol_table_name",
"table_list : table_list ',' symbol_table_name",
"set_statistics : SET STATISTICS INDEX symbol_index_name",
"comment : COMMENT ON ddl_type0 IS ddl_desc",
"comment : COMMENT ON ddl_type1 symbol_ddl_name IS ddl_desc",
"comment : COMMENT ON ddl_type2 symbol_ddl_name ddl_subname IS ddl_desc",
"comment : COMMENT ON ddl_type3 ddl_qualified_name ddl_subname IS ddl_desc",
"comment : COMMENT ON ddl_type4 ddl_qualified_name IS ddl_desc",
"comment : comment_on_user",
"$$123 :",
"$$124 :",
"comment_on_user : COMMENT ON USER symbol_user_name $$123 $$124 opt_use_plugin IS ddl_desc",
"opt_use_plugin :",
"opt_use_plugin : $$124 use_plugin",
"ddl_type0 : DATABASE",
"ddl_type1 : DOMAIN",
"ddl_type1 : TABLE",
"ddl_type1 : VIEW",
"ddl_type1 : TRIGGER",
"ddl_type1 : FILTER",
"ddl_type1 : EXCEPTION",
"ddl_type1 : GENERATOR",
"ddl_type1 : SEQUENCE",
"ddl_type1 : INDEX",
"ddl_type1 : ROLE",
"ddl_type1 : CHARACTER SET",
"ddl_type1 : COLLATION",
"ddl_type1 : PACKAGE",
"ddl_type2 : COLUMN",
"ddl_type3 : PARAMETER",
"ddl_type3 : PROCEDURE PARAMETER",
"ddl_type3 : FUNCTION PARAMETER",
"ddl_type4 : PROCEDURE",
"ddl_type4 : EXTERNAL FUNCTION",
"ddl_type4 : FUNCTION",
"ddl_subname : '.' symbol_ddl_name",
"ddl_qualified_name : symbol_ddl_name",
"ddl_qualified_name : symbol_ddl_name '.' symbol_ddl_name",
"ddl_desc : utf_string",
"ddl_desc : NULL",
"select : select_expr for_update_clause lock_clause",
"for_update_clause :",
"for_update_clause : FOR UPDATE for_update_list",
"for_update_list :",
"for_update_list : OF column_list",
"lock_clause :",
"lock_clause : WITH LOCK",
"select_expr : with_clause select_expr_body order_clause_opt rows_clause",
"select_expr : with_clause select_expr_body order_clause_opt result_offset_clause fetch_first_clause",
"with_clause :",
"with_clause : WITH RECURSIVE with_list",
"with_clause : WITH with_list",
"with_list : with_item",
"with_list : with_list ',' with_item",
"with_item : symbol_table_alias_name derived_column_list AS '(' select_expr ')'",
"column_select : select_expr",
"column_singleton : column_select",
"select_expr_body : query_term",
"select_expr_body : select_expr_body UNION distinct_noise query_term",
"select_expr_body : select_expr_body UNION ALL query_term",
"query_term : query_spec",
"query_spec : SELECT limit_clause distinct_clause select_list from_clause where_clause group_clause having_clause named_windows_clause plan_clause",
"limit_clause :",
"limit_clause : first_clause skip_clause",
"limit_clause : first_clause",
"limit_clause : skip_clause",
"first_clause : FIRST long_integer",
"first_clause : FIRST '(' value ')'",
"first_clause : FIRST parameter",
"skip_clause : SKIP long_integer",
"skip_clause : SKIP '(' value ')'",
"skip_clause : SKIP parameter",
"distinct_clause : DISTINCT",
"distinct_clause : all_noise",
"select_list : select_items",
"select_list : '*'",
"select_items : select_item",
"select_items : select_items ',' select_item",
"select_item : value_opt_alias",
"select_item : symbol_table_alias_name '.' '*'",
"value_opt_alias : value",
"value_opt_alias : value as_noise symbol_item_alias_name",
"as_noise :",
"as_noise : AS",
"from_clause : FROM from_list",
"from_list : table_reference",
"from_list : from_list ',' table_reference",
"table_reference : joined_table",
"table_reference : table_primary",
"table_primary : table_proc",
"table_primary : derived_table",
"table_primary : lateral_derived_table",
"table_primary : '(' joined_table ')'",
"derived_table : '(' select_expr ')' as_noise correlation_name derived_column_list",
"lateral_derived_table : LATERAL derived_table",
"correlation_name :",
"correlation_name : symbol_table_alias_name",
"derived_column_list :",
"derived_column_list : '(' alias_list ')'",
"alias_list : symbol_item_alias_name",
"alias_list : alias_list ',' symbol_item_alias_name",
"joined_table : cross_join",
"joined_table : natural_join",
"joined_table : qualified_join",
"cross_join : table_reference CROSS JOIN table_primary",
"natural_join : table_reference NATURAL join_type JOIN table_primary",
"qualified_join : table_reference join_type JOIN table_reference join_condition",
"qualified_join : table_reference join_type JOIN table_reference named_columns_join",
"join_condition : ON search_condition",
"named_columns_join : USING '(' column_list ')'",
"table_proc : symbol_procedure_name table_proc_inputs as_noise symbol_table_alias_name",
"table_proc : symbol_procedure_name table_proc_inputs",
"table_proc : symbol_package_name '.' symbol_procedure_name table_proc_inputs as_noise symbol_table_alias_name",
"table_proc : symbol_package_name '.' symbol_procedure_name table_proc_inputs",
"table_proc_inputs :",
"table_proc_inputs : '(' value_list ')'",
"table_name : simple_table_name",
"table_name : symbol_table_name as_noise symbol_table_alias_name",
"simple_table_name : symbol_table_name",
"join_type :",
"join_type : INNER",
"join_type : LEFT outer_noise",
"join_type : RIGHT outer_noise",
"join_type : FULL outer_noise",
"outer_noise :",
"outer_noise : OUTER",
"group_clause :",
"group_clause : GROUP BY group_by_list",
"group_by_list : group_by_item",
"group_by_list : group_by_list ',' group_by_item",
"group_by_item : value",
"having_clause :",
"having_clause : HAVING search_condition",
"where_clause :",
"where_clause : WHERE search_condition",
"named_windows_clause :",
"named_windows_clause : WINDOW window_definition_list",
"window_definition_list : window_definition",
"window_definition_list : window_definition_list ',' window_definition",
"window_definition : symbol_window_name AS '(' window_clause ')'",
"symbol_window_name_opt :",
"symbol_window_name_opt : symbol_window_name",
"plan_clause :",
"plan_clause : PLAN plan_expression",
"$$125 :",
"$$126 :",
"plan_expression : plan_type $$125 '(' $$126 plan_item_list ')'",
"plan_type :",
"plan_type : JOIN",
"plan_type : SORT MERGE",
"plan_type : MERGE",
"plan_type : HASH",
"plan_type : SORT",
"plan_item_list : plan_item",
"plan_item_list : plan_item_list ',' plan_item",
"plan_item : table_or_alias_list access_type",
"plan_item : plan_expression",
"table_or_alias_list : symbol_table_name",
"table_or_alias_list : table_or_alias_list symbol_table_name",
"access_type : NATURAL",
"$$127 :",
"$$128 :",
"access_type : INDEX $$127 '(' $$128 index_list ')'",
"$$129 :",
"access_type : ORDER $$129 symbol_index_name $$128 extra_indices_opt",
"index_list : symbol_index_name",
"index_list : index_list ',' symbol_index_name",
"extra_indices_opt :",
"$$130 :",
"extra_indices_opt : INDEX '(' $$130 index_list ')'",
"order_clause_opt :",
"order_clause_opt : order_clause",
"order_clause : ORDER BY order_list",
"order_list : order_item",
"order_list : order_list ',' order_item",
"order_item : value order_direction nulls_clause",
"order_direction :",
"order_direction : ASC",
"order_direction : DESC",
"nulls_clause :",
"nulls_clause : NULLS nulls_placement",
"nulls_placement : FIRST",
"nulls_placement : LAST",
"rows_clause : ROWS value",
"rows_clause : ROWS value TO value",
"rows_clause_optional :",
"rows_clause_optional : rows_clause",
"row_noise : ROW",
"row_noise : ROWS",
"result_offset_clause :",
"result_offset_clause : OFFSET simple_value_spec row_noise",
"first_next_noise : FIRST",
"first_next_noise : NEXT",
"fetch_first_clause :",
"fetch_first_clause : FETCH first_next_noise simple_value_spec row_noise ONLY",
"fetch_first_clause : FETCH first_next_noise row_noise ONLY",
"$$131 :",
"insert : insert_start $$131 ins_column_parens_opt override_opt VALUES '(' value_or_default_list ')' returning_clause",
"insert : insert_start $$131 ins_column_parens_opt override_opt select_expr returning_clause",
"insert : insert_start DEFAULT VALUES returning_clause",
"insert_start : INSERT INTO simple_table_name",
"override_opt :",
"override_opt : OVERRIDING USER VALUE",
"override_opt : OVERRIDING SYSTEM VALUE",
"value_or_default_list : value_or_default",
"value_or_default_list : value_or_default_list ',' value_or_default",
"value_or_default : value",
"value_or_default : DEFAULT",
"$$132 :",
"$$133 :",
"merge : MERGE INTO table_name USING table_reference ON search_condition $$132 $$133 merge_when_clause returning_clause",
"merge_when_clause : $$133 merge_when_matched_clause",
"merge_when_clause : $$133 merge_when_not_matched_clause",
"$$134 :",
"merge_when_clause : merge_when_clause $$134 merge_when_matched_clause",
"merge_when_clause : merge_when_clause $$134 merge_when_not_matched_clause",
"$$135 :",
"$$136 :",
"$$137 :",
"merge_when_matched_clause : WHEN MATCHED $$135 $$136 $$137 merge_update_specification",
"$$138 :",
"$$139 :",
"merge_when_not_matched_clause : WHEN NOT MATCHED $$138 $$139 merge_insert_specification",
"$$140 :",
"merge_update_specification : THEN UPDATE SET $$140 update_assignments",
"$$141 :",
"merge_update_specification : AND search_condition THEN UPDATE SET $$141 update_assignments",
"merge_update_specification : THEN DELETE",
"merge_update_specification : AND search_condition THEN DELETE",
"$$142 :",
"merge_insert_specification : THEN INSERT $$142 ins_column_parens_opt override_opt VALUES '(' value_or_default_list ')'",
"$$143 :",
"merge_insert_specification : AND search_condition THEN INSERT $$143 ins_column_parens_opt override_opt VALUES '(' value_or_default_list ')'",
"delete : delete_searched",
"delete : delete_positioned",
"delete_searched : DELETE FROM table_name where_clause plan_clause order_clause_opt rows_clause_optional returning_clause",
"delete_positioned : DELETE FROM table_name cursor_clause returning_clause",
"update : update_searched",
"update : update_positioned",
"$$144 :",
"update_searched : UPDATE table_name SET $$144 update_assignments where_clause plan_clause order_clause_opt rows_clause_optional returning_clause",
"update_positioned : UPDATE table_name SET $$144 update_assignments cursor_clause returning_clause",
"$$145 :",
"$$146 :",
"$$147 :",
"update_or_insert : UPDATE OR INSERT INTO simple_table_name $$145 $$146 ins_column_parens_opt override_opt VALUES '(' value_or_default_list ')' $$147 update_or_insert_matching_opt returning_clause",
"update_or_insert_matching_opt :",
"$$148 :",
"update_or_insert_matching_opt : MATCHING $$148 ins_column_parens",
"returning_clause :",
"returning_clause : RETURNING select_list",
"returning_clause : RETURNING select_list INTO variable_list",
"cursor_clause : WHERE CURRENT OF symbol_cursor_name",
"assignment : update_column_name '=' value",
"$$149 :",
"update_assignments : $$149 update_assignment",
"$$150 :",
"update_assignments : update_assignments ',' $$150 update_assignment",
"update_assignment : update_column_name '=' value",
"update_assignment : update_column_name '=' DEFAULT",
"exec_function : udf",
"exec_function : non_aggregate_function",
"column_parens_opt :",
"column_parens_opt : column_parens",
"column_parens : '(' column_list ')'",
"column_list : simple_column_name",
"column_list : column_list ',' simple_column_name",
"ins_column_parens_opt :",
"$$151 :",
"ins_column_parens_opt : $$151 ins_column_parens",
"ins_column_parens : '(' $$148 ins_column_list ')'",
"ins_column_list : update_column_name",
"ins_column_list : ins_column_list ',' update_column_name",
"column_name : simple_column_name",
"column_name : symbol_table_alias_name '.' symbol_column_name",
"column_name : ':' symbol_table_alias_name '.' symbol_column_name",
"simple_column_name : symbol_column_name",
"update_column_name : simple_column_name",
"update_column_name : symbol_table_alias_name '.' symbol_column_name",
"search_condition : value",
"boolean_value_expression : predicate",
"boolean_value_expression : value OR value",
"boolean_value_expression : value AND value",
"boolean_value_expression : NOT value",
"boolean_value_expression : '(' boolean_value_expression ')'",
"boolean_value_expression : value IS boolean_literal",
"boolean_value_expression : value IS NOT boolean_literal",
"predicate : comparison_predicate",
"predicate : distinct_predicate",
"predicate : between_predicate",
"predicate : binary_pattern_predicate",
"predicate : ternary_pattern_predicate",
"predicate : in_predicate",
"predicate : null_predicate",
"predicate : quantified_predicate",
"predicate : exists_predicate",
"predicate : singular_predicate",
"predicate : trigger_action_predicate",
"predicate : session_reset_predicate",
"comparison_predicate : value comparison_operator value",
"comparison_operator : '='",
"comparison_operator : '<'",
"comparison_operator : '>'",
"comparison_operator : GEQ",
"comparison_operator : LEQ",
"comparison_operator : NOT_GTR",
"comparison_operator : NOT_LSS",
"comparison_operator : NEQ",
"quantified_predicate : value comparison_operator quantified_flag '(' column_select ')'",
"quantified_flag : ALL",
"quantified_flag : SOME",
"quantified_flag : ANY",
"distinct_predicate : value IS DISTINCT FROM value",
"distinct_predicate : value IS NOT DISTINCT FROM value",
"between_predicate : value BETWEEN value_special AND value_special",
"between_predicate : value NOT BETWEEN value_special AND value_special",
"binary_pattern_predicate : value binary_pattern_operator value",
"binary_pattern_predicate : value NOT binary_pattern_operator value",
"binary_pattern_operator : CONTAINING",
"binary_pattern_operator : STARTING",
"binary_pattern_operator : STARTING WITH",
"ternary_pattern_predicate : value LIKE value",
"ternary_pattern_predicate : value LIKE value ESCAPE value",
"ternary_pattern_predicate : value NOT LIKE value",
"ternary_pattern_predicate : value NOT LIKE value ESCAPE value",
"ternary_pattern_predicate : value SIMILAR TO value",
"ternary_pattern_predicate : value SIMILAR TO value ESCAPE value",
"ternary_pattern_predicate : value NOT SIMILAR TO value",
"ternary_pattern_predicate : value NOT SIMILAR TO value ESCAPE value",
"in_predicate : value IN in_predicate_value",
"in_predicate : value NOT IN in_predicate_value",
"exists_predicate : EXISTS '(' select_expr ')'",
"singular_predicate : SINGULAR '(' select_expr ')'",
"trigger_action_predicate : INSERTING",
"trigger_action_predicate : UPDATING",
"trigger_action_predicate : DELETING",
"session_reset_predicate : RESETTING",
"null_predicate : value IS NULL",
"null_predicate : value IS UNKNOWN",
"null_predicate : value IS NOT NULL",
"null_predicate : value IS NOT UNKNOWN",
"in_predicate_value : table_subquery",
"in_predicate_value : '(' value_list ')'",
"table_subquery : '(' column_select ')'",
"$$152 :",
"create_user_clause : symbol_user_name $$152 $$124 user_fixed_list_opt",
"$$153 :",
"alter_user_clause : symbol_user_name set_noise $$153 $$124 user_fixed_list_opt",
"$$154 :",
"alter_cur_user_clause : set_noise $$154 $$124 user_fixed_list_opt",
"$$155 :",
"replace_user_clause : symbol_user_name set_noise $$155 $$124 user_fixed_list_opt",
"set_noise :",
"set_noise : SET",
"user_fixed_list_opt :",
"user_fixed_list_opt : $$124 user_fixed_list",
"user_fixed_list : $$124 user_fixed_option",
"$$156 :",
"user_fixed_list : user_fixed_list $$156 user_fixed_option",
"user_fixed_option : FIRSTNAME utf_string",
"user_fixed_option : MIDDLENAME utf_string",
"user_fixed_option : LASTNAME utf_string",
"user_fixed_option : PASSWORD utf_string",
"user_fixed_option : GRANT ADMIN ROLE",
"user_fixed_option : REVOKE ADMIN ROLE",
"user_fixed_option : ACTIVE",
"user_fixed_option : INACTIVE",
"user_fixed_option : $$124 use_plugin",
"$$157 :",
"user_fixed_option : TAGS '(' $$157 user_var_list ')'",
"use_plugin : USING PLUGIN valid_symbol_name",
"user_var_list : $$124 user_var_option",
"user_var_list : user_var_list ',' $$157 user_var_option",
"user_var_option : valid_symbol_name '=' utf_string",
"user_var_option : DROP valid_symbol_name",
"$$158 :",
"$$159 :",
"$$160 :",
"create_map_clause : $$158 map_clause $$159 $$160 map_to",
"$$161 :",
"$$162 :",
"alter_map_clause : $$161 map_clause $$162 $$160 map_to",
"$$163 :",
"$$164 :",
"replace_map_clause : $$163 map_clause $$164 $$160 map_to",
"drop_map_clause : map_name",
"$$165 :",
"$$166 :",
"$$167 :",
"map_clause : map_name $$165 USING $$166 map_using FROM $$167 map_from",
"map_name : valid_symbol_name",
"map_from : map_from_symbol_name map_logoninfo",
"map_from : ANY map_from_symbol_name",
"map_from_symbol_name : valid_symbol_name",
"map_from_symbol_name : USER",
"map_from_symbol_name : GROUP",
"map_logoninfo : sql_string",
"map_logoninfo : valid_symbol_name",
"map_using : PLUGIN valid_symbol_name map_in",
"map_using : ANY PLUGIN map_in",
"map_using : ANY PLUGIN SERVERWIDE",
"map_using : MAPPING map_in",
"map_using : '*' map_in",
"map_in :",
"map_in : IN valid_symbol_name",
"map_to : TO map_role valid_symbol_name",
"map_to : TO map_role",
"map_role : ROLE",
"map_role : USER",
"value : value_primary",
"value : boolean_value_expression",
"value_special : value_primary",
"value_special : '(' boolean_value_expression ')'",
"value_primary : nonparenthesized_value",
"value_primary : '(' value_primary ')'",
"simple_value_spec : constant",
"simple_value_spec : variable",
"simple_value_spec : parameter",
"nonparenthesized_value : column_name",
"nonparenthesized_value : array_element",
"nonparenthesized_value : function",
"nonparenthesized_value : u_constant",
"nonparenthesized_value : boolean_literal",
"nonparenthesized_value : parameter",
"nonparenthesized_value : variable",
"nonparenthesized_value : cast_specification",
"nonparenthesized_value : case_expression",
"nonparenthesized_value : next_value_expression",
"nonparenthesized_value : udf",
"nonparenthesized_value : '-' value_special",
"nonparenthesized_value : '+' value_special",
"nonparenthesized_value : value_special '+' value_special",
"nonparenthesized_value : value_special CONCATENATE value_special",
"nonparenthesized_value : value_special COLLATE symbol_collation_name",
"nonparenthesized_value : value_special AT LOCAL",
"nonparenthesized_value : value_special AT TIME ZONE value_special",
"nonparenthesized_value : value_special '-' value_special",
"nonparenthesized_value : value_special '*' value_special",
"nonparenthesized_value : value_special '/' value_special",
"nonparenthesized_value : '(' column_singleton ')'",
"nonparenthesized_value : current_user",
"nonparenthesized_value : current_role",
"nonparenthesized_value : internal_info",
"nonparenthesized_value : recordKeyType",
"nonparenthesized_value : symbol_table_alias_name '.' recordKeyType",
"nonparenthesized_value : VALUE",
"nonparenthesized_value : datetime_value_expression",
"nonparenthesized_value : null_value",
"recordKeyType : DB_KEY",
"recordKeyType : RDB_RECORD_VERSION",
"datetime_value_expression : CURRENT_DATE",
"datetime_value_expression : LOCALTIME time_precision_opt",
"datetime_value_expression : CURRENT_TIME time_precision_opt",
"datetime_value_expression : LOCALTIMESTAMP timestamp_precision_opt",
"datetime_value_expression : CURRENT_TIMESTAMP timestamp_precision_opt",
"time_precision_opt :",
"time_precision_opt : '(' nonneg_short_integer ')'",
"timestamp_precision_opt :",
"timestamp_precision_opt : '(' nonneg_short_integer ')'",
"array_element : column_name '[' value_list ']'",
"value_list_opt :",
"value_list_opt : value_list",
"value_list : value",
"value_list : value_list ',' value",
"constant : u_constant",
"constant : '-' ul_numeric_constant",
"constant : '-' LIMIT64_INT",
"constant : '-' LIMIT64_NUMBER",
"constant : '-' u_constant_128",
"constant : boolean_literal",
"u_numeric_constant : ul_numeric_constant",
"u_numeric_constant : LIMIT64_NUMBER",
"u_numeric_constant : LIMIT64_INT",
"u_numeric_constant : u_constant_128",
"u_constant_128 : NUM128",
"ul_numeric_constant : NUMBER32BIT",
"ul_numeric_constant : FLOAT_NUMBER",
"ul_numeric_constant : DECIMAL_NUMBER",
"ul_numeric_constant : NUMBER64BIT",
"ul_numeric_constant : SCALEDINT",
"u_constant : u_numeric_constant",
"u_constant : sql_string",
"u_constant : DATE STRING",
"u_constant : TIME STRING",
"u_constant : TIMESTAMP STRING",
"boolean_literal : FALSE",
"boolean_literal : TRUE",
"parameter : '?'",
"current_user : USER",
"current_user : CURRENT_USER",
"current_role : CURRENT_ROLE",
"internal_info : CURRENT_CONNECTION",
"internal_info : CURRENT_TRANSACTION",
"internal_info : GDSCODE",
"internal_info : SQLCODE",
"internal_info : SQLSTATE",
"internal_info : ROW_COUNT",
"internal_info : RDB_ERROR '(' error_context ')'",
"error_context : GDSCODE",
"error_context : SQLCODE",
"error_context : SQLSTATE",
"error_context : EXCEPTION",
"error_context : MESSAGE",
"sql_string : STRING",
"sql_string : INTRODUCER STRING",
"utf_string : sql_string",
"signed_short_integer : nonneg_short_integer",
"signed_short_integer : '-' neg_short_integer",
"nonneg_short_integer : NUMBER32BIT",
"neg_short_integer : NUMBER32BIT",
"pos_short_integer : nonneg_short_integer",
"unsigned_short_integer : NUMBER32BIT",
"signed_long_integer : long_integer",
"signed_long_integer : '-' long_integer",
"long_integer : NUMBER32BIT",
"function : aggregate_function",
"function : non_aggregate_function",
"function : over_clause",
"non_aggregate_function : numeric_value_function",
"non_aggregate_function : string_value_function",
"non_aggregate_function : system_function_expression",
"aggregate_function : aggregate_function_prefix",
"aggregate_function : aggregate_function_prefix FILTER '(' WHERE search_condition ')'",
"aggregate_function_prefix : COUNT '(' '*' ')'",
"aggregate_function_prefix : COUNT '(' all_noise value ')'",
"aggregate_function_prefix : COUNT '(' DISTINCT value ')'",
"aggregate_function_prefix : SUM '(' all_noise value ')'",
"aggregate_function_prefix : SUM '(' DISTINCT value ')'",
"aggregate_function_prefix : AVG '(' all_noise value ')'",
"aggregate_function_prefix : AVG '(' DISTINCT value ')'",
"aggregate_function_prefix : MINIMUM '(' all_noise value ')'",
"aggregate_function_prefix : MINIMUM '(' DISTINCT value ')'",
"aggregate_function_prefix : MAXIMUM '(' all_noise value ')'",
"aggregate_function_prefix : MAXIMUM '(' DISTINCT value ')'",
"aggregate_function_prefix : LIST '(' all_noise value delimiter_opt ')'",
"aggregate_function_prefix : LIST '(' DISTINCT value delimiter_opt ')'",
"aggregate_function_prefix : STDDEV_SAMP '(' value ')'",
"aggregate_function_prefix : STDDEV_POP '(' value ')'",
"aggregate_function_prefix : VAR_SAMP '(' value ')'",
"aggregate_function_prefix : VAR_POP '(' value ')'",
"aggregate_function_prefix : COVAR_SAMP '(' value ',' value ')'",
"aggregate_function_prefix : COVAR_POP '(' value ',' value ')'",
"aggregate_function_prefix : CORR '(' value ',' value ')'",
"aggregate_function_prefix : REGR_AVGX '(' value ',' value ')'",
"aggregate_function_prefix : REGR_AVGY '(' value ',' value ')'",
"aggregate_function_prefix : REGR_COUNT '(' value ',' value ')'",
"aggregate_function_prefix : REGR_INTERCEPT '(' value ',' value ')'",
"aggregate_function_prefix : REGR_R2 '(' value ',' value ')'",
"aggregate_function_prefix : REGR_SLOPE '(' value ',' value ')'",
"aggregate_function_prefix : REGR_SXX '(' value ',' value ')'",
"aggregate_function_prefix : REGR_SXY '(' value ',' value ')'",
"aggregate_function_prefix : REGR_SYY '(' value ',' value ')'",
"window_function : DENSE_RANK '(' ')'",
"window_function : RANK '(' ')'",
"window_function : PERCENT_RANK '(' ')'",
"window_function : CUME_DIST '(' ')'",
"window_function : ROW_NUMBER '(' ')'",
"window_function : FIRST_VALUE '(' value ')'",
"window_function : LAST_VALUE '(' value ')'",
"window_function : NTH_VALUE '(' value ',' value ')' nth_from",
"window_function : LAG '(' value ',' value ',' value ')'",
"window_function : LAG '(' value ',' value ')'",
"window_function : LAG '(' value ')'",
"window_function : LEAD '(' value ',' value ',' value ')'",
"window_function : LEAD '(' value ',' value ')'",
"window_function : LEAD '(' value ')'",
"window_function : NTILE '(' ntile_arg ')'",
"nth_from :",
"nth_from : FROM FIRST",
"nth_from : FROM LAST",
"ntile_arg : u_numeric_constant",
"ntile_arg : variable",
"ntile_arg : parameter",
"aggregate_window_function : aggregate_function",
"aggregate_window_function : window_function",
"over_clause : aggregate_window_function OVER symbol_window_name",
"over_clause : aggregate_window_function OVER '(' window_clause ')'",
"window_clause : symbol_window_name_opt window_partition_opt order_clause_opt window_frame_extent window_frame_exclusion_opt",
"window_partition_opt :",
"window_partition_opt : PARTITION BY value_list",
"window_frame_extent :",
"$$168 :",
"$$169 :",
"window_frame_extent : RANGE $$168 $$169 window_frame",
"$$170 :",
"window_frame_extent : ROWS $$170 $$169 window_frame",
"window_frame : window_frame_start",
"window_frame : BETWEEN window_frame_between_bound1 AND window_frame_between_bound2",
"window_frame_start : UNBOUNDED PRECEDING",
"window_frame_start : CURRENT ROW",
"window_frame_start : value PRECEDING",
"window_frame_between_bound1 : UNBOUNDED PRECEDING",
"window_frame_between_bound1 : CURRENT ROW",
"window_frame_between_bound1 : value PRECEDING",
"window_frame_between_bound1 : value FOLLOWING",
"window_frame_between_bound2 : UNBOUNDED FOLLOWING",
"window_frame_between_bound2 : CURRENT ROW",
"window_frame_between_bound2 : value PRECEDING",
"window_frame_between_bound2 : value FOLLOWING",
"window_frame_exclusion_opt :",
"window_frame_exclusion_opt : EXCLUDE NO OTHERS",
"window_frame_exclusion_opt : EXCLUDE CURRENT ROW",
"window_frame_exclusion_opt : EXCLUDE GROUP",
"window_frame_exclusion_opt : EXCLUDE TIES",
"delimiter_opt :",
"delimiter_opt : ',' value",
"numeric_value_function : extract_expression",
"numeric_value_function : length_expression",
"extract_expression : EXTRACT '(' timestamp_part FROM value ')'",
"length_expression : bit_length_expression",
"length_expression : char_length_expression",
"length_expression : octet_length_expression",
"bit_length_expression : BIT_LENGTH '(' value ')'",
"char_length_expression : CHAR_LENGTH '(' value ')'",
"char_length_expression : CHARACTER_LENGTH '(' value ')'",
"octet_length_expression : OCTET_LENGTH '(' value ')'",
"system_function_expression : system_function_std_syntax '(' value_list_opt ')'",
"system_function_expression : system_function_special_syntax",
"system_function_std_syntax : ABS",
"system_function_std_syntax : ACOS",
"system_function_std_syntax : ACOSH",
"system_function_std_syntax : ASCII_CHAR",
"system_function_std_syntax : ASCII_VAL",
"system_function_std_syntax : ASIN",
"system_function_std_syntax : ASINH",
"system_function_std_syntax : ATAN",
"system_function_std_syntax : ATAN2",
"system_function_std_syntax : ATANH",
"system_function_std_syntax : BASE64_DECODE",
"system_function_std_syntax : BASE64_ENCODE",
"system_function_std_syntax : BIN_AND",
"system_function_std_syntax : BIN_NOT",
"system_function_std_syntax : BIN_OR",
"system_function_std_syntax : BIN_SHL",
"system_function_std_syntax : BIN_SHR",
"system_function_std_syntax : BIN_XOR",
"system_function_std_syntax : CEIL",
"system_function_std_syntax : CHAR_TO_UUID",
"system_function_std_syntax : COS",
"system_function_std_syntax : COSH",
"system_function_std_syntax : COT",
"system_function_std_syntax : EXP",
"system_function_std_syntax : FLOOR",
"system_function_std_syntax : GEN_UUID",
"system_function_std_syntax : HEX_DECODE",
"system_function_std_syntax : HEX_ENCODE",
"system_function_std_syntax : LEFT",
"system_function_std_syntax : LN",
"system_function_std_syntax : LOG",
"system_function_std_syntax : LOG10",
"system_function_std_syntax : LPAD",
"system_function_std_syntax : MAKE_DBKEY",
"system_function_std_syntax : MAXVALUE",
"system_function_std_syntax : MINVALUE",
"system_function_std_syntax : MOD",
"system_function_std_syntax : PI",
"system_function_std_syntax : POWER",
"system_function_std_syntax : RAND",
"system_function_std_syntax : RDB_GET_CONTEXT",
"system_function_std_syntax : RDB_GET_TRANSACTION_CN",
"system_function_std_syntax : RDB_ROLE_IN_USE",
"system_function_std_syntax : RDB_SET_CONTEXT",
"system_function_std_syntax : REPLACE",
"system_function_std_syntax : REVERSE",
"system_function_std_syntax : RIGHT",
"system_function_std_syntax : ROUND",
"system_function_std_syntax : RPAD",
"system_function_std_syntax : RSA_PRIVATE",
"system_function_std_syntax : RSA_PUBLIC",
"system_function_std_syntax : SIGN",
"system_function_std_syntax : SIN",
"system_function_std_syntax : SINH",
"system_function_std_syntax : SQRT",
"system_function_std_syntax : TAN",
"system_function_std_syntax : TANH",
"system_function_std_syntax : TRUNC",
"system_function_std_syntax : UUID_TO_CHAR",
"system_function_std_syntax : QUANTIZE",
"system_function_std_syntax : TOTALORDER",
"system_function_std_syntax : NORMALIZE_DECFLOAT",
"system_function_std_syntax : COMPARE_DECFLOAT",
"system_function_special_syntax : DATEADD '(' value timestamp_part TO value ')'",
"system_function_special_syntax : DATEADD '(' timestamp_part ',' value ',' value ')'",
"system_function_special_syntax : DATEDIFF '(' timestamp_part FROM value TO value ')'",
"system_function_special_syntax : DATEDIFF '(' timestamp_part ',' value ',' value ')'",
"system_function_special_syntax : encrypt_decrypt '(' value USING valid_symbol_name crypt_opt_mode KEY value crypt_opt_iv crypt_opt_counter_type crypt_opt_counter ')'",
"system_function_special_syntax : FIRST_DAY '(' of_first_last_day_part FROM value ')'",
"system_function_special_syntax : HASH '(' value ')'",
"system_function_special_syntax : hash_func '(' value USING valid_symbol_name ')'",
"system_function_special_syntax : LAST_DAY '(' of_first_last_day_part FROM value ')'",
"system_function_special_syntax : OVERLAY '(' value PLACING value FROM value FOR value ')'",
"system_function_special_syntax : OVERLAY '(' value PLACING value FROM value ')'",
"system_function_special_syntax : POSITION '(' value IN value ')'",
"system_function_special_syntax : POSITION '(' value_list_opt ')'",
"system_function_special_syntax : rsa_encrypt_decrypt '(' value KEY value crypt_opt_lparam crypt_opt_hash ')'",
"system_function_special_syntax : RSA_SIGN_HASH '(' value KEY value crypt_opt_hash crypt_opt_saltlen ')'",
"system_function_special_syntax : RSA_VERIFY_HASH '(' value SIGNATURE value KEY value crypt_opt_hash crypt_opt_saltlen ')'",
"system_function_special_syntax : RDB_SYSTEM_PRIVILEGE '(' valid_symbol_name ')'",
"hash_func : HASH",
"hash_func : CRYPT_HASH",
"rsa_encrypt_decrypt : RSA_DECRYPT",
"rsa_encrypt_decrypt : RSA_ENCRYPT",
"crypt_opt_lparam :",
"crypt_opt_lparam : LPARAM value",
"crypt_opt_hash :",
"crypt_opt_hash : HASH valid_symbol_name",
"crypt_opt_saltlen :",
"crypt_opt_saltlen : SALT_LENGTH value",
"crypt_opt_mode :",
"crypt_opt_mode : MODE valid_symbol_name",
"crypt_opt_iv :",
"crypt_opt_iv : IV value",
"crypt_opt_counter_type :",
"crypt_opt_counter_type : crypt_counter_type",
"crypt_counter_type : CTR_BIG_ENDIAN",
"crypt_counter_type : CTR_LITTLE_ENDIAN",
"crypt_opt_counter :",
"crypt_opt_counter : crypt_counter_name value",
"crypt_counter_name : COUNTER",
"crypt_counter_name : CTR_LENGTH",
"encrypt_decrypt : ENCRYPT",
"encrypt_decrypt : DECRYPT",
"of_first_last_day_part : OF YEAR",
"of_first_last_day_part : OF MONTH",
"of_first_last_day_part : OF WEEK",
"string_value_function : substring_function",
"string_value_function : trim_function",
"string_value_function : UPPER '(' value ')'",
"string_value_function : LOWER '(' value ')'",
"substring_function : SUBSTRING '(' value FROM value string_length_opt ')'",
"substring_function : SUBSTRING '(' value SIMILAR value ESCAPE value ')'",
"string_length_opt :",
"string_length_opt : FOR value",
"trim_function : TRIM '(' trim_specification value FROM value ')'",
"trim_function : TRIM '(' value FROM value ')'",
"trim_function : TRIM '(' trim_specification FROM value ')'",
"trim_function : TRIM '(' value ')'",
"trim_specification : BOTH",
"trim_specification : TRAILING",
"trim_specification : LEADING",
"udf : symbol_UDF_call_name '(' value_list ')'",
"udf : symbol_UDF_call_name '(' ')'",
"udf : symbol_package_name '.' symbol_UDF_name '(' value_list ')'",
"udf : symbol_package_name '.' symbol_UDF_name '(' ')'",
"cast_specification : CAST '(' value AS data_type_descriptor ')'",
"case_expression : case_abbreviation",
"case_expression : case_specification",
"case_abbreviation : NULLIF '(' value ',' value ')'",
"case_abbreviation : IIF '(' search_condition ',' value ',' value ')'",
"case_abbreviation : COALESCE '(' value ',' value_list ')'",
"case_abbreviation : DECODE '(' value ',' decode_pairs ')'",
"case_abbreviation : DECODE '(' value ',' decode_pairs ',' value ')'",
"case_specification : simple_case",
"case_specification : searched_case",
"$$171 :",
"$$172 :",
"$$173 :",
"simple_case : CASE case_operand $$171 $$172 $$173 simple_when_clause else_case_result_opt END",
"simple_when_clause : WHEN when_operand THEN case_result",
"simple_when_clause : simple_when_clause WHEN when_operand THEN case_result",
"else_case_result_opt :",
"else_case_result_opt : ELSE case_result",
"searched_case : CASE searched_when_clause END",
"searched_case : CASE searched_when_clause ELSE case_result END",
"searched_when_clause : WHEN search_condition THEN case_result",
"searched_when_clause : searched_when_clause WHEN search_condition THEN case_result",
"when_operand : value",
"case_operand : value",
"case_result : value",
"decode_pairs : value ',' value",
"decode_pairs : decode_pairs ',' value ',' value",
"next_value_expression : NEXT VALUE FOR symbol_generator_name",
"next_value_expression : GEN_ID '(' symbol_generator_name ',' value ')'",
"timestamp_part : YEAR",
"timestamp_part : MONTH",
"timestamp_part : DAY",
"timestamp_part : HOUR",
"timestamp_part : MINUTE",
"timestamp_part : SECOND",
"timestamp_part : MILLISECOND",
"timestamp_part : TIMEZONE_HOUR",
"timestamp_part : TIMEZONE_MINUTE",
"timestamp_part : WEEK",
"timestamp_part : WEEKDAY",
"timestamp_part : YEARDAY",
"all_noise :",
"all_noise : ALL",
"distinct_noise :",
"distinct_noise : DISTINCT",
"null_value : NULL",
"null_value : UNKNOWN",
"symbol_UDF_call_name : SYMBOL",
"symbol_UDF_name : valid_symbol_name",
"symbol_blob_subtype_name : valid_symbol_name",
"symbol_blob_subtype_name : BINARY",
"symbol_character_set_name : valid_symbol_name",
"symbol_character_set_name : BINARY",
"symbol_collation_name : valid_symbol_name",
"symbol_column_name : valid_symbol_name",
"symbol_constraint_name : valid_symbol_name",
"symbol_cursor_name : valid_symbol_name",
"symbol_domain_name : valid_symbol_name",
"symbol_exception_name : valid_symbol_name",
"symbol_filter_name : valid_symbol_name",
"symbol_gdscode_name : valid_symbol_name",
"symbol_generator_name : valid_symbol_name",
"symbol_index_name : valid_symbol_name",
"symbol_item_alias_name : valid_symbol_name",
"symbol_label_name : valid_symbol_name",
"symbol_ddl_name : valid_symbol_name",
"symbol_procedure_name : valid_symbol_name",
"symbol_role_name : valid_symbol_name",
"symbol_table_alias_name : valid_symbol_name",
"symbol_table_name : valid_symbol_name",
"symbol_trigger_name : valid_symbol_name",
"symbol_user_name : valid_symbol_name",
"symbol_variable_name : valid_symbol_name",
"symbol_view_name : valid_symbol_name",
"symbol_savepoint_name : valid_symbol_name",
"symbol_package_name : valid_symbol_name",
"symbol_window_name : valid_symbol_name",
"valid_symbol_name : SYMBOL",
"valid_symbol_name : non_reserved_word",
"non_reserved_word : ACTION",
"non_reserved_word : CASCADE",
"non_reserved_word : FREE_IT",
"non_reserved_word : RESTRICT",
"non_reserved_word : ROLE",
"non_reserved_word : TYPE",
"non_reserved_word : BREAK",
"non_reserved_word : DESCRIPTOR",
"non_reserved_word : SUBSTRING",
"non_reserved_word : COALESCE",
"non_reserved_word : LAST",
"non_reserved_word : LEAVE",
"non_reserved_word : LOCK",
"non_reserved_word : NULLIF",
"non_reserved_word : NULLS",
"non_reserved_word : STATEMENT",
"non_reserved_word : FIRST",
"non_reserved_word : SKIP",
"non_reserved_word : BLOCK",
"non_reserved_word : BACKUP",
"non_reserved_word : DIFFERENCE",
"non_reserved_word : IIF",
"non_reserved_word : SCALAR_ARRAY",
"non_reserved_word : WEEKDAY",
"non_reserved_word : YEARDAY",
"non_reserved_word : SEQUENCE",
"non_reserved_word : NEXT",
"non_reserved_word : RESTART",
"non_reserved_word : COLLATION",
"non_reserved_word : RETURNING",
"non_reserved_word : IGNORE",
"non_reserved_word : LIMBO",
"non_reserved_word : UNDO",
"non_reserved_word : REQUESTS",
"non_reserved_word : TIMEOUT",
"non_reserved_word : ABS",
"non_reserved_word : ACCENT",
"non_reserved_word : ACOS",
"non_reserved_word : ALWAYS",
"non_reserved_word : ASCII_CHAR",
"non_reserved_word : ASCII_VAL",
"non_reserved_word : ASIN",
"non_reserved_word : ATAN",
"non_reserved_word : ATAN2",
"non_reserved_word : BIN_AND",
"non_reserved_word : BIN_OR",
"non_reserved_word : BIN_SHL",
"non_reserved_word : BIN_SHR",
"non_reserved_word : BIN_XOR",
"non_reserved_word : CEIL",
"non_reserved_word : COS",
"non_reserved_word : COSH",
"non_reserved_word : COT",
"non_reserved_word : DATEADD",
"non_reserved_word : DATEDIFF",
"non_reserved_word : DECODE",
"non_reserved_word : EXP",
"non_reserved_word : FLOOR",
"non_reserved_word : GEN_UUID",
"non_reserved_word : GENERATED",
"non_reserved_word : HASH",
"non_reserved_word : LIST",
"non_reserved_word : LN",
"non_reserved_word : LOG",
"non_reserved_word : LOG10",
"non_reserved_word : LPAD",
"non_reserved_word : MATCHED",
"non_reserved_word : MATCHING",
"non_reserved_word : MAXVALUE",
"non_reserved_word : MILLISECOND",
"non_reserved_word : MINVALUE",
"non_reserved_word : MOD",
"non_reserved_word : OVERLAY",
"non_reserved_word : PAD",
"non_reserved_word : PI",
"non_reserved_word : PLACING",
"non_reserved_word : POWER",
"non_reserved_word : PRESERVE",
"non_reserved_word : RAND",
"non_reserved_word : REPLACE",
"non_reserved_word : REVERSE",
"non_reserved_word : ROUND",
"non_reserved_word : RPAD",
"non_reserved_word : SIGN",
"non_reserved_word : SIN",
"non_reserved_word : SINH",
"non_reserved_word : SPACE",
"non_reserved_word : SQRT",
"non_reserved_word : TAN",
"non_reserved_word : TANH",
"non_reserved_word : TEMPORARY",
"non_reserved_word : TRUNC",
"non_reserved_word : AUTONOMOUS",
"non_reserved_word : CHAR_TO_UUID",
"non_reserved_word : FIRSTNAME",
"non_reserved_word : MIDDLENAME",
"non_reserved_word : LASTNAME",
"non_reserved_word : MAPPING",
"non_reserved_word : OS_NAME",
"non_reserved_word : UUID_TO_CHAR",
"non_reserved_word : GRANTED",
"non_reserved_word : CALLER",
"non_reserved_word : COMMON",
"non_reserved_word : DATA",
"non_reserved_word : SOURCE",
"non_reserved_word : TWO_PHASE",
"non_reserved_word : BIN_NOT",
"non_reserved_word : ACTIVE",
"non_reserved_word : AFTER",
"non_reserved_word : ASC",
"non_reserved_word : AUTO",
"non_reserved_word : BEFORE",
"non_reserved_word : COMMITTED",
"non_reserved_word : COMPUTED",
"non_reserved_word : CONDITIONAL",
"non_reserved_word : CONTAINING",
"non_reserved_word : CSTRING",
"non_reserved_word : DATABASE",
"non_reserved_word : DESC",
"non_reserved_word : DO",
"non_reserved_word : DOMAIN",
"non_reserved_word : ENTRY_POINT",
"non_reserved_word : EXCEPTION",
"non_reserved_word : EXIT",
"non_reserved_word : FILE",
"non_reserved_word : GENERATOR",
"non_reserved_word : GEN_ID",
"non_reserved_word : IF",
"non_reserved_word : INACTIVE",
"non_reserved_word : INPUT_TYPE",
"non_reserved_word : ISOLATION",
"non_reserved_word : KEY",
"non_reserved_word : LENGTH",
"non_reserved_word : LEVEL",
"non_reserved_word : MANUAL",
"non_reserved_word : MODULE_NAME",
"non_reserved_word : NAMES",
"non_reserved_word : OPTION",
"non_reserved_word : OUTPUT_TYPE",
"non_reserved_word : OVERFLOW",
"non_reserved_word : PAGE",
"non_reserved_word : PAGES",
"non_reserved_word : PAGE_SIZE",
"non_reserved_word : PASSWORD",
"non_reserved_word : PRIVILEGES",
"non_reserved_word : PROTECTED",
"non_reserved_word : READ",
"non_reserved_word : RESERVING",
"non_reserved_word : RETAIN",
"non_reserved_word : SEGMENT",
"non_reserved_word : SHADOW",
"non_reserved_word : SHARED",
"non_reserved_word : SINGULAR",
"non_reserved_word : SIZE",
"non_reserved_word : SNAPSHOT",
"non_reserved_word : SORT",
"non_reserved_word : STABILITY",
"non_reserved_word : STARTING",
"non_reserved_word : STATISTICS",
"non_reserved_word : SUB_TYPE",
"non_reserved_word : SUSPEND",
"non_reserved_word : TRANSACTION",
"non_reserved_word : UNCOMMITTED",
"non_reserved_word : WAIT",
"non_reserved_word : WEEK",
"non_reserved_word : WORK",
"non_reserved_word : WRITE",
"non_reserved_word : ABSOLUTE",
"non_reserved_word : ACOSH",
"non_reserved_word : ASINH",
"non_reserved_word : ATANH",
"non_reserved_word : BODY",
"non_reserved_word : CONTINUE",
"non_reserved_word : DDL",
"non_reserved_word : DECRYPT",
"non_reserved_word : ENCRYPT",
"non_reserved_word : ENGINE",
"non_reserved_word : IDENTITY",
"non_reserved_word : NAME",
"non_reserved_word : PACKAGE",
"non_reserved_word : PARTITION",
"non_reserved_word : PRIOR",
"non_reserved_word : RELATIVE",
"non_reserved_word : DENSE_RANK",
"non_reserved_word : FIRST_VALUE",
"non_reserved_word : NTH_VALUE",
"non_reserved_word : LAST_VALUE",
"non_reserved_word : LAG",
"non_reserved_word : LEAD",
"non_reserved_word : RANK",
"non_reserved_word : ROW_NUMBER",
"non_reserved_word : USAGE",
"non_reserved_word : LINGER",
"non_reserved_word : TAGS",
"non_reserved_word : PLUGIN",
"non_reserved_word : SERVERWIDE",
"non_reserved_word : INCREMENT",
"non_reserved_word : TRUSTED",
"non_reserved_word : BASE64_DECODE",
"non_reserved_word : BASE64_ENCODE",
"non_reserved_word : BIND",
"non_reserved_word : CLEAR",
"non_reserved_word : COUNTER",
"non_reserved_word : COMPARE_DECFLOAT",
"non_reserved_word : CONNECTIONS",
"non_reserved_word : CONSISTENCY",
"non_reserved_word : CRYPT_HASH",
"non_reserved_word : CTR_BIG_ENDIAN",
"non_reserved_word : CTR_LENGTH",
"non_reserved_word : CTR_LITTLE_ENDIAN",
"non_reserved_word : CUME_DIST",
"non_reserved_word : DEFINER",
"non_reserved_word : DISABLE",
"non_reserved_word : ENABLE",
"non_reserved_word : EXCESS",
"non_reserved_word : EXCLUDE",
"non_reserved_word : EXTENDED",
"non_reserved_word : FIRST_DAY",
"non_reserved_word : FOLLOWING",
"non_reserved_word : HEX_DECODE",
"non_reserved_word : HEX_ENCODE",
"non_reserved_word : IDLE",
"non_reserved_word : INCLUDE",
"non_reserved_word : INVOKER",
"non_reserved_word : IV",
"non_reserved_word : LAST_DAY",
"non_reserved_word : LEGACY",
"non_reserved_word : LIFETIME",
"non_reserved_word : LPARAM",
"non_reserved_word : MAKE_DBKEY",
"non_reserved_word : MESSAGE",
"non_reserved_word : MODE",
"non_reserved_word : NATIVE",
"non_reserved_word : NORMALIZE_DECFLOAT",
"non_reserved_word : NTILE",
"non_reserved_word : NUMBER",
"non_reserved_word : OLDEST",
"non_reserved_word : OTHERS",
"non_reserved_word : OVERRIDING",
"non_reserved_word : PERCENT_RANK",
"non_reserved_word : POOL",
"non_reserved_word : PRECEDING",
"non_reserved_word : PRIVILEGE",
"non_reserved_word : QUANTIZE",
"non_reserved_word : RANGE",
"non_reserved_word : RESET",
"non_reserved_word : RSA_DECRYPT",
"non_reserved_word : RSA_ENCRYPT",
"non_reserved_word : RSA_PRIVATE",
"non_reserved_word : RSA_PUBLIC",
"non_reserved_word : RSA_SIGN_HASH",
"non_reserved_word : RSA_VERIFY_HASH",
"non_reserved_word : SALT_LENGTH",
"non_reserved_word : SECURITY",
"non_reserved_word : SESSION",
"non_reserved_word : SIGNATURE",
"non_reserved_word : SQL",
"non_reserved_word : SYSTEM",
"non_reserved_word : TIES",
"non_reserved_word : TOTALORDER",
"non_reserved_word : TRAPS",
"non_reserved_word : ZONE",
};
#endif

#ifdef YYREDUCEPOSNFUNC
#define YYCALLREDUCEPOSN(e)   \
	if (reduce_posn) {     \
	  YYREDUCEPOSNFUNC(yyps->pos, &(yyps->psp)[1 - yym], &(yyps->vsp)[1 - yym], \
			   yym, yyps->psp - yyps->ps, yychar, yyposn, e);       \
	  reduce_posn = 0;    \
	}

#ifndef YYCALLREDUCEPOSNARG
#define YYCALLREDUCEPOSNARG yyps->val
#endif


#define YYPOSNARG(n) ((yyps->psp)[1 - yym + (n) - 1])
#define YYPOSNOUT    (yyps->pos)
#endif

// If delete function is not defined by the user, do not deletions.
#ifndef YYDELETEVAL
#define YYDELETEVAL(v, n)
#endif

// If delete function is not defined by the user, do not deletions.
#ifndef YYDELETEPOSN
#define YYDELETEPOSN(v, n)
#endif

#define yyclearin (yychar = (-1))

#define yyerrok (yyps->errflag = 0)

#ifndef YYSTACKGROWTH
#define YYSTACKGROWTH 128
#endif

#ifndef YYDEFSTACKSIZE
#define YYDEFSTACKSIZE 12
#endif

#ifdef YYDEBUG
int yydebug;
#endif


#define YYABORT  goto yyabort
#define YYACCEPT goto yyaccept
#define YYVALID         do { if (yyps->save)          goto yyvalid; } while(0)
#define YYVALID_NESTED  do { if (yyps->save && \
                                 yyps->save->save == 0) goto yyvalid; } while(0)
//
// For use in generated program
//
#define yytrial (yyps->save)
#define yyvsp   (yyps->vsp)
#define yyval   (yyps->val)
#define yydepth (yyps->ssp - yyps->ss)




void Parser::yySCopy(YYSTYPE* to, YYSTYPE* from, int size)
{
  int i;
  for (i = size - 1; i >= 0; i--) {
    to[i] = from[i];
  }
}


void Parser::yyPCopy(YYPOSN* to, YYPOSN* from, int size)
{
  int i;
  for (i = size - 1; i >= 0; i--) {
    to[i] = from[i];
  }
}


void Parser::yyMoreStack(yyparsestate* yyps)
{
  int p = yyps->ssp - yyps->ss;
  Yshort  *tss = yyps->ss;
  YYSTYPE *tvs = yyps->vs;
  YYPOSN  *tps = yyps->ps;
  yyps->ss = FB_NEW_POOL(getPool()) Yshort [yyps->stacksize + YYSTACKGROWTH];
  yyps->vs = FB_NEW_POOL(getPool()) YYSTYPE[yyps->stacksize + YYSTACKGROWTH];
  yyps->ps = FB_NEW_POOL(getPool()) YYPOSN [yyps->stacksize + YYSTACKGROWTH];
  memcpy(yyps->ss, tss, yyps->stacksize * sizeof(Yshort));
  yySCopy(yyps->vs, tvs, yyps->stacksize);
  yyPCopy(yyps->ps, tps, yyps->stacksize);
  yyps->stacksize += YYSTACKGROWTH;
  delete[] tss;
  delete[] tvs;
  delete[] tps;
  yyps->ssp = yyps->ss + p;
  yyps->vsp = yyps->vs + p;
  yyps->psp = yyps->ps + p;
}


Parser::yyparsestate* Parser::yyNewState(int size)
{
  yyparsestate *p = FB_NEW_POOL(getPool()) yyparsestate;
  p->stacksize = size + 4;
  p->ss = FB_NEW_POOL(getPool()) Yshort [size + 4];
  p->vs = FB_NEW_POOL(getPool()) YYSTYPE[size + 4];
  p->ps = FB_NEW_POOL(getPool()) YYPOSN [size + 4];
  memset(&p->vs[0], 0, (size + 4) * sizeof(YYSTYPE));
  memset(&p->ps[0], 0, (size + 4) * sizeof(YYPOSN));
  return p;
}


void Parser::yyFreeState(Parser::yyparsestate* p)
{
  delete[] p->ss;
  delete[] p->vs;
  delete[] p->ps;
  delete p;
}


void Parser::setNodeLineColumn(Node* node)
{
	// Check if we are not accessing the position stack below its base.
	if (&YYPOSNARG(1) >= yyps->ps)
	{
		node->line = YYPOSNARG(1).firstLine;
		node->column = YYPOSNARG(1).firstColumn;
	}
}


int Parser::parseAux()
{
  // ASF: yym moved to Parser.h
  int /*yym, */yyn, yystate, yychar, yynewerrflag;
  yyparsestate *yyerrctx = NULL;
  int reduce_posn;

#if YYDEBUG
  const char *yys;

  if ((yys = getenv("YYDEBUG"))) {
    yyn = *yys;
    if (yyn >= '0' && yyn <= '9')
      yydebug = yyn - '0';
  }
#endif

  yyps = yyNewState(YYDEFSTACKSIZE);
  yyps->save = 0;
  yynerrs = 0;
  yyps->errflag = 0;
  yychar = (-1);

  yyps->ssp = yyps->ss;
  yyps->vsp = yyps->vs;
  yyps->psp = yyps->ps;
  *(yyps->ssp) = yystate = 0;


  //
  // Main parsing loop
  //
 yyloop:
  if ((yyn = yydefred[yystate])) {
    goto yyreduce;
  }

  //
  // Read one token
  //
  if (yychar < 0) {
    if ((yychar = yylex1()) < 0) yychar = 0;
#if YYDEBUG
    if (yydebug) {
      yys = 0;
      if (yychar <= YYMAXTOKEN) yys = yyname[yychar];
      if (!yys) yys = "illegal-symbol";
      printf("yydebug[%d,%p]: state %d, reading %d (%s)",
	     int(yydepth), yytrial, yystate, yychar, yys);
#ifdef YYDBPR
      printf("<");
      YYDBPR(yylval);
      printf(">");
#endif
      printf("\n");
    }
#endif
  }

  //
  // Do we have a conflict?
  //
  if ((yyn = yycindex[yystate]) &&
      (yyn += yychar) >= 0 &&
      yyn <= YYTABLESIZE &&
      yycheck[yyn] == yychar) {
    int ctry;

    if (yypath) {
#if YYDEBUG
      if (yydebug) {
        printf("yydebug[%d,%p]: CONFLICT in state %d: following successful trial parse\n",
	       int(yydepth), yytrial, yystate);
      }
#endif
      // Switch to the next conflict context
      yyparsestate *save = yypath;
      yypath = save->save;
      ctry = save->ctry;
      if (save->state != yystate)
        goto yyabort;
      yyFreeState(save);

    } else {

#if YYDEBUG
      if (yydebug) {
        printf("yydebug[%d,%p]: CONFLICT in state %d. ",
	       int(yydepth), yytrial, yystate);
        if (yyps->save) {
          printf("ALREADY in conflict. Continue trial parse.");
        } else {
          printf("Start trial parse.");
        }
        printf("\n");
      }
#endif
      yyparsestate *save = yyNewState(yyps->ssp - yyps->ss);
      save->save    = yyps->save;
      save->state   = yystate;
      save->errflag = yyps->errflag;
      save->ssp     = save->ss + (yyps->ssp - yyps->ss);
      save->vsp     = save->vs + (yyps->vsp - yyps->vs);
      save->psp     = save->ps + (yyps->psp - yyps->ps);
      memcpy (save->ss, yyps->ss, (yyps->ssp - yyps->ss + 1) * sizeof(Yshort));
      yySCopy(save->vs, yyps->vs, (yyps->ssp - yyps->ss + 1));
      yyPCopy(save->ps, yyps->ps, (yyps->ssp - yyps->ss + 1));
      ctry = yytable[yyn];
      if (yyctable[ctry] == -1) {
#if YYDEBUG
        if (yydebug && yychar >= 0)
          printf("yydebug[%p]: backtracking 1 token\n", yytrial);
#endif
        ctry++;
      }
      save->ctry = ctry;
      if (!yyps->save) {
        // If this is a first conflict in the stack, start saving lexemes
        if (!yylexemes) {
          yylexemes = FB_NEW_POOL(getPool()) Yshort[YYSTACKGROWTH];
          yylvals = FB_NEW_POOL(getPool()) YYSTYPE[YYSTACKGROWTH];
          yylvlim = yylvals + YYSTACKGROWTH;
          yylpsns = FB_NEW_POOL(getPool()) YYPOSN[YYSTACKGROWTH];
          yylplim = yylpsns + YYSTACKGROWTH;
        }
        if (yylvp == yylve) {
          yylvp = yylve = yylvals;
	  yylpp = yylpe = yylpsns;
          yylexp = yylexemes;
          if (yychar >= 0) {
            *yylve++ = yylval;
            *yylpe++ = yyposn;
            *yylexp = yychar;
            yychar = -1;
          }
        }
      }
      if (yychar >= 0) {
        yylvp--, yylpp--, yylexp--;
        yychar = -1;
      }
      save->lexeme = yylvp - yylvals;
      yyps->save = save;
    }
    if (yytable[yyn] == ctry) {
#if YYDEBUG
      if (yydebug)
        printf("yydebug[%d,%p]: state %d, shifting to state %d\n",
               int(yydepth), yytrial, yystate, yyctable[ctry]);
#endif
      if (yychar < 0)
        yylvp++, yylpp++, yylexp++;
      yychar = -1;
      if (yyps->errflag > 0) --yyps->errflag;
      yystate = yyctable[ctry];
      goto yyshift;
    } else {
      yyn = yyctable[ctry];
      goto yyreduce;
    }
  }

  //
  // Is action a shift?
  //
  if ((yyn = yysindex[yystate]) &&
      (yyn += yychar) >= 0 &&
      yyn <= YYTABLESIZE &&
      yycheck[yyn] == yychar) {
#if YYDEBUG
    if (yydebug)
      printf("yydebug[%d,%p]: state %d, shifting to state %d\n",
             int(yydepth), yytrial, yystate, yytable[yyn]);
#endif
    yychar = (-1);
    if (yyps->errflag > 0)  --yyps->errflag;
    yystate = yytable[yyn];
  yyshift:
    if (yyps->ssp >= yyps->ss + yyps->stacksize - 1) {
      yyMoreStack(yyps);
    }
    *++(yyps->ssp) = yystate;
    *++(yyps->vsp) = yylval;
    *++(yyps->psp) = yyposn;
    goto yyloop;
  }
  if ((yyn = yyrindex[yystate]) &&
      (yyn += yychar) >= 0 &&
      yyn <= YYTABLESIZE &&
      yycheck[yyn] == yychar) {
    yyn = yytable[yyn];
    goto yyreduce;
  }

  //
  // Action: error
  //
  if (yyps->errflag) goto yyinrecovery;
  while (yyps->save) {
    int ctry;
    yyparsestate *save = yyps->save;
#if YYDEBUG
    if (yydebug)
      printf("yydebug[%d,%p]: ERROR in state %d, CONFLICT BACKTRACKING to state %d, %d tokens\n",
             int(yydepth), yytrial, yystate, yyps->save->state, int(yylvp - yylvals - yyps->save->lexeme));
#endif
    // Memorize most forward-looking error state in case
    // it's really an error.
    if (yyerrctx == NULL || yyerrctx->lexeme < yylvp-yylvals) {
      // Free old saved error context state
      if (yyerrctx) yyFreeState(yyerrctx);
      // Create and fill out new saved error context state
      yyerrctx = yyNewState(yyps->ssp - yyps->ss);
      yyerrctx->save = yyps->save;
      yyerrctx->state = yystate;
      yyerrctx->errflag = yyps->errflag;
      yyerrctx->ssp = yyerrctx->ss + (yyps->ssp - yyps->ss);
      yyerrctx->vsp = yyerrctx->vs + (yyps->vsp - yyps->vs);
      yyerrctx->psp = yyerrctx->ps + (yyps->psp - yyps->ps);
      memcpy (yyerrctx->ss, yyps->ss, (yyps->ssp - yyps->ss + 1) * sizeof(Yshort));
      yySCopy(yyerrctx->vs, yyps->vs, (yyps->ssp - yyps->ss + 1));
      yyPCopy(yyerrctx->ps, yyps->ps, (yyps->ssp - yyps->ss + 1));
      yyerrctx->lexeme = yylvp - yylvals;
    }
    yylvp  = yylvals   + save->lexeme;
    yylpp  = yylpsns   + save->lexeme;
    yylexp = yylexemes + save->lexeme;
    yychar = -1;
    yyps->ssp = yyps->ss + (save->ssp - save->ss);
    yyps->vsp = yyps->vs + (save->vsp - save->vs);
    yyps->psp = yyps->ps + (save->psp - save->ps);
    memcpy (yyps->ss, save->ss, (yyps->ssp - yyps->ss + 1) * sizeof(Yshort));
    yySCopy(yyps->vs, save->vs,  yyps->vsp - yyps->vs + 1);
    yyPCopy(yyps->ps, save->ps,  yyps->psp - yyps->ps + 1);
    ctry = ++save->ctry;
    yystate = save->state;
    // We tried shift, try reduce now
    if ((yyn = yyctable[ctry]) >= 0) {
      goto yyreduce;
    }
    yyps->save = save->save;
    yyFreeState(save);
    //
    // Nothing left on the stack -- error
    //
    if (!yyps->save) {
#if YYDEBUG
      if (yydebug) {
        printf("yydebug[%p]: trial parse FAILED, entering ERROR mode\n",
	       yytrial);
      }
#endif
      // Restore state as it was in the most forward-advanced error
      yylvp  = yylvals   + yyerrctx->lexeme;
      yylpp  = yylpsns   + yyerrctx->lexeme;
      yylexp = yylexemes + yyerrctx->lexeme;
      yychar = yylexp[-1];
      yylval = yylvp[-1];
      yyposn = yylpp[-1];
      yyps->ssp = yyps->ss + (yyerrctx->ssp - yyerrctx->ss);
      yyps->vsp = yyps->vs + (yyerrctx->vsp - yyerrctx->vs);
      yyps->psp = yyps->ps + (yyerrctx->psp - yyerrctx->ps);
      memcpy (yyps->ss, yyerrctx->ss, (yyps->ssp - yyps->ss + 1) * sizeof(Yshort));
      yySCopy(yyps->vs, yyerrctx->vs,  yyps->vsp - yyps->vs + 1);
      yyPCopy(yyps->ps, yyerrctx->ps,  yyps->psp - yyps->ps + 1);
      yystate = yyerrctx->state;
      yyFreeState(yyerrctx);
      yyerrctx = NULL;
    }
  }
  yyerror_detailed("syntax error", yychar, yylval, yyposn);
  ++yynerrs;

 yyinrecovery:
  if (yyps->errflag < 3) {
    yyps->errflag = 3;
    for (;;) {
      if ((yyn = yysindex[*(yyps->ssp)]) &&
	  (yyn += YYERRCODE) >= 0 &&
          yyn <= YYTABLESIZE &&
	  yycheck[yyn] == YYERRCODE) {
#if YYDEBUG
        if (yydebug)
          printf("yydebug[%d,%p]: state %d, ERROR recovery shifts to state %d\n",
                 int(yydepth), yytrial, *(yyps->ssp), yytable[yyn]);
#endif
        yystate = yytable[yyn];
        goto yyshift;
      } else {
#if YYDEBUG
        if (yydebug)
          printf("yydebug[%d,%p]: ERROR recovery discards state %d\n",
                 int(yydepth), yytrial, *(yyps->ssp));
#endif
        if (yyps->ssp <= yyps->ss) {
	  goto yyabort;
	}
	if (!yytrial) {
	  YYDELETEVAL(yyps->vsp[0], 1);
	  YYDELETEPOSN(yyps->psp[0], 1);
	}
        --(yyps->ssp);
        --(yyps->vsp);
        --(yyps->psp);
      }
    }
  } else {
    if (yychar == 0) goto yyabort;
#if YYDEBUG
    if (yydebug) {
      yys = 0;
      if (yychar <= YYMAXTOKEN) yys = yyname[yychar];
      if (!yys) yys = "illegal-symbol";
      printf("yydebug[%d,%p]: state %d, ERROR recovery discards token %d (%s)\n",
             int(yydepth), yytrial, yystate, yychar, yys);
    }
#endif
    if (!yytrial) {
      YYDELETEVAL(yylval, 0);
      YYDELETEPOSN(yyposn, 0);
    }
    yychar = (-1);
    goto yyloop;
  }

  //
  // Reduce the rule
  //
yyreduce:
  yym = yylen[yyn];
#if YYDEBUG
  if (yydebug) {
    printf("yydebug[%d,%p]: state %d, reducing by rule %d (%s)",
           int(yydepth), yytrial, yystate, yyn, yyrule[yyn]);
#ifdef YYDBPR
    if (yym) {
      int i;
      printf("<");
      for (i = yym; i > 0; i--) {
        if (i != yym) printf(", ");
        YYDBPR((yyps->vsp)[1 - i]);
      }
      printf(">");
    }
#endif
    printf("\n");
  }
#endif
  if (yyps->ssp + 1 - yym >= yyps->ss + yyps->stacksize) {
    yyMoreStack(yyps);
  }

  /* default action - assign last argument as in standard yacc */
  yyps->val = yyvsp[1 - yym];

  /* default reduced position is NULL -- no position at all.
     no position will be assigned at trial time and if no position handling is present */
  memset(&yyps->pos, 0, sizeof(yyps->pos));

  reduce_posn = TRUE;

  switch (yyn) {

case 1:
  if (!yytrial)
{ DSQL_parse = yyvsp[0].dsqlReq; }
break;
case 2:
  if (!yytrial)
{ DSQL_parse = yyvsp[-1].dsqlReq; }
break;
case 3:
  if (!yytrial)
{ yyval.dsqlReq = FB_NEW_POOL(getStatementPool()) DsqlDmlRequest(getStatementPool(), yyvsp[0].stmtNode); }
break;
case 4:
  if (!yytrial)
{ yyval.dsqlReq = FB_NEW_POOL(getStatementPool()) DsqlDdlRequest(getStatementPool(), yyvsp[0].ddlNode); }
break;
case 5:
  if (!yytrial)
{ yyval.dsqlReq = FB_NEW_POOL(getStatementPool()) DsqlTransactionRequest(getStatementPool(), yyvsp[0].traNode); }
break;
case 6:
  if (!yytrial)
{ yyval.dsqlReq = FB_NEW_POOL(getStatementPool()) DsqlSessionManagementRequest(getStatementPool(), yyvsp[0].mngNode); }
break;
case 7:
  if (!yytrial)
{ yyval.stmtNode = yyvsp[0].stmtNode; }
break;
case 8:
  if (!yytrial)
{ yyval.stmtNode = yyvsp[0].storeNode; }
break;
case 9:
  if (!yytrial)
{ yyval.stmtNode = yyvsp[0].mergeNode; }
break;
case 10:
  if (!yytrial)
{ yyval.stmtNode = yyvsp[0].stmtNode; }
break;
case 11:
  if (!yytrial)
{ yyval.stmtNode = yyvsp[0].execBlockNode; }
break;
case 12:
  if (!yytrial)
{ yyval.stmtNode = yyvsp[0].stmtNode; }
break;
case 13:
  if (!yytrial)
{ yyval.stmtNode = yyvsp[0].selectNode; }
break;
case 14:
  if (!yytrial)
{ yyval.stmtNode = yyvsp[0].stmtNode; }
break;
case 15:
  if (!yytrial)
{ yyval.stmtNode = yyvsp[0].updInsNode; }
break;
case 16:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].ddlNode; }
break;
case 17:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].ddlNode; }
break;
case 18:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].ddlNode; }
break;
case 19:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].ddlNode; }
break;
case 20:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].ddlNode; }
break;
case 21:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].ddlNode; }
break;
case 22:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].grantRevokeNode; }
break;
case 23:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].ddlNode; }
break;
case 24:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].grantRevokeNode; }
break;
case 25:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].ddlNode; }
break;
case 26:
  if (!yytrial)
{ yyval.traNode = yyvsp[0].setTransactionNode; }
break;
case 27:
  if (!yytrial)
{ yyval.traNode = yyvsp[0].traNode; }
break;
case 28:
  if (!yytrial)
{ yyval.traNode = yyvsp[0].traNode; }
break;
case 29:
  if (!yytrial)
{ yyval.mngNode = yyvsp[0].setDecFloatRoundNode; }
break;
case 30:
  if (!yytrial)
{ yyval.mngNode = yyvsp[0].setDecFloatTrapsNode; }
break;
case 31:
  if (!yytrial)
{ yyval.mngNode = yyvsp[0].setSessionNode; }
break;
case 32:
  if (!yytrial)
{ yyval.mngNode = yyvsp[0].setRoleNode; }
break;
case 33:
  if (!yytrial)
{ yyval.mngNode = yyvsp[0].sessionResetNode; }
break;
case 34:
  if (!yytrial)
{ yyval.mngNode = yyvsp[0].mngNode; }
break;
case 35:
  if (!yytrial)
{ yyval.mngNode = yyvsp[0].setBindNode; }
break;
case 36:
  if (!yytrial)
{ yyval.grantRevokeNode = newNode<GrantRevokeNode>(true); }
break;
case 37:
yyval.grantRevokeNode = yyvsp[0].grantRevokeNode;
break;
case 38:
  if (!yytrial)
{ yyval.grantRevokeNode = yyvsp[-2].grantRevokeNode; }
break;
case 39:
yyval.privilegeArray = NOTRIAL(&yyvsp[0].grantRevokeNode->privileges);
break;
case 40:
yyval.granteeArray = NOTRIAL(&yyvsp[-6].grantRevokeNode->users);
break;
case 41:
  if (!yytrial)
{
			yyvsp[-10].grantRevokeNode->object = newNode<GranteeClause>(obj_relation, *yyvsp[-5].metaNamePtr);
			yyvsp[-10].grantRevokeNode->grantAdminOption = yyvsp[-1].boolVal;
			yyvsp[-10].grantRevokeNode->grantor = yyvsp[0].metaNamePtr;
		}
break;
case 42:
  if (!yytrial)
{
			yyvsp[-10].grantRevokeNode->object = newNode<GranteeClause>(obj_procedure, *yyvsp[-5].metaNamePtr);
			yyvsp[-10].grantRevokeNode->grantAdminOption = yyvsp[-1].boolVal;
			yyvsp[-10].grantRevokeNode->grantor = yyvsp[0].metaNamePtr;
		}
break;
case 43:
  if (!yytrial)
{
			yyvsp[-10].grantRevokeNode->object = newNode<GranteeClause>(obj_udf, *yyvsp[-5].metaNamePtr);
			yyvsp[-10].grantRevokeNode->grantAdminOption = yyvsp[-1].boolVal;
			yyvsp[-10].grantRevokeNode->grantor = yyvsp[0].metaNamePtr;
		}
break;
case 44:
  if (!yytrial)
{
			yyvsp[-10].grantRevokeNode->object = newNode<GranteeClause>(obj_package_header, *yyvsp[-5].metaNamePtr);
			yyvsp[-10].grantRevokeNode->grantAdminOption = yyvsp[-1].boolVal;
			yyvsp[-10].grantRevokeNode->grantor = yyvsp[0].metaNamePtr;
		}
break;
case 45:
  if (!yytrial)
{
			yyvsp[-10].grantRevokeNode->object = newNode<GranteeClause>(obj_exception, *yyvsp[-5].metaNamePtr);
			yyvsp[-10].grantRevokeNode->grantAdminOption = yyvsp[-1].boolVal;
			yyvsp[-10].grantRevokeNode->grantor = yyvsp[0].metaNamePtr;
		}
break;
case 46:
  if (!yytrial)
{
			yyvsp[-10].grantRevokeNode->object = newNode<GranteeClause>(obj_generator, *yyvsp[-5].metaNamePtr);
			yyvsp[-10].grantRevokeNode->grantAdminOption = yyvsp[-1].boolVal;
			yyvsp[-10].grantRevokeNode->grantor = yyvsp[0].metaNamePtr;
		}
break;
case 47:
  if (!yytrial)
{
			yyvsp[-10].grantRevokeNode->object = newNode<GranteeClause>(obj_generator, *yyvsp[-5].metaNamePtr);
			yyvsp[-10].grantRevokeNode->grantAdminOption = yyvsp[-1].boolVal;
			yyvsp[-10].grantRevokeNode->grantor = yyvsp[0].metaNamePtr;
		}
break;
case 48:
yyval.granteeArray = NOTRIAL(&yyvsp[-4].grantRevokeNode->users);
break;
case 49:
  if (!yytrial)
{
			yyvsp[-8].grantRevokeNode->object = yyvsp[-5].granteeClause;
			yyvsp[-8].grantRevokeNode->grantAdminOption = yyvsp[-1].boolVal;
			yyvsp[-8].grantRevokeNode->grantor = yyvsp[0].metaNamePtr;
			yyvsp[-8].grantRevokeNode->isDdl = true;
		}
break;
case 50:
  if (!yytrial)
{
			yyvsp[-8].grantRevokeNode->object = newNode<GranteeClause>(obj_database, get_object_name(obj_database));
			yyvsp[-8].grantRevokeNode->grantAdminOption = yyvsp[-1].boolVal;
			yyvsp[-8].grantRevokeNode->grantor = yyvsp[0].metaNamePtr;
			yyvsp[-8].grantRevokeNode->isDdl = true;
		}
break;
case 51:
yyval.grantRevokeNode = NOTRIAL(yyvsp[0].grantRevokeNode);
break;
case 52:
yyval.granteeArray = NOTRIAL(&yyvsp[-3].grantRevokeNode->users);
break;
case 53:
  if (!yytrial)
{
			yyvsp[-7].grantRevokeNode->grantAdminOption = yyvsp[-1].boolVal;
			yyvsp[-7].grantRevokeNode->grantor = yyvsp[0].metaNamePtr;
		}
break;
case 54:
  if (!yytrial)
{ yyval.granteeClause = newNode<GranteeClause>(obj_relations, get_object_name(obj_relations)); }
break;
case 55:
  if (!yytrial)
{ yyval.granteeClause = newNode<GranteeClause>(obj_views, get_object_name(obj_views)); }
break;
case 56:
  if (!yytrial)
{ yyval.granteeClause = newNode<GranteeClause>(obj_procedures, get_object_name(obj_procedures)); }
break;
case 57:
  if (!yytrial)
{ yyval.granteeClause = newNode<GranteeClause>(obj_functions, get_object_name(obj_functions)); }
break;
case 58:
  if (!yytrial)
{ yyval.granteeClause = newNode<GranteeClause>(obj_packages, get_object_name(obj_packages)); }
break;
case 59:
  if (!yytrial)
{ yyval.granteeClause = newNode<GranteeClause>(obj_generators, get_object_name(obj_generators)); }
break;
case 60:
  if (!yytrial)
{ yyval.granteeClause = newNode<GranteeClause>(obj_generators, get_object_name(obj_generators)); }
break;
case 61:
  if (!yytrial)
{ yyval.granteeClause = newNode<GranteeClause>(obj_domains, get_object_name(obj_domains)); }
break;
case 62:
  if (!yytrial)
{ yyval.granteeClause = newNode<GranteeClause>(obj_exceptions, get_object_name(obj_exceptions)); }
break;
case 63:
  if (!yytrial)
{ yyval.granteeClause = newNode<GranteeClause>(obj_roles, get_object_name(obj_roles)); }
break;
case 64:
  if (!yytrial)
{ yyval.granteeClause = newNode<GranteeClause>(obj_charsets, get_object_name(obj_charsets)); }
break;
case 65:
  if (!yytrial)
{ yyval.granteeClause = newNode<GranteeClause>(obj_collations, get_object_name(obj_collations)); }
break;
case 66:
  if (!yytrial)
{ yyval.granteeClause = newNode<GranteeClause>(obj_filters, get_object_name(obj_filters)); }
break;
case 69:
  if (!yytrial)
{ yyvsp[-1].privilegeArray->add(PrivilegeClause('A', NULL)); }
break;
case 70:
  if (!yytrial)
{ yyvsp[-2].privilegeArray->add(PrivilegeClause('A', NULL)); }
break;
case 71:
yyval.privilegeArray = yyvsp[0].privilegeArray;
break;
case 74:
yyval.privilegeArray = yyvsp[-2].privilegeArray;
break;
case 76:
  if (!yytrial)
{ yyvsp[-1].privilegeArray->add(PrivilegeClause('X', NULL)); }
break;
case 77:
  if (!yytrial)
{ yyvsp[-1].privilegeArray->add(PrivilegeClause('G', NULL)); }
break;
case 78:
  if (!yytrial)
{ yyvsp[-1].privilegeArray->add(PrivilegeClause('S', NULL)); }
break;
case 79:
  if (!yytrial)
{ yyvsp[-1].privilegeArray->add(PrivilegeClause('I', NULL)); }
break;
case 80:
  if (!yytrial)
{ yyvsp[-1].privilegeArray->add(PrivilegeClause('D', NULL)); }
break;
case 81:
  if (!yytrial)
{ yyvsp[-2].privilegeArray->add(PrivilegeClause('U', yyvsp[0].valueListNode)); }
break;
case 82:
  if (!yytrial)
{ yyvsp[-2].privilegeArray->add(PrivilegeClause('R', yyvsp[0].valueListNode)); }
break;
case 83:
  if (!yytrial)
{
			yyvsp[-2].privilegeArray->add(PrivilegeClause('C', NULL));
			yyvsp[-2].privilegeArray->add(PrivilegeClause('L', NULL));
			yyvsp[-2].privilegeArray->add(PrivilegeClause('O', NULL));
		}
break;
case 89:
  if (!yytrial)
{ yyvsp[-1].privilegeArray->add(PrivilegeClause('C', NULL)); }
break;
case 90:
  if (!yytrial)
{ yyvsp[-2].privilegeArray->add(PrivilegeClause('L', NULL)); }
break;
case 91:
  if (!yytrial)
{ yyvsp[-2].privilegeArray->add(PrivilegeClause('O', NULL)); }
break;
case 92:
  if (!yytrial)
{
			yyvsp[-2].privilegeArray->add(PrivilegeClause('L', NULL));
			yyvsp[-2].privilegeArray->add(PrivilegeClause('O', NULL));
		}
break;
case 96:
  if (!yytrial)
{ yyvsp[-1].privilegeArray->add(PrivilegeClause('C', NULL)); }
break;
case 97:
  if (!yytrial)
{ yyvsp[-1].privilegeArray->add(PrivilegeClause('L', NULL)); }
break;
case 98:
  if (!yytrial)
{ yyvsp[-1].privilegeArray->add(PrivilegeClause('O', NULL)); }
break;
case 99:
  if (!yytrial)
{ yyval.boolVal = false; }
break;
case 100:
  if (!yytrial)
{ yyval.boolVal = true; }
break;
case 101:
  if (!yytrial)
{ yyval.boolVal = false; }
break;
case 102:
  if (!yytrial)
{ yyval.boolVal = true; }
break;
case 103:
  if (!yytrial)
{ yyval.metaNamePtr = NULL; }
break;
case 104:
  if (!yytrial)
{ yyval.metaNamePtr = yyvsp[0].metaNamePtr; }
break;
case 108:
  if (!yytrial)
{ yyval.metaNamePtr = yyvsp[0].metaNamePtr; }
break;
case 109:
  if (!yytrial)
{ yyval.grantRevokeNode = newNode<GrantRevokeNode>(false); }
break;
case 110:
  if (!yytrial)
{ yyval.grantRevokeNode = yyvsp[-2].grantRevokeNode; }
break;
case 111:
yyval.privilegeArray = NOTRIAL(&yyvsp[-1].grantRevokeNode->privileges);
break;
case 112:
yyval.granteeArray = NOTRIAL(&yyvsp[-7].grantRevokeNode->users);
break;
case 113:
  if (!yytrial)
{
			yyvsp[-10].grantRevokeNode->object = newNode<GranteeClause>(obj_relation, *yyvsp[-4].metaNamePtr);
			yyvsp[-10].grantRevokeNode->grantAdminOption = yyvsp[-9].boolVal;
			yyvsp[-10].grantRevokeNode->grantor = yyvsp[0].metaNamePtr;
		}
break;
case 114:
  if (!yytrial)
{
			yyvsp[-10].grantRevokeNode->object = newNode<GranteeClause>(obj_procedure, *yyvsp[-4].metaNamePtr);
			yyvsp[-10].grantRevokeNode->grantAdminOption = yyvsp[-9].boolVal;
			yyvsp[-10].grantRevokeNode->grantor = yyvsp[0].metaNamePtr;
		}
break;
case 115:
  if (!yytrial)
{
			yyvsp[-10].grantRevokeNode->object = newNode<GranteeClause>(obj_udf, *yyvsp[-4].metaNamePtr);
			yyvsp[-10].grantRevokeNode->grantAdminOption = yyvsp[-9].boolVal;
			yyvsp[-10].grantRevokeNode->grantor = yyvsp[0].metaNamePtr;
		}
break;
case 116:
  if (!yytrial)
{
			yyvsp[-10].grantRevokeNode->object = newNode<GranteeClause>(obj_package_header, *yyvsp[-4].metaNamePtr);
			yyvsp[-10].grantRevokeNode->grantAdminOption = yyvsp[-9].boolVal;
			yyvsp[-10].grantRevokeNode->grantor = yyvsp[0].metaNamePtr;
		}
break;
case 117:
  if (!yytrial)
{
			yyvsp[-10].grantRevokeNode->object = newNode<GranteeClause>(obj_exception, *yyvsp[-4].metaNamePtr);
			yyvsp[-10].grantRevokeNode->grantAdminOption = yyvsp[-9].boolVal;
			yyvsp[-10].grantRevokeNode->grantor = yyvsp[0].metaNamePtr;
		}
break;
case 118:
  if (!yytrial)
{
			yyvsp[-10].grantRevokeNode->object = newNode<GranteeClause>(obj_generator, *yyvsp[-4].metaNamePtr);
			yyvsp[-10].grantRevokeNode->grantAdminOption = yyvsp[-9].boolVal;
			yyvsp[-10].grantRevokeNode->grantor = yyvsp[0].metaNamePtr;
		}
break;
case 119:
  if (!yytrial)
{
			yyvsp[-10].grantRevokeNode->object = newNode<GranteeClause>(obj_generator, *yyvsp[-4].metaNamePtr);
			yyvsp[-10].grantRevokeNode->grantAdminOption = yyvsp[-9].boolVal;
			yyvsp[-10].grantRevokeNode->grantor = yyvsp[0].metaNamePtr;
		}
break;
case 120:
yyval.granteeArray = NOTRIAL(&yyvsp[-5].grantRevokeNode->users);
break;
case 121:
  if (!yytrial)
{
			yyvsp[-8].grantRevokeNode->object = yyvsp[-4].granteeClause;
			yyvsp[-8].grantRevokeNode->grantAdminOption = yyvsp[-7].boolVal;
			yyvsp[-8].grantRevokeNode->grantor = yyvsp[0].metaNamePtr;
			yyvsp[-8].grantRevokeNode->isDdl = true;
		}
break;
case 122:
  if (!yytrial)
{
			yyvsp[-8].grantRevokeNode->object = newNode<GranteeClause>(obj_database, get_object_name(obj_database));
			yyvsp[-8].grantRevokeNode->grantAdminOption = yyvsp[-7].boolVal;
			yyvsp[-8].grantRevokeNode->grantor = yyvsp[0].metaNamePtr;
			yyvsp[-8].grantRevokeNode->isDdl = true;
		}
break;
case 123:
yyval.grantRevokeNode = NOTRIAL(yyvsp[-1].grantRevokeNode);
break;
case 124:
  if (!yytrial)
{
			yyvsp[-7].grantRevokeNode->grantAdminOption = yyvsp[-6].boolVal;
			yyvsp[-7].grantRevokeNode->grantor = yyvsp[0].metaNamePtr;
		}
break;
case 126:
  if (!yytrial)
{ yyval.boolVal = false; }
break;
case 127:
  if (!yytrial)
{ yyval.boolVal = true; }
break;
case 128:
  if (!yytrial)
{ yyval.boolVal = false; }
break;
case 129:
  if (!yytrial)
{ yyval.boolVal = true; }
break;
case 130:
yyval.granteeArray = yyvsp[0].granteeArray;
break;
case 133:
yyval.granteeArray = yyvsp[-2].granteeArray;
break;
case 136:
  if (!yytrial)
{ yyvsp[-2].granteeArray->add(GranteeClause(obj_procedure, *yyvsp[0].metaNamePtr)); }
break;
case 137:
  if (!yytrial)
{ yyvsp[-2].granteeArray->add(GranteeClause(obj_udf, *yyvsp[0].metaNamePtr)); }
break;
case 138:
  if (!yytrial)
{ yyvsp[-2].granteeArray->add(GranteeClause(obj_package_header, *yyvsp[0].metaNamePtr)); }
break;
case 139:
  if (!yytrial)
{ yyvsp[-2].granteeArray->add(GranteeClause(obj_trigger, *yyvsp[0].metaNamePtr)); }
break;
case 140:
  if (!yytrial)
{ yyvsp[-2].granteeArray->add(GranteeClause(obj_view, *yyvsp[0].metaNamePtr)); }
break;
case 141:
  if (!yytrial)
{ yyvsp[-2].granteeArray->add(GranteeClause(obj_sql_role, *yyvsp[0].metaNamePtr)); }
break;
case 142:
  if (!yytrial)
{ yyvsp[-3].granteeArray->add(GranteeClause(obj_privilege, *yyvsp[0].metaNamePtr)); }
break;
case 143:
  if (!yytrial)
{ yyvsp[-1].granteeArray->add(GranteeClause(obj_user_or_role, *yyvsp[0].metaNamePtr)); }
break;
case 144:
  if (!yytrial)
{ yyvsp[-2].granteeArray->add(GranteeClause(obj_user, *yyvsp[0].metaNamePtr)); }
break;
case 145:
  if (!yytrial)
{ yyvsp[-2].granteeArray->add(GranteeClause(obj_user_group, *yyvsp[0].metaNamePtr)); }
break;
case 147:
yyval.grantRevokeNode = yyvsp[-2].grantRevokeNode;
break;
case 149:
  if (!yytrial)
{
			yyvsp[-1].grantRevokeNode->roles.add(GranteeClause(obj_sql_role, *yyvsp[0].metaNamePtr));
			yyvsp[-1].grantRevokeNode->defaultRoles.add(false);
		}
break;
case 150:
  if (!yytrial)
{
			yyvsp[-2].grantRevokeNode->roles.add(GranteeClause(obj_sql_role, *yyvsp[0].metaNamePtr));
			yyvsp[-2].grantRevokeNode->defaultRoles.add(true);
		}
break;
case 153:
  if (!yytrial)
{ yyvsp[-1].granteeArray->add(GranteeClause(obj_user_or_role, *yyvsp[0].metaNamePtr)); }
break;
case 154:
  if (!yytrial)
{ yyvsp[-2].granteeArray->add(GranteeClause(obj_user, *yyvsp[0].metaNamePtr)); }
break;
case 155:
  if (!yytrial)
{ yyvsp[-2].granteeArray->add(GranteeClause(obj_sql_role, *yyvsp[0].metaNamePtr)); }
break;
case 156:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].ddlNode;}
break;
case 157:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].ddlNode; }
break;
case 158:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].createAlterFunctionNode; }
break;
case 159:
  if (!yytrial)
{ yyval.createAlterFunctionNode = newNode<CreateAlterFunctionNode>(*yyvsp[0].metaNamePtr); }
break;
case 160:
yyval.parametersClause = NOTRIAL(&yyvsp[0].createAlterFunctionNode->parameters);
break;
case 161:
yyval.createAlterFunctionNode = yyvsp[-3].createAlterFunctionNode;
break;
case 162:
  if (!yytrial)
{
				yyval.createAlterFunctionNode = yyvsp[-9].createAlterFunctionNode;
				yyval.createAlterFunctionNode->external = newNode<ExternalClause>();
				yyval.createAlterFunctionNode->external->name = *yyvsp[-2].stringPtr;
				yyval.createAlterFunctionNode->external->udfModule = *yyvsp[0].stringPtr;
			}
break;
case 164:
  if (!yytrial)
{
			yyval.legacyField = newNode<dsql_fld>();
			yyval.legacyField->dtype = dtype_blob;
			yyval.legacyField->length = sizeof(bid);
		}
break;
case 165:
  if (!yytrial)
{
			yyval.legacyField = newNode<dsql_fld>();
			yyval.legacyField->dtype = dtype_cstring;
			yyval.legacyField->charLength = (USHORT) yyvsp[-2].int32Val;
			if (yyvsp[0].metaNamePtr)
				yyval.legacyField->charSet = *yyvsp[0].metaNamePtr;
		}
break;
case 167:
yyval.parametersClause = yyvsp[0].parametersClause;
break;
case 169:
yyval.parametersClause = yyvsp[-1].parametersClause;
break;
case 172:
yyval.parametersClause = yyvsp[-2].parametersClause;
break;
case 174:
  if (!yytrial)
{
			yyvsp[-2].parametersClause->add(newNode<ParameterClause>(yyvsp[-1].legacyField, MetaName()));
			yyvsp[-2].parametersClause->back()->udfMechanism = yyvsp[0].nullableIntVal;
		}
break;
case 175:
  if (!yytrial)
{ yyval.nullableIntVal = Nullable<int>::empty(); }
break;
case 176:
  if (!yytrial)
{ yyval.nullableIntVal = Nullable<int>::val(FUN_descriptor); }
break;
case 177:
  if (!yytrial)
{ yyval.nullableIntVal = Nullable<int>::val(FUN_scalar_array); }
break;
case 178:
  if (!yytrial)
{ yyval.nullableIntVal = Nullable<int>::val(FUN_ref_with_null); }
break;
case 179:
yyval.createAlterFunctionNode = yyvsp[0].createAlterFunctionNode;
break;
case 181:
yyval.createAlterFunctionNode = yyvsp[-1].createAlterFunctionNode;
break;
case 183:
  if (!yytrial)
{
			yyvsp[-2].createAlterFunctionNode->returnType = newNode<ParameterClause>(yyvsp[-1].legacyField, MetaName());
			yyvsp[-2].createAlterFunctionNode->returnType->udfMechanism = yyvsp[0].int32Val;
		}
break;
case 184:
  if (!yytrial)
{ yyvsp[-2].createAlterFunctionNode->udfReturnPos = yyvsp[0].int32Val; }
break;
case 185:
  if (!yytrial)
{ yyval.int32Val = FUN_reference; }
break;
case 186:
  if (!yytrial)
{ yyval.int32Val = FUN_value; }
break;
case 187:
  if (!yytrial)
{ yyval.int32Val = FUN_descriptor; }
break;
case 188:
  if (!yytrial)
{ yyval.int32Val = -1 * FUN_reference; }
break;
case 189:
  if (!yytrial)
{ yyval.int32Val = -1 * FUN_descriptor; }
break;
case 190:
  if (!yytrial)
{
				CreateFilterNode* node = newNode<CreateFilterNode>(*yyvsp[-8].metaNamePtr);
				node->inputFilter = yyvsp[-6].filterNameNumber;
				node->outputFilter = yyvsp[-4].filterNameNumber;
				node->entryPoint = *yyvsp[-2].stringPtr;
				node->moduleName = *yyvsp[0].stringPtr;
				yyval.ddlNode = node;
			}
break;
case 191:
  if (!yytrial)
{ yyval.filterNameNumber = newNode<CreateFilterNode::NameNumber>(*yyvsp[0].metaNamePtr); }
break;
case 192:
  if (!yytrial)
{ yyval.filterNameNumber = newNode<CreateFilterNode::NameNumber>(yyvsp[0].int32Val); }
break;
case 193:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].ddlNode; }
break;
case 194:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].createAlterExceptionNode; }
break;
case 195:
  if (!yytrial)
{
				CreateIndexNode* node = newNode<CreateIndexNode>(*yyvsp[-2].metaNamePtr);
				node->unique = yyvsp[-5].boolVal;
				node->descending = yyvsp[-4].boolVal;
				node->relation = yyvsp[0].relSourceNode;
				yyval.ddlNode = node;
			}
break;
case 196:
yyval.createIndexNode = static_cast<CreateIndexNode*>(yyvsp[0].ddlNode);
break;
case 197:
  if (!yytrial)
{
				yyval.ddlNode = yyvsp[-2].ddlNode;
			}
break;
case 198:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].createAlterFunctionNode; }
break;
case 199:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].createAlterProcedureNode; }
break;
case 200:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].createRelationNode; }
break;
case 201:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].createRelationNode; }
break;
case 202:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].createAlterTriggerNode; }
break;
case 203:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].createAlterViewNode; }
break;
case 204:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].createAlterSequenceNode; }
break;
case 205:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].createAlterSequenceNode; }
break;
case 206:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].alterDatabaseNode; }
break;
case 207:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].createDomainNode; }
break;
case 208:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].createShadowNode; }
break;
case 209:
  if (!yytrial)
{ yyvsp[0].createAlterRoleNode->createFlag = true; yyval.ddlNode = yyvsp[0].createAlterRoleNode; }
break;
case 210:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].createCollationNode; }
break;
case 211:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].createAlterUserNode; }
break;
case 212:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].createAlterPackageNode; }
break;
case 213:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].createPackageBodyNode; }
break;
case 214:
yyval.boolVal = false;
break;
case 215:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].mappingNode; }
break;
case 216:
yyval.boolVal = true;
break;
case 217:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].mappingNode; }
break;
case 218:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].ddlNode; }
break;
case 219:
  if (!yytrial)
{ yyval.ddlNode = newNode<RecreateProcedureNode>(yyvsp[0].createAlterProcedureNode); }
break;
case 220:
  if (!yytrial)
{ yyval.ddlNode = newNode<RecreateFunctionNode>(yyvsp[0].createAlterFunctionNode); }
break;
case 221:
  if (!yytrial)
{ yyval.ddlNode = newNode<RecreateTableNode>(yyvsp[0].createRelationNode); }
break;
case 222:
  if (!yytrial)
{ yyval.ddlNode = newNode<RecreateTableNode>(yyvsp[0].createRelationNode); }
break;
case 223:
  if (!yytrial)
{ yyval.ddlNode = newNode<RecreateViewNode>(yyvsp[0].createAlterViewNode); }
break;
case 224:
  if (!yytrial)
{ yyval.ddlNode = newNode<RecreateTriggerNode>(yyvsp[0].createAlterTriggerNode); }
break;
case 225:
  if (!yytrial)
{ yyval.ddlNode = newNode<RecreatePackageNode>(yyvsp[0].createAlterPackageNode); }
break;
case 226:
  if (!yytrial)
{ yyval.ddlNode = newNode<RecreatePackageBodyNode>(yyvsp[0].createPackageBodyNode); }
break;
case 227:
  if (!yytrial)
{ yyval.ddlNode = newNode<RecreateExceptionNode>(yyvsp[0].createAlterExceptionNode); }
break;
case 228:
  if (!yytrial)
{ yyval.ddlNode = newNode<RecreateSequenceNode>(yyvsp[0].createAlterSequenceNode); }
break;
case 229:
  if (!yytrial)
{ yyval.ddlNode = newNode<RecreateSequenceNode>(yyvsp[0].createAlterSequenceNode); }
break;
case 230:
  if (!yytrial)
{ yyval.ddlNode = newNode<RecreateUserNode>(yyvsp[0].createAlterUserNode); }
break;
case 231:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].ddlNode; }
break;
case 232:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].createAlterProcedureNode; }
break;
case 233:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].createAlterFunctionNode; }
break;
case 234:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].createAlterTriggerNode; }
break;
case 235:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].createAlterPackageNode; }
break;
case 236:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].ddlNode; }
break;
case 237:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].createAlterExceptionNode; }
break;
case 238:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].createAlterSequenceNode; }
break;
case 239:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].createAlterSequenceNode; }
break;
case 240:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].createAlterUserNode; }
break;
case 241:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].mappingNode; }
break;
case 242:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].mappingNode; }
break;
case 243:
  if (!yytrial)
{ yyval.createAlterExceptionNode = newNode<CreateAlterExceptionNode>(*yyvsp[-1].metaNamePtr, yyvsp[0].intlStringPtr->getString()); }
break;
case 244:
  if (!yytrial)
{
			CreateAlterExceptionNode* node = newNode<CreateAlterExceptionNode>(*yyvsp[-1].metaNamePtr, yyvsp[0].intlStringPtr->getString());
			node->alter = true;
			yyval.createAlterExceptionNode = node;
		}
break;
case 245:
  if (!yytrial)
{
			CreateAlterExceptionNode* node = newNode<CreateAlterExceptionNode>(*yyvsp[-1].metaNamePtr, yyvsp[0].intlStringPtr->getString());
			node->create = false;
			node->alter = true;
			yyval.createAlterExceptionNode = node;
		}
break;
case 246:
  if (!yytrial)
{ yyval.boolVal = false; }
break;
case 247:
  if (!yytrial)
{ yyval.boolVal = true; }
break;
case 248:
  if (!yytrial)
{ yyvsp[-1].createIndexNode->columns = yyvsp[0].valueListNode; }
break;
case 249:
  if (!yytrial)
{ yyvsp[-1].createIndexNode->columns = yyvsp[0].valueListNode; }
break;
case 250:
  if (!yytrial)
{
 			yyvsp[-4].createIndexNode->computed = newNode<ValueSourceClause>();
			yyvsp[-4].createIndexNode->computed->value = yyvsp[-1].valueExprNode;
			yyvsp[-4].createIndexNode->computed->source = makeParseStr(YYPOSNARG(2), YYPOSNARG(4));
		}
break;
case 251:
  if (!yytrial)
{
	 			yyval.createShadowNode = newNode<CreateShadowNode>(yyvsp[-4].int32Val);
		 		yyval.createShadowNode->manual = yyvsp[-3].boolVal;
		 		yyval.createShadowNode->conditional = yyvsp[-2].boolVal;
		 		yyval.createShadowNode->files.add(newNode<DbFileClause>(*yyvsp[-1].stringPtr));
		 		yyval.createShadowNode->files.front()->length = yyvsp[0].int32Val;
	 		}
break;
case 252:
yyval.dbFilesClause = NOTRIAL(&yyvsp[0].createShadowNode->files);
break;
case 253:
  if (!yytrial)
{ yyval.createShadowNode = yyvsp[-2].createShadowNode; }
break;
case 254:
  if (!yytrial)
{ yyval.boolVal = false; }
break;
case 255:
  if (!yytrial)
{ yyval.boolVal = true; }
break;
case 256:
  if (!yytrial)
{ yyval.boolVal = false; }
break;
case 257:
  if (!yytrial)
{ yyval.boolVal = false; }
break;
case 258:
  if (!yytrial)
{ yyval.boolVal = true; }
break;
case 259:
  if (!yytrial)
{ yyval.int32Val = 0; }
break;
case 260:
  if (!yytrial)
{ yyval.int32Val = yyvsp[-1].int32Val; }
break;
case 262:
yyval.dbFilesClause = yyvsp[0].dbFilesClause;
break;
case 264:
  if (!yytrial)
{ yyvsp[-1].dbFilesClause->add(yyvsp[0].dbFileClause); }
break;
case 265:
  if (!yytrial)
{ yyvsp[-2].dbFilesClause->add(yyvsp[0].dbFileClause); }
break;
case 266:
  if (!yytrial)
{
				yyvsp[-1].legacyField->fld_name = *yyvsp[-3].metaNamePtr;
				yyval.createDomainNode = newNode<CreateDomainNode>(
					newNode<ParameterClause>(yyvsp[-1].legacyField, MetaName(), yyvsp[0].valueSourceClause));
			}
break;
case 267:
yyval.createDomainNode = yyvsp[0].createDomainNode;
break;
case 268:
  if (!yytrial)
{
				yyval.createDomainNode = yyvsp[-3].createDomainNode;
				if (yyvsp[0].metaNamePtr)
					yyval.createDomainNode->nameType->type->collate = *yyvsp[0].metaNamePtr;
			}
break;
case 272:
yyval.createDomainNode = yyvsp[-1].createDomainNode;
break;
case 274:
  if (!yytrial)
{ setClause(yyvsp[-1].createDomainNode->notNull, "NOT NULL"); }
break;
case 275:
  if (!yytrial)
{ setClause(yyvsp[-1].createDomainNode->check, "DOMAIN CHECK CONSTRAINT", yyvsp[0].boolSourceClause); }
break;
case 278:
  if (!yytrial)
{
			ValueSourceClause* clause = newNode<ValueSourceClause>();
			clause->value = yyvsp[0].valueExprNode;
			clause->source = makeParseStr(YYPOSNARG(1), YYPOSNARG(2));
			yyval.valueSourceClause = clause;
		}
break;
case 279:
  if (!yytrial)
{ yyval.valueSourceClause = NULL; }
break;
case 282:
  if (!yytrial)
{
			BoolSourceClause* clause = newNode<BoolSourceClause>();
			clause->value = yyvsp[-1].boolExprNode;
			clause->source = makeParseStr(YYPOSNARG(1), YYPOSNARG(4));
			yyval.boolSourceClause = clause;
		}
break;
case 283:
  if (!yytrial)
{ yyval.createAlterSequenceNode = newNode<CreateAlterSequenceNode>(*yyvsp[0].metaNamePtr); }
break;
case 284:
yyval.createAlterSequenceNode = yyvsp[0].createAlterSequenceNode;
break;
case 285:
  if (!yytrial)
{ yyval.createAlterSequenceNode = yyvsp[-2].createAlterSequenceNode; }
break;
case 287:
yyval.createAlterSequenceNode = yyvsp[-2].createAlterSequenceNode;
break;
case 291:
  if (!yytrial)
{
			setClause(yyvsp[-3].createAlterSequenceNode->value, "START WITH", yyvsp[0].int64Val);
			setClause(yyvsp[-3].createAlterSequenceNode->restartSpecified, "RESTART", true);
		}
break;
case 292:
  if (!yytrial)
{ setClause(yyvsp[-3].createAlterSequenceNode->step, "INCREMENT BY", yyvsp[0].int32Val); }
break;
case 295:
  if (!yytrial)
{
			CreateAlterSequenceNode* node = newNode<CreateAlterSequenceNode>(*yyvsp[0].metaNamePtr);
			node->alter = true;
			yyval.createAlterSequenceNode = node;
		}
break;
case 296:
  if (!yytrial)
{
			/* Remove this to implement CORE-5137*/
			if (!yyvsp[-2].createAlterSequenceNode->restartSpecified && !yyvsp[-2].createAlterSequenceNode->step.specified)
				yyerrorIncompleteCmd(YYPOSNARG(3));
			yyval.createAlterSequenceNode = yyvsp[-2].createAlterSequenceNode;
		}
break;
case 299:
  if (!yytrial)
{
			setClause(yyvsp[-1].createAlterSequenceNode->restartSpecified, "RESTART", true);
		}
break;
case 302:
  if (!yytrial)
{
			CreateAlterSequenceNode* node = newNode<CreateAlterSequenceNode>(*yyvsp[0].metaNamePtr);
			node->create = false;
			node->alter = true;
			yyval.createAlterSequenceNode = node;
		}
break;
case 303:
  if (!yytrial)
{
			if (!yyvsp[-2].createAlterSequenceNode->restartSpecified && !yyvsp[-2].createAlterSequenceNode->value.specified && !yyvsp[-2].createAlterSequenceNode->step.specified)
				yyerrorIncompleteCmd(YYPOSNARG(3));
			yyval.createAlterSequenceNode = yyvsp[-2].createAlterSequenceNode;
		}
break;
case 308:
  if (!yytrial)
{
			setClause(yyvsp[-2].createAlterSequenceNode->restartSpecified, "RESTART", true);
			setClause(yyvsp[-2].createAlterSequenceNode->value, "RESTART WITH", yyvsp[0].nullableInt64Val);
		}
break;
case 309:
  if (!yytrial)
{ yyval.nullableInt64Val = BaseNullable<SINT64>::empty(); }
break;
case 310:
  if (!yytrial)
{ yyval.nullableInt64Val = BaseNullable<SINT64>::val(yyvsp[0].int64Val); }
break;
case 311:
  if (!yytrial)
{
			CreateAlterSequenceNode* node = newNode<CreateAlterSequenceNode>(*yyvsp[-2].metaNamePtr);
			node->create = false;
			node->alter = true;
			node->legacy = true;
			node->restartSpecified = true;
			node->value = BaseNullable<SINT64>::val(yyvsp[0].int64Val);
			yyval.createAlterSequenceNode = node;
		}
break;
case 312:
  if (!yytrial)
{ yyval.int64Val = yyvsp[0].int32Val; }
break;
case 313:
  if (!yytrial)
{
			SINT64 signedNumber = (SINT64) yyvsp[0].scaledNumber.number;

			if (!yyvsp[0].scaledNumber.hex && yyvsp[0].scaledNumber.number > MAX_SINT64)
			{
				ERRD_post(
					Arg::Gds(isc_sqlerr) << Arg::Num(-104) <<
					Arg::Gds(isc_arith_except) <<
					Arg::Gds(isc_numeric_out_of_range));
			}

			yyval.int64Val = signedNumber;
		}
break;
case 314:
  if (!yytrial)
{
			SINT64 signedNumber = (SINT64) yyvsp[0].scaledNumber.number;

			if (yyvsp[0].scaledNumber.hex && signedNumber == MIN_SINT64)
				ERRD_post(Arg::Gds(isc_exception_integer_overflow));

			yyval.int64Val = -signedNumber;
		}
break;
case 315:
  if (!yytrial)
{
			yyval.int64Val = MIN_SINT64;
		}
break;
case 316:
  if (!yytrial)
{ yyval.createAlterRoleNode = newNode<CreateAlterRoleNode>(*yyvsp[0].metaNamePtr); }
break;
case 317:
yyval.createAlterRoleNode = yyvsp[0].createAlterRoleNode;
break;
case 318:
  if (!yytrial)
{ yyval.createAlterRoleNode = yyvsp[-2].createAlterRoleNode; }
break;
case 322:
yyval.createAlterRoleNode = yyvsp[-4].createAlterRoleNode;
break;
case 324:
  if (!yytrial)
{ yyvsp[-3].createAlterRoleNode->sysPrivDrop = true; }
break;
case 326:
yyval.createAlterRoleNode = yyvsp[-2].createAlterRoleNode;
break;
case 328:
  if (!yytrial)
{ yyvsp[-1].createAlterRoleNode->addPrivilege(yyvsp[0].metaNamePtr); }
break;
case 329:
  if (!yytrial)
{ yyval.createCollationNode = newNode<CreateCollationNode>(*yyvsp[-2].metaNamePtr, *yyvsp[0].metaNamePtr); }
break;
case 330:
yyval.createCollationNode = yyvsp[0].createCollationNode;
break;
case 331:
yyval.createCollationNode = yyvsp[-2].createCollationNode;
break;
case 332:
yyval.createCollationNode = yyvsp[-4].createCollationNode;
break;
case 333:
  if (!yytrial)
{ yyval.createCollationNode = yyvsp[-6].createCollationNode; }
break;
case 335:
  if (!yytrial)
{ yyvsp[-2].createCollationNode->fromName = *yyvsp[0].metaNamePtr; }
break;
case 336:
  if (!yytrial)
{ yyvsp[-5].createCollationNode->fromExternal = *yyvsp[-1].stringPtr; }
break;
case 340:
yyval.createCollationNode = yyvsp[-1].createCollationNode;
break;
case 345:
  if (!yytrial)
{ yyvsp[-2].createCollationNode->unsetAttribute(TEXTTYPE_ATTR_PAD_SPACE); }
break;
case 346:
  if (!yytrial)
{ yyvsp[-2].createCollationNode->setAttribute(TEXTTYPE_ATTR_PAD_SPACE); }
break;
case 347:
  if (!yytrial)
{ yyvsp[-2].createCollationNode->unsetAttribute(TEXTTYPE_ATTR_CASE_INSENSITIVE); }
break;
case 348:
  if (!yytrial)
{ yyvsp[-2].createCollationNode->setAttribute(TEXTTYPE_ATTR_CASE_INSENSITIVE); }
break;
case 349:
  if (!yytrial)
{ yyvsp[-2].createCollationNode->unsetAttribute(TEXTTYPE_ATTR_ACCENT_INSENSITIVE); }
break;
case 350:
  if (!yytrial)
{ yyvsp[-2].createCollationNode->setAttribute(TEXTTYPE_ATTR_ACCENT_INSENSITIVE); }
break;
case 352:
  if (!yytrial)
{
			const string& s = *yyvsp[0].stringPtr;
			yyvsp[-1].createCollationNode->specificAttributes.clear();
			yyvsp[-1].createCollationNode->specificAttributes.add((const UCHAR*) s.begin(), s.length());
		}
break;
case 353:
  if (!yytrial)
{ yyval.ddlNode = newNode<AlterCharSetNode>(*yyvsp[-4].metaNamePtr, *yyvsp[0].metaNamePtr); }
break;
case 354:
  if (!yytrial)
{ yyval.ddlNode = newNode<AlterEDSPoolSetNode>(AlterEDSPoolSetNode::POOL_SIZE, yyvsp[0].int32Val); }
break;
case 355:
  if (!yytrial)
{ yyval.ddlNode = newNode<AlterEDSPoolSetNode>(AlterEDSPoolSetNode::POOL_LIFETIME, yyvsp[-1].int32Val * yyvsp[0].intVal); }
break;
case 356:
  if (!yytrial)
{ yyval.ddlNode = newNode<AlterEDSPoolClearNode>(AlterEDSPoolClearNode::POOL_DB, yyvsp[0].intlStringPtr->getString()); }
break;
case 357:
  if (!yytrial)
{ yyval.ddlNode = newNode<AlterEDSPoolClearNode>(AlterEDSPoolClearNode::POOL_ALL); }
break;
case 358:
  if (!yytrial)
{ yyval.ddlNode = newNode<AlterEDSPoolClearNode>(AlterEDSPoolClearNode::POOL_OLDEST); }
break;
case 359:
  if (!yytrial)
{ yyval.intVal = 3600; }
break;
case 360:
  if (!yytrial)
{ yyval.intVal = 60; }
break;
case 361:
  if (!yytrial)
{ yyval.intVal = 1; }
break;
case 362:
  if (!yytrial)
{
				yyval.alterDatabaseNode = newNode<AlterDatabaseNode>();
				yyval.alterDatabaseNode->create = true;
			}
break;
case 363:
yyval.alterDatabaseNode = yyvsp[0].alterDatabaseNode;
break;
case 364:
yyval.alterDatabaseNode = yyvsp[-2].alterDatabaseNode;
break;
case 365:
  if (!yytrial)
{ yyval.alterDatabaseNode = yyvsp[-4].alterDatabaseNode; }
break;
case 372:
yyval.alterDatabaseNode = yyvsp[-1].alterDatabaseNode;
break;
case 381:
  if (!yytrial)
{ yyvsp[-4].alterDatabaseNode->createLength = yyvsp[-1].int32Val; }
break;
case 386:
  if (!yytrial)
{ yyvsp[-1].alterDatabaseNode->files.add(yyvsp[0].dbFileClause); }
break;
case 387:
  if (!yytrial)
{ yyvsp[-4].alterDatabaseNode->setDefaultCharSet = *yyvsp[0].metaNamePtr; }
break;
case 388:
  if (!yytrial)
{
			yyvsp[-6].alterDatabaseNode->setDefaultCharSet = *yyvsp[-2].metaNamePtr;
			yyvsp[-6].alterDatabaseNode->setDefaultCollation = *yyvsp[0].metaNamePtr;
		}
break;
case 389:
  if (!yytrial)
{ yyvsp[-3].alterDatabaseNode->differenceFile = *yyvsp[0].stringPtr; }
break;
case 390:
  if (!yytrial)
{
				DbFileClause* clause = newNode<DbFileClause>(*yyvsp[0].stringPtr);
				yyval.dbFileClause = clause;
			}
break;
case 391:
yyval.dbFileClause = yyvsp[0].dbFileClause;
break;
case 392:
  if (!yytrial)
{ yyval.dbFileClause = yyvsp[-2].dbFileClause; }
break;
case 396:
yyval.dbFileClause = yyvsp[-1].dbFileClause;
break;
case 398:
  if (!yytrial)
{ yyvsp[-3].dbFileClause->start = yyvsp[0].int32Val; }
break;
case 399:
  if (!yytrial)
{ yyvsp[-4].dbFileClause->length = yyvsp[-1].int32Val; }
break;
case 406:
  if (!yytrial)
{
				yyval.createRelationNode = newNode<CreateRelationNode>(yyvsp[-1].relSourceNode, yyvsp[0].stringPtr);
			}
break;
case 407:
yyval.createRelationNode = yyvsp[-1].createRelationNode;
break;
case 408:
yyval.relationNode = yyvsp[-4].createRelationNode;
break;
case 409:
  if (!yytrial)
{
				yyval.createRelationNode = yyvsp[-6].createRelationNode;
			}
break;
case 411:
yyval.relationNode = yyvsp[0].relationNode;
break;
case 412:
yyval.relationNode = yyvsp[-2].relationNode;
break;
case 414:
  if (!yytrial)
{ setClause(yyvsp[-1].relationNode->ssDefiner, "SQL SECURITY", yyvsp[0].boolVal); }
break;
case 415:
  if (!yytrial)
{ setClause(yyvsp[-1].relationNode->replicationState, "PUBLICATION", yyvsp[0].boolVal); }
break;
case 416:
  if (!yytrial)
{ yyval.boolVal = true; }
break;
case 417:
  if (!yytrial)
{ yyval.boolVal = false; }
break;
case 418:
  if (!yytrial)
{ yyval.nullableBoolVal = Nullable<bool>::empty(); }
break;
case 419:
  if (!yytrial)
{ yyval.nullableBoolVal = Nullable<bool>::val(yyvsp[0].boolVal); }
break;
case 420:
  if (!yytrial)
{ yyval.boolVal = true; }
break;
case 421:
  if (!yytrial)
{ yyval.boolVal = false; }
break;
case 422:
  if (!yytrial)
{
				yyval.createRelationNode = newNode<CreateRelationNode>(yyvsp[0].relSourceNode);
				yyval.createRelationNode->relationType = Nullable<rel_t>::empty();
			}
break;
case 423:
yyval.createRelationNode = yyvsp[-4].createRelationNode;
break;
case 424:
  if (!yytrial)
{
				yyval.createRelationNode = yyvsp[-6].createRelationNode;
				if (!yyval.createRelationNode->relationType.specified)
					yyval.createRelationNode->relationType = rel_global_temp_delete;
			}
break;
case 425:
yyval.createRelationNode = yyvsp[0].createRelationNode;
break;
case 427:
yyval.createRelationNode = yyvsp[-2].createRelationNode;
break;
case 430:
  if (!yytrial)
{ setClause(yyvsp[-1].createRelationNode->ssDefiner, "SQL SECURITY", yyvsp[0].nullableBoolVal); }
break;
case 431:
  if (!yytrial)
{ setClause(yyvsp[-4].createRelationNode->relationType, "ON COMMIT DELETE ROWS", rel_global_temp_delete); }
break;
case 432:
  if (!yytrial)
{ setClause(yyvsp[-4].createRelationNode->relationType, "ON COMMIT PRESERVE ROWS", rel_global_temp_preserve); }
break;
case 433:
  if (!yytrial)
{ yyval.stringPtr = NULL; }
break;
case 434:
  if (!yytrial)
{ yyval.stringPtr = yyvsp[0].stringPtr; }
break;
case 435:
  if (!yytrial)
{ yyval.stringPtr = yyvsp[0].stringPtr; }
break;
case 438:
yyval.relationNode = yyvsp[0].createRelationNode;
break;
case 441:
  if (!yytrial)
{
				RelationNode::AddColumnClause* clause = yyval.addColumnClause =
					newNode<RelationNode::AddColumnClause>();
				clause->field = yyvsp[-1].legacyField;
				clause->field->fld_name = *yyvsp[-2].metaNamePtr;
				clause->defaultValue = yyvsp[0].valueSourceClause;
				yyvsp[-3].relationNode->clauses.add(clause);
			}
break;
case 442:
yyval.addColumnClause = NOTRIAL(yyvsp[0].addColumnClause);
break;
case 443:
  if (!yytrial)
{
				if (yyvsp[0].metaNamePtr)
					yyvsp[-3].addColumnClause->collate = *yyvsp[0].metaNamePtr;
			}
break;
case 444:
  if (!yytrial)
{
				RelationNode::AddColumnClause* clause = yyval.addColumnClause =
					newNode<RelationNode::AddColumnClause>();
				clause->field = yyvsp[-1].legacyField;
				clause->field->fld_name = *yyvsp[-2].metaNamePtr;
				clause->identityOptions = yyvsp[0].identityOptions;
				yyvsp[-3].relationNode->clauses.add(clause);
			}
break;
case 445:
  if (!yytrial)
{
				if (yyvsp[0].metaNamePtr)
					yyvsp[-3].addColumnClause->collate = *yyvsp[0].metaNamePtr;
			}
break;
case 446:
  if (!yytrial)
{
			RelationNode::AddColumnClause* clause = newNode<RelationNode::AddColumnClause>();
			clause->field = yyvsp[-1].legacyField;
			clause->field->fld_name = *yyvsp[-2].metaNamePtr;
			clause->computed = yyvsp[0].valueSourceClause;
			yyvsp[-3].relationNode->clauses.add(clause);
			clause->field->flags |= FLD_computed;
		}
break;
case 447:
  if (!yytrial)
{
			RelationNode::AddColumnClause* clause = newNode<RelationNode::AddColumnClause>();
			clause->field = newNode<dsql_fld>();
			clause->field->fld_name = *yyvsp[-1].metaNamePtr;
			clause->computed = yyvsp[0].valueSourceClause;
			yyvsp[-2].relationNode->clauses.add(clause);
			clause->field->flags |= FLD_computed;
		}
break;
case 448:
  if (!yytrial)
{ yyval.identityOptions = newNode<RelationNode::IdentityOptions>(yyvsp[-2].identityType); }
break;
case 449:
yyval.identityOptions = yyvsp[0].identityOptions;
break;
case 450:
  if (!yytrial)
{ yyval.identityOptions = yyvsp[-2].identityOptions; }
break;
case 451:
  if (!yytrial)
{ yyval.identityType = IDENT_TYPE_BY_DEFAULT; }
break;
case 452:
  if (!yytrial)
{ yyval.identityType = IDENT_TYPE_ALWAYS; }
break;
case 454:
yyval.identityOptions = yyvsp[-1].identityOptions;
break;
case 458:
  if (!yytrial)
{ setClause(yyvsp[-3].identityOptions->startValue, "START WITH", yyvsp[0].int64Val); }
break;
case 459:
  if (!yytrial)
{ setClause(yyvsp[-3].identityOptions->increment, "INCREMENT BY", yyvsp[0].int32Val); }
break;
case 460:
  if (!yytrial)
{
			ValueSourceClause* clause = newNode<ValueSourceClause>();
			clause->value = yyvsp[-1].valueExprNode;
			clause->source = makeParseStr(YYPOSNARG(2), YYPOSNARG(4));
			yyval.valueSourceClause = clause;
		}
break;
case 467:
  if (!yytrial)
{
			yyval.legacyField = newNode<dsql_fld>();
			yyval.legacyField->typeOfName = *yyvsp[0].metaNamePtr;
		}
break;
case 468:
  if (!yytrial)
{ yyval.metaNamePtr = NULL; }
break;
case 469:
  if (!yytrial)
{ yyval.metaNamePtr = yyvsp[0].metaNamePtr; }
break;
case 471:
  if (!yytrial)
{
			yyval.legacyField = newNode<dsql_fld>();
			yyval.legacyField->typeOfName = *yyvsp[0].metaNamePtr;
		}
break;
case 472:
  if (!yytrial)
{
			yyval.legacyField = newNode<dsql_fld>();
			yyval.legacyField->typeOfTable = *yyvsp[-2].metaNamePtr;
			yyval.legacyField->typeOfName = *yyvsp[0].metaNamePtr;
		}
break;
case 473:
  if (!yytrial)
{
			yyval.legacyField = newNode<dsql_fld>();
			yyval.legacyField->typeOfName = *yyvsp[0].metaNamePtr;
			yyval.legacyField->fullDomain = true;
		}
break;
case 474:
  if (!yytrial)
{ yyval.valueExprNode = yyvsp[0].valueExprNode; }
break;
case 475:
  if (!yytrial)
{ yyval.valueExprNode = yyvsp[0].valueExprNode; }
break;
case 476:
  if (!yytrial)
{ yyval.valueExprNode = yyvsp[0].valueExprNode; }
break;
case 477:
  if (!yytrial)
{ yyval.valueExprNode = yyvsp[0].valueExprNode; }
break;
case 478:
  if (!yytrial)
{ yyval.valueExprNode = yyvsp[0].valueExprNode; }
break;
case 479:
  if (!yytrial)
{ yyval.valueExprNode = yyvsp[0].valueExprNode; }
break;
case 481:
yyval.addColumnClause = yyvsp[0].addColumnClause;
break;
case 484:
yyval.addColumnClause = yyvsp[-1].addColumnClause;
break;
case 486:
  if (!yytrial)
{
			if (yyvsp[-2].metaNamePtr)
				yyvsp[-3].addColumnClause->constraints.back()->name = *yyvsp[-2].metaNamePtr;
		}
break;
case 487:
  if (!yytrial)
{
			setClause(yyvsp[-1].addColumnClause->notNullSpecified, "NOT NULL");
			RelationNode::AddConstraintClause& constraint = yyvsp[-1].addColumnClause->constraints.add();
			constraint.constraintType = RelationNode::AddConstraintClause::CTYPE_NOT_NULL;
		}
break;
case 488:
  if (!yytrial)
{
			RelationNode::AddConstraintClause& constraint = yyvsp[-1].addColumnClause->constraints.add();
			constraint.constraintType = RelationNode::AddConstraintClause::CTYPE_CHECK;
			constraint.check = yyvsp[0].boolSourceClause;
		}
break;
case 489:
  if (!yytrial)
{
			RelationNode::AddConstraintClause& constraint = yyvsp[-5].addColumnClause->constraints.add();
			constraint.constraintType = RelationNode::AddConstraintClause::CTYPE_FK;

			constraint.columns.add(yyvsp[-5].addColumnClause->field->fld_name);
			constraint.refRelation = *yyvsp[-3].metaNamePtr;
			constraint.refAction = yyvsp[-1].refActionClause;

			const ValueListNode* refColumns = yyvsp[-2].valueListNode;
			if (refColumns)
			{
				const NestConst<ValueExprNode>* ptr = refColumns->items.begin();

				for (const NestConst<ValueExprNode>* const end = refColumns->items.end(); ptr != end; ++ptr)
					constraint.refColumns.add(nodeAs<FieldNode>(*ptr)->dsqlName);
			}

			constraint.index = yyvsp[0].indexConstraintClause;
		}
break;
case 490:
  if (!yytrial)
{
			RelationNode::AddConstraintClause& constraint = yyvsp[-2].addColumnClause->constraints.add();
			constraint.constraintType = RelationNode::AddConstraintClause::CTYPE_UNIQUE;
			constraint.index = yyvsp[0].indexConstraintClause;
		}
break;
case 491:
  if (!yytrial)
{
			RelationNode::AddConstraintClause& constraint = yyvsp[-3].addColumnClause->constraints.add();
			constraint.constraintType = RelationNode::AddConstraintClause::CTYPE_PK;
			constraint.index = yyvsp[0].indexConstraintClause;
		}
break;
case 492:
yyval.relationNode = yyvsp[-1].relationNode;
break;
case 493:
  if (!yytrial)
{
			if (yyvsp[-2].metaNamePtr)
			{
				static_cast<RelationNode::AddConstraintClause*>(
					yyvsp[-3].relationNode->clauses.back().getObject())->name = *yyvsp[-2].metaNamePtr;
			}
		}
break;
case 494:
  if (!yytrial)
{ yyval.metaNamePtr = NULL; }
break;
case 495:
  if (!yytrial)
{ yyval.metaNamePtr = yyvsp[0].metaNamePtr; }
break;
case 496:
  if (!yytrial)
{
			RelationNode::AddConstraintClause& constraint = *newNode<RelationNode::AddConstraintClause>();
			constraint.constraintType = RelationNode::AddConstraintClause::CTYPE_UNIQUE;

			const ValueListNode* columns = yyvsp[-1].valueListNode;
			const NestConst<ValueExprNode>* ptr = columns->items.begin();

			for (const NestConst<ValueExprNode>* const end = columns->items.end(); ptr != end; ++ptr)
				constraint.columns.add(nodeAs<FieldNode>(*ptr)->dsqlName);

			constraint.index = yyvsp[0].indexConstraintClause;

			yyvsp[-3].relationNode->clauses.add(&constraint);
		}
break;
case 497:
  if (!yytrial)
{
			RelationNode::AddConstraintClause& constraint = *newNode<RelationNode::AddConstraintClause>();
			constraint.constraintType = RelationNode::AddConstraintClause::CTYPE_PK;

			const ValueListNode* columns = yyvsp[-1].valueListNode;
			const NestConst<ValueExprNode>* ptr = columns->items.begin();

			for (const NestConst<ValueExprNode>* const end = columns->items.end(); ptr != end; ++ptr)
				constraint.columns.add(nodeAs<FieldNode>(*ptr)->dsqlName);

			constraint.index = yyvsp[0].indexConstraintClause;

			yyvsp[-4].relationNode->clauses.add(&constraint);
		}
break;
case 498:
  if (!yytrial)
{
			RelationNode::AddConstraintClause& constraint = *newNode<RelationNode::AddConstraintClause>();
			constraint.constraintType = RelationNode::AddConstraintClause::CTYPE_FK;

			const ValueListNode* columns = yyvsp[-5].valueListNode;
			const NestConst<ValueExprNode>* ptr = columns->items.begin();

			for (const NestConst<ValueExprNode>* const end = columns->items.end(); ptr != end; ++ptr)
				constraint.columns.add(nodeAs<FieldNode>(*ptr)->dsqlName);

			constraint.refRelation = *yyvsp[-3].metaNamePtr;
			constraint.refAction = yyvsp[-1].refActionClause;

			const ValueListNode* refColumns = yyvsp[-2].valueListNode;
			if (refColumns)
			{
				const NestConst<ValueExprNode>* ptr = refColumns->items.begin();

				for (const NestConst<ValueExprNode>* const end = refColumns->items.end(); ptr != end; ++ptr)
					constraint.refColumns.add(nodeAs<FieldNode>(*ptr)->dsqlName);
			}

			constraint.index = yyvsp[0].indexConstraintClause;

			yyvsp[-8].relationNode->clauses.add(&constraint);
		}
break;
case 499:
  if (!yytrial)
{
			RelationNode::AddConstraintClause* constraint = newNode<RelationNode::AddConstraintClause>();
			constraint->constraintType = RelationNode::AddConstraintClause::CTYPE_CHECK;
			constraint->check = yyvsp[0].boolSourceClause;
			yyvsp[-1].relationNode->clauses.add(constraint);
		}
break;
case 500:
  if (!yytrial)
{ yyval.indexConstraintClause = newNode<RelationNode::IndexConstraintClause>(); }
break;
case 501:
  if (!yytrial)
{
			RelationNode::IndexConstraintClause* clause = yyval.indexConstraintClause =
				newNode<RelationNode::IndexConstraintClause>();
			clause->descending = yyvsp[-2].boolVal;
			clause->name = *yyvsp[0].metaNamePtr;
		}
break;
case 502:
  if (!yytrial)
{ yyval.refActionClause = NULL; }
break;
case 503:
  if (!yytrial)
{ yyval.refActionClause = newNode<RelationNode::RefActionClause>(yyvsp[0].uintVal, 0); }
break;
case 504:
  if (!yytrial)
{ yyval.refActionClause = newNode<RelationNode::RefActionClause>(0, yyvsp[0].uintVal); }
break;
case 505:
  if (!yytrial)
{ yyval.refActionClause = newNode<RelationNode::RefActionClause>(yyvsp[0].uintVal, yyvsp[-1].uintVal); }
break;
case 506:
  if (!yytrial)
{ yyval.refActionClause = newNode<RelationNode::RefActionClause>(yyvsp[-1].uintVal, yyvsp[0].uintVal); }
break;
case 507:
  if (!yytrial)
{ yyval.uintVal = yyvsp[0].uintVal;}
break;
case 508:
  if (!yytrial)
{ yyval.uintVal = yyvsp[0].uintVal;}
break;
case 509:
  if (!yytrial)
{ yyval.uintVal = RelationNode::RefActionClause::ACTION_CASCADE; }
break;
case 510:
  if (!yytrial)
{ yyval.uintVal = RelationNode::RefActionClause::ACTION_SET_DEFAULT; }
break;
case 511:
  if (!yytrial)
{ yyval.uintVal = RelationNode::RefActionClause::ACTION_SET_NULL; }
break;
case 512:
  if (!yytrial)
{ yyval.uintVal = RelationNode::RefActionClause::ACTION_NONE; }
break;
case 515:
  if (!yytrial)
{
			yyval.createAlterProcedureNode = yyvsp[-4].createAlterProcedureNode;
			yyval.createAlterProcedureNode->ssDefiner = yyvsp[-3].nullableBoolVal;
			yyval.createAlterProcedureNode->source = makeParseStr(YYPOSNARG(4), YYPOSNARG(5));
			yyval.createAlterProcedureNode->localDeclList = yyvsp[-1].compoundStmtNode;
			yyval.createAlterProcedureNode->body = yyvsp[0].stmtNode;
		}
break;
case 516:
  if (!yytrial)
{
			yyval.createAlterProcedureNode = yyvsp[-2].createAlterProcedureNode;
			yyval.createAlterProcedureNode->external = yyvsp[-1].externalClause;
			if (yyvsp[0].stringPtr)
				yyval.createAlterProcedureNode->source = *yyvsp[0].stringPtr;
		}
break;
case 517:
  if (!yytrial)
{ yyval.createAlterProcedureNode = newNode<CreateAlterProcedureNode>(*yyvsp[0].metaNamePtr); }
break;
case 518:
yyval.parametersClause = NOTRIAL(&yyvsp[0].createAlterProcedureNode->parameters);
break;
case 519:
yyval.parametersClause = NOTRIAL(&yyvsp[-2].createAlterProcedureNode->returns);
break;
case 520:
  if (!yytrial)
{ yyval.createAlterProcedureNode = yyvsp[-4].createAlterProcedureNode; }
break;
case 521:
  if (!yytrial)
{
			yyval.createAlterProcedureNode = yyvsp[0].createAlterProcedureNode;
			yyval.createAlterProcedureNode->alter = true;
			yyval.createAlterProcedureNode->create = false;
		}
break;
case 522:
  if (!yytrial)
{
			yyval.createAlterProcedureNode = yyvsp[0].createAlterProcedureNode;
			yyval.createAlterProcedureNode->alter = true;
		}
break;
case 530:
  if (!yytrial)
{ yyvsp[-3].parametersClause->add(newNode<ParameterClause>(yyvsp[-2].legacyField, optName(yyvsp[-1].metaNamePtr), yyvsp[0].valueSourceClause)); }
break;
case 533:
  if (!yytrial)
{ yyvsp[-2].parametersClause->add(newNode<ParameterClause>(yyvsp[-1].legacyField, optName(yyvsp[0].metaNamePtr))); }
break;
case 534:
  if (!yytrial)
{
			yyval.legacyField = yyvsp[0].legacyField;
			yyval.legacyField->fld_name = *yyvsp[-1].metaNamePtr;
		}
break;
case 535:
  if (!yytrial)
{ yyval.valueSourceClause = NULL; }
break;
case 536:
  if (!yytrial)
{
			ValueSourceClause* clause = newNode<ValueSourceClause>();
			clause->value = yyvsp[0].valueExprNode;
			clause->source = makeParseStr(YYPOSNARG(1), YYPOSNARG(2));
			yyval.valueSourceClause = clause;
		}
break;
case 537:
  if (!yytrial)
{
			ValueSourceClause* clause = newNode<ValueSourceClause>();
			clause->value = yyvsp[0].valueExprNode;
			clause->source = makeParseStr(YYPOSNARG(1), YYPOSNARG(2));
			yyval.valueSourceClause = clause;
		}
break;
case 540:
  if (!yytrial)
{
			yyval.createAlterFunctionNode = yyvsp[-4].createAlterFunctionNode;
			yyval.createAlterFunctionNode->ssDefiner = yyvsp[-3].nullableBoolVal;
			yyval.createAlterFunctionNode->source = makeParseStr(YYPOSNARG(4), YYPOSNARG(5));
			yyval.createAlterFunctionNode->localDeclList = yyvsp[-1].compoundStmtNode;
			yyval.createAlterFunctionNode->body = yyvsp[0].stmtNode;
		}
break;
case 541:
  if (!yytrial)
{
			yyval.createAlterFunctionNode = yyvsp[-2].createAlterFunctionNode;
			yyval.createAlterFunctionNode->external = yyvsp[-1].externalClause;
			if (yyvsp[0].stringPtr)
				yyval.createAlterFunctionNode->source = *yyvsp[0].stringPtr;
		}
break;
case 542:
  if (!yytrial)
{ yyval.createAlterFunctionNode = newNode<CreateAlterFunctionNode>(*yyvsp[0].metaNamePtr); }
break;
case 543:
  if (!yytrial)
{
				yyval.createAlterFunctionNode = yyvsp[-6].createAlterFunctionNode;
				yyval.createAlterFunctionNode->returnType = newNode<ParameterClause>(yyvsp[-2].legacyField, optName(yyvsp[-1].metaNamePtr));
				yyval.createAlterFunctionNode->deterministic = yyvsp[0].boolVal;
			}
break;
case 544:
  if (!yytrial)
{ yyval.boolVal = false; }
break;
case 545:
  if (!yytrial)
{ yyval.boolVal = false; }
break;
case 546:
  if (!yytrial)
{ yyval.boolVal = true; }
break;
case 547:
  if (!yytrial)
{
			yyval.externalClause = newNode<ExternalClause>();
			yyval.externalClause->name = *yyvsp[-2].stringPtr;
			yyval.externalClause->engine = *yyvsp[0].metaNamePtr;
		}
break;
case 548:
  if (!yytrial)
{
			yyval.externalClause = newNode<ExternalClause>();
			yyval.externalClause->engine = *yyvsp[0].metaNamePtr;
		}
break;
case 549:
  if (!yytrial)
{ yyval.stringPtr = NULL; }
break;
case 550:
  if (!yytrial)
{ yyval.stringPtr = yyvsp[0].stringPtr; }
break;
case 551:
  if (!yytrial)
{
			yyval.createAlterFunctionNode = yyvsp[0].createAlterFunctionNode;
			yyval.createAlterFunctionNode->alter = true;
			yyval.createAlterFunctionNode->create = false;
		}
break;
case 552:
  if (!yytrial)
{
			yyval.createAlterFunctionNode = yyvsp[0].createAlterFunctionNode;
			yyval.createAlterFunctionNode->alter = true;
		}
break;
case 553:
  if (!yytrial)
{
			CreateAlterPackageNode* node = newNode<CreateAlterPackageNode>(*yyvsp[-5].metaNamePtr);
			node->ssDefiner = yyvsp[-4].nullableBoolVal;
			node->source = makeParseStr(YYPOSNARG(4), YYPOSNARG(6));
			node->items = yyvsp[-1].packageItems;
			yyval.createAlterPackageNode = node;
		}
break;
case 555:
  if (!yytrial)
{ yyval.packageItems = newNode<Array<CreateAlterPackageNode::Item> >(); }
break;
case 556:
  if (!yytrial)
{
			yyval.packageItems = newNode<Array<CreateAlterPackageNode::Item> >();
			yyval.packageItems->add(yyvsp[0].packageItem);
		}
break;
case 557:
  if (!yytrial)
{
			yyval.packageItems = yyvsp[-1].packageItems;
			yyval.packageItems->add(yyvsp[0].packageItem);
		}
break;
case 558:
  if (!yytrial)
{ yyval.packageItem = CreateAlterPackageNode::Item::create(yyvsp[-1].createAlterFunctionNode); }
break;
case 559:
  if (!yytrial)
{ yyval.packageItem = CreateAlterPackageNode::Item::create(yyvsp[-1].createAlterProcedureNode); }
break;
case 560:
  if (!yytrial)
{
			yyval.createAlterPackageNode = yyvsp[0].createAlterPackageNode;
			yyval.createAlterPackageNode->alter = true;
			yyval.createAlterPackageNode->create = false;
		}
break;
case 561:
  if (!yytrial)
{
			yyval.createAlterPackageNode = yyvsp[0].createAlterPackageNode;
			yyval.createAlterPackageNode->alter = true;
		}
break;
case 562:
  if (!yytrial)
{
			CreatePackageBodyNode* node = newNode<CreatePackageBodyNode>(*yyvsp[-5].metaNamePtr);
			node->source = makeParseStr(YYPOSNARG(3), YYPOSNARG(6));
			node->declaredItems = yyvsp[-2].packageItems;
			node->items = yyvsp[-1].packageItems;
			yyval.createPackageBodyNode = node;
		}
break;
case 563:
  if (!yytrial)
{
			CreatePackageBodyNode* node = newNode<CreatePackageBodyNode>(*yyvsp[-4].metaNamePtr);
			node->source = makeParseStr(YYPOSNARG(3), YYPOSNARG(5));
			node->items = yyvsp[-1].packageItems;
			yyval.createPackageBodyNode = node;
		}
break;
case 564:
  if (!yytrial)
{ yyval.packageItems = newNode<Array<CreateAlterPackageNode::Item> >(); }
break;
case 566:
  if (!yytrial)
{
			yyval.packageItems = newNode<Array<CreateAlterPackageNode::Item> >();
			yyval.packageItems->add(yyvsp[0].packageItem);
		}
break;
case 567:
  if (!yytrial)
{
			yyval.packageItems = yyvsp[-1].packageItems;
			yyval.packageItems->add(yyvsp[0].packageItem);
		}
break;
case 568:
  if (!yytrial)
{ yyval.packageItem = CreateAlterPackageNode::Item::create(yyvsp[0].createAlterFunctionNode); }
break;
case 569:
  if (!yytrial)
{ yyval.packageItem = CreateAlterPackageNode::Item::create(yyvsp[-1].createAlterFunctionNode); }
break;
case 570:
  if (!yytrial)
{ yyval.packageItem = CreateAlterPackageNode::Item::create(yyvsp[0].createAlterProcedureNode); }
break;
case 571:
  if (!yytrial)
{ yyval.packageItem = CreateAlterPackageNode::Item::create(yyvsp[-1].createAlterProcedureNode); }
break;
case 572:
  if (!yytrial)
{
			CompoundStmtNode* forward = yyvsp[-1].compoundStmtNode;
			CompoundStmtNode* nonForward = yyvsp[0].compoundStmtNode;

			if (!forward)
				yyval.compoundStmtNode = nonForward;
			else
			{
				if (nonForward)
					forward->statements.add(nonForward->statements.begin(), nonForward->statements.getCount());

				yyval.compoundStmtNode = forward;
			}
		}
break;
case 573:
  if (!yytrial)
{ yyval.compoundStmtNode = NULL; }
break;
case 575:
  if (!yytrial)
{
			yyval.compoundStmtNode = newNode<CompoundStmtNode>();
			yyval.compoundStmtNode->statements.add(yyvsp[0].stmtNode);
		}
break;
case 576:
  if (!yytrial)
{
			yyvsp[-1].compoundStmtNode->statements.add(yyvsp[0].stmtNode);
			yyval.compoundStmtNode = yyvsp[-1].compoundStmtNode;
		}
break;
case 577:
  if (!yytrial)
{ yyval.stmtNode = yyvsp[-1].declareSubProcNode; }
break;
case 578:
  if (!yytrial)
{ yyval.stmtNode = yyvsp[-1].declareSubFuncNode; }
break;
case 579:
  if (!yytrial)
{ yyval.compoundStmtNode = NULL; }
break;
case 581:
  if (!yytrial)
{
			yyval.compoundStmtNode = newNode<CompoundStmtNode>();
			yyval.compoundStmtNode->statements.add(yyvsp[0].stmtNode);
		}
break;
case 582:
  if (!yytrial)
{
			yyvsp[-1].compoundStmtNode->statements.add(yyvsp[0].stmtNode);
			yyval.compoundStmtNode = yyvsp[-1].compoundStmtNode;
		}
break;
case 583:
  if (!yytrial)
{
			yyval.stmtNode = yyvsp[-1].stmtNode;
			yyval.stmtNode->line = YYPOSNARG(1).firstLine;
			yyval.stmtNode->column = YYPOSNARG(1).firstColumn;
		}
break;
case 584:
  if (!yytrial)
{
			DeclareSubProcNode* node = yyvsp[-3].declareSubProcNode;
			node->dsqlBlock = newNode<ExecBlockNode>();
			node->dsqlBlock->parameters = node->dsqlParameters;
			node->dsqlBlock->returns = node->dsqlReturns;
			node->dsqlBlock->localDeclList = yyvsp[-1].compoundStmtNode;
			node->dsqlBlock->body = yyvsp[0].stmtNode;

			for (FB_SIZE_T i = 0; i < node->dsqlBlock->parameters.getCount(); ++i)
				node->dsqlBlock->parameters[i]->parameterExpr = make_parameter();

			yyval.stmtNode = node;
		}
break;
case 585:
  if (!yytrial)
{
			DeclareSubFuncNode* node = yyvsp[-3].declareSubFuncNode;
			node->dsqlBlock = newNode<ExecBlockNode>();
			node->dsqlBlock->parameters = node->dsqlParameters;
			node->dsqlBlock->returns = node->dsqlReturns;
			node->dsqlBlock->localDeclList = yyvsp[-1].compoundStmtNode;
			node->dsqlBlock->body = yyvsp[0].stmtNode;

			for (FB_SIZE_T i = 0; i < node->dsqlBlock->parameters.getCount(); ++i)
				node->dsqlBlock->parameters[i]->parameterExpr = make_parameter();

			yyval.stmtNode = node;
		}
break;
case 586:
  if (!yytrial)
{ yyval.declareSubProcNode = newNode<DeclareSubProcNode>(NOTRIAL(*yyvsp[0].metaNamePtr)); }
break;
case 587:
yyval.parametersClause = NOTRIAL(&yyvsp[0].declareSubProcNode->dsqlParameters);
break;
case 588:
yyval.parametersClause = NOTRIAL(&yyvsp[-2].declareSubProcNode->dsqlReturns);
break;
case 589:
  if (!yytrial)
{ yyval.declareSubProcNode = yyvsp[-4].declareSubProcNode; }
break;
case 590:
  if (!yytrial)
{ yyval.declareSubFuncNode = newNode<DeclareSubFuncNode>(NOTRIAL(*yyvsp[0].metaNamePtr)); }
break;
case 591:
yyval.parametersClause = NOTRIAL(&yyvsp[0].declareSubFuncNode->dsqlParameters);
break;
case 592:
  if (!yytrial)
{
				yyval.declareSubFuncNode = yyvsp[-6].declareSubFuncNode;
				yyval.declareSubFuncNode->dsqlReturns.add(newNode<ParameterClause>(yyvsp[-2].legacyField, optName(yyvsp[-1].metaNamePtr)));
				yyval.declareSubFuncNode->dsqlDeterministic = yyvsp[0].boolVal;
			}
break;
case 595:
  if (!yytrial)
{
			DeclareVariableNode* node = newNode<DeclareVariableNode>();
			node->dsqlDef = newNode<ParameterClause>(yyvsp[-2].legacyField, optName(yyvsp[-1].metaNamePtr), yyvsp[0].valueSourceClause);
			yyval.stmtNode = node;
		}
break;
case 598:
  if (!yytrial)
{
			DeclareCursorNode* node = newNode<DeclareCursorNode>(*yyvsp[-6].metaNamePtr,
				DeclareCursorNode::CUR_TYPE_EXPLICIT);
			node->dsqlScroll = yyvsp[-5].boolVal;
			node->dsqlSelect = yyvsp[-1].selectNode;
			yyval.stmtNode = node;
		}
break;
case 599:
  if (!yytrial)
{ yyval.boolVal = false; }
break;
case 600:
  if (!yytrial)
{ yyval.boolVal = false; }
break;
case 601:
  if (!yytrial)
{ yyval.boolVal = true; }
break;
case 604:
  if (!yytrial)
{ yyval.stmtNode = newNode<LineColumnNode>(YYPOSNARG(1).firstLine, YYPOSNARG(1).firstColumn, yyvsp[-1].stmtNode); }
break;
case 605:
  if (!yytrial)
{ yyval.stmtNode = newNode<CompoundStmtNode>(); }
break;
case 606:
  if (!yytrial)
{
			BlockNode* node = newNode<BlockNode>();
			node->action = yyvsp[0].compoundStmtNode;
			yyval.stmtNode = node;
		}
break;
case 607:
  if (!yytrial)
{
			BlockNode* node = newNode<BlockNode>();
			node->action = yyvsp[-1].compoundStmtNode;
			node->handlers = yyvsp[0].compoundStmtNode;
			yyval.stmtNode = node;
		}
break;
case 608:
  if (!yytrial)
{
			yyval.compoundStmtNode = newNode<CompoundStmtNode>();
			yyval.compoundStmtNode->statements.add(yyvsp[0].stmtNode);
		}
break;
case 609:
  if (!yytrial)
{
			yyvsp[-1].compoundStmtNode->statements.add(yyvsp[0].stmtNode);
			yyval.compoundStmtNode = yyvsp[-1].compoundStmtNode;
		}
break;
case 610:
{YYVALID;}
  if (!yytrial)
{ yyval.stmtNode = newNode<LineColumnNode>(YYPOSNARG(1).firstLine, YYPOSNARG(1).firstColumn, yyvsp[-1].stmtNode); }
break;
case 611:
{YYVALID;}
  if (!yytrial)
{ yyval.stmtNode = newNode<LineColumnNode>(YYPOSNARG(1).firstLine, YYPOSNARG(1).firstColumn, yyvsp[0].stmtNode); }
break;
case 613:
  if (!yytrial)
{ yyval.stmtNode = yyvsp[0].storeNode; }
break;
case 614:
  if (!yytrial)
{ yyval.stmtNode = yyvsp[0].mergeNode; }
break;
case 616:
  if (!yytrial)
{ yyval.stmtNode = yyvsp[0].updInsNode; }
break;
case 620:
  if (!yytrial)
{ yyval.stmtNode = yyvsp[0].execStatementNode; }
break;
case 621:
  if (!yytrial)
{ yyval.stmtNode = yyvsp[0].execStatementNode; }
break;
case 623:
  if (!yytrial)
{ yyval.stmtNode = yyvsp[0].stmtNode; }
break;
case 629:
  if (!yytrial)
{ yyval.stmtNode = newNode<SuspendNode>(); }
break;
case 630:
  if (!yytrial)
{ yyval.stmtNode = newNode<ExitNode>(); }
break;
case 631:
  if (!yytrial)
{ yyval.stmtNode = newNode<ReturnNode>(yyvsp[0].valueExprNode); }
break;
case 632:
  if (!yytrial)
{ yyval.stmtNode = newNode<SessionManagementWrapperNode>(yyvsp[0].mngNode, makeParseStr(YYPOSNARG(1), YYPOSNARG(1))); }
break;
case 634:
  if (!yytrial)
{ yyval.stmtNode = yyvsp[0].stmtNode; }
break;
case 638:
  if (!yytrial)
{ yyval.stmtNode = yyvsp[0].forNode; }
break;
case 639:
  if (!yytrial)
{ yyval.stmtNode = yyvsp[0].execStatementNode; }
break;
case 640:
  if (!yytrial)
{
			InAutonomousTransactionNode* node = newNode<InAutonomousTransactionNode>();
			node->action = yyvsp[0].stmtNode;
			yyval.stmtNode = node;
		}
break;
case 641:
  if (!yytrial)
{ yyval.stmtNode = newNode<ExceptionNode>(*yyvsp[0].metaNamePtr); }
break;
case 642:
  if (!yytrial)
{ yyval.stmtNode = newNode<ExceptionNode>(*yyvsp[-1].metaNamePtr, yyvsp[0].valueExprNode); }
break;
case 643:
  if (!yytrial)
{ yyval.stmtNode = newNode<ExceptionNode>(*yyvsp[-4].metaNamePtr, (ValueExprNode*) NULL, yyvsp[-1].valueListNode); }
break;
case 644:
  if (!yytrial)
{ yyval.stmtNode = newNode<ExceptionNode>(); }
break;
case 645:
  if (!yytrial)
{
				ForNode* node = newNode<ForNode>();
				node->dsqlLabelName = yyvsp[-2].metaNamePtr;
				node->dsqlSelect = yyvsp[0].selectNode;
				yyval.forNode = node;
			}
break;
case 646:
yyval.forNode = yyvsp[0].forNode;
break;
case 647:
  if (!yytrial)
{
				ForNode* node = yyvsp[-4].forNode;
				node->statement = yyvsp[0].stmtNode;
				yyval.forNode = node;
			}
break;
case 648:
  if (!yytrial)
{
			yyvsp[-2].forNode->dsqlInto = yyvsp[-1].valueListNode;
			yyvsp[-2].forNode->dsqlCursor = yyvsp[0].declCursorNode;
		}
break;
case 649:
  if (!yytrial)
{
			yyvsp[-2].forNode->dsqlInto = yyvsp[-1].valueListNode;
			yyvsp[-2].forNode->dsqlCursor = yyvsp[0].declCursorNode;
		}
break;
case 650:
  if (!yytrial)
{ yyval.valueListNode = NULL; }
break;
case 652:
  if (!yytrial)
{ yyval.valueListNode = yyvsp[0].valueListNode; }
break;
case 653:
  if (!yytrial)
{ yyval.execStatementNode = newNode<ExecStatementNode>(); }
break;
case 654:
yyval.execStatementNode = yyvsp[0].execStatementNode;
break;
case 655:
yyval.execStatementNode = yyvsp[-2].execStatementNode;
break;
case 656:
  if (!yytrial)
{
			yyval.execStatementNode = yyvsp[-4].execStatementNode;
		}
break;
case 657:
  if (!yytrial)
{
			yyval.execStatementNode = yyvsp[-2].execStatementNode;
			yyval.execStatementNode->outputs = yyvsp[0].valueListNode;
		}
break;
case 658:
  if (!yytrial)
{
			yyval.execStatementNode = yyvsp[-2].execStatementNode;
			yyval.execStatementNode->dsqlLabelName = yyvsp[-4].metaNamePtr;
			yyval.execStatementNode->innerStmt = yyvsp[0].stmtNode;
		}
break;
case 659:
  if (!yytrial)
{ yyvsp[-1].execStatementNode->sql = yyvsp[0].valueExprNode; }
break;
case 660:
yyval.execStatementNode = yyvsp[-4].execStatementNode;
break;
case 661:
  if (!yytrial)
{ yyvsp[-7].execStatementNode->sql = yyvsp[-5].valueExprNode; }
break;
case 662:
  if (!yytrial)
{ yyvsp[-7].execStatementNode->sql = yyvsp[-5].valueExprNode; }
break;
case 665:
  if (!yytrial)
{
			if (!yyvsp[-3].execStatementNode->inputNames)
				yyvsp[-3].execStatementNode->inputNames = FB_NEW_POOL(getPool()) EDS::ParamNames(getPool());

			yyvsp[-3].execStatementNode->inputNames->add(yyvsp[-2].metaNamePtr);

			if (!yyvsp[-3].execStatementNode->inputs)
				yyvsp[-3].execStatementNode->inputs = newNode<ValueListNode>(yyvsp[0].valueExprNode);
			else
				yyvsp[-3].execStatementNode->inputs->add(yyvsp[0].valueExprNode);
		}
break;
case 666:
  if (!yytrial)
{
			if (!yyvsp[-4].execStatementNode->inputNames)
				yyvsp[-4].execStatementNode->inputNames = FB_NEW_POOL(getPool()) EDS::ParamNames(getPool());

			if (!yyvsp[-4].execStatementNode->excessInputs)
				yyvsp[-4].execStatementNode->excessInputs = FB_NEW_POOL(getPool()) EDS::ParamNumbers(getPool());

			yyvsp[-4].execStatementNode->excessInputs->add(yyvsp[-4].execStatementNode->inputNames->getCount());

			yyvsp[-4].execStatementNode->inputNames->add(yyvsp[-2].metaNamePtr);

			if (!yyvsp[-4].execStatementNode->inputs)
				yyvsp[-4].execStatementNode->inputs = newNode<ValueListNode>(yyvsp[0].valueExprNode);
			else
				yyvsp[-4].execStatementNode->inputs->add(yyvsp[0].valueExprNode);
		}
break;
case 669:
  if (!yytrial)
{
			if (!yyvsp[-1].execStatementNode->inputs)
				yyvsp[-1].execStatementNode->inputs = newNode<ValueListNode>(yyvsp[0].valueExprNode);
			else
				yyvsp[-1].execStatementNode->inputs->add(yyvsp[0].valueExprNode);
		}
break;
case 672:
yyval.execStatementNode = yyvsp[-1].execStatementNode;
break;
case 675:
  if (!yytrial)
{ setClause(yyvsp[-5].execStatementNode->dataSource, "EXTERNAL DATA SOURCE", yyvsp[0].valueExprNode); }
break;
case 676:
  if (!yytrial)
{ setClause(yyvsp[-3].execStatementNode->dataSource, "EXTERNAL DATA SOURCE", yyvsp[0].valueExprNode); }
break;
case 677:
  if (!yytrial)
{ setClause(yyvsp[-3].execStatementNode->userName, "USER", yyvsp[0].valueExprNode); }
break;
case 678:
  if (!yytrial)
{ setClause(yyvsp[-2].execStatementNode->password, "PASSWORD", yyvsp[0].valueExprNode); }
break;
case 679:
  if (!yytrial)
{ setClause(yyvsp[-2].execStatementNode->role, "ROLE", yyvsp[0].valueExprNode); }
break;
case 680:
  if (!yytrial)
{ setClause(yyvsp[-3].execStatementNode->traScope, "TRANSACTION", EDS::traAutonomous); }
break;
case 681:
  if (!yytrial)
{ setClause(yyvsp[-3].execStatementNode->traScope, "TRANSACTION", EDS::traCommon); }
break;
case 682:
  if (!yytrial)
{ setClause(yyvsp[-3].execStatementNode->useCallerPrivs, "CALLER PRIVILEGES"); }
break;
case 683:
  if (!yytrial)
{
			IfNode* node = newNode<IfNode>();
			node->condition = yyvsp[-5].boolExprNode;
			node->trueAction = yyvsp[-2].stmtNode;
			node->falseAction = yyvsp[0].stmtNode;
			yyval.stmtNode = node;
		}
break;
case 684:
  if (!yytrial)
{
			IfNode* node = newNode<IfNode>();
			node->condition = yyvsp[-3].boolExprNode;
			node->trueAction = yyvsp[0].stmtNode;
			yyval.stmtNode = node;
		}
break;
case 685:
  if (!yytrial)
{
			PostEventNode* node = newNode<PostEventNode>();
			node->event = yyvsp[-1].valueExprNode;
			node->argument = yyvsp[0].valueExprNode;
			yyval.stmtNode = node;
		}
break;
case 686:
  if (!yytrial)
{ yyval.valueExprNode = NULL; }
break;
case 687:
  if (!yytrial)
{
			ForNode* node = newNode<ForNode>();
			node->dsqlSelect = yyvsp[-2].selectNode;
			node->dsqlInto = yyvsp[0].valueListNode;
			yyval.stmtNode = node;
		}
break;
case 688:
  if (!yytrial)
{
			VariableNode* node = newNode<VariableNode>();
			node->dsqlName = *yyvsp[0].metaNamePtr;
			yyval.valueExprNode = node;
		}
break;
case 689:
  if (!yytrial)
{ yyval.valueListNode = newNode<ValueListNode>(yyvsp[0].valueExprNode); }
break;
case 690:
  if (!yytrial)
{ yyval.valueListNode = newNode<ValueListNode>(yyvsp[0].fieldNode); }
break;
case 691:
  if (!yytrial)
{ yyval.valueListNode = yyvsp[-2].valueListNode->add(yyvsp[0].fieldNode); }
break;
case 692:
  if (!yytrial)
{ yyval.valueListNode = yyvsp[-2].valueListNode->add(yyvsp[0].valueExprNode); }
break;
case 693:
  if (!yytrial)
{
			LoopNode* node = newNode<LoopNode>();
			node->dsqlLabelName = yyvsp[-6].metaNamePtr;
			node->dsqlExpr = yyvsp[-3].boolExprNode;
			node->statement = yyvsp[0].stmtNode;
			yyval.stmtNode = node;
		}
break;
case 694:
  if (!yytrial)
{ yyval.metaNamePtr = NULL; }
break;
case 695:
  if (!yytrial)
{ yyval.metaNamePtr = yyvsp[-1].metaNamePtr; }
break;
case 696:
  if (!yytrial)
{ yyval.stmtNode = newNode<ContinueLeaveNode>(blr_leave); }
break;
case 697:
  if (!yytrial)
{
			ContinueLeaveNode* node = newNode<ContinueLeaveNode>(blr_leave);
			node->dsqlLabelName = yyvsp[0].metaNamePtr;
			yyval.stmtNode = node;
		}
break;
case 698:
  if (!yytrial)
{
			ContinueLeaveNode* node = newNode<ContinueLeaveNode>(blr_continue_loop);
			node->dsqlLabelName = yyvsp[0].metaNamePtr;
			yyval.stmtNode = node;
		}
break;
case 699:
  if (!yytrial)
{ yyval.metaNamePtr = NULL; }
break;
case 701:
  if (!yytrial)
{ yyval.declCursorNode = NULL; }
break;
case 703:
  if (!yytrial)
{ yyval.declCursorNode = newNode<DeclareCursorNode>(*yyvsp[0].metaNamePtr, DeclareCursorNode::CUR_TYPE_FOR); }
break;
case 704:
  if (!yytrial)
{
			yyval.compoundStmtNode = newNode<CompoundStmtNode>();
			yyval.compoundStmtNode->statements.add(yyvsp[0].stmtNode);
		}
break;
case 705:
  if (!yytrial)
{
			yyvsp[-1].compoundStmtNode->statements.add(yyvsp[0].stmtNode);
			yyval.compoundStmtNode = yyvsp[-1].compoundStmtNode;
		}
break;
case 706:
  if (!yytrial)
{ yyval.errorHandlerNode = newNode<ErrorHandlerNode>(); }
break;
case 707:
yyval.exceptionArray = NOTRIAL(&yyvsp[0].errorHandlerNode->conditions);
break;
case 708:
  if (!yytrial)
{
				ErrorHandlerNode* node = yyvsp[-4].errorHandlerNode;
				node->action = yyvsp[0].stmtNode;
				yyval.stmtNode = node;
			}
break;
case 709:
yyval.exceptionArray = yyvsp[0].exceptionArray;
break;
case 711:
yyval.exceptionArray = yyvsp[-2].exceptionArray;
break;
case 713:
  if (!yytrial)
{
			ExceptionItem& item = yyvsp[-2].exceptionArray->add();
			item.type = ExceptionItem::SQL_CODE;
			item.code = yyvsp[0].int32Val;
		}
break;
case 714:
  if (!yytrial)
{
			ExceptionItem& item = yyvsp[-2].exceptionArray->add();
			item.type = ExceptionItem::SQL_STATE;
			item.name = yyvsp[0].intlStringPtr->getString();
		}
break;
case 715:
  if (!yytrial)
{
			ExceptionItem& item = yyvsp[-2].exceptionArray->add();
			item.type = ExceptionItem::GDS_CODE;
			item.name = yyvsp[0].metaNamePtr->c_str();
		}
break;
case 716:
  if (!yytrial)
{
			ExceptionItem& item = yyvsp[-2].exceptionArray->add();
			item.type = ExceptionItem::XCP_CODE;
			item.name = yyvsp[0].metaNamePtr->c_str();
		}
break;
case 717:
  if (!yytrial)
{
			ExceptionItem& item = yyvsp[-1].exceptionArray->add();
			item.type = ExceptionItem::XCP_DEFAULT;
		}
break;
case 721:
  if (!yytrial)
{ yyval.stmtNode = newNode<CursorStmtNode>(blr_cursor_open, *yyvsp[0].metaNamePtr); }
break;
case 722:
  if (!yytrial)
{ yyval.stmtNode = newNode<CursorStmtNode>(blr_cursor_close, *yyvsp[0].metaNamePtr); }
break;
case 723:
  if (!yytrial)
{ yyval.cursorStmtNode = newNode<CursorStmtNode>(blr_cursor_fetch_scroll); }
break;
case 724:
yyval.cursorStmtNode = yyvsp[0].cursorStmtNode;
break;
case 725:
  if (!yytrial)
{
			CursorStmtNode* cursorStmt = yyvsp[-5].cursorStmtNode;
			cursorStmt->dsqlName = *yyvsp[-1].metaNamePtr;
			cursorStmt->dsqlIntoStmt = yyvsp[0].valueListNode;
			yyval.stmtNode = cursorStmt;
		}
break;
case 726:
  if (!yytrial)
{ yyval.stmtNode = newNode<CursorStmtNode>(blr_cursor_fetch, *yyvsp[-1].metaNamePtr, yyvsp[0].valueListNode); }
break;
case 727:
  if (!yytrial)
{ yyvsp[-1].cursorStmtNode->scrollOp = blr_scroll_bof; }
break;
case 728:
  if (!yytrial)
{ yyvsp[-1].cursorStmtNode->scrollOp = blr_scroll_eof; }
break;
case 729:
  if (!yytrial)
{ yyvsp[-1].cursorStmtNode->scrollOp = blr_scroll_backward; }
break;
case 730:
  if (!yytrial)
{ yyvsp[-1].cursorStmtNode->scrollOp = blr_scroll_forward; }
break;
case 731:
  if (!yytrial)
{
			yyvsp[-2].cursorStmtNode->scrollOp = blr_scroll_absolute;
			yyvsp[-2].cursorStmtNode->scrollExpr = yyvsp[0].valueExprNode;
		}
break;
case 732:
  if (!yytrial)
{
			yyvsp[-2].cursorStmtNode->scrollOp = blr_scroll_relative;
			yyvsp[-2].cursorStmtNode->scrollExpr = yyvsp[0].valueExprNode;
		}
break;
case 733:
  if (!yytrial)
{ yyval.stmtNode = newNode<ExecProcedureNode>(QualifiedName(*yyvsp[-2].metaNamePtr), yyvsp[-1].valueListNode, yyvsp[0].valueListNode); }
break;
case 734:
  if (!yytrial)
{ yyval.stmtNode = newNode<ExecProcedureNode>(QualifiedName(*yyvsp[-2].metaNamePtr, *yyvsp[-4].metaNamePtr), yyvsp[-1].valueListNode, yyvsp[0].valueListNode); }
break;
case 735:
  if (!yytrial)
{ yyval.valueListNode = NULL; }
break;
case 736:
  if (!yytrial)
{ yyval.valueListNode = yyvsp[0].valueListNode; }
break;
case 737:
  if (!yytrial)
{ yyval.valueListNode = yyvsp[-1].valueListNode; }
break;
case 738:
  if (!yytrial)
{ yyval.valueListNode = NULL; }
break;
case 739:
  if (!yytrial)
{ yyval.valueListNode = yyvsp[0].valueListNode; }
break;
case 740:
  if (!yytrial)
{ yyval.valueListNode = yyvsp[-1].valueListNode; }
break;
case 741:
  if (!yytrial)
{ yyval.execBlockNode = newNode<ExecBlockNode>(); }
break;
case 742:
yyval.parametersClause = NOTRIAL(&yyvsp[0].execBlockNode->parameters);
break;
case 743:
yyval.parametersClause = NOTRIAL(&yyvsp[-2].execBlockNode->returns);
break;
case 744:
  if (!yytrial)
{
			ExecBlockNode* node = yyvsp[-7].execBlockNode;
			node->localDeclList = yyvsp[-1].compoundStmtNode;
			node->body = yyvsp[0].stmtNode;
			yyval.execBlockNode = node;
		}
break;
case 749:
  if (!yytrial)
{ yyvsp[-4].parametersClause->add(newNode<ParameterClause>(yyvsp[-3].legacyField, optName(yyvsp[-2].metaNamePtr), (ValueSourceClause*) NULL, yyvsp[0].valueExprNode)); }
break;
case 750:
  if (!yytrial)
{
			CreateAlterViewNode* node = newNode<CreateAlterViewNode>(yyvsp[-4].relSourceNode, yyvsp[-3].valueListNode, yyvsp[-1].selectExprNode);
			node->source = makeParseStr(YYPOSNARG(4), YYPOSNARG(5));
			node->withCheckOption = yyvsp[0].boolVal;
			yyval.createAlterViewNode = node;
		}
break;
case 751:
  if (!yytrial)
{
			yyvsp[0].createAlterViewNode->alter = true;
			yyval.ddlNode = yyvsp[0].createAlterViewNode;
		}
break;
case 752:
  if (!yytrial)
{
			yyvsp[0].createAlterViewNode->alter = true;
			yyvsp[0].createAlterViewNode->create = false;
			yyval.ddlNode = yyvsp[0].createAlterViewNode;
		}
break;
case 753:
  if (!yytrial)
{ yyval.boolVal = false; }
break;
case 754:
  if (!yytrial)
{ yyval.boolVal = true; }
break;
case 755:
  if (!yytrial)
{
			yyval.createAlterTriggerNode = yyvsp[-4].createAlterTriggerNode;
			yyval.createAlterTriggerNode->ssDefiner = yyvsp[-3].nullableSqlSecurityVal;
			yyval.createAlterTriggerNode->source = makeParseStr(YYPOSNARG(3), YYPOSNARG(5));
			yyval.createAlterTriggerNode->localDeclList = yyvsp[-1].compoundStmtNode;
			yyval.createAlterTriggerNode->body = yyvsp[0].stmtNode;
		}
break;
case 756:
  if (!yytrial)
{
			yyval.createAlterTriggerNode = yyvsp[-2].createAlterTriggerNode;
			yyval.createAlterTriggerNode->external = yyvsp[-1].externalClause;
			if (yyvsp[0].stringPtr)
				yyval.createAlterTriggerNode->source = *yyvsp[0].stringPtr;
		}
break;
case 757:
  if (!yytrial)
{ yyval.createAlterTriggerNode = newNode<CreateAlterTriggerNode>(*yyvsp[0].metaNamePtr); }
break;
case 758:
yyval.createAlterTriggerNode = NOTRIAL(yyvsp[0].createAlterTriggerNode);
break;
case 759:
  if (!yytrial)
{ yyval.createAlterTriggerNode = yyvsp[-2].createAlterTriggerNode; }
break;
case 760:
yyval.createAlterTriggerNode = NOTRIAL(yyvsp[-1].createAlterTriggerNode);
break;
case 761:
  if (!yytrial)
{
			yyvsp[-4].createAlterTriggerNode->active = yyvsp[-3].nullableBoolVal;
			yyvsp[-4].createAlterTriggerNode->type = yyvsp[-1].uint64Val;
			setClause(yyvsp[-4].createAlterTriggerNode->position, "POSITION", yyvsp[0].nullableIntVal);
		}
break;
case 762:
  if (!yytrial)
{
			yyvsp[-5].createAlterTriggerNode->relationName = *yyvsp[-3].metaNamePtr;
			yyvsp[-5].createAlterTriggerNode->active = yyvsp[-2].nullableBoolVal;
			yyvsp[-5].createAlterTriggerNode->type = yyvsp[-1].uint64Val;
			setClause(yyvsp[-5].createAlterTriggerNode->position, "POSITION", yyvsp[0].nullableIntVal);
		}
break;
case 763:
  if (!yytrial)
{
			yyval.createAlterTriggerNode = yyvsp[0].createAlterTriggerNode;
			yyval.createAlterTriggerNode->alter = true;
		}
break;
case 764:
  if (!yytrial)
{ yyval.nullableBoolVal = Nullable<bool>::val(true); }
break;
case 765:
  if (!yytrial)
{ yyval.nullableBoolVal = Nullable<bool>::val(false); }
break;
case 766:
  if (!yytrial)
{ yyval.nullableBoolVal = Nullable<bool>::empty(); }
break;
case 767:
  if (!yytrial)
{
			yyval.uint64Val = yyvsp[-3].uint64Val;
			setClause(yyvsp[-4].createAlterTriggerNode->position, "POSITION", yyvsp[-2].nullableIntVal);
			yyvsp[-4].createAlterTriggerNode->relationName = *yyvsp[0].metaNamePtr;
		}
break;
case 768:
  if (!yytrial)
{ yyval.uint64Val = yyvsp[0].uint64Val; }
break;
case 769:
  if (!yytrial)
{ yyval.uint64Val = yyvsp[-1].uint64Val + yyvsp[0].uint64Val; }
break;
case 770:
  if (!yytrial)
{ yyval.uint64Val = yyvsp[-1].uint64Val + yyvsp[0].uint64Val - 1; }
break;
case 771:
  if (!yytrial)
{ yyval.uint64Val = TRIGGER_TYPE_DB | DB_TRIGGER_CONNECT; }
break;
case 772:
  if (!yytrial)
{ yyval.uint64Val = TRIGGER_TYPE_DB | DB_TRIGGER_DISCONNECT; }
break;
case 773:
  if (!yytrial)
{ yyval.uint64Val = TRIGGER_TYPE_DB | DB_TRIGGER_TRANS_START; }
break;
case 774:
  if (!yytrial)
{ yyval.uint64Val = TRIGGER_TYPE_DB | DB_TRIGGER_TRANS_COMMIT; }
break;
case 775:
  if (!yytrial)
{ yyval.uint64Val = TRIGGER_TYPE_DB | DB_TRIGGER_TRANS_ROLLBACK; }
break;
case 777:
  if (!yytrial)
{
			yyval.uint64Val = TRIGGER_TYPE_DDL | DDL_TRIGGER_ANY;
		}
break;
case 778:
  if (!yytrial)
{ yyval.uint64Val = TRIGGER_TYPE_DDL | (1LL << DDL_TRIGGER_CREATE_TABLE); }
break;
case 779:
  if (!yytrial)
{ yyval.uint64Val = TRIGGER_TYPE_DDL | (1LL << DDL_TRIGGER_ALTER_TABLE); }
break;
case 780:
  if (!yytrial)
{ yyval.uint64Val = TRIGGER_TYPE_DDL | (1LL << DDL_TRIGGER_DROP_TABLE); }
break;
case 781:
  if (!yytrial)
{ yyval.uint64Val = TRIGGER_TYPE_DDL | (1LL << DDL_TRIGGER_CREATE_PROCEDURE); }
break;
case 782:
  if (!yytrial)
{ yyval.uint64Val = TRIGGER_TYPE_DDL | (1LL << DDL_TRIGGER_ALTER_PROCEDURE); }
break;
case 783:
  if (!yytrial)
{ yyval.uint64Val = TRIGGER_TYPE_DDL | (1LL << DDL_TRIGGER_DROP_PROCEDURE); }
break;
case 784:
  if (!yytrial)
{ yyval.uint64Val = TRIGGER_TYPE_DDL | (1LL << DDL_TRIGGER_CREATE_FUNCTION); }
break;
case 785:
  if (!yytrial)
{ yyval.uint64Val = TRIGGER_TYPE_DDL | (1LL << DDL_TRIGGER_ALTER_FUNCTION); }
break;
case 786:
  if (!yytrial)
{ yyval.uint64Val = TRIGGER_TYPE_DDL | (1LL << DDL_TRIGGER_DROP_FUNCTION); }
break;
case 787:
  if (!yytrial)
{ yyval.uint64Val = TRIGGER_TYPE_DDL | (1LL << DDL_TRIGGER_CREATE_TRIGGER); }
break;
case 788:
  if (!yytrial)
{ yyval.uint64Val = TRIGGER_TYPE_DDL | (1LL << DDL_TRIGGER_ALTER_TRIGGER); }
break;
case 789:
  if (!yytrial)
{ yyval.uint64Val = TRIGGER_TYPE_DDL | (1LL << DDL_TRIGGER_DROP_TRIGGER); }
break;
case 790:
  if (!yytrial)
{ yyval.uint64Val = TRIGGER_TYPE_DDL | (1LL << DDL_TRIGGER_CREATE_EXCEPTION); }
break;
case 791:
  if (!yytrial)
{ yyval.uint64Val = TRIGGER_TYPE_DDL | (1LL << DDL_TRIGGER_ALTER_EXCEPTION); }
break;
case 792:
  if (!yytrial)
{ yyval.uint64Val = TRIGGER_TYPE_DDL | (1LL << DDL_TRIGGER_DROP_EXCEPTION); }
break;
case 793:
  if (!yytrial)
{ yyval.uint64Val = TRIGGER_TYPE_DDL | (1LL << DDL_TRIGGER_CREATE_VIEW); }
break;
case 794:
  if (!yytrial)
{ yyval.uint64Val = TRIGGER_TYPE_DDL | (1LL << DDL_TRIGGER_ALTER_VIEW); }
break;
case 795:
  if (!yytrial)
{ yyval.uint64Val = TRIGGER_TYPE_DDL | (1LL << DDL_TRIGGER_DROP_VIEW); }
break;
case 796:
  if (!yytrial)
{ yyval.uint64Val = TRIGGER_TYPE_DDL | (1LL << DDL_TRIGGER_CREATE_DOMAIN); }
break;
case 797:
  if (!yytrial)
{ yyval.uint64Val = TRIGGER_TYPE_DDL | (1LL << DDL_TRIGGER_ALTER_DOMAIN); }
break;
case 798:
  if (!yytrial)
{ yyval.uint64Val = TRIGGER_TYPE_DDL | (1LL << DDL_TRIGGER_DROP_DOMAIN); }
break;
case 799:
  if (!yytrial)
{ yyval.uint64Val = TRIGGER_TYPE_DDL | (1LL << DDL_TRIGGER_CREATE_ROLE); }
break;
case 800:
  if (!yytrial)
{ yyval.uint64Val = TRIGGER_TYPE_DDL | (1LL << DDL_TRIGGER_ALTER_ROLE); }
break;
case 801:
  if (!yytrial)
{ yyval.uint64Val = TRIGGER_TYPE_DDL | (1LL << DDL_TRIGGER_DROP_ROLE); }
break;
case 802:
  if (!yytrial)
{ yyval.uint64Val = TRIGGER_TYPE_DDL | (1LL << DDL_TRIGGER_CREATE_SEQUENCE); }
break;
case 803:
  if (!yytrial)
{ yyval.uint64Val = TRIGGER_TYPE_DDL | (1LL << DDL_TRIGGER_ALTER_SEQUENCE); }
break;
case 804:
  if (!yytrial)
{ yyval.uint64Val = TRIGGER_TYPE_DDL | (1LL << DDL_TRIGGER_DROP_SEQUENCE); }
break;
case 805:
  if (!yytrial)
{ yyval.uint64Val = TRIGGER_TYPE_DDL | (1LL << DDL_TRIGGER_CREATE_USER); }
break;
case 806:
  if (!yytrial)
{ yyval.uint64Val = TRIGGER_TYPE_DDL | (1LL << DDL_TRIGGER_ALTER_USER); }
break;
case 807:
  if (!yytrial)
{ yyval.uint64Val = TRIGGER_TYPE_DDL | (1LL << DDL_TRIGGER_DROP_USER); }
break;
case 808:
  if (!yytrial)
{ yyval.uint64Val = TRIGGER_TYPE_DDL | (1LL << DDL_TRIGGER_CREATE_INDEX); }
break;
case 809:
  if (!yytrial)
{ yyval.uint64Val = TRIGGER_TYPE_DDL | (1LL << DDL_TRIGGER_ALTER_INDEX); }
break;
case 810:
  if (!yytrial)
{ yyval.uint64Val = TRIGGER_TYPE_DDL | (1LL << DDL_TRIGGER_DROP_INDEX); }
break;
case 811:
  if (!yytrial)
{ yyval.uint64Val = TRIGGER_TYPE_DDL | (1LL << DDL_TRIGGER_CREATE_COLLATION); }
break;
case 812:
  if (!yytrial)
{ yyval.uint64Val = TRIGGER_TYPE_DDL | (1LL << DDL_TRIGGER_DROP_COLLATION); }
break;
case 813:
  if (!yytrial)
{ yyval.uint64Val = TRIGGER_TYPE_DDL | (1LL << DDL_TRIGGER_ALTER_CHARACTER_SET); }
break;
case 814:
  if (!yytrial)
{ yyval.uint64Val = TRIGGER_TYPE_DDL | (1LL << DDL_TRIGGER_CREATE_PACKAGE); }
break;
case 815:
  if (!yytrial)
{ yyval.uint64Val = TRIGGER_TYPE_DDL | (1LL << DDL_TRIGGER_ALTER_PACKAGE); }
break;
case 816:
  if (!yytrial)
{ yyval.uint64Val = TRIGGER_TYPE_DDL | (1LL << DDL_TRIGGER_DROP_PACKAGE); }
break;
case 817:
  if (!yytrial)
{ yyval.uint64Val = TRIGGER_TYPE_DDL | (1LL << DDL_TRIGGER_CREATE_PACKAGE_BODY); }
break;
case 818:
  if (!yytrial)
{ yyval.uint64Val = TRIGGER_TYPE_DDL | (1LL << DDL_TRIGGER_DROP_PACKAGE_BODY); }
break;
case 819:
  if (!yytrial)
{ yyval.uint64Val = TRIGGER_TYPE_DDL | (1LL << DDL_TRIGGER_CREATE_MAPPING); }
break;
case 820:
  if (!yytrial)
{ yyval.uint64Val = TRIGGER_TYPE_DDL | (1LL << DDL_TRIGGER_ALTER_MAPPING); }
break;
case 821:
  if (!yytrial)
{ yyval.uint64Val = TRIGGER_TYPE_DDL | (1LL << DDL_TRIGGER_DROP_MAPPING); }
break;
case 822:
  if (!yytrial)
{ yyval.uint64Val = yyvsp[-2].uint64Val | yyvsp[0].uint64Val; }
break;
case 823:
  if (!yytrial)
{ yyval.uint64Val = 0; }
break;
case 824:
  if (!yytrial)
{ yyval.uint64Val = 1; }
break;
case 825:
  if (!yytrial)
{ yyval.uint64Val = trigger_type_suffix(1, 0, 0); }
break;
case 826:
  if (!yytrial)
{ yyval.uint64Val = trigger_type_suffix(2, 0, 0); }
break;
case 827:
  if (!yytrial)
{ yyval.uint64Val = trigger_type_suffix(3, 0, 0); }
break;
case 828:
  if (!yytrial)
{ yyval.uint64Val = trigger_type_suffix(1, 2, 0); }
break;
case 829:
  if (!yytrial)
{ yyval.uint64Val = trigger_type_suffix(1, 3, 0); }
break;
case 830:
  if (!yytrial)
{ yyval.uint64Val = trigger_type_suffix(2, 1, 0); }
break;
case 831:
  if (!yytrial)
{ yyval.uint64Val = trigger_type_suffix(2, 3, 0); }
break;
case 832:
  if (!yytrial)
{ yyval.uint64Val = trigger_type_suffix(3, 1, 0); }
break;
case 833:
  if (!yytrial)
{ yyval.uint64Val = trigger_type_suffix(3, 2, 0); }
break;
case 834:
  if (!yytrial)
{ yyval.uint64Val = trigger_type_suffix(1, 2, 3); }
break;
case 835:
  if (!yytrial)
{ yyval.uint64Val = trigger_type_suffix(1, 3, 2); }
break;
case 836:
  if (!yytrial)
{ yyval.uint64Val = trigger_type_suffix(2, 1, 3); }
break;
case 837:
  if (!yytrial)
{ yyval.uint64Val = trigger_type_suffix(2, 3, 1); }
break;
case 838:
  if (!yytrial)
{ yyval.uint64Val = trigger_type_suffix(3, 1, 2); }
break;
case 839:
  if (!yytrial)
{ yyval.uint64Val = trigger_type_suffix(3, 2, 1); }
break;
case 840:
  if (!yytrial)
{ yyval.nullableIntVal = Nullable<int>::empty(); }
break;
case 841:
  if (!yytrial)
{ yyval.nullableIntVal = Nullable<int>::val(yyvsp[0].int32Val); }
break;
case 842:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].ddlNode; }
break;
case 843:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].createAlterSequenceNode; }
break;
case 844:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].createAlterExceptionNode; }
break;
case 845:
  if (!yytrial)
{ yyval.ddlNode = newNode<AlterRelationNode>(yyvsp[0].relSourceNode); }
break;
case 846:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[-2].relationNode; }
break;
case 847:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].ddlNode; }
break;
case 848:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].createAlterTriggerNode; }
break;
case 849:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].createAlterProcedureNode; }
break;
case 850:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].createAlterPackageNode; }
break;
case 851:
  if (!yytrial)
{ yyval.alterDatabaseNode = newNode<AlterDatabaseNode>(); }
break;
case 852:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[-2].alterDatabaseNode; }
break;
case 853:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].alterDomainNode; }
break;
case 854:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].ddlNode; }
break;
case 855:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].ddlNode; }
break;
case 856:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].createAlterFunctionNode; }
break;
case 857:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].ddlNode; }
break;
case 858:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].createAlterUserNode; }
break;
case 859:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].createAlterUserNode; }
break;
case 860:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].ddlNode; }
break;
case 861:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].createAlterSequenceNode; }
break;
case 862:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].createAlterSequenceNode; }
break;
case 863:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].mappingNode; }
break;
case 864:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].mappingNode; }
break;
case 865:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].ddlNode; }
break;
case 866:
  if (!yytrial)
{ yyval.alterDomainNode = newNode<AlterDomainNode>(*yyvsp[0].metaNamePtr); }
break;
case 867:
yyval.alterDomainNode = yyvsp[0].alterDomainNode;
break;
case 868:
  if (!yytrial)
{ yyval.alterDomainNode = yyvsp[-2].alterDomainNode; }
break;
case 870:
yyval.alterDomainNode = yyvsp[-1].alterDomainNode;
break;
case 872:
  if (!yytrial)
{ setClause(yyvsp[-2].alterDomainNode->setDefault, "DOMAIN DEFAULT", yyvsp[0].valueSourceClause); }
break;
case 873:
  if (!yytrial)
{ setClause(yyvsp[-3].alterDomainNode->setConstraint, "DOMAIN CONSTRAINT", yyvsp[0].boolSourceClause); }
break;
case 874:
  if (!yytrial)
{ setClause(yyvsp[-2].alterDomainNode->setConstraint, "DOMAIN CONSTRAINT", yyvsp[0].boolSourceClause); }
break;
case 875:
  if (!yytrial)
{ setClause(yyvsp[-2].alterDomainNode->dropDefault, "DOMAIN DROP DEFAULT"); }
break;
case 876:
  if (!yytrial)
{ setClause(yyvsp[-2].alterDomainNode->dropConstraint, "DOMAIN DROP CONSTRAINT"); }
break;
case 877:
  if (!yytrial)
{ setClause(yyvsp[-3].alterDomainNode->notNullFlag, "{SET | DROP} NOT NULL", false); }
break;
case 878:
  if (!yytrial)
{ setClause(yyvsp[-3].alterDomainNode->notNullFlag, "{SET | DROP} NOT NULL", true); }
break;
case 879:
  if (!yytrial)
{ setClause(yyvsp[-2].alterDomainNode->renameTo, "DOMAIN NAME", *yyvsp[0].metaNamePtr); }
break;
case 880:
  if (!yytrial)
{
			/*// FIXME: ALTER DOMAIN doesn't support collations, and altered domain's*/
			/*// collation is always lost.*/
			dsql_fld* type = yyvsp[0].legacyField;
			type->collate = "";
			setClause(yyvsp[-2].alterDomainNode->type, "DOMAIN TYPE", type);
		}
break;
case 883:
  if (!yytrial)
{
			RelationNode::DropColumnClause* clause = newNode<RelationNode::DropColumnClause>();
			clause->name = *yyvsp[-1].metaNamePtr;
			clause->cascade = yyvsp[0].boolVal;
			yyvsp[-3].relationNode->clauses.add(clause);
		}
break;
case 884:
  if (!yytrial)
{
			RelationNode::DropConstraintClause* clause = newNode<RelationNode::DropConstraintClause>();
			clause->name = *yyvsp[0].metaNamePtr;
			yyvsp[-3].relationNode->clauses.add(clause);
		}
break;
case 887:
  if (!yytrial)
{
			RelationNode::AlterColPosClause* clause = newNode<RelationNode::AlterColPosClause>();
			clause->name = *yyvsp[-2].metaNamePtr;
			clause->newPos = yyvsp[0].int32Val;
			yyvsp[-4].relationNode->clauses.add(clause);
		}
break;
case 888:
  if (!yytrial)
{
			RelationNode::AlterColNameClause* clause = newNode<RelationNode::AlterColNameClause>();
			clause->fromName = *yyvsp[-2].metaNamePtr;
			clause->toName = *yyvsp[0].metaNamePtr;
			yyvsp[-4].relationNode->clauses.add(clause);
		}
break;
case 889:
  if (!yytrial)
{
			RelationNode::AlterColNullClause* clause = newNode<RelationNode::AlterColNullClause>();
			clause->name = *yyvsp[-3].metaNamePtr;
			clause->notNullFlag = false;
			yyvsp[-5].relationNode->clauses.add(clause);
		}
break;
case 890:
  if (!yytrial)
{
			RelationNode::AlterColNullClause* clause = newNode<RelationNode::AlterColNullClause>();
			clause->name = *yyvsp[-3].metaNamePtr;
			clause->notNullFlag = true;
			yyvsp[-5].relationNode->clauses.add(clause);
		}
break;
case 891:
  if (!yytrial)
{
			RelationNode::AlterColTypeClause* clause = newNode<RelationNode::AlterColTypeClause>();
			clause->field = yyvsp[0].legacyField;
			clause->field->fld_name = *yyvsp[-2].metaNamePtr;
			yyvsp[-4].relationNode->clauses.add(clause);
		}
break;
case 892:
  if (!yytrial)
{
			RelationNode::AlterColTypeClause* clause = newNode<RelationNode::AlterColTypeClause>();
			clause->field = yyvsp[-1].legacyField;
			clause->field->fld_name = *yyvsp[-3].metaNamePtr;
			clause->computed = yyvsp[0].valueSourceClause;
			yyvsp[-5].relationNode->clauses.add(clause);
			clause->field->flags |= FLD_computed;
		}
break;
case 893:
  if (!yytrial)
{
			RelationNode::AlterColTypeClause* clause = newNode<RelationNode::AlterColTypeClause>();
			clause->field = newNode<dsql_fld>();
			clause->field->fld_name = *yyvsp[-1].metaNamePtr;
			clause->computed = yyvsp[0].valueSourceClause;
			yyvsp[-3].relationNode->clauses.add(clause);
			clause->field->flags |= FLD_computed;
		}
break;
case 894:
  if (!yytrial)
{
			RelationNode::AlterColTypeClause* clause = newNode<RelationNode::AlterColTypeClause>();
			clause->field = newNode<dsql_fld>();
			clause->field->fld_name = *yyvsp[-2].metaNamePtr;
			clause->defaultValue = yyvsp[0].valueSourceClause;
			yyvsp[-4].relationNode->clauses.add(clause);
		}
break;
case 895:
  if (!yytrial)
{
			RelationNode::AlterColTypeClause* clause = newNode<RelationNode::AlterColTypeClause>();
			clause->field = newNode<dsql_fld>();
			clause->field->fld_name = *yyvsp[-2].metaNamePtr;
			clause->dropDefault = true;
			yyvsp[-4].relationNode->clauses.add(clause);
		}
break;
case 896:
  if (!yytrial)
{ yyval.identityOptions = newNode<RelationNode::IdentityOptions>(); }
break;
case 897:
  if (!yytrial)
{
				RelationNode::AlterColTypeClause* clause = newNode<RelationNode::AlterColTypeClause>();
				clause->field = newNode<dsql_fld>();
				clause->field->fld_name = *yyvsp[-3].metaNamePtr;
				clause->identityOptions = yyvsp[-2].identityOptions;
				yyvsp[-5].relationNode->clauses.add(clause);
			}
break;
case 898:
  if (!yytrial)
{
			RelationNode::AlterColTypeClause* clause = newNode<RelationNode::AlterColTypeClause>();
			clause->field = newNode<dsql_fld>();
			clause->field->fld_name = *yyvsp[-2].metaNamePtr;
			clause->dropIdentity = true;
			yyvsp[-4].relationNode->clauses.add(clause);
		}
break;
case 899:
  if (!yytrial)
{
			setClause(yyvsp[-4].relationNode->ssDefiner, "SQL SECURITY", true);
			RelationNode::Clause* clause =
				newNode<RelationNode::Clause>(RelationNode::Clause::TYPE_ALTER_SQL_SECURITY);
			yyvsp[-4].relationNode->clauses.add(clause);
		}
break;
case 900:
  if (!yytrial)
{
			setClause(yyvsp[-4].relationNode->ssDefiner, "SQL SECURITY", false);
			RelationNode::Clause* clause =
				newNode<RelationNode::Clause>(RelationNode::Clause::TYPE_ALTER_SQL_SECURITY);
			yyvsp[-4].relationNode->clauses.add(clause);
		}
break;
case 901:
  if (!yytrial)
{
			setClause(yyvsp[-3].relationNode->ssDefiner, "SQL SECURITY", Nullable<bool>::empty());
			RelationNode::Clause* clause =
				newNode<RelationNode::Clause>(RelationNode::Clause::TYPE_ALTER_SQL_SECURITY);
			yyvsp[-3].relationNode->clauses.add(clause);
		}
break;
case 902:
  if (!yytrial)
{
			setClause(yyvsp[-2].relationNode->replicationState, "PUBLICATION", true);
			RelationNode::Clause* clause =
				newNode<RelationNode::Clause>(RelationNode::Clause::TYPE_ALTER_PUBLICATION);
			yyvsp[-2].relationNode->clauses.add(clause);
		}
break;
case 903:
  if (!yytrial)
{
			setClause(yyvsp[-2].relationNode->replicationState, "PUBLICATION", false);
			RelationNode::Clause* clause =
				newNode<RelationNode::Clause>(RelationNode::Clause::TYPE_ALTER_PUBLICATION);
			yyvsp[-2].relationNode->clauses.add(clause);
		}
break;
case 1002:
  if (!yytrial)
{
			yyval.legacyField = newNode<dsql_fld>();
			yyval.legacyField->typeOfName = *yyvsp[0].metaNamePtr;
		}
break;
case 1003:
yyval.identityOptions = yyvsp[-2].identityOptions;
break;
case 1006:
  if (!yytrial)
{ yyvsp[-3].identityOptions->type = IDENT_TYPE_ALWAYS; }
break;
case 1007:
  if (!yytrial)
{ yyvsp[-4].identityOptions->type = IDENT_TYPE_BY_DEFAULT; }
break;
case 1012:
  if (!yytrial)
{
			setClause(yyvsp[-2].identityOptions->restart, "RESTART");
			yyvsp[-2].identityOptions->startValue = yyvsp[0].nullableInt64Val;
		}
break;
case 1013:
  if (!yytrial)
{ setClause(yyvsp[-4].identityOptions->increment, "SET INCREMENT BY", yyvsp[0].int32Val); }
break;
case 1014:
  if (!yytrial)
{ yyval.boolVal = false; }
break;
case 1015:
  if (!yytrial)
{ yyval.boolVal = false; }
break;
case 1016:
  if (!yytrial)
{ yyval.boolVal = true; }
break;
case 1017:
  if (!yytrial)
{ yyval.ddlNode = newNode<AlterIndexNode>(*yyvsp[-1].metaNamePtr, true); }
break;
case 1018:
  if (!yytrial)
{ yyval.ddlNode = newNode<AlterIndexNode>(*yyvsp[-1].metaNamePtr, false); }
break;
case 1019:
  if (!yytrial)
{
			AlterExternalFunctionNode* node = newNode<AlterExternalFunctionNode>(*yyvsp[-2].metaNamePtr);
			if (yyvsp[-1].stringPtr)
				node->clauses.name = *yyvsp[-1].stringPtr;
			if (yyvsp[0].stringPtr)
				node->clauses.udfModule = *yyvsp[0].stringPtr;
			yyval.ddlNode = node;
		}
break;
case 1020:
  if (!yytrial)
{ yyval.stringPtr = NULL; }
break;
case 1021:
  if (!yytrial)
{ yyval.stringPtr = yyvsp[0].stringPtr; }
break;
case 1022:
  if (!yytrial)
{ yyval.stringPtr = NULL; }
break;
case 1023:
  if (!yytrial)
{ yyval.stringPtr = yyvsp[0].stringPtr; }
break;
case 1024:
  if (!yytrial)
{
			MappingNode* mn = newNode<MappingNode>(MappingNode::MAP_RPL, "AutoAdminImplementationMapping");
			mn->op = yyvsp[-3].boolVal ? MappingNode::MAP_RPL : MappingNode::MAP_DROP;
			mn->from = newNode<IntlString>(FB_DOMAIN_ANY_RID_ADMINS);
			mn->fromType = newNode<MetaName>(FB_PREDEFINED_GROUP);
			mn->mode = 'P';
			mn->plugin = newNode<MetaName>("Win_Sspi");
			mn->role = true;
			mn->to = yyvsp[-4].metaNamePtr;
			mn->validateAdmin();
			yyval.ddlNode = mn;
		}
break;
case 1025:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].createAlterRoleNode; }
break;
case 1026:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].ddlNode; }
break;
case 1027:
  if (!yytrial)
{ yyval.boolVal = true; }
break;
case 1028:
  if (!yytrial)
{ yyval.boolVal = false; }
break;
case 1031:
yyval.dbFilesClause = NOTRIAL(&yyvsp[-1].alterDatabaseNode->files);
break;
case 1033:
  if (!yytrial)
{ yyvsp[-4].alterDatabaseNode->differenceFile = *yyvsp[0].stringPtr; }
break;
case 1034:
  if (!yytrial)
{ yyvsp[-3].alterDatabaseNode->clauses |= AlterDatabaseNode::CLAUSE_DROP_DIFFERENCE; }
break;
case 1035:
  if (!yytrial)
{ yyvsp[-2].alterDatabaseNode->clauses |= AlterDatabaseNode::CLAUSE_BEGIN_BACKUP; }
break;
case 1036:
  if (!yytrial)
{ yyvsp[-2].alterDatabaseNode->clauses |= AlterDatabaseNode::CLAUSE_END_BACKUP; }
break;
case 1037:
  if (!yytrial)
{ yyvsp[-5].alterDatabaseNode->setDefaultCharSet = *yyvsp[0].metaNamePtr; }
break;
case 1038:
yyval.alterDatabaseNode = yyvsp[-3].alterDatabaseNode;
break;
case 1039:
  if (!yytrial)
{
			setClauseFlag(yyvsp[-5].alterDatabaseNode->clauses, AlterDatabaseNode::CLAUSE_CRYPT, "CRYPT");
			yyvsp[-5].alterDatabaseNode->cryptPlugin = *yyvsp[-2].metaNamePtr;
		}
break;
case 1040:
  if (!yytrial)
{ setClauseFlag(yyvsp[-1].alterDatabaseNode->clauses, AlterDatabaseNode::CLAUSE_CRYPT, "CRYPT"); }
break;
case 1041:
  if (!yytrial)
{ yyvsp[-4].alterDatabaseNode->linger = yyvsp[0].int32Val; }
break;
case 1042:
  if (!yytrial)
{ yyvsp[-2].alterDatabaseNode->linger = 0; }
break;
case 1043:
  if (!yytrial)
{ yyvsp[-3].alterDatabaseNode->ssDefiner = yyvsp[0].boolVal; }
break;
case 1044:
  if (!yytrial)
{ yyvsp[-2].alterDatabaseNode->clauses |= AlterDatabaseNode::CLAUSE_ENABLE_PUB; }
break;
case 1045:
  if (!yytrial)
{ yyvsp[-2].alterDatabaseNode->clauses |= AlterDatabaseNode::CLAUSE_DISABLE_PUB; }
break;
case 1046:
  if (!yytrial)
{ yyvsp[-5].alterDatabaseNode->clauses |= AlterDatabaseNode::CLAUSE_PUB_INCL_TABLE; }
break;
case 1047:
  if (!yytrial)
{ yyvsp[-5].alterDatabaseNode->clauses |= AlterDatabaseNode::CLAUSE_PUB_EXCL_TABLE; }
break;
case 1049:
  if (!yytrial)
{ yyvsp[-2].alterDatabaseNode->keyName = *yyvsp[0].metaNamePtr; }
break;
case 1054:
  if (!yytrial)
{ yyvsp[-1].alterDatabaseNode->pubTables.add(*yyvsp[0].metaNamePtr); }
break;
case 1055:
  if (!yytrial)
{
			yyval.createAlterTriggerNode = newNode<CreateAlterTriggerNode>(*yyvsp[-7].metaNamePtr);
			yyval.createAlterTriggerNode->alter = true;
			yyval.createAlterTriggerNode->create = false;
			yyval.createAlterTriggerNode->active = yyvsp[-6].nullableBoolVal;
			yyval.createAlterTriggerNode->type = yyvsp[-5].nullableUint64Val;
			yyval.createAlterTriggerNode->position = yyvsp[-4].nullableIntVal;
			yyval.createAlterTriggerNode->ssDefiner = yyvsp[-3].nullableSqlSecurityVal;
			yyval.createAlterTriggerNode->source = makeParseStr(YYPOSNARG(6), YYPOSNARG(8));
			yyval.createAlterTriggerNode->localDeclList = yyvsp[-1].compoundStmtNode;
			yyval.createAlterTriggerNode->body = yyvsp[0].stmtNode;
		}
break;
case 1056:
  if (!yytrial)
{
			yyval.createAlterTriggerNode = newNode<CreateAlterTriggerNode>(*yyvsp[-5].metaNamePtr);
			yyval.createAlterTriggerNode->alter = true;
			yyval.createAlterTriggerNode->create = false;
			yyval.createAlterTriggerNode->active = yyvsp[-4].nullableBoolVal;
			yyval.createAlterTriggerNode->type = yyvsp[-3].nullableUint64Val;
			yyval.createAlterTriggerNode->position = yyvsp[-2].nullableIntVal;
			yyval.createAlterTriggerNode->external = yyvsp[-1].externalClause;
			if (yyvsp[0].stringPtr)
				yyval.createAlterTriggerNode->source = *yyvsp[0].stringPtr;
		}
break;
case 1057:
  if (!yytrial)
{
			yyval.createAlterTriggerNode = newNode<CreateAlterTriggerNode>(*yyvsp[-4].metaNamePtr);
			yyval.createAlterTriggerNode->alter = true;
			yyval.createAlterTriggerNode->create = false;
			yyval.createAlterTriggerNode->active = yyvsp[-3].nullableBoolVal;
			yyval.createAlterTriggerNode->type = yyvsp[-2].nullableUint64Val;
			yyval.createAlterTriggerNode->position = yyvsp[-1].nullableIntVal;
			yyval.createAlterTriggerNode->ssDefiner = yyvsp[0].nullableSqlSecurityVal;
		}
break;
case 1058:
  if (!yytrial)
{ yyval.nullableUint64Val = Nullable<FB_UINT64>::val(yyvsp[-1].uint64Val + yyvsp[0].uint64Val - 1); }
break;
case 1059:
  if (!yytrial)
{ yyval.nullableUint64Val = Nullable<FB_UINT64>::empty(); }
break;
case 1060:
  if (!yytrial)
{ yyval.nullableSqlSecurityVal = Nullable<TriggerDefinition::SqlSecurity>::empty(); }
break;
case 1061:
  if (!yytrial)
{ yyval.nullableSqlSecurityVal = Nullable<TriggerDefinition::SqlSecurity>::val(TriggerDefinition::SS_DEFINER); }
break;
case 1062:
  if (!yytrial)
{ yyval.nullableSqlSecurityVal = Nullable<TriggerDefinition::SqlSecurity>::val(TriggerDefinition::SS_INVOKER); }
break;
case 1063:
  if (!yytrial)
{ yyval.nullableSqlSecurityVal = Nullable<TriggerDefinition::SqlSecurity>::val(TriggerDefinition::SS_DROP); }
break;
case 1064:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].ddlNode; }
break;
case 1065:
  if (!yytrial)
{ yyval.ddlNode = newNode<DropExceptionNode>(*yyvsp[0].metaNamePtr); }
break;
case 1066:
  if (!yytrial)
{ yyval.ddlNode = newNode<DropIndexNode>(*yyvsp[0].metaNamePtr); }
break;
case 1067:
  if (!yytrial)
{ yyval.ddlNode = newNode<DropProcedureNode>(*yyvsp[0].metaNamePtr); }
break;
case 1068:
  if (!yytrial)
{ yyval.ddlNode = newNode<DropRelationNode>(*yyvsp[0].metaNamePtr, false); }
break;
case 1069:
  if (!yytrial)
{ yyval.ddlNode = newNode<DropTriggerNode>(*yyvsp[0].metaNamePtr); }
break;
case 1070:
  if (!yytrial)
{ yyval.ddlNode = newNode<DropRelationNode>(*yyvsp[0].metaNamePtr, true); }
break;
case 1071:
  if (!yytrial)
{ yyval.ddlNode = newNode<DropFilterNode>(*yyvsp[0].metaNamePtr); }
break;
case 1072:
  if (!yytrial)
{ yyval.ddlNode = newNode<DropDomainNode>(*yyvsp[0].metaNamePtr); }
break;
case 1073:
  if (!yytrial)
{ yyval.ddlNode = newNode<DropFunctionNode>(*yyvsp[0].metaNamePtr); }
break;
case 1074:
  if (!yytrial)
{ yyval.ddlNode = newNode<DropFunctionNode>(*yyvsp[0].metaNamePtr); }
break;
case 1075:
  if (!yytrial)
{ yyval.ddlNode = newNode<DropShadowNode>(yyvsp[-1].int32Val, yyvsp[0].boolVal); }
break;
case 1076:
  if (!yytrial)
{ yyval.ddlNode = newNode<DropRoleNode>(*yyvsp[0].metaNamePtr); }
break;
case 1077:
  if (!yytrial)
{ yyval.ddlNode = newNode<DropSequenceNode>(*yyvsp[0].metaNamePtr); }
break;
case 1078:
  if (!yytrial)
{ yyval.ddlNode = newNode<DropSequenceNode>(*yyvsp[0].metaNamePtr); }
break;
case 1079:
  if (!yytrial)
{ yyval.ddlNode = newNode<DropCollationNode>(*yyvsp[0].metaNamePtr); }
break;
case 1080:
  if (!yytrial)
{ yyval.ddlNode = newNode<DropUserNode>(*yyvsp[-3].metaNamePtr, yyvsp[0].metaNamePtr); }
break;
case 1081:
  if (!yytrial)
{ yyval.ddlNode = newNode<DropUserNode>(*yyvsp[0].metaNamePtr); }
break;
case 1082:
  if (!yytrial)
{ yyval.ddlNode = newNode<DropPackageNode>(*yyvsp[0].metaNamePtr); }
break;
case 1083:
  if (!yytrial)
{ yyval.ddlNode = newNode<DropPackageBodyNode>(*yyvsp[0].metaNamePtr); }
break;
case 1084:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].mappingNode; }
break;
case 1085:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].mappingNode; }
break;
case 1086:
  if (!yytrial)
{ yyval.boolVal = false; }
break;
case 1087:
  if (!yytrial)
{ yyval.boolVal = true; }
break;
case 1088:
  if (!yytrial)
{ yyval.boolVal = false; }
break;
case 1092:
  if (!yytrial)
{
			yyval.legacyField = yyvsp[-2].legacyField;
			yyval.legacyField->notNull = true;
		}
break;
case 1095:
  if (!yytrial)
{
			yyval.legacyField = newNode<dsql_fld>();
			yyval.legacyField->typeOfName = *yyvsp[0].metaNamePtr;
		}
break;
case 1096:
  if (!yytrial)
{
			yyval.legacyField = newNode<dsql_fld>();
			yyval.legacyField->typeOfName = *yyvsp[0].metaNamePtr;
			yyval.legacyField->typeOfTable = *yyvsp[-2].metaNamePtr;
		}
break;
case 1097:
  if (!yytrial)
{
			yyval.legacyField = newNode<dsql_fld>();
			yyval.legacyField->typeOfName = *yyvsp[0].metaNamePtr;
			yyval.legacyField->fullDomain = true;
		}
break;
case 1100:
  if (!yytrial)
{
			yyvsp[-3].legacyField->ranges = yyvsp[-1].valueListNode;
			yyvsp[-3].legacyField->dimensions = yyvsp[-3].legacyField->ranges->items.getCount() / 2;
			yyvsp[-3].legacyField->elementDtype = yyvsp[-3].legacyField->dtype;
			yyval.legacyField = yyvsp[-3].legacyField;
		}
break;
case 1101:
  if (!yytrial)
{
			yyvsp[-4].legacyField->ranges = yyvsp[-2].valueListNode;
			yyvsp[-4].legacyField->dimensions = yyvsp[-4].legacyField->ranges->items.getCount() / 2;
			yyvsp[-4].legacyField->elementDtype = yyvsp[-4].legacyField->dtype;
			if (yyvsp[0].metaNamePtr)
				yyvsp[-4].legacyField->charSet = *yyvsp[0].metaNamePtr;
			yyval.legacyField = yyvsp[-4].legacyField;
		}
break;
case 1103:
  if (!yytrial)
{ yyval.valueListNode = yyvsp[-2].valueListNode->add(yyvsp[0].valueListNode->items[0])->add(yyvsp[0].valueListNode->items[1]); }
break;
case 1104:
  if (!yytrial)
{
			if (yyvsp[0].int32Val < 1)
		 		yyval.valueListNode = newNode<ValueListNode>(MAKE_const_slong(yyvsp[0].int32Val))->add(MAKE_const_slong(1));
			else
		 		yyval.valueListNode = newNode<ValueListNode>(MAKE_const_slong(1))->add(MAKE_const_slong(yyvsp[0].int32Val));
		}
break;
case 1105:
  if (!yytrial)
{ yyval.valueListNode = newNode<ValueListNode>(MAKE_const_slong(yyvsp[-2].int32Val))->add(MAKE_const_slong(yyvsp[0].int32Val)); }
break;
case 1107:
  if (!yytrial)
{
			yyval.legacyField = yyvsp[-1].legacyField;
			if (yyvsp[0].metaNamePtr)
			{
				yyval.legacyField->charSet = *yyvsp[0].metaNamePtr;
				yyval.legacyField->flags |= FLD_has_chset;
			}
		}
break;
case 1113:
  if (!yytrial)
{
			yyval.legacyField = newNode<dsql_fld>();

			if (client_dialect < SQL_DIALECT_V6_TRANSITION)
			{
				ERRD_post (Arg::Gds(isc_sqlerr) << Arg::Num(-104) <<
					Arg::Gds(isc_sql_dialect_datatype_unsupport) << Arg::Num(client_dialect) <<
																	Arg::Str("BIGINT"));
			}

			if (db_dialect < SQL_DIALECT_V6_TRANSITION)
			{
				ERRD_post (Arg::Gds(isc_sqlerr) << Arg::Num(-104) <<
					Arg::Gds(isc_sql_db_dialect_dtype_unsupport) << Arg::Num(db_dialect) <<
																Arg::Str("BIGINT"));
			}

			yyval.legacyField->dtype = dtype_int64;
			yyval.legacyField->length = sizeof(SINT64);
			yyval.legacyField->flags |= FLD_has_prec;
		}
break;
case 1114:
  if (!yytrial)
{
			yyval.legacyField = newNode<dsql_fld>();
			yyval.legacyField->dtype = dtype_int128;
			yyval.legacyField->length = sizeof(Int128);
			yyval.legacyField->flags |= FLD_has_prec;
		}
break;
case 1115:
  if (!yytrial)
{
			yyval.legacyField = newNode<dsql_fld>();
			yyval.legacyField->dtype = dtype_long;
			yyval.legacyField->length = sizeof(SLONG);
			yyval.legacyField->flags |= FLD_has_prec;
		}
break;
case 1116:
  if (!yytrial)
{
			yyval.legacyField = newNode<dsql_fld>();
			yyval.legacyField->dtype = dtype_short;
			yyval.legacyField->length = sizeof(SSHORT);
			yyval.legacyField->flags |= FLD_has_prec;
		}
break;
case 1117:
  if (!yytrial)
{
			yyval.legacyField = newNode<dsql_fld>();
			stmt_ambiguous = true;

			if (client_dialect <= SQL_DIALECT_V5)
			{
				/* Post warning saying that DATE is equivalent to TIMESTAMP*/
				ERRD_post_warning(Arg::Warning(isc_sqlwarn) << Arg::Num(301) <<
								  Arg::Warning(isc_dtype_renamed));
				yyval.legacyField->dtype = dtype_timestamp;
				yyval.legacyField->length = sizeof(GDS_TIMESTAMP);
			}
			else if (client_dialect == SQL_DIALECT_V6_TRANSITION)
				yyabandon(YYPOSNARG(1), -104, isc_transitional_date);
			else
			{
				yyval.legacyField->dtype = dtype_sql_date;
				yyval.legacyField->length = sizeof(ULONG);
			}
			yyval.legacyField->flags |= FLD_has_prec;
		}
break;
case 1118:
  if (!yytrial)
{
			yyval.legacyField = newNode<dsql_fld>();

			checkTimeDialect();
			yyval.legacyField->dtype = dtype_sql_time;
			yyval.legacyField->length = sizeof(SLONG);
			yyval.legacyField->flags |= FLD_has_prec;
		}
break;
case 1119:
  if (!yytrial)
{
			yyval.legacyField = newNode<dsql_fld>();

			checkTimeDialect();
			yyval.legacyField->dtype = dtype_sql_time_tz;
			yyval.legacyField->length = sizeof(ISC_TIME_TZ);
			yyval.legacyField->flags |= FLD_has_prec;
		}
break;
case 1120:
  if (!yytrial)
{
			yyval.legacyField = newNode<dsql_fld>();
			yyval.legacyField->dtype = dtype_timestamp;
			yyval.legacyField->length = sizeof(GDS_TIMESTAMP);
			yyval.legacyField->flags |= FLD_has_prec;
		}
break;
case 1121:
  if (!yytrial)
{
			yyval.legacyField = newNode<dsql_fld>();
			yyval.legacyField->dtype = dtype_timestamp_tz;
			yyval.legacyField->length = sizeof(ISC_TIMESTAMP_TZ);
			yyval.legacyField->flags |= FLD_has_prec;
		}
break;
case 1122:
  if (!yytrial)
{
			yyval.legacyField = newNode<dsql_fld>();
			yyval.legacyField->dtype = dtype_boolean;
			yyval.legacyField->length = sizeof(UCHAR);
		}
break;
case 1127:
  if (!yytrial)
{ yyval.legacyField = newNode<dsql_fld>(); }
break;
case 1128:
yyval.legacyField = NOTRIAL(yyvsp[0].legacyField);
break;
case 1129:
  if (!yytrial)
{
			yyval.legacyField = yyvsp[-4].legacyField;
			yyval.legacyField->dtype = dtype_blob;
			yyval.legacyField->length = sizeof(ISC_QUAD);
			yyval.legacyField->segLength = yyvsp[-1].uintVal;
			if (yyvsp[0].metaNamePtr)
			{
				yyval.legacyField->charSet = *yyvsp[0].metaNamePtr;
				yyval.legacyField->flags |= FLD_has_chset;
			}
		}
break;
case 1130:
  if (!yytrial)
{
			yyval.legacyField = newNode<dsql_fld>();
			yyval.legacyField->dtype = dtype_blob;
			yyval.legacyField->length = sizeof(ISC_QUAD);
			yyval.legacyField->segLength = (USHORT) yyvsp[-1].int32Val;
			yyval.legacyField->subType = 0;
		}
break;
case 1131:
  if (!yytrial)
{
			yyval.legacyField = newNode<dsql_fld>();
			yyval.legacyField->dtype = dtype_blob;
			yyval.legacyField->length = sizeof(ISC_QUAD);
			yyval.legacyField->segLength = (USHORT) yyvsp[-3].int32Val;
			yyval.legacyField->subType = (USHORT) yyvsp[-1].int32Val;
			yyval.legacyField->flags |= FLD_has_sub;
		}
break;
case 1132:
  if (!yytrial)
{
			yyval.legacyField = newNode<dsql_fld>();
			yyval.legacyField->dtype = dtype_blob;
			yyval.legacyField->length = sizeof(ISC_QUAD);
			yyval.legacyField->segLength = 80;
			yyval.legacyField->subType = (USHORT) yyvsp[-1].int32Val;
			yyval.legacyField->flags |= FLD_has_sub;
		}
break;
case 1133:
  if (!yytrial)
{ yyval.uintVal = (USHORT) 80; }
break;
case 1134:
  if (!yytrial)
{ yyval.uintVal = (USHORT) yyvsp[0].int32Val; }
break;
case 1135:
  if (!yytrial)
{ yyvsp[0].legacyField->subType = (USHORT) 0; }
break;
case 1136:
  if (!yytrial)
{ yyvsp[-2].legacyField->subType = (USHORT) yyvsp[0].int32Val; yyvsp[-2].legacyField->flags |= FLD_has_sub; }
break;
case 1137:
  if (!yytrial)
{ yyvsp[-2].legacyField->subTypeName = *yyvsp[0].metaNamePtr; yyvsp[-2].legacyField->flags |= FLD_has_sub; }
break;
case 1138:
  if (!yytrial)
{ yyval.metaNamePtr = NULL; }
break;
case 1139:
  if (!yytrial)
{ yyval.metaNamePtr = yyvsp[0].metaNamePtr; }
break;
case 1140:
  if (!yytrial)
{
			yyval.legacyField = newNode<dsql_fld>();
			yyval.legacyField->dtype = dtype_text;
			yyval.legacyField->charLength = (USHORT) yyvsp[-1].int32Val;
			yyval.legacyField->flags |= (FLD_national | FLD_has_len);
		}
break;
case 1141:
  if (!yytrial)
{
			yyval.legacyField = newNode<dsql_fld>();
			yyval.legacyField->dtype = dtype_text;
			yyval.legacyField->charLength = 1;
			yyval.legacyField->flags |= FLD_national;
		}
break;
case 1142:
  if (!yytrial)
{
			yyval.legacyField = newNode<dsql_fld>();
			yyval.legacyField->dtype = dtype_varying;
			yyval.legacyField->charLength = (USHORT) yyvsp[-1].int32Val;
			yyval.legacyField->flags |= (FLD_national | FLD_has_len);
		}
break;
case 1143:
  if (!yytrial)
{
			yyval.legacyField = newNode<dsql_fld>();
			yyval.legacyField->dtype = dtype_text;
			yyval.legacyField->charLength = (USHORT) yyvsp[-1].int32Val;
			yyval.legacyField->length = (USHORT) yyvsp[-1].int32Val;
			yyval.legacyField->textType = ttype_binary;
			yyval.legacyField->charSetId = CS_BINARY;
			yyval.legacyField->subType = fb_text_subtype_binary;
			yyval.legacyField->flags |= (FLD_has_len | FLD_has_chset);
		}
break;
case 1144:
  if (!yytrial)
{
			yyval.legacyField = newNode<dsql_fld>();
			yyval.legacyField->dtype = dtype_text;
			yyval.legacyField->charLength = 1;
			yyval.legacyField->length = 1;
			yyval.legacyField->textType = ttype_binary;
			yyval.legacyField->charSetId = CS_BINARY;
			yyval.legacyField->subType = fb_text_subtype_binary;
			yyval.legacyField->flags |= FLD_has_chset;
		}
break;
case 1145:
  if (!yytrial)
{
			yyval.legacyField = newNode<dsql_fld>();
			yyval.legacyField->dtype = dtype_varying;
			yyval.legacyField->charLength = (USHORT) yyvsp[-1].int32Val;
			yyval.legacyField->length = (USHORT) yyvsp[-1].int32Val + sizeof(USHORT);
			yyval.legacyField->textType = ttype_binary;
			yyval.legacyField->charSetId = CS_BINARY;
			yyval.legacyField->subType = fb_text_subtype_binary;
			yyval.legacyField->flags |= (FLD_has_len | FLD_has_chset);
		}
break;
case 1146:
  if (!yytrial)
{
			yyval.legacyField = newNode<dsql_fld>();
			yyval.legacyField->dtype = dtype_text;
			yyval.legacyField->charLength = (USHORT) yyvsp[-1].int32Val;
			yyval.legacyField->flags |= FLD_has_len;
		}
break;
case 1147:
  if (!yytrial)
{
			yyval.legacyField = newNode<dsql_fld>();
			yyval.legacyField->dtype = dtype_text;
			yyval.legacyField->charLength = 1;
		}
break;
case 1148:
  if (!yytrial)
{
			yyval.legacyField = newNode<dsql_fld>();
			yyval.legacyField->dtype = dtype_varying;
			yyval.legacyField->charLength = (USHORT) yyvsp[-1].int32Val;
			yyval.legacyField->flags |= FLD_has_len;
		}
break;
case 1160:
  if (!yytrial)
{
			SLONG precision = yyvsp[0].int32Val;

			if (precision != 0 && precision != 16 && precision != 34)
				yyabandon(YYPOSNARG(2), -842, isc_decprecision_err);	/* DecFloat precision must be 16 or 34.*/

			yyval.legacyField = newNode<dsql_fld>();
			if (precision)
				yyval.legacyField->flags |= FLD_has_prec;
			yyval.legacyField->precision = precision == 0 ? 34 : (USHORT) precision;
			yyval.legacyField->dtype = precision == 16 ? dtype_dec64 : dtype_dec128;
			yyval.legacyField->length = precision == 16 ? sizeof(Decimal64) : sizeof(Decimal128);
		}
break;
case 1161:
  if (!yytrial)
{
			yyval.legacyField = yyvsp[0].legacyField;
			yyval.legacyField->subType = dsc_num_type_numeric;
			yyval.legacyField->flags |= FLD_has_sub;
		}
break;
case 1162:
  if (!yytrial)
{
			yyval.legacyField = yyvsp[0].legacyField;
			yyval.legacyField->subType = dsc_num_type_decimal;
			yyval.legacyField->flags |= FLD_has_sub;

			if (yyval.legacyField->dtype == dtype_short)
			{
				yyval.legacyField->dtype = dtype_long;
				yyval.legacyField->length = sizeof(SLONG);
			}
		}
break;
case 1163:
  if (!yytrial)
{
			yyval.legacyField = newNode<dsql_fld>();
			yyval.legacyField->dtype = dtype_long;
			yyval.legacyField->length = sizeof(SLONG);
			yyval.legacyField->precision = 9;
		}
break;
case 1164:
  if (!yytrial)
{
			yyval.legacyField = newNode<dsql_fld>();
			yyval.legacyField->flags |= FLD_has_prec;

			if (yyvsp[-1].int32Val < 1 || yyvsp[-1].int32Val > 38)
				yyabandon(YYPOSNARG(2), -842, Arg::Gds(isc_precision_err2) << Arg::Num(1) << Arg::Num(38));
																/* Precision must be between 1 and 38*/

			if (yyvsp[-1].int32Val > 18)
			{
				yyval.legacyField->dtype = dtype_int128;
				yyval.legacyField->length = sizeof(Int128);
			}
			else if (yyvsp[-1].int32Val > 9)
			{
				if ( ( (client_dialect <= SQL_DIALECT_V5) && (db_dialect > SQL_DIALECT_V5) ) ||
					( (client_dialect > SQL_DIALECT_V5) && (db_dialect <= SQL_DIALECT_V5) ) )
				{
					ERRD_post(Arg::Gds(isc_sqlerr) << Arg::Num(-817) <<
							  Arg::Gds(isc_ddl_not_allowed_by_db_sql_dial) << Arg::Num(db_dialect));
				}

				if (client_dialect <= SQL_DIALECT_V5)
				{
					yyval.legacyField->dtype = dtype_double;
					yyval.legacyField->length = sizeof(double);
				}
				else
				{
					if (client_dialect == SQL_DIALECT_V6_TRANSITION)
					{
						ERRD_post_warning(Arg::Warning(isc_dsql_warn_precision_ambiguous));
						ERRD_post_warning(Arg::Warning(isc_dsql_warn_precision_ambiguous1));
						ERRD_post_warning(Arg::Warning(isc_dsql_warn_precision_ambiguous2));
					}

					yyval.legacyField->dtype = dtype_int64;
					yyval.legacyField->length = sizeof(SINT64);
				}
			}
			else
			{
				if (yyvsp[-1].int32Val < 5)
				{
					yyval.legacyField->dtype = dtype_short;
					yyval.legacyField->length = sizeof(SSHORT);
				}
				else
				{
					yyval.legacyField->dtype = dtype_long;
					yyval.legacyField->length = sizeof(SLONG);
				}
			}

			yyval.legacyField->precision = (USHORT) yyvsp[-1].int32Val;
		}
break;
case 1165:
  if (!yytrial)
{
			yyval.legacyField = newNode<dsql_fld>();
			yyval.legacyField->flags |= (FLD_has_prec | FLD_has_scale);

			if (yyvsp[-3].int32Val < 1 || yyvsp[-3].int32Val > 38)
				yyabandon(YYPOSNARG(2), -842, Arg::Gds(isc_precision_err2) << Arg::Num(1) << Arg::Num(38));
																/* Precision must be between 1 and 38*/

			if (yyvsp[-1].int32Val > yyvsp[-3].int32Val || yyvsp[-1].int32Val < 0)
				yyabandon(YYPOSNARG(4), -842, isc_scale_nogt);	/* Scale must be between 0 and precision*/

			if (yyvsp[-3].int32Val > 18)
			{
				yyval.legacyField->dtype = dtype_int128;
				yyval.legacyField->length = sizeof(Int128);
			}
			else if (yyvsp[-3].int32Val > 9)
			{
				if ( ( (client_dialect <= SQL_DIALECT_V5) && (db_dialect > SQL_DIALECT_V5) ) ||
					( (client_dialect > SQL_DIALECT_V5) && (db_dialect <= SQL_DIALECT_V5) ) )
				{
					ERRD_post(Arg::Gds(isc_sqlerr) << Arg::Num(-817) <<
							  Arg::Gds(isc_ddl_not_allowed_by_db_sql_dial) << Arg::Num(db_dialect));
				}

				if (client_dialect <= SQL_DIALECT_V5)
				{
					yyval.legacyField->dtype = dtype_double;
					yyval.legacyField->length = sizeof(double);
				}
				else
				{
					if (client_dialect == SQL_DIALECT_V6_TRANSITION)
					{
						ERRD_post_warning(Arg::Warning(isc_dsql_warn_precision_ambiguous));
						ERRD_post_warning(Arg::Warning(isc_dsql_warn_precision_ambiguous1));
						ERRD_post_warning(Arg::Warning(isc_dsql_warn_precision_ambiguous2));
					}
					/* client_dialect >= SQL_DIALECT_V6*/
					yyval.legacyField->dtype = dtype_int64;
					yyval.legacyField->length = sizeof(SINT64);
				}
			}
			else
			{
				if (yyvsp[-3].int32Val < 5)
				{
					yyval.legacyField->dtype = dtype_short;
					yyval.legacyField->length = sizeof(SSHORT);
				}
				else
				{
					yyval.legacyField->dtype = dtype_long;
					yyval.legacyField->length = sizeof(SLONG);
				}
			}

			yyval.legacyField->precision = (USHORT) yyvsp[-3].int32Val;
			yyval.legacyField->scale = - (SSHORT) yyvsp[-1].int32Val;
		}
break;
case 1168:
  if (!yytrial)
{
		    /* Precision is binary digits of the significand: 1-24 for 32 bit single precision, 25-53 for 64 bit double precision*/
			/* Precision 0 is the 'no precision specified' case, which defaults to 32 bit single precision*/
			SLONG precision = yyvsp[0].int32Val;

			if (precision != 0 && (precision < 1 || precision > 53))
				yyabandon(YYPOSNARG(2), -842, Arg::Gds(isc_precision_err2) << Arg::Num(1) << Arg::Num(53));
																/* Precision must be between 1 and 53*/

			yyval.legacyField = newNode<dsql_fld>();

			if (precision > 24)
			{
				yyval.legacyField->dtype = dtype_double;
				yyval.legacyField->length = sizeof(double);
			}
			else
			{
				yyval.legacyField->dtype = dtype_real;
				yyval.legacyField->length = sizeof(float);
			}
		}
break;
case 1169:
  if (!yytrial)
{
			/* Precision is binary digits of the significand: 1-53 for 64 bit double precision*/
			/* Precision 0 is the 'no precision specified case', which defaults to 64 bit double precision*/
			SLONG precision = yyvsp[0].int32Val;

			if (precision != 0 && (precision < 1 || precision > 53))
				yyabandon(YYPOSNARG(3), -842, Arg::Gds(isc_precision_err2) << Arg::Num(1) << Arg::Num(53));
																/* Precision must be between 1 and 53*/

			yyval.legacyField = newNode<dsql_fld>();
			yyval.legacyField->dtype = dtype_double;
			yyval.legacyField->length = sizeof(double);
		}
break;
case 1170:
  if (!yytrial)
{
			yyval.legacyField = newNode<dsql_fld>();
			yyval.legacyField->dtype = dtype_real;
			yyval.legacyField->length = sizeof(float);
		}
break;
case 1171:
  if (!yytrial)
{
			yyval.legacyField = newNode<dsql_fld>();
			yyval.legacyField->dtype = dtype_double;
			yyval.legacyField->length = sizeof(double);
		}
break;
case 1172:
  if (!yytrial)
{ yyval.int32Val = 0; }
break;
case 1173:
  if (!yytrial)
{ yyval.int32Val = yyvsp[-1].int32Val; }
break;
case 1177:
  if (!yytrial)
{
			UserSavepointNode* node = newNode<UserSavepointNode>();
			node->command = UserSavepointNode::CMD_SET;
			node->name = *yyvsp[0].metaNamePtr;
			yyval.stmtNode = node;
		}
break;
case 1178:
  if (!yytrial)
{
			UserSavepointNode* node = newNode<UserSavepointNode>();
			node->command = (yyvsp[0].boolVal ? UserSavepointNode::CMD_RELEASE_ONLY : UserSavepointNode::CMD_RELEASE);
			node->name = *yyvsp[-1].metaNamePtr;
			yyval.stmtNode = node;
		}
break;
case 1179:
  if (!yytrial)
{ yyval.boolVal = false; }
break;
case 1180:
  if (!yytrial)
{ yyval.boolVal = true; }
break;
case 1181:
  if (!yytrial)
{
			UserSavepointNode* node = newNode<UserSavepointNode>();
			node->command = UserSavepointNode::CMD_ROLLBACK;
			node->name = *yyvsp[0].metaNamePtr;
			yyval.stmtNode = node;
		}
break;
case 1184:
  if (!yytrial)
{ yyval.traNode = newNode<CommitRollbackNode>(CommitRollbackNode::CMD_COMMIT, yyvsp[0].boolVal); }
break;
case 1185:
  if (!yytrial)
{ yyval.traNode = newNode<CommitRollbackNode>(CommitRollbackNode::CMD_ROLLBACK, yyvsp[0].boolVal); }
break;
case 1188:
  if (!yytrial)
{ yyval.boolVal = false; }
break;
case 1189:
  if (!yytrial)
{ yyval.boolVal = true; }
break;
case 1192:
  if (!yytrial)
{ yyval.setTransactionNode = newNode<SetTransactionNode>(); }
break;
case 1193:
yyval.setTransactionNode = yyvsp[0].setTransactionNode;
break;
case 1194:
  if (!yytrial)
{ yyval.setTransactionNode = yyvsp[-2].setTransactionNode; }
break;
case 1195:
  if (!yytrial)
{ yyval.sessionResetNode = newNode<SessionResetNode>(); }
break;
case 1196:
  if (!yytrial)
{ yyval.setRoleNode = newNode<SetRoleNode>(yyvsp[0].metaNamePtr); }
break;
case 1197:
  if (!yytrial)
{ yyval.setRoleNode = newNode<SetRoleNode>(); }
break;
case 1198:
  if (!yytrial)
{ yyval.setDecFloatRoundNode = newNode<SetDecFloatRoundNode>(yyvsp[0].metaNamePtr); }
break;
case 1199:
  if (!yytrial)
{ yyval.setDecFloatTrapsNode = newNode<SetDecFloatTrapsNode>(); }
break;
case 1200:
yyval.setDecFloatTrapsNode = yyvsp[0].setDecFloatTrapsNode;
break;
case 1201:
  if (!yytrial)
{ yyval.setDecFloatTrapsNode = yyvsp[-2].setDecFloatTrapsNode; }
break;
case 1202:
  if (!yytrial)
{ yyval.setBindNode = newNode<SetBindNode>(); yyval.setBindNode->from = yyvsp[-2].legacyField; yyval.setBindNode->to = yyvsp[0].legacyField; }
break;
case 1204:
  if (!yytrial)
{
			yyval.legacyField = newNode<dsql_fld>();
			yyval.legacyField->dtype = dtype_timestamp_tz;
			yyval.legacyField->length = 0;
		}
break;
case 1206:
  if (!yytrial)
{
			yyval.legacyField = newNode<dsql_fld>();
			yyval.legacyField->dtype = dtype_varying;
			yyval.legacyField->charLength = 0;
		}
break;
case 1207:
  if (!yytrial)
{
			yyval.legacyField = yyvsp[0].legacyField;
		}
break;
case 1208:
  if (!yytrial)
{
			yyval.legacyField = newNode<dsql_fld>();
			yyval.legacyField->flags = FLD_legacy;
		}
break;
case 1209:
  if (!yytrial)
{
			yyval.legacyField = newNode<dsql_fld>();
			yyval.legacyField->flags = FLD_native;
		}
break;
case 1210:
  if (!yytrial)
{
			yyval.legacyField = newNode<dsql_fld>();
			yyval.legacyField->flags = FLD_extended;
		}
break;
case 1211:
  if (!yytrial)
{
			yyval.legacyField = newNode<dsql_fld>();
			checkTimeDialect();
			yyval.legacyField->dtype = dtype_ex_time_tz;
			yyval.legacyField->length = sizeof(ISC_TIME_TZ_EX);
			yyval.legacyField->flags |= FLD_has_prec;
		}
break;
case 1212:
  if (!yytrial)
{
			yyval.legacyField = newNode<dsql_fld>();
			yyval.legacyField->dtype = dtype_ex_timestamp_tz;
			yyval.legacyField->length = sizeof(ISC_TIMESTAMP_TZ_EX);
			yyval.legacyField->flags |= FLD_has_prec;
		}
break;
case 1216:
yyval.setDecFloatTrapsNode = yyvsp[-2].setDecFloatTrapsNode;
break;
case 1218:
  if (!yytrial)
{ yyvsp[-1].setDecFloatTrapsNode->trap(yyvsp[0].metaNamePtr); }
break;
case 1219:
  if (!yytrial)
{ yyval.setSessionNode = newNode<SetSessionNode>(SetSessionNode::TYPE_IDLE_TIMEOUT, yyvsp[-1].int32Val, yyvsp[0].blrOp); }
break;
case 1220:
  if (!yytrial)
{ yyval.setSessionNode = newNode<SetSessionNode>(SetSessionNode::TYPE_STMT_TIMEOUT, yyvsp[-1].int32Val, yyvsp[0].blrOp); }
break;
case 1221:
  if (!yytrial)
{ yyval.blrOp = blr_extract_minute; }
break;
case 1222:
  if (!yytrial)
{ yyval.blrOp = blr_extract_hour; }
break;
case 1223:
  if (!yytrial)
{ yyval.blrOp = blr_extract_minute; }
break;
case 1224:
  if (!yytrial)
{ yyval.blrOp = blr_extract_second; }
break;
case 1225:
  if (!yytrial)
{ yyval.blrOp = blr_extract_second; }
break;
case 1226:
  if (!yytrial)
{ yyval.blrOp = blr_extract_hour; }
break;
case 1227:
  if (!yytrial)
{ yyval.blrOp = blr_extract_minute; }
break;
case 1228:
  if (!yytrial)
{ yyval.blrOp = blr_extract_second; }
break;
case 1229:
  if (!yytrial)
{ yyval.blrOp = blr_extract_millisecond; }
break;
case 1230:
  if (!yytrial)
{ yyval.mngNode = yyvsp[0].mngNode; }
break;
case 1231:
  if (!yytrial)
{ yyval.mngNode = newNode<SetTimeZoneNode>(yyvsp[0].intlStringPtr->getString()); }
break;
case 1232:
  if (!yytrial)
{ yyval.mngNode = newNode<SetTimeZoneNode>(); }
break;
case 1236:
yyval.setTransactionNode = yyvsp[-1].setTransactionNode;
break;
case 1238:
  if (!yytrial)
{ setClause(yyvsp[-2].setTransactionNode->readOnly, "READ {ONLY | WRITE}", true); }
break;
case 1239:
  if (!yytrial)
{ setClause(yyvsp[-2].setTransactionNode->readOnly, "READ {ONLY | WRITE}", false); }
break;
case 1240:
  if (!yytrial)
{ setClause(yyvsp[-1].setTransactionNode->wait, "[NO] WAIT", true); }
break;
case 1241:
  if (!yytrial)
{ setClause(yyvsp[-2].setTransactionNode->wait, "[NO] WAIT", false); }
break;
case 1242:
  if (!yytrial)
{ setClause(yyvsp[-2].setTransactionNode->isoLevel, "ISOLATION LEVEL", yyvsp[0].uintVal); }
break;
case 1243:
  if (!yytrial)
{ setClause(yyvsp[-3].setTransactionNode->noAutoUndo, "NO AUTO UNDO", true); }
break;
case 1244:
  if (!yytrial)
{ setClause(yyvsp[-2].setTransactionNode->ignoreLimbo, "IGNORE LIMBO", true); }
break;
case 1245:
  if (!yytrial)
{ setClause(yyvsp[-2].setTransactionNode->restartRequests, "RESTART REQUESTS", true); }
break;
case 1246:
  if (!yytrial)
{ setClause(yyvsp[-2].setTransactionNode->autoCommit, "AUTO COMMIT", true); }
break;
case 1247:
  if (!yytrial)
{ setClause(yyvsp[-3].setTransactionNode->lockTimeout, "LOCK TIMEOUT", (USHORT) yyvsp[0].int32Val); }
break;
case 1248:
  if (!yytrial)
{ checkDuplicateClause(yyvsp[-1].setTransactionNode->reserveList, "RESERVING"); }
break;
case 1249:
yyval.setTransactionNode = yyvsp[-2].setTransactionNode;
break;
case 1251:
  if (!yytrial)
{ yyval.uintVal = yyvsp[0].uintVal;}
break;
case 1253:
  if (!yytrial)
{ yyval.uintVal = yyvsp[0].uintVal; }
break;
case 1254:
  if (!yytrial)
{ yyval.uintVal = yyvsp[0].uintVal; }
break;
case 1255:
  if (!yytrial)
{ yyval.uintVal = yyvsp[0].uintVal; }
break;
case 1256:
  if (!yytrial)
{ yyval.uintVal = SetTransactionNode::ISO_LEVEL_CONCURRENCY; }
break;
case 1257:
  if (!yytrial)
{
			setClause(yyvsp[-4].setTransactionNode->atSnapshotNumber, "SNAPSHOT AT NUMBER", (CommitNumber) yyvsp[0].int64Val);
			yyval.uintVal = SetTransactionNode::ISO_LEVEL_CONCURRENCY;
		}
break;
case 1258:
  if (!yytrial)
{ yyval.uintVal = SetTransactionNode::ISO_LEVEL_CONSISTENCY; }
break;
case 1259:
  if (!yytrial)
{ yyval.uintVal = SetTransactionNode::ISO_LEVEL_CONSISTENCY; }
break;
case 1260:
  if (!yytrial)
{ yyval.int64Val = yyvsp[0].int32Val; }
break;
case 1261:
  if (!yytrial)
{ yyval.int64Val = yyvsp[0].scaledNumber.number; }
break;
case 1262:
  if (!yytrial)
{ yyval.uintVal = SetTransactionNode::ISO_LEVEL_READ_COMMITTED_NO_REC_VERSION; }
break;
case 1263:
  if (!yytrial)
{ yyval.uintVal = SetTransactionNode::ISO_LEVEL_READ_COMMITTED_REC_VERSION; }
break;
case 1264:
  if (!yytrial)
{ yyval.uintVal = SetTransactionNode::ISO_LEVEL_READ_COMMITTED_NO_REC_VERSION; }
break;
case 1265:
  if (!yytrial)
{ yyval.uintVal = SetTransactionNode::ISO_LEVEL_READ_COMMITTED_READ_CONSISTENCY; }
break;
case 1266:
  if (!yytrial)
{ yyval.uintVal = 0; }
break;
case 1267:
  if (!yytrial)
{ yyval.uintVal = SetTransactionNode::LOCK_MODE_SHARED; }
break;
case 1268:
  if (!yytrial)
{ yyval.uintVal = SetTransactionNode::LOCK_MODE_PROTECTED; }
break;
case 1269:
  if (!yytrial)
{ yyval.uintVal = SetTransactionNode::LOCK_MODE_READ; }
break;
case 1270:
  if (!yytrial)
{ yyval.uintVal = SetTransactionNode::LOCK_MODE_WRITE; }
break;
case 1271:
  if (!yytrial)
{ yyvsp[-1].setTransactionNode->reserveList.add(yyvsp[0].setTransactionRestrictionClause); }
break;
case 1272:
  if (!yytrial)
{ yyvsp[-3].setTransactionNode->reserveList.add(yyvsp[0].setTransactionRestrictionClause); }
break;
case 1273:
  if (!yytrial)
{ yyval.setTransactionRestrictionClause = newNode<SetTransactionNode::RestrictionOption>(yyvsp[-1].metaNameArray, yyvsp[0].uintVal); }
break;
case 1274:
  if (!yytrial)
{ yyval.uintVal = 0; }
break;
case 1275:
  if (!yytrial)
{ yyval.uintVal = yyvsp[-1].uintVal | yyvsp[0].uintVal; }
break;
case 1276:
  if (!yytrial)
{
			ObjectsArray<MetaName>* node = newNode<ObjectsArray<MetaName> >();
			node->add(*yyvsp[0].metaNamePtr);
			yyval.metaNameArray = node;
		}
break;
case 1277:
  if (!yytrial)
{
			ObjectsArray<MetaName>* node = yyvsp[-2].metaNameArray;
			node->add(*yyvsp[0].metaNamePtr);
			yyval.metaNameArray = node;
		}
break;
case 1278:
  if (!yytrial)
{ yyval.ddlNode = newNode<SetStatisticsNode>(*yyvsp[0].metaNamePtr); }
break;
case 1279:
  if (!yytrial)
{ yyval.ddlNode = newNode<CommentOnNode>(yyvsp[-2].intVal, QualifiedName(""), "", *yyvsp[0].stringPtr); }
break;
case 1280:
  if (!yytrial)
{ yyval.ddlNode = newNode<CommentOnNode>(yyvsp[-3].intVal, QualifiedName(*yyvsp[-2].metaNamePtr), "", *yyvsp[0].stringPtr); }
break;
case 1281:
  if (!yytrial)
{ yyval.ddlNode = newNode<CommentOnNode>(yyvsp[-4].intVal, QualifiedName(*yyvsp[-3].metaNamePtr), *yyvsp[-2].metaNamePtr, *yyvsp[0].stringPtr); }
break;
case 1282:
  if (!yytrial)
{ yyval.ddlNode = newNode<CommentOnNode>(yyvsp[-4].intVal, *yyvsp[-3].qualifiedNamePtr, *yyvsp[-2].metaNamePtr, *yyvsp[0].stringPtr); }
break;
case 1283:
  if (!yytrial)
{ yyval.ddlNode = newNode<CommentOnNode>(yyvsp[-3].intVal, *yyvsp[-2].qualifiedNamePtr, "", *yyvsp[0].stringPtr); }
break;
case 1284:
  if (!yytrial)
{ yyval.ddlNode = yyvsp[0].createAlterUserNode; }
break;
case 1285:
  if (!yytrial)
{
				yyval.createAlterUserNode = newNode<CreateAlterUserNode>(CreateAlterUserNode::USER_MOD, *yyvsp[0].metaNamePtr);
			}
break;
case 1286:
yyval.createAlterUserNode = yyvsp[0].createAlterUserNode;
break;
case 1287:
  if (!yytrial)
{
				CreateAlterUserNode* node = yyval.createAlterUserNode = yyvsp[-4].createAlterUserNode;
				node->comment = yyvsp[0].stringPtr;
			}
break;
case 1290:
  if (!yytrial)
{ yyval.intVal = obj_database; }
break;
case 1291:
  if (!yytrial)
{ yyval.intVal = obj_field; }
break;
case 1292:
  if (!yytrial)
{ yyval.intVal = obj_relation; }
break;
case 1293:
  if (!yytrial)
{ yyval.intVal = obj_view; }
break;
case 1294:
  if (!yytrial)
{ yyval.intVal = obj_trigger; }
break;
case 1295:
  if (!yytrial)
{ yyval.intVal = obj_blob_filter; }
break;
case 1296:
  if (!yytrial)
{ yyval.intVal = obj_exception; }
break;
case 1297:
  if (!yytrial)
{ yyval.intVal = obj_generator; }
break;
case 1298:
  if (!yytrial)
{ yyval.intVal = obj_generator; }
break;
case 1299:
  if (!yytrial)
{ yyval.intVal = obj_index; }
break;
case 1300:
  if (!yytrial)
{ yyval.intVal = obj_sql_role; }
break;
case 1301:
  if (!yytrial)
{ yyval.intVal = obj_charset; }
break;
case 1302:
  if (!yytrial)
{ yyval.intVal = obj_collation; }
break;
case 1303:
  if (!yytrial)
{ yyval.intVal = obj_package_header; }
break;
case 1304:
  if (!yytrial)
{ yyval.intVal = obj_relation; }
break;
case 1305:
  if (!yytrial)
{ yyval.intVal = obj_parameter; }
break;
case 1306:
  if (!yytrial)
{ yyval.intVal = obj_procedure; }
break;
case 1307:
  if (!yytrial)
{ yyval.intVal = obj_udf; }
break;
case 1308:
  if (!yytrial)
{ yyval.intVal = obj_procedure; }
break;
case 1309:
  if (!yytrial)
{ yyval.intVal = obj_udf; }
break;
case 1310:
  if (!yytrial)
{ yyval.intVal = obj_udf; }
break;
case 1311:
  if (!yytrial)
{ yyval.metaNamePtr = yyvsp[0].metaNamePtr; }
break;
case 1312:
  if (!yytrial)
{ yyval.qualifiedNamePtr = newNode<QualifiedName>(*yyvsp[0].metaNamePtr); }
break;
case 1313:
  if (!yytrial)
{ yyval.qualifiedNamePtr = newNode<QualifiedName>(*yyvsp[0].metaNamePtr, *yyvsp[-2].metaNamePtr); }
break;
case 1314:
  if (!yytrial)
{ yyval.stringPtr = yyvsp[0].stringPtr; }
break;
case 1315:
  if (!yytrial)
{ yyval.stringPtr = newString(""); }
break;
case 1316:
  if (!yytrial)
{
			SelectNode* node = newNode<SelectNode>();
			node->dsqlExpr = yyvsp[-2].selectExprNode;
			node->dsqlForUpdate = yyvsp[-1].boolVal;
			node->dsqlWithLock = yyvsp[0].boolVal;
			yyval.selectNode = node;
		}
break;
case 1317:
  if (!yytrial)
{ yyval.boolVal = false; }
break;
case 1318:
  if (!yytrial)
{ yyval.boolVal = true; /* for_update_list is ignored */ }
break;
case 1319:
  if (!yytrial)
{ yyval.valueListNode = NULL; }
break;
case 1320:
  if (!yytrial)
{ yyval.valueListNode = yyvsp[0].valueListNode; }
break;
case 1321:
  if (!yytrial)
{ yyval.boolVal = false; }
break;
case 1322:
  if (!yytrial)
{ yyval.boolVal = true; }
break;
case 1323:
  if (!yytrial)
{
			SelectExprNode* node = yyval.selectExprNode = newNode<SelectExprNode>();
			node->querySpec = yyvsp[-2].recSourceNode;
			node->orderClause = yyvsp[-1].valueListNode;
			node->rowsClause = yyvsp[0].rowsClause;
			node->withClause = yyvsp[-3].withClause;
		}
break;
case 1324:
  if (!yytrial)
{
			SelectExprNode* node = yyval.selectExprNode = newNode<SelectExprNode>();
			node->querySpec = yyvsp[-3].recSourceNode;
			node->orderClause = yyvsp[-2].valueListNode;

			if (yyvsp[-1].valueExprNode || yyvsp[0].valueExprNode)
			{
				RowsClause* rowsNode = newNode<RowsClause>();
				rowsNode->skip = yyvsp[-1].valueExprNode;
				rowsNode->length = yyvsp[0].valueExprNode;
				node->rowsClause = rowsNode;
			}

			node->withClause = yyvsp[-4].withClause;
		}
break;
case 1325:
  if (!yytrial)
{ yyval.withClause = NULL; }
break;
case 1326:
  if (!yytrial)
{
			yyval.withClause = yyvsp[0].withClause;
			yyval.withClause->recursive = true;
		}
break;
case 1327:
  if (!yytrial)
{ yyval.withClause = yyvsp[0].withClause; }
break;
case 1328:
  if (!yytrial)
{
			yyval.withClause = newNode<WithClause>();
			yyval.withClause->add(yyvsp[0].selectExprNode);
		}
break;
case 1329:
  if (!yytrial)
{
			yyval.withClause = yyvsp[-2].withClause;
			yyval.withClause->add(yyvsp[0].selectExprNode);
		}
break;
case 1330:
  if (!yytrial)
{
			yyval.selectExprNode = yyvsp[-1].selectExprNode;
			yyval.selectExprNode->dsqlFlags |= RecordSourceNode::DFLAG_DERIVED;
			yyval.selectExprNode->alias = yyvsp[-5].metaNamePtr->c_str();
			yyval.selectExprNode->columns = yyvsp[-4].metaNameArray;
		}
break;
case 1331:
  if (!yytrial)
{
			yyval.selectExprNode = yyvsp[0].selectExprNode;
			yyval.selectExprNode->dsqlFlags |= RecordSourceNode::DFLAG_VALUE;
		}
break;
case 1332:
  if (!yytrial)
{
			yyvsp[0].selectExprNode->dsqlFlags |= RecordSourceNode::DFLAG_SINGLETON;
			yyval.valueExprNode = newNode<SubQueryNode>(blr_via, yyvsp[0].selectExprNode);
		}
break;
case 1333:
  if (!yytrial)
{ yyval.recSourceNode = yyvsp[0].rseNode; }
break;
case 1334:
  if (!yytrial)
{
			UnionSourceNode* node = nodeAs<UnionSourceNode>(yyvsp[-3].recSourceNode);
			if (node && !node->dsqlAll)
				node->dsqlClauses->add(yyvsp[0].rseNode);
			else
			{
				node = newNode<UnionSourceNode>();
				node->dsqlClauses = newNode<RecSourceListNode>(yyvsp[-3].recSourceNode)->add(yyvsp[0].rseNode);
			}
			yyval.recSourceNode = node;
		}
break;
case 1335:
  if (!yytrial)
{
			UnionSourceNode* node = nodeAs<UnionSourceNode>(yyvsp[-3].recSourceNode);
			if (node && node->dsqlAll)
				node->dsqlClauses->add(yyvsp[0].rseNode);
			else
			{
				node = newNode<UnionSourceNode>();
				node->dsqlAll = true;
				node->dsqlClauses = newNode<RecSourceListNode>(yyvsp[-3].recSourceNode)->add(yyvsp[0].rseNode);
			}
			yyval.recSourceNode = node;
		}
break;
case 1337:
  if (!yytrial)
{
			RseNode* rse = newNode<RseNode>();
			rse->dsqlFirst = yyvsp[-8].valueListNode ? yyvsp[-8].valueListNode->items[1] : NULL;
			rse->dsqlSkip = yyvsp[-8].valueListNode ? yyvsp[-8].valueListNode->items[0] : NULL;
			rse->dsqlDistinct = yyvsp[-7].valueListNode;
			rse->dsqlSelectList = yyvsp[-6].valueListNode;
			rse->dsqlFrom = yyvsp[-5].recSourceListNode;
			rse->dsqlWhere = yyvsp[-4].boolExprNode;
			rse->dsqlGroup = yyvsp[-3].valueListNode;
			rse->dsqlHaving = yyvsp[-2].boolExprNode;
			rse->dsqlNamedWindows = yyvsp[-1].namedWindowsClause;
			rse->rse_plan = yyvsp[0].planNode;
			yyval.rseNode = rse;
		}
break;
case 1338:
  if (!yytrial)
{ yyval.valueListNode = NULL; }
break;
case 1339:
  if (!yytrial)
{ yyval.valueListNode = newNode<ValueListNode>(yyvsp[0].valueExprNode)->add(yyvsp[-1].valueExprNode); }
break;
case 1340:
  if (!yytrial)
{ yyval.valueListNode = newNode<ValueListNode>(1u)->add(yyvsp[0].valueExprNode); }
break;
case 1341:
  if (!yytrial)
{ yyval.valueListNode = newNode<ValueListNode>(yyvsp[0].valueExprNode)->add(NULL); }
break;
case 1342:
  if (!yytrial)
{ yyval.valueExprNode = MAKE_const_slong(yyvsp[0].int32Val); }
break;
case 1343:
  if (!yytrial)
{ yyval.valueExprNode = yyvsp[-1].valueExprNode; }
break;
case 1344:
  if (!yytrial)
{ yyval.valueExprNode = yyvsp[0].valueExprNode; }
break;
case 1345:
  if (!yytrial)
{ yyval.valueExprNode = MAKE_const_slong(yyvsp[0].int32Val); }
break;
case 1346:
  if (!yytrial)
{ yyval.valueExprNode = yyvsp[-1].valueExprNode; }
break;
case 1347:
  if (!yytrial)
{ yyval.valueExprNode = yyvsp[0].valueExprNode; }
break;
case 1348:
  if (!yytrial)
{ yyval.valueListNode = newNode<ValueListNode>(0); }
break;
case 1349:
  if (!yytrial)
{ yyval.valueListNode = NULL; }
break;
case 1350:
  if (!yytrial)
{ yyval.valueListNode = yyvsp[0].valueListNode; }
break;
case 1351:
  if (!yytrial)
{ yyval.valueListNode = NULL; }
break;
case 1352:
  if (!yytrial)
{ yyval.valueListNode = newNode<ValueListNode>(yyvsp[0].valueExprNode); }
break;
case 1353:
  if (!yytrial)
{ yyval.valueListNode = yyvsp[-2].valueListNode->add(yyvsp[0].valueExprNode); }
break;
case 1355:
  if (!yytrial)
{
			FieldNode* fieldNode = newNode<FieldNode>();
			fieldNode->dsqlQualifier = *yyvsp[-2].metaNamePtr;
			yyval.valueExprNode = fieldNode;
		}
break;
case 1357:
  if (!yytrial)
{ yyval.valueExprNode = newNode<DsqlAliasNode>(*yyvsp[0].metaNamePtr, yyvsp[-2].valueExprNode); }
break;
case 1360:
  if (!yytrial)
{ yyval.recSourceListNode = yyvsp[0].recSourceListNode; }
break;
case 1361:
  if (!yytrial)
{ yyval.recSourceListNode = newNode<RecSourceListNode>(yyvsp[0].recSourceNode); }
break;
case 1362:
  if (!yytrial)
{ yyval.recSourceListNode = yyvsp[-2].recSourceListNode->add(yyvsp[0].recSourceNode); }
break;
case 1366:
  if (!yytrial)
{ yyval.recSourceNode = yyvsp[0].selectExprNode; }
break;
case 1367:
  if (!yytrial)
{ yyval.recSourceNode = yyvsp[0].selectExprNode; }
break;
case 1368:
  if (!yytrial)
{ yyval.recSourceNode = yyvsp[-1].recSourceNode; }
break;
case 1369:
  if (!yytrial)
{
			yyval.selectExprNode = yyvsp[-4].selectExprNode;
			yyval.selectExprNode->dsqlFlags |= RecordSourceNode::DFLAG_DERIVED;
			if (yyvsp[-1].metaNamePtr)
				yyval.selectExprNode->alias = yyvsp[-1].metaNamePtr->c_str();
			yyval.selectExprNode->columns = yyvsp[0].metaNameArray;
		}
break;
case 1370:
  if (!yytrial)
{
			yyval.selectExprNode = yyvsp[0].selectExprNode;
			yyval.selectExprNode->dsqlFlags |= RecordSourceNode::DFLAG_LATERAL;
		}
break;
case 1371:
  if (!yytrial)
{ yyval.metaNamePtr = NULL; }
break;
case 1373:
  if (!yytrial)
{ yyval.metaNameArray = NULL; }
break;
case 1374:
  if (!yytrial)
{ yyval.metaNameArray = yyvsp[-1].metaNameArray; }
break;
case 1375:
  if (!yytrial)
{
			ObjectsArray<MetaName>* node = newNode<ObjectsArray<MetaName> >();
			node->add(*yyvsp[0].metaNamePtr);
			yyval.metaNameArray = node;
		}
break;
case 1376:
  if (!yytrial)
{
			ObjectsArray<MetaName>* node = yyvsp[-2].metaNameArray;
			node->add(*yyvsp[0].metaNamePtr);
			yyval.metaNameArray = node;
		}
break;
case 1380:
  if (!yytrial)
{
			RseNode* rse = newNode<RseNode>();
			rse->dsqlExplicitJoin = true;
			rse->rse_jointype = blr_inner;
			rse->dsqlFrom = newNode<RecSourceListNode>(yyvsp[-3].recSourceNode)->add(yyvsp[0].recSourceNode);
			yyval.recSourceNode = rse;
		}
break;
case 1381:
  if (!yytrial)
{
			RseNode* rse = newNode<RseNode>();
			rse->dsqlExplicitJoin = true;
			rse->rse_jointype = yyvsp[-2].blrOp;
			rse->dsqlFrom = newNode<RecSourceListNode>(yyvsp[-4].recSourceNode)->add(yyvsp[0].recSourceNode);
			rse->dsqlJoinUsing = newNode<ValueListNode>(0u);	/* using list with size 0 -> natural*/
			yyval.recSourceNode = rse;
		}
break;
case 1382:
  if (!yytrial)
{
			RseNode* rse = newNode<RseNode>();
			rse->dsqlExplicitJoin = true;
			rse->rse_jointype = yyvsp[-3].blrOp;
			rse->dsqlFrom = newNode<RecSourceListNode>(yyvsp[-4].recSourceNode);
			rse->dsqlFrom->add(yyvsp[-1].recSourceNode);
			rse->dsqlWhere = yyvsp[0].boolExprNode;
			yyval.recSourceNode = rse;
		}
break;
case 1383:
  if (!yytrial)
{
			RseNode* rse = newNode<RseNode>();
			rse->dsqlExplicitJoin = true;
			rse->rse_jointype = yyvsp[-3].blrOp;
			rse->dsqlFrom = newNode<RecSourceListNode>(yyvsp[-4].recSourceNode);
			rse->dsqlFrom->add(yyvsp[-1].recSourceNode);
			rse->dsqlJoinUsing = yyvsp[0].valueListNode;
			yyval.recSourceNode = rse;
		}
break;
case 1384:
  if (!yytrial)
{ yyval.boolExprNode = yyvsp[0].boolExprNode; }
break;
case 1385:
  if (!yytrial)
{ yyval.valueListNode = yyvsp[-1].valueListNode; }
break;
case 1386:
  if (!yytrial)
{
			ProcedureSourceNode* node = newNode<ProcedureSourceNode>(QualifiedName(*yyvsp[-3].metaNamePtr));
			node->sourceList = yyvsp[-2].valueListNode;
			node->alias = yyvsp[0].metaNamePtr->c_str();
			yyval.recSourceNode = node;
		}
break;
case 1387:
  if (!yytrial)
{
			ProcedureSourceNode* node = newNode<ProcedureSourceNode>(QualifiedName(*yyvsp[-1].metaNamePtr));
			node->sourceList = yyvsp[0].valueListNode;
			yyval.recSourceNode = node;
		}
break;
case 1388:
  if (!yytrial)
{
			ProcedureSourceNode* node = newNode<ProcedureSourceNode>(
				QualifiedName(*yyvsp[-3].metaNamePtr, *yyvsp[-5].metaNamePtr));
			node->sourceList = yyvsp[-2].valueListNode;
			node->alias = yyvsp[0].metaNamePtr->c_str();
			yyval.recSourceNode = node;
		}
break;
case 1389:
  if (!yytrial)
{
			ProcedureSourceNode* node = newNode<ProcedureSourceNode>(
				QualifiedName(*yyvsp[-1].metaNamePtr, *yyvsp[-3].metaNamePtr));
			node->sourceList = yyvsp[0].valueListNode;
			yyval.recSourceNode = node;
		}
break;
case 1390:
  if (!yytrial)
{ yyval.valueListNode = NULL; }
break;
case 1391:
  if (!yytrial)
{ yyval.valueListNode = yyvsp[-1].valueListNode; }
break;
case 1393:
  if (!yytrial)
{
			RelationSourceNode* node = newNode<RelationSourceNode>(*yyvsp[-2].metaNamePtr);
			node->alias = yyvsp[0].metaNamePtr->c_str();
			yyval.relSourceNode = node;
		}
break;
case 1394:
  if (!yytrial)
{ yyval.relSourceNode = newNode<RelationSourceNode>(*yyvsp[0].metaNamePtr); }
break;
case 1395:
  if (!yytrial)
{ yyval.blrOp = blr_inner; }
break;
case 1396:
  if (!yytrial)
{ yyval.blrOp = blr_inner; }
break;
case 1397:
  if (!yytrial)
{ yyval.blrOp = blr_left; }
break;
case 1398:
  if (!yytrial)
{ yyval.blrOp = blr_right; }
break;
case 1399:
  if (!yytrial)
{ yyval.blrOp = blr_full; }
break;
case 1402:
  if (!yytrial)
{ yyval.valueListNode = NULL; }
break;
case 1403:
  if (!yytrial)
{ yyval.valueListNode = yyvsp[0].valueListNode; }
break;
case 1404:
  if (!yytrial)
{ yyval.valueListNode = newNode<ValueListNode>(yyvsp[0].valueExprNode); }
break;
case 1405:
  if (!yytrial)
{ yyval.valueListNode = yyvsp[-2].valueListNode->add(yyvsp[0].valueExprNode); }
break;
case 1407:
  if (!yytrial)
{ yyval.boolExprNode = NULL; }
break;
case 1408:
  if (!yytrial)
{ yyval.boolExprNode = yyvsp[0].boolExprNode; }
break;
case 1409:
  if (!yytrial)
{ yyval.boolExprNode = NULL; }
break;
case 1410:
  if (!yytrial)
{ yyval.boolExprNode = yyvsp[0].boolExprNode; }
break;
case 1411:
  if (!yytrial)
{ yyval.namedWindowsClause = NULL; }
break;
case 1412:
  if (!yytrial)
{ yyval.namedWindowsClause = yyvsp[0].namedWindowsClause; }
break;
case 1413:
  if (!yytrial)
{
			NamedWindowsClause* node = newNode<NamedWindowsClause>();
			node->add(*yyvsp[0].namedWindowClause);
			yyval.namedWindowsClause = node;
		}
break;
case 1414:
  if (!yytrial)
{
			NamedWindowsClause* node = yyvsp[-2].namedWindowsClause;
			node->add(*yyvsp[0].namedWindowClause);
			yyval.namedWindowsClause = node;
		}
break;
case 1415:
  if (!yytrial)
{
			yyval.namedWindowClause = newNode<NamedWindowClause>(*yyvsp[-4].metaNamePtr, yyvsp[-1].windowClause);
		}
break;
case 1416:
  if (!yytrial)
{ yyval.metaNamePtr = NULL; }
break;
case 1418:
  if (!yytrial)
{ yyval.planNode = NULL; }
break;
case 1419:
  if (!yytrial)
{ yyval.planNode = yyvsp[0].planNode; }
break;
case 1420:
  if (!yytrial)
{ yyval.planNode = newNode<PlanNode>(PlanNode::TYPE_RETRIEVE); }
break;
case 1421:
yyval.planNode = yyvsp[-1].planNode;
break;
case 1422:
  if (!yytrial)
{ yyval.planNode = yyvsp[-4].planNode; }
break;
case 1429:
  if (!yytrial)
{ yyvsp[-1].planNode->subNodes.add(yyvsp[0].planNode); }
break;
case 1430:
  if (!yytrial)
{ yyvsp[-3].planNode->subNodes.add(yyvsp[0].planNode); }
break;
case 1431:
  if (!yytrial)
{
			yyval.planNode = newNode<PlanNode>(PlanNode::TYPE_RETRIEVE);
			yyval.planNode->dsqlNames = yyvsp[-1].metaNameArray;
			yyval.planNode->accessType = yyvsp[0].accessType;
		}
break;
case 1433:
  if (!yytrial)
{
			ObjectsArray<MetaName>* node = newNode<ObjectsArray<MetaName> >();
			node->add(*yyvsp[0].metaNamePtr);
			yyval.metaNameArray = node;
		}
break;
case 1434:
  if (!yytrial)
{
			ObjectsArray<MetaName>* node = yyvsp[-1].metaNameArray;
			node->add(*yyvsp[0].metaNamePtr);
			yyval.metaNameArray = node;
		}
break;
case 1435:
  if (!yytrial)
{ yyval.accessType = newNode<PlanNode::AccessType>(PlanNode::AccessType::TYPE_SEQUENTIAL); }
break;
case 1436:
  if (!yytrial)
{ yyval.accessType = newNode<PlanNode::AccessType>(PlanNode::AccessType::TYPE_INDICES); }
break;
case 1437:
yyval.accessType = yyvsp[-1].accessType;
break;
case 1438:
  if (!yytrial)
{ yyval.accessType = yyvsp[-4].accessType; }
break;
case 1439:
  if (!yytrial)
{ yyval.accessType = newNode<PlanNode::AccessType>(PlanNode::AccessType::TYPE_NAVIGATIONAL); }
break;
case 1440:
  if (!yytrial)
{
			yyval.accessType = yyvsp[-3].accessType;
			yyval.accessType->items.insert(0).indexName = *yyvsp[-2].metaNamePtr;
		}
break;
case 1441:
  if (!yytrial)
{
			PlanNode::AccessItem& item = yyvsp[-1].accessType->items.add();
			item.indexName = *yyvsp[0].metaNamePtr;
		}
break;
case 1442:
  if (!yytrial)
{
			PlanNode::AccessItem& item = yyvsp[-3].accessType->items.add();
			item.indexName = *yyvsp[0].metaNamePtr;
		}
break;
case 1444:
yyval.accessType = yyvsp[-2].accessType;
break;
case 1446:
  if (!yytrial)
{ yyval.valueListNode = NULL; }
break;
case 1448:
  if (!yytrial)
{ yyval.valueListNode = yyvsp[0].valueListNode; }
break;
case 1449:
  if (!yytrial)
{ yyval.valueListNode = newNode<ValueListNode>(yyvsp[0].valueExprNode); }
break;
case 1450:
  if (!yytrial)
{ yyval.valueListNode = yyvsp[-2].valueListNode->add(yyvsp[0].valueExprNode); }
break;
case 1451:
  if (!yytrial)
{
			OrderNode* node = newNode<OrderNode>(yyvsp[-2].valueExprNode);
			node->descending = yyvsp[-1].boolVal;
			node->nullsPlacement = yyvsp[0].nullsPlacement;
			yyval.valueExprNode = node;
		}
break;
case 1452:
  if (!yytrial)
{ yyval.boolVal = false; }
break;
case 1453:
  if (!yytrial)
{ yyval.boolVal = false; }
break;
case 1454:
  if (!yytrial)
{ yyval.boolVal = true; }
break;
case 1455:
  if (!yytrial)
{ yyval.nullsPlacement = OrderNode::NULLS_DEFAULT; }
break;
case 1456:
  if (!yytrial)
{ yyval.nullsPlacement = yyvsp[0].nullsPlacement; }
break;
case 1457:
  if (!yytrial)
{ yyval.nullsPlacement = OrderNode::NULLS_FIRST; }
break;
case 1458:
  if (!yytrial)
{ yyval.nullsPlacement = OrderNode::NULLS_LAST; }
break;
case 1459:
  if (!yytrial)
{
			yyval.rowsClause = newNode<RowsClause>();
			yyval.rowsClause->length = yyvsp[0].valueExprNode;
		}
break;
case 1460:
  if (!yytrial)
{
			yyval.rowsClause = newNode<RowsClause>();
			yyval.rowsClause->skip = newNode<ArithmeticNode>(blr_subtract, true, yyvsp[-2].valueExprNode, MAKE_const_slong(1));
			yyval.rowsClause->length = newNode<ArithmeticNode>(blr_add, true,
				newNode<ArithmeticNode>(blr_subtract, true, yyvsp[0].valueExprNode, yyvsp[-2].valueExprNode), MAKE_const_slong(1));
		}
break;
case 1461:
  if (!yytrial)
{ yyval.rowsClause = NULL; }
break;
case 1465:
  if (!yytrial)
{ yyval.valueExprNode = NULL; }
break;
case 1466:
  if (!yytrial)
{ yyval.valueExprNode = yyvsp[-1].valueExprNode; }
break;
case 1469:
  if (!yytrial)
{ yyval.valueExprNode = NULL; }
break;
case 1470:
  if (!yytrial)
{ yyval.valueExprNode = yyvsp[-2].valueExprNode; }
break;
case 1471:
  if (!yytrial)
{ yyval.valueExprNode = MAKE_const_slong(1); }
break;
case 1472:
yyval.nestFieldArray = NOTRIAL(&yyvsp[0].storeNode->dsqlFields);
break;
case 1473:
  if (!yytrial)
{
			StoreNode* node = yyval.storeNode = yyvsp[-8].storeNode;
			node->overrideClause = yyvsp[-5].nullableOverrideClause;
			node->dsqlValues = yyvsp[-2].valueListNode;
			node->dsqlReturning = yyvsp[0].returningClause;
		}
break;
case 1474:
  if (!yytrial)
{
			StoreNode* node = yyval.storeNode = yyvsp[-5].storeNode;
			node->overrideClause = yyvsp[-2].nullableOverrideClause;
			node->dsqlRse = yyvsp[-1].selectExprNode;
			node->dsqlReturning = yyvsp[0].returningClause;
			yyval.storeNode = node;
		}
break;
case 1475:
  if (!yytrial)
{
			StoreNode* node = yyval.storeNode = yyvsp[-3].storeNode;
			node->dsqlReturning = yyvsp[0].returningClause;
			yyval.storeNode = node;
		}
break;
case 1476:
  if (!yytrial)
{
			StoreNode* node = newNode<StoreNode>();
			node->dsqlRelation = yyvsp[0].relSourceNode;
			yyval.storeNode = node;
		}
break;
case 1477:
  if (!yytrial)
{ yyval.nullableOverrideClause = Nullable<OverrideClause>::empty(); }
break;
case 1478:
  if (!yytrial)
{ yyval.nullableOverrideClause = Nullable<OverrideClause>::val(OverrideClause::USER_VALUE); }
break;
case 1479:
  if (!yytrial)
{ yyval.nullableOverrideClause = Nullable<OverrideClause>::val(OverrideClause::SYSTEM_VALUE); }
break;
case 1480:
  if (!yytrial)
{ yyval.valueListNode = newNode<ValueListNode>(yyvsp[0].valueExprNode); }
break;
case 1481:
  if (!yytrial)
{ yyval.valueListNode = yyvsp[-2].valueListNode->add(yyvsp[0].valueExprNode); }
break;
case 1483:
  if (!yytrial)
{ yyval.valueExprNode = NULL; }
break;
case 1484:
  if (!yytrial)
{
				MergeNode* node = yyval.mergeNode = newNode<MergeNode>();
				node->relation = yyvsp[-4].relSourceNode;
				node->usingClause = yyvsp[-2].recSourceNode;
				node->condition = yyvsp[0].boolExprNode;
			}
break;
case 1485:
yyval.mergeNode = yyvsp[0].mergeNode;
break;
case 1486:
  if (!yytrial)
{
				MergeNode* node = yyval.mergeNode = yyvsp[-3].mergeNode;
				node->returning = yyvsp[0].returningClause;
			}
break;
case 1489:
yyval.mergeNode = yyvsp[-1].mergeNode;
break;
case 1492:
  if (!yytrial)
{ yyval.mergeMatchedClause = &yyvsp[-2].mergeNode->whenMatched.add(); }
break;
case 1493:
yyval.mergeMatchedClause = NOTRIAL(yyvsp[0].mergeMatchedClause);
break;
case 1494:
yyval.metaNamePtr =  NOTRIAL(&yyvsp[-4].mergeNode->relation->dsqlName);
break;
case 1496:
  if (!yytrial)
{ yyval.mergeNotMatchedClause = &yyvsp[-3].mergeNode->whenNotMatched.add(); }
break;
case 1497:
yyval.mergeNotMatchedClause = NOTRIAL(yyvsp[0].mergeNotMatchedClause);
break;
case 1499:
yyval.metaNamePtr = NOTRIAL(yyvsp[-3].metaNamePtr);
break;
case 1500:
  if (!yytrial)
{ yyvsp[-6].mergeMatchedClause->assignments = yyvsp[0].compoundStmtNode; }
break;
case 1501:
yyval.metaNamePtr = NOTRIAL(yyvsp[-5].metaNamePtr);
break;
case 1502:
  if (!yytrial)
{
			yyvsp[-8].mergeMatchedClause->condition = yyvsp[-5].boolExprNode;
			yyvsp[-8].mergeMatchedClause->assignments = yyvsp[0].compoundStmtNode;
		}
break;
case 1504:
  if (!yytrial)
{ yyvsp[-5].mergeMatchedClause->condition = yyvsp[-2].boolExprNode; }
break;
case 1505:
yyval.nestFieldArray = NOTRIAL(&yyvsp[-2].mergeNotMatchedClause->fields);
break;
case 1506:
  if (!yytrial)
{
			yyvsp[-9].mergeNotMatchedClause->overrideClause = yyvsp[-4].nullableOverrideClause;
			yyvsp[-9].mergeNotMatchedClause->values = yyvsp[-1].valueListNode;
		}
break;
case 1507:
yyval.nestFieldArray = NOTRIAL(&yyvsp[-4].mergeNotMatchedClause->fields);
break;
case 1508:
  if (!yytrial)
{
			yyvsp[-11].mergeNotMatchedClause->overrideClause = yyvsp[-4].nullableOverrideClause;
			yyvsp[-11].mergeNotMatchedClause->values = yyvsp[-1].valueListNode;
			yyvsp[-11].mergeNotMatchedClause->condition = yyvsp[-9].boolExprNode;
		}
break;
case 1511:
  if (!yytrial)
{
			EraseNode* node = newNode<EraseNode>();
			node->dsqlRelation = yyvsp[-5].relSourceNode;
			node->dsqlBoolean = yyvsp[-4].boolExprNode;
			node->dsqlPlan = yyvsp[-3].planNode;
			node->dsqlOrder = yyvsp[-2].valueListNode;
			node->dsqlRows = yyvsp[-1].rowsClause;
			node->dsqlReturning = yyvsp[0].returningClause;
			yyval.stmtNode = node;
		}
break;
case 1512:
  if (!yytrial)
{
			EraseNode* node = newNode<EraseNode>();
			node->dsqlRelation = yyvsp[-2].relSourceNode;
			node->dsqlCursorName = *yyvsp[-1].metaNamePtr;
			node->dsqlReturning = yyvsp[0].returningClause;
			yyval.stmtNode = node;
		}
break;
case 1515:
yyval.metaNamePtr = NOTRIAL(&yyvsp[-1].relSourceNode->dsqlName);
break;
case 1516:
  if (!yytrial)
{
			ModifyNode* node = newNode<ModifyNode>();
			node->dsqlRelation = yyvsp[-8].relSourceNode;
			node->statement = yyvsp[-5].compoundStmtNode;
			node->dsqlBoolean = yyvsp[-4].boolExprNode;
			node->dsqlPlan = yyvsp[-3].planNode;
			node->dsqlOrder = yyvsp[-2].valueListNode;
			node->dsqlRows = yyvsp[-1].rowsClause;
			node->dsqlReturning = yyvsp[0].returningClause;
			yyval.stmtNode = node;
		}
break;
case 1517:
  if (!yytrial)
{
			ModifyNode* node = newNode<ModifyNode>();
			node->dsqlRelation = yyvsp[-5].relSourceNode;
			node->statement = yyvsp[-2].compoundStmtNode;
			node->dsqlCursorName = *yyvsp[-1].metaNamePtr;
			node->dsqlReturning = yyvsp[0].returningClause;
			yyval.stmtNode = node;
		}
break;
case 1518:
  if (!yytrial)
{
				UpdateOrInsertNode* node = yyval.updInsNode = newNode<UpdateOrInsertNode>();
				node->relation = yyvsp[0].relSourceNode;
			}
break;
case 1519:
yyval.nestFieldArray = NOTRIAL(&yyvsp[0].updInsNode->fields);
break;
case 1520:
yyval.nestFieldArray = NOTRIAL(&yyvsp[-7].updInsNode->matching);
break;
case 1521:
  if (!yytrial)
{
				UpdateOrInsertNode* node = yyval.updInsNode = yyvsp[-10].updInsNode;
				node->overrideClause = yyvsp[-7].nullableOverrideClause;
				node->values = yyvsp[-4].valueListNode;
				node->returning = yyvsp[0].returningClause;
			}
break;
case 1523:
yyval.nestFieldArray = yyvsp[-1].nestFieldArray;
break;
case 1525:
  if (!yytrial)
{ yyval.returningClause = NULL; }
break;
case 1526:
  if (!yytrial)
{
			yyval.returningClause = FB_NEW_POOL(getPool()) ReturningClause(getPool());
			yyval.returningClause->first = yyvsp[0].valueListNode;
		}
break;
case 1527:
  if (!yytrial)
{
			yyval.returningClause = FB_NEW_POOL(getPool()) ReturningClause(getPool());
			yyval.returningClause->first = yyvsp[-2].valueListNode;
			yyval.returningClause->second = yyvsp[0].valueListNode;
		}
break;
case 1528:
  if (!yytrial)
{ yyval.metaNamePtr = newNode<MetaName>(*yyvsp[0].metaNamePtr); }
break;
case 1529:
  if (!yytrial)
{
			AssignmentNode* node = newNode<AssignmentNode>();
			node->asgnTo = yyvsp[-2].fieldNode;
			node->asgnFrom = yyvsp[0].valueExprNode;
			yyval.stmtNode = node;
		}
break;
case 1530:
yyval.metaNamePtr = yyvsp[0].metaNamePtr;
break;
case 1531:
  if (!yytrial)
{
			yyval.compoundStmtNode = newNode<CompoundStmtNode>();
			yyval.compoundStmtNode->statements.add(yyvsp[0].stmtNode);
		}
break;
case 1532:
yyval.metaNamePtr = yyvsp[-2].metaNamePtr;
break;
case 1533:
  if (!yytrial)
{
			yyvsp[-3].compoundStmtNode->statements.add(yyvsp[0].stmtNode);
			yyval.compoundStmtNode = yyvsp[-3].compoundStmtNode;
		}
break;
case 1534:
  if (!yytrial)
{
			AssignmentNode* node = newNode<AssignmentNode>();
			node->asgnTo = yyvsp[-2].fieldNode;
			node->asgnFrom = yyvsp[0].valueExprNode;
			yyval.stmtNode = node;
		}
break;
case 1535:
  if (!yytrial)
{
			AssignmentNode* node = newNode<AssignmentNode>();
			node->asgnTo = yyvsp[-2].fieldNode;
			node->asgnFrom = newNode<DefaultNode>(*yyvsp[-3].metaNamePtr, yyvsp[-2].fieldNode->dsqlName);
			yyval.stmtNode = node;
		}
break;
case 1536:
  if (!yytrial)
{
			AssignmentNode* node = newNode<AssignmentNode>();
			node->asgnTo = NullNode::instance();
			node->asgnFrom = yyvsp[0].valueExprNode;
			yyval.stmtNode = node;
		}
break;
case 1537:
  if (!yytrial)
{
			AssignmentNode* node = newNode<AssignmentNode>();
			node->asgnTo = NullNode::instance();
			node->asgnFrom = yyvsp[0].valueExprNode;
			yyval.stmtNode = node;
		}
break;
case 1538:
  if (!yytrial)
{ yyval.valueListNode = NULL; }
break;
case 1540:
  if (!yytrial)
{ yyval.valueListNode = yyvsp[-1].valueListNode; }
break;
case 1541:
  if (!yytrial)
{ yyval.valueListNode = newNode<ValueListNode>(yyvsp[0].fieldNode); }
break;
case 1542:
  if (!yytrial)
{ yyval.valueListNode = yyvsp[-2].valueListNode->add(yyvsp[0].fieldNode); }
break;
case 1544:
yyval.nestFieldArray = yyvsp[0].nestFieldArray;
break;
case 1547:
  if (!yytrial)
{ yyvsp[-1].nestFieldArray->add(yyvsp[0].fieldNode); }
break;
case 1548:
  if (!yytrial)
{ yyvsp[-3].nestFieldArray->add(yyvsp[0].fieldNode); }
break;
case 1550:
  if (!yytrial)
{
			FieldNode* fieldNode = newNode<FieldNode>();
			fieldNode->dsqlQualifier = *yyvsp[-2].metaNamePtr;
			fieldNode->dsqlName = *yyvsp[0].metaNamePtr;
			yyval.fieldNode = fieldNode;
		}
break;
case 1551:
  if (!yytrial)
{
			FieldNode* fieldNode = newNode<FieldNode>();
			fieldNode->dsqlQualifier = *yyvsp[-2].metaNamePtr;
			fieldNode->dsqlName = *yyvsp[0].metaNamePtr;
			fieldNode->dsqlCursorField = true;
			yyval.fieldNode = fieldNode;
		}
break;
case 1552:
  if (!yytrial)
{
			FieldNode* fieldNode = newNode<FieldNode>();
			fieldNode->dsqlName = *yyvsp[0].metaNamePtr;
			yyval.fieldNode = fieldNode;
		}
break;
case 1554:
  if (!yytrial)
{
			FieldNode* fieldNode = newNode<FieldNode>();
			fieldNode->dsqlQualifier = *yyvsp[-2].metaNamePtr;
			fieldNode->dsqlName = *yyvsp[0].metaNamePtr;
			yyval.fieldNode = fieldNode;
		}
break;
case 1555:
  if (!yytrial)
{ yyval.boolExprNode = valueToBool(yyvsp[0].valueExprNode); }
break;
case 1557:
  if (!yytrial)
{ yyval.boolExprNode = newNode<BinaryBoolNode>(blr_or, valueToBool(yyvsp[-2].valueExprNode), valueToBool(yyvsp[0].valueExprNode)); }
break;
case 1558:
  if (!yytrial)
{ yyval.boolExprNode = newNode<BinaryBoolNode>(blr_and, valueToBool(yyvsp[-2].valueExprNode), valueToBool(yyvsp[0].valueExprNode)); }
break;
case 1559:
  if (!yytrial)
{ yyval.boolExprNode = newNode<NotBoolNode>(valueToBool(yyvsp[0].valueExprNode)); }
break;
case 1560:
  if (!yytrial)
{ yyval.boolExprNode = yyvsp[-1].boolExprNode; }
break;
case 1561:
  if (!yytrial)
{
			ComparativeBoolNode* node = newNode<ComparativeBoolNode>(blr_equiv, yyvsp[-2].valueExprNode, yyvsp[0].valueExprNode);
			node->dsqlCheckBoolean = true;
			yyval.boolExprNode = node;
		}
break;
case 1562:
  if (!yytrial)
{
			ComparativeBoolNode* node = newNode<ComparativeBoolNode>(blr_equiv, yyvsp[-3].valueExprNode, yyvsp[0].valueExprNode);
			node->dsqlCheckBoolean = true;
			yyval.boolExprNode = newNode<NotBoolNode>(node);
		}
break;
case 1575:
  if (!yytrial)
{ yyval.boolExprNode = newNode<ComparativeBoolNode>(yyvsp[-1].blrOp, yyvsp[-2].valueExprNode, yyvsp[0].valueExprNode); }
break;
case 1576:
  if (!yytrial)
{ yyval.blrOp = blr_eql; }
break;
case 1577:
  if (!yytrial)
{ yyval.blrOp = blr_lss; }
break;
case 1578:
  if (!yytrial)
{ yyval.blrOp = blr_gtr; }
break;
case 1579:
  if (!yytrial)
{ yyval.blrOp = blr_geq; }
break;
case 1580:
  if (!yytrial)
{ yyval.blrOp = blr_leq; }
break;
case 1581:
  if (!yytrial)
{ yyval.blrOp = blr_leq; }
break;
case 1582:
  if (!yytrial)
{ yyval.blrOp = blr_geq; }
break;
case 1583:
  if (!yytrial)
{ yyval.blrOp = blr_neq; }
break;
case 1584:
  if (!yytrial)
{
			ComparativeBoolNode* node = newNode<ComparativeBoolNode>(yyvsp[-4].blrOp, yyvsp[-5].valueExprNode);
			node->dsqlFlag = yyvsp[-3].cmpBoolFlag;
			node->dsqlSpecialArg = yyvsp[-1].selectExprNode;
			yyval.boolExprNode = node;
		}
break;
case 1585:
  if (!yytrial)
{ yyval.cmpBoolFlag = ComparativeBoolNode::DFLAG_ANSI_ALL; }
break;
case 1586:
  if (!yytrial)
{ yyval.cmpBoolFlag = ComparativeBoolNode::DFLAG_ANSI_ANY; }
break;
case 1587:
  if (!yytrial)
{ yyval.cmpBoolFlag = ComparativeBoolNode::DFLAG_ANSI_ANY; }
break;
case 1588:
  if (!yytrial)
{
			ComparativeBoolNode* node = newNode<ComparativeBoolNode>(blr_equiv, yyvsp[-4].valueExprNode, yyvsp[0].valueExprNode);
			yyval.boolExprNode = newNode<NotBoolNode>(node);
		}
break;
case 1589:
  if (!yytrial)
{ yyval.boolExprNode = newNode<ComparativeBoolNode>(blr_equiv, yyvsp[-5].valueExprNode, yyvsp[0].valueExprNode); }
break;
case 1590:
  if (!yytrial)
{
			yyval.boolExprNode = newNode<ComparativeBoolNode>(blr_between, yyvsp[-4].valueExprNode, yyvsp[-2].valueExprNode, yyvsp[0].valueExprNode);
		}
break;
case 1591:
  if (!yytrial)
{
			ComparativeBoolNode* node = newNode<ComparativeBoolNode>(blr_between, yyvsp[-5].valueExprNode, yyvsp[-2].valueExprNode, yyvsp[0].valueExprNode);
			yyval.boolExprNode = newNode<NotBoolNode>(node);
		}
break;
case 1592:
  if (!yytrial)
{ yyval.boolExprNode = newNode<ComparativeBoolNode>(yyvsp[-1].blrOp, yyvsp[-2].valueExprNode, yyvsp[0].valueExprNode); }
break;
case 1593:
  if (!yytrial)
{
			ComparativeBoolNode* cmpNode = newNode<ComparativeBoolNode>(yyvsp[-1].blrOp, yyvsp[-3].valueExprNode, yyvsp[0].valueExprNode);
			yyval.boolExprNode = newNode<NotBoolNode>(cmpNode);
		}
break;
case 1594:
  if (!yytrial)
{ yyval.blrOp = blr_containing; }
break;
case 1595:
  if (!yytrial)
{ yyval.blrOp = blr_starting; }
break;
case 1596:
  if (!yytrial)
{ yyval.blrOp = blr_starting; }
break;
case 1597:
  if (!yytrial)
{ yyval.boolExprNode = newNode<ComparativeBoolNode>(blr_like, yyvsp[-2].valueExprNode, yyvsp[0].valueExprNode); }
break;
case 1598:
  if (!yytrial)
{ yyval.boolExprNode = newNode<ComparativeBoolNode>(blr_like, yyvsp[-4].valueExprNode, yyvsp[-2].valueExprNode, yyvsp[0].valueExprNode); }
break;
case 1599:
  if (!yytrial)
{
			ComparativeBoolNode* node = newNode<ComparativeBoolNode>(blr_like, yyvsp[-3].valueExprNode, yyvsp[0].valueExprNode);
			yyval.boolExprNode = newNode<NotBoolNode>(node);
		}
break;
case 1600:
  if (!yytrial)
{
			ComparativeBoolNode* node = newNode<ComparativeBoolNode>(blr_like, yyvsp[-5].valueExprNode, yyvsp[-2].valueExprNode, yyvsp[0].valueExprNode);
			yyval.boolExprNode = newNode<NotBoolNode>(node);
		}
break;
case 1601:
  if (!yytrial)
{ yyval.boolExprNode = newNode<ComparativeBoolNode>(blr_similar, yyvsp[-3].valueExprNode, yyvsp[0].valueExprNode); }
break;
case 1602:
  if (!yytrial)
{ yyval.boolExprNode = newNode<ComparativeBoolNode>(blr_similar, yyvsp[-5].valueExprNode, yyvsp[-2].valueExprNode, yyvsp[0].valueExprNode); }
break;
case 1603:
  if (!yytrial)
{
			ComparativeBoolNode* node = newNode<ComparativeBoolNode>(blr_similar, yyvsp[-4].valueExprNode, yyvsp[0].valueExprNode);
			yyval.boolExprNode = newNode<NotBoolNode>(node);
		}
break;
case 1604:
  if (!yytrial)
{
			ComparativeBoolNode* node = newNode<ComparativeBoolNode>(blr_similar, yyvsp[-6].valueExprNode, yyvsp[-2].valueExprNode, yyvsp[0].valueExprNode);
			yyval.boolExprNode = newNode<NotBoolNode>(node);
		}
break;
case 1605:
  if (!yytrial)
{
			ComparativeBoolNode* node = newNode<ComparativeBoolNode>(blr_eql, yyvsp[-2].valueExprNode);
			node->dsqlFlag = ComparativeBoolNode::DFLAG_ANSI_ANY;
			node->dsqlSpecialArg = yyvsp[0].exprNode;
			yyval.boolExprNode = node;
		}
break;
case 1606:
  if (!yytrial)
{
			ComparativeBoolNode* node = newNode<ComparativeBoolNode>(blr_eql, yyvsp[-3].valueExprNode);
			node->dsqlFlag = ComparativeBoolNode::DFLAG_ANSI_ANY;
			node->dsqlSpecialArg = yyvsp[0].exprNode;
			yyval.boolExprNode = newNode<NotBoolNode>(node);
		}
break;
case 1607:
  if (!yytrial)
{ yyval.boolExprNode = newNode<RseBoolNode>(blr_any, yyvsp[-1].selectExprNode); }
break;
case 1608:
  if (!yytrial)
{ yyval.boolExprNode = newNode<RseBoolNode>(blr_unique, yyvsp[-1].selectExprNode); }
break;
case 1609:
  if (!yytrial)
{
			yyval.boolExprNode = newNode<ComparativeBoolNode>(blr_eql,
					newNode<InternalInfoNode>(MAKE_const_slong(INFO_TYPE_TRIGGER_ACTION)),
					MAKE_const_slong(1));
		}
break;
case 1610:
  if (!yytrial)
{
			yyval.boolExprNode = newNode<ComparativeBoolNode>(blr_eql,
					newNode<InternalInfoNode>(MAKE_const_slong(INFO_TYPE_TRIGGER_ACTION)),
					MAKE_const_slong(2));
		}
break;
case 1611:
  if (!yytrial)
{
			yyval.boolExprNode = newNode<ComparativeBoolNode>(blr_eql,
					newNode<InternalInfoNode>(MAKE_const_slong(INFO_TYPE_TRIGGER_ACTION)),
					MAKE_const_slong(3));
		}
break;
case 1612:
  if (!yytrial)
{
			yyval.boolExprNode = newNode<ComparativeBoolNode>(blr_eql,
					newNode<InternalInfoNode>(MAKE_const_slong(INFO_TYPE_SESSION_RESETTING)),
					MAKE_const_slong(1));
		}
break;
case 1613:
  if (!yytrial)
{ yyval.boolExprNode = newNode<MissingBoolNode>(yyvsp[-2].valueExprNode); }
break;
case 1614:
  if (!yytrial)
{ yyval.boolExprNode = newNode<MissingBoolNode>(yyvsp[-2].valueExprNode, true); }
break;
case 1615:
  if (!yytrial)
{ yyval.boolExprNode = newNode<NotBoolNode>(newNode<MissingBoolNode>(yyvsp[-3].valueExprNode)); }
break;
case 1616:
  if (!yytrial)
{ yyval.boolExprNode = newNode<NotBoolNode>(newNode<MissingBoolNode>(yyvsp[-3].valueExprNode, true)); }
break;
case 1617:
  if (!yytrial)
{ yyval.exprNode = yyvsp[0].selectExprNode; }
break;
case 1618:
  if (!yytrial)
{ yyval.exprNode = yyvsp[-1].valueListNode; }
break;
case 1619:
  if (!yytrial)
{ yyval.selectExprNode = yyvsp[-1].selectExprNode; }
break;
case 1620:
  if (!yytrial)
{
			yyval.createAlterUserNode = newNode<CreateAlterUserNode>(CreateAlterUserNode::USER_ADD, *yyvsp[0].metaNamePtr);
		}
break;
case 1621:
  if (!yytrial)
{
			yyval.createAlterUserNode = yyvsp[-2].createAlterUserNode;
		}
break;
case 1622:
  if (!yytrial)
{
			yyval.createAlterUserNode = newNode<CreateAlterUserNode>(CreateAlterUserNode::USER_MOD, *yyvsp[-1].metaNamePtr);
		}
break;
case 1623:
  if (!yytrial)
{
			yyval.createAlterUserNode = yyvsp[-2].createAlterUserNode;
		}
break;
case 1624:
  if (!yytrial)
{
			yyval.createAlterUserNode = newNode<CreateAlterUserNode>(CreateAlterUserNode::USER_MOD, "");
		}
break;
case 1625:
  if (!yytrial)
{
			yyval.createAlterUserNode = yyvsp[-2].createAlterUserNode;
		}
break;
case 1626:
  if (!yytrial)
{
			yyval.createAlterUserNode = newNode<CreateAlterUserNode>(CreateAlterUserNode::USER_RPL, *yyvsp[-1].metaNamePtr);
		}
break;
case 1627:
  if (!yytrial)
{
			yyval.createAlterUserNode = yyvsp[-2].createAlterUserNode;
		}
break;
case 1633:
yyval.createAlterUserNode = yyvsp[-1].createAlterUserNode;
break;
case 1635:
  if (!yytrial)
{ setClause(yyvsp[-2].createAlterUserNode->firstName, "FIRSTNAME", yyvsp[0].stringPtr); }
break;
case 1636:
  if (!yytrial)
{ setClause(yyvsp[-2].createAlterUserNode->middleName, "MIDDLENAME", yyvsp[0].stringPtr); }
break;
case 1637:
  if (!yytrial)
{ setClause(yyvsp[-2].createAlterUserNode->lastName, "LASTNAME", yyvsp[0].stringPtr); }
break;
case 1638:
  if (!yytrial)
{ setClause(yyvsp[-2].createAlterUserNode->password, "PASSWORD", yyvsp[0].stringPtr); }
break;
case 1639:
  if (!yytrial)
{ setClause(yyvsp[-3].createAlterUserNode->adminRole, "ADMIN ROLE", true); }
break;
case 1640:
  if (!yytrial)
{ setClause(yyvsp[-3].createAlterUserNode->adminRole, "ADMIN ROLE", false); }
break;
case 1641:
  if (!yytrial)
{ setClause(yyvsp[-1].createAlterUserNode->active, "ACTIVE/INACTIVE", true); }
break;
case 1642:
  if (!yytrial)
{ setClause(yyvsp[-1].createAlterUserNode->active, "ACTIVE/INACTIVE", false); }
break;
case 1644:
yyval.createAlterUserNode = yyvsp[-2].createAlterUserNode;
break;
case 1646:
  if (!yytrial)
{ setClause(yyvsp[-3].createAlterUserNode->plugin, "USING PLUGIN", yyvsp[0].metaNamePtr); }
break;
case 1649:
  if (!yytrial)
{
			yyvsp[-3].createAlterUserNode->addProperty(yyvsp[-2].metaNamePtr, yyvsp[0].stringPtr);
		}
break;
case 1650:
  if (!yytrial)
{
			yyvsp[-2].createAlterUserNode->addProperty(yyvsp[0].metaNamePtr);
		}
break;
case 1651:
yyval.mappingOp = MappingNode::MAP_ADD;
break;
case 1652:
  if (!yytrial)
{
				yyval.mappingNode = yyvsp[0].mappingNode;
				yyval.mappingNode->global = yyvsp[-2].boolVal;
			}
break;
case 1653:
yyval.mappingNode = yyvsp[0].mappingNode;
break;
case 1654:
  if (!yytrial)
{
				yyval.mappingNode = yyvsp[-2].mappingNode;
			}
break;
case 1655:
yyval.mappingOp = MappingNode::MAP_MOD;
break;
case 1656:
  if (!yytrial)
{
				yyval.mappingNode = yyvsp[0].mappingNode;
				yyval.mappingNode->global = yyvsp[-2].boolVal;
			}
break;
case 1657:
  if (!yytrial)
{
				yyval.mappingNode = yyvsp[-2].mappingNode;
			}
break;
case 1658:
yyval.mappingOp = MappingNode::MAP_RPL;
break;
case 1659:
  if (!yytrial)
{
				yyval.mappingNode = yyvsp[0].mappingNode;
				yyval.mappingNode->global = yyvsp[-2].boolVal;
			}
break;
case 1660:
  if (!yytrial)
{
				yyval.mappingNode = yyvsp[-2].mappingNode;
			}
break;
case 1661:
  if (!yytrial)
{
 			MappingNode* node = newNode<MappingNode>(MappingNode::MAP_DROP, *yyvsp[0].metaNamePtr);
			node->global = yyvsp[-1].boolVal;
			yyval.mappingNode = node;
		}
break;
case 1662:
  if (!yytrial)
{
				yyval.mappingNode = newNode<MappingNode>(yyvsp[-1].mappingOp, *yyvsp[0].metaNamePtr);
			}
break;
case 1663:
yyval.mappingNode = yyvsp[-1].mappingNode;
break;
case 1664:
yyval.mappingNode = yyvsp[-4].mappingNode;
break;
case 1665:
  if (!yytrial)
{
				yyval.mappingNode = yyvsp[-6].mappingNode;
			}
break;
case 1666:
  if (!yytrial)
{ yyval.metaNamePtr = yyvsp[0].metaNamePtr; }
break;
case 1667:
  if (!yytrial)
{
			yyvsp[-2].mappingNode->fromType = yyvsp[-1].metaNamePtr;
			yyvsp[-2].mappingNode->from = yyvsp[0].intlStringPtr;
		}
break;
case 1668:
  if (!yytrial)
{
			yyvsp[-2].mappingNode->fromType = yyvsp[0].metaNamePtr;
			yyvsp[-2].mappingNode->from = newNode<IntlString>("*");
		}
break;
case 1670:
  if (!yytrial)
{ yyval.metaNamePtr = newNode<MetaName>("USER"); }
break;
case 1671:
  if (!yytrial)
{ yyval.metaNamePtr = newNode<MetaName>("GROUP"); }
break;
case 1673:
  if (!yytrial)
{ yyval.intlStringPtr = newNode<IntlString>(yyvsp[0].metaNamePtr->c_str()); }
break;
case 1674:
  if (!yytrial)
{
			yyvsp[-3].mappingNode->mode = 'P';
			yyvsp[-3].mappingNode->plugin = yyvsp[-1].metaNamePtr;
			yyvsp[-3].mappingNode->db = yyvsp[0].metaNamePtr;
		}
break;
case 1675:
  if (!yytrial)
{
			yyvsp[-3].mappingNode->mode = 'P';
			yyvsp[-3].mappingNode->db = yyvsp[0].metaNamePtr;
		}
break;
case 1676:
  if (!yytrial)
{
			yyvsp[-3].mappingNode->mode = 'S';
		}
break;
case 1677:
  if (!yytrial)
{
			yyvsp[-2].mappingNode->mode = 'M';
			yyvsp[-2].mappingNode->db = yyvsp[0].metaNamePtr;
		}
break;
case 1678:
  if (!yytrial)
{
			yyvsp[-2].mappingNode->mode = '*';
			yyvsp[-2].mappingNode->db = yyvsp[0].metaNamePtr;
		}
break;
case 1679:
  if (!yytrial)
{ yyval.metaNamePtr = NULL; }
break;
case 1680:
  if (!yytrial)
{ yyval.metaNamePtr = yyvsp[0].metaNamePtr; }
break;
case 1681:
  if (!yytrial)
{
			yyvsp[-3].mappingNode->role = yyvsp[-1].boolVal;
			yyvsp[-3].mappingNode->to = yyvsp[0].metaNamePtr;
		}
break;
case 1682:
  if (!yytrial)
{
			yyvsp[-2].mappingNode->role = yyvsp[0].boolVal;
		}
break;
case 1683:
  if (!yytrial)
{ yyval.boolVal = true; }
break;
case 1684:
  if (!yytrial)
{ yyval.boolVal = false; }
break;
case 1686:
  if (!yytrial)
{ yyval.valueExprNode = newNode<BoolAsValueNode>(yyvsp[0].boolExprNode); }
break;
case 1688:
  if (!yytrial)
{ yyval.valueExprNode = newNode<BoolAsValueNode>(yyvsp[-1].boolExprNode); }
break;
case 1690:
  if (!yytrial)
{ yyval.valueExprNode = yyvsp[-1].valueExprNode; }
break;
case 1694:
  if (!yytrial)
{ yyval.valueExprNode = yyvsp[0].fieldNode; }
break;
case 1696:
  if (!yytrial)
{ yyval.valueExprNode = yyvsp[0].valueExprNode; }
break;
case 1703:
  if (!yytrial)
{ yyval.valueExprNode = yyvsp[0].valueExprNode; }
break;
case 1704:
  if (!yytrial)
{ yyval.valueExprNode = yyvsp[0].valueExprNode; }
break;
case 1705:
  if (!yytrial)
{ yyval.valueExprNode = newNode<NegateNode>(yyvsp[0].valueExprNode); }
break;
case 1706:
  if (!yytrial)
{ yyval.valueExprNode = yyvsp[0].valueExprNode; }
break;
case 1707:
  if (!yytrial)
{ yyval.valueExprNode = newNode<ArithmeticNode>(blr_add, (client_dialect < SQL_DIALECT_V6_TRANSITION), yyvsp[-2].valueExprNode, yyvsp[0].valueExprNode); }
break;
case 1708:
  if (!yytrial)
{ yyval.valueExprNode = newNode<ConcatenateNode>(yyvsp[-2].valueExprNode, yyvsp[0].valueExprNode); }
break;
case 1709:
  if (!yytrial)
{ yyval.valueExprNode = newNode<CollateNode>(yyvsp[-2].valueExprNode, *yyvsp[0].metaNamePtr); }
break;
case 1710:
  if (!yytrial)
{ yyval.valueExprNode = newNode<AtNode>(yyvsp[-2].valueExprNode, nullptr); }
break;
case 1711:
  if (!yytrial)
{ yyval.valueExprNode = newNode<AtNode>(yyvsp[-4].valueExprNode, yyvsp[0].valueExprNode); }
break;
case 1712:
  if (!yytrial)
{ yyval.valueExprNode = newNode<ArithmeticNode>(blr_subtract, (client_dialect < SQL_DIALECT_V6_TRANSITION), yyvsp[-2].valueExprNode, yyvsp[0].valueExprNode); }
break;
case 1713:
  if (!yytrial)
{ yyval.valueExprNode = newNode<ArithmeticNode>(blr_multiply, (client_dialect < SQL_DIALECT_V6_TRANSITION), yyvsp[-2].valueExprNode, yyvsp[0].valueExprNode); }
break;
case 1714:
  if (!yytrial)
{ yyval.valueExprNode = newNode<ArithmeticNode>(blr_divide, (client_dialect < SQL_DIALECT_V6_TRANSITION), yyvsp[-2].valueExprNode, yyvsp[0].valueExprNode); }
break;
case 1715:
  if (!yytrial)
{ yyval.valueExprNode = yyvsp[-1].valueExprNode; }
break;
case 1716:
  if (!yytrial)
{ yyval.valueExprNode = yyvsp[0].valueExprNode; }
break;
case 1717:
  if (!yytrial)
{ yyval.valueExprNode = yyvsp[0].valueExprNode; }
break;
case 1718:
  if (!yytrial)
{ yyval.valueExprNode = yyvsp[0].valueExprNode; }
break;
case 1719:
  if (!yytrial)
{ yyval.valueExprNode = newNode<RecordKeyNode>(yyvsp[0].blrOp); }
break;
case 1720:
  if (!yytrial)
{ yyval.valueExprNode = newNode<RecordKeyNode>(yyvsp[0].blrOp, *yyvsp[-2].metaNamePtr); }
break;
case 1721:
  if (!yytrial)
{ yyval.valueExprNode = newNode<DomainValidationNode>(); }
break;
case 1722:
  if (!yytrial)
{ yyval.valueExprNode = yyvsp[0].valueExprNode; }
break;
case 1723:
  if (!yytrial)
{ yyval.valueExprNode = yyvsp[0].valueExprNode; }
break;
case 1724:
  if (!yytrial)
{ yyval.blrOp = blr_dbkey; }
break;
case 1725:
  if (!yytrial)
{ yyval.blrOp = blr_record_version2; }
break;
case 1726:
  if (!yytrial)
{
			if (client_dialect < SQL_DIALECT_V6_TRANSITION)
			{
				ERRD_post(Arg::Gds(isc_sqlerr) << Arg::Num(-104) <<
						  Arg::Gds(isc_sql_dialect_datatype_unsupport) << Arg::Num(client_dialect) <<
						  												  Arg::Str("DATE"));
			}

			if (db_dialect < SQL_DIALECT_V6_TRANSITION)
			{
				ERRD_post(Arg::Gds(isc_sqlerr) << Arg::Num(-104) <<
						  Arg::Gds(isc_sql_db_dialect_dtype_unsupport) << Arg::Num(db_dialect) <<
						  												  Arg::Str("DATE"));
			}

			yyval.valueExprNode = newNode<CurrentDateNode>();
		}
break;
case 1727:
  if (!yytrial)
{
			if (client_dialect < SQL_DIALECT_V6_TRANSITION)
			{
				ERRD_post(Arg::Gds(isc_sqlerr) << Arg::Num(-104) <<
						  Arg::Gds(isc_sql_dialect_datatype_unsupport) << Arg::Num(client_dialect) <<
						  												  Arg::Str("TIME"));
			}

			if (db_dialect < SQL_DIALECT_V6_TRANSITION)
			{
				ERRD_post(Arg::Gds(isc_sqlerr) << Arg::Num(-104) <<
						  Arg::Gds(isc_sql_db_dialect_dtype_unsupport) << Arg::Num(db_dialect) <<
						  												  Arg::Str("TIME"));
			}

			yyval.valueExprNode = newNode<LocalTimeNode>(yyvsp[0].uintVal);
		}
break;
case 1728:
  if (!yytrial)
{
			if (client_dialect < SQL_DIALECT_V6_TRANSITION)
			{
				ERRD_post(Arg::Gds(isc_sqlerr) << Arg::Num(-104) <<
						  Arg::Gds(isc_sql_dialect_datatype_unsupport) << Arg::Num(client_dialect) <<
						  												  Arg::Str("TIME"));
			}

			if (db_dialect < SQL_DIALECT_V6_TRANSITION)
			{
				ERRD_post(Arg::Gds(isc_sqlerr) << Arg::Num(-104) <<
						  Arg::Gds(isc_sql_db_dialect_dtype_unsupport) << Arg::Num(db_dialect) <<
						  												  Arg::Str("TIME"));
			}

			yyval.valueExprNode = newNode<CurrentTimeNode>(yyvsp[0].uintVal);
		}
break;
case 1729:
  if (!yytrial)
{ yyval.valueExprNode = newNode<LocalTimeStampNode>(yyvsp[0].uintVal); }
break;
case 1730:
  if (!yytrial)
{ yyval.valueExprNode = newNode<CurrentTimeStampNode>(yyvsp[0].uintVal); }
break;
case 1731:
  if (!yytrial)
{ yyval.uintVal = DEFAULT_TIME_PRECISION; }
break;
case 1732:
  if (!yytrial)
{ yyval.uintVal = yyvsp[-1].int32Val; }
break;
case 1733:
  if (!yytrial)
{ yyval.uintVal = DEFAULT_TIMESTAMP_PRECISION; }
break;
case 1734:
  if (!yytrial)
{ yyval.uintVal = yyvsp[-1].int32Val; }
break;
case 1735:
  if (!yytrial)
{
			ArrayNode* node = newNode<ArrayNode>(yyvsp[-3].fieldNode);
			node->field->dsqlIndices = yyvsp[-1].valueListNode;
			yyval.valueExprNode = node;
		}
break;
case 1736:
  if (!yytrial)
{ yyval.valueListNode = newNode<ValueListNode>(0); }
break;
case 1737:
  if (!yytrial)
{ yyval.valueListNode = yyvsp[0].valueListNode; }
break;
case 1738:
  if (!yytrial)
{ yyval.valueListNode = newNode<ValueListNode>(yyvsp[0].valueExprNode); }
break;
case 1739:
  if (!yytrial)
{ yyval.valueListNode = yyvsp[-2].valueListNode->add(yyvsp[0].valueExprNode); }
break;
case 1741:
  if (!yytrial)
{ yyval.valueExprNode = newNode<NegateNode>(yyvsp[0].valueExprNode); }
break;
case 1742:
  if (!yytrial)
{ yyval.valueExprNode = MAKE_const_sint64(MIN_SINT64, 0); }
break;
case 1743:
  if (!yytrial)
{ yyval.valueExprNode = MAKE_const_sint64(MIN_SINT64, yyvsp[0].lim64ptr->getScale()); }
break;
case 1744:
  if (!yytrial)
{ yyval.valueExprNode = newNode<NegateNode>(yyvsp[0].valueExprNode); }
break;
case 1746:
  if (!yytrial)
{ yyval.valueExprNode = yyvsp[0].valueExprNode; }
break;
case 1747:
  if (!yytrial)
{ yyval.valueExprNode = MAKE_constant(yyvsp[0].lim64ptr->c_str(), CONSTANT_NUM128, yyvsp[0].lim64ptr->getScale()); }
break;
case 1748:
  if (!yytrial)
{ yyval.valueExprNode = MAKE_constant(yyvsp[0].lim64ptr->c_str(), CONSTANT_NUM128); }
break;
case 1750:
  if (!yytrial)
{ yyval.valueExprNode = MAKE_constant(yyvsp[0].lim64ptr->c_str(), CONSTANT_NUM128, yyvsp[0].lim64ptr->getScale()); }
break;
case 1751:
  if (!yytrial)
{ yyval.valueExprNode = MAKE_const_slong(yyvsp[0].int32Val); }
break;
case 1752:
  if (!yytrial)
{ yyval.valueExprNode = MAKE_constant(yyvsp[0].stringPtr->c_str(), CONSTANT_DOUBLE); }
break;
case 1753:
  if (!yytrial)
{ yyval.valueExprNode = MAKE_constant(yyvsp[0].stringPtr->c_str(), CONSTANT_DECIMAL); }
break;
case 1754:
  if (!yytrial)
{
			SINT64 signedNumber = (SINT64) yyvsp[0].scaledNumber.number;

			if (yyvsp[0].scaledNumber.hex && signedNumber < 0)
				yyval.valueExprNode = newNode<NegateNode>(MAKE_const_sint64(-signedNumber, yyvsp[0].scaledNumber.scale));
			else
				yyval.valueExprNode = MAKE_const_sint64(signedNumber, yyvsp[0].scaledNumber.scale);
		}
break;
case 1755:
  if (!yytrial)
{ yyval.valueExprNode = MAKE_const_sint64((SINT64) yyvsp[0].scaledNumber.number, yyvsp[0].scaledNumber.scale); }
break;
case 1757:
  if (!yytrial)
{ yyval.valueExprNode = MAKE_str_constant(yyvsp[0].intlStringPtr, lex.att_charset); }
break;
case 1758:
  if (!yytrial)
{
			if (client_dialect < SQL_DIALECT_V6_TRANSITION)
			{
				ERRD_post(Arg::Gds(isc_sqlerr) << Arg::Num(-104) <<
						  Arg::Gds(isc_sql_dialect_datatype_unsupport) << Arg::Num(client_dialect) <<
						  												  Arg::Str("DATE"));
			}
			if (db_dialect < SQL_DIALECT_V6_TRANSITION)
			{
				ERRD_post(Arg::Gds(isc_sqlerr) << Arg::Num(-104) <<
						  Arg::Gds(isc_sql_db_dialect_dtype_unsupport) << Arg::Num(db_dialect) <<
						  												  Arg::Str("DATE"));
			}
			yyval.valueExprNode = MAKE_constant(yyvsp[0].intlStringPtr->getString().c_str(), CONSTANT_DATE);
		}
break;
case 1759:
  if (!yytrial)
{
			if (client_dialect < SQL_DIALECT_V6_TRANSITION)
			{
				ERRD_post(Arg::Gds(isc_sqlerr) << Arg::Num(-104) <<
						  Arg::Gds(isc_sql_dialect_datatype_unsupport) << Arg::Num(client_dialect) <<
						  												  Arg::Str("TIME"));
			}
			if (db_dialect < SQL_DIALECT_V6_TRANSITION)
			{
				ERRD_post(Arg::Gds(isc_sqlerr) << Arg::Num(-104) <<
						  Arg::Gds(isc_sql_db_dialect_dtype_unsupport) << Arg::Num(db_dialect) <<
						  												  Arg::Str("TIME"));
			}
			yyval.valueExprNode = MAKE_constant(yyvsp[0].intlStringPtr->getString().c_str(), CONSTANT_TIME);
		}
break;
case 1760:
  if (!yytrial)
{ yyval.valueExprNode = MAKE_constant(yyvsp[0].intlStringPtr->getString().c_str(), CONSTANT_TIMESTAMP); }
break;
case 1761:
  if (!yytrial)
{ yyval.valueExprNode = MAKE_constant("", CONSTANT_BOOLEAN); }
break;
case 1762:
  if (!yytrial)
{ yyval.valueExprNode = MAKE_constant("1", CONSTANT_BOOLEAN); }
break;
case 1763:
  if (!yytrial)
{ yyval.valueExprNode = make_parameter(); }
break;
case 1764:
  if (!yytrial)
{ yyval.valueExprNode = newNode<CurrentUserNode>(); }
break;
case 1765:
  if (!yytrial)
{ yyval.valueExprNode = newNode<CurrentUserNode>(); }
break;
case 1766:
  if (!yytrial)
{ yyval.valueExprNode = newNode<CurrentRoleNode>(); }
break;
case 1767:
  if (!yytrial)
{ yyval.valueExprNode = newNode<InternalInfoNode>(MAKE_const_slong(INFO_TYPE_CONNECTION_ID)); }
break;
case 1768:
  if (!yytrial)
{ yyval.valueExprNode = newNode<InternalInfoNode>(MAKE_const_slong(INFO_TYPE_TRANSACTION_ID)); }
break;
case 1769:
  if (!yytrial)
{ yyval.valueExprNode = newNode<InternalInfoNode>(MAKE_const_slong(INFO_TYPE_GDSCODE)); }
break;
case 1770:
  if (!yytrial)
{ yyval.valueExprNode = newNode<InternalInfoNode>(MAKE_const_slong(INFO_TYPE_SQLCODE)); }
break;
case 1771:
  if (!yytrial)
{ yyval.valueExprNode = newNode<InternalInfoNode>(MAKE_const_slong(INFO_TYPE_SQLSTATE)); }
break;
case 1772:
  if (!yytrial)
{ yyval.valueExprNode = newNode<InternalInfoNode>(MAKE_const_slong(INFO_TYPE_ROWS_AFFECTED)); }
break;
case 1773:
  if (!yytrial)
{ yyval.valueExprNode = newNode<InternalInfoNode>(MAKE_const_slong(yyvsp[-1].int32Val)); }
break;
case 1774:
  if (!yytrial)
{ yyval.int32Val = INFO_TYPE_GDSCODE; }
break;
case 1775:
  if (!yytrial)
{ yyval.int32Val = INFO_TYPE_SQLCODE; }
break;
case 1776:
  if (!yytrial)
{ yyval.int32Val = INFO_TYPE_SQLSTATE; }
break;
case 1777:
  if (!yytrial)
{ yyval.int32Val = INFO_TYPE_EXCEPTION; }
break;
case 1778:
  if (!yytrial)
{ yyval.int32Val = INFO_TYPE_ERROR_MSG; }
break;
case 1780:
  if (!yytrial)
{
			yyval.intlStringPtr = yyvsp[0].intlStringPtr;
			yyval.intlStringPtr->setCharSet(*yyvsp[-1].metaNamePtr);

			StrMark* mark = strMarks.get(yyvsp[0].intlStringPtr);

			if (mark)	/* hex string is not in strMarks*/
				mark->introduced = true;
		}
break;
case 1781:
  if (!yytrial)
{ yyval.stringPtr = newString(yyvsp[0].intlStringPtr->toUtf8(scratch)); }
break;
case 1783:
  if (!yytrial)
{ yyval.int32Val = -yyvsp[0].int32Val; }
break;
case 1784:
  if (!yytrial)
{
			if (yyvsp[0].int32Val > SHRT_POS_MAX)
				yyabandon(YYPOSNARG(1), -842, isc_expec_short);	/* Short integer expected*/

			yyval.int32Val = yyvsp[0].int32Val;
		}
break;
case 1785:
  if (!yytrial)
{
			if (yyvsp[0].int32Val > SHRT_NEG_MAX)
				yyabandon(YYPOSNARG(1), -842, isc_expec_short);	/* Short integer expected*/

			yyval.int32Val = yyvsp[0].int32Val;
		}
break;
case 1786:
  if (!yytrial)
{
			if (yyvsp[0].int32Val == 0)
				yyabandon(YYPOSNARG(1), -842, isc_expec_positive);	/* Positive number expected*/

			yyval.int32Val = yyvsp[0].int32Val;
		}
break;
case 1787:
  if (!yytrial)
{
			if (yyvsp[0].int32Val > SHRT_UNSIGNED_MAX)
				yyabandon(YYPOSNARG(1), -842, isc_expec_ushort);	/* Unsigned short integer expected*/

			yyval.int32Val = yyvsp[0].int32Val;
		}
break;
case 1789:
  if (!yytrial)
{ yyval.int32Val = -yyvsp[0].int32Val; }
break;
case 1790:
  if (!yytrial)
{ yyval.int32Val = yyvsp[0].int32Val;}
break;
case 1791:
  if (!yytrial)
{ yyval.valueExprNode = yyvsp[0].aggNode; }
break;
case 1798:
  if (!yytrial)
{
			yyval.aggNode = yyvsp[-5].aggNode;

			if (yyval.aggNode->aggInfo.blr == blr_agg_count2 && !yyval.aggNode->arg)	/* count(*)*/
				yyval.aggNode->arg = newNode<ValueIfNode>(yyvsp[-1].boolExprNode, MAKE_const_slong(1), NullNode::instance());
			else
			{
				fb_assert(yyval.aggNode->arg);
				yyval.aggNode->arg = newNode<ValueIfNode>(yyvsp[-1].boolExprNode, yyval.aggNode->arg, NullNode::instance());
			}
		}
break;
case 1799:
  if (!yytrial)
{ yyval.aggNode = newNode<CountAggNode>(false, (client_dialect < SQL_DIALECT_V6_TRANSITION)); }
break;
case 1800:
  if (!yytrial)
{ yyval.aggNode = newNode<CountAggNode>(false, (client_dialect < SQL_DIALECT_V6_TRANSITION), yyvsp[-1].valueExprNode); }
break;
case 1801:
  if (!yytrial)
{ yyval.aggNode = newNode<CountAggNode>(true, (client_dialect < SQL_DIALECT_V6_TRANSITION), yyvsp[-1].valueExprNode); }
break;
case 1802:
  if (!yytrial)
{
			yyval.aggNode = newNode<SumAggNode>(false,
				(client_dialect < SQL_DIALECT_V6_TRANSITION), yyvsp[-1].valueExprNode);
		}
break;
case 1803:
  if (!yytrial)
{
			yyval.aggNode = newNode<SumAggNode>(true,
				(client_dialect < SQL_DIALECT_V6_TRANSITION), yyvsp[-1].valueExprNode);
		}
break;
case 1804:
  if (!yytrial)
{
			yyval.aggNode = newNode<AvgAggNode>(false,
				(client_dialect < SQL_DIALECT_V6_TRANSITION), yyvsp[-1].valueExprNode);
		}
break;
case 1805:
  if (!yytrial)
{
			yyval.aggNode = newNode<AvgAggNode>(true,
				(client_dialect < SQL_DIALECT_V6_TRANSITION), yyvsp[-1].valueExprNode);
		}
break;
case 1806:
  if (!yytrial)
{ yyval.aggNode = newNode<MaxMinAggNode>(MaxMinAggNode::TYPE_MIN, yyvsp[-1].valueExprNode); }
break;
case 1807:
  if (!yytrial)
{ yyval.aggNode = newNode<MaxMinAggNode>(MaxMinAggNode::TYPE_MIN, yyvsp[-1].valueExprNode); }
break;
case 1808:
  if (!yytrial)
{ yyval.aggNode = newNode<MaxMinAggNode>(MaxMinAggNode::TYPE_MAX, yyvsp[-1].valueExprNode); }
break;
case 1809:
  if (!yytrial)
{ yyval.aggNode = newNode<MaxMinAggNode>(MaxMinAggNode::TYPE_MAX, yyvsp[-1].valueExprNode); }
break;
case 1810:
  if (!yytrial)
{ yyval.aggNode = newNode<ListAggNode>(false, yyvsp[-2].valueExprNode, yyvsp[-1].valueExprNode); }
break;
case 1811:
  if (!yytrial)
{ yyval.aggNode = newNode<ListAggNode>(true, yyvsp[-2].valueExprNode, yyvsp[-1].valueExprNode); }
break;
case 1812:
  if (!yytrial)
{ yyval.aggNode = newNode<StdDevAggNode>(StdDevAggNode::TYPE_STDDEV_SAMP, yyvsp[-1].valueExprNode); }
break;
case 1813:
  if (!yytrial)
{ yyval.aggNode = newNode<StdDevAggNode>(StdDevAggNode::TYPE_STDDEV_POP, yyvsp[-1].valueExprNode); }
break;
case 1814:
  if (!yytrial)
{ yyval.aggNode = newNode<StdDevAggNode>(StdDevAggNode::TYPE_VAR_SAMP, yyvsp[-1].valueExprNode); }
break;
case 1815:
  if (!yytrial)
{ yyval.aggNode = newNode<StdDevAggNode>(StdDevAggNode::TYPE_VAR_POP, yyvsp[-1].valueExprNode); }
break;
case 1816:
  if (!yytrial)
{ yyval.aggNode = newNode<CorrAggNode>(CorrAggNode::TYPE_COVAR_SAMP, yyvsp[-3].valueExprNode, yyvsp[-1].valueExprNode); }
break;
case 1817:
  if (!yytrial)
{ yyval.aggNode = newNode<CorrAggNode>(CorrAggNode::TYPE_COVAR_POP, yyvsp[-3].valueExprNode, yyvsp[-1].valueExprNode); }
break;
case 1818:
  if (!yytrial)
{ yyval.aggNode = newNode<CorrAggNode>(CorrAggNode::TYPE_CORR, yyvsp[-3].valueExprNode, yyvsp[-1].valueExprNode); }
break;
case 1819:
  if (!yytrial)
{ yyval.aggNode = newNode<RegrAggNode>(RegrAggNode::TYPE_REGR_AVGX, yyvsp[-3].valueExprNode, yyvsp[-1].valueExprNode); }
break;
case 1820:
  if (!yytrial)
{ yyval.aggNode = newNode<RegrAggNode>(RegrAggNode::TYPE_REGR_AVGY, yyvsp[-3].valueExprNode, yyvsp[-1].valueExprNode); }
break;
case 1821:
  if (!yytrial)
{ yyval.aggNode = newNode<RegrCountAggNode>(yyvsp[-3].valueExprNode, yyvsp[-1].valueExprNode); }
break;
case 1822:
  if (!yytrial)
{ yyval.aggNode = newNode<RegrAggNode>(RegrAggNode::TYPE_REGR_INTERCEPT, yyvsp[-3].valueExprNode, yyvsp[-1].valueExprNode); }
break;
case 1823:
  if (!yytrial)
{ yyval.aggNode = newNode<RegrAggNode>(RegrAggNode::TYPE_REGR_R2, yyvsp[-3].valueExprNode, yyvsp[-1].valueExprNode); }
break;
case 1824:
  if (!yytrial)
{ yyval.aggNode = newNode<RegrAggNode>(RegrAggNode::TYPE_REGR_SLOPE, yyvsp[-3].valueExprNode, yyvsp[-1].valueExprNode); }
break;
case 1825:
  if (!yytrial)
{ yyval.aggNode = newNode<RegrAggNode>(RegrAggNode::TYPE_REGR_SXX, yyvsp[-3].valueExprNode, yyvsp[-1].valueExprNode); }
break;
case 1826:
  if (!yytrial)
{ yyval.aggNode = newNode<RegrAggNode>(RegrAggNode::TYPE_REGR_SXY, yyvsp[-3].valueExprNode, yyvsp[-1].valueExprNode); }
break;
case 1827:
  if (!yytrial)
{ yyval.aggNode = newNode<RegrAggNode>(RegrAggNode::TYPE_REGR_SYY, yyvsp[-3].valueExprNode, yyvsp[-1].valueExprNode); }
break;
case 1828:
  if (!yytrial)
{ yyval.aggNode = newNode<DenseRankWinNode>(); }
break;
case 1829:
  if (!yytrial)
{ yyval.aggNode = newNode<RankWinNode>(); }
break;
case 1830:
  if (!yytrial)
{ yyval.aggNode = newNode<PercentRankWinNode>(); }
break;
case 1831:
  if (!yytrial)
{ yyval.aggNode = newNode<CumeDistWinNode>(); }
break;
case 1832:
  if (!yytrial)
{ yyval.aggNode = newNode<RowNumberWinNode>(); }
break;
case 1833:
  if (!yytrial)
{ yyval.aggNode = newNode<FirstValueWinNode>(yyvsp[-1].valueExprNode); }
break;
case 1834:
  if (!yytrial)
{ yyval.aggNode = newNode<LastValueWinNode>(yyvsp[-1].valueExprNode); }
break;
case 1835:
  if (!yytrial)
{ yyval.aggNode = newNode<NthValueWinNode>(yyvsp[-4].valueExprNode, yyvsp[-2].valueExprNode, yyvsp[0].valueExprNode); }
break;
case 1836:
  if (!yytrial)
{ yyval.aggNode = newNode<LagWinNode>(yyvsp[-5].valueExprNode, yyvsp[-3].valueExprNode, yyvsp[-1].valueExprNode); }
break;
case 1837:
  if (!yytrial)
{ yyval.aggNode = newNode<LagWinNode>(yyvsp[-3].valueExprNode, yyvsp[-1].valueExprNode, NullNode::instance()); }
break;
case 1838:
  if (!yytrial)
{ yyval.aggNode = newNode<LagWinNode>(yyvsp[-1].valueExprNode, MAKE_const_slong(1), NullNode::instance()); }
break;
case 1839:
  if (!yytrial)
{ yyval.aggNode = newNode<LeadWinNode>(yyvsp[-5].valueExprNode, yyvsp[-3].valueExprNode, yyvsp[-1].valueExprNode); }
break;
case 1840:
  if (!yytrial)
{ yyval.aggNode = newNode<LeadWinNode>(yyvsp[-3].valueExprNode, yyvsp[-1].valueExprNode, NullNode::instance()); }
break;
case 1841:
  if (!yytrial)
{ yyval.aggNode = newNode<LeadWinNode>(yyvsp[-1].valueExprNode, MAKE_const_slong(1), NullNode::instance()); }
break;
case 1842:
  if (!yytrial)
{ yyval.aggNode = newNode<NTileWinNode>(yyvsp[-1].valueExprNode); }
break;
case 1843:
  if (!yytrial)
{ yyval.valueExprNode = MAKE_const_slong(NthValueWinNode::FROM_FIRST); }
break;
case 1844:
  if (!yytrial)
{ yyval.valueExprNode = MAKE_const_slong(NthValueWinNode::FROM_FIRST); }
break;
case 1845:
  if (!yytrial)
{ yyval.valueExprNode = MAKE_const_slong(NthValueWinNode::FROM_LAST); }
break;
case 1851:
  if (!yytrial)
{ yyval.valueExprNode = newNode<OverNode>(yyvsp[-2].aggNode, yyvsp[0].metaNamePtr); }
break;
case 1852:
  if (!yytrial)
{ yyval.valueExprNode = newNode<OverNode>(yyvsp[-4].aggNode, yyvsp[-1].windowClause); }
break;
case 1853:
  if (!yytrial)
{
			yyval.windowClause = newNode<WindowClause>(yyvsp[-4].metaNamePtr, yyvsp[-3].valueListNode, yyvsp[-2].valueListNode, yyvsp[-1].windowClauseFrameExtent, yyvsp[0].windowClauseExclusion);
		}
break;
case 1854:
  if (!yytrial)
{ yyval.valueListNode = NULL; }
break;
case 1855:
  if (!yytrial)
{ yyval.valueListNode = yyvsp[0].valueListNode; }
break;
case 1856:
  if (!yytrial)
{ yyval.windowClauseFrameExtent = NULL; }
break;
case 1857:
  if (!yytrial)
{ yyval.windowClauseFrameExtent = newNode<WindowClause::FrameExtent>(WindowClause::FrameExtent::Unit::RANGE); }
break;
case 1858:
yyval.windowClauseFrameExtent = yyvsp[0].windowClauseFrameExtent;
break;
case 1859:
  if (!yytrial)
{ yyval.windowClauseFrameExtent = yyvsp[-2].windowClauseFrameExtent; }
break;
case 1860:
  if (!yytrial)
{ yyval.windowClauseFrameExtent = newNode<WindowClause::FrameExtent>(WindowClause::FrameExtent::Unit::ROWS); }
break;
case 1861:
  if (!yytrial)
{ yyval.windowClauseFrameExtent = yyvsp[-2].windowClauseFrameExtent; }
break;
case 1862:
  if (!yytrial)
{
			yyvsp[-1].windowClauseFrameExtent->frame1 = yyvsp[0].windowClauseFrame;
			yyvsp[-1].windowClauseFrameExtent->frame2 =
				newNode<WindowClause::Frame>(WindowClause::Frame::Bound::CURRENT_ROW);
		}
break;
case 1863:
  if (!yytrial)
{
			yyvsp[-4].windowClauseFrameExtent->frame1 = yyvsp[-2].windowClauseFrame;
			yyvsp[-4].windowClauseFrameExtent->frame2 = yyvsp[0].windowClauseFrame;
		}
break;
case 1864:
  if (!yytrial)
{ yyval.windowClauseFrame = newNode<WindowClause::Frame>(WindowClause::Frame::Bound::PRECEDING); }
break;
case 1865:
  if (!yytrial)
{ yyval.windowClauseFrame = newNode<WindowClause::Frame>(WindowClause::Frame::Bound::CURRENT_ROW); }
break;
case 1866:
  if (!yytrial)
{ yyval.windowClauseFrame = newNode<WindowClause::Frame>(WindowClause::Frame::Bound::PRECEDING, yyvsp[-1].valueExprNode); }
break;
case 1867:
  if (!yytrial)
{ yyval.windowClauseFrame = newNode<WindowClause::Frame>(WindowClause::Frame::Bound::PRECEDING); }
break;
case 1868:
  if (!yytrial)
{ yyval.windowClauseFrame = newNode<WindowClause::Frame>(WindowClause::Frame::Bound::CURRENT_ROW); }
break;
case 1869:
  if (!yytrial)
{ yyval.windowClauseFrame = newNode<WindowClause::Frame>(WindowClause::Frame::Bound::PRECEDING, yyvsp[-1].valueExprNode); }
break;
case 1870:
  if (!yytrial)
{ yyval.windowClauseFrame = newNode<WindowClause::Frame>(WindowClause::Frame::Bound::FOLLOWING, yyvsp[-1].valueExprNode); }
break;
case 1871:
  if (!yytrial)
{ yyval.windowClauseFrame = newNode<WindowClause::Frame>(WindowClause::Frame::Bound::FOLLOWING); }
break;
case 1872:
  if (!yytrial)
{ yyval.windowClauseFrame = newNode<WindowClause::Frame>(WindowClause::Frame::Bound::CURRENT_ROW); }
break;
case 1873:
  if (!yytrial)
{ yyval.windowClauseFrame = newNode<WindowClause::Frame>(WindowClause::Frame::Bound::PRECEDING, yyvsp[-1].valueExprNode); }
break;
case 1874:
  if (!yytrial)
{ yyval.windowClauseFrame = newNode<WindowClause::Frame>(WindowClause::Frame::Bound::FOLLOWING, yyvsp[-1].valueExprNode); }
break;
case 1875:
  if (!yytrial)
{ yyval.windowClauseExclusion = WindowClause::Exclusion::NO_OTHERS; }
break;
case 1876:
  if (!yytrial)
{ yyval.windowClauseExclusion = WindowClause::Exclusion::NO_OTHERS; }
break;
case 1877:
  if (!yytrial)
{ yyval.windowClauseExclusion = WindowClause::Exclusion::CURRENT_ROW; }
break;
case 1878:
  if (!yytrial)
{ yyval.windowClauseExclusion = WindowClause::Exclusion::GROUP; }
break;
case 1879:
  if (!yytrial)
{ yyval.windowClauseExclusion = WindowClause::Exclusion::TIES; }
break;
case 1880:
  if (!yytrial)
{ yyval.valueExprNode = MAKE_str_constant(newIntlString(","), lex.att_charset); }
break;
case 1881:
  if (!yytrial)
{ yyval.valueExprNode = yyvsp[0].valueExprNode; }
break;
case 1884:
  if (!yytrial)
{ yyval.valueExprNode = newNode<ExtractNode>(yyvsp[-3].blrOp, yyvsp[-1].valueExprNode); }
break;
case 1888:
  if (!yytrial)
{ yyval.valueExprNode = newNode<StrLenNode>(blr_strlen_bit, yyvsp[-1].valueExprNode); }
break;
case 1889:
  if (!yytrial)
{ yyval.valueExprNode = newNode<StrLenNode>(blr_strlen_char, yyvsp[-1].valueExprNode); }
break;
case 1890:
  if (!yytrial)
{ yyval.valueExprNode = newNode<StrLenNode>(blr_strlen_char, yyvsp[-1].valueExprNode); }
break;
case 1891:
  if (!yytrial)
{ yyval.valueExprNode = newNode<StrLenNode>(blr_strlen_octet, yyvsp[-1].valueExprNode); }
break;
case 1892:
  if (!yytrial)
{ yyval.valueExprNode = newNode<SysFuncCallNode>(*yyvsp[-3].metaNamePtr, yyvsp[-1].valueListNode); }
break;
case 1893:
  if (!yytrial)
{ yyval.valueExprNode = yyvsp[0].sysFuncCallNode; }
break;
case 1957:
  if (!yytrial)
{
			yyval.sysFuncCallNode = newNode<SysFuncCallNode>(*yyvsp[-6].metaNamePtr,
				newNode<ValueListNode>(yyvsp[-4].valueExprNode)->add(MAKE_const_slong(yyvsp[-3].blrOp))->add(yyvsp[-1].valueExprNode));
			yyval.sysFuncCallNode->dsqlSpecialSyntax = true;
		}
break;
case 1958:
  if (!yytrial)
{
			yyval.sysFuncCallNode = newNode<SysFuncCallNode>(*yyvsp[-7].metaNamePtr,
				newNode<ValueListNode>(yyvsp[-3].valueExprNode)->add(MAKE_const_slong(yyvsp[-5].blrOp))->add(yyvsp[-1].valueExprNode));
			yyval.sysFuncCallNode->dsqlSpecialSyntax = true;
		}
break;
case 1959:
  if (!yytrial)
{
			yyval.sysFuncCallNode = newNode<SysFuncCallNode>(*yyvsp[-7].metaNamePtr,
				newNode<ValueListNode>(MAKE_const_slong(yyvsp[-5].blrOp))->add(yyvsp[-3].valueExprNode)->add(yyvsp[-1].valueExprNode));
			yyval.sysFuncCallNode->dsqlSpecialSyntax = true;
		}
break;
case 1960:
  if (!yytrial)
{
			yyval.sysFuncCallNode = newNode<SysFuncCallNode>(*yyvsp[-7].metaNamePtr,
				newNode<ValueListNode>(MAKE_const_slong(yyvsp[-5].blrOp))->add(yyvsp[-3].valueExprNode)->add(yyvsp[-1].valueExprNode));
			yyval.sysFuncCallNode->dsqlSpecialSyntax = true;
		}
break;
case 1961:
  if (!yytrial)
{
			yyval.sysFuncCallNode = newNode<SysFuncCallNode>(*yyvsp[-11].metaNamePtr,
				newNode<ValueListNode>(yyvsp[-9].valueExprNode)->add(MAKE_str_constant(newIntlString(yyvsp[-7].metaNamePtr->c_str()), CS_ASCII))->
					add(MAKE_str_constant(newIntlString(yyvsp[-6].metaNamePtr->c_str()), CS_ASCII))->add(yyvsp[-4].valueExprNode)->add(yyvsp[-3].valueExprNode)->
					add(MAKE_str_constant(newIntlString(yyvsp[-2].metaNamePtr->c_str()), CS_ASCII))->add(yyvsp[-1].valueExprNode));
			yyval.sysFuncCallNode->dsqlSpecialSyntax = true;
		}
break;
case 1962:
  if (!yytrial)
{
			yyval.sysFuncCallNode = newNode<SysFuncCallNode>(*yyvsp[-5].metaNamePtr,
				newNode<ValueListNode>(MAKE_const_slong(yyvsp[-3].blrOp))->add(yyvsp[-1].valueExprNode));
			yyval.sysFuncCallNode->dsqlSpecialSyntax = true;
		}
break;
case 1963:
  if (!yytrial)
{ yyval.sysFuncCallNode = newNode<SysFuncCallNode>(*yyvsp[-3].metaNamePtr, newNode<ValueListNode>(yyvsp[-1].valueExprNode)); }
break;
case 1964:
  if (!yytrial)
{
			yyval.sysFuncCallNode = newNode<SysFuncCallNode>(*yyvsp[-5].metaNamePtr,
				newNode<ValueListNode>(yyvsp[-3].valueExprNode)->add(MAKE_str_constant(newIntlString(yyvsp[-1].metaNamePtr->c_str()), CS_ASCII)));
			yyval.sysFuncCallNode->dsqlSpecialSyntax = true;
		}
break;
case 1965:
  if (!yytrial)
{
			yyval.sysFuncCallNode = newNode<SysFuncCallNode>(*yyvsp[-5].metaNamePtr,
				newNode<ValueListNode>(MAKE_const_slong(yyvsp[-3].blrOp))->add(yyvsp[-1].valueExprNode));
			yyval.sysFuncCallNode->dsqlSpecialSyntax = true;
		}
break;
case 1966:
  if (!yytrial)
{
			yyval.sysFuncCallNode = newNode<SysFuncCallNode>(*yyvsp[-9].metaNamePtr,
				newNode<ValueListNode>(yyvsp[-7].valueExprNode)->add(yyvsp[-5].valueExprNode)->add(yyvsp[-3].valueExprNode)->add(yyvsp[-1].valueExprNode));
			yyval.sysFuncCallNode->dsqlSpecialSyntax = true;
		}
break;
case 1967:
  if (!yytrial)
{
			yyval.sysFuncCallNode = newNode<SysFuncCallNode>(*yyvsp[-7].metaNamePtr,
				newNode<ValueListNode>(yyvsp[-5].valueExprNode)->add(yyvsp[-3].valueExprNode)->add(yyvsp[-1].valueExprNode));
			yyval.sysFuncCallNode->dsqlSpecialSyntax = true;
		}
break;
case 1968:
  if (!yytrial)
{
			yyval.sysFuncCallNode = newNode<SysFuncCallNode>(*yyvsp[-5].metaNamePtr, newNode<ValueListNode>(yyvsp[-3].valueExprNode)->add(yyvsp[-1].valueExprNode));
			yyval.sysFuncCallNode->dsqlSpecialSyntax = true;
		}
break;
case 1969:
  if (!yytrial)
{ yyval.sysFuncCallNode = newNode<SysFuncCallNode>(*yyvsp[-3].metaNamePtr, yyvsp[-1].valueListNode); }
break;
case 1970:
  if (!yytrial)
{
			yyval.sysFuncCallNode = newNode<SysFuncCallNode>(*yyvsp[-7].metaNamePtr,
				newNode<ValueListNode>(yyvsp[-5].valueExprNode)->add(yyvsp[-3].valueExprNode)->add(yyvsp[-2].valueExprNode)->
					add(MAKE_str_constant(newIntlString(yyvsp[-1].metaNamePtr->c_str()), CS_ASCII)));
			yyval.sysFuncCallNode->dsqlSpecialSyntax = true;
		}
break;
case 1971:
  if (!yytrial)
{
			yyval.sysFuncCallNode = newNode<SysFuncCallNode>(*yyvsp[-7].metaNamePtr,
				newNode<ValueListNode>(yyvsp[-5].valueExprNode)->add(yyvsp[-3].valueExprNode)->
					add(MAKE_str_constant(newIntlString(yyvsp[-2].metaNamePtr->c_str()), CS_ASCII))->add(yyvsp[-1].valueExprNode));
			yyval.sysFuncCallNode->dsqlSpecialSyntax = true;
		}
break;
case 1972:
  if (!yytrial)
{
			yyval.sysFuncCallNode = newNode<SysFuncCallNode>(*yyvsp[-9].metaNamePtr,
				newNode<ValueListNode>(yyvsp[-7].valueExprNode)->add(yyvsp[-5].valueExprNode)->add(yyvsp[-3].valueExprNode)->
					add(MAKE_str_constant(newIntlString(yyvsp[-2].metaNamePtr->c_str()), CS_ASCII))->add(yyvsp[-1].valueExprNode));
			yyval.sysFuncCallNode->dsqlSpecialSyntax = true;
		}
break;
case 1973:
  if (!yytrial)
{
			ValueExprNode* v = MAKE_system_privilege(yyvsp[-1].metaNamePtr->c_str());
			yyval.sysFuncCallNode = newNode<SysFuncCallNode>(*yyvsp[-3].metaNamePtr, newNode<ValueListNode>(v));
		}
break;
case 1978:
  if (!yytrial)
{ yyval.valueExprNode = MAKE_str_constant(newIntlString(""), CS_ASCII); }
break;
case 1979:
  if (!yytrial)
{ yyval.valueExprNode = yyvsp[0].valueExprNode; }
break;
case 1980:
  if (!yytrial)
{ yyval.metaNamePtr = newNode<MetaName>(""); }
break;
case 1981:
  if (!yytrial)
{ yyval.metaNamePtr = yyvsp[0].metaNamePtr; }
break;
case 1982:
  if (!yytrial)
{ yyval.valueExprNode = MAKE_str_constant(newIntlString(""), CS_ASCII); }
break;
case 1983:
  if (!yytrial)
{ yyval.valueExprNode = yyvsp[0].valueExprNode; }
break;
case 1984:
  if (!yytrial)
{ yyval.metaNamePtr = newNode<MetaName>(""); }
break;
case 1985:
  if (!yytrial)
{ yyval.metaNamePtr = yyvsp[0].metaNamePtr; }
break;
case 1986:
  if (!yytrial)
{ yyval.valueExprNode = MAKE_str_constant(newIntlString(""), CS_ASCII); }
break;
case 1987:
  if (!yytrial)
{ yyval.valueExprNode = yyvsp[0].valueExprNode; }
break;
case 1988:
  if (!yytrial)
{ yyval.metaNamePtr = newNode<MetaName>(""); }
break;
case 1989:
  if (!yytrial)
{ yyval.metaNamePtr = yyvsp[0].metaNamePtr; }
break;
case 1992:
  if (!yytrial)
{ yyval.valueExprNode = MAKE_str_constant(newIntlString(""), CS_ASCII); }
break;
case 1993:
  if (!yytrial)
{ yyval.valueExprNode = yyvsp[0].valueExprNode; }
break;
case 1998:
  if (!yytrial)
{ yyval.blrOp = blr_extract_year; }
break;
case 1999:
  if (!yytrial)
{ yyval.blrOp = blr_extract_month; }
break;
case 2000:
  if (!yytrial)
{ yyval.blrOp = blr_extract_week; }
break;
case 2003:
  if (!yytrial)
{ yyval.valueExprNode = newNode<StrCaseNode>(blr_upcase, yyvsp[-1].valueExprNode); }
break;
case 2004:
  if (!yytrial)
{ yyval.valueExprNode = newNode<StrCaseNode>(blr_lowcase, yyvsp[-1].valueExprNode); }
break;
case 2005:
  if (!yytrial)
{
			/* SQL spec requires numbering to start with 1,*/
			/* hence we decrement the first parameter to make it*/
			/* compatible with the engine's implementation*/
			ArithmeticNode* subtractNode = newNode<ArithmeticNode>(
				blr_subtract, true, yyvsp[-2].valueExprNode, MAKE_const_slong(1));

			yyval.valueExprNode = newNode<SubstringNode>(yyvsp[-4].valueExprNode, subtractNode, yyvsp[-1].valueExprNode);
		}
break;
case 2006:
  if (!yytrial)
{ yyval.valueExprNode = newNode<SubstringSimilarNode>(yyvsp[-5].valueExprNode, yyvsp[-3].valueExprNode, yyvsp[-1].valueExprNode); }
break;
case 2007:
  if (!yytrial)
{ yyval.valueExprNode = NULL; }
break;
case 2008:
  if (!yytrial)
{ yyval.valueExprNode = yyvsp[0].valueExprNode; }
break;
case 2009:
  if (!yytrial)
{ yyval.valueExprNode = newNode<TrimNode>(yyvsp[-4].blrOp, yyvsp[-1].valueExprNode, yyvsp[-3].valueExprNode); }
break;
case 2010:
  if (!yytrial)
{ yyval.valueExprNode = newNode<TrimNode>(blr_trim_both, yyvsp[-1].valueExprNode, yyvsp[-3].valueExprNode); }
break;
case 2011:
  if (!yytrial)
{ yyval.valueExprNode = newNode<TrimNode>(yyvsp[-3].blrOp, yyvsp[-1].valueExprNode); }
break;
case 2012:
  if (!yytrial)
{ yyval.valueExprNode = newNode<TrimNode>(blr_trim_both, yyvsp[-1].valueExprNode); }
break;
case 2013:
  if (!yytrial)
{ yyval.blrOp = blr_trim_both; }
break;
case 2014:
  if (!yytrial)
{ yyval.blrOp = blr_trim_trailing; }
break;
case 2015:
  if (!yytrial)
{ yyval.blrOp = blr_trim_leading; }
break;
case 2016:
  if (!yytrial)
{ yyval.valueExprNode = newNode<UdfCallNode>(QualifiedName(*yyvsp[-3].metaNamePtr, ""), yyvsp[-1].valueListNode); }
break;
case 2017:
  if (!yytrial)
{ yyval.valueExprNode = newNode<UdfCallNode>(QualifiedName(*yyvsp[-2].metaNamePtr, ""), newNode<ValueListNode>(0)); }
break;
case 2018:
  if (!yytrial)
{ yyval.valueExprNode = newNode<UdfCallNode>(QualifiedName(*yyvsp[-3].metaNamePtr, *yyvsp[-5].metaNamePtr), yyvsp[-1].valueListNode); }
break;
case 2019:
  if (!yytrial)
{ yyval.valueExprNode = newNode<UdfCallNode>(QualifiedName(*yyvsp[-2].metaNamePtr, *yyvsp[-4].metaNamePtr), newNode<ValueListNode>(0)); }
break;
case 2020:
  if (!yytrial)
{ yyval.valueExprNode = newNode<CastNode>(yyvsp[-3].valueExprNode, yyvsp[-1].legacyField); }
break;
case 2023:
  if (!yytrial)
{
			ComparativeBoolNode* condition = newNode<ComparativeBoolNode>(blr_eql, yyvsp[-3].valueExprNode, yyvsp[-1].valueExprNode);
			yyval.valueExprNode = newNode<ValueIfNode>(condition, NullNode::instance(), yyvsp[-3].valueExprNode);
		}
break;
case 2024:
  if (!yytrial)
{ yyval.valueExprNode = newNode<ValueIfNode>(yyvsp[-5].boolExprNode, yyvsp[-3].valueExprNode, yyvsp[-1].valueExprNode); }
break;
case 2025:
  if (!yytrial)
{ yyval.valueExprNode = newNode<CoalesceNode>(yyvsp[-1].valueListNode->addFront(yyvsp[-3].valueExprNode)); }
break;
case 2026:
  if (!yytrial)
{
			ValueListNode* list = yyvsp[-1].valueListNode;
			ValueListNode* conditions = newNode<ValueListNode>(list->items.getCount() / 2);
			ValueListNode* values = newNode<ValueListNode>(list->items.getCount() / 2);

			for (FB_SIZE_T i = 0; i < list->items.getCount(); i += 2)
			{
				conditions->items[i / 2] = list->items[i];
				values->items[i / 2] = list->items[i + 1];
			}

			yyval.valueExprNode = newNode<DecodeNode>(yyvsp[-3].valueExprNode, conditions, values);
		}
break;
case 2027:
  if (!yytrial)
{
			ValueListNode* list = yyvsp[-3].valueListNode;
			ValueListNode* conditions = newNode<ValueListNode>(list->items.getCount() / 2);
			ValueListNode* values = newNode<ValueListNode>(list->items.getCount() / 2 + 1);

			for (FB_SIZE_T i = 0; i < list->items.getCount(); i += 2)
			{
				conditions->items[i / 2] = list->items[i];
				values->items[i / 2] = list->items[i + 1];
			}

			values->items[list->items.getCount() / 2] = yyvsp[-1].valueExprNode;

			yyval.valueExprNode = newNode<DecodeNode>(yyvsp[-5].valueExprNode, conditions, values);
		}
break;
case 2028:
  if (!yytrial)
{ yyval.valueExprNode = yyvsp[0].decodeNode; }
break;
case 2029:
  if (!yytrial)
{ yyval.valueExprNode = yyvsp[0].valueExprNode; }
break;
case 2030:
  if (!yytrial)
{ yyval.decodeNode = newNode<DecodeNode>(yyvsp[0].valueExprNode, newNode<ValueListNode>(0u), newNode<ValueListNode>(0u)); }
break;
case 2031:
yyval.valueListNode = NOTRIAL(yyvsp[0].decodeNode->conditions);
break;
case 2032:
yyval.valueListNode =  NOTRIAL(yyvsp[-1].decodeNode->values);
break;
case 2033:
  if (!yytrial)
{
				DecodeNode* node = yyval.decodeNode = yyvsp[-5].decodeNode;
				node->label = "CASE";
				if (yyvsp[-1].valueExprNode)
					node->values->add(yyvsp[-1].valueExprNode);
			}
break;
case 2034:
  if (!yytrial)
{
			yyvsp[-5].valueListNode->add(yyvsp[-2].valueExprNode);
			yyvsp[-4].valueListNode->add(yyvsp[0].valueExprNode);
		}
break;
case 2035:
  if (!yytrial)
{
			yyvsp[-6].valueListNode->add(yyvsp[-2].valueExprNode);
			yyvsp[-5].valueListNode->add(yyvsp[0].valueExprNode);
		}
break;
case 2036:
  if (!yytrial)
{ yyval.valueExprNode = NULL; }
break;
case 2037:
  if (!yytrial)
{ yyval.valueExprNode = yyvsp[0].valueExprNode; }
break;
case 2038:
  if (!yytrial)
{ yyval.valueExprNode = yyvsp[-1].valueIfNode; }
break;
case 2039:
  if (!yytrial)
{
			ValueIfNode* last = yyvsp[-3].valueIfNode;
			ValueIfNode* next;

			while ((next = nodeAs<ValueIfNode>(last->falseValue)))
				last = next;

			fb_assert(nodeIs<NullNode>(last->falseValue));

			last->falseValue = yyvsp[-1].valueExprNode;
			yyval.valueExprNode = yyvsp[-3].valueIfNode;
		}
break;
case 2040:
  if (!yytrial)
{ yyval.valueIfNode = newNode<ValueIfNode>(yyvsp[-2].boolExprNode, yyvsp[0].valueExprNode, NullNode::instance()); }
break;
case 2041:
  if (!yytrial)
{
			ValueIfNode* cond = newNode<ValueIfNode>(yyvsp[-2].boolExprNode, yyvsp[0].valueExprNode, NullNode::instance());
			ValueIfNode* last = yyvsp[-4].valueIfNode;
			ValueIfNode* next;

			while ((next = nodeAs<ValueIfNode>(last->falseValue)))
				last = next;

			fb_assert(nodeIs<NullNode>(last->falseValue));

			last->falseValue = cond;
			yyval.valueIfNode = yyvsp[-4].valueIfNode;
		}
break;
case 2045:
  if (!yytrial)
{ yyval.valueListNode = newNode<ValueListNode>(0u)->add(yyvsp[-2].valueExprNode)->add(yyvsp[0].valueExprNode); }
break;
case 2046:
  if (!yytrial)
{ yyval.valueListNode = yyvsp[-4].valueListNode->add(yyvsp[-2].valueExprNode)->add(yyvsp[0].valueExprNode); }
break;
case 2047:
  if (!yytrial)
{
			yyval.valueExprNode = newNode<GenIdNode>((client_dialect < SQL_DIALECT_V6_TRANSITION),
				*yyvsp[0].metaNamePtr, ((Jrd::ValueExprNode*) NULL), true, false);
		}
break;
case 2048:
  if (!yytrial)
{
			yyval.valueExprNode = newNode<GenIdNode>((client_dialect < SQL_DIALECT_V6_TRANSITION),
				*yyvsp[-3].metaNamePtr, yyvsp[-1].valueExprNode, false, false);
		}
break;
case 2049:
  if (!yytrial)
{ yyval.blrOp = blr_extract_year; }
break;
case 2050:
  if (!yytrial)
{ yyval.blrOp = blr_extract_month; }
break;
case 2051:
  if (!yytrial)
{ yyval.blrOp = blr_extract_day; }
break;
case 2052:
  if (!yytrial)
{ yyval.blrOp = blr_extract_hour; }
break;
case 2053:
  if (!yytrial)
{ yyval.blrOp = blr_extract_minute; }
break;
case 2054:
  if (!yytrial)
{ yyval.blrOp = blr_extract_second; }
break;
case 2055:
  if (!yytrial)
{ yyval.blrOp = blr_extract_millisecond; }
break;
case 2056:
  if (!yytrial)
{ yyval.blrOp = blr_extract_timezone_hour; }
break;
case 2057:
  if (!yytrial)
{ yyval.blrOp = blr_extract_timezone_minute; }
break;
case 2058:
  if (!yytrial)
{ yyval.blrOp = blr_extract_week; }
break;
case 2059:
  if (!yytrial)
{ yyval.blrOp = blr_extract_weekday; }
break;
case 2060:
  if (!yytrial)
{ yyval.blrOp = blr_extract_yearday; }
break;
case 2065:
  if (!yytrial)
{ yyval.valueExprNode = NullNode::instance(); }
break;
case 2066:
  if (!yytrial)
{
			dsql_fld* field = newNode<dsql_fld>();
			field->dtype = dtype_boolean;
			field->length = sizeof(UCHAR);

			CastNode* castNode = newNode<CastNode>(NullNode::instance(), field);
			castNode->dsqlAlias = "CONSTANT";
			yyval.valueExprNode = castNode;
		}
break;

  default:
    break;
  }

#if YYDEBUG && defined(YYDBPR)
  if (yydebug) {
    printf("yydebug[%d]: after reduction, result is ", yytrial);
    YYDBPR(yyps->val);
    printf("\n");
  }
#endif

  // Perform user-defined position reduction
#ifdef YYREDUCEPOSNFUNC
  if (!yytrial) {
    YYCALLREDUCEPOSN(YYREDUCEPOSNFUNCARG);
  }
#endif

  yyps->ssp -= yym;
  yystate = *(yyps->ssp);
  yyps->vsp -= yym;
  yyps->psp -= yym;

  yym = yylhs[yyn];
  if (yystate == 0 && yym == 0) {
#if YYDEBUG
    if (yydebug) {
      printf("yydebug[%d,%p]: after reduction, shifting from state 0 to state %d\n",
	     int(yydepth), yytrial, YYFINAL);
    }
#endif
    yystate = YYFINAL;
    *++(yyps->ssp) = YYFINAL;
    *++(yyps->vsp) = yyps->val;
    yyretlval = yyps->val;	// return value of root non-terminal to yylval
    *++(yyps->psp) = yyps->pos;
    yyretposn = yyps->pos;	// return value of root position to yyposn
    if (yychar < 0) {
      if ((yychar = yylex1()) < 0) {
        yychar = 0;
      }
#if YYDEBUG
      if (yydebug) {
        yys = 0;
        if (yychar <= YYMAXTOKEN) yys = yyname[yychar];
        if (!yys) yys = "illegal-symbol";
        printf("yydebug[%d,%p]: state %d, reading %d (%s)\n",
	       int(yydepth), yytrial, YYFINAL, yychar, yys);
      }
#endif
    }
    if (yychar == 0) goto yyaccept;
    goto yyloop;
  }

  if ((yyn = yygindex[yym]) && (yyn += yystate) >= 0 &&
      yyn <= YYTABLESIZE && yycheck[yyn] == yystate) {
    yystate = yytable[yyn];
  } else {
    yystate = yydgoto[yym];
  }
#if YYDEBUG
  if (yydebug)
    printf("yydebug[%d,%p]: after reduction, shifting from state %d to state %d\n",
           int(yydepth), yytrial, *(yyps->ssp), yystate);
#endif
  if (yyps->ssp >= yyps->ss + yyps->stacksize - 1) {
    yyMoreStack(yyps);
  }
  *++(yyps->ssp) = yystate;
  *++(yyps->vsp) = yyps->val;
  *++(yyps->psp) = yyps->pos;
  goto yyloop;


  //
  // Reduction declares that this path is valid.
  // Set yypath and do a full parse
  //
yyvalid:
  if (yypath) {
    goto yyabort;
  }
  while (yyps->save) {
    yyparsestate *save = yyps->save;
    yyps->save = save->save;
    save->save = yypath;
    yypath = save;
  }
#if YYDEBUG
  if (yydebug)
    printf("yydebug[%d,%p]: CONFLICT trial successful, backtracking to state %d, %d tokens\n",
           int(yydepth), yytrial, yypath->state, int(yylvp - yylvals - yypath->lexeme));
#endif
  if (yyerrctx) {
    yyFreeState(yyerrctx);
    yyerrctx = NULL;
  }
  yychar = -1;
  yyps->ssp = yyps->ss + (yypath->ssp - yypath->ss);
  yyps->vsp = yyps->vs + (yypath->vsp - yypath->vs);
  yyps->psp = yyps->ps + (yypath->psp - yypath->ps);
  memcpy (yyps->ss, yypath->ss, (yyps->ssp - yyps->ss + 1) * sizeof(Yshort));
  yySCopy(yyps->vs, yypath->vs,  yyps->vsp - yyps->vs + 1);
  yyPCopy(yyps->ps, yypath->ps,  yyps->psp - yyps->ps + 1);
  yylvp = yylvals + yypath->lexeme;
  yylpp = yylpsns + yypath->lexeme;
  yylexp = yylexemes + yypath->lexeme;
  yystate = yypath->state;
  goto yyloop;


yyabort:
  if (yyerrctx) {
    yyFreeState(yyerrctx);
    yyerrctx = NULL;
  }

  YYSTYPE *pv;
  for(pv = yyps->vs; pv < yyps->vsp; pv++) {
    YYDELETEVAL(*pv, 2);
  }

  YYPOSN *pp;
  for(pp = yyps->ps; pp < yyps->psp; pp++) {
    YYDELETEPOSN(*pp, 2);
  }

  while (yyps) {
    yyparsestate *save = yyps;
    yyps = save->save;
    yyFreeState(save);
  }
  while (yypath) {
    yyparsestate *save = yypath;
    yypath = save->save;
    yyFreeState(save);
  }
  return (1);


yyaccept:
  if (yyps->save) goto yyvalid;
  if (yyerrctx) {
    yyFreeState(yyerrctx);
    yyerrctx = NULL;
  }
  while (yyps) {
    yyparsestate *save = yyps;
    yyps = save->save;
    yyFreeState(save);
  }
  while (yypath) {
    yyparsestate *save = yypath;
    yypath = save->save;
    yyFreeState(save);
  }
  return (0);
}


int Parser::yylex1()
{
  if (yylvp < yylve) {
    yylval = *yylvp++;
    yyposn = *yylpp++;
    return *yylexp++;
  } else {
    if (yyps->save) {
      if (yylvp == yylvlim) {
	yyexpand();
      }
      *yylexp = yylex();
      *yylvp++ = yylval;
      yylve++;
      *yylpp++ = yyposn;
      yylpe++;
      return *yylexp++;
    } else {
      return yylex();
    }
  }
}


int Parser::yyexpand()
{
  int p = yylvp - yylvals;
  int s = yylvlim - yylvals;
  s += YYSTACKGROWTH;
  { Yshort  *tl = yylexemes;
    YYSTYPE *tv = yylvals;
    YYPOSN  *tp = yylpsns;
    yylvals = FB_NEW_POOL(getPool()) YYSTYPE[s];
    yylpsns = FB_NEW_POOL(getPool()) YYPOSN[s];
    yylexemes = FB_NEW_POOL(getPool()) Yshort[s];
    memcpy(yylexemes, tl, (s - YYSTACKGROWTH) * sizeof(Yshort));
    yySCopy(yylvals, tv, s - YYSTACKGROWTH);
    yyPCopy(yylpsns, tp, s - YYSTACKGROWTH);
    delete[] tl;
    delete[] tv;
    delete[] tp;
  }
  yylvp = yylve = yylvals + p;
  yylvlim = yylvals + s;
  yylpp = yylpe = yylpsns + p;
  yylplim = yylpsns + s;
  yylexp = yylexemes + p;
  return 0;
}

